# By, including this file in another makefile, dependencies and object
# code will be made for the source code in that other makefile,
# libraries will be made, certain variables will be set depending on
# the architecture to help you compile and link on the various
# architectures, and executables will be made.

# You should have a line something like, "include
# c_to_o_and_depend_and_link.mak" in your other makefile.  Before that
# include line, you must set the variables SPEEDESHOME, CC_SOURCE,
# EXE, and SRC_DIR.  Even if you think that you do not need to set
# those variables, set them anyway.  If you want them to be a blank
# value, then you must explicitly say so, by having a line something
# like, "EXE=".  It is true that variables have blank values by
# default, if no one has already set them.  But you should not assume
# that no one has set them.  Just set them yourself.  The only
# exception, is that SPEEDESHOME should ONLY be set if it is not
# defined.

# SPEEDESHOME should be set to be the directory which contains among other
# things, the Admin directory and the src directory.

# CC_SOURCE is a list of C++ source files.

# Set SRC_DIR to the absolute directory prefix of the items in CC_SOURCE

# EXE is a list of executables you would like to have linked for you.

# This makefile has a rule that will make any .a or .so file.

# You may optionally append compiler options to the variable CCFLAGS.
# Do not overwrite it with the = or := make command assignment
# operators.  Instead you should append to it with the make command +=
# operator.

# The executables in $(EXE) will be made and linked with the
# dependencies you list in your own makefiles.  In your own makefile,
# you should not have any rules for these executable targets, since
# the rules are in this file.

# This makefile sets a variable called CC_OBJ which is a list of the .o
# files corresponding to the source files listed in the CC_SOURCE
# variable.  The .o files in the variable are listed with the path to the
# correct ArchitectureDirs.  After the include line of this file, you
# can use this variable.  For example, you can list it as a dependent
# after the colon, like:

# target: $(CC_OBJ)

# This makefile uses a method described in the gnumake info manual to
# automatically create dependency files.  You do NOT have to type "make
# depend" to create the dependencies.  If gnumake sees that the
# dependency files do not exist, then it will create them.  If they
# get out of date it will remake them.  The dependency files have the
# extension of .d, and there is one .d file for each .C file.  The .d
# file has the dependencies for the corresponding .C file.

# In addition to the usual method of having the .o file to the left of
# the colon and the .h files to the right, the name of the .d file is
# also listed to the left of the colon, so that the .d file will
# depend on the same things that the .o file depends on.

ifneq ($(origin SPEEDESHOME),file)

  # I wish the make program had a command you could run to print an
  # error message, but I could not find such a command, so I just used
  # the include command with a file whose name is the error message I want
  # to display.
  ifeq ($(origin SPEEDESHOME),command line)
    include "Error:_Do_not_define_SPEEDESHOME_on_the_command_line!"
  else
    ifeq ($(origin SPEEDESHOME),undefined)
      include "Error:_SPEEDESHOME_is_undefined."
    else
      ifeq "$(findstring environment,$(origin SPEEDESHOME))" ""
        include "Internal_error:_strange_origin_of_SPEEDESHOME"
      else
        include "Do_NOT_define_SPEEDESHOME_in_the_environment!"
      endif
    endif
  endif

else

include $(SPEEDESHOME)/Admin/include_only_once.mak

%/.since_this_file_exists_the_directory_containing_this_file_must_exist:
	@echo Making directory $(@D:$(SPEEDESHOME)/%=%)
	@mkdir -p $(@D)
	@touch $@

CC_OBJ = $(CC_SOURCE:%.C=$(SRC_DIR)/$(ARCHDIR)/%.o)

########## Dependencies ################################

ifneq ($(CC_DEPS),)
-include $(CC_DEPS_include_argument)
  ifeq ($(spmake_requires),deps)
    $(CC_OBJ): %.o: %.d
  endif
endif

ifdef depend_command
# I need this ifdef because without it gnumake would remake the
# depend_command makefile and then remake the .d files, and only then
# would it reread each makefile.  I.e. it would NOT remake
# depend_command, then reread it, and then remake the .d files.  It
# does it this way because the manual says that the rule is that it
# makes all makefiles, and THEN rereads them.  But if depend_command
# file does not exist then variable depend_command is not defined, so
# it will not attempt to make the .d files.  It just makes the
# depend_command makefile, then rereads it, and then it sees the .d
# rules and makes them.

$(CC_DEPS): $(SPEEDESHOME)/exe/$(ARCHDIR)/depend_command $(dir_done)
$(CC_DEPS): $(file_whose_date_never_changes1)
$(CC_DEPS): $(SRC_DIR)/$(ARCHDIR)/%.d: $(SRC_DIR)/%.C
	@echo Making dependencies for $(<:$(SPEEDESHOME)/%=%)
	@# If a dependency is in the current directory in which
	@# makedepend is run, then it will list only the filename of
	@# the dependency, without the path of the dependency.  Then If
	@# a make is done from a different directory then the
	@# dependency is no longer in the current directory, meaning
	@# make will say that it does not know how to make it.
	@# Therefore I have this cd / so that the makedepend runs in a
	@# different directory than where the dependencies are.  That
	@# causes makedepend to put the absolute path before the
	@# dependencies.
	$(ECHO)cd /;$(depend_command)

endif


########### Compiling ###################################

phony_obj_targets=$(CC_SOURCE:%.C=$(SRC_DIR)/%)

.PHONY: $(phony_obj_targets)
$(phony_obj_targets): $(SRC_DIR)/%: $(SRC_DIR)/$(ARCHDIR)/%.o

# $(dir_done) used to be dependencies in the following rule.  When
# using the -j option, gnumake would make those dependencies in
# parallel.  But one of those dependencies might be /a/b/f and another
# dependency might be /a/f.  If it makes those two in parallel, then each
# will attempt to mkdir /a.  Two commands trying to make the same thing
# resulted in an error message from the OS.  Therefore I decided not to
# have them be dependencies.  Instead, they are included, which means
# they will NOT be made in parallel!

-include $(dir_done)

$(CC_OBJ): $(SRC_DIR)/$(ARCHDIR)/%.o: $(SRC_DIR)/%.C $(dir_done)
ifeq ($(spmake_requires),deps)
	@
else
    ifeq ($(origin DEBUG_PROCESS_MODEL),undefined)
	@echo Compiling $(<:$(SPEEDESHOME)/%=%)
	$(ECHO)if $(insure_CXX) $(internal_CC_options) $(include_path) \
		$(specific_CCFLAGS) $(EXTRA_CC_OPTIONS) $(unique_CCFLAGS) \
		-o $@ -c $< 1> $@.output 2>&1; \
		then $(show_output) 0; \
		else $(show_output) 1; fi
	@#The following line can be used if the previous line fails
	@#because the compiler does not support the -o option which says
	@#where to put the .o file.
	@#cd $(ARCHDIR) ; $(CXX) $(unique_CCFLAGS) -c ../../$<
    else
	@rm -f $(@:.o=.i)
	$(ECHO)$(CXX) -E $(internal_CC_options) $(include_path) $(specific_CCFLAGS) $(EXTRA_CC_OPTIONS) \
		$(unique_CCFLAGS) -c $< > $(@:.o=.i)
	$(ECHO)if grep _M_C_ $(@:.o=.i) 1>/dev/null 2>&1; then \
		grep -v '^#' $(@:.o=.i) > $(@:.o=.itemp) ; \
		mv $(@:.o=.itemp) $(@:.o=.i) ; \
		echo Telling emacs to add newlines and indenting to \
			$(@:$(SPEEDESHOME)/%.o=%.i) ; \
		emacs -batch -eval '(setq i-file  "$(@:.o=.i)")' -load \
			$(SPEEDESHOME)/Admin/emacsNewLineAndIndent \
			1>/dev/null 2>&1 ; \
		chmod a-w $(@:.o=.i) ; \
	fi
	@echo Compiling $(@:$(SPEEDESHOME)/%.o=%.i)
	$(ECHO)$(insure_CXX) $(internal_CC_options) $(include_path) $(specific_CCFLAGS) $(EXTRA_CC_OPTIONS) $(unique_CCFLAGS) \
		-o $@ -c $(@:.o=.i)
    endif
endif

########## library step ###########################################

%.so:
ifneq ($(filter $(spmake_requires), deps),)
	@
else
	@echo Making $(@F)
	@rm -f $@
	$(ECHO)$(CXX) -o $@ $(make_so_flags) $+
endif

%.a:
ifneq ($(filter $(spmake_requires), deps),)
	@
else
	@echo Making $(@F)
	@rm -f $@
	$(ECHO)ar cr$(local_option) $@ $+
endif


ifdef EXE
 


########## link step ###########################################

$(EXE):
ifneq ($(filter $(spmake_requires), deps),)
	@
else
	@echo Linking $(@F)
	$(ECHO)cd $(@D); $(purify_command) $(insure_link) $(EXTRA_CC_OPTIONS) $(unique_CCFLAGS) -o $(@F) $(link_args)
endif

endif

endif
