ifndef already_included
already_included = this variable is now set

# Use := here instead of = so that time is not wasted by repeated
# execution of the the uname command:
ARCH := $(subst -,,$(shell uname))

COMM = shMemTCP
ifeq ($(COMM),shMemTCP)
   COMMLIB=SpShMemTCP
else
   COMMLIB=Sp$(COMM)
endif

ARCHDIR = ArchitectureDirs/$(ARCH)$(EXTRA_ARCH_DIR)

INCLUDES = \
	import/src/util \
	import/src/compile \
	import/runtime \
	src/stdLibExt \
	src/obsoleteNames \
	src/obsoleteFunctionality \
	src/comm/speedesServer \
	src/comm/shMemTCP/clientServer \
	src/comm/shMemTCP/shMem \
	src/comm/shMemTCP/TCP \
	src/comm/shMemTCP \
	src/util/persistence \
	src/util/internal \
	src/util/misc \
	src/util/containers \
	src/util/parser \
	src/engine/builtInEvents \
	src/engine/eventManagement \
	src/engine/misc \
	src/engine/modelFramework \
	src/engine/modelFramework/processModel \
	src/engine/timeManagement \
	src/stateMgr \
	src/HLA \
	src/HLA/DDM \
	src/HLA/DDM/routingSpace \
	src/HLA/OM/proxy \
	src/HLA/OM/events \
	src/HLA/OM/attributes \
	src/HLA/OM/attributes/dynamic/position \
	src/HLA/OM/attributes/dynamic/float

header_list := $(wildcard $(INCLUDES:%=$(SPEEDESHOME)/%/*.H))

incdir = $(SPEEDESHOME)/include

header_symbolic_link_list = $(addprefix $(incdir)/, $(notdir $(header_list)))

# The process of making .d files reads the symbolic links in the
# include directory.  Therefore, these links must be made BEFORE the
# .d files.  Normally, when you want step A to happen before step B,
# you make step A be a dependent of step B, but if I did that in this
# case, then it would mean making all .d files be dependent on all of
# the symbolic links in the include directory.  That means if you
# change any one file to which a link points, then ALL .d files would
# be remade.  My solution to this problem is to have an intermediate
# file in the chain of dependencies.  The .d files depend on the
# intermediate file and the intermediate file depends on the symbolic
# links.  This still means the links get made before the .d files, but
# if you change a file to which one of the links points, then it means
# the intermediate file must be remade.  But because the intermediate
# file has no rule to make it, it does not get updated.  This means
# that the .d files which depend on it do not get updated either.  Of
# course I still have the other dependency specifications that the
# compiler or makedepend command gives me, and those dependencies
# could cause an update of the .d file, but only if the .d file really
# does depend on the modified .H file.  I need to choose an
# intermediate file that already exists since as I mentioned above, I
# have no rule to make the intermediate file.  I chose the ln
# executable as the intermediate file.  BTW, the .o files still depend
# on the actual .H files, NOT the links to the .H files.  This means
# that when you get error messages it will show the location according
# to our organizational scheme.  I use a similar scheme to cause
# "mkdir include" to be run.

file_whose_date_never_changes1 := /bin/ln
file_whose_date_never_changes2 := /bin/ls

$(file_whose_date_never_changes1): $(header_symbolic_link_list)

place_to_which_link_points = $(filter %/$(@F), $(header_list))

$(file_whose_date_never_changes2): $(SPEEDESHOME)/include/.since_this_file_exists_the_directory_containing_this_file_must_exist

$(header_symbolic_link_list): $(file_whose_date_never_changes2)
	@rm -f $@
	$(ECHO)ln -s $(place_to_which_link_points) $@

include_path = $(INCLUDES:%=-I$(SPEEDESHOME)/%)

# The $($(<:.i=.C)_CCFLAGS) in the following specific_CCFLAGS
# assignment line uses the "Computed Variable Names" feature of
# gnumake.  Users may put lines like the following in their makefiles:

# $(SRC_DIR)/G_Makefile.C_CCFLAGS := -Dflag

# The previous line would mean that when compiling
# $(SRC_DIR)/G_Makefile.C, the flag -Dflag
# should be passed to the compiler, but that flag would NOT be passed
# when compiling any other .C file.

specific_CCFLAGS = $($(<:.i=.C)_CCFLAGS)
GREP_OUT = 2>&1| grep -v
show_output = cat $@.output $($(<:.i=.C)_GREP_OUT); rm $@.output; exit

# This line may seem to be self-contradictory, but it really does make
# sense.  If the user enters "make NODASHG=", then this line which
# sets it will be overridden, and instead this variable will have an
# empty string.  This means that -g will not be used.  But if the user
# does not use that command line option option, then instead of the
# empty string this variable will have the string of "-g".

NODASHG=-g

internal_CC_options = -D$(ARCH) $(NODASHG) $(pic_flags)

# I do not really care if it is sorted.  What I care about is the side
# effect of the sorted function.  That side effect removes duplicates.
unique_CCFLAGS = $(sort $(CCFLAGS))

ifneq ($(origin DEBUG_PROCESS_MODEL),undefined)
  internal_CC_options += -DDEBUG_PROCESS_MODEL
endif

ECHO=@

LIBEXT=so

.DELETE_ON_ERROR:

dir_done = $(addsuffix .since_this_file_exists_the_directory_containing_this_file_must_exist, $(sort $(dir $(CC_OBJ) $(EXE) $(SPEEDESHOME)/exe/$(ARCHDIR)/depend_command $(SPEEDESHOME)/lib/$(ARCHDIR)/dummy_file_name)))

ifeq ($(origin CXX),default)
CXX = CC
endif


insure_CXX=$(CXX)
insure_link=$(CXX)

ifeq ($(INSURE),c)
  insure_link=insure
  insure_CXX=insure
endif

ifeq ($(INSURE),l)
  insure_link=insure
endif


CC_DEPS = $(CC_SOURCE:%.C=$(SRC_DIR)/$(ARCHDIR)/%.d)

ifneq ($(filter $(spmake_requires), objs deps),)
  comment_out = \#
endif

ifdef spmake_requires
  CC_DEPS_include_argument=$(subst $(SPEEDESHOME),$(SPEEDESHOME)/exe/$(ARCHDIR)/loop,$(CC_DEPS))
else
  CC_DEPS_include_argument=$(CC_DEPS)
endif

ifeq ($(origin NODEPEND),undefined)

notfile_with_escaped_forward_slashes = $(subst /,\\\/,$(@D)/)
makedepend_sedexp = \-e /:/s/\^$(subst /,\\\/,$(<D)/)/$(notfile_with_escaped_forward_slashes)/

$(SPEEDESHOME)/exe/$(ARCHDIR)/depend_command: \
			$(SPEEDESHOME)/Admin/create_depend_command_file
	@echo Determining how to make dependencies.
	$(ECHO)mkdir -p $(@D)
	$(ECHO)rm -f $(@D)/loop
	$(ECHO)ln -s $(SPEEDESHOME) $(@D)/loop
	$(ECHO)$< $(@D) $(CXX) $@


-include $(SPEEDESHOME)/exe/$(ARCHDIR)/depend_command

endif


# For the HP CC, the following code sets the variable extra_libs to
# "-Wl,+FPZ".  That tells the compiler to pass option "+FPZ" to the
# linker.  That linker option means that the program will crash when a
# division by zero occurs.  Without that option the program would NOT
# crash.  Rather it would merely calculate the result of the division
# to be infinity, and store infinity in the variable that was being
# assigned the result of the division.  It uses a special
# representation for infinity that when printed shows up as "++".  The
# other architectures do a similar thing, but the flags needed to
# cause a crash are different.

# Also notice that the -z option is used for HPUX which causes the
# program to crash if a null pointer is dereferenced.  Without that
# option the program might not crash for quite a while.  SunOS and
# IRIX do not need options for this since their *default* is to crash.

ifeq ($(ARCH),SunOS)
   extra_libs = -lsocket -lnsl
   ifneq ($(origin ERROR_CHECK),undefined)
      internal_CC_options += $(findstring -fnonstd,$(shell $(CXX) -help))
   endif
   pic_flags = -PIC
   make_so_flags = -G
endif

ifeq ($(ARCH),HPUX)
   #internal_CC_options += +a1
   null_dereference = -z
   ifeq (,$(findstring CenterLine,$(CXX)))
      pass_to_linker = Wl,# HP system compiler
   else
      pass_to_linker = el # CenterLine compiler
   endif
   internal_CC_options \
      += -DHPUX$(subst .,,$(shell uname -r))
   ifneq ($(origin ERROR_CHECK),undefined)
      extra_libs = -$(pass_to_linker)+FPZ
   endif
   ifdef purifyhome
     debug_links = ${purifyhome}/purify_stubs.a #for g++: /usr/lib/end.o
   endif
   extra_libs += /lib/libm.a /lib/libPW.a -Wl,+vnocompatwarnings
   ifneq ($(filter $(shell hostname),tuvok dax),)
     LIBEXT=a
     local_option=l
   else
     pic_flags = +z
     make_so_flags = -b
   endif
endif

ifeq (RCC,$(CXX))
   internal_CC_options += -T
endif

ifneq ($(filter $(ARCH), IRIX IRIX64),)
   ifeq ($(shell hostname),janeway)
     extra_libs += -lm
   else
     ifeq ($(origin ERROR_CHECK),undefined)
       extra_libs += -lm
     else
       extra_libs = -lfpe
       ifeq ($(insure_link),insure)
         extra_libs += -lmx -lm
       else
         extra_libs += -lmx
       endif
     endif
   endif
   extra_libs += -n32 -multigot -G4
   internal_CC_options += -n32 -G4
   make_so_flags = -n32 -shared -soname $@
else
   extra_libs += -lm
endif

ifeq ($(ARCH),Linux)
   extra_libs = -lstdc++
   pic_flags = -fpic
   make_so_flags = -shared
endif

ifneq ($(origin ERROR_CHECK),undefined)
   internal_CC_options += -DERROR_CHECK
   ifneq ($(filter $(CXX),gcc g++),)
#      internal_CC_options += -g3 -Werror -ansi -pedantic-errors \
#	-fno-nonnull-objects -Wtraditional -Wshadow -Wpointer-arith \
#	-Wbad-function-cast -Wcast-qual -Wstrict-prototypes \
#	-Wmissing-prototypes -Wmissing-declarations -Wredundant-decls \
#	-Wnested-externs -Woverloaded-virtual -Wsynth
# -Wall -W -Wwrite-strings -Wconversion -Waggregate-return -O 
   endif
endif

ifeq ($(ARCH),CYGWIN32_NT)
   extra_libs = -lstdc++
endif

ifeq ($(insure_link),insure)
   extra_libs += -Zsl
endif

#if we are using MPI comm rather than the built in SPEEDES comm
ifeq ($(COMM),MPI)
   extra_libs = -L$(MPI_LIB_DIR) $(MPI_EXTRA_LIBS)
   internal_CC_options += -I$(MPI_INCLUDE_DIR)
endif

# The following cancels the implicit rules.
%.o: %.C
%: %.C


link_args = $(NODASHG) $+ $(debug_links) \
	$(extra_libs) $(null_dereference)

SPEEDESLIBDIR = $(SPEEDESHOME)/lib/$(ARCHDIR)
.PHONY: always

test = $(comment_out)$(SPEEDESHOME)/Admin/Test $(<F) $(<D) \
	$@ $(@:test_passed=output)

test_deps = $(SPEEDESHOME)/Admin/Test $(SRC_DIR)/GNUmakefile \
	$(SPEEDESHOME)/Admin/c_to_o_and_depend_and_link.mak

endif
