# You may include this file in your own GNUmakefile.  A file not named
# GNUmakefile should NOT include this file.  Before, you include it,
# set the variable CC_SOURCE to a list of .C files that you want to be
# linked into an executable.  Or instead of setting CC_SOURCE, you may
# set ID_SOURCE to a list of .id files that you would like to have
# compiled by the parallel import (pi) compiler.  After pi compiles
# your .id files and generates C++ files, the C++ files will be
# compiled automatically for you.  If you set PI_MULTIPLE to some
# non-empty value, then pi will make multiple C++ files.  Otherwise it
# will make only one C++ file.

# A target will be made for an executable that will be named the same
# as the test directory's name.  Another target will be made to
# test the executable to see if it crashes or returns an error status
# code.  Also, it will optionally be tested to see if it outputs
# certain regular expressions.

# If you include your own GNUmakefile in another makefile, then set
# the variable TESTDIR to the path of the directory containing the
# paths in CC_SOURCE or in ID_SOURCE.  The executable will be tested
# with the current working directory set to TESTDIR.

# The executable will be linked with libSpCore.so.  If you prefer it to
# be linked with libSpHLA.so, then set the variable USE_HLA_LIB to
# some non-empty value.

# If you want to make sure that the test outputs certain lines, then
# set the variable GREP_EXPS to a list of regular expressions that you
# expect to be in the output.  If the output does not contain one or more
# of those expressions, then an error message will be printed.

# The executable will be tested with 2 nodes, but if you set NUM_NODES
# to a number then that number will be the number of nodes that will be
# used.

# In the directory of the GNUmakefile which includes this file, you may
# enter, "make" to link the executable.  But if you also want to test
# it, you may enter, "make test".  You may also enter "make exe",
# where exe is the name of the executable without any path.  That
# command will just link.  You may also enter, "make exe_test" to have
# it run the test.

# Implementation notes follow:

# Below is a recursive include.  In general with any language whether
# it be C++, lisp, or the language of GNU make, you must have a base
# case of the recursion.  The base case is the case where it does not
# recurse.  I follow this rule here too.  The goal of this recursion
# is to determine the correct value of SPEEDESHOME.  Finding the
# correct value is the base case of the recursion.  The other case
# which recurses, does not find the correct value.  Instead that case
# has a path that is too long.  Therefore, the line above the include
# removes the last directory component, and recurses.  Maybe by removing
# that directory component, the next recursion will find the correct
# value.  If not, then it keeps on recursing.

ifeq ($(wildcard $(SPEEDESHOME)/Admin/test.mak),)
  ifndef SPEEDESHOME
    make_called_from_low_level := this variable is set
    SPEEDESHOME := $(shell pwd)
    TESTDIR:=$(SPEEDESHOME)/
  else
    SPEEDESHOME := $(patsubst %/,%,$(dir $(SPEEDESHOME)))
  endif
  include GNUmakefile
else
  EXE_FILE = $(notdir $(TESTDIR:%/=%))

  PHONY_TEST = $(EXE_FILE)_test

  ifdef make_called_from_low_level
    $(EXE_FILE):
    test: $(PHONY_TEST)

    .PHONY: $(EXE_FILE) test
  endif

# If TESTDIR is realtive, then make it absolute:
ifeq ($(filter /%,$(TESTDIR)),)
  TESTDIR := $(SPEEDESHOME)/$(TESTDIR)
endif

SRC_DIR= $(TESTDIR:%/=%)

EXEDIR = $(SPEEDESHOME)/exe/$(ARCHDIR)

ifdef ID_SOURCE
  include $(SPEEDESHOME)/import/src/GNUmakefile

  deps_except_pi = $(filter-out $(EXEDIR)/pi,$+)

  cc_targets = $(addprefix $(SRC_DIR)/,$(CC_SOURCE))

  ifdef PI_MULTIPLE
    CC_SOURCE = $(ID_SOURCE:%.id=$(ARCHDIR)/%.C) $(ARCHDIR)/Main.C

    # For an explanation of this rule, see the documentation for the make
    # command under the section titled "Empty Target Files to Record
    # Events"

    target = $(SRC_DIR)/$(ARCHDIR)/CPP_files_depend_on_this_empty_target_which_depends_on_id_files

    $(cc_targets): $(target) ;
  else
    CC_SOURCE = $(ARCHDIR)/$(EXE_FILE).C
    target = $(cc_targets)
  endif

  $(target): $(EXEDIR)/pi $(addprefix $(TESTDIR),$(ID_SOURCE))
  ifeq ($(filter $(spmake_requires), deps),)
	@echo pi is compiling $(notdir $(deps_except_pi))
    ifdef PI_MULTIPLE
	$(ECHO)cd $(@D) && $<    -c -p $(SPEEDESHOME) $(deps_except_pi)
    else
	$(ECHO)cd $(@D) && $< -s -c -p $(SPEEDESHOME) $(deps_except_pi) && mv a.C $@
    endif
	@touch $@
  else
	@
  endif
endif

LIB=
EXE=$(EXEDIR)/$(EXE_FILE)
include $(SPEEDESHOME)/Admin/c_to_o_and_depend_and_link.mak

$(EXE_FILE): $(EXE)

ifndef TEMP_LIBEXT
  TEMP_LIBEXT=$(LIBEXT)
endif

ifdef USE_HLA_LIB
 LIBS = $(SPEEDESLIBDIR)/libSpHLA.$(TEMP_LIBEXT) $(SPEEDESLIBDIR)/libSpCore.$(TEMP_LIBEXT)
else
 LIBS = $(SPEEDESLIBDIR)/libSpCore.$(TEMP_LIBEXT)
endif

$(EXE): $(CC_OBJ) $(LIBS)

.PHONY: $(EXE_FILE) $(PHONY_TEST)

$(PHONY_TEST): $(EXE).test_passed

# A feature of GNU make is that a variable's final value is computed
# when read if that variable's dereference is made in targets or
# dependencies.  But for the target's shell commands, the final value
# is computed just before the execution of those command.  Therefore, if I
# put in the command, the line "cd $(TESTDIR)" then all of those
# commands would cd to the same directory.  It would be whatever
# directory was last put into the value of the variable TESTDIR.  But I
# need to use a different directory each time.  Therefore, here is the
# solution:

# This line uses the GNU make feature of computed variable names.  The
# left side of the line expands into the name of the executable that
# should be tested with the string "_TEST_DIR" appended to it.  That
# computed variable name's value is the current working directory that
# should be used when testing that executable.  That value is accessed
# in the following shell command by using the left side of the following
# line as the variable name.  A similar system is used for the GREP_EXPS.

$(EXE)_TESTDIR := $(TESTDIR)

$(EXE)_GREP_EXPS := $(GREP_EXPS)

ifndef NUM_NODES
  NUM_NODES = 2
endif

$(EXE)_NUM_NODES := $(NUM_NODES)

$(EXE).test_passed: $(EXE) $(test_deps)
	$(ECHO)cd $($<_TESTDIR); $(test) $($<_NUM_NODES) $($<_GREP_EXPS)

ifdef USE_HLA_LIB
  include $(SPEEDESHOME)/src/HLA/GNUmakefile
endif
include $(SPEEDESHOME)/src/GNUmakefile

EXE=
SRC_DIR=
USE_HLA_LIB=
GREP_EXPS=
TESTDIR=
CC_SOURCE=
ID_SOURCE=
PI_MULTIPLE=
NUM_NODES=
TEMP_LIBEXT=

endif
