#include "RoutingSpaceFreeList.H"
extern ROUTING_SPACE_FREELIST *RoutingSpaceFreeList;

#include "SpIostream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "O_Grid.H"

#define DEBUG_SpaceMan

#include "O_GridTest.H"
#include "SpScheduleEvent.H"
#include "SpMsg.H"

class E_SpGridTest : public SpEvent {
    virtual void process()  { O_GridTest((O_SpGrid*)SIMOBJ);
    }
};

DEFINE_EVENT(GridTest, E_SpGridTest, SpMsg, 1);

// This class is used in order to have a test of my O_Grid class in
// which O_Grid is NOT the first manager to create objects.  It is
// also used to schedule an initial event that acts on and tests the
// O_Grid.
class O_TestMan : public C_OBJMAN
{
  public:
    O_TestMan();
    void init_events()  {
      ScheduleEvent(1, GridTest, Grid, -1, SpComm_GetNodeId());
    }
};

O_TestMan::O_TestMan()
{
  N_LOC = 729;
  TOTAL_OBJECTS += N_LOC;
  reserve_n_objects(N_LOC);
 
  for (int i=0; i<N_LOC; i++)
    define_object(new C_SIMOBJ);
}

#include "Speedes.H"

DEFINE_OBJECT_MANAGER(TestMan, O_TestMan);

extern void PlugInDDM();

int main (int argc, char **argv)
{
   PLUG_IN_EVENT(GridTest);
   PLUG_IN_OBJECT_MANAGER(TestMan);
   PlugInDDM();
   ExecuteSpeedes(argc, argv);
   return 0;
}

#include "SpErrno.H"


#include "GeoFunctions.H"
#include "speedes_comm.H"

#include "SpaceLookupItem.H"
#include "AltitudeLookupItem.H"
#include "DoubleLookupItem.H"
#include "LatLngLookupItem.H"
#include "EnumLookupItem.H"
#include "S_Grid.H"

#define db_var(var) lookup_id_length << (#var) << " " << (var) << endl

#include "O_GridLookupOutput.H"
#include "PubSubItem.H"

static const char * const hi_lo_res_filename = "hi_lo_res.out";

static void output_results(const char * const description,
			   ostream &lookup_id_length,
			   const O_Grid * const o_grid,		   
			   const SpaceLookupItem &spaceLookupItem)
{
  C_QUEUE results;
  
  lookup_id_length << "\nTest: " << description << ':' << endl;

  lookup_id_length << spaceLookupItem;
  int spaceMan_HiGrid_Lookup_return_value
    = spaceLookupItem.HiGrid_Lookup(&results);
  db_var(spaceMan_HiGrid_Lookup_return_value);
  db_var(results.get_length());
  
  O_GridLookupOutput *grid = (O_GridLookupOutput *)results.ptop();
  const int results_length = results.get_length();
  for (int id_of_grid_in_results = 0;
       id_of_grid_in_results < results_length;
       id_of_grid_in_results++)
  {
    grid->get_MainPubSubItem()->set_global_id(0);
    lookup_id_length << "Region = "  << resetiosflags(ios::left) << setw(11)
		     << id_of_grid_in_results
		     << ": grid->get_gridCovered_p() = "
		     << grid->get_gridCovered_p() << "\n"
		     << (*(grid->get_MainPubSubItem())) << '\n';
    
    if (grid->get_object_node() == SpComm_GetNodeId())
    {
      ofstream hi_lo_res(hi_lo_res_filename, ios::app);
      assert(hi_lo_res.good());
  
      if (id_of_grid_in_results == 0)
	hi_lo_res << "Test: " << description << ':' << endl;
      
      hi_lo_res << "======================================\n";
      hi_lo_res << "Region = " << setw(11) << id_of_grid_in_results
		<< ": grid->get_gridCovered_p() = "
		<< grid->get_gridCovered_p() << "\n";

      HI_GRID &HiGrid = (HI_GRID&) (((o_grid->spaceObj())
				    [grid->get_id_in_node_managed_grids()])
				   .GetHiGrid());

      hi_lo_res << "Space Name: " << HiGrid.get_SpaceName() << endl;
      for (int id_of_dim_in_HiGrid=0;
	   id_of_dim_in_HiGrid<HiGrid.get_Ndimensions();
	   id_of_dim_in_HiGrid++ )
      {
	hi_lo_res << setprecision(6) << resetiosflags(ios::left)
		  << HiGrid.get_DimName(id_of_dim_in_HiGrid) << ": Lo "
		  << setw(7) << HiGrid.get_Lo(id_of_dim_in_HiGrid)
		  << ", Hi "
		  << setw(7) << HiGrid.get_Hi(id_of_dim_in_HiGrid)
		  << ", Resolution [";
	for (int id_of_res_in_dim=0;
	     id_of_res_in_dim<HiGrid.get_Nresolution(id_of_dim_in_HiGrid);
	     id_of_res_in_dim++)
	{
	  hi_lo_res << setw(7) <<
	    ((HiGrid.get_Resolution(id_of_dim_in_HiGrid))[id_of_res_in_dim]);
	  if (id_of_res_in_dim
	      != HiGrid.get_Nresolution(id_of_dim_in_HiGrid)-1)
	    hi_lo_res << ", ";
	}
	hi_lo_res << "]" << endl;
      }

      if (id_of_grid_in_results == results_length-1)
	hi_lo_res << "\n\n\n\n";
    }
    //b_sync();
    SpComm_BarrierSync();
    grid = (O_GridLookupOutput *)grid->get_link();
  }
  
  lookup_id_length << '\n' << endl;

  //This line should be changed to use SpGridLookupOutput::Destroy:
  //RoutingSpaceFreeList->delete_list(&results);
}

static void lookup_shaped_almost_like_one_specific_grid
  (const char * const description,
   ostream &lookup_id_length,
   const O_Grid * const o_grid,
   double epsilon)
{
  SpaceLookupItem SpaceA("SpaceA");
    
  SpaceA.InsertDimLookupItem
    (new DoubleLookupItem("SpaceLevelDimensionA",
			  2.0+(2.0/3.0) - epsilon,
			  6.0+(1.0/3.0) + epsilon, 10));
  SpaceA.InsertDimLookupItem
    (new DoubleLookupItem("SpaceLevelDimensionB",
			  -21 - epsilon,
			  -8  + epsilon, 20));
  SpaceA.InsertDimLookupItem
    (new DoubleLookupItem("TheaterDim1",
			  -0.985 - epsilon,
			  -0.4775 + epsilon, 20));
  SpaceA.InsertDimLookupItem
    (new DoubleLookupItem("TheaterDim2",
			  -18 - epsilon,
			  -14 + epsilon, 20));

  output_results(description,
		 lookup_id_length, o_grid, SpaceA);
}

static void lookup_shaped_almost_like_Theater2
  (const char * const description,
   ostream &lookup_id_length,
   const O_Grid * const o_grid,
   double epsilon)
{
  SpaceLookupItem NoSpaceLevelDims("NoSpaceLevelDims");

  NoSpaceLevelDims.InsertDimLookupItem
    (new DoubleLookupItem("Dim1", 15 - epsilon, 18 + epsilon, 10));
  NoSpaceLevelDims.InsertDimLookupItem
    (new DoubleLookupItem("Dim2", 25 - epsilon, 28 + epsilon, 10));
  NoSpaceLevelDims.InsertDimLookupItem
    (new DoubleLookupItem("Dim3", 35 - epsilon, 38 + epsilon, 20));

  output_results(description,
		 lookup_id_length, o_grid, NoSpaceLevelDims);
}

void general(const O_Grid * const o_grid, ostream &lookup_id_length)
{
  {
    DoubleLookupItem
      *TheaterDim1 = new DoubleLookupItem("TheaterDim1", -2, 188.8, 100),
      *TheaterDim2 = new DoubleLookupItem("TheaterDim2");

    TheaterDim2->set_Lo(-22);
    TheaterDim2->set_Hi(-6);
    TheaterDim2->set_Res(4);
    
    SpaceLookupItem SpaceA("SpaceA");

    SpaceA.InsertDimLookupItem
      (new DoubleLookupItem("SpaceLevelDimensionA", -1, 10, 10));
    // Notice how I do NOT insert a LookupItem for SpaceLevelDimensionB.
    // This means it is a Don't Care.
    SpaceA.InsertDimLookupItem
      (TheaterDim1);
    SpaceA.InsertDimLookupItem
      (TheaterDim2);

    output_results("Covers entire space",
		   lookup_id_length, o_grid, SpaceA);
    
  
    TheaterDim1->set_Hi(.03);
    TheaterDim2->set_Hi(-10);
    output_results("Covers all of TheaterA, but none of TheaterB, and\n"
		   "fits in TheaterA",
		   lookup_id_length, o_grid, SpaceA);

    
    TheaterDim1->set_Lo(.04);
    TheaterDim1->set_Hi(188.8);
    TheaterDim2->set_Hi(-6);
    output_results("Covers all of TheaterB, but none of TheaterA, but does\n"
		   "NOT fit in TheaterB",
		   lookup_id_length, o_grid, SpaceA);
  }
  
  {
    DoubleLookupItem
      *TheaterDim1 = new DoubleLookupItem("TheaterDim1", .03, .03, 100),
      *TheaterDim2 = new DoubleLookupItem("TheaterDim2", -10, -10, 4);

    SpaceLookupItem SpaceA("SpaceA", NULL);

    SpaceA.InsertDimLookupItem
      (new DoubleLookupItem("SpaceLevelDimensionA", -1, -1, 10));
    SpaceA.InsertDimLookupItem
      (new DoubleLookupItem("SpaceLevelDimensionB", -21, -21, 20));
    SpaceA.InsertDimLookupItem(TheaterDim1);
    SpaceA.InsertDimLookupItem(TheaterDim2);

    output_results("Covers a point which means it covers zero area or\n"
		   "zero volume or zero whatever",
		   lookup_id_length, o_grid, SpaceA);


    TheaterDim1->set_Hi(109);
    TheaterDim2->set_Hi(-9);
    output_results("Still covers zero area, but covers two points, each\n"
		   "in different grids",
		   lookup_id_length, o_grid, SpaceA);
  }

  // Overlaps only one grid, but doesn't completely cover it.
  lookup_shaped_almost_like_one_specific_grid
    ("Smaller than grid", lookup_id_length, o_grid, -pow(10.0, -10));

  // Overlaps the previous grid and completely covers it, and it
  // partially covers 53 other grids.  It extends the lookup in all
  // four dimensions.  One dimension has only two grids, so there are
  // two grids it covers in that dimension, and in the other dimensions it
  // covers 3 grids (the center, and the two to either side).  So
  // there are a total of 2*3*3*3=54 grids.
  lookup_shaped_almost_like_one_specific_grid
    ("Bigger than grid", lookup_id_length, o_grid, +pow(10.0, -10));

  {
    SpaceLookupItem NoTheaters("NoTheaters");

    // Notice how I insert them out of order.
    NoTheaters.InsertDimLookupItem
      (new DoubleLookupItem("OneGridDim", 0, .31, 20));
    NoTheaters.InsertDimLookupItem
      (new DoubleLookupItem("SpaceLevelDimensionB", -9, 6, 10));

    output_results("NoTheaters - basic test",
		   lookup_id_length, o_grid, NoTheaters);
  }

  {
    SpaceLookupItem NoTheaters("NoTheaters");

    NoTheaters.InsertDimLookupItem
      (new DoubleLookupItem("SpaceLevelDimensionB", -9, -7, 10));
    NoTheaters.InsertDimLookupItem
      (new DoubleLookupItem("OneGridDim", .22, .27, 20));

    output_results("NoTheaters - fits in SpaceSubSet",
		   lookup_id_length, o_grid, NoTheaters);
  }

  {
    SpaceLookupItem NoTheaters
      ("NoTheaters",
       new DoubleLookupItem("SpaceLevelDimensionB", -9, -7, 10),
       NULL);

    NoTheaters.InsertDimLookupItem
      (new DoubleLookupItem("OneGridDim", -3, .1, 20));

    output_results("NoTheaters - No grids match",
		   lookup_id_length, o_grid, NoTheaters);
  }

  lookup_shaped_almost_like_Theater2
    ("NoSpaceLevelDims - all grids - does NOT fit",
     lookup_id_length, o_grid, pow(10.0, -10));
  
  lookup_shaped_almost_like_Theater2
    ("NoSpaceLevelDims - all grids - fits",
     lookup_id_length, o_grid, -pow(10.0, -10));
  
  {
    SpaceLookupItem NoSpaceLevelDims
      ("NoSpaceLevelDims",
       new DoubleLookupItem("Dim1", 16.6, 16.6, 10),
       new DoubleLookupItem("Dim2", 26.6, 26.6, 10),
       new DoubleLookupItem("Dim3", 36.6, 36.6, 20),
       NULL);

    output_results("NoSpaceLevelDims - Covers just a point (i.e. zero area),\n"
		   "but all Theaters cover this point",
		   lookup_id_length,
		   o_grid, NoSpaceLevelDims);
  }
}

void LatLng(const O_Grid * const o_grid, ostream &lookup_id_length)
{
  {
    LatLngLookupItem * const lat_lng_dim = new LatLngLookupItem
      (DEG_to_RAD(75), .99 * PI, 0, 1.1 * radius_of_earth, 10);
    
    SpaceLookupItem SpaceWithLatLng
      ("SpaceWithLatLng",
       lat_lng_dim,
       // Note that the following dimension will be replaced by the
       // later argument:
       new DoubleLookupItem("Frequency", 123, 123, 10),
       // Notice how the code does not crash even though I have an
       // altitude of -4 to 10.1 here, but an altitude centered at 0
       // in lat_lng_dim.  The reason this works is because the
       // AltitudeLookupItem is independent of LatLngLookupItem,
       // because that is the way that the general SpaceSubSet code is
       // written.  In practical uses of this you will probably want
       // to write a class that holds the pointers to the two lookup
       // items and when you set alt in this class it sets the alt in
       // LatLngLookupItem and adds the radius to it to get the hi to
       // set into the AltitudeLookupItem and subtract the radius from
       // it to get the lo to set into the AltitudeLookupItem.
       new AltitudeLookupItem(-4, 10.1, 20),
       new DoubleLookupItem("Frequency", 0, 11, 10),
       NULL);

    output_results("Covers the whole earth",
		   lookup_id_length, o_grid, SpaceWithLatLng);

    lat_lng_dim->set_KmSizeRadius(.9 * radius_of_earth);
    output_results("Shows that you must pass in radius, not ciurcumference "
		   "and not diameter\nsince the two grids in CoversAllLng "
		   "are no longer covered",
		   lookup_id_length, o_grid, SpaceWithLatLng);
    
    lat_lng_dim->set_KmSizeRadius(.25 * radius_of_earth);
    output_results("Now the lookup is small enough that it no longer "
		   "contains the earth's axis",
		   lookup_id_length, o_grid, SpaceWithLatLng);
    

    lat_lng_dim->set_RadLat(DEG_to_RAD(65));
    const double
      km_size = sin(DEG_to_RAD(70) - lat_lng_dim->get_RadLat())
      * radius_of_earth,

      epsilon = .0001;

    lat_lng_dim->set_KmSizeRadius(km_size * (1 - epsilon));
    output_results("Covers nothing",
		   lookup_id_length, o_grid, SpaceWithLatLng);

    lat_lng_dim->set_KmSizeRadius(km_size * (1 + epsilon));
    output_results("Now big enough to intersect",
		   lookup_id_length, o_grid, SpaceWithLatLng);
  }

  {
    LatLngLookupItem * const lat_lng_dim
      = new LatLngLookupItem(DEG_to_RAD(15),
			     DEG_to_RAD(15), 0, 600, 0);

    SpaceLookupItem TicTacToeSpace("TicTacToeSpace", lat_lng_dim, NULL);

    output_results("All 9 grids are intersected, but only the "
		   "middle grid is covered",
		   lookup_id_length, o_grid, TicTacToeSpace);

    lat_lng_dim->set_KmSizeRadius(1800);
    output_results("All 9 grids are intersected, and now the radius is big\n"
		   "enough to also cover them all",
		   lookup_id_length, o_grid, TicTacToeSpace);
  }
  
  {
    LatLngLookupItem * const lat_lng = new LatLngLookupItem;
    lat_lng
      ->set_RadLat(0)
      .set_RadLng(0)
      .set_KmAlt(-6000)
      .set_KmSizeRadius(10)
      .set_KmResolutionRadius(10);

    SpaceLookupItem SpaceWithLatLng("SpaceWithLatLng");

    // Notice how I insert them out of order:
    SpaceWithLatLng.InsertDimLookupItem(lat_lng);
    SpaceWithLatLng.InsertDimLookupItem
      (new DoubleLookupItem("Frequency", 5, 5, 10));
    SpaceWithLatLng.InsertDimLookupItem(new AltitudeLookupItem(-2, -2, 20));

    output_results("Small centered on 0 lng",
		   lookup_id_length, o_grid, SpaceWithLatLng);

    lat_lng->set_RadLng(PI);
    output_results("Small centered on pi lng",
		   lookup_id_length, o_grid, SpaceWithLatLng);

    lat_lng->set_RadLng(-PI);
    output_results("Small centered on -pi lng",
		   lookup_id_length, o_grid, SpaceWithLatLng);

    lat_lng->set_RadLng(PI/2);
    output_results("Small centered on pi/2 lng",
		   lookup_id_length, o_grid, SpaceWithLatLng);
  }

  {
    const double alt = 2000.3;
    LatLngLookupItem * const lat_lng_dim = new LatLngLookupItem;
    SpaceLookupItem AlmostAllLngSpace
      ("AlmostAllLngSpace",
       lat_lng_dim,
       NULL);

    lat_lng_dim
      ->set_RadLat(DEG_to_RAD(9))
      .set_RadLng(DEG_to_RAD(35))
      .set_KmResolutionRadius(10);
    
    const double distance_from_center_of_sphere_to_axis_of_earth
      = (radius_of_earth + alt) * cos(lat_lng_dim->get_RadLat());

    lat_lng_dim
      ->set_KmSizeRadius(distance_from_center_of_sphere_to_axis_of_earth
			 * sin(DEG_to_RAD(5)));
    
    lat_lng_dim->set_KmAlt(alt + .1);
    output_results("To high of an altitude to intersect",
		   lookup_id_length, o_grid, AlmostAllLngSpace);
    
    lat_lng_dim->set_KmAlt(alt - .1);
    // There are two O_GridLookupOutputs returned that are for the same grid
    // but they have different MAIN_PUBSUB_ITEMs.  One
    // MAIN_PUBSUB_ITEM is for one end of the dimension and the other
    // MAIN_PUBSUB_ITEM is for the other end of the dimension.  The
    // middle of the dimension is not covered.
    output_results("Now we have a low enough of an altitude to intersect",
		   lookup_id_length, o_grid, AlmostAllLngSpace);
  }

  {
    LatLngLookupItem * const lat_lng_dim
      = new LatLngLookupItem(0, 0, -radius_of_earth, .001, 0);
    SpaceLookupItem CoversEntireEarthSpace
      ("CoversEntireEarthSpace",
       lat_lng_dim,
       NULL);

    output_results("Center of the earth",
		   lookup_id_length, o_grid, CoversEntireEarthSpace);

    lat_lng_dim
      ->set_RadLat(north_pole_rad_lat)
      .set_KmAlt(0);
    output_results("On North pole",
		   lookup_id_length, o_grid, CoversEntireEarthSpace);
    lat_lng_dim
      ->set_RadLat(south_pole_rad_lat);
    output_results("On South pole",
		   lookup_id_length, o_grid, CoversEntireEarthSpace);
  }
}

void Enum(const O_Grid * const o_grid, ostream &lookup_id_length)
{
  {
    SpaceLookupItem SpaceWithEnum("SpaceWithEnum");
    output_results("All don't cares",
		   lookup_id_length, o_grid, SpaceWithEnum);


    EnumLookupItem *enum_itemB = new EnumLookupItem("SpaceLevelEnumB",
						   SpaceWithEnum);
    enum_itemB->assign_enums("B1", 0, // later changed to 1
			     "B3", 0, // later changed to 1
			     "B2", 0, // later changed to 1
			     "B3", 1,
			     "B2", 1,
			     "B1", 1,
			     NULL);
    SpaceWithEnum.InsertDimLookupItem(enum_itemB);
    SpaceWithEnum.InsertDimLookupItem
      (new DoubleLookupItem("TheaterDim", 7, 8, 0));
    output_results("All B enums set and one grid in TheaterDim",
		   lookup_id_length, o_grid, SpaceWithEnum);


    enum_itemB->clear_enums("B2",
			     NULL);
    output_results("Only B2 enum is unset and there is one grid in TheaterDim",
		   lookup_id_length, o_grid, SpaceWithEnum);


    EnumLookupItem *enum_itemA = (EnumLookupItem *) SpaceWithEnum
      [SpaceWithEnum.dim_name_and_type_to_id("SpaceLevelEnumA",
					       DimensionLookupItem::Enum)];
    const int
      A1 = enum_itemA->name_to_id_in_dim("A1"),
      A2 = enum_itemA->name_to_id_in_dim("A2"),
      A3 = enum_itemA->name_to_id_in_dim("A3"),
      A4 = enum_itemA->name_to_id_in_dim("A4");

    enum_itemB->assign_enums("B3", 0,
			     NULL);
    // Notice how a spaceLookupItem is NOT passed in before the
    // assign_id_in_dim function is called:
    (enum_itemA = new EnumLookupItem("SpaceLevelEnumA"))
      ->assign_id_in_dim(A3, 1)
      .assign_id_in_dim(A4, 1)
      .assign_id_in_dim(A1, 1);

    enum_itemA->is_for_SpaceLookupItem(SpaceWithEnum);
    SpaceWithEnum.InsertDimLookupItem(enum_itemA);
    output_results("NOT all A enums are set, which means it can NOT be\n"
		   "described by only one HiGridRegion (also notice that\n"
		   "because A2 was not mentioned but other enums were\n"
		   "mentioned, A2 is not set).",
		   lookup_id_length, o_grid, SpaceWithEnum);


    enum_itemA->set_enums("A2", NULL);
    output_results("All A enums are set and it is not distributed, which\n"
		   "means it can be described by only one HiGridRegion",
		   lookup_id_length, o_grid, SpaceWithEnum);

    enum_itemA->assign_id_in_dim(A2, 0).assign_id_in_dim(A4, 0);
    enum_itemB->clear_enums("B2", "B3", "B1", NULL);
    output_results("No grids",
		   lookup_id_length, o_grid, SpaceWithEnum);
  }
}

void O_GridTest(const O_Grid * const o_grid)
{
  // I have the purify option -mail-to-user set, but it only sends an
  // e-mail report if there are errors.  But if there are no errors, I
  // cannot be sure that purify actually even ran!  So I sometimes
  // use this new char line to cause me to get a memory leak, which
  // means I get a purify report.

  //new char;

  //purify_logfile_printf("Node %d\n", SpComm_GetNodeId());

  // This if expression ensures that unlink is done only once per
  // simulation run, instead of once per node:
  if (SpComm_GetNodeId() == 0)
  {
    const int error_p = unlink(hi_lo_res_filename);
    assert(!error_p || errno == ENOENT); // ENOENT means file doesn't exist
  }
  //b_sync();
  SpComm_BarrierSync();
  
  char buf[100];
  sprintf(buf, "lookup_id_length%d.out", SpComm_GetNodeId());

  ofstream lookup_id_length(buf);
  assert(lookup_id_length.good());

  db_var(o_grid->get_StartingUniqueId());
  
  lookup_id_length << *o_grid << endl;

  const int size = 10000;
  char string[size];
  ostrstream space_printing(string, size);
  space_printing << *o_grid << ends;
  assert(space_printing.good());

  if (strstr(string, "SpaceA"))
    general(o_grid, lookup_id_length);

  else if (strstr(string, "SpaceWithLatLng"))
    LatLng(o_grid, lookup_id_length);

  else if (strstr(string, "SpaceWithEnum"))
    Enum(o_grid, lookup_id_length);

  else
    abort();
}
