// RanMotMan.C method file

#include "O_RanMot.H"

#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpStdlib.H"
#include "SpString.H"

#include "M_HLA_event_SCAN.H"
#include "SpScheduleEvent.H"

EVENT_TYPE (TEST_RANMOT);
EVENT_TYPE (BROADCAST_PROXY);
EVENT_TYPE (SCAN);

/************************************************************************
* O_RANDOM_MOTION : construct a random motion manager          		*
************************************************************************/
O_RANDOM_MOTION::O_RANDOM_MOTION() {
  int Nobjects;
  int Replication;
  int RandomSeed;
  double ScanTime;
  int offset;
  int i;
  int GlobalId;
  char s[256];
  OBJECT_PROXY *ObjectProxy;

  cout << "Constructing RANDOM_MOTION Manager" << endl;

/*
** Read in the input file using the Data Parser and extract parameters
*/

  RandomMotionParser = new DATA_PARSER("RanMotion.par");

  RandomMotionParser->GoTo("parameters", NULL);
  Nobjects = RandomMotionParser->GetInt("Nobjects");
  Replication = RandomMotionParser->GetInt("Replication");
  ScanTime = RandomMotionParser->GetFloat("ScanTime");

  RandomSeed = Replication*Nobjects + 12345;

/*
** Determine how many objects to create on this node
*/

  N_TOT = Nobjects;
  N_LOC = deal_me(N_TOT, offset);

/*
** Create N_LOC objects for this node
*/

  RandomMotion = new S_RANDOM_MOTION[N_LOC];
  GlobalId = TOTAL_OBJECTS + offset;

/*
** Loop through each object that has been created and initialize them
*/

  for (i=0; i<N_LOC; i++) {

    RandomMotion[i].set_GLOBAL_ID(GlobalId);
    ObjectProxy = RandomMotion[i].GetObjectProxy();
    ObjectProxy->SetGlobalId(GlobalId);

    RandomMotion[i].set_SEED(GlobalId + RandomSeed);
    RandomMotion[i].InitMotion();

    RandomMotion[i].SetScanTime(ScanTime);

    RandomMotion[i].SetTestInt(0);
    RandomMotion[i].SetTestFloat(0.0);
    RandomMotion[i].SetTestString("RandomMotion0");
    RandomMotion[i].SetTestLogical(LOGICAL_TRUE);

    sprintf(s, "Random Motion %d", GlobalId);
    RandomMotion[i].set_NAME(s);
    RandomMotion[i].SetEntityName(s);

    GlobalId += N_NODES;

  }

/*
** Tell SPEEDES about the local objects that have been created
*/

  reserve_n_objects(N_LOC);

  for (i=0; i<N_LOC; i++) {
    define_object(&RandomMotion[i]);
  }

  TOTAL_OBJECTS += N_TOT;

}

/************************************************************************
* init_events : initialize events 				        *
************************************************************************/
void O_RANDOM_MOTION::init_events() {

  cout << "Initializing events for RANDOM_MOTION" << endl;

/*
** Start up a test event for each object
*/

  for (int i=0; i<N_LOC; i++) {
    schedule(1.0, TEST_RANMOT, OBJECT_TYPE, i, NODE);
    //schedule(0.0, BROADCAST_PROXY, OBJECT_TYPE, i, NODE);
    schedule(1.0, SCAN, OBJECT_TYPE, i, NODE);
  }

}

/************************************************************************
* terminate : termination method 				        *
************************************************************************/
void O_RANDOM_MOTION::terminate(double tend) {

  if (NODE == 0) {
    cout << "Terminating RANDOM_MOTION at time " << tend
         << " on node " << NODE << endl;
  }

  //delete [] RandomMotion;
}








