// S_SBIRS.C source file

#include "S_SBIRS.H"
#include "CircularOrbit.H"

#include "FreeDynamicAttributes.H"
extern FREE_DYNAMIC_ATTRIBUTES *FreeDynamicAttributes;

/******************************************************************************
* S_SBIRS constructor                                                         *
******************************************************************************/
S_SBIRS::S_SBIRS() : S_MOVING_ENTITY ("SBIRS") {
  DEFINE_ATTRIBUTE(ScanTime, "ScanTime");
}
// <END:   Constructor>

// <BEGIN: Define public virtual methods>
/******************************************************************************
* terminate: executed at the end of the simulation                            *
******************************************************************************/
void S_SBIRS::terminate(double /*tend*/) {

}
// <END:   Define public virtual methods>

// <BEGIN: Define public methods>
/******************************************************************************
* init: Initializes the object using the parser                               *
******************************************************************************/
void S_SBIRS::init(DATA_PARSER *Parser) {
  ScanTime = Parser->GetFloat("ScanTime");
  double PhaseAngle = Parser->GetFloat("PhaseAngle") * RADDEG;
  void *Mark;
  double X1[3];
  double V1[3];
  double A1[3];
  double t;

  Mark = Parser->GetMark();
  Parser->GoToFirstReference(Mark, "RING");

  if (Parser->GetLogical("Circular")) {

    double Inclination = Parser->GetFloat("Inclination") * RADDEG;
    double A = Parser->GetFloat("LongOfAscendingNode") * RADDEG;
    double Altitude = Parser->GetFloat("Altitude");

    CIRCULAR_ORBIT *CircularOrbit;
    CircularOrbit = (CIRCULAR_ORBIT *)
                     FreeDynamicAttributes->new_object(CIRCULAR_ORBIT_ID);
    CircularOrbit->init(Inclination, A, Altitude, PhaseAngle, 0.0);
    t = 100.0;
    (*CircularOrbit)(t, ECI, X1, V1, A1);
/*
    cout << "Circular Position X: "
         << X1[0] << " " << X1[1] << " " << X1[2] << endl;
    cout << "Circular Position V: "
         << V1[0] << " " << V1[1] << " " << V1[2] << endl;
    cout << "Circular Position A: "
         << A1[0] << " " << A1[1] << " " << A1[2] << endl;
*/
    CircularOrbit->CheckMotion(t);
    CircularOrbit->CheckUnitVectors();

    Position += CircularOrbit;

  }else{
    cout << "Error (SBIRS): must be circular orbit for now" << endl;
  }

}
// <END:   Define public methods>

