// S_ScenEl.C source file

#include "S_ScenEl.H"
#include "ExtrapolateMotion.H"
#include "AdaToCppFuncs.H"
#include "CppToAdaFuncs.H"
#include "FreeDynamicAttributes.H"

extern FREE_DYNAMIC_ATTRIBUTES *FreeDynamicAttributes;
extern C_SIMOBJ *CurrentSimobj;

// <BEGIN: Constructor>
/******************************************************************************
* S_ScenEl constructor                                                        *
******************************************************************************/
S_ScenEl::S_ScenEl(char *ProxyType) : S_HLA (ProxyType) {

  DEFINE_ATTRIBUTE(Position, "Position");
  DEFINE_ATTRIBUTE(Alive, "Alive");

  Alive = 1;

}
// <END:   Constructor>

// <BEGIN: Define public virtual methods>
/******************************************************************************
* terminate: executed at the end of the simulation                            *
******************************************************************************/
void S_ScenEl::terminate(double /*tend*/) {

}
// <END:   Define public virtual methods>

// <BEGIN: Define public methods>
/******************************************************************************
* init: Initializes the object using the parser                               *
******************************************************************************/

void S_ScenEl::init(DATA_PARSER *Parser) {
  int Status;

  CurrentSimobj = this;

  ScenElId = Parser->GetInt("ScenElId");
  cout << NAME << " " << ScenElId << endl;

  Status = CreateScenEl(NAME, Parser->GetString("SSR"), ScenElId, 1);

  if (Status < 0) {
    cout << "Error, bad status returned from CreateScenEl(" << NAME << ", " 
         << Parser->GetString("SSR") << ", "
         << ScenElId << ")" << endl;
  }

#ifdef CPP_ONLY

  double X0[3];
  double V0[3];
  double A0[3];

  for (int i=0; i<3; i++) {
    X0[i] = 1.1;
    V0[i] = 2.2;
    A0[i] = 3.3;
  }

  EXTRAPOLATE_MOTION *ExtrapolateMotion = (EXTRAPOLATE_MOTION *)
    FreeDynamicAttributes->new_object(EXTRAPOLATE_MOTION_ID);

  ExtrapolateMotion->Init(0.0, 100.0, X0, V0, A0, ECR);
  Position += ExtrapolateMotion;
  (*ExtrapolateMotion)(100.0, ECR, X0, V0, A0);
  ExtrapolateMotion->CheckMotion(50.0);

  ExtrapolateMotion = (EXTRAPOLATE_MOTION *)
    FreeDynamicAttributes->new_object(EXTRAPOLATE_MOTION_ID);

  ExtrapolateMotion->Init(100.0, 1000000000.0, X0, V0, A0, ECR);
  Position += ExtrapolateMotion;
  ExtrapolateMotion->CheckMotion(150.0);

  Position(0.0, ECR, X0, V0, A0);
  cout << "X(0) = [" << X0[0] << ", " << X0[1] << ", " << X0[2] << "]"
       << endl;
  cout << "V(0) = [" << V0[0] << ", " << V0[1] << ", " << V0[2] << "]"
       << endl;
  cout << "A(0) = [" << A0[0] << ", " << A0[1] << ", " << A0[2] << "]"
       << endl;

  Position(50.0, ECR, X0, V0, A0);
  cout << "X(50) = [" << X0[0] << ", " << X0[1] << ", " << X0[2] << "]"
       << endl;
  cout << "V(50) = [" << V0[0] << ", " << V0[1] << ", " << V0[2] << "]"
       << endl;
  cout << "A(50) = [" << A0[0] << ", " << A0[1] << ", " << A0[2] << "]"
       << endl;

  Position(100.0, ECR, X0, V0, A0);
  cout << "X(100) = [" << X0[0] << ", " << X0[1] << ", " << X0[2] << "]"
       << endl;
  cout << "V(100) = [" << V0[0] << ", " << V0[1] << ", " << V0[2] << "]"
       << endl;
  cout << "A(100) = [" << A0[0] << ", " << A0[1] << ", " << A0[2] << "]"
       << endl;

  Position(150.0, ECR, X0, V0, A0);
  cout << "X(150) = [" << X0[0] << ", " << X0[1] << ", " << X0[2] << "]"
       << endl;
  cout << "V(150) = [" << V0[0] << ", " << V0[1] << ", " << V0[2] << "]"
       << endl;
  cout << "A(150) = [" << A0[0] << ", " << A0[1] << ", " << A0[2] << "]"
       << endl;

  ObjectProxy->PrintStream();

  SetExtrapolatedMotion(0.0, 3600.0, X0, V0, A0, ECR);

#endif

  ObjectProxy->PrintStream();

}
// <END:   Define public methods>
