// file: current_funs.C

#include "current_funs.H"
#include "simtime.H"
/*
   extern int UPDATE_EVENT_TYPE;
*/
/*
** Schedule - basic variable-sized message interface. Note, here the message
** is passed in and nothing is returned. The buffer is copied into the message
** and then the caller is expected to delete the buffer.
** (Inlined version implemented in ScheduleEvent.H)
*/

extern "C" int event_priority[1]; /* allows access to ada array */ 

void ScheduleEvent(SIMTIME& time,
        int EventType,
        int SimobjType,
        int SimobjId,
        int SimobjNode,
        int BufferSize,
        char *Buffer) {

  char *Buff;
  CurrentEvent->schedule(time,
       EventType,
       SimobjType,
       SimobjId,
       SimobjNode,
       BufferSize,
       Buff);

  memcpy(Buff, Buffer, BufferSize);

}


/*
** Schedule - Variable-sized message interface that returns a cancel handle
** (Inlined version implemented in ScheduleEvent.H)
*/

void ScheduleEvent(void *CancelHandle,
        SIMTIME& time,
        int EventType,
        int SimobjType,
        int SimobjId,
        int SimobjNode,
        int BufferSize,
        char *&Buffer) {

  char *Buff;
  CurrentEvent->schedule(CancelHandle,
       time,
       EventType,
       SimobjType,
       SimobjId,
       SimobjNode,
       BufferSize,
       Buff);

  memcpy(Buff, Buffer, BufferSize);

}

/*
** Schedule - basic fixed-sized message interface
*/

C_HEADER *ScheduleEvent(SIMTIME& time,
			int EventType,
			int SimobjType,
			int SimobjId,
			int SimobjNode) {

  return (C_HEADER *)CurrentEvent->schedule(time,
					    EventType,
					    SimobjType,
					    SimobjId,
					    SimobjNode);
}

/*
** Schedule - Fixed-sized message interface that returns a cancel handle
*/

C_HEADER *ScheduleEvent(void *CancelHandle,
			SIMTIME& time,
			int EventType,
			int SimobjType,
			int SimobjId,
			int SimobjNode) {

  return (C_HEADER *)CurrentEvent->schedule(CancelHandle,
					    time,
					    EventType,
					    SimobjType,
					    SimobjId,
					    SimobjNode);
}

/*
* from objman.C
* get the node and local id given an object name   *
* call C_OBJECTS::get_nodeid using object returned by get_objman from CurrentEvent.SIMOBJ
*/

//  extern "C" void get_node_and_local_id (char *name, int &node, int &localid) {
//  C_OBJMAN* om;
//  om = (C_OBJMAN*)(CurrentEvent->get_SIMOBJ())->get_manager(current_object_type());
//  om->get_nodid(name, node, localid);

extern "C" void current_cout_line (
     char *Str     // input pointer to null terminated string to output
     ) { 

  cout << Str << endl;
}

extern "C" void current_RB_cout_line (
     char *Str     // input pointer to null terminated string to output
     ) { 

 RB_cout << Str << endl;
}

extern "C" void get_node_and_local_id (
     char *name,     // input name of sim object
     int objtype,    // input object type of sim object
     int &node,      // output node
     int &localid) { // output local od on node

  C_OBJMAN* om;
  om = (C_OBJMAN*)(CurrentEvent->get_SIMOBJ()->get_manager(objtype));
  om->get_nodid(name, node, localid);
}

/*   NOTE FOR schedules below
     Set priority to assure uniqueness - this is a temp fix.
     Should allow caller to pass p1, p2....


     could use current_object_ instead of what is passed in...??
     See cms_sm_priorities.a for hints
  
    if Left.Scheduled_Time < Right.Scheduled_Time then
      return FALSE;
    elsif Left.Scheduled_Time > Right.Scheduled_Time then
      return TRUE;
    elsif Left.Execution_Priority > Right.Execution_Priority then
        return FALSE;

      elsif Right.Execution_Priority > Left.Execution_Priority then
        return TRUE;

      elsif Left.Destination_Component_Value >
           Right.Destination_Component_Value then
        return FALSE;

      elsif Left.Destination_Component_Value <
           Right.Destination_Component_Value then
        return TRUE;

      elsif Left.Source_Component_PC >
           Right.Source_Component_PC then
        return FALSE;
      elsif Right.Source_Component_PC >
             Left.Source_Component_PC then
        return TRUE;
      elsif Left.Source_Component_Value >
           Right.Source_Component_Value then
        return FALSE;

      elsif Left.Source_Component_Value <
           Right.Source_Component_Value then
        return TRUE;

    else
      return TRUE;
    end if;
*/

/*
* Implementation of call from Ada for Non-Cancelable, Variable-length messages
*/

extern "C" void schedule_with_variable_message(
       double for_time,
       int event_type,
       int Prio1,
       int Prio2,
       int sim_object_type,
       int sim_object_id, 
       int sim_object_node,
       int msg_buffer_size,
       char *msg_buffer) {

    SIMTIME sim_time;
    sim_time.SetTime     (for_time);
    sim_time.SetPriority (Prio1, Prio2);
/*
printf ("event_priority[UPDATE]%d\n", event_priority[UPDATE_EVENT_TYPE]);
printf ("event_priority[5]%d\n", event_priority[5]);
    sim_time.SetPriority (
       event_priority[event_type], 
       sim_object_id*1000 + sim_object_node*1000000 +
       current_object_type()*1000000000);
*/

    ScheduleEvent(
       sim_time,
       event_type, 
       sim_object_type,
       sim_object_id,
       sim_object_node,
       msg_buffer_size,
       msg_buffer);
  }


/*
* Implementation of Schedule event call from Ada for Cancelable, Variable-length messages
*/

extern "C" void schedule_with_variable_message_and_cancel(
       void *cancel_handle,
       double for_time,
       int event_type,
       int Prio1,
       int Prio2,
       int sim_object_type,
       int sim_object_id, 
       int sim_object_node,
       int msg_buffer_size,
       char *msg_buffer) {

    SIMTIME sim_time;
    sim_time.SetTime     (for_time);
    sim_time.SetPriority (Prio1, Prio2);
/*
    sim_time.SetPriority (
       event_priority[event_type], 
       sim_object_id*1000 + sim_object_node*1000000 +
       current_object_type()*1000000000);
*/
    ScheduleEvent(
       cancel_handle,
       sim_time,
       event_type, 
       sim_object_type,
       sim_object_id,
       sim_object_node,
       msg_buffer_size,
       msg_buffer);
  }

/*
* Implementation of call from Ada for Non-Cancelable, fixed-length messages
*/

extern "C" void* schedule_from_c (
       double for_time,
       int event_type,
       int Prio1,
       int Prio2,
       int sim_object_type,
       int sim_object_id, 
       int sim_object_node ) {

    SIMTIME sim_time;
    sim_time.SetTime     (for_time);
    sim_time.SetPriority (Prio1, Prio2);
/*
    sim_time.SetPriority (
       event_priority[event_type], 
       sim_object_id*1000 + sim_object_node*1000000 +
       current_object_type()*1000000000);
*/
    return (void*) ScheduleEvent(
       sim_time,
       event_type, 
       sim_object_type,
       sim_object_id,
       sim_object_node);
  }


/*
* Implementation of Schedule event call from Ada for Cancelable, fixed-length messages
*/

extern "C" void* schedule_with_cancel(
       void *cancel_handle,
       double for_time,
       int event_type,
       int Prio1,
       int Prio2,
       int sim_object_type,
       int sim_object_id, 
       int sim_object_node) {

    SIMTIME sim_time;
    sim_time.SetTime     (for_time);
    sim_time.SetPriority (Prio1, Prio2);

/*
    sim_time.SetPriority (
       event_priority[event_type], 
       sim_object_id*1000 + sim_object_node*1000000 +
       current_object_type()*1000000000);
*/
    return (void*) ScheduleEvent(
       cancel_handle,
       sim_time,
       event_type, 
       sim_object_type,
       sim_object_id,
       sim_object_node);
  }

