// file: current_funs.H
// current_funs.H header file defines
// functions that are not inherited from events or sim objects,
// but call the methods for the current event object or sim object.

#ifndef CURRENT_FUNCS_C
#define CURRENT_FUNCS_C

#include "headers.H"
#include "event.H"

extern C_EVENT *CurrentEvent;
// later...? extern C_SIMOBJ *CurrentSimobj;

/* EVENT_TYPE, OBJECT_TYPE, LOCAL_ID, NODE);
    OBJECT_TYPE = 0
    EVENT_TYPE = 0
    GLOBAL_ID = 0
*/

#include "rollbackStreams.H"

// event object data requests
    extern "C" int current_node()        {return CurrentEvent->get_NODE();}
    extern "C" int current_n_nodes()     {return CurrentEvent->get_N_NODES();}
    extern "C" int current_object_type() {return CurrentEvent->get_OBJECT_TYPE();}
    extern "C" int current_event_type()  {return CurrentEvent->get_EVENT_TYPE();}
    extern "C" int current_global_id()   {return CurrentEvent->get_GLOBAL_ID();}
    extern "C" int current_local_id()    {return CurrentEvent->get_LOCAL_ID();}

// event simobj data requests
    //extern "C" int current_node()         {return CurrentEvent->get_SIMOBJ()->get_NODE();}
    //extern "C" int current_n_nodes()      {return CurrentEvent->get_SIMOBJ()->get_N_NODES();}
    //extern "C" int current_global_id()    {return CurrentEvent->get_SIMOBJ()->get_GLOBAL_ID();}
    //extern "C" int current_local_id()     {return CurrentEvent->get_SIMOBJ()->get_LOCAL_ID();}
    //extern "C" int current_object_type()  {return CurrentEvent->get_SIMOBJ()->get_OBJECT_TYPE();}
    extern "C" char  *current_object_name()         {return CurrentEvent->get_SIMOBJ()->get_NAME();}
    extern "C" int    current_n_object_types() {return CurrentEvent->get_SIMOBJ()->get_N_OBJECT_TYPES();}
    //extern "C" double current_time()        {return CurrentEvent->get_SIMOBJ()->get_TIME();}
    extern "C" double current_time()        {return CurrentEvent->get_time_tag();}
    extern "C" int    current_total_objects()  {return CurrentEvent->get_SIMOBJ()->get_TOTAL_OBJECTS();}

    /* for flat dynamic data block memory management - rollbackable */
  //  extern "C" void*  current_rb_new   (void* addr)  {return (CurrentEvent->GetRBQ())->RB_new(addr);}
  //  extern "C" void   current_rb_delete(void* addr)  {       (CurrentEvent->GetRBQ())->RB_delete(addr);}
    extern "C" void   current_rb_memcpy(char* to_addr, char* from_addr, int bytes)
                                                    {RB_memcpy(to_addr, from_addr, bytes);}

    /* cancel_handle access */
    extern "C" void*  current_rb_new_cancel_handle()  {return RB_NEW_CANCEL_HANDLE();}
    extern "C" void   current_rb_delete_cancel_handle(SpCancelHandle* handle) {RB_DELETE_CANCEL_HANDLE( handle);}

//    extern "C" void*  current_cancel_handle()  {return &(CurrentEvent->get_SIMOBJ()->cancel_handle);}

    /* Cancel Event */
    extern "C" void current_cancel (void *CancelHandle) {CurrentEvent->cancel(CancelHandle);}

/*  B E G I N     BIG CHUNK COMMENT !!!!!!!!!!!!!!!!!!!!!!!!
* from objman.C
* get the node and local id given an object name   *
* call C_OBJECTS::get_nodeid using object returned by get_objman from CurrentEvent.SIMOBJ

extern "C" void get_node_and_local_id (char *name, int &node, int &localid);


** Schedule - basic variable-sized message interface. Note, here the message
** is passed in and nothing is returned. The buffer is copied into the message
** and then the caller is expected to delete the buffer.
** (This is also implemented inlined in ScheduleEvent.H)

void ScheduleEvent(SIMTIME& time,
        int EventType,
        int SimobjType,
        int SimobjId,
        int SimobjNode,
        int BufferSize,
        char *Buffer);

** Schedule - Variable-sized message interface that returns a cancel handle
** (Inlined version implemented in ScheduleEvent.H)

void ScheduleEvent(void *CancelHandle,
        SIMTIME& time,
        int EventType,
        int SimobjType,
        int SimobjId,
        int SimobjNode,
        int BufferSize,
        char *&Buffer);

E N D     BIG CHUNK COMMENT !!!!!!!!!!!!!!!!!!!!!!!! */

#endif
