// Detection.H defines a generic detection from a sensor

#ifndef DetectObj
#define DetectObj

#include <iostream.h>
#include "Convert.H"


class DETECTION {

  private:

  protected:

    double Time;

    double TrueSensorPosition[3];
    double TrueTargetPosition[3];

    double TrueSensorVelocity[3];
    double TrueTargetVelocity[3];

    int GlobalId;
    int Node;
    int Type;
    int LocalId;

    int GoodDetection;

  public:

    void SetTime(double time) {Time = time;}
    double GetTime() {return Time;}

    void SetTrueSensorPosition(double trueSensorPosition[3]) {
      TrueSensorPosition[0] = trueSensorPosition[0];
      TrueSensorPosition[1] = trueSensorPosition[1];
      TrueSensorPosition[2] = trueSensorPosition[2];
    }

    double *GetTrueSensorPosition() {return TrueSensorPosition;}

    void SetTrueTargetPosition(double trueTargetPosition[3]) {
      TrueTargetPosition[0] = trueTargetPosition[0];
      TrueTargetPosition[1] = trueTargetPosition[1];
      TrueTargetPosition[2] = trueTargetPosition[2];
    }

    double *GetTrueTargetPosition() {return TrueTargetPosition;}

    void SetTrueSensorVelocity(double trueSensorVelocity[3]) {
      TrueSensorVelocity[0] = trueSensorVelocity[0];
      TrueSensorVelocity[1] = trueSensorVelocity[1];
      TrueSensorVelocity[2] = trueSensorVelocity[2];
    }

    double *GetTrueSensorVelocity() {return TrueSensorVelocity;}

    void SetTrueTargetVelocity(double trueTargetVelocity[3]) {
      TrueTargetVelocity[0] = trueTargetVelocity[0];
      TrueTargetVelocity[1] = trueTargetVelocity[1];
      TrueTargetVelocity[2] = trueTargetVelocity[2];
    }

    double *GetTrueTargetVelocity() {return TrueTargetVelocity;}

    void SetGlobalId(int globalId) {GlobalId = globalId;}
    int GetGlobalId() {return GlobalId;}

    void SetNode(int node) {Node = node;}
    int GetNode() {return Node;}

    void SetType(int type) {Type = type;}
    int GetType() {return Type;}

    void SetLocalId(int localId) {LocalId = LocalId;}
    int GetLocalId() {return LocalId;}

    void SetGoodDetection() {GoodDetection = 1;}
    void ResetGoodDetection() {GoodDetection = 0;}
    int GetGoodDetection() {return GoodDetection;}

    virtual void Print(ostream &out) {
      out << "Detection of " << GlobalId << " at time " << Time << endl;
    }

};

RB_DEFINE_CLASS(DETECTION);

class RADAR_DETECTION : public DETECTION {

  private:

  protected:

    double TargetCrossSection;

    double MeasuredRange;
    double MeasuredRangeRate;
    double MeasuredDirectionVector[3];

    double SigmaRange;
    double SigmaRangeRate;
    double SigmaAngle;

  public:

    void SetTargetCrossSection(double targetCrossSection) {
      TargetCrossSection = targetCrossSection;
    }

    double GetTargetCrossSection() {return TargetCrossSection;}

    void SetMeasuredRange(double measuredRange) {
      MeasuredRange = measuredRange;
    }
    double GetMeasuredRange() {return MeasuredRange;}

    void SetMeasuredRangeRate(double measuredRangeRate) {
      MeasuredRangeRate = measuredRangeRate;
    }
    double GetMeasuredRangeRate() {return MeasuredRangeRate;}

    void SetMeasuredDirectionVector(double measuredDirectionVector[3]) {
      MeasuredDirectionVector[0] = measuredDirectionVector[0];
      MeasuredDirectionVector[1] = measuredDirectionVector[1];
      MeasuredDirectionVector[2] = measuredDirectionVector[2];
    }
    double *GetMeasuredDirectionVector() {return MeasuredDirectionVector;}

    void SetSigmaRange(double sigmaRange) {
      SigmaRange = sigmaRange;
    }
    double GetSigmaRange() {return SigmaRange;}

    void SetSigmaRangeRate(double sigmaRangeRate) {
      SigmaRangeRate = sigmaRangeRate;
    }
    double GetSigmaRangeRate() {return SigmaRangeRate;}

    void SetSigmaAngle(double sigmaAngle) {
      SigmaAngle = sigmaAngle;
    }
    double GetSigmaAngle() {return SigmaAngle;}

    void GetMeasuredPosition(double measuredPosition[3]) {
      for (int i=0; i<3; i++) {
	measuredPosition[i] = TrueSensorPosition[i]
	  + MeasuredDirectionVector[i] * MeasuredRange;
      }
    }

    virtual void Print(ostream &out) {
      double measuredPosition[3];
      double R1[3];
      double R2[3];

      out << "Detection of " << GlobalId << " at time " << Time << endl;
      out << "True Position = " << TrueTargetPosition[0]
	  << " " << TrueTargetPosition[1]
	  << " " << TrueTargetPosition[2]
	  << endl;
      GetMeasuredPosition(measuredPosition);
      out << "Meas Position = " << measuredPosition[0]
	  << " " << measuredPosition[1]
	  << " " << measuredPosition[2]
	  << endl;
      out << "Error ="
	  << " " << measuredPosition[0] - TrueTargetPosition[0]
	  << " " << measuredPosition[1] - TrueTargetPosition[1]
	  << " " << measuredPosition[2] - TrueTargetPosition[2]
	  << endl;

      Difference(TrueTargetPosition, TrueSensorPosition, R1);
      Difference(measuredPosition, TrueSensorPosition, R2);

      out << "True Range = " << Magnitude(R1)
	  << ", Measured Range = " << Magnitude(R2)
	  << endl;

      out << endl;
    }

};

RB_DEFINE_CLASS(RADAR_DETECTION);

#endif
