// EntityDB.C

#include "EntityDB.H"

int EntityDB::done = 0;

BINARY_TREE *EntityDB::EntitiesList = NULL;
BINARY_TREE *EntityDB::EntityTypesList = NULL;

DATA_PARSER *EntityDB::SensorComponents = NULL;
DATA_PARSER *EntityDB::SensorTypes = NULL;

DATA_PARSER *EntityDB::TrackCorrelationComponents = NULL;
DATA_PARSER *EntityDB::TrackCorrelationTypes = NULL;

DATA_PARSER *EntityDB::WeaponComponents = NULL;
DATA_PARSER *EntityDB::WeaponTypes = NULL;

DATA_PARSER *EntityDB::PlatformComponents = NULL;
DATA_PARSER *EntityDB::PlatformTypes = NULL;

DATA_PARSER *EntityDB::ComnmandControlComponents = NULL;
DATA_PARSER *EntityDB::CommandControlTypes = NULL;

DATA_PARSER *EntityDB::MotionComponents = NULL;
DATA_PARSER *EntityDB::MotionTypes = NULL;

DATA_PARSER *EntityDB::CommunicationComponents = NULL;
DATA_PARSER *EntityDB::CommunicationTypes = NULL;

EntityDB::EntityDB(char *entityFileName, char *entityTypeFileName) {

  CurrentEntities = new DATA_PARSER(entityFileName);
  CurrentEntityTypes = new DATA_PARSER(entityTypeFileName);

  Mark = CurrentEntities->GoTo("Objects", NULL);
  NumObjects = CurrentEntities->GetNumReferences(Mark, "OBJECT");

  if (!done) {
    done = 1;

    EntitiesList = new BINARY_TREE();
    EntityTypesList = new BINARY_TREE();

    SensorComponents = new DATA_PARSER("SensorComponents.par");
    SensorTypes = new DATA_PARSER("SensorTypes.par");
    
    TrackCorrelationComponents = new DATA_PARSER("TrackCorrelationComponents.par");
    TrackCorrelationTypes = new DATA_PARSER("TrackCorrelationTypes.par");

    WeaponComponents = new DATA_PARSER("WeaponComponents.par");
    WeaponTypes = new DATA_PARSER("WeaponTypes.par");

    PlatformComponents = new DATA_PARSER("PlatformComponents.par");
    PlatformTypes = new DATA_PARSER("PlatformTypes.par");

    ComnmandControlComponents = new DATA_PARSER("ComnmandControlComponents.par");
    CommandControlTypes = new DATA_PARSER("CommandControlTypes.par");

    MotionComponents = new DATA_PARSER("MotionComponents.par");
    MotionTypes = new DATA_PARSER("MotionTypes.par");

    CommunicationComponents = new DATA_PARSER("CommunicationComponents.par");
    CommunicationTypes = new DATA_PARSER("CommunicationTypes.par");
  }

  EntitiesList->Insert(CurrentEntities, entityFileName);
  EntityTypesList->Insert(CurrentEntityTypes, entityTypeFileName);

}


S_HLA *EntityDB::CreateFirstEntity(int numNodes, int offset,
				      int replication, int &done) {
  S_HLA *entity;

  if (NumObjects == 0) {
    entity = NULL;
    done = 1;
    return entity;
  }

  Index = 0;

  CurrentEntities->GoToFirstReference(Mark, "OBJECT");
  if (ForMyNode(Index, offset, numNodes)) {
    entity = CreateEntity();
    entity->set_SEED(10000*replication + 12345 + Index);
    entity->set_NAME(CurrentEntities->GetReferenceName());
    CreateComponents(entity);
  }else{
    entity = NULL;
  }

  if (Index == NumObjects-1) {
    done = 1;
  }else{
    done = 0;
  }

  return entity;

}


S_HLA *EntityDB::CreateNextEntity(int numNodes, int offset,
				     int replication, int &done) {
  S_HLA *entity;

  Index++;

  if (NumObjects == Index) {
    entity = NULL;
    done = 1;
    return entity;
  }else{
    done = 0;
  }

  CurrentEntities->GoToNextReference(Mark, "OBJECT");
  if (ForMyNode(Index, offset, numNodes)) {
    entity = CreateEntity();
    entity->set_SEED(10000*replication + 12345 + Index);
    entity->set_NAME(CurrentEntities->GetReferenceName());
    CreateComponents(entity);
  }else{
    entity = NULL;
  }

  return entity;

}
