
/************************************************************************/
/*                                                                      */
/*                              Metron Inc.                             */
/*                          11911 Freedom Drive                         */
/*                              Suite 800                               */
/*                           Reston, Virginia                           */
/*                            (703) 787-8700                            */
/*                                                                      */
/*                       Copyright 1995 by METRON INC.                  */
/*                           ALL RIGHTS RESERVED                        */
/*                                                                      */
/************************************************************************/


#if !defined( LATLONdotH )
#define LATLONdotH

#include "LatLonConstants.H"
#include "SpMath.H"

inline double asin_degrees(double d) {
  return 180/LATLON_PI*asin(d);
}

inline double acos_degrees(double d) {
  return 180/LATLON_PI*acos(d);
}

inline double atan_degrees(double d) {
  return atan(d)*180/LATLON_PI;
}

inline double cos_degrees(double d) {
  return cos(LATLON_PI/180*d);
}

inline double sin_degrees(double d) {
  return sin(LATLON_PI/180*d);
}


//Utility Material//////////////////////////////////////////////////////
double FMOD(double a, double mod);
double NormLon(double lon);
double aTrigd(double y, double x, int which);

class LatLon;
class LatLonAlt;

//Vect3//////////////////////////////////////////////////////////////
class Vect3
{
private:
	double x, y, z;
public:
	Vect3();
	Vect3(double x, double y, double z, int normalize=0);
	Vect3(const LatLon &ltLn);
	Vect3(const LatLonAlt &ltLnAlt);
	double X() const {return x;}
	double Y() const {return y;}
	double Z() const {return z;}
	double VSize() const;
	double operator,(const Vect3 &v) const {return x*v.x + y*v.y + z*v.z;}
	const Vect3 operator+(const Vect3 &v) const
	{
		return Vect3(x+v.x, y+v.y, z+v.z);
	};
	const Vect3 operator-(const Vect3 &v) const
	{
		return Vect3(x-v.x, y-v.y, z-v.z);
	};
	const Vect3 operator*(const Vect3 &v) const
	{
		return Vect3(y*v.z-z*v.y, z*v.x-x*v.z, x*v.y-y*v.x);
	};
	const Vect3 operator*(double alpha) const
	{
		return Vect3(alpha*x, alpha*y, alpha*z);
	};
};
inline const Vect3 operator*(double alpha, const Vect3 &v)
{
	return v*alpha;
};
//ostream&  operator<<(ostream& out, const Vect3 &v);

//RangeAzimuth//////////////////////////////////////////////////////////
class RangeAzimuth
{
private:
	double range, azimuth;
public:
	RangeAzimuth();
	RangeAzimuth(double r, double a);
	double Range() const {return range;};
	double Azimuth() const {return azimuth;};
};
//ostream& operator<<(ostream& out, const RangeAzimuth &ra);

//LatLon////////////////////////////////////////////////////////////
class LatLon
{
private:
	double lat, lon;
public:
	LatLon(double lat, double lon);
	LatLon(const Vect3 &v);
	double Lat() const {return lat;};
	double Lon() const {return lon;};
	int operator==(const LatLon &ltLn) const;
	int operator!=(const LatLon &ltLn) const {return !operator==(ltLn);};
	const LatLon operator+(const RangeAzimuth &ra) const;
};
//ostream& operator<<(ostream& out, const LatLon &ltLn);

//LatLonAlt////////////////////////////////////////////////////////////
struct LatLonAlt
{
	LatLon ll;
	double alt;
	LatLonAlt(const LatLon &llIn, double altIn);
	LatLonAlt(const Vect3 &v);
};
//ostream& operator<<(ostream& out, const LatLonAlt &ltLnAlt);

//Global functions involving LatLons and RangeAzimuths.
RangeAzimuth rangeAzimuthFromFirstToSecond(
	const LatLon &latLon0, const LatLon &latLon1
	);
RangeAzimuth rangeAzimuthToArc(
	const LatLon &startingLatLon,
	const LatLon &endpoint0, const LatLon &endpoint1
);
RangeAzimuth twoDRangeAzimuthToArc(
	const LatLon &startingLatLon,
	const LatLon &endpoint0, const LatLon &endpoint1
);
void getRngeAzElAng(
	double lat1, double lon1, double alt1,
	double lat2, double lon2, double alt2,
	double &range, double &elAngle, double &azimuth
	);
void getRelElAz(
	double az1, double elAng1, double az2, double elAng2,
	double &relEl12, double &relAz12
	);
double whenHitAzimuth(
	const LatLonAlt &lla0, const LatLonAlt &lla1, const LatLonAlt &lla2,
	double azimuthToHit, int &returnCode
);
void whenHitElAng(
	const LatLonAlt &lla0, const LatLonAlt &lla1, const LatLonAlt &lla2,
	double elAngToHit, double &p1, double &p2, int &returnCode
);
LatLonAlt endAt(LatLonAlt lla, double range, double azimuth, double elAng);

#endif
