// O_RanMot.C method file

#include "O_RanMot.H"
#include "ScheduleEvent.H"
#include "E_CheckRange.H"

#include <iostream.h>
#include <stdlib.h>
#include <string.h>

/************************************************************************
* O_RanMot : construct a random motion manager          		*
************************************************************************/
O_RanMot::O_RanMot() {
  int Nobjects, Nvmax;
  int offset;
  int i;
  int GlobalId;
  char s[256];
  OBJECT_PROXY *ObjectProxy;
  double *Vmax;

  cout << "Constructing O_RanMot" << endl;

/*
** Read in the input file using the Data Parser and extract parameters
*/

  RanMotParser = new DATA_PARSER("RanMot.par");

  RanMotParser->GoTo("parameters", NULL);
  Nobjects = RanMotParser->GetInt("Nobjects");

  Nvmax = RanMotParser->GetArraySize("Vmax");
  assert(Nvmax>0);
  Vmax = new double [Nvmax];
  for (i=0; i<Nvmax; i++){
    Vmax[i] = RanMotParser->GetFloat("Vmax", i);
  }

/*
** Determine how many objects to create on this node
*/

  N_TOT = Nobjects;
  N_LOC = deal_me(N_TOT, offset);

/*
** Create N_LOC objects for this node
*/

  RanMot = new S_RanMot[N_LOC];
  GlobalId = TOTAL_OBJECTS + offset;

/*
** Loop through each object that has been created and initialize them
*/

  int j = 0;
  int v = 0;

cout << "O_RanMot node " << NODE << " nloc " << N_LOC << endl;

  for (i=0; i<N_TOT; i++) {
    if ((i-offset+N_NODES)%N_NODES == 0) {
      RanMot[j].set_GLOBAL_ID(GlobalId);
      ObjectProxy = RanMot[j].GetObjectProxy();
      ObjectProxy->SetGlobalId(GlobalId);

      int seed = 10000 + i + 54321;

      //cout << "O_RanMot::O_RanMot: NODE= " << NODE << ", GlobalId="
      //<< GlobalId << ", seed=" << seed << endl ;

      RanMot[j].set_SEED(seed);
      RanMot[j].SetVmax(Vmax[v++]);
      if (v==Nvmax) v = 0;

      RanMot[j].InitParser(RanMotParser);

      sprintf(s, "RanMot %d", GlobalId);
      RanMot[j].set_NAME(s);

      GlobalId += N_NODES;

      j++;

    }
  }

/*
** Tell SPEEDES about the local objects that have been created
*/

  reserve_n_objects(N_LOC);

  for (i=0; i<N_LOC; i++) {
    define_object(&RanMot[i]);
  }

  TOTAL_OBJECTS += N_TOT;

#ifdef VALIDATE_DDM
  if (N_NODES==1){
    SpSimTime event_time(0.0);
    event_time.SetPriority(MaxPriority1);
    ScheduleEvent(event_time, CheckRange, get_OBJECT_TYPE(), get_LOCAL_ID(), get_NODE());
  }
#endif
}

/************************************************************************
* init_events : initialize events 				        *
************************************************************************/
void O_RanMot::init_events() {

  cout << "Initializing events for RanMot objects" << endl;

  for (int i=0; i<N_LOC; i++) {
    RanMot[i].InitEvents();
  }

}

/************************************************************************
* terminate : termination method 				        *
************************************************************************/
void O_RanMot::terminate(double tend) {

}
