// P_RadarPrecomp.C

#include "P_RadarPrecomp.H"

#include "S_Entity.H"
#include "EntityProxy.H"
#include "process.H"
#include "RB.H"
#include "F_ProxyItem.H"

#include "SpEnterExitRange.H"

extern int DebugRadarPrecomp;

int FindRemoteProxy(S_ENTITY *entity, F_PROXY_ITEM *proxy_item){
  int i, found;
  F_SpProxyItem *proxy;
  C_XQUEUE *proxies;

  proxies = entity->GetRemoteObjectProxies();
  proxy   = (F_SpProxyItem *) proxies->get_top();
  for (found=0, i=0; i<proxies->get_length() && !found; i++){
    if (proxy==proxy_item) found = 1;
    proxy = (F_SpProxyItem *) proxy->get_link();
  }

  return (found);
}

#define _DEBUG
//#define OBJECT_ID_ENABLED

/*
** Local variables (LV) that are persistent with this process are defined
** here. The PROCESS_LV and BEGIN_PROCESS macros really define the variables
** in a locally defined class called LV. This is why you need to use the LV->
** sytax for accessing these variables.
*/

PROCESS_LV (P_RadarPrecomp, process)
RB_int HappyVariable;
int LoopForever;
RADAR_COMPONENT *RadarComponent;
F_PROXY_ITEM *ProxyItem;
double RadarRange;
double ScanTime;
RB_BINARY_TREE *DetectableProxies;
SpCounterSem *NumDetectableProxies;
OBJECT_ID *objectId;
LogicalSemaphore RemoteProxySemaphore;
int successFlag;
RB_int InRangeFlag;
LogicalSemaphore BogusSemaphore;
BEGIN_PROCESS (P_RadarPrecomp, process, 9)

/*
** These variables are defined as temporary. You should not assume their
** values are set correctly after waking up after time has elapsed. The
** reason we define temporary variables is for efficiency. Loop variables,
** for example, which do not pass time in the loop are much more efficient
** than their rollbackable counterparts.
*/

S_ENTITY *Entity;
ENTITY_PROXY *EntityProxy;
ENTITY_PROXY *RemoteEntityProxy;
double X1[3], V1[3], A1[3];
double X2[3], V2[3], A2[3];
double R[3];
double Range;
F_PROXY_ITEM *ProxyItem1;
M_RadarPrecomp *radarPrecompMsg;
//double EarliestTime;

/*
** The first thing we do when entering into a process is to access whatever
** data was passed in through the MESSAGE and store it in the LV variables.
** This is only done once, when the process is first entered. After than,
** there is no user data passed in the message.
*/

radarPrecompMsg = (M_RadarPrecomp *)MESSAGE;
LV->ProxyItem = radarPrecompMsg->ProxyItem;
LV->RadarComponent = radarPrecompMsg->RadarComponent;

LV->RadarRange = (LV->RadarComponent)->GetRange();
LV->ScanTime = (LV->RadarComponent)->GetScanTime();
LV->DetectableProxies = (LV->RadarComponent)->GetDetectableProxies();
LV->NumDetectableProxies = (LV->RadarComponent)->GetNumDetectableProxies();

/*
** Add the remote proxy semaphore here (we will have to make sure it is removed
** before exiting this process)
*/

RemoteEntityProxy = (ENTITY_PROXY *)(LV->ProxyItem)->GetObjectProxy();
(LV->RadarComponent)->InsertRemoteProxySemaphore
  (&(LV->RemoteProxySemaphore),
   RemoteEntityProxy->GetGlobalId());

/*
** First time in, check if the remote proxy is already in range. If so, we
** need to add the proxy to the detectable proxy list, etc.
*/

*(LV->RadarComponent->GetNumEnterProxies()) += 1;

Entity = (S_ENTITY *)SIMOBJ;

EntityProxy = (ENTITY_PROXY *)Entity->GetObjectProxy();
EntityProxy->Get_Position (X1, V1, A1, ECR, TIME_TAG);

RemoteEntityProxy = (ENTITY_PROXY *)(LV->ProxyItem)->GetObjectProxy();

RemoteEntityProxy->Get_Position (X2, V2, A2, ECR, TIME_TAG);

Difference(X1,X2,R);
Range = Magnitude(R);

if (Range <= LV->RadarRange) {

  LV->InRangeFlag = 1;

  ProxyItem1 = (F_PROXY_ITEM *)(LV->DetectableProxies)->Find
    (RemoteEntityProxy->GetGlobalId());

  if (ProxyItem1) {
    if (ProxyItem1 != LV->ProxyItem) {
//      cout << "Error, bad found proxy item" << endl;
    }
  }else{

    (*(LV->NumDetectableProxies))++;

    (*(LV->DetectableProxies)).Insert
      (LV->ProxyItem, RemoteEntityProxy->GetGlobalId());

#ifdef OBJECT_ID_ENABLED
    LV->objectId = RB_NEW_OBJECT_ID();
    (LV->objectId)->SetObjectId(RemoteEntityProxy->GetGlobalId());
    (LV->objectId)->SetObjectName(RemoteEntityProxy->GetName());
    (LV->RadarComponent)->AddObjectId(LV->objectId);
#endif

    if (DebugRadarPrecomp) {
      RB_cout << "Adding proxy item" << endl;
    }

  }

}else{

  LV->InRangeFlag = 0;

}

/*
** Here we loop forever (note that the LoopForever variable must be a of the
** LV type since it is used even when time passes.
*/

LV->LoopForever = 1;
while (LV->LoopForever) {

/*
** First get the position, velocity, and acceleration of my object
** (done through my proxy)
*/

  Entity = (S_ENTITY *)SIMOBJ;

  EntityProxy = (ENTITY_PROXY *)Entity->GetObjectProxy();
//  EntityProxy->Get_Position (X1, V1, A1, ECR, TIME_TAG);

/*
** Then, get the position, velocity, and acceleration of the remote entity
** through its proxy.
*/

  RemoteEntityProxy = (ENTITY_PROXY *)(LV->ProxyItem)->GetObjectProxy();
//  RemoteEntityProxy->Get_Position (X2, V2, A2, ECR, TIME_TAG);

/*
** Call the SpEnterExitRange function to determine when the remote object
** enters or exits my range. This is just for testing right now...
*/

  double time;
  int flag;

  SpEnterExitRange(EntityProxy,
		   RemoteEntityProxy,
		   EntityProxy->GetVmax(),
		   RemoteEntityProxy->GetVmax(),
		   TIME_TAG,
		   LV->RadarRange,
		   36000.0,
		   time,
		   flag);

  if (DebugRadarPrecomp) {
    cout << ">>> SpEnterExitRange time = " << time
	 << ", flag = " << flag << endl;
  }

  LV->RemoteProxySemaphore = 0;

  switch (flag) {

/*
** This is the case where there is no solution
*/
    case RANGE_NO_SOLUTION:

      if (DebugRadarPrecomp) {
	RB_cout << "RANGE_NO_SOLUTION" << endl;
      }

      WAIT_FOR_LOGICAL(1,
		       LV->RemoteProxySemaphore,
		       -1,
		       LV->successFlag);

      break;

/*
** This is the case where we should recompute later since the earliest
** intersection occurs too far out into the future to worry about it now
*/
    case RANGE_COMPUTE_LATER:

      if (DebugRadarPrecomp) {
	RB_cout << "RANGE_COMPUTE LATER at time " << time << endl;
      }

      WAIT_FOR_LOGICAL(2,
		       LV->RemoteProxySemaphore,
		       time-TIME_TAG,
		       LV->successFlag);

      if (!FindRemoteProxy((S_ENTITY *) SIMOBJ, LV->ProxyItem)){
        LV->BogusSemaphore = 0;
        WAIT_FOR_LOGICAL(3, LV->BogusSemaphore, -1, LV->successFlag);
      }

      if (LV->RemoteProxySemaphore) {
	if (DebugRadarPrecomp) {
	  RB_cout << "Wait for logical was interrupted "
		  << LV->RemoteProxySemaphore << endl;
	}
      }
      break;

/*
** This is the case where the remote proxy enters my range
*/
    case RANGE_ENTER:

      if (DebugRadarPrecomp) {
	RB_cout << "RANGE_ENTER at time " << time << endl;
      }

      if (LV->InRangeFlag == 1) {
//	RB_cout << "Error, entering when already in range" << endl;
      }

      WAIT_FOR_LOGICAL(4,
		       LV->RemoteProxySemaphore,
		       time-TIME_TAG,
		       LV->successFlag);

      if (!FindRemoteProxy((S_ENTITY *) SIMOBJ, LV->ProxyItem)){
        LV->BogusSemaphore = 0;
        WAIT_FOR_LOGICAL(5, LV->BogusSemaphore, -1, LV->successFlag);
      }

      if (LV->RemoteProxySemaphore) {
	if (DebugRadarPrecomp) {
	  RB_cout << "Wait for logical was interrupted "
		  << LV->RemoteProxySemaphore << endl;
	}
	break;
      }

      Entity = (S_ENTITY *)SIMOBJ;

      EntityProxy = (ENTITY_PROXY *)Entity->GetObjectProxy();
      EntityProxy->Get_Position (X1, V1, A1, ECR, TIME_TAG+0.001);

      RemoteEntityProxy = (ENTITY_PROXY *)(LV->ProxyItem)->GetObjectProxy();
      RemoteEntityProxy->Get_Position (X2, V2, A2, ECR, TIME_TAG+0.001);

      Difference(X1,X2,R);
      Range = Magnitude(R);

      if (Range >= LV->RadarRange) {
//	RB_cout << "Error, bad RANGE_ENTER" << endl;
	break;
      }

      ProxyItem1 = (F_PROXY_ITEM *)(LV->DetectableProxies)->Find
	(RemoteEntityProxy->GetGlobalId());

      if (ProxyItem1) {
	if (ProxyItem1 != LV->ProxyItem) {
//	  cout << "Error, bad found proxy item" << endl;
	}
      }else{

	LV->InRangeFlag = 1;

	(*(LV->NumDetectableProxies))++;

	(*(LV->DetectableProxies)).Insert
	  (LV->ProxyItem, RemoteEntityProxy->GetGlobalId());

#ifdef OBJECT_ID_ENABLED
	LV->objectId = RB_NEW_OBJECT_ID();
	(LV->objectId)->SetObjectId(RemoteEntityProxy->GetGlobalId());
	(LV->objectId)->SetObjectName(RemoteEntityProxy->GetName());
	(LV->RadarComponent)->AddObjectId(LV->objectId);
#endif

	if (DebugRadarPrecomp) {
	  RB_cout << "Adding proxy item" << endl;
	}

      }

      break;

/*
** This is the case where the remote proxy exits my range
*/
    case RANGE_EXIT:

      if (DebugRadarPrecomp) {
	RB_cout << "RANGE_EXIT at time " << time << endl;
      }

      if (LV->InRangeFlag == 0) {
//	RB_cout << "Error, exiting when not in range" << endl;
      }

      WAIT_FOR_LOGICAL(6,
		       LV->RemoteProxySemaphore,
		       time-TIME_TAG,
		       LV->successFlag);

      if (!FindRemoteProxy((S_ENTITY *) SIMOBJ, LV->ProxyItem)){
        LV->BogusSemaphore = 0;
        WAIT_FOR_LOGICAL(7, LV->BogusSemaphore, -1, LV->successFlag);
      }

      if (LV->RemoteProxySemaphore) {
	if (DebugRadarPrecomp) {
	  RB_cout << "Wait for logical was interrupted "
		  << LV->RemoteProxySemaphore << endl;
	}
	break;
      }

      Entity = (S_ENTITY *)SIMOBJ;

      EntityProxy = (ENTITY_PROXY *)Entity->GetObjectProxy();
      EntityProxy->Get_Position (X1, V1, A1, ECR, TIME_TAG+0.001);

      RemoteEntityProxy = (ENTITY_PROXY *)(LV->ProxyItem)->GetObjectProxy();
      RemoteEntityProxy->Get_Position (X2, V2, A2, ECR, TIME_TAG+0.001);

      Difference(X1,X2,R);
      Range = Magnitude(R);

      if (Range <= LV->RadarRange) {
//	RB_cout << "Error, bad RANGE_EXIT" << endl;
	break;
      }

      ProxyItem1 = (F_PROXY_ITEM *)(LV->DetectableProxies)->Remove
	(RemoteEntityProxy->GetGlobalId());

      if (ProxyItem1) {
        *(LV->RadarComponent->GetNumExitProxies()) += 1;

	(*(LV->NumDetectableProxies))--;

	LV->InRangeFlag = 0;

#ifdef OBJECT_ID_ENABLED
	(LV->RadarComponent)->RemoveObjectId(LV->objectId);
	RB_DELETE_OBJECT_ID(LV->objectId);
#endif

	if (ProxyItem1 != LV->ProxyItem) {
	  if (DebugRadarPrecomp) {
//	    RB_cout << "Error,  bad removed proxy item" << endl;
	  }
	}else{
	  if (DebugRadarPrecomp) {
	    RB_cout << "Removing proxy item" << endl;
	  }
	}

      }

      break;

/*
** This is the case where the range computationi reports an error (it should
** never occur)
*/
    case RANGE_ERROR:

      if (DebugRadarPrecomp) {
	RB_cout << "RANGE_ERROR" << endl;
      }

      WAIT_FOR_LOGICAL(8,
		       LV->RemoteProxySemaphore,
		       -1,
		       LV->successFlag);

      break;

  }

  if (!FindRemoteProxy((S_ENTITY *) SIMOBJ, LV->ProxyItem)){
    LV->BogusSemaphore = 0;
    WAIT_FOR_LOGICAL(9, LV->BogusSemaphore, -1, LV->successFlag);
  }

}

END_PROCESS (P_RadarPrecomp, process)


