// Query.C file

#include <iostream.h>
#include <stdlib.h>
#include <signal.h>

#include "EmHostUser.H"
#include "Defines.H"

/*
** Create the External Module Host User object. It assumes that there is a
** speedes.par file to tell it SpeedesServer information.
*/

EM_HOST_USER EmHostUser;

void sigintfunc(int) {
  cout << "Interupt" << endl;
  signal(SIGINT, sigintfunc);
  exit(1);
}

/*
*****************************************************************************
* Query <ObjectName>: queries the SPEEDES object and prints out the response
*****************************************************************************
*/

int main(int argc, char **argv) {
  NAME_VALUE_LIST *NameValueList;

  signal(SIGINT, sigintfunc);

/*
** Make sure that there is only one or two arguments:
*/

  if ((argc != 2) && (argc != 3)) {
    cout << "usage: " << argv[0] << " objectName [queryName]" << endl;
    exit(1);
  }

  char *objectName = argv[1];
  char *queryName = NULL;

/*
** Do the query and then print out the result.
*/

  if (argc == 2) {
    NameValueList = EmHostUser.Query(objectName);
  }else{
    queryName = argv[2];
    NameValueList = EmHostUser.Query(objectName, queryName);
  }

/*
** Interpretation of the contents of NameValueList depends on queryName (see S_Ship.C):
*/

  if (NameValueList) {

    if (queryName) {
      char *name;
      double time;
      NameValueList->GetNameValue(name, time);


      if (strcmp(queryName, "Vmax") == 0) {
	double vmax_kms;
	NameValueList->GetNameValue(name, vmax_kms);
	double vmax_knots = vmax_kms * NauticalMiles_per_Kilometer *
	  Seconds_per_Hour;

	cout << endl << "Vmax of " << objectName << " at time " << time 
	     << " is " << vmax_kms << " km/s = " << vmax_knots << " knots" 
	     << endl << endl;
      }


      else if (!strcmp(queryName, "Position")) {
	double x[3], v[3], a[3];

	NameValueList->GetNameValue(name, x[0]);
	NameValueList->GetNameValue(name, x[1]);
	NameValueList->GetNameValue(name, x[2]);
	NameValueList->GetNameValue(name, v[0]);
	NameValueList->GetNameValue(name, v[1]);
	NameValueList->GetNameValue(name, v[2]);
	NameValueList->GetNameValue(name, a[0]);
	NameValueList->GetNameValue(name, a[1]);
	NameValueList->GetNameValue(name, a[2]);
	
	double lat = x[0] * DEGRAD;
	double lon = x[1] * DEGRAD;
	double alt = x[2];

	cout << endl << "Position of " << objectName << " at time " << time
	     << " is: lat=" << lat << "   lon=" << lon << "   alt=" << alt 
	     << endl << endl;
      }


      else if (!strcmp(queryName, "Alive")) {
	double alive;
	NameValueList->GetNameValue(name, alive);

	if (alive) {
	  cout << endl << objectName << " is alive at time " << time
	       << endl << endl;
	}

	else {
	  cout << endl << objectName << " is not alive at time " << time
	       << endl << endl;
	}
      }


      else if (!strcmp(queryName, "PlatformComponents")) {
	double num;
	NameValueList->GetNameValue(name, num);

	cout << endl << "Number of platform components on " << objectName 
	     << " at time " << time << " is " << num
	     << endl << endl;
      }


      else if (!strcmp(queryName, "MotionComponents")) {
	double num;
	NameValueList->GetNameValue(name, num);

	cout << endl << "Number of motion components on " << objectName 
	     << " at time " << time << " is " << num
	     << endl << endl;
      }


      else if (!strcmp(queryName, "WeaponComponents")) {
	double num;
	NameValueList->GetNameValue(name, num);

	cout << endl << "Number of weapon components on " << objectName 
	     << " at time " << time << " is " << num
	     << endl << endl;
      }


      else if (!strcmp(queryName, "SensorComponents")) {
	double numComponents;
	NameValueList->GetNameValue(name, numComponents);

	cout << endl << "Number of sensor components on " << objectName 
	     << " at time " << time << " is " << numComponents << ":"
	     << endl << endl;

	for (int component = 0; component < numComponents; component++) {
	  double range;
	  NameValueList->GetNameValue(name, range);
	  double scan_time;
	  NameValueList->GetNameValue(name, scan_time);
	  double not_used;
	  char *radar_type;
	  NameValueList->GetNameValue(radar_type, not_used);
	  double sensor_id;
	  NameValueList->GetNameValue(name, sensor_id);

	  cout << radar_type << ":" << endl
	       << "  Range: " << range << endl
	       << "  Scan Time: " << scan_time << endl
	       << "  Sensor ID: " << sensor_id
	       << endl << endl;
	}
      }

      else if (!strcmp(queryName, "TrackerCorrelationComponents")) {
	double num;
	NameValueList->GetNameValue(name, num);

	cout << endl << "Number of tracker/correlator components on " << objectName 
	     << " at time " << time << " is " << num
	     << endl << endl;
      }


      else if (!strcmp(queryName, "CommandControlComponents")) {
	double num;
	NameValueList->GetNameValue(name, num);

	cout << endl << "Number of C2 components on " << objectName 
	     << " at time " << time << " is " << num
	     << endl << endl;
      }


      else if (!strcmp(queryName, "CommunicationComponents")) {
	double num;
	NameValueList->GetNameValue(name, num);

	cout << endl << "Number of communication components on " << objectName 
	     << " at time " << time << " is " << num
	     << endl << endl;
      }


      else {
	cout << "Query name \"" << queryName << "\" is unknown." << endl;
      }
    }

    else {
    }

    delete NameValueList; 
  }	

  return (0);
}
