// RadarComponent.C file

#include "RadarComponent.H"
#include "ScheduleEvent.H"
#include "M_RadarScan.H"

EVENT_TYPE(RadarScan);

RADAR_COMPONENT::RADAR_COMPONENT(S_ENTITY *entity,
				 DATA_PARSER *parser,
				 char *radarType,
				 int sensorId)
  : SENSOR_COMPONENT(entity) {

  S_Entity = entity;
  RadarType = radarType;
  SensorId = sensorId;

  parser->GoTo(radarType, NULL);

  ScanTime = parser->GetFloat("ScanTime");
  Range = parser->GetFloat("Range");

  RadarSensorModel.SetRange(Range);
  RadarSensorModel.SetScanTime(ScanTime);

  RadarSensorModel.SetSigmaRange(parser->GetFloat("SigmaRange"));
  RadarSensorModel.SetSigmaRangeRate(parser->GetFloat("SigmaRangeRate"));
  RadarSensorModel.SetSigmaAngle(parser->GetFloat("SigmaAngle"));

  DiscoverObject.SetRadarComponent(this);
  ReflectAttributes.SetRadarComponent(this);
  UpdateAttributes.SetRadarComponent(this);

  AddEventHandler("Update Attributes",  &UpdateAttributes);
  AddEventHandler("Reflect Attributes", &ReflectAttributes);
  AddEventHandler("Discover Object",    &DiscoverObject);

//  cout << "Creating Radar Component of type " << radarType << endl;

}

void RADAR_COMPONENT::InitEvents() {
  M_RadarScan radarScanMsg;

#ifdef NOISY
  cout << "RadarScan=" << RadarScan
       << ", ObjectType=" << ObjectType
       << ", LocalId=" << LocalId
       << ", Node=" << Node
       << endl;
#endif

  radarScanMsg.RadarComponent = this;
  ScheduleProcess(0.0, RadarScan, ObjectType, LocalId, Node, &radarScanMsg);
}
