// S_RanMot.C source file

#define ENABLE_DDM

#include "S_RanMot.H"
#include "GreatCircle.H"
#include "ConstMotion.H"
#include "RadarComponent.H"
#include "Defines.H"
#ifdef ENABLE_DDM
#include "SpDDMRegion.H"
#include "SpGeoFunctions.H"
#endif

extern int DebugObjectProxy;
extern int DebugRemoteObjectProxy;

#include "FreeDynamicAttributes.H"
extern FREE_DYNAMIC_ATTRIBUTES *FreeDynamicAttributes;

extern DATA_PARSER *speedes_parser;
extern DATA_PARSER *SensorParser;

#include "F_ProxyItem.H"
#include "SpSubscribeSpace.H"
#include "SpPublishSpace.H"
extern C_EVENT *CurrentEvent;

static void NewObjProxy(SpSubscribeSpace *space, F_SpProxyItem *item){
  RB_cout << "simobj " << space->GetSimObj()->get_GLOBAL_ID()
       << " node " << space->GetSimObj()->get_NODE()
       << " sub space " << space->GetName() << " rcv'd new proxy "
       << item->GetGlobalId() << " @ " << CurrentEvent->get_MESSAGE()->time_tag << endl;
}

static void RemoveObjProxy(SpSubscribeSpace *space, int globalId){
  RB_cout << "simobj " << space->GetSimObj()->get_GLOBAL_ID()
       << " node " << space->GetSimObj()->get_NODE()
       << " sub space " << space->GetName() << " rcv'd rem proxy " << globalId 
       << " @ " << CurrentEvent->get_MESSAGE()->time_tag << endl;
}


S_RanMot::S_RanMot() : S_MOVING_ENTITY("RanMot"){

//  cout << "Creating Random Motion Object" << endl;

}

int S_RanMot::max_leg = 0;

void S_RanMot::InitParser(DATA_PARSER *parser) {
  GREAT_CIRCLE *GreatCircle;
  CONSTANT_MOTION *ConstantMotion;
  double Lat, Lon, Alt;
  double OldLat, OldLon, OldAlt, OldTime;
  int Nsensors = 0;

/*
** Create a RanMotRadar component
*/

  RADAR_COMPONENT *RadarComponent = new RADAR_COMPONENT
    (this, SensorParser, "RanMotRadar", Nsensors);
  SensorComponents.Insert(RadarComponent, Nsensors);
  AddComponent(RadarComponent);

/*
** Generate random great circle motion
*/

  RANDOM->set_seed(SEED);

  RANDOM->set_float_limits(-PI_OVER2, PI_OVER2);
  OldLat = RANDOM->get_random_float();

  RANDOM->set_float_limits(-PI, PI);
  OldLon = RANDOM->get_random_float();

  RANDOM->set_float_limits(0.0, 10.0);
  OldAlt = RANDOM->get_random_float();

  OldTime = 0.0;

//cout << "SIMOBJ " << get_GLOBAL_ID() << " (" << OldLat << ", " << OldLon << ", " << OldAlt << ")" << endl;

  speedes_parser->GoTo("parameters", NULL);
  double tend = speedes_parser->GetFloat("tend");

#define MEMORY_PROBLEM
#ifdef  MEMORY_PROBLEM
  int num_leg = 0;
  const int max_turn = 100;
  while ((OldTime < tend) && (++num_leg<max_turn)){
#else
  while (OldTime < tend) {
#endif

    GreatCircle = (GREAT_CIRCLE *)
      FreeDynamicAttributes->new_object(GREAT_CIRCLE_ID);

    RANDOM->set_float_limits(-PI_OVER2, PI_OVER2);
    Lat = RANDOM->get_random_float();

    RANDOM->set_float_limits(-PI, PI);
    Lon = RANDOM->get_random_float();

    RANDOM->set_float_limits(0.0, 10.0);
    Alt = RANDOM->get_random_float();

    GreatCircle->init_Vconstant(OldLat, OldLon, OldAlt, OldTime,
				Lat, Lon, Alt, Vmax);

    OldLat = Lat;
    OldLon = Lon;
    OldAlt = Alt;
    OldTime = GreatCircle->GetEndTime();

    Position += GreatCircle;
  }
  if (num_leg>max_leg) max_leg = num_leg;

  ConstantMotion = (CONSTANT_MOTION *)
    FreeDynamicAttributes->new_object(CONSTANT_MOTION_ID);

  ConstantMotion->SetEARTH();
  ConstantMotion->SetStartTime(OldTime);
  ConstantMotion->SetEndTime(1.0e20);
  ConstantMotion->SetPosition(OldLat, OldLon, OldAlt);
  Position += ConstantMotion;

#ifdef ENABLE_DDM
  InitDDM(parser);
#endif

  SEED = RANDOM->get_seed();
}

void 
S_RanMot::InitDDM(DATA_PARSER *parser){
#ifdef ENABLE_DDM
  char enum_val[16];
  float lo, hi;
  float sub_region_width;

  SpSimTime event_time(0.0);
  SpPublishSpace *publish_space = PublishSpace(event_time, "Earth");

  float expand_range = (publish_space->GetUpdateFreqHour() + 
                        publish_space->GetLookAheadHour()) * GetMaxSpeedKnots();

  const int total_obj = parser->GetInt("Nobjects");
  const float re = 6357; // unit : km
  const float sensor_range = NauticalMiles_to_Kilometers(GetMaxSensorRangeNmi() + expand_range);
  const float earth_area  = 4/3.0 * PI * (re * re);
  const float sensor_area = PI * (sensor_range * sensor_range); 
  float avg_num_detect = total_obj * (sensor_area / earth_area);

  avg_num_detect  /= 10.0;	// this is due to the "RandomEnumDim"
  int enable_filters = (avg_num_detect<1.0) ? 0 : 1; 
  sub_region_width = sensor_range / avg_num_detect;

  static int show_param = 1;

  if (show_param && get_NODE()==0){
    cout << "DDM param : sensor range " << sensor_range << " earth area " << earth_area 
         << " sensor area " << sensor_area << " avg detect " << avg_num_detect 
         << " subscribe region width " << sub_region_width << " total obj " << total_obj 
         << " sensor expand " << NauticalMiles_to_Kilometers(expand_range) 
         << " update freq " << publish_space->GetUpdateFreqHour()
         << " lookahead " << publish_space->GetLookAheadHour() << " enable filters " << enable_filters << endl;
    show_param = 0;
  }

  if (enable_filters){
    SpPublishRegion *publish_region = publish_space->CreateRegion("A");
    RANDOM->set_int_limits(1, 10);
    sprintf(enum_val, "%d", RANDOM->get_random_int());
    RANDOM->set_float_limits(0.0, sensor_range);
    lo = hi = RANDOM->get_random_float();
    publish_region->CreateDimension("RandomEnumDim", enum_val, (char *) 0);
    publish_region->CreateDimension("RandomNormalDim", lo, hi, (char *) 0);
  }
  publish_space->Update();

  SpSubscribeSpace *subscribe_space = SubscribeSpace(event_time, "Earth");
  if (enable_filters){
    SpSubscribeRegion *subscribe_region = subscribe_space->CreateRegion("A");
    RANDOM->set_int_limits(1, 10);
    sprintf(enum_val, "%d", RANDOM->get_random_int());
    RANDOM->set_float_limits(0.0, sensor_range-sub_region_width);
    lo = hi = RANDOM->get_random_float();
    hi += sub_region_width; if (hi>sensor_range) hi = sensor_range;
    subscribe_region->CreateDimension("RandomEnumDim", enum_val, (char *) 0);
    subscribe_region->CreateDimension("RandomNormalDim", lo, hi, (char *) 0);
  }
//subscribe_space->RegisterNewObjectProxyCallback(NewObjProxy);
//subscribe_space->RegisterRemoveObjectProxyCallback(RemoveObjProxy);
  subscribe_space->Update();
#endif
}

void S_RanMot::InitEvents() {
  SENSOR_COMPONENT *SensorComponent =
    (SENSOR_COMPONENT *)SensorComponents.GetFirstElement();
  while (SensorComponent) {
    SensorComponent->InitEvents();
    SensorComponent = (SENSOR_COMPONENT *) ++SensorComponents;
  }
}

void S_RanMot::terminate(double tend) {

/*
  if (GetRemoteObjectProxies()->get_length()>0)
    cout << "RanMot remote proxies " << get_GLOBAL_ID() << " " 
         << GetRemoteObjectProxies()->get_length() << endl;

  SpSubscribeSpace *subscribe_space = FindSubscribeSpace("Earth");
  SpPublishSpace *publish_space     = FindPublishSpace("Earth");
  int num_publish_grids   = publish_space->GetNumGrids();
  int num_subscribe_grids = subscribe_space->GetNumGrids();
  if ((num_publish_grids+num_subscribe_grids)>0)
    cout << "simobj pubsub grids " << get_GLOBAL_ID() << " "
         << num_publish_grids << " " << num_subscribe_grids << endl;
*/

  if (DebugObjectProxy) {

/*
** Print my object proxy
*/

    cout << endl;
    cout << "----------------------------------------------------------"
         << endl;
    cout << "Printing my proxy" << endl;

    ObjectProxy->PrintStream(cout);

/*
** Print each of the remote object proxies I have received
*/

    if (DebugRemoteObjectProxy) {
      cout << endl;
      cout << "Printing the remote proxies" << endl;

      F_PROXY_ITEM *ProxyItem = (F_PROXY_ITEM *)RemoteObjectProxies.get_top();

      for (int i=0; i<RemoteObjectProxies.get_length(); i++) {
        OBJECT_PROXY *RemoteProxy = ProxyItem->GetObjectProxy();
        RemoteProxy->PrintStream(cout);
        ProxyItem = (F_PROXY_ITEM *)ProxyItem->get_link();
      }

      cout << "----------------------------------------------------------"
           << endl;

    }

  }

}

void S_RanMot::NamedQuery(double t, NAME_VALUE_LIST &NameValueList,
                        char *queryName){

  if (!strcmp(queryName, "Vmax")) {
    NameValueList.SetNameValue("Time", t);
    NameValueList.SetNameValue("Vmax", GetVmax());
  }
  else if (!strcmp(queryName, "Position")) {
    NameValueList.SetNameValue("Time", t);

    double x[3], v[3], a[3];

    Position(t, EARTH, x, v, a);

//  Clamp altitude to 0:

    if (x[2] < 0.01) x[2] = 0.0;

    NameValueList.SetNameValue("x0", x[0]);  // Lat (rad)
    NameValueList.SetNameValue("x1", x[1]);  // Long (rad)
    NameValueList.SetNameValue("x2", x[2]);  // Alt (km?)
    NameValueList.SetNameValue("v0", v[0]);  // Vlat (km/s?)
    NameValueList.SetNameValue("v1", v[1]);  // Vlong (km/s?)
    NameValueList.SetNameValue("v2", v[2]);  // Valt (km/s?)
    NameValueList.SetNameValue("a0", a[0]);  // Alat (km/s^2?)
    NameValueList.SetNameValue("a1", a[1]);  // Along (km/s^2?)
    NameValueList.SetNameValue("a2", a[2]);  // Aalt (km/s^2?)
  }
}

///////////////////////////////////////////////////////////////////////////////////////////

double 
S_RanMot::GetMaxSpeedKnots(void){
  NAME_VALUE_LIST name_value_list;
  double vmax;
  int found;
  char *name = "Vmax";

  NamedQuery(0.0, name_value_list, name);
  vmax = name_value_list.GetValue(name, found);

  if (!found) vmax = 0.0;
  vmax = Kilometers_to_NauticalMiles(vmax * 3600.0);

  return (vmax);
}

#include "SpGeoFunctions.H"
#include "LatLon.H"

void 
S_RanMot::GetPositionRadKm(double time, double &latRad, double &lonRad, double &altKm){
  NAME_VALUE_LIST name_value_list;
  char *name;
  int found;
  double v;

  name = "Position"; NamedQuery(time, name_value_list, name);
  name = "x0"; v = name_value_list.GetValue(name, found); latRad = (found) ? v : 0.0;
  name = "x1"; v = name_value_list.GetValue(name, found); lonRad = (found) ? v : 0.0;
  name = "x2"; v = name_value_list.GetValue(name, found); altKm  = (found) ? v : 0.0;

/////////////////////////////////////////////////////////////

  const double epsilon = 1E-5;
  if (latRad>north_pole_rad_lat) latRad = latRad - north_pole_rad_lat;
  if (fabs(latRad-north_pole_rad_lat)<epsilon) latRad = north_pole_rad_lat-epsilon;
  if (latRad<south_pole_rad_lat) latRad = south_pole_rad_lat - (latRad - south_pole_rad_lat);
  if (fabs(latRad-south_pole_rad_lat)<epsilon) latRad = south_pole_rad_lat+epsilon;
  lonRad = DEG_to_RAD(NormLon(RAD_to_DEG(lonRad)));

/////////////////////////////////////////////////////////////
}

double 
S_RanMot::GetLookAheadHour(void){
  return (GetMinRescheduleTimeHour()/4.0);
}

double 
S_RanMot::GetMinExpansionNmi(void){
  return (10.0);
}

double 
S_RanMot::GetMinRescheduleTimeHour(void){
  return (1.0/6.0);
}
