// S_Ship.C source file

#define ENABLE_DDM

#include "S_Ship.H"
#include "RadarComponent.H"
#include "ScriptedMotionComponent.H"
#include "F_ProxyItem.H"
#include "ScheduleEvent.H"
#include "MovingEntityProxy.H"
#include "RB.H"

extern DATA_PARSER *SensorParser;
extern int DebugObjectProxy;
extern int DebugRemoteObjectProxy;

/******************************************************************************
* S_Ship constructor                                                          *
******************************************************************************/
S_Ship::S_Ship() : S_MOVING_ENTITY ("Ship") {
  cout << "Creating a Ship" << endl;
}

/******************************************************************************
* terminate: executed at the end of the simulation                            *
******************************************************************************/
void S_Ship::terminate(double /*tend*/) {

  if (DebugObjectProxy) {

/*
** Print my object proxy
*/

    cout << endl;
    cout << "----------------------------------------------------------"
	 << endl;
    cout << "Printing my proxy" << endl;

    ObjectProxy->PrintStream(cout);

/*
** Print each of the remote object proxies I have received
*/

    if (DebugRemoteObjectProxy) {
      cout << endl;
      cout << "Printing the remote proxies" << endl;

      F_PROXY_ITEM *ProxyItem = (F_PROXY_ITEM *)RemoteObjectProxies.get_top();

      for (int i=0; i<RemoteObjectProxies.get_length(); i++) {
	OBJECT_PROXY *RemoteProxy = ProxyItem->GetObjectProxy();
	RemoteProxy->PrintStream(cout);
	ProxyItem = (F_PROXY_ITEM *)ProxyItem->get_link();
      }

      cout << "----------------------------------------------------------"
	   << endl;

    }

  }

}

// <END:   Define public virtual methods>

#include "SpSubscribeSpace.H"
#include "SpPublishSpace.H"

static void NewObjProxy(SpSubscribeSpace *space, F_SpProxyItem *item){
  cout << space->GetSimObj()->get_NAME() 
       << " subscribe to space "
       << space->GetName() << " rcv'd new proxy "
       << item->GetName()
       << " @ " << CurrentEvent->get_time_tag() << endl;
}

static void RemoveObjProxy(SpSubscribeSpace *space, int globalId){
  cout << space->GetSimObj()->get_NAME() 
       << " subscribe to space "
       << space->GetName() << " rcv'd rem proxy " << globalId
       << " @ " << CurrentEvent->get_time_tag() << endl;
}

// <BEGIN: Define public methods>
/******************************************************************************
* init: Initializes the object using the parser                               *
******************************************************************************/
void S_Ship::init(DATA_PARSER *Parser) {
  int i;
  char *DefName;
  char *DefValue;
  int Ndefines;
  int Nsensors = 0;
  char *MotionType;
  void *Mark;

  Mark = Parser->GetMark();

/*
** Create my sensor components
*/

  Ndefines = Parser->GetNdefines();
  Parser->GetFirstDefine(DefName, DefValue);

  for (i=0; i<Ndefines; i++) {
    cout << "Define " << DefName << " " << DefValue << endl;

/*
** Load up SENSOR components
*/

    if (!strcmp("SENSOR", DefName)) {

      while (1) {

// AirSearchRadar1
	if (!strcmp("AirSearchRadar1", DefValue)) {
//	  SensorParser->GoTo(DefValue, NULL);
	  RADAR_COMPONENT *RadarComponent = new RADAR_COMPONENT
	    (this, SensorParser, DefValue, Nsensors);
	  SensorComponents.Insert(RadarComponent, Nsensors);
	  AddComponent(RadarComponent);
//	  RadarComponent->InitEvents();
	  break;
	}

// SeaSurfaceSearchRadar1
	if (!strcmp("SeaSurfaceSearchRadar1", DefValue)) {
//	  SensorParser->GoTo(DefValue, NULL);
	  RADAR_COMPONENT *RadarComponent = new RADAR_COMPONENT
	    (this, SensorParser, DefValue, Nsensors);
	  SensorComponents.Insert(RadarComponent, Nsensors);
	  AddComponent(RadarComponent);
//	  RadarComponent->InitEvents();
	  break;
	}

// SeaSufaceSearchRadar1
	if (!strcmp("SeaSurfaceSearchRadar2", DefValue)) {
//	  SensorParser->GoTo(DefValue, NULL);
	  RADAR_COMPONENT *RadarComponent = new RADAR_COMPONENT
	    (this, SensorParser, DefValue, Nsensors);
	  SensorComponents.Insert(RadarComponent, Nsensors);
	  AddComponent(RadarComponent);
//	  RadarComponent->InitEvents();
	  break;
	}

	cout << "Error, unknown SENSOR type " << DefValue << endl;
	exit(1);

      }

      Nsensors++;
    }

    Parser->GetNextDefine(DefName, DefValue);
  }

/*
** Create my motion components
*/

  cout << "Number of MOTIONS " << Parser->GetNumReferences(Mark, "MOTION")
       << endl;
  Parser->GoToFirstReference(Mark, "MOTION");

  while (1) {
    MotionType = Parser->GetReferenceName();
    if (MotionType == NULL) break;

// Scripted motion
    if (!strcmp("Scripted", MotionType)) {
      SCRIPTED_MOTION_COMPONENT *ScriptedMotionComponent =
	new SCRIPTED_MOTION_COMPONENT(&Position, Parser);
      MotionComponents.Insert(ScriptedMotionComponent, "ScriptedMotion");
      AddComponent(ScriptedMotionComponent);

      Vmax = ScriptedMotionComponent->GetVmax();

      break;
    }

// Other kinds of motion

    break;

  }

#ifdef ENABLE_DDM
  SpSimTime event_time(0.0);
  SpPublishSpace *publish_space;
  publish_space = PublishSpace(event_time, "Earth");
  publish_space->Update();

  SpSubscribeSpace *subscribe_space;
  subscribe_space = SubscribeSpace(event_time, "Earth");
  subscribe_space->RegisterNewObjectProxyCallback(NewObjProxy);
  subscribe_space->RegisterRemoveObjectProxyCallback(RemoveObjProxy);
  subscribe_space->Update();
#endif
}

// <END:   Define public methods>

/************************************************************************
* Query : Query virtual function					*
************************************************************************/
void S_Ship::Query(double t, NAME_VALUE_LIST &NameValueList) {
  int i;
  char s[120];

  if (NAME) NameValueList.SetNameOfList(NAME);

  for (i=0; i<100; i++) {
    sprintf(s, "Query Test %d", i);
    NameValueList.SetNameValue(s, i);
  }

}

/************************************************************************
* NamedQuery : Query virtual function					*
************************************************************************/
void S_Ship::NamedQuery(double t, NAME_VALUE_LIST &NameValueList,
		       char *queryName){
 if (!strcmp(queryName, "Vmax")) {
    NameValueList.SetNameValue("Time", t);
    NameValueList.SetNameValue("Vmax", GetVmax());
  }


  else if (!strcmp(queryName, "Position")) {
    NameValueList.SetNameValue("Time", t);
	
    double x[3], v[3], a[3];

    Position(t, EARTH, x, v, a);

//  Clamp altitude to 0:

    if (x[2] < 0.01) x[2] = 0.0;

    NameValueList.SetNameValue("x0", x[0]);  // Lat (rad)
    NameValueList.SetNameValue("x1", x[1]);  // Long (rad)
    NameValueList.SetNameValue("x2", x[2]);  // Alt (km?)
    NameValueList.SetNameValue("v0", v[0]);  // Vlat (km/s?)
    NameValueList.SetNameValue("v1", v[1]);  // Vlong (km/s?)
    NameValueList.SetNameValue("v2", v[2]);  // Valt (km/s?)
    NameValueList.SetNameValue("a0", a[0]);  // Alat (km/s^2?)
    NameValueList.SetNameValue("a1", a[1]);  // Along (km/s^2?)
    NameValueList.SetNameValue("a2", a[2]);  // Aalt (km/s^2?)
  }


  else if (!strcmp(queryName, "Alive")) {
    NameValueList.SetNameValue("Time", t);
    NameValueList.SetNameValue("Alive", GetAlive());
  }


  else if (!strcmp(queryName, "PlatformComponents")) {
    NameValueList.SetNameValue("Time", t);
    NameValueList.SetNameValue("N", PlatformComponents.GetNumElements());
    
  }


  else if (!strcmp(queryName, "MotionComponents")) {
    NameValueList.SetNameValue("Time", t);
    NameValueList.SetNameValue("N", MotionComponents.GetNumElements());

    SCRIPTED_MOTION_COMPONENT *scripted_motion_component = 
      (SCRIPTED_MOTION_COMPONENT*) MotionComponents.GetFirstElement();

    while (scripted_motion_component) {
      scripted_motion_component = (SCRIPTED_MOTION_COMPONENT*) 
	++MotionComponents;
    }
  }


  else if (!strcmp(queryName, "WeaponComponents")) {
    NameValueList.SetNameValue("Time", t);
    NameValueList.SetNameValue("N", WeaponComponents.GetNumElements());
  }


  else if (!strcmp(queryName, "SensorComponents")) {
    NameValueList.SetNameValue("Time", t);
    NameValueList.SetNameValue("N", SensorComponents.GetNumElements());

    RADAR_COMPONENT *radar_component = 
      (RADAR_COMPONENT*) SensorComponents.GetFirstElement();

    while (radar_component) {
      NameValueList.SetNameValue("Range", radar_component->GetRange());
      NameValueList.SetNameValue("ScanTime", radar_component->GetRange());
      NameValueList.SetNameValue(radar_component->GetRadarType(), 0);
      NameValueList.SetNameValue("SensorId", radar_component->GetSensorId());
      radar_component = (RADAR_COMPONENT*) ++SensorComponents;
    }
  }


  else if (!strcmp(queryName, "TrackerCorrelationComponents")) {
    NameValueList.SetNameValue("Time", t);
    NameValueList.SetNameValue("N", 
			       TrackCorrelationComponents.GetNumElements());
  }


  else if (!strcmp(queryName, "CommandControlComponents")) {
    NameValueList.SetNameValue("Time", t);
    NameValueList.SetNameValue("N", 
			       CommandControlComponents.GetNumElements());
  }


  else if (!strcmp(queryName, "CommunicationComponents")) {
    NameValueList.SetNameValue("Time", t);
    NameValueList.SetNameValue("N", CommunicationComponents.GetNumElements());
  }
}

void S_Ship::InitEvents() {
  SENSOR_COMPONENT *SensorComponent =
    (SENSOR_COMPONENT *)SensorComponents.GetFirstElement();
  while (SensorComponent) {
    SensorComponent->InitEvents();
    SensorComponent = (SENSOR_COMPONENT *) ++SensorComponents;
  }

#ifdef COMMENTOUT
  SpObjectHandle objectHandle;
  objectHandle.NodeId = NODE;
  objectHandle.ObjManId = OBJECT_TYPE;
  objectHandle.SimObjId = LOCAL_ID;

  Schedule_ToggleMotion(0.001,
			objectHandle,
			100.0, 0);
#endif

}

void S_Ship::ToggleMotion(double time, int iteration) {
#ifdef COMMENTOUT
  RB_cout << "ToggleMotion event at time "
	  << CurrentEvent->get_time_tag()
	  << endl;

  RB_cout << "time = " << time << ", iteration = " << iteration << endl;
#endif

  SpObjectHandle objectHandle;
  objectHandle.NodeId = NODE;
  objectHandle.ObjManId = OBJECT_TYPE;
  objectHandle.SimObjId = LOCAL_ID;

//#ifdef COMMENTOUT
  if (iteration % 2) {
    ProcessEventHandlers("ChangeScriptedMotion",
			 CurrentEvent->get_time_tag(),
			 "TOKYO",
			 strlen("TOKYO") + 1,
			 -1);
  }else{
    ProcessEventHandlers("ChangeScriptedMotion",
			 CurrentEvent->get_time_tag(),
			 "HONOLULU",
			 strlen("HONOLULU") + 1,
			 -1);
  }
//#endif

  Schedule_ToggleMotion(CurrentEvent->get_time_tag() + time,
			objectHandle,
			time, iteration+1);
}

double 
S_Ship::GetMaxSpeedKnots(void){
  NAME_VALUE_LIST name_value_list;
  double vmax;
  int found;
  char *name = "Vmax";

  NamedQuery(0.0, name_value_list, name);
  vmax = name_value_list.GetValue(name, found);

  if (!found) vmax = 0.0;
  vmax = Kilometers_to_NauticalMiles(vmax * 3600.0);

  return (vmax);
}

#include "SpGeoFunctions.H"

void 
S_Ship::GetPositionRadKm(double time, double &latRad, double &lonRad, double &altKm){
  NAME_VALUE_LIST name_value_list;
  char *name;
  int found;
  double v;

  name = "Position"; NamedQuery(time, name_value_list, name);
  name = "x0"; v = name_value_list.GetValue(name, found); latRad = (found) ? v : 0.0;
  name = "x1"; v = name_value_list.GetValue(name, found); lonRad = (found) ? v : 0.0;
  name = "x2"; v = name_value_list.GetValue(name, found); altKm  = (found) ? v : 0.0;
}

double 
S_Ship::GetLookAheadHour(void){
  return (GetMinRescheduleTimeHour()/4.0);
//  return (1/3600.0);
}

double 
S_Ship::GetMinExpansionNmi(void){
  return (1.0);
}

double 
S_Ship::GetMinRescheduleTimeHour(void){
  return (1.0/6.0);
}
