// ScriptedMotionComponient.C file

#include "ScriptedMotionComponent.H"
#include "GreatCircle.H"
#include "ConstMotion.H"
#include "Convert.H"

#include "FreeDynamicAttributes.H"
extern FREE_DYNAMIC_ATTRIBUTES *FreeDynamicAttributes;
extern DATA_PARSER *LocationParser;

SCRIPTED_MOTION_COMPONENT::SCRIPTED_MOTION_COMPONENT
( DYNAMIC_POSITION_ATTRIBUTE *position,
  DATA_PARSER *parser) :
  MOTION_COMPONENT(position) {

  int i;
  int NumSegments;
  char *MotionType;
  char *Location1;
  char *Location2;
  double StartTime;
  GREAT_CIRCLE *GreatCircle;
  CONSTANT_MOTION *ConstantMotion;
  void *Mark;
  double Lat1, Lon1, Alt1;
  double Lat2, Lon2, Alt2;
  double Speed;
  int UnitDistance;
  int UnitTime;

/*
** Get units for distance
*/

  UnitDistance = GetUnitDistance(parser->GetString("UnitDistance"));
  UnitTime = GetUnitTime(parser->GetString("UnitTime"));

  Vmax = KilometersPerSecond(parser->GetFloat("Vmax"),
			     UnitDistance, UnitTime);

/*
** Create the event handler to change my motion
*/

  ChangeScriptedMotion = new H_CHANGE_SCRIPTED_MOTION(Position);

  AddEventHandler("ChangeScriptedMotion", ChangeScriptedMotion);

  cout << "Adding ChangeScriptedMotion event handler" << endl;

/*
** Extract the start time from the parser
*/

  StartTime = parser->GetFloat("StartTime");
  cout << "StartTime " << StartTime << endl;

  Mark = parser->GetMark();
  NumSegments = parser->GetNumReferences(Mark, "SEGMENT");
  cout << "Number of SEGMENTS " << NumSegments << endl;

  if (NumSegments) {

    parser->GoToFirstReference(Mark, "SEGMENT");
    for (i=0; i<NumSegments; i++) {

      MotionType = parser->GetString("MotionType");
      cout << "MotionType " << MotionType << endl;

      while (1) {

	if (!strcmp("GreatCircle",MotionType)) {

/*
** Get the two locations
*/

	  Location1 = parser->GetString("Location1");
	  Location2 = parser->GetString("Location2");
	  Speed = KilometersPerSecond(parser->GetFloat("Speed"),
					UnitDistance, UnitTime);

	  cout << "Locations " << Location1 << ", " << Location2 << endl;

/*
** Compute the first location
*/

	  LocationParser->GoTo(Location1, NULL);

	  Lat1 = (LocationParser->GetFloat("LatDeg")
	        + LocationParser->GetFloat("LatMin") / 60.0
	        + LocationParser->GetFloat("LatSec") / 3600.0)
	        * (PI/180);
	  if (!LocationParser->GetLogical("North")) Lat1 = -Lat1;

	  Lon1 = (LocationParser->GetFloat("LonDeg")
	        + LocationParser->GetFloat("LonMin") / 60.0
	        + LocationParser->GetFloat("LonSec") / 3600.0)
	        * (PI/180);
	  if (!LocationParser->GetLogical("East")) Lon1 = -Lon1;

	  Alt1 = LocationParser->GetFloat("Altitude");;

/*
** Get the initial location if this is the first position
*/

	  if (i==0) {
	    ConstantMotion = (CONSTANT_MOTION *)
	      FreeDynamicAttributes->new_object(CONSTANT_MOTION_ID);
	    ConstantMotion->SetEARTH();
	    ConstantMotion->SetStartTime(-1.0e20);
	    ConstantMotion->SetEndTime(StartTime);
	    ConstantMotion->SetPosition(Lat1, Lon1, Alt1);
	    *Position += ConstantMotion;
	  }

/*
** Compute the second location
*/

	  LocationParser->GoTo(Location2, NULL);

	  Lat2 = (LocationParser->GetFloat("LatDeg")
	        + LocationParser->GetFloat("LatMin") / 60.0
	        + LocationParser->GetFloat("LatSec") / 3600.0)
	        * (PI/180);
	  if (!LocationParser->GetLogical("North")) Lat2 = -Lat2;

	  Lon2 = (LocationParser->GetFloat("LonDeg")
	        + LocationParser->GetFloat("LonMin") / 60.0
	        + LocationParser->GetFloat("LonSec") / 3600.0)
	        * (PI/180);
	  if (!LocationParser->GetLogical("East")) Lon2 = -Lon2;

	  Alt2 = LocationParser->GetFloat("Altitude");;

/*
** Construct the great circle motion
*/

	  GreatCircle = (GREAT_CIRCLE *)
	    FreeDynamicAttributes->new_object(GREAT_CIRCLE_ID);

	  GreatCircle->init_Vconstant(Lat1, Lon1, Alt1, StartTime,
				      Lat2, Lon2, Alt2, Speed);

/*
** Check that things make sense
*/
          int Status;
	  if (Status = GreatCircle->CheckMotion()) {
	    cout << "Bad CheckMotion" << Status << endl;
	  }

	  double X1[3], X2[3], Earth[3], R[3], Distance;
	  Earth[0] = Lat1; Earth[1] = Lon1; Earth[2] = Alt1;
	  EARTH_to_ECI(StartTime, Earth, X1);

	  Earth[0] = Lat2; Earth[1] = Lon2; Earth[2] = Alt2;
	  EARTH_to_ECI(StartTime, Earth, X2);
	  Difference(X1, X2, R);
	  Distance = Magnitude(R);

	  double Vel = Distance
	    / (GreatCircle->GetEndTime() - GreatCircle->GetStartTime());

	  cout << "Distance between "
	       << Location1 << " and "
	       << Location2 << " = "
	       << Distance
	       << " (velocity = " << Vel << ")" << endl;

/*
** Add the great circle to the dynamic position attribute
*/

	  *Position += GreatCircle;
	  StartTime = GreatCircle->GetEndTime();

	  if (i==NumSegments-1) {
	    ConstantMotion = (CONSTANT_MOTION *)
	      FreeDynamicAttributes->new_object(CONSTANT_MOTION_ID);
	    ConstantMotion->SetEARTH();
	    ConstantMotion->SetStartTime(StartTime);
	    ConstantMotion->SetEndTime(1.0e20);
	    ConstantMotion->SetPosition(Lat2, Lon2, Alt2);
	    *Position += ConstantMotion;
	  }


	  break;
	}

      }

      parser->GoToNextReference(Mark, "SEGMENT");

    }


  }

}
