// SensorComponent.H header file

#ifndef SENSOR_COMP_OBJ
#define SENSOR_COMP_OBJ

#include "Component.H"
#include "Semaphore.H"
#include "SensorListElement.H"
#include "S_Entity.H"
#include "RB_BinaryTree.H"

class SENSOR_COMPONENT : public COMPONENT {

  private:

  protected:

    LIST_ATTRIBUTE *SensorList;
    SENSOR_LIST_ELEMENT SensorListElement;
    RB_BINARY_TREE DetectableProxies;
    SpCounterSem NumDetectableProxies;
    RB_BINARY_TREE RemoteProxySemaphores;

    RB_int NumEnterProxies, NumExitProxies;
 
  public:

    SENSOR_COMPONENT(S_ENTITY *entity) {
      SensorList = entity->GetSensorList();
      *SensorList += &SensorListElement;
      NumEnterProxies = NumExitProxies = 0;
    }

    virtual void InitEvents(){;}
    virtual double GetRange(void) const = 0;

/*
** Detectable proxy methods
*/

    SpCounterSem *GetNumDetectableProxies() {return &NumDetectableProxies;}
    RB_BINARY_TREE *GetDetectableProxies() {return &DetectableProxies;}

    RB_int *GetNumEnterProxies(void)  { return (&NumEnterProxies); }
    RB_int *GetNumExitProxies(void)   { return (&NumExitProxies); }
    int GetNumRemoteProxySemaphores(void) { return (RemoteProxySemaphores.GetNumElements()); }

/*
** Object Id methods
*/

    void AddObjectId(OBJECT_ID *objectId) {
      SensorListElement.AddObjectId(objectId);
    }

    void RemoveObjectId(OBJECT_ID *objectId) {
      SensorListElement.RemoveObjectId(objectId);
    }

/*
** Remote proxy semaphore methods
*/

    LogicalSemaphore *FindRemoteProxySemaphore(int globalId) {
      LogicalSemaphore *remoteProxySemaphore;
      remoteProxySemaphore = (LogicalSemaphore *)
	RemoteProxySemaphores.Find(globalId);
      return remoteProxySemaphore;
    }

    void InsertRemoteProxySemaphore(LogicalSemaphore *remoteProxySemaphore,
				    int globalId) {
      RemoteProxySemaphores.Insert(remoteProxySemaphore, globalId);
    }

    LogicalSemaphore *RemoveRemoteProxySemaphore(int globalId) {
      return ((LogicalSemaphore *)RemoteProxySemaphores.Remove(globalId));
    }

    void SetRemoteProxySemaphore(int globalId=-1) {
      LogicalSemaphore *remoteProxySemaphore;
      if (globalId < 0) {

	remoteProxySemaphore = (LogicalSemaphore *)
	  RemoteProxySemaphores.GetFirstElement();

	while (remoteProxySemaphore) {
	  *remoteProxySemaphore = 1;
	  remoteProxySemaphore = (LogicalSemaphore *) ++RemoteProxySemaphores;
	}

      }else{
	remoteProxySemaphore = FindRemoteProxySemaphore(globalId);
	if (remoteProxySemaphore) {
	  *remoteProxySemaphore = 1;
	}else{
	  RB_cout << "Error, could not find remote proxy semaphore "
		  << globalId << endl;
	}
      }

    }

};

#endif
