// app1man.C method file

#include "SpStdio.H"
#include "app1man.H"
#include "nodeman.H"
#include "start_app1.H"
#include "SpBasetype.H"

/************************************************************************
* C_APP1MAN : construct a app1man object				*
************************************************************************/
C_APP1MAN::C_APP1MAN() {
  C_NODEMAN *nodeman;
  C_BASETYPE *parameters;
  int i;

//...... get application1 information out of the applications parser

  parameters = applications_parser->get_basetype("app1");

  fanout = parameters->get_int("fanout");
  bytes = parameters->get_int("bytes");
  test_flag = parameters->get_logical("test_flag");
  compute_time = parameters->get_float("compute_time");

//...... get information from the node manager and create app1 objects

  nodeman = (C_NODEMAN *)get_manager(HYPE_NODE);
  N_TOT = nodeman->get_N_TOT();
  N_LOC = nodeman->get_N_LOC();
  app1obj = new C_APP1OBJ[N_LOC];

//...... initialize the app1 objects

  HYPER_n_nodes = nodeman->get_HYPER_n_nodes();

  for (i=0; i<N_LOC; i++) {
    app1obj[i].set_compute_time(compute_time);
    app1obj[i].set_fanout(fanout);
    app1obj[i].set_bytes(bytes);
    app1obj[i].set_HYPER_n_nodes(HYPER_n_nodes);
    app1obj[i].set_HYPER_node(nodeman->get_HYPER_node(i));
    app1obj[i].set_SEED(app1obj[i].get_HYPER_node() + 56789);
    app1obj[i].init_time();
  }

//...... tell SPEEDES about the node objects

  TOTAL_OBJECTS += N_TOT;
  reserve_n_objects(N_LOC);
  for (i=0; i<N_LOC; i++) define_object(&app1obj[i]);
  set_interact();

  cout << "APP1MAN object created with " << N_TOT << " objects ("
       << N_LOC << " loc)\n";

}
/************************************************************************
* init_events : initialize events 				        *
************************************************************************/
void C_APP1MAN::init_events() {
  int i;
  double tim;

  printf("APPLICATION 1 events initializing ...\n");

  if (test_flag) {
    if (NODE == 0) {
      tim = app1obj[0].get_time();
      schedule (tim, START_APP1, APP1, 0, 0);
    }
  }else{
    for (i=0; i<N_LOC; i++) {
      tim = app1obj[i].get_time();
      schedule (tim, START_APP1, APP1, i, NODE);
    }
  }

}

void C_APP1MAN::terminate(double /*tend*/) {
  delete [] app1obj;
}
