// nodeman.C method file

#include "SpStdio.H"
#include "SpScheduleEvent.H"
#include "nodeman.H"

#include "RB_SpFrameworkFuncs.H"
#include "SpBasetype.H"

EVENT_TYPE (TEST_CANCEL)

/************************************************************************
* C_NODEMAN : construct a nodeman object				*
************************************************************************/
C_NODEMAN::C_NODEMAN() {
  C_BASETYPE *parameters;
  C_BASETYPE *test_parameters;
  C_BASETYPE *priorities;
  int i;
  int offset;

//...... create the free list object

  freelist = new C_FREELIST();
  RB_DEFINE_FREE(freelist);

//...... get information out of the hypercube parser

  parameters = hypercube_parser->get_basetype("parameters");

  HYPER_n_nodes = parameters->get_int("n_nodes");
  HYPER_card_deal = parameters->get_logical("card_deal");
  HYPER_bandwidth = parameters->get_float("bandwidth");
  HYPER_send_latency = parameters->get_float("send_latency");
  HYPER_receive_latency = parameters->get_float("receive_latency");
  spin = parameters->get_int("spin");

  priorities = hypercube_parser->get_basetype("priorities");
  n_priorities = priorities->get_enum("N_PRIORITIES");

  test_parameters = hypercube_parser->get_basetype("test_parameters");
  test_cancel = test_parameters->get_logical("test_cancel");

//...... set up some SPEEDES variables

  N_TOT = HYPER_n_nodes;
  N_LOC = deal_me(N_TOT, offset);
  nodeobj = new C_NODEOBJ[N_LOC];
  node0 = (NODE - offset + N_NODES) % N_NODES;
  if (node0 != (TOTAL_OBJECTS%N_NODES)) {
    fprintf(stderr,"Error, bad node0\n");
  }

//...... smart decomposition

  if (!HYPER_card_deal) {
    if (HYPER_n_nodes % N_NODES) {
      printf("Warning, (nodeman) card dealing\n");
      HYPER_card_deal = 1;
    }else{
      mask1 = HYPER_n_nodes / N_NODES;
      mask2 = NODE * mask1;
    }
  }

  for (i=0; i<N_LOC; i++) {

    if (HYPER_card_deal) {
      nodeobj[i].set_HYPER_node((N_NODES*i)+offset);
    }else{
      nodeobj[i].set_HYPER_node(mask2+i);
    }

    nodeobj[i].set_SEED(nodeobj[i].get_HYPER_node() + 12345);

    nodeobj[i].set_HYPER_n_nodes(HYPER_n_nodes);
    nodeobj[i].set_HYPER_bandwidth(HYPER_bandwidth);
    nodeobj[i].set_HYPER_send_latency(HYPER_send_latency);
    nodeobj[i].set_HYPER_receive_latency(HYPER_receive_latency);
    nodeobj[i].set_time(0.0);
    nodeobj[i].set_spin(spin);
    nodeobj[i].create_mqueues(n_priorities);

  }

//...... test if routing is ok

/*
  int j;

  for (i=0; i<N_LOC; i++) {
    for (j=0; j<HYPER_n_nodes; j++) {
      if (i!= j) {
	nodeobj[i].get_route(j);
      }
    }
  }
*/

//...... tell SPEEDES about the node objects

  TOTAL_OBJECTS += N_TOT;
  reserve_n_objects(N_LOC);
  for (i=0; i<N_LOC; i++) define_object(&nodeobj[i]);
  set_interact();

  cout << "NODEMAN object created with " << N_TOT
       << " objects (" << N_LOC << " loc)\n";

}
/************************************************************************
* init_events : initialize events 				        *
************************************************************************/
void C_NODEMAN::init_events() {
  int i;

  printf("NODE events initializing ...\n");

  if (test_cancel) {
    for (i=0; i<N_LOC; i++) {
      schedule (0.0, TEST_CANCEL, HYPE_NODE, i, NODE);
    }
  }

}

/************************************************************************
* terminate : termination method 				        *
************************************************************************/
void C_NODEMAN::terminate(double) {

  if (NODE == 0) {
    cout << endl << "Printing freelist memory" << endl;
    freelist->print();
  }

  delete [] nodeobj;
  delete freelist;
}

/************************************************************************
* get_nodeid : get the local node and id given a global node	        *
************************************************************************/
void C_NODEMAN::get_nodeid(int global_node, int &local_node, int &local_id) {

  if (HYPER_card_deal) {
    local_node = (global_node + node0) % N_NODES;
    local_id = global_node / N_NODES;
  }else{
    local_node = global_node / mask1;
    local_id = global_node % mask1;
  }
}

