// nodeobj.C method file

#include "SpStdio.H"
#include "nodeobj.H"

int C_NODEOBJ::HYPER_n_nodes = 1;
int C_NODEOBJ::HYPER_dimension = 0;
int C_NODEOBJ::n_priorities = 1;
int C_NODEOBJ::spin = 0;

/************************************************************************
* C_NODEOBJ : construct a nodeobj object				*
************************************************************************/
C_NODEOBJ::C_NODEOBJ() {

  busy = 0;
  cancel_handle = NULL;

}

/************************************************************************
* calculate_HYPER_dimension : calculate the dimension of the cube	*
************************************************************************/
void C_NODEOBJ::calculate_HYPER_dimension() {
  int i;

  HYPER_dimension = 0;
  for (i=1; i!=HYPER_n_nodes; i+=i) HYPER_dimension++;

}

/************************************************************************
* create_mqueues : create the message queues				*
************************************************************************/
void C_NODEOBJ::create_mqueues(int p) {

  n_priorities = p;
  message_queue = new C_XQUEUE[p];

}

/************************************************************************
* get_n_messages : return the number of pending messages		*
************************************************************************/
int C_NODEOBJ::get_n_messages() {
  int i;
  int n;

  n = 0;
  for (i=0; i<n_priorities; i++) n+=message_queue[i].get_length();
  return n;

}

/************************************************************************
* get_next_message_queue : get the highest pri. non-empty message queue	*
************************************************************************/
C_XQUEUE *C_NODEOBJ::get_next_message_queue() {
  int i;

  for (i=0; i<n_priorities; i++) {
    if (message_queue[i].get_length()) return &message_queue[i];
  }

  return NULL;

}

/************************************************************************
* exchange_cancel_handle : exchange the cancel handle			*
************************************************************************/
void C_NODEOBJ::exchange_cancel_handle(void *&ch) {
  // 	I have to do some tricky casting
  // to avoid the compiler warning which says
  // that a temporary value will be passed
  // which means that the exchange function
  // will not be able to modify the member
  // variable.
  void **v = (void **) (&cancel_handle);

  EXCHANGE(*v, ch);
}

/************************************************************************
* get_route : figure out what node to send the message			*
************************************************************************/
int C_NODEOBJ::get_route(int destination) {
  int channels;
  int cmask;

//...... return -1 if this is already the node

  if (HYPER_node == destination) return -1;

//...... compute the next node

  channels = HYPER_node ^ destination;
  for (cmask=1; cmask<=HYPER_n_nodes; cmask+=cmask) {
    if (cmask & channels) {
      return (HYPER_node ^ cmask);
    }
  }

  fprintf(stderr,"Error, (nodeobj) could not find a route\n");
  return -1;

}

/************************************************************************
* terminate : this routine is called by SPEEDES at the end of the run	*
************************************************************************/
void C_NODEOBJ::terminate(double) {

//  fprintf(stderr,"Terminating node %d at time %f\n", HYPER_node, tend);
  delete [] message_queue;
}
