// read_app1.C method file

#include "SpStdio.H"

#include "read_app1.H"
#include "app1obj.H"
#include "msg.H"
#include "SpScheduleEvent.H"
#include "send_mess.H"

#include "RB_SpContainerFuncs.H"

EVENT_TYPE (SEND);
EVENT_TYPE (START_APP1);
OBJ_MGR_TYPE (HYPE_NODE);
OBJ_MGR_TYPE (APP1);

/************************************************************************
* C_READ_APP1 : construct a read_app1 object				*
************************************************************************/
C_READ_APP1::C_READ_APP1() {
  set_lazy();
}

/************************************************************************
* exchange : exchange state variables                                   *
************************************************************************/
void C_READ_APP1::exchange() {
  C_APP1OBJ *app1obj;

  app1obj = (C_APP1OBJ *)SIMOBJ;
  app1obj->exchange_time(time);
  app1obj->exchange_n_received(n_received);

}

/************************************************************************
* lazy : lazy cancellation routine                                      *
************************************************************************/
int C_READ_APP1::lazy() {
  C_APP1OBJ *app1obj;
  C_XQUEUE *message_queue;
  C_MSG *msg;
  double t;

  app1obj = (C_APP1OBJ *)SIMOBJ;

  t = app1obj->get_time();
  if (TIME_TAG > t) t = TIME_TAG;
  if (t != lazy_time) return 0;

  message_queue = app1obj->get_message_queue();
  msg = (C_MSG *)message_queue->get_top();
  if (msg != lazy_msg) return 0;

  if (lazy_n_received != app1obj->get_n_received()) return 0;

  return 1;

}

/************************************************************************
* temp_process : temporarily process event				*
************************************************************************/
void C_READ_APP1::temp_process() {
  C_APP1OBJ *app1obj;
  C_XQUEUE *message_queue;
  C_MSG *msg;
  SEND_MESS *send_mess;

  app1obj = (C_APP1OBJ *)SIMOBJ;

//...... compute the application time

  time = app1obj->get_time();
  if (TIME_TAG > time) time = TIME_TAG;
  lazy_time = time;

//...... get the message

  message_queue = app1obj->get_message_queue();
  msg = (C_MSG *)message_queue->get_top();
  lazy_msg = msg;
  *message_queue -= msg;

//...... get the number of messages returned

  lazy_n_received = app1obj->get_n_received();
  n_received = lazy_n_received;

//...... process the message

  if (msg->get_source() == app1obj->get_HYPER_node()) {

    n_received++;

    if (n_received == app1obj->get_fanout()) {

      time += 0.000000001;
      schedule (
		time,
		START_APP1,
		APP1,
		LOCAL_ID,
		NODE	);
    }

  }else{

    time += app1obj->get_compute_time();
    send_mess = (SEND_MESS *) schedule (
		time,
		SEND,
		HYPE_NODE,
		LOCAL_ID,
		NODE	);

    send_mess->destination = msg->get_source();
    send_mess->source = msg->get_source();
    send_mess->application = int(APP1);
    send_mess->read_event = int(READ_APP1);
    send_mess->priority = msg->get_priority();
    send_mess->dummy_bytes = msg->get_dummy_bytes();

  }

  RB_FREE_DELETE(msg);

}

/************************************************************************
* perm_process : permanently process event				*
************************************************************************/
void C_READ_APP1::perm_process() {

}

/************************************************************************
* cleanup : clean up data structures used				*
************************************************************************/
void C_READ_APP1::cleanup() {

}



