// route.C method file

#include "SpStdio.H"

#include "nodeobj.H"
#include "msg.H"
#include "freelist.H"
#include "SpScheduleEvent.H"
#include "route.H"
#include "receive_mess.H"
#include "send_mess.H"
#include "nodeman.H"

#include "RB_SpContainerFuncs.H"

EVENT_TYPE (RECEIVE);
EVENT_TYPE (SEND);

/************************************************************************
* C_ROUTE : construct a route object					*
************************************************************************/
C_ROUTE::C_ROUTE() {
  set_lazy();
}

/************************************************************************
* init : initialize the event		                                *
************************************************************************/
void C_ROUTE::init(C_HEADER *) {
  
}

/************************************************************************
* exchange : exchange state variables                                   *
************************************************************************/
void C_ROUTE::exchange() {
  C_NODEOBJ *nodeobj;

  SIMOBJ->EXCHANGE_SEED(seed);

  nodeobj = (C_NODEOBJ *)SIMOBJ;
  nodeobj->exchange_busy(busy);
  nodeobj->exchange_time(time);

}

/************************************************************************
* lazy : lazy cancellation routine                                      *
************************************************************************/
int C_ROUTE::lazy() {
  C_NODEOBJ *nodeobj;
  C_XQUEUE *message_queue;
  C_MSG *msg;
  int temp;
  double t;

  nodeobj = (C_NODEOBJ *)SIMOBJ;

  if (lazy_seed != nodeobj->get_SEED()) return 0;
  if (lazy_bandwidth != nodeobj->get_HYPER_bandwidth()) return 0;

  message_queue = nodeobj->get_next_message_queue();
  msg = (C_MSG *)message_queue->get_top();
  if (lazy_msg != msg) return 0;

  t = TIME_TAG;
  if (t < nodeobj->get_time()) t = nodeobj->get_time();
  if (lazy_time != t) return 0;

  temp = 0;
  if (!lazy_nmess && !(nodeobj->get_n_messages())) temp = 1;
  if (lazy_nmess && (nodeobj->get_n_messages())) temp = 1;

  return temp;

}

/************************************************************************
* temp_process : temporarily process event				*
************************************************************************/
void C_ROUTE::temp_process() {
  C_NODEMAN *nodeman;
  C_NODEOBJ *nodeobj;
  C_XQUEUE *message_queue;
  C_MSG *msg;
  SEND_MESS *send_mess;
  RECEIVE_MESS *receive_mess;
  int next_node;
  int nd;
  int lid;

  lazy_seed = SIMOBJ->get_SEED();
  seed = lazy_seed;
  RANDOM->set_seed(seed);
  RANDOM->set_float_limits(0.000000001, 0.000000002);

  nodeman = (C_NODEMAN *)get_manager(HYPE_NODE);
  nodeobj = (C_NODEOBJ *)SIMOBJ;

//...... remove the next message from the priority message queues

  message_queue = nodeobj->get_next_message_queue();
  msg = (C_MSG *)message_queue->get_top();
  *message_queue -= msg;

  lazy_msg = msg;

//...... compute the time to get the message off this node

  time = TIME_TAG;
  lazy_time = nodeobj->get_time();

//...... get the lazy bandwidth

  lazy_bandwidth = nodeobj->get_HYPER_bandwidth();

  if (time < lazy_time) time = lazy_time;

//...... compute the next node for routing

  next_node = nodeobj->get_route(msg->get_destination());
  if (next_node == -1) {

    time += 0.000000001;

    receive_mess = (RECEIVE_MESS *) schedule (
		time + RANDOM->get_random_float(),
		RECEIVE,
		msg->get_application(),
		LOCAL_ID,
		NODE	);

    receive_mess->destination = msg->get_destination();
    receive_mess->source = msg->get_source();
    receive_mess->application = msg->get_application();
    receive_mess->read_event = msg->get_read_event();
    receive_mess->priority = msg->get_priority();
    receive_mess->dummy_bytes = msg->get_dummy_bytes();

  }else{

    nodeman->get_nodeid(next_node,nd,lid);

    time += nodeobj->get_HYPER_send_latency() +
	    (msg->get_dummy_bytes() / lazy_bandwidth);

    send_mess = (SEND_MESS *) schedule (
		time
		+ nodeobj->get_HYPER_receive_latency()
		+ RANDOM->get_random_float(),
		SEND,
		HYPE_NODE,
		lid,
		nd	);

    send_mess->destination = msg->get_destination();
    send_mess->source = msg->get_source();
    send_mess->application = msg->get_application();
    send_mess->read_event = msg->get_read_event();
    send_mess->priority = msg->get_priority();
    send_mess->dummy_bytes = msg->get_dummy_bytes();

  }

  RB_FREE_DELETE(msg);

  lazy_nmess = nodeobj->get_n_messages();
  if (lazy_nmess) {
    busy = 1;
    schedule (time + RANDOM->get_random_float(), ROUTE,
		OBJECT_TYPE, LOCAL_ID, NODE);
  }else{
    busy = 0;
  }

//...... spin loop for higher granularity event processing

  spin_loop(nodeobj->get_spin());

//...... save the seed for exchange

  seed = RANDOM->get_seed();

}

/************************************************************************
* perm_process : permanently process event				*
************************************************************************/
void C_ROUTE::perm_process() {

  if (lazy_bandwidth == 1234.0) {
    fprintf(stderr,"Error (route) bad bandwidth\n");
  }

}

/************************************************************************
* cleanup : clean up data structures used				*
************************************************************************/
void C_ROUTE::cleanup() {

}



