#include "SpStdio.H"
#include "SpStdlib.H"
#include "SpMemory.H"
#include "SpSignal.H" 
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpAssert.H"

#include "pause_mess.H"
#include "host_user.H"
#include "crash_on_divide_by_zero.H"

int stop_flag;
int cntrl_c;
int PauseFlag = 1;

void sigintfunc(int)
{
    signal( SIGINT, sigintfunc ); 
    cntrl_c = 1;
    if (stop_flag) exit(0);
} 

void get_id();
void query();
void monitor();
void multi_monitor();
void pause_continue();

C_HOST_USER host_user;

int main() {
  crash_on_divide_by_zero();

  int option;
  C_EM_HEADER *EM_header;

  signal( SIGINT, sigintfunc ); 

  fprintf(stderr,"\f");

  while (1) {

    host_user.reset();
    stop_flag = 1;
    cntrl_c = 0;

    fprintf(stderr,"Interactive Options\n");
    fprintf(stderr,"-------------------\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"0. Exit\n");
    fprintf(stderr,"1. Get ID\n");
    fprintf(stderr,"2. Query\n");
    fprintf(stderr,"3. Monitor\n");
    fprintf(stderr,"4. Multi-Monitor\n");
    fprintf(stderr,"5. Pause/Continue Simulation\n");
    fprintf(stderr,"\n");

    fprintf(stderr,"Enter option(0,1,2,3,4,5): ");
    fflush(stderr);
    
    char line[256];
    if (fgets(line, 256, stdin) == NULL || line[0] == '\0' || line[1] != '\0'
	|| line[0] < '0' || line[0] > '5')
    {
      fprintf(stderr, "Invalid input.\n");
      continue;
    }
    
    option = line[0] - '0';

//...... before doing anything, flush out all of the messages

    EM_header = NULL;
    EM_header = host_user.get_message();
    while (EM_header != NULL) {
      delete EM_header;
      EM_header = host_user.get_message();
    }

    if (option == 0) break;

    switch (option) {

      case 1:
	get_id();
        break;

      case 2:
	query();
        break;

      case 3:
	monitor();
        break;

      case 4:
	multi_monitor();
        break;

      case 5:
	pause_continue();
        break;

      default:
	assert(0);
	break;

    }

  }

  return 0;
}

/************************************************************************
* get_id - get the id of an object given its name			*
************************************************************************/
void get_id() {
  int id;
  char name[40];

  fprintf(stderr,"ID> Enter object name: ");
  fflush(stderr);
  fgets(name, 40, stdin);

  id = host_user.getid(name);
  fprintf(stderr, "%s has ID %d\n",name,id);

}

/************************************************************************
* query - query an object based on its id				*
************************************************************************/
void query() {
  int id;
  C_EM_HEADER *EM_header;

  fprintf(stderr,"QUERY> Enter object id: ");
  fflush(stderr);

  char line[256];
  fgets(line, 256, stdin);
  id = atoi(line);
  
  EM_header = host_user.query(id);

  if (EM_header != NULL) {
    delete EM_header;
  }

}

/************************************************************************
* monitor - monitor an object based on its id				*
************************************************************************/
void monitor() {
  int id;
  C_EM_HEADER *EM_header;

  fprintf(stderr,"MONITOR> Enter object id: ");
  fflush(stderr);

  char line[256];
  fgets(line, 256, stdin);
  id = atoi(line);

//...... loop until control c monitoring the object

  stop_flag = 0;
  while (1) {

    EM_header = host_user.monitor(id);

    if (EM_header != NULL) {
      delete EM_header;
    }

    if (cntrl_c == 1) {
      cntrl_c = 0;
      host_user.disconnect();
      break;
    }

  }

}

/************************************************************************
* multi_monitor - monitor multiple objects based on names		*
************************************************************************/
void multi_monitor() {
  int i;
  int n;
  C_EM_HEADER EM_header;
  C_EM_HEADER *EM_header1;
  C_HOST_USER *host_user_array[500];
  int id[500];
  char name[120];

  EM_header.time_tag = -1.0e20;
  EM_header.evtype = -1;
  EM_header.obtype = -1;
  EM_header.ext = 1;
  EM_header.rbid = -1;
  EM_header.EM_socket = -1;
  EM_header.EM_interaction = (EM_MONITOR | EM_CONNECT | EM_AGGRESSIVE);
  EM_header.data_bytes = 0;

  n = 0;
  while (1) {
    fprintf(stderr,"MONITOR> Enter object name (quit): ");
    fflush(stderr);
    cin >> name;
    if (!strcmp(name,"quit")) break;
    host_user_array[n] = new C_HOST_USER();
    id[n] = host_user_array[n]->getid(name);
    cout << "ID = " << id[n] << " name = " << name << endl;
    n++;
  }

  for (i=0; i<n; i++) {
    EM_header.objid = id[i];
    host_user_array[i]->send_message(&EM_header);
  }

//...... loop until control c monitoring the objects

  stop_flag = 0;
  while (1) {
    for (i=0; i<n; i++) {
      EM_header1 = host_user_array[i]->get_message();
      if (EM_header1 != NULL) {
        host_user_array[i]->print_query(EM_header1);
	delete EM_header1;
      }
    }

    if (cntrl_c == 1) {
      cntrl_c = 0;
      EM_header.EM_interaction = (EM_MONITOR | EM_DISCONNECT | EM_AGGRESSIVE);
      for (i=0; i<n; i++) {
        EM_header.objid = id[i];
	host_user_array[i]->send_message(&EM_header);
	delete host_user_array[i];
      }
      break;
    }

  }

}

/************************************************************************
* pause_continue - pause or continue the execution of the simulation	*
************************************************************************/
void pause_continue() {
  PAUSE_MESS pause_mess;

  pause_mess.objid = 0;
  pause_mess.PauseFlag = PauseFlag;

  if (PauseFlag) {
    host_user.send_module(&pause_mess);
    PauseFlag = 0;
  }else{
    pause_mess.EM_interaction = (EM_MODULE | EM_BLOCKING | EM_AGGRESSIVE);
    host_user.send_message(&pause_mess);
    PauseFlag = 1;
  }

}

