// serve_event.C method file

#include "SpStdio.H"
#include "SpAssert.H"

#include "serve_event.H"
#include "servman.H"
#include "servobj.H"

/************************************************************************
* C_SERVE_EVENT : construct a serve_event object			*
************************************************************************/
C_SERVE_EVENT::C_SERVE_EVENT() {

// turn on lazy cancellation for this event type

  set_lazy();

}

/************************************************************************
* exchange : exchange state variables                                   *
************************************************************************/
void C_SERVE_EVENT::exchange() {

  SIMOBJ->EXCHANGE_SEED(seed);

}

/************************************************************************
* lazy : lazy cancellation routine                                      *
************************************************************************/
int C_SERVE_EVENT::lazy() {
  C_SERVOBJ *servobj;

// pass test if the seeds are the same

  if (lazy_seed != SIMOBJ->get_SEED()) return 0;

// pass test if both previous and present server times are less than the
// time tag of this event (i.e., the customer's arrival time). Otherwise
// the server times must be identicle

  servobj = (C_SERVOBJ *)SIMOBJ;

  if (lazy_server_time < TIME_TAG) {
    if (servobj->get_server_time() > TIME_TAG) return 0;
  }else{
    if (servobj->get_server_time() != lazy_server_time) return 0;
  }

// pass this test if n_served is the same

  if (lazy_n_served != servobj->get_n_served()) return 0;

// pass this test if n_served is the same

  if (lazy_wait_time != servobj->get_wait_time());

// if we got this far, then we passed all of the tests so roll the event
// forward (i.e., return 1)

  return 1;

}

/************************************************************************
* process : optimistically process event				*
************************************************************************/
void C_SERVE_EVENT::process() {
  C_SERVOBJ *servobj;
  C_SERVMAN *servman;
  int n_servers;
  int next_server;
  int next_local_id;
  int next_node;
  double service_time;
  double travel_time;

  double server_time;	// time of the server
  int n_served;		// total number of customers served
  double wait_time;	// total time spent waiting

  // DATA is the variable length buffer, but since this event has no
  // variable length buffer I expect DATA to be NULL.
  assert(DATA == NULL);

  if (TIME_TAG != MESSAGE->time_tag) {
    cout << "Major problem in serve_event, bad time tag" << endl;
    cout << "TIME_TAG = " << TIME_TAG << endl;
    cout << "time_tag = " << MESSAGE->time_tag << endl;
    cout << endl;
  }

// cast the SIMOBJ to SERVOBJ because we know what kind of object this
// event acts on. Also, get the servman object manager

  servobj = (C_SERVOBJ *)SIMOBJ;
  servman = (C_SERVMAN *)get_manager(SERVER);

// get data from the the server object and store values for later lazy
// cancellation use

  seed = SIMOBJ->get_SEED();
  lazy_seed = seed;

  server_time = servobj->get_server_time();
  lazy_server_time = server_time;

  n_served = servobj->get_n_served();
  lazy_n_served = n_served;

  wait_time = servobj->get_wait_time();
  lazy_wait_time = wait_time;

// compute the service and travel times

  if (servman->get_const_serve_time()) {
    service_time = servman->get_serve_time();
  }else{
    service_time = servobj->service_time(seed) + servman->get_serve_time();
  }

  if (servman->get_const_travel_time()) {
    travel_time = servman->get_travel_time();
  }else{
    travel_time = servobj->travel_time(seed) + servman->get_travel_time();
  }

// compute the server's new time

  if (TIME_TAG < server_time) {
    server_time += service_time;
  }else{
    server_time = TIME_TAG + service_time;
  }

// update the n_served and wait_time variables (which will be exchanged with
// the servobj in the exchange method for this event

  n_served++;
  wait_time += server_time - TIME_TAG;

// choose the next server for this customer

  n_servers = servman->get_n_servers();
  if (servman->get_fully_interconnected()) {
    RANDOM->set_seed(seed);
    RANDOM->set_int_limits(0, n_servers-1);
    next_server = RANDOM->get_random_int();
    seed = RANDOM->get_seed();
  }else{
    next_server = (LOCAL_ID + 1) % n_servers;
  }

// get the node and local id from the servman for the next server

  servman->get_nodid(next_server, next_node, next_local_id);

// schedule this event again

  schedule(	server_time+travel_time,
		EVENT_TYPE,
		OBJECT_TYPE,
		next_local_id,
		next_node
	);

  servobj->set_server_time(server_time);
  servobj->set_n_served(n_served);
  servobj->set_wait_time(wait_time);

}
