#ifdef TEST_HIGRID
#include "random.H"

class TEST_PUBSUB_ITEM : public MAIN_PUBSUB_ITEM {
  public:
    double Xlo;
    double Xhi;
    double Xres;
    double Ylo;
    double Yhi;
    double Yres;
    double Flo;
    double Fhi;
    double Fres;

    virtual void print() {
      cout << "Global Id " << id << ": Xlo Xhi Xres "
	   << Xlo << " " << Xhi << " " << Xres
	   << ",  Ylo Yhi Yres "
	   << Ylo << " " << Yhi << " " << Yres
	   << ",  Flo Fhi Fres "
	   << Flo << " " << Fhi << " " << Fres
	   << endl;
    }

};

#include "SpCrashOnDivZero.H"

/************************************************************************
* Standalone test program to verify that the hierarchical grid data	*
* structure is working correctly.					*
************************************************************************/
int main() {
  crash_on_divide_by_zero();

  SpHiGrid *HiGrid;
  int j;
  int global_id;
  double X[2], Xlo, Xhi, Xr, Xres[3];
  double Y[2], Ylo, Yhi, Yr, Yres[2];
  double F[2], Flo, Fhi, Fr, Fres[3];
  SpPubSubItem *PubSubItem;
  TEST_PUBSUB_ITEM *TestPubSubItem;
  TEST_PUBSUB_ITEM *TestPub;
  TEST_PUBSUB_ITEM *TestSub;
  SpStopWatch StopWatch("Publisher");
  C_QUEUE Qsub;
  C_QUEUE Qpub;
  C_RANDOM Random;
  SpDimInfo *DimInfo;

/*
** Set up arguments for the PlayBox HiGrid
*/

  char *SpaceName = strdup("PlayBox");
  int SpaceNum = 4;

  int NumDims = 3;

  char *Dim1 = strdup("X");
  char *Dim2 = strdup("Y");
  char *Dim3 = strdup("Frequency");

  int XdimNum = 0;
  int YdimNum = 1;
  int FdimNum = 2;

  Xlo = 0; 	Xhi = 10000; 	Xres[0] = 900; 	Xres[1] = 120; 	Xres[2] = 15;
  Ylo = 3000; 	Yhi = 8000; 	Yres[0] = 600; 	Yres[1] = 65;
  Flo = 100; 	Fhi = 1000; 	Fres[0] = 80; 	Fres[1] = 7; 	Fres[2] = 1;

/*
** Set up the DimInfo object for the normal (non VA) interfaces
*/

  DimInfo = new SpDimInfo[NumDims];

  DimInfo[XdimNum].SetDimName(Dim1);
  DimInfo[XdimNum].SetDimNum(XdimNum);
  DimInfo[XdimNum].SetLo(Xlo);
  DimInfo[XdimNum].SetHi(Xhi);
  DimInfo[XdimNum].SetNresolutions(3);
  DimInfo[XdimNum].SetResolution(0, Xres[0]);
  DimInfo[XdimNum].SetResolution(1, Xres[1]);
  DimInfo[XdimNum].SetResolution(2, Xres[2]);

  DimInfo[YdimNum].SetDimName(Dim2);
  DimInfo[YdimNum].SetDimNum(YdimNum);
  DimInfo[YdimNum].SetLo(Ylo);
  DimInfo[YdimNum].SetHi(Yhi);
  DimInfo[YdimNum].SetNresolutions(2);
  DimInfo[YdimNum].SetResolution(0, Yres[0]);
  DimInfo[YdimNum].SetResolution(1, Yres[1]);

  DimInfo[FdimNum].SetDimName(Dim3);
  DimInfo[FdimNum].SetDimNum(FdimNum);
  DimInfo[FdimNum].SetLo(Flo);
  DimInfo[FdimNum].SetHi(Fhi);
  DimInfo[FdimNum].SetNresolutions(3);
  DimInfo[FdimNum].SetResolution(0, Fres[0]);
  DimInfo[FdimNum].SetResolution(1, Fres[1]);
  DimInfo[FdimNum].SetResolution(2, Fres[2]);

  cout << "testing HiGrid" << endl;

/*
** Create the HiGrid object and then initialize it. Note that the arguments
** passed into the init routine use variable argument lists. Each dimension
** is terminated with an END_DIM argument
*/

  HiGrid = new SpHiGrid();

#ifdef TEST_VA_INTERFACES
  HiGrid->va_init(SpaceName, SpaceNum, NumDims,
	    Dim2, YdimNum, Ylo, Yhi, Yres[0], Yres[1], END_DIM,
            Dim1, XdimNum, Xlo, Xhi, Xres[0], Xres[1], Xres[2], END_DIM,
	    Dim3, FdimNum, Flo, Fhi, Fres[0], Fres[1], Fres[2], END_DIM);
#else
  HiGrid->init(SpaceName, SpaceNum, NumDims, DimInfo);
#endif

  delete [NumDims] DimInfo;

/*
** Randomly add publishers and subscribers to the HiGrid with different
** resolutions.
*/

  StopWatch.StartTimer();
  for (global_id=0; global_id<N_ITERATIONS; global_id++) {

// X dimension

    Random.set_float_limits(Xlo, Xhi);
    X[0] = X[1] = Random.get_random_float();

    Random.set_int_limits(0,2);
    j = Random.get_random_int();

    Random.set_float_limits(0, 2*Xres[j]);
    X[0] -= Random.get_random_float();
    X[1] += Random.get_random_float();
    Xr = X[1] - X[0] + Random.get_random_float();

    if (X[0] < Xlo) X[0] = Xlo;
    if (X[1] > Xhi) X[1] = Xhi;


// Y dimension

    Random.set_float_limits(Ylo, Yhi);
    Y[0] = Y[1] = Random.get_random_float();

    Random.set_int_limits(0,1);
    j = Random.get_random_int();

    Random.set_float_limits(0, 2*Yres[j]);
    Y[0] -= Random.get_random_float();
    Y[1] += Random.get_random_float();
    Yr = Y[1] - Y[0] + Random.get_random_float();

    if (Y[0] < Ylo) Y[0] = Ylo;
    if (Y[1] > Yhi) Y[1] = Yhi;


// Frequency dimension

    Random.set_float_limits(Flo, Fhi);
    F[0] = F[1] = Random.get_random_float();

    Random.set_int_limits(0,2);
    j = Random.get_random_int();

    Random.set_float_limits(0, 2*Fres[j]);
    F[0] -= Random.get_random_float();
    F[1] += Random.get_random_float();
    Fr = F[1] - F[0] + Random.get_random_float();

    if (F[0] < Flo) F[0] = Flo;
    if (F[1] > Fhi) F[1] = Fhi;

// New publisher and subscriber

#ifndef HIGRID_TIMINGS
    cout << "Adding Global Id " << global_id
         << ":  [X, (" << X[0] << "," << X[1] << "), Res " << Xr << "]"
	 << "  [Y, (" << Y[0] << "," << Y[1] << "), Res " << Yr << "]"
	 << "  [F, (" << F[0] << "," << F[1] << "), Res " << Fr << "]"
	 << endl;
#endif

    TestPubSubItem = new TEST_PUBSUB_ITEM();

    TestPubSubItem->set_global_id(global_id);
    TestPubSubItem->set_node(0);
    TestPubSubItem->set_local_id(3);
    TestPubSubItem->set_type(2);

    TestPubSubItem->Xlo = X[0];
    TestPubSubItem->Xhi = X[1];
    TestPubSubItem->Xres = Xr;

    TestPubSubItem->Ylo = Y[0];
    TestPubSubItem->Yhi = Y[1];
    TestPubSubItem->Yres = Yr;

    TestPubSubItem->Flo = F[0];
    TestPubSubItem->Fhi = F[1];
    TestPubSubItem->Fres = Fr;

    TestPubSubItem->SetNdimensions(NumDims);
    TestPubSubItem->SetLHR(XdimNum, X[0], X[1], Xr);
    TestPubSubItem->SetLHR(YdimNum, Y[0], Y[1], Yr);
    TestPubSubItem->SetLHR(FdimNum, F[0], F[1], Fr);

    if (global_id == 34) {
      TestPub = TestPubSubItem;
    }

#ifdef TEST_VA_INTERFACES
    HiGrid->va_AddPublisher(TestPubSubItem,
			 XdimNum, X[0], X[1], Xr,
			 YdimNum, Y[0], Y[1], Yr,
			 FdimNum, F[0], F[1], Fr,
			 END_SPACE);
#else
    HiGrid->AddPublisher(TestPubSubItem);
#endif

    TestPubSubItem = new TEST_PUBSUB_ITEM();

    TestPubSubItem->set_global_id(global_id);
    TestPubSubItem->set_node(0);
    TestPubSubItem->set_local_id(3);
    TestPubSubItem->set_type(2);

    TestPubSubItem->Xlo = X[0];
    TestPubSubItem->Xhi = X[1];
    TestPubSubItem->Xres = Xr;

    TestPubSubItem->Ylo = Y[0];
    TestPubSubItem->Yhi = Y[1];
    TestPubSubItem->Yres = Yr;

    TestPubSubItem->Flo = F[0];
    TestPubSubItem->Fhi = F[1];
    TestPubSubItem->Fres = Fr;

    TestPubSubItem->SetNdimensions(NumDims);
    TestPubSubItem->SetLHR(XdimNum, X[0], X[1], Xr);
    TestPubSubItem->SetLHR(YdimNum, Y[0], Y[1], Yr);
    TestPubSubItem->SetLHR(FdimNum, F[0], F[1], Fr);

    if (global_id == 68) {
      TestSub = TestPubSubItem;
    }

#ifdef TEST_VA_INTERFACES
    HiGrid->va_AddSubscriber(TestPubSubItem,
			 XdimNum, X[0], X[1], Xr,
			 YdimNum, Y[0], Y[1], Yr,
			 FdimNum, F[0], F[1], Fr,
			 END_SPACE);
#else
    HiGrid->AddSubscriber(TestPubSubItem);
#endif

  }

  StopWatch.StopTimer();
  cout << "Time to add an item is "
       << StopWatch.get_CumulativeTime()/N_ITERATIONS << endl;;

#ifndef HIGRID_TIMINGS
  HiGrid->RecursivePrint();
#endif

/*
** Get publishers and subscribers to verify that things are working
*/

  StopWatch.StartTimer();
#ifdef TEST_VA_INTERFACES
  HiGrid->va_GetPublishers(&Qpub,
		  XdimNum, 5400.0, 5800.0,
		  YdimNum, 2900.0, 3200.0,
		  FdimNum, 420.0, 430.0,
		  END_SPACE);
#else
  HiGrid->GetPublishers(TestPub, &Qpub);
#endif

#ifndef HIGRID_TIMINGS
  cout << Qpub.get_length() << " Publishers" << endl;
  Qpub.print();
#endif

#ifdef TEST_VA_INTERFACES
  HiGrid->va_GetSubscribers(&Qsub,
		  XdimNum, 5400.0, 5800.0,
		  YdimNum, 2900.0, 3200.0,
		  FdimNum, 420.0, 430.0,
		  END_SPACE);
#else
  HiGrid->GetSubscribers(TestSub, &Qsub);
#endif

#ifndef HIGRID_TIMINGS
  cout << Qsub.get_length() << " Subscribers" << endl;
  Qsub.print();
#endif

  cout << "Time to get publishers or subscribers: "
       << StopWatch.StopTimer()/2.0 << endl;

/*
** Check the integrity of the HiGrid
*/

  HiGrid->CheckIntegrity();

/*
** Delete everything and cleanup
*/

  StopWatch.StartTimer();
  for (global_id=0; global_id<N_ITERATIONS; global_id++) {
    PubSubItem = HiGrid->RemoveSubscriber(global_id);
    delete PubSubItem;
    PubSubItem = HiGrid->RemovePublisher(global_id);
    delete PubSubItem;
  }
  cout << "Time to remove publishers and subscribers: "
       << StopWatch.StopTimer()/(2.0*N_ITERATIONS) << endl;

  HiGrid->print();
  HiGrid->RecursivePrint();
  delete HiGrid;

}

#endif




