#include "SpServiceGroup.H"

/************ EXAMPLE SERVICE GROUP CLASS **************** 

  This is an example of how to define a service group class, an
  instance of which will be created by the MESSAGE_SERVER and will
  respond to a set of arriving messages.  The essential idea is this:
  you must write a method capable of handling each possible message
  type.  When the server receives a message intended for your object,
  it will read an int off the message header and this will tell it
  which one of the methods to invoke; so each handler-method must be
  identified by an associated int (see [1] below).

  You may have as many other methods or data members as you
  want--that's the idea of encapsulating this all in an object.  After
  you handler methods have been defined, you must write a constructor
  that will "register" these handlers so that the MESSAGE_SERVER can
  call them (by putting pointers to the methods into an array).  This
  requires two steps--one to allocate storage, and one to store a
  pointer for each method (see [2] and [3] below).

  Finally, you must write a global function to create an instance of
  your class (see [4]).  This is so the MESSAGE_SERVER can
  instantinate your obj.  It may have other SERVICE_GROUP class defs
  like yours.  When it receives a message addressed to a
  SERVICE_GROUP, it does the following: if the group exists, it
  invokes the appropriate method to handle the message; if the group
  does NOT exist, it creates an instance of the right kind of group
  (determined by the serviceType field in the msg header), and then
  calls the handler as above.

**********************/


class BogusServer : public SERVICE_GROUP {
public:
  enum { F0=0, F1, F2, F3, F4, NUMHANDLERS }; // [1] ints associated w/handlers
  
  int Knt;
  int* LocalNodeIDs;
  int* BoxNumbers;
  int* LocalNodeZeros;
  CONNECTION * SendSocket;
  void Message(CONNECTION *conn, GENERIC_MSG_HEADER *mh);
  
  void GlobalSum(CONNECTION *conn, GENERIC_MSG_HEADER *mh);
  
  void GlobalMin(CONNECTION *conn, GENERIC_MSG_HEADER *mh);
  
  void GlobalMax(CONNECTION *conn, GENERIC_MSG_HEADER *mh);
  
  void nb_Redop(CONNECTION *conn, GENERIC_MSG_HEADER *mh);
  
  void Redop(CONNECTION *conn, GENERIC_MSG_HEADER *mh);
  
  void Startup(CONNECTION *conn, GENERIC_MSG_HEADER *mh);

  ~BogusServer(){
    if (SendSocket != NULL){
      delete [] SendSocket;
      delete [] LocalNodeZeros;
      delete [] BoxNumbers;
      delete [] LocalNodeIDs;
    }
  }
  
  BogusServer(void) {
    Knt = 0;
    SendSocket = NULL;
    LocalNodeIDs = NULL;
    BoxNumbers == NULL;
    LocalNodeZeros == NULL;
    allocateHandlerStorage(NUMHANDLERS);// [2] Allocate storage for your msg handlers
    
      // [3] Use this macro to store pointer, and associate an int with the pointer
    //     (just puts pointer in Nth slot in an array):
    STORE_MSG_HANDLER(NORMAL_MESSAGE, BogusServer, Message);
    STORE_MSG_HANDLER(GLOBAL_SUM, BogusServer, GlobalSum);
    STORE_MSG_HANDLER(GLOBAL_MIN, BogusServer, GlobalMin);
    STORE_MSG_HANDLER(GLOBAL_MAX, BogusServer, GlobalMax);
    STORE_MSG_HANDLER(NON_BLOCKING_SYNCHRONIZATION, BogusServer, nb_Redop);
    STORE_MSG_HANDLER(BLOCKING_SYNCHRONIZATION, BogusServer, Redop);
    STORE_MSG_HANDLER(TCPIP_STARTUP, BogusServer, Startup);
  }
};

// [4] Fcn to create new BogusServer (must take void arg list):
BogusServer *makeBogusServer(void);
