;; import-mode.el --- IMPORT editing support package
;;
;; Author: Alan Whitehurst <whitehur@cs.uiuc.edu>
;; Based upon the ModSim mode
;; Keywords: languages
;;
;; A major mode for editing IMPORT code.  It provides convenient abbrevs
;; for IMPORT keywords, knows about the standard layout rules, and supports
;; a native compile command.

(defun import-mode-help ()
  (interactive)
  (pop-to-buffer "--- import-mode Help ---")
  (if (= (buffer-size) 0)
      (insert-string
       "

   Keystroke	Function

     Control-i	Tabs to the correct IMPORT tab position.
     Control-j	NewLine, then move to same column as preceding line.
     Meta-q	Indent entire module pointed at by cursor.
     \C-ca	Inserts the template for an ASK method declaration.
     \C-cb	Inserts the template for a BLOCK structure.
     \C-cc	Inserts the template for a CONSTRUCTOR method declaration.
     \C-cd	Inserts the template for a DESTRUCTOR method declaration.
     \C-ce	Inserts the template for an ELSEIF  within the body of an IF structure.
     \C-cf	Inserts the template for a FOR structure.
     \C-ch	Inserts the template for a comment header into a module.
     \C-ci	Inserts the template for an IF structure.
     \C-ck	Inserts the template for a CASE structure.
     \C-cl	Inserts the template for a LOOP structure.
     \C-cm	Inserts the template for a MODULE declaration.
     \C-co	Inserts the template for an OBJECT structure.
     \C-cp	Inserts the template for a FUNCTION structure.
     \C-cq	Inserts the template for a QUERY method declaration;
     \C-cr	Inserts the template for a REPEAT structure.
     \C-ct	Inserts the template for a TELL method delcaration;
     \C-cw	Inserts the template for the WHILE structure.
     \C-c{	Inserts the symbol(s) for beginning a multi-line comment.
     \C-c}	Inserts the symbol(s) for ending a multi-line comment.

     \C-c\C-c	Invokes the IMPORT compiler.
     \C-c\C-h	Invokes the IMPORT Help Buffer.
     \C-c\C-p	Invokes the IMPORT Paint Mode (colors text).
     \C-c\C-s	Invokes the compiler SETUP utility.
     \C-c\C-t   Invokes the IMPORT TypeChecker
     \C-c\C-z	Suspends emacs."
    ))
  (goto-char (point-min))
  (set-buffer-modified-p nil)
  (setq buffer-read-only t)
  (other-window 1))


;;; Variable declarations:

(defvar import-mode-syntax-table nil
  "Syntax table in use in IMPORT buffers.")

(defvar import-last-compiled-buffer ""
  "Name of the last IMPORT buffer to be compiled")

;;; Variables that can be changed interactively:

(defvar import-indent 3 
  "The amount of each tab-stop in import-mode")
(put 'import-indent 'variable-interactive "nAmount to indent for each tab: ")

;*** remove import-bin-dir
;(defvar import-bin-dir "/opt/net1/asset/isle/import/bin"
;  "Path to the binary directory for import.")
;(put 'import-bin-dir 'variable-interactive "sPath where IMPORT compiler is found: ");
;
;(defvar import-runtime-dir "/home/Asset/src/import/runtime"
;  "Path to the runtime directory for import.")
;(put 'import-runtime-dir 'variable-interactive "sPath of IMPORT runtime directory: ")
;

(defvar import-compile-command "pi -p"
  "Command to compile IMPORT programs")
(put 'import-compile-command 'variable-interactive "Command to compile IMPORT programs: ")

(defvar import-typecheck-command "pi -t"
  "Command to typecheck IMPORT programs without compiling them.")

;(defvar import-cpp-include-dirs import-runtime-dir
;  "Names of C++ include directories for IMPORT programs.")
;(put 'import-cpp-include-dirs 'variable-interactive "sName of include directory of C++ header files: ")
;
;(defvar import-link-lib-paths "-L/opt/net1/asset/isle/import/runtime"
;  "Path to modules for library routines")
;(put 'import-link-lib-paths 'variable-interactive "sPath to IMPORT libraries: ")

;(defvar import-link-cpp-libraries "-lBuiltIn -lMath -ltask -lm -lsocket -lnsl" 
;  "String containing standard libraries for linking.")
;(put 'import-link-cpp-libraries 'variable-interactive "sImport standard libraries to link string: ")

;(defvar import-link-command "CC"
;  "Command to C++ compile and link IMPORT programs")
;(put 'import-link-command 'variable-interactive "sCommand to C++ compile and link IMPORT-generated code: ")

(defvar import-link-name "process"
  "Name of the executable.")
(put 'import-link-name 'variable-interactive "sName for the executable: ")


;;; mode tables

(if import-mode-syntax-table
    ()
  (let ((table (make-syntax-table))
	(i 0))
    (while (< i ?0)
      (modify-syntax-entry i "_   " table)
      (setq i (1+ i)))
    (setq i (1+ ?9))
    (while (< i ?A)
      (modify-syntax-entry i "_   " table)
      (setq i (1+ i)))
    (setq i (1+ ?Z))
    (while (< i ?a)
      (modify-syntax-entry i "_   " table)
      (setq i (1+ i)))
    (setq i (1+ ?z))
    (while (< i 128)
      (modify-syntax-entry i "_   " table)
      (setq i (1+ i)))
    (modify-syntax-entry ?\\ "\\  " table) 
    (modify-syntax-entry ?\( "()  " table)
    (modify-syntax-entry ?\) ")(  " table)
    (modify-syntax-entry ?\[ "(]  " table)
    (modify-syntax-entry ?\] ")[  " table)
    (modify-syntax-entry ?\t "    " table)
    (modify-syntax-entry ?\  "    " table)
    (modify-syntax-entry ?{ "<   " table)
    (modify-syntax-entry ?} ">   " table)
    (modify-syntax-entry ?: ".   " table)
    (modify-syntax-entry ?\; ".   " table)
    (modify-syntax-entry ?. ".   " table)
    (modify-syntax-entry ?* ".   " table)
    (modify-syntax-entry ?+ ".   " table)
    (modify-syntax-entry ?- ". 2 " table)
    (modify-syntax-entry ?= ".   " table)
    (modify-syntax-entry ?% ".   " table)
    (modify-syntax-entry ?< ". 1 " table)
    (modify-syntax-entry ?> ".   " table)
    (modify-syntax-entry ?\' "\"   " table)
    (modify-syntax-entry ?\n ">   " table)
    (modify-syntax-entry ?\f ">   " table)
    (setq import-mode-syntax-table table)))

(defvar import-mode-map nil
  "Keymap used in IMPORT mode.")

(if import-mode-map ()
  (let ((map (make-sparse-keymap)))
    (define-key map "\^i"   'import-indent-line)
    (define-key map "\^j"   'import-newline)
    ;;
    (define-key map [menu-bar tools]
      (cons "Tools" (make-sparse-keymap "Tools")))
    (define-key map "\M-q" 'import-indent-module)
    (define-key map [menu-bar tools format] '("Format Module" . import-indent-module))
    (define-key map "\C-c\C-s" 'import-compile-setup)
    (define-key map [menu-bar tools setup] '("Setup" . import-compile-setup))
;    (define-key map "\C-c\C-l" 'import-link)
;    (define-key map [menu-bar tools link] '("Link" . import-link))
    (define-key map "\C-c\C-c" 'import-compile)
    (define-key map [menu-bar tools compile] '("Compile" . import-compile))
    (define-key map "\C-c\C-t" 'import-typecheck)
    (define-key map [menu-bar tools typecheck] '("Typecheck" . import-typecheck))
    ;;
    (define-key map [menu-bar methods]
      (cons "Methods" (make-sparse-keymap "Methods")))
    (define-key map "\C-cp" 'import-function)
    (define-key map [menu-bar methods function] '("FUNCTION" . import-function))
    (define-key map "\C-ct" 'import-method-tell)
    (define-key map [menu-bar methods tell] '("TELL Method" . import-method-tell))
    (define-key map "\C-cq" 'import-method-query)
    (define-key map [menu-bar methods query] '("QUERY Method" . import-method-query))
    (define-key map "\C-cd" 'import-method-destr)
    (define-key map [menu-bar methods destr] '("DESTRUCTOR Method" . import-method-destr))
    (define-key map "\C-cc" 'import-method-const)
    (define-key map [menu-bar methods const] '("CONSTRUCTOR Method" . import-method-const))
    (define-key map "\C-ca" 'import-method-ask)
    (define-key map [menu-bar methods ask] '("ASK Method" . import-method-ask))
    ;;
    (define-key map [menu-bar templates]
      (cons "Templates" (make-sparse-keymap "Templates")))
    (define-key map "\C-cw" 'import-while)
    (define-key map [menu-bar templates while] '("WHILE" . import-while))
;    (define-key map "\C-cv" 'import-wait)
;    (define-key map [menu-bar templates wait] '("WAIT" . import-wait))
    (define-key map "\C-cr" 'import-repeat)
    (define-key map [menu-bar templates repeat] '("REPEAT" . import-repeat))
    (define-key map "\C-co" 'import-object)
    (define-key map [menu-bar templates object] '("OBJECT" . import-object))
    (define-key map "\C-cm" 'import-module)
    (define-key map [menu-bar templates module] '("MODULE" . import-module))
    (define-key map "\C-cl" 'import-loop)
    (define-key map [menu-bar templates loop] '("LOOP" . import-loop))
    (define-key map "\C-ci" 'import-if)
    (define-key map [menu-bar templates if] '("IF" . import-if))
    (define-key map "\C-cf" 'import-for)
    (define-key map [menu-bar templates for] '("FOR" . import-for))
    (define-key map "\C-ck" 'import-case)
    (define-key map [menu-bar templates case] '("CASE" . import-case))
    (define-key map "\C-cb" 'import-block)
    (define-key map [menu-bar templates block] '("BLOCK" . import-block))
    ;;
    (define-key map "\C-ce" 'import-elseif)
    (define-key map "\C-ch" 'import-header)
    (define-key map "\C-c{" 'import-begin-comment)
    (define-key map "\C-c}" 'import-end-comment)
    (define-key map "\C-c\C-h" 'import-mode-help)
    (define-key map "\C-c\C-p" 'import-do-highlight)
    (define-key map "\C-c\C-z" 'suspend-emacs)
    ;;
    (setq import-mode-map map)))


;;; major mode setup
  

(defun import-mode ()
  "This is a mode intended to support program development in IMPORT.
All control templates of IMPORT syntax can be reached by typing C-c
followed by the first character of the construct. \\<import-mode-map>
Type \\[import-mode-help] for a listing of key combinations and functions."
  (interactive)
  (kill-all-local-variables)
  (use-local-map import-mode-map)
  (setq major-mode 'import-mode)
  (setq mode-name "IMPORT")
  (setq selective-display t)
  (make-local-variable 'comment-column)
  (setq comment-column 41)
  (make-local-variable 'end-comment-column)
  (setq end-comment-column 75)
  (set-syntax-table import-mode-syntax-table)
  (make-local-variable 'paragraph-start)
  (setq paragraph-start (concat "^$\\|" page-delimiter))
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate paragraph-start)
  (make-local-variable 'paragraph-ignore-fill-prefix)
  (setq paragraph-ignore-fill-prefix t)
  (make-local-variable 'require-final-newline)
  (setq require-final-newline t)
  (make-local-variable 'comment-start)
  (setq comment-start "{")
  (make-local-variable 'comment-end)
  (setq comment-end "}")
  (make-local-variable 'comment-start-skip)
  (setq comment-start-skip "\\(<--[ \t]*\\|\\\\/[ \t]*\\)")
  (make-local-variable 'comment-indent-function)
  (setq comment-indent-function 'c-comment-indent)
  (make-local-variable 'parse-sexp-ignore-comments)
  (setq parse-sexp-ignore-comments t)
  (run-hooks 'import-mode-hook))

(defun import-newline ()
  "Insert a newline and indent following line like previous line."
  (interactive)
  (let ((hpos (current-indentation)))
    (newline)
    (indent-to hpos)))

(defun import-tab ()
  "Indent to next tab stop."
  (interactive)
  (indent-to (* (1+ (/ (current-indentation) import-indent)) import-indent)))

(defun import-newline-untab (&optional n)
  "Insert a newline and indent the line (n * import-indent) less than
previous tab stop."
  (interactive)
  (let ((hpos (current-indentation)))
    (newline)
    (if n
	(indent-to (max 0 (* (- (/ hpos import-indent) n) import-indent)))
      (indent-to (max 0 (* (1- (/ hpos import-indent)) import-indent))))))

(defun import-newline-tab (&optional &rest n)

  "For each element in n, insert a newline and indent.  If (car n)=0,
then indent to the same column as the previous line.  If (car n)>0,
then indent (car n) positions more than previous line, if (car n)<0,
then indent (car n) positions less than previous line up to the
beginning of the line. n defaults to (1)"

  (interactive)
  (let ((hpos (current-indentation)))

    (if (not n)
	(setq n '(1)))
    (while n
      (newline)
      (indent-to (max 0 (* (+ (/ hpos import-indent) (car n)) import-indent)))
      (setq n (cdr n)))))


(defun import-get-previous-indent ()
  "Return the position of the first non-whitespace character on the
preceeding line.  If no non-whitespace character exists, return the
position of the end-of-line."
  (interactive)
  (save-excursion
    (if (eq (forward-line -1) -1)
	0
      (current-indentation))))

(defun import-indent-relative ()
  "Indent one more position than the previous line."
  (interactive)
  (let ((spoint (import-get-previous-indent)))
    (beginning-of-line)
    (fixup-whitespace)
    (indent-to (* (1+ (/ spoint import-indent)) import-indent))))


(defun import-correct-indent ()
  "Indent this line depending upon its type and the indentation of previous line.
We are assuming that we always begin on the beginning of the line (ie, column 0)
to be formatted."
  (interactive)
  ;; first, see if we are on the first line of the buffer;
  ;; if so, just indent to 0
  (if (eq (point) 1)
      (fixup-whitespace)
    ;; otherwise, set up the target indentation, and the modifier
    (let ((indentfact 0)
	  (target 0))
      ;; first, see if the line that we are indenting is a line that would cause an
      ;; unindent to take place; if so, record that by decrementing indentfact
      (if (looking-at "^[ \t]*\\(BEGIN\\|END\\|ELSE\\|ELSEIF\\|WHEN\\|OTHERWISE\\|THEN\\|UNTIL\\)")
	  (setq indentfact (1- indentfact)))
      ;; next, if it wasn't passed in, record the indentation of the previous line,
      ;; and check to see if that line is one that would cause an extra level of indentation
      (if (eq (forward-line -1) -1)
	  (setq target (current-indentation)) ; use current line's indent if no previous line
	(setq target (current-indentation))
	(if (looking-at "^[ \t]*\\(EXPORT\\|PUBLIC\\|PRIVATE\\)?[ \t]*\\(BEGIN\\|BLOCK\\|DOME\\|IF\\|ELSE\\|ELSEIF\\|CASE\\|WHILE\\|REPEAT\\|FOR\\|LOOP\\|MODULE\\|OBJECT\\|WHEN\\|OTHERWISE\\|[A-Za-z]+[A-Za-z0-9_]*[ \t]*[=][ \t]*OBJECT\\)")
	    (setq indentfact (1+ indentfact))
	  ;; otherwise, check to see if the previous line was a method declaration.
	  ;; remember, QUERY methods never have a body, so they never cause indentation.
	  ;; Other method declaration won't cause indentation within interface objects.
	  (if (looking-at "^[ \t]*\\(EXPORT\\|PUBLIC\\|PRIVATE\\)?[ \t]*\\(ASK[ \\t]+METHOD\\|TELL[ \\t]+METHOD\\|CONSTRUCTOR[ \\t]+METHOD\\|DESTRUCTOR[ \\t]+METHOD\\)")
	      (if (search-backward "INTERFACE" 1 t)
		  t
		(setq indentfact (1+ indentfact)))))
	(forward-line))
      ;; finally, indent appropriately
      (fixup-whitespace)
      (indent-to (* (+ indentfact (/ target import-indent)) import-indent)))))


(defun import-indent-line ()
  (interactive)
  (beginning-of-line)
  (import-correct-indent))

(defun import-indent-region (beg end)
  (interactive "r")
  (if (> beg end) (let (mid) (setq mid beg beg end end mid)))
  (save-excursion
    (save-restriction
      (narrow-to-region beg end)
      (goto-char beg)
      (while (not (eobp))
	(import-correct-indent)
	(forward-line)))))

(defun import-indent-module ()
  (interactive)
  (save-excursion
    (let ((modbeg (search-backward-regexp "^[ \t]*\\(EXPORT\\|PUBLIC\\|PRIVATE\\)?[ \t]*MODULE" 0 t))
	  (modend (search-forward-regexp "^[ \t]*END[ \t]*MODULE")))
      (if (and modbeg modend)
	  (progn
	    (message "Reformatting Module...")
	    (import-indent-region modbeg modend)
	    (message "Reformatting Module...done."))
	(error "Not within a MODULE boundary.")))))

;; ****************************************************************************************
;;
;; IMPORT/DOME STATEMENT TEMPLATES
;;
;; ****************************************************************************************

(defun import-block ()
  "Insert a BLOCK..END BLOCK template."
  (interactive)
  (insert "BLOCK")
  (import-newline)
  (insert "BEGIN")
  (import-newline-tab 1 0)
  (insert "END BLOCK;")
  (end-of-line 0))

(defun import-case (cname)
  "Build skeleton CASE statment, prompting for the <expression>."
  (interactive "sCase Expression: ")
  (let ((ccond "")
	(ccount 0))
    (insert "CASE " cname )
    (import-newline)
    (setq ccond (read-string "Case Condition: "))
    (while (not (string-equal ccond ""))
      (setq ccount (1+ ccount))
      (insert "WHEN " ccond ":")
      (import-newline-tab 1 0)
      (setq ccond (read-string "Case Condition: ")))
    (insert "OTHERWISE")
    (import-newline-tab 1 0)
    (insert "END CASE; <--  " cname)
    (end-of-line (* -2 ccount))))

(defun import-elseif ()
  "Insert ELSEIF keyword and indent for next line."
  (interactive)
  (import-newline)
  (backward-delete-char-untabify import-indent ())
  (insert "ELSEIF ")
  (insert (read-string "Condition: "))
  (import-newline-tab))

(defun import-for (importvar importinit importlimit importbystr)
  "Build skeleton FOR loop statment, prompting for the loop parameters."
  (interactive "sLoop Variable: \nsInitial Value: \nsLimit: \nsBy: ")
  (insert "FOR " importvar ":=" importinit 
	  " TO " importlimit)
  (if (not (string-equal importbystr ""))
      (insert " BY " importbystr))
  (insert " DO")
  (import-newline-tab 1 0)
  (insert "END FOR; <-- "
	  importvar " from " importinit	  " to " importlimit)
  (end-of-line 0)
  (message "Check \"TO\" keyword for appropriateness (may need \"DOWNTO\")."))

(defun import-function (fname)
  (interactive "sFunction Name: ")
  (insert "FUNCTION " fname "(")
  (let (args)
    (insert (read-string "Arguments: ") ")")
    (setq args (read-string "Result Type: "))
    (if (not (string-equal args ""))
	(insert " : " args))
    (insert ";")
    (import-newline)
    (insert "BEGIN")
    (import-newline-tab 1 0)
    (insert "END FUNCTION; <--" name)
    (end-of-line 0)))


(defun import-header (&optional importtitle)
  "Insert a comment block containing the module title, author, etc."
  (interactive "sTitle: ")
  (insert "{\n     Title: \t")
  (if importtitle
      (insert importtitle)
    (insert (read-string "Title: ")))
  (insert "\n    Created:\t")
  (insert (current-time-string))
  (insert "\n    Author: \t")
  (if (string-equal (user-full-name) "unknown")
      (insert (read-string "Enter name of module author: "))
    (insert (user-full-name)))
  (insert (concat "\n\t\t<" (user-login-name) "@" (system-name) ">\n"))
  (insert "}\n\n"))

(defun import-if (bcond)
  "Insert skeleton IF statment, prompting for <boolean-expression>."
  (interactive "s BOOLEAN Expression: ")
  (insert "IF " bcond " THEN ")
  (import-newline-tab 1 0)
  (insert "ELSE")
  (import-newline-tab 1 0)
  (insert "END IF;")
  (end-of-line -2))

(defun import-loop ()
  "Build skeleton LOOP (with END)."
  (interactive)
  (insert "LOOP")
  (import-newline-tab 1 0)
  (insert "END LOOP;")
  (end-of-line 0))

(defun import-method (importmtype importname importargs importrtype)
  "Insert an <MTYPE> METHOD keyword."
  (insert importmtype " METHOD " importname "(" importargs ")")
  (if (not (string-equal importrtype ""))
      (insert " : " importrtype))
  (import-newline)
  (insert "BEGIN")
  (import-newline-tab 1 0)
  (insert "END METHOD; <-- " (downcase importmtype) " method " importname)
  (end-of-line 0))

(defun import-method-ask (importname importargs importrtype)
  (interactive "sName: \nsArguments: \nsResult Type: ")
  (import-method "ASK" importname importargs importrtype))

(defun import-method-const ()
  (interactive)
  (let* ((importname
	  (save-excursion
	    (if (search-backward-regexp "^[ \t]*\\(EXPORT\\|PUBLIC\\|PRIVATE\\)?[ \t]*[A-Za-z]+[A-Za-z0-9_]*[ \t]*[=][ \t]*OBJECT" 0 t)
		(current-word)
	      (read-string "Method Name: "))))
	 (importargs (read-string "Arguments: ")))
    (import-method "CONSTRUCTOR" importname importargs "")))

(defun import-method-destr ()
  (interactive)
  (let* ((importname
	  (save-excursion
	    (if (search-backward-regexp "^[ \t]*\\(EXPORT\\|PUBLIC\\|PRIVATE\\)?[ \t]*[A-Za-z]+[A-Za-z0-9_]*[ \t]*[=][ \t]*OBJECT" 0 t)
		(current-word)
	      (read-string "Method Name: ")))))
    (import-method "DESTRUCTOR" importname "" "")))

(defun import-method-query (importname importargs)
  (interactive "sName: \nsArguments: ")
  (import-method "QUERY" importname importargs "BOOLEAN"))

(defun import-method-tell (importname importargs)
  (interactive "sName: \nsArguments: ")
  (import-method "TELL" importname importargs ""))

(defun import-module (mname)
  "Build skeleton MODULE, prompting for <module-name>, et all."
  (interactive "sModule Name: ")
  (if (not (string-equal mname ""))
      (progn
	(import-header)
	(import-newline)
	(insert "MODULE " mname)
	(import-newline-tab)
	(let ((imodname (read-string "Import Module Name: ")))
	  (while (not (string-equal imodname ""))
	    (insert "IMPORT " imodname ";")
	    (import-newline)
	    (setq imodname (read-string "Import Module Name: "))))
	(import-newline-tab 0 0 -1)
	(insert "END MODULE; <-- " mname)
	(end-of-line -1))))

(defun import-object (oname)
  "Build skeleton object declaration, prompting for <object-name>."
  (interactive "sObject Name: ")
  (let ((hpos (current-indentation))
	inherits)
    (insert oname " = OBJECT")
    (setq inherits (read-string "Superclasses (comma separated): "))
    (if (not (string-equal inherits ""))
	(insert " (" inherits ")"))
    (import-newline-tab 1 1 1 0)
    (insert "END OBJECT; <-- " oname)
    (end-of-line -1)))

(defun import-repeat (bcond)
  (interactive "sBoolean Expression: ")
  (insert "REPEAT")
  (import-newline-tab 1 0)
  (insert "UNTIL " bcond " END REPEAT;")
  (end-of-line -1))

;(defun import-wait ()
;  (interactive)
;  (insert "WAIT ")
;  (let ((wreason (read-string "Wait Reason: ")))
;    (insert wreason)
;    (import-newline-tab 1 0)
;    (insert "ON INTERRUPT")
;    (import-newline-tab 1 0)
;    (insert "END WAIT;")
;    (end-of-line -2)))

(defun import-while (bcond)
  (interactive "sBoolean Expression: ")
  (insert "WHILE " bcond " DO")
  (import-newline-tab 1 0)
  (insert "END WHILE; <-- " bcond)
  (end-of-line -1))

;;
;;
;; IMPORT/DOME COMMENTS
;;
;;

(defun import-begin-comment ()
  (interactive)
  (if (not (bolp))
      (indent-to comment-column 0))
  (insert "{  "))

(defun import-end-comment ()
  (interactive)
  (if (not (bolp))
      (indent-to end-comment-column))
  (insert "}"))

;;
;; INFORMATION HIDING: the following functions implement the concept
;;   of information hiding by utilizing the 'selective-display
;;   feature of emacs.  When a buffer's local selective-display is
;;   non-nil, all lines ending in a 
 are replaced by "..."
;;

(defun import-hide-region (from to)
  "Hides lines from FROM to TO"
  (let (buffer-read-only)
    (goto-char from)
    (insert "
")
    (subst-char-in-region from to ?\n ?\
 t)))

(defun import-show-region (from to)
  "Shows (unhides) lines from FROM to TO"
  (let (buffer-read-only)
    (goto-char from)
    (if (looking-at "[
]")
	(delete-char 1))
    (subst-char-in-region from to ?\
 ?\n t)))

(defvar import-private-start 
  '("^[ \t]*\\(PRIVATE\\)?[ \t]+[A-Za-z][A-Za-z0-9_]*[ \t]*[:][=]"
    "^[ \t]*\\(PRIVATE\\)?[ \t]+MODULE[ \t]+"
    "^[ \t]*\\(PRIVATE\\)?[ \t]+IMPORT[ \t]+"
    "^[ \t]*\\(PRIVATE\\)?[ \t]+OBJECT[ \t]+"
    "^[ \t]*\\(PRIVATE\\)?[ \t]+[A-Za-z][A-Za-z0-9_]*[ \t]*[=][ \t]*OBJECT"
    "^[ \t]*\\(PRIVATE\\)?[ \t]+[A-Za-z][A-Za-z0-9_]*[ \t]*[=][ \t]*"
    "^[ \t]*\\(PRIVATE\\)?[ \t]+\\(ASK\\|CONSTRUCTOR\\|DESTRUCTOR\\|TELL\\)[ \t]+METHOD[ \t]+"
    "^[ \t]*\\(PRIVATE\\)?[ \t]+QUERY[ \t]+METHOD[ \t]+"
    "^[ \t]*\\(PRIVATE\\)?[ \t]+FUNCTION[ \t]+"
    "^[ \t]*\\(PRIVATE\\)?[ \t]+[A-Za-z][A-Za-z0-9_]*[ \t]*[=][=]"
    "^[ \t]*\\(PRIVATE\\)?[ \t]+[A-Za-z][A-Za-z0-9_]*\\([ \t\n]*[,][ \t\n]*[A-Za-z][A-Za-z0-9_]*\\)*[ \t]*[:][ \t]*[^=]"
;    "^[ \t]*PRIVATE[ \t]*$"
    )
  "List of regular-expressions signaling the start of a private section.")

(defvar import-private-end 
  '(("[;]" (point))
    ("END[ \t]+MODULE[ \t]*[;]" (point))
    ("[;]" (point))
    ("END[ \t]+OBJECT[ \t]*[;]" (point))
    ("[;]" (point))
    ("END[ \t]+METHOD[ \t]*[;]" (point))
    ("BOOLEAN[ \t]*[;]" (point))
    ("END[ \t]+FUNCTION[ \t]*[;]" (point))
    ("[;]" (point))
    ("[;]" (point))
;    ("^[ \t]*EXPORT\\|PUBLIC" (progn (end-of-line 0)(point)))
    )
  "List of lists of regular-expressions signaling the end of a private section,
and a function to move from the position at the end of search to the position
of the desired end of the hidden region.")

(defun import-find-file-hook ()
  (if buffer-read-only
      (save-excursion
	;; first, hide all the PRIVATE declarations matched by the regular
	;; expressions in 'import-private-start
	(let ((ips import-private-start)
	      (ipe import-private-end)
	      cur-start
	      cur-end
	      end-fun)
	  (while ips
	    (setq cur-start (car ips)
		  cur-end (car (car ipe))
		  end-fun (car (cdr (car ipe)))
		  ips (cdr ips)
		  ipe (cdr ipe))
	    (goto-char (point-min))
	    (while (search-forward-regexp cur-start (point-max) t)
	      (let* ((cur-point (point))
		     (priv-start (search-backward-regexp cur-start)))
		(goto-char cur-point)
		(if (search-forward-regexp cur-end (point-max) t)
		    (import-hide-region priv-start (eval end-fun)))))))
	;; next, hide method bodies, unless this is an interface module
	(goto-char (point-min))	
	(if (not (search-forward "INTERFACE" nil t))
	    (while (search-forward-regexp "^[ \t]*\\(\\(EXPORT\\|PUBLIC\\|PRIVATE\\)?[ \t]+\\)?\\(ASK\\|CONSTRUCTOR\\|DESTRUCTOR\\|FUNCTION\\|TELL\\)[ \t]+METHOD" (point-max) t)
	      (let ((priv-start (search-forward-regexp "[)][ \t]*[^\n]*")))
		(search-forward-regexp "END[ \t]+\\(METHOD\\|FUNCTION\\)[ \t]*[;]")
		(import-hide-region priv-start (point)))))
	(not-modified))))

;*****************************************************
; the following line, when uncommented, implements
; visibility:
;
;(add-hook 'find-file-hooks 'import-find-file-hook t)
;*****************************************************		    

;;
;;
;; IMPORT/DOME INTERFACE TO COMPILER TOOLS
;;
;;

;; assume that each module is contained in a separate file.  

(defun import-compile-setup ()
  "Display and allow the user to change values that effect
the compilation of IMPORT programs."
  (interactive)
  (setq import-typecheck-command (read-string "Command for typechecking IMPORT programs: " "import -t"))
  (setq import-compile-command (read-string "Command for compiling IMPORT programs: " "import"))
  (setq import-link-name (read-string "Name of executable to produce: " ""))
  )

(defun import-chase-import (filename)
  (let ((impfiles "")
	(cbuf (current-buffer)))
    (if (file-exists-p filename)
	(progn
	  (set-buffer (find-file-noselect filename))
	  (goto-char (point-min))
	  (while (search-forward-regexp "^[ \t]*\\(EXPORT\\|PUBLIC\\|PRIVATE\\)?[ \t]*IMPORT[ \t]+" nil t)
	    (let ((imported-file (concat (current-word) ".id")))
	      (if (not (string-match imported-file impfiles))
		  (setq impfiles (concat impfiles (import-chase-import imported-file))))))
	  (if (not (string-match filename impfiles))
	      (setq impfiles (concat impfiles " " filename)))
	  (set-buffer cbuf)))
    impfiles))


(defun import-compile ()
  "Invoke the IMPORT compiler on the file in the current buffer,
using the MODULE/OBJECT under the cursor as the object to invoke
at start-up."
  (interactive)
  (save-excursion
    (save-some-buffers nil nil)
    (let* ((bufname (buffer-name))
	   (mypoint (point))
	   (objname (if (re-search-backward "[A-Za-z][A-Za-z0-9_]*[ \t]+\=[ \t]+OBJECT" 1 t)
			(current-word)
		      nil))
	   (modname (if (re-search-backward "MODULE[ \t]+[A-Za-z][A-Za-z0-9_]*" 1 t)
			(progn
			  (forward-word 1)
			  (forward-char 1)
			  (current-word))
		      nil))
	   (outfile (concat (substring bufname 0 -3) ".cxx"))
	   (cresult nil))
      (if (and modname objname)
	  (progn
	    (setq import-last-compiled-buffer bufname)
	    (if (file-exists-p outfile)
		(if (yes-or-no-p (concat outfile " already exists! Overwrite? "))
		    (delete-file outfile)
		  (keyboard-quit)))
	    ;; now, we want to handle the case of importation.  We will assume that
	    ;; each module has its own file, and that the filename matches the
	    ;; module name with an ".id" suffix added.
	    (compile
	      (concat import-compile-command ;; JBco: " -s\"DISPOSE NEW " 
		      ;; JBco: modname "." objname ";\"" 
		      " -o " (if (not (string-equal import-link-name ""))
				 import-link-name
			       modname)
		      (import-chase-import bufname))))
	(error "Not Pointing at a MODULE.OBJECT declaration.")))))

(defun import-typecheck ()
  "Invoke the IMPORT parser/typechecker on the file in the current buffer."
  (interactive)
  (save-excursion
    (save-some-buffers nil nil)
    (let* ((bufname (buffer-name)))
      (setq import-last-compiled-buffer (buffer-name))
      ;; now, we want to handle the case of importation.  We will assume that
      ;; each module has its own file, and that the filename matches the
      ;; module name with an ".id" suffix added.
      (compile
       (concat import-typecheck-command (import-chase-import bufname))))))

;(defun import-link ()
;  "Compile the generated C++ and link it into an executable."
;  (interactive)
;  (save-excursion
;    (save-some-buffers nil nil)
;    (let*
;	((bufname (buffer-name))
;	 (srcname (concat (substring bufname 0 -3) ".cxx"))
;	 (outname (if (not (string-equal "" import-link-name))
;		      import-link-name
;		    (substring bufname 0 -3))))
;      (if (file-exists-p outname)
;	  (if (yes-or-no-p (concat outname " already exists! Overwrite? "))
;	      (compile (concat 
;			import-link-command " " srcname " -o " outname 
;			" -I" import-cpp-include-dirs " "
;			import-link-lib-paths " " 
;			import-link-cpp-libraries)))
;	(compile (concat 
;		  import-link-command " " srcname " -o " outname 
;		  " -I" import-cpp-include-dirs " "
;		  import-link-lib-paths " " 
;		  import-link-cpp-libraries))))))
;    

(defun execute-monitor-command (command)
  (let* ((shell shell-file-name)
	 (csh (equal (file-name-nondirectory shell) "csh")))
    (call-process shell nil t t "-cf" (concat "exec " command))))

(defun import-next-error-finish (markerpair)
  (if (or (not (markerp (car markerpair)))
	  (not (markerp (cdr markerpair))))
      nil
    (let ((error-buf (marker-buffer (car markerpair)))
	  (source-buf (marker-buffer (cdr markerpair)))
	  (offset 0)
	  (savebuf (current-buffer)))
      (save-excursion
	(set-buffer error-buf)
	(let ((eolpos (progn (end-of-line)(point))))
	  (beginning-of-line)
	  (if (search-forward-regexp ":[0-9]+\\.[0-9]+" eolpos t)
	      (setq offset (string-to-number (current-word))))))
      (set-buffer source-buf)
      (goto-char (1- (+ (point) offset)))
      (set-buffer savebuf))))

(setq next-error-finish-function (function import-next-error-finish))

(defun import-do-highlight ()
  "Load Highlighting for IMPORT Buffers."
  (interactive)
  (require 'hilit19)
  (cond (window-system
	 (set-cursor-color "red")
	 (scroll-bar-mode t)
	 (menu-bar-mode t)
	 (hilit-set-mode-patterns
	  'import-mode
	  '(
	    ;; comments
	    ("\\{" "\\}" comment)
	    ("\\\\/.*$" nil comment)
	    ("<--.*$" nil comment)
	    ;; strings
	    ("\"" "[^\\]\"" string)
	    ;; keywords
	    ("\\<\\(ABOUT\\|AND\\|AM\\|ANYOBJ\\|ARRAY\\|AT\\|ASK\\|BEGIN\\|\
BLOCK\\|BY\\|CASE\\|CONST\\|CONSTRUCTOR\\|DESTRUCTOR\\|\
DISPOSE\\|DO\\|DOME\\|DOWNTO\\|DONE\\|DURATION\\|ELSE\\|ELSEIF\\|END\\|EXIT\\|\
FALSE\\|FOR\\|FROM\\|FUNCTION\\|IF\\|IMPORT\\|IN\\|INIT\\|INOUT\\|\
INTERFACE\\|INTERRUPT\\|ISA\\|IS\\|LOOP\\|METHOD\\|MOD\\|\
MODULE\\|NEW\\|NOT\\|NULL\\|OF\\|OBJECT\\|OR\\|OTHERWISE\\|OUT\\|PM\\|\
PRIVATE\\|EXPORT\\|PUBLIC\\|QUERY\\|REPEAT\\|RETURN\\|SUBCLASS\\|SELF\\|\
TELL\\|THEN\\|TO\\|TRUE\\|TYPE\\|UNTIL\\|VAR\\|VIEWED\\|WAIT\\|\
WHEN\\|WHILE\\)\\>" nil defun)
	    ;; builtin procedures
	    ("\\<\\(ABS\\|CAP\\|CHARTOSTR\\|CHR\\|DEC\\|DISPOSE\\|\
DISPOSEOBJ\\|FLOAT\\|HALT\\|HIGH\\|ID\\|INC\\|INPUT\\|INSERT\\|INTTOSTR\\|\
LOW\\|LOWER\\|MAX\\|MAXOF\\|MIN\\|MINOF\\|NEW\\|NEWOBJ\\|ODD\\|ORD\\|\
OUTPUT\\|POSITION\\|REALTOSTR\\|REPLACE\\|ROUND\\|SCHAR\\|SIZEOF\\|\
STRLEN\\|STRTOCHAR\\|STRTOINT\\|STRTOREAL\\|SUBSTR\\|TIME\\|UPPER\\|TRUNC\\|\
VAL\\|out\\)\\>"  nil decl)
	    ;; standard types
	    ("\\<\\(BOOLEAN\\|CHAR\\|INTEGER\\|LIST\\|REAL\\|SEMAPHORE\\|\
STRING\\|TAGGED\\)\\>" nil crossref)
	    ;; module and object declarations
	    ("\\(MODULE[ \t]+[A-Za-z]+[A-Za-z0-9_]*\\|(OBJECT[ \t]+[A-Za-z]+[A-Za-z0-9_]*\\|[A-Za-z]+[A-Za-z0-9_]*[ \t]*[=][ \t]*OBJECT\\)"
	     nil struct))))))

;;; IMPORT.el ends here

(message "I *really* finished!")
