/*****************************************************************************/ 
/*                               CodeGen.h                                   */
/*****************************************************************************/

/* AUTHOR: Vance Morrison
     DATE: 6/1/94		*/
/*****************************************************************************/

#ifndef _CodeGen_h
#define _CodeGen_h

//#include "Depend.H"

#include "SpIostream.H"
#include "SpString.H"
#include "CharStream.H"

// JB: for indenting
class PiIndent
{
 public:
   PiIndent(): IndentLevel(0) {}
   PiIndent(int indentLevel): IndentLevel(indentLevel) {}
   void CopyOverIndentBuffer(char *Buffer)
      {
	 char indentBuff[64];
	 for (int i=0; i<IndentLevel*IndentSpacing; i++) indentBuff[i] = ' ';
	 indentBuff[IndentLevel*IndentSpacing] = '\0';
	 strcpy(Buffer, indentBuff);
      }
   PiIndent operator++ () { IndentLevel++; return *this; } 
   PiIndent operator++ /*postfix*/ (int)
      {
	 PiIndent indent = *this;
	 IndentLevel++;
	 return indent;
      }
   PiIndent operator-- () { IndentLevel--; return *this; } 
   PiIndent operator-- /*postfix*/ (int)
      {
	 PiIndent indent = *this;
	 IndentLevel--;
	 return indent;
      }
   PiIndent SendIncrement()
      {
	 PiIndent indent = *this;
	 indent.IndentLevel++;
	 return indent;
      }
   PiIndent SendDecrement()
      {
	 PiIndent indent = *this;
	 indent.IndentLevel--;
	 return indent;
      }
 private:
   int IndentLevel;
   const static int IndentSpacing;
};

ostream& operator<< (ostream &out, PiIndent &indent);

class PiList;

class Decl;
class Type;
class Type;
class TDecl;
class VDecl;
class Exp;
class Stmt;
class Term;
class Method;

class ModDecl;
class ImportDecl;
class EnumDecl;
class ObjectDecl;
class TypeDecl;
class ConstDecl;
class VarDecl;
class FunDecl;
class FunArgDecl;

class DeferedBody;
class MethodBody;

class IdentType;
class FunType;
class ArrayType;
class AtomicType;

class AsExp;
class IsAExp;
class CastExp;
class SelfExp;
class IdentExp;
class ArrayExp;
class ApplyExp;
class NullExp;
class IntExp;
class RealExp;
class StringExp;
class CharExp;
class BoolExp;
class VarExp;
class NewExp;

class AssignStmt;
class IfStmt;
class WhileStmt;
class RepeatStmt;
class ForStmt;
class LoopStmt;
class ReturnStmt;
class ExitStmt;
class ExpStmt;
class DisposeStmt;
class BlockStmt;
class CaseStmt;

class FunctorTerm;
class VarTerm;
class ConstTerm;
class ImportTerm;
class ListTerm;

enum FileType_t { CFILE, _HFILE };

enum CodeGen_t { NORMAL, SINGLE_FILE, MAIN };

enum cond_t
{
   FIRST_COLLECT,
   INTERIM_COLLECT,
   PRINT_DEFINES,
   PRINT_SCHEDULE_DECLARES,
   PRINT_MAIN
};

enum qualifier_t
{
   BY_POINTER,
   BY_REFERENCE,
   BY_VALUE
};

struct objMgrCont_t
{
   objMgrCont_t(const char *name, int totSimObj): TotSimObj(totSimObj)
      { Name = strdup(name); }
   char *Name;
   int TotSimObj;
};

class AssocInstDecl;
class Clause;
class AsyncStmt;

/*****************************************************************************/
class CodeGen {
public:
    CodeGen() {}
   void genCode(const PiList* decls, const PiList* stdDecls, ostream& out,
		enum FileType_t fileType, const Stmt* startStmt = NULL,
		enum CodeGen_t type = NORMAL);

private:
   void CollectObjMgrAndEventNamesAndPrint(const PiList* decls, ostream& out,
					   enum cond_t cond);
   void CollectDefines(const PiList* decls, ostream& out);
    void genDeclObj(const Decl* decl, ostream& out, PiIndent indent);
    void genDeclObjs(const PiList* decls, ostream& out, PiIndent indent);

    void genDecl(const Decl* decl, ostream& out, PiIndent indent,
		 Bool overRide = FALSE, Bool rollbackable = FALSE);
    void genDeclDef(const Decl* decl, ostream& out, PiIndent indent);

    void genDecls(const PiList* decls, ostream& out, PiIndent indent,
		  int skipSpaces = 0, Bool overRide = FALSE);
    void genDeclDefs(const PiList* decls, ostream& out, PiIndent indent);
    void genDefines(const PiList* decls, ostream& out);
    void genScheduleDeclares(const PiList* decls, ostream& out);
    void genMain(const PiList* decls, ostream& out);

    const char* canonname(const Decl* decl, char* buff, int len);
    void canonName(const Decl* decl, ostream& out);

    void genType(const Type* type, const char* cName, ostream& out,
		 PiIndent indent, Bool rollbackable = FALSE,
		 qualifier_t qualifier = BY_VALUE);
    void genAtomicType(const AtomicType* type, const char* cName, 
		       ostream& out, PiIndent indent, int rollbackable = 0);
    void genFunType(const FunType* type, const char* cName,
		    ostream& out, PiIndent indent);

    void genName(const AtomicType* type, ostream& out, PiIndent indent);

    void genTypeTag(const Type* type, const Exp* exp, ostream& out,
		    PiIndent indent);
    void genTypeUnTag(const Type* type, const Exp* exp, ostream& out,
		      PiIndent indent);
    void genTypeCheck(const Type* type, const Exp* exp, ostream& out,
		      PiIndent indent);

    void genAsyncDecl(const FunDecl* fun, ostream& out, PiIndent indent);

    void genFunDeclDef(const FunDecl* decl, ostream& out, PiIndent indent);
    void genObjectDeclDef(const ObjectDecl* decl, ostream& out, PiIndent indent);
    void genEnumDeclDef(const EnumDecl* decl, ostream& out, PiIndent indent);
    void genProverCall(const FunType* fun, const char* name, ostream& out,
		       PiIndent indent);
    void genDomeVar(const Type* type, const char* name, 
		    const Exp* exp, ostream& out, PiIndent indent);

    void genInterfaceDef(const Decl* decl, ostream& out, PiIndent indent);
    void genAsyncDef(const FunDecl* fun, ostream& out, PiIndent indent);

    void genModDecl(const ModDecl* decl, ostream& out, PiIndent indent);
		    
    void genEnumDecl(const EnumDecl* decl, ostream& out, PiIndent indent);
    void genObjectDecl(const ObjectDecl* decl, ostream& out, PiIndent indent);
		       
    void genTypeDecl(const TypeDecl* decl, ostream& out, PiIndent indent);
    void genConstDecl(const ConstDecl* decl, ostream& out, PiIndent indent);
    void genVarDecl(const VarDecl* decl, ostream& out, PiIndent indent,
		    Bool rollbackable = FALSE);
    void genFunDecl(const FunDecl* decl, ostream& out, PiIndent indent,
		    Bool overRide = FALSE);
    void genFunArgDecl(const FunArgDecl* decl, ostream& out, PiIndent indent);
    void genAssocInstDecl(const AssocInstDecl* decl, ostream& out,
			  PiIndent indent);

    void genMethodDef(const Method* method, const char* funHead, ostream& out,
		      PiIndent indent);
    void genClause(const Clause* clause, int& curName, ostream& out,
		   PiIndent indent);
    void genTerm(const Term* term, int& curName, ostream& out, PiIndent indent);
    void genListTerm(const ListTerm* term, int& curName, ostream& out,
		     PiIndent indent);
    void genTerms(const PiList* terms, int& curName, ostream& out,
		  PiIndent indent);

    void genAuxClause(const Clause* clause, int& curName, ostream& out,
		      PiIndent indent);
    void genAuxTerm(const Term* term, int& curName, ostream& out,
		    PiIndent indent);
    void genImportAuxTerm(const ImportTerm* term, int& curName, ostream& out,
			  PiIndent indent);
    void genAuxTerms(const PiList* terms, int& curName, ostream& out,
		     PiIndent indent);

    void genExp(const Exp* exp, const Type* expectType, ostream& out,
		PiIndent indent, qualifier_t qualifier = BY_POINTER); 
    void genToCExp(const Exp* exp, const FunArgDecl* argDecl, ostream& out,
		   PiIndent indent); 
    void genExps(const PiList* exps, const PiList* expectTypes, ostream& out, 
		 const char* sep, PiIndent indent,
		 qualifier_t qualifier = BY_POINTER); 
    void genToCExps(const PiList* exps, const PiList* expectTypes, ostream& out, 
		    const char* sep, PiIndent indent); 
    void genIsAExp(const IsAExp* exp, ostream& out, PiIndent indent); 
    void genCastExp(const CastExp* exp, ostream& out, PiIndent indent); 
    void genApplyExp(const ApplyExp* exp, ostream& out, PiIndent indent); 
    void genApply_Exp(const ApplyExp* exp, ostream& out, PiIndent indent);
		      
    void genNewExp(const NewExp* stmt, ostream& out, PiIndent indent); 

    void genStmt(const Stmt* stmt, ostream& out, PiIndent indent); 
    void genStmts(const PiList* stmts, ostream& out, PiIndent indent); 
    void genAssignStmt(const AssignStmt* stmt, ostream& out, PiIndent indent);
		       
    void genIfStmt(const IfStmt* stmt, ostream& out, PiIndent indent); 
    void genWhileStmt(const WhileStmt* stmt, ostream& out, PiIndent indent);
		      
    void genRepeatStmt(const RepeatStmt* stmt, ostream& out, PiIndent indent);
		       
    void genForStmt(const ForStmt* stmt, ostream& out, PiIndent indent); 
    void genLoopStmt(const LoopStmt* stmt, ostream& out, PiIndent indent);
		     
    void genReturnStmt(const ReturnStmt* stmt, ostream& out, PiIndent indent);
		       
    void genExitStmt(const ExitStmt* stmt, ostream& out, PiIndent indent);
		     
    void genDisposeStmt(const DisposeStmt* stmt, ostream& out, PiIndent indent);
			
    void genBlockStmt(const BlockStmt* stmt, ostream& out, PiIndent indent);
		      
    void genCaseStmt(const CaseStmt* stmt, ostream& out, PiIndent indent);
		     
    void genAsyncStmt(const AsyncStmt* stmt, ostream& out, PiIndent indent);

private:
   // JBco: DeclSet* deps;
};


#endif
