/**************************************************************************/
/*                         Format.h    Format.h                           */
/**************************************************************************/

/* Format abstraction for a tree structure 

	FormatNode	- A node to be printed 

	Formatter	- The pretty printer.  Responsible for converting
			  an AstNode into a list of strings and AstNodes
			  (its children).   Thus AstNodes need not be 
			  unparsed all the way 
*/

/* AUTHOR: Vance Morrison
   DATE  : 9/19/94                  */
/**************************************************************************/

#ifndef _Format_h
#define _Format_h

#ifndef Bool
#define Bool int
#define TRUE 1
#define FALSE 0
#endif 

/**************************************************************************/
class Formatter;
class FormatNode {
public:
	// print the node to the Formatter.  Note that only one level
	// of the heirarchy should be 'printed'.  subNodes should not
	// be expanded but passed to the formater using the 'subObj' function
    virtual void format(Formatter& out) const=0;

	// may not actually return full (deep) copy, but it will insure that
	// the returned FormatNode is 'owned' by the caller (that
	// is the caller must call `delete' on what is returned.
    virtual FormatNode* copy()=0;
    virtual ~FormatNode() {}
};

/**************************************************************************/
class Formatter {
public:
    virtual ~Formatter() {}
	// these two functions indicate the prefered indentation pattern
	// for the code.  
    virtual void shiftIn()=0;
    virtual void shiftOut()=0;

	// it is useful for formatting purposes to know what items are simply
	// a list (for example so you can group them and perform formatting
	// operations on the group as a whole)
    virtual void listBegin() {};
    virtual void listEnd() {};

	// When the 'canBreak' lines can be brocken at any space.
    virtual Bool canBreak(Bool) { return(1); };

	// used to print strings
    virtual Formatter& operator<<(const char *str)=0;
    Formatter& operator<<(char c);
    Formatter& operator<<(int i);
    Formatter& operator<<(double d);

	// used to 'print' sub objects of a node 
	// ownership of `obj' does NOT pass to the callee, the
	// callee should use `copy' if ownership should persist
    virtual void subObj(const FormatNode& obj)=0; 
};

#endif

