/*****************************************************************************/ 
/*                             IDAST.c                               */
/*****************************************************************************/

/* import.h contains the description of the abstract syntax tree for the
   imort language */

/* AUTHOR: Vance Morrison
   DATE:  10/20/93		*/
/*****************************************************************************/

#include "IDAST.H"
#include "SpIostream.H"
#include "TypeEnv.H"
#include "NewWhere.H"

#define OPT_COPY(x, where) ((x==0)?0:x->copy(where))

static const AtomicType aObjectType(Type::objectType);
static const AtomicType aEnumType(Type::enumType);
static const AtomicType aModuleType(Type::moduleType);
static const AtomicType aSetType(Type::setType);
static const AtomicType aAframeType(Type::aframeType);
static const AtomicType aAssocType(Type::assocType);
static const AtomicType aAssocInstType(Type::assocInstType);
static const AtomicType aBoolType(Type::boolType);
static const AtomicType aUnknownObjType(Type::unknownObjType);
static const AtomicType aIntType(Type::intType);
static const AtomicType aRealType(Type::realType);
static const AtomicType aStringType(Type::stringType);
static const AtomicType aCharType(Type::charType);
static const AtomicType aPiListType(Type::listType);
static const AtomicType aTaggedType(Type::taggedType);

/*****************************************************************************/
const PiList* Decl::isInterface() const {

    const Decl* env = envDecl;
    while (env != 0) {
        const ModDecl* asMod = env->toModDecl();
	if (asMod != 0 && asMod->getInterface() != 0)
	    return(asMod->getInterface());
        env = env->envDecl;
        }
    return(0);
}

/*****************************************************************************/
const Type* ImportDecl::getType() const {
    return(0);	// should never access this information
}

/*****************************************************************************/
const Type* AFrameDecl::getType() const {
    return(&aAframeType);
}

/*****************************************************************************/
const Type* AssocInstDecl::getType() const {
    return(&aAssocInstType);
}

/*****************************************************************************/
const Type* AssocDecl::getType() const {
    return(&aAssocType);
}

/*****************************************************************************/
const Type* ModDecl::getType() const {
    return(&aModuleType);
}

/*****************************************************************************/
const Type* AssocArgDecl::getType() const {
    return(&aObjectType);
}

/*****************************************************************************/
const Type* TypeDecl::getType() const {
    return(&aSetType);
}

/*****************************************************************************/
const Type* EnumDecl::getType() const {
    return(&aEnumType);
}

/*****************************************************************************/
const Type* ObjectDecl::getType() const {
    return(&aObjectType);
}

/*****************************************************************************/
const Type* IsAExp::getType() const {
    return(&aBoolType);
}

/*****************************************************************************/
const Type* NullExp::getType() const {
    return(&aUnknownObjType);
}

/*****************************************************************************/
const Type* ApplyExp::getType() const {
    return(decl->getType()->toFunType()->getRetType());
}

/*****************************************************************************/
const Type* IntExp::getType() const {
    return(&aIntType);
}

/*****************************************************************************/
const Type* RealExp::getType() const {
    return(&aRealType);
}

/*****************************************************************************/
const Type* StringExp::getType() const {
    return(&aStringType);
}

/*****************************************************************************/
const Type* CharExp::getType() const {
    return(&aCharType);
}

/*****************************************************************************/
const Type* BoolExp::getType() const {
    return(&aBoolType);
}

/*****************************************************************************/
const Type* VarExp::getType() const {
    return(&aTaggedType);
}

/*****************************************************************************/
char* str(const char* s, const void* where) {

    char* newStr=new(where, "char", strlen(s)+1) char[strlen(s)+1];
    strcpy(newStr, s);
    return(newStr);
}

/*****************************************************************************/
PiList* copyStrs(const PiList* elems, const void* where) {
    PiList* ret = 0;
    while (elems != 0) {
        char* elem = (char*) elems->first();
	ret = new(where, "PiList") PiList(str(elem, where), ret);
        elems = elems->rest();
        }
    return(reverse(ret));
}

/*****************************************************************************/
PiList* copyStrPiList(const PiList* elems, const void* where) {
    PiList* ret = 0;
    while (elems != 0) {
        PiList* elem = (PiList*) elems->first();
	ret = new(where, "PiList") PiList(copyStrs(elem, where), ret);
        elems = elems->rest();
        }
    return(reverse(ret));
}

/*****************************************************************************/
PiList* copyDecls(const PiList* elems, const void* where) {
    PiList* ret = 0;
    while (elems != 0) {
        Decl* elem = (Decl*) elems->first();
	ret = new(where, "PiList") PiList(elem->copy(where), ret);
        elems = elems->rest();
        }
    return(reverse(ret));
}

/*****************************************************************************/
PiList* copyTerms(const PiList* elems, const void* where) {
    PiList* ret = 0;
    while (elems != 0) {
        Term* elem = (Term*) elems->first();
	ret = new(where, "PiList") PiList(elem->copy(where), ret);
        elems = elems->rest();
        }
    return(reverse(ret));
}

/*****************************************************************************/
PiList* copyClauses(const PiList* elems, const void* where) {
    PiList* ret = 0;
    while (elems != 0) {
        Clause* elem = (Clause*) elems->first();
	ret = new(where, "PiList") PiList(elem->copy(where), ret);
        elems = elems->rest();
        }
    return(reverse(ret));
}

/*****************************************************************************/
PiList* copyStmts(const PiList* elems, const void* where) {
    PiList* ret = 0;
    while (elems != 0) {
        Stmt* elem = (Stmt*) elems->first();
	ret = new(where, "PiList") PiList(elem->copy(where), ret);
        elems = elems->rest();
        }
    return(reverse(ret));
}

/*****************************************************************************/
PiList* copyExps(const PiList* elems, const void* where) {
    PiList* ret = 0;
    while (elems != 0) {
        Exp* elem = (Exp*) elems->first();
	ret = new(where, "PiList") PiList(elem->copy(where), ret);
        elems = elems->rest();
        }
    return(reverse(ret));
}

/*****************************************************************************/
PiList* copyTypes(const PiList* elems, const void* where) {
    PiList* ret = 0;
    while (elems != 0) {
        Type* elem = (Type*) elems->first();
	ret = new(where, "PiList") PiList(elem->copy(where), ret);
        elems = elems->rest();
        }
    return(reverse(ret));
}

/*****************************************************************************/
PiList* copyIdents(const PiList* elems, const void* where) {
    PiList* ret = 0;
    while (elems != 0) {
        Ident* elem = (Ident*) elems->first();
	ret = new(where, "PiList") PiList(elem->copy(where), ret);
        elems = elems->rest();
        }
    return(reverse(ret));
}

/*****************************************************************************/
PiList* copyWhens(const PiList* elems, const void* where) {
    PiList* ret = 0;
    while (elems != 0) {
        When* elem = (When*) elems->first();
	ret = new(where, "PiList") PiList(elem->copy(where), ret);
        elems = elems->rest();
        }
    return(reverse(ret));
}

/*****************************************************************************/
Type* AtomicType::copy(const void* where) const {

    return(new(where, "AtomicType") AtomicType(kind));
}

/*****************************************************************************/
Type* FunType::copy(const void* where) const {

    return(new(where, "FunType") 
	FunType(kind, copyDecls(args, where), retType->copy(where),
	OPT_COPY(objType, where)));
}

/*****************************************************************************/
Ident* Ident::copy(const void* where) const {

    return(new(where, "Ident") 
	Ident(str(ident, where), OPT_COPY(mod, where)));
}

/*****************************************************************************/
Type* IdentType::copy(const void* where) const {

    return(new(where, "IdentType") IdentType(ident->copy(where)));
}

/*****************************************************************************/
Type* ArrayType::copy(const void* where) const {
    return(new(where, "ArrayType") 
	ArrayType(elemType->copy(where), OPT_COPY(bound, where)));
}

/*****************************************************************************/
Type* FutureType::copy(const void* where) const {

    return(new(where, "FutureType") FutureType(subType->copy(where)));
}

/*****************************************************************************/
Type* SynType::copy(const void* where) const {

    return(new(where, "SynType") 
	SynType(str(name, where), typeRef->copy(where)));
}

/*****************************************************************************/
Type* EnumType::copy(const void* where) const {

    return(new(where, "EnumType") 
	EnumType(str(name, where), copyDecls(enums, where)));
}

/*****************************************************************************/
Type* ObjectType::copy(const void* where) const {

    return(new(where, "ObjectType") 
	ObjectType(str(name, where), copyIdents(supers, where), 
	copyDecls(decls, where), copyClauses(clauses, where)));
}

/*****************************************************************************/
Decl* CommentDecl::copy(const void* where) const {

    return(new(where, "CommentDecl") 
	CommentDecl(kind, str(msg, where), OPT_COPY(decl, where)));
}

/*****************************************************************************/
Decl* ModDecl::copy(const void* where) const {

    return(new(where, "ModDecl") 
	ModDecl(str(name, where), copyDecls(decls, where), 
	copyStrPiList(interface, where), vis));
}

/*****************************************************************************/
Decl* FlashBackDecl::copy(const void* where) const {

    return(new(where, "FlashBackDecl") 
	FlashBackDecl(decl->copy(where), time->copy(where), 
	(work == 0 ? 0 : str(work, where))));
}

/*****************************************************************************/
Decl* RenameDecl::copy(const void* where) const {

    return(new(where, "RenameDecl") 
	RenameDecl(str(name,where), ident->copy(where), OPT_COPY(type, where), vis));
}

/*****************************************************************************/
Decl* ImportDecl::copy(const void* where) const {

    return(new(where, "ImportDecl") ImportDecl(ident->copy(where), vis));
}

/*****************************************************************************/
Decl* ConstDecl::copy(const void* where) const {

    return(new(where, "ConstDecl") 
	ConstDecl(str(name, where), value->copy(where), vis));
}

/*****************************************************************************/
Decl* TypeDecl::copy(const void* where) const {

    return(new(where, "TypeDecl") 
	TypeDecl(str(name, where), typeRef->copy(where), vis));
}

/*****************************************************************************/
Decl* VarDecl::copy(const void* where) const {

    return(new(where, "VarDecl") 
	VarDecl(str(name, where), type->copy(where), vis));
}

/*****************************************************************************/
Decl* FunDecl::copy(const void* where) const {

    return(new(where, "FunDecl") 
	FunDecl(str(name, where), (FunType*) type->copy(where)->toFunType(), 
	vis, OPT_COPY(body, where)));
}

/*****************************************************************************/
Decl* EnumValDecl::copy(const void* where) const {

    return(new(where, "EnumValDecl") EnumValDecl(str(name, where)));
}

/*****************************************************************************/
Decl* EnumDecl::copy(const void* where) const {

    return(new(where, "EnumDecl") 
        EnumDecl(str(name, where), 
	(EnumType*) typeRef->copy(where)->toEnumType(), vis));
}

/*****************************************************************************/
Decl* ObjectDecl::copy(const void* where) const {

    return(new(where, "ObjectDecl") 
        ObjectDecl(str(name, where), 
        (ObjectType*) typeRef->copy(where)->toObjectType(), vis));
}


/*****************************************************************************/
Decl* FunArgDecl::copy(const void* where) const {

    return(new(where, "FunArgDecl") 
	FunArgDecl(kind, str(name, where), type->copy(where)));
}

/*****************************************************************************/
Decl* AFrameDecl::copy(const void* where) const {

    return(new(where, "AFrameDecl") 
	AFrameDecl(str(name, where), copyDecls(decls, where), vis));
}

/*****************************************************************************/
Decl* AssocInstDecl::copy(const void* where) const {

    return(new(where, "AssocInstDecl") 
	AssocInstDecl(str(name, where), ident->copy(where),
	copyTypes(args, where), vis));
}

/*****************************************************************************/
Decl* AssocDecl::copy(const void* where) const {

    return(new(where, "AssocDecl") 
	AssocDecl(str(name, where), copyDecls(args, where),
	copyIdents(supers, where), copyDecls(reqs, where),
	copyClauses(specs, where), copyDecls(roles, where), vis));
}

/*****************************************************************************/
RangeSpec* RangeSpec::copy(const void* where) const {

    return(new(where, "RangeSpec") 
	RangeSpec(new int(*lValue), new int(*uValue)));
}

/*****************************************************************************/
Decl* AssocArgDecl::copy(const void* where) const {

    return(new(where, "AssocArgDecl") 
	AssocArgDecl(str(name, where), type->copy(where), 
	OPT_COPY(range, where)));
}

/*****************************************************************************/
Exp* NewExp::copy(const void* where) const {

    return(new(where, "NewExp") 
	NewExp(copyExps(newArgs, where), type->copy(where), 
	copyExps(args, where)));
}

/*****************************************************************************/
Exp* AsExp::copy(const void* where) const {

    return(new(where, "AsExp") AsExp(body->copy(where), type->copy(where)));
}

/*****************************************************************************/
Exp* IsAExp::copy(const void* where) const {

    return(new(where, "IsAExp") IsAExp(body->copy(where),subType->copy(where)));
}

/*****************************************************************************/
Exp* CastExp::copy(const void* where) const {

    return(new(where, "CastExp") CastExp(body->copy(where), type->copy(where)));
}

/*****************************************************************************/
Exp* SelfExp::copy(const void* where) const {

    return(new(where, "SelfExp") SelfExp());
}

/*****************************************************************************/
Exp* IdentExp::copy(const void* where) const {

    return(new(where, "IdentExp") IdentExp(ident->copy(where)));
}

/*****************************************************************************/
Exp* ArrayExp::copy(const void* where) const {

    return(new(where, "ArrayExp") 
	ArrayExp(array->copy(where), index->copy(where)));
}

/*****************************************************************************/
Exp* ApplyExp::copy(const void* where) const {

    return(new(where, "ApplyExp") 
	ApplyExp(kind, OPT_COPY(obj, where), name->copy(where),
	copyExps(args, where), OPT_COPY(from, where)));
}

/*****************************************************************************/
Exp* NullExp::copy(const void* where) const {

    return(new(where, "NullExp") NullExp());
}

/*****************************************************************************/
Exp* IntExp::copy(const void* where) const {

    return(new(where, "IntExp") IntExp(new int(*value)));
}

/*****************************************************************************/
Exp* RealExp::copy(const void* where) const {

	
    return(new(where, "RealExp") RealExp(new double(*value)));
}

/*****************************************************************************/
Exp* StringExp::copy(const void* where) const {

    return(new(where, "StringExp") StringExp(str(value, where)));
}

/*****************************************************************************/
Exp* CharExp::copy(const void* where) const {

	
    return(new(where, "CharExp") CharExp(new char(*value)));
}

/*****************************************************************************/
Exp* BoolExp::copy(const void* where) const {

    return(new(where, "BoolExp") BoolExp(val));
}

/*****************************************************************************/
Exp* VarExp::copy(const void* where) const {

    return(new(where, "VarExp") VarExp(str(var, where)));
}

/*****************************************************************************/
Stmt* CommentStmt::copy(const void* where) const {

    return(new(where, "CommentStmt") 
	CommentStmt(kind, str(msg, where), OPT_COPY(stmt, where)));
}

/*****************************************************************************/
Stmt* AsyncStmt::copy(const void* where) const {

    return(new(where, "AsyncStmt") 
	AsyncStmt(lhs->copy(where), 
	(ApplyExp*) apply->copy(where)->toApplyExp()));
}

/*****************************************************************************/
Stmt* AssignStmt::copy(const void* where) const { 

    return(new(where, "AssignStmt") 
	AssignStmt(lhs->copy(where), rhs->copy(where)));
}

/*****************************************************************************/
Stmt* IfStmt::copy(const void* where) const {

    return(new(where, "IfStmt") 
	IfStmt(cond->copy(where), copyStmts(then, where),
	copyStmts(else_, where)));
}

/*****************************************************************************/
Stmt* WhileStmt::copy(const void* where) const {

    return(new(where, "WhileStmt") 
	WhileStmt(cond->copy(where), copyStmts(body, where)));
}

/*****************************************************************************/
Stmt* RepeatStmt::copy(const void* where) const {

    return(new(where, "RepeatStmt") 
	RepeatStmt(cond->copy(where), copyStmts(body, where)));
}

/*****************************************************************************/
Stmt* ForStmt::copy(const void* where) const {

    return(new(where, "ForStmt") 
	ForStmt(index->copy(where), from->copy(where), dir, to->copy(where),  
	OPT_COPY(by, where), copyStmts(body, where)));
}

/*****************************************************************************/
Stmt* LoopStmt::copy(const void* where) const {

    return(new(where, "LoopStmt") LoopStmt(copyStmts(body, where)));
}

/*****************************************************************************/
Stmt* ReturnStmt::copy(const void* where) const {

    return(new(where, "ReturnStmt") ReturnStmt(OPT_COPY(val, where)));
}

/*****************************************************************************/
Stmt* ExitStmt::copy(const void* where) const {

    return(new(where, "ExitStmt") ExitStmt());
}

/*****************************************************************************/
Stmt* ExpStmt::copy(const void* where) const {

    return(new(where, "ExpStmt") ExpStmt(exp->copy(where)));
}

/*****************************************************************************/
Stmt* DisposeStmt::copy(const void* where) const {

    return(new(where, "DisposeStmt") DisposeStmt(body->copy(where)));
}

/*****************************************************************************/
Stmt* BlockStmt::copy(const void* where) const {

    return(new(where, "BlockStmt") 
	BlockStmt(copyDecls(decls, where), copyStmts(body, where)));
}

/*****************************************************************************/
When* When::copy(const void* where) const {

    return(new(where, "When") 
	When(copyExps(whenLabels, where), copyStmts(stmts, where)));
}

/*****************************************************************************/
Stmt* CaseStmt::copy(const void* where) const {

    return(new(where, "CaseStmt") 
	CaseStmt(expm->copy(where), copyWhens(otherwise, where),
	copyStmts(otherwise, where)));
}

/*****************************************************************************/
Term* FunctorTerm::copy(const void* where) const {

    return(new(where, "FunctorTerm") 
	FunctorTerm( str(functor, where), copyTerms(args, where)));
}

/*****************************************************************************/
Term* VarTerm::copy(const void* where) const {

    return(new(where, "VarTerm") VarTerm(str(var, where)));
}

/*****************************************************************************/
Term* ConstTerm::copy(const void* where) const {

    return(new(where, "ConstTerm") ConstTerm(exp->copy(where)));
}

/*****************************************************************************/
Term* ImportTerm::copy(const void* where) const {

    return(new(where, "ImportTerm") ImportTerm(copyStmts(stmts, where)));
}

/*****************************************************************************/
Term* ListTerm::copy(const void* where) const {

    return(new(where, "ListTerm") 
	ListTerm(copyTerms(firstTerms, where), 
	last==0?0:((VarTerm*) last->copy(where)->toVarTerm())));
}

/*****************************************************************************/
Clause* PropClause::copy(const void* where) const {

    return(new(where, "PropClause") 
	PropClause(copyExps(terms, where), 
	(FunctorTerm*) goal->copy(where)->toFunctorTerm()));
}

/*****************************************************************************/
Clause* CommentClause::copy(const void* where) const {

    return(new(where, "CommentClause") 
	CommentClause(kind, str(msg, where), OPT_COPY(clause, where)));
}

/*****************************************************************************/
Method* DeferedBody::copy(const void* where) const {

    return(new(where, "DeferedBody") DeferedBody());
}

/*****************************************************************************/
Method* MethodBody::copy(const void* where) const {

    return(new(where, "MethodBody") 
	MethodBody(copyDecls(decls, where), copyStmts(stmts, where),
	copyExps(constrInits, where)));
}

/*****************************************************************************/
TimeSpec* IdentTimeSpec::copy(const void* where) const {

    return(new(where, "IdentTimeSpec") IdentTimeSpec(str(name, where)));
}

/*****************************************************************************/
TimeSpec* LitTimeSpec::copy(const void* where) const {

    return(new(where, "LitTimeSpec") LitTimeSpec(ticks));
}

/*****************************************************************************/
Bool FunType::canBe(const Type* t) const {

    t = t->normType();			// normalize the type
    if (t->toFutureType() != 0)		// it is a future
	t = t->toFutureType()->getSubType()->normType();

    if (t->isKind(anyType))		// every type can be an anyType
	return(TRUE);
    if (t->isKind(taggedType))		// this type can be an taggedType
	return(TRUE);

    const FunType* asFun = t->toFunType();
    if (asFun == 0)
	return(FALSE);

    if (asFun == this)			// shortcut
	return(TRUE);

    if (asFun->kind != kind) {
		// queries are equivalent to ask
	if (!((asFun->kind == FunType::query && kind == FunType::ask) ||
	      (asFun->kind == FunType::ask && kind == FunType::query)))
	    return(FALSE);
	}

    if (kind != ftn) {		// if it is a method, check the objects
	const Type* anObj = objType;
	if (anObj == 0) 
	anObj = obj;
	if (asFun->objType != 0) {
	    if(!asFun->objType->canBe(anObj))
		return(FALSE);
	    }
	else {
	    if(!asFun->obj->canBe(anObj))
		return(FALSE);
	    }
	}

    const PiList* args1 = asFun->args;
    const PiList* args2 = args;
    while(args1 != 0 && args2 != 0) {
	FunArgDecl* arg1 = (FunArgDecl*) args1->first();
	FunArgDecl* arg2 = (FunArgDecl*) args2->first();

	if (arg2->getKind()!=FunArgDecl::in && arg1->getKind()==FunArgDecl::in)
	    return(FALSE);

	if (!arg1->getType()->canBe(arg2->getType()))
	    return(FALSE);
	args1 = args1->rest();
	args2 = args2->rest();
	}
    if (args1 != 0 || args2 != 0)
	return(FALSE);

    return (retType->canBe(asFun->retType));
}

/*****************************************************************************/
Bool ArrayType::canBe(const Type* t) const {

    t = t->normType();			// normalize the type
    if (t->toFutureType() != 0)		// it is a future
	t = t->toFutureType()->getSubType()->normType();

    if (t->isKind(anyType))		// every type can be an anyType
	return(TRUE);
    if (t->isKind(taggedType))		// this type can be an taggedType
	return(TRUE);

    const ArrayType* asArray = t->toArrayType();
    if (asArray == 0)
	return(FALSE);

    if (asArray == this)	// shortcut
	return(TRUE);

    if (!asArray->indexType->canBe(indexType))
	return(FALSE);

    if (!elemType->canBe(asArray->elemType))
	return(FALSE);
	
	// bigger arrays can be smaller arrays
    if (asArray->evalBound == 0)
	return(TRUE);

    return(evalBound == asArray->evalBound);
}

/*****************************************************************************/
Bool FutureType::canBe(const Type* t) const {

    t = t->normType();			// normalize the type
    if (t->isKind(anyFutureType))
	return(TRUE);
    return(subType->canBe(t)); 
}

/*****************************************************************************/
Bool EnumType::canBe(const Type* t) const {

    t = t->normType();			// normalize the type
    if (t->toFutureType() != 0)		// it is a future
	t = t->toFutureType()->getSubType()->normType();

    if (t->isKind(anyType))		// every type can be an anyType
	return(TRUE);
    if (t->isKind(taggedType))		// this type can be an taggedType
	return(TRUE);

    if (t->isKind(anyEnumType))
	return(TRUE);
    return(t->toEnumType() == this); 
}

/*****************************************************************************/
Bool ObjectType::canBe(const Type* t) const { 

    t = t->normType();			// normalize the type
    if (t->toFutureType() != 0)		// it is a future
	t = t->toFutureType()->getSubType()->normType();

    if (t->isKind(anyType))		// every type can be an anyType
	return(TRUE);
    if (t->isKind(taggedType))		// this type can be a taggedType
	return(TRUE);
    if (t->isKind(anyObjType))
	return(TRUE);

    const ObjectType* obj = t->toObjectType();   // is it an object type?
    if (obj == 0)
	return(FALSE);

    if (this == obj)		// if its me, then it is true
	return(TRUE);

    PiList* superPtr = superClasses;
    while(superPtr != 0) {
	ObjectType* super = (ObjectType*) superPtr->first();
	if (super->canBe(obj))
	    return(TRUE);
	superPtr = superPtr->rest();
	}
    return(FALSE);
}

/*****************************************************************************/
Bool AtomicType::canBe(const Type* t) const {

    t = t->normType();			// normalize the type
    if (t->toFutureType() != 0)		// it is a future
	t = t->toFutureType()->getSubType()->normType();

    if (t->isKind(Type::anyType))	// every type can be an anyType
	return(TRUE);
    if (t->isKind(taggedType)) 		// most types can be a taggedType
	return(kind <= taggedType);

    if (kind == unknownObjType)	{	// semantics of unknownType say it
	if (t->toObjectType() != 0)	// can be any object type
	    return(TRUE);
	return(t->isKind(anyObjType));
	}

    const AtomicType* asAtom = t->toAtomicType();
    if (asAtom == 0) return(FALSE);
	
    switch(kind) {
	case intType:
	    if (asAtom->kind == realType) 	// integers can be reals
		return(TRUE);
	    break;
	case moduleType:
	case aframeType:
	case assocInstType:
	    if (asAtom->kind == modType) 	// these can be mods
		return(TRUE);
	    break;
	case objectType:
	    if (asAtom->kind == modType || asAtom->kind == setType)
		return(TRUE);
	    break;
	case enumType:
	    if (asAtom->kind == setType)
		return(TRUE);
	    break;
	}
    return(asAtom->kind == kind);
}

/*****************************************************************************/
Bool AtomicType::indexAble() const {

    switch(kind) {
	case Type::boolType:
	case Type::charType:
	case Type::intType:
	    return(TRUE);
	default:
	    return(FALSE);
	}
}

/*****************************************************************************/
const Decl* Decl::normDecl() const {

    switch (declKind()) {
	case dCommentDecl: {
	    const Decl* ret = toCommentDecl()->getDecl();
	    if (ret == 0)
		return(this);
	    return(ret->normDecl());
	    }
	case dRenameDecl:
	    return(toRenameDecl()->getOrigDecl()->normDecl());
	case dFlashBackDecl:
	    return(toFlashBackDecl()->getDecl()->normDecl());
	default:
	    return(this);
	}
}

/*****************************************************************************/
const Type* Type::normType() const {

    switch (typeKind()) {
	case tIdentType:
	    return(toIdentType()->getDef()->getTypeDef()->normType());
	case tSynType:
	    return(toSynType()->getTypeRef()->normType());
	default:
	    return(this);
	}
}

/*****************************************************************************/
Bool IdentExp::isLValue() const {
   const FunArgDecl* asFunArg = def->toFunArgDecl();
   if (asFunArg != 0)
       return(asFunArg->getKind() != FunArgDecl::in);

   return(def->toVarDecl() != 0);
}
	
/*****************************************************************************/
TimeSpec::CmpResult IdentTimeSpec::cmp(Time*) const {
    // FIX NOW
    return(TIME_SPEC_UNKNOWN);
}

/*****************************************************************************/
TimeSpec::CmpResult LitTimeSpec::cmp(Time*) const {
    // FIX NOW
    return(TIME_SPEC_UNKNOWN);
}

/*****************************************************************************/
const Type* FunDecl::retType(Bool async) const {

	// FIX: This whole function seems like a kludge
    const Type* ret = getType()->toFunType()->getRetType();
    if (async) {
	if (asyncRet == 0)
	    ((FunDecl*) this)->asyncRet = new FutureType(ret->copy(this));
	return(asyncRet);
	}
    else
	return(ret);
}

/*****************************************************************************/
const Exp* IdentExp::getConstant() const {
    switch(def->declKind()) {
	case Decl::dConstDecl:
	    return(def->toConstDecl()->getValue()->getConstant()); 
	case Decl::dEnumValDecl:
	    return(this);
	default:
	    return(0);
	}
}

/*****************************************************************************/
void unlinkNodePiList(PiList* elems) {

    PiList* elemPtr = elems;
    while (elemPtr != 0) {
        Node* elem = (Node*) elemPtr->first();
	elem->unlink();
        elemPtr = elemPtr->rest();
        }
   delete elems;
}

/*****************************************************************************/
void ModDecl::destroy() {

    delete [] name;
    unlinkNodePiList(decls); 
    PiList* interPtr = interface;
    while (interPtr != 0) {
        PiList* tuple = (PiList*) interPtr->first();
	destroyStrPiList(tuple);
        interPtr = interPtr->rest();
        }
    delete interface;
    delete this;
}

/*****************************************************************************/
void CaseStmt::destroy() {

    expm->unlink();
    PiList* whenPtr = whens;
    while (whenPtr != 0) {
        When* when = (When*) whenPtr->first();
	when->destroy();
        whenPtr = whenPtr->rest();
        }
    delete whens;
    unlinkNodePiList(otherwise);
    delete this;
}

/*****************************************************************************/
void FunDecl::destroy() {
    delete [] name;
    type->unlink();
    if (body != 0) body->unlink();
    delete this;
}

/*****************************************************************************/
ModDecl::~ModDecl() {
   if (scope != 0) scope->unlink();
}

/*****************************************************************************/
ObjectType::ObjectType(char* aName, PiList* aSupers, PiList* aDecls, PiList* aClauses,
		       Bool simobj, int totSimobj)
    : name(aName), decls(aDecls), supers(aSupers), clauses(aClauses), 
      superClasses(0), deferedMeths(0), scope(0), 
      externalMeths(0), simObj(simobj), totSimObj(totSimobj) {

   scope = new(this, "TypeScope") TypeScope(this);
}

/*****************************************************************************/
ObjectType::~ObjectType() {

    if (externalMeths != 0)
	delete externalMeths;
    if (deferedMeths != 0)
	delete deferedMeths;
    if (superClasses != 0)
	delete superClasses;
    if (scope != 0) scope->unlink();
}

/*****************************************************************************/
AFrameDecl::~AFrameDecl() {
   if (scope != 0) scope->unlink();
}

/*****************************************************************************/
AssocInstDecl::~AssocInstDecl() {
   if (scope != 0) scope->unlink();
    unlinkNodePiList(decls);
    unlinkNodePiList(argDecls);
    unlinkNodePiList(reqDecls);
    unlinkNodePiList(assocDecls);
}

/*****************************************************************************/
AssocDecl::~AssocDecl() {
    if     (supers != 0) unlinkNodePiList(supers);
    if   (testArgs != 0) unlinkNodePiList(testArgs);
    if (assocDecls != 0) unlinkNodePiList(assocDecls);
    if      (scope != 0) scope->unlink();
    if   (assocEnv != 0) assocEnv->destroy();
}

/*****************************************************************************/
MethodBody::~MethodBody() {
   if (localScope != 0) localScope->unlink();
}
