/**************************************************************************/
/*                           IDFormat.h    IDFormat.h                     */
/**************************************************************************/

/* This file defines the formater (Pretty Printer) for the Import/Dome
   Language
*/

/* AUTHOR: Vance Morrison
   DATE  : 9/19/94                  */
/**************************************************************************/

#ifndef _IDFormat_h
#define _IDFormat_h

#include "IDAST.H"
#include "Format.H"
#include "SpIostream.H"

/**************************************************************************/
class IDFormat : public FormatNode {
public:
    IDFormat(const Node* aNode=0, int aPrec=0) : myNode(aNode), precm(aPrec) {
	if (myNode != 0) ((Node*) myNode)->link(); }
    IDFormat(const IDFormat& fmt) : myNode(fmt.myNode), precm(fmt.precm) {
	if (myNode != 0) ((Node*) myNode)->link(); }
    ~IDFormat() { if (myNode != 0) ((Node*) myNode)->unlink(); }

    void format(Formatter& out) const;
    FormatNode* copy();

public:
    void formatDecl(const Decl* node, Formatter& out) const;
    void formatDeclHead(const Decl* node, Formatter& out) const;

    void formatRenameDecl(const RenameDecl* node, Formatter& out) const;
    void formatFlashBackDecl(const FlashBackDecl* node, Formatter& out) const;
    void formatEnumDecl(const EnumDecl* node, Formatter& out) const;
    void formatObjectDecl(const ObjectDecl* node, Formatter& out) const;
    void formatModDecl(const ModDecl* node, Formatter& out) const;
    void formatTypeDecl(const TypeDecl* node, Formatter& out) const;
    void formatCommentDecl(const CommentDecl* node, Formatter& out) const;

    void formatFunDecl(const FunDecl* node, Formatter& out) const;
    void formatFunDeclHead(const FunDecl* node, Formatter& out) const;
    void formatFunArgDecl(const FunArgDecl* node, Formatter& out) const;

    void formatType(const Type* node, Formatter& out) const;
    void formatAtomicType(const AtomicType* node, Formatter& out) const;
    void formatFunType(const FunType* node, Formatter& out) const;

    void formatExp(const Exp* node, Formatter& out) const;
    void formatNewExp(const NewExp* node, Formatter& out) const;


    void formatTerm(const Term* node, Formatter& out) const;
    void formatTerms(const PiList* terms, const char* sep, Formatter& out) const;

    void formatStmt(const Stmt* node, Formatter& out) const;
    void formatAssignStmt(const AssignStmt* node, Formatter& out) const;
    void formatIfStmt(const IfStmt* node, Formatter& out) const;
    void formatWhileStmt(const WhileStmt* node, Formatter& out) const;
    void formatRepeatStmt(const RepeatStmt* node, Formatter& out) const;
    void formatForStmt(const ForStmt* node, Formatter& out) const;
    void formatBlockStmt(const BlockStmt* node, Formatter& out) const;
    void formatCaseStmt(const CaseStmt* node, Formatter& out) const;
    void formatCommentStmt(const CommentStmt* node, Formatter& out) const;

    void formatMethod(const Method* node, Formatter& out) const;
    void formatTimeSpec(const TimeSpec* node, Formatter& out) const;
    void formatIdent(const Ident* node, Formatter& out) const;
    void formatClause(const Clause* node, Formatter& out) const;
    void formatCommentClause(const CommentClause* node, Formatter& out) const;

    void formatDecls(const PiList* decls, Formatter& out) const;
    void formatStmts(const PiList* stmts, Formatter& out) const;
    void formatExps(const PiList* exps, Formatter& out) const;
    void formatClauses(const PiList* clauses, Formatter& out) const;

    void formatApplyExp(const ApplyExp* apply, Formatter& out) const;

    Bool isInfix(const Ident* name, int args, int& prec, int& assoc) const;

private:
    const Node* myNode;
    int precm;
};

// these are here mostly for calling from the debugger

void print(const Node* node, ostream& out);
void printDecls(const PiList* decls, ostream& out);
void printStmts(const PiList* stmts, ostream& out);
void printExps(const PiList* exps, ostream& out);

#endif

