/**************************************************************************/
/*                             IndentFormat.c                             */
/**************************************************************************/

/* a formatter that simply prints its output to a ostream */

/* AUTHOR: Vance Morrison
   DATE  : 7/19/93                  */
/**************************************************************************/

#include "IndentFormat.H"
#include "SpIostream.H"
#include "SpString.H"
#include "SpMemory.H"


/**************************************************************************/
IndentFormatter::~IndentFormatter() {

   if (buffIdx ==  0) 
	return;
   out->write(buff, buffIdx);
   delete [] buff;
}

/**************************************************************************/
void IndentFormatter::flush() {

   if (buffIdx < 20) 
	return;
   out->write(buff, buffIdx);
   buffIdx = 0;
}

/**************************************************************************/
Bool IndentFormatter::canBreak(Bool val) {

   Bool ret = canBreak_;
   canBreak_ = val;
   if (canBreak_) {
       out->write(buff, buffIdx);
       buffIdx = 0;
       }
   return(ret);
}

/**************************************************************************/
inline void IndentFormatter::output(char c) {

    buff[buffIdx++] = c;
    if (buffIdx >= maxIdx) {
	unsigned flushAmount = buffIdx / 2 + 1;
        out->write(buff, flushAmount);
	buffIdx -= flushAmount;
	memcpy(buff, &buff[flushAmount], buffIdx);
	}
}

/**************************************************************************/
void IndentFormatter::clip() {

    int search = 0;
    for(int i=buffIdx-2; i >= 0; --i) {
	if (buff[i] == ' ') {
	    out->write(buff, i);
	    *out << '\n';
	    for(int j=0; j < indent+1; j++) 
		*out << "    ";
	    i++;
	    buffIdx -= i;
            memcpy(buff, &buff[i], buffIdx);
	    curCol = (indent+1)*4+buffIdx;
	    return;
	    }
	search++;
	if (search > maxCol/2)
	    return;
	}
}

/**************************************************************************/
Formatter& IndentFormatter::operator<<(const char *str) {

    register const char* ptr = str;
    for(;;) {
	switch(*ptr) {
	    case '\n':
		curCol = 0;
		flush();
		break;
	    case '\0':
		return(*this);
	    default:
		if (curCol == 0) {
		    for(int i=0; i < indent*4; i++) 
			output(' ');
		    curCol = indent*4;
		    }
		curCol++;
	    }
	if (*ptr == '\t') 
	    do {
	       curCol++;
	       output(' ');
	    } while (curCol % 8 != 0);
	else
	    output(*ptr);
	if (canBreak_) {
	    if (curCol == maxCol && indent*4 < maxCol * 2/3)
		clip();
	    else if (curCol > maxCol && (*ptr == ' ' || *ptr == '\t')) {
		output('\n');
		curCol = 0;
		flush();
		}
	    }
	ptr++;
	}
}

