/**************************************************************************/
/*                             IndentFormat.h                             */
/**************************************************************************/

/* AUTHOR: Vance Morrison
   DATE  : 7/19/93                  */
/**************************************************************************/

#ifndef _IndentFormat_h
#define _IndentFormat_h

#include "Format.H"

#ifndef Bool
#define Bool int
#endif
#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

#include "SpIostream.H"


/**************************************************************************/
/* IndentFormatter contains all the state information to do nice indenting
   of program text */

class IndentFormatter : public Formatter {
public:
    IndentFormatter(ostream& aOut, unsigned aMaxCol=1000000)
	: out(&aOut), indent(0), curCol(0), maxCol(aMaxCol), 
	  maxIdx(256), buffIdx(0), canBreak_(1) { 
	  if (maxCol != 0) buff = new char[maxIdx]; }
    ~IndentFormatter();

        //  this causes transfer of ownership of the childNodes to
        // the caller (that is AstEd will not delete them on cleanup

    void shiftIn()              { indent++; }
    void shiftOut()             { if (indent > 0) --indent; }
    Bool canBreak(Bool val);

    void listBegin()            {}
    void listEnd()              {}
    Formatter& operator<<(const char *str);
    void subObj(const FormatNode& obj) { obj.format(*this); }

private:
    void clip();
    void flush();
    void output(char c);

private:
    ostream* out; 
    unsigned indent;
    unsigned curCol;
    unsigned maxCol;
    unsigned buffIdx;
    unsigned maxIdx;
    char *buff;
    Bool canBreak_;
};

#endif

