#include "SpIostream.H"
#include "SpFstream.H"
#include "IDAST.H"
#include "IOStream.H"
#include "CharArrayStream.H"
#include "Yacc.H"
#include "TypeEnv.H"
#include "TypeStatus.H"
#include "CodeGen.H"
#include "RunProg.H"
#include "SpPoDataBase.H"
//#include <execDir.h>
//#include <getInterface.h>
#include "NewWhere.H"	// extended new operator
#include "CharManipulation.H"	// string manipuliation
#ifndef __IBMCPP__
//#include <dirent.h>	// for opendir ...
//#include <FileMapHeap.h>

#include <pwd.h>	// for getpwent
#include <unistd.h>	// for getuid access
#include <sys/types.h>	// for stat
#include <sys/stat.h>	// for stat
#endif
#include "SpStdlib.H"	// for exit strtol
#include "SpCtype.H"	// for isdigit
#include "SpOS.H"

enum FlagVals { VERBOSE = 1,
		DONT_USE_ENV_DB = 2,
		CREATE_ENV_DB = 4,
		USING_VISUAL_AGE_COMPILER=8
};

/**************************************************************************/
int hasTail(const char* str, const char* tail) {

    int sLen = strlen(str);
    int tLen = strlen(tail);
    if (tLen > sLen)
        return(0);
    else
        return(strcmp(&str[sLen-tLen], tail) == 0);
}

/**************************************************************************/
/* get the priority of a filename.  If the filename has the form
   xxx.ddd.xxx where ddd are signed digits, then the last such pattern
   is the priority of the filename.  
   Note that the priority of a number is interpreted as a decimal fraction
   multiplied by 100 million.  Thus -.20.id <.100.id < .20.id */

int getPriority(const char* str) {

    const char* prio = str + strlen(str);
    int inPrio = 0;
    for(;;) {
	if (prio < str) return(0);
	if(!inPrio) 
	    inPrio = (*prio == '.');
	else if (*prio == '.')
	    break;
	else 
	    inPrio = isdigit(*prio);
	--prio;
	}

    int sign = (prio > str) && (prio[-1] == '-');
    prio++;
    int ret = 0;
    for(int i=0; i < 8; i++) {
	if (isdigit(*prio)) {
	    ret = ret + *prio - '0';
	    prio++;
	    }
	ret = ret * 10;
	}
    if (sign) ret = -ret;
    return(ret);
}

/**************************************************************************/
int fileCmp(const void* item1, const void* item2) {

    const char* str1 = *((const char**) item1);
    const char* str2 = *((const char**) item2);
    int prio1 = getPriority(str1);
    int prio2 = getPriority(str2);

    if (prio1 > prio2)
	return(1);
    if (prio1 < prio2)
	return(-1);
    return(strcmp(str1, str2));
}

/**************************************************************************/
/* JBco: not portable
PiList* findSuffix(const char* name, const char* tail, PiList* lst=0) {

    DIR* dir = opendir(name);
    if (dir == 0)
	return(lst);

    for(;;) {
	struct dirent* ent = readdir(dir);
	if (ent == 0)
	    break;
	const char* fileName = ent->d_name;
	if (strcmp(fileName, ".") != 0 && strcmp(fileName, "..") != 0) {
	    if (hasTail(fileName, tail))
                lst = new PiList(concat(name, "/", fileName), lst);
	    }
	}
    closedir(dir);
    lst = sort(lst, fileCmp);
    return(lst);
}
*/
/**************************************************************************/
/* expands the name 'name' into all the logical files it represents.  
   in particular, if 'id' is a directory it represents all the '.id' files
   inside it (recursively). These files are prepended to 'lst'.  Note
   that 'name' must be a heap string and ownership is passed to expand
   (it becomres part of the list of names) */

   /* JBco: not portable
PiList* expand(char* name, PiList* lst=0) {

    DIR* dir = opendir(name);
    if (dir == 0) {
	lst = new PiList(name, lst);
	return(lst);
	}
    closedir(dir);
    
    PiList* entPtr = reverse(findSuffix(name, ".id"));
    while (entPtr != 0) {
	char* ent = (char*) entPtr->first();
	lst = expand(ent, lst);
	entPtr = entPtr->pop();
	}
    delete [] name;		// we don't need this name 
    return(lst);
}
   */

/*****************************************************************************/
/* compares two files and returns 0 if they are identical */

int cmpFiles(const char* fileName1, const char* fileName2) {

    ifstream file1(fileName1);
    ifstream file2(fileName2);
    if (!file1 || !file2)
	return(-1);

    char buff1[8192];
    char buff2[8192];
    for(;;) {
	file1.read(buff1, 8192); int count1 = file1.gcount();
	file2.read(buff2, 8192); int count2 = file2.gcount();
	if (count1 != 8192  || count2 != 8192) {
	    if (count1 != count2)
		return(1);
	    if (memcmp(buff1, buff2, count1) != 0)
		return(1);
	    break;
	    }
   	if (memcmp(buff1, buff2, 8192) != 0)
	    return(1);
	}
    return(0);
}

/******************************************************************************/
PiList* compileStream(InStream *in, char* name, void* where,
		    const TypeEnv& env, TypeScope* scope, int flags,
		    ostream* dump, int compileFlag = 0)
{
   char* outName = strrchr(name, '/');
   if (outName == 0) outName = name; else outName++;
   if (flags & VERBOSE) cerr << "File '" << outName << "'  ";

   if (flags & VERBOSE) cerr << "Parse ";
   IDParser myParse(where);
   if(myParse.parse(*in) != IDParser::SUCCESS)
   {
      const IDParserStatus* ret = myParse.status();
      cerr << "\n" << name << ":" << ret->pos.row  << "." << ret->pos.col
	   << " Syntax error parsing token " << ret->token << "\n";
      return(0);
   }
   
   if (flags & VERBOSE) cerr << "TypeCheck ";
   if ((flags & VERBOSE & compileFlag) == 1) cerr << "Decls ";
   if ((flags & VERBOSE & compileFlag) == 2) cerr << "Defs ";
   // note that there is no type recursion on the top level
   TypeStatus typeStatus;
   TypeEnv myEnv(scope, &env);
   PiList* decls = (PiList*) myParse.ret();
   PiList* declPtr = decls;
   while (declPtr != 0)
   {
      Decl* decl = (Decl*) declPtr->first();

      /* there is a problem right now with the compileFlag strategy */
      
      if (compileFlag != 2) /* 2 == decls only */
      {      
	 decl->addObjDeclToScope(scope, &typeStatus);
	 if (!typeStatus.ok()) break;
      }

      if (compileFlag != 1) /* 1 == defs only */
      {
	 decl->typeDecl(&myEnv, &typeStatus);
	 if (!typeStatus.ok()) break;
      }
      
      if (compileFlag != 2)
      {      
	 decl->addDeclToScope(scope, &typeStatus);
	 if (!typeStatus.ok()) break;
      }

      if (compileFlag != 1)
      {      
	 decl->typeDeclBody(&myEnv, &typeStatus);
	 if (!typeStatus.ok()) break;
      }
      
      declPtr = declPtr->rest();
   }
   
   if (!typeStatus.ok())
   {
      const TermExtent* where = myParse.findExtent(typeStatus.getAt());
      cerr << "\n" << name << ":";
      if (where != 0) 
	cerr << where->begin.row << "." << where->begin.col
	     << " through "
	     << where->end.row << "." << where->end.col;
      else
	cerr << "<Unknown Position>";
      cerr << " Type checking Error\n";
      cerr << "        " << typeStatus.msg() << "\n";
      if (dump != 0 && typeStatus.getEnv() != 0)
      {
	 cerr << "Dumping environment\n";
	 typeStatus.getEnv()->print(*dump);
      }
      return(0);
   }
   if (flags & VERBOSE) cerr << "Done\n";
   return(decls);
} 

/******************************************************************************/
PiList* compileStreamDecls(InStream *in, char* name, void* where,
			    const TypeEnv& env, TypeScope* scope, int flags,
			    ostream* dump)
{
   return compileStream(in, name, where, env, scope, flags, dump, 1 /* decls only */);
}
   
/******************************************************************************/
PiList* compileStreamDeclDefs(InStream *in, char* name, void* where,
			    const TypeEnv& env, TypeScope* scope, int flags,
			    ostream* dump)
{
   return compileStream(in, name, where, env, scope, flags, dump, 2 /* defs only */);
}

/*****************************************************************************/
Stmt* compileStartStmt(char* str, void* where, TypeEnv& env, int flags,
    ostream* dump) {

    if (flags & VERBOSE) cerr << "StartStmt Parse ";
    StrInStream in(str);
    IDParser myParse(where);
    if(myParse.parseStmt(in) != IDParser::SUCCESS) {
	const IDParserStatus* ret = myParse.status();
        cerr << "\nSTART:" << ret->pos.row  << "." << ret->pos.col
             << " Syntax error parsing token " << ret->token << "\n";
	return(0);
	}

    if (flags & VERBOSE) cerr << "TypeCheck ";
    TypeStatus typeStatus;
    Stmt* stmt = (Stmt*) myParse.ret();
    stmt->typeStmt(&env, &typeStatus);

    if (!typeStatus.ok()) {
        const TermExtent* where = myParse.findExtent(typeStatus.getAt());
        cerr << "\nStartStmt TypeCheck Error: ";
        if (where != 0) 
            cerr << where->begin.row << "." << where->begin.col
                 << " through "
                 << where->end.row << "." << where->end.col;
        else
            cerr << "Unknown Position";
        cerr << "\n";
        cerr << "        " << typeStatus.msg() << "\n";
	if (dump != 0 && typeStatus.getEnv() != 0) {
	    cerr << "Dumping environment\n";
	    typeStatus.getEnv()->print(*dump);
	    }
	return(0);
	}
    if (flags & VERBOSE) cerr << "Done\n";
    return(stmt);
} 

/*****************************************************************************/
/* compile the standard environment.  Return a list of declarations.  
   'bltDecls' is also a set of declarations but differs from the return
   value in that it should NOT be run through the code generator (for
   typechecking only).  Note that the return value and the bldDecls should
   NOT be deleted since they might be from a OO database */

PiList* compileStdEnv(TypeEnv& env, TypeScope* stdScope,
		    //JB co: FileMapHeap& heap,
		    const char* runTime, PiList*& bltDecls, int flags)
{
    void* db = 0;

   /* JB co: (we are not doing environment databases for now ...)

    if (!(flags & DONT_USE_ENV_DB)) {
		// get modify time for executable so we don't use a old database
	char* executable = concat(runTime, "/../bin/import");
	struct stat execStat; 
	execStat.st_mtime = 0;
	stat(executable, &execStat);
	delete [] executable;

	//char* stdEnvHeapName = concat(runTime, "/stdEnv.db");

	if ((flags & CREATE_ENV_DB)) {
	    unlink(stdEnvHeapName);
	    if (access(stdEnvHeapName, F_OK) != 0) 
	      heap.open(stdEnvHeapName);
	    if (!heap) 
		cerr << "Could not create standard environment database\n";
	    else {
		db = new(heap) int(0);
		heap.setRoot((void*) execStat.st_mtime, 2);// store modify time
		}
	    }
	else {
	    heap.open(stdEnvHeapName, 0, FileMapHeap::PRIVATE_COPY);
	    if (!heap) 
		cerr << "Could not create standard environment database\n";
	    else {
		PiList* stdDecls = (PiList*) heap.getRoot();
		bltDecls = (PiList*) heap.getRoot(1);
		if (stdDecls != 0 && bltDecls != 0 && 
		    heap.getRoot(2) == (void*) execStat.st_mtime) {
		    if (flags & VERBOSE) 
			cerr << "Using Standard Environment database\n";
		    TypeStatus typeStatus;
		    PiList* declPtr = bltDecls;
		    while (declPtr != 0) {
			Decl* decl = (Decl*) declPtr->first();
			decl->addObjDeclToScope(stdScope, &typeStatus);
			if (!typeStatus.ok()) break;
			decl->addDeclToScope(stdScope, &typeStatus);
			if (!typeStatus.ok()) break;
			declPtr = declPtr->rest();
			}
		    declPtr = stdDecls;
		    while (declPtr != 0) {
			Decl* decl = (Decl*) declPtr->first();
			if (!typeStatus.ok()) break;
			decl->addObjDeclToScope(stdScope, &typeStatus);
			if (!typeStatus.ok()) break;
			decl->addDeclToScope(stdScope, &typeStatus);
			declPtr = declPtr->rest();
			}
		    if (!typeStatus.ok()) {
			cerr << "Error in standard environment DB\n";
		   	exit(-1);
			}
		    return(stdDecls);
		    }
		cerr << "Error reading standard environment DB\n";
		}
	    }
	delete [] stdEnvHeapName;
	}
   */
	
   if (flags & VERBOSE) cerr << "Using Standard Environment files\n";
   PiList* stdDecls = 0;

    // parse standard environment

    // JBco: not portable: PiList* files = findSuffix(runTime, ".idh");
    // for now, use instead:
    char *temp;

    temp = concat(runTime,"/REALS.idh");
    PiList *files = new PiList(temp, NULL);
    temp = concat(runTime, "/INTEGERS.idh");
    files = new PiList(temp, files);
    temp = concat(runTime, "/BOOLEANS.idh");
    files = new PiList(temp, files);
    temp = concat(runTime, "/PROCESS_MODEL.idh");
    files = new PiList(temp, files);
    temp = concat(runTime, "/IO.id");
    files = new PiList(temp, files);
    temp = concat(runTime, "/STD.id");
    files = new PiList(temp, files);

    files = reverse(files);
    
    PiList* filePtr = files;
    while (filePtr != 0)
    {
	char* fileName = (char*) filePtr->first();

	ifstream file(fileName);
	IOInStream fileIO(file);
	PiList* ret = compileStream(&fileIO, fileName, db, env, stdScope,
				  flags, 0);
	bltDecls = append(bltDecls, ret);
	if (ret == 0)
	{
	   if (flags & VERBOSE)
	     cerr << "problem with compiling standard environment"
		  << endl;
	   exit(-1);
	}

	filePtr = filePtr->rest();
    }
    destroyStrPiList(files);

    /* JBco: not using .id's in runtime yet
    files = findSuffix(runTime, ".id");
    filePtr = files;
    while (filePtr != 0)
    {
       char* fileName = (char*) filePtr->first();
       
       ifstream file(fileName);
       IOInStream fileIO(file);
       PiList* ret = compileStream(&fileIO, fileName, db, env, stdScope, 
				 flags, 0);
       if (ret == 0) exit(-1);
       stdDecls = append(stdDecls, ret);
       filePtr = filePtr->rest();
    }
    destroyStrPiList(files);
    */
    
    /* JB co:
    if ((flags & CREATE_ENV_DB) && !!heap) {
	heap.setRoot(stdDecls);
	heap.setRoot(bltDecls, 1);
	if (flags & VERBOSE) 
	    cerr << "Successfully generated new standard env DB\n";
	}
    */
    
    return(stdDecls);
    }

int GenCFile(char *outFileName, PiList* decls, PiList* stdDecls, Stmt* start,
	     char*, enum CodeGen_t type = NORMAL,
	     char **headerFiles = NULL, int numHeaderFiles = 0)
{
   CodeGen aCodeGen;
   char *CFileName = concat(outFileName, ".C");
   ofstream CFile(CFileName);
   if (!CFile)
   {
      cerr << "Can't open file '" << CFileName << "'\n";
      delete CFileName;
      return(-1);
   }
   delete [] CFileName;

   /* JBco: not portable
   PiList* files = findSuffix(runTime, ".idh.h");
   PiList* filePtr = files;
   int runTimeLen = strlen(runTime);
   */

   
   CFile << "// " << outFileName << ".C file" << endl
	 << "// Automatically generated by Parallel IMPORT"
	 << endl;

   /* JBco: not portable
   while (filePtr != 0)
   {
      char* fileName = (char*) filePtr->first();
      CFile << "#include <" << &fileName[runTimeLen+1] << ">\n";
      filePtr = filePtr->rest();
   }
   destroyStrPiList(files);
   */
   
   if (type == NORMAL) CFile << "#include \"" << outFileName << ".H\"" << endl;
   if (type == MAIN) for (int i=0; i<numHeaderFiles; i++)
	CFile << "#include \"" << headerFiles[i] << ".H\"" << endl;
   CFile << endl;
   aCodeGen.genCode(decls, stdDecls, CFile, CFILE, start, type);
   CFile.flush();
   return 1;
}

int GenHFile(char *outFileName, PiList* decls, PiList* stdDecls,
	     char **headerFiles = NULL, int numHeaderFiles = 0)
{
   CodeGen aCodeGen;
   char *HFileName = concat(outFileName, ".H");
   ofstream HFile(HFileName);

   if (!HFile)
   {
      cerr << "Can't open file '" << HFileName << "'\n";
      delete HFileName;
      return(-1);
   }
   delete [] HFileName;

   HFile << "// " << outFileName << ".H file" << endl
	 << "// Automatically generated by Parallel IMPORT"
	 << endl << endl
	 << "#ifndef " << outFileName << "_GATE" << endl
	 << "#define " << outFileName << "_GATE" << endl << endl;

   for (int i=0; i<numHeaderFiles; i++)
     if (strcmp(outFileName, headerFiles[i]))
       HFile << "#include \"" << headerFiles[i] << ".H\"" << endl;
   
   aCodeGen.genCode(decls, stdDecls, HFile, _HFILE);
   HFile << "#endif" << endl;   
   HFile.flush();
   return 1;
}


/***************************************************************************/
int doCodeGen(char* outFileName, PiList* decls, PiList* stdDecls, Stmt* start, 
	      char* runTime, int flags, enum CodeGen_t type = NORMAL,
	      char** headerFiles = NULL, int numHeaderFiles = 0)
{
   if (type == NORMAL)
   {
      if (flags & VERBOSE) cerr << "File '" << outFileName << ".[CH]'  CodeGen ";
      GenHFile(outFileName, decls, stdDecls, headerFiles, numHeaderFiles);
      GenCFile(outFileName, decls, stdDecls, start, runTime, NORMAL,
	       headerFiles, numHeaderFiles);
   }
   else
   {
      if (flags & VERBOSE) cerr << "File '" << outFileName << ".C'  CodeGen ";
      if (type == SINGLE_FILE)
	GenCFile(outFileName, decls, stdDecls, start, runTime, SINGLE_FILE);
      else // (type == MAIN)
	GenCFile(outFileName, decls, stdDecls, start, runTime, MAIN,
		 headerFiles, numHeaderFiles);
   }

   if (flags & VERBOSE) cerr << "Done\n";
   return(0);
}

/***************************************************************************/
int doCompile(char* outFileName, const PiList*, char*, int genDebug,
	      int flags)
{
   // JBco: char* cmd = concat(runTime, "/CC -c -I. -I", runTime, " ");
   // JBco: cmd = concatDel1(cmd, "-I", runTime, "/os ");

    /* JBco: PiList* interPtr = getInterfaceDecls(decls, 0);
    while (interPtr != 0) {
	PiList* interface = (PiList*) interPtr->first();
	while (interface != 0) {
	    PiList* tuple  = (PiList*) interface->first();
	    char* tag  = (char*) tuple->first();
	    char* val = (char*) tuple->rest()->first(); 
	    if (strcmp(tag, "headerDir") == 0) {
		if (access(val, R_OK) == 0)
		    cmd = concatDel1(cmd, "-I", val, " ");
		else {
		    char* fullName = concat(runTime, "/", val);
		    if (access(fullName, R_OK) == 0)
			cmd = concatDel1(cmd, "-I", fullName, " ");
		    else 
			cerr<<"Warning, could not find directory `"<<val<<"'\n";
		    delete fullName;
		    }
		 }
	    interface = interface->rest();
	    }
	interPtr = interPtr->pop();
	}
    */
   char *cmd = new char[1];
   cmd[0] = '\0';
   
    if (genDebug)
	cmd = concatDel1(cmd, "-g ");
    cmd = concatDel1(cmd, outFileName, ".C");

    if (flags & VERBOSE) cerr << "Executing command `" << cmd << "'\n";
    int ret = runProg(cmd);
    delete [] cmd;
    return(ret);
}

/***************************************************************************/
int doLink(char*, PiList*, char*, int,
	int)
{

    /* JBCO:
    char* link = concat(runTime, "/LINK -o ", outFileName, " ");
    if (genDebug)
	link = concatDel1(link, "-g ");
    link = concatDel1(link, outFileName, ".o");

    PiList* interPtr = getInterfaceDecls(decls, 0);
    while (interPtr != 0) {
	PiList* interface = (PiList*) interPtr->first();
	while (interface != 0) {
	    PiList* tuple  = (PiList*) interface->first();
	    char* tag  = (char*) tuple->first();
	    char* val = (char*) tuple->rest()->first(); 
	    if (strcmp(tag, "library") == 0) {
		if (access(val, R_OK) == 0)
		    link = concatDel1(link, " ", val);
		else {
		    char* fullName = concat(runTime, "/", val);
		    if (access(fullName, R_OK) == 0)
			link = concatDel1(link, " ", fullName);
		    else 
			link = concatDel1(link, " ", val);
		    delete fullName;
		    }
		 }
	    else if (strcmp(tag, "libraryDir") == 0) {
		if (access(val, R_OK) == 0)
		    link = concatDel1(link, " -L", val);
		else {
		    char* fullName = concat(runTime, "/", val);
		    if (access(fullName, R_OK) == 0)
			link = concatDel1(link, " -L", fullName);
		    else 
			cerr<<"Warning, could not find directory `"<<val<<"'\n";
		    delete fullName;
		    }
		}
	    interface = interface->rest();
	    }
	interPtr = interPtr->pop();
	}
    */

   /* JBco:
    PiList* libs = findSuffix(runTime, ".a");
    PiList* libPtr = libs;
    while (libPtr != 0) {
	char* libName = (char*) libPtr->first();
	link = concatDel1(link, " ", libName);
	libPtr = libPtr->rest();
	}
    destroyStrPiList(libs);
    
    libs = findSuffix(runTime, ".so");
    libPtr = libs;
    while (libPtr != 0) {
	char* libName = (char*) libPtr->first();
	link = concatDel1(link, " ", libName);
	libPtr = libPtr->rest();
	}
    destroyStrPiList(libs);
    
    // link = concatDel1(link, " -R", runTime);
    
    if (flags & VERBOSE) cerr << "Executing command `" << link << "'\n";
    int ret = runProg(link);
    delete link;
    return(ret);
   */
  return 0;
}

static inline char *strdupNew(char *s) {
  return strcpy(new char[strlen(s) + 1], s);
}


static inline void OutputIcc(FILE* outfile,
			     const char* str1,
			     const char* str2 = NULL,
			     const char* str3 = NULL,
			     const char* str4 = NULL,
			     const char* str5 = NULL,
			     const char* str6 = NULL){
  if (outfile){
    fwrite(str1, sizeof(char), strlen(str1), outfile);
    if (str2){
      fwrite(str2, sizeof(char), strlen(str2), outfile);
      if (str3){
	fwrite(str3, sizeof(char), strlen(str3), outfile);
	if (str4){
	  fwrite(str4, sizeof(char), strlen(str4), outfile);
	  if (str5){
	    fwrite(str5, sizeof(char), strlen(str5), outfile);
	    if (str6){
	      fwrite(str6, sizeof(char), strlen(str6), outfile);
	    }
	  }
	}
      }
    }
  }
  else{
    cout<<str1;
    if (str2){
      cout<<str2;
      if (str3){
	cout<<str3;
	if (str4){
	  cout<<str4;
	  if (str5){
	    cout<<str5;
	    if (str6){
	      cout<<str6;
	    }
	  }
	}
      }
    }
  }
}

char * fixSlashes(const char* path){
  char* rc = new char[2*strlen(path) + 1];
  int rcPos = 0;

  for (int i=0; i < strlen(path); ++i){
    rc[rcPos] = path[i];
    if ((path[i] == '/') || (path[i] == '\\')){
      rc[rcPos++] = '\\';
      rc[rcPos++] = '\\';
    }
    else{
      rc[rcPos++] = path[i];
    }
  }
  rc[rcPos] = 0;
  if ((rc[strlen(rc) - 1] != '\\') && (rc[strlen(rc) - 1] != '/')){
    strcat(rc, "\\\\");
  }
  return rc;
}


inline void includeIcc(FILE* iccFile, const char* path, const char* dir)
{
  OutputIcc(iccFile, "  incl(searchpath, ", "\"", path, "src\\\\", dir,"\"),\n");
}

char*  GenerateVisualAge(const char* speedesPath,
			 const char** outFileName,
			 const int totFiles,
			 const char* exeName){
  char* fname = new char[strlen(exeName) + strlen(".icc") + 1];
  sprintf(fname, "%s.icc", exeName);
  FILE* iccFile = fopen(fname, "w+");
  if (!iccFile){
    cout<<"Unable to open "<<fname<<" for writing."<<endl
	<<"The file that would have been written and compiled will be written"
	<<" to the console"<<endl;
  }
  char* path = fixSlashes(speedesPath);

  OutputIcc(iccFile, " {\n");
  OutputIcc(iccFile, "  option\n");
  includeIcc(iccFile, path, "comm\\\\SpeedesServer");
  includeIcc(iccFile, path, "stdLibExt");
  includeIcc(iccFile, path, "comm\\\\shMemTCP\\\\ClientServer");
  includeIcc(iccFile, path, "util\\\\containers");
  includeIcc(iccFile, path, "util\\\\misc");
  includeIcc(iccFile, path, "util\\\\internal");
  includeIcc(iccFile, path, "obsoleteFunctionality");
  includeIcc(iccFile, path, "obsoleteNames");
  includeIcc(iccFile, path, "engine\\\\modelFramework");
  includeIcc(iccFile, path, "engine\\\\modelFramework\\\\processModel");
  includeIcc(iccFile, path, "engine\\\\misc");
  includeIcc(iccFile, path, "engine\\\\builtInEvents");
  includeIcc(iccFile, path, "engine\\\\eventManagement");
  includeIcc(iccFile, path, "HLA\\\\DDM\\\\routingSpace");
  includeIcc(iccFile, path, "HLA\\\\DDM");
  includeIcc(iccFile, path, "HLA\\\\OM\\\\proxy");
  includeIcc(iccFile, path, "HLA\\\\OM\\\\events");
  includeIcc(iccFile, path, "HLA\\\\OM\\\\attributes\\\\dynamic\\\\position");
  includeIcc(iccFile, path, "HLA\\\\OM\\\\attributes\\\\dynamic\\\\float");
  includeIcc(iccFile, path, "HLA\\\\OM\\\\attributes\\\\dynamic");
  includeIcc(iccFile, path, "HLA\\\\OM\\\\attributes");
  includeIcc(iccFile, path, "HLA");
  includeIcc(iccFile, path, "util\\\\persistence");
  includeIcc(iccFile, path, "..\\\\import\\\\runtime");
  OutputIcc(iccFile, "  define(__SPEEDES_NT__)\n");
  OutputIcc(iccFile, "  {\n");
  OutputIcc(iccFile, "    target \"", exeName, ".exe\"\n");
  OutputIcc(iccFile, "    {\n");
  if (totFiles){
    for (int i=0; i < totFiles; ++i){
      OutputIcc(iccFile, "     source type ('cpp') \"");
      OutputIcc(iccFile, outFileName[i], ".C\"\n");
    }
  }
  else{
    OutputIcc(iccFile, "     source type (\"cpp\") \"", exeName, ".C\"\n");
  }
  OutputIcc(iccFile, "     source type(\"lib\") ", "\"ws2_32.lib\"\n");
  OutputIcc(iccFile, "     source type(\"lib\") ", "\"ole32.lib\"\n");
  OutputIcc(iccFile, "     source type(\"lib\") \"", path, "lib\\\\SpShmemTCP.lib\"\n");
  OutputIcc(iccFile, "     source type(\"lib\") \"", path, "lib\\\\SpEngine.lib\"\n");
  OutputIcc(iccFile, "     source type(\"lib\") \"", path, "lib\\\\SpUtil.lib\"\n");
  OutputIcc(iccFile, "     source type(\"lib\") \"", path, "lib\\\\SpHLA.lib\"\n");
  OutputIcc(iccFile, "    }\n");
  OutputIcc(iccFile, "  }\n");
  OutputIcc(iccFile, "}\n");
  fclose(iccFile);
  return fname;
}

// This function returns the portion of the pathname that is not the
// directory portion.  It returns a pointer that points to within the
// passed in string.  Therefore, you should probably not in general
// delete or free what this function returns.  For example, if you
// pass it "/home/bjacob/t1.id", it returns "t1.id".
static char *NotDir(char *pathname)
{
  char *not_dir;
  for (not_dir = pathname; *pathname != '\0'; ++pathname)
    if (*pathname == '/')
      not_dir = pathname + 1;
  return not_dir;
}








/***************************************************************************/
int main(int argc, char *argv[])
{
    extern SpPoDataBase* DefaultDataBase;
    DefaultDataBase = new SpPoDataBase;
    //char* argv0 = argv[0];
    --argc; argv++;

    // JB:
    char* speedesPath = NULL;
    char* cppCompilerExe = NULL;
    
    char* exeName = NULL;
    ofstream dumpFile;
    ostream* dFile=0;
    int genDebug = 0;
    int execParse = 1;
    int execCodeGen = 1;
    int execCompile = 1;
    int execLink = 1;
    int giveRevision = 0;
    int flags = 0;
    int includeStdEnv = 1;
    int singleFile = 0;
    char* startStmt = 0;
    // JB co:    FileMapHeap stdEnvHeap;
    while (argc > 0 && argv[0][0] == '-') {
	switch (argv[0][1]) {
	   case 'p': // JB
	        speedesPath = &argv[0][2];
		if (*speedesPath == '\0' && argc > 0)
		{
		   --argc; argv++;
		   speedesPath = *argv;
		}
		break;
	   case 'C': // JB
	      if (argv[0][2] == 'C')
	      {
		 cppCompilerExe = &argv[0][3];
		 if (*cppCompilerExe == '\0' && argc > 0)
		 {
		    --argc; argv++;
		    cppCompilerExe = *argv;
		 }
	      }
	      else execParse = execCodeGen = execLink = 0;
	      break;
	   case 'X':
		flags |= CREATE_ENV_DB;
		break;
	    case 'x':
		flags |= DONT_USE_ENV_DB;
		break;
	    case 's':
	       singleFile = 1;
	       /* JBco:
		startStmt = &argv[0][2];
		if (*startStmt == '\0' && argc > 0) {
		    --argc; argv++;
		    startStmt = *argv;
		    }
	       */
		break;
	    case 't':
		execCodeGen = execCompile = execLink = 0;
		break;
	    case 'c':
		execCompile = execLink = 0;
		break;
	    case 'g':
		genDebug = 1;
		break;
	    case 'l':
		execLink = 0;
		break;
	    case 'L':
		execCodeGen = execCompile = 0;
		break;
	    case 'n':
		includeStdEnv = 0;
		break;
	    case 'v':
		flags |= VERBOSE;
		break;
   	    case 'V':
	        flags |= USING_VISUAL_AGE_COMPILER;
		cout<<"Building with Visual Age C++"<<endl;
		break;
	    case 'r':
		giveRevision = 1;
		break;
	    case 'd': {
	        const char* dFileName = &argv[0][2];
		if (*dFileName == '\0' && argc > 0) {
		    --argc; argv++;
		    dFileName = *argv;
		    }
		if (dFileName[0] == 0)
		    dFile = &cerr;
		else {
		    dumpFile.open(dFileName);
		    if (!dumpFile) {
			cerr << "Error: could not open '" <<dFileName << "'\n";
			return(-1);
			}
		    dFile = &dumpFile;
		    }
		} break;
	    case 'o': 
	        exeName = &argv[0][2];
		if (*exeName == '\0' && argc > 0) {
		    --argc; argv++;
		    exeName = *argv;
		    }
		break;
 	    default: 
		cerr << "Unknown option '" << argv[0] << "'\n";
		return(-1);
	    }
	--argc; argv++;
	}
    if (!exeName) exeName = strdup("a");
    if (!speedesPath)
    {
       cerr << "Must specify a SPEEDES path with the -p option" << endl;
       return (-1);
    }
    if (!cppCompilerExe) cppCompilerExe = strdup("CC");
    if (giveRevision != 0) {
	cerr << "Parallel Import Compiler: Revision: 0.7 $\n";
	cerr << "Date: January 11, 1999\n";
	return(0);
	}
    /* JBco:
    if (startStmt == 0 && execCodeGen) {
	cerr << "Error: Must specify a start stmt with the -s qualifier\n";
	return(-1);
	}
    if (exeName == 0 && (execCodeGen || execCompile || execLink)) {
	cerr << "Error: Must specify an output file with -o qualifier\n";
	return(-1);
	}

        char* execDir = normExecDir(argv0);
        char* runTime1 = concat(execDir, "/../runtime");
        char* runTime = normalizePath(runTime1);	
    if (runTime == 0) {
	cerr << "import: Could not find runtime directory " << runTime1 << "\n";
	return(-1);
	}
    delete [] runTime1;
    delete [] execDir;
    if (flags & VERBOSE) 
	cerr << "Run time library = '" << runTime << "'\n";
    */

    // use this simpler idea instead for now -JB
    char *runTime = strdupNew(speedesPath);
    runTime = concatDel1(runTime, "/import/runtime");

    int ret = 0;
    PiList *decls = 0;
    static PiList *stdDecls = 0;	// making static keeps purify happy
    static PiList *bltDecls = 0;	// making static keeps purify happy
    Stmt* start = 0;

    PiList* filePiList = 0;
    PiList* filePtr;
    int totFiles;
    char**  inFileName;
    char**  inFileNameFullPath;
    char** outFileName;
    //int fileNum = 0;

    if (execParse) {
	void* db = 0;	// 0 means allocate in transient memory 
	TypeScope* stdScope = new(db, "TypeScope") TypeScope();
	TypeEnv stdEnv(stdScope, 0);
	stdScope->unlink();

	if (includeStdEnv)
	  stdDecls = compileStdEnv(stdEnv, stdScope, // JBco: stdEnvHeap, 
				   runTime, bltDecls, flags);

				   
	TypeScope* topScope = new(db, "TypeScope") TypeScope();
	TypeEnv topEnv(topScope, &stdEnv);
	topScope->unlink();

	for(int i=argc-1; i >=0; --i) 
	    filePiList = new PiList(copy(argv[i]), filePiList);
	// used to be: for(int i=argc-1; i >=0; --i)
	//                 expand(copy(argv[i]), filePiList);

	if (startStmt != 0 && *startStmt != 0)
	{
	   start = compileStartStmt(startStmt, 0, topEnv, flags, dFile);
	   if (start == 0)
	   {
	      if (flags & VERBOSE) cerr << "problem with start statement"
					<< endl;
	      return(-1);
	   }
	}

	filePtr = filePiList;
	totFiles = filePtr->len();
	int fileNum = 0;

	inFileName = (char **) malloc (totFiles * sizeof (char*));
	inFileNameFullPath = (char **) malloc (totFiles * sizeof (char*));

	// for Main.C (later also speedes.par,etc.)
	totFiles++;
	outFileName = (char **) malloc (totFiles * sizeof (char*));

	ifstream **file = new ifstream*[totFiles];
	IOInStream **in = new IOInStream*[totFiles];
	while (fileNum < (totFiles-1))
	{
	   inFileNameFullPath[fileNum] = (char*) filePtr->first();
	   inFileName[fileNum] = NotDir(inFileNameFullPath[fileNum]);
	   
	   file[fileNum] = new ifstream(inFileNameFullPath[fileNum]);
	   if (!*file[fileNum])
	   {
	      cerr << "Error: could not open file '"
		   << inFileName[fileNum]
		   << "'\n";
	      return(-1);
	   }
	   in[fileNum] = new IOInStream(*file[fileNum]);

	   PiList* streamDecls = compileStream(in[fileNum], inFileName[fileNum],
					     db, stdEnv, topScope, flags,
					     dFile);
	   if (streamDecls == 0)
	   {
	      if (flags & VERBOSE)
		cerr << "problem with compiling stream decls from '"
		     << inFileName[fileNum]
		     << "'"
		     << endl;
	      return(-1);
	   }

	   decls = append(decls, streamDecls);

	   filePtr = filePtr->pop();
	   fileNum++;
	}
	
	fileNum = 0;
	ifstream **file2 = new ifstream*[totFiles];
	IOInStream **in2 = new IOInStream*[totFiles];
	if (singleFile) ret = doCodeGen(exeName, decls, stdDecls, start,
					runTime, flags, SINGLE_FILE);
	else
	{
	   // create out file names:
	   for (int i=0; i<(totFiles-1); i++)
	   {
	      outFileName[i] = strdup (inFileName[i]);
	      for (int j=(strlen(outFileName[i])-1); j>0; j--)
	      {
		 if (outFileName[i][j] == '.')
		 {
		    outFileName[i][j] = '\0';
		    break;
		 }
	      }
	   }
	   outFileName[totFiles-1] = strdup("Main");
	   PiList *streamDeclDefs;
	   while (fileNum < (totFiles-1))
	   {
	      file2[fileNum] = new ifstream(inFileNameFullPath[fileNum]);
	      in2[fileNum] = new IOInStream(*file2[fileNum]);
	      streamDeclDefs = compileStreamDeclDefs(in2[fileNum],
						     inFileName[fileNum],
						     db, stdEnv, topScope,
						     flags, dFile);
	      if (streamDeclDefs == 0)
	      {
		 if (flags & VERBOSE)
		   cerr << "problem with compiling decl defs from '"
			<< inFileName[fileNum]
			<< "'"
			<< endl;
		 return(-1);
	      }
	      
	      if (ret == 0 && execCodeGen)
	      {
		 // note: right now, header files are all fileNames
		 // previous to this one command-line-wise (suboptimal)
		 ret = doCodeGen(outFileName[fileNum],
				 streamDeclDefs,
				 stdDecls,
				 start,
				 runTime,
				 flags,
				 NORMAL,
				 fileNum > 0 ?
				 &outFileName[fileNum-1] :
				 ((char**)NULL),
				 fileNum > 0 ? 1 : 0);
	      }
	      fileNum++;
	   }

	   // Generate Main.C (NOTE: right now, header files are all
	   // non-Main.C fileNames (suboptimal))
	   ret = doCodeGen(outFileName[totFiles-1], decls, stdDecls,
			   start, runTime, flags, MAIN, outFileName,
			   totFiles - 1);

	}
    }
    PiList* codeDecls = append(copy(stdDecls), copy(decls));
    if (ret == 0 && execCompile && !(flags & USING_VISUAL_AGE_COMPILER))
    {
       char* cmd = strdup(cppCompilerExe);
       if (genDebug) cmd = concatDel1(cmd, " -g");
       cmd = concatDel1(cmd, " -c -I", speedesPath, "/include");
       if (singleFile) cmd = concatDel1(cmd, " ", exeName, ".C");
       else for (int i=0; i<totFiles; i++)
	 cmd = concatDel1(cmd, " ", outFileName[i], ".C");
       if (flags & VERBOSE) cerr << "Compile command = " << endl << cmd << endl;
       runProg(cmd);
    }
    if (ret == 0 && execLink && !(flags & USING_VISUAL_AGE_COMPILER))
    {
       char* cmd = strdup(cppCompilerExe);
       if (genDebug) cmd = concatDel1(cmd, " -g");
       if (strcmp(exeName, "a")) cmd = concatDel1(cmd, " -o ", exeName);
       cmd = concatDel1(cmd, " ");
       if (singleFile) cmd = concatDel1(cmd, exeName, ".o ");
       else for (int i=0; i<totFiles; i++)
	 cmd = concatDel1(cmd, outFileName[i], ".o ");
       cmd = concatDel1(cmd,
			speedesPath,
			"/lib/ArchitectureDirs/Linux/libSpeedes.so");
       if (flags & VERBOSE) cerr << "Link command = " << endl << cmd << endl;
       runProg(cmd);
    }
    if ((ret==0) &&
	(flags & USING_VISUAL_AGE_COMPILER) &&
	(execCompile || execLink)){
      if (singleFile){
	totFiles = 0;
      }
      char* iccFName = GenerateVisualAge(speedesPath,
					 (const char**)outFileName,
					 totFiles,
					 exeName);
      SpeedesSpawnWait("vacbld.exe",
		       "vacbld.exe",
		       "-showp=1",
		       "-sev=error",
		       iccFName,
		       NULL);
      delete [] iccFName;
    }
      
    // delete [] runTime;
    delete codeDecls;
    unlinkNodePiList(decls);
    if (start != 0) start->unlink();

    return(ret);
}
