/*****************************************************************************/
/*                              runProg.c                                    */
/*****************************************************************************/

/* this file defines a program like 'system' which will run a particular
   program.   It is differen than the 'system command because
	1) It returns the error code of the program
	2) It allows stdin, stdout stderr to be set if desired
	3) It does NOT spawn a shell (no meta character replacement)
		(if you need it, just spawn a shell instead)	      */

/* AUTHOR: Vance Morrison
   DATE:  11/1/94              */
/*****************************************************************************/

#include "SpString.H"
#include "SpStdio.H"
#include "SpStdlib.H"	// for getenv

#ifndef __IBMCPP__
#ifndef __SPEEDES_NT__
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>	// for access
#endif
#else

//Plug in some stubs for Visual Age since these are not called with Visual Age
const static int X_OK = 1;
int access(const char*, int){return 1;}
int execvp(const char*, char**){return 0;}
typedef int pid_t;
int fork(){return 1;}
int dup2(int, int){return 1;}
int waitpid(pid_t, int*, int){return 0;}
#endif

#include "CharStr.H"	// for strcopy copy etc

/***************************************************************************/
int runProg(const char* command, int stdIn, int stdOut, int stdErr) {

    char* argBuff = new char[strlen(command)+1];
    strcpy(argBuff, command);

    int numArgs = 0;
    char* argPtr = argBuff;
    for(;;) {
	numArgs++;
	argPtr = strchr(argPtr, ' ');
	if (argPtr == 0)
	    break;
	while(*argPtr == ' ')
	   argPtr++;
	if (*argPtr == 0)
	    break;
	}
    char** args = new char*[numArgs+1];

    argPtr = argBuff;
    numArgs=0;
    for(;;) {
	args[numArgs++] = argPtr;
	argPtr = strchr(argPtr, ' ');
	if (argPtr == 0) {
	    break;
	    }
	while(*argPtr == ' ')
	   *argPtr++ = 0;
	if (*argPtr == 0)
	    break;
	}
    args[numArgs] = 0;

    pid_t child = fork();
    if (child == 0) {
	if (stdIn != 0)
	    dup2(stdIn, 0);
	if (stdOut != 1)
	    dup2(stdOut, 1);
	if (stdErr != 2)
	    dup2(stdErr, 2);
	int ret = execvp(args[0], args);
	perror("exec failed");
	_exit(ret);
	}
    int status = 0;
    waitpid(child, &status, 0);
    int ret = status >> 8;
    if (status != 0 && ret == 0)
	ret = -status;		// child killed by interrupt 

    delete [] args;
    delete [] argBuff;
    return(ret);
}

/***************************************************************************/
char* getCommandPath(const char* command) {

    const char* path = getenv("PATH");
    if (path == 0)
	return(0);

    const char* ptr = path;
    char* buff = new char [strlen(path)+strlen(command)+2];
    for(;;) {
	ptr = strchr(path, ':');
	if (ptr == 0) ptr = path + strlen(path);
	if (ptr == path)
	    break;
	strcopy(strcopy(strcopy(buff, path, ptr-path), "/"), command);
	if (access(buff, X_OK) == 0) {	// this is not quite right for dirs
	    char* ret = copy(buff);
	    delete [] buff;
	    return(ret);
	    }
	path = ptr;
	if (*path == ':') path++;
	}
    delete [] buff;
    return(0);
}

/***************************************************************************/
int runCommand(const char* command, int stdIn, int stdOut, int stdErr) {

    char* justCommand = copy(command);
    char* endCommand = strchr(justCommand, ' ');
    if (endCommand != 0) *endCommand = 0;
    char* path = getCommandPath(justCommand);
    delete [] justCommand;
    if (path == 0)
	return(-1);
    int ret = runProg(command, stdIn, stdOut, stdErr);
    delete [] path;
    return(ret);
}
