/*****************************************************************************/
/*                               Yacc.H                                  */
/*****************************************************************************/

/* this file defines the 'Parser' class. this class acts as an interface
   between the YACC++ parser and the rest of the program.  The rest of
   the program should not need to know anything about YACC++, all references
   to the YACC++ code pass though here first  */

/* AUTHOR: Vance Morrison
   DATE:  10/12/93		*/
/*****************************************************************************/

#ifndef _Yacc_H
#define _Yacc_H

#include "IDAST.H"
#include "CharStream.H"
#include "Lex.H"
#include "SpIostream.H"
extern "C" int yylex(void);

class Ident;
class Lexer;
class StrOutStream;
class Node;


class PiList;
class SpHashTree;

typedef int VOID;	// AT&T templates cant instantiate void so we use VOID

/*****************************************************************************/
struct IDParserStatus {
    IDParserStatus() : retCode(0), msg("") { token[0] = 0; }

    int retCode;			// return code from YACC++
    TermPos pos;
    char token[16];
    const char *msg;
    };

class PiList;

/*****************************************************************************/
class IDParser {
public:
    enum error_codes { SUCCESS = 0, FAIL = -1, SYNTAX_ERROR = -2 };

    IDParser(void* aWhere);
    ~IDParser();
    int operator!() const 	{ return(retStatus.retCode); }
    int parse(InStream& in) { firstTok = 0; return(doParse(in)); }
    int parseExp(InStream& in);
    int parseStmt(InStream& in);
    const IDParserStatus* status() { return(&retStatus); }
    void* ret()			 { return(retVal); }
	// ownership does NOT transfer to call with findExtent
    const TermExtent* findExtent(const void* node) const; 

    friend int yyparse();
    friend void yyerror(char* msg);
    friend int yylex();
    friend PiList* postCommentVarDecl(PiList* varDecls, char* comment);
    friend PiList* preCommentVarDecl(PiList* varDecls, char* comment);
    friend PiList* newFunArgs(FunArgDecl::ArgType kind,PiList* ids,Type* aType);
    friend PiList* newVarDecls(PiList* ids,Type* aType, Visibility aVis,
    			     Bool byPointer = FALSE/* JB */);
    friend PiList* fixImports(char* mod, PiList* decls);
    friend void ext(void* node, void* begin, void* end);
    friend Ident* ident(char* str, void* where);

    static void* GetWhere(){return parser->where;};

private:
    void syntaxError(const char* msg);	// indicates a syntax error 
    void delStr(char* str);
    int doParse(InStream& in);	// does most of the work
    static IDParser* parser;              // link to yylex etc

private:
    void* where;			// where to allocate the parse tree
    Lexer* lexer;                	// the lexer that this interpetor uses
    StrOutStream* tokString;            // place to place the token string
    int curTok;                		// last token read by lexer
    int firstTok;			// The first token the lexer returns
    int tokId;				// Generates unique ids for tokens	
    IDParserStatus retStatus;
    void* retVal;			// what we return 
    SpHashTree* extents;
    PiList* delStrings;			// strings to delete when parser dies
};

#endif
