/****************************************************************************/ 
/*			       IDParser.y                                   */
/****************************************************************************/ 
/*  Copyright (C) 1992,	 Vance Morrison	  */ 
/*  All rights reserved, see Copyright file for more information */ 
 
/* this file does all the interfacing between YACC and the rest of the
   program.  */
    
/* Author Vance Morrison 
   Date	  6/5/92	       */ 
/****************************************************************************/ 
 
 
/*************************** C++  Declarations ******************************/ 
%{ 
#include "SpIostream.H"	        // for debugging
#include "IDAST.H"              // for the defintions of the Abstract syntax
#include "Yacc.H" 	        // interface to the IDParser object
#include "SpStdlib.H"             // for ator atoi etc
#include "SpString.H"             // for strncpy etc
#include "SpHashTree.H"          // HashTable
#include "NewWhere.H"         // for new

static double toFloat(char* beforeDot, char* afterDot);
static char* mkString(const char* str, void* where);
#define NEWNODE(aType) 	new(IDParser::GetWhere(),#aType,1) aType
#define STR(str) 	mkString(str, IDParser::GetWhere())
%} 
 
/*************************** Yacc Declarations ******************************/ 
%union { 
    FunKind asFunKind;
    Visibility asVisibility;
    unsigned int asNum;      
    double* asFloat;      
    int*   asInt;
    char*  asChar;	// unlike asString this points to a single character
    char*  asString;
    PiList*  asList;
    FunArgDecl* asFunArgDecl;
    Decl* asDecl;
    When* asWhen;
    Method* asMethod;
    Stmt* asStmt;
    Type*  asType;
    NewExp*  asNewExp;
    ApplyExp*  asApplyExp;
    Exp*  asExp;
    RangeSpec*  asRangeSpec;
    Ident*  asIdent;
    Clause*  asClause;
    Term*  asTerm;
    VarTerm*  asVarTerm;
    TimeSpec*  asTimeSpec;
    void*  asVoid;
    void* null;	    /* for tokens that dont' have a value */
    }
     
/************************************************************************/
/* define all tokens from the lexer.  Most commonly when you add new
   Keywords to the language, their tokens need to be put here */

    /* tokens with values */
%token <asString> COMMENT_TOK 
%token <asString> PRECOMMENT_TOK 
%token <asString> POSTCOMMENT_TOK 
%token <asString> SQUOTE_TOK 
%token <asString> QUOTE_TOK 
%token <asString> NUM_TOK 
%token <asChar> CHAR_TOK 
%token <asString> ID_TOK 

%token <null>  ERROR_TOK	/* returned on lexer errors */

    /* alphabetic keywords go here.  Note that you must also add the token
       defintions to the 'keyWords' structure in the function 'parser'
       defined below.  Otherwise the lexer will never generate them  */

    /* the following keywords were added to support associations:
       AFRAME, ASSOCIATION, FIELD, REQUIRES, ROLES, SPECIFICATION
       RAW - 25 Jan 95 */

%token <null> ABOUT AFRAME AM ANY ARRAY ASSOCIATION AT ASK BEGIN BLOCK BOOLEAN 
%token <null> BY CASE CHAR CONST CONSTRUCTOR DESTRUCTOR DISPOSE DO DOWNTO 
%token <null> DOME DURATION ELSE ELSEIF ENUM END_ EXIT EXPORT FALSE_ FIELD 
%token <null> FOR FROM FUTURE
%token <null> FUNCTION IF IMPORT IN INTEGER INIT INOUT INTERFACE INTERRUPT
%token <null> IS LIST LOOP METHOD MODULE NEW_ NULL_ OBJECT OF 
%token <null> OTHERWISE OUT PM PRIVATE PUBLIC QUERY REAL REPEAT REQUIRES
%token <null> RETURN ROLES SELF SPECIFICATION STRING SUBCLASS 
%token <null> TAGGED TELL THEN TO TRUE_ TYPE UNTIL VAR VIEWED WAIT WHEN WHILE 

    /* single character symbolic tokens that don't need precedence */
%token <null> '~' '`' '\\' '{' '}' 
%token <null> '|' ':' ';' '?' ',' '(' ')' 

%token <null> LSSBAR BARGTR LSSLSSBAR BARGTRGTR

    /* delcaration of precedence (from lowest to highest)  */
%left <null> ASSIGN 
%left <null> DGTR DLSS
%left <null> OR 
%left <null> AND 
%right <null> NOT
%left <null> '=' '<' '>' LEQ GEQ NEQ DEQUAL
%left <null> '+' '-' 
%left <null> '*' '/' MOD
%right <null> '#' '^' 
%left <null> '%' '&'
%right <null> '!' '$' UMINUS
%left <null> ISA
%left <null> ASA
%left <null> DCOLON DDASH
%left <null> '[' ']'
%left <null> RARROW
%left <null> '.' 

    /* these tokens are 'fake' they are used only to trick YACC into
       allowing multiple starting points to the grammar.  Basically if
       we don't want the default start token to be used, we force the
       lexer to generate one of these tokens before lexing the input
       stream.  This forces the lexer to the desired state.  See the
       rules on the 'Start' symbol to see how this works */
%token <null> PARSE_EXP PARSE_STMT PARSE_TYPE
/***********************************************************************/ 
/* nonTerminals */
 
%type <asTimeSpec> Time
%type <asNum> Num PosAmPm PosSec PosYear 

%type <asList> Decls VarDecl
%type <asDecl> Decl FullDecl

%type <asFunKind> FunKind ConFunKind MethFunKind
%type <asVisibility> PosVis PosPubPriv
%type <null> PosCVT
%type <asList> PosInterface 
%type <asList> PosSupers 
%type <asList> PosClauses 
%type <asList> FormalArgs FormalArgs1
%type <asList> FormalArg
%type <asType> RetType AddType PosType
%type <asList> ConstrInits 
%type <asList> ConstrInits1
%type <asNewExp> ConstrInit

%type <asMethod> Body FunBody

%type <asType> Type 
%type <asList> Types Types1
%type <asList> Idents1 IdRefs IdRefs1
%type <asList> EnumIdents EnumIdents1 
%type <asIdent> IdRef

%type <asExp> Literal 
%type <asApplyExp> Apply
%type <asIdent> PosFrom
%type <asString> PosView
%type <asExp> Exp
%type <asList> Exps Exps1
%type <asExp> PosRet

%type <asStmt> Stmt
%type <asList> Stmts
%type <asList> ElseIfs
%type <asExp> PosBy
%type <asExp> PosExp
%type <asList> PosArgs
%type <asWhen> When
%type <asList> Whens
%type <asList> PosOtherwise

%type <asFloat> Float
%type <asInt> Integer
%type <asList> InterVals InterVals1
%type <asList> InterVal 
 
	/* association non-terminals */
%type <asList> AssArgs1
%type <asDecl> AssArg
%type <asRangeSpec> PosAssMult
%type <asList> PosAssReqs 

%type <asList> PosAssSpecs AssSpecs
%type <asClause> AssSpec

%type <asList> PosAssRoles 

	/* dome non-terminals */
%type <asClause> Clause
%type <asList> Clauses
%type <asTerm> Term Goal
%type <asString> Variable
%type <asList> Terms1 Goals1 
%type <null> PosIf PosIsTrue
%type <asVarTerm> PosBarTerm

%type <null> Start
%type <asList> DefaultStart
%type <asVoid> StartChoices
%start Start 

%% 
/**************************** Rule section *******************************/ 
 
Start	    :	StartChoices	{ IDParser::parser->retVal = $1; }
	    ;

StartChoices: DefaultStart		{ $$ = $1; }
	    | PARSE_EXP Exp		{ $$ = $2; }
	    | PARSE_STMT Stmt		{ $$ = $2; }
	    | PARSE_TYPE Type		{ $$ = $2; }
	    ;

DefaultStart: Decls			{ $$ = $1; }
	    ;

Decls 	    : /* empty */		{ $$ = 0; } 
	    | FullDecl Decls		{ $$ = NEWNODE(PiList)($1, $2); }
	    | VarDecl Decls		{ $$ = append($1, $2); }
	    | COMMENT_TOK Decls 	{ $$ = NEWNODE(PiList)
		    (NEWNODE(CommentDecl)(Comment::replace, $1, 0), $2); }
            | '%' COMMENT_TOK '%' Decls { $$ = NEWNODE(PiList)
                    (NEWNODE(CommentDecl)(Comment::passThrough, $2, 0), $4); }
	    ;

FullDecl    : Decl 			{ $$ = $1; }
	    | AT '(' Time ')' PosView Decl
	    { 	    $$ = NEWNODE(FlashBackDecl)($6, $3, $5); ext($$,$1,$6); }
	    | FullDecl POSTCOMMENT_TOK 	
	    {       $$ = NEWNODE(CommentDecl)(Comment::postfix, $2, $1); }
	    | PRECOMMENT_TOK FullDecl	
	    {       $$ = NEWNODE(CommentDecl)(Comment::prefix, $1, $2); }
	    ;

Decl        : PosVis ID_TOK ASSIGN Exp ';'
            {       $$ = NEWNODE(ConstDecl)($2, $4, $1); ext($$,$2,$4); }
            | PosVis MODULE ID_TOK PosInterface Decls END_ MODULE ';'
            {       $$ = NEWNODE(ModDecl)($3,$5,$4,$1); ext($$,$2,$7); }
            /*  aframe declaration - RAW (25 Jan 95) */
            | PosVis AFRAME ID_TOK Decls END_ AFRAME ';'
            {       $$ = NEWNODE(AFrameDecl)($3,$4,$1);ext($$,$2,$6); }
            /*  association instantiation declaration - RAW (26 Jan 95) */
            | PosVis ID_TOK '=' IdRef '(' Types ')' ';'
            {       $$ = NEWNODE(AssocInstDecl)($2, $4, $6, $1);ext($$,$2,$8); }
            | PosVis IMPORT IdRef ';'
            {       $$= NEWNODE(ImportDecl)($3,$1); ext($$,$2,$3); }
            | PosVis ID_TOK '=' Type ';'
            {       SynType* t = NEWNODE(SynType)($2, $4); ext(t,$2,$4);
                    $$ = NEWNODE(TypeDecl)(STR($2), t, $1); ext($$,$2,$4); }
            | PosVis ID_TOK '=' '(' EnumIdents ')' ';'
            {       EnumType* t = NEWNODE(EnumType)($2, $5); ext(t, $2, $6);
                    $$ = NEWNODE(EnumDecl)(STR($2), t, $1); ext($$,$2,$6); }
            | PosVis ID_TOK '=' OBJECT PosSupers
                Decls PosClauses END_ OBJECT ';'
	    {       ObjectType*	t = NEWNODE(ObjectType)($2, $5, $6, $7);
                    $$ = NEWNODE(ObjectDecl)(STR($2), t, $1); 
		    ext($$,$2,$9); ext(t,$2,$9);  }
            | PosVis OBJECT ID_TOK PosSupers 
                Decls PosClauses END_ OBJECT ';'
	    {       ObjectType*	t = NEWNODE(ObjectType)($3, $4, $5, $6);
                    $$ = NEWNODE(ObjectDecl)(STR($3), t, $1); 
		    ext($$,$3,$8); ext(t,$3,$8);  }

/* JB: begin SpSimObj OBJECT declaration syntax: */

            | PosVis ID_TOK '[' Integer ']' '=' OBJECT PosSupers
                Decls PosClauses END_ OBJECT ';'
	    {       ObjectType*	t = NEWNODE(ObjectType)($2, $8, $9, $10, 
							TRUE, *$4);
                    $$ = NEWNODE(ObjectDecl)(STR($2), t, $1); 
		    ext($$,$2,$12); ext(t,$2,$12);  }
            | PosVis OBJECT ID_TOK '[' Integer ']' PosSupers 
                Decls PosClauses END_ OBJECT ';'
	    {       ObjectType*	t = NEWNODE(ObjectType)($3, $7, $8, $9,
							TRUE, *$5);
                    $$ = NEWNODE(ObjectDecl)(STR($3), t, $1); 
		    ext($$,$3,$11); ext(t,$3,$11);  }

/* JB: end SpSimObj OBJECT declaration syntax: */

             | PosVis MethFunKind PosType METHOD ID_TOK 
		'(' FormalArgs ')' RetType Body ';'
            {       $$ = NEWNODE(FunDecl)($5,
                    NEWNODE(FunType)($2,$7,$9,$3), $1, $10); ext($$,$4,$11); }
            | PosVis ConFunKind PosType METHOD 
		'(' FormalArgs ')' RetType Body ';'
            {       $$ = NEWNODE(FunDecl)(STR(""),
                    NEWNODE(FunType)($2,$6,$8,$3), $1, $9); ext($$,$4,$10); }
            | PosVis FUNCTION ID_TOK '(' FormalArgs ')' RetType FunBody ';'
            {       $$ = NEWNODE(FunDecl)($3,
                    NEWNODE(FunType)(FunType::ftn,$5,$7), $1, $8);
                    ext($$,$2,$6); }
            | PosVis ID_TOK DEQUAL IdRef AddType ';'
            {       $$ = NEWNODE(RenameDecl)($2, $4, $5, $1); ext($$,$2,$4); }
            /*  association declaration - RAW (25 Jan 95) */
            | PosVis ID_TOK '(' AssArgs1 ')' '=' ASSOCIATION PosSupers 
		PosAssReqs PosAssSpecs PosAssRoles END_ ASSOCIATION ';'
            {       $$ = NEWNODE(AssocDecl)($2, $4, $8, $9, $10, $11, $1);
			 ext($$,$2,$14); }
            ;

VarDecl     : PosVis Idents1 ':' Type ';' 
            {           $$ = newVarDecls($2, $4, $1); }
            | PosVis Idents1 ':' '(' Type ')' ';'
            {           $$ = newVarDecls($2, $5, $1, TRUE); } /* JB*/
	    | VarDecl POSTCOMMENT_TOK { $$ = postCommentVarDecl($1, $2); }
	    | PRECOMMENT_TOK VarDecl { $$ = preCommentVarDecl($2, $1); }
            ;

Type	    : IdRef		
	    {	    $$ = NEWNODE(IdentType)($1); ext($$,$1,$1); }
            | ARRAY '[' Exp ']' OF Type 
	    {	    $$ = NEWNODE(ArrayType)($6, $3); ext($$,$1,$6); }
            | ARRAY '[' ']' OF Type 
	    {	    $$ = NEWNODE(ArrayType)($5, 0); ext($$,$1,$5); }
	    | FunKind PosType METHOD '(' FormalArgs ')' RetType 
            {       $$ = NEWNODE(FunType)($1,$5,$7,$2); ext($$,$3,$6); }
	    | FUNCTION '(' FormalArgs ')' RetType 
            {       $$ = NEWNODE(FunType)(FunType::ftn,$3,$5); ext($$,$1,$4); }
	    | INTEGER 		
	    { 	    $$ = NEWNODE(AtomicType)(Type::intType); ext($$,$1,$1); }
	    | REAL 		
	    {	    $$ = NEWNODE(AtomicType)(Type::realType); ext($$,$1,$1); }
	    | STRING 		
	    { 	    $$ = NEWNODE(AtomicType)(Type::stringType); ext($$,$1,$1); }
	    | BOOLEAN 		
	    { 	    $$ = NEWNODE(AtomicType)(Type::boolType); ext($$,$1,$1); }
	    | CHAR 		
	    { 	    $$ = NEWNODE(AtomicType)(Type::charType); ext($$,$1,$1); }
	    | LIST 		
	    { 	    $$ = NEWNODE(AtomicType)(Type::listType); ext($$,$1,$1); }
	    | TAGGED 		
	    { 	    $$ = NEWNODE(AtomicType)(Type::taggedType); ext($$,$1,$1); }
	    | OBJECT 		
	    { 	    $$ = NEWNODE(AtomicType)(Type::anyObjType); ext($$,$1,$1); }
	    | ENUM 		
	    { 	    $$ = NEWNODE(AtomicType)(Type::anyEnumType);ext($$,$1,$1); }
	    | FUTURE 		
	    { 	    $$= NEWNODE(AtomicType)(Type::anyFutureType);ext($$,$1,$1);}
	    | FUTURE '(' Type ')'
	    { 	    $$ = NEWNODE(FutureType)($3);ext($$,$1,$4); }
	    | ANY 		
	    { 	    $$ = NEWNODE(AtomicType)(Type::anyType); ext($$,$1,$1); }
	    ;

FunKind     : ConFunKind		{ $$ = $1; }
	    | MethFunKind		{ $$ = $1; }
	    ;

MethFunKind : ASK 			{ $$ = FunType::ask; }
	    | TELL 			{ $$ = FunType::ask; 
		  TermExtent* extent= (TermExtent*) IDParser::parser->
		    extents->Find((int) $1);
		  cerr<< "\n:" << extent->begin.row << "." << extent->begin.col 
		      << "   Warning: TELL methods same as ASK,"
		      << " please use ASK.  "; }
	    | QUERY 			{ $$ = FunType::query; }
	    ;

ConFunKind  : CONSTRUCTOR 		{ $$ = FunType::constructor; }
	    | DESTRUCTOR 		{ $$ = FunType::destructor; }
	    ;

RetType	    : /* empty */		
	    { 	    $$ = NEWNODE(AtomicType)(Type::voidType); }
	    | ':' Type			{ $$ = $2; }
	    ;

AddType	    : /* empty */		{ $$ = 0; }
	    | DCOLON Type		{ $$ = $2; }
	    ;

PosType	    : /* empty */		{ $$ = 0; }
	    | Type			{ $$ = $1; }
	    ;

FormalArg   : IN Idents1 ':' Type	
	    { 	    $$ = newFunArgs(FunArgDecl::in, $2, $4); delete $2; }
	    | OUT Idents1 ':' Type	
	    { 	    $$ = newFunArgs(FunArgDecl::out, $2, $4); delete $2; }
	    | INOUT Idents1 ':' Type	
	    { 	    $$ = newFunArgs(FunArgDecl::inout, $2, $4); delete $2; }
	    ;

Body        : /* empty */ 		{ $$ = 0; }
	    | SUBCLASS 
	    {	    $$ = NEWNODE(DeferedBody)(); ext($$,$1,$1); }
	    | ConstrInits Decls BEGIN Stmts END_ METHOD 
	    {	    $$ = NEWNODE(MethodBody)($2, $4, $1); ext($$,$3,$6); }
	    ;

FunBody     : /* empty */ 		{ $$ = 0; }
	    | Decls BEGIN Stmts END_ FUNCTION 
	    {	    $$ = NEWNODE(MethodBody)($1, $3, 0); ext($$,$2,$5); }
	    ;

ConstrInit  : IdRef '(' Exps ')'	
	    {	    $$ = NEWNODE(NewExp)(0, NEWNODE(IdentType)($1), $3); 
		    ext($$,$1,$4); }
	    ;

/***********************************************************************/
/*                              DOME                                   */
/***********************************************************************/

Clause      : Goals1 THEN Goal 
	    { FunctorTerm* goal = (FunctorTerm*) $3->toFunctorTerm();
	      if (goal == 0) {  yyerror("not functor\n"); YYERROR; }
	      $$ = NEWNODE(PropClause)($1, goal); 
		   ext($$,($1!=0)?$1->first():$2,$3); }
	    | Goal 
	    { FunctorTerm* goal = (FunctorTerm*) $1->toFunctorTerm();
	      if (goal == 0) {  yyerror("not functor\n"); YYERROR; }
	      $$ = NEWNODE(PropClause)(0, goal); 
		   ext($$,$1,$1); }
	    | Clause POSTCOMMENT_TOK
	    {       $$ = NEWNODE(CommentClause)(Comment::postfix, $2, $1); }
	    | PRECOMMENT_TOK Clause
	    {       $$ = NEWNODE(CommentClause)(Comment::prefix, $1, $2); }
            ; 

Goal	    : PosIf Term PosIsTrue	 		{ $$ = $2; }
	    ;

Term        : ID_TOK 			
	    {	    $$ = NEWNODE(FunctorTerm)($1, 0); ext($$,$1,$1); }
            | ID_TOK '(' Terms1 ')'	
	    {	    $$ = NEWNODE(FunctorTerm)($1, $3); ext($$,$1,$4); }
            | Variable                  
	    {	    $$ = NEWNODE(VarTerm)($1); ext($$,$1,$1); }
            | '[' ']'			
	    {	    $$ = NEWNODE(ListTerm)(0, 0); ext($$,$1,$2); }
            | '[' Terms1 PosBarTerm ']'	
	    {	    $$ = NEWNODE(ListTerm)($2, $3); ext($$,$1,$4); }
            | LSSLSSBAR Stmts BARGTRGTR
	    {	    $$ = NEWNODE(ImportTerm)($2); ext($$,$1,$3); }
	    | LSSBAR Exp BARGTR
	    {	    $$ = NEWNODE(ConstTerm)($2); ext($$,$1,$3); }
	    | Literal
	    {	    $$ = NEWNODE(ConstTerm)($1); ext($$,$1,$1); }
            ;

Variable    : '?' ID_TOK                { $$ = $2; }
            ;
/***********************************************************************/
/*                         ASSOCIATIONS                                */
/*                       RAW - 25 Jan 1995                             */
/***********************************************************************/

AssArgs1    : AssArg                    { $$ = NEWNODE(PiList)($1, 0); }
            | AssArg ';' AssArgs1       { $$ = NEWNODE(PiList)($1, $3); }
            ;

AssArg      : ID_TOK ':' Type PosAssMult
            {       $$ = NEWNODE(AssocArgDecl)($1,$3,$4); ;ext($$,$1,$3); }
            ;

PosAssMult  : /* empty */               { $$ = 0; }
            | '[' Integer ']'           
            {       $$=NEWNODE(RangeSpec)(NEWNODE(int)(*$2),$2); }
            | '[' '*' ']'
            {       $$ = NEWNODE(RangeSpec)(); }
            | '[' Integer DDASH Integer ']'
            {       $$ = NEWNODE(RangeSpec)($2, $4); }
            | '[' Integer DDASH '*' ']'
            {       $$ = NEWNODE(RangeSpec)($2); }
            ;

PosAssReqs  : /* empty */              	    { $$ = 0; }
            | REQUIRES Decls 		    { $$ = $2; }
            ;

PosAssSpecs : /* empty */                       { $$ = 0; }
            | SPECIFICATION AssSpecs 		{ $$ = $2; }
            ;

AssSpecs    : /* empty */	  	    { $$ = 0; }
            | AssSpec ';' AssSpecs          { $$ = NEWNODE(PiList)($1, $3); }
            ;

AssSpec     : Clause 			    { $$ = $1; }
            ;

PosAssRoles : /* empty */                   { $$ = 0; }
            | ROLES Decls 		    { $$ = $2; }
            ;

/***********************************************************************/

Stmt	    : BLOCK Decls BEGIN Stmts END_ BLOCK ';'
	    {	    $$ = NEWNODE(BlockStmt)($2, $4); ext($$,$1,$7); }
	    | Exp ';' 
	    { 	    $$ = NEWNODE(ExpStmt)($1); ext($$,$1,$2); }
	    | TELL Exp TO IdRef '(' Exps ')' PosFrom ';'
	    {	    $$ = NEWNODE(AsyncStmt)(0, 
			 NEWNODE(ApplyExp)(FunType::tell, $2, $4, $6, $8)); 
			 ext($$,$1,$9); }

/* JB: begin syntax for TELL (temporarily not availbable: PosRet in front of 
       TELL both before and after this comment) */

	    | IN Exp TELL Exp TO IdRef '(' Exps ')' PosFrom ';'
	    {	    $$ = NEWNODE(AsyncStmt)(0, 
			 NEWNODE(ApplyExp)(FunType::tell, $4, $6, $8, $10),
		                           $4, AsyncStmt::awWait, $2);
			 ext($$,$1,$9); }

	    | AT Exp TELL Exp TO IdRef '(' Exps ')' PosFrom ';'
	    {	    $$ = NEWNODE(AsyncStmt)(0, 
			 NEWNODE(ApplyExp)(FunType::tell, $4, $6, $8, $10),
	                                   0, AsyncStmt::awWaitUntil, $2);
			 ext($$,$1,$9); }

	    | TELL Exp '[' Exp ']' TO IdRef '(' Exps ')' PosFrom ';'
	    {	    $$ = NEWNODE(AsyncStmt)(0, 
			 NEWNODE(ApplyExp)(FunType::tell, $2, $7, $9, $11), $4);
			 ext($$,$1,$12); }

	    | IN Exp TELL Exp '[' Exp ']' TO IdRef '(' Exps ')' PosFrom ';'
	    {	    $$ = NEWNODE(AsyncStmt)(0, 
			 NEWNODE(ApplyExp)(FunType::tell, $4, $9, $11, $13), 
					    $6, AsyncStmt::awWait, $2);
			 ext($$,$1,$12); }

	    | AT Exp TELL Exp '[' Exp ']' TO IdRef '(' Exps ')' 
                 PosFrom ';'
	    {	    $$ = NEWNODE(AsyncStmt)(0, 
			 NEWNODE(ApplyExp)(FunType::tell, $4, $9, $11, $13), 
					    $6, AsyncStmt::awWaitUntil, $2);
			 ext($$,$1,$12); }

/* JB: end syntax */

	    | IF Exp THEN Stmts ElseIfs END_ IF ';'
	    { 	    $$ = NEWNODE(IfStmt)($2, $4, $5); ext($$,$1,$8); }
	    | CASE Exp Whens PosOtherwise END_ CASE ';'
	    { 	    $$ = NEWNODE(CaseStmt)($2, $3, $4); ext($$,$1,$7); }
	    | WHILE Exp DO Stmts END_ WHILE ';'
	    { 	    $$ = NEWNODE(WhileStmt)($2, $4); ext($$,$1,$7); }
	    | REPEAT Stmts UNTIL Exp  END_ REPEAT ';'
	    { 	    $$ = NEWNODE(RepeatStmt)($4, $2); ext($$,$1,$7); }
	    | FOR Exp ASSIGN Exp TO Exp PosBy DO Stmts END_ FOR ';'
	    { 	    $$ = NEWNODE(ForStmt)($2,$4,1,$6,$7,$9); ext($$,$1,$12); }
	    | FOR Exp ASSIGN Exp DOWNTO Exp PosBy DO Stmts END_ FOR ';'
	    { 	    $$ = NEWNODE(ForStmt)($2,$4,-1,$6,$7,$9); ext($$,$1,$12); }
	    | LOOP Stmts END_ LOOP ';'
	    { 	    $$ = NEWNODE(LoopStmt)($2); ext($$,$1,$5); }
	    | RETURN PosExp ';'
	    { 	    $$ = NEWNODE(ReturnStmt)($2); ext($$,$1,$3); }
	    | Exp ASSIGN Exp ';'
	    { 	    $$ = NEWNODE(AssignStmt)($1, $3); ext($$,$1,$4); }
	    | EXIT ';'
	    { 	    $$ = NEWNODE(ExitStmt)(); ext($$,$1,$2); }
	    | DISPOSE Exp ';'
	    { 	    $$ = NEWNODE(DisposeStmt)($2); ext($$,$1,$3); }
	    | Stmt POSTCOMMENT_TOK
	    {       $$ = NEWNODE(CommentStmt)(Comment::postfix, $2, $1); }
	    | PRECOMMENT_TOK Stmt
	    {       $$ = NEWNODE(CommentStmt)(Comment::prefix, $1, $2); }
	    ;

ElseIfs     : /* empty */	   	   { $$ = 0; }
	    | ELSEIF Exp THEN Stmts ElseIfs 
	    { 	    Stmt* stmt = NEWNODE(IfStmt)($2, $4, $5); 
		    ext(stmt,$1,$3);
	    	    $$ = NEWNODE(PiList) (stmt, 0); }
	    | ELSE Stmts 		   { $$ = $2; ext($$,$1,$1); }
	    ; 

When	    : WHEN  Exps ':' Stmts 	   { $$ = NEWNODE(When)($2, $4); }
	    ;

Exp	    : '(' Exp ')' 		{ $$ = $2; }
	    | Literal			{ $$ = $1; }
	    | Apply			{ $$ = $1; }
	    | Variable	
	    {	    $$ = NEWNODE(VarExp)($1); ext($$,$1,$1); }
	    | Exp DCOLON Type		 
	    {	    $$ = NEWNODE(AsExp)($1, $3); ext($$,$1,$3); }
	    | Exp ISA Type		 
	    {	    $$ = NEWNODE(IsAExp)($1, $3); ext($$,$1,$3); }
	    | Exp ISA VAR
	    {	    $$ = NEWNODE(IsAExp)($1, 0); ext($$,$1,$3); }
	    | Exp ASA Type
	    {	    $$ = NEWNODE(CastExp)($1, $3); ext($$,$1,$3); }
	    | SELF			
	    {	    $$ = NEWNODE(SelfExp)(); ext($$,$1,$1); }
	    | IdRef			
	    {	    $$ = NEWNODE(IdentExp)($1); ext($$,$1,$1); }
            | Exp '[' Exp ']'           
	    {	    $$ = NEWNODE(ArrayExp)($1, $3); ext($$,$1,$3); }
	    | NEW_ PosArgs Type PosArgs
	    { 	    $$ = NEWNODE(NewExp)($2, $3, $4); ext($$,$1,$3); }
	    ;

Literal	    : Integer	    { $$ = NEWNODE(IntExp)($1); ext($$,$1,$1); }
	    | Float	    { $$ = NEWNODE(RealExp)($1); ext($$,$1,$1); }
	    | CHAR_TOK	    { $$ = NEWNODE(CharExp)($1); ext($$,$1,$1); }
	    | QUOTE_TOK	    { $$ = NEWNODE(StringExp)($1); ext($$,$1,$1); }
	    | TRUE_	    { $$ = NEWNODE(BoolExp)(TRUE); ext($$,$1,$1); }
	    | FALSE_	    { $$ = NEWNODE(BoolExp)(FALSE); ext($$,$1,$1); }
	    | NULL_	    { $$ = NEWNODE(NullExp)(); ext($$,$1,$1); }
	    ;

Apply       : ASK Exp TO IdRef '(' Exps ')' PosFrom
	    { 	    $$ = NEWNODE(ApplyExp)(FunType::ask, $2, $4, $6, $8); 
		    ext($$,$1,$7); }

/* JB: begin WAIT, WAIT UNTIL, etc. */

            | WAIT Exp
	    {	    $$ = NEWNODE(ApplyExp)(ident("wait", $1),
		    NEWNODE(PiList)($2, 0)); ext($$,$1,$2); }

            | WAIT UNTIL Exp
	    {	    $$ = NEWNODE(ApplyExp)(ident("waitUntil", $1),
		    NEWNODE(PiList)($3, 0)); ext($$,$1,$3); }

            | WAIT FOR Exp
	    {	    double *negOne = new(double);
	            *negOne = -1;
		    $$ = NEWNODE(ApplyExp)(ident("waitFor", $1),
					   NEWNODE(PiList)($3, 
							 NEWNODE(PiList)
							 (NEWNODE(RealExp)
							  (negOne), 0))); 
	            ext($$,$1,$3); }

            | WAIT Exp FOR Exp
	    {	    $$ = NEWNODE(ApplyExp)(ident("waitFor", $1),
		    NEWNODE(PiList)($4, NEWNODE(PiList)($2, 0))); ext($$,$1,$4); }

/* JB: end WAIT, WAIT UNTIL, etc. */

	    | Exp RARROW IdRef '(' Exps ')' PosFrom
	    { 	    $$ = NEWNODE(ApplyExp)(FunType::ask, $1, $3, $5, $7); 
		    ext($$,$1,$6); }
	    | QUERY Exp ABOUT IdRef '(' Exps ')' PosFrom 
	    {	    $$ = NEWNODE(ApplyExp)(FunType::query, $2, $4, $6, $8); 
		    ext($$,$1,$7); }
	    | IdRef '(' Exps ')'	
	    {	    $$ = NEWNODE(ApplyExp)($1, $3); ext($$,$1,$4); }
	    | Exp '+' Exp 		
	    {	    $$ = NEWNODE(ApplyExp)(ident("+", $2),
		    NEWNODE(PiList)($1, NEWNODE(PiList)($3, 0))); ext($$,$1,$3); }
	    | Exp '-' Exp 		
	    {	    $$ = NEWNODE(ApplyExp)(ident("-", $2),
		    NEWNODE(PiList)($1, NEWNODE(PiList)($3, 0))); ext($$,$1,$3); }
	    | Exp '*' Exp 		
	    {	    $$ = NEWNODE(ApplyExp)(ident("*", $2),
		    NEWNODE(PiList)($1, NEWNODE(PiList)($3, 0))); ext($$,$1,$3); }
	    | Exp '%' Exp 		
	    {	    $$ = NEWNODE(ApplyExp)(ident("%", $2),
		    NEWNODE(PiList)($1, NEWNODE(PiList)($3, 0))); ext($$,$1,$3); }
	    | Exp '&' Exp 		
	    {	    $$ = NEWNODE(ApplyExp)(ident("&", $2),
		    NEWNODE(PiList)($1, NEWNODE(PiList)($3, 0))); ext($$,$1,$3); }
	    | Exp '#' Exp 		
	    {	    $$ = NEWNODE(ApplyExp)(ident("#", $2),
		    NEWNODE(PiList)($1, NEWNODE(PiList)($3, 0))); ext($$,$1,$3); }
	    | Exp '^' Exp 		
	    {	    $$ = NEWNODE(ApplyExp)(ident("^", $2),
		    NEWNODE(PiList)($1, NEWNODE(PiList)($3, 0))); ext($$,$1,$3); }
	    | Exp '/' Exp 		
	    {	    $$ = NEWNODE(ApplyExp)(ident("/", $2),
		    NEWNODE(PiList)($1, NEWNODE(PiList)($3, 0))); ext($$,$1,$3); }
	    | Exp MOD Exp 		
	    {	    $$ = NEWNODE(ApplyExp)(ident("mod", $2),
		    NEWNODE(PiList)($1, NEWNODE(PiList)($3, 0))); ext($$,$1,$3); }
	    | Exp NEQ Exp 		
	    {	    $$ = NEWNODE(ApplyExp)(ident("!=", $2),
		    NEWNODE(PiList)($1, NEWNODE(PiList)($3, 0))); ext($$,$1,$3); }
	    | Exp '=' Exp 		
	    {	    $$ = NEWNODE(ApplyExp)(ident("=", $2),
		    NEWNODE(PiList)($1, NEWNODE(PiList)($3, 0))); ext($$,$1,$3); }
	    | Exp LEQ Exp 		
	    {	    $$ = NEWNODE(ApplyExp)(ident("<=", $2),
		    NEWNODE(PiList)($1, NEWNODE(PiList)($3, 0))); ext($$,$1,$3); }
	    | Exp GEQ Exp 		
	    {	    $$ = NEWNODE(ApplyExp)(ident(">=", $2),
		    NEWNODE(PiList)($1, NEWNODE(PiList)($3, 0))); ext($$,$1,$3); }
	    | Exp '>' Exp 		
	    {	    $$ = NEWNODE(ApplyExp)(ident(">", $2),
		    NEWNODE(PiList)($1, NEWNODE(PiList)($3, 0))); ext($$,$1,$3); }
	    | Exp '<' Exp 		
	    {	    $$ = NEWNODE(ApplyExp)(ident("<", $2),
		    NEWNODE(PiList)($1, NEWNODE(PiList)($3, 0))); ext($$,$1,$3); }
	    | Exp DEQUAL Exp 
	    {	    $$ = NEWNODE(ApplyExp)(ident("==", $2),
		    NEWNODE(PiList)($1, NEWNODE(PiList)($3, 0))); ext($$,$1,$3); }
	    | Exp AND Exp 		
	    {	    $$ = NEWNODE(ApplyExp)(ident("and", $2),
		    NEWNODE(PiList)($1, NEWNODE(PiList)($3, 0))); ext($$,$1,$3); }
	    | Exp OR Exp 		
	    {	    $$ = NEWNODE(ApplyExp)(ident("or", $2),
		    NEWNODE(PiList)($1, NEWNODE(PiList)($3, 0))); ext($$,$1,$3); }
	    | Exp DLSS Exp 		
	    {	    $$ = NEWNODE(ApplyExp)(ident("<<", $2),
		    NEWNODE(PiList)($1, NEWNODE(PiList)($3, 0))); ext($$,$1,$3); }
	    | Exp DGTR Exp 		
	    {	    $$ = NEWNODE(ApplyExp)(ident(">>", $2),
		    NEWNODE(PiList)($1, NEWNODE(PiList)($3, 0))); ext($$,$1,$3); }
	    | NOT Exp 			
	    {	    $$ = NEWNODE(ApplyExp)(ident("not", $1),
		    NEWNODE(PiList)($2, 0)); ext($$,$1,$2); }
	    | '-' Exp 			%prec UMINUS
	    {	    $$ = NEWNODE(ApplyExp)(ident("-", $1),
		    NEWNODE(PiList)($2, 0)); ext($$,$1,$2); }
	    | '!' Exp 
	    {	    $$ = NEWNODE(ApplyExp)(ident("!", $1),
		    NEWNODE(PiList)($2, 0)); ext($$,$1,$2); }
	    | '$' Exp 
	    {	    $$ = NEWNODE(ApplyExp)(ident("$", $1),
		    NEWNODE(PiList)($2, 0)); ext($$,$1,$2); }
	    ;

IdRef	    : ID_TOK
	    {	    $$ = NEWNODE(Ident)($1); ext($$,$1,$1); }
	    | IdRef '.' ID_TOK
	    {	    $$ = NEWNODE(Ident)($3, $1); ext($$,$1,$3); }
	    ;

Time	    : ID_TOK
	    { 	    $$ = NEWNODE(IdentTimeSpec)($1); ext($$,$1,$1); }
	    | Num '/' Num PosYear 
	    { 	    unsigned long ticks = toTicks($1,$3,$4,0,0,0,0);
		    if (ticks == ((unsigned long)-1)) {  
			yyerror("bad date\n"); YYERROR; }
		    $$ = NEWNODE(LitTimeSpec)(ticks);  ext($$,$2,$2); }
	    | Num '/' Num PosYear Num ':' Num PosSec PosAmPm
	    { 	    unsigned long ticks = toTicks($1,$3,$4,$5,$7,$8,$9);
		    if (ticks == ((unsigned long)-1)) {  
			yyerror("bad date\n"); YYERROR; }
		    $$ = NEWNODE(LitTimeSpec)(ticks);  ext($$,$2,$6); }
	    ;

InterVal    : ID_TOK ':' QUOTE_TOK 
	    { $$ = NEWNODE(PiList)($1, NEWNODE(PiList)($3, 0)); ext($$,$1,$3); }

Num	    : NUM_TOK
	    {	    $$ = atoi($1); IDParser::parser->delStr($1); }
	    ;

Integer	    : NUM_TOK
	    {	    if ($1[0] == '0' && ($1[1] == 'X' || $1[1] == 'X'))
			$$ = NEWNODE(int)((int) strtol($1+2, 0, 16));
		    else
			$$ = NEWNODE(int)(atoi($1));
		    ext((Node*) $$, $1, $1); IDParser::parser->delStr($1); }
	    ;

Float	    : NUM_TOK '.' NUM_TOK	{ $$ = NEWNODE(double);
					  *$$ = toFloat($1, $3); 
					  ext((Node*) $$, $1, $3); 
					  IDParser::parser->delStr($1);
					  IDParser::parser->delStr($3); }
	    | '.' NUM_TOK 		{ $$ = NEWNODE(double);
					  *$$ = toFloat(0, $2); 
					  ext((Node*) $$, $1, $2); 
					  IDParser::parser->delStr($2); }
	    | NUM_TOK '.'  		{ $$ = NEWNODE(double);
					  *$$ = toFloat($1, 0); 
					  ext((Node*) $$, $1, $2);  
					  IDParser::parser->delStr($1); }
	    ;


/**************************************************************************/
/* definitions of lists and optional parts go here, (since they are not 
   usually interesting at all) */

PosView	    : /* empty */		{ $$ = 0; }
	    | VIEWED '(' ID_TOK ')' 	{ $$ = $3; }
	    ;

PosAmPm     : AM			{ $$ = 0; }
	    | PM			{ $$ = 12; }
	    ;

PosSec	    : /* empty */		{ $$ = 0; }
	    | ':' Num			{ $$ = $2; }
	    ;

PosYear	    : /* empty */		{ $$ = 0; } 
	    | '/' Num			{ $$ = $2; }
	    ;

PosVis	    : PosCVT PosPubPriv		{ $$ = $2; }
	    ;

PosPubPriv  : /* empty */		{ $$ = Decl::defaultVis; }
	    | EXPORT			{ $$ = Decl::publicVis; }
	    | PUBLIC			{ $$ = Decl::publicVis; }
	    | PRIVATE			{ $$ = Decl::privateVis; }
	    ;

PosCVT	    : /* empty */		{ $$ = 0; }
	    | VAR			{ $$ = 0; }
	    | CONST			{ $$ = 0; }
	    | TYPE			{ $$ = 0; }
	    ;

PosFrom     : /* empty */               { $$ = 0; }
            | FROM IdRef                { $$ = $2; }
            ;
 
PosRet	    : /* empty */		{ $$ = 0; }
	    | Exp ASSIGN  		{ $$ = $1; }
	    ;

PosSupers   : /* empty */		{ $$ = 0; }
	    | '(' IdRefs ')'  		{ $$ = $2; }
	    ;

PosBy       : /* empty */	   	{ $$ = 0; }
	    | BY Exp 			{ $$ = $2; }
	    ; 

PosArgs     : /* empty */	   	{ $$ = 0; }
	    | '(' Exps ')'		{ $$ = $2; }
	    ; 

PosExp     : /* empty */	   	{ $$ = 0; }
	    | Exp 			{ $$ = $1; }
	    ; 

PosOtherwise: /* empty */	   	{ $$ = 0; }
	    | OTHERWISE Stmts 		{ $$ = $2; }
	    ; 

PosInterface: /* empty */	   	{ $$ = 0; }
	    | INTERFACE '(' InterVals ')' { $$ = $3; }
	    ; 

PosClauses  : /* empty */	   		{ $$ = 0; }
	    | DOME BEGIN Clauses END_ DOME ';'	{ $$ = $3; }
	    ; 

PosBarTerm  : /* empty */	   	{ $$ = 0; }
            | '|' Variable              { $$ = NEWNODE(VarTerm)($2); }
	    ; 

PosIf       : /* empty */	   	{ $$ = 0; }
	    | IF 			{ $$ = 0; }
	    ; 

PosIsTrue   : /* empty */	   	{ $$ = 0; }
	    | IS TRUE_ 			{ $$ = 0; }
	    ; 

Whens       : /* empty */	   	{ $$ = 0; }
	    | When Whens 	   	{ $$ = NEWNODE(PiList)($1, $2); }
	    ; 

Stmts       : /* empty */	   	{ $$ = 0; }
	    | Stmt Stmts	    	{ $$ = NEWNODE(PiList)($1, $2); }
	    | COMMENT_TOK Stmts 	{ $$ = NEWNODE(PiList)(
		     NEWNODE(CommentStmt)(Comment::replace, $1, 0), $2); }
	    ; 

FormalArgs: /* empty */           	{ $$ = 0; }
       	    | FormalArgs1	      	{ $$ = $1; }
       	    ;

FormalArgs1: FormalArg           	{ $$ = $1; }
            | FormalArg ';' FormalArgs1	{ $$ = append($1, $3); }
            ;

ConstrInits : /* Empty */			{ $$ = 0; }
	    | INIT ConstrInits1 ';'		{ $$ = $2; }
	    ;

ConstrInits1 :	ConstrInit			{ $$ = NEWNODE(PiList)($1, 0); }
	    | ConstrInit ',' ConstrInits1	{ $$ = NEWNODE(PiList)($1, $3); }
	    ;

IdRefs	    : /* empty */              	{ $$ = 0; }
            | IdRefs1			{ $$ = $1; }
            ;

IdRefs1	    : IdRef	              	{ $$ = NEWNODE(PiList)($1, 0); }
            | IdRef ',' IdRefs1    	{ $$ = NEWNODE(PiList)($1, $3); }
            ;

EnumIdents  : /* empty */           	{ $$ = 0; }
       	    | EnumIdents1	        { $$ = $1; }
       	    ;

EnumIdents1 : ID_TOK                   	
	    { 		$$ = NEWNODE(PiList)(NEWNODE(EnumValDecl)($1), 0); }
            | ID_TOK ',' EnumIdents1   	
	    { 		$$ = NEWNODE(PiList)(NEWNODE(EnumValDecl)($1), $3); }
            ;

Idents1	    : ID_TOK                   	{ $$ = NEWNODE(PiList)($1, 0); }
            | ID_TOK ',' Idents1      	{ $$ = NEWNODE(PiList)($1, $3); }
            ;

Exps	    : /* empty */           	{ $$ = 0; }
       	    | Exps1	            	{ $$ = $1; }
       	    ;

Exps1	    : Exp                   	{ $$ = NEWNODE(PiList)($1, 0); }
            | Exp ',' Exps1      	{ $$ = NEWNODE(PiList)($1, $3); }
            ;

Types	    : /* empty */           	{ $$ = 0; }
       	    | Types1	            	{ $$ = $1; }
       	    ;

Types1	    : Type                   	{ $$ = NEWNODE(PiList)($1, 0); }
            | Type ',' Types1      	{ $$ = NEWNODE(PiList)($1, $3); }
            ;

Goals1      : Goal                      { $$ = NEWNODE(PiList)($1, 0); }
            | Goal ',' Goals1           { $$ = NEWNODE(PiList)($1, $3); }
            | Goal AND Goals1           { $$ = NEWNODE(PiList)($1, $3); }
            ;

Terms1      : Term                      { $$ = NEWNODE(PiList)($1, 0); }
            | Term ',' Terms1           { $$ = NEWNODE(PiList)($1, $3); }
            ;

Clauses     : /* empty */               { $$ = 0; }
            | Clause '.' Clauses        { $$ = NEWNODE(PiList)($1, $3); }
	    | COMMENT_TOK Clauses 	{ $$ = NEWNODE(PiList)(
		     NEWNODE(CommentClause)(Comment::replace, $1, 0), $2); }
            ;

InterVals   : /* empty */           	{ $$ = 0; }
       	    | InterVals1	            	{ $$ = $1; }
       	    ;

InterVals1  : InterVal                 { $$ = NEWNODE(PiList)($1, 0); }
            | InterVal ',' InterVals1    { $$ = NEWNODE(PiList)($1, $3); }
            ;

%% 

/************************** Program section *******************************/ 

/* this code establishes the link to the lexer and provides the interface
   to the rest of the program */

#include "Lex.H" 
#include "CharArrayStream.H" 

int yyparse();

	// this is the pointer that the non-object-oriented routines like 
	// yyparse, yylex, yyerror use to access the parser object
IDParser* IDParser::parser=0;  

/*****************************************************************************/
/* helper function needed to delete the Extent table */

static void* delExtents(const int*, const TermExtent* value, void*) {

    delete (TermExtent*) value;
    return(0);
}

/**************************************************************************/ 
/* this function takes a list of identifiers and creates a list of 
   FunArgDecls that have the given ArgType and Type */

PiList* newFunArgs(FunArgDecl::ArgType kind, PiList* ids, Type* aType) {

    if (ids == 0) {
	delete aType;
	return(0);
	}

    char* id = (char*) ids->first();
    FunArgDecl* newFunArg = NEWNODE(FunArgDecl) (kind, id, aType);
        ext(newFunArg, id, id);

    ids = ids->rest();
        if (ids == 0)
    return(NEWNODE(PiList) (newFunArg, 0));

    aType = aType->copy(IDParser::GetWhere());
    return(NEWNODE(PiList) (newFunArg, newFunArgs(kind, ids, aType)));
}

/**************************************************************************/ 
PiList* postCommentVarDecl(PiList* varDecls, char* comment) {

    PiList* ret = 0;

    PiList* varPtr = varDecls;
    while (varPtr->rest() != 0) {
	ret = NEWNODE(PiList)(varPtr->first(), ret);
        varPtr = varPtr->rest();
        }
    ret = NEWNODE(PiList)(NEWNODE(CommentDecl)(Comment::postfix, 
	comment, (Decl*) varDecls->first()), ret);

    delete varDecls;
    return(reverse(ret));
}

/**************************************************************************/ 
PiList* preCommentVarDecl(PiList* varDecls, char* comment) {

    PiList* ret = NEWNODE(PiList)(NEWNODE(CommentDecl)(Comment::prefix, 
	comment, (Decl*) varDecls->first()), 0);

    PiList* varPtr = varDecls->rest();
    while (varPtr != 0) {
	ret = NEWNODE(PiList)(varPtr->first(), ret);
        varPtr = varPtr->rest();
        }

    delete varDecls;
    return(reverse(ret));
}

/**************************************************************************/ 
/* this function takes a list of identifiers and creates a list of 
   VarDecl that have the given Type */

PiList* newVarDecls(PiList* ids, Type* aType, Visibility aVis, 
		  Bool byPointer /* JB */)
{

    PiList* ret = 0;
    PiList* idPtr = ids;
    while (idPtr != 0)
    {
        char* id = (char*) idPtr->first();
	VarDecl* newVarDecl = NEWNODE(VarDecl) (id, aType, aVis, byPointer);

	ext(newVarDecl, id, id);
        ret = NEWNODE(PiList) (newVarDecl, ret);
        idPtr = idPtr->rest();
	if (idPtr != 0) aType = aType->copy(IDParser::GetWhere());
    }
    delete ids;	// we are through with this list
    return(ret);
}

/**************************************************************************/ 
/* Creats an identifier for a given string */

Ident* ident(char* str, void* where) {

    Ident* ret = NEWNODE(Ident)(STR(str));
    ext(ret, where, where); 
    return(ret);
}

/**************************************************************************/ 
/* converts a string to a float, strings can be NULL  */

static double toFloat(char* beforeDot, char* afterDot) {

    static char emptyString[] = "";
    if (beforeDot == 0)
	beforeDot = emptyString;
    if (afterDot == 0)
	afterDot = emptyString;

    int totlen = strlen(beforeDot) + strlen(afterDot) + 2;
    char* real = new char[totlen];
    strcpy(real, beforeDot);
    strcat(real, ".");
    strcat(real, afterDot);
    double ret = atof(real);
    delete [] real; 
    return(ret);
}

/**************************************************************************/ 
/* allocate a string on the heap */

static char* mkString(const char* str, void* where=0) {
    int len = strlen(str)+1;
    char *ret;
    if (where != 0)
	ret = new(where, "char", len) char[len];
    else
	ret = new char[len];
    strcpy(ret, str);
    return(ret);
}

/**************************************************************************/ 
/* ext updates the extent database.  For each node in the AST that the
   parser creates, the lexer also keeps track of the extent (so that
   error messages can be related back to the input text */

void ext(void* node, void* begin, void* end) {
    TermExtent* begExtent = 
      (TermExtent*) IDParser::parser->extents->Find((int) begin);
    TermExtent* endExtent = 
      (TermExtent*) IDParser::parser->extents->Find((int) end);
    if (begExtent == 0 || endExtent == 0) {
	cerr << "error looking up extents beg " << begin 	// DBG
	     << " end " << end << "\n" ;	// DBG
	if (begExtent != 0) {			// DBG
	     cerr << "begin extent = "		// DBG
	     << begExtent->begin.row << ":" << begExtent->begin.col << " end "//DBG
	     << begExtent->end.row << ":" << begExtent->end.col << "\n"; //DBG
	     }
	if (endExtent != 0) {			// DBG
	     cerr << "end extent = "		// DBG
	     << endExtent->begin.row << ":" << endExtent->begin.col << " end "//DBG
	     << endExtent->end.row << ":" << endExtent->end.col << "\n"; //DBG
	     }
	return;
	}
    TermExtent* newExtent = new TermExtent;
    newExtent->begin = begExtent->begin;
    newExtent->end = endExtent->end;
//    cerr << "creating extent for obj " << node << " begin " 	// DBG
//  	 << newExtent->begin.row << ":" << newExtent->begin.col << " end "//DBG
//	 << newExtent->end.row << ":" << newExtent->end.col << "\n"; //DBG
    
	IDParser::parser->extents->Insert((void *) newExtent, (int) node);
	// JBco and changed:     newExtent = (TermExtent*) IDParser::parser->
	//  extents->GetPreviousElement();
	//if (newExtent != 0) {
//	cerr << "object put in table twice looking up extents\n";	// DBG
//	cerr << "previous object = " << (void*) newExtent << "  "
//	     << newExtent->begin.row << ":" << newExtent->begin.col << " end "//DBG
//	     << newExtent->end.row << ":" << newExtent->end.col << "\n"; //DBG
	//delete newExtent;
	//	}
}

/**************************************************************************/ 
/* yacc's lexer.  here we do some simple translation from the lexer 
   to what yacc expects */

int yylex() { 

	/* We allow the parser to designate a token to be the first token
	   in the input stream if so desired.  This can be used to force
	   YACC into a desired state */
    if (IDParser::parser->firstTok != 0) {
	int ret = IDParser::parser->firstTok;
	IDParser::parser->firstTok = 0;
	return(ret);
	}
	
	// create a unique default id.   keep tokId odd, so it can never be a
	// valid pointer (returned by new) because of alignment.
    yylval.asVoid = (void*) IDParser::parser->tokId; 	
    IDParser::parser->tokId += 2;

    TermExtent* extent = new TermExtent;

    IDParser::parser->tokString->reset();
    IDParser::parser->curTok = 
	IDParser::parser->lexer->get(
	    *IDParser::parser->tokString, extent); 
    IDParser::parser->tokString->done();
    char* tokStr = IDParser::parser->tokString->val();
    // cerr<<"Tok "<<IDParser::parser->curTok<<" '"<<tokStr<<"'\n"; // DBG

    if (IDParser::parser->curTok > 0) {// single char keyword token
	IDParser::parser->extents->Insert(
	    (void *) extent, (int) yylval.asVoid);
	return(IDParser::parser->curTok);
	}

    switch(IDParser::parser->curTok) { 
	case Lexer::QUOTE: 
	    yylval.asString = STR(tokStr);
	    IDParser::parser->extents->Insert(
		(void *) extent, (int) yylval.asVoid);
	    return(QUOTE_TOK);
	case Lexer::SQUOTE:	    
		// if not a one character string
	    if (tokStr[0] == 0 || tokStr[1] != 0) {
		delete extent;
		return(ERROR_TOK);
		}
	    yylval.asChar = NEWNODE(char)(tokStr[0]);
	    IDParser::parser->extents->Insert(
		(void *) extent, (int) yylval.asVoid);
	    return(CHAR_TOK);
	case Lexer::NUM: 
		    // note that we do NOT allocate persistantly 
		    // but the yacc rules above must delete it
	    yylval.asString = new char[strlen(tokStr)+1];
	    strcpy(yylval.asString, tokStr);
	    IDParser::parser->extents->Insert(
		(void*) extent, (int) yylval.asVoid);
	    return(NUM_TOK);
	case Lexer::ID:	    // as identifiers 
	    yylval.asString = STR(tokStr);
	    IDParser::parser->extents->Insert(
		(void *) extent, (int) yylval.asVoid);
	    return(ID_TOK);
	case Lexer::COMMENT: 
	    yylval.asString = STR(tokStr);
	    IDParser::parser->extents->Insert(
		(void*) extent, (int) yylval.asVoid);
	    return(COMMENT_TOK);
	case Lexer::POSTCOMMENT: 
	    yylval.asString = STR(tokStr);
	    IDParser::parser->extents->Insert(
		(void *) extent, (int) yylval.asVoid);
	    return(POSTCOMMENT_TOK);
	case Lexer::PRECOMMENT: 
	    yylval.asString = STR(tokStr);
	    IDParser::parser->extents->Insert(
		(void *) extent, (int) yylval.asVoid);
	    return(PRECOMMENT_TOK);
	case Lexer::END: 
	    delete extent;
	    return(0);	    // yaccs end symbol
	default:
	    delete extent;
	    return(ERROR_TOK);
	} 
} 
     
/**************************************************************************/ 
void yyerror(char* msg) { 
    IDParser::parser->syntaxError(msg);
	/* FIX - we really should clean up all the objects we allocated */
}

/**************************************************************************/ 
void IDParser::syntaxError(const char* message) {

    retVal = 0;
    retStatus.retCode = SYNTAX_ERROR;
    lexer->curPos(&retStatus.pos);
    retStatus.msg = message;
    StrOutStream out(retStatus.token, 16);
    lexer->printTok(curTok, out); 
    out.done();
} 
 
/**************************************************************************/ 
int IDParser::parseExp(InStream& in) {

    firstTok = PARSE_EXP;
    return(doParse(in));
}

/**************************************************************************/ 
int IDParser::parseStmt(InStream& in) {

    firstTok = PARSE_STMT;
    return(doParse(in));
}

/**************************************************************************/ 
/* parse the input stream and save the results.   Return the error status
   as well as saving it. */
 
int IDParser::doParse(InStream& in) {

    if (parser != 0) {		// only one parser can be running at a time
	retStatus.retCode = FAIL;
	return(FAIL);
	}
    parser = this;		// attach yyparse, yylex routines to me

    Lexer myLexer(in); 
    lexer = &myLexer; 
    StrOutStream myString;  
    tokString = &myString;  
    retStatus.retCode = SUCCESS;
    retStatus.msg = "";
    retStatus.token[0] = 0;
    retVal = 0;

    tokId = 1;	// set to a value that a pointer new would never return
    if (extents != 0) {
       // might create a leak - JB: FIX by deleting each item in the tree 
       // extents->forAll(delExtents, 0);
	delete extents;
	}
    extents = new SpHashTree(10297);

    lexer->addKeyWord("ABOUT", ABOUT);
    lexer->addKeyWord("AFRAME",AFRAME); /* RAW */
    lexer->addKeyWord("AND", AND);
    lexer->addKeyWord("AM", AM);
    lexer->addKeyWord("ANY", ANY);
    lexer->addKeyWord("ARRAY", ARRAY);
    lexer->addKeyWord("ASSOCIATION",ASSOCIATION); /* RAW */
    lexer->addKeyWord("AT", AT);
    lexer->addKeyWord("ASK", ASK);
    lexer->addKeyWord("BEGIN", BEGIN);
    lexer->addKeyWord("BLOCK", BLOCK);
    lexer->addKeyWord("BOOLEAN", BOOLEAN);
    lexer->addKeyWord("BY", BY);
    lexer->addKeyWord("CASE", CASE);
    lexer->addKeyWord("ASA", ASA);
    lexer->addKeyWord("CHAR", CHAR);
    lexer->addKeyWord("CONST", CONST);
    lexer->addKeyWord("CONSTRUCTOR", CONSTRUCTOR);
    lexer->addKeyWord("DESTRUCTOR", DESTRUCTOR);
    lexer->addKeyWord("DISPOSE", DISPOSE);
    lexer->addKeyWord("DO", DO);
    lexer->addKeyWord("DOWNTO", DOWNTO);
    lexer->addKeyWord("DOME", DOME);
    lexer->addKeyWord("DURATION", DURATION);
    lexer->addKeyWord("ELSE", ELSE);
    lexer->addKeyWord("ELSEIF", ELSEIF);
    lexer->addKeyWord("END", END_);
    lexer->addKeyWord("ENUM", ENUM);
    lexer->addKeyWord("EXIT", EXIT);
    lexer->addKeyWord("EXPORT", EXPORT);
    lexer->addKeyWord("FALSE", FALSE_);
    lexer->addKeyWord("FIELD",FIELD); /* RAW */
    lexer->addKeyWord("FOR", FOR);
    lexer->addKeyWord("FROM", FROM);
    lexer->addKeyWord("FUNCTION", FUNCTION);
    lexer->addKeyWord("FUTURE", FUTURE);
    lexer->addKeyWord("IF", IF);
    lexer->addKeyWord("IMPORT", IMPORT);
    lexer->addKeyWord("IN", IN);
    lexer->addKeyWord("INIT", INIT);
    lexer->addKeyWord("INOUT", INOUT);
    lexer->addKeyWord("INTEGER", INTEGER);
    lexer->addKeyWord("INTERFACE", INTERFACE);
    lexer->addKeyWord("INTERRUPT", INTERRUPT);
    lexer->addKeyWord("ISA", ISA);
    lexer->addKeyWord("IS", IS);
    lexer->addKeyWord("LIST", LIST);
    lexer->addKeyWord("LOOP", LOOP);
    lexer->addKeyWord("METHOD", METHOD);
    lexer->addKeyWord("MOD", MOD);
    lexer->addKeyWord("MODULE", MODULE);
    lexer->addKeyWord("NEW", NEW_);
    lexer->addKeyWord("NOT", NOT);
    lexer->addKeyWord("NULL", NULL_);
    lexer->addKeyWord("OBJECT", OBJECT);
    lexer->addKeyWord("OF", OF);
    lexer->addKeyWord("OR", OR);
    lexer->addKeyWord("OTHERWISE", OTHERWISE);
    lexer->addKeyWord("OUT", OUT);
    lexer->addKeyWord("PM", PM);
    lexer->addKeyWord("PRIVATE", PRIVATE);
    lexer->addKeyWord("PUBLIC", PUBLIC);
    lexer->addKeyWord("QUERY", QUERY);
    lexer->addKeyWord("REAL", REAL);
    lexer->addKeyWord("REPEAT", REPEAT);
    lexer->addKeyWord("REQUIRES",REQUIRES); /* RAW */
    lexer->addKeyWord("RETURN", RETURN);
    lexer->addKeyWord("ROLES",ROLES); /* RAW */
    lexer->addKeyWord("SUBCLASS", SUBCLASS);
    lexer->addKeyWord("SELF", SELF);
    lexer->addKeyWord("SPECIFICATION",SPECIFICATION); /* RAW */
    lexer->addKeyWord("STRING", STRING);
    lexer->addKeyWord("TAGGED", TAGGED);
    lexer->addKeyWord("TELL", TELL);
    lexer->addKeyWord("THEN", THEN);
    lexer->addKeyWord("TO", TO);
    lexer->addKeyWord("TRUE", TRUE_);
    lexer->addKeyWord("TYPE", TYPE);
    lexer->addKeyWord("UNTIL", UNTIL);
    lexer->addKeyWord("VAR", VAR);
    lexer->addKeyWord("WAIT", WAIT);
    lexer->addKeyWord("WHEN", WHEN);
    lexer->addKeyWord("WHILE", WHILE);
    lexer->addKeyWord("VIEWED", VIEWED);

    lexer->addSymbol("<|", LSSBAR);
    lexer->addSymbol("|>", BARGTR);
    lexer->addSymbol("<<|", LSSLSSBAR);
    lexer->addSymbol("|>>", BARGTRGTR);
    lexer->addSymbol("==", DEQUAL);
    lexer->addSymbol("--", DDASH); /* RAW: Association Multiplicity */
    lexer->addSymbol(">=", GEQ);
    lexer->addSymbol("<=", LEQ);
    lexer->addSymbol("!=", NEQ);
    lexer->addSymbol(":=", ASSIGN);
    lexer->addSymbol(">>", DGTR);
    lexer->addSymbol("<<", DLSS);
    lexer->addSymbol("::", DCOLON);
    lexer->addSymbol("->", RARROW);

    yyparse();				    // run the configuration 
    parser = 0;
    return(retStatus.retCode); 
} 

/**************************************************************************/ 
IDParser::IDParser(void* aWhere) : where(aWhere), delStrings(0) {
    extents = 0;

}

/**************************************************************************/ 
IDParser::~IDParser() {

    if (extents != 0) {
       // may create a leak - JB: FIX by deleting each item in the tree
       //extents->forAll(delExtents, 0);
	delete extents;
	}
    destroyStrPiList(delStrings);
}

/**************************************************************************/ 
/* when some strings are freed we can not immediately return them to the
   heap because the pointer to them is used to carry extent information
   and if we returned it to the heap, it may be reallocated as another
   token and cause a clash.  Thus we save all the deleted strings and
   delete them when the parser dies.  */

void IDParser::delStr(char* str) {
    delStrings = new PiList(str, delStrings);
}

/**************************************************************************/ 
const TermExtent* IDParser::findExtent(const void* node) const {

    return((const TermExtent*) extents->Find((int) node));
}
