/**************************************************************************/ 
/*		     ArrayStream.h    ArrayStream.h			  */ 
/**************************************************************************/ 
/*  Copyright (C) 1992,	 Vance Morrison	  */ 
/*  All rights reserved, see Copyright file for more information */ 
 
/* AUTHOR: Vance Morrison  
   DATE	 : 10/23/93		     */ 
/**************************************************************************/ 
 
#ifndef _CharArrayStream_H
#define _CharArrayStream_H
 
#include "CharStream.H"

/**************************************************************************/ 
class CharArrayInStream : public InStream {
public:
    CharArrayInStream(const char* aArray, int aBound) 
    : array(aArray), bound(aBound) {}
    const char& cur()		{ return(*array); }
    Bool isEnd()		{ return(bound == 0); }
    void next()			{ --bound; array++; }

private:
    int bound; 
    const char* array; 
};

/**************************************************************************/ 
class CharArrayOutStream : public OutStream {
public:
    CharArrayOutStream(int initSize=32) : 
	array(0),bound(initSize),cur(0),dynamic(1),dealloc(1) {
	array = new char[initSize]; 
	}
    CharArrayOutStream(char* aArray, int aBound) 
	: array(aArray), bound(aBound),cur(0),dynamic(0),dealloc(0) {}
    virtual ~CharArrayOutStream() { if (dealloc) delete [] array; };

    /* ownership of the array does NOT pass to the caller 
       thus when this object dies so does the array */
    char* val()		    { return(array); }
    /* ownership of the array now passes to the caller.	 it is
       now invalid to do more 'puts' on the stream */
    char* ret()		    { dealloc = FALSE; return(array); }
    /* reuse the array */
    void reset() { cur = 0; } 
    Bool put(const char& val);

    void done()			    { }
    int size()			    { return(cur); }

protected:
    Bool dealloc;
    Bool dynamic;
    int cur; 
    int bound; 
    char* array; 
};

/**************************************************************************/ 
/* strings are a bit different than simple character arrays because we
   want the null termination.  Also because the elements are characters
   we can do other stuff */

class StrInStream : public InStream {
public:
    StrInStream(const char* aStr) : str(aStr) {}
    const char& cur()		{ return(*str); }
    Bool isEnd()		{ return(*str == 0); }
    void next()			{ str++; }

private:
    const char* str; 
};

/**************************************************************************/ 
class StrOutStream : public CharArrayOutStream {
public:
    StrOutStream(char* aStr, int aBound) 
    : terminated(FALSE), CharArrayOutStream(aStr, aBound-1) {}
    StrOutStream(int initSize = 32) 
    : terminated(FALSE), CharArrayOutStream(initSize) {}

    char* val()			    { if (!terminated) 
	  done();
	   return(array); }
    void reset()		    { terminated = FALSE; 
	  ((CharArrayOutStream*) this)->reset(); }
    void done()			    { if (put('\0')) 
	  --cur;
	   else 
	  array[cur] = 0;
	  terminated = TRUE; }
private:
    Bool terminated;
};

#endif

