/*****************************************************************************/
/*                              charStr.h                                    */
/*****************************************************************************/

/* defined some useful strings manipulation routines on char* (as apposed
   to a new String structure */

/* AUTHOR: Vance Morrison
   DATE:  11/1/94              */
/*****************************************************************************/

#ifndef _CharStr_h
#define _CharStr_h

#include "SpString.H"

/****************************************************************************/
inline char* strcopy(char* buff, const char* str, int n) {

    int len = strlen(str);
    if (len > n) len = n;

    strncpy(buff, str, len);
    return(buff+len);
}

/****************************************************************************/
inline char* strcopy(char* buff, const char* str) {

    strcpy(buff, str);
    return(buff+strlen(buff));
}

/****************************************************************************/
/* create a copy of 'str' on the heap */

inline char* copy(const char* str) {

    char* ret = new char[strlen(str)+1];
    strcpy(ret, str);
    return(ret);
}
    
/****************************************************************************/
/* concatinate two strings.  Caller must free returned buffer */

inline char* concat(const char* str1, const char* str2) {

    char* ret = new char[strlen(str1)+strlen(str2)+1];
    strcpy(strcopy(ret, str1), str2);
    return(ret);
}
    
/****************************************************************************/
/* concatinate three strings.  Caller must free returned buffer */

inline char* concat(const char* str1, const char* str2, const char* str3) {

    char* ret = new char[strlen(str1)+strlen(str2)+strlen(str3)+1];
    strcpy(strcopy(strcopy(ret, str1), str2), str3);
    return(ret);
}
    
/****************************************************************************/
/* concatinate four strings.  Caller must free returned buffer */

inline char* concat(const char* str1, const char* str2, const char* str3,			    const char* str4) {

    char* ret = new char[strlen(str1)+strlen(str2)+strlen(str3)+strlen(str4)+1];
    strcpy(strcopy(strcopy(strcopy(ret, str1), str2), str3), str4);
    return(ret);
}
    
/****************************************************************************/
/* concatinate five strings.  Caller must free returned buffer */

inline char* concat(const char* str1, const char* str2, const char* str3,			    const char* str4, const char* str5) {

    char* ret = new char[strlen(str1)+strlen(str2)+strlen(str3)+
			 strlen(str4)+strlen(str5)+1];
    strcpy(strcopy(strcopy(strcopy(strcopy(ret,str1),str2),str3),str4),str5);
    return(ret);
}
    
/****************************************************************************/
/* concatinate two strings and delete the first arg */

inline char* concatDel1(char* str1, const char* str2) {

    char* ret = new char[strlen(str1)+strlen(str2)+1];
    strcpy(strcopy(ret, str1), str2);
    delete [] str1;
    return(ret);
}
    
/****************************************************************************/
/* concatinate three strings and delete the first arg */

inline char* concatDel1(char* str1, const char* str2, const char* str3) {

    char* ret = new char[strlen(str1)+strlen(str2)+strlen(str3)+1];
    strcpy(strcopy(strcopy(ret, str1), str2), str3);
    delete [] str1;
    return(ret);
}
    
/****************************************************************************/
/* concatinate four strings and delete the first arg */

inline char* concatDel1(char* str1, const char* str2, const char* str3,
    const char* str4) {

    char* ret = new char[strlen(str1)+strlen(str2)+strlen(str3)+strlen(str4)+1];
    strcpy(strcopy(strcopy(strcopy(ret, str1), str2), str3), str4);
    delete [] str1;
    return(ret);
}
    
#endif
