/**************************************************************************/ 
/*			  Stream.c    Stream.c				  */ 
/**************************************************************************/ 
/*  Copyright (C) 1992,	 Vance Morrison	  */ 
/*  All rights reserved, see Copyright file for more information */ 
 
/* Many, many times, you are manipulating a list (often of characters) one 
   at a time.  Usually you are reading in this stream or writing this
   stream.  This file contains classes that abstract this very common
   case */
 
/* AUTHOR: Vance Morrison  
   DATE	 : 10/23/93		     */ 
/**************************************************************************/ 
 
#include "CharStream.H"

/*************************************************************************/
/* get returns the current value and fetches the next.	It returns 0 if
   cur would be undefined */
const char* InStream:: get() { 
    if (isEnd()) 
    return(0);
    else {
    register const char* ret = &cur();
    next();
    return(ret);
    }
};

/*************************************************************************/
/* reads a whole array of values in */

int InStream::read(char* buff, int n) {
    int ret = 0;
    while (!isEnd() && ret < n) {
    *buff = cur();
    next();
    buff++;
    ret++;
    }
    return(ret);
};

/**************************************************************************/ 
int InStream::
readWhile(char* buff, int n, Bool (*predicate)(const char&, void* context), 
    void* context) {
    int ret = 0;
    while (!isEnd() && ret < n) {
    const char* v = &cur();
    if(!(*predicate)(*v, context))
    break;
    *buff = *v;
    next();
    buff++;
    ret++;
    }
    return(ret);
};

/**************************************************************************/ 
int OutStream::write(const char* val, int n) {
    int ret = 0;
    while (ret < n) {
    if (!put(*val))
    break;
    val++;
    ret++;
    }
    return(ret);
};

