/**************************************************************************/ 
/*			  Stream.h    Stream.h				  */ 
/**************************************************************************/ 
/*  Copyright (C) 1992,	 Vance Morrison	  */ 
/*  All rights reserved, see Copyright file for more information */ 
 
/* Many, many times, you are manipulating a list (often of characters) one 
   at a time.  Usually you are reading in this stream or writing this
   stream.  This file contains classes that abstract this very common
   case */
 
/* AUTHOR: Vance Morrison  
   DATE	 : 10/23/93		     */ 
/**************************************************************************/ 
 
#ifndef _Stream_h 
#define _Stream_h 1 
 
#ifndef Bool
#define Bool int
#endif
#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

/**************************************************************************/ 
/* an input stream that is read once with no bells or whistles.	  */

class InStream { 
public:
    // these first three functions are the only mandatory ones
    virtual Bool isEnd()=0;	    
    // note that the pointer returned from a previous read is NOT 
    // guarenteed to be legal after another.  Copy it if you need it!!
    virtual const char& cur()=0;	// undefined unless isEnd is false
    virtual void next()=0;		// advance one element 

    /*****************************************************************/
    /* these functions are built on top of the ones above (but can be
       overridden if you have a more efficient implmentation) */

    virtual const char* get();	    // returns null on end of stream
    virtual int read(char* buff, int n);  // fill an array, returns count

    /* gets values up to n or while predicate is true, context is 
       passed to the predicate	*/
    virtual int readWhile(char* buff, int n, 
    Bool (*predicate)(const char&, void* context), void* context);
    /*****************************************************************/
    /* shorthand functions */
    const char& operator*()	{   return(cur()); }
    void operator++(int)	{   next();	    }
};

/**************************************************************************/ 
/* an output stream that is written once with no bells or whistles */

class OutStream {
public:
    /* the first two functions are the only mandatory ones */
    virtual Bool put(const char& val)=0; // ret = true means success 
    virtual void done()=0;		   // indicate that we are done

    /* these functions can be built up from the previous two */
    virtual int write(const char* val, int n);

};

#endif
