/**************************************************************************/ 
/*		               NewWhere.h				  */
/**************************************************************************/ 
/*  Copyright (C) 1994,	 Vance Morrison	  */ 
/*  All rights reserved, see Copyright file for more information */ 
 
/* This file defines a new operator that can makes persistance really
   orthogonal.  What is done is that the new operator can be given a
   void* pointer that declares where (persistant or not) the new object
   should be allocated.  The new object will be allocated in the same
   pointer space as the given void* object.  This way if the given
   object is persistant, the new object will be (and thus can be used
   as a component of that object safely).    */

/* AUTHOR: Vance Morrison  
   DATE	 : 2/7/94		     */ 
/**************************************************************************/ 
 
#ifndef _NewWhere_h 
#define _NewWhere_h 1

#include "SpStdlib.H"

#ifdef MULT_HEAPS
/**************************************************************************/ 


    // here is the extended new operator  (where=0 means transient memory)
void* operator new(size_t sz, const void* where, const char* tName,int num=1);

   // returns true if ptr1 has guarenteed shorter lifetime than ptr2
   // or is in the same pointer space as ptr2
int memLessEq(const void* ptr1, const void* ptr2);

#ifdef __GNUG__
#ifndef NO_ARRAY_NEW
inline
void* operator new[](size_t sz, const void* where, const char* tName,int num=1){
    return(operator new(sz, where, tName, num));
}
#endif
#endif

#else	/* MULT_HEAPS */
/**************************************************************************/
/* turn off persistance.  The persistant new operator is just a normal one */
void* operator new(size_t sz, const void*, const char*, int n=1);

static inline int memLessEq(const void*, const void*) { return(1); } 

// __GNUG__ is defined by g++
// _ABIN32 is defined by the SGI compiler if the -n32 option is used
// __rational__ is defined by the Rational Apex compiler named RCC
#if defined (__GNUG__) || defined (_ABIN32) || defined (__rational__) || defined(__IBMCPP__)
#ifndef NO_ARRAY_NEW
#define YES_ARRAY_NEW
#endif
#endif

#ifdef YES_ARRAY_NEW
void* operator new[](size_t sz, const void*, const char*,int n=1);
#endif

#endif	/* MULT_HEAPS */
#endif  /* _NewWhere_h */

