/**************************************************************************/
/*                                Number.c                                */
/**************************************************************************/

/* Miscellaneous Numeric Support functions */

/* AUTHOR: Vance Morrison
     DATE: 7/15/94		*/

/**************************************************************************/

#include "SpString.H"

/**************************************************************************/
/* converts a integer to a string and places it in the buffer 'str',
   it returns true if it succeeded (did not need to truncate the number)  */

int int2str(char* str, int len, int i) {
	
    if (len <= 0)
	return(0);
    --len;		// save room for terminator;
    if (len <= 0) {
	*str = 0;
	return(0);
	}
	
    char buff[64];
    char* ptr = buff;
    
    if (i < 0) {
	*str++ = '-';
	--len;
	i = -i;
	}

    do {
	*ptr++ = '0' + i % 10;
	i = i / 10;
    } while (i != 0);
	
    --ptr;
    while(ptr >= buff) {
	if (len <= 0) {
	   *str = 0;
	   return(0);
	   }
	--len;
	*str++ = *ptr;
	--ptr;
	}
    *str = 0;
    return(1);
}

#include "SpIostream.H"

/**************************************************************************/
/* converts a floating point number to a string and places it in the 
   buffer 'str'.  Note that this is a simplistic routine and does not
   support scientific notation.  returns true if successful */

int float2str(char* str, int len, double d, int afterDot) {

    if (d > 2000000000	|| d < -2000000000) {
	if (len > 0) *str = 0;
	return(0);
	}

    if (!int2str(str, len, (int) d))		//  convert the first part
	return(0);
    int slen = strlen(str);
    str += slen;
    len -= slen;
    --len;
    if (len > 0) {
	*str++ = '.';
	--len;
	}

    if (d < 0) d = -d;
    d = d - (int) d + .000005;	// get just the part after the decimal point
    if (afterDot > 5) afterDot = 5;
    for(int i=0; i < afterDot; i++) {
	if (len <= 0) break;
	d = d * 10;
	*str++ = '0' + ((int) d);
	d = d - ((int) d) - .000045; 
	--len;
	if (d < .00001) break;
	}
    *str = 0;
    return(1);
}
