/*****************************************************************************/ 
/*                             List.c   	                             */
/*****************************************************************************/

/* A simple, linked list implmentation */

/* AUTHOR: Vance Morrison
   DATE:  10/20/93		*/
/*****************************************************************************/

#include "PiList.H"

PiList::~PiList() { if (this != 0) delete(tail); } // note does NOT delete elems

/*****************************************************************************/
int PiList::len() const {

    int ret = 0;
    const PiList* ptr = this;

    while(ptr != 0) {
	ret++;
	ptr = ptr->rest();
	}
    return(ret);
}

/*****************************************************************************/
void destroyStrPiList(PiList* elems) {

    PiList* elemPtr = elems;
    while (elemPtr != 0) {
        char* elem = (char*) elemPtr->first();
        delete [] elem;
        elemPtr = elemPtr->rest();
        }
   delete elems;
}

/*****************************************************************************/
PiList* reverse(PiList* list) {

    PiList* ptr = 0;
    while (list != 0) {
        PiList* tmp = list->tail;
        list->tail = ptr;
        ptr = list;
        list = tmp;
        }
    return(ptr);
}

/*****************************************************************************/
PiList* PiRemove(PiList* lst, void* elem) {

    PiList** ptr = &lst;
    while (*ptr != 0) {
	if ((*ptr)->first() == elem) {
	    PiList* tmp = *ptr;
	    *ptr = (*ptr)->tail;
	    tmp->pop();
	    return(lst);
	    }
        ptr = &(*ptr)->tail;
        }
    return(lst);
}

/*****************************************************************************/
PiList* copy(const PiList* list) {

    PiList* ret = 0;
    while (list != 0) {
	ret = new PiList((void*) list->first(), ret);
	list = list->rest();
	}
    return(reverse(ret));
}

/*****************************************************************************/
PiList* append(PiList* front, PiList* back) {

    if (front == 0)
        return(back);

    PiList** ptr = &front->tail;
    while (*ptr != 0) {
        ptr = &(*ptr)->tail;
        }
    *ptr = back;
    return(front);
}

#ifndef NO_SORT

#include "SpStdlib.H"

/*****************************************************************************/
/* destructively sorts a list.  Mergesort might be faster than using
   an array, but frankly I don't think so */

PiList* sort(PiList* list, int (*cmp)(const void*, const void*)) {

    int i;
  
    int len = list->len();
    if (len == 0)
	return(0);

	// put list in array
    void** array = new void*[len];
    PiList* ptr = list;
    for (i=0; i < len; i++) {
	array[i] = ptr->first();
	ptr = ptr->rest();
	}

    qsort(array, len, sizeof(void*), cmp);

	// put array in list 
    ptr = list;
    for (i=0; i < len; i++) {
	ptr->head = array[i];
	ptr = ptr->rest();
	}
    delete [] array;
    return(list);
}

#endif /* NO_SORT */
