// S_Radar.C

#include "S_Radar.H"
#include "SpIostream.H"
#include "SpString.H"
#include "F_SpProxyItem.H"
#include "E_RadarBroadcastEvent.H"
#include "E_RadarQueryEvent.H"
#include "O_Ship.H"

S_Radar::S_Radar()
  : S_SpHLA("Radar") {
  char s[120];
  sprintf(s, "Radar%d", NODE+LOCAL_ID);
  NAME = strdup(s);
}

void 
S_Radar::InitEvents() {
  int node;
  M_RadarBroadcastEvent msg;

  for (node=0; node<get_N_NODES(); node++){
    msg.ObjType  = get_OBJECT_TYPE();
    msg.LocalId  = get_LOCAL_ID();
    msg.Node     = get_NODE();
    msg.GlobalId = get_GLOBAL_ID();
    ScheduleEvent(0.0, RadarBroadcastEvent, Ship, -1, node, msg);
  }

  ScheduleEvent(25.0, RadarQueryEvent,
                get_OBJECT_TYPE(), get_LOCAL_ID(), get_NODE());
  ScheduleEvent(75.0, RadarQueryEvent,
                get_OBJECT_TYPE(), get_LOCAL_ID(), get_NODE());
  ScheduleEvent(125.0, RadarQueryEvent,
                get_OBJECT_TYPE(), get_LOCAL_ID(), get_NODE());
}

void S_Radar::Debug(const SpSimTime &time) {
  C_XQUEUE *remote_proxies = GetRemoteObjectProxies();
  int num_proxy            = remote_proxies->get_length();
  F_PROXY_ITEM *proxy_item;

  RB_cout << time << " Radar " << get_GLOBAL_ID() << " proxy : ";
  proxy_item = (F_PROXY_ITEM *) remote_proxies->get_top();
  for (int i=0; i<num_proxy; i++){
    RB_cout << proxy_item->GetGlobalId() << ", ";
    proxy_item = (F_PROXY_ITEM *) proxy_item->get_link();
  }
  RB_cout << endl;
}

void 
S_Radar::terminate(double tend){
  C_XQUEUE *remote_proxies = GetRemoteObjectProxies();
  int num_proxy            = remote_proxies->get_length();
  if (num_proxy!=100){
    cout << "ERROR: simobj " << get_NAME() << "(" << get_GLOBAL_ID() << ")" << endl;
    abort();
  }
}
