// S_Ship.C

#include "S_Ship.H"
#include "SpIostream.H"
#include "SpString.H"
#include "SpDistList.H"
#include "SpProxyObjHandle.H"
//#include "E_ShipCreateProxyDistListEvent.H"
//#include "E_ShipRemoveProxyDistListEvent.H"

#include "RB_SpDefineClass.H"

RB_DEFINE_CLASS(SpProxyObjHandle);

S_Ship::S_Ship()
  : S_SpHLA("Ship") {
  char s[120];
  sprintf(s, "Ship%d", NODE+LOCAL_ID);
  NAME = strdup(s);
}

void 
S_Ship::InitEvents(){

  SpObjectHandle handle;

  handle.NodeId   = get_NODE();
  handle.ObjManId = get_OBJECT_TYPE();
  handle.SimObjId = get_LOCAL_ID();

  ScheduleShipCreateProxyDistListEvent(10.0, handle);
  ScheduleShipRemoveProxyDistListEvent(50.0, handle);
  ScheduleShipAddProxyDistListEvent(100.0,   handle);
}

void 
S_Ship::CreateProxyDistList() {
  SpProxyObjHandle *handle;
  const char *list1_name = "List 1";
  const char *list2_name = "List 2";
  SpDistList *distlist_1 = CreateDistList(list1_name);
  SpDistList *distlist_2 = CreateDistList(list2_name);
  int count = 0;

  DestroyDistList(list1_name);
  DestroyDistList(list2_name);

  distlist_1 = GetDistList(list1_name);
  distlist_2 = GetDistList(list2_name);

  if (distlist_1) abort();
  if (distlist_2) abort();

  distlist_1 = CreateDistList(list1_name);
  distlist_2 = CreateDistList(list2_name);

  handle = (SpProxyObjHandle *) Radars.GetFirstElement();
  while (handle) {
    if ((count++ % 2)==0)
      distlist_1->Insert(handle);
    else
      distlist_2->Insert(handle);
    handle = (SpProxyObjHandle *) Radars.GetNextElement();
  }
}

void 
S_Ship::RemoveProxyDistList() {
  SpProxyObjHandle *handle;
  SpDistList *distlist_1 = GetDistList("List 1");
  SpDistList *distlist_2 = GetDistList("List 2");
  int count = 0;
  int removed = 0;

  handle = (SpProxyObjHandle *) Radars.GetFirstElement();
  while (handle) {
    if ((count++%2)==0){
      removed = distlist_1->Remove(handle);
      if (!removed)
        removed = distlist_2->Remove(handle);
      if (!removed) abort();
      RemoveRadars.Insert(handle, handle->GetGlobalId());
    }
    handle = (SpProxyObjHandle *) Radars.GetNextElement();
  }
}

void 
S_Ship::AddProxyDistList() {
  SpProxyObjHandle *handle;
  SpDistList *distlist_1 = GetDistList("List 1");
  SpDistList *distlist_2 = GetDistList("List 2");
  int count = 0;

  while ((handle = (SpProxyObjHandle *) Radars.RemoveFirstElement())){
    if ((count++ % 2)==0)
      distlist_1->Insert(handle);
    else
      distlist_2->Insert(handle);
  }
}

void 
S_Ship::AddRadar(int local_id, int obj_type, int node, int global_id){

  SpProxyObjHandle *handle = new SpProxyObjHandle(local_id, node, obj_type, global_id);
  Radars.Insert(RB_NEW_SpProxyObjHandle(handle), handle->GetGlobalId());
}
