// If you define TEST_THIS_TEST then this test code will remove an
// item front the list that the SpHiGrid returns.  This will make it
// appear that the SpHiGrid SHOULD have returned this item, but failed
// to do so.  I want to make sure that this test code will catch this
// kind of mistake.  I.e., I want to test this test.

//#define TEST_THIS_TEST

#define DEBUG_HIGRID

#include "RB_SpHiGrid.H"
#include "SpRandom.H"
#include "SpAssert.H"
#include <values.h>
#include "SpCrashOnDivZero.H"
#include "SpStrstream.H"
#include "RB_SpList.H"
#include "SpAltMgr.H"
#include "FreeObjectProxy.H"
#include "SpList.H"

/*
** Constructor where free list is set up
*/

FREE_OBJECT_PROXY::FREE_OBJECT_PROXY(int n) {

/*
** Tell the free list base class how many free lists to create
*/

  assert(n == 0);
  set_ntypes(n);

}
#include "RB_SpDefineClass.H"

RB_DEFINE_CLASS(MAIN_PUBSUB_ITEM);

static SpRandom Random;
static int
  wasteful_return,
  no_wasteful,
  number_of_times_hiGrid_and_lookup_do_not_intersect,
  item_res_more_than_all_HiGrid_res,
  item_res_less_than_all_HiGrid_res,
  item_res_between_HiGrid_resolutions,
  hiGrid_lookup_and_at_least_one_item_intersect_but_lookup_does_not_return_everything;

#define db_var(var) cerr << (#var) << "=" << (var) << endl;

// Returns true iff the first argument has a MAIN_PUBSUB_ITEM whose
// UniqueId matches the UniqueId of the second argument.
int Has_UniqueId(RB_SpList &list, MAIN_PUBSUB_ITEM *whatWeAreLookingFor)
{
  for (MAIN_PUBSUB_ITEM *list_item
	 = (MAIN_PUBSUB_ITEM *) list.GetFirstElement();
       list_item;
       list_item = (MAIN_PUBSUB_ITEM *) list.GetNextElement())
    if (          list_item->GetUniqueKey() ==
	whatWeAreLookingFor->GetUniqueKey())
      return 1;

  return 0;
}

int Has_UniqueId(SpList &list, MAIN_PUBSUB_ITEM *whatWeAreLookingFor)
{
  for (MAIN_PUBSUB_ITEM *list_item
	 = (MAIN_PUBSUB_ITEM *) list.GetFirstElement();
       list_item;
       list_item = (MAIN_PUBSUB_ITEM *) list.GetNextElement())
    if (          list_item->GetUniqueKey() ==
	whatWeAreLookingFor->GetUniqueKey())
      return 1;

  return 0;
}

// Sort the arguments in increasing order.
inline void Sort(double &a, double &b)
{
  if (a > b)
  {
    double t = a;
    a = b;
    b = t;
  }
  assert(a <= b);
}

SpHiGrid::SpPubSub GetRandomPubOrSub()
{
  if (Random.GenerateDouble(0, 1) > .5)
    return SpHiGrid::SUBSCRIBER;
  else
    return SpHiGrid::PUBLISHER;
}

void Get(RB_SpHiGrid &hiGrid, RB_SpList *list[2], SpHiGrid::SpPubSub e,
	 MAIN_PUBSUB_ITEM *lookup, SpList *hiGrid_lookup_result)
{
  hiGrid.Get(e, lookup, hiGrid_lookup_result);
#ifdef TEST_THIS_TEST
  if (Random.GenerateDouble(0, 1) > .99)
    hiGrid_lookup_result->RemoveTop();
#endif
  for (MAIN_PUBSUB_ITEM *list_item
	 = (MAIN_PUBSUB_ITEM *) hiGrid_lookup_result->GetFirstElement();
       list_item;
       list_item =
	 (MAIN_PUBSUB_ITEM *) hiGrid_lookup_result->GetNextElement())
  {
    if (!Has_UniqueId(*list[e], list_item))
    {
      cerr << "SpHiGrid returned " << e << " " << (*list_item)
	   << ", but that item was not found in the RB_SpList." << endl;
      assert(0);
    }
  }
}

static SpHiGrid::SpPubSub GetRandomItemDummyDefualt;
// This function randomly chooses one of the two lists passed into the
// first argument, and indicates the index of the list it chose in
// the 2nd argument which is an output argument.  It then chooses a
// random item from that list to return.  If both lists have 0 items,
// then this function just returns NULL.
inline MAIN_PUBSUB_ITEM *GetRandomItem
  (SpList *list[2], SpHiGrid::SpPubSub &pubOrSub = GetRandomItemDummyDefualt)
{
  if (list[SpHiGrid::PUBLISHER]->get_length() == 0)
    pubOrSub = SpHiGrid::SUBSCRIBER;
  else if (list[SpHiGrid::SUBSCRIBER]->get_length() == 0)
    pubOrSub = SpHiGrid::PUBLISHER;
  else 
    pubOrSub = GetRandomPubOrSub();

  void *l = list[pubOrSub]->GetFirstElement();

  for (int i = Random.GenerateInt(1, list[pubOrSub]->GetNumElements());
       i > 1;
       i--)
    l = list[pubOrSub]->GetNextElement();

  return (MAIN_PUBSUB_ITEM*) l;
}

// This function randomly chooses one of the two lists passed into the
// first argument, and indicates the index of the list it chose in
// the 2nd argument which is an output argument.  It then chooses a
// random item from that list to return.  If both lists have 0 items,
// then this function just returns NULL.
inline MAIN_PUBSUB_ITEM *GetRandomItem
  (RB_SpList *list[2], SpHiGrid::SpPubSub &pubOrSub = GetRandomItemDummyDefualt)
{
  if (list[SpHiGrid::PUBLISHER]->get_length() == 0)
    pubOrSub = SpHiGrid::SUBSCRIBER;
  else if (list[SpHiGrid::SUBSCRIBER]->get_length() == 0)
    pubOrSub = SpHiGrid::PUBLISHER;
  else 
    pubOrSub = GetRandomPubOrSub();

  void *l = list[pubOrSub]->GetFirstElement();

  for (int i = Random.GenerateInt(1, list[pubOrSub]->GetNumElements());
       i > 1;
       i--)
    l = list[pubOrSub]->GetNextElement();

  return (MAIN_PUBSUB_ITEM*) l;
}

// These Remove and Add functions perform the same operations on two
// different data structures.  I.e., if we add something to a
// RB_SpHiGrid, then we add the same thing to a RB_SpList pointer
// array.  And if we remove something from a RB_SpHiGrid, then we
// remove the same thing from the RB_SpList pointer array.  Later
// when we access stuff from the RB_SpHiGrid, we use our RB_SpList
// pointer array to calculate what we expect the RB_SpHiGrid to
// return, as a way of testing that the RB_SpHiGrid is working.  Note
// that these functions may not definitely Add or Remove.  For
// example, if there is nothing to remove then the function just makes
// no changes to the SpHiGrid or RB_SpList pointer array.  Or if the
// item does not intersect the RB_SpHiGrid, then we do not add anything
// to it.  Instead, we just return, having changed nothing.

// Returns a successful flag.  I.e. returns true iff it had anything to remove.
inline int Remove(RB_SpHiGrid &hiGrid, RB_SpList *list[2])
{
  if (Random.GenerateDouble(0, 1) < .2)
  {
    // Maybe I should try removing some other id which I know to not
    // be there.  And maybe it should be a positive id or one that was
    // formerly there.
    assert(hiGrid.Remove(SpHiGrid::PUBLISHER , -1) == NULL);
    assert(hiGrid.Remove(SpHiGrid::SUBSCRIBER, -1) == NULL);
    return 0;
  }
  else
  {
    SpHiGrid::SpPubSub pubOrSub;
  
    RB_SpList *normal_list[2];
    normal_list[0] = list[0];
    normal_list[1] = list[1];

    MAIN_PUBSUB_ITEM *l = GetRandomItem(normal_list, pubOrSub);

    if (l == NULL)
      return 0;
    else
    {
      //cout << "Removing UniqueId: " << l->GetUniqueKey() << endl;

      SpList returned_items2;
      Get(hiGrid, list, pubOrSub, l, &returned_items2);
      assert(Has_UniqueId(returned_items2, l));

      list[pubOrSub]->Remove(l);
      MAIN_PUBSUB_ITEM *h = hiGrid.Remove(pubOrSub, l->GetUniqueKey());

      assert(h == l);

      // Make sure that the item we just removed really is no longer
      // in the HiGrid.  We verify this in two different ways.  The
      // first way is to call remove again.  This time it should not
      // find the thing we try to remove.  The HiGrid is documented to
      // return NULL in this case.
      assert(hiGrid.Remove(pubOrSub, l->GetUniqueKey()) == NULL);

      // The second way to verify is to ask for the item using the item.
      SpList returned_items;
      Get(hiGrid, list, pubOrSub, l, &returned_items);
      assert(!Has_UniqueId(returned_items, l));

      RB_DELETE_MAIN_PUBSUB_ITEM(l);

      return 1;
    }
  }
}

int SpRound(const double d)
{
  if (double(-MAXINT) >= d || d >= double(MAXINT))
  {
    cerr << "OverFlow!" << endl;
    assert(0);
  }
  return int(d + .5);
}

void CreateCareLookup(RB_SpHiGrid &hiGrid, const int dimension,
		      SpMainPubSubItem *added_item)
{
  double res;

  if (Random.GenerateDouble(0, 1) < .5)
  {
    // special case of the inserted item having the exact same
    // resolution as one of the resolutions in the HiGrid.
    res = hiGrid.get_Resolution(dimension)
      [Random.GenerateInt(0, hiGrid.get_Nresolution(dimension) - 1)];
  }
  else
  {
    // This is the other case where the inserted item probably
    // does not have the exact same resolution as one of the
    // resolutions in the HiGrid.  It will probably be between two
    // of the HiGrid's resolutions or maybe greater than all of
    // them or less than all of them.  It would not be a good test
    // if all of my inserts were between the HiGrid's 2nd and 3rd
    // resolutions.  I should also test inserting with a resolution
    // between the 3rd and 4th, and the 4th and 5th, etc.
    // I should also test inserting with a resolution less than all
    // of the HiGrid's resolution or greater than all the HiGrid's
    // resolutions.  All of these cases have the same probability
    // of being tested because I assign an int to all of these
    // cases and generate a random int to choose which case to test.

    const int Case = Random.GenerateInt(0, hiGrid.get_Nresolution(dimension));
    double Rhi, Rlo;
	
    if (Case == 0)
    {
      Rlo = hiGrid.get_Resolution(dimension)[Case];
      Rhi = Rlo * 2;
      item_res_more_than_all_HiGrid_res = 1;
    }
    else if (Case == hiGrid.get_Nresolution(dimension))
    {
      Rlo = 0;
      Rhi = hiGrid.get_Resolution(dimension)[Case - 1];
      item_res_less_than_all_HiGrid_res = 1;
    }
    else
    {
      Rlo = hiGrid.get_Resolution(dimension)[Case];
      Rhi = hiGrid.get_Resolution(dimension)[Case - 1];
      item_res_between_HiGrid_resolutions = 1;
    }

    res = Random.GenerateDouble(Rlo, Rhi);
  }

  double lo, hi;
  if (Random.GenerateDouble(0, 1) < .5)
  {
    const int num_bins = SpRound((hiGrid.get_Hi(dimension) -
				  hiGrid.get_Lo(dimension))
				 / res);
    lo = hiGrid.get_Lo(dimension) + Random.GenerateInt(0, num_bins) * res;
    hi = hiGrid.get_Lo(dimension) + Random.GenerateInt(0, num_bins) * res;
    Sort(lo, hi);
  } else {
    const double
      delta = Random.GenerateDouble(0, 10 * res),
      center = Random.GenerateDouble(-2, 2);
	
    lo = center - delta;
    hi = center + delta;
  }

  added_item->SetLHR(dimension, lo, hi, res);
}

double Don_tCareResolution()
{
  const double res = Random.GenerateDouble(-MAXDOUBLE, -MINDOUBLE);
  assert(res < 0);
  return res;
}

double CareResolution()
{
  const double res = Random.GenerateDouble(0, MAXDOUBLE);
  assert(res >= 0);
  return res;
}

void Add(RB_SpHiGrid &hiGrid, RB_SpList *list[2])
{
  for (;;)
  {
    MAIN_PUBSUB_ITEM *added_item = RB_NEW_MAIN_PUBSUB_ITEM();
    added_item->AllocateNumDimensions(hiGrid.get_Ndimensions());
    int added_item_and_hiGrid_intersect_p = 1;
    for (int dimension = 0; dimension < hiGrid.get_Ndimensions(); dimension++)
    {
      if (Random.GenerateDouble(0, 1) < .2)
      {
	// Since it is a don't care, I do not set hi or lo.
	added_item->SetRes(dimension, Don_tCareResolution());
	assert(added_item->GetRes(dimension) < 0);
      }
      else
      {
	CreateCareLookup(hiGrid, dimension, added_item);
	if (added_item->GetLo(dimension) > hiGrid.get_Hi(dimension) ||
	    added_item->GetHi(dimension) < hiGrid.get_Lo(dimension))
	  added_item_and_hiGrid_intersect_p = 0;
	assert(added_item->GetRes(dimension) >= 0);
      }
    }

    SpHiGrid::SpPubSub pubOrSub = GetRandomPubOrSub();

    const int hiGrid_thinks_it_added_p = hiGrid.Add(pubOrSub, added_item);

    assert(hiGrid_thinks_it_added_p == added_item_and_hiGrid_intersect_p);
    
    if (added_item_and_hiGrid_intersect_p) {
      list[pubOrSub]->Insert(added_item);
      //cout << "Adding   UniqueId: " << added_item->GetUniqueKey() << endl;
    }

    // Make sure that the added_item really is in the HiGrid.  Maybe
    // instead of looking for that added_item by using that
    // added_item, I should make a new item to look for it.  This new
    // item might cover the entire HiGrid or I might choose to iterate
    // through the added_item and for each dimension, I could select
    // the end point of the added_item or maybe choose a random place in
    // the added_item.  And maybe have the lookup item have 0 range in each
    // dimension.
    SpList returned_items;
    Get(hiGrid, list, pubOrSub, added_item, &returned_items);
    assert(Has_UniqueId(returned_items, added_item)
	   == added_item_and_hiGrid_intersect_p);

    if (added_item_and_hiGrid_intersect_p)
      break;
    else
      RB_DELETE_MAIN_PUBSUB_ITEM(added_item);
  }
}

void Init(RB_SpHiGrid &hiGrid, const int numDims, const int maxNumRes,
	  const int maxNumBins)
{
  DIM_INFO *DimInfo = new DIM_INFO[numDims];

  int d;
  
  for (d = 0; d < numDims; d++)
  {
    ostrstream os;
    os << "Dim" << d << '!' << ends;
    DimInfo[d].SetDimName(os.str());
    DimInfo[d].SetDimNum(d);
    DimInfo[d].SetNresolutions(d == 0 ? maxNumRes
			       : Random.GenerateInt(1, maxNumRes));
    double
      lo = Random.GenerateDouble(-1, 1),
      hi = Random.GenerateDouble(-1, 1);
    Sort(lo, hi);
    DimInfo[d].SetLo(lo);
    DimInfo[d].SetHi(hi);
    double range = hi - lo;
    for (int r = 0; r < DimInfo[d].GetNresolutions(); r++)
    {
      DimInfo[d].SetResolution(r, range / Random.GenerateDouble(1,
								maxNumBins));
      range = DimInfo[d].GetResolution(r);
    }
  }

  const int SpaceNum = 0;

  hiGrid.init("SpaceName", SpaceNum, numDims, DimInfo);

  delete [] DimInfo;
}

inline void CheckMatchesOneByOne(RB_SpHiGrid &hiGrid, RB_SpList &list,
				 MAIN_PUBSUB_ITEM *lookup,
				 SpList &hiGrid_lookup_result)
{
  int exact_number_of_macthes = 0;
  for (MAIN_PUBSUB_ITEM *list_item
	 = (MAIN_PUBSUB_ITEM *) list.GetFirstElement();
       list_item;
       list_item = (MAIN_PUBSUB_ITEM *) list.GetNextElement())
  {
    int list_item_and_lookup_intersect_p = 1;
    for (int d = 0; d < hiGrid.get_Ndimensions(); d++)
    {
      if (lookup->GetRes(d) >= 0 &&
	  list_item->GetRes(d) >= 0 &&
	  (lookup->GetLo(d) > list_item->GetHi(d) ||
	   lookup->GetHi(d) < list_item->GetLo(d)))
      {
	list_item_and_lookup_intersect_p = 0;
	break;
      }
    }
    if (list_item_and_lookup_intersect_p)
    {
      ++exact_number_of_macthes;
      if (!Has_UniqueId(hiGrid_lookup_result, list_item))
      {
	assert(hiGrid.CheckIntegrity());
	hiGrid.print();
	hiGrid.RecursivePrint();

	db_var(*list_item);
	db_var(*lookup);
	
	cerr << "The RB_SpHiGrid should have returned an item with id "
	     << list_item->GetUniqueKey() << " but as you can see,\n"
	  "that id is not in this list of ids it returned:" << endl;
	
	db_var(hiGrid_lookup_result.GetNumElements());
	for (list_item
	       = (MAIN_PUBSUB_ITEM *) hiGrid_lookup_result.GetFirstElement();
	     list_item;
	     list_item =
	       (MAIN_PUBSUB_ITEM *) hiGrid_lookup_result.GetNextElement())
	  db_var(list_item->GetUniqueKey());

	assert(0);
      }
    }
  }

  if (exact_number_of_macthes >= 1
      && hiGrid_lookup_result.get_length() < list.get_length())
    hiGrid_lookup_and_at_least_one_item_intersect_but_lookup_does_not_return_everything = 1;
    
  // The RB_SpHiGrid class is a conservative data structure.  That means
  // that if it should return something, it definitely will, but it
  // sometimes return extra things just to make sure that it returns
  // everything it should return.  Users have to check to see if each
  // thing it returns really does intersect.  You might say, "What good
  // is it if it returns things it should NOT return?"  The answer is that
  // it often avoids returning many things that do not need to be
  // returned.  This saves you the time of checking if those things
  // really intersect.  I want to make sure I have some test cases
  // that are conservative and wasteful in this sense, but I also want
  // some test cases that just happen to not be wasteful.  Therefore, I
  // keep track of such cases here:
  if (hiGrid_lookup_result.get_length() > exact_number_of_macthes)
  {
    ++wasteful_return;
  }
  else if (hiGrid_lookup_result.get_length() == exact_number_of_macthes)
    ++no_wasteful;
  else
  {
    assert(hiGrid_lookup_result.get_length() < exact_number_of_macthes);
    assert(0);
  }
}

inline void Lookup(RB_SpHiGrid &hiGrid, RB_SpList *list[2])
{
  MAIN_PUBSUB_ITEM *lookup, created;
  
  RB_SpList *normal_list[2];
  normal_list[0] = list[0];
  normal_list[1] = list[1];

  int hiGrid_and_lookup_intersect_p = 1;

  if (Random.GenerateDouble(0, 1) < .5
      || (lookup = GetRandomItem(normal_list)) == NULL)
  {
    lookup = &created;
    created.AllocateNumDimensions(hiGrid.get_Ndimensions());
    for (int d = 0; d < hiGrid.get_Ndimensions(); d++)
    {
      if (Random.GenerateDouble(0, 1) < .5)
      {
	double
	  lo = Random.GenerateDouble(-2, 2),
	  hi = Random.GenerateDouble(-2, 2);

	Sort(lo, hi);

	created.SetLo(d, lo);
	created.SetHi(d, hi);
	created.SetRes(d, CareResolution());
	if (lookup->GetLo(d) > hiGrid.get_Hi(d) ||
	    lookup->GetHi(d) < hiGrid.get_Lo(d))
	  hiGrid_and_lookup_intersect_p = 0;
      } else {
	created.SetRes(d, Don_tCareResolution());
      }
    }
  }

  const SpHiGrid::SpPubSub pubOrSub = GetRandomPubOrSub();
  
  SpList hiGrid_lookup_result;
  assert(hiGrid_lookup_result.get_length() == 0);
  Get(hiGrid, list, pubOrSub, lookup, &hiGrid_lookup_result);
  
  if (hiGrid_and_lookup_intersect_p)
  {
    CheckMatchesOneByOne
      (hiGrid, *list[pubOrSub], lookup, hiGrid_lookup_result);
  }
  else
  {
    assert(hiGrid_lookup_result.get_length() == 0);
    ++number_of_times_hiGrid_and_lookup_do_not_intersect;
  }
    
  if (Random.GenerateDouble(0, 1) < .1)
  {
    MAIN_PUBSUB_ITEM mathces_everything;
    mathces_everything.AllocateNumDimensions(hiGrid.get_Ndimensions());
    for (int d = 0; d < hiGrid.get_Ndimensions(); d++)
    {
      mathces_everything.SetLo(d, -MAXDOUBLE);
      mathces_everything.SetHi(d, MAXDOUBLE);
    }

    const int old = hiGrid_lookup_result.get_length();
    Get(hiGrid, list, pubOrSub, &mathces_everything, &hiGrid_lookup_result);
    if (hiGrid_lookup_result.get_length() - old
	!= list[pubOrSub]->get_length())
    {
      hiGrid.print();
      db_var(pubOrSub);
      db_var(hiGrid_lookup_result.get_length());
      db_var(old);
      db_var(list[pubOrSub]->get_length());
      hiGrid.Get(pubOrSub, &mathces_everything, &hiGrid_lookup_result);
      assert(0);
    }
  }
}

void Test(const int numDims, const int maxNumRes, const int maxNumBins,
	  const int maxNumItemsToAdd, const int numCyclesToTest,
	  const float probabilityOfAddingToHiGrid)
{
  wasteful_return                                        = 0;
  no_wasteful                                            = 0;
  number_of_times_hiGrid_and_lookup_do_not_intersect    = 0;
  item_res_more_than_all_HiGrid_res                      = 0;
  item_res_less_than_all_HiGrid_res                      = 0;
  item_res_between_HiGrid_resolutions                    = 0;
  
  RB_SpHiGrid hiGrid;
  Init(hiGrid, numDims, maxNumRes, maxNumBins);

  RB_SpList pub_sub_item_objects[2];
  RB_SpList *pub_sub_items[2];
  pub_sub_items[0] = pub_sub_item_objects;
  pub_sub_items[1] = pub_sub_item_objects + 1;
  
  const float
    probability_of_removing_from_HiGrid = 1 - probabilityOfAddingToHiGrid;
  float current_P_add = probabilityOfAddingToHiGrid;
  int current_num_cycles = 0, current_target = maxNumItemsToAdd;
  // Iterations of this loop usually add items to the hiGrid, but the
  // loop iteration might also sometimes remove items.  Since it adds
  // more often than removes, the number of items currently in the
  // hiGrid will tend to grow quickly over the long term of many
  // iterations.  When it reaches a maximum that is passed to this
  // function, it will then make it so that removing is more likely
  // than adding.  This will cause the number of items in the hiGrid
  // to decrease quickly over the long term.  When it gets down to 0
  // items, then we switch back to the old value of the probability of
  // adding and start growing again, and when we reach the maximum, we
  // start shrinking again, etc.  This goes on for numCyclesToTest.
  //
  // Not only does each iteration add, but each iteration also calls
  // GetPublishers or GetSubscribers, and verifies that the result is
  // correct.
  for (;;)
  {
    //#ifdef __INSURE__
    //hiGrid.RecursivePrint();
    //#endif

    const double d = Random.GenerateDouble(0, 1);
    if (d < .1)
      SpCurrentAltMgr->Rollback();
    else if (d < .2)
      SpCurrentAltMgr->Cleanup();
    
    Lookup(hiGrid, pub_sub_items);

    if (Random.GenerateDouble(0, 1) <= current_P_add
	|| (!Remove(hiGrid, pub_sub_items)))
      Add(hiGrid, pub_sub_items);

    if (pub_sub_items[SpHiGrid::PUBLISHER ]->GetNumElements() +
	pub_sub_items[SpHiGrid::SUBSCRIBER]->GetNumElements()
	== current_target)
    {
      if (current_target == 0)
      {
	if ((++current_num_cycles) == numCyclesToTest)
	  break;
	current_target = maxNumItemsToAdd;
	current_P_add = probabilityOfAddingToHiGrid;
      }
      else
      {
	current_target = 0;
	current_P_add = probability_of_removing_from_HiGrid;
      }
    }
  }

  assert(wasteful_return == 0);

  assert(no_wasteful > 0);
  assert(number_of_times_hiGrid_and_lookup_do_not_intersect > 0);
  
  assert(item_res_more_than_all_HiGrid_res);
  assert(item_res_less_than_all_HiGrid_res);
  if (maxNumRes >= 2)
    assert(item_res_between_HiGrid_resolutions);
  
  for (int d = 0; d < numDims; d++)
  {
    delete [] hiGrid.get_DimName(d);
  }

  SpCurrentAltMgr->Cleanup();
}

int main()
{
  crash_on_divide_by_zero();

  // I do this so it will call the constructor since it is the
  // constructor that initializes some static variables in this class.
  // Another purpose of this is to delete memory which happens when this
  // object's destructor is called.  It is implicitly called at the end
  // of this function
  SpAltMgr AltMgr;
  SpCurrentAltMgr = &AltMgr;

  SpCurrentAltMgr->SetOptimistic();
  
  SpMainPubSubItem::InitSpMainPubSubItemClass(0, 1);

  hiGrid_lookup_and_at_least_one_item_intersect_but_lookup_does_not_return_everything = 0;

  Test(3, 3, 4,  3, 2,  1);
  Test(3, 3, 4,  3, 1, .8);
#ifndef __INSURE__
  Test(2, 2, 7, 20, 1, .9);
  Test(1, 2, 3, 30, 4, .8);
  Test(5, 4, 4,  8, 3,  1);
  assert(hiGrid_lookup_and_at_least_one_item_intersect_but_lookup_does_not_return_everything);
#endif
   
  {
    // The following code tests whether it is possible to destruct a
    // hiGrid with an item still in it.
    RB_SpList pub_sub_item_objects[2];
    RB_SpList *pub_sub_items[2];
    pub_sub_items[0] = pub_sub_item_objects;
    pub_sub_items[1] = pub_sub_item_objects + 1;

    RB_SpHiGrid hiGrid;
    Init(hiGrid, 2, 2, 3);
    Add(hiGrid, pub_sub_items);
    for (int d = 0; d < 2; d++)
      delete [] hiGrid.get_DimName(d);

    delete (SpMainPubSubItem*) pub_sub_items[SpHiGrid::PUBLISHER ]->pop_top();
    delete (SpMainPubSubItem*) pub_sub_items[SpHiGrid::SUBSCRIBER]->pop_top();
  
    SpCurrentAltMgr->Cleanup();
  }

  return 0;
}
