// Program for running point tests on SpSet class

#include "SpStdlib.H"
#include "SpSignal.H"
#include "SpIostream.H"
#include "SpSetTest.H"
#include "SpSet.H"

int main()
{

  SpSetTest noargconstructor("no-argument constructor");
  SpSet a;
  noargconstructor.Verify(noargconstructor.Success());

  SpSetTest inttest("Insert(int,char *) method");
  a.Insert(5,"myint");
  inttest.Verify(a.GetInt("myint")==5);

  SpSetTest floattest("Insert(double,char *) method");
  a.Insert(5.1,"myfloat");
  floattest.Verify(a.GetFloat("myfloat")==5.1);

  SpSetTest stringtest("Insert(char *,char*) method");
  a.Insert("some old string","mystring");
  stringtest.Verify(!strcmp(a.GetString("mystring"),"some old string"));
 
  SpSetTest arraytest("AddArray(char *,ElementType type,int Length) method");
  a.AddArray("myarray",INT_ET,10);
  arraytest.Verify(arraytest.Success());

  SpSetTest getarraytest("GetArray(char *) method");
  SpSetArray *myarray = a.GetArray("myarray");
  int *intarray = *myarray;
  intarray[5] = 21;
  getarraytest.Verify(10==myarray->GetLength());

  SpSetTest defaultinttest("GetInt(char *, int, int &) method");
  int getintstatus;
  defaultinttest.Verify(21 == a.GetInt("dontexist",21,getintstatus));

  SpSetTest embedtest("Embed(char *) method");
  SpSet *mysubset = a.Embed("mysubset");
  embedtest.Verify(embedtest.Success());

  SpSetTest subsettest("Insert(int,char *) method on subset");
  mysubset->Insert(23,"myint");
  subsettest.Verify(23==mysubset->GetInt("myint"));

//   SpSetTest setnonint("Insert(int,char *) method on non-int");
//   if (sigsetjmp(SpSetTestenv,1)) {
//     cout << "abort() called as expected..." << endl;
//   }
//   else {
//     mysubset->Insert(24.3,"myint");
//     cout << "Got to line before Verify() call " << endl;
//     setnonint.Verify(setnonint.Success());
//   }
//   setnonint.Verify(!setnonint.Success());

  SpSetTest setlogical("Insert(bool,char *) method");
  if (sigsetjmp(SpSetTestenv,1)) {
    cout << "abort() called in last test" << endl;
  }
  else {
    mysubset->Insert(true,"mylogical");
    setlogical.Verify(mysubset->GetLogical("mylogical")==true);

  }
  

  SpSetTest setpack("Pack() method");
  if (sigsetjmp(SpSetTestenv,1)) {
    cout << "abort() called in last test" << endl;
  }
  else {
    char *buffer = a.Pack();
    SpSet b(buffer);
    SpSet *mss = b.GetSet("mysubset");

    
    setpack.Verify(setpack.Success());
    SpSetTest setrestore("Restore() method");
    setrestore.Verify((int)mss);
    free(buffer);
  }

  if (SpSetTest::Failed()) {
    cerr << "Failed: SpSetTest" << endl;
  }
  else {
    cerr << "Passed: SpSetTest" << endl;
  }
 
  return 0;
}
