// Defines an SpSetTest class to assist point tests 
// of SpSet container class

#ifndef _SpSetTest_H
#define _SpSetTest_H

#include "SpSetjmp.H"
#include "SpString.H"
#include "SpIostream.H"
#include "SpStdio.H"
#include "SpStdlib.H"
#include "SpErrno.H"
#include "SpSignal.H"

#define FAILURE 0
#define SUCCESS 1

sigjmp_buf SpSetTestenv;

class SpSetTest
{
public:
  SpSetTest(const char *testName,int initStatus=SUCCESS) {
    LastTestStatus = initStatus;
    CurrentTestObj = this;
    if (CurrentTestName) {
      free(CurrentTestName);
    }
    CurrentTestName = strdup(testName);

    NTestObjs++;
    if (SIG_ERR == signal(SIGSEGV, err_handler)) {
      cerr << "Couldn't set signal handler for SIGSEGV" << endl;
      perror("SpSetTest");
      abort();
    }  
    if (SIG_ERR == signal(SIGABRT, err_handler)) {
      cerr << "Couldn't set signal handler for SIGABRT" << endl;
      perror("SpSetTest");
      abort();
    }
  }

  virtual ~SpSetTest() {
    NTestObjs--;
    if (!NTestObjs) {
      cout << NPassed << " tests passed, ";
      cout << NFailed << " failed ";
      cout << "out of " << NPassed + NFailed << " attempted." << endl;
      free(CurrentTestName); 
      if (NFailed) {
         exit(1);
      }
      else {
         exit(0);
      }
    }

  }

  

  void Verify(int condition) {

    if (condition) {
      LastTestStatus = 1;
      NPassed++;
      cout << "Passed: " << CurrentTestName << endl;
      do_specific_behavior(1);
    }
    else {
      LastTestStatus = 0;
      NFailed++;
      cout << "Failed: " << CurrentTestName << endl;
      do_specific_behavior(0);
    }
  }

  inline int WillAbort() {
    cout << "willabort() first go-around..." << endl;
    if (sigsetjmp(SpSetTestenv,1)) {
      return 1;
    }
    else {
      return 0;
    }
  }

  virtual int Success() {
    return LastTestStatus;
  }

  static int Passed() { return NPassed; }
  static int Failed() { return NFailed; } 

protected:
  static char *CurrentTestName;
  static SpSetTest *CurrentTestObj;
  static int LastTestStatus;
  static int NPassed;
  static int NFailed;
  static int NTestObjs;

  // The argument to be passed  this method is nonzero to indicate success, 
  // zero for failure
  virtual void do_specific_behavior(int) {} 
  static void err_handler(int signum) {
    LastTestStatus = FAILURE;
    if (signum==SIGSEGV) cerr << "Failed: " << CurrentTestName << endl;
    CurrentTestObj->do_specific_behavior(0);
    if (signum == SIGABRT) siglongjmp(SpSetTestenv,1);
  }

};

int SpSetTest::NPassed = 0;
int SpSetTest::NFailed = 0;
int SpSetTest::NTestObjs = 0;
int SpSetTest::LastTestStatus = SUCCESS;
char *SpSetTest::CurrentTestName = NULL;
SpSetTest *SpSetTest::CurrentTestObj = NULL;

#endif // #ifndef _SpSetTest_H
