// S_TestRB_Container.C source file

#include "S_TestRB_Container.H"

#define TEST_ARRAY
#define TEST_TREE
#define TEST_HASH
#define TEST_PRIORITY

//#define NOISY

#define N_REMOVES 5

RB_DEFINE_CLASS_PRIMATIVE(int);

/******************************************************************************
 * DoTest - does the test to make sure that the rollbackable dynamic array    *
 *          works                                                             *
******************************************************************************/
void S_TestRB_Container::DoTest(double) {
  int *element;
  int *e;
  int i;
  unsigned int seed;
  int key;

#ifdef NOISY
  RB_cout << "Testing RB_Containers at time " << time << endl;
#endif

  element = RB_NEW_int();
  *element = RB_Counter;
/*
** Test the dynamic array
*/

#ifdef TEST_ARRAY

#ifdef NOISY
  RB_cout << "Testing Dynamic Array" << endl;
#endif

  if ((void *)DynArray[RB_Counter] != NULL) {
    RB_cout << "Error, DynArray[RB_Counter] != NULL" << endl;
  }

  DynArray[RB_Counter] = element;

  for (i=0; i<=RB_Counter; i++) {
    e = (int *)(void *)DynArray[i];
    if (*e != i) {
      ErrorFlag = 1;
      RB_cout << "Error, bad DynArray element" << endl;
    }
  }

#ifdef NOISY
  RB_cout << "Done" << endl;
#endif

#endif

/*
** Test the binary tree
*/

#ifdef TEST_TREE

#ifdef NOISY
  RB_cout << "Testing Tree" << endl;
#endif

  seed = *element;
  srand(seed);
  if (RB_Counter % 2) {
    BinaryTree.SetSplayTreeMode();
  }else{
    BinaryTree.SetBalancedTreeMode();
  }

  srand(RB_Counter);
  if (seed)
    key = rand();
  else
    key = 0;
  BinaryTree.Insert(element, key);
  if (!BinaryTree.CheckIntegrity()) {
    RB_cout << "Error, bad integrity for binary tree" << endl;
    ErrorFlag = 1;
  }

  e = (int *)BinaryTree.Find(key);
  if (!e) {
    ErrorFlag = 1;
    RB_cout << "Error, could not find element after inserting in tree"
	    << endl;
  }

  e = (int *)BinaryTree.GetFirstElement();
  if (*e != 0) {
    RB_cout << "Error, bad first element in tree" << endl;
    ErrorFlag = 1;
  }

  i = 0;
  while (e) {
    if (*e > RB_Counter) {
      ErrorFlag = 1;
      RB_cout << "Error, bad element in binary tree" << endl;
    }
    i++;
    e = (int *)++BinaryTree;
  }

  if (i != RB_Counter+1) {
    ErrorFlag = 1;
    RB_cout << "Error, bad number of elements in binary tree" << endl;
  }

  e = (int *)BinaryTree.Find(key);
  if (e) {
    if (*e != RB_Counter) {
      ErrorFlag = 1;
      RB_cout << "Error, found wrong element in tree" << endl;
    }
  }else{
    ErrorFlag = 1;
    RB_cout << "Error, could not find element in tree" << endl;
  }

  double dkey;
  e = (int *)BinaryTree.RemoveFirstElement(dkey);
  if (!e) {
    ErrorFlag = 1;
    RB_cout << "Error, could not remove first element" << endl;
  }else{
    BinaryTree.Insert(e, dkey);
  }
  
  e = (int *)BinaryTree.Remove(key);
  if (!e) {
    ErrorFlag = 1;
    RB_cout << "Error, could not remove " << key << endl;
  }else{
    if (BinaryTree.Remove(key)) {
      ErrorFlag = 1;
      RB_cout << "Error, was able to remove an element twice" << endl;
    }
    BinaryTree.Insert(e, key);
  }


  if (!BinaryTree.CheckIntegrity()) {
    RB_cout << "Error, bad integrity at the end of the binary tree test"
	    << endl;
    ErrorFlag = 1;
  }

#ifdef NOISY
  RB_cout << "Done" << endl;
#endif

#endif

/*
** Test the hash tree
*/

#ifdef TEST_HASH

#ifdef NOISY
  RB_cout << "Testing Hash" << endl;
#endif

  seed = *element;
  srand(seed);
  if (RB_Counter % 2) {
    HashTree.SetSplayTreeMode();
  }else{
    HashTree.SetBalancedTreeMode();
  }

  srand(RB_Counter);
  if (seed)
    key = rand();
  else
    key = 0;
  HashTree.Insert(element, key);

  e = (int *)HashTree.Remove(key);
  if (e!= element) {
    ErrorFlag = 1;
    RB_cout << "Error, could not remove element from hash" << endl;
  }else{
    HashTree.Insert(element, key);
  }

  e = (int *)HashTree.GetFirstElement();
  i = 0;
  while (e) {
    if (*e > RB_Counter) {
      ErrorFlag = 1;
      RB_cout << "Error, bad element in hash tree" << endl;
    }
    i++;
    e = (int *)++HashTree;
  }

  if (i != RB_Counter+1) {
    ErrorFlag = 1;
    RB_cout << "Error, bad traversal of hash tree" << endl;
    RB_cout << "Should have traversed " << RB_Counter+1
	    << ", but only traversed " << i << endl;
    RB_cout << "NumElements = " << HashTree.GetNumElements() << endl;
  }

  e = (int *)HashTree.Find(key);
  if (e) {
    if (*e != RB_Counter) {
      ErrorFlag = 1;
      RB_cout << "Error, found wrong element in hash tree" << endl;
    }
  }else{
    ErrorFlag = 1;
    RB_cout << "Error, could not find element in hash tree" << endl;
  }

  if (!HashTree.CheckIntegrity()) {
    cout << "Error, bad integrity for Hash Tree" << endl;
    ErrorFlag = 1;
  }

#ifdef NOISY
  RB_cout << "Done" << endl;
#endif

#endif

/*
** Test the priority tree
*/

#ifdef TEST_PRIORITY

#ifdef NOISY
  RB_cout << "Testing Priority" << endl;
#endif

  seed = RB_Counter;
  srand(seed);
  if (seed)
    key = rand();
  else
    key = 0;
  double priority = key;
  double p;

  PriorityTree.Insert(element, priority);

  if (PriorityTree.GetCurrentPriority() != 0) {
    ErrorFlag = 1;
    RB_cout << "Bad first priority = " << PriorityTree.GetCurrentPriority()
	    << endl;
  }

  int nTest;
  int *eTest[N_REMOVES];
  double pTest[N_REMOVES];
  void *rTest[N_REMOVES];

  nTest = 0;
  while (nTest < N_REMOVES) {
    e = (int *)PriorityTree.Remove(p);
    eTest[nTest] = e;
    pTest[nTest] = p;
    if (e == NULL) break;

#ifdef NOISY
    RB_cout << "Priority[" << nTest << "] = " << pTest[nTest]
	    << ", Value = " << *e << endl;
#endif
    
    nTest++;
  }
  
  for (i=0; i<nTest; i++) {
    if (eTest[i] == NULL) {
      ErrorFlag = 1;
      cout << "Something is really wrong, eTest[i] = NULL" << endl;
    }
    e = eTest[i];
    p = pTest[i];
    rTest[i] = PriorityTree.Insert(e, p);
  }

  e = (int *)PriorityTree.Find(pTest[nTest-1]);
  if (!e) {
    RB_cout << "Error, did not successfully find element" << endl;
    ErrorFlag = 1;
  }

  int status = PriorityTree.Retract(rTest[nTest-1]);
  if (status != 1) {
    RB_cout << "Error, could not retract element" << endl;
    ErrorFlag = 1;
  }

  e = (int *)PriorityTree.Find(pTest[nTest-1]);
  if (e) {
    RB_cout << "Error, did not successfully retract element" << endl;
    ErrorFlag = 1;
  }

  PriorityTree.Insert(eTest[nTest-1], pTest[nTest-1]);

  if (!PriorityTree.CheckIntegrity()) {
    RB_cout << "Error, bad integrity for Priority Tree" << endl;
    ErrorFlag = 1;
  }

#ifdef NOISY
  RB_cout << "Done" << endl;
#endif

#endif

}
