// Test of SpParser class for parsing speedes par files.
// Expects to see certain arbitrary elements in its input file. 

#include "SpIostream.H"
#include "SpParser.H"
#include "SpSet.H"

int main(/*int argc,char **argv*/) 
{
  istream *inputfile;
//   if (argc>1) {
//     inputfile = new ifstream(argv[1]);
//     if (inputfile->bad()) {
//       inputfile = &cin;
//     }
//   }
//   else {
    //    inputfile = &cin;
    inputfile = new ifstream("test.par");
//   }

  SpParser parser(inputfile);
  SpSet *inputset = parser.Parse();
  int nelements = inputset->GetNumElements();
  cout << "Number of elements in inputset = " << nelements << endl;
  SpSet *parameterset = inputset->GetSet("parameters");
  if (parameterset) {
    cout << "FOUND set named \"parameters\" in top level set" << endl;
    int getintstatus;
    int retval = parameterset->GetInt("multiroute_thresh",23,
				      getintstatus);
    if (getintstatus) {
      cout << "Couldn't find integer \"multiroute_thresh\" ";
      cout << "in set \"parameters\"" << endl;
      exit(-1);
    }
    else {
      cout << "Value of integer \"multiroute_thresh\" ";
      cout << "in set \"parameters\" = " << retval << endl;
    }

    SpSetArray *intarray = parameterset->GetArray("intarray");
    if (intarray) {
      if ((int)(*intarray)[10] == 320) {
	cout << "intarray[10] = 320, as expected" << endl;
      }
      else {
	cout << "Unexpected value " <<  (int)(*intarray)[10] << " for ";
	cout << "intarray[10]" << endl;
	exit(-1);
      }
    }
    else {
      cout << "Couldn't find intarray in parameterset" << endl;
      exit(-1);
    }

    SpSet *gvtref = parameterset->GetSet("gvt_parameters");
    if (gvtref) {
      cout << "Retrieved gvt_parameters reference from ";
      cout << "parameters" << endl;
      int refstatus;
      int ngvt = gvtref->GetInt("Ngvt",5,refstatus);
      if (refstatus) {
	cout << "Failed to access Ngvt member from reference ";
	cout << "to set gvt_parameters" << endl;
	exit(-1);
      }
      else {
	if (ngvt == 1000) {
	  cout << "Ngvt in refset had expected value " << ngvt << endl;
	}
	else {
	  cout << "Ngvt in refset had unexpected value " << ngvt << endl;
	  exit(-1);
	}
      }
      int hlaint = gvtref->GetInt("hlaint",42,refstatus);
      if (refstatus) {
	cout << "Failed to access hlaint member from reference ";
	cout << "to set gvt_parameters (which in turn came ";
	cout << "from included set HLA" << endl;
	exit(-1);
      }
      else {
	if (hlaint == 53401) {
	  cout << "hlaint in refset (from included set HLA) ";
	  cout << "had expected value " << hlaint << endl;
	}
	else {
	  cout << "hlaint in refset (from included set HLA) had ";
	  cout << "unexpected value " << hlaint << endl;
	  exit(-1);
	}
      }
    }
  }
  else {
    cout << "COULDN'T find set named \"parameters\" in top level set";
    cout << endl;
    exit(-1);
  }
  
  int gotthreelevel=0;
  int thirdleveli;
  SpSet *SpeedesServerSet = inputset->GetSet("SpeedesServer");
  if (SpeedesServerSet) {
    SpSet *HostRouterSet = SpeedesServerSet->GetSet("HostRouter");
    if (HostRouterSet) {
      SpSet *thirdlevelset = HostRouterSet->GetSet("threelevel");
      if (thirdlevelset) {
	gotthreelevel = 1;
	thirdleveli = thirdlevelset->GetInt("i");
      }
    }
  }
  else {
    cout << "Couldn't find set \"SpeedesServer\" in toplevel" << endl;
    exit(-1);
  }



  if (gotthreelevel) {
    cout << "Got threelevel set" << endl;
    cout << "value of integer i in threelevel set = " << thirdleveli << endl;
  }
  else {
    cout << "Couldn't find threelevel set" << endl;
    exit(-1);
  }
  
  cout << "\n\nPassed: SpParser test\n" << endl;
  return 0;
}
