// S_LogSemC.C source file

#include "S_LogSem.H"
#include "O_LogSem.H"
#include "SpProc.H"
#include "print.H"

RB_DEFINE_CLASS(SpProcStackElement)

void S_LogSem::LogSemControllerProcess(const int proId) // ID of control proc
P_VAR
   P_LV(O_LogSem*, manager)
   P_LV(int, noisy)
   P_LV(double, end)
   P_LV(int, objId) // ID of Logical Semaphore simobj
   P_LV(int, semId) // ID of Logical Semaphore being requested
   P_LV(int, open)
   P_LV(int, waitValue)
P_BEGIN(1)
   manager   = (O_LogSem *) get_manager();
   noisy     = manager->noisy;
   end       = manager->end;
   objId     = get_LOCAL_ID();
   if (manager->numSemsPerObject){
     semId     = proId % manager->numSemsPerObject;
   }
   else{
     cout<<" Bad semid since manager->numSemsPerObject is 0 Choosing a "
	 << "semid of 0"<<endl;
     semId = 0;
   }
   open      = 0;
   waitValue = manager->waitValueC + proId;
     
   PRINT("****** BEGIN: LogicalSem controller ");

// There used to be a "while" loop here, but Insure++ cannot handle
// some kinds of conditional expressions being in "while" or "for"
// loops.  But it can handle those expressions being in if statements.
// Therefore I made the loop condition be always true, and moved the
// condition inside the loop into an if statement, but of course I had
// to prepend a boolean "not" symbol of the exclamation point to the
// expression.  Note that the kinds of expressions Insure++ cannot
// handle are demonstrated by a bug report I sent to parasoft.
   for (;;)
   {
     if (!(GetTime() <= end)) break;
      LS[semId].Toggle();
      open = 1 - open;
      if (open) PRINT("opened"); else PRINT("closed");
      WAIT (1, waitValue);
   }
P_END

void S_LogSem::LogSemRequesterProcess(const int proId) // ID of control proc
P_VAR
   P_LV(O_LogSem*, manager)
   P_LV(int, noisy)
   P_LV(double, end)
   P_LV(int, success)
   P_LV(int, objId) // ID of Logical Semaphore simobj
   P_LV(int, semId) // ID of Logical Semaphore being requested
   P_LV(int, timeOutValue)
   P_LV(int, waitValue)
P_BEGIN(2)
   manager      = (O_LogSem *) get_manager();
   noisy        = manager->noisy;
   end          = manager->end;
   objId        = get_LOCAL_ID();
   semId        = proId % manager->numSemsPerObject;
   timeOutValue = manager->timeOutValue;
   waitValue    = manager->waitValueR + proId;
   PRINT("****** BEGIN: LogicalSem requester ");
// There used to be a "while" loop here, but Insure++ cannot handle
// some kinds of conditional expressions being in "while" or "for"
// loops.  But it can handle those expressions being in if statements.
// Therefore I made the loop condition be always true, and moved the
// condition inside the loop into an if statement, but of course I had
// to prepend a boolean "not" symbol of the exclamation point to the
// expression.  Note that the kinds of expressions Insure++ cannot
// handle are demonstrated by a bug report I sent to parasoft.
   for (;;)
   {
      if (!(GetTime() <= end)) break;
      PRINT ("waiting for gate  ");
      WAIT_FOR_LOGICAL(1, LS[semId], timeOutValue, success);
      if (success) PRINT ("going through gate");
      else PRINT ("gave up waiting for gate");
      WAIT(2, waitValue);
   }
P_END
  
