// config.C file

#include "config.H"
#include "SpAssert.H"

int CONFIG::Replication = 0;

void Exit(char *msg)
{
   cout << "!!!! procMod Terminal error: " << msg << endl;
   exit(-1);
}

CONFIG::CONFIG(): Parser("test.par"), ConfigureTestSuccessful(0)
{
   if (Replication == 0)
   {
      if (!Parser.MoveIn("Parameters")) Exit("No Parameters Group");
      if (!(Replication = Parser.GetInt("Replication")))
	Exit("No Replication number in Parameters Group");
      cout << "Executing Replication " << Replication << endl;
      Parser.MoveOutCompletely();
   }
   if (!Parser.MoveIn("Tests")) Exit("No Tests Group") ;
   force_global_noisy = Parser.GetLogical("force_global_noisy");
   force_global_quiet = Parser.GetLogical("force_global_quiet");
}

int CONFIG::ConfigureTest(char *name, int &numSimObj,
			  double &start, double &end)
{
   int found=0;

   Parser.MoveOutCompletely();
   Parser.MoveIn("Tests");
   if (!Parser.MoveToFirstSetInGroup("TEST"))
     Exit("No tests selected in test.par file");
   do
   {
      if (!strcmp (name, Parser.GetSetName()))
      {
	 found = 1;
	 break;
      }
   } while (Parser.MoveToNextSetInGroup());
   if (!found) return (ConfigureTestSuccessful = 0);

   strcpy (TestName, name);
   if (!(numSimObj = Parser.GetInt("numSimObj")))
     Exit("Must have non-zero value for 'numSimObj' in test.par file");
   if (!(end = Parser.GetFloat("end")))
     Exit("Must have non-zero value for 'end' in test.par file");
   start = Parser.GetFloat("start");
   return (ConfigureTestSuccessful = 1);
}

int CONFIG::ConfigureSubtest(char *name)
{
   if (!ConfigureTestSuccessful) Exit("CONFIG error. Must "
				      "successfully call ConfigureTest() "
				      "before ConfigureSubtest()");
   
   Parser.MoveOutCompletely();
   if (!Parser.MoveIn(TestName))
   {
      char str[64];
      sprintf(str, "No '%s' subtest in test.par file\n", TestName);
      Exit(str);
   }
   Parser.MoveToFirstSetInGroup(TestName);
   do
   {
      if (!strcmp (name, Parser.GetSetName())) return 1;
   } while (Parser.MoveToNextSetInGroup());
   return 0;
}

int CONFIG::GetSubtestParameter(char *name)
{
   return Parser.GetInt(name);
}

int CONFIG::GetNoisy()
{
   if (force_global_noisy) return 1;
   if (force_global_quiet) return 0;
   return Parser.GetLogical("noisy");
}

int CONFIG::GetReplication()
{
   return Replication;
}
