// S_TestCancel.C source file

//#define NOISY_TEST_CANCEL

#include "S_TestCancel.H"

O_TestCancel *MgrTestCancel = NULL;
extern SpEvent *CurrentEvent;

//----------------------------------------------------------------------------
// Init - Initializes my SimObj
//----------------------------------------------------------------------------
void S_TestCancel::Init() {

  MyEventCounter = 0;

  MgrTestCancel =
    (O_TestCancel *)SpGetSimObjMgr(OBJECT_TYPE);

  cout << "Initializing " << NAME
       << " <" << NODE << ", " << OBJECT_TYPE << ", " << LOCAL_ID << ">"
       << endl;

  SpObjHandle objHandle = MgrTestCancel->GetObjHandle(NUM_ID);

  if (objHandle.NodeId != NODE) {
    cout << "Error, bad NODE lookup" << endl;
  }

  if (objHandle.ObjManId != OBJECT_TYPE) {
    cout << "Error, bad OBJECT_TYPE lookup" << endl;
  }

  if (objHandle.SimObjId != LOCAL_ID) {
    cout << "Error, bad LOCAL_ID lookup" << endl;
  }

  SpRandom random = int(SEED);
  ScheduleScheduleMyEventTC(0.0, objHandle, random);

  random.GenerateBit();
  ScheduleMyEvilEventTC(0.0, objHandle, random);

}

//----------------------------------------------------------------------------
// ScheduleMyEventTC - Schedules my event into the future, retracts it, and
// then reschedules it for the future.
//----------------------------------------------------------------------------
void S_TestCancel::ScheduleMyEventTC(SpRandom random) {

  if (98765 == random) {
    RB_cout << "Error, event should have been canceled "
	    << CurrentEvent->get_cancel_id() << endl;
    return;
  }

#ifdef NOISY_TEST_CANCEL
  RB_cout << "ScheduleMyEventTC NODE:" << NODE
	  << " SimTime:" << CurrentEvent->get_time_tag()
	  << endl;
#endif

  MyEventCounter++;

/*
** Schedule the event that we will be retracting for a random object
*/

  int destination = random.GenerateInt(0,GetNumTC()-1);
  SpObjHandle objHandle = MgrTestCancel->GetObjHandle(destination);

  SpSimTime time = CurrentEvent->get_time_tag() + 10*DtSched;


  SpRandom ran = 98765;
  SpCancelHandle cancelHandle =
    ScheduleScheduleMyEventTC(time, objHandle, ran);

/*
** Schedule the event for me that will retract the event just scheduled
*/

  time = CurrentEvent->get_time_tag() + 10*DtSched - 1.0;
  objHandle = MgrTestCancel->GetObjHandle(NUM_ID);
  ScheduleRetractMyEventTC(time, objHandle, cancelHandle);

/*
** Reschedule this event for me to occur every 10 units of time
*/

  time = CurrentEvent->get_time_tag() + DtSched;
  objHandle = MgrTestCancel->GetObjHandle(NUM_ID);
  ScheduleScheduleMyEventTC(time, objHandle, random);

}

//----------------------------------------------------------------------------
// RetractMyEvent - Retracts my event.
//----------------------------------------------------------------------------
void S_TestCancel::RetractMyEventTC(SpCancelHandle cancelHandle) {
#ifdef NOISY_TEST_CANCEL
  RB_cout << "RetractMyEventTC NODE: " << NODE
	  << ", cancel_id " << cancelHandle.get_rbid() << endl;
#endif
  CancelEvent(&cancelHandle);
}

//----------------------------------------------------------------------------
// MyEvilEvent - Event that bounces around causing rollbacks.
//----------------------------------------------------------------------------
void S_TestCancel::MyEvilEventTC(SpRandom random) {

#ifdef NOISY_TEST_CANCEL
  RB_cout << "MyEvilEventTC NODE:" << NODE
	  << " SimTime:" << CurrentEvent->get_time_tag()
	  << endl;
#endif

  int destination = random.GenerateInt(0,GetNumTC()-1);
  SpObjHandle objHandle = MgrTestCancel->GetObjHandle(destination);

  SpSimTime time = CurrentEvent->get_time_tag()
    + random.GenerateDouble(0.0, 100.0);

  ScheduleMyEvilEventTC(time, objHandle, random);

}
