// S_TestCreate.C source file

//#define NOISY_TEST_CANCEL

#include "S_TestCreate.H"

O_TestCreate *MgrTestCreate = NULL;
extern SpEvent *CurrentEvent;

//----------------------------------------------------------------------------
// Init - Initializes my SimObj
//----------------------------------------------------------------------------
void S_TestCreate::Init() {

  MyEventCounter = 0;

  MgrTestCreate =
    (O_TestCreate *)SpGetSimObjMgr(OBJECT_TYPE);

  cout << "Initializing " << NAME
       << " <" << NODE << ", " << OBJECT_TYPE << ", " << LOCAL_ID << ">"
       << endl;

  SpObjHandle objHandle = MgrTestCreate->GetObjHandle(NUM_ID);

  if (objHandle.NodeId != NODE) {
    cout << "Error, bad NODE lookup" << endl;
  }

  if (objHandle.ObjManId != OBJECT_TYPE) {
    cout << "Error, bad OBJECT_TYPE lookup" << endl;
  }

  if (objHandle.SimObjId != LOCAL_ID) {
    cout << "Error, bad LOCAL_ID lookup" << endl;
  }

  SpRandom random = int(SEED);
  ScheduleMyRepeatedEventTCR(0.0, objHandle, random);

  random.GenerateBit();
  ScheduleMyEvilEventTCR(0.0, objHandle, random);

}

//----------------------------------------------------------------------------
// InitData - Initializes my SimObj using Data passed in
//----------------------------------------------------------------------------
void S_TestCreate::InitData(char *Data) {
  RB_cout << "Initializing with Data: " << Data << endl;
  MyEventCounter = -1;
}

//----------------------------------------------------------------------------
// ScheduleMyEventTC - Schedules my event into the future, retracts it, and
// then reschedules it for the future.
//----------------------------------------------------------------------------
void S_TestCreate::MyRepeatedEventTCR(SpRandom random) {

/*
** Reschedule this event for me to occur every DtSched units of time
*/

  MyEventCounter++;

  SpSimTime time = CurrentEvent->get_time_tag() + DtSched;
  SpObjHandle objHandle = MgrTestCreate->GetObjHandle(NUM_ID);
  ScheduleMyRepeatedEventTCR(time, objHandle, random);

/*
** Create a SimObj and then destroy it later
*/

  CreateSimObj(time,
	       OBJECT_TYPE,
	       (NODE+1)%N_NODES,
	       objHandle,
	       "Hello World",
	       strlen("Hello World")+1);

//  cout << "objHandle.SimObjId = " << objHandle.SimObjId << endl;

  time += 1.0;
  DestroySimObj(time, objHandle);

}


//----------------------------------------------------------------------------
// MyEvilEvent - Event that bounces around causing rollbacks.
//----------------------------------------------------------------------------
void S_TestCreate::MyEvilEventTCR(SpRandom random) {

#ifdef NOISY_TEST_CANCEL
  RB_cout << "MyEvilEventTC NODE:" << NODE
	  << " SimTime:" << CurrentEvent->get_time_tag()
	  << endl;
#endif

  int destination = random.GenerateInt(0,GetNumTC()-1);
  SpObjHandle objHandle = MgrTestCreate->GetObjHandle(destination);

  SpSimTime time = CurrentEvent->get_time_tag()
    + random.GenerateDouble(0.0, 100.0);

  ScheduleMyEvilEventTCR(time, objHandle, random);

}
