// serve_event.C method file

// I moved the following #include line to before all other #includes
// so that it would compile on the HP Convex/Exemplar computer named
// dax.cmf.nrl.navy.mil.  The old order caused the HP compiler to say
// "signal 11" with no other explanation of what could be wrong, which
// is why I suspect it is a bug in the HP compiler.  There were about
// a total of 38 files in which I had to change the order of includes
// to get it to compile.
#include "SpSimObj.H"
#include "SpStdio.H"
#include "SpAssert.H"
#include "SpIostream.H"

#include "E_PersistenceTest.H"
#include "M_PersistenceTest.H"
#include "O_PersistenceTestObject.H"
#include "S_PersistenceTestObject.H"
#include "PersistenceTestList.H"
#include "SpScheduleEvent.H"
#include "SpFoo.H"


extern SpSimObj* CurrentSimobj;


void GetFirstGetNextTest(S_PersistenceTestObject * simObj){
  SpFoo* foo = PO_NEW_ARRAY_SpFoo(5);
  SpFoo* foo17 = PO_NEW_ARRAY_SpFoo(19);

  //Test some of the generic add and delete functions
  DefaultDataBase->ResetPOErrno();
  if (foo[2].DoTest()){
    simObj->SetError(S_PersistenceTestObject::SPFOO_ERROR);
  }

  SpFoo *foo1, *foo2;

  //Test the get first and get next, comparing generic with specific
  foo1 = PO_GET_FIRST_SpFoo();
  foo2 = (SpFoo*)PO_GET_FIRST("SpFoo");
  if ((foo1 != foo2) && (foo1 != foo)){
    simObj->SetError(S_PersistenceTestObject::GET_FIRST_ERROR);
  }
  foo2 = (SpFoo*)PO_GET_NEXT("SpFoo");
  foo1 = PO_GET_FIRST_SpFoo();
  foo1 = PO_GET_NEXT_SpFoo();
  if ((foo1 != foo2) && (foo1 != &(foo[1]))){
    simObj->SetError(S_PersistenceTestObject::GET_NEXT_ERROR);
  }

  //Test the get first and get next, comparing generic with specific
  foo1 = PO_GET_FIRST_BLOCK_SpFoo();
  foo2 = (SpFoo*)PO_GET_FIRST_BLOCK("SpFoo");
  if ((foo1 != foo2) && (foo1 != foo)){
    simObj->SetError(S_PersistenceTestObject::GET_FIRST_BLOCK_ERROR);
  }
  foo2 = (SpFoo*)PO_GET_NEXT_BLOCK("SpFoo");
  foo1 = PO_GET_FIRST_BLOCK_SpFoo();
  foo1 = PO_GET_NEXT_BLOCK_SpFoo();
  if ((foo1 != foo2) && (foo1 != foo17)){
    simObj->SetError(S_PersistenceTestObject::GET_NEXT_BLOCK_ERROR);
  }

  PO_DELETE_ARRAY_SpFoo(foo);
  PO_DELETE_ARRAY_SpFoo(foo17);
}

class TopPointerClass{
public:
  int junk;
  PersistenceListItem* Top;
  char otherjunk;
};

PO_DEFINE_CLASS(TopPointerClass)

/************************************************************************
* process : optimistically process event				*
************************************************************************/
void E_PersistenceTest::process() {
  S_PersistenceTestObject * simObj = (S_PersistenceTestObject*)CurrentSimobj;
  char* data = get_DATA();
  char* dbBuffer = data + 2*sizeof(int);
  int listLen = (data) ? *(int*) data : 0;
  int AddRemove = (data) ? *(int*) (data+sizeof(int)) : 1;
  int trueListLength;
  double AddProb = simObj->GetAddProb();
  const int MaxListLength = simObj->GetMaxListLength();;
  char itemName[1000];
  M_PersistenceTest header;
  TopPointerClass* topPointer;

  PO_REGISTER_CLASS_ContainedClass();
  PO_REGISTER_CLASS_DerivedClass();
  PO_REGISTER_CLASS_PersistenceListItem();
  PO_REGISTER_CLASS_TopPointerClass();

  SpPoDataBase* db = DefaultDataBase;
  DefaultDataBase = new SpPoDataBase;

  if (AddRemove == 0){
    AddProb = 1-AddProb;
  }
  RB_cout<<"At time "<<(int)TIME_TAG<<" list has length "<<listLen;
  if (data){
    PO_RECONSTRUCT_BUFFER(dbBuffer);
    RB_cout<<" and persistence buffer has size "
	   <<PO_SIZEOF_BUFFER(dbBuffer)<<endl;
  }
  else{
    RB_cout<<" and persistence buffer has size 0"<<endl;
  }

  topPointer = PO_LOOKUP_NAME_TopPointerClass("TopPointer");
  if (topPointer){
    ListTop = topPointer->Top;
  }
  else{
    ListTop = NULL;
  }
  if (ListTop != PO_LOOKUP_NAME_PersistenceListItem("Top")){
    simObj->SetError(S_PersistenceTestObject::VOID_POINTER_LOST);
  }
  PO_DELETE_TopPointerClass(topPointer);

  trueListLength = PersistenceListItem::GetListLength();
  if (listLen != trueListLength){
    RB_cout<<"Wrong length "<<listLen<<" "
	   <<PersistenceListItem::GetListLength()<<endl;
    simObj->SetError(S_PersistenceTestObject::WRONG_LIST_LENGTH);
    return;
  }

  if ((listLen != PO_GET_NUMBER_OF("PersistenceListItem")) ||
      (listLen != PO_GET_NUMBER_OF_PersistenceListItem())){
    simObj->SetError(S_PersistenceTestObject::GET_NUMBER_OF);
  }

  PersistenceListItem::VerifyList();
  
  listLen = trueListLength;
  int loopLim = simObj->GetRandom()->GenerateInt(0, 100);
  for (int i=0; i < loopLim; ++i){
    double randomThrow = (simObj->GetRandom())->GenerateDouble();
    int where = simObj->GetRandom()->GenerateInt(0, listLen-1);
    PersistenceListItem* walker = ListTop;
    for (int j=0; j < where; ++j){
      walker = walker->GetNext();
    }
    if (randomThrow < AddProb){
      ContainedClass* Item;
      if (simObj->GetRandom()->GenerateInt(0, 1)){
	Item = PO_NEW_ContainedClass();
      }
      else{
	Item = (ContainedClass*)PO_NEW_DerivedClass();
      }
      if (ListTop == NULL){
	strcpy(itemName, "0");
      }
      else{
	sprintf(itemName, "%d", where+1);
      }
      Item->reset();
      Item->SetName(itemName);
      PersistenceListItem::InsertAfter(walker, Item);
      ++listLen;
    }
    else{
      ContainedClass* jj = PersistenceListItem::Remove(walker);
      if (jj){
	if (jj->GetType() == CONTAINED_CLASS){
	  PO_DELETE_ContainedClass(jj);
	}
	else{
	  PO_DELETE_DerivedClass((DerivedClass*)jj);
	}
	--listLen;
      }
    }
    if ((listLen > MaxListLength) && (AddProb > 0.5)){
      AddProb = 1-AddProb;
    }
    if ((listLen == 0) && (AddProb < 0.5)){
      AddProb = 1 - AddProb;
    }
  }

  topPointer = new TopPointerClass();
  PO_ADD(topPointer, sizeof(TopPointerClass));
  topPointer->Top = ListTop;
  PO_ATTACH_PTR(topPointer->Top);
  PO_SET_NAME(topPointer, "TopPointer");

  char* buffer = PO_CREATE_BUFFER();
  char* message = new char[PO_SIZEOF_BUFFER(buffer) + 2*sizeof(int)];
  memcpy(message, &listLen, sizeof(int));
  int jj = 1;
  if (AddProb < 0.5){
    jj = 0;
  }
  memcpy(message+sizeof(int), &jj, sizeof(int));
  memcpy(message+2*sizeof(int), buffer,PO_SIZEOF_BUFFER(buffer));

  ScheduleEvent(TIME_TAG+1,
		EVENT_TYPE,
		OBJECT_TYPE,
		0,
		(NODE+1)%N_NODES, //Schedule for next node
		&header,
		message,
		PO_SIZEOF_BUFFER(buffer) + 2*sizeof(int));

  delete [] buffer;
  delete [] message;

  PO_REMOVE(topPointer);
  delete topPointer;

  for (int kk = 0; kk < listLen; ++kk){
    ContainedClass* jj = PersistenceListItem::Remove(ListTop);
    if (jj->GetType() == CONTAINED_CLASS){
      PO_DELETE_ContainedClass(jj);
    }
    else{
      PO_DELETE_DerivedClass((DerivedClass*)jj);
    }
  }

  if (PO_GET_NUMBER_OF_PersistenceListItem()){
    RB_cout<<"There are still "<<PO_GET_NUMBER_OF_PersistenceListItem()
	   <<" persistence list items in the database"<<endl;
    simObj->SetError(S_PersistenceTestObject::TOO_MANY_PERSISTENCE_LIST_ITEMS);
  }

  if (PO_GET_NUMBER_OF_DerivedClass()){
    RB_cout<<"There are still "<<PO_GET_NUMBER_OF_DerivedClass()
	   <<" DerivedClass items in the database"<<endl;
    simObj->SetError(S_PersistenceTestObject::TOO_MANY_DERIVED_CLASS_ITEMS);
  }

  if (PO_GET_NUMBER_OF_ContainedClass()){
    RB_cout<<"There are still "<<PO_GET_NUMBER_OF_ContainedClass()
	   <<" ContainedClass items in the database"<<endl;
    simObj->SetError(S_PersistenceTestObject::TOO_MANY_CONTAINED_CLASS_ITEMS);
  }


  GetFirstGetNextTest(simObj);

  delete DefaultDataBase;
  DefaultDataBase = db;

  if (SpPoDataBase::PO_errno){
    simObj->SetError(S_PersistenceTestObject::UNKNOWN_GENERAL_ERROR);
  }
  SpPoDataBase::PO_errno = 0;
}
