// E_SpAddSubscriberSpace.C source file

#include "E_SpAddSubscriberSpace.H"
#include "S_SpHLA.H"
#include "SpDDMGrid.H"
#include "SpPubSubItem.H"
#include "SpPublishSpace.H"
#include "SpProxyObjHandle.H"


E_SpAddSubscriberSpace::E_SpAddSubscriberSpace()
  : valid(0), errcode(0) {
  set_lazy();
}

int E_SpAddSubscriberSpace::lazy() {
  M_SpAddSubscriberSpace *msg = (M_SpAddSubscriberSpace *) get_MESSAGE();

  // TODO : result in error

  int result = (msg->IsGridReply()) ? 1 : 0;
  result = 0;  
  return (result);
}

void 
E_SpAddSubscriberSpace::process() {
  S_SpHLA *hla                = (S_SpHLA *)                SIMOBJ;
  M_SpAddSubscriberSpace *msg = (M_SpAddSubscriberSpace *) get_MESSAGE();
  const char *data            = (const char *)             get_DATA(); 
  SpPublishSpace *space       = hla->GetPublishSpace(msg->SpaceId);
  SpPublishGrid *grid;
  int nbytes;
  SpList subscribers;
  SpProxyObjHandle *obj_handle;

  valid   = 1;
  errcode = 0;

  grid = space->FindGrid(msg->GridId);
  if (grid==(SpPublishGrid *)0){
    valid   = 0;
    errcode = 1;
    return;
  }

  SpProxyObjHandle::CreateObj(data, msg->NumSubscribers, nbytes, subscribers);

  if (msg->IsGridReply()){
    while ((obj_handle = (SpProxyObjHandle *) subscribers.RemoveTop())){
      grid->AddLookupSubscriber(obj_handle->GetNode(),    obj_handle->GetObjType(),
                                obj_handle->GetLocalId(), obj_handle->GetGlobalId());
      SpProxyObjHandle::Destroy(obj_handle);
    }
    space->ReceivePublishReply(1);
  }
  else{
    while ((obj_handle = (SpProxyObjHandle *) subscribers.RemoveTop())){
      grid->AddNewSubscriber(obj_handle->GetNode(),    obj_handle->GetObjType(),
                             obj_handle->GetLocalId(), obj_handle->GetGlobalId());
      SpProxyObjHandle::Destroy(obj_handle);
    }
    SpBinaryTree grids;
    grids.Insert(grid, grid->GetGlobalId());
#define DISABLE_AUTO_TIE_BREAKING
#ifdef  DISABLE_AUTO_TIE_BREAKING
    const SpSimTime &t = get_MESSAGE()->time_tag;
    SpSimTime sim_time(t.GetTime(), t.GetPriority1(), t.GetPriority2());
    valid &= SpPublishGrid::AddSubscriber(sim_time, msg->LookAhead, hla, msg->SpaceId, grids);
#else
    valid &= SpPublishGrid::AddSubscriber(get_MESSAGE()->time_tag, msg->LookAhead, hla, msg->SpaceId, grids);
#endif
    while (grids.RemoveFirstElement());
    if (!valid) errcode = 2;
  }
}

void 
E_SpAddSubscriberSpace::commit(){
  M_SpAddSubscriberSpace *msg = (M_SpAddSubscriberSpace *) get_MESSAGE();
  const char *data            = (const char *)             get_DATA(); 
  int nbytes;
  SpList subscribers;
  SpProxyObjHandle *obj_handle;

  if (!valid){
    cout << time_tag
         << " ERROR::E_SpAddSubscriberSpace event processed invalid code " << errcode;
    cout << " isGridReply " << msg->IsGridReply() 
         << " simobj " << SIMOBJ->get_GLOBAL_ID() << " grid " << msg->GridId;
    cout << " subscribers (node, type, local, global) : ";
    SpProxyObjHandle::CreateObj(data, msg->NumSubscribers, nbytes, subscribers);
    while ((obj_handle = (SpProxyObjHandle *) subscribers.RemoveTop())){
      cout << " (" << obj_handle->GetNode()    << ", " << obj_handle->GetObjType() << ", "
                   << obj_handle->GetLocalId() << ", " << obj_handle->GetGlobalId() << ") ,";
      SpProxyObjHandle::Destroy(obj_handle);
    }
    cout << endl;
    abort();
  }
}
