// E_SpPublishGrid.C source file

#include "E_SpPublishGrid.H"
#include "SpDDMGrid.H"
#include "S_SpGrid.H"
#include "SpScheduleEvent.H"
#include "SpGridLookupOutput.H"
#include "SpList.H"
#include "SpPubSubItem.H"
#include "SpProxyObjHandle.H"
#include "SpExecute.H"
#include "RB_SpDefineClass.H"

#ifdef TODO
EVENT_TYPE (AddSubscriberSpace);
#else
#include "E_AddSubscriberSpace.H"
#endif
 
RB_DEFINE_CLASS(MAIN_PUBSUB_ITEM);

//#define DDM_DEBUG

E_SpPublishGrid::E_SpPublishGrid() {
//  set_lazy();
}

int E_SpPublishGrid::lazy() {
  return (0);
}

void E_SpPublishGrid::process() {
  S_SpGrid *s_grid     = (S_SpGrid *)        SIMOBJ;
  M_SpPublishGrid *msg = (M_SpPublishGrid *) get_MESSAGE();
  const char *data     = (const char *)      get_DATA();
  MAIN_PUBSUB_ITEM *item, *publish_item;
  int nbytes, buffer_len;
  NET_INT *remove_lookup;
  SpList subscribers;
  SpPubSubItem *s;
  char *buffer;
  SpBinaryTree new_subscribers;
  SpProxyObjHandle *obj_handle;
  SpList remove_lookups;
  SpList rb_lookups;

  new_subscribers.SetBalancedTreeMode();

  SpDDMGrid::CreateObj(data, nbytes, remove_lookups, rb_lookups);

  while (remove_lookups.GetNumElements()>0){
    remove_lookup = (NET_INT *) remove_lookups.pop_top();    
    item = (MAIN_PUBSUB_ITEM *)
           s_grid->GetHiGrid()->RemovePublisher(*remove_lookup);
    if (item)
      RB_DELETE_MAIN_PUBSUB_ITEM(item);
    delete remove_lookup;
  }

#ifdef DDM_DEBUG
  cout << time_tag << " E_SpPublishGrid simobj " << msg->GlobalId 
       << " lookup grid " << s_grid->GetGlobalId() 
       << " num elem " << lookups.GetNumElements()
       << " node simobj/grid : " << msg->Node << "/" << get_NODE() 
       << " subscribers ";
#endif

  while (rb_lookups.GetNumElements()>0){
    publish_item = (MAIN_PUBSUB_ITEM *) rb_lookups.pop_top();
    publish_item->init(msg->Node, msg->Type, msg->LocalId, msg->GlobalId, msg->ClassId);
    s_grid->GetHiGrid()->AddPublisherAndGet(publish_item, &subscribers);
    while ((s = (SpPubSubItem *) subscribers.RemoveTop())){
#ifdef DDM_DEBUG
      cout << s->get_global_id() << ", ";
#endif
      if (s->get_global_id()==msg->GlobalId)        continue;
      if (new_subscribers.Find(s->get_global_id())) continue;
      new_subscribers.Insert(new SpProxyObjHandle(s->get_local_id(), s->get_node(),
                                                  s->get_type(), s->get_global_id()),
                             s->get_global_id());
    }
  }

#ifdef DDM_DEBUG
  cout << endl;
#endif

  SpSimTime event_time(get_MESSAGE()->time_tag);
  event_time += msg->GetLookAhead();

  buffer     = (char *) 0;
  buffer_len = 0;

  SpProxyObjHandle::CreateMessage(new_subscribers, buffer, buffer_len);

  M_SpAddSubscriberSpace add_subscriber_msg(msg->SpaceId, s_grid->GetGlobalId(), 1);
  add_subscriber_msg.SetNumSubscribers(new_subscribers.GetNumElements());
  ScheduleEvent(event_time, AddSubscriberSpace, 
                msg->Type, msg->LocalId, msg->Node,
                &add_subscriber_msg, buffer, buffer_len);  

  if (buffer) delete [] buffer;

  while ((obj_handle = (SpProxyObjHandle *) new_subscribers.RemoveFirstElement()))
    delete obj_handle;
}

void E_SpPublishGrid::commit() {
}
