// E_SpSubscribeGrid.C source file

#include "E_SpSubscribeGrid.H"

#include "SpDDMGrid.H"
#include "S_SpGrid.H"
#include "SpScheduleEvent.H"
#include "SpGridLookupOutput.H"
#include "queue.H"
#include "SpProxyObjHandle.H"
#include "SpDDMUtils.H"
#include "SpPubSubItem.H"
#include "RB_SpDefineClass.H"
#include "SpList.H"

RB_DEFINE_CLASS(MAIN_PUBSUB_ITEM);
RB_DEFINE_CLASS(SpPublishGrid);

#ifdef TODO
EVENT_TYPE (AddSubscriberSpace);
EVENT_TYPE (UnSubscriberSpace);
#else
#include "E_AddSubscriberSpace.H"
#include "E_SpUnSubscribeSpace.H"
#endif

//#define DDM_DEBUG

E_SpSubscribeGrid::E_SpSubscribeGrid() {
//  set_lazy();
}

int E_SpSubscribeGrid::lazy() {
  return (0);
}

void E_SpSubscribeGrid::process() {
  S_SpGrid *s_grid       = (S_SpGrid *)          SIMOBJ;
  M_SpSubscribeGrid *msg = (M_SpSubscribeGrid *) get_MESSAGE();
  const char *data       = (const char *)        get_DATA();
  MAIN_PUBSUB_ITEM *item, *subscribe_item;
  SpBinaryTree old_publishers, new_publishers;
  SpList add_publishers, remove_publishers;
  NET_INT *remove_lookup;
  SpList publishers;
  SpPubSubItem *pubsub_item;
  SpProxyObjHandle *obj_handle;
  char *buffer;
  int buffer_len, nbytes;
  int id;
  SpList remove_lookups;
  SpList rb_lookups;

  SpDDMGrid::CreateObj(data, nbytes, remove_lookups, rb_lookups);

  old_publishers.SetBalancedTreeMode();
  new_publishers.SetBalancedTreeMode();

  SpSimTime event_time(get_MESSAGE()->time_tag);
  event_time += msg->GetLookAhead();

  // retrieve list of publishers based on previous subscription region
  // to be used for compare with the new publishers based on the new subscription 
  // region.

  while (remove_lookups.GetNumElements()>0){
    remove_lookup = (NET_INT *) remove_lookups.pop_top();
    id = *remove_lookup;
    s_grid->GetHiGrid()->GetPublishersWithKey(id, &publishers);
    while ((pubsub_item = (SpPubSubItem *) publishers.RemoveTop())){
      if (pubsub_item->get_global_id()==msg->GlobalId)       continue;
      if (old_publishers.Find(pubsub_item->get_global_id())) continue;
      obj_handle = new SpProxyObjHandle(pubsub_item->get_local_id(), 
                                        pubsub_item->get_node(), 
                                        pubsub_item->get_type(), 
                                        pubsub_item->get_global_id());
      old_publishers.Insert(obj_handle, obj_handle->GetGlobalId());
    }
    item = (MAIN_PUBSUB_ITEM *) 
           s_grid->GetHiGrid()->RemoveSubscriber(id);
    if (item) 
      RB_DELETE_MAIN_PUBSUB_ITEM(item);
    delete remove_lookup;
  }

#ifdef DDM_DEBUG
  cout << time_tag << " E_SpSubscribeGrid simobj " << msg->GlobalId 
       << " lookup grid " << s_grid->GetGlobalId() 
       << " num elem " << rb_lookups.GetNumElements() << " publishers ";
#endif

  while (rb_lookups.GetNumElements()>0){
    subscribe_item = (MAIN_PUBSUB_ITEM *) rb_lookups.pop_top();
    subscribe_item->init(msg->Node, msg->Type, msg->LocalId, msg->GlobalId, msg->ClassId);

    s_grid->GetHiGrid()->AddSubscriberAndGet(subscribe_item, &publishers);

    while ((pubsub_item = (SpPubSubItem *) publishers.RemoveTop())){
      if (pubsub_item->get_global_id()==msg->GlobalId)       continue;
      if (new_publishers.Find(pubsub_item->get_global_id())) continue;
      new_publishers.Insert(pubsub_item, pubsub_item->get_global_id());
#ifdef DDM_DEBUG
      cout << pubsub_item->get_global_id() << ", ";
#endif
    }
  }

  Compare(&old_publishers, &new_publishers, 
          SpProxyObjHandle::GetGlobalId, SpPubSubItem::GetGlobalId,
          &add_publishers, &remove_publishers);

  SpProxyObjHandle new_subscriber(msg->LocalId, msg->Node, msg->Type, msg->GlobalId);
  new_subscriber.CreateMessage(buffer, buffer_len);

  M_SpAddSubscriberSpace add_subscriber_msg(msg->SpaceId, s_grid->GetGlobalId(), 0);
  add_subscriber_msg.SetNumSubscribers(1);
  add_subscriber_msg.SetLookAhead(msg->GetLookAhead());

  while ((pubsub_item = (SpPubSubItem *) add_publishers.RemoveTop())){
    ScheduleEvent(event_time, AddSubscriberSpace, 
                  pubsub_item->get_type(), pubsub_item->get_local_id(), pubsub_item->get_node(),
                  &add_subscriber_msg, buffer, buffer_len);
  }

  if (buffer)
    delete [] buffer;

  M_SpRemSubscriberSpace rem_subscriber_space_msg(msg->SpaceId, msg->GlobalId, 1);
  rem_subscriber_space_msg.SetLookAhead(msg->GetLookAhead());

  NET_INT grid_global_id = s_grid->get_GLOBAL_ID();

#ifdef DDM_DEBUG
  cout << " remove ";
#endif
  // notify publisher
  while (obj_handle = (SpProxyObjHandle *) remove_publishers.RemoveTop()){
    ScheduleEvent(event_time, RemSubscriberSpace,
                  obj_handle->GetObjType(),
                  obj_handle->GetLocalId(),
                  obj_handle->GetNode(),
                  &rem_subscriber_space_msg, (char *) &grid_global_id, sizeof(NET_INT));
#ifdef DDM_DEBUG
  cout << obj_handle->GetGlobalId() << ", ";
#endif
  }
#ifdef DDM_DEBUG
  cout << endl;
#endif

  while ((obj_handle = (SpProxyObjHandle *) old_publishers.RemoveFirstElement()))
    delete obj_handle;

  while (new_publishers.RemoveFirstElement());
}

void E_SpSubscribeGrid::commit() {
}
