// P_SpPublishSpace.C 

//#include "SpProc.H"
// should eventually refactor code to allow #include "SpProc.H"
#include "process.H"

#include "P_SpPublishSpace.H"
#include "SpScheduleGridEvents.H"
#include "SpDDMGrid.H"
#include "S_SpHLA.H"
#include "SpPublishSpace.H"
#include "RB_SpDefineClass.H"
#include "RB_SpList.H"
 
RB_DEFINE_CLASS(SpPublishGrid);

//#define DDM_DEBUG
#define PROXY_LOOKAHEAD

extern SpEvent *CurrentEvent;

PROCESS_LV (P_SpPublishSpace, process)

  int success_flag;
  int loop_forever;
  SpPublishSpace *space;
  RB_SpBinaryTree current_grids;
  RB_SpList delete_grids, lookup_grids;
  SpList grids_added, grids_removed, grids_reprocess;
  SpSimTime event_time, unpublish_time;
  RB_int first_cycle;		// TRUE, until interacted with the grid manager at least once.
  RB_int scheduled_events;	// TRUE, if interacted with the grid manager in this cycle.
  RB_double cycle_start_time;
  RB_double look_ahead;
  SpSimTime sim_time;

BEGIN_PROCESS (P_SpPublishSpace, process, 4)

  double delta_t;
  int node, object_type, local_id, global_id, class_id;
  SpPublishGrid *grid_item, *update_item;

  LV->current_grids.SetBalancedTreeMode();

  LV->space = ((M_SpPublishSpace *) MESSAGE)->PublishSpace;

  if (!LV->space->GetUpdateSemaphore()->GetState()){
    WAIT_FOR_LOGICAL(1,
                     *((LV->space)->GetUpdateSemaphore()),
                     -1.0,
                     LV->success_flag);
  }

#ifdef DDM_DEBUG
  cout << CurrentEvent->get_MESSAGE()->time_tag
       << " P_SpPublishSpace simobj " << LV->space->GetSimObj()->get_GLOBAL_ID()
       << " updateFreq hour " << LV->space->GetUpdateFreqHour()
       << " sec "   << (LV->space->GetUpdateFreqHour() * 3600.0)
       << " knots " << LV->space->GetSimObj()->GetMaxSpeedKnots() << endl;
#endif

  LV->loop_forever  = 1;
  LV->first_cycle   = 1;
  while (LV->loop_forever) {
    LV->cycle_start_time = CurrentEvent->get_MESSAGE()->time_tag;
    LV->space->GetCurrentGrids(CurrentEvent->get_MESSAGE()->time_tag, LV->current_grids);

    SpDDMGrid::CompareGlobalId(*(LV->space->GetGrids()), LV->current_grids, 
                         LV->grids_added, LV->grids_removed, LV->grids_reprocess);

    // zero-look ahead for the first iteration. 
#ifdef PROXY_LOOKAHEAD
    LV->look_ahead = ( (LV->first_cycle) || LV->space->GetUpdateSemaphore()->GetState() )
                     ? 0.0 : (LV->space->GetLookAheadHour()/3.0)*3600.0;
#else
    LV->look_ahead = ( (LV->first_cycle) || LV->space->GetUpdateSemaphore()->GetState() )
                     ? 0.0 : (LV->space->GetLookAheadHour()/2.0)*3600.0;
#endif
    LV->event_time = CurrentEvent->get_MESSAGE()->time_tag + LV->look_ahead;

    node        = LV->space->GetSimObj()->get_NODE();
    object_type = LV->space->GetSimObj()->get_OBJECT_TYPE();
    local_id    = LV->space->GetSimObj()->get_LOCAL_ID();
    global_id   = LV->space->GetSimObj()->get_GLOBAL_ID();
    class_id    = LV->space->GetSimObj()->GetObjectProxy()->GetObjectTemplate()->GetClassID();

    // Compile a list of grids to be lookup which contains : add, and reprocess grids.
    grid_item = (SpPublishGrid *) LV->grids_added.GetFirstElement(); 
    while (grid_item){
      LV->lookup_grids.Insert(grid_item);
      grid_item = (SpPublishGrid *) LV->grids_added.GetNextElement(); 
    }

    // update the grid that are to be reprocess
    while ((grid_item = (SpPublishGrid *) LV->grids_reprocess.RemoveTop())){
      update_item = (SpPublishGrid *) LV->current_grids.Remove(grid_item->GetGlobalId());
      grid_item->Update(update_item);
      RB_DELETE_SpPublishGrid(update_item);
      LV->lookup_grids.Insert(grid_item);
    }

    // transfer the new items from the "grids_added" to the space's published grids, 
    // and delete other items "in current_grids"
    grid_item = (SpPublishGrid *) LV->grids_added.GetFirstElement(); 
    while (grid_item){
      LV->space->AddGrid(grid_item);
      grid_item = (SpPublishGrid *) LV->grids_added.GetNextElement(); 
    }
    while ((grid_item = (SpPublishGrid *) LV->grids_added.RemoveTop()))
      LV->current_grids.Remove(grid_item->GetGlobalId());
    while ((grid_item = (SpPublishGrid *) LV->current_grids.RemoveFirstElement()))
      RB_DELETE_SpPublishGrid(grid_item);
    while ((grid_item = (SpPublishGrid *) LV->grids_removed.RemoveTop()))
      LV->delete_grids.Insert(grid_item);

    LV->event_time.IncrementPriority1(2);
    ScheduleUnPublishGridEvents(LV->event_time, node, object_type, local_id, global_id, 
                 LV->space->GetId(), LV->look_ahead, LV->delete_grids);
    LV->unpublish_time  = LV->event_time;
    LV->unpublish_time += LV->look_ahead;
//  LV->unpublish_time.SetPriority2(MaxPriority2);
    LV->event_time.DecrementPriority1(2);

    LV->event_time.IncrementPriority1(4); 
    SchedulePublishGridEvents(LV->event_time, node, object_type, local_id, global_id, class_id,
                 LV->space->GetId(), LV->look_ahead, LV->lookup_grids);
    LV->event_time.DecrementPriority1(4);

#ifdef DDM_DEBUG
    cout << LV->event_time << " P_SpPublishSpace simobj " << global_id << " grids lookup ";
    grid_item = (SpPublishGrid *) LV->lookup_grids.GetFirstElement();
    while (grid_item){
      cout << grid_item->GetGlobalId() << ", ";
      grid_item = (SpPublishGrid *) LV->lookup_grids.GetNextElement();
    }
    cout << " remove ";
    grid_item = (SpPublishGrid *) LV->delete_grids.GetFirstElement();
    while (grid_item){
      cout << grid_item->GetGlobalId() << ", ";
      grid_item = (SpPublishGrid *) LV->delete_grids.GetNextElement();
    }
    cout << endl;
#endif

    LV->scheduled_events = 0;

    if (LV->lookup_grids.GetNumElements()>0) LV->scheduled_events = 1;
    LV->space->SetNumPublishReplyPending(LV->lookup_grids.GetNumElements());
    REQUEST_RESOURCE(2,
                     *((LV->space)->GetNumPublishReply()),
                     LV->lookup_grids.GetNumElements(),
                     -1,
                     LV->success_flag);
    LV->space->SetNumPublishReplyPending(0);

#ifdef DDM_DEBUG
    cout << CurrentEvent->get_MESSAGE()->time_tag << " P_SpPublishSpace simobj " 
         << LV->space->GetSimObj()->get_GLOBAL_ID() << " publish add/remove grids ";
    grid_item = (SpPublishGrid *) LV->lookup_grids.GetFirstElement();
    while (grid_item){
      cout << grid_item->GetGlobalId() << ", ";
      grid_item = (SpPublishGrid *) LV->lookup_grids.GetNextElement();
    }
    cout << endl;
#endif

#define DISABLE_AUTO_TIE_BREAKING
#ifdef  DISABLE_AUTO_TIE_BREAKING
    LV->sim_time.SetValues(CurrentEvent->get_MESSAGE()->time_tag.GetTime(), 
                       CurrentEvent->get_MESSAGE()->time_tag.GetPriority1(), 
                       CurrentEvent->get_MESSAGE()->time_tag.GetPriority2());
    SpPublishGrid::AddRemoveSubscriber(LV->sim_time,
#else
    SpPublishGrid::AddRemoveSubscriber(CurrentEvent->get_MESSAGE()->time_tag,
#endif
                                 LV->look_ahead, LV->space->GetSimObj(), 
                                 LV->space->GetId(), LV->lookup_grids);
    while (LV->lookup_grids.GetNumElements()>0)
      LV->lookup_grids.pop_top();

    if (LV->delete_grids.GetNumElements()>0) LV->scheduled_events = 1;
    WAIT_UNTIL(3, LV->unpublish_time); 
    SpPublishGrid::RemoveSubscriber(CurrentEvent->get_MESSAGE()->time_tag, 
                                    LV->look_ahead, LV->space->GetSimObj(), 
                                    LV->space->GetId(), LV->delete_grids);
    while (LV->delete_grids.GetNumElements()>0){
      grid_item = (SpPublishGrid *) LV->delete_grids.pop_top();
      LV->space->DeleteGrid(grid_item->GetGlobalId());
    }

    delta_t = LV->space->GetUpdateFreqHour() * 3600.0;
    if (delta_t>0.0){
      if (LV->scheduled_events){
        delta_t -= (CurrentEvent->get_MESSAGE()->time_tag - LV->cycle_start_time);
      }
    }

    if (LV->first_cycle)
      LV->first_cycle = !LV->scheduled_events;

    // I can be interrupted if any of my other dimensions in the routing space 
    // get changed.(i.e, Update gets called).
    LV->space->SetUpdateSemaphore(0);
    WAIT_FOR_LOGICAL(4,
                     *((LV->space)->GetUpdateSemaphore()),
                     delta_t,
                     LV->success_flag);

  }

END_PROCESS (P_SpPublishSpace, process)

///////////////////////////////////////////////////////////////////////////////
