// SpDDMDim.C

#include "SpDDMDim.H"
#include "RB_SpBinaryTree.H"
#include "RB_SpDefineClass.H"

RB_DEFINE_CLASS(SpDDMEnum);
RB_DEFINE_CLASS(SpDDMDoubleDimension);
RB_DEFINE_CLASS(SpDDMEnumDimension);

/////////////////////////////////////////////////////////////////////////////////////////

SpDDMDim::SpDDMDim()
  : name((char *) 0) {
}

SpDDMDim::SpDDMDim(const char *_name){
  int len = strlen(_name);
  name = (len>0) ? new char [len+1] : (char *) 0;
  if (len>0) strcpy(name, _name);
}
  
SpDDMDim::~SpDDMDim() {
  if (name!=(char *) 0) delete [] name;
}

/////////////////////////////////////////////////////////////////////////////////////////

SpDDMDoubleDimension::SpDDMDoubleDimension() {
  lo  = 0.0;
  hi  = 0.0;
  res = 0.0;
}
SpDDMDoubleDimension::SpDDMDoubleDimension(const char *_name, 
                                     double _lo, double _hi, double _res)
  : SpDDMDim(_name) {
  lo  = _lo;
  hi  = _hi;
  res = _res;
}

SpDDMDoubleDimension::~SpDDMDoubleDimension() {}

/////////////////////////////////////////////////////////////////////////////////////////

SpDDMEnum::SpDDMEnum()
  : name((char *) 0) {
}

SpDDMEnum::SpDDMEnum(const char *_name){
  int len = strlen(_name);
  name = (len>0) ? new char [len+1] : (char *) 0;
  if (len>0) strcpy(name, _name);
}

SpDDMEnum::~SpDDMEnum(){
  if (name!=(char *) 0) delete [] name;
}

/////////////////////////////////////////////////////////////////////////////////////////

SpDDMEnumDimension::SpDDMEnumDimension():Enums(new RB_SpBinaryTree) {}

SpDDMEnumDimension::SpDDMEnumDimension(const char *_dim_name)
  : SpDDMDim(_dim_name),Enums(new RB_SpBinaryTree)  {}

SpDDMEnumDimension::SpDDMEnumDimension(const char *_dim_name, 
                                       const char *_enum_name, va_list &va)
  : SpDDMDim(_dim_name) {
  Add(_enum_name, va);
}

SpDDMEnumDimension::~SpDDMEnumDimension(){
  SpDDMEnum *region_enum;

  while ((region_enum = (SpDDMEnum *) Enums->RemoveFirstElement()))
    delete region_enum;
  delete Enums;
}

void 
SpDDMEnumDimension::Add(const char *_enum_name, ...){
  va_list va;

  va_start(va, _enum_name);
  Add(_enum_name, va);
  va_end(va);
}

void 
SpDDMEnumDimension::Add(const char *_enum_name, va_list &va){
  const char *enum_name;
  SpDDMEnum *region_enum;

  region_enum = (SpDDMEnum *) Enums->Find((char *) _enum_name); 
  if (region_enum==(SpDDMEnum *) 0){
    region_enum = RB_NEW_SpDDMEnum(new SpDDMEnum(_enum_name));
    Enums->Insert(region_enum, (char *) region_enum->GetName());
  }

  for (; (enum_name=va_arg(va, char*)) ;){
    if (Enums->Find((char *) enum_name)) continue;
    region_enum = RB_NEW_SpDDMEnum(new SpDDMEnum(enum_name));
    Enums->Insert(region_enum, (char *) region_enum->GetName());
  }
}

void 
SpDDMEnumDimension::Remove(const char *_enum_name, ...){
  va_list va;

  va_start(va, _enum_name);
  Remove(_enum_name, va);
  va_end(va);
}

void 
SpDDMEnumDimension::Remove(const char *_enum_name, va_list &va){
  const char *enum_name;
  SpDDMEnum *region_enum;

  region_enum = (SpDDMEnum *) Enums->Remove((char *) _enum_name);
  if (region_enum) RB_DELETE_SpDDMEnum(region_enum);

  for (; (enum_name=va_arg(va, char*)) ;){
    region_enum = (SpDDMEnum *) Enums->Remove((char *) enum_name);
    if (region_enum) RB_DELETE_SpDDMEnum(region_enum);
  }
}

/////////////////////////////////////////////////////////////////////////////////////////



void Sp_Register_SpDDMDoubleDimension()
{
  RB_REGISTER_CLASS_SpDDMDoubleDimension();
}
void Sp_Register_SpDDMEnumDimension()
{
  RB_REGISTER_CLASS_SpDDMEnumDimension();
}
void Sp_Register_SpDDMEnum()
{
  RB_REGISTER_CLASS_SpDDMEnum();
}
