// SpDDMDim.H 

#ifndef DDMDIMENSION_H
#define DDMDIMENSION_H

class RB_SpBinaryTree;
#include "RB_Types.H"
#include "SpDimLookupItem.H"

typedef SpDimLookupItem::DimensionTypes SpDimensionTypes;

class SpDDMDim {
public:
  SpDDMDim(void);
  SpDDMDim(const char *_name);
  virtual ~SpDDMDim(void);
  const char *GetName(void) const { return (name); }

  virtual SpDimensionTypes GetType(void) const = 0;

private:
  char *name;
};

class SpDDMDoubleDimension : public SpDDMDim {
public:
  SpDDMDoubleDimension(void);
  SpDDMDoubleDimension(const char *_name, double _lo, double _hi, double _res);
  ~SpDDMDoubleDimension(void);

  SpDimensionTypes GetType(void) const { return (SpDimLookupItem::Double); }

  double GetLo(void)  { return (lo);  }
  double GetHi(void)  { return (hi);  }
  double GetRes(void) { return (res); }

  void SetLo(double v)  { lo  = v; }
  void SetHi(double v)  { hi  = v; }
  void SetRes(double v) { res = v; }

private:
  RB_double lo, hi, res;
};

//RB_DEFINE_CLASS(SpDDMDoubleDimension);

class SpDDMEnum {
public:
  SpDDMEnum(void);
  SpDDMEnum(const char *name);
  ~SpDDMEnum(void);

  const char *GetName(void) const { return (name); }

private:
  char  *name;
};

//RB_DEFINE_CLASS(SpDDMEnum);

class SpDDMEnumDimension : public SpDDMDim {
  friend class SpSubscribeRegion;
  friend class SpDDMRegion;
public:
  SpDDMEnumDimension(void);
  SpDDMEnumDimension(const char *_dim_name);
  SpDDMEnumDimension(const char *_dim_name, const char *_enum_name, va_list &va);
  ~SpDDMEnumDimension(void);

  SpDimensionTypes GetType(void) const { return (SpDimLookupItem::Enum); }

    // NULL terminated
  void Remove(const char *_enum_name, ...);
  void Add   (const char *_enum_name, ...);

  RB_SpBinaryTree *GetEnums(void) { return (Enums); }

private:
  RB_SpBinaryTree* Enums;

  void Add(const char *_enum_name, va_list &va);
  void Remove(const char *_enum_name, va_list &va);
};

//RB_DEFINE_CLASS(SpDDMEnumDimension);

#endif
