// DDMGrid.C

#include "SpDDMGrid.H"
#include "SpGridLookupOutput.H"
#include "SpSubscriberHandle.H"
#include "SpProxyObjHandle.H"
#include "S_SpHLA.H"
#include "F_SpProxyItem.H"
#include "SpDDMUtils.H"
#include "SpNetTypes.H"
#include "E_SpValidateDDMSubscriber.H"
#include "E_SpValidateDDMPublisher.H"
#include "SpScheduleEvent.H"
#include "RB_SpDefineClass.H"
#include "SpPubSubItem.H"

RB_DEFINE_CLASS(SpProxyObjHandle);
RB_DEFINE_CLASS(SpDDMGrid);
RB_DEFINE_CLASS(SpPublishGrid);
RB_DEFINE_CLASS(SpSubscribeGrid);
RB_DEFINE_CLASS(MAIN_PUBSUB_ITEM);

//#define DDM_DEBUG
#define PROXY_LOOKAHEAD

SpDDMGrid::SpDDMGrid(void) {}

SpDDMGrid::SpDDMGrid(SpGridLookupOutput *grid){
  Lookups.Insert(grid);
}

SpDDMGrid::~SpDDMGrid(void) {

  SpGridLookupOutput *lookup;
  while (Lookups.GetNumElements()>0){
    lookup = (SpGridLookupOutput *) Lookups.pop_top();
    SpGridLookupOutput::Destroy(lookup);
  }

  NET_INT *remove_lookup;
  while ((remove_lookup = (NET_INT *) RemoveLookups.RemoveTop()))
    delete remove_lookup;
}

int
SpDDMGrid::Update(SpDDMGrid *grid){
  SpGridLookupOutput *lookup;
  NET_INT *remove_lookup;

  while ((remove_lookup = (NET_INT *) RemoveLookups.RemoveTop()))
    delete remove_lookup;

  // record the my lookup items for removal
  while (Lookups.GetNumElements()>0){
    lookup = (SpGridLookupOutput *) Lookups.pop_top();
    RemoveLookups.Insert(new NET_INT(lookup->get_MainPubSubItem()->GetUniqueKey()));
    SpGridLookupOutput::Destroy(lookup);
  }

  // transfer the other grid's lookup item to me 
  while (grid->Lookups.GetNumElements()>0){
    lookup = (SpGridLookupOutput *) grid->Lookups.pop_top();
    Lookups.Insert(lookup);
  }

  return (1);
}

int 
SpDDMGrid::GetGlobalId(void){ 
  SpGridLookupOutput *lookup = GetLookup();
  return (lookup ? lookup->get_id() : -1); 
}

int 
SpDDMGrid::GetLocalId(void){
  SpGridLookupOutput *lookup = GetLookup();
  return (lookup ? lookup->getLocalId() : -1); 
}

int 
SpDDMGrid::GetNode(void){
  SpGridLookupOutput *lookup = GetLookup();
  return (lookup ? lookup->getNode() : -1); 
}

int 
SpDDMGrid::GetObjectType(void){
  SpGridLookupOutput *lookup = GetLookup();
  return (lookup ? lookup->getObjType() : -1); 
}

SpGridLookupOutput *
SpDDMGrid::GetLookup(void){
  SpGridLookupOutput *lookup = (SpGridLookupOutput *) Lookups.GetFirstElement();
  return (lookup);
}

void
SpDDMGrid::CreateMsgLookupId(SpDDMGrid *grid, char* &msg, int &len){
  int num_lookups;  
  SpGridLookupOutput *item;
  NET_INT id;
  char *cp;

  num_lookups = grid->Lookups.GetNumElements();

  len = sizeof(NET_INT) * num_lookups;
  msg = (len>0) ? new char [len] : (char *) 0;
  cp  = msg;
  
  item = (SpGridLookupOutput *) grid->Lookups.GetFirstElement();
  while (item){
    id = item->get_MainPubSubItem()->GetUniqueKey();
    memcpy(cp, &id, sizeof(NET_INT)); cp += sizeof(NET_INT);
    item = (SpGridLookupOutput *) grid->Lookups.GetNextElement();
  }
}

void
SpDDMGrid::CreateMessage(char* &msg, int &len){
  char *cp;
  char **grid_msg;
  int *grid_msg_len, total_msg_len;
  SpGridLookupOutput *grid;
  int i, n;
  NET_INT *remove_lookup, num_lookups, num_remove_lookups;

  msg = (char *) 0;
  len = 0;
  
  n = num_lookups = Lookups.GetNumElements();

  grid_msg     = (n>0) ? new char *[n] : (char **) 0;
  grid_msg_len = (n>0) ? new int   [n] : (int *)   0;

  i = 0;
  total_msg_len = 0;
  grid = (SpGridLookupOutput *) Lookups.GetFirstElement();
  while (grid){
    grid->get_MainPubSubItem()->CreateMessage(grid_msg[i], grid_msg_len[i]);
    total_msg_len += grid_msg_len[i];
    i++;
    grid = (SpGridLookupOutput *) Lookups.GetNextElement();
  }

  num_remove_lookups = RemoveLookups.GetNumElements();

  len = sizeof(NET_INT) + total_msg_len + 
        sizeof(NET_INT) + (sizeof(NET_INT) * num_remove_lookups);
  msg = new char [len]; 
  cp  = msg;
  memcpy(cp, &num_lookups, sizeof(NET_INT)); cp += sizeof(NET_INT);
  for (i=0; i<num_lookups; i++){
    memcpy(cp, grid_msg[i], grid_msg_len[i]); cp += grid_msg_len[i];
  }

  memcpy(cp, &num_remove_lookups, sizeof(NET_INT)); cp += sizeof(NET_INT);

  remove_lookup = (NET_INT *) RemoveLookups.GetFirstElement();
  while (remove_lookup){
    memcpy(cp, remove_lookup, sizeof(NET_INT)); cp += sizeof(NET_INT);
    remove_lookup = (NET_INT *) RemoveLookups.GetNextElement();
  } 

  if (num_lookups>0)
    delete [] grid_msg_len;
  for (i=0; i<num_lookups; i++)
    delete [] grid_msg[i];
  if (num_lookups>0)
    delete [] grid_msg;
}

void 
SpDDMGrid::CreateMsgGlobalId(SpDDMGrid *grid, char* &msg, int &len){
  ((SpObsObjHandle *) grid->GetLookup())->CreateMessageGlobalId(msg, len);
}

void 
SpDDMGrid::CreateMsgObjHandle(SpDDMGrid *grid, char* &msg, int &len){
  ((SpObsObjHandle *) grid->GetLookup())->CreateMessage(msg, len);
}

void
SpDDMGrid::CreateObj(const char *msg, int &obj_bytes, 
                     SpList &remove_lookups, SpList &rb_lookups){
  const char *cp = msg;
  int i, nbytes;
  NET_INT num_lookups, num_remove_lookups, *remove_lookup;
  MAIN_PUBSUB_ITEM *lookup;

  obj_bytes = 0;

  nbytes = sizeof(NET_INT);
  memcpy(&num_lookups, cp, nbytes); cp += nbytes; obj_bytes += nbytes;
  for (i=0; i<num_lookups; i++){
    lookup = RB_NEW_MAIN_PUBSUB_ITEM();
    MAIN_PUBSUB_ITEM::CreateObject(cp, nbytes, *lookup); cp += nbytes; obj_bytes += nbytes;
    rb_lookups.Insert(lookup);
  }

  nbytes = sizeof(NET_INT);
  memcpy(&num_remove_lookups, cp, nbytes); cp += nbytes; obj_bytes += nbytes;
  for (i=0; i<num_remove_lookups; i++){
    remove_lookup = new NET_INT;
    memcpy(remove_lookup, cp, nbytes); cp += nbytes; obj_bytes += nbytes;
    remove_lookups.Insert(remove_lookup);
  }
}

int
SpDDMGrid::Add(SpGridLookupOutput *newLookup){
#ifdef ORIG
  int match = 0;
  SpGridLookupOutput *lookup;
 
  lookup = (SpGridLookupOutput *) Lookups.GetFirstElement();
  while (lookup){
    if (lookup->get_MainPubSubItem()->get_global_id()==newLookup->get_MainPubSubItem()->get_global_id()){
//    if (*lookup==*newLookup){
      match = 1;
      break;
    }
    lookup = (SpGridLookupOutput *) Lookups.GetNextElement();
  }
  if (!match)
    Lookups.Insert(newLookup);

  return (!match);
#else
  Lookups.Insert(newLookup);
  return (1);
#endif
}

void 
SpDDMGrid::CompareGlobalId(RB_SpBinaryTree &orig_grids,	// input
                   RB_SpBinaryTree &new_grids,		// input
                   SpList &grids_add,			// output
                   SpList &grids_remove,		// output
                   SpList &grids_reprocess){		// output

  Compare(&orig_grids, &new_grids,
          SpDDMGrid::GetGlobalId, SpDDMGrid::GetGlobalId,
          &grids_add, &grids_remove, &grids_reprocess);
}

int
SpDDMGrid::operator==(SpDDMGrid &g){

  cout << "Invoked SpDDMGrid's == operator" << endl;
  abort();
  return (1);
}

void
SpDDMGrid::Debug(ostream &os){
//#ifdef NOT_USED
  SpGridLookupOutput *lookup;
  
  lookup = (SpGridLookupOutput *) Lookups.GetFirstElement();
  while (lookup){
    os << "grid id, local id, node, type / mainpubsub uniqueKey : " 
       << lookup->getGlobalId() << ", "
       << lookup->getLocalId() << ", "
       << lookup->getNode() << ", "
       << lookup->getObjType() << ", " 
       << lookup->get_MainPubSubItem()->GetUniqueKey() << endl;
    lookup->get_MainPubSubItem()->PrintThis(os);
    lookup = (SpGridLookupOutput *) Lookups.GetNextElement();
  }
//#else
//  os << "\t\tgrid : " << GetGlobalId() << endl;
//#endif
}

////////////////////////////////////////////////////////////////////////////////////////

SpPublishGrid::SpPublishGrid() {}

SpPublishGrid::SpPublishGrid(SpGridLookupOutput *_grid)
  : SpDDMGrid(_grid) {}

SpPublishGrid::~SpPublishGrid() {
  SpProxyObjHandle *subscriber;

  while (subscribers.RemoveFirstElement());

  while ((subscriber = (SpProxyObjHandle *) lookup_subscribers.RemoveFirstElement()))
    RB_DELETE_SpProxyObjHandle(subscriber);

  while ((subscriber = (SpProxyObjHandle *) new_subscribers.RemoveFirstElement()))
    RB_DELETE_SpProxyObjHandle(subscriber);
}

void
SpPublishGrid::DeleteLookupSubscribers(){
  SpProxyObjHandle *subscriber;

  while ((subscriber = (SpProxyObjHandle *) lookup_subscribers.RemoveFirstElement()))
    RB_DELETE_SpProxyObjHandle(subscriber);
}

int 
SpPublishGrid::AddLookupSubscriber(int _node, int _type, int _local_id, int _global_id){
  SpProxyObjHandle *subscriber;

  subscriber = (SpProxyObjHandle *) lookup_subscribers.Find(_global_id);
  if (subscriber) return (0);
  
  subscriber = new SpProxyObjHandle(_local_id, _node, _type, _global_id);
  lookup_subscribers.Insert(RB_NEW_SpProxyObjHandle(subscriber), _global_id);

  return (1);
}

void
SpPublishGrid::DeleteNewSubscribers(){
  SpProxyObjHandle *subscriber;

  while ((subscriber = (SpProxyObjHandle *) new_subscribers.RemoveFirstElement()))
    RB_DELETE_SpProxyObjHandle(subscriber);
}

int
SpPublishGrid::AddNewSubscriber(int _node, int _type, int _local_id, int _global_id){
  SpProxyObjHandle *subscriber;

  subscriber = (SpProxyObjHandle *) new_subscribers.Find(_global_id);
  if (subscriber) return (0);
 
  subscriber = new SpProxyObjHandle(_local_id, _node, _type, _global_id);
  new_subscribers.Insert(RB_NEW_SpProxyObjHandle(subscriber), _global_id);

  return (1);
}

int 
SpPublishGrid::AddSubscriber(SpSubscriberHandle *subscriber){
  if (subscribers.Find(subscriber->GetGlobalId())) 
    return (0);
  subscribers.Insert(subscriber, subscriber->GetGlobalId());
  return (1);
}

extern SpEvent *CurrentEvent;

SpSubscriberHandle * 
SpPublishGrid::RemoveSubscriber(int global_id){
  SpSubscriberHandle *h = (SpSubscriberHandle *) GetSubscribers().Remove(global_id);
  return (h);
}

void
SpPublishGrid::Destroy(SpPublishGrid *grid){
  RB_DELETE_SpPublishGrid(grid);
}

SpDDMGrid * 
SpPublishGrid::RBCreate(SpGridLookupOutput *item){
  return (RB_NEW_SpPublishGrid(new SpPublishGrid(item)));
}

int 
SpPublishGrid::RemoveSubscriber(const SpSimTime &current_time, 
                                float look_ahead, S_SpHLA *hla, int space_id, int global_id){
  SpBinaryTree remove_subscribers;
  SpBinaryTree remove_subscriber_holder, undeliver_subscriber_holder;
  SpSubscriberHandle *s = RemoveSubscriber(global_id);

  if (s==(SpSubscriberHandle *)0) return (0);

#ifdef PROXY_LOOKAHEAD
  SpSimTime event_time(current_time);
  event_time += look_ahead;
#else
  const SpSimTime &event_time = current_time;
  look_ahead = 0.0;
#endif

#ifdef DDM_DEBUG
    cout << event_time << " publish simobj " << hla->get_GLOBAL_ID()
         << " grid " << GetGlobalId() << " remove " << global_id << endl;
#endif

  SpProxyObjHandle subscriber(s->GetLocalId(), s->GetNode(),
                              s->GetObjType(), s->GetGlobalId());
  remove_subscribers.Insert((void *) &subscriber, subscriber.GetGlobalId());

  hla->RemoveSubscriber(SpProxySubscriber::DDM, GetGlobalId(), 
		remove_subscribers,
		&remove_subscriber_holder, &undeliver_subscriber_holder);

  hla->RemoveProxy(event_time, look_ahead, space_id, 
		&remove_subscriber_holder, &undeliver_subscriber_holder);

  S_SpHLA::ClearSubscriberHolder(remove_subscriber_holder);
  S_SpHLA::ClearSubscriberHolder(undeliver_subscriber_holder);

  while (remove_subscribers.RemoveFirstElement());

  return (1);
}

int 
SpPublishGrid::RemoveSubscriber(const SpSimTime &current_time, float look_ahead,
                                S_SpHLA *hla, int space_id, RB_SpList &grids){
  SpPublishGrid *grid;
  SpBinaryTree remove_subscriber_holder, undeliver_subscriber_holder;
  SpProxyObjHandle *subscriber, *grid_subscriber;
  SpBinaryTree grid_subscribers;
  SpList remove_subscribers;

#ifdef PROXY_LOOKAHEAD
  SpSimTime event_time(current_time);
  event_time += look_ahead;
#else
  const SpSimTime &event_time = current_time;
  look_ahead = 0.0;
#endif

#ifdef DDM_DEBUG
  cout << event_time << " simobj " << hla->get_GLOBAL_ID() << " remove";
#endif

  grid = (SpPublishGrid *) grids.GetFirstElement();
  while (grid){
#ifdef DDM_DEBUG
  cout << " grid " << grid->GetGlobalId() << " subscribers ";
#endif
    while ((subscriber = (SpProxyObjHandle *) grid->GetSubscribers().RemoveFirstElement())){
#ifdef DDM_DEBUG
  cout << subscriber->GetGlobalId() << ", ";
#endif
      grid_subscriber = new SpProxyObjHandle(subscriber->GetLocalId(), 
                                             subscriber->GetNode(),
                                             subscriber->GetObjType(), 
                                             subscriber->GetGlobalId());
      grid_subscribers.Insert((void *) grid_subscriber, grid_subscriber->GetGlobalId());
      remove_subscribers.InsertTop((void *) grid_subscriber);
    }

    hla->RemoveSubscriber(SpProxySubscriber::DDM, grid->GetGlobalId(), 
                          grid_subscribers,
                          &remove_subscriber_holder, &undeliver_subscriber_holder);

    while (grid_subscribers.RemoveFirstElement());

    grid = (SpPublishGrid *) grids.GetNextElement();
  }
#ifdef DDM_DEBUG
  cout << " remove "    << remove_subscriber_holder.GetNumElements()
       << " undeliver " << undeliver_subscriber_holder.GetNumElements() << endl;
#endif

  hla->RemoveProxy(event_time, look_ahead, space_id, 
                   &remove_subscriber_holder, &undeliver_subscriber_holder);

  while ((subscriber = (SpProxyObjHandle *) remove_subscribers.Remove()))
    delete subscriber;

  S_SpHLA::ClearSubscriberHolder(remove_subscriber_holder);
  S_SpHLA::ClearSubscriberHolder(undeliver_subscriber_holder);

  return (1);
}

int 
SpPublishGrid::AddRemoveSubscriber(const SpSimTime &current_time, float look_ahead,
                                   S_SpHLA *hla, int space_id, RB_SpList &grids){
  SpPublishGrid *grid;
  SpSubscriberHandle *sub_handle;
  SpProxyObjHandle *obj_handle;
  SpList add_subscribers, remove_subscribers;
  SpBinaryTree new_subscriber_holder, add_subscriber_holder, dup_subscriber_holder;
  SpBinaryTree remove_subscriber_holder, undeliver_subscriber_holder;
  SpBinaryTree subscriber_handles;
  SpBinaryTree remove_grids;
  SpBinaryTree *current_grid;
  int grid_id;

#ifdef PROXY_LOOKAHEAD
  SpSimTime event_time(current_time);
  event_time += look_ahead;
#else
  const SpSimTime &event_time = current_time;
  look_ahead = 0.0;
#endif

  grid = (SpPublishGrid *) grids.GetFirstElement();
  while (grid){
    Compare(&(grid->subscribers), &(grid->lookup_subscribers),
            SpSubscriberHandle::GetGlobalId, SpProxyObjHandle::GetGlobalId,
            &add_subscribers, &remove_subscribers);

#ifdef DDM_DEBUG
    cout << event_time << " publish simobj " << hla->get_GLOBAL_ID() 
         << " grid " << grid->GetGlobalId() << " current ";
    sub_handle = (SpSubscriberHandle *) grid->subscribers.GetFirstElement();
    while (sub_handle){
      cout << sub_handle->GetGlobalId() << ", ";
      sub_handle = (SpSubscriberHandle *) grid->subscribers.GetNextElement();
    }
    cout << " lookup ";
    obj_handle = (SpProxyObjHandle *) grid->lookup_subscribers.GetFirstElement();
    while (obj_handle){
      cout << obj_handle->GetGlobalId() << ", ";
      obj_handle = (SpProxyObjHandle *) grid->lookup_subscribers.GetNextElement();
    }
    cout << " add ";
    obj_handle = (SpProxyObjHandle *) add_subscribers.GetFirstElement();
    while (obj_handle){
      cout << obj_handle->GetGlobalId() << ", ";
      obj_handle = (SpProxyObjHandle *) add_subscribers.GetNextElement();
    }
    cout << " remove ";
    sub_handle = (SpSubscriberHandle *) remove_subscribers.GetFirstElement();
    while (sub_handle){
      cout << sub_handle->GetGlobalId() << ", ";
      sub_handle = (SpSubscriberHandle *) remove_subscribers.GetNextElement();
    }
    cout << endl;
#endif

    hla->AddSubscriber(SpProxySubscriber::DDM, grid->GetGlobalId(), add_subscribers,
            &new_subscriber_holder, &add_subscriber_holder, &dup_subscriber_holder, 
            &subscriber_handles);

    while ((sub_handle=(SpSubscriberHandle *) subscriber_handles.RemoveFirstElement()))
      grid->AddSubscriber(sub_handle);

    while (add_subscribers.RemoveTop());

    current_grid = new SpBinaryTree;
    remove_grids.Insert(current_grid, grid->GetGlobalId());
    while ((obj_handle=(SpProxyObjHandle*) remove_subscribers.RemoveTop())){
      grid->RemoveSubscriber(obj_handle->GetGlobalId());
      current_grid->Insert(obj_handle, obj_handle->GetGlobalId());
    }

    grid = (SpPublishGrid *) grids.GetNextElement();
  }

  current_grid = (SpBinaryTree *) remove_grids.GetFirstElement();
  while (current_grid){
    grid_id = int (remove_grids.GetCurrentKey());

    hla->RemoveSubscriber(SpProxySubscriber::DDM, grid_id, *current_grid,
                          &remove_subscriber_holder, &undeliver_subscriber_holder);

    while (current_grid->RemoveFirstElement());

    current_grid = (SpBinaryTree *) remove_grids.GetNextElement();
  }

  while ((current_grid = (SpBinaryTree *) remove_grids.RemoveFirstElement()))
    delete current_grid;

  hla->AddProxy(event_time, look_ahead, space_id, 
                &new_subscriber_holder, &add_subscriber_holder, &dup_subscriber_holder);
  hla->RemoveProxy(event_time, look_ahead, space_id, 
                &remove_subscriber_holder, &undeliver_subscriber_holder);

  grid = (SpPublishGrid *) grids.GetFirstElement();
  while (grid){
    grid->DeleteLookupSubscribers();
    grid = (SpPublishGrid *) grids.GetNextElement();
  }

  S_SpHLA::ClearSubscriberHolder(new_subscriber_holder);
  S_SpHLA::ClearSubscriberHolder(add_subscriber_holder);
  S_SpHLA::ClearSubscriberHolder(dup_subscriber_holder);
  S_SpHLA::ClearSubscriberHolder(remove_subscriber_holder);
  S_SpHLA::ClearSubscriberHolder(undeliver_subscriber_holder);

  return (1);
}

int 
SpPublishGrid::AddSubscriber(SpSimTime &current_time, float look_ahead,
                             S_SpHLA *hla, int space_id, SpBinaryTree &grids){
  SpPublishGrid *grid;
  SpSubscriberHandle *sub_handle;
  SpBinaryTree new_subscriber_holder, add_subscriber_holder, dup_subscriber_holder;
  SpBinaryTree handles;
  SpList add_subscribers, remove_subscribers;
  int valid = 1;

#ifdef PROXY_LOOKAHEAD
  SpSimTime event_time(current_time);
  event_time += look_ahead;
#else
  SpSimTime &event_time = current_time;
  look_ahead = 0.0;
#endif

  grid = (SpPublishGrid *) grids.GetFirstElement();
  while (grid){

    Compare(&(grid->subscribers), &(grid->new_subscribers),
            SpSubscriberHandle::GetGlobalId, SpProxyObjHandle::GetGlobalId,
            &add_subscribers, &remove_subscribers);

#ifdef DDM_DEBUG
    cout << event_time << " publish sub simobj " << hla->get_GLOBAL_ID() << " grid " 
         << grid->GetGlobalId() << " current ";
    sub_handle = (SpSubscriberHandle *) grid->subscribers.GetFirstElement();
    while (sub_handle){
      cout << sub_handle->GetGlobalId() << ", ";
      sub_handle = (SpSubscriberHandle *) grid->subscribers.GetNextElement();
    }
    cout << " add ";
    SpProxyObjHandle *obj_handle = (SpProxyObjHandle *) add_subscribers.GetFirstElement();
    while (obj_handle){
      cout << obj_handle->GetGlobalId() << ", ";
      obj_handle = (SpProxyObjHandle *) add_subscribers.GetNextElement();
    }
    if (add_subscribers.GetNumElements()<=0){
      cout << " DUPLICATE ";
      obj_handle = (SpProxyObjHandle *) grid->new_subscribers.GetFirstElement();
      while (obj_handle){
        cout << obj_handle->GetGlobalId() << ", ";
        obj_handle = (SpProxyObjHandle *) grid->new_subscribers.GetNextElement();
      }
    }
    cout << endl;
#endif

    if (add_subscribers.GetNumElements()<=0){
      valid = 0;
      break;
    }

    hla->AddSubscriber(SpProxySubscriber::DDM, grid->GetGlobalId(), add_subscribers,
                       &new_subscriber_holder, &add_subscriber_holder, &dup_subscriber_holder, 
                       &handles);

    while (add_subscribers.RemoveTop());
    while (remove_subscribers.RemoveTop());

    while ((sub_handle=(SpSubscriberHandle *) handles.RemoveFirstElement()))
      grid->AddSubscriber(sub_handle);

    grid = (SpPublishGrid *) grids.GetNextElement();
  }

  if (valid)
    hla->AddProxy(event_time, look_ahead, space_id, 
                  &new_subscriber_holder, &add_subscriber_holder, &dup_subscriber_holder);

  grid = (SpPublishGrid *) grids.GetFirstElement();
  while (grid){
    grid->DeleteNewSubscribers();
    grid = (SpPublishGrid *) grids.GetNextElement();
  }

  S_SpHLA::ClearSubscriberHolder(new_subscriber_holder);
  S_SpHLA::ClearSubscriberHolder(add_subscriber_holder);
  S_SpHLA::ClearSubscriberHolder(dup_subscriber_holder);

  return (valid);
}

void
SpPublishGrid::Debug(ostream &os){
  SpSubscriberHandle *subscriber;

  SpDDMGrid::Debug(os);
  os << "\t\t\tsubscribers (" << subscribers.GetNumElements() << ") : ";
 
  subscriber = (SpSubscriberHandle *) subscribers.GetFirstElement();
  while (subscriber){
    os << subscriber->GetGlobalId() << ", ";
    subscriber = (SpSubscriberHandle *) subscribers.GetNextElement();
  }
  os << endl;
}

void
SpPublishGrid::Validate(SIMTIME &time, int publisher_global_id, const char *space_name){
  SpSubscriberHandle *subscriber;
  M_SpValidateDDMSubscriber msg(publisher_global_id);

  subscriber = (SpSubscriberHandle *) subscribers.GetFirstElement();
  while (subscriber){
    ScheduleEvent(time, ValidateDDMSubscriber, 
                  subscriber->GetObjType(), subscriber->GetLocalId(), subscriber->GetNode(),
                  msg, space_name, strlen(space_name)+1);
    subscriber = (SpSubscriberHandle *) subscribers.GetNextElement();
  }
}

////////////////////////////////////////////////////////////////////////////////////////

SpSubscribeGrid::SpSubscribeGrid() {}

SpSubscribeGrid::SpSubscribeGrid(SpGridLookupOutput *_grid)
  : SpDDMGrid(_grid) {}

SpSubscribeGrid::~SpSubscribeGrid() {

  while (RemoteObjectProxies.RemoveFirstElement());
}

SpDDMGrid * 
SpSubscribeGrid::RBCreate(SpGridLookupOutput *item){
  return (RB_NEW_SpSubscribeGrid(new SpSubscribeGrid(item)));
}

int
SpSubscribeGrid::Add(F_SpProxyItem *item){
  F_SpProxyItem *p = (F_SpProxyItem *) RemoteObjectProxies.Find(item->GetGlobalId());
  if (p) return (0);

  RemoteObjectProxies.Insert(item, item->GetGlobalId());
  return (1);
}

F_SpProxyItem *
SpSubscribeGrid::Find(int global_id){
  return ((F_SpProxyItem *) RemoteObjectProxies.Find(global_id));
}

int
SpSubscribeGrid::Remove(int global_id){
  F_SpProxyItem *p = (F_SpProxyItem *) RemoteObjectProxies.Find(global_id);
  if (!p) return (0);
  RemoteObjectProxies.Remove(global_id);
  return (1);
}

void
SpSubscribeGrid::Destroy(SpSubscribeGrid *grid){
  RB_DELETE_SpSubscribeGrid(grid);
}

void
SpSubscribeGrid::Debug(ostream &os){
  F_SpProxyItem *proxy;

//  SpDDMGrid::Debug(os);
//  os << "\t\t\tremote proxy (" << RemoteObjectProxies.GetNumElements() << ") : ";

  proxy = (F_SpProxyItem *) RemoteObjectProxies.GetFirstElement();
  while (proxy){
    os << proxy->GetGlobalId() << ", ";
    proxy = (F_SpProxyItem *) RemoteObjectProxies.GetNextElement();
  }
//  os << endl;
}

void
SpSubscribeGrid::Validate(SIMTIME &time, int subscriber_global_id, const char *space_name){
  M_SpValidateDDMPublisher msg(subscriber_global_id, GetGlobalId());

  F_SpProxyItem *proxy = (F_SpProxyItem *) RemoteObjectProxies.GetFirstElement();
  while (proxy){
    ScheduleEvent(time, ValidateDDMPublisher, 
                  proxy->GetObjType(), proxy->GetLocalId(), proxy->GetNode(),
                  msg, space_name, strlen(space_name)+1);
    proxy = (F_SpProxyItem *) RemoteObjectProxies.GetNextElement();
  }
}

////////////////////////////////////////////////////////////////////////////////////////

void RB_Register_SpDDMGrid()
{
  RB_REGISTER_CLASS_SpDDMGrid();
}

void RB_Register_SpPublishGrid()
{
  RB_REGISTER_CLASS_SpPublishGrid();
}

void RB_Register_SpSubscribeGrid()
{
  RB_REGISTER_CLASS_SpSubscribeGrid();
}
