// SpDDMRegion.H header file defines an HLA region which can be used both for
// publishing and subscribing.

#ifndef DDMREGION_H
#define DDMREGION_H

class RB_SpBinaryTree;
#include "RB_Types.H"
#include "SpOS.H"

class SpDDMSpace;
class SpDDMDim;
class SpDDMEnumDimension;

class SpDDMRegion {
public:
  SpDDMRegion(void);
  SpDDMRegion(SpDDMSpace *space, const char *name);
  ~SpDDMRegion(void);

    // Add 1 or more double (name/lo/hi) dimension (null terminated)
  void CreateDimension(const char *dimName, double lo, double hi, ...);
    // Modify 1 or more double (name/lo/hi) dimension (null terminated)
  void ModifyDimension(const char *dimName, double lo, double hi, ...);

    // Add ONE enumerated dimension : add 1 or more enumerated (enum_name) value (null terminated)
  void CreateDimension(const char *dimName, const char *enum_name, ...);
    // Modify the enumerated dimension : add 1 or more enumerated value (null terminated)
  void ModifyDimension(const char *dimName, const char *enum_name, ...);

  SpDDMDim *FindDimension(const char *dimName);

  void DeleteDimension(const char *dimName);

  RB_SpBinaryTree &GetDimensions(void);
  int GetNumDimensions(void);

  const char *GetName(void) const { return (Name); }

protected:
  SpDDMEnumDimension *CreateEnumDimension(const char *dim_name);
  
private:
  char *Name;
  SpDDMSpace *space;
  RB_SpBinaryTree* Dimensions;

  void CreateDoubleDimension(const char *dimName, double lo, double hi, double res);

  int  UpdateDimension(const char *dimName, double lo, double hi, double res);
  void ModifyDimension(const char *dimName, double lo, double hi, va_list &va);

  int ValidateEnumDimension(const char *dimName);
  int ValidateDoubleDimension(const char *dimName);

  void DeleteDimension(SpDDMDim *dimension);
};

//RB_DEFINE_CLASS(SpDDMRegion);

class SpPublishRegion : public SpDDMRegion {
public:
  SpPublishRegion(void);
  SpPublishRegion(SpDDMSpace *space, const char *name);
  ~SpPublishRegion(void);
};

//RB_DEFINE_CLASS(SpPublishRegion);

class SpSubscribeRegion : public SpDDMRegion {
public:
  SpSubscribeRegion(void);
  SpSubscribeRegion(SpDDMSpace *space, const char *name);
  ~SpSubscribeRegion(void);

  void AddClass   (const char *className, ...);
  void RemoveClass(const char *className, ...);

private:
  SpDDMEnumDimension *class_name_dimension;
};

//RB_DEFINE_CLASS(SpSubscribeRegion);

#endif
