// SpPublishRoutingSpace.H header file defines the base class for
// managing DDM routing spaces used for publishing

#ifndef SpPubRoutSpaceGate
#define SpPubRoutSpaceGate

#include "SpPublishSpace.H"
#include "RB_Types.H"
#include "SpSimTime.H"

class SpDDMRegion;
class SpSpace;
class SpSpaceLookupItem;
class SpDDMGrid;
class SpGridLookupOutput;
class RB_SpBinaryTree;

class SpPublishRoutingSpace : public SpPublishSpace {
  friend class S_SpHLA;
  friend class SpSubscribeRoutingSpace;
  friend class SpSubscribeTheaterSpace;
public:
  SpPublishRoutingSpace(void);
  ~SpPublishRoutingSpace(void);

protected:
  SpPublishRoutingSpace(const SpSimTime &time, S_SpHLA *hla, const SpSpace *space);

  void GetCurrentGrids(const SpSimTime &time, RB_SpBinaryTree &grids);

  static void CreateSpaceLookupItem(const SpSimTime &time, SpDDMSpace *space, SpDDMRegion *region,
                                    SpSpaceLookupItem* &space_lookup);

  static void GetCurrentGrids(const SpSimTime &time,
                              SpDDMSpace *space, S_SpHLA *hla, RB_SpBinaryTree &regions,
                              void (*CreateSpaceLookupItem) (const SpSimTime &time,
                                                  SpDDMSpace *space, SpDDMRegion *region,
                                                  SpSpaceLookupItem* &space_lookup),
                              SpDDMGrid * (*RBCreate) (SpGridLookupOutput *),
                              RB_SpBinaryTree &grids);

  static void GetCurrentGrids(const SpSimTime &time,
                              SpDDMSpace *space, S_SpHLA *hla, SpDDMRegion *region,
                              void (*CreateSpaceLookupItem) (const SpSimTime &time,
                                                  SpDDMSpace *space, SpDDMRegion *region,
                                                  SpSpaceLookupItem* &space_lookup),
                              SpDDMGrid * (*RBCreate) (SpGridLookupOutput *),
                              RB_SpBinaryTree &grids);

  SpPublishRoutingSpace(S_SpHLA *hla, const SpSpace *space);

  double GetUpdateFreqHour(void) { return (-1.0); }
  double GetLookAheadHour(void)  { return ( 0.0); }
};

#endif
