// PublishSpace.C file

#include "SpPublishSpace.H"
#include "RB_Types.H"
#include "SpDDMGrid.H"
#include "SpDDMRegion.H"
#include "S_SpHLA.H"
#include "RB_SpBinaryTree.H"
#include "RB_SpDefineClass.H"

RB_DEFINE_CLASS(SpPublishGrid);
RB_DEFINE_CLASS(SpPublishRegion);


SpPublishSpace::SpPublishSpace(void)
  : UpdateSemaphore(0), NumPublishReply(0), NumPublishReplyPending(0),
    Grids(new RB_SpBinaryTree), Regions(new RB_SpBinaryTree){}

SpPublishSpace::SpPublishSpace(S_SpHLA *hla, const SpSpace *space)
  : SpDDMSpace(hla, space), 
    UpdateSemaphore(0), NumPublishReply(0), NumPublishReplyPending(0),
    Grids(new RB_SpBinaryTree), Regions(new RB_SpBinaryTree){}

SpPublishSpace::~SpPublishSpace(void) {
  SpPublishRegion *region;
  SpPublishGrid   *grid;

  while ((grid = (SpPublishGrid *) Grids->RemoveFirstElement()))
    delete grid;
  delete Grids;
  while ((region = (SpPublishRegion *) Regions->RemoveFirstElement()))
    delete region;
  delete Regions;
}

SpPublishGrid *
SpPublishSpace::FindGrid(int global_id){
  SpPublishGrid *grid = (SpPublishGrid *) Grids->Find(global_id);
  return (grid);
}

int
SpPublishSpace::DeleteGrid(int global_id){
  int deleted = 0;
  SpPublishGrid *grid = (SpPublishGrid *) Grids->Remove(global_id);
  if (grid){
    RB_DELETE_SpPublishGrid(grid);
    deleted = 1;
  }
  return (deleted);
}

SpPublishGrid *
SpPublishSpace::RemoveGrid(int global_id){
  return ((SpPublishGrid *) Grids->Remove(global_id));
}

void
SpPublishSpace::AddGrid(SpPublishGrid *grid){
  if (Grids->Find(grid->GetGlobalId())){
    cerr << "SpPublishSpace - grid already exist : " << grid->GetGlobalId() << endl;
    abort();
  }
  Grids->Insert(grid, grid->GetGlobalId());
}

void
SpPublishSpace::Update() {
  SetUpdateSemaphore(1);
}

SpPublishRegion *
SpPublishSpace::CreateRegion(const char *name) {
  SpPublishRegion *region = FindRegion(name);

  if (region==(SpPublishRegion *) 0){
    region = RB_NEW_SpPublishRegion(new SpPublishRegion(this, name));
    Regions->Insert(region, (char *) region->GetName());
  }

  return (region);
}

SpPublishRegion *
SpPublishSpace::FindRegion(const char *name) {
  return ((SpPublishRegion *) Regions->Find((char *) name));
}

int
SpPublishSpace::DeleteRegion(const char *name) {
  int deleted = 0;
  SpPublishRegion *region = (SpPublishRegion *) Regions->Remove((char *) name);

  if (region){
    RB_DELETE_SpPublishRegion(region);
    deleted = 1;
  }
  return (deleted);
}

void
SpPublishSpace::Debug(ostream &os){
  os << "\t" << "sim obj         : " 
     << GetSimObj()->get_NAME() << " (" << GetSimObj()->get_GLOBAL_ID() << ")" << endl;
  os << "\t" << "pub space name  : " << GetName() << endl;
  os << "\t" << "num regions     : " << Regions->GetNumElements() << endl;
  os << "\t" << "num grids       : " << Grids->GetNumElements() << endl;
  SpPublishGrid *grid = (SpPublishGrid *) Grids->GetFirstElement();
  while (grid){
    grid->Debug(os);
    grid = (SpPublishGrid *) Grids->GetNextElement();
  }
  os << endl;
}

void
SpPublishSpace::Validate(SIMTIME &time){
  const int publisher_global_id = GetSimObj()->get_GLOBAL_ID();
  SpPublishGrid *grid = (SpPublishGrid *) Grids->GetFirstElement();
  while (grid){
    grid->Validate(time, publisher_global_id, GetName());
    grid = (SpPublishGrid *) Grids->GetNextElement();
  }
}

int 
SpPublishSpace::GetNumGrids(void) const{
  return ((Grids) ? Grids->GetNumElements() : 0);
}
