// SpPublishTheaterSpace.C file

#include "SpPublishRoutingSpace.H"
#include "SpPublishTheaterSpace.H"
#include "S_SpHLA.H"
#include "M_SpPublishSpace.H"
#include "SpScheduleEvent.H"
#include "SpGridLookupOutput.H"
#include "SpSpaceLookupItem.H"
#include "SpLatLngLookupItem.H"
#include "SpDDMGrid.H"
#include "SpConvert.H"
#include "SpProxyDefines.H"
#include "RB_SpDefineClass.H"

RB_DEFINE_CLASS(SpPublishTheaterSpace);

#ifdef TODO
EVENT_TYPE (PublishSpace);
#else
#include "P_PublishSpace.H"
#endif

extern O_SpGrid *Grid_OBJECT_MANAGER;

float SpPublishTheaterSpace::ResolutionScaleFactor = -1.0;

SpPublishTheaterSpace::SpPublishTheaterSpace() {}

SpPublishTheaterSpace::SpPublishTheaterSpace(const SpSimTime &time, 
                                             S_SpHLA *hla, const SpSpace *space)
  : SpPublishRoutingSpace(hla, space) {

  M_SpPublishSpace msg(this);

  ScheduleProcess(time, PublishSpace, ObjectType, LocalId, Node, &msg);
}

SpPublishTheaterSpace::~SpPublishTheaterSpace() {}

double
SpPublishTheaterSpace::GetUpdateFreqHour(){
  const double e = 1E-5;
  double expansion_interval_hour;
  S_SpHLA *hla = GetSimObj();

  expansion_interval_hour = (hla->GetMaxSpeedKnots()<e)
                            ? hla->GetMinRescheduleTimeHour()
                            : hla->GetMinExpansionNmi() / hla->GetMaxSpeedKnots();
  if (expansion_interval_hour<hla->GetMinRescheduleTimeHour())
    expansion_interval_hour = hla->GetMinRescheduleTimeHour();

  return (expansion_interval_hour);
}

void
SpPublishTheaterSpace::CreateSpaceLookupItem(const SpSimTime &time, SpDDMSpace *space, SpDDMRegion *region,
                                             SpSpaceLookupItem* &space_lookup) {
  SpDimLookupItem *lookup_item;
  double expansion_radius_nmi, resolution_radius_nmi;
  S_SpHLA *hla = space->GetSimObj();
  double latRad, lonRad, altKm;
  double max_sensor_range = 0.0; // hla->GetMaxSensorRangeNmi();

  if (ResolutionScaleFactor<0.0){
    ResolutionScaleFactor = Grid_OBJECT_MANAGER->GetResolutionScaleFactor();
  }

  SpPublishRoutingSpace::CreateSpaceLookupItem(time, space, region, space_lookup);

  expansion_radius_nmi = (space->GetUpdateFreqHour() + space->GetLookAheadHour()) * hla->GetMaxSpeedKnots();
  expansion_radius_nmi += max_sensor_range;
  resolution_radius_nmi = expansion_radius_nmi * ResolutionScaleFactor;

  if (expansion_radius_nmi<1.0)  expansion_radius_nmi  = 1.0;
  if (resolution_radius_nmi<1.0) resolution_radius_nmi = 1.0;

  hla->GetPositionRadKm(time.GetTime(), latRad, lonRad, altKm);

  lookup_item = new SpLatLngLookupItem(latRad, lonRad, altKm,
                                       NauticalMiles_to_Kilometers(expansion_radius_nmi),
                                       NauticalMiles_to_Kilometers(resolution_radius_nmi));

  space_lookup->InsertDimLookupItem(lookup_item);
}

void 
SpPublishTheaterSpace::GetCurrentGrids(const SpSimTime &time, RB_SpBinaryTree &grids){
  if (Regions->GetNumElements()<=0)
    SpPublishRoutingSpace::GetCurrentGrids(time, this, (S_SpHLA *) SimObj, (SpDDMRegion *) 0,
                                         &SpPublishTheaterSpace::CreateSpaceLookupItem,
                                         &SpPublishGrid::RBCreate, grids);
  else
    SpPublishRoutingSpace::GetCurrentGrids(time, this, (S_SpHLA *) SimObj, *Regions,
                                         &SpPublishTheaterSpace::CreateSpaceLookupItem,
                                         &SpPublishGrid::RBCreate, grids);
}

double
SpPublishTheaterSpace::GetLookAheadHour(){
  return (GetSimObj()->GetLookAheadHour());
}

void RB_Register_SpPublishTheaterSpace()
{
  RB_REGISTER_CLASS_SpPublishTheaterSpace();
}
