// SpScheduleGridEvents.C 

#include "M_SpGrid.H"
#include "M_SpPublishGrid.H"
#include "M_SpUnPublishGrid.H"
#include "M_SpUnSubscribeGrid.H"
#include "M_SpSubscribeGrid.H"

#include "SpScheduleEvent.H"
#include "RB_SpList.H"
#include "SpDDMGrid.H"
#include "F_SpProxyItem.H"
#include "SpProxyObjHandle.H"
#include "O_SpGrid.H"

#ifdef TODO
EVENT_TYPE (PublishGrid);
EVENT_TYPE (UnPublishGrid);
EVENT_TYPE (SubscribeGrid);
EVENT_TYPE (UnSubscribeGrid);
#else
#include "E_SpPublishGrid.H"
#include "E_SpUnPublishGrid.H"
#include "E_SpSubscribeGrid.H"
#include "E_SpUnSubscribeGrid.H"
#endif

void SchedulePublishGridEvents(const SpSimTime &time, 
                               int node, int type, int local_id, int global_id, int class_id, 
                               int space_id,
                               double look_ahead,
                               RB_SpList &grids){
  char *buffer;
  int buffer_len;
  SpPublishGrid *grid;

  M_SpPublishGrid msg(node, type, local_id, global_id, space_id, class_id);
  msg.SetLookAhead(look_ahead);

#define DISABLE_AUTO_TIE_BREAKING
#ifdef  DISABLE_AUTO_TIE_BREAKING
  SpSimTime event_time(time);
  if (look_ahead>0)
    event_time.DisableAutoTieBreaking();
#endif

  grid = (SpPublishGrid *) grids.GetFirstElement();
  while (grid){
//  SpPublishGrid::CreateMsg(grid, buffer, buffer_len);
    grid->CreateMessage(buffer, buffer_len);
#ifdef  DISABLE_AUTO_TIE_BREAKING
    ScheduleEvent(event_time, PublishGrid,
#else
    ScheduleEvent(time, PublishGrid,
#endif
                  grid->GetObjectType(), grid->GetLocalId(), grid->GetNode(), 
                  msg, buffer, buffer_len);
    if (buffer) delete [] buffer;
    grid = (SpPublishGrid *) grids.GetNextElement();
  }
}

void ScheduleUnPublishGridEvents(const SpSimTime &time, 
                               int /* node */, int /* type */, int /* local_id */, int global_id, 
                               int /* space_id */,
                               double look_ahead,
                               RB_SpList &grids){
  char *buffer;
  int buffer_len;
  SpPublishGrid *grid;

  M_SpUnPublishGrid msg(global_id);

#ifdef  DISABLE_AUTO_TIE_BREAKING
  SpSimTime event_time(time);
  if (look_ahead>0)
    event_time.DisableAutoTieBreaking();
#endif

  grid = (SpPublishGrid *) grids.GetFirstElement();
  while (grid){

    SpDDMGrid::CreateMsgLookupId(grid, buffer, buffer_len);
    msg.SetNumLookupItem(grid->GetNumLookup());
    
#ifdef  DISABLE_AUTO_TIE_BREAKING
    ScheduleEvent(event_time, UnPublishGrid,
#else
    ScheduleEvent(time, UnPublishGrid,
#endif
                  grid->GetObjectType(), grid->GetLocalId(), grid->GetNode(), 
                  msg, buffer, buffer_len);

    if (buffer) delete [] buffer;
    grid = (SpPublishGrid *) grids.GetNextElement();
  }
}

void ScheduleSubscribeGridEvents(const SpSimTime &time,
                               int node, int type, int local_id, int global_id, int class_id,
                               int space_id,
                               double look_ahead,
                               RB_SpList &grids){
  char *buffer;
  int buffer_len;
  SpSubscribeGrid *grid;

  M_SpSubscribeGrid msg(node, type, local_id, global_id, space_id, class_id);
  msg.SetLookAhead(look_ahead);

#ifdef  DISABLE_AUTO_TIE_BREAKING
  SpSimTime event_time(time);
  if (look_ahead>0)
    event_time.DisableAutoTieBreaking();
#endif

  grid = (SpSubscribeGrid *) grids.GetFirstElement();
  while (grid){
//    SpSubscribeGrid::CreateMsg(grid, buffer, buffer_len);
    grid->CreateMessage(buffer, buffer_len);
#ifdef  DISABLE_AUTO_TIE_BREAKING
    ScheduleEvent(event_time, SubscribeGrid,
#else
    ScheduleEvent(time, SubscribeGrid,
#endif
                  grid->GetObjectType(), grid->GetLocalId(), grid->GetNode(), 
                  msg, buffer, buffer_len);
    if (buffer) delete [] buffer;
    grid = (SpSubscribeGrid *) grids.GetNextElement();
  }
}

void ScheduleUnSubscribeGridEvents(const SpSimTime &time,
                                   int global_id, int space_id, double look_ahead,
                                   RB_SpList &grids){
  char *buffer;
  int buffer_len;
  SpSubscribeGrid *grid;

  M_SpUnSubscribeGrid msg(global_id, space_id);
  msg.SetLookAhead(look_ahead);

#ifdef  DISABLE_AUTO_TIE_BREAKING
  SpSimTime event_time(time);
  event_time.DisableAutoTieBreaking();
#endif

  grid = (SpSubscribeGrid *) grids.GetFirstElement();
  while (grid){

    SpDDMGrid::CreateMsgLookupId(grid, buffer, buffer_len);
    msg.SetNumLookupItem(grid->GetNumLookup());

#ifdef  DISABLE_AUTO_TIE_BREAKING
    ScheduleEvent(event_time, UnSubscribeGrid,
#else
    ScheduleEvent(time, UnSubscribeGrid,
#endif
                  grid->GetObjectType(), grid->GetLocalId(), grid->GetNode(), 
                  msg, buffer, buffer_len);
    if (buffer) delete [] buffer;
    grid = (SpSubscribeGrid *) grids.GetNextElement();
  }
}
