// SpSubscribeTheaterSpace.C file

#include "SpPublishTheaterSpace.H"
#include "SpSubscribeTheaterSpace.H"
#include "M_SpSubscribeSpace.H"
#include "SpScheduleEvent.H"
#include "SpPublishRoutingSpace.H"
#include "SpDDMGrid.H"
#include "S_SpHLA.H"
#include "RB_SpDefineClass.H"
#include "SpDimLookupItem.H"
#include "O_SpGrid.H"
#include "SpLatLngLookupItem.H"
#include "SpSpaceLookupItem.H"

RB_DEFINE_CLASS(SpSubscribeTheaterSpace);

#ifdef TODO
EVENT_TYPE (SubscribeSpace);
#else
#include "P_SubscribeSpace.H"
#endif

extern O_SpGrid *Grid_OBJECT_MANAGER;

float SpSubscribeTheaterSpace::ResolutionScaleFactor = -1.0;

SpSubscribeTheaterSpace::SpSubscribeTheaterSpace() {}

SpSubscribeTheaterSpace::SpSubscribeTheaterSpace(const SpSimTime &time, S_SpHLA *hla, 
                                                 const SpSpace *space)
  : SpSubscribeRoutingSpace(hla, space) {

  M_SpSubscribeSpace msg(this);
  ScheduleProcess(time, SubscribeSpace, ObjectType, LocalId, Node, &msg);
}

SpSubscribeTheaterSpace::~SpSubscribeTheaterSpace() {}

void
SpSubscribeTheaterSpace::GetCurrentGrids(const SpSimTime &time, RB_SpBinaryTree &grids){

  if (Regions->GetNumElements()<=0)
    SpPublishRoutingSpace::GetCurrentGrids(time, this, (S_SpHLA *) SimObj, (SpDDMRegion *) 0,
                  &SpSubscribeTheaterSpace::CreateSpaceLookupItem,
                  &SpSubscribeGrid::RBCreate, grids);
  else
    SpPublishRoutingSpace::GetCurrentGrids(time, this, (S_SpHLA *) SimObj, *Regions,
                  &SpSubscribeTheaterSpace::CreateSpaceLookupItem,
                  &SpSubscribeGrid::RBCreate, grids);
}

void
SpSubscribeTheaterSpace::CreateSpaceLookupItem(const SpSimTime &time, SpDDMSpace *space, SpDDMRegion *region,
                                               SpSpaceLookupItem* &space_lookup) {
  SpDimLookupItem *lookup_item;
  double expansion_radius_nmi, resolution_radius_nmi;
  S_SpHLA *hla = space->GetSimObj();
  double latRad, lonRad, altKm;
  double max_sensor_range = hla->GetMaxSensorRangeNmi();

  if (ResolutionScaleFactor<0.0){
    ResolutionScaleFactor = Grid_OBJECT_MANAGER->GetResolutionScaleFactor();
  }

  if (max_sensor_range<1.0) max_sensor_range = 1.0;

  SpPublishRoutingSpace::CreateSpaceLookupItem(time, space, region, space_lookup);

  expansion_radius_nmi = (space->GetUpdateFreqHour() + space->GetLookAheadHour()) * hla->GetMaxSpeedKnots();
  expansion_radius_nmi += max_sensor_range;
  resolution_radius_nmi = expansion_radius_nmi * ResolutionScaleFactor;

  hla->GetPositionRadKm(time.GetTime(), latRad, lonRad, altKm);

  lookup_item = new SpLatLngLookupItem(latRad, lonRad, altKm,
                                       NauticalMiles_to_Kilometers(expansion_radius_nmi),
                                       NauticalMiles_to_Kilometers(resolution_radius_nmi));

  space_lookup->InsertDimLookupItem(lookup_item);
}

double
SpSubscribeTheaterSpace::GetUpdateFreqHour(){
  const double e = 1E-5;
  double expansion_interval_hour;
  S_SpHLA *hla = GetSimObj();

  expansion_interval_hour = (hla->GetMaxSpeedKnots()<e)
                            ? hla->GetMinRescheduleTimeHour()
                            : hla->GetMinExpansionNmi() / hla->GetMaxSpeedKnots();
  if (expansion_interval_hour<hla->GetMinRescheduleTimeHour())
    expansion_interval_hour = hla->GetMinRescheduleTimeHour();

  return (expansion_interval_hour);
}

double
SpSubscribeTheaterSpace::GetLookAheadHour(){
  return (GetSimObj()->GetLookAheadHour());
}

void RB_Register_SpSubscribeTheaterSpace()
{
  RB_REGISTER_CLASS_SpSubscribeTheaterSpace();
}

