// DDMGrid.H header file 

#ifndef DDMGRID_H
#define DDMGRID_H

#include "RB_SpBinaryTree.H"
#include "RB_SpList.H"
#include "SpList.H"
#include "RB_Types.H"
#include "SpSimTime.H"
#include "SpList.H"

class SpGridLookupOutput;
class MAIN_PUBSUB_ITEM;
class SpSubscriberHandle;
class S_SpHLA;
class F_SpProxyItem;

class SpDDMGrid {
  public:
    SpDDMGrid(void);
//    SpDDMGrid(SpGridLookupOutput *_grid);
//    SpDDMGrid(int local_id, int node, int object_type, int global_id);
    SpDDMGrid(int local_id, int node, int object_type);

    virtual ~SpDDMGrid(void);

    virtual void CreateMessage(char* &msg, int &len);

    int Add(SpGridLookupOutput *grid);

//  int GetGlobalId(void);
    int GetLocalId(void)    const { return (LocalId);    }
    int GetNode(void)       const { return (Node);       }
    int GetObjectType(void) const { return (ObjectType); }

//    static void CompareGlobalId(RB_SpBinaryTree &orig_grids,	// input
//                        RB_SpBinaryTree &new_grids,		// input
//                        SpList &grids_add,			// output
//                        SpList &grids_remove,			// output
//                        SpList &grids_reprocess);		// output

    static void CreateObj(const char *msg, int &obj_bytes, SpDDMGrid* &_grid);
//    static void CreateMsgObjHandle(SpDDMGrid *grid, char* &msg, int &len);
//    static void CreateMsgGlobalId(SpDDMGrid *grid, char* &msg, int &len);
    static void CreateMsgLookupId(SpDDMGrid *grid, char* &msg, int &len);

    int GetNumLookup(void) { return (Lookups.GetNumElements()); }

    RB_SpList &GetLookups(void)       { return (Lookups); }
    SpList    &GetRemoveLookups(void) { return (RemoveLookups); }

//    int operator==(SpDDMGrid& g);

    int Update(SpDDMGrid *grid);

    void Debug(ostream &o);

//    static int GetGlobalId(void *item) { return (((SpDDMGrid *) item)->GetGlobalId()); }

  protected:
    RB_SpList Lookups;
    SpList    RemoveLookups;

  private:
    int LocalId, Node, ObjectType;
//  int GetGridCoverP(void);

//    SpGridLookupOutput *GetLookup(void);
};

class SpPublishGrid : public SpDDMGrid {
  public:
    SpPublishGrid(void);
//    SpPublishGrid(SpGridLookupOutput *_grid);
//    SpPublishGrid(int local_id, int node, int object_type, int global_id);
    SpPublishGrid(int local_id, int node, int object_type);
    ~SpPublishGrid(void);

    RB_SpBinaryTree &GetSubscribers(void) { return (subscribers); }

    int AddLookupSubscriber(int _node, int _type, int _local_id, int _global_id);
    void DeleteLookupSubscribers(void);
    int GetNumLookupSubscribers(void) { return (lookup_subscribers.GetNumElements()); }

    int AddNewSubscriber(int _node, int _type, int _local_id, int _global_id);
    void DeleteNewSubscribers(void);
    int GetNumNewSubscribers(void) { return (new_subscribers.GetNumElements()); }

    int RemoveSubscriber(const SpSimTime &time, float look_ahead,
                        S_SpHLA *hla, int space_id, int global_id);

    static int RemoveSubscriber(const SpSimTime &time, float look_ahead,
			S_SpHLA *hla, int space_id, RB_SpList &grids);

    static int AddSubscriber(SpSimTime &time, float look_ahead,
			S_SpHLA *hla, int space_id, SpBinaryTree &grids);
    static int AddRemoveSubscriber(const SpSimTime &time, float look_ahead,
			S_SpHLA *hla, int space_id, RB_SpList &grids);

    void CreateMessage(char* &msg, int &len);

    static void CreateMsg(SpPublishGrid *grid, char* &msg, int &len);
    static void CreateMsg(SpBinaryTree &items, char* &buffer, int &buffer_len);
//    static void CreateMsgObjHandle(SpBinaryTree &items, char* &buffer, int &buffer_len);
//    static void CreateMsgGlobalId(SpBinaryTree &items, char* &buffer, int &buffer_len);

    static void CreateObj(const char* &buffer, int num_items, int &obj_bytes, SpBinaryTree &items);
    static void CreateObj(const char *msg, int &obj_bytes, SpPublishGrid* &item);

    static SpDDMGrid * RBCreate(SpGridLookupOutput *item);

    void Debug(ostream &os);
    void Validate(SIMTIME &time, int publisher_global_id, const char *space_name);

    static void Destroy(SpPublishGrid *grid);

  protected:

  private:

      // contains pointers to SpSubscriberHandle in ProxySubscriber of S_SpHLA
    RB_SpBinaryTree subscribers;

      // new subscribers to this grid (SpProxyObjectHandle)
    RB_SpBinaryTree lookup_subscribers;

///////////////////////////////
// PENDING
    RB_SpBinaryTree new_subscribers;
///////////////////////////////

//    static void CreateMsgObjHandle(SpPublishGrid *grid, char* &msg, int &len);
//    static void CreateMsgGlobalId(SpPublishGrid *grid, char* &msg, int &len);
    static void CreateMsg(SpBinaryTree &items,
                          void (*CreateMsg) (SpPublishGrid *grid, char* &msg, int &len),
                          char* &buffer, int &buffer_len);

    int AddSubscriber(SpSubscriberHandle *subscriber);
    SpSubscriberHandle *AddSubscriber(int global_id);
    SpSubscriberHandle *RemoveSubscriber(int global_id);

};

class SpSubscribeGrid : public SpDDMGrid {
  public:
    SpSubscribeGrid(void);
//    SpSubscribeGrid(SpGridLookupOutput *_grid);

    ~SpSubscribeGrid(void);

    void CreateMessage(char* &msg, int &len);

    int Add(F_SpProxyItem *item);
    int Remove(int global_id);
    F_SpProxyItem *Find(int global_id);
    int GetNumRemoteObjectProxies(void) { return (RemoteObjectProxies.GetNumElements()); }

    RB_SpBinaryTree &GetPublishers(void) { return (Publishers); }

    static void CreateMsg(SpSubscribeGrid *grid, char* &msg, int &len);
    static void CreateMsg(SpBinaryTree &items, char* &buffer, int &buffer_len);

    static void CreateObj(const char* &buffer, int num_items, int &obj_bytes, SpBinaryTree &items);
    static void CreateObj(const char *msg, int &obj_bytes, SpSubscribeGrid* &item);

    static SpDDMGrid * RBCreate(SpGridLookupOutput *item);

    static void Destroy(SpSubscribeGrid *grid);

    void Debug(ostream &os);
    void Validate(SIMTIME &time, int subscriber_global_id, const char *space_name);

  protected:

  private:
    RB_SpBinaryTree RemoteObjectProxies;
    RB_SpBinaryTree Publishers;

    static void CreateMsg(SpBinaryTree &items,
                          void (*CreateMsg) (SpSubscribeGrid *grid, char* &msg, int &len),
                          char* &buffer, int &buffer_len);
};

#endif

