// SpPublishRoutingSpace.C file

#include "SpScheduleEvent.H"
#include "SpPublishRoutingSpace.H"
#include "S_SpHLA.H"
#include "SpDDMRegion.H"
#include "SpSpaceLookupItem.H"
#include "SpDoubleLookupItem.H"
#include "SpEnumLookupItem.H"
#include "SpAltitudeLookupItem.H"
#include "SpGridLookupOutput.H"
#include "SpDDMGrid.H"
#include "M_SpPublishSpace.H"
#include "SpDDMDim.H"
#include "RB_SpBinaryTree.H"
#include "RB_SpDefineClass.H"

RB_DEFINE_CLASS(SpPublishRoutingSpace);

#ifdef TODO
EVENT_TYPE (PublishSpace);
#else
#include "P_PublishSpace.H"
#endif

extern O_SpGrid *Grid_OBJECT_MANAGER;

SpPublishRoutingSpace::SpPublishRoutingSpace() {}

SpPublishRoutingSpace::SpPublishRoutingSpace(S_SpHLA *hla, const SpSpace *space)
  : SpPublishSpace(hla, space) {}

SpPublishRoutingSpace::SpPublishRoutingSpace(const SpSimTime &time, 
                                             S_SpHLA *hla, const SpSpace *space)
  : SpPublishSpace(hla, space) {
  M_SpPublishSpace msg(this);
  ScheduleProcess(time, PublishSpace, ObjectType, LocalId, Node, &msg);
}

SpPublishRoutingSpace::~SpPublishRoutingSpace() {}

void
SpPublishRoutingSpace::CreateSpaceLookupItem(const SpSimTime &, SpDDMSpace *space, SpDDMRegion *region,
                                             SpSpaceLookupItem* &space_lookup) {
  SpDDMDim             *dimension;
  SpDDMDoubleDimension *double_dimension;
  SpDDMEnumDimension   *enum_dimension;
  SpDimLookupItem      *lookup_item;
  SpEnumLookupItem     *enum_lookup_item;
  SpDDMEnumDimension   *region_enum;
  RB_SpBinaryTree      *enums;

  space_lookup = new SpSpaceLookupItem(space->GetSpace()->get_SpaceName());

  if (region==(SpDDMRegion *) 0) return;

  RB_SpBinaryTree &dimensions = region->GetDimensions();
  dimension = (SpDDMDim *) dimensions.GetFirstElement();
  while (dimension){
    switch (dimension->GetType()){
      case SpDimLookupItem::Double:
        double_dimension = (SpDDMDoubleDimension *) dimension;
        lookup_item = new SpDoubleLookupItem(double_dimension->GetName(), 
                                           double_dimension->GetLo(), 
                                           double_dimension->GetHi(), 
                                           double_dimension->GetRes());
        break;
      case SpDimLookupItem::Enum:
        enum_dimension = (SpDDMEnumDimension *) dimension;
        enum_lookup_item = new SpEnumLookupItem(enum_dimension->GetName(), *space_lookup);
        enums = enum_dimension->GetEnums();
        region_enum = (SpDDMEnumDimension *) enums->GetFirstElement();
        while (region_enum){
          enum_lookup_item->assign_enums(region_enum->GetName(), 1, (void *) 0);
          region_enum = (SpDDMEnumDimension *) enums->GetNextElement();
        } 
        lookup_item = enum_lookup_item;
        break;
      default:;
    }

    space_lookup->InsertDimLookupItem(lookup_item);
    dimension = (SpDDMDim *) dimensions.GetNextElement();
  }
}

void
SpPublishRoutingSpace::GetCurrentGrids(const SpSimTime &time, RB_SpBinaryTree &grids){
  int use_default_region = (Regions->GetNumElements()<=0);
  const char *default_region_name = "Default Region";

  if (use_default_region)
    CreateRegion(default_region_name);

  GetCurrentGrids(time, this, (S_SpHLA *) SimObj, *Regions,
                  &CreateSpaceLookupItem,
                  &SpPublishGrid::RBCreate, grids);

  if (use_default_region)
    DeleteRegion(default_region_name);
}

void 
SpPublishRoutingSpace::GetCurrentGrids(const SpSimTime &time,
                         SpDDMSpace *space, S_SpHLA *hla, RB_SpBinaryTree &regions, 
                         void (*CreateSpaceLookupItem) (const SpSimTime &, SpDDMSpace *, 
                                                        SpDDMRegion *, SpSpaceLookupItem* &),
                         SpDDMGrid* (*RBCreate) (SpGridLookupOutput *), RB_SpBinaryTree &grids){
  SpDDMRegion *region;

  region = (SpDDMRegion *) regions.GetFirstElement();
  while (region){
    if (region->GetNumDimensions()<=0){
      region = (SpDDMRegion *) regions.GetNextElement();
      continue;
    }
    GetCurrentGrids(time, space, hla, region, CreateSpaceLookupItem, RBCreate, grids);
    region = (SpDDMRegion *) regions.GetNextElement();
  }
}

void 
SpPublishRoutingSpace::GetCurrentGrids(const SpSimTime &time,
                         SpDDMSpace *space, S_SpHLA *hla, SpDDMRegion *region,
                         void (*CreateSpaceLookupItem) (const SpSimTime &, SpDDMSpace *, 
                                                        SpDDMRegion *, SpSpaceLookupItem* &),
                         SpDDMGrid* (*RBCreate) (int, int, int), RB_SpBinaryTree &grids){
  SpSpaceLookupItem *space_lookup;
  C_QUEUE lookup_items;
  SpDDMGrid *grid, *match_grid;
  SpGridLookupOutput *item;
  int unique_id = 0;

  CreateSpaceLookupItem(time, space, region, space_lookup);
  Grid_OBJECT_MANAGER->HiGrid_Lookup(*space_lookup, &lookup_items);

  while (lookup_items.length()>0){
    item = (SpGridLookupOutput *) lookup_items.pop_top();
    item->get_MainPubSubItem()->init(hla->get_NODE(), hla->get_OBJECT_TYPE(),
               hla->get_LOCAL_ID(), hla->get_GLOBAL_ID(),
               hla->GetObjectProxy()->GetObjectTemplate()->GetClassID());

    match_grid = (SpDDMGrid *) 0;
    grid = (SpDDMGrid *) grids.GetFirstElement();
    while (grid){
      if ( (grid->GetLocalId()   ==item->getLocalId()) &&
           (grid->GetNode()      ==item->getNode())    &&
           (grid->GetObjectType()==item->getObjType()) ){
        match_grid = grid;
        break;
      }
      grid = (SpDDMGrid *) grids.GetNextElement();
    }

    if (match_grid){
      if (!grid->Add(item)) SpGridLookupOutput::Destroy(item);
    }
    else{
      grid = RBCreate(item->getLocalId(), item->getNode(), item->getObjType());
      grids.Insert(grid, grid->GetGlobalId());
    }
  }

  delete space_lookup;
}


void RB_Register_SpPublishRoutingSpace()
{
  RB_REGISTER_CLASS_SpPublishRoutingSpace();
}

