// E_SpPublishGrid.C source file

#include "E_SpPublishGrid.H"
#include "SpDDMGrid.H"
#include "S_SpGrid.H"
#include "SpScheduleEvent.H"
#include "SpGridLookupOutput.H"
#include "SpList.H"
#include "SpPubSubItem.H"
#include "SpProxyObjHandle.H"
#include "SpExecute.H"
#include "RB_SpDefineClass.H"

#ifdef TODO
EVENT_TYPE (AddSubscriberSpace);
#else
#include "E_AddSubscriberSpace.H"
#endif
 
RB_DEFINE_CLASS(MAIN_PUBSUB_ITEM);


//#define DDM_DEBUG

E_SpPublishGrid::E_SpPublishGrid() {
//  set_lazy();
}

int E_SpPublishGrid::lazy() {
  return (0);
}

void E_SpPublishGrid::process() {
  S_SpGrid *s_grid     = (S_SpGrid *)        SIMOBJ;
  M_SpPublishGrid *msg = (M_SpPublishGrid *) get_MESSAGE();
  const char *data     = (const char *)      get_DATA();
  MAIN_PUBSUB_ITEM *item, *lookup_item, *publish_item;
  int i, nbytes, buffer_len;
  SpPublishGrid *grid;
  NET_INT *remove_lookup;
  SpList subscribers;
  SpPubSubItem *s;
  char *buffer;
  SpGridLookupOutput *lookup;
  SpBinaryTree new_subscribers;
  SpProxyObjHandle *obj_handle;

  SpPublishGrid::CreateObj(data, nbytes, grid);

  SpList &remove_lookups = grid->GetRemoveLookups();
  remove_lookup = (NET_INT *) remove_lookups.GetFirstElement();    
  while (remove_lookup){
    item = (MAIN_PUBSUB_ITEM *)
           s_grid->GetHiGrid()->RemovePublisher(*remove_lookup);
    if (item)
      RB_DELETE_MAIN_PUBSUB_ITEM(item);
    remove_lookup = (NET_INT *) remove_lookups.GetNextElement();
  }

  RB_SpBinaryTree &lookups = grid->GetLookups();

#ifdef DDM_DEBUG
  cout << time_tag << " E_SpPublishGrid simobj " << msg->GlobalId 
       << " lookup grid " << grid->GetGlobalId() 
       << " node simobj/grid : " << msg->Node << "/" << get_NODE() 
       << " subscribers ";
#endif

  lookup = (SpGridLookupOutput *) lookups.GetFirstElement();
  while (lookup){
    lookup_item  = lookup->get_MainPubSubItem();

    publish_item = RB_NEW_MAIN_PUBSUB_ITEM(new MAIN_PUBSUB_ITEM(lookup_item->GetUniqueKey()));
    publish_item->init(msg->Node, msg->Type, msg->LocalId, msg->GlobalId, msg->ClassId);
    publish_item->AllocateNumDimensions(lookup_item->GetNdimensions());
    for (i=0; i<lookup_item->GetNdimensions(); i++)
      publish_item->SetLHR(i, lookup_item->GetLo(i), lookup_item->GetHi(i), lookup_item->GetRes(i));

    s_grid->GetHiGrid()->AddPublisherAndGet(publish_item, &subscribers);

    while ((s = (SpPubSubItem *) subscribers.RemoveTop())){
#ifdef DDM_DEBUG
      cout << s->get_global_id() << ", ";
#endif
      if (s->get_global_id()==msg->GlobalId) continue;
      if (new_subscribers.Find(s->get_global_id())){
        continue;
      }
      new_subscribers.Insert(new SpProxyObjHandle(s->get_local_id(), s->get_node(), 
                                                  s->get_type(), s->get_global_id()),
                             s->get_global_id());
    }

    lookup = (SpGridLookupOutput *) lookups.GetNextElement();
  }// lookup
#ifdef DDM_DEBUG
  cout << endl;
#endif

  SpSimTime event_time(get_MESSAGE()->time_tag);
  event_time += msg->GetLookAhead();

  buffer     = (char *) 0;
  buffer_len = 0;

  SpProxyObjHandle::CreateMessage(new_subscribers, buffer, buffer_len);

  M_SpAddSubscriberSpace add_subscriber_msg(msg->SpaceId, grid->GetGlobalId(), 1);
  add_subscriber_msg.SetNumSubscribers(new_subscribers.GetNumElements());
  ScheduleEvent(event_time, AddSubscriberSpace, 
                msg->Type, msg->LocalId, msg->Node,
                &add_subscriber_msg, buffer, buffer_len);  

  if (buffer) delete [] buffer;

  while ((obj_handle = (SpProxyObjHandle *) new_subscribers.RemoveFirstElement()))
    delete obj_handle;

  SpPublishGrid::Destroy(grid);
}

void E_SpPublishGrid::commit() {
}
