// E_SpSubscribeGrid.C source file

#include "E_SpSubscribeGrid.H"

#include "SpDDMGrid.H"
#include "S_SpGrid.H"
#include "SpScheduleEvent.H"
#include "SpGridLookupOutput.H"
#include "queue.H"
#include "SpProxyObjHandle.H"
#include "SpDDMUtils.H"
#include "SpPubSubItem.H"
#include "RB_SpDefineClass.H"

RB_DEFINE_CLASS(MAIN_PUBSUB_ITEM);
RB_DEFINE_CLASS(SpPublishGrid);

#ifdef TODO
EVENT_TYPE (AddSubscriberSpace);
EVENT_TYPE (UnSubscriberSpace);
#else
#include "E_AddSubscriberSpace.H"
#include "E_SpUnSubscribeSpace.H"
#endif

//#define DDM_DEBUG

E_SpSubscribeGrid::E_SpSubscribeGrid() {
//  set_lazy();
}

int E_SpSubscribeGrid::lazy() {
  return (0);
}

void E_SpSubscribeGrid::process() {
  S_SpGrid *s_grid       = (S_SpGrid *)          SIMOBJ;
  M_SpSubscribeGrid *msg = (M_SpSubscribeGrid *) get_MESSAGE();
  const char *data       = (const char *)        get_DATA();
  MAIN_PUBSUB_ITEM *item, *lookup_item, *subscribe_item;
  SpBinaryTree old_publishers, new_publishers, add_publishers, remove_publishers;
  NET_INT *remove_lookup;
  SpList publishers;
  SpSubscribeGrid *grid;
  SpGridLookupOutput *lookup;
  SpPubSubItem *pubsub_item;
  SpProxyObjHandle *obj_handle;
  char *buffer;
  int i, buffer_len, nbytes;
  int id;

  SpSubscribeGrid::CreateObj(data, nbytes, grid);

  SpSimTime event_time(get_MESSAGE()->time_tag);
  event_time += msg->GetLookAhead();

  // retrieve list of publishers based on previous subscription region
  // to be used for compare with the new publishers based on the new subscription 
  // region.

  SpList &remove_lookups = grid->GetRemoveLookups();
  remove_lookup = (NET_INT *) remove_lookups.GetFirstElement();
  while (remove_lookup){
    id = *remove_lookup;
    s_grid->GetHiGrid()->GetPublishersWithKey(id, &publishers);
    while ((pubsub_item = (SpPubSubItem *) publishers.RemoveTop())){
      if (pubsub_item->get_global_id()==msg->GlobalId) continue;
      obj_handle = new SpProxyObjHandle(pubsub_item->get_local_id(), 
                                        pubsub_item->get_node(), 
                                        pubsub_item->get_type(), 
                                        pubsub_item->get_global_id());
      old_publishers.Insert(obj_handle, obj_handle->GetGlobalId());
    }
    item = (MAIN_PUBSUB_ITEM *) 
           s_grid->GetHiGrid()->RemoveSubscriber(id);
    if (item) 
      RB_DELETE_MAIN_PUBSUB_ITEM(item);
    remove_lookup = (NET_INT *) remove_lookups.GetNextElement();
  }

  RB_SpBinaryTree &lookups = grid->GetLookups();

#ifdef DDM_DEBUG
  cout << time_tag << " E_SpSubscribeGrid simobj " << msg->GlobalId 
       << " lookup grid " << grid->GetGlobalId() << " publishers ";
#endif

  lookup = (SpGridLookupOutput *) lookups.GetFirstElement();
  while (lookup){

    lookup_item  = lookup->get_MainPubSubItem();

    subscribe_item = RB_NEW_MAIN_PUBSUB_ITEM(new MAIN_PUBSUB_ITEM(lookup_item->GetUniqueKey()));
    subscribe_item->init(msg->Node, msg->Type, msg->LocalId, msg->GlobalId, msg->ClassId);
    subscribe_item->AllocateNumDimensions(lookup_item->GetNdimensions());
    for (i=0; i<lookup_item->GetNdimensions(); i++)
      subscribe_item->SetLHR(i, lookup_item->GetLo(i), lookup_item->GetHi(i), lookup_item->GetRes(i));

    s_grid->GetHiGrid()->AddSubscriberAndGet(subscribe_item, &publishers);

    while ((pubsub_item = (SpPubSubItem *) publishers.RemoveTop())){
#ifdef DDM_DEBUG
      cout << pubsub_item->get_global_id() << ", ";
#endif
      if (pubsub_item->get_global_id()==msg->GlobalId) continue;
      if (new_publishers.Find(pubsub_item->get_global_id())){
         continue;
      }
      new_publishers.Insert(pubsub_item, pubsub_item->get_global_id());
    }

    lookup = (SpGridLookupOutput *) lookups.GetNextElement();
  }// lookup
#ifdef DDM_DEBUG
  cout << endl;
#endif

  Compare(&old_publishers, &new_publishers, 
          SpProxyObjHandle::GetGlobalId, SpPubSubItem::GetGlobalId,
          &add_publishers, &remove_publishers);

  SpProxyObjHandle new_subscriber(msg->LocalId, msg->Node, msg->Type, msg->GlobalId);
  new_subscriber.CreateMessage(buffer, buffer_len);

  M_SpAddSubscriberSpace add_subscriber_msg(msg->SpaceId, grid->GetGlobalId(), 0);
  add_subscriber_msg.SetNumSubscribers(1);
  add_subscriber_msg.SetLookAhead(msg->GetLookAhead());

  while ((pubsub_item = (SpPubSubItem *) add_publishers.RemoveFirstElement())){
    ScheduleEvent(event_time, AddSubscriberSpace, 
                  pubsub_item->get_type(), pubsub_item->get_local_id(), pubsub_item->get_node(),
                  &add_subscriber_msg, buffer, buffer_len);
  }

  if (buffer)
    delete [] buffer;

  M_SpRemSubscriberSpace rem_subscriber_space_msg(msg->SpaceId, msg->GlobalId, 1);
  rem_subscriber_space_msg.SetLookAhead(msg->GetLookAhead());

  NET_INT grid_global_id = s_grid->get_GLOBAL_ID();

  // notify publisher
  while (obj_handle = (SpProxyObjHandle *) remove_publishers.RemoveFirstElement()){
    ScheduleEvent(event_time, RemSubscriberSpace,
                  obj_handle->GetObjType(),
                  obj_handle->GetLocalId(),
                  obj_handle->GetNode(),
                  &rem_subscriber_space_msg, (char *) &grid_global_id, sizeof(NET_INT));
  }

  while ((obj_handle = (SpProxyObjHandle *) old_publishers.RemoveFirstElement()))
    delete obj_handle;

  SpSubscribeGrid::Destroy(grid);
}

void E_SpSubscribeGrid::commit() {
}
