// E_SpUnSubscribeGrid.C source file

#include "E_SpUnSubscribeGrid.H"

#include "O_SpGrid.H"
#include "S_SpGrid.H"
#include "SpBinaryTree.H"
#include "SpDDMGrid.H"
#include "M_SpUnSubscribeSpace.H"
#include "SpGridLookupOutput.H"
#include "SpScheduleEvent.H"
#include "RB_SpDefineClass.H"
 
RB_DEFINE_CLASS(MAIN_PUBSUB_ITEM);

#ifdef TODO
EVENT_TYPE (RemSubscriberSpace);
#else
#include "E_UnSubscribeSpace.H"
#endif

//#define DDM_DEBUG

E_SpUnSubscribeGrid::E_SpUnSubscribeGrid() 
  : valid(0) {
//  set_lazy();
}

int E_SpUnSubscribeGrid::lazy() { 
  return (0);
}

void E_SpUnSubscribeGrid::process() {
  S_SpGrid *s_grid         = (S_SpGrid *)            SIMOBJ;
  M_SpUnSubscribeGrid *msg = (M_SpUnSubscribeGrid *) get_MESSAGE();
  const NET_INT *id        = (NET_INT *)             get_DATA();
  SpPubSubItem *publisher;
  MAIN_PUBSUB_ITEM *item;
  SpList publishers;
  NET_INT grid_global_id;
  SpBinaryTree grid_publishers;
  int i, num_remove;

#ifdef DDM_DEBUG
  cout << time_tag << " E_SpUnSubscribeGrid simobj " << msg->GlobalId
       << " unsubscribe grid " << s_grid->get_GLOBAL_ID() << " publishers ";
#endif

  for (i=0; i<msg->NumLookupItem; i++){
    s_grid->GetHiGrid()->GetPublishersWithKey(id[i], &publishers);
    while ((publisher = (SpPubSubItem *) publishers.RemoveTop())){
#ifdef DDM_DEBUG
    cout << publisher->get_global_id() << ", ";
#endif
      if (publisher->get_global_id()==msg->GlobalId) continue;
      if (grid_publishers.Find(publisher->get_global_id())){
        continue;
      }
      grid_publishers.Insert(publisher, publisher->get_global_id());
    }
  }
#ifdef DDM_DEBUG
  cout << endl;
#endif

  SpSimTime event_time(get_MESSAGE()->time_tag);
  event_time += msg->GetLookAhead();

  M_SpRemSubscriberSpace rem_subscriber_space_msg(msg->SpaceId, msg->GlobalId, 1);
  rem_subscriber_space_msg.SetLookAhead(msg->GetLookAhead());

  grid_global_id = s_grid->get_GLOBAL_ID();

  // notify publisher
  while (publisher = (SpPubSubItem *) grid_publishers.RemoveFirstElement()){
    ScheduleEvent(event_time, RemSubscriberSpace, 
                  publisher->get_type(),
                  publisher->get_local_id(),
                  publisher->get_node(),
                  &rem_subscriber_space_msg, (char *) &grid_global_id, sizeof(NET_INT));
  }

  num_remove = 0;
  for (i=0; i<msg->NumLookupItem; i++){
    item = (MAIN_PUBSUB_ITEM *) 
           s_grid->GetHiGrid()->RemoveSubscriber(id[i]);
    if (item){
      RB_DELETE_MAIN_PUBSUB_ITEM(item);
      num_remove++;
    }
  }

  valid = (num_remove==msg->NumLookupItem);
}

void E_SpUnSubscribeGrid::commit() {
  S_SpGrid *s_grid         = (S_SpGrid *)            SIMOBJ;
  M_SpUnSubscribeGrid *msg = (M_SpUnSubscribeGrid *) get_MESSAGE();

  if (!valid){
    cout << time_tag << " ERROR::E_SpUnSubscribeGrid event processed invalid simobj " 
         << msg->GlobalId << " grid " << s_grid->get_GLOBAL_ID() << endl;
    abort();
  }
}
