// E_SpRemSubscriberSpace.C source file

#include "E_SpUnSubscribeSpace.H"
#include "S_SpHLA.H"
#include "SpDDMGrid.H"
#include "SpPublishSpace.H"


/******************************************************************************
* E_SpRemSubscriberSpace constructor                                          *
******************************************************************************/
E_SpRemSubscriberSpace::E_SpRemSubscriberSpace()
  : valid(0), errcode(0) {
//  set_lazy();
}

/******************************************************************************
* process: Optimistically process the event                                   *
******************************************************************************/

int E_SpRemSubscriberSpace::lazy() { 
  return (0);
}

void E_SpRemSubscriberSpace::process() { 
  const SpSimTime &time          = get_MESSAGE()->time_tag; 
  S_SpHLA *hla                   = (S_SpHLA *)              SIMOBJ;
  M_SpRemSubscriberSpace *msg    = (M_SpRemSubscriberSpace *) get_MESSAGE();
  const char *data               = (const char *)           get_DATA();
  SpPublishSpace *space          = hla->GetPublishSpace(msg->SpaceId);
  const NET_INT  *grid_global_id = (const NET_INT *) data;
  SpPublishGrid  *publish_grid;
  int ok;

  errcode = 0;
  ok   = (msg->NumGrid>0);
  for (int i=0; i<msg->NumGrid && ok; i++){
    publish_grid = space->FindGrid(grid_global_id[i]);

    if (publish_grid){
      ok &= publish_grid->RemoveSubscriber(time, msg->LookAhead, hla, msg->SpaceId, msg->GlobalId);
      if (!ok) errcode = 1;
    }
    else{
      errcode = 2;
      ok      = 0;
    }
  }

  valid = ok;
}

/******************************************************************************
* commit: commit the event                                                    *
******************************************************************************/
void E_SpRemSubscriberSpace::commit() {
  S_SpHLA *hla                   = (S_SpHLA *)              SIMOBJ;
  M_SpRemSubscriberSpace *msg    = (M_SpRemSubscriberSpace *) get_MESSAGE();
  const char *data               = (const char *)           get_DATA();
  const NET_INT  *grid_global_id = (const NET_INT *) data;

  if (!valid){
    cout << time_tag 
         << " ERROR::E_SpRemSubscriberSpace event processed invalid publisher " 
         << hla->get_GLOBAL_ID() << " remove subscriber " << msg->GlobalId 
         << " error code " << errcode 
         << " grid ";
    for (int i=0; i<msg->NumGrid; i++)
      cout << grid_global_id[i] << ", ";
    cout << endl;
    abort();
  }
}
