// P_SpSubscribeSpace.C

//#include "SpProc.H"
// should eventually refactor code to allow #include "SpProc.H"
#include "process.H"

#include "SpProcSem.H"
#include "P_SpSubscribeSpace.H"
#include "RB_SpBinaryTree.H"
#include "SpScheduleGridEvents.H"
#include "SpDDMGrid.H"
#include "S_SpHLA.H"
#include "SpSubscribeSpace.H"
#include "RB_SpDefineClass.H"
 
RB_DEFINE_CLASS(SpSubscribeGrid);

extern SpEvent *CurrentEvent;

//#define DDM_DEBUG
#define PROXY_LOOKAHEAD

PROCESS_LV (P_SpSubscribeSpace, process)

  int success_flag;
  int loop_forever;
  SpSubscribeSpace *space;
  RB_SpBinaryTree current_grids, lookup_grids, delete_grids;
  SpBinaryTree grids_added, grids_removed, grids_reprocess;
  SIMTIME event_time, unsubscribe_time, subscribe_time;
  RB_int first_cycle;		// TRUE, until interacted with the grid manager at least once.
  RB_int scheduled_events;	// TRUE, if interacted with the grid manager in this cycle.
  RB_double cycle_start_time;

BEGIN_PROCESS (P_SpSubscribeSpace, process, 4)

  double delta_t;
  SpSubscribeGrid *update_item, *grid_item;
  int node, object_type, local_id, global_id, class_id;
  double look_ahead;

  LV->space = ((M_SpSubscribeSpace *) MESSAGE)->SpSpace;

  if (!LV->space->GetUpdateSemaphore()->GetState()){
    WAIT_FOR_LOGICAL(1,
  	   *((LV->space)->GetUpdateSemaphore()),
	   -1.0,
	   LV->success_flag);
  }

  LV->loop_forever = 1;
  LV->first_cycle  = 1;
  while (LV->loop_forever) {

    LV->cycle_start_time = CurrentEvent->get_MESSAGE()->time_tag;
    LV->space->GetCurrentGrids(CurrentEvent->get_MESSAGE()->time_tag, LV->current_grids);

    SpDDMGrid::CompareGlobalId(*(LV->space->GetGrids()), LV->current_grids,
                       LV->grids_added, LV->grids_removed, LV->grids_reprocess);
#ifdef PROXY_LOOKAHEAD
    look_ahead     = ( LV->first_cycle || LV->space->GetUpdateSemaphore()->GetState() )
                     ? 0.0 : (LV->space->GetLookAheadHour() / 3.0) * 3600.0;
#else
    look_ahead     = ( LV->first_cycle || LV->space->GetUpdateSemaphore()->GetState() )
                     ? 0.0 : (LV->space->GetLookAheadHour() / 2.0) * 3600.0;
#endif
    LV->event_time = CurrentEvent->get_MESSAGE()->time_tag + look_ahead;
    node        = LV->space->GetSimObj()->get_NODE();
    object_type = LV->space->GetSimObj()->get_OBJECT_TYPE();
    local_id    = LV->space->GetSimObj()->get_LOCAL_ID();
    global_id   = LV->space->GetSimObj()->get_GLOBAL_ID();
    class_id    = LV->space->GetSimObj()->GetObjectProxy()->GetObjectTemplate()->GetClassID();

    // Compile a list of grids to be lookup which contains : 
    // added grids, and reprocess grids.
    grid_item = (SpSubscribeGrid *) LV->grids_added.GetFirstElement();
    while (grid_item){
      LV->space->AddGrid(grid_item);
      LV->lookup_grids.Insert(grid_item, grid_item->GetGlobalId());
      grid_item = (SpSubscribeGrid *) LV->grids_added.GetNextElement();
    }
    while ((grid_item = (SpSubscribeGrid *) LV->grids_reprocess.RemoveFirstElement())){
      update_item = (SpSubscribeGrid *) LV->current_grids.Remove(grid_item->GetGlobalId());
      grid_item->Update(update_item);
      RB_DELETE_SpSubscribeGrid(update_item);
      LV->lookup_grids.Insert(grid_item, grid_item->GetGlobalId());
    }
    while ((grid_item = (SpSubscribeGrid *) LV->grids_added.RemoveFirstElement()))
      LV->current_grids.Remove(grid_item->GetGlobalId());
    while ((grid_item = (SpSubscribeGrid *) LV->current_grids.RemoveFirstElement()))
      RB_DELETE_SpSubscribeGrid(grid_item);
    while ((grid_item = (SpSubscribeGrid *) LV->grids_removed.RemoveFirstElement()))
      LV->delete_grids.Insert(grid_item, grid_item->GetGlobalId());

    LV->event_time.IncrementPriority1(1);
    ScheduleUnSubscribeGridEvents(LV->event_time, global_id, 
                 LV->space->GetId(), look_ahead, LV->delete_grids);
    LV->unsubscribe_time = LV->event_time;
#ifdef PROXY_LOOKAHEAD
    LV->unsubscribe_time += (look_ahead * 2.0);
#else
    LV->unsubscribe_time += look_ahead;
#endif
    LV->unsubscribe_time.SetPriority2(MaxPriority2);
    LV->event_time.DecrementPriority1(1);

    LV->event_time.IncrementPriority1(3);
    ScheduleSubscribeGridEvents(LV->event_time, 
                 node, object_type, local_id, global_id, class_id,
                 LV->space->GetId(), look_ahead, LV->lookup_grids);
    LV->subscribe_time = LV->event_time;
#ifdef PROXY_LOOKAHEAD
    LV->subscribe_time += (look_ahead * 2.0);
#else
    LV->subscribe_time += look_ahead;
#endif
    LV->subscribe_time.SetPriority2(MaxPriority2);
    LV->event_time.DecrementPriority1(3);

#ifdef DDM_DEBUG
    cout << LV->event_time << " P_SpSubscribeSpace simobj " << global_id << " grids lookup ";
    grid_item = (SpSubscribeGrid *) LV->lookup_grids.GetFirstElement();
    while (grid_item){
      cout << grid_item->GetGlobalId() << ", ";
      grid_item = (SpSubscribeGrid *) LV->lookup_grids.GetNextElement();
    }
    cout << " remove ";
    grid_item = (SpSubscribeGrid *) LV->delete_grids.GetFirstElement();
    while (grid_item){
      cout << grid_item->GetGlobalId() << ", ";
      grid_item = (SpSubscribeGrid *) LV->delete_grids.GetNextElement();
    }
    cout << endl;
#endif

    LV->scheduled_events = 0;
    if (LV->delete_grids.GetNumElements()>0){
      WAIT_UNTIL(2, LV->unsubscribe_time);
#ifdef DDM_DEBUG
      cout << CurrentEvent->get_MESSAGE()->time_tag << " P_SpSubscribeSpace simobj "
           << LV->space->GetSimObj()->get_GLOBAL_ID() << " unsubscribe grid ";
      grid_item = (SpSubscribeGrid *) LV->delete_grids.GetFirstElement();
      while (grid_item){
        cout << grid_item->GetGlobalId() << ", ";
        grid_item = (SpSubscribeGrid *) LV->delete_grids.GetNextElement();
      }
      cout << endl;
#endif
      while ((grid_item = (SpSubscribeGrid *) LV->delete_grids.RemoveFirstElement()))
        LV->space->DeleteGrid(grid_item->GetGlobalId());
      LV->scheduled_events = 1;
    }

    if (LV->lookup_grids.GetNumElements()>0){
      WAIT_UNTIL(3, LV->subscribe_time);
#ifdef DDM_DEBUG
      cout << CurrentEvent->get_MESSAGE()->time_tag << " P_SpSubscribeSpace simobj "
           << LV->space->GetSimObj()->get_GLOBAL_ID() << " subscribe add/remove grid ";
      grid_item = (SpSubscribeGrid *) LV->lookup_grids.GetFirstElement();
      while (grid_item){
        cout << grid_item->GetGlobalId() << ", ";
        grid_item = (SpSubscribeGrid *) LV->lookup_grids.GetNextElement();
      }
      cout << endl;
#endif
      while (LV->lookup_grids.RemoveFirstElement());
      LV->scheduled_events = 1;
    }

    delta_t = LV->space->GetUpdateFreqHour() * 3600.0;
    if (delta_t>0.0){
      if (LV->scheduled_events){
        delta_t -= (CurrentEvent->get_MESSAGE()->time_tag - LV->cycle_start_time);
      }
    }

    if (LV->first_cycle)
      LV->first_cycle = !LV->scheduled_events;

    // I can be interrupted (i.e., delta_t does not time out) if any of my
    // other dimensions in the routing space get changed. (i.e, Update gets called).

    LV->space->SetUpdateSemaphore(0);
    WAIT_FOR_LOGICAL(4,
  	   *((LV->space)->GetUpdateSemaphore()),
	   delta_t,
	   LV->success_flag);

  }// loop_forever

END_PROCESS (P_SpSubscribeSpace, process)
