#include "SpDDMUtils.H"
#include "SpBinaryTree.H"
#include "RB_SpBinaryTree.H"


void 
Compare(RB_SpBinaryTree *origItems, RB_SpBinaryTree *newItems,
        int (*getOrigItemId) (void *), int (*getNewItemId) (void *),
        SpBinaryTree *addItems, SpBinaryTree *removeItems, SpBinaryTree *matchItems) {
  SpBinaryTree orig_tree, new_tree;
  void *item;
  int key;

  item = origItems->GetFirstElement();
  while (item){
    key = getOrigItemId(item);
    orig_tree.Insert(item, key);
    item = origItems->GetNextElement();
  }

  item = newItems->GetFirstElement();
  while (item){
    key = getOrigItemId(item);
    new_tree.Insert(item, key);
    item = newItems->GetNextElement();
  }

  Compare(&orig_tree, &new_tree, getOrigItemId, getNewItemId, addItems, removeItems, matchItems);
}

void 
Compare(RB_SpBinaryTree *origItems, RB_SpBinaryTree *newItems,
        int (*getOrigItemId) (void *), int (*getNewItemId) (void *),
        SpBinaryTree *addItems, SpBinaryTree *removeItems) {
  SpBinaryTree matchItems;
  SpBinaryTree orig_tree, new_tree;
  void *item;
  int key;

  item = origItems->GetFirstElement();
  while (item){
    key = getOrigItemId(item);
    orig_tree.Insert(item, key);
    item = origItems->GetNextElement();
  }

  item = newItems->GetFirstElement();
  while (item){
    key = getOrigItemId(item);
    new_tree.Insert(item, key);
    item = newItems->GetNextElement();
  }

  Compare(&orig_tree, &new_tree, getOrigItemId, getNewItemId,
          addItems, removeItems, &matchItems);
}

void 
Compare(SpBinaryTree *origItems, SpBinaryTree *newItems,
        int (*getOrigItemId) (void *), int (*getNewItemId) (void *),
        SpBinaryTree *addItems, SpBinaryTree *removeItems) {
  SpBinaryTree matchItems;

  Compare(origItems, newItems, getOrigItemId, getNewItemId,
          addItems, removeItems, &matchItems);
}

void 
Compare(SpBinaryTree *origItems, SpBinaryTree *newItems,
        int (*getOrigItemId) (void *), int (*getNewItemId) (void *),
        SpBinaryTree *addItems, SpBinaryTree *removeItems, SpBinaryTree *matchItems) {
  void *orig_item, *new_item;
  int orig_id, new_id;

  orig_item = origItems->GetFirstElement();
  new_item  = newItems->GetFirstElement();

  while (orig_item || new_item) {

    if (orig_item)
      orig_id = getOrigItemId(orig_item);
    if (new_item)
      new_id  = getNewItemId(new_item);

    if (!orig_item){
      addItems->Insert(new_item, new_id);
      new_item = newItems->GetNextElement();
      continue;
    }

    if (!new_item){
      removeItems->Insert(orig_item, orig_id);
      orig_item = origItems->GetNextElement();
      continue;
    }

    if (orig_id==new_id){		// stay in the same grid
      matchItems->Insert(orig_item, orig_id);
      orig_item = origItems->GetNextElement();
      new_item  = newItems->GetNextElement();
    }
    else if (orig_id<new_id){		// moved out of previous grid
      removeItems->Insert(orig_item, orig_id);
      orig_item = origItems->GetNextElement();
    }
    else {				// moved into new grid
      addItems->Insert(new_item, new_id);
      new_item = newItems->GetNextElement();
    }
  }// while
}

