// PublishSpace.H header file defines the base class for managing
// DDM spaces used for publishing

#ifndef PUBLISH_SPACE_H
#define PUBLISH_SPACE_H

#include "SpComponent.H"
#include "SpProcSem.H"
#include "SpDDMSpace.H"
#include "SpSimTime.H"

class SpPublishGrid;
class S_SpHLA;
class SpPublishRegion;
class RB_SpBinaryTree;

class SpPublishSpace : public SpDDMSpace {
  friend class S_SpHLA;
  friend class P_SpPublishSpace;
  friend class E_SpAddSubscriberSpace;
  friend class E_SpUnPublishSpace;
  friend class E_SpRemSubscriberSpace;
  friend class E_SpValidateDDMPublisher;

  public:

    SpPublishRegion *CreateRegion(const char *name);
    SpPublishRegion *FindRegion(const char *name);
    int              DeleteRegion(const char *name);
    void             Update(void);

    void Debug(ostream &os);

  protected:

    SpPublishSpace(void);
    SpPublishSpace(S_SpHLA *hla, const SpSpace *space);
    virtual ~SpPublishSpace(void);

    RB_SpBinaryTree *GetGrids(void) { return (Grids); }
    SpPublishGrid   *FindGrid(int global_id);

    void AddGrid(SpPublishGrid *grid);
    int DeleteGrid(int global_id);
    SpPublishGrid *RemoveGrid(int global_id);

    LogicalSemaphore *GetUpdateSemaphore(void)  { return (&UpdateSemaphore); }
    void              SetUpdateSemaphore(int v) { UpdateSemaphore = v; }

    virtual void GetCurrentGrids(const SpSimTime &time, RB_SpBinaryTree &grids) = 0;

    IntegerSemaphore *GetNumPublishReply(void)     { return (&NumPublishReply);   }
    void              ReceivePublishReply(int v)   { NumPublishReply += v;        }

    int  GetNumPublishReplyPending(void) const { return (NumPublishReplyPending); }
    void SetNumPublishReplyPending(int v) { NumPublishReplyPending = v; }

    void Validate(SIMTIME &time);

    RB_SpBinaryTree  *Regions;

  private:
    RB_SpBinaryTree  *Grids;
    LogicalSemaphore UpdateSemaphore;
    IntegerSemaphore NumPublishReply;
    RB_int NumPublishReplyPending;
};

#endif
