// SpScheduleGridEvents.C 

#include "M_SpGrid.H"
#include "M_SpPublishGrid.H"
#include "M_SpUnPublishGrid.H"
#include "M_SpUnSubscribeGrid.H"
#include "M_SpSubscribeGrid.H"

#include "SpScheduleEvent.H"
#include "RB_SpBinaryTree.H"
#include "SpDDMGrid.H"
#include "F_SpProxyItem.H"
#include "SpProxyObjHandle.H"
#include "O_SpGrid.H"

#ifdef TODO
EVENT_TYPE (PublishGrid);
EVENT_TYPE (UnPublishGrid);
EVENT_TYPE (SubscribeGrid);
EVENT_TYPE (UnSubscribeGrid);
#else
#include "E_SpPublishGrid.H"
#include "E_SpUnPublishGrid.H"
#include "E_SpSubscribeGrid.H"
#include "E_SpUnSubscribeGrid.H"
#endif

void SchedulePublishGridEvents(const SpSimTime &time, 
                               int node, int type, int local_id, int global_id, int class_id, 
                               int space_id,
                               double look_ahead,
                               RB_SpBinaryTree &grids){
  char *buffer;
  int buffer_len;
  SpPublishGrid *grid;

  M_SpPublishGrid msg(node, type, local_id, global_id, space_id, class_id);
  msg.SetLookAhead(look_ahead);

  grid = (SpPublishGrid *) grids.GetFirstElement();
  while (grid){
    SpPublishGrid::CreateMsg(grid, buffer, buffer_len);
    ScheduleEvent(time, PublishGrid,
                  grid->GetObjectType(), grid->GetLocalId(), grid->GetNode(), 
                  msg, buffer, buffer_len);
    if (buffer) delete [] buffer;
    grid = (SpPublishGrid *) grids.GetNextElement();
  }
}

void ScheduleUnPublishGridEvents(const SpSimTime &time, 
                               int /* node */, int /* type */, int /* local_id */, int global_id, 
                               int /* space_id */,
                               double /* look_ahead */,
                               RB_SpBinaryTree &grids){
  char *buffer;
  int buffer_len;
  SpPublishGrid *grid;

  M_SpUnPublishGrid msg(global_id);

  grid = (SpPublishGrid *) grids.GetFirstElement();
  while (grid){

    SpDDMGrid::CreateMsgLookupId(grid, buffer, buffer_len);
    msg.SetNumLookupItem(grid->GetNumLookup());
    
    ScheduleEvent(time, UnPublishGrid,
                  grid->GetObjectType(), grid->GetLocalId(), grid->GetNode(), 
                  msg, buffer, buffer_len);

    if (buffer) delete [] buffer;
    grid = (SpPublishGrid *) grids.GetNextElement();
  }
}

void ScheduleSubscribeGridEvents(const SpSimTime &time,
                               int node, int type, int local_id, int global_id, int class_id,
                               int space_id,
                               double look_ahead,
                               RB_SpBinaryTree &grids){
  char *buffer;
  int buffer_len;
  SpSubscribeGrid *grid;

  M_SpSubscribeGrid msg(node, type, local_id, global_id, space_id, class_id);
  msg.SetLookAhead(look_ahead);

  grid = (SpSubscribeGrid *) grids.GetFirstElement();
  while (grid){
    SpSubscribeGrid::CreateMsg(grid, buffer, buffer_len);
    ScheduleEvent(time, SubscribeGrid,
                  grid->GetObjectType(), grid->GetLocalId(), grid->GetNode(), 
                  msg, buffer, buffer_len);
    if (buffer) delete [] buffer;
    grid = (SpSubscribeGrid *) grids.GetNextElement();
  }
}

void ScheduleUnSubscribeGridEvents(const SpSimTime &time,
                                   int global_id, int space_id, double look_ahead,
                                   RB_SpBinaryTree &grids){
  char *buffer;
  int buffer_len;
  SpSubscribeGrid *grid;

  M_SpUnSubscribeGrid msg(global_id, space_id);
  msg.SetLookAhead(look_ahead);

  grid = (SpSubscribeGrid *) grids.GetFirstElement();
  while (grid){

    SpDDMGrid::CreateMsgLookupId(grid, buffer, buffer_len);
    msg.SetNumLookupItem(grid->GetNumLookup());

    ScheduleEvent(time, UnSubscribeGrid,
                  grid->GetObjectType(), grid->GetLocalId(), grid->GetNode(), 
                  msg, buffer, buffer_len);
    if (buffer) delete [] buffer;
    grid = (SpSubscribeGrid *) grids.GetNextElement();
  }
}
