// SpSubscribeSpace.H header file defines the base class for managing
// DDM spaces used for publishing

#ifndef SUBSCRIBE_SPACE_H
#define SUBSCRIBE_SPACE_H

#include "SpComponent.H"
#include "SpProcSem.H"
#include "SpDDMSpace.H"
#include "SpSimTime.H"
#include "EventHandlerMethod.H"

class S_SpHLA;
class SpSubscribeGrid;
class SpSubscribeRegion;
class F_SpProxyItem;
class RB_SpBinaryTree;

class SpSubscribeSpace;
typedef void (*NewObjProxyCallbackFunction)    (SpSubscribeSpace *, F_SpProxyItem *);
typedef void (*RemoveObjProxyCallbackFunction) (SpSubscribeSpace *, int);

class SpSubscribeSpace : public SpDDMSpace {
    friend class H_SpRemoveProxySpace;
    friend class H_SpAddProxySpace;
    friend class P_SpSubscribeSpace;
    friend class S_SpHLA;
  public:

    SpSubscribeRegion *CreateRegion(const char *name);
    SpSubscribeRegion *FindRegion(const char *name);
    int                DeleteRegion(const char *name);
    void               Update(void);

      // return a RB_BinaryTree of SpSubscribeProxy objects
    RB_SpBinaryTree   &GetSubscribeObjectProxies(void) { return (*SubscribeObjectProxies); }
    void RegisterNewObjectProxyCallback(void (*NewObjProxy) (SpSubscribeSpace *, F_SpProxyItem *));
    void RegisterRemoveObjectProxyCallback(void (*RemoveObjProxy) (SpSubscribeSpace *, int global_id));

    void Debug(ostream &);

  protected:

    SpSubscribeSpace(void);
    SpSubscribeSpace(S_SpHLA *hla, const SpSpace *space);
    virtual ~SpSubscribeSpace(void);

    RB_SpBinaryTree *GetGrids(void) { return (Grids); }
    SpSubscribeGrid *FindGrid(int global_id);

    void AddGrid(SpSubscribeGrid *grid);
    int DeleteGrid(int global_id);

    LogicalSemaphore *GetUpdateSemaphore(void)      { return (&UpdateSemaphore); }
    void              SetUpdateSemaphore(int v)     { UpdateSemaphore = v; }

    virtual void GetCurrentGrids(const SpSimTime &time, RB_SpBinaryTree &grids) = 0;

    void Validate(SIMTIME &time);

    RB_SpBinaryTree *Regions;

  private:
    RB_SpBinaryTree                *Grids;
    LogicalSemaphore               UpdateSemaphore;
    RB_SpBinaryTree                *SubscribeObjectProxies;
    NewObjProxyCallbackFunction    NewObjectProxyCallback;
    RemoveObjProxyCallbackFunction RemoveObjectProxyCallback;

    EVENT_HANDLER_METHOD_H (SpSubscribeSpace, AddRemoteProxy);
    EVENT_HANDLER_METHOD_H (SpSubscribeSpace, RemoveRemoteProxy);

    void Add(const SpSimTime &time, int grid_id, F_SpProxyItem *item, int &added, int &notify);
    void Remove(const SpSimTime &time, int grid_id, int global_id, int &removed, int &notify);
};

class SpSubscribeProxy {
    friend class SpSubscribeSpace;
  public:
    SpSubscribeProxy(void);
    SpSubscribeProxy(F_SpProxyItem *proxyItem);
    ~SpSubscribeProxy(void) {}

    int GetGlobalId(void)             const { return (GlobalId);  }
    F_SpProxyItem *GetProxyItem(void) const { return (ProxyItem); }

  private:
    int GlobalId;
    RB_int NumGrid;
    F_SpProxyItem *ProxyItem;

    int GetNumGrid(void)              const { return (NumGrid);   }
    int IncrNumGrid(void)                   { return (++NumGrid); }
    int DecrNumGrid(void)                   { return (--NumGrid); }
};

#endif
