#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpDim.H"
#include "SpDataParser.H"
#include "SpDataParserTypes.H"

SpDim::SpDim(SpDataParser &dataParser)
  : DimName(dataParser.GetReferenceName()),
    Distribute_p(dataParser.GetLogical("Distribute"))
{
}

SpDim::SpDim(const char *name, int distribute)
  :  Distribute_p(distribute)
{
  int len = strlen(name);
  DimName = (len>0) ? new char [len+1] : (char *) 0;
  if (len>0)
    strcpy(DimName, name);
}

void SpDim::get_resolution
  (SpDataParser &dataParser, int &nresolutions, double* &resolution)
{
  nresolutions = dataParser.GetArraySize("Resolution");
  assert(nresolutions > 0);
  resolution = new double[nresolutions];

  for (int IdOfResInDim = 0; IdOfResInDim < nresolutions; IdOfResInDim++)
  {
    resolution[IdOfResInDim] =
      dataParser.GetFloat("Resolution", IdOfResInDim);
    assert(resolution[IdOfResInDim] > 0);
  }
}

char *SpDim::SpHiGrid_DimName(const char *ChildName) const
{
  char *buf = new char[strlen(ChildName) + strlen(DimName) + 4];
  sprintf(buf, "%s - %s", ChildName, DimName);
  return buf;
}

void SpDim::print(ostream &os) const
{
  os << "      Dimension \"" << DimName
     << "\" has " << NIdsInDim << " grids." << endl
     << "         Distribute_p = "
     << (Distribute_p ? "True" : "False") << endl;
  print_work(os);
}
