#ifndef __SpDimLookupItem__
#define __SpDimLookupItem__

//class ostream;
#include "SpIostream.H"
#include "SpStdlib.H"
#include "SpNetTypes.H"


//   Descendants of this class contain lookup data for only one
// dimension in the overall lookup.  The overall lookup is represented
// by class SpSpaceLookupItem.
class SpDimLookupItem
{
  public:
    virtual ~SpDimLookupItem() {};

    enum DimensionTypes {Double, Enum, LatLng, Altitude};
    virtual DimensionTypes get_DimType() const=0;
    const char *get_LookupItemName() const { return LookupItemName; }

    virtual void print_work(ostream &os) const = 0;
    
    int get_DontCare_p() const { return DontCare_p; }

    virtual void CreateMessage(char* &msg, int &len);
    static void CreateObj(const char *msg, int &obj_bytes, SpDimLookupItem* &lookup_item);
    
  protected:
    //   The descendent constructors which take lookup specification
    // info like los or his should pass in 0 for the dontCare_p
    // argument since if the user specifies his or los then he cares
    // about this dimension.
    //   The descendents' set functions should call
    // "set_DontCare_p(0);" since if the user sets them then the user
    // must care about them.  The descendents' accesser functions
    // should call "assert(!get_DontCare_p());" since the data should
    // not be accessed if the dimension is a Don't Care.
    SpDimLookupItem(const char * const lookupItemName, int dontCare_p)
      : LookupItemName(lookupItemName), DontCare_p(dontCare_p) {}

    SpDimLookupItem() : LookupItemName(NULL), DontCare_p(-1234567)
	{abort();}

    void set_DontCare_p(int in) { DontCare_p = in; }
    
  private:
    SpDimLookupItem(SpDimLookupItem&) : LookupItemName(NULL),
      DontCare_p(-1234567) {abort();}
    
    virtual void operator=(SpDimLookupItem&) {abort();}
  
    const char * const LookupItemName;

    // If the following if TRUE, then this SpDimLookupItem
    // exactly fits into any SpDim.
    NET_INT DontCare_p;
};

ostream & operator<<
  (ostream &os, const SpDimLookupItem *dimensionLookupItem);
ostream & operator <<
  (ostream &os, const SpDimLookupItem::DimensionTypes dimType);

typedef SpDimLookupItem DimensionLookupItem;

#endif
