#include "SpEnumLookupItem.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpPubSubItem.H"
#include "SpSpaceLookupItem.H"
#include "SpEnumDim.H"

SpEnumLookupItem &SpEnumLookupItem::is_for_SpaceLookupItem
  (const SpSpaceLookupItem &spaceLookupItem)
{
  assert(Enums == NULL);

  const SpEnumLookupItem * const eli = ((SpEnumLookupItem*) spaceLookupItem
				      [spaceLookupItem.dim_name_and_type_to_id
				      (get_LookupItemName(), get_DimType())]);

  Enums = eli->Enums;
  NEnums = eli->NEnums;

  return *this;
}

#define ASSIGN_ENUMS(FUNCTION_NAME, VALUE)                     \
void SpEnumLookupItem::FUNCTION_NAME(const char *enum_name, ...) \
{                                                              \
  va_list ap;                                                  \
                                                               \
  for(va_start(ap, enum_name);                                 \
      enum_name;                                               \
      enum_name = va_arg(ap, const char *))                    \
    assign_id_in_dim(name_to_id_in_dim(enum_name), VALUE);            \
                                                               \
  va_end(ap);                                                  \
}

ASSIGN_ENUMS(assign_enums, va_arg(ap, int))
ASSIGN_ENUMS(set_enums, 1)
ASSIGN_ENUMS(clear_enums, 0)

SpEnumLookupItem &SpEnumLookupItem::assign_id_in_dim(int id_in_dim, int value)
{
  set_DontCare_p(0);

  const BitVector one = 1;

  BitVector shifted = one << id_in_dim;
  if (value)
    EnumBitVector |= shifted;
  else
    EnumBitVector &= ~shifted;

  return *this;
}

int SpEnumLookupItem::name_to_id_in_dim(const char *enum_name) const
{
  assert(Enums != NULL);

  int id_in_dim;
  for (id_in_dim = 0; id_in_dim < NEnums; id_in_dim++)
  {
    if (0 == strcmp(Enums[id_in_dim], enum_name))
      break;
  }
  if (id_in_dim == NEnums)
  {
    cerr << "Error: attempt to use unknown enum name \"" << enum_name
	 << "\" in\nEnum Dimension \""
	 << get_LookupItemName() << "\"." << endl;
    abort();
  }

  return id_in_dim;
}

void SpEnumLookupItem::print_work(ostream &os) const
{
  assert(Enums != NULL);

  const BitVector one = 1;

  BitVector enumBitVector_copy = EnumBitVector;
  int found_p = 0;
  for (int IdInDimension = 0; IdInDimension < NEnums; IdInDimension++)
  {
    if (enumBitVector_copy & one)
    {
      os << Enums[IdInDimension];
      if (IdInDimension != NEnums - 1)
	os << ", ";
      found_p = 1;
    }
    enumBitVector_copy >>= 1;
  }

  if (!found_p)
    os << "Warning: No enums are set!";
  
  os << endl;
}
