#include "SpString.H"
#include "RB_SpFrameworkFuncs.H"

#include "SpGridLookupOutput.H"

#include "SpRoutingSpaceFreeList.H"
#include "RB_SpDefineClass.H"

RB_DEFINE_CLASS(SpGridLookupOutput);

extern SpRoutingSpaceFreeList *RoutingSpaceFreeList;

#define db_var(var) os << #var << " " << var << endl;

ostream & operator<<(ostream &os, const SpGridLookupOutput &grid)
{
  db_var(grid.get_id_in_speedes());
  os << (*((SpObsObjHandle*) (&grid)));
  db_var(grid.get_gridCovered_p());
  
  return os;
}

void 
SpGridLookupOutput::CreateMessage(char* &msg, int &len){
  char *cp;
  char *main_pub_sub_item_msg, *obj_handle_msg;
  int   main_pub_sub_item_len, obj_handle_msg_len;

  SpObsObjHandle::CreateMessage(obj_handle_msg,       obj_handle_msg_len);
  MainPubSubItem.CreateMessage(main_pub_sub_item_msg, main_pub_sub_item_len);

  len = obj_handle_msg_len + main_pub_sub_item_len;
  msg = (len>0) ? new char [len] : (char *) 0;
  cp  = msg;

  if (obj_handle_msg_len>0){
    memcpy(cp, obj_handle_msg, obj_handle_msg_len);
    cp += obj_handle_msg_len;
    delete [] obj_handle_msg;
  }

  if (main_pub_sub_item_len>0){
    memcpy(cp, main_pub_sub_item_msg, main_pub_sub_item_len);
    delete [] main_pub_sub_item_msg;
  }
}

void 
SpGridLookupOutput::CreateObject(const char *msg, int &obj_bytes, SpGridLookupOutput* &item){
  int nbytes;
  const char *p = msg;

  obj_bytes = 0;
  item      = Create();
  SpObsObjHandle::CreateObject(p, nbytes, *item); 
  p += nbytes; obj_bytes += nbytes;
  MAIN_PUBSUB_ITEM::CreateObject(p, nbytes, item->MainPubSubItem);
               obj_bytes += nbytes;
}

SpGridLookupOutput *
SpGridLookupOutput::Create(int local_id, int node, int object_type, int global_id){
  SpGridLookupOutput *item = Create();
  item->init(local_id, node, object_type, global_id);
  return (item);
}

SpGridLookupOutput *
SpGridLookupOutput::Create(void){
#define NOT_USED_RB_FREE_NEW
#ifdef  NOT_USED_RB_FREE_NEW
  SpGridLookupOutput *item = (SpGridLookupOutput *)
                             RB_NEW_SpGridLookupOutput(new SpGridLookupOutput);
#else
  SpGridLookupOutput *item = (SpGridLookupOutput *) 
                             RB_FREE_NEW(RoutingSpaceFreeList, FREE_SpGridLookupOutput);
#endif
  item->get_MainPubSubItem()->SetUniqueKey();
  return (item);
}

void
SpGridLookupOutput::Destroy(SpGridLookupOutput *grid){
#ifdef NOT_USED_RB_FREE_NEW
  RB_DELETE_SpGridLookupOutput(grid);
#else
  RB_FREE_DELETE(RoutingSpaceFreeList, grid);
#endif
}

int 
SpGridLookupOutput::operator==(const SpGridLookupOutput &l) const{
  return (MainPubSubItem==l.MainPubSubItem);
}


void RB_Register_SpGridLookupOutput()
{
  RB_REGISTER_CLASS_SpGridLookupOutput();
}


