#ifndef __SpGridLookupOutput__
#define __SpGridLookupOutput__

#include "SpObsObjHandle.H"
#include "SpStdlib.H"
#include "SpPubSubItem.H"

class SpGridLookupOutput : public SpObsObjHandle
{
  private:

    SpGridLookupOutput(SpGridLookupOutput&) {abort();}

    int GridCovered_p;

    // You will probably want to pass this object to
    // SpHiGrid::AddPublisher or SpHiGrid::AddSubscriber.  However,
    // before you pass it, you will probably want to initialize some
    // fields in it.  For example, if a detectable simulation entity
    // is publishing itself in a sensor space, then it will probably
    // want to set the fields node, type, and local_id, so that when a
    // sensor later calls O_SpGrid::HiGrid_Lookup and then calls
    // SpHiGrid::GetPublishers, it will know where to find the returned
    // publishers so that it can ask them for their position, speed,
    // etc., so that it can know if it can detect those publishers.
    // The main point of this is that the sensor needs to know the
    // node, local_id, etc. of the publisher in order to find the
    // publisher, but this data is not set in this object, so you must
    // set it yourself.  Note that I am talking about setting the
    // fields node, type, and local_id in this MAIN_PUBSUB_ITEM, NOT
    // the similar fields in the SpObsObjHandle which this
    // SpGridLookupOutput inherits.  Those fields in the inherited
    // SpObsObjHandle describe where this grid is.  But the
    // MAIN_PUBSUB_ITEM's fields describe where the publisher is.  Of
    // course this publisher is in the simulated physical space that
    // the grid manages, but the publisher might be MANAGED by a
    // totally different node than the node that manages the grid!
    MAIN_PUBSUB_ITEM MainPubSubItem;
    
  public:

    SpGridLookupOutput(void)
      : GridCovered_p(0) {}

    void set_id_in_speedes(int ui) {id = ui;}
    void set_gridCovered_p(int gridCovered_p) {GridCovered_p = gridCovered_p;}

    int get_id_in_speedes() const {return id;}
    int get_gridCovered_p() const {return GridCovered_p;}

    MAIN_PUBSUB_ITEM *get_MainPubSubItem() { return &MainPubSubItem; }

    void CreateMessage(char* &msg, int &len);
    static void CreateObject(const char *msg, int &obj_bytes, SpGridLookupOutput* &item);

    static SpGridLookupOutput *Create(int local_id, int node, int object_type, int global_id);
    static SpGridLookupOutput *Create(void);
    static void                Destroy(SpGridLookupOutput *grid);

    int operator==(const SpGridLookupOutput &l) const;
};

//RB_DEFINE_CLASS(SpGridLookupOutput);

ostream & operator<<(ostream &os, const SpGridLookupOutput &grid);

typedef SpGridLookupOutput O_GridLookupOutput;

#endif
