#ifndef __SpLatLngLookupItem__
#define __SpLatLngLookupItem__

#include "SpAssert.H"

#include "SpDimLookupItem.H"

// For looking up spheres.  If you want to use an nss cylindrical
// sensor, do not use the formula Jacob made for testing the coverage
// code in nss.  That formula cannot be used here.  Instead we will
// have to use the triangle shown in function
// get_delta_rad_lat_of_sphere but we know A and solve for
// sphere_radius.
class SpLatLngLookupItem : public SpDimLookupItem
{
  public:
    // All points within kmSizeRadius kilometers of the point (radLat,
    // radLng, kmAlt) will be covered.  This distance of kmSizeRadius
    // kilometers is a straight line distance.  It is NOT measured
    // over the curve of the surface of the earth or over any other
    // curve.  kmResolutionRadius is also a straight line distance.
    SpLatLngLookupItem(const double radLat,
		     const double radLng,
		     const double kmAlt,
		     const double kmSizeRadius,
		     const double kmResolutionRadius);
    
    SpLatLngLookupItem() : SpDimLookupItem("LatLng", 1) {}

    SpLatLngLookupItem &set_RadLat(const double radLat);

    // radLng may be between -pi and 2*pi.  If it is more than pi,
    // then radLng - 2*pi will be stored:
    SpLatLngLookupItem &set_RadLng(double radLng);

    SpLatLngLookupItem &set_KmAlt(double kmAlt);
    SpLatLngLookupItem &set_KmSizeRadius(double in);
    SpLatLngLookupItem &set_KmResolutionRadius(const double in);
    
    double get_RadLat() const { assert(!get_DontCare_p()); return RadLat; }
    double get_RadLng() const { assert(!get_DontCare_p()); return RadLng; }
    
    double get_KmAlt() const { assert(!get_DontCare_p()); return KmAlt; }
    double get_KmSizeRadius() const
	{ assert(!get_DontCare_p()); return KmSizeRadius; }
    double get_KmResolutionRadius() const
	{ assert(!get_DontCare_p()); return KmResolutionRadius; }

  private:
    SpLatLngLookupItem(SpLatLngLookupItem&) {abort();}

    // Km means kilometer:
    double RadLat, RadLng, KmAlt, KmSizeRadius, KmResolutionRadius;

    DimensionTypes get_DimType() const { return LatLng; }

    void print_work(ostream &os) const;
};

typedef SpLatLngLookupItem LatLngLookupItem;

#endif
