#include "SpString.H"
#include "SpObsObjHandle.H"
#include "RB_SpDefineClass.H"
RB_DEFINE_CLASS(SpObsObjHandle);

#define db_var(var) os << #var << " " << var << endl;

ostream & operator<<(ostream &os, const SpObsObjHandle &oh)
{
  db_var(oh.get_object_node());
  db_var(oh.get_id_in_node_managed_grids());
  
  return os;
}

void 
SpObsObjHandle::CreateMessageGlobalId(char* &msg, int &len) const{
  len = sizeof(NET_INT);
  msg = new char [len];

  memcpy(msg, &id, sizeof(NET_INT));
}

void 
SpObsObjHandle::CreateMessage(char* &msg, int &len) const{
  len = sizeof(NET_INT) * 4;
  msg = new char [len];
  char *p = msg;

  memcpy(p, &object_node, sizeof(NET_INT)); p += sizeof(NET_INT);
  memcpy(p, &object_type, sizeof(NET_INT)); p += sizeof(NET_INT);
  memcpy(p, &local_id,    sizeof(NET_INT)); p += sizeof(NET_INT);
  memcpy(p, &id,          sizeof(NET_INT));
}

void 
SpObsObjHandle::CreateMessage(SpBinaryTree &objs, char* &msg, int &len){
  char *cp;
  int i, total_len, num_objs;
  char **obj_msg;
  int   *obj_msg_len;
  SpObsObjHandle *obj;

  msg = (char *) 0;
  len = 0;

  num_objs = objs.GetNumElements();

  if (num_objs<=0) return;

  obj_msg     = new char * [num_objs];
  obj_msg_len = new int    [num_objs];
  total_len   = 0;

  i = 0;
  obj = (SpObsObjHandle *) objs.GetFirstElement();
  while (obj){
    obj->CreateMessage(obj_msg[i], obj_msg_len[i]);
    total_len += obj_msg_len[i];
    i++;
    obj = (SpObsObjHandle *) objs.GetNextElement();
  }

  msg = (total_len>0) ? new char [total_len] : (char *) 0;
  len = total_len;
  cp  = msg;

  for (i=0; i<num_objs; i++){
    memcpy(cp, obj_msg[i], obj_msg_len[i]); cp += obj_msg_len[i];
    delete [] obj_msg[i];
  }

  delete [] obj_msg;
  delete [] obj_msg_len;
}

void 
SpObsObjHandle::CreateObject(const char *msg, int &obj_bytes, SpObsObjHandle &item){
  const char *p = msg;
  NET_INT node, type, local_id, global_id;

  obj_bytes = 0;

  memcpy(&node,      p, sizeof(NET_INT)); p += sizeof(NET_INT); obj_bytes += sizeof(NET_INT);
  memcpy(&type,      p, sizeof(NET_INT)); p += sizeof(NET_INT); obj_bytes += sizeof(NET_INT);
  memcpy(&local_id,  p, sizeof(NET_INT)); p += sizeof(NET_INT); obj_bytes += sizeof(NET_INT);
  memcpy(&global_id, p, sizeof(NET_INT));                       obj_bytes += sizeof(NET_INT);

  item.set_object_node(node);
  item.set_object_type(type);
  item.set_local_id(local_id);
  item.set_id(global_id);
}

void 
SpObsObjHandle::CreateObj(const char* &buffer, int num_items, int &nbytes, SpBinaryTree &items){
  int i, obj_bytes;
  const char *cp;
  SpObsObjHandle *object_handle;

  nbytes = 0;
  for (obj_bytes=0, cp=buffer, i=0; i<num_items; i++, cp+=obj_bytes){
    object_handle = new SpObsObjHandle;
    SpObsObjHandle::CreateObject(cp, obj_bytes, *object_handle); nbytes += obj_bytes;
    items.Insert(object_handle, object_handle->get_id());
  }
}


void RB_Register_SpObsObjHandle()
{
  RB_REGISTER_CLASS_SpObsObjHandle();
}


