// SpObjHandle.H header file

#ifndef ObjectHandle_object
#define ObjectHandle_object

#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpStdlib.H"

#include "SpBinaryTree.H"

#include "qitem.H"

/************************************************************************
*	SpObsObjHandle object 	       					*
************************************************************************/
class SpObsObjHandle : public C_ITEM {

  private:
    SpObsObjHandle (SpObsObjHandle&) {abort();}
    virtual void operator=(SpObsObjHandle&) {abort();}
    
    NET_INT object_node;
    NET_INT local_id;
    NET_INT object_type;

  public:
    SpObsObjHandle (void)
      : object_node(-1), local_id(-1), object_type(-1) {
      set_id(-1);
    }
    SpObsObjHandle (int _local_id, int _node, int _object_type, int _global_id)
      : object_node(_node), local_id(_local_id), object_type(_object_type) {
      set_id(_global_id);
    }

    void init(int _local_id, int _node, int _object_type, int _global_id){
      object_node = _node;
      local_id    = _local_id;
      object_type = _object_type;
      set_id(_global_id);
    }
 
    void set_object_node(int on)              { object_node = on; }
    void set_local_id(int v)                  { local_id = v;     }
    void set_id_in_node_managed_grids(int oi) { set_local_id(oi); }
    void set_object_type(int ot)              { object_type = ot; }

    int get_object_node(void)              const { return object_node;    }
    int get_local_id(void)                 const { return (local_id);     }
    int get_id_in_node_managed_grids(void) const { return get_local_id(); }
    int get_object_type(void)              const { return object_type;    }

    int getNode(void)      const { return (object_node); }
    int getObjType(void)   const { return (object_type); }
    int getLocalId(void)   const { return (local_id);    }
    int getGlobalId(void)  const { return (get_id()); }

    void setGlobalId(int i) { set_id(i); }


    void CreateMessage(char* &msg, int &len) const;
    void CreateMessageGlobalId(char* &msg, int &len) const;
    static void CreateObject(const char *msg, int &obj_bytes, SpObsObjHandle &item);
    static void CreateObj(const char* &buffer, int num_items, int &nbytes, SpBinaryTree &items);
    static void CreateMessage(SpBinaryTree &objs, char* &msg, int &len);
};

//RB_DEFINE_CLASS(SpObsObjHandle);

ostream & operator<<(ostream &os, const SpObsObjHandle &oh);


typedef SpObsObjHandle OBJECT_HANDLE;

#endif
