#ifndef __SpSpace__
#define __SpSpace__

#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpStdlib.H"

class SpHiGrid;
class C_QUEUE;
class SpDataParser;
class SpSpaceLookupItem;

#include "SpSpaceSubSet.H"
#include "SpDimLookupItem.H"

class SpSpace
{
  public:
    SpSpace() {}
    void InitSpace(SpDataParser &DataParser, const int idOfSpaceInMan);

    // Initializes HiGrid with appropriate lo's, hi's, resolutions,
    // names, etc.  It uses IdOfGridInSpace to calculate the los, and
    // his.
    void InitSpHiGrid(const int IdOfGridInSpace, SpHiGrid * const HiGrid) const;

    //   This function will calculate all of the ids in this SpSpace
    // which correspond to grids whose area overlaps the area of
    // spaceLookupItem.
    //   All SpaceSubSets are checked for matches.
    //   This function returns TRUE iff spaceLookupItem fits entirely
    // within the boundaries of at least one of the SpaceSubSets
    // contained in this SpSpace.  This means that if there are two
    // SpaceSubSets in this SpSpace, and one of them contains half of
    // spaceLookupItem, and the other contains exactly the opposite
    // (maybe complementary is a better word) half, then this function
    // returns FALSE since no *one* SpSpaceSubSet by *itself* contains
    // all of spaceLookupItem.
    //   This function will dynamically (i.e. using new) allocate a
    // SpHiGridRegionVec and put the new ids in it.  The address of
    // this SpHiGridRegionVec is returned by reference in argument
    // output_ids_of_grids_in_SpSpace.  The caller is responsible
    // for deleting this SpHiGridRegionVec.
    int HiGrid_Lookup
      (const SpSpaceLookupItem &spaceLookupItem,
       SpHiGridRegionVec * &output_ids_of_grids_in_SpSpace) const;

    void print(ostream &os) const;
    
    void GetDontCareLookupItems
      (int &NDimLookupItems, SpDimLookupItem ** &DimLookupItems) const;
    
    const char *get_SpaceName() const { return SpaceName; }
    int get_SpaceId(void) const { return (IdOfSpaceInMan); }
    int get_NIdsInSpace() const { return NIdsInSpace; }
    int get_CumNIdsInSpace() const { return CumNIdsInSpace; }
    void set_CumNIdsInSpace(int in) { CumNIdsInSpace = in; }
    int HasTheaterDimension(void) const;
    void FindDimensionType(const char *name, 
                          int &found, SpDimLookupItem::DimensionTypes &type) const;
  private:
    SpSpace         (SpSpace&) {abort();}
    void operator=(SpSpace&) {abort();}

    char *SpaceName;
    int IdOfSpaceInMan;

    int NIdsInSpace;     // Number of grids this space wants distributed
    int CumNIdsInSpace;  // Number of grids all previous spaces want
                         //   distributed.  Only O_SpGrid should use this.

    int NSpaceSubSets;
    SpSpaceSubSet *SpaceSubSets;

    
    // Holds dimensions listed in the space section, not the sections
    // referenced by the string "THEATER" in the .par file.  Each
    // SpSpaceSubSet in the array SpaceSubSets is considered to contain
    // these space level dimensions, but I did not want to duplicate
    // these common space level dimensions in every element of that
    // array because that duplication would duplicate not only space,
    // but also execution time (that time being the time to process
    // these dimensions in various ways).  So the common dimensions
    // are stored in only one place, namely the following variable:
    SpSpaceSubSet SpaceLevelDimensions;

    // The number of SpHiGrid (as opposed to class SpDim)
    // dimensions in this space.
    int NumDimRegions;
};

inline ostream & operator<<(ostream &os, const SpSpace &space)
{
  space.print(os);
  return os;
}

typedef SpSpace Space;

#endif
