#include "SpSpaceLookupItem.H"
#include "SpDimLookupItem.H"
#include "SpAssert.H"
#include "SpHiGrid.H"
#include "SpStdarg.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"

SpSpaceLookupItem::SpSpaceLookupItem
  (const char * const spaceName, SpDimLookupItem *dimLookupItem, ...)
    : SpaceName(spaceName)
{
  Initialize();
  va_list ap;

  for(va_start(ap, dimLookupItem);
      dimLookupItem;
      dimLookupItem = va_arg(ap, SpDimLookupItem *))
  {
    InsertDimLookupItem(dimLookupItem);
  }
  
  va_end(ap);
}

SpSpaceLookupItem::~SpSpaceLookupItem()
{
  for (int i = 0; i < get_NDimLookupItems(); i++)
    delete DimLookupItems[i];

  delete [] DimLookupItems;
}

int SpSpaceLookupItem::dim_name_and_type_to_id
  (const char * const dim_name,
   SpDimLookupItem::DimensionTypes dim_type) const
{
  int i;
  for (i = 0; i < get_NDimLookupItems(); i++)
  {
    if (0 == strcmp(DimLookupItems[i]->get_LookupItemName(), dim_name)
	&& DimLookupItems[i]->get_DimType() == dim_type)
      return i;
  }

  fflush(stdout);
  fflush(stderr);
  
  const int WName = 45, WType = 32;
  
  cerr.setf(ios::left);
  
  cerr << "\nError: SpSpaceLookupItem: Attempt to find a SpDimLookupItem\n"
    "named \"" << dim_name << "\" of type \"" << dim_type << "\" in space \""
       << SpaceName << "\",\nbut that space has only the following ";

  if (get_NDimLookupItems() > 1)
    cerr << get_NDimLookupItems() << ' ';

  cerr << "SpDimLookupItem"
       << (get_NDimLookupItems() > 1 ? "s" : "") << ":\n\n"

       << setw(WName) << "Name"
       << setw(WType) << "Type" << "\n"
       << setw(WName) << "============================="
       << setw(WType) << "=============================" << '\n';

  for (i = 0; i < get_NDimLookupItems(); i++)
  {
    cerr << setw(WName) << DimLookupItems[i]->get_LookupItemName()
	 << setw(WType) << DimLookupItems[i]->get_DimType() << '\n';
  }

  cerr.flush();
  
  abort();

  return -1;
}

void SpSpaceLookupItem::InsertDimLookupItem
  (SpDimLookupItem * const dimLookupItem)
{
  int id = dim_name_and_type_to_id(dimLookupItem->get_LookupItemName(),
				   dimLookupItem->get_DimType());
  delete DimLookupItems[id];
  DimLookupItems[id] = dimLookupItem;
  return;
}

const SpDimLookupItem *SpSpaceLookupItem::operator[](const int index) const
{
  assert(0 <= index && index < get_NDimLookupItems());
  return DimLookupItems[index];
}

ostream & operator<<(ostream &os, const SpSpaceLookupItem &spaceLookupItem)
{
  os << "This SpaceLookupItem is for looking up regions in Space \""
     << spaceLookupItem.get_SpaceName()
     << "\".\nIt has these DimensionLookupItems:\n";
  for (int i = 0; i < spaceLookupItem.get_NDimLookupItems(); i++)
    os << spaceLookupItem[i];
  return  os;
}

const O_SpGrid *SpSpaceLookupItem::SpSpace_Man = NULL;

void SpSpaceLookupItem::set_SpSpace_Man(const O_SpGrid * const space_Man)
{
  // The assert assures that this function is only called once.
  assert(SpSpace_Man == NULL);
  SpSpace_Man = space_Man;
}
