// O_SpGrid.C method file

#include "O_SpGrid.H"
#include "SpStopWatch.H"
#include "S_SpGrid.H"
#include "SpGridLookupOutput.H"
#include "SpSpace.H"
#include "SpSpaceLookupItem.H"
#include "SpHiGridRegionVec.H"
#include "SpAssert.H"


//#define DEBUG_SpaceMan
//#define TEST_O_SpGrid

/*
** Define any global free lists here
*/

#include "SpRoutingSpaceFreeList.H"
SpRoutingSpaceFreeList *RoutingSpaceFreeList = NULL;

O_SpGrid *Grid_OBJECT_MANAGER = NULL;

/************************************************************************
* O_SpGrid : construct a O_SpGrid object					*
************************************************************************/
O_SpGrid::O_SpGrid(const char * const SpDataParser_file_name)
  : DataParser(SpDataParser_file_name)
{
  if (RoutingSpaceFreeList == NULL)
    RoutingSpaceFreeList = new SpRoutingSpaceFreeList();
  
  int i;
  SpStopWatch StopWatch("O_SpGrid");

  StopWatch.StartTimer();

  Grid_OBJECT_MANAGER = this;
  
  void *InterestSpacesMark = DataParser.GoTo("InterestSpaces", NULL);
  if (InterestSpacesMark == NULL)
  {
    cerr << "Could not find InterestSpaces section in file InterestSpace.par"
	 << endl;
    abort();
  }

  Nspaces = DataParser.GetNtypes("SPACE");
  assert(Nspaces >= 0);
  if (Nspaces == 0)
  {
    cerr << "There are no SPACEs in the InterestSpaces section." << endl;
    abort();
  }
  Spaces = new SpSpace[Nspaces];
  
  int IdOfSpaceInMan = 0;
  for (void *not_done_p = DataParser.GoToFirstElement(InterestSpacesMark);
       not_done_p;
       not_done_p = DataParser.GoToNextElement(InterestSpacesMark))
  {
    Spaces[IdOfSpaceInMan].InitSpace(DataParser, IdOfSpaceInMan);
    for (i = 0; i < IdOfSpaceInMan; i++)
    {
      if (0 == strcmp(Spaces[IdOfSpaceInMan].get_SpaceName(),
		      Spaces[i].get_SpaceName()))
      {
	cerr << "Error: There are two spaces named \""
	     << Spaces[i].get_SpaceName() << "\"." << endl;
	abort();
      }
    }
    ++IdOfSpaceInMan;
  }
  assert(Nspaces == IdOfSpaceInMan);

/*
** Now that the parser information has been completely read in and stored,
** deal out the HiGrid objects for each space.
*/

  N_TOT = 0;
  for (i=0; i<Nspaces; i++) {
    Spaces[i].set_CumNIdsInSpace(N_TOT);
    N_TOT += Spaces[i].get_NIdsInSpace();
  }

  int offset;
  N_LOC = deal_me(N_TOT, offset);
  const int n_loc = N_LOC;
  SpaceObj = new S_SpGrid[n_loc];

/*
** Initialize the objects
*/

  IdOfSpaceInMan = 0;
  
  for (int SpaceObjIndex = 0; SpaceObjIndex < N_LOC; SpaceObjIndex++) {
    const int IdOfGridInManager = offset + SpaceObjIndex * N_NODES;
    
    // This loop could be replaced with a binary search, but there might
    // never be enough spaces to make much of a difference anyway.
    while ((Spaces[IdOfSpaceInMan].get_CumNIdsInSpace()
	    + Spaces[IdOfSpaceInMan].get_NIdsInSpace())
	   <= IdOfGridInManager)
    {
      IdOfSpaceInMan++;
      assert(IdOfSpaceInMan < Nspaces);
    }

    // IdOfSpaceInMan is now the index of the SpSpace which
    // holds grid IdOfGridInManager
    const int IdOfGridInSpace
      = IdOfGridInManager - Spaces[IdOfSpaceInMan].get_CumNIdsInSpace();

#ifdef DEBUG_SpaceMan
    cout << "Node " << NODE
	 << " IdOfSpaceInMan = " << IdOfSpaceInMan
	 << " SpaceObjIndex = " << SpaceObjIndex
	 << " IdOfGridInManager = " << IdOfGridInManager
	 << " IdOfGridInSpace = " << IdOfGridInSpace << endl;
#endif

    Spaces[IdOfSpaceInMan]
      .InitSpHiGrid(IdOfGridInSpace, SpaceObj[SpaceObjIndex].GetHiGrid());
  }

/*
** Define these objects for SPEEDES
*/

  StartingUniqueId = TOTAL_OBJECTS;
  TOTAL_OBJECTS += N_TOT;
  reserve_n_objects(N_LOC);
  for (i=0; i<N_LOC; i++){
    define_object(&SpaceObj[i]);
  }

/*
** Test that the object manager is working properly
*/

#ifdef TEST_O_SpGrid
  TestO_Grid();
#endif 

  SpSpaceLookupItem::set_SpSpace_Man(this);

  //if (get_node() == 0) SpComm
  if (SpComm_GetNodeId() == 0)
    cout << "The Manager named O_SpGrid has " << N_TOT << " grid objects." << endl;
  
  StopWatch.StopTimer();
  //  StopWatch.print();

}

const SpSpace *
O_SpGrid::FindSpace(const char *_name) const{
  for (int i=0; i<Nspaces; i++){
    if (strcmp(_name, Spaces[i].get_SpaceName())==0)
      return (Spaces+i);
  }
  return ((const SpSpace *) 0);
}

const SpSpace *
O_SpGrid::GetSpace(int _id) const{
  for (int i=0; i<Nspaces; i++){
    if (_id==Spaces[i].get_SpaceId())
      return (Spaces+i);
  }
  abort();
  return NULL;
}

S_SpGrid *
O_SpGrid::GetGrid(int _global_id){
  for (int i=0; i<N_LOC; i++)
    if (SpaceObj[i].get_GLOBAL_ID()==_global_id) return (SpaceObj+i);

  cout << "ERROR: Unable to find grid : " << _global_id 
       << " on node " << get_NODE() << endl;
  for (int x=0; x<N_LOC; x++)
    cout << x << ") " << SpaceObj[x].get_GLOBAL_ID() << ", " << SpaceObj[x].get_LOCAL_ID() << endl;
  abort();
  return NULL;
}

/************************************************************************
* init_events : initialize events 				        *
************************************************************************/
void O_SpGrid::init_events() {

}

/************************************************************************
* terminate : termination method 				        *
************************************************************************/
void O_SpGrid::terminate(double
#ifdef DEBUG_SpaceMan
			 tend
#endif
			 ) {

#ifdef DEBUG_SpaceMan
  cout << "Terminating O_SpGrid at time " << tend
       << " on node " << NODE << endl;
  print();
#endif

}

void O_SpGrid::Print(ostream &os) const {
  int i;

  for (i=0; i<Nspaces; i++) {
    os << Spaces[i];
  }
}

void O_SpGrid::InitSpaceLookupItem
  (const char * const SpaceName, int &IdOfSpaceInManager,
   int &NDimLookupItems, SpDimLookupItem ** &DimLookupItems) const
{
  for (IdOfSpaceInManager = 0;
       IdOfSpaceInManager < Nspaces;
       IdOfSpaceInManager++)
  {
    if (0 == strcmp(Spaces[IdOfSpaceInManager].get_SpaceName(), SpaceName))
    {
      Spaces[IdOfSpaceInManager].GetDontCareLookupItems
	(NDimLookupItems, DimLookupItems);
      return;
    }
  }
  abort();
}

/************************************************************************
* HiGrid_Lookup : get a list of the HiGrids that were distributed	*
* based	on the interest region passed in through the SpSpaceLookupItem	*
* object.								*
************************************************************************/
int O_SpGrid::HiGrid_Lookup(const SpSpaceLookupItem &spaceLookupItem,
			    C_QUEUE * const ids_of_grids_in_speedes) const
{
  assert(ids_of_grids_in_speedes->length() == 0);
  SpHiGridRegionVec *ids_of_grids_in_SpSpace;
  const int lookup_fits_in_some_SpSpaceSubSet_p
    = Spaces[spaceLookupItem.get_IdOfSpaceInManager()]
    .HiGrid_Lookup(spaceLookupItem, ids_of_grids_in_SpSpace);
  
  // For each grid, set more info, mostly used for finding the grid no
  // matter which node holds it.
  SpHiGridRegionVecReader iterator(ids_of_grids_in_SpSpace);
  while (++iterator)
  {
    const int
      IdOfGridInSpace = iterator.id(),

      IdOfGridInManager
      = IdOfGridInSpace + Spaces[spaceLookupItem.get_IdOfSpaceInManager()]
      .get_CumNIdsInSpace(),

      IdOfGridInSpeedes = IdOfGridInManager + StartingUniqueId;

    SpGridLookupOutput * const grid = SpGridLookupOutput::Create();

    grid->set_id_in_speedes(IdOfGridInSpeedes);
    grid->set_object_node(IdOfGridInSpeedes % N_NODES);
    grid->set_id_in_node_managed_grids(IdOfGridInManager / N_NODES);
    grid->set_object_type(OBJECT_TYPE);
    grid->set_gridCovered_p(iterator.gridCovered_p());
    iterator.Init_PubSub_Item(grid->get_MainPubSubItem());
								   
    ids_of_grids_in_speedes->push_bot(grid);
  }

  delete ids_of_grids_in_SpSpace;
  
  return lookup_fits_in_some_SpSpaceSubSet_p;
}

/************************************************************************
* TestO_Grid : test that the O_SpGrid object manager is working	*
* properly			 				        *
************************************************************************/
/*
void O_SpGrid::TestO_Grid() {
  int i,j;
  MAIN_PUBSUB_ITEM *MainPubSubItem;
  C_QUEUE Q;
  double L;
  double H;
  int LoBin;
  int HiBin;
  double HmL;
  double Ldiff;
  double Hdiff;
  SpObsObjHandle *ObjectHandle;
  SpStopWatch StopWatch("TestO_Grid");

#ifdef DEBUG_SpaceMan
  cout << "Testing O_SpGrid on node " << NODE << endl;
#endif

  for (i=0; i<Nspaces; i++) {

#ifdef DEBUG_SpaceMan
    cout << "Constructing MainPubSubItem" << endl;
#endif

    MainPubSubItem = (MAIN_PUBSUB_ITEM *)
      RoutingSpaceFreeList->new_object(FREE_MAIN_PUBSUB_ITEM);

    MainPubSubItem->SetNdimensions(Spaces[i].Ndims);
    for (j=0; j<Spaces[i].Ndims; j++) {
      L = Spaces[i].DimInfo[j].Lo;
      H = Spaces[i].DimInfo[j].Hi;
      MainPubSubItem->SetLHR(j, L+(H-L)/3, H-(H-L)/3, 0.0);

      HmL = Spaces[i].DimInfo[j].Hi - L;
      Ldiff = MainPubSubItem->GetLo(j) - L;
      Hdiff = MainPubSubItem->GetHi(j) - L;

      LoBin = int(Spaces[i].DimInfo[j].Ndist * (Ldiff/HmL));
      HiBin = int(Spaces[i].DimInfo[j].Ndist * (Hdiff/HmL));

#ifdef DEBUG_SpaceMan
      cout << "Dimension " << j
	   << " True Lo/Hi (" << Spaces[i].DimInfo[j].Lo
	   << ", " << Spaces[i].DimInfo[j].Hi << ")"
	   << " Region Lo/Hi (" << MainPubSubItem->GetLo(j)
	   << ", " << MainPubSubItem->GetHi(j) << ") " 
	   << " Bin(0," << Spaces[i].DimInfo[j].Ndist << ")=("
	   << LoBin << ", " << HiBin << ")" << endl;
#endif

    }

#ifdef DEBUG_SpaceMan
    cout << "Testing HiGrid_Lookup" << endl;
#endif

    StopWatch.Reset();
    StopWatch.StartTimer();
    HiGrid_Lookup(i, MainPubSubItem, &Q);
    StopWatch.StopTimer();
    StopWatch.print();

    RoutingSpaceFreeList->delete_object(MainPubSubItem);

#ifdef DEBUG_SpaceMan
    Q.print();
#endif

#ifdef DEBUG_SpaceMan
    cout << "May have some Verify Region errors" << endl;
    for (i=0; i<N_LOC; i++) {
      SpaceObj[i].VerifyRegion(MainPubSubItem);
    }
#endif

#ifdef DEBUG_SpaceMan
    cout << "Verifying Good Regions" << endl;
#endif

    while (ObjectHandle = (SpObsObjHandle *)Q.pop_top()) {
      if (ObjectHandle->get_object_node() == NODE) {
        SpaceObj[ObjectHandle->get_object_id()].VerifyRegion(MainPubSubItem);
      }
      RoutingSpaceFreeList->delete_object(ObjectHandle);
    }

#ifdef DEBUG_SpaceMan
    cout << endl;
#endif

  }

#ifdef DEBUG_SpaceMan
  cout << "Done testing O_SpGrid on node " << NODE << endl;
#endif

}
*/
