// O_SpGrid.H header file

#ifndef O_SpGrid_object
#define O_SpGrid_object

#include "SpStdlib.H"
#include "SpSimObjMgr.H"
#include "SpDataParser.H"

class SpSpaceLookupItem;
#include "SpSpace.H"
class S_SpGrid;
class SpDataParser;

/************************************************************************
*	O_SpGrid object - manager of interest spaces			*
************************************************************************/
class O_SpGrid : public SpSimObjMgr
{

  private:
    O_SpGrid      (O_SpGrid&) : DataParser(NULL) {abort();}
    void operator=(O_SpGrid&) {abort();}
   // O_SpGrid() : DataParser(NULL) { abort(); }

    SpDataParser DataParser;		// Parser input file

    int Nspaces;			// Number of interest spaces
    SpSpace *Spaces;	                // Information for each interest space

    S_SpGrid *SpaceObj;			// Locally created space objects
    int StartingUniqueId;		// Starting unique id for HiGrids

  public:

    const SpSpace *FindSpace(const char *_name) const;
    const SpSpace *GetSpace(int _id) const;

    S_SpGrid *GetGrid(int _global_id);

    void InitSpaceLookupItem
      (const char * const SpaceName, int &IdOfSpaceInManager,
       int &NDimLookupItems, SpDimLookupItem ** &DimLookupItems) const;

    int get_StartingUniqueId() const { return StartingUniqueId; }

/*
** Methods that SPEEDES calls
*/
    O_SpGrid(const char * const SpDataParser_file_name = "InterestSpaces.par");
    
    const S_SpGrid *spaceObj() const { return SpaceObj; }

    virtual void init_events();
    virtual void terminate(double tend);

    void print() { Print(cout); }
    void Print(ostream &os) const;

/*
** Methods for getting HiGrid lookup information given a region in the
** interest space
*/

    //   This function will calculate all of the ids in the SpSpace
    // specified by spaceLookupItem which correspond to grids whose
    // area overlaps the area of spaceLookupItem.  In addition to the
    // ids of grids, this function also calculates the node which
    // manages these grids, and the rest of the members of class
    // SpGridLookupOutput and its ancestors (the C_QUEUE output argument is
    // a list of SpGridLookupOutput objects).
    //   All SpaceSubSets of the specified SpSpace are checked for
    // matches.
    //   This function returns TRUE iff spaceLookupItem fits entirely
    // within the boundaries of at least one of the SpaceSubSets
    // contained in the specified SpSpace.  This means that if there are
    // two SpaceSubSets in the specified SpSpace, and one of them
    // contains half of spaceLookupItem, and the other contains
    // exactly the opposite (maybe complementary is a better word)
    // half, then this function returns FALSE since no *one*
    // SpSpaceSubSet by *itself* contains all of spaceLookupItem.
    //   If a SpDimLookupItem in spaceLookupItem is a "Don't
    // Care", then that lookup item is considered to exactly fit
    // within all SpaceSubSets, and will NOT cause this function to
    // return FALSE (of course FALSE could still be returned if
    // another SpDimLookupItem is not a "Don't Care", and falls
    // out side all SpaceSubSets).
    //   If the return value is FALSE, then maybe part and maybe even
    // all of the spaceLookupItem is NOT being accounted for by the
    // specified space.  If you later do another lookup on this space,
    // and it also returns FALSE, then it is possible that there are
    // intersections between the two lookup items.  If so, then those
    // intersections will NOT be reported to you.  If you think this
    // would be bad, then you may want to check the return value and
    // if it is 0, then you may want to try to figure out how to
    // modify your code until you no longer get a return value of 0.
    // However, if you don't care if the lookup doesn't fit in any
    // SpaceSubSets, then, you can just ignore the return value.
    int HiGrid_Lookup(const SpSpaceLookupItem &spaceLookupItem,
		      C_QUEUE * const ids_of_grids_in_speedes) const;
};

inline ostream & operator<<(ostream &os, const O_SpGrid &o_grid)
{
  o_grid.Print(os);
  return os;
}

DEFINE_OBJECT_MANAGER(Grid, O_SpGrid);

typedef O_SpGrid O_Grid;

#endif
