#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpLatLngLookupItem.H"
#include "SpPubSubItem.H"
#include "SpGeoFunctions.H"

SpLatLngLookupItem::SpLatLngLookupItem(const double radLat,
				   const double radLng,
				   const double kmAlt,
				   const double kmSizeRadius,
				   const double kmResolutionRadius)
  : SpDimLookupItem("LatLng", 0)
{
  set_RadLat(radLat);
  set_RadLng(radLng);
  set_KmAlt(kmAlt);
  set_KmSizeRadius(kmSizeRadius);
  set_KmResolutionRadius(kmResolutionRadius);
}

SpLatLngLookupItem &SpLatLngLookupItem::set_RadLat(const double radLat)
{
  // This line was not here for about 2 years until I noticed its
  // absence and finally added it.  I do not know why it used to not
  // be here.
  set_DontCare_p(0);

  assert(south_pole_rad_lat <= radLat && radLat <= north_pole_rad_lat);
  RadLat = radLat;
  return *this;
}

SpLatLngLookupItem &SpLatLngLookupItem::set_RadLng(double radLng)
{
  set_DontCare_p(0);

  const double
    min_angle = -PI,
    max_angle =  2 * PI;
  
  assert(min_angle <= radLng && radLng <= max_angle);

  if (radLng > PI)
    radLng -= TWO_PI;
  
  RadLng = radLng;

  return *this;
}

SpLatLngLookupItem &SpLatLngLookupItem::set_KmAlt(double kmAlt)
{
  set_DontCare_p(0);
  assert(kmAlt >= -radius_of_earth);
  KmAlt = kmAlt;
  return *this;
}

SpLatLngLookupItem &SpLatLngLookupItem::set_KmSizeRadius(double in)
{
  set_DontCare_p(0);
  assert(in >= 0);
  KmSizeRadius = in;
  return *this;
}

SpLatLngLookupItem &SpLatLngLookupItem::set_KmResolutionRadius(const double in)
{
  set_DontCare_p(0);
  assert(in >= 0);
  KmResolutionRadius = in;
  return *this;
}

void SpLatLngLookupItem::print_work(ostream &os) const
{
  os << setprecision(8) << resetiosflags(ios::left)
     << "Lat = " << setw(9) << RAD_to_DEG(RadLat)
     << " degrees, Lng = " 
     << setw(9) << RAD_to_DEG(RadLng)
     << " degrees,\n                        Alt = "
     << setw(9) << KmAlt << " km, SizeRadius = "
     << setw(9) << KmSizeRadius
     << " km,\n                        ResolutionRadius = "
     << setw(9) << KmResolutionRadius << " km" << endl;
}
