// Attribute.C Source File

#include "SpAddPaddingToAlign.H"

#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"


#include "SpOS.H"
#include "SpExportAttribute.H"
#include "SpObjProxy.H"
#include "SpObjFactory.H"
#include "F_SpUpdateListItem.H"
#include "SpScheduleEvent.H"

#include "RB_SpContainerFuncs.H"

#include "SpEvent.H"
#include "S_SpHLA.H"

extern SpSimObj *CurrentSimobj;
extern SpEvent *CurrentEvent;

extern SpObjFactory *ObjFact;

#include "SpFreeDynAttributes.H"
extern SpFreeDynAttributes *FreeDynamicAttributes;

#include "SpFreeHLA.H"
extern SpFreeHLA *FreeHLA;

int OBJECT_ATTRIBUTE::Counter = 0;

//proxy:  need to initialze obj proxy slot, InObjectAttribute, with something
//      that isn't NULL;  make bogus pointer for this purpose:
static int BogusInt;
static OBJECT_ATTRIBUTE *BogusObjAttPtr = (OBJECT_ATTRIBUTE *) &BogusInt;

EVENT_TYPE (SCHEDULE_UPDATE);
EVENT_TYPE (REFLECT_ATTRIBUTES);
EVENT_TYPE (BROADCAST_PROXY);
EVENT_TYPE (DELIVER_PROXY);
EVENT_TYPE (DELIVER_PROXY_PTR);
EVENT_TYPE (TOUCH_PROXY);

EVENT_TYPE (ADD_PROXY_LOCAL_PROXYMAN);
EVENT_TYPE (NEW_PROXY_LOCAL_PROXYMAN);
EVENT_TYPE (REMOVE_PROXY_LOCAL_PROXYMAN);
EVENT_TYPE (ADD_SUBSCRIBER_HLA);
EVENT_TYPE (DEFAULT_PUBLISH_HLA);
EVENT_TYPE (DEFAULT_SUBSCRIBE_HLA);
EVENT_TYPE (UNDELIVER_PROXY_PTR);
EVENT_TYPE (UNPUBLISH_HLA);
EVENT_TYPE (UNSUBSCRIBE_HLA);

EVENT_TYPE (PUBLISH_DECMAN);
EVENT_TYPE (SUBSCRIBE_DECMAN);
EVENT_TYPE (UNPUBLISH_DECMAN);
EVENT_TYPE (UNSUBSCRIBE_DECMAN);

EVENT_TYPE (AddSubscriberSpace);
EVENT_TYPE (PublishGrid);
EVENT_TYPE (SubscribeGrid);
EVENT_TYPE (UnPublishGrid);
EVENT_TYPE (UnPublishSpace);
EVENT_TYPE (UnSubscribeGrid);
EVENT_TYPE (UnSubscribeSpace);

EVENT_TYPE (DeliverProxyPtrSpace);
EVENT_TYPE (UndeliverProxyPtrSpace);

#define case_and_output(type) case type: os << #type ; break
ostream & operator << (ostream &os, const ATTRIBUTE_TYPE type)
{
  switch (type)
  {
    case_and_output(STATIC_INT_ATTRIBUTE_TYPE);
    case_and_output(STATIC_FLOAT_ATTRIBUTE_TYPE);
    case_and_output(STATIC_LOGICAL_ATTRIBUTE_TYPE);
    case_and_output(STATIC_POSITION_ATTRIBUTE_TYPE);

    case_and_output(DYNAMIC_INT_ATTRIBUTE_TYPE);
    case_and_output(DYNAMIC_FLOAT_ATTRIBUTE_TYPE);
    case_and_output(DYNAMIC_LOGICAL_ATTRIBUTE_TYPE);
    case_and_output(DYNAMIC_OBJECT_ATTRIBUTE_TYPE);
    case_and_output(DYNAMIC_POSITION_ATTRIBUTE_TYPE);

    case_and_output(STRING_ATTRIBUTE_TYPE);

    case_and_output(OBJECT_ATTRIBUTE_TYPE);
    case_and_output(LIST_ATTRIBUTE_TYPE);

    default:
      cerr << "unknown ATTRIBUTE_TYPE";
      abort();
  }
  return  os;
}

int MapAttribute(char *name) {

  if (!strcasecmp(name,"STATIC_INT"))
    return STATIC_INT_ATTRIBUTE_TYPE;
  if (!strcasecmp(name,"STATIC_FLOAT"))
    return STATIC_FLOAT_ATTRIBUTE_TYPE;
  if (!strcasecmp(name,"STATIC_LOGICAL"))
    return STATIC_LOGICAL_ATTRIBUTE_TYPE;
  if (!strcasecmp(name,"STATIC_POSITION"))
    return STATIC_POSITION_ATTRIBUTE_TYPE;

  if (!strcasecmp(name,"INT"))
    return STATIC_INT_ATTRIBUTE_TYPE;
  if (!strcasecmp(name,"FLOAT"))
    return STATIC_FLOAT_ATTRIBUTE_TYPE;
  if (!strcasecmp(name,"LOGICAL"))
    return STATIC_LOGICAL_ATTRIBUTE_TYPE;
  if (!strcasecmp(name,"POSITION"))
    return STATIC_POSITION_ATTRIBUTE_TYPE;

  if (!strcasecmp(name,"DYNAMIC_INT"))
    return DYNAMIC_INT_ATTRIBUTE_TYPE;
  if (!strcasecmp(name,"DYNAMIC_FLOAT"))
    return DYNAMIC_FLOAT_ATTRIBUTE_TYPE;
  if (!strcasecmp(name,"DYNAMIC_LOGICAL"))
    return DYNAMIC_LOGICAL_ATTRIBUTE_TYPE;
  if (!strcasecmp(name,"DYNAMIC_OBJECT"))
    return DYNAMIC_OBJECT_ATTRIBUTE_TYPE;
  if (!strcasecmp(name,"DYNAMIC_POSITION"))
    return DYNAMIC_POSITION_ATTRIBUTE_TYPE;

  if (!strcasecmp(name,"STRING")) return STRING_ATTRIBUTE_TYPE;

  if (!strcasecmp(name,"OBJECT")) return OBJECT_ATTRIBUTE_TYPE;
  if (!strcasecmp(name,"LIST")) return LIST_ATTRIBUTE_TYPE;

  cout << "Error (MapAttribute): Bad argument " << name << endl;
  return -1;
}

char *MapAttribute(int index) {

  if ((index < 0) || (index >= N_ATTRIBUTE_TYPES)) {
    cout << "Error (MappAttribute): Bad argument " << index << endl;
    return NULL;
  }

  switch (index) {

/*
** STATIC ATTRIBUTE TYPES
*/

    case STATIC_INT_ATTRIBUTE_TYPE:
      return "INT";

    case STATIC_FLOAT_ATTRIBUTE_TYPE:
      return "FLOAT";

    case STATIC_LOGICAL_ATTRIBUTE_TYPE:
      return "LOGICAL";

    case STATIC_POSITION_ATTRIBUTE_TYPE:
      return "POSITION";

/*
** DYNAMIC ATTRIBUTE TYPES
*/

    case DYNAMIC_INT_ATTRIBUTE_TYPE:
      return "DYNAMIC_INT";

    case DYNAMIC_FLOAT_ATTRIBUTE_TYPE:
      return "DYNAMIC_FLOAT";

    case DYNAMIC_LOGICAL_ATTRIBUTE_TYPE:
      return "DYNAMIC_LOGICAL";

    case DYNAMIC_OBJECT_ATTRIBUTE_TYPE:
      return "DYNAMIC_OBJECT";

    case DYNAMIC_POSITION_ATTRIBUTE_TYPE:
      return "DYNAMIC_POSITION";

/*
** OTHER ATTRIBUTE TYPES
*/

    case STRING_ATTRIBUTE_TYPE:
      return "STRING";

    case OBJECT_ATTRIBUTE_TYPE:
      return "OBJECT";

    case LIST_ATTRIBUTE_TYPE:
      return "LIST";


  }
  cout << "Error, fell through switch (MappAttribute): Bad argument " << index << endl;
  return NULL;

}

void STATIC_INT_ATTRIBUTE::EmptyMessage(const char *buff) {
  INT_ATTRIBUTE_MESSAGE_INFO *info = (INT_ATTRIBUTE_MESSAGE_INFO *)buff;
  attribute = info->GetValue();
}

void STATIC_FLOAT_ATTRIBUTE::FillMessage(char *buff) {
  FLOAT_ATTRIBUTE_MESSAGE_INFO *info
    = (FLOAT_ATTRIBUTE_MESSAGE_INFO *)buff;
  info->SetNbytes(GetMessageSize());
  info->SetAttributeIndex(GetAttributeIndex());
  info->SetValue(attribute);
}

void STATIC_FLOAT_ATTRIBUTE::EmptyMessage(const char *buff) {
  FLOAT_ATTRIBUTE_MESSAGE_INFO *info =
    (FLOAT_ATTRIBUTE_MESSAGE_INFO *)buff;
  attribute = info->GetValue();
}

void STATIC_LOGICAL_ATTRIBUTE::FillMessage(char *buff) {
  LOGICAL_ATTRIBUTE_MESSAGE_INFO *info =
    (LOGICAL_ATTRIBUTE_MESSAGE_INFO *)buff;
  info->SetNbytes(GetMessageSize());
  info->SetAttributeIndex(GetAttributeIndex());
  info->SetValue(GetValue());
}

void STATIC_LOGICAL_ATTRIBUTE::EmptyMessage(const char *buff) {
  LOGICAL_ATTRIBUTE_MESSAGE_INFO *info =
    (LOGICAL_ATTRIBUTE_MESSAGE_INFO *)buff;
  attribute = info->GetValue();
}

void STATIC_POSITION_ATTRIBUTE::FillMessage(char *buff) {
  POSITION_ATTRIBUTE_MESSAGE_INFO *info =
    (POSITION_ATTRIBUTE_MESSAGE_INFO *)buff;
  info->SetNbytes(GetMessageSize());
  info->SetAttributeIndex(GetAttributeIndex());
  double Pos[3];
  Pos[0] = Position[0];
  Pos[1] = Position[1];
  Pos[2] = Position[2];
  info->SetValue(Pos, PositionType);
}

void STATIC_POSITION_ATTRIBUTE::EmptyMessage(const char *buff) {
  POSITION_ATTRIBUTE_MESSAGE_INFO *info =
    (POSITION_ATTRIBUTE_MESSAGE_INFO *)buff;
  double Pos[3];
  int PosType;
  info->GetValue(Pos, &PosType);
  Position[0] = Pos[0];
  Position[1] = Pos[1];
  Position[2] = Pos[2];
  PositionType = PosType;
}



//proxy: new constructor
OBJECT_ATTRIBUTE::OBJECT_ATTRIBUTE(char *className)
{
  Active = 1; 
  if (className!=NULL) {
    SetClassName(className); // <-- CREATES right proxy in ObjectProxy slot 
  }
  else {
    ObjectProxy = NULL;
  }
  /*
  StartTime = -1.0e20;
  EndTime = 1.0e20;
  */
  ListId = Counter;
  Counter++;

}


OBJECT_ATTRIBUTE *LIST_ATTRIBUTE::operator() (char *n) {
  OBJECT_ATTRIBUTE *ObjectAttribute;
  ObjectAttribute = (OBJECT_ATTRIBUTE *)get_top();
  while (ObjectAttribute) {
    if (!strcmp(ObjectAttribute->GetObjectName(), n)) {
      return ObjectAttribute;
    }else{
      ObjectAttribute = (OBJECT_ATTRIBUTE *)++list;
    }
  }
  return NULL;
}

/*
OBJECT_ATTRIBUTE *LIST_ATTRIBUTE::operator() (double t) {
  OBJECT_ATTRIBUTE *ObjectAttribute;
  ObjectAttribute = (OBJECT_ATTRIBUTE *)get_top();
  while (ObjectAttribute) {
    if ((ObjectAttribute->GetStartTime() < t) &&
	(t < ObjectAttribute->GetEndTime())) {
      return ObjectAttribute;
    }else{
      ObjectAttribute = (OBJECT_ATTRIBUTE *)++list;
    }
  }
  return NULL;
}
*/

//proxy: add new method
// If possible, take the updateList ptr from the proxy that contains "addedItem"
// and insert it into addItem's proxy (and all subsidiary proxies below).
void LIST_ATTRIBUTE::InsertUpdateList(OBJECT_ATTRIBUTE *addedItem,
				      SpObjProxy *containingProxy) {
  RB_UpdateXqueue *updateList = containingProxy->GetAttributeUpdateList();
  if (updateList != NULL) {
    SpObjProxy *objProxy = addedItem->GetObjectProxy();
    if (objProxy==NULL) {
      cout << "ERROR adding item to LIST_ATTRIB:  new OBJ_ATT has no proxy!"
	   << endl;
      return;
    }
    
    objProxy->SetAttributeUpdateList(updateList); // store ptr to updatelist
    // recursively store updateList in all subsidiary proxies:
    objProxy->SetProxyUpdatePtrs(updateList);
  }
  // else:  if updateList has not been inserted into the containing proxy, it means
  //        we haven't yet hit the root proxy;  if/when we do, the updateList will
  //        be created, and we'll recurse down to this proxy, inserting the ptr then.
}


OBJECT_ATTRIBUTE *LIST_ATTRIBUTE::FindObjectAttribute(int ListId) {
  OBJECT_ATTRIBUTE *ObjectAttribute;
  ObjectAttribute = (OBJECT_ATTRIBUTE *)get_top();
  while (ObjectAttribute) {
    if (ObjectAttribute->GetListId() == ListId) {
      return ObjectAttribute;
    }else{
      ObjectAttribute = (OBJECT_ATTRIBUTE *)++list;
    }
  }
  return NULL;
}

void LIST_ATTRIBUTE::outstream(ostream &out, int level){
  OBJECT_ATTRIBUTE *ObjectAttribute;
  out << "List has " << list.get_length() << " elements";
  if (list.get_length() == 0) out << endl;
  ObjectAttribute = (OBJECT_ATTRIBUTE *)get_top();
  while (ObjectAttribute) {
    ObjectAttribute->outstream(out, level);
    ObjectAttribute = (OBJECT_ATTRIBUTE *)++list;
  }
}

int LIST_ATTRIBUTE::GetMessageSize() {
  OBJECT_ATTRIBUTE *ObjectAttribute;
  int s = sizeof(LIST_ATTRIBUTE_MESSAGE_INFO);

  ObjectAttribute = (OBJECT_ATTRIBUTE *)list.get_top();
  while (ObjectAttribute) {
    s += ObjectAttribute->GetMessageSize();
    ObjectAttribute = (OBJECT_ATTRIBUTE *)++list;
  }
  return s;
}

void LIST_ATTRIBUTE::FillMessage(char *buff) {
  OBJECT_ATTRIBUTE *ObjectAttribute;
  char *buff1;
  LIST_ATTRIBUTE_MESSAGE_INFO *info = (LIST_ATTRIBUTE_MESSAGE_INFO *)buff;

  info->SetNbytes(GetMessageSize());
  info->SetAttributeIndex(GetAttributeIndex());
  info->SetLength(list.get_length());

  buff1 = (char *)&info[1];
  ObjectAttribute = (OBJECT_ATTRIBUTE *)list.get_top();

  while (ObjectAttribute) {
    ObjectAttribute->FillMessage(buff1);
    buff1 += ((OBJECT_ATTRIBUTE_MESSAGE_INFO *)buff1)->GetNbytes();
    ObjectAttribute = (OBJECT_ATTRIBUTE *)++list;
  }

}

void LIST_ATTRIBUTE::EmptyMessage(const char *buff) {
  OBJECT_ATTRIBUTE *ObjectAttribute;
  char *buff1;
  LIST_ATTRIBUTE_MESSAGE_INFO *info = (LIST_ATTRIBUTE_MESSAGE_INFO *)buff;

  int Length = info->GetLength();
  buff1 = (char *)&info[1];

  for (int i=0; i<Length; i++) {
    ObjectAttribute = new OBJECT_ATTRIBUTE();
    ObjectAttribute->EmptyMessage(buff1);
    *this += ObjectAttribute;
    buff1 += ((OBJECT_ATTRIBUTE_MESSAGE_INFO *)buff1)->GetNbytes();
  }

}

/************************************************************************
 * OBJECT_ATTRIBUTE - destructor for object attribute			*
 ************************************************************************/
OBJECT_ATTRIBUTE::~OBJECT_ATTRIBUTE() {
  ObjFact->PartialDestroy(ObjectProxy); 
  /*
  if ((!(ObjectProxy->GetDefined())) || (ObjectProxy->FullDestroy_p))
    ObjFact->Destroy(ObjectProxy); 
  else
    ObjFact->PartialDestroy(ObjectProxy);
    */
}


/************************************************************************
 * BASE_ATTRIBUTE - GetDepth: Determines the depth in terms of Object	*
 * Proxies.								*
 ************************************************************************/
int BASE_ATTRIBUTE::GetDepth() {
  OBJECT_ATTRIBUTE *InObjectAttribute;

  if (InObjectProxy) {

    InObjectAttribute =
      InObjectProxy->GetInObjectAttribute();

    if (InObjectAttribute) {
      return 1 + InObjectAttribute->GetDepth();
    }else{
      return 1;
    }

  }else{

    cout << "Error (GetDepth): Attribute not in an Object Proxy" << endl;
    return 0;

  }

}

/************************************************************************
 * OBJECT_ATTRIBUTE - GetDepth: Determines the depth in terms of Object	*
 * Proxies.								*
 ************************************************************************/
int OBJECT_ATTRIBUTE::GetDepth() {
  OBJECT_ATTRIBUTE *InObjectAttribute;

  if (InObjectProxy) {

    InObjectAttribute = (OBJECT_ATTRIBUTE *)
      InObjectProxy->GetInObjectAttribute();

    if (InObjectAttribute) {
      return 1 + InObjectAttribute->GetDepth();
    }else{
      return 1;
    }

  }

  if (InListAttribute) {
    return InListAttribute->GetDepth();
  }

  cerr << "InObjectProxy and InObjectAttribute are both FALSE!" << endl;
  abort();
  return 0;
}

char *BASE_ATTRIBUTE::PackLevelInfo(const char *Buff, int NlevelInfo) {
  PROXY_LEVEL_INFO *ProxyLevelInfo;
  OBJECT_ATTRIBUTE *InObjectAttribute;

  if (InObjectProxy) {
    InObjectAttribute =
      InObjectProxy->GetInObjectAttribute();

    if (InObjectAttribute) {

      ProxyLevelInfo = (PROXY_LEVEL_INFO *)
	InObjectAttribute->PackLevelInfo(Buff, NlevelInfo);

    }else{

      ProxyLevelInfo = (PROXY_LEVEL_INFO *)Buff;
    }

    ProxyLevelInfo->Level = NlevelInfo;
    ProxyLevelInfo->AttributeIndex = AttributeIndex;
    ProxyLevelInfo->ArrayIndex = ArrayIndex;
    ProxyLevelInfo->ListId = -1;

    return (char *)&ProxyLevelInfo[1];

  }else{

    cout << "Error (PackLevelInfo): Attribute not in an Object Proxy" << endl;
    return NULL;

  }

}

char *OBJECT_ATTRIBUTE::PackLevelInfo(const char *Buff, int NlevelInfo) {
  PROXY_LEVEL_INFO *ProxyLevelInfo;
  OBJECT_ATTRIBUTE * InObjectAttribute;
  char *b;

  if (InObjectProxy) {

    InObjectAttribute =
      InObjectProxy->GetInObjectAttribute();

    if (InObjectAttribute) {

      ProxyLevelInfo = (PROXY_LEVEL_INFO *)
	InObjectAttribute->PackLevelInfo(Buff, NlevelInfo);

    }else{

      ProxyLevelInfo = (PROXY_LEVEL_INFO *)Buff;
    }

    ProxyLevelInfo->Level = NlevelInfo;
    ProxyLevelInfo->AttributeIndex = AttributeIndex;
    ProxyLevelInfo->ArrayIndex = ArrayIndex;
    ProxyLevelInfo->ListId = ListId;

    return (char *)&ProxyLevelInfo[1];

  }

  if (InListAttribute) {

    b = InListAttribute->PackLevelInfo(Buff, NlevelInfo);
    ProxyLevelInfo = (PROXY_LEVEL_INFO *)Buff;
    ProxyLevelInfo->ListId = ListId;

    return b;
  }

  cerr << "InObjectProxy and InListAttribute are both FALSE!" << endl;
  abort();
  return NULL;
}

/************************************************************************
 * OBJECT_ATTRIBUTE - SetInObjectProxy: Attribute is in Object Proxy	*
 ************************************************************************/
void OBJECT_ATTRIBUTE::SetInObjectProxy(SpObjProxy *o) {
  //SpObjTemplate *ObjectTemplate;

  InObjectProxy = o;
  InListAttribute = NULL;

  if (ObjectProxy) {
    ObjectProxy->SetInObjectAttribute(this);
  }else{
    /*
    ObjectTemplate = InObjectProxy->GetObjectTemplate();
    cout << "Error (OBJECT_ATTRIBUTE): Defining Object Attribute "
	 << ObjectTemplate->GetAttributeName(AttributeIndex)
	 <<  " Before it was created" << endl;
    */
  }

}

/************************************************************************
 * OBJECT_ATTRIBUTE - SetInListAttribute: Attribute is in list		*
 ************************************************************************/
void OBJECT_ATTRIBUTE::SetInListAttribute(LIST_ATTRIBUTE *l) {
  SpObjTemplate *ObjectTemplate;

  InListAttribute = l;
  InObjectProxy = NULL;

  ArrayIndex = 0;
  AttributeIndex = 0;

  if (ObjectProxy) {
    ObjectProxy->SetInObjectAttribute(this);
  }else{
    ObjectTemplate = InObjectProxy->GetObjectTemplate();
    cout << "Error (OBJECT_ATTRIBUTE): Defining Object Attribute "
	 << ObjectTemplate->GetAttributeName(AttributeIndex)
	 <<  " Before it was created" << endl;
  }

}

/************************************************************************
 * OBJECT_ATTRIBUTE - SetClassName: constructs the ObjectProxy to be	*
 * the right kind of object.						*
 ************************************************************************/
//proxy: new version:
void OBJECT_ATTRIBUTE::SetClassName(char *ClassName) {
  ObjectProxy = ObjFact->Create(ClassName);
  // Temporarily insert BogusObjAttPtr to mark this proxy as
  // belonging to an OBJECT_ATTRIBUTE
  ObjectProxy->SetInObjectAttribute(BogusObjAttPtr);
  DEFINE_ATTRIBUTE(Active, "Active");
}

/************************************************************************
 * OBJECT_ATTRIBUTE - ObjProxyGetClassName: returns the class name of the	*
 * object								*
 ************************************************************************/
char *OBJECT_ATTRIBUTE::ObjProxyGetClassName() {
  return ObjectProxy->ObjProxyGetClassName();
}

/************************************************************************
 * OBJECT_ATTRIBUTE - SetObjectName: sets the name of the object	*
 ************************************************************************/
void OBJECT_ATTRIBUTE::SetObjectName(char *ObjectName) {
  ObjectProxy->SetObjectName(ObjectName);
}

/************************************************************************
 * OBJECT_ATTRIBUTE - GetObjectName: returns the name of the object	*
 ************************************************************************/
char *OBJECT_ATTRIBUTE::GetObjectName() {
  return ObjectProxy->GetObjectName();
}

/************************************************************************
 * OBJECT_ATTRIBUTE - outstream prints out the object to the outstream	*
 * variabale								*
 ************************************************************************/
void OBJECT_ATTRIBUTE::outstream(ostream &out, int level) {
  if (ObjectProxy) {
//    out << ObjectProxy->GetObjectName();
    ObjectProxy->PrintStream(out, level+1);
  }else{
    out << "NULL";
  }
}

/************************************************************************
 * OBJECT_ATTRIBUTE - GetMessageSize returns the size of the Object 	*
 * Proxy Message size. This is used for packing an object proxy into	*
 * a message for transmision.						*
 ************************************************************************/
int OBJECT_ATTRIBUTE::GetMessageSize() {

  int size = sizeof(OBJECT_ATTRIBUTE_MESSAGE_INFO) +
    ObjFact->GetMessageSize(ObjectProxy);

//  cout << "Size of object = " << size << " bytes" << endl;

  return size;
}

/************************************************************************
 * OBJECT_ATTRIBUTE - FillMessage packs a message with Object Proxy	*
 * data.       								*
 ************************************************************************/
void OBJECT_ATTRIBUTE::FillMessage(char *buff) {
  SpObjProxyMsg *ObjectProxyMessage =
    ObjFact->GenerateMessage(ObjectProxy);

  OBJECT_ATTRIBUTE_MESSAGE_INFO *info = (OBJECT_ATTRIBUTE_MESSAGE_INFO *)buff;

  // I commented these two lines out since they are no longer necessary
  // since I made the people who call this code set these values.  In
  // addition to not now needing these things, I also got rid of them
  // because they seem to set the wrong value, and I did not know how
  // to fix it.
  info->SetNbytes(GetMessageSize());
  info->SetAttributeIndex(GetAttributeIndex());

  info->SetListId(ListId);

/*
  cout << "Nbytes in message = " << ObjectProxyMessage->GetNbytes() << endl;
  cout << "Nbytes in object info = " << info->GetNbytes() << endl;
*/

  info->SetValue((char *)ObjectProxyMessage, ObjectProxyMessage->GetNbytes());

  delete [] ((char*)ObjectProxyMessage);

}

/************************************************************************
 * OBJECT_ATTRIBUTE - EmptyMessage takes a message and constructs the	*
 * Object Proxy.       	.						*
 ************************************************************************/
void OBJECT_ATTRIBUTE::EmptyMessage(const char *buff) {
  OBJECT_ATTRIBUTE_MESSAGE_INFO *info = (OBJECT_ATTRIBUTE_MESSAGE_INFO *)buff;
  SpObjProxyMsg *ObjectProxyMessage  = (SpObjProxyMsg *)&info[1];
  ObjectProxy = ObjFact->GenerateObjectProxy(ObjectProxyMessage);

  ObjectProxy->SetInObjectAttribute(this);
  ListId = info->GetListId();

//  ObjectProxy->PrintStream();
}

/************************************************************************
 * BASE_DYNAMIC_ATTRIBUTE - destructor		.			*
 ************************************************************************/
BASE_DYNAMIC_ATTRIBUTE::~BASE_DYNAMIC_ATTRIBUTE() {
  SpDynItem *DynamicItem;

  while ((DynamicItem = (SpDynItem *)list.pop_top())) {
    FreeDynamicAttributes->delete_object(DynamicItem);
  }

}

SpDynItem *BASE_DYNAMIC_ATTRIBUTE::FindDynamicItem(double t) {
  SpDynItem *DynamicItem;
  DynamicItem = (SpDynItem *)list.get_top();
  while (DynamicItem) {
    if (DynamicItem->GoodTime(t)) {return DynamicItem;}
    DynamicItem = (SpDynItem *)(++list);
  }
  cout << "Error (FindDynamicItem): Could not find item at time "
       << t << endl;
  return NULL;
}

SpDynItem *BASE_DYNAMIC_ATTRIBUTE::FindDynamicItem(int i){
  SpDynItem *DynamicItem;
  DynamicItem = (SpDynItem *)list.get_top();
  while (DynamicItem) {
    if (DynamicItem->GetCountId() == i) {return DynamicItem;}
    DynamicItem = (SpDynItem *)(++list);
  }
  cout << "Error (FindDynamicItem): Could not find item with id "
       << i << endl;
  return NULL;
}

double BASE_DYNAMIC_ATTRIBUTE::GetStartTime() {
  double StartTime = INFINITY;
  SpDynItem *DynamicItem;
  DynamicItem = (SpDynItem *)list.get_top();
  while (DynamicItem) {
    if (DynamicItem->GetStartTime() < StartTime) {
      StartTime = DynamicItem->GetStartTime();
    }
    DynamicItem = (SpDynItem *)(++list);
  }
  return StartTime;
}

double BASE_DYNAMIC_ATTRIBUTE::GetEndTime() {
  double EndTime = -INFINITY;
  SpDynItem *DynamicItem;
  DynamicItem = (SpDynItem *)list.get_top();
  while (DynamicItem) {
    if (DynamicItem->GetEndTime() > EndTime) {
      EndTime = DynamicItem->GetEndTime();
    }
    DynamicItem = (SpDynItem *)(++list);
  }
  return EndTime;
}

void BASE_DYNAMIC_ATTRIBUTE::outstream(ostream &out, int level) {
  out << list.get_length() << " items";
  if (list.get_length()) out << endl;
  SpDynItem *DynamicItem = (SpDynItem *)list.get_top();
  while (DynamicItem) {
    for (int i=0; i<=level; i++) out << "     ";
    out << DynamicItem->GetType() << ": Time = ["
	<< DynamicItem->GetStartTime() << ", "
	<< DynamicItem->GetEndTime() << "]";
    DynamicItem = (SpDynItem *)(++list);
    if (DynamicItem) out << endl;
  }
}

/************************************************************************
 * BASE_DYNAMIC_ATTRIBUTE - GetMessageSize returns the size of the	*
 * buffer required to store the list of dynamic items.			*
 ************************************************************************/
int BASE_DYNAMIC_ATTRIBUTE::GetMessageSize() {
  SpDynItem *DynamicItem;
  int s = sizeof(DYNAMIC_ATTRIBUTE_MESSAGE_INFO);

  DynamicItem = (SpDynItem *)list.get_top();
  while (DynamicItem) {
    s += FreeDynamicAttributes->get_size(DynamicItem->get_freeid());
    DynamicItem = (SpDynItem *)++list;
  }

  return s;
}

/************************************************************************
 * BASE_DYNAMIC_ATTRIBUTE - FillMessage fills the buffer with the list	*
 * of dynamic items				      			*
 ************************************************************************/
void BASE_DYNAMIC_ATTRIBUTE::FillMessage(char *buff) {
  SpDynItem *DynamicItem;
  char *buff1;
  int s;
  DYNAMIC_ATTRIBUTE_MESSAGE_INFO *info =
    (DYNAMIC_ATTRIBUTE_MESSAGE_INFO *)buff;

  info->SetNbytes(GetMessageSize());
  info->SetAttributeIndex(GetAttributeIndex());
  info->SetLength(list.get_length());

  buff1 = (char *)&info[1];
  DynamicItem = (SpDynItem *)list.get_top();
  while (DynamicItem) {
    s = FreeDynamicAttributes->get_size(DynamicItem->get_freeid());
    memcpy(buff1, (char *)DynamicItem, s);
    buff1 += s;
    DynamicItem = (SpDynItem *)++list;
  }

}

/************************************************************************
 * BASE_DYNAMIC_ATTRIBUTE - EmptyMessage takes the message and creates	*
 * dynamic items which are added to the xlist.				*
 ************************************************************************/
void BASE_DYNAMIC_ATTRIBUTE::EmptyMessage(const char *buff) {
  SpDynItem *DynamicItem;
  char *buff1;
  int s;
  DYNAMIC_ATTRIBUTE_MESSAGE_INFO *info =
    (DYNAMIC_ATTRIBUTE_MESSAGE_INFO *)buff;

  int Length =info->GetLength();
  buff1 = (char *)&info[1];

  for (int i=0; i<Length; i++) {
    DynamicItem = (SpDynItem *)FreeDynamicAttributes->generate(buff1);
    list += DynamicItem;
    s = FreeDynamicAttributes->get_size(DynamicItem->get_freeid());
    buff1 += s;
  }

}

extern int EXTERNAL_OBJECT_PROXY;

/************************************************************************
 * ScheduleUpdate - This routine is called whenever an attribute is	*
 * modified. It schedules an UpdateAttribute event in SPEEDES for the	*
 * first attribute that is updated in an event.				*
 ************************************************************************/
void BASE_ATTRIBUTE::ScheduleUpdate
  (ADD_REMOVE AddRemove, ATTRIBUTE_LIST_ELEMENT ptr) {
  if (EXTERNAL_OBJECT_PROXY) return;
  int i, len;
  F_UPDATE_LIST_ITEM *UpdateListItem;
  RB_UpdateXqueue *UpdateList;  // proxy: new list type
  static int done = 0;


  // proxy: NOTE: this huge bunch of if's will all be removed (down to "end proxy")
  //         they're here now so i can run BmdSim (see "AmSavingUpdates" below)


  if (CurrentEvent == NULL) return;

  if (!done) {
    if (CurrentEvent->get_time_tag() == -1.0e20) return;
    done = 1;
  }

  if (CurrentEvent->get_EVENT_TYPE() == REFLECT_ATTRIBUTES) {
//    cout << "Warning: Reflect Attributes invoking ScheduleUpdate" << endl;
    return;
  }

  if (CurrentEvent->get_EVENT_TYPE() == BROADCAST_PROXY) {
//    cout << "Warning: Broadcast Proxy invoking ScheduleUpdate" << endl;
    return;
  }

  if (CurrentEvent->get_EVENT_TYPE() == DELIVER_PROXY) {
//    cout << "Warning: Deliver Proxy invoking ScheduleUpdate" << endl;
    return;
  }

  if (CurrentEvent->get_EVENT_TYPE() == DELIVER_PROXY_PTR) {
//    cout << "Warning: Deliver Proxy Ptr invoking ScheduleUpdate" << endl;
    return;
  }

  if (CurrentEvent->get_EVENT_TYPE() == TOUCH_PROXY) {
//    cout << "Warning: Touch Proxy invoking ScheduleUpdate" << endl;
    return;
  }

    // Declaration Management events
  if (CurrentEvent->get_EVENT_TYPE()==ADD_PROXY_LOCAL_PROXYMAN)
    return;
  if (CurrentEvent->get_EVENT_TYPE()==NEW_PROXY_LOCAL_PROXYMAN)
    return;
  if (CurrentEvent->get_EVENT_TYPE()==REMOVE_PROXY_LOCAL_PROXYMAN)
    return;
  if (CurrentEvent->get_EVENT_TYPE()==ADD_SUBSCRIBER_HLA)
    return;
  if (CurrentEvent->get_EVENT_TYPE()==DEFAULT_PUBLISH_HLA)
    return;
  if (CurrentEvent->get_EVENT_TYPE()==DEFAULT_SUBSCRIBE_HLA)
    return;
  if (CurrentEvent->get_EVENT_TYPE()==UNDELIVER_PROXY_PTR)
    return;
  if (CurrentEvent->get_EVENT_TYPE()==UNPUBLISH_HLA)
    return;
  if (CurrentEvent->get_EVENT_TYPE()==UNSUBSCRIBE_HLA)
    return;
  if (CurrentEvent->get_EVENT_TYPE()==PUBLISH_DECMAN)
    return;
  if (CurrentEvent->get_EVENT_TYPE()==SUBSCRIBE_DECMAN)
    return;
  if (CurrentEvent->get_EVENT_TYPE()==UNPUBLISH_DECMAN)
    return;
  if (CurrentEvent->get_EVENT_TYPE()==UNSUBSCRIBE_DECMAN)
    return;

    // Data Distribution Management events
  if (CurrentEvent->get_EVENT_TYPE()==PublishGrid)
    return;
  if (CurrentEvent->get_EVENT_TYPE()==UnPublishGrid)
    return;
  if (CurrentEvent->get_EVENT_TYPE()==SubscribeGrid)
    return;
  if (CurrentEvent->get_EVENT_TYPE()==UnSubscribeGrid)
    return;
  if (CurrentEvent->get_EVENT_TYPE()==UnSubscribeSpace)
    return;
  if (CurrentEvent->get_EVENT_TYPE()==UnPublishSpace)
    return;
  if (CurrentEvent->get_EVENT_TYPE()==AddSubscriberSpace)
    return;
  if (CurrentEvent->get_EVENT_TYPE()==DeliverProxyPtrSpace)
    return;
  if (CurrentEvent->get_EVENT_TYPE()==UndeliverProxyPtrSpace)
    return;

  
  // end proxy note

  //proxy:  replace this:
  //  S_SpHLA *HLAobj = (S_SpHLA *)CurrentSimobj;
  //  UpdateList = HLAobj->GetUpdateList();
  //  len = UpdateList->get_length();

  // The ptr to the updateList is now in the proxy that owns this
  // attribute.  If this attrib has not been Defined into a proxy, there's
  // nothing we can do except bail out:
  SpObjProxy *ownerProxy = GetInObjectProxy(); //proxy:
  if (ownerProxy==NULL) {
    return;
  }

			
  UpdateList = ownerProxy->GetAttributeUpdateList();  //proxy:

  // proxy:  
  if (UpdateList==NULL) {
    return;
  }


  //proxy: future change
  //  if (!UpdateList->AmSavingUpdates()) {
  //    return;
  //  }


/*
** If there is nothing in the UpdateList, then we need to schedule an event
** immediately following this event to send out the updated attributes.
** Otherwise, the event has already been scheduled so we don't want to
** schedule it again.
*/


  len = UpdateList->get_length(); // proxy:

  // proxy:  NOTE:  following block to be removed
  if (len == 0) {


//    cout << "Scheduling Update event for Attribute Type " << GetType()
//	 << " at time " << CurrentEvent->get_time_tag() << endl;


    CurrentEvent->schedule(SpSimTime(CurrentEvent->get_time_tag())
			   .IncrementPriority2(), 
			   SCHEDULE_UPDATE,
			   CurrentEvent->get_OBJECT_TYPE(),
			   CurrentEvent->get_LOCAL_ID(),
			   CurrentEvent->get_NODE());
  }


/*
** Go through the UpdateList to check if the same attribute was already
** updated. If so, then don't put another update list item in the list.
*/

  if (AddRemove == NEITHER) {
    UpdateListItem = (F_UPDATE_LIST_ITEM *)UpdateList->get_top();
    for (i=0; i<len; i++) {
      if (UpdateListItem->GetAttribute() == this) {
	UpdateListItem->SetAddRemove(AddRemove);
	return;
      }
      UpdateListItem = (F_UPDATE_LIST_ITEM *)UpdateListItem->get_link();
    }
  }

  UpdateListItem = (F_UPDATE_LIST_ITEM *)
    RB_FREE_NEW(FreeHLA, UPDATE_LIST_ITEM);
  UpdateListItem->SetAddRemove(AddRemove);
  UpdateListItem->SetPtr(ptr);

//  UpdateListItem = (F_UPDATE_LIST_ITEM *)RB_NEW(new F_UPDATE_LIST_ITEM());

  UpdateListItem->SetAttribute(this);
  RB_INSERT(UpdateList, UpdateListItem);
//  cout << "Length of UpdateList = " << UpdateList->get_length() << endl;


}

double DYNAMIC_FLOAT_ATTRIBUTE::operator () (double t){
  DYNAMIC_FLOAT_ITEM *item = (DYNAMIC_FLOAT_ITEM *)FindDynamicItem(t);
  if (item) {
    return (*item)(t);
  }else{
    return -1.0e20;
  }
}

double DYNAMIC_FLOAT_ATTRIBUTE::operator () (double t, double &V) {
  DYNAMIC_FLOAT_ITEM *item = (DYNAMIC_FLOAT_ITEM *)FindDynamicItem(t);
  if (item) {
    return (*item)(t,V);
  }else{
    V = -1.0e20;
    return -1.0e20;
  }
}

double DYNAMIC_FLOAT_ATTRIBUTE::operator () (double t, double &V, double &A) {
  DYNAMIC_FLOAT_ITEM *item = (DYNAMIC_FLOAT_ITEM *)FindDynamicItem(t);
  if (item) {
    return (*item)(t,V,A);
  }else{
    V = -1.0e20;
    A = -1.0e20;
    return -1.0e20;
  }
}

DYNAMIC_OBJECT *DYNAMIC_OBJECT_ATTRIBUTE::operator () (double t) {
  DYNAMIC_OBJECT *item =
    (DYNAMIC_OBJECT *)FindDynamicItem(t);
  if (item) {
    return item;
  }else{
    cout << "Warning: could not find item at time " << t << endl;
    return NULL;
  }
}

void DYNAMIC_POSITION_ATTRIBUTE::operator () (double t, int PT, double x[3]) {
  DYNAMIC_POSITION_ITEM *item =
    (DYNAMIC_POSITION_ITEM *)FindDynamicItem(t);
  if (item) {
    (*item)(t, PT, x);
  }else{
    for (int i=0; i<3; i++) {
      x[i] = 0.0;
    }
    cout << "Warning: could not find item at time " << t << endl;
  }
}

void DYNAMIC_POSITION_ATTRIBUTE::operator () (double t, int PT, double x[3], double v[3]) {
  DYNAMIC_POSITION_ITEM *item =
    (DYNAMIC_POSITION_ITEM *)FindDynamicItem(t);
  if (item) {
    (*item)(t, PT, x, v);
  }else{
    for (int i=0; i<3; i++) {
      x[i] = 0.0;
      v[i] = 0.0;
    }
    cout << "Warning: could not find item at time " << t << endl;
  }
}

void DYNAMIC_POSITION_ATTRIBUTE::operator () (double t, int PT, double x[3], double v[3],
					      double a[3]){
  DYNAMIC_POSITION_ITEM *item =
    (DYNAMIC_POSITION_ITEM *)FindDynamicItem(t);
  if (item) {
    (*item)(t, PT, x, v, a);
  }else{
    for (int i=0; i<3; i++) {
      x[i] = 0.0;
      v[i] = 0.0;
      a[i] = 0.0;
    }
    cout << "Warning: could not find item at time " << t << endl;
  }
}

int DYNAMIC_POSITION_ATTRIBUTE::CheckMotion() {
  int RC_Status = 0;
  DYNAMIC_POSITION_ITEM *DPI = (DYNAMIC_POSITION_ITEM *)get_top();
  while (DPI) {
    RC_Status += DPI->CheckMotion();
    DPI = (DYNAMIC_POSITION_ITEM *)++list;
  }
  return RC_Status;
}

int DYNAMIC_POSITION_ATTRIBUTE::CheckTime() {
  int RC_Status = 0;
  DYNAMIC_POSITION_ITEM *DPI = (DYNAMIC_POSITION_ITEM *)get_top();
  double Tcheck = DPI->GetStartTime();
  while (DPI) {
    if (Tcheck != DPI->GetStartTime()) RC_Status++;
    Tcheck = DPI->GetEndTime();
    DPI = (DYNAMIC_POSITION_ITEM *)++list;
  }
  if (RC_Status) cout << "Warning (CheckTime): Non-continuous times" << endl;
  return RC_Status;
}

/************************************************************************
 * STATIC_POSITION_METHODS		.				*
 ************************************************************************/
void STATIC_POSITION_ATTRIBUTE::GetEARTH(double &Lat, double &Lon,
					 double &Alt, double t) {
  double x[3];
  double Pos[3];
  if (int(PositionType) == EARTH) {
    Lat = Position[0];
    Lon = Position[1];
    Alt = Position[2];
    return;
  }
  Pos[0] = Position[0];
  Pos[1] = Position[1];
  Pos[2] = Position[2];
  if (int(PositionType) == ECR) {
    ECR_to_EARTH(Pos, x);
    Lat = x[0];
    Lon = x[1];
    Alt = x[2];
    return;
  }
  if (int(PositionType) == ECI) {
    ECI_to_EARTH(t, Pos, x);
    Lat = x[0];
    Lon = x[1];
    Alt = x[2];
    return;
  }
}

void STATIC_POSITION_ATTRIBUTE::GetECI(double x[3], double t) {
  double v[3];
  double Pos[3];
  if (int(PositionType) == EARTH) {
    Pos[0] = Position[0];
    Pos[1] = Position[1];
    Pos[2] = Position[2];
    EARTH_to_ECI(t, Pos, x);
    return;
  }
  if (int(PositionType) == ECR) {
    x[0] = Position[0];
    x[1] = Position[1];
    x[2] = Position[2];
    v[0] = 0;
    v[1] = 0;
    v[2] = 0;
    ECR_to_ECI(t, x, v);
    return;
  }
  if (int(PositionType) == ECI) {
    x[0] = Position[0];
    x[1] = Position[1];
    x[2] = Position[2];
    return;
  }
}

void STATIC_POSITION_ATTRIBUTE::GetECR(double x[3], double t) {
  double v[3];
  double Pos[3];
  if (int(PositionType) == EARTH) {
    Pos[0] = Position[0];
    Pos[1] = Position[1];
    Pos[2] = Position[2];
    EARTH_to_ECR(Pos, x);
    return;
  }
  if (int(PositionType) == ECR) {
    x[0] = Position[0];
    x[1] = Position[1];
    x[2] = Position[2];
    return;
  }
  if (int(PositionType) == ECI) {
    x[0] = Position[0];
    x[1] = Position[1];
    x[2] = Position[2];
    v[0] = 0;
    v[1] = 0;
    v[2] = 0;
    ECI_to_ECR(t, x, v);
    return;
  }
}

void STATIC_POSITION_ATTRIBUTE::GetValue(double x[3], int PT, double t) {
  if (PT == EARTH) {
    GetEARTH(x[0], x[1], x[2], t);
  }
  if (PT == ECI) {
    GetECI(x,t);
  }
  if (PT == ECR) {
    GetECR(x,t);
  }
}

void STATIC_POSITION_ATTRIBUTE::outstream(ostream &out, int){
  if (int(PositionType) == ECI) out << "ECI: ";
  if (int(PositionType) == ECR) out << "ECR: ";
  if (int(PositionType) == EARTH) out << "EARTH: ";
  out << Position[0] << " " << Position[1] << " " << Position[2];
}

void OBJECT_ATTRIBUTE::Copy(BASE_ATTRIBUTE *a) {
  OBJECT_ATTRIBUTE *Oa = (OBJECT_ATTRIBUTE *)a;
  memcpy((char *)this, (char *)Oa, sizeof(OBJECT_ATTRIBUTE));
  SpObjProxy *Op = Oa->GetObjectProxy();
  ObjectProxy = ObjFact->Clone(Op);
  ObjectProxy->SetInObjectAttribute(this);//proxy: fix
}

LIST_ATTRIBUTE::~LIST_ATTRIBUTE()
{
  OBJECT_ATTRIBUTE *ObjectAttribute;
  while ((ObjectAttribute = (OBJECT_ATTRIBUTE *)list.pop_top())) {
    delete ObjectAttribute;
  }
}

void LIST_ATTRIBUTE::Copy(BASE_ATTRIBUTE *a) {
  LIST_ATTRIBUTE *La = (LIST_ATTRIBUTE *)a;
  OBJECT_ATTRIBUTE *Oa = (OBJECT_ATTRIBUTE *)La->get_top();

  while (Oa) {
    //OBJECT_ATTRIBUTE *ObjectAttribute = (OBJECT_ATTRIBUTE *) //bad lines
    //  ObjFact->CreateAttribute(OBJECT_ATTRIBUTE_TYPE, 1);

    OBJECT_ATTRIBUTE *ObjectAttribute = new OBJECT_ATTRIBUTE; // better line

    ObjectAttribute->Copy(Oa);
    push_bot(ObjectAttribute);
    ObjectAttribute->SetInListAttribute(this); //proxy: fix

    Oa = (OBJECT_ATTRIBUTE *)(++(*La));
  }

}

void BASE_DYNAMIC_ATTRIBUTE::Copy(BASE_ATTRIBUTE *a) {
  BASE_DYNAMIC_ATTRIBUTE *Da = (BASE_DYNAMIC_ATTRIBUTE *)a;
  int FreeId;
  SpDynItem *Di = (SpDynItem *)Da->get_top();

  while (Di) {
    FreeId = Di->get_freeid();
    SpDynItem *DynamicItem = (SpDynItem *)
      FreeDynamicAttributes->new_object(FreeId);
    memcpy((char *)DynamicItem,
	   (char *)Di,
	   FreeDynamicAttributes->get_size(FreeId));
    push_bot(DynamicItem);
    Di = (SpDynItem *)(++(*Da));
  }
}

void STRING_ATTRIBUTE::outstream(ostream &out, int) {
  char *val = attribute;
  if (val){
    out << val;
  }
  else{
    out<<"(null)";
  }
}

int STRING_ATTRIBUTE::GetMessageSize(char *value) {
  int s = sizeof(STRING_ATTRIBUTE_MESSAGE_INFO)
    + (value == NULL
       ? 0
       : strlen(value) + 1);

  // What if a double follows this?  If that can happen then maybe
  // we should change int to double in this formula.
  return SpAddPaddingToAlign(s, sizeof(int));
}

void STRING_ATTRIBUTE::EmptyMessage(const char *buff) {
  STRING_ATTRIBUTE_MESSAGE_INFO *info =
    (STRING_ATTRIBUTE_MESSAGE_INFO *)buff;
  attribute = info->GetValue();
}
