#include "SpPolyMotion.H"
#include "SpMatrixEqu.H"

void POLY_N_MOTION::AddPoints(double ts, double te,
			      double t, double v0, double v1, double v2,
			      double err2, ...) {
  va_list va;
  int i,j;
  double Xi0, Xi1, Xi2, Ti, Si;
  BIG_FLOAT C0[2*N_MAX_COEFF_POLY_MOTION];
  BIG_FLOAT C1[2*N_MAX_COEFF_POLY_MOTION];
  BIG_FLOAT C2[2*N_MAX_COEFF_POLY_MOTION];

  StartTime = ts;
  EndTime = te;

  int n = GetDim();

  if (A0 == NULL) {

    A0 = new NET_DOUBLE *[n];
    A1 = new NET_DOUBLE *[n];
    A2 = new NET_DOUBLE *[n];

    for (i=0; i<n; i++) {
      A0[i] = new NET_DOUBLE[n];
      A1[i] = new NET_DOUBLE[n];
      A2[i] = new NET_DOUBLE[n];
    }

    B0 = new NET_DOUBLE[n];
    B1 = new NET_DOUBLE[n];
    B2 = new NET_DOUBLE[n];

    for (i=0; i<n; i++) {
      B0[i] = 0.0;
      B1[i] = 0.0;
      B2[i] = 0.0;
      for (j=0; j<n; j++) {
	A0[i][j] = 0.0;
	A1[i][j] = 0.0;
	A2[i][j] = 0.0;
      }
    }

  }

  va_start(va, err2);

  Xi0 = v0;
  Xi1 = v1;
  Xi2 = v2;

  Ti = t;
  Si = err2;

  for (;;) {

    Ti -= StartTime;

    C0[0] = 1.0/Si;
    for (i=1; i<2*n; i++) {
      C0[i] = C0[i-1]*Ti;
    }
    for (i=0; i<n; i++) {
      B0[i] += Xi0*C0[i];
      for (j=0; j<n; j++) {
	A0[i][j] += C0[i+j];
      }
    }

    C1[0] = 1.0/Si;
    for (i=1; i<2*n; i++) {
      C1[i] = C1[i-1]*Ti;
    }
    for (i=0; i<n; i++) {
      B1[i] += Xi1*C1[i];
      for (j=0; j<n; j++) {
	A1[i][j] += C1[i+j];
      }
    }

    C2[0] = 1.0/Si;
    for (i=1; i<2*n; i++) {
      C2[i] = C2[i-1]*Ti;
    }
    for (i=0; i<n; i++) {
      B2[i] += Xi2*C2[i];
      for (j=0; j<n; j++) {
	A2[i][j] += C2[i+j];
      }
    }

    Ti = va_arg(va,double);
    if (Ti == END_POLY_MOTION) break;

    Xi0 = va_arg(va,double);
    Xi1 = va_arg(va,double);
    Xi2 = va_arg(va,double);

    Si = va_arg(va,double);

  }

  va_end(va);

}

