#ifndef DynSpline3Motion_Objects
#define DynSpline3Motion_Objects

#include "SpStdarg.H"
#include "SpMath.H"

#include "SpDynItem.H"
#include "SpNetTypes.H"

class SPLINE3_MOTION : public DYNAMIC_POSITION_ITEM {
  private:
    NET_DOUBLE X0[3];
    NET_DOUBLE V0[3];
    NET_DOUBLE a[3];
    NET_DOUBLE b[3];

  protected:

  public:

    void init(double t0, double x0[3], double v0[3],
	      double t1, double x1[3], double v1[3]) {
      double dt1 = t1-t0;
      double dt2 = dt1*dt1;
      double dt3 = dt2*dt1;

      StartTime = t0;
      EndTime = t1;
      for (int i=0; i<3; i++) {
        X0[i] = x0[i];
	V0[i] = v0[i];
	a[i] = -(v1[i]+2*v0[i])/dt1 + 3.0*(x1[i]-x0[i])/dt2;
	b[i] = (v1[i]+v0[i])/dt2 - 2.0*(x1[i]-x0[i])/dt3;
      }

    }

    virtual void operator () (double t, int PT, double X[3]) {
      double dt1 = t-StartTime;
      double dt2 = dt1*dt1;
      double dt3 = dt2*dt1;
      //double V[3];
      for (int i=0; i<3; i++) {
	X[i] = X0[i] + V0[i]*dt1 + a[i]*dt2 + b[i]*dt3;
      }
      Convert(X, PositionType, X, PT, t);
    }

    virtual void operator () (double t, int PT, double X[3], double V[3]) {
      double dt1 = t-StartTime;
      double dt2 = dt1*dt1;
      double dt3 = dt2*dt1;
      for (int i=0; i<3; i++) {
	X[i] = X0[i] + V0[i]*dt1 + a[i]*dt2 + b[i]*dt3;
	V[i] = V0[i] + 2.0*a[i]*dt1 + 3.0*b[i]*dt2;
      }
      Convert(X, V, PositionType, X, V, PT, t);
    }

    virtual void operator () (double t, int PT, double X[3], double V[3],
			      double A[3]) {
      double dt1 = t-StartTime;
      double dt2 = dt1*dt1;
      double dt3 = dt2*dt1;
      for (int i=0; i<3; i++) {
	X[i] = X0[i] + V0[i]*dt1 + a[i]*dt2 + b[i]*dt3;
	V[i] = V0[i] + 2.0*a[i]*dt1 + 3.0*b[i]*dt2;
	A[i] = 2.0*a[i] + 6.0*b[i]*dt1;
      }
      Convert(X, V, A, PositionType, X, V, A, PT, t);
    }

    virtual char *GetType() {return "Dynamic Spline3 Motion Type";}
};

#endif
