#ifndef DynSpline6Motion_Objects
#define DynSpline6Motion_Objects

#include "SpStdarg.H"
#include "SpMath.H"

#include "SpDynItem.H"
#include "SpNetTypes.H"

class SPLINE6_MOTION : public DYNAMIC_POSITION_ITEM {
  private:
    NET_DOUBLE X0[3];
    NET_DOUBLE V0[3];
    NET_DOUBLE A0[3];
    NET_DOUBLE a[3];
    NET_DOUBLE b[3];
    NET_DOUBLE c[3];

  protected:

  public:

    void init(double t0, double x0[3], double v0[3], double a0[3],
	      double t1, double x1[3], double v1[3], double a1[3]) {
      double alpha[3];
      double beta[3];
      double gamma[3];

      double dt1 = t1-t0;
      double dt2 = dt1*dt1;
      double dt3 = dt2*dt1;

      StartTime = t0;
      EndTime = t1;
      for (int i=0; i<3; i++) {
        X0[i] = x0[i];
	V0[i] = v0[i];
	A0[i] = a0[i];

	alpha[i] = x1[i] - X0[i] - V0[i]*dt1 - A0[i]*dt2/2.0;
	beta[i] = v1[i] - V0[i] - A0[i]*dt1;
	gamma[i] = a1[i] - A0[i];

	a[i] = (alpha[i]*10.0 - beta[i]*4.0*dt1 + gamma[i]*dt2/2.0) / dt3;
	b[i] = (-alpha[i]*15.0/dt1 + beta[i]*7.0 - gamma[i]*dt1) / dt3;
	c[i] = (alpha[i]*6.0/dt2 - beta[i]*3.0/dt1 + gamma[i]/2.0) / dt3;
      }

    }

    virtual void operator () (double t, int PT, double X[3]) {
      double dt1 = t-StartTime;
      double dt2 = dt1*dt1;
      double dt3 = dt2*dt1;
      double dt4 = dt3*dt1;
      double dt5 = dt4*dt1;

      for (int i=0; i<3; i++) {
	X[i] = X0[i] + V0[i]*dt1 + A0[i]*dt2/2.0 + a[i]*dt3 + b[i]*dt4
	  + c[i]*dt5;
      }
      Convert(X, PositionType, X, PT, t);
    }

    virtual void operator () (double t, int PT, double X[3], double V[3]) {
      double dt1 = t-StartTime;
      double dt2 = dt1*dt1;
      double dt3 = dt2*dt1;
      double dt4 = dt3*dt1;
      double dt5 = dt4*dt1;

      for (int i=0; i<3; i++) {
	X[i] = X0[i] + V0[i]*dt1 + A0[i]*dt2/2.0 + a[i]*dt3 + b[i]*dt4
	  + c[i]*dt5;
	V[i] = V0[i] + A0[i]*dt1 + 3.0*a[i]*dt2 + 4.0*b[i]*dt3 + 5.0*c[i]*dt4;
      }
      Convert(X, V, PositionType, X, V, PT, t);
    }

    virtual void operator () (double t, int PT, double X[3], double V[3],
			      double A[3]) {
      double dt1 = t-StartTime;
      double dt2 = dt1*dt1;
      double dt3 = dt2*dt1;
      double dt4 = dt3*dt1;
      double dt5 = dt4*dt1;

      for (int i=0; i<3; i++) {
	X[i] = X0[i] + V0[i]*dt1 + A0[i]*dt2/2.0 + a[i]*dt3 + b[i]*dt4
	  + c[i]*dt5;
	V[i] = V0[i] + A0[i]*dt1 + 3.0*a[i]*dt2 + 4.0*b[i]*dt3 + 5.0*c[i]*dt4;
	A[i] = A0[i] + 6.0*a[i]*dt1 + 12.0*b[i]*dt2 + 20.0*c[i]*dt3;
      }
      Convert(X, V, A, PositionType, X, V, A, PT, t);
    }

    virtual char *GetType() {return "Dynamic Spline6 Motion Type";}
};

#endif
