// BroadcastProxy.C event object file

#include "E_SpBroadcastProxy.H"
#include "M_SpDeliverProxy.H"
#include "S_SpHLA.H"
#include "SpScheduleEvent.H"
#include "O_SpLocalProxyMgr.H"

#include "RB_SpFrameworkFuncs.H"

#ifdef TODO
EVENT_TYPE (DELIVER_PROXY);
#else
#include "E_SpDeliverProxy.H"
#endif

extern SpObjFactory *ObjFact;

/************************************************************************
* E_SpBroadcastProxy : constructor for the BroadcastProxy event		*
************************************************************************/
E_SpBroadcastProxy::E_SpBroadcastProxy() {
  set_lazy();
}

/************************************************************************
* process : optimistically process event				*
************************************************************************/
void E_SpBroadcastProxy::process() {
  char *Buff;
  int i;
  SpDestination *Destination;
  SpDestination Destination1(N_NODES);

/*
** Convert the SIMOBJ to an HLA object and then get the LOCAL_PROXYMAN object
** manager
*/

  S_SpHLA *HLAobj = (S_SpHLA *)SIMOBJ;
  O_SpLocalProxyMgr *LocalProxyMan =
    (O_SpLocalProxyMgr *)get_manager(LOCAL_PROXYMAN);

/*
** Set the destination distribution to all nodes
*/

  for (i=0; i<N_NODES; i++) Destination1.Set(i);
  Destination = HLAobj->GetDestination();
  RB_memcpy(	Destination->GetDestination(),
		Destination1.GetDestination(),
		Destination1.GetNbytes()	);

/*
** Make a message out of the object proxy and deliver it to each node
*/

  SpObjProxy *ObjectProxy = HLAobj->GetObjectProxy();

  SpObjProxyMsg *ObjectProxyMessage = 
    ObjFact->GenerateMessage(ObjectProxy);

  int NumDataBytes = ObjectProxyMessage->GetNbytes()
    + sizeof(SpObjProxyMsg);

  for (i=0; i<N_NODES; i++) {

    M_SpDeliverProxy *DeliverProxyMessage = (M_SpDeliverProxy *)
    schedule(SpSimTime(time_tag).IncrementPriority2(),
	     DELIVER_PROXY,
	     LOCAL_PROXYMAN,
	     LOCAL_ID % LocalProxyMan->get_N_LOC(),
	     i,
	     NumDataBytes,
	     Buff);

    DeliverProxyMessage->Broadcast = 1;

    if (Buff-(char *)&DeliverProxyMessage[1] != 0) {
      cout << "MAJOR LEAGE ERROR IN BROADCAST_PROXY" << endl;
    }

    memcpy(Buff, (char *)ObjectProxyMessage,
	   NumDataBytes - sizeof(SpObjProxyMsg));
  }

  delete [] ((char*) ObjectProxyMessage);

}
