// DeliverProxy.C event object file

#include "E_SpDeliverProxy.H"
#include "M_SpDeliverProxyPtr.H"
#include "S_SpLocalProxyMgr.H"

#include "O_SpLocalProxyMgr.H"

#include "SpObjFactory.H"
#include "RB_SpObjProxyFuncs.H"
#include "RB_SpContainerFuncs.H"

#include "F_SpProxyListItem.H"
#include "F_SpSimObjRef.H"
#include "SpScheduleEvent.H"
#include "RB_SpFrameworkFuncs.H"

#ifdef TODO
EVENT_TYPE (DELIVER_PROXY_PTR);
#else
#include "E_SpDeliverProxyPtr.H"
#endif

extern SpObjFactory *ObjFact;

#include "SpFreeHLA.H"
extern SpFreeHLA *FreeHLA;

/************************************************************************
* E_SpDeliverProxy : constructor for the DeliverProxy event		*
************************************************************************/
E_SpDeliverProxy::E_SpDeliverProxy() {
  set_lazy();
}

/************************************************************************
* process : optimistically process event				*
************************************************************************/
void E_SpDeliverProxy::process() {
  M_SpDeliverProxy *DeliverProxyMessage = (M_SpDeliverProxy *)MESSAGE;
  S_SpLocalProxyMgr *LocalProxyMan = (S_SpLocalProxyMgr *)SIMOBJ;

/*
  char s[256];
  double t = TIME_TAG;
  sprintf(s, "DELIVER_PROXY at time %f\n", t);
  RB_PRINT(stdout, s);
*/

/*
** Get the object proxy message from the input message and generate an
** object proxy
*/

  SpObjProxyMsg *ObjectProxyMessage = 
    (SpObjProxyMsg *)&DeliverProxyMessage[1];

  SpObjProxy *ObjectProxy = RB_GenerateProxy(ObjectProxyMessage);

/*
  cout << "Checking Proxy in DeliverProxy" << endl;
  int status = ObjectProxy->CheckProxy();
  cout << "Done Checking Proxy in DeliverProxy: status = " << status << endl;
*/

/*
** Put the object proxy in the LocalProxyMan's list of object proxies
*/

  C_XQUEUE *ProxyList = LocalProxyMan->GetProxyList();

  F_SpProxyListItem *ProxyListItem =
    (F_SpProxyListItem *)RB_FREE_NEW(FreeHLA, PROXY_LIST_ITEM);

  ProxyListItem->SetObjectProxy(ObjectProxy);
  ProxyListItem->SetGlobalId(ObjectProxy->GetGlobalId());

  RB_INSERT(ProxyList, ProxyListItem);

/*
** Now make the ProxyList mean the list of entities receiving the proxy
*/

  ProxyList = ProxyListItem->GetProxyList();

/*
** If we are broadcasting the proxy to everybody, then find out who needs the
** proxy and then schedule an event to give them the proxy
*/

  if (DeliverProxyMessage->Broadcast) {
    char *Name;

    O_SpLocalProxyMgr *o_LocalProxyMan =
      (O_SpLocalProxyMgr *)get_manager(LOCAL_PROXYMAN);

    SpDataParser *Parser = o_LocalProxyMan->GetProxyManParser();
//    Parser->print(cout);

    char *b;
    Parser->GoTo("parameters", NULL);
    Parser->GetFirstDefine(b, Name);

    while (Name) {
      int Otype = object_type(Name);
      SpSimObjMgr *ObjMan = (SpSimObjMgr *)get_manager(Otype);
      for (int i=0; i<ObjMan->get_N_LOC(); i++) {

/*
	cout << "BROADCAST OBJECT TYPE: " << Name 
	     << " LOCAL_ID " << i << endl;
*/

	if (	(ObjectProxy->GetObjectType() != Otype) || 
		(ObjectProxy->GetLocalId() != i) ||
		(ObjectProxy->GetNode() != NODE)	) {

	  F_SpSimObjRef *SimObjRef = (F_SpSimObjRef *)
	    RB_FREE_NEW(FreeHLA, SIMOBJ_REF);

	  SimObjRef->SetNode(NODE);
	  SimObjRef->SetType(Otype);
	  SimObjRef->SetLocalID(i);
	  SimObjRef->SetGlobalID(-1);

	  RB_INSERT(ProxyList, SimObjRef);

	  M_SpDeliverProxyPtr *DeliverProxyPtrMessage = 
	    (M_SpDeliverProxyPtr *)schedule(SpSimTime(time_tag).IncrementPriority2(),
					  DELIVER_PROXY_PTR,
					  Otype,
					  i,
					  NODE);

	  DeliverProxyPtrMessage->ObjectProxy = ObjectProxy;
	  DeliverProxyPtrMessage->SpaceId     = -1;
	  DeliverProxyPtrMessage->NumGrid     =  0;

	}

      }

      Parser->GetNextDefine(b, Name);

    }
  }

}
