// E_SpDeliverProxyPtrSpace.C 

#include "E_SpDeliverProxyPtrSpace.H"
#include "S_SpHLA.H"
#include "SpSubscribeSpace.H"
#include "F_SpProxyItem.H"
#include "H_SpAddProxySpace.H"


/************************************************************************
* E_SpDeliverProxyPtrSpace : constructor for the DeliverProxyPointer 	*
************************************************************************/
E_SpDeliverProxyPtrSpace::E_SpDeliverProxyPtrSpace()
  : valid(0) {
//  set_lazy();
}

/************************************************************************
* process : optimistically process event				*
************************************************************************/

int E_SpDeliverProxyPtrSpace::lazy() {
  return (0);
}

void E_SpDeliverProxyPtrSpace::process() {
  S_SpHLA *hla                  = (S_SpHLA *) SIMOBJ;
  M_SpDeliverProxyPtrSpace *msg = (M_SpDeliverProxyPtrSpace *) get_MESSAGE();
  char *data                    = (char *) get_DATA();
  C_XQUEUE *remote_proxies      = hla->GetRemoteObjectProxies(); 
  int num_proxy                 = remote_proxies->get_length();
  F_PROXY_ITEM *proxy_item, *add_proxy_item;
  int buffer_len, len;
  char *buffer, *cp;

  add_proxy_item = (F_PROXY_ITEM *) 0;
  proxy_item     = (F_PROXY_ITEM *) remote_proxies->get_top();
  for (int i=0; i<num_proxy; i++){
    if (proxy_item->GetGlobalId()==msg->ProxyGlobalId){
      add_proxy_item = proxy_item;
      break;
    }
    proxy_item = (F_PROXY_ITEM *) proxy_item->get_link();
  }

  valid = (add_proxy_item) ? 1 : 0;
  errcode = 1;

  if (valid){
    SpSubscribeSpaceAddProxy add_proxy(msg->SpaceId, add_proxy_item, msg->NumGrid);
    buffer_len = sizeof(SpSubscribeSpaceAddProxy) + (msg->NumGrid * sizeof(int));
    cp = buffer = new char [buffer_len];
    len = sizeof(SpSubscribeSpaceAddProxy); memcpy(cp, &add_proxy, len); cp += len;
    len = msg->NumGrid * sizeof(int);       memcpy(cp, data, len); 
    hla->ProcessEventHandlers("SubscribeSpace Add Proxy Ptr",
                              TIME_TAG,
                              (char *) buffer, 
                              buffer_len,
                              -1);
    SpSubscribeSpaceAddProxy *added_proxy = (SpSubscribeSpaceAddProxy *) buffer;
    valid &= (added_proxy->GetNumGridAdd()==msg->NumGrid);
    errcode = 2;
    delete [] buffer;
  }
}

void 
E_SpDeliverProxyPtrSpace::commit() {
  S_SpHLA *hla                  = (S_SpHLA *) SIMOBJ;
  M_SpDeliverProxyPtrSpace *msg = (M_SpDeliverProxyPtrSpace *) get_MESSAGE();
  int *grid_id                  = (int *) get_DATA();

  if (!valid){
    cout << time_tag << " ERROR::E_SpDeliverProxyPtrSpace event processed invalid code " << errcode
         << " simobj " << hla->get_GLOBAL_ID() << " proxy " << msg->ProxyGlobalId 
         << " num grid " << msg->NumGrid << " grids ";
    for (int i=0; i<msg->NumGrid; i++)
      cout << grid_id[i] << ", ";
    cout << endl;  
    abort();
  }
}
