// E_ReflectAttributes.C method file

// I moved the following #include line to before all other #includes
// so that it would compile on the HP Convex/Exemplar computer named
// dax.cmf.nrl.navy.mil.  The old order caused the HP compiler to say
// "signal 11" with no other explanation of what could be wrong, which
// is why I suspect it is a bug in the HP compiler.  There were about
// a total of 38 files in which I had to change the order of includes
// to get it to compile.
#include "O_SpFedGhost.H"
#include "E_SpReflectAttributes.H"
#include "S_SpLocalProxyMgr.H"
#include "F_SpProxyListItem.H"
#include "F_SpSimObjRef.H"
#include "SpExportAttribute.H"

#include "RB_SpObjProxyFuncs.H"
#include "M_SpTouchProxy.H"

#include "SpObjFactory.H"
extern SpObjFactory *ObjFact;

#include "SpDynItem.H"
#include "SpFreeDynAttributes.H"
extern SpFreeDynAttributes *FreeDynamicAttributes;
#include "SpScheduleEvent.H"

#ifdef TODO
EVENT_TYPE (TOUCH_PROXY)
#else
#include "E_SpTouchProxy.H"
#endif

#include "E_SpReflectAttributesData.H"

/************************************************************************
* REFLECT_ATTRIBUTES : Construct a REFLECT_ATTRIBUTES event object     	*
************************************************************************/
E_SpReflectAttributes::E_SpReflectAttributes() {}

/************************************************************************
* process : optimistically process event				*
************************************************************************/
void E_SpReflectAttributes::process() {
  int i;
  int len;
  int Natts;
  int Nbytes;
  SpObjProxy *ObjectProxy;
  SpObjProxy *Op;
  ATTRIBUTE_MESSAGE_INFO *Info;
  const char *Buff;
  PROXY_LEVEL_INFO *ProxyLevelInfo;
  int Level;
  SpObjProxy *OldObjectProxy;
  int OpWarning = 0;
  M_SpTouchProxy TouchProxyMessage;
  M_SpReflectAttributesData ReflectAttrDataMessage;

/*
** Get the reference to the LocalProxyMan and the input message
*/

  S_SpLocalProxyMgr *LocalProxyMan = (S_SpLocalProxyMgr *)SIMOBJ;

  M_SpReflectAttributes *ReflectAttributesMessage =
    (M_SpReflectAttributes *)MESSAGE;

/*
** Get the proxy list item for the referenced object and then get its
** proxy list
*/

  C_XQUEUE *ProxyList = LocalProxyMan->GetProxyList();

/*
  cout << "Reflecting Attributes for object "
       << ReflectAttributesMessage->GlobalId
       << endl;
*/

  F_SpProxyListItem *ProxyListItem = (F_SpProxyListItem *)
    ProxyList->find(ReflectAttributesMessage->GlobalId);

  if (ProxyListItem == NULL) {
    char s[120];
    sprintf(s,"%s%d%s%d\n",
	    "Error (REFLECT_ATTRIBUTES): Could not find ProxyListItem ",
	    int(ReflectAttributesMessage->GlobalId), //pxy fix: int cast
	    " in ProxyList of length ",
	    int(ProxyList->get_length()) );
    RB_cout << s << endl;
    return;
  }

/*
** SPEEDES has a rule that says that events must act on or modify
** the state of only one simobj.  That simobj is the one for which the
** event was scheduled.  The proxy code violates the rule of acting on
** only one simobj.  If the proxy code were rewritten to NOT violate
** that rule, then it would probably work by scheduling an event for
** all S_SpHLA simobjs that use the updated proxy.  The message for the
** event would contain the new proxy.  But because that message
** contains the proxy, it is a big message, which means it is
** inefficient.  Therefore the S_SpHLA simobjs do not have their own
** copy of the proxy.  Instead they all have pointers that point
** to the same proxy.  This way we do not have to send them a new
** proxy.  We just change the one proxy to which they all point.  But
** when we do that we are changing the state of those multiple S_SpHLA
** simobjs.  This violates the SPEEDES rule of changing only one
** simobj.  This rule violation means that the following problem can
** occur:  Suppose at time 2, we change the proxy to which the S_SpHLA
** simobjs point.  Then an event from another node comes in to act on
** one of those S_SpHLA simobjs at time 1.  SPEEDES will roll back only
** events that were processed on this simobj.  SPEEDES will NOT roll
** back events (like this E_SpReflectAttributes event) processed for
** other simobjs.  And it this E_SpReflectAttributes event that made
** the change to the proxy to which the rolled back S_SpHLA points.
** Therefore the change to the proxy will NOT be rolled back.  To
** solve this problem, we need to schedule an event on the rolled-back
** S_SpHLA.  Such an event WOULD be rolled back in the example I'm
** using.  This event will give a new pointer to the S_SpHLA.  This new
** pointer points to a proxy that contains the updates.  To make this
** new proxy we clone the old one and modify the clone.
**
** Making an ObjectProxyMessage is Not the most efficient way to make
** a copy of the old Object Proxy, but it is reliable. The RB_Clone
** should be must faster - and is also reliable.
*/

  OldObjectProxy = ProxyListItem->GetObjectProxy();

  if (0) {
    SpObjProxyMsg *ObjectProxyMessage;
    ObjectProxyMessage = ObjFact->GenerateMessage(OldObjectProxy);
    ObjectProxy = RB_GenerateProxy(ObjectProxyMessage);
    delete [] ((char*) ObjectProxyMessage);
  }else{
    ObjectProxy = RB_Clone(OldObjectProxy);
  }

  RB_DestroyProxy(OldObjectProxy);
  ProxyListItem->SetObjectProxy(ObjectProxy);

/*
  cout << "Checking Proxy in ReflectAttriutes" << endl;
  int status = ObjectProxy->CheckProxy();
  cout << "Done Checking Proxy in ReflectAttributes: status = "
       << status << endl;
*/

/*
** Go through the message and update the attributes for the Object Proxy
*/

  Natts = ReflectAttributesMessage->Natts;
  Nbytes = ReflectAttributesMessage->bytes;
  Buff = DATA; //(char *)&ReflectAttributesMessage[1];

//  cout << "Number of reflected attributes = " << Natts << endl;

  for (i=0; i<Natts; i++) {
    SpObjProxy::EMPTY_MESSAGE_RESULT r = ObjectProxy->EmptyMessage(Buff);
    Buff = r.Buff;
    if (r.OpWarning)
      OpWarning = 1;
  }

  if (OpWarning) {

    Buff = DATA; //(char *)&ReflectAttributesMessage[1];
    for (i=0; i<Natts; i++) {

      ProxyLevelInfo = (PROXY_LEVEL_INFO *)Buff;
      Level = ProxyLevelInfo->Level;

      Buff += Level*sizeof(PROXY_LEVEL_INFO);
      Info = (ATTRIBUTE_MESSAGE_INFO *)Buff;

      // Jacob commented these variables out since they are never read
      // after this set:
      //AttributeIndex = Info->GetIndex();
      //ArrayIndex = Info->GetArrayIndex();

      Op = ObjectProxy->FindProxy(ProxyLevelInfo);

      if (Op == NULL) {
//#define ERROR_CHECK1
#ifdef ERROR_CHECK1
	cout << "Warning (ReflectAttributes): OpWarning set" << endl;
#endif
      }

      Buff += Info->GetNbytes();
    }

  }

  if (int(Buff-get_DATA()/*(char *)&ReflectAttributesMessage[1]*/) != Nbytes) {
    cout << "Major League Error in ReflectAttributes" << endl;
  }

/*
** Schedule a Touch event for every local object in the list
*/

  ProxyList = ProxyListItem->GetProxyList();
  len = ProxyList->get_length();
//  cout << "ProxyList length = " << len << endl;

  F_SpSimObjRef *SimObjRef = (F_SpSimObjRef *)ProxyList->get_top();
  for (i=0; i<len; i++) {

    TouchProxyMessage.ObjectProxy = ObjectProxy;
    TouchProxyMessage.globalId = ReflectAttributesMessage->GlobalId;

    ScheduleEvent(SpSimTime(time_tag).IncrementPriority2(),
		  TOUCH_PROXY,
		  SimObjRef->GetType(),
		  SimObjRef->GetLocalID(),
		  NODE,
		  TouchProxyMessage);

    /* If the target is a FedGhost, we need to ship the changed attrs
     * to it as well as the TouchProxy message.
     */
    if (SimObjRef->GetType() == SpFedGhost) {

      ReflectAttrDataMessage.GlobalId = ReflectAttributesMessage->GlobalId;

      ScheduleEvent(SpSimTime(time_tag).IncrementPriority2(),
		    REFLECT_ATTRIBUTES_DATA,
		    SimObjRef->GetType(),
		    SimObjRef->GetLocalID(),
		    NODE,
		    ReflectAttrDataMessage,
		    DATA,
		    Nbytes);
    }

    SimObjRef = (F_SpSimObjRef *)SimObjRef->get_link();

  }

}
