// E_SpRemoveProxy.C

#include "M_SpNewProxy.H"
#include "E_SpRemoveProxy.H"
#include "M_SpUndeliverProxyPtr.H"
#include "SpScheduleEvent.H"
#include "S_SpLocalProxyMgr.H"

#include "F_SpProxyListItem.H"
#include "F_SpSimObjRef.H"
#include "RB_SpObjProxyFuncs.H"
#include "RB_SpContainerFuncs.H"
#include "F_SpProxyItem.H"
#include "xqueue.H"

#include "SpFreeHLA.H"
extern SpFreeHLA *FreeHLA;

#ifdef TODO
EVENT_TYPE (UNDELIVER_PROXY_PTR)
EVENT_TYPE (UndeliverProxyPtrSpace)
#else
#include "E_SpUndeliverProxyPtr.H"
#include "E_SpUndeliverProxyPtrSpace.H"
#endif


/************************************************************************
* E_SpRemoveProxy : constructor for the NEW_PROXY event	*
************************************************************************/
E_SpRemoveProxy::E_SpRemoveProxy()
  : valid(0), errcode(0) {
//  set_lazy();
}

/************************************************************************
* process : optimistically process event				*
************************************************************************/

int E_SpRemoveProxy::lazy() {
  M_SpRemoveProxy *msg = (M_SpRemoveProxy *) get_MESSAGE();
  S_SpLocalProxyMgr *local_proxy_man = (S_SpLocalProxyMgr *) SIMOBJ;
  RB_cout << time_tag << " E_SpRemoveProxy lazy proxy " << msg->ProxyGlobalId 
       << " node "      << local_proxy_man->get_NODE()
       << " remove "    << msg->NumRemoveDest
       << " undeliver " << msg->NumUndeliverDest << endl;
  return (0);
}

void E_SpRemoveProxy::process() {
  S_SpLocalProxyMgr *local_proxy_man = (S_SpLocalProxyMgr *) SIMOBJ;
  char *data           = (char *)            get_DATA();
  M_SpRemoveProxy *msg = (M_SpRemoveProxy *) get_MESSAGE();
  F_SpProxyListItem *proxy_list_item;
  C_XQUEUE *proxy_list;
  SpProxyDestination *destination;
  int ok = 1;

  // data storage : [remove dest] + [undeliver dest]

  proxy_list      = local_proxy_man->GetProxyList();
  proxy_list_item = (F_SpProxyListItem *) proxy_list->find(msg->ProxyGlobalId);
  ok &= (proxy_list_item) ? 1 : 0;

  if (!ok) errcode = 1;

//float look_ahead = msg->LookAhead;
  float look_ahead = 0.0;

  destination = (SpProxyDestination *) 
                (data + (sizeof(SpProxyDestination)*msg->NumRemoveDest));
  if (ok){
    ok &= ProcessUndeliver(proxy_list_item, msg->SpaceId, look_ahead,
		msg->NumUndeliverDest, destination);
    if (!ok) errcode = 2;
  }

  destination = (SpProxyDestination *) data;

  if (ok){
    ok &= ProcessRemove(proxy_list, proxy_list_item, msg->SpaceId, look_ahead,
		msg->NumRemoveDest, destination);
    if (!ok) errcode = 3;
  }

  if (ok) errcode = 0;

  valid = ok;

////////////////////////////
/*
    int i;
    cout << time_tag << " E_SpRemoveProxy proxy " << msg->ProxyGlobalId;

    destination = (SpProxyDestination *) 
                  (data + (sizeof(SpProxyDestination)*msg->NumRemoveDest));
    cout << " undeliver : ";
    for (i=0; i<msg->NumUndeliverDest; i++)
      cout << destination[i].GlobalId << "(" << destination[i].GridId << ")" << ", ";

    cout << " remove : ";
    destination = (SpProxyDestination *) data;
    for (i=0; i<msg->NumRemoveDest; i++)
      cout << destination[i].GlobalId << "(" << destination[i].GridId << ")" << ", ";
    cout << " valid " << valid << " errcode " << errcode;
    cout << endl;
*/
//////////////////////////////

}

int 
E_SpRemoveProxy::ProcessRemove(C_XQUEUE *proxy_list, F_SpProxyListItem *proxy_list_item, int space_id,
                               float look_ahead,
                               int num_destination, SpProxyDestination *destination) {
  C_XQUEUE *local_proxy_list;
  F_SpSimObjRef **subscriber_ref;
  int i;
  int ok = 1;
  
  if (num_destination<=0) return (ok);

  subscriber_ref = (num_destination>0) ? new F_SpSimObjRef * [num_destination] 
                                       : (F_SpSimObjRef **) 0;

  local_proxy_list = proxy_list_item->GetProxyList();
  for (i=0; i<num_destination && ok; i++){
    subscriber_ref[i] = (F_SpSimObjRef *) local_proxy_list->find(destination[i].GlobalId);
    ok &= (subscriber_ref[i]!=(F_SpSimObjRef*)0);
  }

  SpSimTime event_time(time_tag);
  event_time += look_ahead;

#define  DISABLE_AUTO_TIE_BREAKING
#ifdef   DISABLE_AUTO_TIE_BREAKING
event_time.DisableAutoTieBreaking();
event_time.IncrementPriority2(2);
#endif

  local_proxy_list = proxy_list_item->GetProxyList();
  for (i=0; i<num_destination && ok; i++){
    // notify local subscriber
    int grid_id = destination[i].GridId;
    M_SpUndeliverProxyPtr undeliver_msg(space_id, proxy_list_item->GetGlobalId(), 1);
    ScheduleEvent(event_time, UNDELIVER_PROXY_PTR,
                  subscriber_ref[i]->GetType(), 
                  subscriber_ref[i]->GetLocalID(), 
                  subscriber_ref[i]->GetNode(),
                  undeliver_msg, (const char *) &grid_id, (undeliver_msg.NumGrid * sizeof(int))); 
    RB_REMOVE(local_proxy_list, subscriber_ref[i]->GetGlobalID());
    RB_FREE_DELETE(FreeHLA, subscriber_ref[i]);
  }

  if (subscriber_ref) delete [] subscriber_ref;

  if (local_proxy_list->get_length()<=0){
    RB_REMOVE(proxy_list, proxy_list_item->GetGlobalId());
    RB_DestroyProxy(proxy_list_item->GetObjectProxy());
    RB_FREE_DELETE(FreeHLA, proxy_list_item);
  } 

  return (ok);
}

int 
E_SpRemoveProxy::ProcessUndeliver(F_SpProxyListItem *proxy_list_item, int space_id,
                                  float look_ahead,
                                  int num_destination, SpProxyDestination *destination) {
  C_XQUEUE *local_proxy_list;
  F_SpSimObjRef *subscriber_ref;
  SpBinaryTree subscribers, *subscriber;
  SpProxyDestination *local_dest;
  int *grid, num_grid;
  int ok = 1;
  int i;

  if (num_destination<=0) return (ok);

  local_proxy_list = proxy_list_item->GetProxyList();
  for (i=0; i<num_destination && ok; i++){
    subscriber_ref = (F_SpSimObjRef *) local_proxy_list->find(destination[i].GlobalId);
    ok &= (subscriber_ref!=(F_SpSimObjRef*)0);
  }

  for (i=0; i<num_destination && ok; i++){
    subscriber = (SpBinaryTree *) subscribers.Find(destination[i].GlobalId);
    if (!subscriber){
      subscriber = new SpBinaryTree();
      subscribers.Insert((void *) subscriber, destination[i].GlobalId);
    }
    subscriber->Insert((void *) (destination+i), destination[i].GridId);
  }

  SpSimTime event_time(time_tag);
  event_time += look_ahead;

#ifdef   DISABLE_AUTO_TIE_BREAKING
event_time.DisableAutoTieBreaking();
event_time.IncrementPriority2(1);
#endif

  subscriber = (SpBinaryTree *) subscribers.GetFirstElement();
  while (subscriber && ok){

    num_grid = subscriber->GetNumElements();
    grid     = (num_grid>0) ? new int [num_grid] : (int *) 0;
    num_grid = 0;

    local_dest = (SpProxyDestination *) subscriber->GetFirstElement();
    while (local_dest){
      grid[num_grid++] = local_dest->GridId;
      local_dest = (SpProxyDestination *) subscriber->GetNextElement();
    }

    local_dest = (SpProxyDestination *) subscriber->GetFirstElement();

    M_SpUndeliverProxyPtrSpace msg(space_id, proxy_list_item->GetGlobalId(), num_grid);
    ScheduleEvent(event_time, UndeliverProxyPtrSpace,
                  local_dest->ObjectType, local_dest->LocalId, local_dest->Node,
                  msg, (const char *) grid, (num_grid*sizeof(int)));

    if (grid) delete [] grid;

    subscriber = (SpBinaryTree *) subscribers.GetNextElement();
  }

  while (subscriber = (SpBinaryTree *) subscribers.RemoveFirstElement())
    delete subscriber;

  return (ok);
}

void E_SpRemoveProxy::commit() {
  char *data           = (char *)            get_DATA();
  M_SpRemoveProxy *msg = (M_SpRemoveProxy *) get_MESSAGE();
  SpProxyDestination *destination;
  int i;
  F_SpProxyListItem *proxy_list_item;
  C_XQUEUE *proxy_list;
  S_SpLocalProxyMgr *local_proxy_man = (S_SpLocalProxyMgr *) SIMOBJ;

  proxy_list      = local_proxy_man->GetProxyList();
  proxy_list_item = (F_SpProxyListItem *) proxy_list->find(msg->ProxyGlobalId);

  if (!valid){
    cout << time_tag << " ERROR::E_SpRemoveProxy event processed invalid code " << errcode
         << " proxy " << msg->ProxyGlobalId;

    destination = (SpProxyDestination *) 
                  (data + (sizeof(SpProxyDestination)*msg->NumRemoveDest));
    cout << " undeliver : ";
    for (i=0; i<msg->NumUndeliverDest; i++)
      cout << destination[i].GlobalId << "(" << destination[i].GridId << ")" << ", ";

    cout << " remove : ";
    destination = (SpProxyDestination *) data;
    for (i=0; i<msg->NumRemoveDest; i++)
      cout << destination[i].GlobalId << "(" << destination[i].GridId << ")" << ", ";

    C_XQUEUE *local_proxy_list = proxy_list_item->GetProxyList();
    cout << " remote : ";
    F_PROXY_ITEM *proxy_item = (F_PROXY_ITEM *) local_proxy_list->get_top();
    for (i=0; i<local_proxy_list->get_length(); i++){
      cout << proxy_item->GetGlobalId() << ", ";
      proxy_item = (F_PROXY_ITEM *) proxy_item->get_link();
    }
    cout << endl;

    abort();
  }
}
