// E_ScheduleUpdate.C method file

#include "E_SpScheduleUpdate.H"
#include "S_SpHLA.H"
#include "F_SpUpdateListItem.H"
#include "O_SpLocalProxyMgr.H"
#include "M_SpReflectAttributes.H"
#include "SpObjProxyMsg.H"
#include "SpScheduleEvent.H"

#include "RB_SpContainerFuncs.H"

#include "SpFreeHLA.H"
extern SpFreeHLA *FreeHLA;

#include "SpHostUser.H"
extern SpHostUser *HostUser;

#ifdef TODO
EVENT_TYPE (REFLECT_ATTRIBUTES);
#else
#include "E_SpReflectAttributes.H"
#endif

/************************************************************************
* SCHEDULE_UPDATE : Construct a SCHEDULE_UPDATE event object   		*
************************************************************************/
E_SpScheduleUpdate::E_SpScheduleUpdate() {}

/************************************************************************
* process : optimistically process event				*
************************************************************************/
void E_SpScheduleUpdate::process() {
  S_SpHLA *HLAobj = (S_SpHLA *)SIMOBJ;
  O_SpLocalProxyMgr *LocalProxyMan =
    (O_SpLocalProxyMgr *)get_manager(LOCAL_PROXYMAN);
  ExtSub = SIMOBJ->CheckSubscription("ObjectProxy");

  // replaced: int Natts = HLAobj->GetUpdateList()->get_length();
  RB_UpdateXqueue *attUpdateList = HLAobj->GetUpdateList(); //pxy
  int Natts = attUpdateList->get_length(); //pxy

  HLAobj->GetObjectProxy()->FillMessage
    ((F_UPDATE_LIST_ITEM*) attUpdateList->get_top(), //pxy
     Natts,
     Buffer,
     Nbytes);
  
/*
** Schedule events for all of the nodes that are in the Destination
*/

  M_SpReflectAttributes ReflectAttributesMessage;
  ReflectAttributesMessage.Natts = Natts;
  ReflectAttributesMessage.GlobalId = GLOBAL_ID;

  int Node = HLAobj->GetDestination()->GetFirstNode();
  while (Node != -1) {
    ScheduleEvent(SpSimTime(time_tag).IncrementPriority2(),
		  REFLECT_ATTRIBUTES,
		  LOCAL_PROXYMAN,
		  LOCAL_ID % LocalProxyMan->get_N_LOC(),
		  Node,
		  &ReflectAttributesMessage,
		  Buffer,
		  Nbytes);

    Node = HLAobj->GetDestination()->GetNextNode(Node);
  }

/*
** Delete all of the items in the update list and delete Buffer
*/

  for (int i=0; i<Natts; i++) {
    F_UPDATE_LIST_ITEM *UpdateListItem
      = (F_UPDATE_LIST_ITEM *)RB_POP(attUpdateList);
    RB_FREE_DELETE(FreeHLA, UpdateListItem);
  }

  SIMOBJ->ProcessEventHandlers("Update Attributes",
			       TIME_TAG,
			       NULL,
			       0,
			       -1);

}

void E_SpScheduleUpdate::commit() {

  if (ExtSub) {
    cout << "Commiting to OPSM" << endl;
    HostUser->SendSubscribedData(TIME_TAG,
				   "ObjectProxy",
				   Buffer,
				   Nbytes,
				   GLOBAL_ID);
  }

  delete [] Buffer;

}
