// E_SpUndeliverProxyPtr.C event object file

#include "E_SpUndeliverProxyPtr.H"
#include "F_SpProxyItem.H"
#include "S_SpHLA.H"
#include "SpSubscribeSpace.H"
#include "H_SpRemoveProxySpace.H"

#include "SpFreeHLA.H"
extern SpFreeHLA *FreeHLA;

#include "RB_SpContainerFuncs.H"
/************************************************************************
* E_SpUndeliverProxyPtr : constructor for the UnDeliverProxyPointer 	*
************************************************************************/
E_SpUndeliverProxyPtr::E_SpUndeliverProxyPtr()
  : valid(0) {
//  set_lazy();
}

/************************************************************************
* process : optimistically process event				*
************************************************************************/

int E_SpUndeliverProxyPtr::lazy() {
  cout << time_tag << " E_SpUndeliverProxyPtr lazy" << endl;
  return (0);
}

void E_SpUndeliverProxyPtr::process() {
  F_SpProxyItem *proxy_item;
  S_SpHLA *sim_obj           = (S_SpHLA *) SIMOBJ;
  M_SpUndeliverProxyPtr *msg = (M_SpUndeliverProxyPtr *) get_MESSAGE();
  char *data                 = (char *) get_DATA();
  C_XQUEUE *remote_proxies   = sim_obj->GetRemoteObjectProxies();
  int buffer_len, len;
  char *buffer, *cp;

  valid   = 1;
  errcode = 0;

  if (msg->ProxyGlobalId == GLOBAL_ID) return;

  proxy_item = (F_SpProxyItem *) remote_proxies->find(msg->ProxyGlobalId);
  if (!proxy_item){
    valid   = 0;
    errcode = 1;
    return;
  }

  // undeliver proxy-ptr to subscribe space ( msg->SpaceId )
  if (msg->SpaceId>=0){
    SpSubscribeSpaceRemoveProxy remove_proxy(msg->SpaceId, msg->ProxyGlobalId, msg->NumGrid);
    buffer_len = sizeof(SpSubscribeSpaceRemoveProxy) + (sizeof(int) * msg->NumGrid);
    cp = buffer = new char [buffer_len];
    len = sizeof(SpSubscribeSpaceRemoveProxy); memcpy(cp, &remove_proxy, len); cp += len;
    len = sizeof(int) * msg->NumGrid;          memcpy(cp, data, len);
    SIMOBJ->ProcessEventHandlers("SubscribeSpace Remove Proxy Ptr",
                                 TIME_TAG,
                                 (char *) buffer, buffer_len, -1);
    SpSubscribeSpaceRemoveProxy *removed_proxy = (SpSubscribeSpaceRemoveProxy *) buffer;
    valid &= (removed_proxy->GetNumGridRemove()==msg->NumGrid);
    if (!valid) errcode = 2;
    delete [] buffer;
  }

#ifdef DDM_DEBUG
  if (valid){
    cout << time_tag << " E_SpUndeliverProxyPtr event processed " 
         << " simobj " << sim_obj->get_GLOBAL_ID() << " proxy " << msg->ProxyGlobalId 
         << " grids ";
    int *grid_id = (int *) get_DATA();
    for (int i=0; i<msg->NumGrid; i++){
      cout << grid_id[i] << ", ";
    }
    cout << endl;
  }
#endif

  RB_REMOVE(remote_proxies, proxy_item->GetGlobalId());
  RB_FREE_DELETE(FreeHLA, proxy_item);
}

void E_SpUndeliverProxyPtr::commit() {
  S_SpHLA *simobj            = (S_SpHLA *) SIMOBJ;
  M_SpUndeliverProxyPtr *msg = (M_SpUndeliverProxyPtr *) get_MESSAGE();
  int *grid_id               = (int *) get_DATA();

  if (!valid){
    cout << time_tag << " ERROR::E_SpUndeliverProxyPtr event processed invalid code " << errcode 
         << " simobj " << simobj->get_GLOBAL_ID() << " proxy " << msg->ProxyGlobalId 
         << " grids ";
    for (int i=0; i<msg->NumGrid; i++){
      cout << grid_id[i] << ", ";
    }
    cout << endl;
    abort();
  }
}
