// E_DeliverProxyPtr event object file

#include "E_SpDeliverProxyPtr.H"
#include "S_SpHLA.H"
#include "SpSubscribeSpace.H"
#include "F_SpProxyItem.H"
#include "H_SpAddProxySpace.H"

#include "RB_SpContainerFuncs.H"

#include "SpFreeHLA.H"
extern SpFreeHLA *FreeHLA;

//#define DDM_DEBUG

/************************************************************************
* E_SpDeliverProxyPtr : constructor for the DeliverProxyPointer	event	*
************************************************************************/
E_SpDeliverProxyPtr::E_SpDeliverProxyPtr()
  : valid(0) {
//  set_lazy();
}

/************************************************************************
* process : optimistically process event				*
************************************************************************/
int E_SpDeliverProxyPtr::lazy() {
  return (0);
}

void E_SpDeliverProxyPtr::process() {
  M_SpDeliverProxyPtr *msg = (M_SpDeliverProxyPtr *)MESSAGE;
  S_SpHLA *HLAobj          = (S_SpHLA *)SIMOBJ;
  char *data               = (char *) get_DATA();
  SpObjProxy *ObjectProxy  = msg->ObjectProxy;
  int buffer_len, len;
  char *buffer, *cp;

  valid   = 1;
  errcode = 0;
  num_grid_added = 0;

  if (ObjectProxy->GetGlobalId() == GLOBAL_ID) {
    RB_cout << "Warning: Getting my own Object Proxy" << endl;
    return;
  }

  C_XQUEUE *remote_proxies = HLAobj->GetRemoteObjectProxies();
  int num_proxy            = remote_proxies->get_length();
  F_PROXY_ITEM *proxy_item;
  int i;

#ifdef DDM_DEBUG
  cout << time_tag << " E_SpDeliverProxyPtr simobj " << HLAobj->get_GLOBAL_ID() 
       << " proxy " << ObjectProxy->GetGlobalId() 
       << " current proxy ";
  proxy_item = (F_PROXY_ITEM *) remote_proxies->get_top();
  for (i=0; i<num_proxy; i++){
    cout << proxy_item->GetGlobalId() << ", ";
    proxy_item = (F_PROXY_ITEM *) proxy_item->get_link();
  }
  cout << endl;
#endif

  proxy_item = (F_PROXY_ITEM *) remote_proxies->get_top();
  for (i=0; i<num_proxy; i++){
    if (proxy_item->GetGlobalId()==ObjectProxy->GetGlobalId()) break;
    proxy_item = (F_PROXY_ITEM *) proxy_item->get_link();
  }
  if (i<num_proxy){
    valid   = 0;
    errcode = 1;
    return;
  }

  F_SpProxyItem *ProxyItem = (F_SpProxyItem *)RB_FREE_NEW(FreeHLA, PROXY_ITEM);
  ProxyItem->SetObjectProxy(msg->ObjectProxy);

  C_XQUEUE *RemoteObjectProxies = HLAobj->GetRemoteObjectProxies();
  RB_INSERT(RemoteObjectProxies, ProxyItem);

  SIMOBJ->ProcessEventHandlers("Discover Object",
			       TIME_TAG,
			       (char *)ProxyItem,
			       sizeof(F_SpProxyItem),
			       -1);

  // deliver proxy ptr to subscribe space
  if (msg->SpaceId>=0){
    SpSubscribeSpaceAddProxy add_proxy(msg->SpaceId, ProxyItem, msg->NumGrid);
    buffer_len = sizeof(SpSubscribeSpaceAddProxy) + (sizeof(int) * msg->NumGrid);
    cp = buffer = new char [buffer_len];
    len = sizeof(SpSubscribeSpaceAddProxy); memcpy(cp, &add_proxy, len); cp += len;
    len = sizeof(int) * msg->NumGrid; memcpy(cp, data, len);
    SIMOBJ->ProcessEventHandlers("SubscribeSpace Add Proxy Ptr",
                              TIME_TAG,
                              (char *) buffer, buffer_len, -1);
    SpSubscribeSpaceAddProxy *added_proxy = (SpSubscribeSpaceAddProxy *) buffer;
    valid &= (added_proxy->GetNumGridAdd()==msg->NumGrid);
    if (!valid){
      errcode = 2;
      num_grid_added = added_proxy->GetNumGridAdd();
    }
    delete [] buffer;
  }
}

void 
E_SpDeliverProxyPtr::commit() {
  M_SpDeliverProxyPtr *msg = (M_SpDeliverProxyPtr *)MESSAGE;
  S_SpHLA *simobj          = (S_SpHLA *)SIMOBJ;
  SpObjProxy *proxy        = msg->ObjectProxy;

  if (!valid){
    cout << time_tag << " ERROR::E_SpDeliverProxyPtr event processed invalid code " << errcode;
    cout << " simobj " << simobj->get_GLOBAL_ID() << " proxy " << proxy->GetGlobalId() 
         << " num grid added " << num_grid_added << "/" << msg->NumGrid << endl;
    abort();
  }
}
