// E_SpNewProxy.C event object file

#include "E_SpNewProxy.H"
#include "M_SpDeliverProxyPtr.H"
#include "SpScheduleEvent.H"
#include "S_SpLocalProxyMgr.H"

#include "SpFreeHLA.H"
extern SpFreeHLA *FreeHLA;

#include "RB_SpContainerFuncs.H"
#include "RB_SpObjProxyFuncs.H"
#include "F_SpProxyListItem.H"
#include "F_SpSimObjRef.H"

#ifdef TODO
EVENT_TYPE (DELIVER_PROXY_PTR);
EVENT_TYPE (DeliverProxyPtrSpace);
#else
#include "E_SpDeliverProxyPtr.H"
#include "E_SpDeliverProxyPtrSpace.H"
#endif


/************************************************************************
* E_SpNewProxy : constructor for the NEW_PROXY event	*
************************************************************************/
E_SpNewProxy::E_SpNewProxy()
  : valid(0), errcode(0) {
//  set_lazy();
}

/************************************************************************
* process : optimistically process event				*
************************************************************************/

void E_SpNewProxy::process() {
  char *data        = (char *)         get_DATA();
  M_SpNewProxy *msg = (M_SpNewProxy *) get_MESSAGE();
  SpObjProxyMsg *proxy_msg;
  SpProxyDestination *new_dest, *add_dest, *dup_dest;
  int num_new_dest, num_add_dest, num_dup_dest;
  int i, *dup_dest_sent;

  // data storage : [Add Dest] + [Dup Dest] + [New Dest] + [New Proxy]

  proxy_msg = (SpObjProxyMsg *) 
              (data + (sizeof(SpProxyDestination) * 
               (msg->NumAddDest + msg->NumDupDest + msg->NumNewDest)));

  num_new_dest = msg->NumNewDest;
  num_add_dest = msg->NumAddDest;
  num_dup_dest = msg->NumDupDest;

  add_dest = (SpProxyDestination *) data;
  new_dest = (SpProxyDestination *) 
             (data + (sizeof(SpProxyDestination) *  
              (msg->NumAddDest + msg->NumDupDest)));
  dup_dest = (SpProxyDestination *) 
             (data + (sizeof(SpProxyDestination) * msg->NumAddDest));

  valid = 1; errcode = 0;

  dup_dest_sent = (num_dup_dest>0) ? new int [num_dup_dest] : (int *) 0;
  for (i=0; i<num_dup_dest; i++) 
    dup_dest_sent[i] = 0;

//float look_ahead = msg->LookAhead;
  float look_ahead = 0.0;

  ProcessNew(msg->SpaceId, look_ahead, num_new_dest, new_dest, 
             num_dup_dest, dup_dest, dup_dest_sent, proxy_msg); 
  ProcessAdd(msg->ProxyGlobalId, msg->SpaceId, look_ahead, num_add_dest, add_dest, 
             num_dup_dest, dup_dest, dup_dest_sent);
  ProcessDup(msg->ProxyGlobalId, msg->SpaceId, look_ahead, num_dup_dest, dup_dest, dup_dest_sent);

  if (dup_dest_sent) delete [] dup_dest_sent;
}

void 
E_SpNewProxy::ProcessNew(int space_id, float look_ahead, 
                         int num_new_dest, SpProxyDestination *new_dest, 
                         int num_dup_dest, SpProxyDestination *dup_dest, int *dup_dest_sent, 
                         SpObjProxyMsg *proxy_msg) {
  S_SpLocalProxyMgr *local_proxy_man = (S_SpLocalProxyMgr *) SIMOBJ;
  SpObjProxy *obj_proxy;
  C_XQUEUE *proxy_list, *local_subscriber_list;
  F_SpSimObjRef *obj_ref;

  if (num_new_dest<=0) return;

  // Get the object proxy message from the input message and generate an
  obj_proxy = RB_GenerateProxy(proxy_msg);

  // Put the object proxy in the LocalProxyMan's list of object proxies
  proxy_list = local_proxy_man->GetProxyList();

  F_SpProxyListItem *proxy_list_item =
    (F_SpProxyListItem *) RB_FREE_NEW(FreeHLA, PROXY_LIST_ITEM);

  proxy_list_item->SetObjectProxy(obj_proxy);
  proxy_list_item->SetGlobalId(obj_proxy->GetGlobalId());

  RB_INSERT(proxy_list, proxy_list_item);

  local_subscriber_list = proxy_list_item->GetProxyList();

  int num_grid_id;
  int *grid_id = new int[num_dup_dest+1];

  SpSimTime event_time(time_tag);
  event_time += look_ahead;

  for (int i=0; i<num_new_dest; i++){
    obj_ref = (F_SpSimObjRef *) RB_FREE_NEW(FreeHLA, SIMOBJ_REF);
    obj_ref->SetNode(new_dest[i].Node);
    obj_ref->SetType(new_dest[i].ObjectType);
    obj_ref->SetLocalID(new_dest[i].LocalId);
    obj_ref->SetGlobalID(new_dest[i].GlobalId);
    RB_INSERT(local_subscriber_list, obj_ref);

    num_grid_id = 0; grid_id[num_grid_id++] = new_dest[i].GridId;
    for (int j=0; j<num_dup_dest; j++){
      if (dup_dest_sent[j]) continue;
      if (new_dest[i].GlobalId==dup_dest[j].GlobalId){
        grid_id[num_grid_id++] = dup_dest[j].GridId;
        dup_dest_sent[j]       =  1;	// marked as processed
      }
    }

    // Deliver to HLA
    M_SpDeliverProxyPtr msg(obj_proxy, space_id, num_grid_id);
    ScheduleEvent(event_time, DELIVER_PROXY_PTR,
                new_dest[i].ObjectType, new_dest[i].LocalId, new_dest[i].Node,
                msg, (const char *) grid_id, (msg.NumGrid * sizeof(int)));
  }

  delete [] grid_id;
}

void 
E_SpNewProxy::ProcessAdd(int proxy_global_id, int space_id, float look_ahead,
                         int num_add_dest, SpProxyDestination *add_dest,
                         int num_dup_dest, SpProxyDestination *dup_dest, int *dup_dest_sent) {
  S_SpLocalProxyMgr *local_proxy_man = (S_SpLocalProxyMgr *) SIMOBJ;
  F_SpProxyListItem *proxy_list_item;
  F_SpSimObjRef *new_obj_ref;
  C_XQUEUE *local_subscriber_list;

  if (num_add_dest<=0) return;

  proxy_list_item = (F_SpProxyListItem *)
                    local_proxy_man->GetProxyList()->find(proxy_global_id);

  if (!proxy_list_item){
    valid   = 0;
    errcode = 1;
    return;
  }

  int num_grid_id;
  int *grid_id = new int[num_dup_dest+1];

  SpSimTime event_time(time_tag);
  event_time += look_ahead;

  for (int i=0; i<num_add_dest; i++){
    new_obj_ref = (F_SpSimObjRef *) RB_FREE_NEW(FreeHLA, SIMOBJ_REF);
    new_obj_ref->SetNode(add_dest[i].Node);
    new_obj_ref->SetType(add_dest[i].ObjectType);
    new_obj_ref->SetLocalID(add_dest[i].LocalId);
    new_obj_ref->SetGlobalID(add_dest[i].GlobalId);

    local_subscriber_list = proxy_list_item->GetProxyList();

    RB_INSERT(local_subscriber_list, new_obj_ref);

    num_grid_id = 0; grid_id[num_grid_id++] = add_dest[i].GridId;
    for (int j=0; j<num_dup_dest; j++){
      if (dup_dest_sent[j]) continue;
      if (add_dest[i].GlobalId==dup_dest[j].GlobalId){
        grid_id[num_grid_id++] = dup_dest[j].GridId;
        dup_dest_sent[j]       = 1;	// marked as processed
      }
    }

    // deliver to HLA
    M_SpDeliverProxyPtr msg(proxy_list_item->GetObjectProxy(), space_id, num_grid_id);
    ScheduleEvent(event_time, DELIVER_PROXY_PTR,
                  add_dest[i].ObjectType, add_dest[i].LocalId, add_dest[i].Node,
                  msg, (const char *) grid_id, (msg.NumGrid * sizeof(int)));
  }

  if (grid_id) delete [] grid_id;
}

void 
E_SpNewProxy::ProcessDup(int proxy_global_id, int space_id, float look_ahead,
                         int num_dest, SpProxyDestination *dest, int *dest_sent) {
  S_SpLocalProxyMgr *local_proxy_man = (S_SpLocalProxyMgr *) SIMOBJ;
  SpBinaryTree subscribers, *subscriber;
  SpProxyDestination *local_dest;
  int *grid, num_grid;
  F_SpProxyListItem *proxy_list_item;

  if (num_dest<=0) return;

  proxy_list_item = (F_SpProxyListItem *)
                    local_proxy_man->GetProxyList()->find(proxy_global_id);

  if (!proxy_list_item){
    valid   = 0;
    errcode = 2;
    return;
  }

  for (int i=0; i<num_dest; i++){
    if (dest_sent[i]) continue;	// if marked as processed, ignore
    subscriber = (SpBinaryTree *) subscribers.Find(dest[i].GlobalId);
    if (!subscriber){
      subscriber = new SpBinaryTree();
      subscribers.Insert((void *) subscriber, dest[i].GlobalId);
    }
    subscriber->Insert((void *) (dest+i), dest[i].GridId);
  }

  SpSimTime event_time(time_tag);
  event_time += look_ahead;

  M_SpDeliverProxyPtrSpace msg(space_id, proxy_global_id);

  subscriber = (SpBinaryTree *) subscribers.GetFirstElement();
  while (subscriber){

    num_grid = subscriber->GetNumElements();
    grid     = (num_grid>0) ? new int [num_grid] : (int *) 0;
    num_grid = 0;

    local_dest = (SpProxyDestination *) subscriber->GetFirstElement();
    while (local_dest){
      grid[num_grid++] = local_dest->GridId;
      local_dest = (SpProxyDestination *) subscriber->GetNextElement();
    } 

    local_dest = (SpProxyDestination *) subscriber->GetFirstElement();

    msg.NumGrid = num_grid;
    ScheduleEvent(event_time, DeliverProxyPtrSpace, 
                  local_dest->ObjectType, local_dest->LocalId, local_dest->Node, 
                  msg, (const char *) grid, (msg.NumGrid*sizeof(int)));
    if (grid) delete [] grid;

    subscriber = (SpBinaryTree *) subscribers.GetNextElement();
  }

  while (subscriber = (SpBinaryTree *) subscribers.RemoveFirstElement())
    delete subscriber;
}

int 
E_SpNewProxy::lazy() {
  return (0); 
}

void
E_SpNewProxy::commit(){
  int num_new_dest, num_add_dest, num_dup_dest;
  SpProxyDestination *new_dest, *add_dest, *dup_dest;
  M_SpNewProxy *msg = (M_SpNewProxy *) get_MESSAGE();
  char *data        = (char *)         get_DATA();
  int i;

  num_new_dest = msg->NumNewDest;
  num_add_dest = msg->NumAddDest;
  num_dup_dest = msg->NumDupDest;

  add_dest = (SpProxyDestination *) data;
  new_dest = (SpProxyDestination *) 
             (data + (sizeof(SpProxyDestination) *  
              (msg->NumAddDest + msg->NumDupDest)));
  dup_dest = (SpProxyDestination *) 
             (data + (sizeof(SpProxyDestination) * msg->NumAddDest));

  if (!valid){
    cout << time_tag << " ERROR::E_SpNewProxy event processed invalid error code " 
         << errcode << " - ";

    cout << "Proxy " << msg->ProxyGlobalId;
    cout << " new ";
    for (i=0; i<num_new_dest; i++)
      cout << new_dest[i].GlobalId << ", ";

    cout << " add ";
    for (i=0; i<num_add_dest; i++)
      cout << add_dest[i].GlobalId << ", ";

    cout << " dup ";
    for (i=0; i<num_dup_dest; i++)
      cout << dup_dest[i].GlobalId << ", ";
    cout << endl;
    abort();
  }
}
