// C_SpSensor.C file implements the sensor component class

#include "C_SpSensor.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"

#include "RB_SpContainerFuncs.H"
#include "RB_SpDefineClass.H"

RB_DEFINE_CLASS(GATE_ITEM);

/*****************************************************************************
* Constructor
******************************************************************************/
C_SpSensor::C_SpSensor() {
 GateHash = NULL;
}

/*****************************************************************************
* AddProxy - Adds a proxy to the DetectableProxyList
******************************************************************************/
void C_SpSensor::AddProxy(F_SpProxyItem *proxyItem) {

  DetectableProxyList += proxyItem;

}

/*****************************************************************************
* RemoveProxy - Removes a proxy from the DetectableProxyList
******************************************************************************/
void C_SpSensor::RemoveProxy(F_SpProxyItem *proxyItem) {

  DetectableProxyList -= proxyItem;
  
}

/*****************************************************************************
* AddGate - Add a gate to the GateHash
******************************************************************************/
void C_SpSensor::AddGate(int proxyId) {

  if (GateHash == NULL) {         
    GateHash = new RB_hash(100);
  }

  GATE_ITEM *GateItem = RB_NEW_GATE_ITEM();    
  GateItem->SetProxyId(proxyId);               
  *GateHash +=GateItem;

}

/*****************************************************************************
* RemoveGate - Removes a gate from the GateHash.  
******************************************************************************/
void C_SpSensor::RemoveGate(int proxyId) {

  GATE_ITEM *GateItem = (GATE_ITEM *) RB_REMOVE(GateHash, proxyId);

  RB_DELETE_GATE_ITEM(GateItem);

}

/*****************************************************************************
* SetGate - Finds the gate for the remote proxy in the hash table and sets
* its value to 1 (occurs when the remote proxy changes)
******************************************************************************/
void C_SpSensor::SetGate(int proxyId) {

  if (GateHash == NULL) {
    RB_cout << "Error (C_SpSensor), NULL GateHash in SetGate"
	    << endl;
    return;
  }

  GATE_ITEM *GateItem = (GATE_ITEM *)GateHash->find(proxyId);
  if (GateItem == NULL) {
    RB_cout << "Error (C_SpSensor), No GateItem in SetGate" << endl;
    return;
  }

  GateItem->Gate = 1;

}

/*****************************************************************************
* SetAllGates - Sets all of the gates for remote proxies to 1 (occurs when
* my object changes)
******************************************************************************/
void C_SpSensor::SetAllGates() {
  int i;

  if (GateHash == NULL) {
    return;
  }

  GateHash->combine();
  C_XQUEUE *qall = GateHash->get_all();

  GATE_ITEM *GateItem = (GATE_ITEM *)qall->get_top();
  for (i=0; i<qall->get_length(); i++) {
    GateItem->Gate = 1;
    GateItem = (GATE_ITEM *)GateItem->get_link();
  }

}

void RB_Register_CLASS_GATE_ITEM()
{
  RB_REGISTER_CLASS_GATE_ITEM();
}

