// C_SpSensor.H header file defines the class that coordinates
// precomputations for sensors

#ifndef SENSOR_COMP_OBJ
#define SENSOR_COMP_OBJ

#include "SpComponent.H"
#include "xhash.H"
#include "xlist.H"
#include "SpProcSem.H"
#include "F_SpProxyItem.H"

/*****************************************************************************
* GATE_ITEM - This class is only used by the sensor component class to
* reference gates in the Gate Hash table
******************************************************************************/
class GATE_ITEM : public C_SQ_ITEM {

  public:

    LogicalSemaphore Gate;

    GATE_ITEM() {;}

    void SetProxyId(int proxyId) {id = proxyId;}
    int GetProxyId() {return id;}

};

//RB_DEFINE_CLASS(GATE_ITEM);

/*****************************************************************************
* C_SpSensor - Applications inherit from this class to define their own
* specific sensor components. This class manages the detectable proxy list
* and a hash table of gates which are used by processes that are doing
* precomputations.
******************************************************************************/
class C_SpSensor : public SpComponent {

  public:
    C_SpSensor();

  protected:

    RB_list DetectableProxyList;
    RB_hash *GateHash;

  private:

/*
** Methods for adding and removing proxies from the Detectable Proxy List
*/
    void AddProxy(F_SpProxyItem *proxyItem);
    void RemoveProxy(F_SpProxyItem *proxyItem);

/*
** Methods involving setting the gate when attributes change (either myself
** or a remote object)
*/

    void AddGate(int proxyId);  // Called the first time this proxy is added 
                                // to the DectableProxyList
    void RemoveGate(int proxyId); // Called when this proxy is removed from 
                                // the object's remoteProxy list 
    void SetGate(int proxyId);  // Gets called when an object changes
    void SetAllGates();         // Gets called when I change

};

typedef C_SpSensor SENSOR_COMPONENT;

#endif
