// LocalProxyMan.C method file

#include "O_SpLocalProxyMgr.H"

#include "SpObjFactory.H"
extern SpObjFactory *ObjFact;

#include "S_SpHLA.H"

#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpStdlib.H"
#include "SpString.H"

/************************************************************************
* LOCAL_PROXYMAN : Construct a Local Proxy Manager object		*
************************************************************************/
O_SpLocalProxyMgr::O_SpLocalProxyMgr() {
  int i;

  cout << "Constructing LOCAL_PROXYMAN Manager" << endl;

  InitializeProxyCode();
 
/*
** Read in the input file using the Data Parser and extract parameters
*/

  ProxyManParser = new SpDataParser("ProxyMan.par");

  ProxyManParser->GoTo("parameters", NULL);
  N_LOC = ProxyManParser->GetInt("Nmanagers");

/*
** Determine how many objects to create on this node
*/

  N_TOT = N_LOC*N_NODES;

/*
** Create N_LOC objects for this node.  If we have only 1 then all
** events for updating proxies, and delivering them, etc. will go to
** this one S_SpLocalProxyMgr, which will result in lots of unnecessary
** rollbacks.  So it makes sense for the N_LOC variable to be a high
** number like maybe 100 so that each S_SpLocalProxyMgr will have a small
** number of events which means less roll backs.  We could also use lazy
** cancellation to prevent the rollbacks, but there may be problems
** with that.
*/

  const int n_loc = N_LOC;
  LocalProxyMan = new S_SpLocalProxyMgr[n_loc];

/*
** Tell SPEEDES about the local objects that have been created
*/

  reserve_n_objects(N_LOC);

  for (i=0; i<N_LOC; i++) {
    define_object(&LocalProxyMan[i]);
  }

  TOTAL_OBJECTS += N_TOT;

}

void O_SpLocalProxyMgr::terminate(double /*tend*/) {
  delete [] LocalProxyMan;
  delete ObjFact;
}
