// S_SpDecMgr.C source file

#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "S_SpDecMgr.H"
#include "SpScheduleEvent.H"
#include "RB_Types.H"
#include "M_SpHlaUnsubscribe.H"
#include "M_SpAddSubscriber.H"
#include "S_SpHLA.H"
#include "SpBinaryTree.H"
#include "SpList.H"
#include "SpProxyObjHandle.H"

#ifdef TODO
EVENT_TYPE (ADD_SUBSCRIBER_HLA);
EVENT_TYPE (UNSUBSCRIBE_HLA);
#else
#include "E_SpHlaUnsubscribe.H"
#include "E_SpAddSubscriber.H"
#endif

#include "RB_SpDefineClass.H"
#include "RB_SpBinaryTree.H"
 
RB_DEFINE_CLASS(SpProxyObjHandle);


S_SpDecMgr::S_SpDecMgr()
  : class_id(-1), 
    publisher(new RB_SpBinaryTree),
    subscriber(new RB_SpBinaryTree){

  publisher->SetBalancedTreeMode();
  subscriber->SetBalancedTreeMode();
}

S_SpDecMgr::~S_SpDecMgr() {
  SpProxyObjHandle *obj_handle; 

  while ((obj_handle = (SpProxyObjHandle *) publisher->RemoveFirstElement()))
    RB_DELETE_SpProxyObjHandle(obj_handle);
  delete publisher;
  while ((obj_handle = (SpProxyObjHandle *) subscriber->RemoveFirstElement()))
    RB_DELETE_SpProxyObjHandle(obj_handle);
  delete subscriber;
}

// terminate: executed at the end of the simulation
void S_SpDecMgr::terminate(double /*tend*/) {
/*
  cout << "S_SpDecMgr class id " << GetClassId() 
       << " global id " << get_GLOBAL_ID()
       << " node " << NODE
       << " critical path : "    << GetCriticalPathTime()
       << " commit : "           << GetCommittedTime() << endl;
*/
}

// invoke from E_SpDecMgrSubscribe
void 
S_SpDecMgr::AddSubscriber(int node, int type, int local_id, int global_id){
  const SpSimTime &time = CurrentEvent->get_MESSAGE()->time_tag;
  SpProxyObjHandle *publisher_obj, *subscriber_obj;
  int publisher_global_id;
  char *new_subscriber_msg;
  int new_subscriber_msg_len;

  subscriber_obj = (SpProxyObjHandle *) subscriber->Find(global_id);
  if (subscriber_obj) return;

  M_SpAddSubscriber msg(1);
  SpProxyObjHandle new_subscriber(local_id, node, type, global_id);
  new_subscriber.CreateMessage(new_subscriber_msg, new_subscriber_msg_len);

  subscriber->Insert(RB_NEW_SpProxyObjHandle(new SpProxyObjHandle(local_id, node, type, global_id)), global_id);

  // send to each publisher, 'this' new subscriber
  publisher_obj = (SpProxyObjHandle *) publisher->GetFirstElement();
  while (publisher_obj){
    publisher_global_id = int(publisher->GetCurrentKey());
    if (publisher_global_id==global_id){
      publisher_obj = (SpProxyObjHandle *) publisher->GetNextElement();
      continue;
    }
    ScheduleEvent(time, ADD_SUBSCRIBER_HLA, publisher_obj->GetObjType(), 
                  publisher_obj->GetLocalId(), publisher_obj->GetNode(),
                  msg, new_subscriber_msg, new_subscriber_msg_len);
    publisher_obj = (SpProxyObjHandle *) publisher->GetNextElement();
  }

  delete [] new_subscriber_msg;
}

// invoke from E_PUBLISH_DECMAN
void 
S_SpDecMgr::AddPublisher(int node, int type, int local_id, int global_id){
  const SpSimTime &time = CurrentEvent->get_MESSAGE()->time_tag;
  int subscriber_global_id;
  SpProxyObjHandle *subscriber_obj, *publish_obj;
  char *buffer;
  int buffer_len;
  SpList subscribers;

  publish_obj = (SpProxyObjHandle *) publisher->Find(global_id);
  if (publish_obj) return;

  publisher->Insert(RB_NEW_SpProxyObjHandle(new SpProxyObjHandle(local_id, node, type, global_id)), global_id);

  if (subscriber->GetNumElements()<=0) return;

  // send to the publisher, a list of 'current' subscriber

  subscriber_obj = (SpProxyObjHandle *) subscriber->GetFirstElement();
  while (subscriber_obj){
    subscriber_global_id = int(subscriber->GetCurrentKey());
    if (subscriber_global_id==global_id){
      subscriber_obj = (SpProxyObjHandle *) subscriber->GetNextElement();
      continue;
    }
    subscribers.Insert(subscriber_obj);
    subscriber_obj = (SpProxyObjHandle *) subscriber->GetNextElement();
  }

  if (subscribers.GetNumElements()>0){
    SpProxyObjHandle::CreateMessage(subscribers, buffer, buffer_len);
    M_SpAddSubscriber msg(subscribers.GetNumElements());
    ScheduleEvent(time, ADD_SUBSCRIBER_HLA, type, local_id, node,
                  msg, buffer, buffer_len);
    delete [] buffer;
    while (subscribers.RemoveTop());
  }
}

// invoke from E_SpDecMgrUnpublish
int 
S_SpDecMgr::RemovePublisher(int, int, int, int global_id){
  int status = 0;
  SpProxyObjHandle *publish_obj;

  publish_obj = (SpProxyObjHandle *) publisher->Find(global_id);
  if (publish_obj){
      // remove from "publisher" list
    publish_obj = (SpProxyObjHandle *) publisher->Remove(global_id);
    if (publish_obj){
      RB_DELETE_SpProxyObjHandle(publish_obj);
      status = 1;
    }
  }

  return (status);
}

// invoke from UNSUBSCRIBE_DECMAN
void
S_SpDecMgr::RemoveSubscriber(int, int global_id){
  int publisher_global_id;
  const SpSimTime &time = CurrentEvent->get_MESSAGE()->time_tag;
  SpProxyObjHandle *publisher_obj, *subscriber_obj;

  subscriber_obj = (SpProxyObjHandle *) subscriber->Find(global_id);
  if (!subscriber_obj) return;

  M_SpHlaUnsubscribe msg(subscriber_obj->GetNode(), subscriber_obj->GetGlobalId(),
                         subscriber_obj->GetLocalId(), subscriber_obj->GetObjType());

    // remove from "subscriber" list
  subscriber_obj = (SpProxyObjHandle *) subscriber->Remove(global_id);
  RB_DELETE_SpProxyObjHandle(subscriber_obj);

    // notify publisher
  publisher_obj = (SpProxyObjHandle *) publisher->GetFirstElement();
  while (publisher_obj){
    publisher_global_id = int(publisher->GetCurrentKey());
    if (publisher_global_id==global_id){
      publisher_obj = (SpProxyObjHandle *) publisher->GetNextElement();
      continue;
    }
    ScheduleEvent(time, UNSUBSCRIBE_HLA, publisher_obj->GetObjType(), 
                  publisher_obj->GetLocalId(), publisher_obj->GetNode(), msg);
    publisher_obj = (SpProxyObjHandle *) publisher->GetNextElement();
  }
}
