// SpObjFactory.C Source File

#include "SpStdlib.H"

#include "SpObjFactory.H"
#include "SpObjProxy.H"

int SpObjFactory::Node = 0;
int SpObjFactory::Nnodes = 0;

int SpObjTemplate::ClassCounter = 0;
void *new_SpObjProxy(int n) 	{return new SpObjProxy[n];}

#include "F_SpProxyItem.H"

// <BEGIN: Global free lists>
#include "SpFreeDynAttributes.H"
SpFreeDynAttributes *FreeDynamicAttributes = NULL;

#include "SpFreeHLA.H"
SpFreeHLA *FreeHLA = NULL;

// <END:   Global free lists>

/*
** This little routine copies the input string to the output string and then
** replaces all occurances of the token with NULLs. The input string can be
** the same as the output string. The number of tokens found is returned from
** the function call.
*/

#define MAXIMUM_STRING_LENGTH 255

int FindTok(char *OutString, char *InString, char tok) {
  int i, j;
  char S[MAXIMUM_STRING_LENGTH+1];

  if (strlen(InString) > MAXIMUM_STRING_LENGTH) {
    cout << "Error FindTok: Input string is too long"
	 << strlen(InString) << endl;
    exit(0);
  }

  strcpy(S, InString);
  strcpy(OutString, S);

  j = 0;
  for (i=0; i<int(strlen(OutString)); i++) {
    if (OutString[i] == tok) OutString[i] = '\0';
    j++;
  }

  return j;

}


/************************************************************************
 * SpObjFactory constructor				.		*
 ************************************************************************/
SpObjFactory::SpObjFactory(char *FileName) {
  void *Reference;
  int n;
  int i;
  int j;
  char *Name;
  SpObjTemplate *Template;
  SpObjTemplate *Parent;
  C_QUEUE *q;

  Parser = new SpDataParser(FileName);
  Templates = new C_HASH(100);

  Reference = Parser->GetFirstType();
  while (Reference) {
    cout << "Constructing template for: "
	 << Parser->GetReferenceName() << endl;
    GetTemplate(Parser->GetReferenceName());
    Reference = Parser->GetNextType(Reference);
  }

/*
** Go through the object names in the free list and see if they are in the
** hash table of object templates. If so, then set the free list id in the
** template and copy the name in the FreeNames array. This lets us go back
** and forth between object name and free list id.
*/

  n = Templates->get_nitems();
  FreeObjectProxy = new FREE_OBJECT_PROXY(n);
  
// <BEGIN: Create global free lists>
  FreeDynamicAttributes = new SpFreeDynAttributes();
  FreeHLA = new SpFreeHLA;
// <END:   Create global free lists>

  FreeNames = new char *[n];
  for (i=0; i<n; i++) FreeNames[i] = NULL;

  for (i=0; i<n; i++) {
    Name = FreeObjectProxy->GetName(i);
    if (Name) {
      Template = (SpObjTemplate *)Templates->get(Name);
      if (Template) {
	Template->set_freeid(i);
	FreeNames[i] = strdup(Name);
      }else{
	cout << "Error (Object Factory): FREE_OBJECT_PROXY::FREE_OBJECT_PROXY"
	  "\ncalled set_type and passed Free list name " << Name
	     << "\nbut that name was not found in Objects.par" << endl;
	exit(1);
      }
    }
  }

/*
** Now, it is possible that not all of the object proxies have user-defined
** free list object counterparts. In this case, we still want to use the
** free list, but we will have to define things ourselves. In this step, we
** go through the templates to find any object types that are not defined
** in the free list. When we find them, we define them for the free list.
*/

  Templates->combine();
  q = Templates->get_all();
  Template = (SpObjTemplate *)q->get_top();
  j = 0;

//  cout << "Testing Free List Name-Id Pairs" << endl;
  for (i=0; i<n; i++) {
    if (Template->get_freeid() == -1) {
      while (FreeNames[j] != NULL) j++;
      FreeNames[j] = strdup(Template->GetName());
      Template->set_freeid(j);

      Parent = Template;
      while ((Parent = Parent->GetParent())) {
	if (Parent->get_freeid() != -1) {
	  break;
	}
      }

      if (Parent) {

// put in parent constructor

/*
	cout << Template->GetName() << " using Parent: "
	     << Parent->GetName() << endl;
*/
	FreeObjectProxy->CopyType(Template->GetName(),
				  j,
				  Parent->get_freeid());
      }else{

	FreeObjectProxy->set_type(Template->GetName(),
				  j,
				  new_SpObjProxy,
				  sizeof(SpObjProxy),
				  100);
      }

    }
//    cout << Template->get_freeid() << ": " << Template->GetName() << endl;
    Template = (SpObjTemplate *)Template->get_link();
  }

  for (i = 0; i < n; i++)
    free(FreeNames[i]);

  delete [] FreeNames;

//  cout << "Testing Free List Name-Id Pairs" << endl;
//  for (i=0; i<n; i++) {
//    cout << i << ": " << FreeNames[i] << endl;
//  }

//  FreeObjectProxy->print(cout);

}

SpObjFactory::~SpObjFactory()
{
  delete FreeHLA;
  delete FreeDynamicAttributes;
  delete FreeObjectProxy;
  delete Parser;
  delete Templates;
}

C_QUEUE *
SpObjFactory::GetObjTemplate() {
  Templates->combine();
  return (Templates->get_all());
}

/************************************************************************
 * GetTemplate - Returns an object template based on the name passed	*
 * It creates the object template if it doesn't already	exist.		*
 ************************************************************************/
SpObjTemplate *SpObjFactory::GetTemplate(char *ObjectType,
					     void *ObjectMark) {
  SpObjTemplate *ObjectTemplate;
  SpObjTemplate *ParentTemplate;
  void *Mark;
  int Ninherits;
  int Nobjects;
  int Ndefines;
  char *DefName;
  char *DefType;
  int i;
  int StartIndex;
  char S[120];
  int Nsubscribe;

/*
** First check if the object template already exists. If so, return it.
** Otherwise, go on and create it.
*/

  ObjectTemplate = (SpObjTemplate *)Templates->get(ObjectType);
  if (ObjectTemplate) {
    return ObjectTemplate;
  }

/*
** Here is where the object template is created and then put in the
** Templates hash table.
*/

  ObjectTemplate = new SpObjTemplate(ObjectType,
				       Templates->get_hashid(ObjectType));
  Templates->add(ObjectTemplate);

/*
** Now actually construct the object template
*/

  if (ObjectMark) {
    Parser->SetMark(ObjectMark);
    Mark = ObjectMark;
  }else{
    Mark = Parser->GoTo(ObjectType, NULL);
  }

  Ninherits = Parser->GetNtypes("INHERIT");
  Nobjects = Parser->GetNtypes("OBJECT");
  Ndefines = Parser->GetNdefines();
  if (Ninherits<=0)
    Ndefines += 1;		// explicitly define "Active" logical attribute
  Nsubscribe = Parser->GetNtypes("SUBSCRIBE");

  if (Ninherits > 1) {
    cout << "Error (SpObjFactory): Multiple inheritance not supported"
	 << endl;
    exit(1);
  }

  ObjectTemplate->SetNinherits(Ninherits);
  ObjectTemplate->SetNobjects(Nobjects);
  ObjectTemplate->SetNdefines(Ndefines);

  if (Ninherits == 1) {

    Parser->GoToFirstReference(Mark, "INHERIT");

    FindTok(S, Parser->GetReferenceName(), '[');
    ParentTemplate = GetTemplate(S);

    ObjectTemplate->Inherit(ParentTemplate);
    StartIndex = ParentTemplate->GetNattributes();

  }else{

    ObjectTemplate->Inherit(NULL);
    StartIndex = 0;

  }

/*
** Add the subscription here
*/

  ObjectTemplate->SetNsubscribe(Nsubscribe);
  if (Nsubscribe>0){
    Parser->GoToFirstReference(Mark, "SUBSCRIBE");
    for (i=0; i<Nsubscribe; i++){
      ObjectTemplate->SetSubscribeName(i, Parser->GetReferenceName());
      Parser->GoToNextReference(Mark, "SUBSCRIBE");
    }
  }

/*
** Add the defined attributes here
*/

  Parser->SetMark(Mark);
  Parser->GetFirstDefine(DefType, DefName);
  int n = (Ninherits<=0) ? Ndefines-1 : Ndefines;
  for (i=0; i<n; i++) {
    ObjectTemplate->SetDefine(StartIndex+i, DefName, DefType);
    Parser->GetNextDefine(DefType, DefName);
  }
  if (Ninherits<=0)
    ObjectTemplate->SetDefine(StartIndex+i, "Active", MapAttribute(STATIC_LOGICAL_ATTRIBUTE_TYPE));

/*
** Add the virtual methods here
*/

  if (Parser->GoToFirstReference(Mark, "VIRTUAL")) {
    ObjectTemplate->CreateMethodArray(Parser->GetNtypes("define"));

    Parser->GetFirstDefine(DefType, DefName);

    for (i=0; i<ObjectTemplate->GetNmethods(); i++) {
      ObjectTemplate->SetMethodName(i,DefName);
      Parser->GetNextDefine(DefType, DefName);
    }
  }

/*
** Return the Object Template now
*/

  return ObjectTemplate;

}

/************************************************************************
 * Define - defines a method for an object class.			*
 ************************************************************************/
void SpObjFactory::Define(void (*Method)(void *, va_list &),
			    char *MethodName,
			    char *ObjectType) {
  SpObjTemplate *ObjectTemplate;

  ObjectTemplate = GetTemplate(ObjectType);
  if (ObjectTemplate == NULL) {
    cout << "Error (Define): Could not find Object Template for "
	 << ObjectType << endl;
    exit(0);
  }
  ObjectTemplate->AddMethod(Method, MethodName);

}

/************************************************************************
 * Create - create an object proxy using the free list.			*
 ************************************************************************/
SpObjProxy *SpObjFactory::Create(char *type) {
  SpObjProxy *ObjectProxy;
  SpObjTemplate *ObjectTemplate;
  int i;

  //cout << "Object Factory creating: " << type << endl;

  ObjectTemplate = (SpObjTemplate *)Templates->get(type);
  if (ObjectTemplate == NULL){
    cout<<" Could not find a template for type "<<type<<"."
	<<" We are now exiting"<<endl;
    exit(-17);
  }
  else{
    i = ObjectTemplate->get_freeid();
    
    ObjectProxy = (SpObjProxy *)FreeObjectProxy->new_object(i);
    ObjectProxy->Construct(type);
  }
  return ObjectProxy;

}


/************************************************************************
 * Destroy - Destroy an object proxy using the free list.			*
 ************************************************************************/
void SpObjFactory::Destroy(SpObjProxy *ObjectProxy) {

/*
** Destruct the Object Proxy and put its shell back into the free list
*/

  ObjectProxy->Destruct();
  FreeObjectProxy->delete_object(ObjectProxy);

}


/************************************************************************
 * PartialDestroy - Partially destroy an object proxy using the free    *
 * list.			                                        *
 ************************************************************************/
void SpObjFactory::PartialDestroy(SpObjProxy *ObjectProxy) {

/*
** Partially destruct the Object Proxy and put its shell back into the\
** free list
*/

  ObjectProxy->PartialDestruct();
  FreeObjectProxy->delete_object(ObjectProxy);

}


/************************************************************************
 * GetMessageSize - returns the number of bytes needed to package up	*
 * an object proxy into a message					*
 ************************************************************************/
int SpObjFactory::GetMessageSize(SpObjProxy *ObjectProxy) {
  SpObjTemplate *Template;
  int i;
  int j;
  int Nats;
  int Nbyts;
  int Nattributes;
  BASE_ATTRIBUTE *Attribute;
  BASE_ATTRIBUTE *Attribute1;
  int Nbytes;
  int Nbytes1;
  char *buff1;
  int ArraySize;

  Template = ObjectProxy->GetObjectTemplate();

  Nbytes1 = FreeObjectProxy->get_size(ObjectProxy->get_freeid())
          - sizeof(SpObjProxy);

  Nbytes = sizeof(SpObjProxyMsg) + Nbytes1;

  int Offset = strlen(ObjectProxy->GetObjectName()) + 1;
  Offset += (sizeof(double) - Offset%sizeof(double))%sizeof(double);
  Nbytes += Offset;

  Nats = 0;
  Nbyts = 0;

  Nattributes = ObjectProxy->GetNattributes();
  for (i=0; i<Nattributes; i++) {
    Attribute = ObjectProxy->GetAttribute(i);
    ArraySize = Template->GetArraySize(i);
    if (Attribute) {
/*
      cout << MapAttribute(Attribute->GetType())
	   << "[" << ArraySize << "]"
	   << ", " << Attribute->GetMessageSize()
	   << endl;
*/
      if (Attribute->GetMessageSize()) {
	for (j=0; j<ArraySize; j++) {
	  buff1 = (char *)Attribute;
	  buff1 += Attribute->GetSize() * j;
	  Attribute1 = (BASE_ATTRIBUTE *)buff1;
	  Nbytes += Attribute1->GetMessageSize();
	  Nbyts += Attribute1->GetMessageSize();
	  Nats++;
	}
      }
    }
  }


//  cout << "Total bytes (GetMessageSize) = " << Nbytes << endl;

  return Nbytes;

}


/************************************************************************
 * GenerateMessage - Generate a message based on the Object Proxy.	*
 *		     works in conjunction with the Generateobjectproxy	*
 *		     method.						*
 ************************************************************************/
SpObjProxyMsg *SpObjFactory::GenerateMessage
  (SpObjProxy *ObjectProxy) {
  SpObjTemplate *Template;
  int i;
  int j;
  int Nats;
  //int Nbyts;
  int Nattributes;
  BASE_ATTRIBUTE *Attribute;
  BASE_ATTRIBUTE *Attribute1;
  int Nbytes;
  int Nbytes1;
  char *buff;
  char *buff1;
  SpObjProxyMsg *ObjectProxyMessage;
  int ArraySize;

  Template = ObjectProxy->GetObjectTemplate();

  Nbytes1 = FreeObjectProxy->get_size(ObjectProxy->get_freeid())
          - sizeof(SpObjProxy);

  Nbytes = sizeof(SpObjProxyMsg) + Nbytes1;

  int Offset = strlen(ObjectProxy->GetObjectName()) + 1;
  Offset += (sizeof(double) - Offset%sizeof(double))%sizeof(double);
  Nbytes += Offset;

  Nats = 0;
  //Nbyts = 0;

  Nattributes = ObjectProxy->GetNattributes();
  for (i=0; i<Nattributes; i++) {
    Attribute = ObjectProxy->GetAttribute(i);
    ArraySize = Template->GetArraySize(i);
    if (Attribute) {

/*
      cout << MapAttribute(Attribute->GetType())
	   << "[" << ArraySize << "]"
	   << ", " << Attribute->GetMessageSize()
	   << endl;
*/

      if (Attribute->GetMessageSize()) {
	for (j=0; j<ArraySize; j++) {
	  buff1 = (char *)Attribute;
	  buff1 += Attribute->GetSize() * j;
	  Attribute1 = (BASE_ATTRIBUTE *)buff1;
	  Nbytes += Attribute1->GetMessageSize();

	  // I commented this out since Nbyts is not used later.
	  //Nbyts += Attribute1->GetMessageSize();

	  Nats++;
	}
      }
    }
  }

/*
  cout << "Nbytes1 (Generate Message) = " << Nbytes1 << endl;
  cout << "Nbyts (Generate Message) = " << Nbyts << endl;
  cout << "Total bytes (Generate Message) = " << Nbytes << endl;
*/

  buff = new char[Nbytes];
  ObjectProxyMessage = (SpObjProxyMsg *)buff;

//  cout << "OBJECT ID " << ObjectProxy->GetObjectId() << endl;

  ObjectProxyMessage->SetNode(ObjectProxy->GetNode());
  ObjectProxyMessage->SetLocalId(ObjectProxy->GetLocalId());
  ObjectProxyMessage->SetType(ObjectProxy->GetObjectType());
  ObjectProxyMessage->SetObjectId(ObjectProxy->GetObjectId());
  ObjectProxyMessage->SetFreeID(Template->get_freeid());
  ObjectProxyMessage->SetNbytes(Nbytes);
  ObjectProxyMessage->SetNattributeUpdates(Nats);

  ObjectProxyMessage->SetObjectName(ObjectProxy->GetObjectName());
  buff = ObjectProxyMessage->GoToFirstAttributeUpdate(ObjectProxy->GetObjectName());

  for (i=0; i<Nattributes; i++) {
    Attribute = ObjectProxy->GetAttribute(i);
    if (Attribute) {
      if (Attribute->GetMessageSize()) {
	ArraySize = Template->GetArraySize(i);

//	cout << "Attribute " << i << " Array Size = " << ArraySize << endl;

	for (j=0; j<ArraySize; j++) {

	  ((ATTRIBUTE_MESSAGE_INFO *)buff)->SetArrayIndex(j);
	  buff1 = (char *)Attribute;
	  buff1 += Attribute->GetSize() * j;
	  Attribute1 = (BASE_ATTRIBUTE *)buff1;
	  ((ATTRIBUTE_MESSAGE_INFO *)buff)->SetAttributeIndex(i);
	  ((ATTRIBUTE_MESSAGE_INFO *)buff)
	    ->SetNbytes(Attribute1->GetMessageSize());
	  Attribute1->FillMessage(buff);
	  buff = ObjectProxyMessage->GoToNextAttributeUpdate();

	  if ((buff-(char *)ObjectProxyMessage) > Nbytes-Nbytes1) {
	    cout << "Really bad Error, overflowing buffer "
		 << int(buff - (char *)ObjectProxyMessage) - Nbytes + Nbytes1
		 << endl;
	  }

	}
      }
    }
  }

  buff1 = (char *)&ObjectProxy[1];
  memcpy(buff, buff1, Nbytes1);

/*
  cout << "Nbyts " << Nbyts << endl;
  cout << "Byte counting should be zero "
       << int(buff - (char *)ObjectProxyMessage) - Nbyts
          - sizeof(SpObjProxyMsg)
       << endl;
*/

  return ObjectProxyMessage;

}

/************************************************************************
 * GenerateObjectProxy - Generate an Object Proxy based on the message	*
 *                       that is passed in. This works in conjunction	*
 *                       with the GenerateMessage function.		*
 ************************************************************************/
SpObjProxy *SpObjFactory::GenerateObjectProxy
  (SpObjProxyMsg *ObjectProxyMessage) {
  SpObjProxy *ObjectProxy;
  ATTRIBUTE_MESSAGE_INFO *AttributeMessageInfo;
  SpObjTemplate *Template;
  BASE_ATTRIBUTE *Attribute;
  BASE_ATTRIBUTE *Attribute1;
  int ClassID;
  char *ClassName;
  int i;
  int AttributeType;
  int AttributeIndex;
  int AttributeArraySize;
  //int ArrayIndex;
  char *Pointer;
  char *buff;

  ClassID = ObjectProxyMessage->GetFreeID();
  ObjectProxy = (SpObjProxy *)FreeObjectProxy->new_object(ClassID);

  if (!ObjectProxy->GetInitialized()) {
    ClassName = FreeObjectProxy->GetName(ClassID);
    ObjectProxy->Construct(ClassName);
  }

  ObjectProxy->SetNode(ObjectProxyMessage->GetNode());
  ObjectProxy->SetLocalId(ObjectProxyMessage->GetLocalId());
  ObjectProxy->SetObjectType(ObjectProxyMessage->GetType());
  ObjectProxy->SetObjectId(ObjectProxyMessage->GetObjectId());
  Template = ObjectProxy->GetObjectTemplate();

  ObjectProxy->SetObjectName(ObjectProxyMessage->GetObjectName());
  Pointer = ObjectProxyMessage->
    GoToFirstAttributeUpdate(ObjectProxyMessage->GetObjectName());

  //proxy: replace: for (i=0; i<ObjectProxyMessage->GetNattributeUpdates(); i++) {
  // replaced by next 3 lines:
  i=0;
  int numAttUpdates = ObjectProxyMessage->GetNattributeUpdates();
  while (i < numAttUpdates) {

    AttributeMessageInfo = ObjectProxyMessage->GetAttributeMessageInfo();

    AttributeIndex = AttributeMessageInfo->GetAttributeIndex();
    AttributeArraySize = Template->GetArraySize(AttributeIndex);
    AttributeType = Template->GetAttributeTypeID(AttributeIndex);

//    cout << "Attribute Type ID = " << AttributeType << endl;

    Attribute = ObjectProxy->GetAttribute(AttributeIndex);
    if (!Attribute) {
      Attribute = CreateAttribute(AttributeType, AttributeArraySize);
//      ObjectProxy->Define(Attribute,AttributeIndex);
    }

    //proxy:  start changes
    int j, attSize = Attribute->GetSize();
    buff = (char *)Attribute;
    for(j=0; j<AttributeArraySize; j++) {
      Attribute1 = (BASE_ATTRIBUTE *)buff;
      Attribute1->EmptyMessage(Pointer);

      Pointer = ObjectProxyMessage->GoToNextAttributeUpdate();
      buff += attSize;
      i++;
    }
    // want to Define ONCE/attrib, and AFTER it has been filled in via EmptyMessage
    ObjectProxy->Define(Attribute,AttributeIndex);
    //proxy: end changes


  }

  return ObjectProxy;
}

/************************************************************************
 * CreateAttribute - Create an attribute object based on type and the	*
 * number of attributes to be created as an array	.		*
 ************************************************************************/
BASE_ATTRIBUTE *SpObjFactory::CreateAttribute(int type, int n) {

  switch(type) {
    case STATIC_INT_ATTRIBUTE_TYPE:
      return new STATIC_INT_ATTRIBUTE[n];

    case STATIC_FLOAT_ATTRIBUTE_TYPE:
      return new STATIC_FLOAT_ATTRIBUTE[n];

    case STATIC_LOGICAL_ATTRIBUTE_TYPE:
      return new STATIC_LOGICAL_ATTRIBUTE[n];

    case STATIC_POSITION_ATTRIBUTE_TYPE:
      return new STATIC_POSITION_ATTRIBUTE[n];

    case DYNAMIC_INT_ATTRIBUTE_TYPE:
      return new DYNAMIC_INT_ATTRIBUTE[n];

    case DYNAMIC_FLOAT_ATTRIBUTE_TYPE:
      return new DYNAMIC_FLOAT_ATTRIBUTE[n];

    case DYNAMIC_LOGICAL_ATTRIBUTE_TYPE:
      return new DYNAMIC_LOGICAL_ATTRIBUTE[n];

    case DYNAMIC_OBJECT_ATTRIBUTE_TYPE:
      return new DYNAMIC_OBJECT_ATTRIBUTE[n];

    case DYNAMIC_POSITION_ATTRIBUTE_TYPE:
      return new DYNAMIC_POSITION_ATTRIBUTE[n];

    case STRING_ATTRIBUTE_TYPE:
      return new STRING_ATTRIBUTE[n];

    case OBJECT_ATTRIBUTE_TYPE:
      return new OBJECT_ATTRIBUTE[n];

    case LIST_ATTRIBUTE_TYPE:
      return new LIST_ATTRIBUTE[n];

    default:

      cout << "Error (CreateAttribute), unknown attribute type "
	   << type << endl;
  }

  return NULL;

}

/************************************************************************
 * Clone - Creates a copy of the Object Proxy that is passed in		*
 ************************************************************************/
SpObjProxy *SpObjFactory::Clone(SpObjProxy *Op) {
  int i, j;
  SpObjProxy *ObjectProxy;
  SpObjTemplate *Template;
  int ClassID;
  char *ClassName;
  int AttributeArraySize;
  int AttributeType;
  BASE_ATTRIBUTE *Attribute1;
  BASE_ATTRIBUTE *Attribute;
  BASE_ATTRIBUTE *AttributeOp;

  ClassID = Op->get_freeid();
  ObjectProxy = (SpObjProxy *)FreeObjectProxy->new_object(ClassID);

  if (!ObjectProxy->GetInitialized()) {
    ClassName = FreeObjectProxy->GetName(ClassID);
    ObjectProxy->Construct(ClassName);
  }

  ObjectProxy->SetNode(Op->GetNode());
  ObjectProxy->SetLocalId(Op->GetLocalId());
  ObjectProxy->SetObjectType(Op->GetObjectType());
  ObjectProxy->SetObjectId(Op->GetObjectId());
  ObjectProxy->SetObjectName(Op->GetObjectName());

  Template = ObjectProxy->GetObjectTemplate();

  memcpy((char *)&ObjectProxy[1],
	 (char *)&Op[1],
	 FreeObjectProxy->get_size(ClassID) - sizeof(SpObjProxy));

  for (i=0; i<Template->GetNattributes(); i++) {

    AttributeArraySize = Template->GetArraySize(i);
    AttributeType = Template->GetAttributeTypeID(i);

    Attribute1 = ObjectProxy->GetAttribute(i);
    if (!Attribute1) {
      Attribute1 = CreateAttribute(AttributeType, AttributeArraySize);
    }

    Attribute = Attribute1;
    for (j=0; j<AttributeArraySize; j++) {

      AttributeOp = Op->GetAttribute(i,j);
      Attribute->Copy(AttributeOp);

      char *Buff = (char *)Attribute;
      Buff += Attribute->GetSize();
      Attribute = (BASE_ATTRIBUTE *)Buff;

    }

    ObjectProxy->Define(Attribute1,i);

  }

  return ObjectProxy;

}
