// SpObjFactory.H Header File

#ifndef ObjectFactory_Object
#define ObjectFactory_Object

#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpString.H"
#include "SpStdlib.H"

#include "SpDataParser.H"
#include "hash.H"
#include "SpExportAttribute.H"
#include "SpFreeObjProxy.H"
#include "SpObjProxy.H"
#include "SpObjTemplate.H"
#include "SpObjProxyMsg.H"

  // forward declaration
class C_QUEUE;


class SpObjFactory {

  private:

    SpDataParser *Parser;             // Reads in Objects.par proxy file
    C_HASH *Templates;               // Name-based hash table for proxy classes
    FREE_OBJECT_PROXY *FreeObjectProxy; // Supports user-defined
                                        // descendants of ObjectProxy
    char **FreeNames;           // Array of proxy class names
                                // Index is the free list id for that proxy

    static int Node;   // my node id
    static int Nnodes; // number of nodes in simulation

  protected:

  public:

    SpObjFactory(char *FileName);
    ~SpObjFactory();

    // return all template
    C_QUEUE *GetObjTemplate(void);
    
    SpObjTemplate *GetTemplate(char *ObjectType, void *ObjectMark = NULL);

    SpObjProxy *Create(char *type);
    void Destroy(SpObjProxy *ObjectProxy);
    void PartialDestroy(SpObjProxy *ObjectProxy);

    void Define(void (*Method)(void *, va_list &),
		char *MethodName,
		char *ObjectType);

    void PrintStream(char *ObjectType, ostream &out = cout) {
      SpObjTemplate *ObjectTemplate = GetTemplate(ObjectType);
      if (ObjectTemplate) {
	ObjectTemplate->PrintStream(out);
      }else{
	out << "Error (printstream): Bad object type " << ObjectType << endl;
      }
    }

    int GetMessageSize(SpObjProxy *ObjectProxy);
    SpObjProxyMsg *GenerateMessage(SpObjProxy *ObjectProxy);
    SpObjProxy *GenerateObjectProxy(SpObjProxyMsg *Message);

    BASE_ATTRIBUTE *CreateAttribute(int Type, int n);

    SpObjProxy *Clone(SpObjProxy *Op);

    static void SetNode(int n) {Node = n;}
    static void SetNnodes(int n) {Nnodes = n;}

};

typedef SpObjFactory OBJECT_FACTORY;

extern SpObjFactory* ObjFact;

#endif
