// SpObjProxy.C Source File

int EXTERNAL_OBJECT_PROXY = 0;

#include "SpObjProxy.H"
#include "SpObjFactory.H"
#include "SpObjProxyMsg.H"
#include "F_SpUpdateListItem.H"
#include "SpDestination.H"
#include "SpScheduleEvent.H"

EVENT_TYPE (REFLECT_TWO_WAY_ATTRIBUTES);

#include "SpFreeHLA.H"
extern SpFreeHLA *FreeHLA;

extern SpObjFactory *ObjFact;
SpObjProxy *CurrentObjectProxy = NULL;

#include "SpDynItem.H"
#include "SpFreeDynAttributes.H"
#include "SpAssert.H" //proxy:
extern SpFreeDynAttributes *FreeDynamicAttributes;

#include "RB_SpDefineClass.H"

RB_DEFINE_CLASS(RB_UpdateXqueue);
#ifdef __IBMCPP__
namespace {
#endif
RB_DEFINE_CLASS_PRIMATIVE(char);
#ifdef __IBMCPP__
}
#endif

MONOTONICALLY_INCREASING_BUFFER SpObjProxy::MIB;

SpObjProxy::SpObjProxy() {
  ObjectTemplate = NULL;
  BaseAttribute = NULL;
  ObjectName = NULL;
  SetObjectName("Unnamed");
  Nattributes = 0;
  Defined = 0;
  //FullDestroy_p = 0;
  InObjectAttribute = NULL;
  ObjectId = -1;
  Node = -1;
  ObjectType = -1;
  LocalId = -1;
  SetAttributeUpdateList(NULL); //proxy: add
}
// proxy:
// Delete list of attribute updates
void SpObjProxy::DeleteUpdateList(void) {
  RB_UpdateXqueue *attUpdateList = GetAttributeUpdateList();
  if (attUpdateList!=NULL) {
    int numUpdateItems = attUpdateList->get_length();
    if (numUpdateItems > 0) {
      cout << "WARNING:  SpObjProxy::DeleteUpdateList() deleting NON-EMPTY update list!" 
	   << endl;


      C_ITEM *updateItem;

      // delete all update items on the list:
      while (numUpdateItems > 0) {
	updateItem = (C_ITEM *) attUpdateList->pop_top();
	RB_FREE_DELETE(FreeHLA, updateItem );
	numUpdateItems--;
      }
    }
    
    /////    delete AttributeUpdateList; RB_NEW_RB_UpdateXqueue();
    RB_DELETE_RB_UpdateXqueue(attUpdateList);

    AttributeUpdateList = NULL;
  }
}


void SpObjProxy::free_ObjectName()
{
  if (ObjectName)
    free(ObjectName);

  ObjectName = NULL;
}

SpObjProxy::~SpObjProxy()
{
  free_ObjectName();
  if (RootProxyP()) { //proxy:
    DeleteUpdateList();
  }
}

void SpObjProxy::Define(BASE_ATTRIBUTE &Attribute, char *Name) {
  Defined = 1;
  int Found = 0;
  for (int i=0; i<Nattributes; i++) {
    if (!strcmp(ObjectTemplate->GetAttributeName(i), Name)) {
      Define(&Attribute, i);
      Found = 1;
      break;
    }
  }
  if (!Found) {
    cout << "Error (Define): Could not find attribute "
	 << Name << " in Object Proxy " << ObjectName << endl;
  }
}

void SpObjProxy::Define(BASE_ATTRIBUTE *Attribute, int AttributeIndex){
  BaseAttribute[AttributeIndex] = Attribute;
  char *Att = (char *)Attribute;
  for (int i=0; i<ObjectTemplate->GetArraySize(AttributeIndex); i++) {
    ((BASE_ATTRIBUTE *)Att)->SetAttributeIndex(AttributeIndex);
    ((BASE_ATTRIBUTE *)Att)->SetArrayIndex(i);
    ((BASE_ATTRIBUTE *)Att)->SetInObjectProxy(this);
    Att += Attribute->GetSize();
  }
  // proxy::  new code here to end-of-fcn:
  // If this is the root proxy, make sure it has an updatelist:
  if ((RootProxyP()) && (GetAttributeUpdateList()==NULL)) {
    RB_UpdateXqueue *updateList =  RB_NEW_RB_UpdateXqueue();
    updateList->reset();
    SetAttributeUpdateList(updateList);
  }

  int attribType = Attribute->GetType();

  // If this attrib contains a proxy (eg OBJECT_ATTRIB), we may need to assign
  // it (and all subsidiary proxies) a ptr to the root proxy's updateList (this
  // is so there is only one updateList for the entire tree).  Do
  // this in 2 cases: 1) this is the root proxy, or 2) we're connecting this
  // proxy subtree onto a proxy that already has the updateList pointer (ie we
  // Define this attrib AFTER the attrib that owns this one has already been
  // Defined).
  if (ContainsProxy(attribType)) {  //proxy:
    if (GetAttributeUpdateList()!=NULL) {
      AssignUpdateListPointers(AttributeIndex, attribType);
    }
  }
}

/************* proxy:  debug code
void SpObjProxy::checkForBadUpdateList(void) {
  RB_UpdateXqueue *attributeUpdateList = GetAttributeUpdateList();
  if (attributeUpdateList!=NULL) {
    int len = attributeUpdateList->get_length();
    if ( (len < 0) || (len > 200) ||
	 ((len > 0) && (attributeUpdateList->get_top()==NULL)) ) {
      cout << "Ick! looks like a bad update list!?" << endl;
    }
  }
}
**************/

//proxy:
// Instantiate the update list, if nec, and store in this (the root) proxy.
// Run down the "branch" of the proxy tree denoted by arg, attribNum, inserting
// a pointer to the updateList in all subsidiary proxies.
void SpObjProxy::AssignUpdateListPointers(int attribNum, int attribTypeID) {
  RB_UpdateXqueue *updateList = GetAttributeUpdateList();
  assert(updateList!=NULL);

  // Find proxies in this branch of the tree (if any), and give them
  // the updateList pointer (so all attrib changes get stored in one place):
  SetAttribUpdatePtrs(updateList, attribNum, attribTypeID);
}

// proxy:
// Run down ALL branches (ie attributes) of this proxy, and insert
// the updateList pointer into all subsidiary proxies
void SpObjProxy::SetProxyUpdatePtrs(RB_UpdateXqueue *updateList) {
  int i, attType;
  for (i=0; i<Nattributes; i++) {
    attType = ObjectTemplate->GetAttributeTypeID(i);
    SetAttribUpdatePtrs(updateList, i, attType);
  }
}

// proxy:
// Climb down branch of "proxy tree", putting the arg, "updateList", into all proxies visited.
// Proxies can form a tree because attributes representing objects (OBJECT_ATTRIBUTE, 
// DYNAMIC_OBJECT_ATTRIBUTE, and LIST_ATTRIBUTE), contain proxies themselves.
// The point of this is that we want to have only one update list for the entire
// proxy tree.  This way, all attrib changes get stored in the one update list.
void SpObjProxy::SetAttribUpdatePtrs(RB_UpdateXqueue *updateList,
				     int attribNum, int attribTypeID) {
  int j, arrSize;
  SpObjProxy* ObjectProxy;
  LIST_ATTRIBUTE *ListAttribute;
  OBJECT_ATTRIBUTE *ObjectAttribute;

  if (BaseAttribute[attribNum]) {
    switch (attribTypeID) {
      case OBJECT_ATTRIBUTE_TYPE:
      case DYNAMIC_OBJECT_ATTRIBUTE_TYPE:
	arrSize = ObjectTemplate->GetArraySize(attribNum);
	for(j=0; j<arrSize; j++) {
	  ObjectAttribute = (OBJECT_ATTRIBUTE *) GetAttribute(attribNum, j);
	  ObjectProxy = ObjectAttribute->GetObjectProxy();
	  if (ObjectProxy == NULL) {
	    cout << "Error (SetAttribUpdatePtrs): No Object Proxy in Attribute["
		 << attribNum << "," << j << "]" << endl;
	    break;
	  }
	  if (ObjectProxy->GetInObjectAttribute() != ObjectAttribute) {
	    ObjectProxy->SetInObjectAttribute(ObjectAttribute);
	    cout << "Warning (SetAttribUpdatePtrs): Bad InAttribute (repaired)" << endl;
	  }
	  ObjectProxy->SetAttributeUpdateList(updateList);
	  ObjectProxy->SetProxyUpdatePtrs(updateList);
	}

	break;

      case LIST_ATTRIBUTE_TYPE:
	ListAttribute = (LIST_ATTRIBUTE *)BaseAttribute[attribNum];
	ObjectAttribute = (OBJECT_ATTRIBUTE *)ListAttribute->get_top();
	while (ObjectAttribute) {

	  ObjectProxy = ObjectAttribute->GetObjectProxy();
	  if (ObjectProxy == NULL) {
	    cout << "Error (SetAttribUpdatePtrs): No Object Proxy in List Attribute "
		 << "(could not repair)"
		 << endl;
	    break;
	  }


	  if (ObjectProxy->GetInObjectAttribute() != ObjectAttribute) {
	    ObjectProxy->SetInObjectAttribute(ObjectAttribute);
	    cout << "Warning (SetAttribUpdatePtrs): Bad List InAttribute (repaired)"
		 << endl;
	  }

	  if (ObjectAttribute->GetInListAttribute() != ListAttribute) {
	    ObjectAttribute->SetInListAttribute(ListAttribute);
	    cout << "Warning (SetAttribUpdatePtrs): Bad List Attribute (repaired)"
		 << endl;
	  }


	  ObjectProxy->SetAttributeUpdateList(updateList);
	  ObjectProxy->SetProxyUpdatePtrs(updateList);

	  ObjectAttribute = (OBJECT_ATTRIBUTE *) ++(*ListAttribute);
	}
	break;

      default:
	break;
    }
  }
  else{
    cout << "Warning (SetAttribUpdatePtrs): No Attribute Defined "
	 << "(could not repair)"
	 << endl;

  }

}


// proxy::  check to make sure update ptrs are correct
void SpObjProxy::CheckProxyUpdatePtrs(RB_UpdateXqueue *updateList) {
  int i;
  for (i=0; i<Nattributes; i++) {
    CheckAttribUpdatePtrs(updateList, i);
  }
}


// proxy::  check to make sure update ptrs are correct
void SpObjProxy::CheckAttribUpdatePtrs(RB_UpdateXqueue *updateList, int attribNum) {
  int j=0, arrSize;
  SpObjProxy* ObjectProxy;
  LIST_ATTRIBUTE *ListAttribute;
  OBJECT_ATTRIBUTE *ObjectAttribute;
  RB_UpdateXqueue *localUpdateList;

  if (BaseAttribute[attribNum]) {
    //    cout << "  << Check Attrib " << ObjectTemplate->ObjProxyGetClassName() << "."
    //	 << ObjectTemplate->GetAttributeName(attribNum) << " >>" << endl;
    switch (ObjectTemplate->GetAttributeTypeID(attribNum)) {
      case OBJECT_ATTRIBUTE_TYPE:
      case DYNAMIC_OBJECT_ATTRIBUTE_TYPE:
	arrSize = ObjectTemplate->GetArraySize(attribNum);
	for(j=0; j<arrSize; j++) {
	  ObjectAttribute = (OBJECT_ATTRIBUTE *) GetAttribute(attribNum, j);
	  ObjectProxy = ObjectAttribute->GetObjectProxy();
	  if (ObjectProxy == NULL) {
	    cout << "Error (CheckAttribUpdatePtrs): No Object Proxy in Attribute# "
		 << attribNum << endl;
	    break;
	  }
	  if (ObjectProxy->GetInObjectAttribute() != ObjectAttribute) {
	    ObjectProxy->SetInObjectAttribute(ObjectAttribute);
	    cout << "Warning (CheckAttribUpdatePtrs): Bad InAttribute (repaired)" << endl;
	  }
	  localUpdateList = ObjectProxy->GetAttributeUpdateList();
	  if (localUpdateList != updateList) {
	    cout << "==> CheckAttribUpdatePtrs error:  obj or dynObj attrib[" << attribNum
		 << "," << j << "] has diff updateList ptr!" << endl;
	  }
	  //	  cout << "--Proxy Type: " << ObjectTemplate->ObjProxyGetClassName() << ", attrib type: "
	  //	       << ObjectTemplate->GetAttributeTypeID(attribNum) << endl;
	  //	  cout << "--local ptr = " << localUpdateList << ";  parent ptr = " << updateList << endl;

	  ObjectProxy->CheckProxyUpdatePtrs(localUpdateList);
	}

	break;

      case LIST_ATTRIBUTE_TYPE:
	ListAttribute = (LIST_ATTRIBUTE *)BaseAttribute[attribNum];
	ObjectAttribute = (OBJECT_ATTRIBUTE *)ListAttribute->get_top();
	while (ObjectAttribute) {

	  ObjectProxy = ObjectAttribute->GetObjectProxy();
	  if (ObjectProxy == NULL) {
	    cout << "Error (CheckAttribUpdatePtrs): No Object Proxy in List Attribute "
		 << "(could not repair)"
		 << endl;
	    break;
	  }

	  if (ObjectProxy->GetInObjectAttribute() != ObjectAttribute) {
	    ObjectProxy->SetInObjectAttribute(ObjectAttribute);
	    cout << "Warning (CheckAttribUpdatePtrs): Bad List InAttribute (repaired)"
		 << endl;
	  }

	  if (ObjectAttribute->GetInListAttribute() != ListAttribute) {
	    ObjectAttribute->SetInListAttribute(ListAttribute);
	    cout << "Warning (CheckAttribUpdatePtrs): Bad List Attribute (repaired)"
		 << endl;
	  }

	  localUpdateList = ObjectProxy->GetAttributeUpdateList();
	  if (localUpdateList != updateList) {
	    cout << "==> CheckAttribUpdatePtrs error:  list attrib[" << attribNum
		 << "," << j << "] has diff updateList ptr!" << endl;
	  }
	  //	  cout << "--Proxy Type: " << ObjectTemplate->ObjProxyGetClassName() << ", attrib type: "
	  //	       << ObjectTemplate->GetAttributeTypeID(attribNum) << endl;
	  //	  cout << "--local ptr = " << localUpdateList << ";  parent ptr = " << updateList << endl;

	  ObjectProxy->CheckProxyUpdatePtrs(localUpdateList);

	  ObjectAttribute = (OBJECT_ATTRIBUTE *) ++(*ListAttribute);
	  j++;
	}
	break;

      default:
	break;
    }
  }
  else{
    cout << "Warning (CheckAttribUpdatePtrs): No Attribute Defined "
	 << "(could not repair)"
	 << endl;

  }

}



void SpObjProxy::Execute(char *MethodName, ...) {
  SpObjTemplate *OT;
  va_list va;
  int status;

  va_start(va, MethodName);

  OT = ObjectTemplate;
  status = 0;
  while (OT) {
    if (OT->Execute(this, MethodName, va)) {
      status = 1;
      break;
    }
    OT = OT->GetParent();
  }

  va_end(va);

  if (status == 0) {
    cout << "Error in SpObjProxy "
	 << ObjectTemplate->ObjProxyGetClassName()
	 << ": Undefined method " << MethodName << endl;
    exit(0);
  }

}

int SpObjProxy::GetReference(const char *name) const{
  for (int i=0; i<Nattributes; i++) {
    if (!strcmp(name, ObjectTemplate->GetAttributeName(i))) {
      return i;
    }
  }
  cout << "Error, could not find " << name << endl;
  return -1;
}

BASE_ATTRIBUTE *SpObjProxy::Find(const int AttributeIndex, int ArrayIndex) const{
  if (0 > AttributeIndex || AttributeIndex >= Nattributes)
    return NULL;

  BASE_ATTRIBUTE *a = BaseAttribute[AttributeIndex];

  if (a == NULL)
    return NULL;

  return (BASE_ATTRIBUTE*) (((char*)a) + ArrayIndex * a->GetSize());
}

void SpObjProxy::FillMessage(F_UPDATE_LIST_ITEM *top,
			       int len,
			       char *&Buffer,
			       int &Nbytes)
{
  int i;
  F_UPDATE_LIST_ITEM *UpdateListItem;
  BASE_ATTRIBUTE *Attribute;
  char *Buff;
  OBJECT_ATTRIBUTE *ObjectAttribute;
  SpDynItem *DynamicItem;
  char *buf;
  //int Natts;
  int *NlevelInfo;
  PROXY_LEVEL_INFO *ProxyLevelInfo;

/*
  cout << "Schedule Update Event being processed for OBJECT_TYPE "
       << OBJECT_TYPE << " LOCAL_ID " << LOCAL_ID
       << " SpSimTime " << TIME_TAG
       << endl;
*/

//  cout << "Schedule Update Event being processed" << endl;

  //Natts = len;

/*
** First, find out how many bytes we need to send the updated attributes
*/

  Nbytes = 0;
  NlevelInfo = new int[len];
  int *NbytesArray = new int[len];
  //for (i=0; i<len; i++) NlevelInfo[i] = 1;

  UpdateListItem = top;
  for (i=0; i<len; i++) {

    Attribute = UpdateListItem->GetAttribute();
    NlevelInfo[i] = Attribute->GetDepth();

//    cout << "NlevelInfo[" << i << "] = " << NlevelInfo[i] << endl;

    Nbytes += NlevelInfo[i]*sizeof(PROXY_LEVEL_INFO);

    if (UpdateListItem->GetAddRemove() == NEITHER) {

// Normal Attribute

      NbytesArray[i] = Attribute->GetMessageSize();

    }else{

      if (Attribute->GetType() == LIST_ATTRIBUTE_TYPE) {

// List Attribute (adding/removing an OBJECT_ATTRIBUTE)

	if (UpdateListItem->GetAddRemove() == OBJECT_PROXY_ADD) {

	  ObjectAttribute = UpdateListItem->GetPtr().ObjectAttribute;
	  NbytesArray[i] =  ObjectAttribute->GetMessageSize();

	}else{
	  assert(UpdateListItem->GetAddRemove() == REMOVE);
	  NbytesArray[i] = sizeof(ATTRIBUTE_MESSAGE_INFO);

	}

      }else{

// Dynamic Attribute (adding/removing a SpDynItem)

	if (UpdateListItem->GetAddRemove() == OBJECT_PROXY_ADD) {

	  DynamicItem = UpdateListItem->GetPtr().DynamicItem;
	  NbytesArray[i] = sizeof(ATTRIBUTE_MESSAGE_INFO) +
	            FreeDynamicAttributes->get_size(DynamicItem);

	}else{

	  assert(UpdateListItem->GetAddRemove() == REMOVE);
	  NbytesArray[i] = sizeof(ATTRIBUTE_MESSAGE_INFO);

	}
      }

    }
    Nbytes += NbytesArray[i];

    UpdateListItem = (F_UPDATE_LIST_ITEM *)UpdateListItem->get_link();

  }

/*
** Then create the message buffer and then fill it in with the updated
** attributes.
*/

//  Buffer = new char[Nbytes];
  Buffer = RB_NEW_ARRAY_char(Nbytes);

  for (i=0; i<Nbytes; i++) Buffer[i] = 0;
  Buff = Buffer;

  UpdateListItem = top;

  for (i=0; i<len; i++) {


    Attribute = UpdateListItem->GetAttribute();
    ProxyLevelInfo = (PROXY_LEVEL_INFO *)Buff;
    Buff = Attribute->PackLevelInfo(Buff, NlevelInfo[i]);


// Check that the level info was packed correctly

/*
    cout << endl;
    for (int j=0; j<NlevelInfo[i]; j++) ProxyLevelInfo[j].print();
*/

#ifdef ERROR_CHECK
    // I added this #ifdef.  Am I correct that this code is
    // just for checking for errors.  I.e. is it merely to make sure
    // that PackLevelInfo ran correctly?
    SpObjProxy *Op = FindProxy(ProxyLevelInfo);
    if (Op) {
      BASE_ATTRIBUTE *Att = Op->GetAttribute(Attribute->GetAttributeIndex(),
					     Attribute->GetArrayIndex());
      if (Att != Attribute) {
	cout << "Error (SpObjProxy::FillMessage): Bad Lookup" << endl;
      }
    }else{
/*
      cout << "Error (SpObjProxy::FillMessage): Could not find Object Proxy"
	   << endl;
*/
    }
#endif
    
    ((ATTRIBUTE_MESSAGE_INFO *)Buff)->
      SetArrayIndex(Attribute->GetArrayIndex());

    ((ATTRIBUTE_MESSAGE_INFO *)Buff)->
      SetAddRemove(UpdateListItem->GetAddRemove());

    if (UpdateListItem->GetAddRemove() == NEITHER) {

// Normal Attribute

      Attribute->FillMessage(Buff);

    }else{

      if (Attribute->GetType() == LIST_ATTRIBUTE_TYPE) {

// List Attribute (adding/removing an OBJECT_ATTRIBUTE)

	ObjectAttribute = UpdateListItem->GetPtr().ObjectAttribute;

	if (UpdateListItem->GetAddRemove() == OBJECT_PROXY_ADD) {

	  ObjectAttribute->
	    FillMessage(Buff);

	}else{
	  assert(UpdateListItem->GetAddRemove() == REMOVE);
	  
	  //((ATTRIBUTE_MESSAGE_INFO *)Buff)->
	  //  SetNbytes(sizeof(ATTRIBUTE_MESSAGE_INFO));

	}

	((ATTRIBUTE_MESSAGE_INFO *)Buff)->
	  SetAddRemoveId(ObjectAttribute->GetListId());

      }else{

// Dynamic Attribute (adding/removing a SpDynItem)

	DynamicItem = UpdateListItem->GetPtr().DynamicItem;

	if (UpdateListItem->GetAddRemove() == OBJECT_PROXY_ADD) {

	  //((ATTRIBUTE_MESSAGE_INFO *)Buff)->
	  //  SetNbytes(sizeof(ATTRIBUTE_MESSAGE_INFO) +
	  //	      FreeDynamicAttributes->get_size(DynamicItem));

	  buf = Buff + sizeof(ATTRIBUTE_MESSAGE_INFO);

	  memcpy(buf, (char *)DynamicItem,
		 FreeDynamicAttributes->get_size(DynamicItem));

	}else{
	  assert(UpdateListItem->GetAddRemove() == REMOVE);

	  //((ATTRIBUTE_MESSAGE_INFO *)Buff)->
	  //  SetNbytes(sizeof(ATTRIBUTE_MESSAGE_INFO));

	}
	((ATTRIBUTE_MESSAGE_INFO *)Buff)->
	  SetAddRemoveId(DynamicItem->GetCountId());

      }

    }

    ((ATTRIBUTE_MESSAGE_INFO *)Buff)->
      SetAttributeIndex(Attribute->GetAttributeIndex());

    ((ATTRIBUTE_MESSAGE_INFO *)Buff)->SetNbytes(NbytesArray[i]);


    Buff += ((ATTRIBUTE_MESSAGE_INFO *)Buff)->GetNbytes();
  
    UpdateListItem = (F_UPDATE_LIST_ITEM *)UpdateListItem->get_link();

  }

  delete [] NbytesArray;
  delete [] NlevelInfo;

  if (int(Buff-Buffer) != Nbytes) {
    cout << "(E_ScheduleUpdate): Major League Error "
	 << int(Buff-Buffer) << " should be " << Nbytes << ", Diff "
	 << int(Buff-Buffer) - Nbytes
	 << endl;
  }

//  delete Buffer;

}

SpObjProxy::EMPTY_MESSAGE_RESULT EmptyMessageResultRC;

SpObjProxy::EMPTY_MESSAGE_RESULT& SpObjProxy::EmptyMessage(const char *Buff)
{
    OBJECT_ATTRIBUTE *ObjectAttribute;
    LIST_ATTRIBUTE *ListAttribute;
    SpDynItem *DynamicItem;
    BASE_DYNAMIC_ATTRIBUTE *DynamicAttribute;
    int Level;
    SpObjProxy *Op;
    int AttributeIndex;
    int ArrayIndex;
    PROXY_LEVEL_INFO *ProxyLevelInfo;
    ATTRIBUTE_MESSAGE_INFO *Info;

    EmptyMessageResultRC.OpWarning = 0;
    EmptyMessageResultRC.Buff = Buff;
    
    ProxyLevelInfo = (PROXY_LEVEL_INFO *)EmptyMessageResultRC.Buff;
    Level = ProxyLevelInfo->Level;

    EmptyMessageResultRC.Buff += Level*sizeof(PROXY_LEVEL_INFO);
    Info = (ATTRIBUTE_MESSAGE_INFO *)EmptyMessageResultRC.Buff;
    
    AttributeIndex = Info->GetAttributeIndex();
    ArrayIndex = Info->GetArrayIndex();

/*
    cout << endl;
    for (int j=0; j<Level; j++) ProxyLevelInfo[j].print();
*/

    Op = FindProxy(ProxyLevelInfo);

    if (Op) {
      EmptyMessageResultRC.Attribute =
	Op->GetAttribute(AttributeIndex, ArrayIndex);
      if (EmptyMessageResultRC.Attribute) {

	if (Info->GetAddRemove() == NEITHER) {
/*
** Handle a Normal Attribute (i.e., anything that does not add or remove an
** item from a list attribute.
*/

/*
	  cout << i << " Reflecting Normal Attribute: AttributeIndex " << AttributeIndex
	       << ", ArrayIndex " << ArrayIndex
	       << ", Bytes " << Info->GetNbytes()
	       << endl;
*/

	  EmptyMessageResultRC.Attribute
	    ->EmptyMessage(EmptyMessageResultRC.Buff);

	}else{

/*
	  cout << i <<  " Reflecting list Attribute: AttributeIndex " << AttributeIndex
	       << ", ArrayIndex " << ArrayIndex
	       << ", Bytes " << Info->GetNbytes()
	       << endl;
*/

	  if (EmptyMessageResultRC.Attribute->GetType() == LIST_ATTRIBUTE_TYPE) {

/*
** Handle a list attribute that contains object attributes.
*/

	    ListAttribute = (LIST_ATTRIBUTE *)EmptyMessageResultRC.Attribute;

	    if (Info->GetAddRemove() == OBJECT_PROXY_ADD) {

//	      cout << "Adding an Object Attribute to a List" << endl;

	      //ObjectAttribute = (OBJECT_ATTRIBUTE *) // bad lines
	      //  ObjFact->CreateAttribute(OBJECT_ATTRIBUTE_TYPE,1);

	      ObjectAttribute = new OBJECT_ATTRIBUTE; // better line
	      ObjectAttribute->EmptyMessage(EmptyMessageResultRC.Buff);

	      *ListAttribute += ObjectAttribute;

	    }else{
	      assert(Info->GetAddRemove() == REMOVE);

//	      cout << "Removing an Object Attribute from a List" << endl;

	      ObjectAttribute =
		ListAttribute->FindObjectAttribute(Info->GetAddRemoveId());

	      if (ObjectAttribute) {
		*ListAttribute -= ObjectAttribute;
	      }else{
/*
		cout << "Error (ReflectAttributes): Object not in list"
		     << endl;
*/
	      }

	    }

	  }else{

/*
** Handle a dynamic attribute that contains dynamic items.
*/

	    DynamicAttribute =
	      (BASE_DYNAMIC_ATTRIBUTE *)EmptyMessageResultRC.Attribute;

	    if (Info->GetAddRemove() == OBJECT_PROXY_ADD) {

//	      cout << "Adding a dynamic item to a Dynamic Attribute" << endl;

	      DynamicItem = (SpDynItem *)
		FreeDynamicAttributes->generate((char *)&Info[1]);

	      *DynamicAttribute += DynamicItem;

	    }else{
/*
	      cout << "Removing a dynamic item from a Dynamic Attribute"
		   << endl;
*/

	      DynamicItem = 
		DynamicAttribute->FindDynamicItem(Info->GetAddRemoveId());

	      if (DynamicItem) {
		*DynamicAttribute -= DynamicItem;
	      }else{
		cout << "Error (ReflectAttributes): "
		     << "Dynamic Item not in Dynamic Attribute"
		     << endl;
	      }

	    }

	  }

	}

      }else{
/*
** Error: Could not find Attribute in Op
*/
	  cout << "Error (ReflectAttributes): Could Not Find Attribute"
	       << endl;
      }

    }else{

/*
** Warning: Could not find Object Proxy (Op) but that may be Ok since the
** Object Proxy will be created later in this event in an Object Attribute.
** This should not be a problem, but we should verify to make sure.
*/

      EmptyMessageResultRC.OpWarning = 1;

    }

/*
** End of the loop where we advance Buff by the number of bytes in the Info
*/

    EmptyMessageResultRC.Buff += Info->GetNbytes();

    return EmptyMessageResultRC;
}

void SpObjProxy::UpdateRemoteAttribute
  (const char * const name, const int index,
   const char * const fill_message_buff, const int fill_mess_buff_size)
{
  assert(!EXTERNAL_OBJECT_PROXY);
  BASE_ATTRIBUTE *Attribute = Find(name, index);
  if (Attribute == NULL)
  {
    cerr << "Warning, could not find ATTRIBUTE " << name << endl;
    return;
  }
  
/*
** First, find out how many bytes we need to send the updated attributes
*/
  const int NlevelInfo = Attribute->GetDepth();
  
  const int message_buffer_size
    = fill_mess_buff_size + NlevelInfo*sizeof(PROXY_LEVEL_INFO);
  
/*
** Then create the message buffer and then fill it in with the updated
** attributes.
*/

  static MONOTONICALLY_INCREASING_BUFFER UpdateRemoteAttribute_MIB;
  const char * const message_buffer
    = UpdateRemoteAttribute_MIB.need(message_buffer_size);
  
  //for (int i=0; i<NBytes; i++) Buffer[i] = 0;
  //r.Buff = Buffer;
  
  //ProxyLevelInfo = (PROXY_LEVEL_INFO *)Buff;
  char *attribute_message_info_buff
    = Attribute->PackLevelInfo(message_buffer, NlevelInfo);

  assert(attribute_message_info_buff - message_buffer + fill_mess_buff_size
	 == message_buffer_size);
  
// Check that the level info was packed correctly
// ...

    SpObjProxy *Op
      = FindProxy((PROXY_LEVEL_INFO*) message_buffer);
    if (Op) {
      BASE_ATTRIBUTE *Att = Op->GetAttribute(Attribute->GetAttributeIndex(),
					     Attribute->GetArrayIndex());
      if (Att != Attribute) {
	cout << "Error (SpObjProxy::FillMessage): Bad Lookup" << endl;
      }
    }else{
      cout << "Error (SpObjProxy::FillMessage): Could not find Object Proxy"
	   << endl;
    }
    
  memcpy(attribute_message_info_buff, fill_message_buff, fill_mess_buff_size);

  ATTRIBUTE_MESSAGE_INFO *attribute_message_info =
    (ATTRIBUTE_MESSAGE_INFO*) attribute_message_info_buff;
  attribute_message_info->SetAttributeIndex(Attribute->GetAttributeIndex());
  attribute_message_info->SetArrayIndex(Attribute->GetArrayIndex());
  attribute_message_info->SetNbytes(fill_mess_buff_size);

  // We will probably have to change this if we get two way proxies to
  // work with lists:
  attribute_message_info->SetAddRemove(NEITHER);

  ScheduleEvent(SpSimTime(CurrentEvent->time_tag).IncrementPriority2(),
		REFLECT_TWO_WAY_ATTRIBUTES,
		ObjectType,
		LocalId,
		Node,
		NULL,
		message_buffer,
		message_buffer_size);
}

int SpObjProxy::GetInt(const char *name, int index) const {
  INT_ATTRIBUTE *IntAtt;
  IntAtt = (INT_ATTRIBUTE *)Find(name);

  if (IntAtt == NULL) {
    cout << "Warning, could not find INT_ATTRIBUTE " << name << endl;
    return 0x8000000;
  }

  return IntAtt[index].GetValue();
}

int SpObjProxy::GetInt(const int reference, int index) const{
  INT_ATTRIBUTE *IntAtt;
  IntAtt = (INT_ATTRIBUTE *)BaseAttribute[reference];
  if (index == -1) {
    return IntAtt->GetValue();
  }else{
    return IntAtt[index].GetValue();
  }
}

int SpObjProxy::GetDynamicInt(char *name, double t, int index) {
  DYNAMIC_INT_ATTRIBUTE *DynamicIntAtt;
  DynamicIntAtt = (DYNAMIC_INT_ATTRIBUTE *)Find(name);

  if (DynamicIntAtt == NULL) {
    cout << "Warning, could not find DYNAMIC_INT_ATTRIBUTE " << name << endl;
    return 0x8000000;
  }

  if (index == -1) {
    return (*DynamicIntAtt)(t);
  }else{
    return DynamicIntAtt[index](t);
  }
}

int SpObjProxy::GetDynamicInt(int reference, double t, int index) {
  DYNAMIC_INT_ATTRIBUTE *DynamicIntAtt;
  DynamicIntAtt = (DYNAMIC_INT_ATTRIBUTE *)BaseAttribute[reference];
  if (index == -1) {
    return (*DynamicIntAtt)(t);
  }else{
    return DynamicIntAtt[index](t);
  }
}

double SpObjProxy::GetFloat(const char *name, int index) const {
  FLOAT_ATTRIBUTE *FloatAtt;
  FloatAtt = (FLOAT_ATTRIBUTE *)Find(name);

  if (FloatAtt == NULL) {
    cout << "Warning, could not find FLOAT_ATTRIBUTE " << name << endl;
    return -1.0e20;
  }

  if (index == -1) {
    return FloatAtt->GetValue();
  }else{
    return FloatAtt[index].GetValue();
  }
}

double SpObjProxy::GetFloat(const int reference, int index) const {
  FLOAT_ATTRIBUTE *FloatAtt;
  FloatAtt = (FLOAT_ATTRIBUTE *)BaseAttribute[reference];
  if (index == -1) {
    return FloatAtt->GetValue();
  }else{
    return FloatAtt[index].GetValue();
  }
}

double SpObjProxy::GetDynamicFloat(char *name, double t, int index) {
  DYNAMIC_FLOAT_ATTRIBUTE *DynamicFloatAtt;
  DynamicFloatAtt = (DYNAMIC_FLOAT_ATTRIBUTE *)Find(name);

  if (DynamicFloatAtt == NULL) {
    cout << "Warning, could not find DYNAMIC_FLOAT_ATTRIBUTE " << name << endl;
    return -1.0e20;
  }

  if (index == -1) {
    return (*DynamicFloatAtt)(t);
  }else{
    return DynamicFloatAtt[index](t);
  }
}

double SpObjProxy::GetDynamicFloat(int reference, double t, int index) {
  DYNAMIC_FLOAT_ATTRIBUTE *DynamicFloatAtt;
  DynamicFloatAtt = (DYNAMIC_FLOAT_ATTRIBUTE *)BaseAttribute[reference];
  if (index == -1) {
    return (*DynamicFloatAtt)(t);
  }else{
    return DynamicFloatAtt[index](t);
  }
}

double SpObjProxy::GetDynamicFloat(char *name, double t, double &V, int index){
  DYNAMIC_FLOAT_ATTRIBUTE *DynamicFloatAtt;
  DynamicFloatAtt = (DYNAMIC_FLOAT_ATTRIBUTE *)Find(name);

  if (DynamicFloatAtt == NULL) {
    cout << "Warning, could not find DYNAMIC_FLOAT_ATTRIBUTE " << name << endl;
    return -1.0e20;
  }

  if (index == -1) {
    return (*DynamicFloatAtt)(t,V);
  }else{
    return DynamicFloatAtt[index](t,V);
  }
}

double SpObjProxy::GetDynamicFloat(int reference, double t, double &V,
		       int index) {
  DYNAMIC_FLOAT_ATTRIBUTE *DynamicFloatAtt;
  DynamicFloatAtt = (DYNAMIC_FLOAT_ATTRIBUTE *)BaseAttribute[reference];
  if (index == -1) {
    return (*DynamicFloatAtt)(t,V);
  }else{
    return DynamicFloatAtt[index](t,V);
  }
}

double SpObjProxy::GetDynamicFloat(char *name, double t, double &V, double &A,
		       int index) {
  DYNAMIC_FLOAT_ATTRIBUTE *DynamicFloatAtt;
  DynamicFloatAtt = (DYNAMIC_FLOAT_ATTRIBUTE *)Find(name);

  if (DynamicFloatAtt == NULL) {
    cout << "Warning, could not find DYNAMIC_FLOAT_ATTRIBUTE " << name << endl;
    return -1.0e20;
  }

  if (index == -1) {
    return (*DynamicFloatAtt)(t,V,A);
  }else{
    return DynamicFloatAtt[index](t,V,A);
  }
}

double SpObjProxy::GetDynamicFloat(int reference, double t, double &V,double &A,
		       int index) {
  DYNAMIC_FLOAT_ATTRIBUTE *DynamicFloatAtt;
  DynamicFloatAtt = (DYNAMIC_FLOAT_ATTRIBUTE *)BaseAttribute[reference];
  if (index == -1) {
    return (*DynamicFloatAtt)(t,V,A);
  }else{
    return DynamicFloatAtt[index](t,V,A);
  }
}

char *SpObjProxy::GetString(const char *name, int index) const{
  STRING_ATTRIBUTE *StringAtt;
  StringAtt = (STRING_ATTRIBUTE *)Find(name);

  if (StringAtt == NULL) {
    cout << "Warning, could not find STRING_ATTRIBUTE " << name << endl;
    return NULL;
  }

  if (index == -1) {
    return StringAtt->GetValue();
  }else{
    return StringAtt[index].GetValue();
  }
}

char *SpObjProxy::GetString(const int reference, int index) const{
  STRING_ATTRIBUTE *StringAtt;
  StringAtt = (STRING_ATTRIBUTE *)BaseAttribute[reference];
  if (index == -1) {
    return StringAtt->GetValue();
  }else{
    return StringAtt[index].GetValue();
  }
}

int SpObjProxy::GetLogical(const char *name, int index) const{
  LOGICAL_ATTRIBUTE *LogicalAtt;
  LogicalAtt = (LOGICAL_ATTRIBUTE *)Find(name);

  if (LogicalAtt == NULL) {
    cout << "Warning, could not find LOGICAL_ATTRIBUTE " << name << endl;
    return 0;
  }

  if (index == -1) {
    return LogicalAtt->GetValue();
  }else{
    return LogicalAtt[index].GetValue();
  }
}

int SpObjProxy::GetLogical(const int reference, int index) const{
  LOGICAL_ATTRIBUTE *LogicalAtt;
  LogicalAtt = (LOGICAL_ATTRIBUTE *)BaseAttribute[reference];
  if (index == -1) {
    return LogicalAtt->GetValue();
  }else{
    return LogicalAtt[index].GetValue();
  }
}

int SpObjProxy::GetDynamicLogical(char *name, double t, int index) {
  DYNAMIC_LOGICAL_ATTRIBUTE *DynamicLogicalAtt;
  DynamicLogicalAtt = (DYNAMIC_LOGICAL_ATTRIBUTE *)Find(name);

  if (DynamicLogicalAtt == NULL) {
    cout << "Warning, could not find DYNAMIC_LOGICAL_ATTRIBUTE "
	 << name << endl;
    return 0;
  }

  if (index == -1) {
    return (*DynamicLogicalAtt)(t);
  }else{
    return DynamicLogicalAtt[index](t);
  }
}

int SpObjProxy::GetDynamicLogical(int reference, double t, int index) {
  DYNAMIC_LOGICAL_ATTRIBUTE *DynamicLogicalAtt;
  DynamicLogicalAtt = (DYNAMIC_LOGICAL_ATTRIBUTE *)
    BaseAttribute[reference];
  if (index == -1) {
    return (*DynamicLogicalAtt)(t);
  }else{
    return DynamicLogicalAtt[index](t);
  }
}

SpObjProxy *SpObjProxy::GetObjectProxy(char *name, int index) {
  OBJECT_ATTRIBUTE *ObjectAtt;
  ObjectAtt = (OBJECT_ATTRIBUTE *)Find(name);

  if (ObjectAtt == NULL) {
    cout << "Warning, could not find OBJECT_ATTRIBUTE " << name << endl;
    return NULL;
  }

  if (index == -1) {
    return ObjectAtt->GetValue();
  }else{
    ObjectAtt = (OBJECT_ATTRIBUTE *)
      ((char *)ObjectAtt + index*ObjectAtt->GetSize());
    return ObjectAtt->GetValue();
  }
}

SpObjProxy *SpObjProxy::GetObjectProxy(int reference, int index) {
  OBJECT_ATTRIBUTE *ObjectAtt;
  ObjectAtt = (OBJECT_ATTRIBUTE *)BaseAttribute[reference];
  if (index == -1) {
    return ObjectAtt->GetValue();
  }else{
    ObjectAtt = (OBJECT_ATTRIBUTE *)
      ((char *)ObjectAtt + index*ObjectAtt->GetSize());
    return ObjectAtt->GetValue();
  }
}

DYNAMIC_OBJECT *SpObjProxy::GetDynamicObject(char *name, double t, int index) {
  DYNAMIC_OBJECT_ATTRIBUTE *DynamicObjectAtt;
  DynamicObjectAtt = (DYNAMIC_OBJECT_ATTRIBUTE *)Find(name);

  if (DynamicObjectAtt == NULL) {
    cout << "Warning, could not find DYNAMIC_OBJECT_ATTRIBUTE "
	 << name << endl;
    return NULL;
  }

  if (index == -1) {
    return (*DynamicObjectAtt)(t);
  }else{
    return DynamicObjectAtt[index](t);
  }
}

DYNAMIC_OBJECT *SpObjProxy::GetDynamicObject(int reference, double t, int index) {
  DYNAMIC_OBJECT_ATTRIBUTE *DynamicObjectAtt;
  DynamicObjectAtt = (DYNAMIC_OBJECT_ATTRIBUTE *)BaseAttribute[reference];
  if (index == -1) {
    return (*DynamicObjectAtt)(t);
  }else{
    return DynamicObjectAtt[index](t);
  }
}

void SpObjProxy::GetPosition(char *name,
			       double x[3],
			       int PT,
			       double t,
			       int index) {
  POSITION_ATTRIBUTE *PositionAtt;
  PositionAtt = (POSITION_ATTRIBUTE *)Find(name);

  if (PositionAtt == NULL) {
    cout << "Warning, could not find POSITION_ATTRIBUTE " << name << endl;
    for (int i=0; i<3; i++) {
      x[i] = 0.0;
    }
    return;
  }

  if (index == -1) {
    PositionAtt->GetValue(x, PT, t);
  }else{
    PositionAtt = (POSITION_ATTRIBUTE *)
      ((char *)PositionAtt + index*PositionAtt->GetSize());
    PositionAtt->GetValue(x, PT, t);
  }
}

void SpObjProxy::GetPosition(char *name,
			       double x[3], double v[3],
			       int PT,
			       double t,
			       int index) {
  POSITION_ATTRIBUTE *PositionAtt;
  PositionAtt = (POSITION_ATTRIBUTE *)Find(name);

  if (PositionAtt == NULL) {
    cout << "Warning, could not find POSITION_ATTRIBUTE " << name << endl;
    for (int i=0; i<3; i++) {
      x[i] = 0.0;
    }
    return;
  }

  if (index == -1) {
    PositionAtt->GetValue(x, PT, t);
  }else{
    PositionAtt = (POSITION_ATTRIBUTE *)
      ((char *)PositionAtt + index*PositionAtt->GetSize());
    PositionAtt->GetValue(x, PT, t);
  }

  v[0] = 0.0;
  v[1] = 0.0;
  v[2] = 0.0;
  Convert(x,v, PositionAtt->GetPositionType(),
	  x, v, PT,
	  t);
}


void SpObjProxy::GetPosition(char *name,
			       double x[3], double v[3], double a[3],
			       int PT,
			       double t,
			       int index) {
  POSITION_ATTRIBUTE *PositionAtt;
  PositionAtt = (POSITION_ATTRIBUTE *)Find(name);

  if (PositionAtt == NULL) {
    cout << "Warning, could not find POSITION_ATTRIBUTE " << name << endl;
    for (int i=0; i<3; i++) {
      x[i] = 0.0;
    }
    return;
  }

  if (index == -1) {
    PositionAtt->GetValue(x, PT, t);
  }else{
    PositionAtt = (POSITION_ATTRIBUTE *)
      ((char *)PositionAtt + index*PositionAtt->GetSize());
    PositionAtt->GetValue(x, PT, t);
  }

  v[0] = 0.0;
  v[1] = 0.0;
  v[2] = 0.0;

  a[0] = 0.0;
  a[1] = 0.0;
  a[2] = 0.0;

  Convert(x,v,a,  PositionAtt->GetPositionType(),
	  x, v, a, PT,
	  t);
}

void SpObjProxy::GetDynamicPosition(char *name, double x[3], int PT,
				      double t,
				      int index) {
  DYNAMIC_POSITION_ATTRIBUTE *DynamicPositionAtt;
  DynamicPositionAtt = (DYNAMIC_POSITION_ATTRIBUTE *)Find(name);

  if (DynamicPositionAtt == NULL) {
    cout << "Warning, could not find DYNAMIC_POSITION_ATTRIBUTE "
	 << name << endl;
    for (int i=0; i<3; i++) {
      x[i] = 0.0;
    }
    return;
  }

  if (index == -1) {
    (*DynamicPositionAtt)(t,PT,x);
  }else{
    DynamicPositionAtt = (DYNAMIC_POSITION_ATTRIBUTE *)
      ((char *)DynamicPositionAtt + index*DynamicPositionAtt->GetSize());
    (*DynamicPositionAtt)(t,PT,x);
  }
}

void SpObjProxy::GetDynamicPosition(char *name, double x[3], double v[3], int PT,
				      double t,
				      int index) {
  DYNAMIC_POSITION_ATTRIBUTE *DynamicPositionAtt;
  DynamicPositionAtt = (DYNAMIC_POSITION_ATTRIBUTE *)Find(name);

  if (DynamicPositionAtt == NULL) {
    cout << "Warning, could not find DYNAMIC_POSITION_ATTRIBUTE "
	 << name << endl;
    for (int i=0; i<3; i++) {
      x[i] = 0.0;
      v[i] = 0.0;
    }
    return;
  }

  if (index == -1) {
    (*DynamicPositionAtt)(t,PT,x,v);
  }else{
    DynamicPositionAtt = (DYNAMIC_POSITION_ATTRIBUTE *)
      ((char *)DynamicPositionAtt + index*DynamicPositionAtt->GetSize());
    (*DynamicPositionAtt)(t,PT,x,v);
  }
}

void SpObjProxy::GetDynamicPosition(char *name, double x[3], double v[3], double a[3],
				      int PT,
				      double t,
				      int index) {
  DYNAMIC_POSITION_ATTRIBUTE *DynamicPositionAtt;
  DynamicPositionAtt = (DYNAMIC_POSITION_ATTRIBUTE *)Find(name);

  if (DynamicPositionAtt == NULL) {
    cout << "Warning, could not find DYNAMIC_POSITION_ATTRIBUTE "
	 << name << endl;
    for (int i=0; i<3; i++) {
      x[i] = 0.0;
      v[i] = 0.0;
      a[i] = 0.0;
    }
    return;
  }

  if (index == -1) {
    (*DynamicPositionAtt)(t,PT,x,v,a);
  }else{
    DynamicPositionAtt = (DYNAMIC_POSITION_ATTRIBUTE *)
      ((char *)DynamicPositionAtt + index*DynamicPositionAtt->GetSize());
    (*DynamicPositionAtt)(t,PT,x,v,a);
  }
}

void SpObjProxy::GetPosition(int reference,
			       double x[3],
			       int PT,
			       double t,
			       int index) {
  POSITION_ATTRIBUTE *PositionAtt;
  PositionAtt = (POSITION_ATTRIBUTE *)BaseAttribute[reference];
  if (index == -1) {
    PositionAtt->GetValue(x, PT, t);
  }else{
    PositionAtt = (POSITION_ATTRIBUTE *)
      ((char *)PositionAtt + index*PositionAtt->GetSize());
    PositionAtt->GetValue(x, PT, t);
  }
}

void SpObjProxy::GetPosition(int reference,
			       double x[3], double v[3],
			       int PT,
			       double t,
			       int index) {
  POSITION_ATTRIBUTE *PositionAtt;
  PositionAtt = (POSITION_ATTRIBUTE *)BaseAttribute[reference];
  if (index == -1) {
    PositionAtt->GetValue(x, PT, t);
  }else{
    PositionAtt = (POSITION_ATTRIBUTE *)
      ((char *)PositionAtt + index*PositionAtt->GetSize());
    PositionAtt->GetValue(x, PT, t);
  }

  v[0] = 0.0;
  v[1] = 0.0;
  v[2] = 0.0;

  Convert(x, v, PositionAtt->GetPositionType(),
	  x, v, PT,
	  t);
}


void SpObjProxy::GetPosition(int reference,
			       double x[3], double v[3], double a[3],
			       int PT,
			       double t,
			       int index) {
  POSITION_ATTRIBUTE *PositionAtt;
  PositionAtt = (POSITION_ATTRIBUTE *)BaseAttribute[reference];
  if (index == -1) {
    PositionAtt->GetValue(x, PT, t);
  }else{
    PositionAtt = (POSITION_ATTRIBUTE *)
      ((char *)PositionAtt + index*PositionAtt->GetSize());
    PositionAtt->GetValue(x, PT, t);
  }

  v[0] = 0.0;
  v[1] = 0.0;
  v[2] = 0.0;

  a[0] = 0.0;
  a[1] = 0.0;
  a[2] = 0.0;

  Convert(x, v, a, PositionAtt->GetPositionType(),
	  x, v, a, PT,
	  t);
}

void SpObjProxy::GetDynamicPosition(int reference, double x[3], int PT,
				      double t,
				      int index) {
  DYNAMIC_POSITION_ATTRIBUTE *DynamicPositionAtt;
  DynamicPositionAtt = (DYNAMIC_POSITION_ATTRIBUTE *)
    BaseAttribute[reference];
  if (index == -1) {
    (*DynamicPositionAtt)(t,PT,x);
  }else{
    DynamicPositionAtt = (DYNAMIC_POSITION_ATTRIBUTE *)
      ((char *)DynamicPositionAtt + index*DynamicPositionAtt->GetSize());
    (*DynamicPositionAtt)(t,PT,x);
  }
}

void SpObjProxy::GetDynamicPosition(int reference, double x[3], double v[3], int PT,
				      double t,
				      int index){
  DYNAMIC_POSITION_ATTRIBUTE *DynamicPositionAtt;
  DynamicPositionAtt = (DYNAMIC_POSITION_ATTRIBUTE *)
    BaseAttribute[reference];
  if (index == -1) {
    (*DynamicPositionAtt)(t,PT,x,v);
  }else{
    DynamicPositionAtt = (DYNAMIC_POSITION_ATTRIBUTE *)
      ((char *)DynamicPositionAtt + index*DynamicPositionAtt->GetSize());
    (*DynamicPositionAtt)(t,PT,x,v);
  }
}

void SpObjProxy::GetDynamicPosition(int reference, double x[3], double v[3],
				      double a[3], int PT,
				      double t,
				      int index) {
  DYNAMIC_POSITION_ATTRIBUTE *DynamicPositionAtt;
  DynamicPositionAtt = (DYNAMIC_POSITION_ATTRIBUTE *)
    BaseAttribute[reference];
  if (index == -1) {
    (*DynamicPositionAtt)(t,PT,x,v,a);
  }else{
    DynamicPositionAtt = (DYNAMIC_POSITION_ATTRIBUTE *)
      ((char *)DynamicPositionAtt + index*DynamicPositionAtt->GetSize());
    (*DynamicPositionAtt)(t,PT,x,v,a);
  }
}

LIST_ATTRIBUTE *SpObjProxy::GetList(char *name, int index){
  LIST_ATTRIBUTE *ListAtt;
  ListAtt = (LIST_ATTRIBUTE *)Find(name);

  if (ListAtt == NULL) {
    cout << "Warning, could not find LIST_ATTRIBUTE "
	 << name << endl;
    return NULL;
  }

  if (index == -1) {
    return ListAtt;
  }else{
    ListAtt = (LIST_ATTRIBUTE *)
      ((char *)ListAtt + index*ListAtt->GetSize());
    return ListAtt;
  }
}

SpObjProxy *SpObjProxy::GetListName(char *name, char *n, int index) {
  LIST_ATTRIBUTE *ListAtt;
  ListAtt = (LIST_ATTRIBUTE *)Find(name);

  if (ListAtt == NULL) {
    cout << "Warning, could not find LIST_ATTRIBUTE "
	 << name << endl;
    return NULL;
  }

  if (index == -1) {
    return ((*ListAtt)(n))->GetObjectProxy();
  }else{
    ListAtt = (LIST_ATTRIBUTE *)
      ((char *)ListAtt + index*ListAtt->GetSize());
    return ((*ListAtt)(n))->GetObjectProxy();
  }
}

/*
SpObjProxy *SpObjProxy::GetListTime(char *name, double t, int index) {
  LIST_ATTRIBUTE *ListAtt;
  ListAtt = (LIST_ATTRIBUTE *)Find(name);

  if (ListAtt == NULL) {
    cout << "Warning, could not find LIST_ATTRIBUTE "
	 << name << endl;
    return NULL;
  }

  if (index == -1) {
    return ((*ListAtt)(t))->GetObjectProxy();
  }else{
    ListAtt = (LIST_ATTRIBUTE *)
      ((char *)ListAtt + index*ListAtt->GetSize());
    return ((*ListAtt)(t))->GetObjectProxy();
  }
}
*/

LIST_ATTRIBUTE *SpObjProxy::GetList(int reference, int index) {
  LIST_ATTRIBUTE *ListAtt;
  ListAtt = (LIST_ATTRIBUTE *)BaseAttribute[reference];
  if (index == -1) {
    return ListAtt;
  }else{
    ListAtt = (LIST_ATTRIBUTE *)
      ((char *)ListAtt + index*ListAtt->GetSize());
    return ListAtt;
  }
}

SpObjProxy *SpObjProxy::GetListName(int reference, char *n, int index){
  LIST_ATTRIBUTE *ListAtt;
  ListAtt = (LIST_ATTRIBUTE *)BaseAttribute[reference];
  if (index == -1) {
    return ((*ListAtt)(n))->GetObjectProxy();
  }else{
    ListAtt = (LIST_ATTRIBUTE *)
      ((char *)ListAtt + index*ListAtt->GetSize());
    return ((*ListAtt)(n))->GetObjectProxy();
  }
}

/*
SpObjProxy *SpObjProxy::GetListTime(int reference, double t, int index){
  LIST_ATTRIBUTE *ListAtt;
  ListAtt = (LIST_ATTRIBUTE *)BaseAttribute[reference];
  if (index == -1) {
    return ((*ListAtt)(t))->GetObjectProxy();
  }else{
    ListAtt = (LIST_ATTRIBUTE *)
      ((char *)ListAtt + index*ListAtt->GetSize());
    return ((*ListAtt)(t))->GetObjectProxy();
  }
}
*/

void SpObjProxy::PrintStream(ostream &out, int level) const{
  BASE_ATTRIBUTE *attptr;
  int i;
  int j;
  out << endl;
  SpaceLevel(out, level);
  out << "Object Proxy for "
      << ObjectTemplate->ObjProxyGetClassName() << ": "
      << ObjectName << ", Id " << ObjectId << endl;
  for (i=0; i<Nattributes; i++) {

    SpaceLevel(out, level);
    out << i << ": "
	<< ObjectTemplate->ObjProxyGetClassName(i) << " "
	<< ObjectTemplate->GetAttributeType(i) << " ";

    out << ObjectTemplate->GetAttributeName(i);

    if (ObjectTemplate->GetArraySize(i) > 1) {
      out << " [" << ObjectTemplate->GetArraySize(i) << "]";
    }

    out << " = ";

    if (BaseAttribute[i]) {
      BaseAttribute[i]->outstream(out, level);
      if (ObjectTemplate->GetArraySize(i) > 1) {
	for (j=1; j<ObjectTemplate->GetArraySize(i); j++) {
	  out << " ";
	  attptr = (BASE_ATTRIBUTE *)((char *)BaseAttribute[i]
				      + j*BaseAttribute[i]->GetSize());
	  attptr->outstream(out, level);
	}
      }
    }else{
      out << "Undefined";
    }

    if ((level == 0) || (i < Nattributes-1)) out << endl;

  }
}

void SpObjProxy::Destruct() {
  int i;

/*
  if (Nattributes) cout << "Destruct Object Proxy "
			<< ObjectTemplate->ObjProxyGetClassName() << endl;
*/

  for (i=0; i<Nattributes; i++) {
    if (BaseAttribute[i]) {
      BaseAttribute[i]->delete_vector_with_child_type();
    }
  }
  delete [] BaseAttribute;

  free_ObjectName();
  if (RootProxyP()) { //proxy:
    DeleteUpdateList(); //proxy:
  }
  SetAttributeUpdateList(NULL); //proxy: add

  ObjectTemplate = NULL;
  BaseAttribute = NULL;
  Nattributes = 0;
  ObjectId = -1;
  Node = -1;
  ObjectType = -1;
  LocalId = -1;
  Defined = 0;

  InObjectAttribute = NULL;
}

void SpObjProxy::PartialDestruct() {

/*
  if (Nattributes) cout << "Partial Destruct Object Proxy "
			<< ObjectTemplate->ObjProxyGetClassName() << endl;
*/

  delete [] BaseAttribute;
  free_ObjectName();
  if (RootProxyP()) { //proxy:
    DeleteUpdateList();
  }
  SetAttributeUpdateList(NULL); //proxy: add

  ObjectTemplate = NULL;
  BaseAttribute = NULL;
  Nattributes = 0;
  ObjectId = -1;
  Node = -1;
  ObjectType = -1;
  LocalId = -1;
  Defined = 0;

  InObjectAttribute = NULL;
}

void SpObjProxy::Construct(char *type) {
  int i;

//  cout << "SpObjProxy constructor creating " << type << endl;

  Destruct();
  SetObjectName("Unnamed");

/*
** Get the Object Template from the Object Factory. This will create the
** template if it has not already been created.
*/

  ObjectTemplate = ObjFact->GetTemplate(type);

/*
** Create (and initialize to NULL) the array of attribute references
*/

  Nattributes = ObjectTemplate->GetNattributes();
  BaseAttribute = new BASE_ATTRIBUTE *[Nattributes];
  for (i=0; i<Nattributes; i++) BaseAttribute[i] = NULL;

/*
** Set the current Object Proxy global reference to this
*/

  CurrentObjectProxy = this;

//  cout << endl << "SpObjProxy constructor done creating " << type << endl;

}

int SpObjProxy::GetReference(const char *name, char *type) const{
  SpObjTemplate *ObjTemp = ObjFact->GetTemplate(type);
  int Natts = ObjTemp->GetNattributes();
  for (int i=0; i<Natts; i++) {
    if (!strcmp(name, ObjTemp->GetAttributeName(i))) {
      return i;
    }
  }
  cout << "Error, could not find " << name << endl;
  return -1;
}

int SpObjProxy::CheckProxy() {
  int i;
  int status = 0;
  SpObjProxy* ObjectProxy;
  LIST_ATTRIBUTE *ListAttribute;
  OBJECT_ATTRIBUTE *ObjectAttribute;

  for (i=0; i<Nattributes; i++) {

    if (BaseAttribute[i]) {

      if (BaseAttribute[i]->GetInObjectProxy() != this) {
	status++;
	BaseAttribute[i]->SetInObjectProxy(this);
	cout << "Warning (CheckProxy): Bad InObjectProxy (repaired)" << endl;
      }

      switch (ObjectTemplate->GetAttributeTypeID(i)) {

	case DYNAMIC_OBJECT_ATTRIBUTE_TYPE:
	  ObjectAttribute = (OBJECT_ATTRIBUTE *)BaseAttribute[i];
	  ObjectProxy = ObjectAttribute->GetObjectProxy();
	  if (ObjectProxy == NULL) {
	    cout << "Error (CheckProxy): No Object Proxy in Attribute "
		 << "(could not repair)"
		 << endl;
	    status++;
	    break;
	  }
	  if (ObjectProxy->GetInObjectAttribute() != ObjectAttribute) {
	    ObjectProxy->SetInObjectAttribute(ObjectAttribute);
	    cout << "Warning (CheckProxy): Bad InAttribute (repaired)" << endl;
	    status++;
	  }
	  status += ObjectProxy->CheckProxy();
	  break;

	case LIST_ATTRIBUTE_TYPE:
	  ListAttribute = (LIST_ATTRIBUTE *)BaseAttribute[i];
	  ObjectAttribute = (OBJECT_ATTRIBUTE *)ListAttribute->get_top();
	  while (ObjectAttribute) {

	    ObjectProxy = ObjectAttribute->GetObjectProxy();
	    if (ObjectProxy == NULL) {
	      cout << "Error (CheckProxy): No Object Proxy in List Attribute "
		   << "(could not repair)"
		   << endl;
	      status++;
	      break;
	    }

	    if (ObjectProxy->GetInObjectAttribute() != ObjectAttribute) {
	      ObjectProxy->SetInObjectAttribute(ObjectAttribute);
	      cout << "Warning (CheckProxy): Bad List InAttribute (repaired)"
		   << endl;
	      status++;
	    }

	    if (ObjectAttribute->GetInListAttribute() != ListAttribute) {
	      ObjectAttribute->SetInListAttribute(ListAttribute);
	      cout << "Warning (CheckProxy): Bad List Attribute (repaired)"
		   << endl;
	      status++;
	    }

	    status += ObjectProxy->CheckProxy();
	    ObjectAttribute = (OBJECT_ATTRIBUTE *) ++(*ListAttribute);
	  }
	  break;

	default:
	  break;

      }

    }else{

      cout << "Warning (CheckProxy): No Attribute Defined "
	   << "(could not repair)"
	   << endl;

    }

  }

  return status;

}

SpObjProxy *SpObjProxy::FindProxy(PROXY_LEVEL_INFO *ProxyLevelInfo) {
  int i;
  SpObjProxy *Op;
  BASE_ATTRIBUTE *Attribute;
  OBJECT_ATTRIBUTE *ObjectAttribute;
  LIST_ATTRIBUTE *ListAttribute;

  Op = this;

  for (i=0; i<ProxyLevelInfo->Level-1; i++) {

    if (Op == NULL) {
//    cout << "Error (FindProxy): Bad Op" << endl;
      return NULL;
    }

    Attribute = Op->GetAttribute(ProxyLevelInfo[i].AttributeIndex,
				 ProxyLevelInfo[i].ArrayIndex);

    switch(Attribute->GetType()) {

      case OBJECT_ATTRIBUTE_TYPE:
	ObjectAttribute = (OBJECT_ATTRIBUTE *)Attribute;
	Op = ObjectAttribute->GetObjectProxy();
	break;

      case LIST_ATTRIBUTE_TYPE:
	ListAttribute = (LIST_ATTRIBUTE *)Attribute;
	ObjectAttribute = ListAttribute->
	  FindObjectAttribute(ProxyLevelInfo[i].ListId);
	if (ObjectAttribute) {
	  Op = ObjectAttribute->GetObjectProxy();
	}else{
/*
	  cout << "Error (FindProxy): Could not find Object Proxy in List"
	       << endl;
*/
	  return NULL;
	}
	break;

      default:
//	cout << "Error (FindProxy): Bad Trail" << endl;
	return NULL;

    }

  }

  return Op;

}

