// SpObjProxy.H Header File

#ifndef ObjectProxy_Object
#define ObjectProxy_Object

#include "SpString.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpStdarg.H"

#include "qitem.H"
#include "SpExportAttribute.H"
#include "SpProxyMacros.H"
#include "SpObjTemplate.H"

#include "RB_UpdateXqueue.H" //proxy:
#include "RB_voidPtr.H" //proxy:

class MONOTONICALLY_INCREASING_BUFFER {
  private:
    char *B;
    int Size;
  public:
    MONOTONICALLY_INCREASING_BUFFER() : B(NULL), Size(0) {}
    ~MONOTONICALLY_INCREASING_BUFFER() { delete [] B; }
    char *need(int size) {
      if (size > Size)
      {
	delete [] B;
	Size = size;
	B = new char[Size];
      }
      return B;
    }
};

class F_UPDATE_LIST_ITEM;

// Note that in this code, the word reference and AttributeIndex are
// often used synonymously.

class SpObjProxy : public C_ITEM {

  private:
    static MONOTONICALLY_INCREASING_BUFFER MIB;

    SpObjTemplate *ObjectTemplate;  // Describes the attributes of this class
    BASE_ATTRIBUTE **BaseAttribute;   // Array of pointers to attributes
                                      // Attributes can be arrays

    OBJECT_ATTRIBUTE *InObjectAttribute; // This OBJECT_ATTRIBUTE contains me

    // proxy: List of attribute changes (moved from S_SpHLA class)
    RB_voidPtr AttributeUpdateList;

    int Nattributes; // size of BaseAttribute array
    int Defined; // TRUE when attributes owned by an entity are mapped into me
                 // FALSE if this is a proxy recontructed remotely
    char *ObjectName; // Instances of objects can have a name

    int ObjectId;   // global id of object
    int Node;       // node of owner of this proxy
    int ObjectType; // type of owner of this proxy
    int LocalId;    // local id of owner of this proxy

    //proxy:
    // Instantiate the update list, if nec, and store in this (the root) proxy.
    // Run down the "branch" of the proxy tree denoted by arg, attribNum, inserting
    // a pointer to the updateList in all subsidiary proxies.
    void AssignUpdateListPointers(int attribNum, int attribTypeID);

    // proxy:
    // Run down ONE branch of this proxy (denoted by attribNum), and insert
    // the updateList pointer into all subsidiary proxies
    void SetAttribUpdatePtrs(RB_UpdateXqueue *updateList, 
			     int attribNum, int attribTypeID);

/******************************************************************************
* Internal Method: UpdateRemoteAttribute
* Summary: general helper function called by many public functions
* Description: Updates the remote attribute specified by the first two
*              arguments with the data from the last two arguments.
*              That data should have been created by
*              BASE_ATTRIBUTE::FillMessage.
* Return: void
******************************************************************************/
    void UpdateRemoteAttribute(const char * const name, const int index,
			       const char * const buff, const int nBytes);

  public:
    //int FullDestroy_p;

    SpObjProxy();
    ~SpObjProxy();

    SpObjProxy(char *type) /* : FullDestroy_p(0) */ {
      AttributeUpdateList = NULL;  //proxy: add
      Construct(type);
    }

    void Construct(char *type);
    virtual void Destruct();
    virtual void PartialDestruct();

  // void checkForBadUpdateList(void); // proxy: temp

    // proxy: get list of updates to attribs
    RB_UpdateXqueue *GetAttributeUpdateList() {
      return (RB_UpdateXqueue *) ((void *) AttributeUpdateList);
    }

    // proxy: set list of attrib updates
    void SetAttributeUpdateList(RB_UpdateXqueue *attributeUpdateList) { 
       AttributeUpdateList = (void *) attributeUpdateList;
    }

    // proxy:
    // return true iff this proxy is the "root"--ie NOT inside an obj attrib
    int RootProxyP(void) { 
      return( GetInObjectAttribute() == NULL );
    }

    // proxy:
    // return true if this attribute type contains an obj proxy
    int ContainsProxy(int attribTypeID) {
      return( (attribTypeID==DYNAMIC_OBJECT_ATTRIBUTE_TYPE) ||
	      (attribTypeID==OBJECT_ATTRIBUTE_TYPE) ||
	      (attribTypeID==LIST_ATTRIBUTE_TYPE) );
    }


    //pxy:
    // return TRUE iff this obj proxy's updateList is saving updates
    int AmSavingUpdates(void) { 
       RB_UpdateXqueue *attUpdateList = GetAttributeUpdateList();
      if (attUpdateList != NULL) {
	return attUpdateList->AmSavingUpdates();
      }
      else {
	cout << "SpObjProxy::AmSavingUpdates ERROR:  Attribute UpdateList is NULL!"
	     << endl;
	return 0;
      }
    }
     

    //proxy:
    // Turn ON collection of attribute updates 
    void CollectAttribUpdates(void) {
      RB_UpdateXqueue *attUpdateList = GetAttributeUpdateList();
      if (attUpdateList != NULL) {
	attUpdateList->SaveUpdates();
      }
      else {
	cout << "SpObjProxy::CollectAttribUpdates ERROR:  Attribute UpdateList is NULL!"
	     << endl;
      }
    }

    // proxy:
    // Turn OFF collection of attribute updates 
    void DiscardAttribUpdates(void) {
      RB_UpdateXqueue *attUpdateList = GetAttributeUpdateList();
      if (attUpdateList != NULL) {
	attUpdateList->DiscardUpdates();
      }
      else {
	cout << "SpObjProxy::DiscardAttribUpdates ERROR:  Attribute UpdateList is NULL!"
	     << endl;
      }
    }
    

    // proxy:
    // Run down ALL branches (ie attributes) of this proxy, and insert
    // the updateList pointer into all subsidiary proxies
    void SetProxyUpdatePtrs(RB_UpdateXqueue *updateList);

    // proxy::  check to make sure update ptrs are correct
    void CheckProxyUpdatePtrs(RB_UpdateXqueue *updateList);

    // proxy::  check to make sure update ptrs are correct
    void CheckAttribUpdatePtrs(RB_UpdateXqueue *updateList, int attribNum);


    BASE_ATTRIBUTE *GetAttribute(int i) {return BaseAttribute[i];}
    BASE_ATTRIBUTE *GetAttribute(int i, int j) {
      char *ba = (char *)BaseAttribute[i];
      if (ba) ba += j * BaseAttribute[i]->GetSize();
      return (BASE_ATTRIBUTE *)ba;
    }

    OBJECT_ATTRIBUTE *GetInObjectAttribute() {return InObjectAttribute;}
    void SetInObjectAttribute(OBJECT_ATTRIBUTE *i) {InObjectAttribute = i;}

    void Define(BASE_ATTRIBUTE *Attribute, char *Name) {
      Defined = 1;
      Define(Attribute[0], Name);
    }

    void Define(BASE_ATTRIBUTE &Attribute, char *Name);
    int CheckProxy();
    SpObjProxy *FindProxy(PROXY_LEVEL_INFO *ProxyLevelInfo);

    void Define(BASE_ATTRIBUTE *Attribute, int AttributeIndex);
    void Execute(char *MethodName, ...);

// returns the integer array index for the attribute
// name (and proxy class type) passed in

    int GetReference(const char *name) const;
    int GetReference(const char *name, char *type) const;

    BASE_ATTRIBUTE *Find(const int reference,    int ArrayIndex = 0) const;
    BASE_ATTRIBUTE *Find(const char *name, int ArrayIndex = 0) const{
      return Find(GetReference(name), ArrayIndex);
    }



/******************************************************************************
* Method: UpdateRemote<Type>
* Summary: Change the value of an attribute in the remote owner of this proxy.
* Description: NOTE: These functions do not immediately update this
*              proxy or the owner of this proxy.  For
*              example, suppose "sample_int" is 2, and you have the
*              following code:
*
*              UpdateRemoteInt("sample_int", 3);
*              cout << GetInt("sample_int");
*
*              The above example code would print out 2.
*              
*              This function is called in a certain event.  The
*              attribute will be updated in this proxy and the owner
*              of this proxy sometime AFTER that event finishes but
*              still at the same physical simulation time, although at
*              possibly higher priorities.  The update might not occur
*              by the time of the very next event, but it will still
*              occur sometime during the same physical time (although
*              at possibly different priorities.).  But once an event
*              with a higher physical time is processed, then the
*              attribute is definitely updated in both this proxy and
*              the owner of this proxy.
*
* Return: void
******************************************************************************/
    /*
      This code is not implemented yet.
    void UpdateRemoteInt(const char *name, int index, int new_value) {
      const int fill_mess_buff_size = INT_ATTRIBUTE::GetMessageSizeStatic();
      char * const fill_message_buff = MIB.need(fill_mess_buff_size);
      INT_ATTRIBUTE::FillMessageWith(fill_message_buff, new_value);
      UpdateRemoteAttribute
	(name, index, fill_message_buff, fill_mess_buff_size);
    }

    void UpdateRemoteInt(char *name, int new_value)
    	{ UpdateRemoteInt(name, 0, new_value); }

    void UpdateRemoteString(char *name, int index, char *new_value) {
      const int fill_mess_buff_size
	= STRING_ATTRIBUTE::GetMessageSize(new_value);
      char * const fill_message_buff = MIB.need(fill_mess_buff_size);
      ((STRING_ATTRIBUTE_MESSAGE_INFO*)fill_message_buff)
	->SetNbytes(fill_mess_buff_size);
      STRING_ATTRIBUTE::FillMessageWith(fill_message_buff, new_value);
      UpdateRemoteAttribute
	(name, index, fill_message_buff, fill_mess_buff_size);
    }

    void UpdateRemoteString(char *name, char *new_value)
    	{ UpdateRemoteString(name, 0, new_value); }

	*/
/*
** Return information based on name of attribute
*/

    void GetTimeInterval(int reference,
			 double &startTime,
			 double &endTime,
			 int index=0)
    {
      BASE_DYNAMIC_ATTRIBUTE *a
	= (BASE_DYNAMIC_ATTRIBUTE*) Find(reference, index);
      startTime = a->GetStartTime();
      endTime = a->GetEndTime();
    }

    void GetTimeInterval(const char *name,
			 double &startTime,
			 double &endTime,
			 int index=0)
    {
      GetTimeInterval(GetReference(name), startTime, endTime, index);
    }

    int GetTimeInterval(int reference,
			 double time,
			 double &startTime,
			 double &endTime,
			 int index=0)
    {
      BASE_DYNAMIC_ATTRIBUTE *BDA =
	(BASE_DYNAMIC_ATTRIBUTE *)Find(reference,index);

      if (BDA) {
	return BDA->GetTimeInterval(time, startTime, endTime);
      }else{
	startTime = -INFINITY;
	endTime = INFINITY;
	return 0;
      }
    }

    int GetTimeInterval(const char *name,
			 double time,
			 double &startTime,
			 double &endTime,
			 int index=0) {
      return GetTimeInterval(GetReference(name), time, startTime,
			     endTime, index);
    }

// GetInt

    int GetInt(const char *name, int index = 0) const;
    int GetInt(const int reference, int index = -1) const;
    int GetDynamicInt(char *name, double t, int index = -1);
    int GetDynamicInt(int reference, double t, int index = -1);
    
// GetFloat

    double GetFloat(const char *name, int index = -1) const;
    double GetFloat(const int reference, int index = -1) const ;
    double GetDynamicFloat(char *name, double t, int index = -1);
    double GetDynamicFloat(int reference, double t, int index = -1);
    double GetDynamicFloat(char *name, double t, double &V, int index = -1);
    double GetDynamicFloat(int reference, double t, double &V,
			   int index = -1);
    double GetDynamicFloat(char *name, double t, double &V, double &A,
		    int index = -1);
    double GetDynamicFloat(int reference, double t, double &V,double &A,
			   int index = -1);
// GetString

    char *GetString(const char *name, int index = -1) const;
    char *GetString(const int reference, int index = -1) const;
    
// GetLogical

    int GetLogical(const char *name, int index = -1) const;
    int GetLogical(const int reference, int index = -1) const;
    int GetDynamicLogical(char *name, double t, int index = -1);
    int GetDynamicLogical(int reference, double t, int index = -1);
    
// GetObject

    SpObjProxy *GetObjectProxy(char *name, int index = -1);
    SpObjProxy *GetObjectProxy(int reference, int index = -1);
    DYNAMIC_OBJECT *GetDynamicObject(char *name, double t, int index = -1);
    DYNAMIC_OBJECT *GetDynamicObject(int reference, double t, int index = -1);
    
// GetPosition

    void GetPosition(char *name,
		     double x[3],
		     int PT,
		     double t=0.0,
		     int index = -1);

    void GetPosition(char *name,
		     double x[3], double v[3],
		     int PT,
		     double t=0.0,
		     int index = -1);

    void GetPosition(char *name,
		     double x[3], double v[3], double a[3],
		     int PT,
		     double t=0.0,
		     int index = -1);

    void GetDynamicPosition(char *name, double x[3], int PT,
			      double t=0.0,
			      int index = -1);
    void GetDynamicPosition(char *name, double x[3], double v[3], int PT,
			      double t=0.0,
			      int index = -1);
    void GetDynamicPosition(char *name, double x[3], double v[3], double a[3],
			      int PT,
			      double t=0.0,
			      int index = -1);

    void GetPosition(int reference,
		     double x[3],
		     int PT,
		     double t=0.0,
		     int index = -1);

    void GetPosition(int reference,
		     double x[3], double v[3],
		     int PT,
		     double t=0.0,
		     int index = -1);

    void GetPosition(int reference,
		     double x[3], double v[3], double a[3],
		     int PT,
		     double t=0.0,
		     int index = -1);

    void GetDynamicPosition(int reference, double x[3], int PT,
			      double t=0.0,
			      int index = -1);
    void GetDynamicPosition(int reference, double x[3], double v[3], int PT,
			      double t=0.0,
			      int index = -1);
    void GetDynamicPosition(int reference, double x[3], double v[3],
			      double a[3], int PT,
			      double t=0.0,
			      int index = -1);


// GetActive
    int GetActive(void) const { return (GetLogical("Active")); }

// GetList

    LIST_ATTRIBUTE *GetList(char *name, int index = -1);
    SpObjProxy *GetListName(char *name, char *n, int index = -1);
    //SpObjProxy *GetListTime(char *name, double t, int index = -1);
    LIST_ATTRIBUTE *GetList(int reference, int index = -1);
    SpObjProxy *GetListName(int reference, char *n, int index = -1);
    //SpObjProxy *GetListTime(int reference, double t, int index = -1);
/*
** Other stuff
*/

    int GetInitialized() {
      int i = 0;
      if (ObjectTemplate) i = 1;
      return i;
    }

    void ResetDefined() {Defined = 0;}
    void SetDefined() {Defined = 1;}
    int GetDefined() {return Defined;}

    void SetObjectId(int i) {ObjectId = i;}
    int GetObjectId() {return ObjectId;}

    void SetGlobalId(int i) {ObjectId = i;}
    int GetGlobalId() {return ObjectId;}

    void SetLocalId(int i) {LocalId = i;}
    int GetLocalId() {return LocalId;}

    void SetObjectType(int i) {ObjectType = i;}
    int GetObjectType() {return ObjectType;}

    void SetNode(int i) {Node = i;}
    int GetNode() {return Node;}

    void SetObjectTemplate(SpObjTemplate *OT) {ObjectTemplate = OT;}
    SpObjTemplate *GetObjectTemplate() {return ObjectTemplate;}

    void SetObjectName(char *n) {
      if (ObjectName) free(ObjectName);
      ObjectName = strdup(n);
    }
    void free_ObjectName();

    //proxy: delete contained update items and the list itself
    void DeleteUpdateList(void);
    
    char *GetObjectName() {return ObjectName;}
    char *GetName() {return ObjectName;}

    int GetNattributes() {return Nattributes;}
    void SetNattributes(int n) {Nattributes = n;}

    char *ObjProxyGetClassName() {return ObjectTemplate->ObjProxyGetClassName();}

    void SpaceLevel(ostream &out, int level) const{
      for (int i=0; i<level; i++) out << "     ";
    }

    void PrintStream(ostream &out = cout, int level=0) const;

/******************************************************************************
* Internal Method: FillMessage
* Summary: Fills in a message representing some attributes of this
*          SpObjProxy.
* Description: Returns by reference in the last two arguments a buffer
*              which can be sent as part of an event message, which
*              represents the attributes specified in the linked list
*              argument called top.  EmptyMessage knows how to
*              interpret the message created by this function.
* Return: void
******************************************************************************/
    void FillMessage(F_UPDATE_LIST_ITEM *top, // IN: attributes to read
		     int len,                 // IN: num of attributes to read
		     char *&message,          // OUT: start of filled message
		     int &nBytes);            // OUT: length of filled message

    struct EMPTY_MESSAGE_RESULT {
	BASE_ATTRIBUTE *Attribute;
	const char *Buff;
	int OpWarning;
      //The following was added to avoid a UMR in Purify
        EMPTY_MESSAGE_RESULT(EMPTY_MESSAGE_RESULT& j){
	  OpWarning = j.OpWarning;
	  Buff = j.Buff;
	  Attribute = j.Attribute;
	}
        EMPTY_MESSAGE_RESULT():Attribute(NULL), Buff(NULL), OpWarning(0){}
    };

/******************************************************************************
* Internal Method: EmptyMessage
* Summary: Empty the input message into this SpObjProxy.
* Description: The input message was made by something like
*              FillMessage.  This function reads that message and
*              modifies the BASE_ATTRIBUTE in this SpObjProxy
*              according to how the message says it should be modified.
* Return: The Attribute field in the return value is the attribute
*         that the message said to modify.  The Buff field is the start
*         of the next message.  This value may be passed to another
*         call of EmptyMessage.  The OpWarning field indicates that
*         an error may have occurred.
******************************************************************************/
    EMPTY_MESSAGE_RESULT& EmptyMessage(const char *message);
};

typedef SpObjProxy OBJECT_PROXY;

#endif

