// SpObjProxyMsg.H Header File

#ifndef ObjPrxMess_Object
#define ObjPrxMess_Object

#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpString.H"
#include "SpStdlib.H"
#include "SpMemory.H"
#include "SpAssert.H"
#include "SpNetTypes.H"

enum ADD_REMOVE {
  REMOVE=-1,
  NEITHER=1,
  OBJECT_PROXY_ADD=0
};

class ATTRIBUTE_MESSAGE_INFO {
  private:
  protected:

    // geb2 start changes
    NET_INT Nbytes;
    NET_INT AttributeIndex;
    NET_INT ArrayIndex;
    NET_INT AddRemove;     // was:ADD_REMOVE AddRemove;
    NET_INT AddRemoveId;
    // geb2 end changes

  public:

    void SetNbytes(int n) {Nbytes = n;}
    int GetNbytes() {return Nbytes;}

    void SetAttributeIndex(int i) {AttributeIndex = i;}
    int GetAttributeIndex() {return AttributeIndex;}

    void SetArrayIndex(int i) {ArrayIndex = i;}
    int GetArrayIndex() {return ArrayIndex;}

    void SetAddRemove(ADD_REMOVE i) {AddRemove = i;}
    ADD_REMOVE GetAddRemove() {
      return (ADD_REMOVE) ((int) AddRemove); //geb2
    }

    void SetAddRemoveId(int i) {AddRemoveId = i;}
    int GetAddRemoveId() {return AddRemoveId;}

};

//geb2:  several changes to INT_ATTRIBUTE_MESSAGE_INFO
class INT_ATTRIBUTE_MESSAGE_INFO : public ATTRIBUTE_MESSAGE_INFO {
  private:
  protected:
    NET_INT Value;
  public:

    // Note:  this class can be embedded within a raw buffer (eg via a socket)
    // so it is NOT necessarily aligned;  hence the gyrations and memcpy's below:
    void SetValue(int v) {
      NET_INT v1 = v;
      memcpy((char *)&Value, (char *)&v1, sizeof(NET_INT));
    }
    int GetValue() {
      NET_INT v;
      memcpy((char *)&v, (char *)&Value, sizeof(NET_INT));
      return (int) v;
    }
};

//geb2: several changes:
class FLOAT_ATTRIBUTE_MESSAGE_INFO : public ATTRIBUTE_MESSAGE_INFO {
  private:
  protected:
    NET_DOUBLE Value;
  public:
    void SetValue(double v) {
      NET_DOUBLE v1 = v;
      memcpy((char *)&Value, (char *)&v1, sizeof(NET_DOUBLE));
    }
    double GetValue() {
      NET_DOUBLE v;
      memcpy((char *)&v, (char *)&Value, sizeof(NET_DOUBLE));
      return (double) v;
    }
};

//geb2: several changes:
class LOGICAL_ATTRIBUTE_MESSAGE_INFO : public ATTRIBUTE_MESSAGE_INFO {
  private:
  protected:
    NET_INT Value;
  public:
    void SetValue(int v) {
      NET_INT v1 = v;
      memcpy((char *)&Value, (char *)&v1, sizeof(NET_INT));
    }
    int GetValue() {
      NET_INT v;
      memcpy((char *)&v, (char *)&Value, sizeof(NET_INT));
      return (int) v;
    }
};

class STRING_ATTRIBUTE_MESSAGE_INFO : public ATTRIBUTE_MESSAGE_INFO {
  private:
  protected:
  public:
    void SetValue(char *s) {
      if (s == NULL)
	assert(Nbytes == sizeof(*this));
      else
      {
	STRING_ATTRIBUTE_MESSAGE_INFO *buff = this;
	strcpy((char *)&buff[1], s);
      }
    }
    char *GetValue() {
      if (Nbytes == sizeof(*this))
	return NULL;
      else
      {
	STRING_ATTRIBUTE_MESSAGE_INFO *buff = this;
	return (char *)&buff[1];
      }
    }
};

//geb2:  many changes:
class POSITION_ATTRIBUTE_MESSAGE_INFO : public ATTRIBUTE_MESSAGE_INFO {
  private:
  protected:
    NET_DOUBLE Position[3];
    NET_INT PositionType;
  public:
    void SetValue(double *P, int newPosnType) {
      NET_DOUBLE location[3];
      for(int i=0; i<3; i++) {
	location[i] = P[i];
      }
      memcpy((char *)Position, (char *)location, 3*sizeof(NET_DOUBLE));
      
      NET_INT posnType(newPosnType);
      memcpy((char *)&PositionType, (char *)&posnType, sizeof(NET_INT));
    }

    void GetValue(double *P, int *PT) {
      NET_DOUBLE location[3];
      memcpy((char *)location, (char *)Position, 3*sizeof(NET_DOUBLE));
      for(int i=0; i<3; i++) {
	 P[i] = ((double) location[i]);
      }

      NET_INT posnType;
      memcpy((char *)&posnType, (char *)&PositionType, sizeof(NET_INT));
      *PT = ((int) posnType);
    }
};

class OBJECT_ATTRIBUTE_MESSAGE_INFO : public ATTRIBUTE_MESSAGE_INFO {
  private:
  protected:
    NET_INT ListId;
  public:

    void SetListId(int id) {
      NET_INT newID(id);
      memcpy((char *)&ListId, (char *)&newID, sizeof(NET_INT));
    }

    int GetListId() {
      NET_INT id;
      memcpy((char *)&id, (char *)&ListId, sizeof(NET_INT));
      return ((int) id);
    }

    void SetValue(char *ObjectProxyMessage, int n) {
      OBJECT_ATTRIBUTE_MESSAGE_INFO *buff = this;
      memcpy((char *)&buff[1], ObjectProxyMessage, n);
    }
    char *GetValue() {
      OBJECT_ATTRIBUTE_MESSAGE_INFO *buff = this;
      return (char *)&buff[1];
    }
};

class LIST_ATTRIBUTE_MESSAGE_INFO : public ATTRIBUTE_MESSAGE_INFO {
  private:
    NET_INT Length;

  protected:
  public:

    void SetLength(int newLen) {
      NET_INT len(newLen);
      memcpy((char *)&Length, (char *)&len, sizeof(NET_INT));
    }

    int GetLength() {
      NET_INT len;
      memcpy((char *)&len, (char *)&Length, sizeof(NET_INT));
      return ((int) len);
    }

};

class DYNAMIC_ATTRIBUTE_MESSAGE_INFO : public LIST_ATTRIBUTE_MESSAGE_INFO {
  private:
  protected:
  public:
};


class PROXY_LEVEL_INFO {
  private:
  protected:
  public:

    NET_INT Level;
    NET_INT AttributeIndex;
    NET_INT ArrayIndex;
    NET_INT ListId;

  /***********  These guys are not needed because this class is at beginning
          of new buffers and is therefore aligned (see 
    void SetLevel(int lv) {
      NET_INT newLevel(lv);
      memcpy((char *)&Level, (char *)&newLevel, sizeof(NET_INT));
    }

    int GetLevel() {
      NET_INT lv;
      memcpy((char *)&lv, (char *)&Level, sizeof(NET_INT));
      return ((int) lv);
    }

    void SetAttributeIndex(int i) {
      NET_INT index(i);
      memcpy((char *)&AttributeIndex, (char *)&index, sizeof(NET_INT));
    }

    int GetAttributeIndex() {
      NET_INT index;
      memcpy((char *)&index, (char *)&AttributeIndex, sizeof(NET_INT));
      return ((int) index);
    }

    void SetArrayIndex(int i) {
      NET_INT index(i);
      memcpy((char *)&ArrayIndex, (char *)&index, sizeof(NET_INT));
    }

    int GetArrayIndex() {
      NET_INT index;
      memcpy((char *)&index, (char *)&ArrayIndex, sizeof(NET_INT));
      return ((int) index);
    }

    void SetListId(int id) {
      NET_INT newID(id);
      memcpy((char *)&ListId, (char *)&newID, sizeof(NET_INT));
    }

    int GetListId() {
      NET_INT id;
      memcpy((char *)&id, (char *)&ListId, sizeof(NET_INT));
      return ((int) id);
    }

    ***********************/

    void print(ostream &out = cout) {
      out << "Level " << Level
	  << ", AttributeIndex " << AttributeIndex
	  << ", ArrayIndex " << ArrayIndex
	  << ", ListId " << ListId
	  << endl;
    }
};


//geb2: This thing is @ beg of new buffer and is therefore 8-byte aligned
// (see SpObjFactory::GenerateMessage)
class SpObjProxyMsg {

  private:

  protected:

    NET_INT ObjectId;
    NET_INT Node;
    NET_INT LocalId;
    NET_INT Type;
    NET_INT FreeID;
    NET_INT Nbytes;
    NET_INT NattributeUpdates;
    char *Pointer;

  public:

    SpObjProxyMsg();


    void SetObjectId(int i) {ObjectId = i;}
    int GetObjectId() {return ObjectId;}

    void SetNode(int n) {Node = n;}
    int GetNode() {return Node;}

    void SetLocalId(int l) {LocalId = l;}
    int GetLocalId() {return LocalId;}

    void SetType(int t) {Type = t;}
    int GetType() {return Type;}

    void SetFreeID(int i) {FreeID = i;}
    int GetFreeID() {return FreeID;}

    void SetNbytes(int n) {Nbytes = n;}
    int GetNbytes() {return Nbytes;}

    void SetNattributeUpdates(int n) {NattributeUpdates = n;}
    int GetNattributeUpdates() {return NattributeUpdates;}

    void SetObjectName(char *ObjectName) {
      strcpy((char *)&this[1], ObjectName);
    }
    char *GetObjectName() {return (char *)&this[1];}

    char *GoToFirstAttributeUpdate(char *ObjectName) {
      Pointer = (char *)&this[1];
      int Offset = strlen(ObjectName) + 1;
      Offset += (sizeof(double) - Offset%sizeof(double)) % sizeof(double);
      Pointer += Offset;
      return Pointer;
    }

    char *GoToNextAttributeUpdate() {
      Pointer = Pointer + ((ATTRIBUTE_MESSAGE_INFO *)Pointer)->GetNbytes();
      return Pointer;
    }

    ATTRIBUTE_MESSAGE_INFO *GetAttributeMessageInfo() {
      return (ATTRIBUTE_MESSAGE_INFO *)Pointer;
    }

    void Print(ostream &out = cout) {
      out << "Printing Object Proxy Message for Object Type "
	  << FreeID << endl;
      GoToFirstAttributeUpdate(GetObjectName());
      for (int i=0; i<NattributeUpdates; i++) {
	out << "Attribute " << i << ": AttributeIndex "
	    << ((ATTRIBUTE_MESSAGE_INFO *)Pointer)->GetAttributeIndex() << endl;
	GoToNextAttributeUpdate();
      }
    }

};

typedef SpObjProxyMsg OBJECT_PROXY_MESSAGE;

#endif
