// SpObjFactory.H Header File

#ifndef ObjectTemplate_Object
#define ObjectTemplate_Object

#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpString.H"
#include "SpStdlib.H"
#include "SpStdarg.H"

#include "qitem.H"
#include "SpExportAttribute.H"
#include "queue.H"

class SpObjTemplate : public C_ITEM {

  private:

    char * const ClassName;
    int ClassID;

    int Ndefines;
    int Nobjects;
    int Ninherits;
    int Nsubscribe;

    int Nattributes;
    char **AttributeName;
    char **AttributeType;
    int *AttributeTypeID;
    char **SubscribeName;
    SpObjTemplate **ClassReference;
    SpObjTemplate *Parent;
    int *ArraySize;

    void (**Method)(void *, va_list &);
    char **MethodName;
    int Nmethods;

    static int ClassCounter;

  protected:

  public:

    SpObjTemplate(char *name, int i);
    ~SpObjTemplate();
    void Inherit(SpObjTemplate *ParentTemplate);

    int GetNmethods() {return Nmethods;}
    void SetMethodName(int i, char *n) {MethodName[i] = strdup(n);}

    void CreateMethodArray(int n) {
      if (n!=0) {
	Nmethods = n;
	typedef void t(void *, va_list &);
 	Method = new t*[n];
	MethodName = new char*[n];
	for (int i=0; i<n; i++) {MethodName[i] = NULL;}
      }
    }

    void AddMethod(void (*m)(void *, va_list &), char *n) {
      int Found = 0;
      for (int i=0; i<Nmethods; i++) {
	if (!strcmp(n, MethodName[i])) {
	  Method[i] = m;
	  Found = 1;
	}
      }
      if (!Found) {
	cout << "Error (AddMethod): Bad method name " << n << endl;
	exit(0);
      }
    }

    int Execute(void *ObjectProxy, char *Mname, va_list &va) {
      for (int i=0; i<Nmethods; i++) {
	if (!strcmp(Mname, MethodName[i])) {
	  Method[i](ObjectProxy, va);
	  return 1;
	}
      }
      return 0;
    }

    SpObjTemplate *GetParent() {return Parent;}

    void SetDefine(int i, char *n, char *t) {
      char *s;
      int a;
      int len;

      s = strchr(n, '[');
      if (s) {
        sscanf(&s[1], "%d", &a);
	ArraySize[i] = a;
	len = strlen(n) - strlen(s);
	AttributeName[i] = new char[len+1];
	memcpy(AttributeName[i], n, len);
	AttributeName[i][len] = 0;
      }else{
        AttributeName[i] = strdup(n);
      }

      AttributeTypeID[i] = MapAttribute(t);
      AttributeType[i] = strdup(MapAttribute(AttributeTypeID[i]));
    }

    void SetObjects(int i, char *n) {
      char *s;
      int a;
      int len;

      s = strchr(n, '[');
      if (s) {
        sscanf(&s[1], "%d", &a);
	ArraySize[i] = a;
	len = strlen(n) - strlen(s);
	AttributeName[i] = new char[len+1];
	memcpy(AttributeName[i], n, len);
	AttributeName[i][len] = 0;
      }else{
        AttributeName[i] = strdup(n);
      }

      AttributeType[i] = strdup("OBJECT");
      AttributeTypeID[i] = MapAttribute("OBJECT");
    }

    void SetSubscribeName(int i, const char *name) {
      int len = strlen(name);
      SubscribeName[i] = (len>0) ? new char [len+1] : (char *) 0;
      memcpy(SubscribeName[i], name, len+1);
    }

    SpObjTemplate *GetClassReference(int i) {return ClassReference[i];}
    int GetArraySize(int i) {return ArraySize[i];}
    void SetArraySize(int i, int s) {ArraySize[i] = s;}

    int GetNdefines() {return Ndefines;}
    int GetNobjects() {return Nobjects;}
    int GetNinherits() {return Ninherits;}
    int GetNsubscribe(void) const { return Nsubscribe; }

    void SetNdefines(int n) {Ndefines = n;}
    void SetNobjects(int n) {Nobjects = n;}
    void SetNinherits(int n) {Ninherits = n;}
    void SetNsubscribe(int n) {
      Nsubscribe = n;
      SubscribeName = (Nsubscribe>0) ? new char * [Nsubscribe] : (char **) 0;
    }
    char *GetSubscribeName(int i) { return (SubscribeName[i]); }

    int GetNattributes() {return Nattributes;}

    void SetAttributeName(int i, char *n) {AttributeName[i] = strdup(n);}
    char *GetAttributeName(int i) {return AttributeName[i];}
    char **GetAttributeName() {return AttributeName;}

    void SetAttributeType(int i, char *t) {AttributeType[i] = strdup(t);}
    char *GetAttributeType(int i) {return AttributeType[i];}
    char **GetAttributeType() {return AttributeType;}

    void SetAttributeTypeID(int i, int t) {AttributeTypeID[i] = t;}
    int GetAttributeTypeID(int i) {return AttributeTypeID[i];}

    char *ObjProxyGetClassName(int i) {return ClassReference[i]->ObjProxyGetClassName();}
    char *ObjProxyGetClassName() {return ClassName;}
    virtual char *GetName() {return ClassName;}

    int GetClassID() {return ClassID;}
    void SetClassID(int i) {ClassID = i;}

    virtual void PrintStream(ostream &out) {
      int i;
      out << endl;
      out << "Printing SpObjTemplate for " << ClassName << " ["
	  << ClassID << "]" << endl;
      for (i=0; i<Nattributes; i++) {
	out << i << ": " << ClassReference[i]->ObjProxyGetClassName() << " "
	    << AttributeType[i] << " ["
	    << AttributeTypeID[i] << "] "
	    << AttributeName[i];
	if (ArraySize[i] > 1) out << " [" << ArraySize[i] << "]";
	cout << endl;
      }

      for (i=0; i<Nmethods; i++) {
	out << "Method[" << i << "] " << MethodName[i] << endl;
      }
    }

};

typedef SpObjTemplate OBJECT_TEMPLATE;

#endif
