#include "SpProxyObjHandle.H"
#include "SpBinaryTree.H"
#include "SpList.H"
#include "RB_SpFrameworkFuncs.H"
#include "RB_SpBinaryTree.H"
#include "RB_SpDefineClass.H"

RB_DEFINE_CLASS(SpProxyObjHandle);


void 
SpProxyObjHandle::CreateMessage(char* &msg, int &len) const{
  len = sizeof(NET_INT) * 4;
  msg = new char [len];

  CreateMessage(msg);
}

void 
SpProxyObjHandle::CreateMessage(char* &msg) const{
  char *p = msg;

  memcpy(p, &NodeId,   sizeof(NET_INT)); p += sizeof(NET_INT);
  memcpy(p, &ObjManId, sizeof(NET_INT)); p += sizeof(NET_INT);
  memcpy(p, &SimObjId, sizeof(NET_INT)); p += sizeof(NET_INT);
  memcpy(p, &GlobalId, sizeof(NET_INT));
}

void 
SpProxyObjHandle::CreateMessage(SpBinaryTree &objs, char* &msg, int &len){
  const int msg_len = sizeof(NET_INT) * 4;
  int num_objs;
  SpProxyObjHandle *obj;
  char *cp;

  msg = (char *) 0;
  len = 0;

  num_objs = objs.GetNumElements();

  if (num_objs<=0) return;

  len = msg_len * num_objs;
  msg = (len>0) ? new char [len] : (char *) 0;

  cp = msg;
  obj = (SpProxyObjHandle *) objs.GetFirstElement();
  while (obj){
    obj->CreateMessage(cp); cp += msg_len;
    obj = (SpProxyObjHandle *) objs.GetNextElement();
  }
}

void 
SpProxyObjHandle::CreateMessage(SpList &objs, char* &msg, int &len){
  const int msg_len = sizeof(NET_INT) * 4;
  int num_objs;
  SpProxyObjHandle *obj;
  char *cp;

  msg = (char *) 0;
  len = 0;

  num_objs = objs.GetNumElements();

  if (num_objs<=0) return;

  len = msg_len * num_objs;
  msg = (len>0) ? new char [len] : (char *) 0;

  cp = msg;
  obj = (SpProxyObjHandle *) objs.GetFirstElement();
  while (obj){
    obj->CreateMessage(cp); cp += msg_len;
    obj = (SpProxyObjHandle *) objs.GetNextElement();
  }
}

void 
SpProxyObjHandle::CreateObject(const char *msg, int &obj_bytes, SpProxyObjHandle &item){
  const char *p = msg;
  NET_INT node, type, local_id, global_id;

  obj_bytes = 0;

  memcpy(&node,      p, sizeof(NET_INT)); p += sizeof(NET_INT); obj_bytes += sizeof(NET_INT);
  memcpy(&type,      p, sizeof(NET_INT)); p += sizeof(NET_INT); obj_bytes += sizeof(NET_INT);
  memcpy(&local_id,  p, sizeof(NET_INT)); p += sizeof(NET_INT); obj_bytes += sizeof(NET_INT);
  memcpy(&global_id, p, sizeof(NET_INT));                       obj_bytes += sizeof(NET_INT);

  item.SetNode(node);
  item.SetObjType(type);
  item.SetLobalId(local_id);
  item.SetGlobalId(global_id);
}

void 
SpProxyObjHandle::CreateObj(const char* &buffer, int num_items, int &nbytes, SpList &items){
  int i, obj_bytes;
  const char *cp;
  SpProxyObjHandle *object_handle;

  nbytes = 0;
  for (obj_bytes=0, cp=buffer, i=0; i<num_items; i++, cp+=obj_bytes){
    object_handle = new SpProxyObjHandle;
    SpProxyObjHandle::CreateObject(cp, obj_bytes, *object_handle); nbytes += obj_bytes;
    items.Insert(object_handle);
  }
}

void 
SpProxyObjHandle::CreateObj(const char* &buffer, int num_items, int &nbytes, RB_SpBinaryTree &items){
  int i, obj_bytes;
  const char *cp;
  SpProxyObjHandle *object_handle;

  nbytes = 0;
  for (obj_bytes=0, cp=buffer, i=0; i<num_items; i++, cp+=obj_bytes){
    object_handle = RB_NEW_SpProxyObjHandle(new SpProxyObjHandle);
    SpProxyObjHandle::CreateObject(cp, obj_bytes, *object_handle); nbytes += obj_bytes;
    items.Insert(object_handle, object_handle->GetGlobalId());
  }
}

void 
SpProxyObjHandle::RBDestroy(SpProxyObjHandle *handle){
  RB_DELETE_SpProxyObjHandle(handle);
}

void 
SpProxyObjHandle::Destroy(SpProxyObjHandle *handle){
  delete handle;
}

void RB_Register_SpProxyObjHandle()
{
  RB_REGISTER_CLASS_SpProxyObjHandle();
}
