// SpProxyObjHandle.H header file

#ifndef PROXY_OBJECT_HANDLE_H
#define PROXY_OBJECT_HANDLE_H

#include "SpNetTypes.H"
#include "SpStdlib.H"
#include "SpObjHandle.H"

class RB_SpBinaryTree;
class SpBinaryTree;
class SpList;


class SpProxyObjHandle : public SpObjHandle {

  private:
    SpProxyObjHandle  (SpProxyObjHandle&) { abort(); }
    SpProxyObjHandle& operator=(SpProxyObjHandle&) { abort(); return (*this); }
   
    NET_INT GlobalId; 
    void CreateMessage(char* &msg) const;

  public:
    SpProxyObjHandle (void)
      : GlobalId(-1) {}

    SpProxyObjHandle (int local_id, int node, int object_type, int global_id)
      : SpObjectHandle(node, object_type, local_id), GlobalId(global_id) {}

    ~SpProxyObjHandle (void) {}

    int GetNode(void)      const { return (NodeId);   }
    int GetObjType(void)   const { return (ObjManId); }
    int GetLocalId(void)   const { return (SimObjId); }
    int GetGlobalId(void)  const { return (GlobalId); }

    void SetNode(int i)     { NodeId   = i; }
    void SetObjType(int i)  { ObjManId = i; }
    void SetLobalId(int i)  { SimObjId = i; }
    void SetGlobalId(int i) { GlobalId = i; }

    void CreateMessage(char* &msg, int &len) const;

    static void CreateObject(const char *msg, int &obj_bytes, SpProxyObjHandle &item);
    static void CreateObj(const char* &buffer, int num_items, int &nbytes, RB_SpBinaryTree &items);
    static void CreateObj(const char* &buffer, int num_items, int &nbytes, SpList &items);
    static void CreateMessage(SpList &objs, char* &msg, int &len);
    static void CreateMessage(SpBinaryTree &objs, char* &msg, int &len);

    static void RBDestroy(SpProxyObjHandle *handle);
    static void Destroy(SpProxyObjHandle *handle);

    static int GetGlobalId(void *item) { return (((SpProxyObjHandle *) item)->GetGlobalId()); }
};

typedef SpProxyObjHandle SpProxyObjectHandle;

#endif
