// SpSubscriberHandle.C source file

#include "SpSubscriberHandle.H"
#include "RB_SpBinaryTree.H"

#include "RB_SpDefineClass.H"
RB_DEFINE_CLASS(SpSubscriberHandle);

class SpSubscribeDistList {
public:
  SpSubscribeDistList(void)
    : Id(-1), NumSubscribe(0) {}

  SpSubscribeDistList(int id)
    : Id(id), NumSubscribe(0) {}

  ~SpSubscribeDistList(void) {}

  int GetId(void)            const { return (Id);             }
  int GetNumSubscribe(void)  const { return (NumSubscribe);   }
  int IncrNumSubscribe(void)       { return (++NumSubscribe); }
  int DecrNumSubscribe(void)       { return (--NumSubscribe); }
  int ResetNumSubscribe(void)      { return (NumSubscribe=0); }
    
private:
  int    Id;
  RB_int NumSubscribe;
};

RB_DEFINE_CLASS(SpSubscribeDistList);

void RB_Register_SpSubscribeDistList(){
  RB_REGISTER_CLASS_SpSubscribeDistList();
}

/////////////////////////////////////////////////////////////////////////////////

SpSubscriberHandle::SpSubscriberHandle()
  : SubscribeDistLists(new RB_SpBinaryTree) {}

SpSubscriberHandle::SpSubscriberHandle(int _local_id, int _node, int _object_type, int _global_id)
  : SpProxyObjHandle(_local_id, _node, _object_type, _global_id),
    SubscribeDistLists(new RB_SpBinaryTree) {}

SpSubscriberHandle::~SpSubscriberHandle() {
  SpSubscribeDistList *manager;
  while ((manager = (SpSubscribeDistList *) SubscribeDistLists->RemoveFirstElement()))
    RB_DELETE_SpSubscribeDistList(manager);
  delete SubscribeDistLists;
}

int 
SpSubscriberHandle::getNumSubscribe(){ 
  int n = 0;

  SpSubscribeDistList *manager = (SpSubscribeDistList *) SubscribeDistLists->GetFirstElement();
  while (manager){
    n += manager->GetNumSubscribe();
    manager = (SpSubscribeDistList *) SubscribeDistLists->GetNextElement();
  }
  return (n);
}

int 
SpSubscriberHandle::getNumSubscribe(int id) { 
  SpSubscribeDistList *dist_list = GetSubscribeDistList(id);
  return (dist_list->GetNumSubscribe());
}

int 
SpSubscriberHandle::incrNumSubscribe(int id) { 
  SpSubscribeDistList *dist_list = GetSubscribeDistList(id);
  return (dist_list->IncrNumSubscribe());
}

int 
SpSubscriberHandle::decrNumSubscribe(int id) { 
  SpSubscribeDistList *dist_list = GetSubscribeDistList(id);
  return (dist_list->DecrNumSubscribe());
}

int 
SpSubscriberHandle::resetNumSubscribe(int id) { 
  SpSubscribeDistList *dist_list = GetSubscribeDistList(id);
  return (dist_list->ResetNumSubscribe());
}

SpSubscribeDistList *
SpSubscriberHandle::GetSubscribeDistList(int id){
  SpSubscribeDistList *manager = (SpSubscribeDistList *) SubscribeDistLists->Find(id);
  if (manager==(SpSubscribeDistList *)0){
    manager = RB_NEW_SpSubscribeDistList(new SpSubscribeDistList(id));
    SubscribeDistLists->Insert(manager, manager->GetId());
  }
  return (manager);
}

void RB_Register_SpSubscriberHandle() {
  RB_REGISTER_CLASS_SpSubscriberHandle();
}
