// SpDistList.C

#include "SpDistList.H"
#include "SpString.H"
#include "S_SpHLA.H"
#include "RB_SpBinaryTree.H"
#include "RB_SpList.H"
#include "RB_SpDefineClass.H"

RB_DEFINE_CLASS(SpDistList);
RB_DEFINE_CLASS(SpProxyObjHandle);

void RB_Register_SpDistList()
{
  RB_REGISTER_CLASS_SpDistList();
}

SpDistList::SpDistList()
  : SimObj((S_SpHLA *) 0), Name((char *) 0), Id(-1),
    ObjHandles(new RB_SpBinaryTree), 
    RemoveHandles(new RB_SpList), AddHandles(new RB_SpList) {}

SpDistList::SpDistList(const char *name, int id, S_SpHLA *simobj)
  : SimObj(simobj), Name((char *) 0), Id(id),
    ObjHandles(new RB_SpBinaryTree), 
    RemoveHandles(new RB_SpList), AddHandles(new RB_SpList) {

  if (name) Name = strdup(name);
}

SpDistList::~SpDistList(){
  if (Name) free(Name);
  while (ObjHandles->RemoveFirstElement());
  delete ObjHandles;
  delete RemoveHandles;
  delete AddHandles;
}

int
SpDistList::Insert(const SpProxyObjHandle *handle){

  if (ObjHandles->Find(handle->GetGlobalId())) return (0);

  ObjHandles->Insert((void *) handle, handle->GetGlobalId());
  AddHandles->Insert((void *) handle);
  Dirty();

  return (1);
}

int
SpDistList::Remove(const SpProxyObjHandle *handle){
  const SpProxyObjHandle *remove = (const SpProxyObjHandle *) 
                                   ObjHandles->Remove(handle->GetGlobalId());

  if (remove==(const SpProxyObjHandle *) 0) return (0);

  RemoveHandles->Insert(RB_NEW_SpProxyObjHandle(new SpProxyObjHandle(remove->GetLocalId(), 
                                                                    remove->GetNode(), 
                                                                    remove->GetObjType(), 
                                                                    remove->GetGlobalId())));
  Dirty();

  return (1);
}

int
SpDistList::GetNumObjHandles(void) {
  return (ObjHandles->GetNumElements());
}

void
SpDistList::Dirty(){
  SimObj->ScheduleProxyDistUpdate();
}

int
SpDistList::IsDirty() {
  int num_modification = AddHandles->GetNumElements() + RemoveHandles->GetNumElements();
  return (num_modification>0);
}

void
SpDistList::GetChanges(SpBinaryTree &add, SpBinaryTree &remove){
  const SpProxyObjHandle *handle;

  handle = (const SpProxyObjHandle *) AddHandles->GetFirstElement();
  while (handle){
    add.Insert((void *) handle, handle->GetGlobalId());
    handle = (const SpProxyObjHandle *) AddHandles->GetNextElement();
  }

  handle = (const SpProxyObjHandle *) RemoveHandles->GetFirstElement();
  while (handle){
    remove.Insert((void *) handle, handle->GetGlobalId());
    handle = (const SpProxyObjHandle *) RemoveHandles->GetNextElement();
  }
}

void
SpDistList::UnDirty(){
  SpProxyObjHandle *handle;

  while (AddHandles->pop_top());

  while ((handle = (SpProxyObjHandle *) RemoveHandles->pop_top()))
    RB_DELETE_SpProxyObjHandle(handle);  
}
