#include "SpProxyObjHandle.H"
#include "SpBinaryTree.H"
#include "RB_SpFrameworkFuncs.H"
#include "RB_SpBinaryTree.H"
#include "RB_SpDefineClass.H"

RB_DEFINE_CLASS(SpProxyObjHandle);

void 
SpProxyObjHandle::CreateMessage(char* &msg, int &len) const{
  len = sizeof(NET_INT) * 4;
  msg = new char [len];
  char *p = msg;

  memcpy(p, &NodeId,   sizeof(NET_INT)); p += sizeof(NET_INT);
  memcpy(p, &ObjManId, sizeof(NET_INT)); p += sizeof(NET_INT);
  memcpy(p, &SimObjId, sizeof(NET_INT)); p += sizeof(NET_INT);
  memcpy(p, &GlobalId, sizeof(NET_INT));
}

void 
SpProxyObjHandle::CreateMessage(SpBinaryTree &objs, char* &msg, int &len){
  char *cp;
  int i, total_len, num_objs;
  char **obj_msg;
  int   *obj_msg_len;
  SpProxyObjHandle *obj;

  msg = (char *) 0;
  len = 0;

  num_objs = objs.GetNumElements();

  if (num_objs<=0) return;

  obj_msg     = new char * [num_objs];
  obj_msg_len = new int    [num_objs];
  total_len   = 0;

  i = 0;
  obj = (SpProxyObjHandle *) objs.GetFirstElement();
  while (obj){
    obj->CreateMessage(obj_msg[i], obj_msg_len[i]);
    total_len += obj_msg_len[i];
    i++;
    obj = (SpProxyObjHandle *) objs.GetNextElement();
  }

  msg = (total_len>0) ? new char [total_len] : (char *) 0;
  len = total_len;
  cp  = msg;

  for (i=0; i<num_objs; i++){
    memcpy(cp, obj_msg[i], obj_msg_len[i]); cp += obj_msg_len[i];
    delete [] obj_msg[i];
  }

  delete [] obj_msg;
  delete [] obj_msg_len;
}

void 
SpProxyObjHandle::CreateObject(const char *msg, int &obj_bytes, SpProxyObjHandle &item){
  const char *p = msg;
  NET_INT node, type, local_id, global_id;

  obj_bytes = 0;

  memcpy(&node,      p, sizeof(NET_INT)); p += sizeof(NET_INT); obj_bytes += sizeof(NET_INT);
  memcpy(&type,      p, sizeof(NET_INT)); p += sizeof(NET_INT); obj_bytes += sizeof(NET_INT);
  memcpy(&local_id,  p, sizeof(NET_INT)); p += sizeof(NET_INT); obj_bytes += sizeof(NET_INT);
  memcpy(&global_id, p, sizeof(NET_INT));                       obj_bytes += sizeof(NET_INT);

  item.SetNode(node);
  item.SetObjType(type);
  item.SetLobalId(local_id);
  item.SetGlobalId(global_id);
}

void 
SpProxyObjHandle::CreateObj(const char* &buffer, int num_items, int &nbytes, SpBinaryTree &items){
  int i, obj_bytes;
  const char *cp;
  SpProxyObjHandle *object_handle;

  nbytes = 0;
  for (obj_bytes=0, cp=buffer, i=0; i<num_items; i++, cp+=obj_bytes){
    object_handle = RB_NEW_SpProxyObjHandle(new SpProxyObjHandle);
    SpProxyObjHandle::CreateObject(cp, obj_bytes, *object_handle); nbytes += obj_bytes;
    items.Insert(object_handle, object_handle->GetGlobalId());
  }
}

void 
SpProxyObjHandle::CreateObj(const char* &buffer, int num_items, int &nbytes, RB_SpBinaryTree &items){
  int i, obj_bytes;
  const char *cp;
  SpProxyObjHandle *object_handle;

  nbytes = 0;
  for (obj_bytes=0, cp=buffer, i=0; i<num_items; i++, cp+=obj_bytes){
    object_handle = RB_NEW_SpProxyObjHandle(new SpProxyObjHandle);
    SpProxyObjHandle::CreateObject(cp, obj_bytes, *object_handle); nbytes += obj_bytes;
    items.Insert(object_handle, object_handle->GetGlobalId());
  }
}

void 
SpProxyObjHandle::Destroy(SpProxyObjHandle *handle){
  RB_DELETE_SpProxyObjHandle(handle);
}


void RB_Register_SpProxyObjHandle()
{
  RB_REGISTER_CLASS_SpProxyObjHandle();
}
