// O_SpFedGhost.C source file

#include "O_SpFedGhost.H"

#include "RB_SpDefineClass.H"
#include "SpExecute.H"

RB_DEFINE_CLASS(S_SpFedGhost);
EVENT_TYPE(CREATE);
//----------------------------------------------------------------------------
// O_SpFedGhost constructor
//----------------------------------------------------------------------------
O_SpFedGhost::O_SpFedGhost() {

  cout << "Creating O_SpFedGhost" << endl;

  AddEventHandler("CreateFedGhost", &CreateFedGhost);

  // temporary measure to handle create/destroy fedex
  if (NODE == 0) {
    AddEventHandler("Create Fedex", &CreateFedEx);
    AddEventHandler("Destroy Fedex", &DestroyFedEx);
    FEDs = NULL;
  }

  N_TOT = 0;
  N_LOC = 0;

  reserve_n_objects(10);

}

void H_SpCreateFedGhost::Process(char *,
				 SIMTIME &timeTag,
				 char *data,
				 int dataBytes,
				 int externalId) {
  
  cout << "H_SpCreateFedGhost creating " << data << endl;
  
  SpObjHandle objHandle;
  
  int buffLen = dataBytes + sizeof(int);
  char *buff = new char[buffLen];
  memcpy(buff, data, dataBytes);
  memcpy(&buff[dataBytes], (char *)&externalId, sizeof(int));
  
#ifndef __SPEEDES_NT__
  CreateSimObj(timeTag, ObjectType, Node, objHandle, buff, buffLen);
#else
  SpSimObjMgr *simObjMgr = (SpSimObjMgr *)CurrentEvent->get_manager(ObjectType);

  objHandle.ObjManId = ObjectType;
  objHandle.NodeId = Node;
  objHandle.SimObjId = simObjMgr->GenerateNewLocalId();

  CREATE_MESS createSimObjMess;
  createSimObjMess.NewLocalId = objHandle.SimObjId;

  return ScheduleEvent(timeTag,
		       CREATE,
		       ObjectType,
		       -1,
		       Node,
		       createSimObjMess,
		       buff,
		       buffLen);

#endif

  
  cout << "LocalId = " << objHandle.SimObjId << endl;
  cout << "ExternalId = " << externalId << endl;
  
  delete [] buff;
  
}

SpSimObj *O_SpFedGhost::CreateSimObj() {return RB_NEW_S_SpFedGhost();}

inline void klugeFunctionForVisualAgeCPP(void* object){
  delete (S_SpFedGhost*)object;
}

void O_SpFedGhost::DestroySimObj(SpSimObj *simObj) {
  //RB_DELETE_S_SpFedGhost((S_SpFedGhost *)simObj);
  RB_DELETE_HELPER(simObj, klugeFunctionForVisualAgeCPP);
}


// temporary measure to handle create/destroy fedex
void H_SpCreateFedEx::Process(char *command, SIMTIME &timeTag,
			      char *data, int dataBytes,
			      int externalId)
{
  O_SpFedGhost *objMgr = (O_SpFedGhost*)SimObj;

  if (!objMgr->FEDs)
    objMgr->FEDs = new SpBinaryTree();

  char *ret = "1";
  int retlen = 2;
  char *fedex = data;
  data += strlen(data)+1;

  if (objMgr->FEDs->Find (fedex)) {
    // fedex already exists
    ret = data;
    retlen = strlen(ret)+1;
  }
  else {
    char *FED = strdup(data);
    objMgr->FEDs->Insert (FED, fedex);
  }
  HostUser->SendNamedData("Create Fedex Return",
			  timeTag,
			  externalId,
			  ret,
			  retlen,
			  GlobalId);
}
void H_SpDestroyFedEx::Process(char *command, SIMTIME &timeTag,
			       char *data, int dataBytes,
			       int externalId)
{
  O_SpFedGhost *objMgr = (O_SpFedGhost*)SimObj;

  char *ret = "1";
  int retlen = 2;

  if ((!objMgr->FEDs) || (objMgr->FEDs->Find (data) == NULL)) {
    // fedex does not exist
    ret = data;
    retlen = dataBytes;
  }
  else {
    char *FED = (char*) objMgr->FEDs->Remove (data);
    free(FED);
  }
  HostUser->SendNamedData("Destroy Fedex Return",
			  timeTag,
			  externalId,
			  ret,
			  retlen,
			  GlobalId);
}
