// S_SpFedGhost.C source file

#include "S_SpFedGhost.H"
#include "SpProc.H"
#include "RB_SpBinaryTree.H"
#include "SpScheduleEvent.H"
#include "SpBarrier.H"
#include "SpExtMsgSimObjIds.H"
#include "O_SpFedGhost.H"
#include "RB_SpDefineClass.H"

RB_DEFINE_CLASS(SpBarrier);

extern RB_SpBinaryTree *Barriers;
extern int SpGateway;
extern SpSimTime *IgnoreTimeGrantGvt;

RB_DEFINE_CLASS(SpSimTime);
RB_DEFINE_CLASS(SpProcStackElement)
RB_DEFINE_CLASS(S_SpFedGhost)
enum {
  BARRIER_REACHED,
  OUTPUT_MESSAGE,
  NER_TIMEOUT
};

//----------------------------------------------------------------------------
// S_SpFedGhost constructor
//----------------------------------------------------------------------------
S_SpFedGhost::S_SpFedGhost() : S_SpHLA("Federate") {


  AddEventHandler("Discover Object", &DiscoverObject);
  AddEventHandler("Reflect Attributes", &ReflectAttributes);

  AddEventHandler("Subscribe Object", &SubscribeObject);
  AddEventHandler("Unsubscribe Object", &UnsubscribeObject);

  AddEventHandler("Subscribe Interaction", &SubInteractionHandler);
  AddEventHandler("Unsubscribe Interaction", &UnsubInteractionHandler);
  AddEventHandler("Send Interaction", &SendInteractionHandler);

  AddEventHandler("Set Time Lag", &SetTimeLagHandler);
  AddEventHandler("Next Event Request", &NextEventRequest);

  AddEventHandler("Resign Fed", &ResignFedHandler);


  NextEventRequestTime = 0.0; // mjp should this be -INFINITY
  MinMsgTime = INFINITY;
  TimeLag = 0.0;
  ExternalId = -1;

  FederationData = NULL;

  TimeRequestInterruptSem = 0;
  TimeRequestSem = 0;

  InitDataFlag = 0;

  if (SpPoDataBase::DoNotAllocateMemoryInConstructor == 1){
    return;
  }

  DEFINE_ATTRIBUTE(FedName, "FedName");
  RB_cout << "Creating S_SpFedGhost on node " << NODE << endl;

}


//----------------------------------------------------------------------------
// InitData() is automatically called when constructing this object
// dynamically
//----------------------------------------------------------------------------
void S_SpFedGhost::InitData(char *data) {

  RB_cout << "Initializing S_SpFedGhost: " << data << endl;

  FedName = RB_strdup(data);
  set_NAME(data);

  if (!strcasecmp(data, "HLA Gateway")) {
    RB_cout << "Setting Gateway flag to 1" << endl;
    if (SpGateway) {
      Gateway = 1;
    }else{
      RB_cout << "Warning, Gateway not enabled in speedes.par file" << endl;
      Gateway = 0;
    }
  }else{
    RB_cout << "Resetting Gateway flag to 0" << endl;
    Gateway = 0;
  }

  if (!Gateway) {
    data += strlen(data) + 1;
    FederationData = (char*)((O_SpFedGhost*)GetSimObjMgr(OBJECT_TYPE))->FEDs->Find(data);
  }

  int externalId;
  memcpy((char *)&externalId, &data[strlen(data)+1], sizeof(int));
  ExternalId = externalId;

  RB_cout << "ExternalId = " << ExternalId << endl;

  TimeLag = 0.0;

  SpObjHandle objHandle;
  objHandle.NodeId = NODE;
  objHandle.ObjManId = OBJECT_TYPE;
  objHandle.SimObjId = LOCAL_ID;

  SpSimTime time = GetTime();
  ScheduleCoordinateTime(time, objHandle, 12345);
}

//----------------------------------------------------------------------------
// CoordinateTime() process manages the next event request / time advance
// grant cycle between SPEEDES and the external federate.
//----------------------------------------------------------------------------
void S_SpFedGhost::CoordinateTime(int dummy) {

  RB_cout << "CoordinateTime() with dummy arg = " << dummy << endl;

P_VAR
  P_LV(int, successFlag);
  P_LV(int, WakeUpReason);
  P_LV(int, LoopForever);
  P_LV(SpSimTime, timeOut);
  P_LV(SpSimTime, BarrierTime);
  P_LV(SpSimTime, CurrentTime);
P_BEGIN(2)

  if (InitDataFlag == 0) {
    SpExtMsgSimObjIds extMsgSimObjIds;
    extMsgSimObjIds.NodeId = NODE;
    extMsgSimObjIds.MgrId = OBJECT_TYPE;
    extMsgSimObjIds.LocalId = LOCAL_ID;
    extMsgSimObjIds.ExternalId = ExternalId;

    char *data = (char *)&extMsgSimObjIds;
    int dataBytes = sizeof(SpExtMsgSimObjIds);

    if (!Gateway) {
      data = new char[dataBytes + strlen(FederationData) + 1];
      memcpy (data, (char *)&extMsgSimObjIds, dataBytes);
      strcpy (data+dataBytes, FederationData);
      dataBytes += strlen(FederationData) + 1;
    }

    GrantedTime = GetTime();
    cerr << "*** Sending SimObj Ids msg at time " << GrantedTime << endl;
    HostUser->SendNamedData("SimObj Ids",
			    GrantedTime,
			    ExternalId,
			    data,
			    dataBytes,
			    GLOBAL_ID);
    InitDataFlag = 1;
  }

  RB_cout << "First entering CoordinateTime() with dummy arg = "
	  << dummy << endl;
 
  SpBarrier *barrier;

  barrier = RB_NEW_SpBarrier();
  barrier->SetSimObj(this);
  barrier->SetGateway(Gateway);
  barrier->SetTimeLag(TimeLag);
  barrier->Set_TIME_GRANT();

  *barrier = GrantedTime;
  *barrier += TimeLag;

  Barriers->Insert(barrier, ExternalId);
  BarrierTime = *barrier;

  LoopForever = 1;
  while(LoopForever) {

    TimeRequestSem = 0;
    RB_cout << "WAIT_FOR_LOGICAL(1) at time " << GetTime() << endl;
    WAIT_FOR_LOGICAL(1, TimeRequestSem, -1, successFlag);
    RB_cout << "WAKEUP(1) at time " << GetTime() << endl;

    while(LoopForever) {

      timeOut = NextEventRequestTime;
      WakeUpReason = NER_TIMEOUT;

      if (MinMsgTime < timeOut) {
	RB_cout << "1. OUTPUT_MESSAGE " << MinMsgTime << endl;
	timeOut = MinMsgTime;
	WakeUpReason = OUTPUT_MESSAGE;
      }

      if (BarrierTime < timeOut) {
	RB_cout << "1. BARRIER_REACHED" << endl;
	timeOut = BarrierTime;
	WakeUpReason = BARRIER_REACHED;
      }

      timeOut -= GetTime();

      if ((0 <= timeOut.GetTime()) || (WakeUpReason == BARRIER_REACHED)) {

	TimeRequestInterruptSem = 0;

	RB_cout << "WAIT_FOR_LOGICAL(2) at time " << GetTime()
		<< " with timeOut = "
		<< timeOut << endl;

	WAIT_FOR_LOGICAL(2,
			 TimeRequestInterruptSem,
			 timeOut.GetTime(),
			 successFlag);

	RB_cout << "WAKEUP(2) at time " << GetTime()
		<< " with TimeRequestInterruptSem "
		<< TimeRequestInterruptSem << endl;

/*
** This is the case where an output message interrupts our wait so we know
** that the grant should be the output message time
*/

	if ((0 != TimeRequestInterruptSem) ||
	    (WakeUpReason == OUTPUT_MESSAGE)) {
	  RB_cout << "2. OUTPUT_MESSAGE " << MinMsgTime << endl;
	  WakeUpReason = OUTPUT_MESSAGE;
	  GrantedTime = MinMsgTime;
	  break;
	}

/*
** This is the case where we reached our barrier so we should remove it
** and wait again until either the next event request time timeouts or an
** output message needs to be sent
*/

	if (WakeUpReason == BARRIER_REACHED) {
	  RB_cout << "2. BARRIER_REACHED" << endl;
	  barrier = (SpBarrier *)Barriers->Remove(ExternalId);
	  if (barrier) {
	    RB_DELETE_SpBarrier(barrier);
	    BarrierTime = 1.0e20;
	  }else{
	    RB_cout << "Error, ExternalId "
		    << ExternalId << "  not in Barriers" << endl;
	  }
	}else{

/*
** This is the case where we timed out meaning that we reached the
** next event request time
*/

	  GrantedTime = NextEventRequestTime;
	  break;

	}

      }else{

/*
** This is the case where the time out is negative (i.e., either the
** next event request time or an output message time is in our past, but we
** should be protected by our barrier from going too far ahead
*/

	RB_cout << "In Past, but Ok" << endl;
	GrantedTime = NextEventRequestTime;
	if(MinMsgTime < NextEventRequestTime)
	  GrantedTime = MinMsgTime;
	break;

      }
    }

/*
** Set up new barrier and grant the time
*/

    BarrierTime = GrantedTime;
    BarrierTime += TimeLag;
    ResetMinMsgTime();

    barrier = (SpBarrier *)Barriers->Remove(ExternalId);
    if (barrier) {
      RB_DELETE_SpBarrier(barrier);
    }

    if (BarrierTime < GetTime())
    {
      CurrentTime = GetTime();
      BarrierTime.SetPriority1(CurrentTime.GetPriority1());
      BarrierTime.SetPriority2(CurrentTime.GetPriority2());
      BarrierTime.SetCounter(CurrentTime.GetCounter());
      BarrierTime.SetUniqueId(CurrentTime.GetUniqueId());
    }

    barrier = RB_NEW_SpBarrier();
    barrier->SetSimObj(this);
    barrier->SetGateway(Gateway);
    barrier->SetTimeLag(TimeLag);
    *barrier = BarrierTime;

    if (WakeUpReason == OUTPUT_MESSAGE) barrier->Set_OUTPUT_MSG();
    if (WakeUpReason == BARRIER_REACHED) barrier->Set_TIME_GRANT();
    if (WakeUpReason == NER_TIMEOUT) barrier->Set_TIME_GRANT();

    Barriers->Insert(barrier, ExternalId);

    RB_cout << NAME << " <Granted time is " << GrantedTime << ">" 
	    << endl;
    HostUser->RB_SendNamedData("TimeAdvanceGrant",
			       GetTime(),
			       ExternalId,
			       (char *)&GrantedTime,
			       sizeof(SpSimTime),
			       GLOBAL_ID);

  }

P_END

}


void
S_SpFedGhost::Resign() {
  /* for now, this method removes any barriers the fed ghost set up
   * and removes all the event handlers
   */

  SpBarrier *barrier = (SpBarrier *)Barriers->Remove(ExternalId);
  if (barrier) {
    RB_DELETE_SpBarrier(barrier);
  }else{
    RB_cout << "Error, ExternalId " << ExternalId << " not in Barriers" 
	    << endl;
  }

  RemoveEventHandler("Discover Object", &DiscoverObject);
  RemoveEventHandler("Reflect Attributes", &ReflectAttributes);

  RemoveEventHandler("Subscribe Object", &SubscribeObject);
  RemoveEventHandler("Unsubscribe Object", &UnsubscribeObject);

  RemoveEventHandler("Subscribe Interaction", &SubInteractionHandler);
  RemoveEventHandler("Unsubscribe Interaction", &UnsubInteractionHandler);
  RemoveEventHandler("Send Interaction", &SendInteractionHandler);

  RemoveEventHandler("Set Time Lag", &SetTimeLagHandler);
  RemoveEventHandler("Next Event Request", &NextEventRequest);

  RemoveEventHandler("Resign Fed", &ResignFedHandler);

  // interaction handlers???
}


void RB_Register_SpSimTime()
{
  RB_REGISTER_CLASS_SpSimTime();
}

void RB_Register_S_SpFedGhost()
{
  RB_REGISTER_CLASS_S_SpFedGhost();
}

