// S_SpFedGhost.H header file defines the ghost federate class

#ifndef S_SpFedGhostGate
#define S_SpFedGhostGate

#include "S_SpHLA.H"
#include "SpProc.H"
#include "RB.H"
#include "Semaphore.H"
#include "SpScheduleEvent.H"

#include "H_SpFedGhostDiscoverObject.H"
#include "H_SpFedGhostReflectAttributes.H"

#include "H_SpFedGhostSubscribeObject.H"
#include "H_SpFedGhostUnsubscribeObject.H"

#include "H_SpFedGhostSubscribeInteraction.H"
#include "H_SpFedGhostUnsubscribeInteraction.H"
#include "H_SpFedGhostSendInteraction.H"

#include "H_SpFedGhostSetTimeLag.H"
#include "H_SpFedGhostNextEventRequest.H"

#include "H_SpFedGhostResign.H"


class S_SpFedGhost : public S_SpHLA {

  private:

    H_SpFedGhostDiscoverObject DiscoverObject;
    H_SpFedGhostReflectAttributes ReflectAttributes;

    H_SpFedGhostSubscribeObject SubscribeObject;
    H_SpFedGhostUnsubscribeObject UnsubscribeObject;

    H_SpFedGhostSubscribeInteraction SubInteractionHandler;
    H_SpFedGhostUnsubscribeInteraction UnsubInteractionHandler;
    H_SpFedGhostSendInteraction SendInteractionHandler;

    H_SpFedGhostSetTimeLag SetTimeLagHandler;
    H_SpFedGhostNextEventRequest NextEventRequest;

    H_SpFedGhostResignFederate ResignFedHandler;

    STRING_ATTRIBUTE FedName;
    LogicalSemaphore TimeRequestSem;
    LogicalSemaphore TimeRequestInterruptSem;

    RB_int ExternalId;
    RB_int Gateway;

    char *FederationData;  // name of FED file, not used with gateway

    SpSimTime NextEventRequestTime;
    SpSimTime MinMsgTime;
    RB_double TimeLag;

    int InitDataFlag;
    SpSimTime GrantedTime;

  protected:

  public:

    S_SpFedGhost();
    void InitData(char *data);
    void CoordinateTime(int dummy);
    int GetExternalId() {return ExternalId;}
    void SetTimeLag(double timeLag) {TimeLag = timeLag;}

    void SetNextEventRequestTime(SpSimTime &time) {
      RB_memcpy((char *)&NextEventRequestTime,
		(char *)&time,
		sizeof(SpSimTime));
      TimeRequestSem = 1;
    }

    void ResetMinMsgTime() {
      SpSimTime time;
      time.SetTime(1.0e20);
      RB_memcpy((char *)&MinMsgTime, (char *)&time, sizeof(SpSimTime));
      if (1 == TimeRequestInterruptSem) {TimeRequestInterruptSem = 0;}
    }

    void SetMinMsgTime(SpSimTime &time) {

      if (0 == TimeRequestInterruptSem) {
	TimeRequestInterruptSem = 1;
      }
      
      if (time < MinMsgTime) {
	RB_memcpy((char *)&MinMsgTime, (char *)&time, sizeof(SpSimTime));
      }
    }

    /* Clears out any barriers and removes event handlers for now
     */
    void Resign();

    /* This is called by SpBTW to determine the minimum outgoing message time
     */
    virtual SpSimTime GetMinGatewayMsgTime() { return MinMsgTime; }
};


DEFINE_EVENT_METHOD_1_ARG(CoordinateTime,
			  10,
			  S_SpFedGhost,
			  CoordinateTime,
			  int);

//RB_DEFINE_CLASS(S_SpFedGhost);

#endif
