// S_SpHLA.C Source File

#include "S_SpHLA.H"
#include "H_SpObjProxy.H"
#include "SpScheduleEvent.H"
#include "SpObjProxy.H"
#include "RB_Types.H"
#include "SpProxySubscriber.H"
#include "SpString.H"
#include "M_SpRemoveProxy.H"
#include "M_SpNewProxy.H"
#include "O_SpLocalProxyMgr.H"
#include "SpSubscriberHandle.H"
#include "M_SpDecMgrUnsubscribe.H"
#include "M_SpDecMgrSubscribe.H"
#include "M_SpDecMgrUnpublish.H"
#include "M_SpDefaultPublish.H"
#include "M_SpDefaultSubscribe.H"
#include "M_SpPublish.H"
#include "O_SpGrid.H"
#include "SpSpace.H"
#include "SpPublishTheaterSpace.H"
#include "SpPublishRoutingSpace.H"
#include "SpSubscribeTheaterSpace.H"
#include "SpSubscribeRoutingSpace.H"
#include "O_SpDecMgr.H"
#include "SpDistList.H"
#include "E_SpNewProxy.H"
#include "E_SpRemoveProxy.H"
#include "SpList.H"

#include "SpPlugInDM.H"
#include "SpPlugInDDM.H"
#include "RB_SpDefineClass.H"
#include "RB_SpBinaryTree.H"

EVENT_TYPE (DEFAULT_PUBLISH_HLA);
EVENT_TYPE (DEFAULT_SUBSCRIBE_HLA);
EVENT_TYPE (PUBLISH_DECMAN);
EVENT_TYPE (UNPUBLISH_DECMAN);
EVENT_TYPE (SUBSCRIBE_DECMAN);
EVENT_TYPE (UNSUBSCRIBE_DECMAN);
//EVENT_TYPE (REM_PROXY_LOCAL_PROXYMAN);
//EVENT_TYPE (NEW_PROXY_LOCAL_PROXYMAN);
EVENT_TYPE (UpdateProxyDistList);

RB_DEFINE_CLASS(SpPublishTheaterSpace)
RB_DEFINE_CLASS(SpPublishRoutingSpace)
RB_DEFINE_CLASS(SpSubscribeTheaterSpace)
RB_DEFINE_CLASS(SpSubscribeRoutingSpace)
RB_DEFINE_CLASS(SpDistList)

// extern from O_SpGrid.C
extern O_SpGrid *Grid_OBJECT_MANAGER;

/*
** Some Global Variables
*/

SpObjFactory *S_SpHLA::ObjectFactory = NULL;
SpObjFactory *ObjFact = NULL;

void InitializeProxyCode()
{
/* 
** Construct the Object Factory if it hasn't already been constructed 
*/ 
 
  ObjFact = S_SpHLA::GetObjectFactory();
 
} 

// forward declaration
static void NewProxyDestination(SpBinaryTree *node, 
                                int &num_proxy, SpProxyDestination** &proxy);
static void DelProxyDestination(int num_proxy, SpProxyDestination** &proxy);


/************************************************************************
* S_SpHLA constructor							*
************************************************************************/

S_SpHLA::S_SpHLA(char *ClassType)
  : PublishSpaces(new RB_SpBinaryTree), SubscribeSpaces(new RB_SpBinaryTree), 
    ProxyDistList(new RB_SpBinaryTree) {

  DirtyProxyDistList = 0;

//  cout << "HLA Simulation Object constructor starting" << endl;

  SpEndianNature::GetEndianNature();

  InitializeProxyCode();

  if (ObjectFactory == NULL)
    ObjectFactory = ObjFact;

  if (SpPoDataBase::DoNotAllocateMemoryInConstructor == 1){
    return;
  }
/*
** Create the Object Proxy for this entity
*/

  ObjectProxy = ObjectFactory->Create(ClassType);

  ObjectProxy->SetGlobalId(GLOBAL_ID);
  ObjectProxy->SetLocalId(LOCAL_ID);
  ObjectProxy->SetObjectType(OBJECT_TYPE);
  ObjectProxy->SetNode(NODE);

/*
** Add the object proxy event handler
*/

  H_SpObjProxy *ObjectProxyEventHandler = new H_SpObjProxy(ObjectProxy);
  AddEventHandler("ObjectProxy", ObjectProxyEventHandler);

/*
** Initialize the rest of the HLA Simulation Object
*/

  Destination = new SpDestination(N_NODES);

  ProxySubscriber = new SpProxySubscriber;

  DefaultPublish();
  DefaultSubscribe();

//  cout << "HLA Simulation Object constructor done" << endl;

  Active = 1; DEFINE_ATTRIBUTE(Active, "Active");
}

S_SpHLA::~S_SpHLA()
{
  free(NAME);
  //ObjectFactory->PartialDestroy(ObjectProxy);
  delete ProxySubscriber;

  SpPublishSpace *publish_space;
  while ((publish_space = (SpPublishSpace *) PublishSpaces->RemoveFirstElement())){
    if (publish_space->GetSpace()->HasTheaterDimension())
      RB_DELETE_SpPublishTheaterSpace((SpPublishTheaterSpace *) publish_space);
    else
      RB_DELETE_SpPublishRoutingSpace((SpPublishRoutingSpace *) publish_space);
  }

  SpSubscribeSpace *subscribe_space;
  while ((subscribe_space = (SpSubscribeSpace *) SubscribeSpaces->RemoveFirstElement())){
    if (subscribe_space->GetSpace()->HasTheaterDimension())
      RB_DELETE_SpSubscribeTheaterSpace((SpSubscribeTheaterSpace *) subscribe_space);
    else
      RB_DELETE_SpSubscribeRoutingSpace((SpSubscribeRoutingSpace *) subscribe_space);
  }

  SpDistList *dist_list;
  while ((dist_list = (SpDistList *) ProxyDistList->RemoveFirstElement()))
    RB_DELETE_SpDistList(dist_list);   
}

SpObjFactory *
S_SpHLA::GetObjectFactory() {
/*
** Construct the Object Factory if it hasn't already been constructed
*/
  if (ObjFact == NULL)
    ObjFact = new SpObjFactory("Objects.par");

  return (ObjFact);
}

void
S_SpHLA::DefaultPublish(){

  if (!SpDMEnabled())
    return;

  M_SpDefaultPublish msg;

  SpSimTime &time = CurrentEvent->get_MESSAGE()->time_tag;
  ScheduleEvent(time, DEFAULT_PUBLISH_HLA, get_OBJECT_TYPE(), get_LOCAL_ID(), get_NODE(), msg);
}

void
S_SpHLA::DefaultSubscribe(){

  if (!SpDMEnabled())
    return;

  M_SpDefaultSubscribe msg;

  SpSimTime &time = CurrentEvent->get_MESSAGE()->time_tag;
  ScheduleEvent(time, DEFAULT_SUBSCRIBE_HLA, get_OBJECT_TYPE(), get_LOCAL_ID(), get_NODE(), msg);
}

// publish myself
//   or
// invoke from event : DEFAULT_PUBLISH_HLA
int
S_SpHLA::Publish(){
  // send 1 msg to :
  //   1) my S_SpDecMgr
  //   2) each of my S_SpDecMgr's ancestor
  int node, local_id;

  if (!SpDMEnabled()){
    cerr << "ERROR: Declaration Management not plugged in - PlugInDM()" << endl;
    abort();
  }

  SpObjTemplate *obj_template = ObjectProxy->GetObjectTemplate();
  M_SpPublish msg(get_NODE(), get_OBJECT_TYPE(), get_LOCAL_ID(), get_GLOBAL_ID());
  SpSimTime time(CurrentEvent->get_MESSAGE()->time_tag);
  time.IncrementPriority1(4);

  while (obj_template){
	// TO DO : don't use name look-up.
    DecMan_OBJECT_MANAGER->get_nodid(obj_template->GetName(), node, local_id);
    ScheduleEvent(time, PUBLISH_DECMAN, DECMAN, local_id, node, msg);
    obj_template = obj_template->GetParent();
  }

  return (1);
}

// unpublish myself
void
S_SpHLA::UnPublish(){
  // send 1 msg to :
  //   1) my S_SpDecMgr
  //   2) each of my S_SpDecMgr's ancestor
  //int num_subscriber;
  int node, local_id;

  if (!SpDMEnabled()){
    cerr << "ERROR: Declaration Management not plugged in - PlugInDM()" << endl;
    abort();
  }

  SpObjTemplate *obj_template = ObjectProxy->GetObjectTemplate();
  SpSimTime time(CurrentEvent->get_MESSAGE()->time_tag);
  time.IncrementPriority1(2);

  M_SpDecMgrUnpublish msg(get_NODE(), get_OBJECT_TYPE(), get_LOCAL_ID(), get_GLOBAL_ID());
  while (obj_template){
	// TO DO : don't use name lookup 
    DecMan_OBJECT_MANAGER->get_nodid(obj_template->GetName(), node, local_id);
    ScheduleEvent(time, UNPUBLISH_DECMAN, DECMAN, local_id, node, msg);
    obj_template = obj_template->GetParent();
  }
}

int
S_SpHLA::Subscribe(char *type){
  // send 1 msg to the S_SpDecMgr that I'm subscribing.
  int node, local_id;

  if (!SpDMEnabled()){
    cerr << "ERROR: Declaration Management not plugged in - PlugInDM()" << endl;
    abort();
  }

    // TO DO : don't use name lookup 
  DecMan_OBJECT_MANAGER->get_nodid(type, node, local_id);

  if ( (node<0) || (local_id<0) ){
    RB_cout << "WARNING:: Unable to subscribe to " << type << " - UNKNOWN TYPE." << endl;
    return (0);
  }

  SpSimTime time(CurrentEvent->get_MESSAGE()->time_tag);
  time.IncrementPriority1(3);

  M_SpDecMgrSubscribe msg(get_NODE(), get_OBJECT_TYPE(), get_LOCAL_ID(), get_GLOBAL_ID());
  ScheduleEvent(time, SUBSCRIBE_DECMAN, DECMAN, local_id, node, msg);

  return (1);
}

int 
S_SpHLA::UnSubscribe(char *type){
  int node, local_id;

  if (!SpDMEnabled()){
    cerr << "ERROR: Declaration Management not plugged in - PlugInDM()" << endl;
    abort();
  }

    // TO DO : don't use name lookup 
  DecMan_OBJECT_MANAGER->get_nodid(type, node, local_id);

  if ( (node<0) || (local_id<0) ){
    RB_cout << "WARNING:: Unable to unsubscribe to " << type << " - UNKNOWN TYPE." << endl;
    return (0);
  }

  M_SpDecMgrUnsubscribe msg(get_NODE(), get_GLOBAL_ID());
  SpSimTime time(CurrentEvent->get_MESSAGE()->time_tag);
  time.IncrementPriority1(1);

  ScheduleEvent(time, UNSUBSCRIBE_DECMAN, DECMAN, local_id, node, msg);

  return (1);
}

SpPublishSpace *
S_SpHLA::PublishSpace(const SpSimTime &time, const char *name){

  if (!SpDDMEnabled()){
    cerr << "ERROR: Data Distribution Management not plugged in - PlugInDDM()" << endl;
    abort();
  }

  SpPublishSpace *publish_space = (SpPublishSpace *) 0;

  if (Grid_OBJECT_MANAGER==(O_SpGrid *) 0){
    cerr << "ERROR: Data Distribution Management object manager has not been initialized." << endl;
    cerr << "PlugInDDM() before PLUG_IN_OBJECT_MANAGER(user-defined object manager)." << endl;
    abort();
  }

  const SpSpace *space = Grid_OBJECT_MANAGER->FindSpace(name);

  if (space==(const SpSpace *) 0)
    return (publish_space);

  publish_space = (SpPublishSpace *) PublishSpaces->Find((char *) name);
  if (publish_space) return (publish_space);

  publish_space = (space->HasTheaterDimension()) 
    ? RB_NEW_SpPublishTheaterSpace(new SpPublishTheaterSpace(time, this, space))
    : RB_NEW_SpPublishRoutingSpace(new SpPublishRoutingSpace(time, this, space));

  PublishSpaces->Insert(publish_space, (char *) publish_space->GetName());
  AddComponent(publish_space);

  return (publish_space);
}

SpSubscribeSpace *
S_SpHLA::SubscribeSpace(const SpSimTime &time, const char *name){

  if (!SpDDMEnabled()){
    cerr << "ERROR: Data Distribution Management not plugged in - PlugInDDM()" << endl;
    abort();
  }

  SpSubscribeSpace *subscribe_space = (SpSubscribeSpace *) 0;

  if (Grid_OBJECT_MANAGER==(O_SpGrid *) 0){
    cerr << "ERROR: Data Distribution Management object manager has not been initialized." << endl;
    cerr << "PlugInDDM() before PLUG_IN_OBJECT_MANAGER(user-defined object manager)." << endl;
    abort();
  }

  const SpSpace *space = Grid_OBJECT_MANAGER->FindSpace(name);

  if (space==(const SpSpace *) 0)
    return (subscribe_space);

  subscribe_space = (SpSubscribeSpace *) SubscribeSpaces->Find((char *) name);
  if (subscribe_space) return (subscribe_space);

  subscribe_space = (space->HasTheaterDimension()) 
    ? RB_NEW_SpSubscribeTheaterSpace(new SpSubscribeTheaterSpace(time, this, space))
    : RB_NEW_SpSubscribeRoutingSpace(new SpSubscribeRoutingSpace(time, this, space));

  SubscribeSpaces->Insert(subscribe_space, (char *) subscribe_space->GetName());
  AddComponent(subscribe_space);

  return (subscribe_space);
}

////////////////////////////////////////////////////////////////////////////////////////

// invoke from event : UNPUBLISH_HLA
int
S_SpHLA::UnPublish(int remove_proxy){
  const SpSimTime &time = CurrentEvent->get_MESSAGE()->time_tag;
  const int grid_id     = -1;
  SpBinaryTree remove_subscriber_holder;
  SpBinaryTree subscriber_nodes;
  SpBinaryTree *current_node, *node, *grid;
  SpProxyObjHandle *subscriber;

  if (!remove_proxy) return (0);

  ProxySubscriber->Remove(SpProxySubscriber::DM, subscriber_nodes);

  current_node = (SpBinaryTree *) subscriber_nodes.GetFirstElement();
  while (current_node){
    node = new SpBinaryTree();
    remove_subscriber_holder.Insert((void *) node, int(subscriber_nodes.GetCurrentKey()));
    grid = new SpBinaryTree();
    node->Insert((void *) grid, grid_id);
    subscriber = (SpProxyObjHandle *) current_node->GetFirstElement();
    while (subscriber){
      grid->Insert((void *) subscriber, subscriber->GetGlobalId());
      subscriber = (SpProxyObjHandle *) current_node->GetNextElement();
    }
    current_node = (SpBinaryTree *) subscriber_nodes.GetNextElement();
  }

  RemoveProxy(time, &remove_subscriber_holder);

  while ((node=(SpBinaryTree *) remove_subscriber_holder.RemoveFirstElement())){
    while ((grid=(SpBinaryTree*) node->RemoveFirstElement()))
      delete grid;
    delete node;
  }

  while ((node=(SpBinaryTree *) subscriber_nodes.RemoveFirstElement())){
    while ((subscriber=(SpProxyObjHandle *) node->RemoveFirstElement()))
      delete subscriber;
    delete node;
  }

  return (1);
}

// invoke from event : DEFAULT_SUBSCRIBE_HLA
void
S_SpHLA::Subscribe(){
  // send 1 msg to the S_SpDecMgr that I'm subscribing.
  SpObjTemplate *obj_template = ObjectProxy->GetObjectTemplate();

  while (obj_template){
    for (int i=0; i<obj_template->GetNsubscribe(); i++)
      Subscribe(obj_template->GetSubscribeName(i));
    obj_template = obj_template->GetParent();
  }
}

// invoke from event : UNSUBSCRIBE_HLA
void
S_SpHLA::RemoveSubscriber(int node, int global_id, int local_id, int obj_type){
  SpBinaryTree remove_subscriber_holder;
  SpBinaryTree subscribers;
  const SpSimTime &time = CurrentEvent->get_MESSAGE()->time_tag;

  SpProxyObjHandle subscriber(local_id, node, obj_type, global_id);
  subscribers.Insert(&subscriber, subscriber.GetGlobalId());
  RemoveSubscriber(SpProxySubscriber::DM, subscribers, &remove_subscriber_holder);
  RemoveProxy(time, &remove_subscriber_holder);

  ClearSubscriberHolder(remove_subscriber_holder);
}

#include "SpWalltime.H"

// invoke from event : ADD_SUBSCRIBER_HLA
void
S_SpHLA::AddSubscriber(SpList &subscribers) {
  SpBinaryTree new_subscriber_holder, add_subscriber_holder;
  SpSimTime time = CurrentEvent->get_MESSAGE()->time_tag;

  new_subscriber_holder.SetBalancedTreeMode();
  add_subscriber_holder.SetBalancedTreeMode();

  AddSubscriber(SpProxySubscriber::DM, subscribers, &new_subscriber_holder, &add_subscriber_holder);
  AddProxy(time, &new_subscriber_holder, &add_subscriber_holder);

  ClearSubscriberHolder(new_subscriber_holder);
  ClearSubscriberHolder(add_subscriber_holder);
}

////////////////////////////////////////////////////////////////////////////////////////

void
S_SpHLA::RemoveSubscriber(int /* proxy_manager */,
                          SpBinaryTree * /* remove_subscriber_holder */){
}

void
S_SpHLA::RemoveSubscriber(int proxy_manager,
                          SpBinaryTree &subscribers,
                          SpBinaryTree *remove_subscriber_holder){
  const int grid_id = -1;

  RemoveSubscriber(proxy_manager, grid_id, subscribers, remove_subscriber_holder);
}

void
S_SpHLA::RemoveSubscriber(int proxy_manager,
                          int grid_id, SpBinaryTree &subscribers,
                          SpBinaryTree *remove_subscriber_holder,
                          SpBinaryTree *undeliver_subscriber_holder){
  SpProxyObjHandle *subscriber;
  int remove_proxy;
  SpList       *current_grid;
  SpBinaryTree *current_node;

  subscriber = (SpProxyObjHandle *) subscribers.GetFirstElement();
  while (subscriber){
    remove_proxy = ProxySubscriber->Remove(proxy_manager, 
                                           subscriber->GetNode(), 
                                           subscriber->GetGlobalId());
    if (remove_proxy){
      current_node = (SpBinaryTree *) remove_subscriber_holder->Find(subscriber->GetNode());
      if (!current_node){
        current_node = new SpBinaryTree();
        remove_subscriber_holder->Insert((void *) current_node, subscriber->GetNode());
      }
      current_grid = (SpList *) current_node->Find(grid_id);
      if (!current_grid){
        current_grid = new SpList();
        current_node->Insert((void *) current_grid, grid_id);
      }
      current_grid->Insert((void *) subscriber);
    }
    else{
      if (undeliver_subscriber_holder){
        current_node = (SpBinaryTree *) undeliver_subscriber_holder->Find(subscriber->GetNode());
        if (!current_node){
          current_node = new SpBinaryTree();
          undeliver_subscriber_holder->Insert((void *) current_node, subscriber->GetNode());
        }
        current_grid = (SpList *) current_node->Find(grid_id);
        if (!current_grid){
          current_grid = new SpList();
          current_node->Insert((void *) current_grid, grid_id);
        }
        current_grid->Insert((void *) subscriber);
      }
    }
    subscriber = (SpProxyObjHandle *) subscribers.GetNextElement();
  }// while
}

void
S_SpHLA::RemoveProxy(const SpSimTime &time, SpBinaryTree *remove_subscriber_holder){
  const float look_ahead = 0.0;
  const int   space_id   = -1;

  RemoveProxy(time, look_ahead, space_id, remove_subscriber_holder);
}

void
S_SpHLA::RemoveProxy(const SpSimTime &time, float look_ahead, int space_id,
                     SpBinaryTree *remove_subscriber_holder,
                     SpBinaryTree *undeliver_subscriber_holder){
  int node_id;
  SpBinaryTree *node;
  int num_remove_proxy_dest, num_undeliver_proxy_dest;
  SpProxyDestination **remove_proxy_dest, **undeliver_proxy_dest;

  for (node_id=0; node_id<N_NODES; node_id++){

    node = (SpBinaryTree *) remove_subscriber_holder->Find(node_id);
    NewProxyDestination(node, num_remove_proxy_dest, remove_proxy_dest);

    num_undeliver_proxy_dest = 0;
    undeliver_proxy_dest     = (SpProxyDestination **) 0;
    if (undeliver_subscriber_holder){
      node = (SpBinaryTree *) undeliver_subscriber_holder->Find(node_id);
      NewProxyDestination(node, num_undeliver_proxy_dest, undeliver_proxy_dest);
    }

    if (!remove_proxy_dest && !undeliver_proxy_dest) continue;

    RemoveProxy(time, look_ahead, space_id, node_id, 
                num_remove_proxy_dest, remove_proxy_dest,
                num_undeliver_proxy_dest, undeliver_proxy_dest);

    DelProxyDestination(num_remove_proxy_dest,    remove_proxy_dest);
    DelProxyDestination(num_undeliver_proxy_dest, undeliver_proxy_dest);

  }

  UpdateProxyDestination();
}

void
S_SpHLA::RemoveProxy(const SpSimTime &time, float look_ahead, int space_id, int node,
                     int num_remove_dest,    SpProxyDestination** &remove_dest,
                     int num_undeliver_dest, SpProxyDestination** &undeliver_dest){
  O_SpLocalProxyMgr *local_proxyman = (O_SpLocalProxyMgr *) get_manager(LOCAL_PROXYMAN);
  char *buffer, *cp;
  int buffer_len;
  int i, len;

  if ((num_remove_dest+num_undeliver_dest)<=0) return;

  buffer_len = 0;

  len = sizeof(SpProxyDestination);
  buffer_len += (len*num_remove_dest);
  buffer_len += (len*num_undeliver_dest);

  buffer = new char [buffer_len];
  cp = buffer;

  if (num_remove_dest>0){
    len = sizeof(SpProxyDestination);
    for (i=0; i<num_remove_dest; i++){
      memcpy(cp, remove_dest[i], len); cp += len;
    }
  }

  if (num_undeliver_dest>0){
    len = sizeof(SpProxyDestination);
    for (i=0; i<num_undeliver_dest; i++){
      memcpy(cp, undeliver_dest[i], len); cp += len;
    }
  }

  M_SpRemoveProxy remove_proxy_msg(get_GLOBAL_ID(), space_id, num_remove_dest, num_undeliver_dest);
  remove_proxy_msg.SetLookAhead(look_ahead);
  ScheduleEvent(time, REM_PROXY_LOCAL_PROXYMAN, LOCAL_PROXYMAN,
                get_LOCAL_ID() % local_proxyman->get_N_LOC(), node,
                remove_proxy_msg, buffer, buffer_len);

  delete [] buffer;
}

void
S_SpHLA::AddSubscriber(int proxy_manager,
                     SpList &subscribers,  	                       // input  : SpProxyObjHandle *
                     SpBinaryTree *new_subscriber_holder,               // output
                     SpBinaryTree *add_subscriber_holder){		// output
  const int grid_id = -1;

  AddSubscriber(proxy_manager, grid_id, subscribers,
                new_subscriber_holder, add_subscriber_holder);
}

void
S_SpHLA::AddSubscriber(int proxy_manager,
                     int grid_id,
                     SpList &subscribers,				// input  : SpProxyObjHandle *
                     SpBinaryTree *new_subscriber_holder,               // output
                     SpBinaryTree *add_subscriber_holder,               // output
                     SpBinaryTree *dup_subscriber_holder,               // output
                     SpBinaryTree *subscriber_handles){			// output
  int deliver_proxy, duplicate_proxy;
  SpProxyObjHandle *subscriber;
  SpSubscriberHandle *subscriber_handle;
  SpList       *current_grid;
  SpBinaryTree *current_node;

  subscriber = (SpProxyObjHandle *) subscribers.GetFirstElement();
  while (subscriber){
    ProxySubscriber->Add(proxy_manager,
                         subscriber->GetNode(),    subscriber->GetObjType(),
                         subscriber->GetLocalId(), subscriber->GetGlobalId(),
                         deliver_proxy, duplicate_proxy, subscriber_handle);

    if (subscriber_handles)
      subscriber_handles->Insert(subscriber_handle, subscriber_handle->GetGlobalId());

    if (duplicate_proxy){
      if (dup_subscriber_holder){
        current_node = (SpBinaryTree *) dup_subscriber_holder->Find(subscriber->GetNode());
        if (!current_node){
          current_node = new SpBinaryTree();
          current_node->SetBalancedTreeMode();
          dup_subscriber_holder->Insert((void *) current_node, subscriber->GetNode());
        }
        current_grid = (SpList *) current_node->Find(grid_id);
        if (!current_grid){
          current_grid = new SpList();
          current_node->Insert((void *) current_grid, grid_id);
        }
        current_grid->Insert((void *) subscriber);
      }
    }
    else{
      if (deliver_proxy){
        current_node = (SpBinaryTree *) new_subscriber_holder->Find(subscriber->GetNode());
        if (!current_node){
          current_node = new SpBinaryTree();
          current_node->SetBalancedTreeMode();
          new_subscriber_holder->Insert((void *) current_node, subscriber->GetNode());
        }
        current_grid = (SpList *) current_node->Find(grid_id);
        if (!current_grid){
          current_grid = new SpList();
          current_node->Insert((void *) current_grid, grid_id);
        }
        current_grid->Insert((void *) subscriber);
      }
      else{
        current_node = (SpBinaryTree *) add_subscriber_holder->Find(subscriber->GetNode());
        if (!current_node){
          current_node = new SpBinaryTree();
          current_node->SetBalancedTreeMode();
          add_subscriber_holder->Insert((void *) current_node, subscriber->GetNode());
        }
        current_grid = (SpList *) current_node->Find(grid_id);
        if (!current_grid){
          current_grid = new SpList();
          current_node->Insert((void *) current_grid, grid_id);
        }
        current_grid->Insert((void *) subscriber);
      }
    }
    subscriber = (SpProxyObjHandle *) subscribers.GetNextElement();
  }// while
}

void
S_SpHLA::AddProxy(const SpSimTime &time,
                  SpBinaryTree *new_subscriber_holder,
                  SpBinaryTree *add_subscriber_holder){
  const float look_ahead = 0.0;
  const int space_id     = -1;

  AddProxy(time, look_ahead, space_id, new_subscriber_holder, add_subscriber_holder);
}

void
S_SpHLA::AddProxy(const SpSimTime &time, float look_ahead, int space_id,
                  SpBinaryTree *new_subscriber_holder,
                  SpBinaryTree *add_subscriber_holder,
                  SpBinaryTree *dup_subscriber_holder){
  int node_id;
  SpBinaryTree *node;
  int num_new_proxy_dest, num_add_proxy_dest, num_dup_proxy_dest;
  SpProxyDestination **new_proxy_dest, **add_proxy_dest, **dup_proxy_dest;
 
  for (node_id=0; node_id<N_NODES; node_id++){

    node = (SpBinaryTree *) new_subscriber_holder->Find(node_id);
    NewProxyDestination(node, num_new_proxy_dest, new_proxy_dest);

    node = (SpBinaryTree *) add_subscriber_holder->Find(node_id);
    NewProxyDestination(node, num_add_proxy_dest, add_proxy_dest);

    if (dup_subscriber_holder){
      node = (SpBinaryTree *) dup_subscriber_holder->Find(node_id);
      NewProxyDestination(node, num_dup_proxy_dest, dup_proxy_dest);
    }
    else{
      num_dup_proxy_dest = 0;
      dup_proxy_dest     = (SpProxyDestination **) 0;
    }

    if (!new_proxy_dest && !add_proxy_dest && !dup_proxy_dest) continue;

    AddProxy(time, look_ahead, space_id, node_id, 
             num_new_proxy_dest, new_proxy_dest,
             num_add_proxy_dest, add_proxy_dest,
             num_dup_proxy_dest, dup_proxy_dest);

    DelProxyDestination(num_new_proxy_dest, new_proxy_dest);
    DelProxyDestination(num_add_proxy_dest, add_proxy_dest);
    DelProxyDestination(num_dup_proxy_dest, dup_proxy_dest);

  }// for node

  UpdateProxyDestination();
}

void
S_SpHLA::AddProxy(const SpSimTime &time, float look_ahead, int space_id, int node, 
                  int num_new_dest, SpProxyDestination** &new_dest,
                  int num_add_dest, SpProxyDestination** &add_dest,
                  int num_dup_dest, SpProxyDestination** &dup_dest){
  int i, len;
  char *buffer, *cp;
  int buffer_len;
  SpObjProxyMsg *data;
  O_SpLocalProxyMgr *local_proxyman = (O_SpLocalProxyMgr *) get_manager(LOCAL_PROXYMAN);

  if ((num_new_dest+num_add_dest+num_dup_dest)<=0) return;

  data       = (SpObjProxyMsg *) 0;
  buffer_len = 0;
  if (num_new_dest>0){
    data = GetObjectFactory()->GenerateMessage(GetObjectProxy());
    buffer_len += data->GetNbytes();
  }
  len = sizeof(SpProxyDestination);
  buffer_len += (len*num_new_dest);
  buffer_len += (len*num_add_dest);
  buffer_len += (len*num_dup_dest);

  buffer = new char [buffer_len];
  cp = buffer;

  if (num_add_dest>0){
    len = sizeof(SpProxyDestination);
    for (i=0; i<num_add_dest; i++){
      memcpy(cp, add_dest[i], len); cp += len;
    }
  }

  if (num_dup_dest>0){
    len = sizeof(SpProxyDestination);
    for (i=0; i<num_dup_dest; i++){
      memcpy(cp, dup_dest[i], len); cp += len;
    }
  }

  if (num_new_dest>0){
    len = sizeof(SpProxyDestination);
    for (i=0; i<num_new_dest; i++){
      memcpy(cp, new_dest[i], len); cp += len;
    }
    len = data->GetNbytes(); memcpy(cp, data, len); cp += len;
  }

  M_SpNewProxy new_proxy_msg(get_GLOBAL_ID(), space_id,  
                             num_new_dest, num_add_dest, num_dup_dest);
  new_proxy_msg.SetLookAhead(look_ahead);

  ScheduleEvent(time, NEW_PROXY_LOCAL_PROXYMAN, LOCAL_PROXYMAN, 
                get_LOCAL_ID() % local_proxyman->get_N_LOC(), node, 
                new_proxy_msg, buffer, buffer_len);

  if (buffer) delete [] buffer;
  if (data)   delete [] ((char*)data);
}

void 
S_SpHLA::ClearSubscriberHolder(SpBinaryTree &holder){
  SpBinaryTree *node;
  SpList *grid;

  while ((node=(SpBinaryTree *) holder.RemoveFirstElement())){
    while ((grid=(SpList*) node->RemoveFirstElement()))
      delete grid;
    delete node;
  }
}

void NewProxyDestination(SpBinaryTree *nodes, 
                         int &num_proxy, SpProxyDestination** &proxy){
  int s;
  int grid_id;
  SpList *grid;
  SpProxyObjHandle *subscriber;

  num_proxy = 0;
  proxy     = (SpProxyDestination **) 0;

  if (!nodes) return;
  
  // count number of subscriber from all grids
  grid = (SpList *) nodes->GetFirstElement();
  while (grid){
    num_proxy += grid->GetNumElements();
    grid = (SpList *) nodes->GetNextElement();
  }

  // collect all subscribers from all grids
  proxy = new SpProxyDestination * [num_proxy];
  s = 0;
  grid = (SpList *) nodes->GetFirstElement();
  while (grid){
    grid_id    = int(nodes->GetCurrentKey());
    subscriber = (SpProxyObjHandle *) grid->GetFirstElement();
    while (subscriber){
      proxy[s++] = new SpProxyDestination(grid_id, *subscriber);
      subscriber = (SpProxyObjHandle *) grid->GetNextElement();
    }
    grid = (SpList *) nodes->GetNextElement();
  }
}

void DelProxyDestination(int num_proxy, SpProxyDestination** &proxy){
  if (!proxy) return;

  for (int i=0; i<num_proxy; i++)
    delete proxy[i];
  delete [] proxy;
}

void
S_SpHLA::UpdateProxyDestination(){
  SpDestination new_destination(get_N_NODES());

  ProxySubscriber->SetDestination(new_destination);
  RB_memcpy(Destination->GetDestination(), 
            new_destination.GetDestination(), 
            new_destination.GetNbytes());
}

SpPublishSpace *
S_SpHLA::GetPublishSpace(int spaceId){
  SpPublishSpace *publish_space = (SpPublishSpace *) PublishSpaces->GetFirstElement();
 
  while (publish_space){
    if (publish_space->GetId()==spaceId) return (publish_space);
    publish_space = (SpPublishSpace *) PublishSpaces->GetNextElement();
  }

  cout << "HLA : " << get_NAME() << " unable to get published space " << spaceId << endl;
  abort();

  return ((SpPublishSpace *) 0);
}

SpPublishSpace *
S_SpHLA::FindPublishSpace(const char *name){
  SpPublishSpace *publish_space = (SpPublishSpace *) PublishSpaces->Find((char *) name);
  return (publish_space);
}

SpPublishSpace *
S_SpHLA::GetPublishSpace(const char *name){
  SpPublishSpace *publish_space = FindPublishSpace(name);

  if (publish_space==(SpPublishSpace *) 0){
    cout << "HLA : " << get_NAME() << " unable to get published space " << name << endl;
    abort();
  }

  return (publish_space);
}

SpSubscribeSpace *
S_SpHLA::GetSubscribeSpace(int spaceId){
  SpSubscribeSpace *subscribe_space = (SpSubscribeSpace *) SubscribeSpaces->GetFirstElement();
 
  while (subscribe_space){
    if (subscribe_space->GetId()==spaceId) return (subscribe_space);
    subscribe_space = (SpSubscribeSpace *) SubscribeSpaces->GetNextElement();
  }

  cout << "HLA : " << get_NAME() << " unable to get subscribed space " << spaceId << endl;
  abort();

  return ((SpSubscribeSpace *) 0);
}

SpSubscribeSpace *
S_SpHLA::FindSubscribeSpace(const char *name){
  SpSubscribeSpace *subscribe_space = (SpSubscribeSpace *) SubscribeSpaces->Find((char *) name);
  return (subscribe_space);
}

SpSubscribeSpace *
S_SpHLA::GetSubscribeSpace(const char *name){
  SpSubscribeSpace *subscribe_space = FindSubscribeSpace(name);

  if (subscribe_space==(SpSubscribeSpace *) 0){
    cout << "HLA : " << get_NAME() << " unable to get subscribed space " << name << endl;
    abort();
  }

  return (subscribe_space);
}

void
S_SpHLA::GetPositionRadKm(double /* time */, double &latRad, double &lonRad, double &altKm){
  latRad = lonRad = altKm = 0.0;
}

double 
S_SpHLA::GetMaxSpeedKnots(void) { 
  return (0.0); 
}

double 
S_SpHLA::GetMinExpansionNmi(void) { 
  return (1.0); 
}

double
S_SpHLA::GetMaxSensorRangeNmi(void) {
  return (0.0);
}

double 
S_SpHLA::GetMinRescheduleTimeHour(void) { 
  return  (1.0); 
}

double 
S_SpHLA::GetLookAheadHour(void) { 
  const double e = 1E-5;
  const double f = 4.0;
  double reschedule_interval = (GetMaxSpeedKnots()<e) 
                                ? GetMinRescheduleTimeHour() 
                                : GetMinExpansionNmi() / GetMaxSpeedKnots();
  return ((reschedule_interval>GetMinRescheduleTimeHour()) 
          ? reschedule_interval/f 
          : GetMinRescheduleTimeHour()/f);
}

SpDistList *
S_SpHLA::CreateDistList(const char *name){
  SpDistList *new_list = (SpDistList *) ProxyDistList->Find(name);
  if (new_list==(SpDistList *) 0){
    new_list = new SpDistList(name, SpProxySubscriber::GetNewDistListId(), this);
    ProxyDistList->Insert(RB_NEW_SpDistList(new_list), (char *) new_list->GetName());
  }
  ScheduleProxyDistUpdate();
  return (new_list);
}

SpDistList *
S_SpHLA::GetDistList(const char *name){
  return ((SpDistList *) ProxyDistList->Find(name));
}

void        
S_SpHLA::DestroyDistList(const char *name){
  SpDistList *del_list = GetDistList(name);
  if (del_list){
    ProxyDistList->Remove((char *) del_list->GetName());
    RB_DELETE_SpDistList(del_list);
  }
}

void
S_SpHLA::ScheduleProxyDistUpdate(){
  if (DirtyProxyDistList) return;

  DirtyProxyDistList = 1;
  ScheduleEvent(CurrentEvent->get_MESSAGE()->time_tag, 
                UpdateProxyDistList, get_OBJECT_TYPE(), get_LOCAL_ID(), get_NODE());
}

void 
S_SpHLA::UnDirtyProxyDistList(void) { 
  DirtyProxyDistList = 0; 
}

void 
S_SpHLA::debug(ostream& os){
  os << CurrentEvent->get_MESSAGE()->time_tag << " " 
     << get_NAME() << " (" << get_GLOBAL_ID() << ")" << endl; 
  ProxySubscriber->debug(os);
}
