// S_SpHLA.H Header File

#ifndef S_SpHLA_OBJECT
#define S_SpHLA_OBJECT

#include "SpObjProxy.H"
#include "SpObjFactory.H"
#include "S_SpProc.H"
#include "SpDestination.H"
#include "SpProxySubscriber.H"
#include "RB_UpdateXqueue.H" //proxy:

void InitializeProxyCode();

  // forward declaration
class SpProxyObjHandle;
class SpSimTime;
class SpPublishSpace;
class SpSubscribeSpace;
class SpProxyDestination;
class RB_SpBinaryTree;

class S_SpHLA : public S_SpProc {

  friend class E_SUBSCRIBE_HLA;
  friend class E_SpAddSubscriberSpace;
  friend class E_SpRemSubscriberSpace;
  friend class E_SpUnPublishSpace;
  friend class E_SpDeliverProxyPtrSpace;
  friend class E_SpHlaUpdateProxyDistList;
  friend class SpPublishGrid;
  friend class SpPublishTheaterSpace;
  friend class SpSubscribeTheaterSpace;
  friend class SpDistList;

  private:

    C_XQUEUE DistributionList;		// List of objects that need my proxy
    SpDestination *Destination;		// Collection of nodes needing my proxy
					// Multiple objects on the same node that
					// require my proxy are represented once.
    // proxy::  the updateList is now inside the obj proxy:
//    C_XQUEUE UpdateList;		// List of my attributes modified in an event

    static SpObjFactory *ObjectFactory; // ObjectProxy creation services

    SpProxySubscriber *ProxySubscriber;	// Maintain list of subscriber to my proxy
    RB_SpBinaryTree *PublishSpaces;	// List of DDM spaces published to
    RB_SpBinaryTree *SubscribeSpaces;	// List of DDM spaces subscribed to
    RB_SpBinaryTree *ProxyDistList;
    RB_int          DirtyProxyDistList; // TRUE if user modified "ProxyDistList"

    LOGICAL_ATTRIBUTE Active;

      // internal methods
    void DefaultPublish(void);
    void DefaultSubscribe(void);

    void AddSubscriber(int proxy_manager,
		SpBinaryTree &subscribers,                         
		SpBinaryTree *new_subscriber_holder,                    
		SpBinaryTree *add_subscriber_holder);
    void AddSubscriber(int proxy_manager,
		int grid_id,
		SpBinaryTree &subscribers,                         
		SpBinaryTree *new_subscriber_holder,                    
		SpBinaryTree *add_subscriber_holder,                    
		SpBinaryTree *dup_subscriber_holder=(SpBinaryTree *) 0,
		SpBinaryTree *subscriber_handles=(SpBinaryTree *) 0);
    void AddProxy(const SpSimTime &time,
		SpBinaryTree *new_subscriber_holder, 
		SpBinaryTree *add_subscriber_holder);
    void AddProxy(const SpSimTime &time, float look_ahead, int space_id, 
		SpBinaryTree *new_subscriber_holder, 
		SpBinaryTree *add_subscriber_holder,
		SpBinaryTree *dup_subscriber_holder=(SpBinaryTree *) 0);
    void AddProxy(const SpSimTime &time, float look_ahead, int space_id, int node,
		int num_new_dest, SpProxyDestination** &new_dest,
		int num_add_dest, SpProxyDestination** &add_dest,
		int num_dup_dest, SpProxyDestination** &dup_dest);

    void RemoveSubscriber(int proxy_manager,
		SpBinaryTree *remove_subscriber_holder);
    void RemoveSubscriber(int proxy_manager,
		SpBinaryTree &subscribers,
		SpBinaryTree *remove_subscriber_holder);
    void RemoveSubscriber(int proxy_manager,
		int grid_id, SpBinaryTree &subscribers,
		SpBinaryTree *remove_subscriber_holder,
		SpBinaryTree *undeliver_subscriber_holder=(SpBinaryTree *) 0);
    void RemoveProxy(const SpSimTime &time, SpBinaryTree *remove_subscriber_holder);
    void RemoveProxy(const SpSimTime &time, float look_ahead, int space_id,
		SpBinaryTree *remove_subscriber_holder,
		SpBinaryTree *undeliver_subscriber_holder=(SpBinaryTree *) 0);
    void RemoveProxy(const SpSimTime &time, float look_ahead, int space_id, int node,
		int num_remove_dest,    SpProxyDestination** &remove_dest,
		int num_undeliver_dest, SpProxyDestination** &undeliver_dest);

    void UnDirtyProxyDistList(void);

    static void ClearSubscriberHolder(SpBinaryTree &holder);

    void UpdateProxyDestination(void);

      // DM event methods

        // DEFAULT_SUBSCRIBE_HLA
    void Subscribe(void);
    friend class E_SpDefaultSubscribe;

        // ADD_SUBSCRIBER_HLA
    void AddSubscriber(SpBinaryTree &subscribers);
    friend class E_SpAddSubscriber;

        // UNSUBSCRIBE_HLA
    void RemoveSubscriber(int _node, int _global_id, int _local_id, int _obj_type);
    friend class E_SpHlaUnsubscribe;

        // UNPUBLISH_HLA
    int UnPublish(int remove_proxy);
    friend class E_SpHlaUnpublish;

      // DDM methods
    SpPublishSpace   *GetPublishSpace(int id);
    SpSubscribeSpace *GetSubscribeSpace(int id);

      // Data distribution methods
    void ScheduleProxyDistUpdate(void);

  protected:

    SpObjProxy *ObjectProxy;      // The proxy repesenting my entity
    C_XQUEUE RemoteObjectProxies; // List of proxies of other objects I need


  public:

    S_SpHLA(char *ClassType);
    ~S_SpHLA();
    
    SpObjProxy *GetObjectProxy() {return ObjectProxy;}
    static SpObjFactory *GetObjectFactory(void);

//////// Declaration management (DM) operations.
    int Publish(void);
    int Subscribe(char *type);
    void UnPublish(void);
    int UnSubscribe(char *type);
//////// Declaration management (DM) operations.

//////// Data distribution management (DDM) operations.
      // Retrieve publish space by name, abort if not found
    SpPublishSpace   *GetPublishSpace(const char *name);
      // Retrieve publish space by name, return NULL if not found
    SpPublishSpace   *FindPublishSpace(const char *name);

      // Retrieve subscribe space by name, abort if not found
    SpSubscribeSpace *GetSubscribeSpace(const char *name);
      // Retrieve subscribe space by name, return NULL if not found
    SpSubscribeSpace *FindSubscribeSpace(const char *name);

    SpPublishSpace   *PublishSpace(const SpSimTime &time, const char *name);
    SpSubscribeSpace *SubscribeSpace(const SpSimTime &time, const char *name);

    virtual double GetLookAheadHour(void);
    virtual double GetMaxSensorRangeNmi(void);
    virtual double GetMaxSpeedKnots(void);
    virtual double GetMinExpansionNmi(void); 	
    virtual double GetMinRescheduleTimeHour(void);
    virtual void   GetPositionRadKm(double time, double &latRad, double &lonRad, double &altKm);
//////// Data distribution management (DDM) operations.

//////// Data distribution operations.
    SpDistList *CreateDistList(const char *name);
    SpDistList *GetDistList(const char *name);
    void        DestroyDistList(const char *name);
//////// Data distribution operations.

    int  GetActive(void) const { return (Active); }
    void SetActive(int v) { Active = v; }

    C_XQUEUE *GetDistributionList() {return &DistributionList;}
    C_XQUEUE *GetRemoteObjectProxies() {return &RemoteObjectProxies;}

    SpDestination *GetDestination() {return Destination;}

    virtual double GetScanTime() {return 0.0;}
    virtual void set_NAME(char *n) {
      ObjectProxy->SetObjectName(n);
      NAME = strdup(n);
    }

  // proxy::  update list is now in the proxy:
  //////    C_XQUEUE *GetUpdateList() {return &UpdateList;}
    RB_UpdateXqueue *GetUpdateList() {
      SpObjProxy *myProxy = GetObjectProxy();
      assert(myProxy!=NULL);
      return myProxy->GetAttributeUpdateList();
    }

    //proxy:
    // Turn ON collection of attribute updates
    void CollectAttribUpdates(void) {
      SpObjProxy *myProxy = GetObjectProxy();
      assert(myProxy!=NULL);
      myProxy->CollectAttribUpdates();
    }

    // proxy:
    // Turn OFF collection of attribute updates
    void DiscardAttribUpdates(void) {
      SpObjProxy *myProxy = GetObjectProxy();
      assert(myProxy!=NULL);
      myProxy->DiscardAttribUpdates();
    }

    void debug(ostream &os);
};

typedef S_SpHLA S_HLA;

#endif
