/* The basic comm library.  Be careful, whenever sending stuff over
   the network, you need to use the global node ID.  However, on any
   shared memory communication, we use the local node ID's */



#include "SpStdlib.H"
#include "SpErrno.H"
#ifdef __SPEEDES_NT__
#include <winsock.h>
#include <winnt.h>
#else
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#endif

#include "SpNetTypes.H"
#include "SpComm.H"
#include "SpCommShMem.H"
#include "SpCommNet.H"
#include "SpCommPar.H"
#include "SpCommShMemCalls.H"
#include "SpOS.H"


#define IMPORTANT_DEBUG_SP_COMM



int Startup = 0;


/*
** These variable correspond to node information.
*/

int *LocalNodeZeros;          /* This is an array of the GIDs of the local node
				 zero's */
int SpComm_Nboxes = 0;        // The total number of physical computers
int *LocalNodeIDs;            // can reroute message appropriately
int *GlobalNodeIDs;          /* This is an array of the GIDs of the local node
				 zero's */
int *NodeBoxNumber;           /* A map from GlobalNodeIDs to which physical
				 computer a node resides on */
int LocalBoxNumber = 0;       // The current node's box number
int SpComm_Node = 0;          // Our global node number
int SpComm_Nnodes = 1;        // The number of global nodes
extern int SpComm_Nlocal;     // The number of local nodes
extern int SpComm_LocalNode;  // Our Local node number
int GroupId = 1;              // Our group ID


int MsgInOutCount = 0; //Used for the blocking sends

/* These are the shared memory segments required in the shared memory
   library */


extern int Nredops;       // Number of reduction operations for this machine
extern int Nstages;       // The number of stages in the reduction
extern int *StageIndex;   // A pointer to the Stage redop area for stage[i]
extern int *StageElements;// The number of nodes particpating in stage[i]
extern int SpRedopShmemId;  // The shared memory ID for the redop memory
extern SHMEM_REDOP* SpRedop;// The pointer to the redop shared memory
extern char SpRedopCreated; // 1 if the redop shared memory is created, 0 o/w

extern SHMEM_MESSAGE **SpComm_ShmemMessage;
char SpComm_ShmemMessageCreated = 0;

/* How to look up a local node ID from a global node ID */

#define LocalNodeID(GlobalNodeID) (LocalNodeIDs[GlobalNodeID])
#define IsLocalID(GlobalNodeID) (NodeBoxNumber[GlobalNodeID] == LocalBoxNumber)
#define LocalNodeZero(GlobalNodeID) (LocalNodeZeros[GlobalNodeID])


int GlobalNodeID(int LocalNodeID) {

  int i;
  for(i=0; ((NodeBoxNumber[i] != LocalBoxNumber) &&
	    (LocalNodeIDs[i] != LocalNodeID)); ++i);
  return i;
}


/*
** At exit function that cleans up shared memory segments
*/



int Shmem_FlushSendQueue();

void SpComm_CleanupShm()
{
#ifdef IRIX
  // This is needed because of a bug in IRIX which causes the atexit
  // function to be called if abort is called.
  static int seen = 0;
  if (seen) return;
  seen = 1;
#endif

  if (SpComm_LocalNode) return;

  int i;

  // We will not wait for the children to exit.  That means
  // we will try to delete shared memory while the children are
  // still using it.  Will this really delete the memory?

  if (SpRedopCreated){
    if (SHMDT((char*)SpRedop) != 0) {
      PERROR_LOCATION("Error: shmdt failed");
    }

    if (SHMCTL(SpRedopShmemId, IPC_RMID, NULL) != 0) {
      PERROR_LOCATION("Error SpComm_AtExit: could not remove redop shared memory");
    }
    
    if(SpComm_ShmemMessageCreated){
      if (SpComm_ShmemMessage) {
	for (i=0; i<SpComm_Nlocal; i++) SpComm_ShmemMessage[i]->DeleteShm();
	for (i=0; i<SpComm_Nlocal; i++) delete SpComm_ShmemMessage[i];
	delete [] SpComm_ShmemMessage;	
      }
    }
  }
}

void SpComm_CleanupOther()
{
  int i;

  // Make sure we flush out all of our messages
  while (ShmemFlushSendQueue()) SpComm_Receive();

  SpComm_Net_Terminate();
  
  if (SpComm_LocalNode) return;

  if (SpRedopCreated){
    for (i=1; i<SpComm_Nlocal; i++){
#ifndef __SPEEDES_NT__
      wait(NULL);
#endif
    }
    
    if (GlobalNodeIDs) delete [] GlobalNodeIDs;
    if (LocalNodeIDs) delete [] LocalNodeIDs;
    if (LocalNodeZeros) delete [] LocalNodeZeros;
  }
  SpComm_Net_Terminate();
}



void DebuggingBreakpointFunction(int i);

char SpParentProcessId[256];

int SpComm_StartUp(int nTot, int nLoc, int group, char * pipeInfo) {
  int i;
  char shmSegmentName[8192]; //Should not be any longer than this
  int id;
  int nodeInfoShmID;
  LocalBoxNumber = 0;
  int parentProcessId = -1;


  if (nTot < 0){
    SpComm_Nnodes = -nTot;
    SpComm_Nlocal = -nLoc;
    strcpy(SpParentProcessId, pipeInfo);
  }
  else{
#ifdef __SPEEDES_NT__
    parentProcessId = CoGetCurrentProcess();
    sprintf(SpParentProcessId, "%d", parentProcessId);
#endif
    SpComm_Nnodes = nTot;
    SpComm_Nlocal = nLoc;
  }


  GroupId = group;

  /* Set up GIDs and LIDs */
  GlobalNodeIDs = new int[SpComm_Nnodes];
  LocalNodeIDs = new int[SpComm_Nnodes];
  for (i=0; i < SpComm_Nnodes; ++i){
    GlobalNodeIDs[i] = i;
    LocalNodeIDs[i] = -1; //Initially, have no routing
  }

  /*
    ** Figure out how many stages are required for the global reductions and
    ** create the arrays, etc., for managing stages
    */
  i = SpComm_Nlocal;
  while (i != 1) {
    i = (i+1) / 2;
    Nstages++;
  }
    
  StageIndex = new int[Nstages+1];
  for (i=0; i<=Nstages; i++) StageIndex[i] = 0;
    
  StageElements = new int[Nstages+1];
  StageElements[0] = SpComm_Nlocal;
  for (i=1; i<=Nstages; i++) StageElements[i] = 0;
    
  Nredops = SpComm_Nlocal;
  i = SpComm_Nlocal;
  Nstages = 1;
  while(i!=1) {
    StageIndex[Nstages] = Nredops;
    i = (i+1) / 2;
    StageElements[Nstages] = i;
    Nredops += i;
    Nstages++;
  }
    
#ifdef DEBUG_SP_COMM
  cout << "Nstages = " << Nstages << endl;
  cout << "Nredops = " << Nredops << endl;
  for (i=0; i<=Nstages; i++) {
    cout << "StageElements[" << i << "] = " << StageElements[i] << endl;
  }
#endif
    
  if (SpComm_Net_Initialize()) return ERROR_SP_COMM;

  if (nTot > 0){
    SpComm_Pre_Fork_Net_GetIDs(&LocalNodeIDs, SpComm_Nboxes,
			       &LocalNodeZeros, &NodeBoxNumber,
			       LocalBoxNumber);


    
    /* Create the shared memory segment for global reductions, attach
    ** it to the SpRedop array pointer, and then call the init method
    ** for each redop to set everything to zero.
    */
    
    sprintf(shmSegmentName, "SPEEDES_REDOP_SEGMENT_GROUP_%d_uniqueId_%d",
	    group, parentProcessId);
    SpRedopShmemId = SHMGET(IPC_PRIVATE, Nredops*sizeof(SHMEM_REDOP), 0600,
			  shmSegmentName);
    if (SpRedopShmemId < 0) {
      perror("SpComm_StartUp: shmget redops failed.\nRun \"clearipcs\" in the speedes/exe directory if you are not running NT.");
      return ERROR_SP_COMM;
    }
    
    
    /*
    ** Create the shared memory message passing buffers for general message
    ** passing.
    */
    
    SpComm_ShmemMessage = new SHMEM_MESSAGE *[SpComm_Nlocal];
    for (i=0; i<SpComm_Nlocal; i++) {
      SpComm_ShmemMessage[i] = new SHMEM_MESSAGE(OUTPUT_BUFFER_SIZE,
						 INPUT_ARRAY_SIZE,
						 SpComm_Nlocal,
						 i);
      if (SpComm_ShmemMessage[i]->GetErrorStatus() == ERROR_SP_COMM){
	return ERROR_SP_COMM;
      }
    }
    SpComm_ShmemMessageCreated = 1;
    
    
    void * tempPtr = SHMAT(SpRedopShmemId, 0, 0);
    SpRedop = (SHMEM_REDOP *)tempPtr;
    if (SpRedop == (SHMEM_REDOP *)(0xFFFFFFFF)) {
      perror("SpComm_StartUp: shmat redops failed");
      return ERROR_SP_COMM;
    }
    for (i=0; i<Nredops; i++){
      SpRedop[i].init();
    }

    if (SHMDT((char*)SpRedop) != 0) {
      perror("SpComm_StartUp: shmdt failed");
      return ERROR_SP_COMM;
    }
    
    /* Create a shared memory segment for telling the child what node it
       is and (under UNIX), passing the id of the shared memory segments */

    sprintf(shmSegmentName, "SharedMemorySegmentForNodeIDs%d_uniqueId_%d",
	    group, parentProcessId);
    nodeInfoShmID = SHMGET(IPC_PRIVATE,
			   (SpComm_Nlocal+4+SpComm_Nnodes*3)*sizeof(int),
			   0600, shmSegmentName);
    char * nodeInfoShmPtr = SHMAT(nodeInfoShmID, 0, 0);
    char * tptr = nodeInfoShmPtr;
    tptr += sizeof(int);	//Skip over where we will write the childs
				//id number.
    memcpy(tptr, (char *)LocalNodeIDs, sizeof(int)*SpComm_Nnodes);
    tptr += sizeof(int)*SpComm_Nnodes;
    memcpy(tptr, (char *)LocalNodeZeros, sizeof(int)*SpComm_Nnodes);
    tptr += sizeof(int)*SpComm_Nnodes;
    memcpy(tptr, (char *)NodeBoxNumber, sizeof(int)*SpComm_Nnodes);
    tptr += sizeof(int)*SpComm_Nnodes;
    memcpy(tptr, (char *)&SpComm_Nboxes, sizeof(int));
    tptr += sizeof(int); 
    memcpy(tptr, (char *)&LocalBoxNumber, sizeof(int));
    tptr += sizeof(int); 

				//The rest is not really necessary
				//for NT but reduces the number of #ifdefs
    memcpy(tptr, (char *)&SpRedopShmemId, sizeof(int));
    tptr += sizeof(int); 
    for (int ii=0; ii < nLoc; ++ii){
      id = SpComm_ShmemMessage[ii]->GetShmemId();
      memcpy(tptr, (char *)&id, sizeof(int));
      tptr += sizeof(int);
    }


    /*
    ** Spawn Nlocal - 1 process (note, the parent process is local node 0)
    */
    
    
    char pipeArg[256];
#ifndef __SPEEDES_NT__
    sprintf(pipeArg, "%d", nodeInfoShmID); //Ignored in WIN32 but does not
				           //cause a problem.
#else
    sprintf(pipeArg, "%d", parentProcessId);
#endif

    SpComm_LocalNode = 0;
    char groupStr[256];
    char negNodes[256];
    char negLocal[256];
    sprintf(negLocal, "%d", -SpComm_Nlocal);
    sprintf(groupStr, "%d", GroupId);
    sprintf(negNodes, "%d", -SpComm_Nnodes);
    for (i=1; i<SpComm_Nlocal; i++) {
      memcpy(nodeInfoShmPtr, (char *)&i, sizeof(int));
      SpeedesSpawn(pipeInfo, pipeInfo,
		   "-nnodes", negNodes,
		   "-lnodes", negLocal,
		   "-group", groupStr,
		   "-pipe", pipeArg, NULL);
      
      while (*(int *)nodeInfoShmPtr != -1){
	SpeedesSleep(0);
      }
    }
  }
  else{ //if (nLoc < 0) which means we are a child
    SpComm_ShmemMessage = new SHMEM_MESSAGE *[SpComm_Nlocal];
#if defined (__SPEEDES_NT__) || defined (__CYGWIN32__)
    sprintf(shmSegmentName, "SharedMemorySegmentForNodeIDs%d_uniqueId_%s",
	    group, pipeInfo);
    nodeInfoShmID = SHMGET(IPC_PRIVATE, (SpComm_Nlocal+3)*sizeof(int), 0600,
			   shmSegmentName);
#else
    nodeInfoShmID = atoi(pipeInfo);
#endif
    int *nodeInfoShmPtr = (int*)SHMAT(nodeInfoShmID, 0, 0);
    int * tPtr = nodeInfoShmPtr;
    SpComm_LocalNode = *nodeInfoShmPtr;
    ++tPtr;
    delete [] LocalNodeIDs;
    LocalNodeIDs = new int[SpComm_Nnodes];
    memcpy(LocalNodeIDs, tPtr, sizeof(int)*SpComm_Nnodes);
    tPtr += SpComm_Nnodes;

    LocalNodeZeros = new int[SpComm_Nnodes];
    memcpy(LocalNodeZeros, tPtr, sizeof(int)*SpComm_Nnodes);
    tPtr += SpComm_Nnodes;

    NodeBoxNumber = new int[SpComm_Nnodes];
    memcpy(NodeBoxNumber, tPtr, sizeof(int)*SpComm_Nnodes);
    tPtr += SpComm_Nnodes;

    memcpy(&SpComm_Nboxes, tPtr, sizeof(int));
    tPtr++;

    memcpy(&LocalBoxNumber, tPtr, sizeof(int));
    tPtr++;


#if defined (__SPEEDES_NT__) || defined (__CYGWIN32__)
    sprintf(shmSegmentName, "SPEEDES_REDOP_SEGMENT_GROUP_%d_uniqueId_%s",
	    GroupId, pipeInfo);
    SpRedopShmemId = SHMGET(-1, Nredops*sizeof(SHMEM_REDOP),-1,shmSegmentName);
    for (int i=0; i < SpComm_Nlocal; ++i){
      SpComm_ShmemMessage[i] = new SHMEM_MESSAGE(OUTPUT_BUFFER_SIZE,
						 INPUT_ARRAY_SIZE,
						 SpComm_Nlocal,
						 i);
      if (SpComm_ShmemMessage[i]->GetErrorStatus() == ERROR_SP_COMM){
	return ERROR_SP_COMM;
      }
    }
#else

    SpRedopShmemId =  *tPtr;
    ++tPtr;
    for (i=0; i < SpComm_Nlocal; ++i){
      SpComm_ShmemMessage[i] = new SHMEM_MESSAGE(OUTPUT_BUFFER_SIZE,
						 INPUT_ARRAY_SIZE,
						 SpComm_Nlocal,
						 i,
						 *tPtr);
      ++tPtr;
    }
#endif

    SpComm_ShmemMessageCreated = 1;
    *(int*)nodeInfoShmPtr = -1;	//Signal to parent that we have initialized
  }

  if (!SpComm_LocalNode){
    SHMCTL(nodeInfoShmID, IPC_RMID, NULL);
  }

  DebuggingBreakpointFunction(i);


  /*
   * Now everyone needs to initiailize the shared memory message passing
   * segments since SHMAT should be called after the fork().  Similarly,
   * we need to call the SHMAT for the SpRedop after we pass the fork.
   * We also need to initialize the SpRedop but only one node should do this
   * and I just picked node 0 to be the one to do it.
   */
  for (i=0; i < SpComm_Nlocal; ++i){
    SpComm_ShmemMessage[i]->Initialize(OUTPUT_BUFFER_SIZE);
    if (i == SpComm_LocalNode){ //Only clear our own buffer
      SpComm_ShmemMessage[SpComm_LocalNode]->ClearBuffers();
    }
  }
  void * tempJJPtr = SHMAT(SpRedopShmemId, 0, 0);
  SpRedop = (SHMEM_REDOP *)tempJJPtr;
  if (SpRedop == (SHMEM_REDOP *)(0xFFFFFFFF)) {
    perror("SpComm_StartUp: shmat redops failed");
    return ERROR_SP_COMM;
  }
  
  SpComm_Post_Fork_Net_GetIDs(&LocalNodeIDs, SpComm_Nboxes,
			      &LocalNodeZeros, &NodeBoxNumber, LocalBoxNumber);

  for (i = 0; i < SpComm_Nnodes; ++i){
    if ((LocalNodeIDs[i] == SpComm_LocalNode) &&
	(NodeBoxNumber[i] == LocalBoxNumber)){
      SpComm_Node = i;
    }
  }


  Startup = 1;



/*
** Do a quick sync to make sure we have all initialized our SpComm_ShmemMessage
** buffers and set Startup=1 before we return a GOOD_SP_COMM
*/

  SpComm_BarrierSync();

  return GOOD_SP_COMM;

}








int SpComm_GetNumNodes() {
  if (!SpComm_Nnodes) {
    cout<<"Error GetNumNodes, number of nodes not yet set in Startup - Exiting"
	<< endl;
    exit(1);
  }
  return SpComm_Nnodes;
}

int SpComm_GetNodeId() {
  if (!Startup) {
    cout << "Error GetNodeId, SpComm_StartUp must be called first - Exiting"
	 << endl;
    exit(1);
  }
  return SpComm_Node;
}



void SpComm_Receive()
{
  SpComm_Net_Receive();   //Do a network receive
  SpComm_Shmem_Receive(); // Now do a local receive
};

char *SpComm_Receive(int type, int &nbytes)
{
  SpComm_Receive();
  return (char*) SpComm_Shmem_Receive(type, nbytes);
};

char *SpComm_GetPendingMessage(int type, int &nbytes)
{
  return (char*) SpComm_Shmem_GetPendingMessage(type, nbytes);
};

void SpComm_Send(int destination, int type, int nbytes, void *buff)
{
  if (IsLocalID(destination))
    SpComm_Shmem_Send(LocalNodeID(destination),type, nbytes, buff);
  else SpComm_Net_Send(destination, type, nbytes, buff);
}

void SpComm_Send(SpDestination *dest, int type, int nbytes, void *buff){

  int Dest;
  static int **DEST;
  static int firsttime = 1;
  int i, j;
  int PresBox;

  if (firsttime){
    DEST = new int *[SpComm_Nboxes];
    for (i=0; i < SpComm_Nboxes; i++){
      DEST[i] = new int[SpComm_Nnodes+1]; /*should really be
					    (#machines/box)+1*/
    }
    firsttime = 0;
  }
  for (i=0; i < SpComm_Nboxes; DEST[i++][0] = 0);
  Dest = dest->GetFirstNode();
  while (Dest != -1){
    PresBox=NodeBoxNumber[Dest];
    DEST[PresBox][0] = DEST[PresBox][0] + 1;
    DEST[PresBox][DEST[PresBox][0]] = Dest;
    Dest = dest->GetNextNode(Dest);
  }
  /* Now we have built up arrays of the destinations, send them on to
     their respective local node zero's for redistribution */

  for (i=0; i < SpComm_Nboxes; ++i){
    if ((DEST[i][0])){ // If we have messages for box number i
      if (i == LocalBoxNumber){  // If local box, use shared memory
	for (j = 1; j < DEST[i][0]+1; ++j){
	  DEST[i][j] = LocalNodeID(DEST[i][j]);
	}
	SpComm_Shmem_Send(DEST[i], type, nbytes, buff);
      }
      else
	SpComm_Net_Send(DEST[i], type, nbytes, buff);
    }
  }
}


/* Broadcast */
void SpComm_Send(int type, int nbytes, void *buff){

  SpComm_Shmem_Send(type, nbytes, buff); // First broadcast to our box
  SpComm_Net_Send(type, nbytes, buff);   // Then broadcast to the other boxes
}


void SpComm_CoordinatedSend(int dest, int nbytes, void *buff){

  ++MsgInOutCount;
  SpComm_Send(dest, SPCOMM_MESSAGE_BLOCKING_SEND, nbytes, buff);
}

void SpComm_CoordinatedSend(SpDestination *dest, int nbytes, void *buff){

  int Node = dest->GetFirstNode();
  while (Node != -1){++MsgInOutCount; Node = dest->GetNextNode(Node);}
  SpComm_Send(dest, SPCOMM_MESSAGE_BLOCKING_SEND, nbytes, buff);
}

void SpComm_CoordinatedSend(int nbytes, void *buff){

  MsgInOutCount += (SpComm_GetNumNodes() - 1);
  SpComm_Send(SPCOMM_MESSAGE_BLOCKING_SEND, nbytes, buff);
}

#define NEW_BEHAVIOUR 0

char *SpComm_CoordinatedReceive(int &nbytes){

  void * RcvBuff = SpComm_GetPendingMessage(SPCOMM_MESSAGE_BLOCKING_SEND,
					    nbytes);
  //static int EnteredFuzzy = 0;
  //static int ExitedFuzzy = 0;
  if (RcvBuff){
    --MsgInOutCount;
    return (char*) RcvBuff;
  }

#if (NEW_BEHAVIOUR)
  if (!EnteredFuzzy){
    SpComm_EnterFuzzyBarrier();
    EnteredFuzzy = 1;
  }
  if (!ExitedFuzzy){
    while (!SpComm_ExitFuzzyBarrier()){
      RcvBuff = SpComm_Receive(SPCOMM_MESSAGE_BLOCKING_SEND, nbytes);
      if (RcvBuff != NULL) {
	--MsgInOutCount;
	return RcvBuff; 
      }
    }
  }
  ExitedFuzzy = 1;
#endif
  /* Now we have exited fuzzy barrier.  This means everyone thinks
     they have received all of their messages.  Now we start checking
     if we are done.  */
  while (SpComm_GlobalSum(MsgInOutCount)){
    RcvBuff = SpComm_Receive(SPCOMM_MESSAGE_BLOCKING_SEND, nbytes);
    if (RcvBuff != NULL) {
      --MsgInOutCount;
      return (char*) RcvBuff; 
    }
  }
  //EnteredFuzzy = 0;
  //ExitedFuzzy = 0;
  MsgInOutCount = 0;
  return NULL;
}



void DebuggingBreakpointFunction(int NodeNumber)
{
  switch(NodeNumber)
    {
    case 0:
      break;
    case 1:
      break;
    case 2:
      break;
    case 3:
      break;
    case 4:
      break;
    case 5:
      break;
    case 6:
      break;
    case 7:
      break;
    case 8:
      break;
    case 9:
      break;
    case 10:
      break;
    case 11:
      break;
    case 12:
      break;
    case 13:
      break;
    case 14:
      break;
    case 15:
      break;
    case 16:
      break;
    case 17:
      break;
    case 18:
      break;
    case 19:
      break;
    case 20:
      break;
    case 21:
      break;
    case 22:
      break;
    case 23:
      break;
    case 24:
      break;
    case 25:
      break;
    case 26:
      break;
    case 27:
      break;
    case 28:
      break;
    case 29:
      break;
    case 30:
      break;
    case 31:
      break;
    case 32:
      break;
    case 33:
      break;
    case 34:
      break;
    case 35:
      break;
    case 36:
      break;
    case 37:
      break;
    case 38:
      break;
    case 39:
      break;
    case 40:
      break;
    case 41:
      break;
    case 42:
      break;
    case 43:
      break;
    case 44:
      break;
    case 45:
      break;
    case 46:
      break;
    case 47:
      break;
    case 48:
      break;
    case 49:
      break;
    case 50:
      break;
    case 51:
      break;
    case 52:
      break;
    case 53:
      break;
    case 54:
      break;
    case 55:
      break;
    case 56:
      break;
    case 57:
      break;
    case 58:
      break;
    case 59:
      break;
    case 60:
      break;
    case 61:
      break;
    case 62:
      break;
    case 63:
      break;
    }
}
