#ifndef COMMSERVERHEADER_H
#define COMMSERVERHEADER_H

#include "SpNetTypes.H"
#include "SpMsgHeader.H"

/*************************************************************************
 * Class: COMM_SERVER_HEADER
 * Summary: Header of all messages sent with the comm server
 * Description: Messages sent with the comm server using the client
 *              server code require that they have a header that is
 *              derived from the GENERIC_MSG_HEADER class.  This
 *              extends that class to have 4 additional pieces of
 *              information that is important for the comm server and
 *              the client for SpComm_NET.H
 * Examples: Look SpCommTCP.C and CommmServer.C
*************************************************************************/
class COMM_SERVER_HEADER : public GENERIC_MSG_HEADER {
private:
  NET_INT Type;
  NET_INT Node;
  NET_INT Data;
  NET_INT Unused;

public:

  /*************************************************************************
   * Method: getType
   * Summary: Returns the type of the message
   * Description: Returns the type of the message
   * Return: Returns the type of the message
   *************************************************************************/
  int getType(void) { return Type;}
  /*************************************************************************
   * Method: setType
   * Summary: Sets the type of the message to t
   * Description: Sets the type of the message to t
   * Return: void
   *************************************************************************/
  void setType(int t //In: The type of the message
	       )
    {
      Type = t; 
      setUnused(0);
    }
  
  /*************************************************************************
   * Method: getNode
   * Summary: Returns the node of the message
   * Description: Returns the node of the message
   * Return: Returns the node of the message
   *************************************************************************/
  int getNode(void) { return Node; }
  /*************************************************************************
   * Method: setNode
   * Summary: Sets the node of the message
   * Description: Sets the node of the message
   * Return: void
   *************************************************************************/
  void setNode(int nodeNum //In: Either the node of the sender or receiver
	                   //    of this message
	       ) { Node = nodeNum; }
  
  /*************************************************************************
   * Method: getData
   * Summary: Returns the value of the Data member for this message
   * Description: Returns the value of the Data member for this message
   * Return: Returns the value of the Data member for this message
   *************************************************************************/
  int getData(void) { return Data; }
  /*************************************************************************
   * Method: setData
   * Summary: Sets the Data member to ival
   * Description: Sets the Data member to ival
   * Return: void
   *************************************************************************/
  void setData(int ival) { Data = ival; }


  /*************************************************************************
   * Method: getUnused
   * Summary: unused
   * Description: not used
   * Return: unused value
   *************************************************************************/
  int getUnused(void) { return Unused; }
  /*************************************************************************
   * Method: setUnused
   * Summary: Not Used
   * Description: Not Used
   * Return: void
   *************************************************************************/
  void setUnused(int ival) { Unused = ival; }


};


#endif
