#include <sys/types.h>
#include "SpStdlib.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpString.H"
#ifndef __SPEEDES_NT__
#include <sys/time.h>
#include <netdb.h>
#endif

#include "SpNetTypes.H"
#include "SpDestination.H"
#include "SpSimTime.H"
#include "SpComm.H"
#include "SpCommNet.H"
#include "SpCommShMem.H"
#include "SpShMemQueue.H"
#include "SpClientSock.H"
#include "SpServerSock.H"
#include "SpCommServerMsgs.H"
#include "SpCommServerHeader.H"
#include "SpServer.H"

CLIENT_SOCK *SendSocket = NULL;
#ifdef ORIG
char * SpServerMachineName = NULL;
int SpServerPortNum = -1;
int SpGatherStatistics = 0;
#else
extern char * SpServerMachineName;
extern int SpServerPortNum;
extern int SpGatherStatistics;
#endif

int SpCounter=0;

extern int GroupId;
extern int LocalBoxNumber;
extern int SpComm_Nnodes;
extern int SpComm_Nlocal;
extern int SpComm_LocalNode;
extern int SpComm_Node;
extern int SpComm_Nlocal;
extern int *GlobalNodeIDs;
extern int *LocalNodeIDs;
extern int *LocalNodeZeros;
extern int *NodeBoxNumber;

int TCPIP_Initialized = 0;
int RedopDone = 0;
int nbRedopDone = 0;
int RedopInfoInt;
double RedopInfoDouble;
SpSimTime RedopSimTime;

#define LocalNodeID(GlobalNode) (LocalNodeIDs[GlobalNode])
#define NodeZero(GlobalNode) (LocalNodeZeros[GlobalNode])
#define FMNode(GlobalNode) (NodeBoxNumber[GlobalNode])


void InitHeader(COMM_SERVER_HEADER * Hdr){
  static COMM_SERVER_HEADER DefaultHeader;
  static int DefaultHeaderInitialized = 0;

  if (!DefaultHeaderInitialized){
    DefaultHeader.setSrvType(SPEEDES_COMM_SERVER_ID);
    DefaultHeader.setGroupNum(GroupId);
    DefaultHeader.setBytes(-1);
    DefaultHeader.setType(-1);
    DefaultHeader.setUnused(-1);
    DefaultHeader.setNode(-1);
    DefaultHeader.setHandlerNum(-1);
    DefaultHeader.setData(-1);
  }
  memcpy(Hdr, &DefaultHeader, sizeof(COMM_SERVER_HEADER));
}

void RedopSend(int *Ival, double *Dval, SpSimTime * Sval, int Handler)
{
  static int nbytes = sizeof(REDOP_INFO) + sizeof(COMM_SERVER_HEADER);
  static char * Message = new char[nbytes];
  COMM_SERVER_HEADER *Header= (COMM_SERVER_HEADER*)Message;
  REDOP_INFO *Rdp = (REDOP_INFO*)&(Header[1]);
  static int first = 1;

  if (first){
    InitHeader(Header);
    Header->setNode(SpComm_GetNodeId());
    Header->setBytes(nbytes);
    first = 0;
  }
  Header->setHandlerNum(Handler);
  Rdp->Init(*Ival, *Dval, *Sval);
  SendSocket->sendMsg_blk(Header, nbytes);
}


void Redop()
{
  /* With Redop, we simply send them our nodeid.  We really don't even
     need to send anything but I send this so that we could use it for
     debugging*/
  static COMM_SERVER_HEADER Header;
  static int first = 1;

  if (first){
    InitHeader(&Header);
    Header.setHandlerNum(BLOCKING_SYNCHRONIZATION);
    Header.setBytes(sizeof(COMM_SERVER_HEADER));
    Header.setNode(SpComm_GetNodeId());
    first = 0;
  }
  SendSocket->sendMsg_blk(&Header, sizeof(COMM_SERVER_HEADER));
}


void nbRedop()
{
  /* With nbRedop, we simply send them our nodeid.  We really don't
     even need to send anything but I send this so that we could use
     it for debugging*/
  static COMM_SERVER_HEADER Header;
  static int first = 1;

  if (first){
    InitHeader(&Header);
    Header.setHandlerNum(NON_BLOCKING_SYNCHRONIZATION);
    Header.setBytes(sizeof(COMM_SERVER_HEADER));
    Header.setNode(SpComm_GetNodeId());
    first = 0;
  }
  SendSocket->sendMsg_blk(&Header, sizeof(COMM_SERVER_HEADER));
}

void SpComm_Net_Terminate()
{
  if (SpComm_Nnodes != SpComm_Nlocal){
    if (SendSocket){
      delete SendSocket;
      SendSocket = NULL;
    }
  }
}




int SpComm_Net_Initialize()
{
  /* This is the basic stuff that is required for starting up TCPIP. */

  SpEndianNature::DetermineEndianNature();
  int retval = 0;
  if (SpComm_Nnodes != SpComm_Nlocal){
    SendSocket = new CLIENT_SOCK(SpServerMachineName, SpServerPortNum);
    if (SendSocket == NULL){
      retval = 0;
    }
  }
  return retval;
}







static inline int Allign8Bytes(int val) {return val + (8 - val%8);}

/* Need to put together a way for everyone to query FM node 0 for their
   node numbers.  FM node 0 continues to poll until everyone as checked in
   and then it returns this info to each of the nodes */


void SpComm_Pre_Fork_Net_GetIDs(int **LocalNodeIdTable, int &Nboxes,
				int ** LNZ, int **BoxTable,
				int & BoxNumber)
{
  Nboxes = 0;
  int i;
  int Nbytes = sizeof(COMM_SERVER_HEADER)+2*sizeof(NET_INT);
  SockMsgType Msg;
  char * buff;

  *LNZ = new int[SpComm_Nnodes];
  *BoxTable = new int[SpComm_Nnodes];
  delete [] *LocalNodeIdTable;
  *LocalNodeIdTable = new int[SpComm_Nnodes];

  if (SpComm_Nnodes != SpComm_Nlocal){
    char * Message = new char[Nbytes];
    COMM_SERVER_HEADER *Header = (COMM_SERVER_HEADER*)Message;
    NET_INT *msg = (NET_INT*)&Header[1];
    InitHeader(Header);
    Header->setHandlerNum(TCPIP_STARTUP);
    Header->setNode(SpComm_Nlocal);
    Header->setBytes(2*sizeof(NET_INT));
    msg[0] = SpComm_Nnodes;
    msg[1] = SpGatherStatistics;
    SendSocket->sendMsg_blk(Header, Nbytes);
    delete [] Message;

    SendSocket->getMessage_blk(&Msg);
    char *MsgBuff = Msg.getMsgBuff();
    buff = MsgBuff + sizeof(COMM_SERVER_HEADER);
    NET_INT *array;

    array = (NET_INT*) buff;
    for (i=0; i < SpComm_Nnodes; ++i){
      (*LocalNodeIdTable)[i] = array[i];
    }
    buff += Allign8Bytes(sizeof(NET_INT)*SpComm_Nnodes);

    array = (NET_INT*) buff;
    for (i=0; i < SpComm_Nnodes; ++i){
      (*LNZ)[i] = array[i];
    }
    buff += Allign8Bytes(sizeof(NET_INT)*SpComm_Nnodes);

    array = (NET_INT*) buff;
    for (i=0; i < SpComm_Nnodes; ++i){
      (*BoxTable)[i] = array[i];
    }
    buff += Allign8Bytes(sizeof(NET_INT)*SpComm_Nnodes);

    Nboxes = *(NET_INT*)buff;
    buff += Allign8Bytes(sizeof(NET_INT));

    BoxNumber = *(NET_INT*)buff;

    delete [] MsgBuff;
  }
  else{
    for (i=0; i < SpComm_Nlocal; ++i){
      LocalNodeIdTable[0][i] = i;
      BoxTable[0][i] = LNZ[0][i] = 0;
    }
    BoxNumber = 0;
    Nboxes = 1;
  }
  

  LocalNodeIDs= *LocalNodeIdTable;
  NodeBoxNumber = *BoxTable;
  LocalNodeZeros = *LNZ;

  
#if 0
  cout<<endl<<"  GID  LID  BOX Node0"<<endl;
  for (i=0; i < SpComm_Nnodes; ++i)
    cout<<"   "<<i<<"    "<<LocalNodeIDs[i]<<"    "
	<<NodeBoxNumber[i]<<"    "<<LocalNodeZeros[i]<<endl;
  cout<<endl;
#endif
}





void SpComm_Post_Fork_Net_GetIDs(int **localNodeIDs,
				 int &,
				 int **localNodeZeros,
				 int **localBoxNumber,
				 int &BoxNumber)
{
  /* Here, all non-node 0's need to contact the comm_server so that the
     comm_server can get a socket port for each person */

  int i, MyNode = 0;
  if (SpComm_Nlocal != SpComm_Nnodes){
    if (SpComm_LocalNode != 0){ //Only non-node0's need to connect now.
      COMM_SERVER_HEADER Header;

      LocalNodeIDs = *localNodeIDs;
      NodeBoxNumber = *localBoxNumber;
      LocalNodeZeros = *localNodeZeros;

      //SendSocket = new CLIENT_SOCK(MachineName, Port);
      
      for (i = 0; i < SpComm_Nnodes; ++i){
	if ((LocalNodeIDs[i] == SpComm_LocalNode) &&
	    (NodeBoxNumber[i] == LocalBoxNumber)){
	  MyNode = i;
	}
      }
      InitHeader(&Header);
      Header.setHandlerNum(TCPIP_STARTUP);
      Header.setNode(MyNode);
      cout<<"My node is "<<MyNode<<endl;
      SendSocket->sendMsg_blk(&Header, sizeof(COMM_SERVER_HEADER));
      
    }
    else{
      SpeedesSleep(1);
    }
  }
  else {
    BoxNumber = 0;
  }
}



/*inline */void BlockingRedopReceive()
{
  static SockMsgType Msg;
  SpComm_Net_Receive(1);
  
}



void  SpComm_Net_Redop()
{
  if (SpComm_Nnodes != SpComm_Nlocal){
    Redop(); //Send a message to node 0
    while (!RedopDone){
      BlockingRedopReceive(); // Wait for the reply
    }
    RedopDone = 0;
  }
}


void SpComm_Net_Max(int * ival, double * dval, SpSimTime * Stime, int)
{
  if (SpComm_Nnodes != SpComm_Nlocal){
    RedopSend(ival, dval, Stime, GLOBAL_MAX);
    
    while (!RedopDone) BlockingRedopReceive();

    *ival = RedopInfoInt;
    *dval = RedopInfoDouble;
    *Stime = RedopSimTime;
  }
  RedopDone = 0;
}


void SpComm_Net_Sum(int * ival, double * dval, SpSimTime * Stime, int)
{
  if (SpComm_Nnodes != SpComm_Nlocal){
    RedopSend(ival, dval, Stime, GLOBAL_SUM);

    while (!RedopDone) BlockingRedopReceive();
    
    *ival = RedopInfoInt;
    *dval = RedopInfoDouble;
    *Stime = RedopSimTime;
  }
  RedopDone = 0;
  }


void SpComm_Net_Min(int * ival, double * dval, SpSimTime * Stime, int)
{
  if (SpComm_Nnodes != SpComm_Nlocal){
    RedopSend(ival, dval, Stime, GLOBAL_MIN);
    
    while (!RedopDone){
      BlockingRedopReceive();
    }
    *ival = RedopInfoInt;
    *dval = RedopInfoDouble;
    *Stime = RedopSimTime;
  }
  RedopDone = 0;
}



static int NonBlockingEntered = 0;

void SpComm_Net_Enter_nbRedop()
{
  if (SpComm_Nnodes != SpComm_Nlocal){
    nbRedop();
    NonBlockingEntered = 1;
  }
}


int SpComm_Net_nbRedop()
{
  if (SpComm_Nnodes != SpComm_Nlocal){
    if (!NonBlockingEntered){
      SpComm_Net_Enter_nbRedop();
    }
    SpComm_Net_Receive();
    if (nbRedopDone){ //If I am the local node zero and got a message
      nbRedopDone = 0;
      NonBlockingEntered = 0;
      return 1;
    }
    return 0;
  }
  nbRedopDone = 0;
  NonBlockingEntered = 0;
  return 1;
}


int GetWholeMessage(SockMsgType * Msg)
{
  int RETVAL = STREAM_SOCK::PARTIAL_MSG_OK;
  int Ntries = 0;

  while ((RETVAL == STREAM_SOCK::PARTIAL_MSG_OK) && (Ntries < 10)){
     RETVAL = SendSocket->getMessage(Msg);
     ++Ntries;

  }
  if (RETVAL == STREAM_SOCK::WHOLE_MSG_OK){
    return 0;
  }
  else{
    return 1;
  }
}

int BlockingGetWholeMessage(SockMsgType * Msg)
{
  int RETVAL = STREAM_SOCK::PARTIAL_MSG_OK;
  int Ntries = 0;

  while ((RETVAL == STREAM_SOCK::PARTIAL_MSG_OK) && (Ntries < 10)){
     RETVAL = SendSocket->getMessage_blk(Msg);
     ++Ntries;

  }
  if (RETVAL == STREAM_SOCK::WHOLE_MSG_OK){
    return 0;
  }
  else{
    return 1;
  }
}


void SpComm_Net_Receive(int blocking)
{
  if (SpComm_Nnodes != SpComm_Nlocal){
    int i;
    SockMsgType Message;
    char * Msg;
    NET_INT * DEST;
    char * TrueMsg;
    REDOP_INFO * RedInf;
    int Nbytes;
    COMM_SERVER_HEADER * Hdr;
    static int *LocalNodeDEST = NULL;

    /* Pulls a single msg off of the network */
    while (1){
      if (blocking){
	if (BlockingGetWholeMessage(&Message)){
	  break;
	}
      }
      else{
	if (GetWholeMessage(&Message)){
	  break;
	}
      }
      //Didn't break since we got a whole message so handle that message
      Hdr = (COMM_SERVER_HEADER *)Message.getMsgBuff();
      Msg = (char *) &Hdr[1];
      if (Msg == NULL) return; //Should not get NULL messages

      switch (Hdr->getHandlerNum()){
      case NORMAL_MESSAGE:
	SpComm_Shmem_Send(LocalNodeID(Hdr->getNode()), Hdr->getType(), 
			  Hdr->getBytes(), Msg);
	break;
	
      case MESSAGE_MULTICAST:
	DEST = (NET_INT *) Msg;
	if (LocalNodeDEST == NULL){
	  LocalNodeDEST = new int[SpComm_GetNumNodes() + 1];
	}
	Nbytes = Hdr->getBytes() - Allign8Bytes(((*DEST)+1)*sizeof(NET_INT));
	TrueMsg = (char *) (Msg +  Allign8Bytes(((*DEST)+1)*sizeof(NET_INT)) );
	for (i=0; i < *DEST; ++i){
	  LocalNodeDEST[i+1] = LocalNodeID(DEST[i+1]);
	}
	LocalNodeDEST[0] = DEST[0];
	SpComm_Shmem_Send(LocalNodeDEST, Hdr->getType(), Nbytes, TrueMsg);
	break;

      case MESSAGE_BROADCAST:
	SpComm_Shmem_Send(SpComm_LocalNode, Hdr->getType(),
			  Hdr->getBytes(), Msg); //Send to myself
	SpComm_Shmem_Send(Hdr->getType(), Hdr->getBytes(), Msg); /* broadcast
								    to box */
	break;
      case GLOBAL_SUM:
      case GLOBAL_MIN:
      case GLOBAL_MAX:
	RedInf = (REDOP_INFO*) Msg;
	RedopInfoInt = RedInf->GetInt();
	RedopInfoDouble = RedInf->GetDouble();
	RedopSimTime = RedInf->GetSimTime();
      case BLOCKING_SYNCHRONIZATION: //We actually do want to fall through here
	RedopDone = 1;
	break;
      case NON_BLOCKING_SYNCHRONIZATION:
	nbRedopDone = 1;
	break;
      default:
	cout<<"ERROR!! Bad message Handler: "<<Hdr->getHandlerNum()
	    <<"  bytes = "<<Hdr->getBytes()<<"  Type = "<<Hdr->getType()
	    <<"  Data = "<<Hdr->getData()<<"  Node = "<<Hdr->getNode()<<endl;
	exit(-2);
	break;
      }
      if (nbRedopDone || RedopDone){
	if (nbRedopDone){
	  //cout<<"Exiting a fuzzy barrier "<<SpCounter<<endl;
	}
	else{
	  //cout<<"Exiting a blocking barrier "<<SpCounter<<endl;
	}
	SpCounter++;
	break;
      }
      delete [] (char *) Hdr;
    }
  } //SpComm_Nnodes != SpComm_Nlocal
}




void SpComm_Net_Send(int dest, int type, int nbytes, void * buff)
{
  /* Each message starts with a header containing the destination,
     type, and number of bytes.  This is so that FM can grab out what
     pieces it needes */

#ifdef DEBUG_SEND_DELAY
  static int UnusedValue = 0x100000*(SpComm_GetNodeId()+1);
  static int counter = 0;
#endif
  if (SpComm_Nnodes != SpComm_Nlocal){
    int Nbytes = sizeof(COMM_SERVER_HEADER) + nbytes;
    char * Msg = new char [Nbytes];
    COMM_SERVER_HEADER *Header = (COMM_SERVER_HEADER *)Msg;
    
    InitHeader(Header);
    Header->setType(type);
    Header->setBytes(nbytes);
    Header->setHandlerNum(NORMAL_MESSAGE);
    Header->setNode(dest);
    memcpy(&(Header[1]), buff, nbytes);

#ifdef DEBUG_SEND_DELAY
    Header->setUnused(UnusedValue | (counter++));
    PrintMessageSendTime(Header->getUnused());
#endif
    SendSocket->sendMsg_blk(Header, Nbytes);
    delete [] Msg;
  }
  else{
    cout<<"error, trying a network send from "<<SpComm_Node<<" to "
	<<dest<<" with only shared memory"<<endl;
  }
}







void SpComm_Net_Send(SpDestination *, int, int, void *)
{
  cout<<"SpComm_Net_Send(SpDestination *, int, int, void *) not working yet "
      <<endl;
  exit(1);
  return;

  //Must complete this one.
}

void SpComm_Net_Send(int *dest, int type, int nbytes, void *buff)
{  

  NET_INT * Dest;
  if (SpComm_Nnodes != SpComm_Nlocal){
    int Ndest = *dest;
    int Nbytes = nbytes + Allign8Bytes((sizeof(NET_INT)*(Ndest+1)));
    char * Msg = new char[Nbytes+sizeof(COMM_SERVER_HEADER)];
    COMM_SERVER_HEADER *Header = (COMM_SERVER_HEADER*)Msg;
    char * ptr = (char *)&(Header[1]);

    Dest = (NET_INT*)ptr;
    for (int j=0; j < Ndest+1; ++j){
      *Dest = *dest;
      ++Dest;
      ++dest;
    }
    ptr += Allign8Bytes(sizeof(NET_INT)*(Ndest+1));

    memcpy(ptr, buff, nbytes);

    InitHeader(Header);
    Header->setType(type);
    Header->setBytes(Nbytes);
    Header->setHandlerNum(MESSAGE_MULTICAST);
    //Header->setData(type);
    Header->setNode(SpComm_GetNodeId());

    SendSocket->sendMsg_blk(Header, sizeof(COMM_SERVER_HEADER)+Nbytes);
    delete [] Msg;
  }
  else{
    cout<<"error, trying a network Multi-send from "<<SpComm_Node
	<<" with only shared memory"<<endl;
  }
}




/* Broadcast to the other boxes */
void SpComm_Net_Send(int type, int nbytes, void *buff)
{
  if (SpComm_Nnodes != SpComm_Nlocal){
    char * Msg = new char[sizeof(COMM_SERVER_HEADER)+nbytes];
    COMM_SERVER_HEADER *Header = (COMM_SERVER_HEADER*)Msg;

    InitHeader(Header);
    Header->setType(type);
    Header->setBytes(nbytes);
    Header->setHandlerNum(MESSAGE_BROADCAST);
    Header->setNode(SpComm_GetNodeId());
    memcpy(&(Header[1]), buff, nbytes);
    SendSocket->sendMsg_blk(Header, sizeof(COMM_SERVER_HEADER)+nbytes);
    delete [] Msg;
  }
}




