#include <sys/types.h>
#ifndef __SPEEDES_NT__
#include <netdb.h>
#include <unistd.h>
#else
#include <windows.h>
#endif

#include "SpErrno.H"
#include "SpStdlib.H"
#include "SpString.H"
#include "SpClientSock.H"
#include "SpOS.H"

CLIENT_SOCK:: CLIENT_SOCK(char *server_name, int server_port, int waitSecs) {
   ServerName = strdup(server_name);

   if (server_port == -1){
     server_port = ServerPort = 10000 + 
#ifndef __SPEEDES_NT__
       int(getuid());
#else
       0;
#endif
   }
   else{
     ServerPort = server_port;
   }
   setSocketStatus(CLIENT);

   if (connectToServer(server_name, server_port, waitSecs)) {

      ServerConn = new CONNECTION(getfd(), server_port);
   }
   else {
      ServerConn = NULL;
//      exit(1);
   }
}

CLIENT_SOCK::~CLIENT_SOCK(void) {
  free(ServerName);
  delete ServerConn;
}

// Connect to server machine named serverName on port serverPort.  Return 1 if success;  0 otherwise.
int CLIENT_SOCK::connectToServer(char *serverMachineName,
				 int serverPort,
				 int WaitSecs) {
   struct sockaddr_in serverAddr;
   hostent *hp;

   hp = gethostbyname(serverMachineName);
   if (hp == NULL) {
      cerr << "Unknown server: " << serverMachineName << endl;
      return 0;
   }

   serverAddr.sin_family = AF_INET;
   memcpy((char *)&serverAddr.sin_addr, (char *)hp->h_addr, hp->h_length);
   serverAddr.sin_port = htons(serverPort);

   return connectToServer(&serverAddr, WaitSecs);
}

int CLIENT_SOCK::connectToServer(struct sockaddr_in *serverAddr,
				 int waitSecs) {
   int needMsg = 1;
   int WaitTime = waitSecs;

   while (waitSecs >= 0) {
      if (connect(getfd(), (sockaddr*) serverAddr, sizeof *serverAddr) < 0) {
	 if (errno == ECONNREFUSED) {
	    if (needMsg) {
	      if (waitSecs > 0){
		cout << "Client refused connection; waiting.  Server name: "
		     << ServerName << endl;
	      }
	       needMsg = 0;
	    }
	    if (waitSecs > 0) {
	      SpeedesSleep(5);
	    }
	    waitSecs -= 5;

	    // If sock was refused, OS closes it, so we need to re-open:
#ifndef __SPEEDES_NT__
	    setfd(socket( AF_INET, SOCK_STREAM, 0 ));
	    if (getfd() < 0 ) {
	      perror( "Couldn't open socket stream" );
	      exit(1);
	    }
	    if (waitSecs > 0){
	      cout << "Re-open sock... "<<waitSecs << endl;
	    }
#endif
	 }
	 else {
	   if (waitSecs > 0){
	     perror("connecting stream socket");
	   }
	   return 0;
	 }
      }
      else {  // Made connection OK:
	return 1;
      }
   }
   
   if (WaitTime > 0){
     cout << "Client STILL refused connection after "
	  << WaitTime << " seconds; quitting." << endl;
   }
   return 0;
}


GENERIC_MSG_HEADER *CLIENT_SOCK::getMsg_blk() {
  static SockMsgType msg;
  ServerConn->getMsg_blk(&msg);
  return (GENERIC_MSG_HEADER *)msg.getMsgBuff();
}
/*********
	  void CLIENT_SOCK::handleBadRecv(int status, CONNECTION *conn) {
	  switch (status) {
	  case NO_MSG_AVAIL:
	  cout << "Conn tried to read msg when none there." << endl;
	  break;
	  case CLOSED_CONN:
	  cout << "handleBadRecv: CLOSED_CONN internal msg" << endl;
	  closeConnection(conn);
	  break;
	  default:
	  cout << "Unknown bad status = " << status << endl;
	  break;
	  }
	  }
**********/
