#ifndef server_socket
#define server_socket

#include "SpStreamSock.H"

#include "SpTime.H"  // For Linux, to define FD_CLR


class SERVER_SOCK : public STREAM_SOCK {

 public:
   SERVER_SOCK(int port=DEFAULT_PORT, int accept_qsize=ACCEPT_Q_LENGTH);

   void listenForConnections(int acceptQsize);
    
   // Accept new connection; insert in cgroup.  return connect fd or
   // -1 if error
   int acceptConnection(void);

   // zero-out bits in fd_sets;  decrement NumConnections
   void closeConnection(int fd) {
      FD_CLR(fd, &Readfds);
      FD_CLR(fd, &Writefds);
      FD_CLR(fd, &Exceptionfds);

      NumConnections--;
      if (fd==MaxFD) MaxFD--;
   }


/* checkForAllConnectionEvents Checks for read, write, and exception
   events on all file descriptors denoted by Readfds, Writefds, and
   Exceptionfds. Returns the # of such events from select call.  The
   descriptors w/events are returned in the data members Copy_readfds,
   Copy_writefds, and Copy_exceptionfds.  To check if file descriptor
   fd0 has a pending READ, use macro FD_ISSET(fd0, &Copy_readfds)
   which is True iff the Copy_readfds bitmask indicates a pending read
   at fd0 (likewise for write, exception).  The args default to a
   BLOCKING select call; to poll (ie wait 0 secs before timing out)
   use checkForAllConnectionEvents(0); otherwise use the call:
   checkForAllConnectionEvents(0, timeOutSecs) where timeOutSecs=FLOAT
   # secs before timing out */
   int checkForAllConnectionEvents(int blockp=1, float waitSecs=-1);


/* checkForAllConnectionReads Checks for read events on all file
   descriptors denoted by Readfds, Returns the # of such events from
   select call.  The descriptors w/events are returned in the data
   member Copy_readfds.  To check if file descriptor fd0 has a pending
   READ, use macro FD_ISSET(fd0, &Copy_readfds) which is True iff the
   Copy_readfds bitmask indicates a pending read at fd0.

   The args default to a BLOCKING select call; to poll (ie wait 0 secs
   before timing out) use checkForConnectionReads(0); otherwise use
   the call: checkForConnectionReads(0, timeOutSecs) where
   timeOutSecs=FLOAT # secs before timing out */

   int checkForConnectionReads(int blockp=1, float waitSecs=-1);

   // Same as above except checks which connections are WRITABLE as well
   // as readable.
   int checkForConnectionReadsAndWrites(int blockp=1, float waitSecs=-1);

   // Is someone trying to connect to us?
   int connectIsWaiting_p(void);

 protected:
   // Sets implemented as array of longs; 1 bit/file descriptor; used
   // by select (see types.h)
   fd_set Readfds, Copy_readfds;
   fd_set Writefds, Copy_writefds;
   fd_set Exceptionfds, Copy_exceptionfds;


 private:
   int NumConnections;
   int MaxFD;  // Max file descriptor# (for use in select calls)

};
 

#endif
