#include "SpString.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpStdlib.H"
#include "SpCircularBuff.H"

  CIRCULAR_BUFFER::CIRCULAR_BUFFER(const int sz) {
    BuffSize = sz;
    Buff = new char [sz];
    I_Created = 1;
    Head = 0;
    Tail = 0;
    Recheck = 1;
    BufferFuzzyEnd = BuffSize-1;
  }

  CIRCULAR_BUFFER::CIRCULAR_BUFFER(const int sz, //In: Size of the circular buffer to be created
		  char * Buffer //Buffer in which the circular buffer will
		                //reside.
		  ) {
    BuffSize = sz;
    Buff = Buffer;
    I_Created = 0;
    Head = 0;
    Tail = 0;
    BufferFuzzyEnd = BuffSize-1;
    Recheck = 1; // Jacob added to prevent UMR
    memset(Buff, sz, 0);
  }

void CIRCULAR_BUFFER::GarbageCollect()
  { // Called by checkspace to free read buffer regions
    
    CIRC_BUFF_HEADER * CBHeader; // Pointer to the present circ buff header
                                 // that is being considered for deletion
    
    // While the read flag is set at the tail
    while (((CBHeader = (CIRC_BUFF_HEADER*)&Buff[Tail])->GetReadFlag()==1) &&
	   (Tail != Head)){
      // If this is a multi send and I am not the last in line to read it
      if (CBHeader->GetOffset() > CBHeaderSize) Tail += CBHeaderSize;
      // Otherwise, I am the last to read this message so skip past it.
      else Tail+=(Allign8Bytes(CBHeaderSize+CBHeader->GetNbytes()));
      //else Tail+=(CBHeaderSize+CBHeader->GetNbytes()+CBHeader->GetNbytes()%8);
      // If I have hit the end of the buffer, run back to zero and start over.
      if (Tail == BufferFuzzyEnd){
	if (Head == BufferFuzzyEnd){
	  Head = 0;
	}
	BufferFuzzyEnd = 0;
	Tail = 0;
      }
    }
  }

  int CIRCULAR_BUFFER::CheckSpace(const int Givenbytes, const int n) {
/* returns an integer giving the location of where one can fit this
   buffer.  On error, it returns -1 */

    int RC = -1;
    int Nbytes = Allign8Bytes(Givenbytes) + n*CBHeaderSize; // total bytes
                                                            // required
    int BetweenSpace;       // space between the head and tail.
                            // -1 to make sure Head+msg != Tail
    int EndSpace;           //Space at the end of the buffer
    int StartSpace;         // Space at the start of the buffer.
                            // -1 to make sure Head+msg != Tail

    if (Head == Tail){ // No messages in the buffer
      Head = Tail = BufferFuzzyEnd = 0;
      if (Nbytes < BuffSize) return Head;
      else{
	cout<<"CIRCULAR_BUFFER.CheckSpace(): Tried to create message "
	    <<"larger than buffer size"<<endl
	    <<"Buffer size is "<<BuffSize
	    <<" and message Size is "<<Nbytes<<endl;
	exit(1);
      }
    }
    /* This is the case where the Head has wrapped around the end of
       the buffer (creating a FuzzyBuffer) and the Tail still has not
       wrapped as well.  */
    if (Head < Tail){
      BetweenSpace = Tail - Head - 1;
      if (BetweenSpace > Nbytes) return Head;
      else GarbageCollect();}
    /* Otherwise, we have a contiguous piece of buffer with the tail
       behind the head and empty space at the begining, end, or
       both.*/
    else {
      EndSpace = BuffSize - Head - 1; //Space at the end of the buffer
      /* First check to see if I can put the message at the end.  The
	 "-1" is just a sort of Kluge to make sure that I don't
	 overflow my buffer */
      if (EndSpace > Nbytes) return Head;
      else{
	/* Then check to see if I can put the message at the begining.
	   This means that I am going to leave some blank space at the
	   end and I need to update the position of the FuzzyBuffer */
	StartSpace = Tail-1;
	if (StartSpace > Nbytes){
	  BufferFuzzyEnd = Head; //Update the "fuzzy" end of the buffer
	  return 0;} //Not enough space at end, so wrap to begining
	else  GarbageCollect();
      }
    }
    /* Now we have fallen all the way through without freeing any
       space.  We have called GarbageCollect() and will now check for
       free space again.  If we don't find free space on a second
       attempt, we return -1 and set ReCheck for the next attempt.
       Notice that this means I will only call GarbageCollect once
       before reporting that there is no space in the buffer.  */
    
    if (Recheck){
      Recheck = 0;
      // GarbageCollect has already been called so maybe there is some room now
      RC = CheckSpace(Givenbytes, n);
    }
    Recheck = 1;
    return RC;
  }

  int CIRCULAR_BUFFER::Write(const int bytes,   //In: Size of message in bytes
	    const char * buff, //In: Message in this buff
	    const int type,    //In: Type of message
	    const int n        //In: Number of recipients
	    ) {
    int i;
    int StartSpace;
    CIRC_BUFF_HEADER *CircBuffHeader;

    if ((StartSpace = CheckSpace(bytes, n)) == -1){
      return -1;
    }
    CircBuffHeader = (CIRC_BUFF_HEADER *)&Buff[StartSpace];

    for (i=0; i<n; i++){
      CircBuffHeader->init(bytes, (n-i)*CBHeaderSize, type);
      ++CircBuffHeader;}
    memcpy((char *)CircBuffHeader, buff, bytes);
    Head = StartSpace + n*CBHeaderSize + Allign8Bytes(bytes);
    return StartSpace;
  }

  char * CIRCULAR_BUFFER::GetBuff(const int index, //In: Index of the header
		 int &Type,       //Out: Type of the message pointed to by the
		                  //     header
		 int &Nbytes      //Out: Size of the message pointed to by the
		                  //     header
		 ) {
    CIRC_BUFF_HEADER *CircBuffHeader = // circular buffer header given at index
      (CIRC_BUFF_HEADER *)&Buff[index];
    Type = CircBuffHeader->GetType();
    Nbytes = CircBuffHeader->GetNbytes();
    if (CircBuffHeader->GetReadFlag()){
      cout<<"Reading an already read message"<<endl;
    }
    return &Buff[index + CircBuffHeader->GetOffset()];
  }

  char *CIRCULAR_BUFFER::ReadBuff(const int index, //In: Index of the header
		 int &Type,       //Out: Type of the message pointed to by the
		                  //     header
		 int &Nbytes      //Out: Size of the message pointed to by the
		                  //     header
		 ) {
    char * buff; // pointer to buffer that will be returned
    CIRC_BUFF_HEADER *CircBuffHeader = // circular buffer header given at index
      (CIRC_BUFF_HEADER *)&Buff[index];
    Type = CircBuffHeader->GetType();
    Nbytes = CircBuffHeader->GetNbytes();
    if (CircBuffHeader->GetReadFlag()){
      cout<<"Reading an already read message"<<endl;
    }
    buff = new char[Nbytes];
    memcpy(buff, &Buff[index + CircBuffHeader->GetOffset()], Nbytes);
    CircBuffHeader->SetReadFlag();
    return buff;
  }
