// SpCommShMem.C file supports all of the shared memory SPEEDES Comm.
#define SPCOMM_SHMEM_C


int DEBUG = 0;
extern int MultiDebug;
extern int BIGDEBUG;
extern int GroupId;		    //From SpComm.C
extern char SpParentProcessId[256]; //From SpComm.C

#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpStdlib.H"
#ifndef __SPEEDES_NT__
#include <sys/time.h>
#include <sys/types.h>
#include "SpSignal.H"
#include <unistd.h>
#endif

#include "SpComm.H"
#include "SpCommShMem.H"
#include "SpCommPar.H"
#include "SpNetTypes.H"
#include "SpCommNet.H"
#include "SpCommShMemCalls.H"
#include "SpOS.H"

//#define DEBUG_SP_COMM
#ifdef DEBUG_SP_COMM
#define Node SpComm_Node
#endif

#define IMPORTANT_DEBUG_SP_COMM

#ifdef VOLATILE_IS_THE_RIGHT_THING_BUT_CENTERLINE_DOES_NOT_SUPPORT_VOLATILE

#define THE_REDOP (*(volatile SHMEM_REDOP **)&SpRedop)
#ifndef NOP_SHOULD_SLEEP
inline NOP_OP(){}
#endif

#else

#define THE_REDOP SpRedop
int NopUsesThisSinceVolatileDoesNotWork = 0;

#ifndef NOP_SHOULD_SLEEP
inline void NOP_OP() {++NopUsesThisSinceVolatileDoesNotWork;}
#endif

#endif // VOLATILE_IS_THE_RIGHT_THING_BUT_CENTERLINE_DOES_NOT_SUPPORT_VOLATILE




#ifdef NOP_SHOULD_SLEEP

/* If NOP_SHOULD_SLEEP is defined, then we will sleep for a microsecond.
 * This could be useful if one is trying to run more SpeedesComm processes
 * than there are processors.  This will force the current process to sleep
 * in the middle of a reduction rather than trying to poll on that value.
 * This tells the OS that it can switch out the present process and allow
 * another (probably SpeedesComm) process to run.  This will _greatly_
 * speedup exectution times although it will still not improve them as much
 * as if you only ran n processes on n processors.
 */

#define NOP {SpeedesMicroSleep(1);}
#else
#define NOP {NOP_OP();}
#endif //NOP_SHOULD_SLEEP



#define MIN(a,b) ( ((a) < (b)) ? (a) : (b))
#define MAX(a,b) ( ((a) > (b)) ? (a) : (b))



/* Globals corresponding to node information */

int SpComm_LocalNode = 0;
int SpComm_Nlocal = 1;
extern int SpComm_Nnodes;

/*
** Static global variables
*/

/*
** These variables are globals used for the reductions.
*/

int Nredops = 0;            // Number of reduction operations for this machine
static int Toggle = 0;      // The present state of the toggle (0 or 1)
int Nstages = 1;            // The number of stages in the reduction
int *StageIndex;            // A pointer to the Stage redop area for stage[i]
int *StageElements;         // The number of nodes particpating in stage[i]
int SpRedopShmemId;         // The shared memory ID for the redop memory
SHMEM_REDOP* SpRedop;       // The pointer to the SpRedop shared memory
char SpRedopCreated = 0;    // 1 if the redop shared memory is created, 0 o/w

/*
** These are for the fuzzy reductions
*/

static int FuzzyStage;           // The stage for the fuzzy reductions
static int FuzzyNetworkWait = 0; /* 1 if node 0 needs to check the netowrk
				    since the local reduction is done. 0
				    otherwise. */



SHMEM_MESSAGE **SpComm_ShmemMessage = NULL;


void PerrorLocation(char *message, char *file, int line) {
  ostrstream os;
  os << message << ": at " << file << ", line " << line << ends;
  perror(os.str());
}




SpSimTime NetDummySTime;
int NetDummyInt;
double NetDummyDouble;

void SHMEM_MESSAGE::DeleteShm()
{
  if (SHMDT((char*)SharedMemoryAddress) != 0) {
    PERROR_LOCATION("SpComm_StartUp: shmdt failed");
    return;
  }

  if (SHMCTL(ShmemId, IPC_RMID, NULL) != 0) {
    PERROR_LOCATION("SpComm_AtExit: could not remove SHMEM_MESSAGE shared memory");
    return;
  }
}

SHMEM_MESSAGE::~SHMEM_MESSAGE() {
  int nbytes, type, dest;

  if (AlreadyDeleted){
    return;
  }

  AlreadyDeleted = 1;
  delete [] InputArray;
  delete [] LastReadIndex;
  delete [] LastWriteIndex;
  delete [] PartialMsgPtr;
  delete [] PartialMsg;
  delete OutgoingMessageBuffer;



  for (int i=0; i < N_MESSAGE_TYPES+SHMEM_MESSAGE_TYPES; ++i){
    while (InQueue[i].Count()){
      delete [] InQueue[i].Remove(nbytes, type, dest);
    }
  }
  while (AnnounceQueue.Count()){
    delete [] AnnounceQueue.Remove(nbytes, type, dest);
  }
  while (TempSendQueue.Count()){
    delete [] TempSendQueue.Remove(nbytes, type, dest);
  }
  if (NopUsesThisSinceVolatileDoesNotWork == 1.5){
    cout<<"This will never happen but prevents this from being optimized out"
	<<endl;
  }
}

void SHMEM_MESSAGE::Initialize(const int OutBufferSize)
{
  int i;
  SharedMemoryAddress = (char *)SHMAT(ShmemId, 0, 0);

  if (SharedMemoryAddress == (char *)(0xFFFFFFFF)) {
    cout<<"ShmemID = "<<ShmemId<<endl;
    perror("SpComm_StartUp: shmat SHMEM_MESSAGE failed");
    ErrorStatus = ERROR_SP_COMM;
    return;
  }

  InputArray = new int *[NlocalNodes];
  for(i=0; i<NlocalNodes; ++i){
    InputArray[i] = (int *)(SharedMemoryAddress + i*MaxMessages*sizeof(int));
  }
  OutputBuffer = (char *)(SharedMemoryAddress +
			  NlocalNodes*MaxMessages*sizeof(int));
  LastReadIndex = new int[NlocalNodes];
  LastWriteIndex = new int[NlocalNodes];
  OutgoingMessageBuffer = new CIRCULAR_BUFFER(OutBufferSize,
					      (char *)OutputBuffer);
  MsgInOutCount = 0;

  PartialMsg = new char*[NlocalNodes];
  PartialMsgPtr = new char*[NlocalNodes];
  for (i=0; i < NlocalNodes; ++i){
    PartialMsg[i] = NULL;
    PartialMsgPtr[i] = NULL;
  }

  PartialSendMsg = NULL;
  PartialMessageType = 0;
}



SHMEM_MESSAGE::SHMEM_MESSAGE(const int OutBufferSize,
			     const int InArraySize,
			     const int Nloc,
			     const int LocId,
			     const int shmemId)
{
  AlreadyDeleted = 0;
  OutputBufferSize = OutBufferSize;
  MaxMessages = InArraySize;
  NlocalNodes = Nloc;
  MyLocalNode = LocId;
  char shmSegmentName[8192];
  ErrorStatus = GOOD_SP_COMM;

  Nbytes = OutputBufferSize + NlocalNodes*MaxMessages*sizeof(int);

  sprintf(shmSegmentName, "SPEEDES_SHMEM_SEGMENT_GROUP_%d_NODE_%d_uniqueId_%s",
	  GroupId, LocId, SpParentProcessId);
  if (shmemId == -1){
    ShmemId = SHMGET(IPC_PRIVATE, Nbytes, 0600, shmSegmentName);
    if (ShmemId < 0) {
      perror("SpComm_StartUp: shmget SHMEM_MESSAGE failed.\nRun \"clearipcs\" in the speedes/exe directory if you are not running NT.");
      ErrorStatus = ERROR_SP_COMM;
      return;
    }
  }
  else{
    ShmemId = shmemId;
  }
  SpRedopCreated = 1;
}





/*
  All of the basic services

  Notice that all of the global reductions (min, max, sum) along with
  the sync() and fuzzy barriers use almost identical code.  For this
  reason, SpComm_BarrierSync() is heavily commented but the remaining
  ones are not.  If you can understand SpComm_BarrierSync(), you
  should be able to understand the rest of them.

 */



void SpComm_BarrierSync() {
  int NextToggle; /* This is the new value for the toggle which alternates
		     between 0 and 1 */
  int stage = 0;  // Indicates how far we have advanced in the reduction
  int index;      // used to locate the children at the present stage
  

/*
  Toggle alternates between 1 and 0.  We switch it at the end of each
  global reduction operation (which are guarranteed not to clash).
  */

  NextToggle = (Toggle+1)%2;


/*
      stage 0        stage 1        stage 2       stage 3      stage 4

      node  0 ------>node  0------->node  0------>node  0----->node  0
                --/            --/             /
      node  1--/   ->node  1--/  _->node  1---/
                 -//            /
                / /            /
      node  2--/ /   node  2--/
               -/    /
      node  3-/   --/
                 /
                /
      node  4--/

      At stage 0, everyone fills in their toggle and then advances to the next
      stage.  At the next stage, half of the nodes drop out of the summation.
      Node 0 waits for node 0 and node 1 to fill in their blanks, and
      Node 1 waits for node 2 and node 3 to fill in their blanks, and
      Node 2 waits for node 4 to fill in it's blank (and no other children
                                                     of that node)
      Nodes 3 and 4 now go to stage 3 and wait for node 0 to fill in the blank
      there.

      Once this has happened, they then advance to the next stage with about
      half of the nodes dropping out at this stage as well.  At this stage,
      Node 0 waits at stage 1 for node 0 and node 1 at stage 0, and
      node 1 waits at stage 1 for node 2 (and no other children) at stage 0.
      Node 2 joins nodes 3 and 4 in waiting for node 0 at stage 3.

      Finally almost everyone has dropped out and node 0 waits at stage 3
      for node 0 and node 1 at stage 2.
      Node 1 has joined nodes 2, 3, and 4 in waiting on node 0 at stage 3.

      The extra node 0 at stage 4 is just a way to keep the logic working.
      Nothing is ever written to that value but it does keep node 0 from
      trying to move on to further stages.

      */      

  // Change our own location's toggle since we have "filled" in it's value 
  SpRedop[SpComm_LocalNode].Toggle = NextToggle;

  // While we participate in the next stage
  while (SpComm_LocalNode < StageElements[stage+1]){
    // index is the location of our two children
    index=2*SpComm_LocalNode+StageIndex[stage];
    // Wait on the first child to fill in its Toggle with NextToggle
    while ((THE_REDOP)[index].Toggle == Toggle) NOP;
    // If we have a second child
    if ((2*SpComm_LocalNode+1) < StageElements[stage]){
      // Wait on the second child to fill in its Toggle with NextToggle
      while ((THE_REDOP)[index+1].Toggle == Toggle) NOP;
    }
    
    // Move on to the next stage
    stage++;
    //If this is the last stage and we are node 0
    if (!StageElements[stage+1] && !SpComm_LocalNode){
      // Query the network for a synchronization
      SpComm_Net_Redop();
    }
    
    /* Toggle our location and we can then return to the top of the
       while loop to begin working on the next stage. */
    SpRedop[StageIndex[stage]+SpComm_LocalNode].Toggle=NextToggle;
  } // while
  
  // This is where we wait on Node 0 to fill in the final toggle.
  while ((THE_REDOP)[Nredops-1].Toggle == Toggle) NOP;
  // Update the toggle for the next go around
  Toggle = NextToggle;
  
  /* If There is only one node on our machine, then we don't go
     through any of the above stuff and we simply go out and
     synchronize with the network */
  if ((SpComm_Nlocal == 1) && (SpComm_Nnodes > 1)){
    SpComm_Net_Redop();
  }
  

}

void SpComm_EnterFuzzyBarrier() {
  int NextToggle;
  int index;
  //cout<<"Entering fuzzy barrier"<<endl;
  // If we are the only local node and there are multiple machines 

  if ((SpComm_Nlocal == 1) && (SpComm_Nlocal != SpComm_Nnodes)){
    FuzzyNetworkWait = 1;
    SpComm_Net_Enter_nbRedop();
    return;
  }
  else FuzzyNetworkWait = 0;
  
  NextToggle = (Toggle+1)%2;
  SpRedop[SpComm_LocalNode].Toggle = NextToggle;

/* FuzzyStage is a global variable that keeps track of our stage.  With the
** Fuzzy barrier, we can break out of this routine and need to know what
** Stage we were in when we try to SpComm_ExitFuzzyBarrier().
*/

  FuzzyStage = 0;

  while(SpComm_LocalNode < StageElements[FuzzyStage+1]){
    index = 2*SpComm_LocalNode+StageIndex[FuzzyStage];

    if (SpRedop[index].Toggle==Toggle) break;
    if ((2*SpComm_LocalNode+1)<StageElements[FuzzyStage]){
      if(SpRedop[index+1].Toggle==Toggle) break;
    }

    FuzzyStage++;

    if (!StageElements[FuzzyStage+1] && !SpComm_LocalNode){
      FuzzyNetworkWait = 1;
      SpComm_Net_Enter_nbRedop();
      break;
    }
    SpRedop[StageIndex[FuzzyStage]+SpComm_LocalNode].Toggle=NextToggle;
  }

#ifdef DEBUG_SP_COMM
  cout << "Finished with the Enter Fuzzy Barrier on node "
       << SpComm_GetNodeId() << endl;
#endif

#undef DEBUG_SP_COMM
}


int SpComm_ExitFuzzyBarrier() {
  int index;
  int NextToggle = (Toggle+1)%2;
  int status;

  while (SpComm_LocalNode < StageElements[FuzzyStage+1]){
    index=2*SpComm_LocalNode+StageIndex[FuzzyStage];

    if (SpRedop[index].Toggle==Toggle){
	break;
    }
    if((2*SpComm_LocalNode+1) < StageElements[FuzzyStage]){
      if (SpRedop[index+1].Toggle == Toggle){
	break;
      }
    }

    FuzzyStage++;

    if (!StageElements[FuzzyStage+1] && !SpComm_LocalNode){
      FuzzyNetworkWait = 1;
      status = SpComm_Net_nbRedop();
      if (status == 1){
	SpRedop[StageIndex[FuzzyStage]].Toggle=NextToggle;
      }
      break;
    }
    
    SpRedop[StageIndex[FuzzyStage]+SpComm_LocalNode].Toggle = NextToggle;

  }

/* We have finished the fuzzy barrier on our machine and now need to check
   with the network.  Notice that FuzzyNetworkWait will be set to 1 only if
   we are local node 0.  This means that everyone else will just fall through
   and check the final node int SpRedop.
   */

  if (FuzzyNetworkWait){
    if (SpRedop[StageIndex[FuzzyStage]+SpComm_LocalNode].Toggle!=NextToggle){
      status=SpComm_Net_nbRedop();
      if(status == 1){
	SpRedop[StageIndex[FuzzyStage]+SpComm_LocalNode].Toggle=NextToggle;
      }
      else{
	return 0;
      }
    }
  }
  if (SpRedop[Nredops-1].Toggle==Toggle)
  {
#ifdef DEBUG_SP_COMM
  cout << "Exiting SpComm_ExitFuzzyBarrier() with rc=0 on node " << Node << endl;
#endif
    return 0;
  }
  Toggle = NextToggle;// (Toggle+1)%2;
#ifdef DEBUG_SP_COMM
  ++foo;
  cout << "Exiting SpComm_ExitFuzzyBarrier() with rc=1 on node "
       <<SpComm_GetNodeId() <<" count number "<<foo<< endl;
#endif
  //cout<<"Exiting fuzzy barrier"<<endl;
  return 1;
}

int SpComm_GlobalSum(int value) {
  int NextToggle;
  int stage = 0;
  int index;
  int temp = value;

  NextToggle = (Toggle+1)%2;
  SpRedop[SpComm_LocalNode].Ivalue = value;
  SpRedop[SpComm_LocalNode].Toggle = NextToggle;

  while (SpComm_LocalNode < StageElements[stage+1]){
    index=2*SpComm_LocalNode+StageIndex[stage];
    while ((THE_REDOP)[index].Toggle == Toggle)NOP;
    temp = SpRedop[index].Ivalue;
    if ((2*SpComm_LocalNode+1) < StageElements[stage]){
      while ((THE_REDOP)[index+1].Toggle == Toggle)NOP;
      temp+=SpRedop[index+1].Ivalue;
    }

    stage++;
    if (!StageElements[stage+1] && !SpComm_LocalNode){
      SpComm_Net_Sum(&temp, &NetDummyDouble, &NetDummySTime, 0);
    }
    while((THE_REDOP)[StageIndex[stage]+SpComm_LocalNode].Toggle!=Toggle){
      NOP;
    }
    SpRedop[StageIndex[stage]+SpComm_LocalNode].Ivalue=temp;
    SpRedop[StageIndex[stage]+SpComm_LocalNode].Toggle=NextToggle;
  }

  if ((SpComm_Nlocal == 1) && (SpComm_Nnodes > 1)){
    SpComm_Net_Sum(&temp, &NetDummyDouble, &NetDummySTime, 0);
    SpRedop[Nredops-1].Ivalue = temp;
  }

  while ((THE_REDOP)[Nredops-1].Toggle == Toggle)NOP;
  value=SpRedop[Nredops-1].Ivalue;
  Toggle = NextToggle;
#ifdef DEBUG_SP_COMM
  cout << "Finished with the SpComm_GlobalSum(int) on node " << Node << endl;
#endif

  return value;
}


double SpComm_GlobalSum(double value) {
  int NextToggle;
  int stage = 0;
  int index;
  double temp = value;

  NextToggle = (Toggle+1)%2;
  SpRedop[SpComm_LocalNode].Dvalue = value;
  SpRedop[SpComm_LocalNode].Toggle = NextToggle;

  while (SpComm_LocalNode < StageElements[stage+1]){
    index=2*SpComm_LocalNode+StageIndex[stage];
    while ((THE_REDOP)[index].Toggle == Toggle)NOP;
    temp = SpRedop[index].Dvalue;
    if ((2*SpComm_LocalNode+1) < StageElements[stage]){
      while ((THE_REDOP)[index+1].Toggle == Toggle)NOP;
      temp+=SpRedop[index+1].Dvalue;
    }

    stage++;
    if (!StageElements[stage+1] && !SpComm_LocalNode){
      SpComm_Net_Sum(&NetDummyInt, &temp, &NetDummySTime, 0);
    }
    while((THE_REDOP)[StageIndex[stage]+SpComm_LocalNode].Toggle!=Toggle)
      NOP;
    SpRedop[StageIndex[stage]+SpComm_LocalNode].Dvalue=temp;
    SpRedop[StageIndex[stage]+SpComm_LocalNode].Toggle=NextToggle;
  }

  if ((SpComm_Nlocal == 1) && (SpComm_Nnodes > 1)){
    SpComm_Net_Sum(&NetDummyInt, &temp, &NetDummySTime, 0);
    SpRedop[Nredops-1].Dvalue = temp;
  }

  while ((THE_REDOP)[Nredops-1].Toggle == Toggle)NOP;
  value=SpRedop[Nredops-1].Dvalue;
  Toggle = NextToggle;
#ifdef DEBUG_SP_COMM
  cout <<"Finished with the SpComm_GlobalSum(double) on node " << Node << endl;
#endif

  return value;
}



double SpComm_GlobalMin(double value) {
  int NextToggle;
  int stage = 0;
  int index;
  double temp = value;

  NextToggle = (Toggle+1)%2;
  SpRedop[SpComm_LocalNode].Dvalue = value;
  SpRedop[SpComm_LocalNode].Toggle = NextToggle;

  while (SpComm_LocalNode < StageElements[stage+1]){
    index=2*SpComm_LocalNode+StageIndex[stage];
    while ((THE_REDOP)[index].Toggle == Toggle)NOP;
    temp = SpRedop[index].Dvalue;
    if ((2*SpComm_LocalNode+1) < StageElements[stage]){
      while ((THE_REDOP)[index+1].Toggle == Toggle)NOP;
      temp=MIN(SpRedop[index+1].Dvalue, temp);
    }

    stage++;
    if (!StageElements[stage+1] && !SpComm_LocalNode){
      SpComm_Net_Min(&NetDummyInt, &temp, &NetDummySTime, 0);
    }
    while((THE_REDOP)[StageIndex[stage]+SpComm_LocalNode].Toggle!=Toggle)
      NOP;
    SpRedop[StageIndex[stage]+SpComm_LocalNode].Dvalue=temp;
    SpRedop[StageIndex[stage]+SpComm_LocalNode].Toggle=NextToggle;
  }

  if ((SpComm_Nlocal == 1) && (SpComm_Nnodes > 1)){
    SpComm_Net_Min(&NetDummyInt, &temp, &NetDummySTime, 0);
    SpRedop[Nredops-1].Dvalue = temp;
  }

  while ((THE_REDOP)[Nredops-1].Toggle == Toggle)NOP;
  value=SpRedop[Nredops-1].Dvalue;
  Toggle = NextToggle;
#ifdef DEBUG_SP_COMM
  cout <<"Finished with the SpComm_GlobalMin(double) on node " << Node << endl;
#endif

  return value;
}

int SpComm_GlobalMin(int value) {
  int NextToggle;
  int stage = 0;
  int index;
  int temp = value;

  NextToggle = (Toggle+1)%2;
  SpRedop[SpComm_LocalNode].Ivalue = value;
  SpRedop[SpComm_LocalNode].Toggle = NextToggle;

  while (SpComm_LocalNode < StageElements[stage+1]){
    index=2*SpComm_LocalNode+StageIndex[stage];
    while ((THE_REDOP)[index].Toggle == Toggle)NOP;
    temp = SpRedop[index].Ivalue;
    if ((2*SpComm_LocalNode+1) < StageElements[stage]){
      while ((THE_REDOP)[index+1].Toggle == Toggle)NOP;
      temp=MIN(SpRedop[index+1].Ivalue, temp);
    }

    stage++;
    if (!StageElements[stage+1] && !SpComm_LocalNode){
      SpComm_Net_Min(&temp, &NetDummyDouble, &NetDummySTime, 0);
    }
    while((THE_REDOP)[StageIndex[stage]+SpComm_LocalNode].Toggle!=Toggle)
      NOP;
    SpRedop[StageIndex[stage]+SpComm_LocalNode].Ivalue=temp;
    SpRedop[StageIndex[stage]+SpComm_LocalNode].Toggle=NextToggle;
  }

  if ((SpComm_Nlocal == 1) && (SpComm_Nnodes > 1)){
    SpComm_Net_Min(&temp, &NetDummyDouble, &NetDummySTime, 0);
    SpRedop[Nredops-1].Ivalue = temp;
  }

  while ((THE_REDOP)[Nredops-1].Toggle == Toggle)NOP;
  value=SpRedop[Nredops-1].Ivalue;
  Toggle = NextToggle;
#ifdef DEBUG_SP_COMM
  cout << "Finished with the SpComm_GlobalMin(int) on node " << Node << endl;
#endif

  return value;
}


int SpComm_GlobalMax(int value) {
  int NextToggle;
  int stage = 0;
  int index;
  int temp = value;

  NextToggle = (Toggle+1)%2;
  SpRedop[SpComm_LocalNode].Ivalue = value;
  SpRedop[SpComm_LocalNode].Toggle = NextToggle;

  while (SpComm_LocalNode < StageElements[stage+1]){
    index=2*SpComm_LocalNode+StageIndex[stage];
    while ((THE_REDOP)[index].Toggle == Toggle)NOP;
    temp = SpRedop[index].Ivalue;
    if ((2*SpComm_LocalNode+1) < StageElements[stage]){
      while ((THE_REDOP)[index+1].Toggle == Toggle)NOP;
      temp=MAX(SpRedop[index+1].Ivalue, temp);
    }

    stage++;
    if (!StageElements[stage+1] && !SpComm_LocalNode){
      SpComm_Net_Max(&temp, &NetDummyDouble, &NetDummySTime, 0);
    }
    while((THE_REDOP)[StageIndex[stage]+SpComm_LocalNode].Toggle!=Toggle)
      NOP;
    SpRedop[StageIndex[stage]+SpComm_LocalNode].Ivalue=temp;
    SpRedop[StageIndex[stage]+SpComm_LocalNode].Toggle=NextToggle;
  }

  if ((SpComm_Nlocal == 1) && (SpComm_Nnodes > 1)){
    SpComm_Net_Max(&temp, &NetDummyDouble, &NetDummySTime, 0);
    SpRedop[Nredops-1].Ivalue = temp;
  }

  while ((THE_REDOP)[Nredops-1].Toggle == Toggle)NOP;
  value=SpRedop[Nredops-1].Ivalue;
  Toggle = NextToggle;
#ifdef DEBUG_SP_COMM
  cout << "Finished with the SpComm_GlobalMax(int) on node " << Node << endl;
#endif

  return value;
}



SpSimTime &SpComm_GlobalMin(SpSimTime &SimTime) {
  int NextToggle;
  int stage = 0;
  int index;
  SpSimTime temp = SimTime;

  NextToggle = (Toggle+1)%2;
  SpRedop[SpComm_LocalNode].SimTime = SimTime;
  SpRedop[SpComm_LocalNode].Toggle = NextToggle;

  while (SpComm_LocalNode < StageElements[stage+1]){
    index=2*SpComm_LocalNode+StageIndex[stage];
    while ((THE_REDOP)[index].Toggle == Toggle)NOP;
    temp = SpRedop[index].SimTime;
    if ((2*SpComm_LocalNode+1) < StageElements[stage]){
      while ((THE_REDOP)[index+1].Toggle == Toggle)NOP;
      if (SpRedop[index+1].SimTime < temp) temp = SpRedop[index+1].SimTime;
    }

    stage++;
    if (!StageElements[stage+1] && !SpComm_LocalNode){
      SpComm_Net_Min(&NetDummyInt, &NetDummyDouble, &temp, 0);
    }
    while((THE_REDOP)[StageIndex[stage]+SpComm_LocalNode].Toggle!=Toggle)
      NOP;
    SpRedop[StageIndex[stage]+SpComm_LocalNode].SimTime=temp;
    SpRedop[StageIndex[stage]+SpComm_LocalNode].Toggle=NextToggle;
  }

  if ((SpComm_Nlocal == 1) && (SpComm_Nnodes > 1)){
    SpComm_Net_Min(&NetDummyInt, &NetDummyDouble, &temp, 0);
    SpRedop[Nredops-1].SimTime = temp;
  }

  while ((THE_REDOP)[Nredops-1].Toggle == Toggle)NOP;
  SimTime=SpRedop[Nredops-1].SimTime;
  Toggle = NextToggle;
#ifdef DEBUG_SP_COMM
  cout <<"Finished with the SpComm_GlobalMin(SpSimTime on node " << Node << endl;
#endif

  return SimTime;
}

double SpComm_GlobalMax(double value) {
  int NextToggle;
  int stage = 0;
  int index;
  double temp = value;

  NextToggle = (Toggle+1)%2;
  SpRedop[SpComm_LocalNode].Dvalue = value;
  SpRedop[SpComm_LocalNode].Toggle = NextToggle;

  while (SpComm_LocalNode < StageElements[stage+1]){
    index=2*SpComm_LocalNode+StageIndex[stage];
    while ((THE_REDOP)[index].Toggle == Toggle)NOP;
    temp = SpRedop[index].Dvalue;
    if ((2*SpComm_LocalNode+1) < StageElements[stage]){
      while ((THE_REDOP)[index+1].Toggle == Toggle)NOP;
      temp=MAX(SpRedop[index+1].Dvalue, temp);
    }

    stage++;
    if (!StageElements[stage+1] && !SpComm_LocalNode){
      SpComm_Net_Max(&NetDummyInt, &temp, &NetDummySTime, 0);
    }
    while((THE_REDOP)[StageIndex[stage]+SpComm_LocalNode].Toggle!=Toggle)
      NOP;
    SpRedop[StageIndex[stage]+SpComm_LocalNode].Dvalue=temp;
    SpRedop[StageIndex[stage]+SpComm_LocalNode].Toggle=NextToggle;
  }

  if ((SpComm_Nlocal == 1) && (SpComm_Nnodes > 1)){
    SpComm_Net_Max(&NetDummyInt, &temp, &NetDummySTime, 0);
    SpRedop[Nredops-1].Dvalue = temp;
  }

  while ((THE_REDOP)[Nredops-1].Toggle == Toggle)NOP;
  value=SpRedop[Nredops-1].Dvalue;
  Toggle = NextToggle;
#ifdef DEBUG_SP_COMM
  cout <<"Finished with the SpComm_GlobalMax(double) on node " << Node << endl;
#endif

  return value;
}

SpSimTime &SpComm_GlobalMax(SpSimTime &SimTime) {
  int NextToggle;
  int stage = 0;
  int index;
  SpSimTime temp = SimTime;

  NextToggle = (Toggle+1)%2;
  SpRedop[SpComm_LocalNode].SimTime = SimTime;
  SpRedop[SpComm_LocalNode].Toggle = NextToggle;

  while (SpComm_LocalNode < StageElements[stage+1]){
    index=2*SpComm_LocalNode+StageIndex[stage];
    while ((THE_REDOP)[index].Toggle == Toggle)NOP;
    temp = SpRedop[index].SimTime;
    if ((2*SpComm_LocalNode+1) < StageElements[stage]){
      while ((THE_REDOP)[index+1].Toggle == Toggle)NOP;
      if (SpRedop[index+1].SimTime > temp) temp = SpRedop[index+1].SimTime;
    }

    stage++;
    if (!StageElements[stage+1] && !SpComm_LocalNode){
      SpComm_Net_Max(&NetDummyInt, &NetDummyDouble, &temp, 0);
    }
    while((THE_REDOP)[StageIndex[stage]+SpComm_LocalNode].Toggle!=Toggle)
      NOP;
    SpRedop[StageIndex[stage]+SpComm_LocalNode].SimTime=temp;
    SpRedop[StageIndex[stage]+SpComm_LocalNode].Toggle=NextToggle;
  }

  if ((SpComm_Nlocal == 1) && (SpComm_Nnodes > 1)){
    SpComm_Net_Max(&NetDummyInt, &NetDummyDouble, &temp, 0);
    SpRedop[Nredops-1].SimTime = temp;
    SpRedop[Nredops-1].Toggle = NextToggle;
  }

  while ((THE_REDOP)[Nredops-1].Toggle == Toggle)NOP;
  SimTime=SpRedop[Nredops-1].SimTime;
  Toggle = NextToggle;
#ifdef DEBUG_SP_COMM
  cout <<"Finished with the SpComm_GlobalMax(SpSimTime) on node "<< Node << endl;
#endif

  return SimTime;
}






/*
  All of these calls do not go out to the network.  SpComm_*() should determine
  if the call is going to a node on this computer or if it is going off of the
  computer.  That way, all of the Shmem*() calls do not need to worry about
  any odd routing.
  */



int SHMEM_MESSAGE::TrySend(const int dest, const int type, const int nbytes,
			   const void *buff)
{
  int Index;
  int RC = MESSAGE_SENT;

  Index = OutgoingMessageBuffer->Write(nbytes, (char *)buff, type);
  if (Index != -1){ //If we were able to write the message
    if (Announce(dest, &Index)){ //If we are unable to announce it
      AnnounceQueue.AddToTail((char *)(&Index), sizeof(int),
			      SPCOMM_MESSAGE_ANNOUNCE, dest);
      RC = MESSAGE_POSTPONED;
    }
  }
  else{
    RC = MESSAGE_FAILED; //Complete send failure
  }
  return RC;
}






int SHMEM_MESSAGE::TrySend(const int type, const int nbytes, const void * buff)
{
  /* In a broadcast, we first write to the buffer and then look to see
     to whom we can announce the message.  If a person cannot
     presently receive the message, we queue up a message to tell them
     about the message */

  int Pos, RC = SHMEM_MULTISEND_ALL, Index, i = 0;
  
  if (SpComm_Nlocal > 1){ /* Only do a local broadcast if there is
			     more than one node on this box */
    Index=OutgoingMessageBuffer->Write(nbytes, (char *)buff, type,
				       SpComm_Nlocal - 1);
    if (Index != -1){
      for (int dest=0; dest < SpComm_Nlocal; ++dest){
	if (dest != MyLocalNode){ // Don't broadcast to self
	  Pos = Index + i*CBHeaderSize; //Get the index of the message
	  ++i;
	  if (Announce(dest, &Pos)){
	    /* Queue up an announcement that the message is waiting
	       for them if unable to announce it */
	    RC = SHMEM_MULTISEND_PARTIAL;
	    AnnounceQueue.AddToTail((char *)(&Pos), sizeof(int),
				    SPCOMM_MESSAGE_ANNOUNCE, dest);
	  }
	}
      }
    }
    else{
      RC = SHMEM_MULTISEND_NONE;
    }
  }
  return RC;
}

int SHMEM_MESSAGE::TrySend(SpDestination * DEST, const int type,
			    const int nbytes, const void * buff)
{
  int Pos, dest, RC = SHMEM_MULTISEND_ALL, i = 1, Index;
  /* In a multicast, we first write to the buffer and then look to see
     to whom we can announce the message.  We tell them the message is
     waiting if they can receive it.  If not, we put a message on the
     temp send queue to remind us to let them know that it is waiting
     for them.*/

  dest = DEST->GetFirstNode();
  int Nnodes;
  for (Nnodes = 1; ((dest = DEST->GetNextNode(dest)) != -1); ++Nnodes);
  Index= OutgoingMessageBuffer->Write(nbytes, (char *)buff, type, Nnodes);
  if (Index != -1){
    RC = SHMEM_MULTISEND_PARTIAL;
    dest = DEST->GetFirstNode();
    for (i=0; dest != -1; ++i){
      Pos = Index + i*CBHeaderSize;
      if (dest == MyLocalNode){
	InQueue[type].AddToTail((char *)buff, nbytes, type);
	OutgoingMessageBuffer->SetReadFlag(Pos);
      }
      else if (Announce(dest, &Pos)){
	AnnounceQueue.AddToTail((char *)(&Pos), sizeof(int),
				 SPCOMM_MESSAGE_ANNOUNCE, dest);
      }
      dest = DEST->GetNextNode(dest);
    }
  }
  else RC = SHMEM_MULTISEND_NONE;
  return RC;
}

int SHMEM_MESSAGE::TrySend(int * DEST, const int type,
			   const int nbytes, const void * buff)
{
  /* In a multi/broad cast, we first write to the buffer and then look
     to see to whom we can announce the message.  If a person cannot
     receive the message, we queue up a message to tell that person
     that he/she has a message waiting.  DEST is an array of
     destinations.  The first element is the number of destinations
     and the the destinations follow.  */

  int Pos, *dest = DEST+1, RC = 0, Nsend = *DEST;

  int Index =OutgoingMessageBuffer->Write(nbytes, (char *)buff, type, *DEST);

  if (Index != -1){
    RC = SHMEM_MULTISEND_PARTIAL;
    for (int i=0; i < Nsend; ++i){
      Pos = Index + i*CBHeaderSize;
      if (*dest == MyLocalNode){ //If send to self
	InQueue[type].AddToTail((char *)buff, nbytes, type);
	OutgoingMessageBuffer->SetReadFlag(Pos);
      }
      else if (Announce(*dest, &Pos)){
	AnnounceQueue.AddToTail((char *)(&Pos), sizeof(int),
				 SPCOMM_MESSAGE_ANNOUNCE, *dest);
      }
      ++dest;
    }
  }
  else{
    RC = SHMEM_MULTISEND_NONE;
  }
  return RC;
}

void SHMEM_MESSAGE::EnQueuePartialMsg(char * Msg,
				      int Dest,
				      int MsgLen,
				      int Type,
				      int MaxMsgLen)

{
  MaxMsgLen  -= 2*sizeof(CIRC_BUFF_HEADER);

  char * TempPtr = new char[MsgLen];
  int * PartialMessageBegin = new int[2];
  int * PartialMessageEnd = new int[2];
  int BytesToSend = MsgLen;

  memcpy(TempPtr, Msg, MsgLen);
  PartialMessageBegin[0] = MsgLen;
  PartialMessageEnd[1] = 0;
  PartialMessageEnd[0] = MsgLen;
  PartialMessageEnd[1] = Type;
  
  TempSendQueue.AddToTail((char *)PartialMessageBegin,
			  2*sizeof(int),
			  SPCOMM_MESSAGE_PARTIAL,
			  Dest, 0); //The '0' means don't allocate memory

  while (BytesToSend > MaxMsgLen){
    TempSendQueue.AddToTail(TempPtr,
			    MaxMsgLen,
			    SPCOMM_MESSAGE_PARTIAL,
			    Dest, 0);
    TempPtr += MaxMsgLen;
    BytesToSend -= MaxMsgLen;
  }

  TempSendQueue.AddToTail(TempPtr,
			  BytesToSend,
			  SPCOMM_MESSAGE_PARTIAL,
			  Dest, 0);

  TempSendQueue.AddToTail((char *)PartialMessageEnd,
			  2*sizeof(int),
			  SPCOMM_MESSAGE_PARTIAL_END,
			  Dest, 0); //The '0' means don't allocate memory
  FlushSendQueue();
}


void SHMEM_MESSAGE::EnQueuePartialMsg(char * Msg,
				      int MsgLen,
				      int Type,
				      int MaxMsgLen)

{
  static int HeaderSize = 2*NlocalNodes*sizeof(CIRC_BUFF_HEADER);
  MaxMsgLen  -= HeaderSize;

  char * TempPtr = new char[MsgLen];
  int * PartialMessageBegin = new int[2];
  int * PartialMessageEnd = new int[2];
  int NumPieces = MsgLen/MaxMsgLen;

  memcpy(TempPtr, Msg, MsgLen);
  PartialMessageBegin[0] = MsgLen;
  PartialMessageEnd[1] = SPCOMM_MESSAGE_BROADCAST;
  PartialMessageEnd[0] = MsgLen;
  PartialMessageEnd[1] = Type;
  
  TempSendQueue.AddToTail((char *)PartialMessageBegin,
			  2*sizeof(int),
			  SPCOMM_MESSAGE_PARTIAL_BROADCAST_BEGIN,
			  -1, 0); //The '0' means don't allocate memory

  for (int i=0; i < NumPieces; ++i){
    TempSendQueue.AddToTail(TempPtr,
			    MaxMsgLen,
			    SPCOMM_MESSAGE_PARTIAL,
			    -1, 0);
    TempPtr += MaxMsgLen;
  }

  TempSendQueue.AddToTail(TempPtr,
			  MsgLen%MaxMsgLen,
			  SPCOMM_MESSAGE_PARTIAL,
			  -1, 0);

  TempSendQueue.AddToTail((char *)PartialMessageEnd,
			  2*sizeof(int),
			  SPCOMM_MESSAGE_PARTIAL_END,
			  -1, 0); //The '0' means don't allocate memory
  FlushSendQueue();
}


void SHMEM_MESSAGE::EnQueuePartialMsg(char * Msg,
				      SpDestination *Dest,
				      int MsgLen,
				      int Type,
				      int MaxMsgLen)

{
  static int HeaderSize = 2*NlocalNodes*sizeof(CIRC_BUFF_HEADER);
  MaxMsgLen -= HeaderSize;

  char * TempPtr = new char[MsgLen];
  int * PartialMessageBegin = new int[2];
  int * PartialMessageEnd = new int[2];
  int NumPieces = MsgLen/MaxMsgLen;
  int nbytes;

  memcpy(TempPtr, Msg, MsgLen);
  PartialMessageBegin[0] = MsgLen;
  PartialMessageEnd[1] = SPCOMM_MESSAGE_MULTICAST;
  PartialMessageEnd[0] = MsgLen;
  PartialMessageEnd[1] = Type;
  
  char * destinationBuffer = Dest->PackBuffer(nbytes);
  TempSendQueue.AddToTail(destinationBuffer,
			  nbytes,
			  SPCOMM_MESSAGE_PARTIAL_DESTINATION,
			  -1, 0);


  TempSendQueue.AddToTail((char *)PartialMessageBegin,
			  2*sizeof(int),
			  SPCOMM_MESSAGE_PARTIAL_MULTICAST_BEGIN,
			  -1, 0); //The '0' means don't allocate memory


  for (int i=0; i < NumPieces; ++i){
    TempSendQueue.AddToTail(TempPtr,
			    MaxMsgLen,
			    SPCOMM_MESSAGE_PARTIAL,
			    -1, 0);
    TempPtr += MaxMsgLen;
  }

  TempSendQueue.AddToTail(TempPtr,
			  MsgLen%MaxMsgLen,
			  SPCOMM_MESSAGE_PARTIAL,
			  -1, 0);

  TempSendQueue.AddToTail((char *)PartialMessageEnd,
			  2*sizeof(int),
			  SPCOMM_MESSAGE_PARTIAL_END,
			  -1, 0); //The '0' means don't allocate memory
  FlushSendQueue();
}





void SHMEM_MESSAGE::Receive()
{
  char * Msg;
  int Type, TempNbytes, Dest, Index;

  FlushSendQueue(); //Attempt to flush the temp send queue of any messages.
  for (int Sender=0; Sender < SpComm_Nlocal; ++Sender){
    if (Sender != MyLocalNode){
      int NextReadIndex = (LastReadIndex[Sender]+1)%MaxMessages;
      while ((Index = InputArray[Sender][NextReadIndex]) != MSG_READ){
	//Index = InputArray[Sender][NextReadIndex];
	Msg = (SpComm_ShmemMessage[Sender]->OutgoingMessageBuffer)->
	  GetBuff(Index,Type,TempNbytes);
	if ((Type != SPCOMM_MESSAGE_PARTIAL) && 
	    (Type != SPCOMM_MESSAGE_PARTIAL_END)){
	  InQueue[Type].AddToTail(Msg, TempNbytes, Type);
	}
	else{
	  if (PartialMsg[Sender] == NULL){ //Start of a new part msg
	    int msglen = *((int *)Msg);    //payload is total message length
	    PartialMsg[Sender] = new char[msglen];
	    PartialMsgPtr[Sender] = PartialMsg[Sender];
	    PartialMsgByteCount[Sender] = 0;
	  }
	  else if (Type != SPCOMM_MESSAGE_PARTIAL_END){ //Continue part msg
	    memcpy(PartialMsgPtr[Sender], Msg, TempNbytes);
	    PartialMsgPtr[Sender] += TempNbytes;
	    PartialMsgByteCount[Sender] += TempNbytes;
	  }
	  else{ //end of partial message.  Queue it up and set partmsg to NULL
	    int PartialLen = *((int *)Msg);
	    int PartialType = ((int *)Msg)[1];
	    InQueue[PartialType].AddToTail(PartialMsg[Sender],
					   PartialLen,
					   PartialType,
					   -1, //Destination is ignored
					   0); //Don't allocate buffer space
	    PartialMsg[Sender] = NULL;
	  }
	}
	(SpComm_ShmemMessage[Sender]->OutgoingMessageBuffer)->
	  SetReadFlag(Index);
	InputArray[Sender][NextReadIndex] = MSG_READ; //Mark index area read
	LastReadIndex[Sender] = NextReadIndex;   //Move the LastReadIndex along
	NextReadIndex = (LastReadIndex[Sender]+1)%MaxMessages;

      }
    }
  }



  if (MyLocalNode == 0){
    void * buff = NULL;
    //Process special messages that only node 0 has to deal with.

    while (InQueue[SPCOMM_MESSAGE_REROUTE].Count()){
      cout<<"Should not be any reroute messages "<<endl;
      /* Reroute messages have been slightly altered.  The buffer is
	 made larger by 8 bytes and then the true destination is then
	 stored in the leading 8 bytes.*/
      buff = InQueue[SPCOMM_MESSAGE_REROUTE].Remove(TempNbytes, Type, Dest);
      int * TrueDest = (int *)buff;
      char * TrueBuff = ((char *)buff)+4*sizeof(int);
      SpComm_Net_Send(*TrueDest, TrueDest[1], TempNbytes-4*sizeof(int),
		      (void *)TrueBuff);
      delete [] buff;
    }
  }
}

static inline int Allign8Bytes(int val) {return val + (8 - val%8);}

int SHMEM_MESSAGE::FlushSendQueue(){
  /* Attempts to send any messages in the TempSendQueue.  Returns 0
     if all messages have been flushed and 1 if not */

  int dest, type, nbytes/*, RC*/;
  char * buff;
  int * Dest;
  static int TotalPartialBytesSent=0;
  
  while (AnnounceQueue.Count()){
    buff = AnnounceQueue.NextValue(nbytes, type, dest);
    if (Announce(dest, ((int *) buff))){
      return 1;
    }
    delete [] AnnounceQueue.Remove(nbytes, type, dest);
  }

  /* Notice we return if the announce queue is not flushed.  This
     means one might think we don't have to have a special check for
     that.  However, we may add messages to the AnnounceQueue though a
     broadcast or multicast attempt.  */

  while (TempSendQueue.Count() && !AnnounceQueue.Count()){
    buff = TempSendQueue.NextValue(nbytes, type, dest);
    switch (type){

      /* In the case of broadcast or multisend messages, the type is
	 stored in dest.  I also need to remove the buffer from the
	 front of the queue rather than just examining it.  This is
	 because TrySend() may be able to partially send the message
	 and, in that case, I would get several announce messages in
	 front of the true broadcast/multicast messaage*/
#define ActualType dest
    case SPCOMM_MESSAGE_BROADCAST:
      if (TrySend(ActualType, nbytes, buff) == SHMEM_MULTISEND_NONE){
	return 1;
      }
      delete [] TempSendQueue.Remove(nbytes, type, dest);
      break;
    case SPCOMM_MESSAGE_MULTICAST:
      /* Failed multi-cast where I was unable to get buffer space so
	 this message is a combination of the destinations and the
	 message.  The first part contains the # of destinations and
	 the destination while the second part contains the real
	 message.  The type is stored in the unused parameter dest. */
      {
	Dest = (int *)buff;
	char * RealBuff = buff + sizeof(int)*Allign8Bytes((*Dest + 1));
	if (TrySend(Dest, ActualType, nbytes, RealBuff)
	    == SHMEM_MULTISEND_NONE){
	  return 1;
	}
	delete [] TempSendQueue.Remove(nbytes, type, dest);
      }
    break;
    case SPCOMM_MESSAGE_PARTIAL_DESTINATION:
      PartialSendDestination.UnpackBuffer(buff);
      delete [] TempSendQueue.Remove(nbytes, type, dest);
      buff = TempSendQueue.NextValue(nbytes, type, dest);//Want to fall through
    case SPCOMM_MESSAGE_PARTIAL_MULTICAST_BEGIN:
      if (TrySend(&PartialSendDestination,SPCOMM_MESSAGE_PARTIAL,nbytes,buff)
	  == SHMEM_MULTISEND_NONE){
	return 1;
      }
      PartialMessageType = SPCOMM_MESSAGE_MULTICAST;
      delete [] TempSendQueue.Remove(nbytes, type, dest);
      PartialSendMsg = TempSendQueue.NextValue(nbytes, type, dest);
      break;
    case SPCOMM_MESSAGE_PARTIAL_BROADCAST_BEGIN:
      if ((TrySend(SPCOMM_MESSAGE_PARTIAL, nbytes, buff))
	  == SHMEM_MULTISEND_NONE){
	return 1;
      }
      PartialMessageType = SPCOMM_MESSAGE_BROADCAST;
      delete [] TempSendQueue.Remove(nbytes, type, dest);
      PartialSendMsg = TempSendQueue.NextValue(nbytes, type, dest);
      break;
    case SPCOMM_MESSAGE_PARTIAL_END:
    case SPCOMM_MESSAGE_PARTIAL:
      switch (PartialMessageType)
	{
	case SPCOMM_MESSAGE_BROADCAST:
	  if (TrySend(type, nbytes, buff) == SHMEM_MULTISEND_NONE){
	    return 1;
	  }
	  TotalPartialBytesSent+= nbytes;
	  break;
	case SPCOMM_MESSAGE_MULTICAST:
	  if (TrySend(&PartialSendDestination,
		      type, nbytes, buff) == SHMEM_MULTISEND_NONE){
	    return 1;
	  }
	  TotalPartialBytesSent+= nbytes;
	  break;
#undef ActualType
	case 0:
	default:
	  if ((/*RC = */TrySend(dest, type, nbytes, buff)) == MESSAGE_FAILED){
	    return 1;
	  }
	  TotalPartialBytesSent+= nbytes;
	  break;
	} //switch (SendingPartialMessage)
      if (!PartialSendMsg){
	TotalPartialBytesSent = 0;
	PartialMessageType = 0; // Signifies normal message
	delete [] TempSendQueue.Remove(nbytes, type, dest);
	PartialSendMsg = TempSendQueue.NextValue(nbytes, type, dest);
      }
      else if (type == SPCOMM_MESSAGE_PARTIAL_END){ // Message has been
	                                            // completely sent
	delete [] TempSendQueue.Remove(nbytes, type, dest);
	delete [] PartialSendMsg;
	PartialSendMsg = NULL;
	PartialMessageType = 0;
	TotalPartialBytesSent = 0;
      }
      else{
	TempSendQueue.Remove(nbytes, type, dest);
      }
      break;
      
    default:
      if ((/*RC = */TrySend(dest, type, nbytes, buff)) == MESSAGE_FAILED){
	return 1;
      }
      delete [] TempSendQueue.Remove(nbytes, type, dest);
      break;
    }
  }
  return (TempSendQueue.Count()+AnnounceQueue.Count());
}
  

int ShmemFlushSendQueue()
{
  return SpComm_ShmemMessage[SpComm_LocalNode]->FlushSendQueue();
}
