#include "SpCommShMemCalls.H"
#include "SpStdio.H"
#include "SpStdlib.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpString.H"
#include "SpErrno.H"

#if defined (__CYGWIN32__) || defined (__SPEEDES_NT__)

class SharedMemoryHandles{
 public:
  HANDLE h;
  HANDLE ProcessHandle;
  char * ptr;
  struct shmid_ds Shmid_DS;
  char * shmSegmentName;
};


SharedMemoryHandles ShmemHandles[1024];
int totShmemHandles = 0;
int totAllocatedHandles = 0;
int totFreeHandles = 0;

HANDLE ParentProcessHandle;

#else

#include <unistd.h>

#endif

int SHMGET(key_t key, int size, int shmflg, char *
#if defined (__CYGWIN32__) || defined (__SPEEDES_NT__)
	   shmSegName
#endif
	   )
{
#if defined (__CYGWIN32__) || defined (__SPEEDES_NT__)

  HANDLE retVal;

  //cout<<"Doing a shmget on name "<<shmSegName<<" with size "<<size<<endl;

  retVal = CreateFileMapping( (HANDLE) 0xFFFFFFFF, NULL, PAGE_READWRITE,
			      0, size, shmSegName);
  
  if (retVal != NULL) {
    if (GetLastError() == ERROR_ALREADY_EXISTS){
      //cout<<"Opened preexisting file mapping"<<endl;
    }
    else{
      //cout<<"Created file mapping"<<endl;
    }
  }
  else{
    cout<<"Unable to create file mapping"<<endl;
    exit(-1);
  }
  ShmemHandles[totShmemHandles].h = retVal;
  ShmemHandles[totShmemHandles].shmSegmentName = strdup(shmSegName);
  
  //cout<<"Handle number is "<<totShmemHandles<<endl;
  ++totShmemHandles;
  return (totShmemHandles-1);
#else
  return shmget(key, size, shmflg);
#endif
}


char * SHMAT(int handle, char * shmaddr, int shmflg)
{
#if defined (__CYGWIN32__) || defined (__SPEEDES_NT__)

  HANDLE ChildMemoryHandle;

  void * memPtr = (void*)MapViewOfFile(ShmemHandles[handle].h,
				       FILE_MAP_ALL_ACCESS, 0, 0, 0);
  if (memPtr == NULL){
    cout<<"Unable to map view of file for handle "<<handle<<endl;
    exit(-1);
  }

  //cout<<"Mapped view of file"<<endl;

  return (char*)memPtr;

#else
  return (char *)shmat(handle, shmaddr, shmflg);
#endif

}


int SHMDT(char * ShmPtr)
{
#if defined (__CYGWIN32__) || defined (__SPEEDES_NT__)

  return 0;
#else
  return shmdt(ShmPtr);
#endif
}


int SHMCTL(int handle, int action, struct shmid_ds * buf)
{
#if defined (__CYGWIN32__) || defined (__SPEEDES_NT__)
  int exists = 0;

  if (handle >= totShmemHandles){
    if (ShmemHandles[handle].h != NULL){
      exists = 1;
    }
    if (exists){
      switch(action){
      case IPC_STAT:
	errno = EINVAL;
	printf("IPC_STAT not implemented yet\n");
	break;
      case IPC_SET:
	errno = EINVAL;
	printf("IPC_SET not implemented yet\n");
	break;
      case IPC_RMID:
	CloseHandle(ShmemHandles[handle].h);
	UnmapViewOfFile(ShmemHandles[handle].h);
	ShmemHandles[handle].h = NULL;
	free (ShmemHandles[handle].shmSegmentName);
	break;
      }
    }
    else{
      errno = EINVAL;
    }
  }

  return 0;

#else
  return shmctl(handle, action, buf);
#endif
}




