#ifndef SHARED_MEMORY_CALLS_H
#define SHARED_MEMORY_CALLS_H

#if defined (__CYGWIN32__) || defined (__SPEEDES_NT__)
#include <windows.h>
#ifdef __SPEEDES_NT__
typedef unsigned char ushort;
#endif

typedef int key_t;

#define IPC_PRIVATE ((key_t) 0)

struct ipc_perm
{
  key_t  key;
  ushort uid;   /* owner euid and egid */
  ushort gid;
  ushort cuid;  /* creator euid and egid */
  ushort cgid;
  ushort mode;  /* access modes see mode flags below */
  ushort seq;   /* sequence number */
};


struct shmid_ds {
  struct    ipc_perm shm_perm;  /* operation perms */
  int  shm_segsz;               /* size of segment (bytes) */
  time_t    shm_atime;          /* last attach time */
  time_t    shm_dtime;          /* last detach time */
  time_t    shm_ctime;          /* last change time */
  unsigned short shm_cpid;      /* pid of creator */
  unsigned short shm_lpid;      /* pid of last operator */
  short     shm_nattch;         /* no. of current attaches */
                                /* the following are private */
  unsigned short   shm_npages;  /* size of segment (pages) */
  unsigned long   *shm_pages;
  struct shm_desc *attaches;    /* descriptors for attaches */
};

/* permission flag for shmget */
#define SHM_R		0400	/* or S_IRUGO from linux/stat.h */
#define SHM_W		0200	/* or S_IWUGO from linux/stat.h */

/* mode for attach */
#define	SHM_RDONLY	010000	/* read-only access */
#define	SHM_RND		020000	/* round attach address to SHMLBA boundary */
#define	SHM_REMAP	040000	/* take-over region on attach */

/* super user shmctl commands */
#define SHM_LOCK 	11
#define SHM_UNLOCK 	12

struct	shminfo {
    int shmmax;	
    int shmmin;	
    int shmmni;	
    int shmseg;	
    int shmall;	
};

/* resource get request flags */
#define IPC_CREAT  00001000   /* create if key is nonexistent */
#define IPC_EXCL   00002000   /* fail if key exists */
#define IPC_NOWAIT 00004000   /* return error on wait */


/* 
 * Control commands used with semctl, msgctl and shmctl 
 * see also specific commands in sem.h, msg.h and shm.h
 */
#define IPC_RMID 0     /* remove resource */
#define IPC_SET  1     /* set ipc_perm options */
#define IPC_STAT 2     /* get ipc_perm options */
#define IPC_INFO 3     /* see ipcs */


#else
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#endif


#include "SpStdlib.H"

int SHMGET(key_t key, int size, int shmflg, char *shmSegName=NULL);
char * SHMAT(int handle, char * shmaddr, int shmflg);
int SHMCTL(int handle, int action,struct shmid_ds * buf);
int SHMDT(char * ShmPtr);



#endif
