#ifndef __SpCommServer__
#define __SpCommServer__

#include "SpServiceGroup.H"
#include "SpCommServerHeader.H"

/************ EXAMPLE SERVICE GROUP CLASS **************** 

  This is an example of how to define a service group class, an
  instance of which will be created by the MESSAGE_SERVER and will
  respond to a set of arriving messages.  The essential idea is this:
  you must write a method capable of handling each possible message
  type.  When the server receives a message intended for your object,
  it will read an int off the message header and this will tell it
  which one of the methods to invoke; so each handler-method must be
  identified by an associated int (see [1] below).

  You may have as many other methods or data members as you
  want--that's the idea of encapsulating this all in an object.  After
  you handler methods have been defined, you must write a constructor
  that will "register" these handlers so that the MESSAGE_SERVER can
  call them (by putting pointers to the methods into an array).  This
  requires two steps--one to allocate storage, and one to store a
  pointer for each method (see [2] and [3] below).

  Finally, you must write a global function to create an instance of
  your class (see [4]).  This is so the MESSAGE_SERVER can
  instantinate your obj.  It may have other SERVICE_GROUP class defs
  like yours.  When it receives a message addressed to a
  SERVICE_GROUP, it does the following: if the group exists, it
  invokes the appropriate method to handle the message; if the group
  does NOT exist, it creates an instance of the right kind of group
  (determined by the serviceType field in the msg header), and then
  calls the handler as above.

**********************/

const static int NumMsgTypes = 300;
const static int MaxMsgSize = 10000;

class CommServer : public SERVICE_GROUP {
  NET_INT* LocalNodeIDs;
  NET_INT* BoxNumbers;
  NET_INT* LocalNodeZeros;
  NET_INT * LNZ;
  char ** MultiSendMessages;
  int NextBoxNumber;
  int TablesCreated;
  int NumCheckedIn;
  int NumInGroup;
  int NumBoxes;
  double ** MessageStatistics;
  int GatherStatistics;
  CONNECTION ** SpComm_Sock;
  //The following are used in the reductions:
  int Ival;
  double Dval;
  SpSimTime Sval;
  // The following are used in the statistics
  int GlobalSumCount;
  int GlobalMinCount;
  int GlobalMaxCount;
  int BlockingCount;
  int FuzzyCount;


public:
  
  CONNECTION * SendSocket;

  void DummyConnectionMessage(CONNECTION *, COMM_SERVER_HEADER){};

  void Message(CONNECTION *conn, COMM_SERVER_HEADER *mh);

  void Broadcast(CONNECTION *conn, COMM_SERVER_HEADER *mh);
  
  void Multicast(CONNECTION *conn, COMM_SERVER_HEADER *mh);
  
  void Reduction(CONNECTION *conn, COMM_SERVER_HEADER *mh);
  
  void Startup(CONNECTION *conn, COMM_SERVER_HEADER *mh);

  int FlushSendQueues();

  int FlushSendQueue(int WhichQueue);

  int Allign8Bytes(int val) {return val + (8 - val%8);}

  void SendMessage(int whichConn, char * msg, int nBytes);

  virtual void CloseConnection(CONNECTION * connectionThatIsClosing);

  ~CommServer();
  
  CommServer();

};

// [4] Fcn to create new CommServer (must take void arg list):
CommServer *makeCommServer(void);

#endif
