#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpStdlib.H"

#include "SpHostRouter.H"
#include "SpNameHandle.H"
#include "SpNameValueList.H"
#include "SpEmConnectionHolder.H"
#include "SpSubscriptionItem.H"
#include "SpSubscriptionName.H"

/*
*****************************************************************************
* HOST_ROUTER constructor
*****************************************************************************
*/

HOST_ROUTER::HOST_ROUTER() {

  cout << "Constructing Host Router object" << endl;
  Nnodes = 0;
  EmCounter = 0;
  NnodesConnected = 0;
  NodeConnection = NULL;
  HashEmConnections = new C_HASH(100);
  NobjectTypes = 0;
  SubscriptionNames = NULL;
  NameByTypeList = NULL;
  GlobalIdHash = NULL;
  TypeHash = NULL;
  NameHash = NULL;
  CancelHandleHash = new SpHashTree(100);
  NextNode = 0;

/*
** Tell the service group parent class how many message handlers exist in
** the Host Router and then pass them in
*/

  allocateHandlerStorage(N_HOST_ROUTER_MESSAGES);

  STORE_MSG_HANDLER(SPEEDES_CONNECTION_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleSpeedesConnection);

  STORE_MSG_HANDLER(EXTERNAL_MODULE_CONNECTION_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleExternalModuleConnection);

  STORE_MSG_HANDLER(DISCONNECT_BARRIER_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleDisconnectBarrier);

  STORE_MSG_HANDLER(NAME_LOOKUP_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleNameLookup);

  STORE_MSG_HANDLER(OBJECT_NAME_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleObjectName);

  STORE_MSG_HANDLER(OBJECT_TYPE_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleObjectType);

  STORE_MSG_HANDLER(NUM_NODES_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleNumNodes);

  STORE_MSG_HANDLER(NEXT_NODE_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleNextNode);

  STORE_MSG_HANDLER(FLUSH_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleFlush);

  STORE_MSG_HANDLER(PAUSE_MESSAGE_ID,
		    HOST_ROUTER,
		    HandlePause);

  STORE_MSG_HANDLER(RESUME_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleResume);

  STORE_MSG_HANDLER(NAMED_PAUSE_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleNamedPause);

  STORE_MSG_HANDLER(NAMED_RESUME_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleNamedResume);

  STORE_MSG_HANDLER(KILL_SPEEDES_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleKillSPEEDES);

  STORE_MSG_HANDLER(QUERY_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleQuery);

  STORE_MSG_HANDLER(QUERY_REPLY_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleQueryReply);

  STORE_MSG_HANDLER(COMMAND_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleCommand);

  STORE_MSG_HANDLER(COMMAND_REPLY_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleCommandReply);

  STORE_MSG_HANDLER(TWO_WAY_COMMAND_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleTwoWayCommand);

  STORE_MSG_HANDLER(TIME_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleTime);

  STORE_MSG_HANDLER(TIME_REPLY_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleTimeReply);

  STORE_MSG_HANDLER(MIN_MSG_TIME_ID,
		    HOST_ROUTER,
		    HandleMinMsgTime);

  STORE_MSG_HANDLER(MIN_MSG_TIME_REPLY_ID,
		    HOST_ROUTER,
		    HandleMinMsgTimeReply);

  STORE_MSG_HANDLER(TIME_ADVANCE_REQUEST_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleTimeAdvanceRequest);

  STORE_MSG_HANDLER(TIME_ADVANCE_GRANT_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleTimeAdvanceGrant);

  STORE_MSG_HANDLER(GO_TO_TIME_REQUEST_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleGoToTimeRequest);

  STORE_MSG_HANDLER(GO_TO_TIME_GRANT_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleTimeAdvanceGrant);

  STORE_MSG_HANDLER(SPEEDES_TO_EXTERNAL_MODULE_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleSpeedesToExternalModule);

  STORE_MSG_HANDLER(SUBSCRIBE_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleSubscribe);

  STORE_MSG_HANDLER(UNSUBSCRIBE_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleUnSubscribe);

  STORE_MSG_HANDLER(SUBSCRIBED_DATA_MESSAGE_ID,
		    HOST_ROUTER,
		    HandleSubscribedData);

  STORE_MSG_HANDLER(SCHEDULE_EVENT_ID,
		        HOST_ROUTER,
		        HandleScheduleEvent);

  STORE_MSG_HANDLER(CANCEL_EVENT_ID,
		        HOST_ROUTER,
		        HandleCancelEvent);

  STORE_MSG_HANDLER(SCHEDULE_EVENT_REPLY_ID,
		        HOST_ROUTER,
		        HandleScheduleEventReply);

  STORE_MSG_HANDLER(CANCEL_EVENT_REPLY_ID,
		        HOST_ROUTER,
		        HandleCancelEventReply);


}

/*
*****************************************************************************
* HOST_ROUTER destructor
*****************************************************************************
*/

HOST_ROUTER::~HOST_ROUTER() {
  SpNameHandle *NameHandle;
  int i, j;
  int len;
  C_QUEUE *qall;
  EM_CONNECTION_HOLDER *EmConnectionHolder;
  SUBSCRIPTION_NAME *SubName;

  cout << "Destructing Host Router object" << endl;

/*
** Delete the NodeConnection array of pointers to connections
*/

  if (NodeConnection != NULL){
    delete [] NodeConnection;
  }

/*
** Delete the NameHandles in the NameList
*/

  len = NameList.get_length();
  for (i=0; i<len; i++) {
    NameHandle = (SpNameHandle *)NameList.pop_top();
    delete(NameHandle);
  
}

/*
** Delete the NameHandles in the TypeList
*/

  len = TypeList.get_length();
  for (i=0; i<len; i++) {
    NameHandle = (SpNameHandle *)TypeList.pop_top();
    delete(NameHandle);
  }

/*
** Delete the NameHandles in the NameByTypeList
*/

  if (NameByTypeList != NULL){
    for (i=0; i<NobjectTypes; i++) {
      len = NameByTypeList[i].get_length();
      for (j=0; j<len; j++) {
	NameHandle = (SpNameHandle *)NameByTypeList[i].pop_top();
	delete(NameHandle);
      }
    }
    delete [] NameByTypeList;
  }

/*
** Delete the NameHandles in the NameHash
*/

  if (NameHash != NULL){
    NameHash->combine();
    qall = NameHash->get_all();
    len = qall->get_length();
    for (i=0; i<len; i++) {
      NameHandle = (SpNameHandle *)qall->pop_top();
      delete NameHandle;
    }
    delete NameHash;	// Memory leak (need to delete array of queues)
  }

/*
** Delete the NameHandles in the GlobalIdHash
*/

  if (GlobalIdHash != NULL){
    GlobalIdHash->combine();
    qall = GlobalIdHash->get_all();
    len = qall->get_length();
    for (i=0; i<len; i++) {
      NameHandle = (SpNameHandle *)qall->pop_top();
      delete NameHandle;
    }
    delete GlobalIdHash;	// Memory leak (need to delete array of queues)
  }

/*
** Delete the NameHandles in the TypeHash
*/

  if (TypeHash != NULL){
    for (i=0; i<Nnodes; i++) {
      TypeHash[i]->combine();
      qall = TypeHash[i]->get_all();
      len = qall->get_length();
      for (j=0; j<len; j++) {
	NameHandle = (SpNameHandle *)qall->pop_top();
	delete NameHandle;
      }
      delete TypeHash[i];
    }
    delete [] TypeHash;
  }

/*
** Delete the connection holders
*/

  if (HashEmConnections != NULL){
    HashEmConnections->combine();
    qall = HashEmConnections->get_all();
    len = qall->get_length();
    for (i=0; i<len; i++) {
      EmConnectionHolder = (EM_CONNECTION_HOLDER *)qall->pop_top();
      delete EmConnectionHolder;
    }
    delete HashEmConnections;	// Memory leak (need to delete array of queues)
  }

/*
** Delete the Subscription Name hash table
*/

  if (SubscriptionNames != NULL){
    SubscriptionNames->combine();
    qall = SubscriptionNames->get_all();
    len = qall->get_length();
    for (i=0; i<len; i++) {
      SubName = (SUBSCRIPTION_NAME *)qall->pop_top();
      delete SubName;
    }
    delete SubscriptionNames;	// Memory leak (need to delete array of queues)
  }
  int *GID = (int*) CancelHandleHash->GetFirstElement();
  while (GID){
    delete GID;
    GID = (int*) CancelHandleHash->GetNextElement();
  }
  delete CancelHandleHash;

}

/*
*****************************************************************************
* HandleCloseConnection
*****************************************************************************
*/
void HOST_ROUTER::CloseConnection(CONNECTION *Connection) {
  EM_CONNECTION_HOLDER *EmConnectionHolder;
  CONNECTION *Conn;
  int i;
  int Node;
  DISCONNECT_BARRIER_MESSAGE DisconnectBarrierMessage;
  //KILL_SPEEDES_MESSAGE KillSPEEDESMessage;
  int BarrierNode;

  cout << "CLOSING CONNECTION" << endl;

/*
** First check if we are disconnecting an external module. If so, we want to
** make sure that we remove any barriers that were set up.
*/

  EmConnectionHolder = (EM_CONNECTION_HOLDER *)
    HashEmConnections->dehash(Connection->get_id());

  if (EmConnectionHolder) {

    Conn = EmConnectionHolder->GetConnection();

    if (Conn != Connection) {
      cout << "Error in removing a connection" << endl;
    }

    UnSubscribeAll(Connection->get_id());

    if (EmConnectionHolder->GetBarrier()) {
      BarrierNode = EmConnectionHolder->GetBarrierNode();

      DisconnectBarrierMessage.serviceType = -1;
      DisconnectBarrierMessage.group = -1;
      DisconnectBarrierMessage.handlerNum = DISCONNECT_BARRIER_MESSAGE_ID;
      DisconnectBarrierMessage.ExternalId=EmConnectionHolder->GetBarrierNode();
      DisconnectBarrierMessage.bytes = sizeof(DISCONNECT_BARRIER_MESSAGE);

      if (NodeConnection[BarrierNode]) {
	NodeConnection[BarrierNode]->scheduleSend(
	     (char *)&DisconnectBarrierMessage,
	     sizeof(DISCONNECT_BARRIER_MESSAGE));
      }
    }
    if (GIDofRecentlyScheduledEvent[Connection->get_id()]){
      delete (int*) GIDofRecentlyScheduledEvent[Connection->get_id()];
    }


    delete EmConnectionHolder;

  }else{

/*
** Check if we are disconnecting a SPEEDES node.
*/

    Node = -1;
    for (i=0; i<Nnodes; i++) {
      if (Connection == NodeConnection[i]) {
	Node = i;
	break;
      }
    }
    if (Node >= 0) {
      cout << "Disconnecting Node " << Node << endl;
      NodeConnection[Node] = NULL;
      int nConnected = 0;
      for (int jj=0; jj < Nnodes; ++jj){
	if (NodeConnection[jj]!= NULL){
	  ++nConnected;
	}
      }
      cout<<"NumConnected is "<<nConnected<<" out of "<<Nnodes<<endl;
      if (nConnected == 0){
	KillExternalModules();
      }
    }else{
      cout << "I have no idea who I am disconnecting\a" << endl;
    }
  }

}

/*
*****************************************************************************
* HandleSpeedesConnection
*****************************************************************************
*/
void HOST_ROUTER::HandleSpeedesConnection(CONNECTION *Connection,
					  GENERIC_MSG_HEADER *Message) {
  SpNameHandle *NameHandle;
  SpNameHandle *NameHandle1;
  SPEEDES_CONNECTION_MESSAGE *SpeedesConnectionMessage;
  char *buff;
  char *buff1;
  int size;
  int i;
  int len;
  C_QUEUE *qall;
  EM_CONNECTION_HOLDER *EmConnectionHolder;
  EXTERNAL_MODULE_CONNECTION_MESSAGE EmConnectionMessage;
  CONNECTION *EmConnection;

  cout << "Handling a SPEEDES connection" << endl;

  SpeedesConnectionMessage = (SPEEDES_CONNECTION_MESSAGE *)Message;
  if (Nnodes) {
    if (Nnodes != SpeedesConnectionMessage->Nnodes) {
      cout << "Error: Bad number of nodes: "
	   << SpeedesConnectionMessage->Nnodes
	   << ", should be " << Nnodes << endl;
    }
  }else{
    Nnodes = SpeedesConnectionMessage->Nnodes;
    NodeConnection = new CONNECTION *[Nnodes];
  }

  NodeConnection[SpeedesConnectionMessage->Node] = Connection;

/*
** Unpack the buffer and create NameHandles that go in the NameList
*/

  buff = (char *)&SpeedesConnectionMessage[1];
  size = SpeedesConnectionMessage->bytes - sizeof(SPEEDES_CONNECTION_MESSAGE);

  buff1 = buff;
  while (buff1-buff-size < 0) {

    NameHandle = new SpNameHandle();
    buff1 = NameHandle->EmptyBuffer(buff1);

    if (NameHandle->GetLocalId() < 0) {
      if (NameHandle->GetNode() == 0) NobjectTypes++;
      TypeList.push_bot(NameHandle);
    }else{
      NameList.push_bot(NameHandle);
    }

//    NameHandle->printstream(cout);

  }

  cout << "Received " << NameList.get_length() << " Name Handles" << endl;

  if (buff1-buff-size != 0) {
    cout << "Major Error in HandleSpeedesConnection: Bad buffer size "
	 << (buff1-buff) << " should be " << size << endl;
  }

/*
** Count the number of nodes connected so far and do some special bookkeeping
** things if all nodes have connected.
*/
  NnodesConnected++;

  if (NnodesConnected == Nnodes) {
    cout << "All " << Nnodes << " SPEEDES nodes have connected" << endl;

    len = NameList.get_length();
    NameHash = new C_HASH(len);
    GlobalIdHash = new C_HASH(len);

    NameByTypeList = new C_QUEUE[NobjectTypes];

    NameHandle = (SpNameHandle *)NameList.get_top();
    for (i=0; i<len; i++) {

// Add to NameHash

      NameHandle1 = new SpNameHandle(NameHandle);
      NameHandle1->set_id(NameHash->get_hashid(NameHandle->GetName()));
      NameHash->add(NameHandle1);

// Add to GlobalIdHash

      NameHandle1 = new SpNameHandle(NameHandle);
      GlobalIdHash->add(NameHandle1);

// Add to NameByTypeList array

      NameHandle1 = new SpNameHandle(NameHandle);
      NameByTypeList[NameHandle->GetObjectType()].push_bot(NameHandle1);

      NameHandle = (SpNameHandle *)NameHandle->get_link();
    }

// Add to TypeHash

    len = TypeList.get_length();
    TypeHash = new C_HASH*[Nnodes];
    for (i=0; i<Nnodes; i++) TypeHash[i] = new C_HASH(NobjectTypes);

    NameHandle = (SpNameHandle *)TypeList.get_top();
    for (i=0; i<len; i++) {
      int nd = NameHandle->GetNode();
      cout << NameHandle->GetName() << " is on Node " << nd << endl;
      NameHandle1 = new SpNameHandle(NameHandle);
      NameHandle1->set_id(TypeHash[nd]->get_hashid(NameHandle->GetName()));
      TypeHash[nd]->add(NameHandle1);

      NameHandle = (SpNameHandle *)NameHandle->get_link();
    }


    for (i=0; i<Nnodes; i++) {
      cout << "TypeHash[" << i << "] has " << TypeHash[i]->get_nitems()
	   << " items" << endl;
    }

/*
** Send messages back to the external modules that connected before SPEEDES
** nodes connected
*/

    EmConnectionMessage.handlerNum = EXTERNAL_MODULE_CONNECTION_MESSAGE_ID;
    EmConnectionMessage.bytes = sizeof(EXTERNAL_MODULE_CONNECTION_MESSAGE);

    HashEmConnections->combine();
    qall = HashEmConnections->get_all();
    len = qall->get_length();
    EmConnectionHolder = (EM_CONNECTION_HOLDER *)qall->get_top();

    for (i=0; i<len; i++) {
      EmConnection = EmConnectionHolder->GetConnection();

      if (EmConnection) {
	EmConnection->scheduleSend((char *)&EmConnectionMessage,
				   sizeof(EXTERNAL_MODULE_CONNECTION_MESSAGE));
      }

      EmConnectionHolder = (EM_CONNECTION_HOLDER *)
	EmConnectionHolder->get_link();
    }

  }

  if (NnodesConnected > Nnodes) {
    cout << "Error handling a SPEEDES connection: Too many nodes" << endl;
  }
  delete [] SpeedesConnectionMessage;

}

/*
*****************************************************************************
* HandleExternalModuleConnection
*****************************************************************************
*/

void HOST_ROUTER::HandleExternalModuleConnection(CONNECTION *Connection,
						 GENERIC_MSG_HEADER *Message) {
  EM_CONNECTION_HOLDER *EmConnectionHolder;

  cout << "Handling an External Module connection" << endl;
  EmConnectionHolder = new EM_CONNECTION_HOLDER(Connection);
  HashEmConnections->add(EmConnectionHolder);

  if ((NnodesConnected == Nnodes) && (Nnodes > 0)) {
    Connection->scheduleSend((char *)Message,
			     sizeof(EXTERNAL_MODULE_CONNECTION_MESSAGE));
  }else{
    cout << "Waiting until all SPEEDES nodes connect..." << endl;
  }
  delete [] (char*)Message;
}

/*
*****************************************************************************
* HandleDisconnectBarrier
*****************************************************************************
*/

void HOST_ROUTER::HandleDisconnectBarrier(CONNECTION *Connection,
					  GENERIC_MSG_HEADER *Message) {
  EM_CONNECTION_HOLDER *EmConnectionHolder;

  cout << "Handling an External Module disconnect "
       << Connection->get_id() << endl;

  EmConnectionHolder = (EM_CONNECTION_HOLDER *)
    HashEmConnections->get(Connection->get_id());

  if (EmConnectionHolder->GetBarrier()) {

    int BarrierNode = EmConnectionHolder->GetBarrierNode();
    EmConnectionHolder->ResetBarrier();

    DISCONNECT_BARRIER_MESSAGE *DisconnectBarrierMessage =
      (DISCONNECT_BARRIER_MESSAGE *)Message;

    DisconnectBarrierMessage->ExternalId = Connection->get_id();
    cout << "Sending disconnection Message ExternalId = "
	 << DisconnectBarrierMessage->ExternalId << endl;
    DisconnectBarrierMessage->bytes = sizeof(DISCONNECT_BARRIER_MESSAGE);

    if (NodeConnection[BarrierNode]) {
      NodeConnection[BarrierNode]->scheduleSend(
	     (char *)DisconnectBarrierMessage,
	     sizeof(DISCONNECT_BARRIER_MESSAGE));
    }

  }

  delete [] (char *)Message;

}

/*
*****************************************************************************
* HandleNameLookup
*****************************************************************************
*/

void HOST_ROUTER::HandleNameLookup(CONNECTION *Connection,
				   GENERIC_MSG_HEADER *Message) {
  SpNameHandle *NameHandle;
  NAME_LOOKUP_MESSAGE ReturnMessage;
  NAME_LOOKUP_MESSAGE *NameLookupMessage = (NAME_LOOKUP_MESSAGE *)Message;

  cout << "Handling a Name Lookup for "
       << (char *)&NameLookupMessage[1]
       << endl;

  NameHandle = (SpNameHandle *)NameHash->get((char *)&NameLookupMessage[1]);
  if (NameHandle) {
    ReturnMessage.GlobalId = NameHandle->GetGlobalId();
  }else{
    cout << "Error, could not lookup the name: "
	 << (char *)&NameLookupMessage[1]
	 << endl;
    ReturnMessage.GlobalId = -1;
  }

  ReturnMessage.handlerNum = NAME_LOOKUP_MESSAGE_ID;
  ReturnMessage.bytes = sizeof(NAME_LOOKUP_MESSAGE);

  Connection->scheduleSend((char *)&ReturnMessage,
			   sizeof(NAME_LOOKUP_MESSAGE));

  delete [] (char*) Message;
}

/*
*****************************************************************************
* HandleObjectName
*****************************************************************************
*/

void HOST_ROUTER::HandleObjectName(CONNECTION *Connection,
				   GENERIC_MSG_HEADER *Message) {
  SpNameHandle *NameHandle;
  int ObjectType;
  C_QUEUE *qall;
  C_QUEUE EmptyQ;
  OBJECT_NAME_MESSAGE *ReturnMessage;
  OBJECT_NAME_MESSAGE *ObjectNameMessage = (OBJECT_NAME_MESSAGE *)Message;

  cout << "Handling an object name request " << endl;

/*
** Figure out what type (if applicable) of objects to return names of
*/

  if (ObjectNameMessage->NameFlag) {
    NameHandle = (SpNameHandle *)
      TypeHash[0]->get((char *)&ObjectNameMessage[1]);
    if (NameHandle) {
      ObjectType = NameHandle->GetObjectType();
      qall = &NameByTypeList[ObjectType];
    }else{
      cout << "Error, bad object type" << endl;
      qall = &EmptyQ;
    }
  }else{
    ObjectType = ObjectNameMessage->ObjectType;
    if (ObjectType < 0) {
      qall = &NameList;
    }else{
      if (ObjectType < NobjectTypes) {
	qall = &NameByTypeList[ObjectType];
      }else{
	qall = &EmptyQ;
      }
    }
  }

  cout << "Number of object names returned " << qall->get_length() << endl;

  NameHandle = (SpNameHandle *)qall->get_top();
  int len = qall->get_length();

/*
** Create the NameValueList that is returned
*/

  SpNameValueList NameValueList(len+1);

/*
** Set the name of the NameValueList
*/

  char s[120];
  if (ObjectNameMessage->NameFlag) {
    sprintf(s, "Object Names of Type %s", (char *)&ObjectNameMessage[1]);
  }else{
    if (ObjectType < 0) {
      sprintf(s, "Object Names of All Types");
    }else{
      sprintf(s, "Object Names of Type %d", ObjectType);
    }
  }
  NameValueList.SetNameOfList(s);

/*
** Fill out the NameValueList
*/

  for (int i=0; i<len; i++) {
    NameValueList.SetNameValue(NameHandle->GetName(),
			       NameHandle->GetGlobalId());

    NameHandle = (SpNameHandle *)NameHandle->get_link();
  }

//  NameValueList.Print();

/*
** Generate the message that is returned
*/

  int size = NameValueList.GetSizeOfBuffer();
  int total_size = size + sizeof(OBJECT_NAME_MESSAGE);
  char *buff = new char[total_size];

  ReturnMessage = (OBJECT_NAME_MESSAGE *)buff;
  memcpy((char *)ReturnMessage,
	 (char *)ObjectNameMessage,
	 sizeof(OBJECT_NAME_MESSAGE));

  NameValueList.PackBuffer((char *)&ReturnMessage[1], size);
  if (len == 0) {
    ReturnMessage->bytes = sizeof(OBJECT_NAME_MESSAGE);;
  }else{
    ReturnMessage->bytes = total_size;
  }

  cout << "Sending total_size " << total_size << endl;

  Connection->scheduleSend(buff, total_size);
  delete [] buff;
  delete [] Message;
  cout << "Done sending object names..." << endl;

}

/*
*****************************************************************************
* HandleObjectType
*****************************************************************************
*/

void HOST_ROUTER::HandleObjectType(CONNECTION *Connection,
				   GENERIC_MSG_HEADER *Message) {
  SpNameHandle *NameHandle;
  OBJECT_TYPE_MESSAGE *ObjectTypeMessage = (OBJECT_TYPE_MESSAGE *)Message;

  cout << "Handling an ObjectType request" << endl;

  SpNameValueList NameValueList(NobjectTypes);
  NameValueList.SetNameOfList("Object Types in SPEEDES");

  TypeHash[0]->combine();
  C_QUEUE *qall = TypeHash[0]->get_all();
  int len = qall->get_length();

  if (len != NobjectTypes) {
    cout << "Error, bad number of object types "
	 << len << " " << NobjectTypes << endl;
  }

  NameHandle = (SpNameHandle *)qall->get_top();
  for (int i=0; i<len; i++) {
    NameValueList.SetNameValue(NameHandle->GetObjectType(),
			       NameHandle->GetName(),
			       NameHandle->GetObjectType());

    NameHandle = (SpNameHandle *)NameHandle->get_link();
  }

  int size = NameValueList.GetSizeOfBuffer();
  int total_size = size + sizeof(OBJECT_TYPE_MESSAGE);
  char *buff = new char[total_size];

  OBJECT_TYPE_MESSAGE *ReturnMessage = (OBJECT_TYPE_MESSAGE *)buff;
  memcpy((char *)ReturnMessage,
	 (char *)ObjectTypeMessage,
	 sizeof(OBJECT_TYPE_MESSAGE));

  NameValueList.PackBuffer((char *)&ReturnMessage[1], size);

  ReturnMessage->setBytes(total_size);

  Connection->scheduleSend(buff, total_size);
  delete [] buff;
  delete [] (char*)Message;

}


/*
*****************************************************************************
* HandleNumNodes
*****************************************************************************
*/

void HOST_ROUTER::HandleNumNodes(CONNECTION *Connection,
				 GENERIC_MSG_HEADER *Message) {

  NUM_NODES_MESSAGE *NumNodesMessage = (NUM_NODES_MESSAGE *)Message;
  NumNodesMessage->NumNodes = Nnodes;

  Connection->scheduleSend((char *)Message, sizeof(NUM_NODES_MESSAGE));
  delete [] (char*)Message;

}

/*
*****************************************************************************
* HandleNextNode
*****************************************************************************
*/

void HOST_ROUTER::HandleNextNode(CONNECTION *Connection,
				 GENERIC_MSG_HEADER *Message) {

  NEXT_NODE_MESSAGE *NextNodeMessage = (NEXT_NODE_MESSAGE *)Message;
  NextNodeMessage->NextNode = NextNode;
  NextNode = (NextNode+1) % Nnodes;

  Connection->scheduleSend((char *)Message, sizeof(NEXT_NODE_MESSAGE));
  delete [] (char*)Message;

}

/*
*****************************************************************************
* HandleFlush
*****************************************************************************
*/

void HOST_ROUTER::HandleFlush(CONNECTION *Connection,
			      GENERIC_MSG_HEADER *Message) {
  cout << "Handling a Flush message" << endl;
  FLUSH_MESSAGE *FlushMessage = (FLUSH_MESSAGE *)Message;
  Connection->scheduleSend((char *)FlushMessage, sizeof(FLUSH_MESSAGE));
  delete [] (char*)Message;
}

/*
*****************************************************************************
* HandlePause
*****************************************************************************
*/

void HOST_ROUTER::HandlePause(CONNECTION *,
			      GENERIC_MSG_HEADER *Message) {
  cout << "Handling a Pause message" << endl;
  PAUSE_MESSAGE *PauseMessage = (PAUSE_MESSAGE *)Message;

  if (NodeConnection[0]) {
    NodeConnection[0]->scheduleSend((char *)PauseMessage,
				    sizeof(PAUSE_MESSAGE));
  }
  delete [] (char*)Message;
}

/*
*****************************************************************************
* HandleResume
*****************************************************************************
*/

void HOST_ROUTER::HandleResume(CONNECTION *,
			       GENERIC_MSG_HEADER *Message) {
  cout << "Handling a Resume message" << endl;
  RESUME_MESSAGE *ResumeMessage = (RESUME_MESSAGE *)Message;

  if (NodeConnection[0]) {
    NodeConnection[0]->scheduleSend((char *)ResumeMessage,
				    sizeof(RESUME_MESSAGE));
  }
  delete [] (char*)Message;
}

/*
*****************************************************************************
* HandleNamedPause
*****************************************************************************
*/

void HOST_ROUTER::HandleNamedPause(CONNECTION *,
				   GENERIC_MSG_HEADER *Message) {

  NAMED_PAUSE_MESSAGE *NamedPauseMessage = (NAMED_PAUSE_MESSAGE *)Message;

  cout << "Handling a Named Pause message "
       << (char *)&NamedPauseMessage[1]
       << " with pause time " << NamedPauseMessage->Time
       << endl;

  if (NodeConnection[0]) {
    NodeConnection[0]->scheduleSend((char *)NamedPauseMessage,
				    sizeof(NAMED_PAUSE_MESSAGE) +
				    strlen((char *)&NamedPauseMessage[1])
				    + 1);
  }
  delete [] (char*)Message;
}

/*
*****************************************************************************
* HandleNamedResume
*****************************************************************************
*/

void HOST_ROUTER::HandleNamedResume(CONNECTION *,
			       GENERIC_MSG_HEADER *Message) {

  NAMED_RESUME_MESSAGE *NamedResumeMessage = (NAMED_RESUME_MESSAGE *)Message;

  cout << "Handling a Named Resume message"
       << (char *)&NamedResumeMessage[1]
       << endl;

  if (NodeConnection[0]) {
    NodeConnection[0]->scheduleSend((char *)NamedResumeMessage,
				    sizeof(NAMED_RESUME_MESSAGE) +
				    strlen((char *)&NamedResumeMessage[1])
				    + 1);
  }
  delete [] (char*)Message;
}


void HOST_ROUTER::KillExternalModules(CONNECTION* Connection)
{
  HashEmConnections->combine();
  C_QUEUE* qall = HashEmConnections->get_all();
  int len = qall->get_length();
  EM_CONNECTION_HOLDER *EmConnectionHolder =
    (EM_CONNECTION_HOLDER *)qall->get_top();
  KILL_SPEEDES_MESSAGE KillSPEEDESMessage;
  KillSPEEDESMessage.handlerNum = KILL_SPEEDES_MESSAGE_ID;
  KillSPEEDESMessage.bytes = sizeof(KILL_SPEEDES_MESSAGE);

  for (int i=0; i<len; i++) {
    CONNECTION* ExtConn = EmConnectionHolder->GetConnection();
    if (ExtConn != Connection) {
      cout << "Killing external module " << ExtConn->get_id() << endl;

      if (ExtConn) {
	ExtConn->scheduleSend((char *)&KillSPEEDESMessage,
			      sizeof(KILL_SPEEDES_MESSAGE));
      }

    }
    EmConnectionHolder = (EM_CONNECTION_HOLDER *)
      EmConnectionHolder->get_link();
  }
}

/*
*****************************************************************************
* HandleKillSPEEDES
*****************************************************************************
*/

void HOST_ROUTER::HandleKillSPEEDES(CONNECTION *,
				    GENERIC_MSG_HEADER *Message) {
  int i;

  cout << "Handling a Kill SPEEDES message" << endl;
  KILL_SPEEDES_MESSAGE *KillSPEEDESMessage = (KILL_SPEEDES_MESSAGE *)Message;

/*
** Send kill messages to all of the external modules
*/

  KillExternalModules();

/*
** Send kill messages to all of the nodes in SPEEDES. The SPEEDES
** nodes will only exit when all nodes have received the message to avoid
** deadlock problems (like what happens when nodes are doing a global reduction
** when some nodes have exited)
*/

  for (i=0; i<Nnodes; i++) {

    if (NodeConnection[i]) {
      NodeConnection[i]->scheduleSend((char *)KillSPEEDESMessage,
				      sizeof(KILL_SPEEDES_MESSAGE));
      NodeConnection[i] = NULL;
    }

  }
  delete [] (char*)Message;
}

/*
*****************************************************************************
* HandleQuery
*****************************************************************************
*/

void HOST_ROUTER::HandleQuery(CONNECTION *Connection,
			      GENERIC_MSG_HEADER *Message) {
  SpNameHandle *NameHandle;

  QUERY_MESSAGE *QueryMessage = (QUERY_MESSAGE *)Message;
  if (QueryMessage->GlobalId == -1) {
    NameHandle = (SpNameHandle *)NameHash->get((char *)&QueryMessage[1]);
  }else{
    NameHandle = (SpNameHandle *)GlobalIdHash->get(QueryMessage->GlobalId);
  }

  if (NameHandle) {

    QueryMessage->LocalId = NameHandle->GetLocalId();
    QueryMessage->ObjectType = NameHandle->GetObjectType();
    QueryMessage->GlobalId = NameHandle->GetGlobalId();
    QueryMessage->ExternalId = Connection->get_id();

    QueryMessage->bytes = sizeof(QUERY_MESSAGE);

    if (NodeConnection[NameHandle->GetNode()]) {
      NodeConnection[NameHandle->GetNode()]->
	scheduleSend((char *)QueryMessage,
		     sizeof(QUERY_MESSAGE) + QueryMessage->bytes);
    }

  }else{

    cout << "Error, bad query Name or Id" << endl;

    QUERY_REPLY_MESSAGE QueryReplyMessage;

    QueryReplyMessage.bytes = 0x7fffffff;
    QueryReplyMessage.serviceType = Message->serviceType;
    QueryReplyMessage.group = Message->group;
    QueryReplyMessage.handlerNum = QUERY_REPLY_MESSAGE_ID;
    QueryReplyMessage.bytes = sizeof(QUERY_REPLY_MESSAGE);

    cout << "Sending error response!" << endl;
    Connection->scheduleSend((char *)&QueryReplyMessage,
			     sizeof(QUERY_REPLY_MESSAGE));
  }
  delete [] (char*)Message;
}

/*
*****************************************************************************
* HandleQueryReply
*****************************************************************************
*/

void HOST_ROUTER::HandleQueryReply(CONNECTION *,
				   GENERIC_MSG_HEADER *Message) {
  EM_CONNECTION_HOLDER *EmConnectionHolder;
  CONNECTION *EmConnection;

  cout << "Handling a query reply message" << endl;

  QUERY_REPLY_MESSAGE *QueryReplyMessage = (QUERY_REPLY_MESSAGE *)Message;

  EmConnectionHolder = (EM_CONNECTION_HOLDER *)
    HashEmConnections->get(QueryReplyMessage->ExternalId);

  if (EmConnectionHolder) {
    EmConnection = EmConnectionHolder->GetConnection();

    if (EmConnection) {
      EmConnection->scheduleSend((char *)QueryReplyMessage,
				 sizeof(QUERY_REPLY_MESSAGE) +
				 QueryReplyMessage->bytes);
    }

  }else{
    cout << "Error, cannot send query reply message to "
	 << QueryReplyMessage->ExternalId << endl;
  }
  delete [] (char*)Message;
}


/*
*****************************************************************************
* HandleCommand
*****************************************************************************
*/

void HOST_ROUTER::HandleCommand(CONNECTION *Connection,
				GENERIC_MSG_HEADER *Message) {
  SpNameHandle *NameHandle = NULL;

  cout << "Handling a Command message" << endl;

  COMMAND_MESSAGE *CommandMessage = (COMMAND_MESSAGE *)Message;
  //char *Buff = (char *)&CommandMessage[1];

  if (CommandMessage->GlobalId == -1) {
    if (CommandMessage->Node == -1) {
      cout << "Getting NameHandle for " << (char *)&CommandMessage[1] << endl;
      NameHandle = (SpNameHandle *)NameHash->get((char *)&CommandMessage[1]);
    }else{
      NameHandle = NULL;
    }
  }else{
    NameHandle = (SpNameHandle *)GlobalIdHash->get(CommandMessage->GlobalId);
  }

  if (NameHandle) {

    CommandMessage->LocalId = NameHandle->GetLocalId();
    CommandMessage->ObjectType = NameHandle->GetObjectType();
    CommandMessage->ExternalId = Connection->get_id();

    if (NodeConnection[NameHandle->GetNode()]) {
      NodeConnection[NameHandle->GetNode()]->
	scheduleSend((char *)CommandMessage, CommandMessage->bytes);
    }

  }else{

    if (CommandMessage->Node == -1) {
      if (CommandMessage->ReturnStatus) {

	COMMAND_REPLY_MESSAGE CommandReplyMessage;

	CommandReplyMessage.serviceType = Message->serviceType;
	CommandReplyMessage.group = Message->group;
	CommandReplyMessage.handlerNum = COMMAND_REPLY_MESSAGE_ID;
	CommandMessage->ExternalId = Connection->get_id();

	CommandReplyMessage.CommandStatus = COMMAND_STATUS_BAD_OBJECT;
	CommandReplyMessage.bytes = sizeof(COMMAND_REPLY_MESSAGE);

	Connection->scheduleSend((char *)&CommandReplyMessage,
				 sizeof(COMMAND_REPLY_MESSAGE));

	cout << "Error, bad object Name or Id" << endl;
      }

    }else{

      CommandMessage->ExternalId = Connection->get_id();
      if (NodeConnection[CommandMessage->Node]) {
	NodeConnection[CommandMessage->Node]->
	  scheduleSend((char *)CommandMessage, CommandMessage->bytes);
      }

    }

  }

  delete [] (char*)Message;

}

/*
*****************************************************************************
* HandleTwoWayCommand
*****************************************************************************
*/

void HOST_ROUTER::HandleTwoWayCommand(CONNECTION *Connection,
				      GENERIC_MSG_HEADER *Message) {
  SpNameHandle *NameHandle;

  cout << "Handling a two-way command message" << endl;

  TWO_WAY_COMMAND_MESSAGE *CommandMessage = (TWO_WAY_COMMAND_MESSAGE *)Message;
  //char *Buff = (char *)&CommandMessage[1];

  if (CommandMessage->GlobalId == -1) {
    NameHandle = (SpNameHandle *)NameHash->get((char *)&CommandMessage[1]);
  }else{
    NameHandle = (SpNameHandle *)GlobalIdHash->get(CommandMessage->GlobalId);
  }

  if (NameHandle) {

    int ExternalId = Connection->get_id();
    int BarrierNode = NameHandle->GetNode();

    EM_CONNECTION_HOLDER *EmConnectionHolder =
      (EM_CONNECTION_HOLDER *)HashEmConnections->get(ExternalId);

    if (EmConnectionHolder) {
      EmConnectionHolder->SetBarrier();
      EmConnectionHolder->SetBarrierNode(BarrierNode);
    }else{
      cout << "Error, could not find EmConnectionHolder for "
	   << ExternalId << endl;
    }

    CommandMessage->LocalId = NameHandle->GetLocalId();
    CommandMessage->ObjectType = NameHandle->GetObjectType();
    CommandMessage->ExternalId = Connection->get_id();

    if (NodeConnection[NameHandle->GetNode()]) {
      NodeConnection[NameHandle->GetNode()]->
	scheduleSend((char *)CommandMessage, CommandMessage->bytes);
    }

  }else{

    COMMAND_REPLY_MESSAGE CommandReplyMessage;

    CommandReplyMessage.serviceType = Message->serviceType;
    CommandReplyMessage.group = Message->group;
    CommandReplyMessage.handlerNum = COMMAND_REPLY_MESSAGE_ID;

    CommandReplyMessage.CommandStatus = COMMAND_STATUS_BAD_OBJECT;
    CommandReplyMessage.bytes = sizeof(COMMAND_REPLY_MESSAGE);

    Connection->scheduleSend((char *)&CommandReplyMessage,
			     sizeof(COMMAND_REPLY_MESSAGE));

    cout << "Error, bad object Name or Id" << endl;

  }
  delete [] (char*)Message;
}


/*
*****************************************************************************
* HandleTime
*****************************************************************************
*/

void HOST_ROUTER::HandleTime(CONNECTION *Connection,
			     GENERIC_MSG_HEADER *Message) {
  static int NodeCounter = 0;

  cout << "Handling a time message" << endl;

  TIME_MESSAGE *TimeMessage =
    (TIME_MESSAGE *)Message;

  TimeMessage->ExternalId = Connection->get_id();

  if (NodeConnection[NodeCounter%Nnodes]) {
    NodeConnection[NodeCounter%Nnodes]->
      scheduleSend((char *)Message, sizeof(TIME_MESSAGE));
  }

  NodeCounter++;
  delete [] (char*)Message;
}

/*
*****************************************************************************
* HandleTimeReply
*****************************************************************************
*/

void HOST_ROUTER::HandleTimeReply(CONNECTION *,
				  GENERIC_MSG_HEADER *Message) {

  cout << "Handling a time reply message" << endl;

  TIME_REPLY_MESSAGE *TimeReplyMessage =
    (TIME_REPLY_MESSAGE *)Message;

  EM_CONNECTION_HOLDER *EmConnectionHolder = (EM_CONNECTION_HOLDER *)
    HashEmConnections->get(TimeReplyMessage->ExternalId);

  if (EmConnectionHolder) {
    CONNECTION *EmConnection = EmConnectionHolder->GetConnection();

    if (EmConnection) {
      EmConnection->scheduleSend((char *)TimeReplyMessage,
				 sizeof(TIME_REPLY_MESSAGE));
    }

  }else{
    cout << "Error, cannot send time reply message to "
	 << TimeReplyMessage->ExternalId << endl;
  }
  delete [] (char*)Message;
}

/*
*****************************************************************************
* HandleMinMsgTime
*****************************************************************************
*/

void HOST_ROUTER::HandleMinMsgTime(CONNECTION *Connection,
				   GENERIC_MSG_HEADER *Message) {
  static int NodeCounter = 0;

  cout << "Handling a MinMsgTime message" << endl;

  MIN_MSG_TIME_MESSAGE *MinMsgTimeMessage =
    (MIN_MSG_TIME_MESSAGE *)Message;

  MinMsgTimeMessage->ExternalId = Connection->get_id();

  if (NodeConnection[NodeCounter%Nnodes]) {
    NodeConnection[NodeCounter%Nnodes]->
      scheduleSend((char *)Message, sizeof(MIN_MSG_TIME_MESSAGE));
  }

  NodeCounter++;
  delete [] (char*)Message;

}

/*
*****************************************************************************
* HandleMinMsgTimeReply
*****************************************************************************
*/

void HOST_ROUTER::HandleMinMsgTimeReply(CONNECTION *,
					GENERIC_MSG_HEADER *Message) {

  cout << "Handling a MinMsgTime reply message" << endl;

  MIN_MSG_TIME_REPLY_MESSAGE *MinMsgTimeReplyMessage =
    (MIN_MSG_TIME_REPLY_MESSAGE *)Message;

  EM_CONNECTION_HOLDER *EmConnectionHolder = (EM_CONNECTION_HOLDER *)
    HashEmConnections->get(MinMsgTimeReplyMessage->ExternalId);

  if (EmConnectionHolder) {
    CONNECTION *EmConnection = EmConnectionHolder->GetConnection();

    if (EmConnection) {
      EmConnection->scheduleSend((char *)MinMsgTimeReplyMessage,
				 sizeof(MIN_MSG_TIME_REPLY_MESSAGE));
    }

  }else{

    cout << "Error, cannot send MinMsgTime reply message to "
	 << MinMsgTimeReplyMessage->ExternalId << endl;

  }

  delete [] (char*)Message;

}

/*
*****************************************************************************
* HandleCommandReply
*****************************************************************************
*/

void HOST_ROUTER::HandleCommandReply(CONNECTION *,
				     GENERIC_MSG_HEADER *Message) {

  cout << "Handling a command reply message" << endl;

  COMMAND_REPLY_MESSAGE *CommandReplyMessage =
    (COMMAND_REPLY_MESSAGE *)Message;

  EM_CONNECTION_HOLDER *EmConnectionHolder = (EM_CONNECTION_HOLDER *)
    HashEmConnections->get(CommandReplyMessage->ExternalId);

  if (EmConnectionHolder) {
    CONNECTION *EmConnection = EmConnectionHolder->GetConnection();

    if (EmConnection) {
      EmConnection->scheduleSend((char *)CommandReplyMessage,
				 sizeof(COMMAND_REPLY_MESSAGE));
    }

  }else{
    cout << "Error, cannot send command reply message to "
	 << CommandReplyMessage->ExternalId << endl;
  }
  delete [] (char*)Message;
}

/*
*****************************************************************************
* HandleGoToTimeRequest
*****************************************************************************
*/

void HOST_ROUTER::HandleGoToTimeRequest(CONNECTION *Connection,
					GENERIC_MSG_HEADER *Message) {

  GO_TO_TIME_REQUEST_MESSAGE *TimeMessage =
    (GO_TO_TIME_REQUEST_MESSAGE *)Message;

  int ExternalId = Connection->get_id();
  int BarrierNode = ExternalId % Nnodes;

  EM_CONNECTION_HOLDER *EmConnectionHolder =
    (EM_CONNECTION_HOLDER *)HashEmConnections->get(ExternalId);

  if (EmConnectionHolder) {
    EmConnectionHolder->SetBarrier();
    EmConnectionHolder->SetBarrierNode(BarrierNode);
  }else{
    cout << "Error, could not find EmConnectionHolder for "
	 << ExternalId << endl;
  }

  TimeMessage->ExternalId = ExternalId;

  if (NodeConnection[BarrierNode]) {
    NodeConnection[BarrierNode]->scheduleSend
      ((char *)Message, sizeof(GO_TO_TIME_REQUEST_MESSAGE));
  }
  delete [] (char*)Message;
}

/*
*****************************************************************************
* HandleGoToTimeGrant
*****************************************************************************
*/

void HOST_ROUTER::HandleGoToTimeGrant(CONNECTION *,
				      GENERIC_MSG_HEADER *Message) {


  GO_TO_TIME_GRANT_MESSAGE *TimeMessage =
    (GO_TO_TIME_GRANT_MESSAGE *)Message;

  EM_CONNECTION_HOLDER *EmConnectionHolder = (EM_CONNECTION_HOLDER *)
    HashEmConnections->get(TimeMessage->ExternalId);

  if (EmConnectionHolder) {
    if (EmConnectionHolder->GetBarrier()) {
      CONNECTION *EmConnection = EmConnectionHolder->GetConnection();

      if (EmConnection) {
	EmConnection->scheduleSend((char *)TimeMessage,
				   sizeof(GO_TO_TIME_GRANT_MESSAGE));
      }

    }
  }else{
    cout << "Error, cannot send go to time grant message to "
	 << TimeMessage->ExternalId << endl;
  }
  delete [] (char*)Message;
}



/*
*****************************************************************************
* HandleTimeAdvanceRequest
*****************************************************************************
*/

void HOST_ROUTER::HandleTimeAdvanceRequest(CONNECTION *Connection,
					   GENERIC_MSG_HEADER *Message) {
  cout << "Handling a time advance request message" << endl;

  TIME_ADVANCE_REQUEST_MESSAGE *TimeMessage =
    (TIME_ADVANCE_REQUEST_MESSAGE *)Message;

  int ExternalId = Connection->get_id();
  int BarrierNode = ExternalId % Nnodes;

  EM_CONNECTION_HOLDER *EmConnectionHolder =
    (EM_CONNECTION_HOLDER *)HashEmConnections->get(ExternalId);

  if (EmConnectionHolder) {
    EmConnectionHolder->SetBarrier();
    EmConnectionHolder->SetBarrierNode(BarrierNode);
  }else{
    cout << "Error, could not find EmConnectionHolder for "
	 << ExternalId << endl;
  }

  TimeMessage->ExternalId = ExternalId;

  if (NodeConnection[BarrierNode]) {
    NodeConnection[BarrierNode]->
      scheduleSend((char *)Message, sizeof(TIME_ADVANCE_REQUEST_MESSAGE));
  }
  delete [] (char*)Message;
}

/*
*****************************************************************************
* HandleTimeAdvanceGrant
*****************************************************************************
*/

void HOST_ROUTER::HandleTimeAdvanceGrant(CONNECTION *,
					 GENERIC_MSG_HEADER *Message) {

  cout << "Handling a time advance grant message" << endl;

  TIME_ADVANCE_GRANT_MESSAGE *TimeMessage =
    (TIME_ADVANCE_GRANT_MESSAGE *)Message;

  EM_CONNECTION_HOLDER *EmConnectionHolder = (EM_CONNECTION_HOLDER *)
    HashEmConnections->get(TimeMessage->ExternalId);

  if (EmConnectionHolder) {
    if (EmConnectionHolder->GetBarrier()) {
      CONNECTION *EmConnection = EmConnectionHolder->GetConnection();

      if (EmConnection) {
	EmConnection->scheduleSend((char *)TimeMessage,
				   sizeof(TIME_ADVANCE_GRANT_MESSAGE));
      }

    }
  }else{
    cout << "Error, cannot send time advance grant message to "
	 << TimeMessage->ExternalId << endl;
  }
  delete [] (char*)Message;
}

/*
*****************************************************************************
* HandleSpeedesToExternalModule
*****************************************************************************
*/

void HOST_ROUTER::HandleSpeedesToExternalModule(CONNECTION *,
						GENERIC_MSG_HEADER *Message) {

  SPEEDES_TO_EXTERNAL_MODULE_MESSAGE *SpEmMessage =
    (SPEEDES_TO_EXTERNAL_MODULE_MESSAGE *)Message;

  EM_CONNECTION_HOLDER *EmConnectionHolder = (EM_CONNECTION_HOLDER *)
    HashEmConnections->get(SpEmMessage->ExternalId);

  cout << "Handling a SPEEDES to external module message to "
       << SpEmMessage->ExternalId << endl;

  if (EmConnectionHolder) {

    CONNECTION *EmConnection = EmConnectionHolder->GetConnection();

    int TotalBytes = SpEmMessage->bytes
      + sizeof(SPEEDES_TO_EXTERNAL_MODULE_MESSAGE);

    if (EmConnection) {
      EmConnection->scheduleSend((char *)SpEmMessage, TotalBytes);
    }

  }else{
    cout << "Error, cannot send message to "
	 << SpEmMessage->ExternalId << endl;
  }
  delete [] (char*)Message;
}

/*
*****************************************************************************
* HandleSubscribe
*****************************************************************************
*/

void HOST_ROUTER::HandleSubscribe(CONNECTION *Connection,
				  GENERIC_MSG_HEADER *Message) {
  SUBSCRIPTION_ITEM *SubItem;
  SUBSCRIPTION_NAME *SubName;
  int ForwardMessageFlag = 1;

  SUBSCRIBE_MESSAGE *SubscribeMessage = (SUBSCRIBE_MESSAGE *)Message;
  SpNameHandle *NameHandle;

  cout << "Handling a subscribe message" << endl;

  SubItem = new SUBSCRIPTION_ITEM();
  SubItem->UnPackBuffer((char *)&SubscribeMessage[1]);
  SubItem->SetExternalId(Connection->get_id());

/*
** Fill out the rest of the SubItem
*/

  if (SubItem->GetObjectName()) {
    NameHandle = (SpNameHandle *)NameHash->get(SubItem->GetObjectName());
    if (NameHandle) {
      SubItem->SetObjectInfo(NameHandle->GetObjectType(),
			     NameHandle->GetLocalId(),
			     NameHandle->GetNode());
    }else{
      cout << "Error, could not lookup the name: "
	 << SubItem->GetObjectName()
	 << endl;

      delete SubItem;
      SubItem = NULL;

      delete [] (char *)Message;
      return;
    }
  }

  SubItem->Print();

  if (SubscriptionNames == NULL) {

    cout << "Creating Subscription Name hash table" << endl;

    SubscriptionNames = new C_HASH(100);

    cout << "Creating SubName " << SubItem->GetName() << endl;
    SubName = new SUBSCRIPTION_NAME(SubItem->GetName(), SubItem);

    SubName->set_id(SubscriptionNames->get_hashid(SubItem->GetName()));
    SubscriptionNames->add(SubName);

  }else{

    SubName = (SUBSCRIPTION_NAME *)SubscriptionNames->get(SubItem->GetName());

    if (SubName) {

      if (!SubName->Compare(SubItem)) {

	cout << "Adding SubItem to SubName" << endl;

	if (SubName->AddSubItem(SubItem)) {
	  ForwardMessageFlag = 0;
	}else{
	  ForwardMessageFlag = 1;
	}

      }else{

	delete SubItem;
	SubItem = NULL;

	ForwardMessageFlag = 0;
	cout << "Warning, trying to add duplicate subscription item" << endl;

      }

    }else{

      cout << "Creating SubName " << SubItem->GetName() << endl;
      SubName = new SUBSCRIPTION_NAME(SubItem->GetName(), SubItem);

      SubName->set_id(SubscriptionNames->get_hashid(SubItem->GetName()));
      SubscriptionNames->add(SubName);

    }

  }

  if (ForwardMessageFlag) {


    SubItem->PackBuffer((char *)&SubscribeMessage[1]);

    if (SubItem->GetNode() == -1) {

      if (NodeConnection[0]) {
	NodeConnection[0]->
	  scheduleSend((char *)SubscribeMessage, SubscribeMessage->bytes +
		       sizeof(SUBSCRIBE_MESSAGE));
      }

    }else{	

      if (NodeConnection[SubItem->GetNode()]) {
	NodeConnection[SubItem->GetNode()]->
	  scheduleSend((char *)SubscribeMessage, SubscribeMessage->bytes +
		       sizeof(SUBSCRIBE_MESSAGE));
      }

    }

  }else{

    cout << "Not forwarding subscription " << endl;

  }

  delete [] (char *)SubscribeMessage;

}


/*
*****************************************************************************
* HandleUnSubscribe
*****************************************************************************
*/

void HOST_ROUTER::HandleUnSubscribe(CONNECTION *Connection,
				    GENERIC_MSG_HEADER *Message) {
  SUBSCRIPTION_ITEM *SubItem;
  SUBSCRIPTION_NAME *SubName;
  int ForwardMessageFlag = 1;

  UNSUBSCRIBE_MESSAGE *UnSubscribeMessage = (UNSUBSCRIBE_MESSAGE *)Message;
  SpNameHandle *NameHandle;

  cout << "Handling an UnSubscribe message" << endl;

  SubItem = new SUBSCRIPTION_ITEM();
  SubItem->UnPackBuffer((char *)&UnSubscribeMessage[1]);
  SubItem->SetExternalId(Connection->get_id());

/*
** Fill out the rest of the SubItem
*/

  if (SubItem->GetObjectName()) {
    NameHandle = (SpNameHandle *)NameHash->get(SubItem->GetObjectName());
    if (NameHandle) {
      SubItem->SetObjectInfo(NameHandle->GetObjectType(),
			     NameHandle->GetLocalId(),
			     NameHandle->GetNode());
    }else{
      cout << "Error, could not lookup the name: "
	 << SubItem->GetObjectName()
	 << endl;

      delete SubItem;
      SubItem = NULL;

      delete [] (char *)Message;
      return;
    }
  }

  SubItem->Print();

  if (SubscriptionNames == NULL) {

    cout << "Error, trying to unsubscribe before subscriptions made" << endl;

  }else{

    SubName = (SUBSCRIPTION_NAME *)SubscriptionNames->get(SubItem->GetName());

    if (SubName) {

      if (SubName->Compare(SubItem)) {

	cout << "Removing SubItem from SubName" << endl;
	ForwardMessageFlag = SubName->RemoveSubItem(SubItem);

      }else{

	cout << "Warning, trying to remove non-existent subscription" << endl;
	delete SubItem;
	SubItem = NULL;
	ForwardMessageFlag = 0;

      }

    }else{

      cout << "Warning, name of subscription is non-existent" << endl;
      delete SubItem;
      SubItem = NULL;
      ForwardMessageFlag = 0;

    }

  }

  if (ForwardMessageFlag) {

    cout << "Forwarding UnSubscription " << SubItem->GetName() << endl;

    SubItem->PackBuffer((char *)&UnSubscribeMessage[1]);

    if (SubItem->GetNode() == -1) {

      if (NodeConnection[0]) {
	NodeConnection[0]->
	  scheduleSend((char *)UnSubscribeMessage, UnSubscribeMessage->bytes);
      }

    }else{

      if (NodeConnection[SubItem->GetNode()]) {
	NodeConnection[SubItem->GetNode()]->
	  scheduleSend((char *)UnSubscribeMessage, UnSubscribeMessage->bytes);
      }

    }

  }else{

    cout << "Not forwarding UnSubscription " << endl;

  }

  delete [] (char *)Message;

}

/*
*****************************************************************************
* UnSubscribeAll
*****************************************************************************
*/

void HOST_ROUTER::UnSubscribeAll(int ExternalId) {
  int i, j;
  int len1, len2;
  SUBSCRIPTION_NAME *SubName;
  C_QUEUE *qall;
  C_QUEUE *SubItems;
  SUBSCRIPTION_ITEM *SubItem;
  UNSUBSCRIBE_MESSAGE *UnSubscribeMessage;
  int Size;

  if (SubscriptionNames != NULL) {

    SubscriptionNames->combine();
    qall = SubscriptionNames->get_all();
    len1 = qall->get_length();

    SubName = (SUBSCRIPTION_NAME *)qall->get_top();
    for (i=0; i<len1; i++) {
      
      SubItems = SubName->GetSubItems();
      len2 = SubItems->get_length();

      for (j=0; j<len2; j++) {
	SubItem = (SUBSCRIPTION_ITEM *)SubItems->pop_top();
	if (SubItem->GetExternalId() == ExternalId) {

	  Size = SubItem->GetBufferSize();
	  UnSubscribeMessage = (UNSUBSCRIBE_MESSAGE *)
	    new char[Size + sizeof(UNSUBSCRIBE_MESSAGE)];

	  SubItem->PackBuffer((char *)&UnSubscribeMessage[1]);

	  UnSubscribeMessage->handlerNum = UNSUBSCRIBE_MESSAGE_ID;
	  UnSubscribeMessage->bytes = Size;

	  if (SubItem->GetNode() == -1) {

	    if (NodeConnection[0]) {
	      NodeConnection[0]->
		scheduleSend((char *)UnSubscribeMessage,
			     sizeof(UNSUBSCRIBE_MESSAGE)+
			     UnSubscribeMessage->bytes);
	    }

	  }else{

	    if (NodeConnection[SubItem->GetNode()]) {
	      NodeConnection[SubItem->GetNode()]->
		scheduleSend((char *)UnSubscribeMessage,
			     sizeof(UNSUBSCRIBE_MESSAGE)+
			     UnSubscribeMessage->bytes);
	    }

	  }

	  delete [] (char *)UnSubscribeMessage;
	  delete SubItem;

	}else{

	  SubItems->push_bot(SubItem);

	}
      }

      SubName = (SUBSCRIPTION_NAME *)SubName->get_link();

    }
  }
}


/*
*****************************************************************************
* HandleSpeedesToExternalModule
*****************************************************************************
*/

void HOST_ROUTER::HandleSubscribedData(CONNECTION *,
				       GENERIC_MSG_HEADER *Message) {

  SUBSCRIBED_DATA_MESSAGE *SpEmMessage = (SUBSCRIBED_DATA_MESSAGE *)Message;
  SUBSCRIPTION_NAME *SubName;
  EM_CONNECTION_HOLDER *EmConnectionHolder;
  CONNECTION *EmConnection;

  char *Name = (char *)&SpEmMessage[1];

  cout << "Handling a SPEEDES subscribed data message of type: "
       << Name << " from " << SpEmMessage->GlobalId << endl;

  if (SubscriptionNames == NULL) {
    cout << "Warning, No subscriptions have been defined" << endl;
    delete [] (char *)Message;
    return;
  }

  SubName = (SUBSCRIPTION_NAME *)SubscriptionNames->get(Name);

  if (SubName == NULL) {
    cout << "Warning, No subscriptions of type " << Name
	 << " have been defined" << endl;
    delete [] (char *)Message;
    return;
  }

  SpNameHandle *NameHandle = (SpNameHandle *)
    GlobalIdHash->get(SpEmMessage->GlobalId);

  if (NameHandle == NULL) {
    cout << "Warning, No name handles with Global Id "
	 << SpEmMessage->GlobalId << endl;
    delete [] (char *)Message;
    return;
  }

  C_QUEUE *SubItems = SubName->GetSubItems();
  int len = SubItems->get_length();
  SUBSCRIPTION_ITEM *SubItem = (SUBSCRIPTION_ITEM *)SubItems->get_top();

  for (int i=0; i<len; i++) {

    if (SubItem->CheckContained(NameHandle->GetObjectType(),
				NameHandle->GetLocalId(),
				NameHandle->GetNode())) {

      EmConnectionHolder = (EM_CONNECTION_HOLDER *)
	HashEmConnections->get(SubItem->GetExternalId());

      if (EmConnectionHolder != NULL) {

	EmConnection = EmConnectionHolder->GetConnection();

	EmConnection->scheduleSend((char *)SpEmMessage,
				   sizeof(SUBSCRIBED_DATA_MESSAGE) +
				   SpEmMessage->bytes
				   + strlen((char *)&SpEmMessage[1]) + 1);

	cout << "Sending Subscribed Data to " << SubItem->GetExternalId()
	     << endl;
      }else{

	cout << "Warning: Connection holder not found" << endl;

      }

    }

    SubItem = (SUBSCRIPTION_ITEM *)SubItem->get_link();

  }
  delete [] (char*)Message;
}

void HOST_ROUTER::HandleScheduleEvent(CONNECTION *Connection,
				      GENERIC_MSG_HEADER *Message)
{
  SCHEDULE_EVENT_MESSAGE *msg = (SCHEDULE_EVENT_MESSAGE*)Message;
  SpNameHandle *NameHandle = NULL;
  char * eventName = (char *)&msg[1];
  char * objectName = eventName + strlen(eventName) + 1 +
    msg->DataBytes + msg->MessageBytes;

  msg->ExternalId = Connection->get_id();
  if (msg->GlobalId != -1){
    NameHandle = (SpNameHandle *)GlobalIdHash->get(msg->GlobalId);
  }
  else {
    NameHandle = (SpNameHandle *)NameHash->get(objectName);
    char * objectName = eventName + strlen(eventName) + 1 +
      msg->DataBytes + msg->MessageBytes;
    NameHash->get(objectName);
  }


  if (NameHandle){
    if (msg->CancelHandleRequested){
      int * GID = new int(NameHandle->GetGlobalId());
      GIDofRecentlyScheduledEvent[Connection->get_id()] = GID;
    }
    msg->LocalId = NameHandle->GetLocalId();
    msg->ObjectType = NameHandle->GetObjectType();
    int node = msg->ObjectNode = NameHandle->GetNode();
    if (NodeConnection[node]){
      NodeConnection[node]->scheduleSend((char*)Message, Message->getBytes());
    }
  }
  else{
    SCHEDULE_EVENT_REPLY_MESSAGE replyMessage;
    replyMessage.handlerNum = SCHEDULE_EVENT_REPLY_ID;
    replyMessage.ErrorCode = OBJECT_SPECIFIED_DOES_NOT_EXIST;
    replyMessage.bytes = sizeof(SCHEDULE_EVENT_REPLY_MESSAGE);
    Connection->scheduleSend((char *)&replyMessage,
			     sizeof(SCHEDULE_EVENT_REPLY_MESSAGE));
  }
  delete [] (char*)Message;
}




void HOST_ROUTER::HandleCancelEvent(CONNECTION *Connection,
				    GENERIC_MSG_HEADER *Message)
{

  CANCEL_EVENT_MESSAGE *msg = (CANCEL_EVENT_MESSAGE*)Message;

  msg->ExternalId = Connection->get_id();
  int *globalId = (int*)CancelHandleHash->Remove(msg->CancelHandle);
  if (globalId){
    msg->GlobalId = *globalId;
    SpNameHandle *NameHandle = (SpNameHandle *)GlobalIdHash->get(msg->GlobalId);
    int node = NameHandle->GetNode();

    if (NodeConnection[node]) {
      NodeConnection[node]->scheduleSend((char *)msg, msg->getBytes());
    }
  }
  else{
    CANCEL_EVENT_REPLY_MESSAGE replyMessage;
    replyMessage.handlerNum = CANCEL_EVENT_REPLY_ID;
    replyMessage.ErrorCode = CANCEL_EVENT_NOT_FOUND;
    replyMessage.bytes = sizeof(CANCEL_EVENT_REPLY_MESSAGE);
    Connection->scheduleSend((char *)&replyMessage,
			     sizeof(CANCEL_EVENT_REPLY_MESSAGE));
  }
  delete [] (char *)Message;
}


void HOST_ROUTER::HandleScheduleEventReply(CONNECTION *,
					   GENERIC_MSG_HEADER *Message)
{
  SCHEDULE_EVENT_REPLY_MESSAGE *msg = (SCHEDULE_EVENT_REPLY_MESSAGE*)Message;
  EM_CONNECTION_HOLDER *EmConnectionHolder = (EM_CONNECTION_HOLDER *)
    HashEmConnections->get(msg->ExternalId);

  if (EmConnectionHolder){
    CONNECTION * extConn = EmConnectionHolder->GetConnection();
    extConn->scheduleSend((char*)msg, sizeof(SCHEDULE_EVENT_REPLY_MESSAGE));

    CancelHandleHash->Insert(GIDofRecentlyScheduledEvent[extConn->get_id()],
			     msg->CancelHandle);
    GIDofRecentlyScheduledEvent[extConn->get_id()] = NULL;
  }
  delete [] (char *)Message;
}

void HOST_ROUTER::HandleCancelEventReply(CONNECTION *,
					 GENERIC_MSG_HEADER *Message)
{
  CANCEL_EVENT_REPLY_MESSAGE *msg = (CANCEL_EVENT_REPLY_MESSAGE*)Message;
  EM_CONNECTION_HOLDER *EmConnectionHolder = (EM_CONNECTION_HOLDER *)
    HashEmConnections->get(msg->ExternalId);

  if (EmConnectionHolder){
    EmConnectionHolder->GetConnection()->
      scheduleSend((char*)msg, sizeof(CANCEL_EVENT_REPLY_MESSAGE));
  }
  delete [] (char *)Message;
}
