// SpHostRouter.H header file services communications between SPEEDES and
// external modules.

#ifndef HOST_ROUTER_OBJECT
#define HOST_ROUTER_OBJECT

#include "SpServiceGroup.H"
#include "SpHostRouterMsgs.H"
#include "hash.H"
#include "queue.H"
#include "SpHashTree.H"
#include "SpDynPtrArray.H"

class HOST_ROUTER : public SERVICE_GROUP {

  private:

    int Nnodes;
    int NnodesConnected;
    int EmCounter;
    int NobjectTypes;
    int NextNode;

    CONNECTION **NodeConnection;
    C_HASH *HashEmConnections;

    C_QUEUE NameList;
    C_QUEUE TypeList;
    C_QUEUE *NameByTypeList;
    C_HASH *NameHash;
    C_HASH *GlobalIdHash;
    C_HASH **TypeHash;

    SpHashTree * CancelHandleHash;
    SpDynPtrArray GIDofRecentlyScheduledEvent;

    C_HASH *SubscriptionNames;

  protected:


  void KillExternalModules(CONNECTION*conn = NULL);

  public:
    HOST_ROUTER();
    ~HOST_ROUTER();

    virtual void CloseConnection(CONNECTION *Connection);

/*
** Message handlers for the Host Router
*/

    void HandleSpeedesConnection(CONNECTION *Connection,
				 GENERIC_MSG_HEADER *Message);

    void HandleExternalModuleConnection(CONNECTION *Connection,
					GENERIC_MSG_HEADER *Message);

    void HandleDisconnectBarrier(CONNECTION *Connection,
				 GENERIC_MSG_HEADER *Message);

    void HandleNameLookup(CONNECTION *Connection, GENERIC_MSG_HEADER *Message);

    void HandleObjectName(CONNECTION *Connection, GENERIC_MSG_HEADER *Message);

    void HandleObjectType(CONNECTION *Connection, GENERIC_MSG_HEADER *Message);

    void HandleNumNodes(CONNECTION *Connection, GENERIC_MSG_HEADER *Message);
    void HandleNextNode(CONNECTION *Connection, GENERIC_MSG_HEADER *Message);

    void HandleFlush(CONNECTION *Connection, GENERIC_MSG_HEADER *Message);

    void HandlePause(CONNECTION *Connection, GENERIC_MSG_HEADER *Message);

    void HandleResume(CONNECTION *Connection, GENERIC_MSG_HEADER *Message);

    void HandleNamedPause(CONNECTION *Connection, GENERIC_MSG_HEADER *Message);

    void HandleNamedResume(CONNECTION *Connection,
			   GENERIC_MSG_HEADER *Message);

    void HandleKillSPEEDES(CONNECTION *Connection,
			   GENERIC_MSG_HEADER *Message);

    void HandleQuery(CONNECTION *Connection, GENERIC_MSG_HEADER *Message);

    void HandleQueryReply(CONNECTION *Connection, GENERIC_MSG_HEADER *Message);

    void HandleCommand(CONNECTION *Connection, GENERIC_MSG_HEADER *Message);

    void HandleCommandReply(CONNECTION *Connection,
			    GENERIC_MSG_HEADER *Message);

    void HandleTwoWayCommand(CONNECTION *Connection,
			     GENERIC_MSG_HEADER *Message);

    void HandleTime(CONNECTION *Connection, GENERIC_MSG_HEADER *Message);
    void HandleTimeReply(CONNECTION *Connection, GENERIC_MSG_HEADER *Message);

    void HandleMinMsgTime(CONNECTION *Connection, GENERIC_MSG_HEADER *Message);

    void HandleMinMsgTimeReply(CONNECTION *Connection,
			       GENERIC_MSG_HEADER *Message);

    void HandleTimeAdvanceRequest(CONNECTION *Connection,
				  GENERIC_MSG_HEADER *Message);

    void HandleTimeAdvanceGrant(CONNECTION *Connection,
				GENERIC_MSG_HEADER *Message);

    void HandleGoToTimeRequest(CONNECTION *Connection,
				  GENERIC_MSG_HEADER *Message);

    void HandleGoToTimeGrant(CONNECTION *Connection,
				GENERIC_MSG_HEADER *Message);

    void HandleSpeedesToExternalModule(CONNECTION *Connection,
				       GENERIC_MSG_HEADER *Message);

    void HandleSubscribe(CONNECTION *Connection,
			 GENERIC_MSG_HEADER *Message);

    void HandleUnSubscribe(CONNECTION *Connection,
			   GENERIC_MSG_HEADER *Message);

    void HandleSubscribedData(CONNECTION *Connection,
			      GENERIC_MSG_HEADER *Message);

    void HandleScheduleEvent(CONNECTION *Connection,
			     GENERIC_MSG_HEADER *Message);

    void HandleCancelEvent(CONNECTION *Connection,
			   GENERIC_MSG_HEADER *Message);

    void HandleScheduleEventReply(CONNECTION *Connection,
				  GENERIC_MSG_HEADER *Message);

    void HandleCancelEventReply(CONNECTION *Connection,
				GENERIC_MSG_HEADER *Message);

    void UnSubscribeAll(int ExternalId);

};

#endif
