// SpHostRouterMsgs.H header file defines all of the message types for the
// Host Router

#ifndef HOST_ROUTER_MESSAGES_HEADER
#define HOST_ROUTER_MESSAGES_HEADER

#include "SpMsgHeader.H"
#include "SpNetTypes.H"
#include "SpSimTime.H"

enum  {
  NOT_SCHEDULED_BECAUSE_IN_PAST,
  EVENT_SCHEDULED,
  EVENT_SCHEDULED_AT_GVT,
  EVENT_HANDLER_SCHEDULED,
  EVENT_HANDLER_SCHEDULED_AT_GVT,
  OBJECT_SPECIFIED_DOES_NOT_EXIST
};

enum {
  CANCELED_EVENT_OK,
  CANCEL_EVENT_NOT_FOUND
};

enum {
  AGGRESSIVE_BARRIER,
  SYNCHRONIZED_BARRIER,
  NO_BARRIER
};

enum {
  COMMAND_STATUS_GOOD,
  COMMAND_STATUS_BAD_OBJECT,
  COMMAND_STATUS_BAD_EVENT,
  COMMAND_STATUS_ABORT
};

enum {
  SUBSCRIBE_STATUS_GOOD,
  SUBSCRIBE_STATUS_REDUNDANT_SUBSCRIPTION,
  SUBSCRIBE_STATUS_BAD_OBJECT,
  SUBSCRIBE_STATUS_BAD_TYPE,
  SUBSCRIBE_STATUS_BAD_SUBSCRIPTION_NAME
};

enum {
  UNSUBSCRIBE_STATUS_GOOD,
  UNSUBSCRIBE_STATUS_REDUNDANT_SUBSCRIPTION,
  UNSUBSCRIBE_STATUS_BAD_OBJECT,
  UNSUBSCRIBE_STATUS_BAD_TYPE,
  UNSUBSCRIBE_STATUS_BAD_SUBSCRIPTION_NAME
};

enum {
  SPEEDES_CONNECTION_MESSAGE_ID,
  EXTERNAL_MODULE_CONNECTION_MESSAGE_ID,
  DISCONNECT_BARRIER_MESSAGE_ID,
  NAME_LOOKUP_MESSAGE_ID,
  OBJECT_NAME_MESSAGE_ID,
  OBJECT_TYPE_MESSAGE_ID,
  NUM_NODES_MESSAGE_ID,
  NEXT_NODE_MESSAGE_ID,
  FLUSH_MESSAGE_ID,
  PAUSE_MESSAGE_ID,
  RESUME_MESSAGE_ID,
  NAMED_PAUSE_MESSAGE_ID,
  NAMED_RESUME_MESSAGE_ID,
  KILL_SPEEDES_MESSAGE_ID,
  QUERY_MESSAGE_ID,
  QUERY_REPLY_MESSAGE_ID,
  COMMAND_MESSAGE_ID,
  COMMAND_REPLY_MESSAGE_ID,
  TWO_WAY_COMMAND_MESSAGE_ID,
  TIME_MESSAGE_ID,
  TIME_REPLY_MESSAGE_ID,
  TIME_ADVANCE_REQUEST_MESSAGE_ID,
  TIME_ADVANCE_GRANT_MESSAGE_ID,
  GO_TO_TIME_REQUEST_MESSAGE_ID,
  GO_TO_TIME_GRANT_MESSAGE_ID,
  SPEEDES_TO_EXTERNAL_MODULE_MESSAGE_ID,
  SUBSCRIBE_MESSAGE_ID,
  UNSUBSCRIBE_MESSAGE_ID,
  SUBSCRIBED_DATA_MESSAGE_ID,
  SCHEDULE_EVENT_ID,
  SCHEDULE_EVENT_REPLY_ID,
  CANCEL_EVENT_ID,
  CANCEL_EVENT_REPLY_ID,
  MIN_MSG_TIME_ID,
  MIN_MSG_TIME_REPLY_ID,
  N_HOST_ROUTER_MESSAGES
};

class REPLY_WITH_ID : public GENERIC_MSG_HEADER {
public:
  NET_INT ExternalId;

  REPLY_WITH_ID():ExternalId(0){}
  void Init(){
    GENERIC_MSG_HEADER::Init();
    ExternalId = 0;
  }
};

class REPLY_WITH_TIME : public GENERIC_MSG_HEADER {
public:
  SpSimTime Time;

  REPLY_WITH_TIME():Time(0){}
  void Init(){
    GENERIC_MSG_HEADER::Init();
    Time = 0;
  }
};

class REPLY_WITH_ID_AND_TIME : public GENERIC_MSG_HEADER{
private:
  int EightByteAlignmentData;
public:
  NET_INT ExternalId;
  SpSimTime Time;

  REPLY_WITH_ID_AND_TIME():EightByteAlignmentData(0), ExternalId(0), Time(0){}
  void Init(){
    GENERIC_MSG_HEADER::Init();
    Time = 0;
    ExternalId = 0;
  }
};


class SPEEDES_CONNECTION_MESSAGE : public GENERIC_MSG_HEADER {
public:
  NET_INT Node;
  NET_INT Nnodes;

  SPEEDES_CONNECTION_MESSAGE():Node(0), Nnodes(0){}
  void Init(){
    GENERIC_MSG_HEADER::Init();
    Node = 0;
    Nnodes = 0;
  }
};

class DISCONNECT_BARRIER_MESSAGE : public REPLY_WITH_ID {
private:
  int EightByteAlignmentData;
public:
  DISCONNECT_BARRIER_MESSAGE():EightByteAlignmentData(0){}
};

class EXTERNAL_MODULE_CONNECTION_MESSAGE : public GENERIC_MSG_HEADER {
};

class NAME_LOOKUP_MESSAGE : public GENERIC_MSG_HEADER {
public:
  NET_INT GlobalId;

  NAME_LOOKUP_MESSAGE():GlobalId(0){}
  void Init(){
    GENERIC_MSG_HEADER::Init();
    GlobalId = 0;
  }
};

class OBJECT_NAME_MESSAGE : public GENERIC_MSG_HEADER {
public:
  NET_INT ObjectType;
  NET_INT NameFlag;

  OBJECT_NAME_MESSAGE():ObjectType(0), NameFlag(0){}
  void Init(){
    GENERIC_MSG_HEADER::Init();
    ObjectType = 0;
    NameFlag = 0;
  }
};

class OBJECT_TYPE_MESSAGE : public GENERIC_MSG_HEADER {
};

class NUM_NODES_MESSAGE : public GENERIC_MSG_HEADER {
  public:
    NET_INT NumNodes;
};

class NEXT_NODE_MESSAGE : public GENERIC_MSG_HEADER {
  public:
    NET_INT NextNode;
};

class FLUSH_MESSAGE : public GENERIC_MSG_HEADER {
};

class PAUSE_MESSAGE : public GENERIC_MSG_HEADER {
};

class RESUME_MESSAGE : public GENERIC_MSG_HEADER {
};

class NAMED_PAUSE_MESSAGE : public REPLY_WITH_TIME {
};

class NAMED_RESUME_MESSAGE : public GENERIC_MSG_HEADER {
};

class KILL_SPEEDES_MESSAGE : public GENERIC_MSG_HEADER {
};

class QUERY_MESSAGE : public REPLY_WITH_ID {
private:
  int EightByteAlignmentData;
public:
  NET_INT ObjectType;
  NET_INT LocalId;
  NET_INT GlobalId;

  QUERY_MESSAGE():EightByteAlignmentData(0), ObjectType(-1), LocalId(-1),
    GlobalId(-1) {}
  void Init(){
    REPLY_WITH_ID::Init();
    ObjectType = -1;
    LocalId = -1;
    GlobalId = -1;
  }
};

class QUERY_REPLY_MESSAGE : public REPLY_WITH_ID {
private:
  int EightByteAlignmentData;
public:
  QUERY_REPLY_MESSAGE():EightByteAlignmentData(0){}
};

class COMMAND_MESSAGE : public QUERY_MESSAGE{
  public:
    NET_INT Node;
    NET_INT ReturnStatus;
    SpSimTime Time;

    COMMAND_MESSAGE():Time(-1e20){}
    void Init(){
      QUERY_MESSAGE::Init();
      Time = -1e20;
    }
};

class TWO_WAY_COMMAND_MESSAGE : public COMMAND_MESSAGE {
public:
  NET_INT BarrierMode;
  NET_DOUBLE TimeLag;

  TWO_WAY_COMMAND_MESSAGE(): BarrierMode(-1), TimeLag(0){}
  void Init(){
    COMMAND_MESSAGE::Init();
    BarrierMode = -1;
    TimeLag = -1;
  }
};

class COMMAND_REPLY_MESSAGE : public REPLY_WITH_ID_AND_TIME {
public:
  NET_INT CommandStatus;

  COMMAND_REPLY_MESSAGE():CommandStatus(-1){}
  void Init(){
    REPLY_WITH_ID_AND_TIME::Init();
    CommandStatus = -1;
  }
};

class TIME_MESSAGE : public REPLY_WITH_ID {
private:
  int EightByteAlignmentData;
public:
  TIME_MESSAGE():EightByteAlignmentData(0){}
};

class TIME_REPLY_MESSAGE : public REPLY_WITH_ID_AND_TIME {
};

class MIN_MSG_TIME_MESSAGE : public REPLY_WITH_ID {
};

class MIN_MSG_TIME_REPLY_MESSAGE : public REPLY_WITH_ID_AND_TIME {
};

class TIME_ADVANCE_REQUEST_MESSAGE : public REPLY_WITH_ID {
private:
  int EightByteAlignmentData;
public:
  SpSimTime OldTime;
  NET_DOUBLE TimeLag;

  TIME_ADVANCE_REQUEST_MESSAGE():EightByteAlignmentData(0),OldTime(0),
    TimeLag(0){}
  void Init(){
    REPLY_WITH_ID::Init();
    OldTime = 0;
    TimeLag = 0;
  }
};

class TIME_ADVANCE_GRANT_MESSAGE : public REPLY_WITH_ID_AND_TIME {
};

class SPEEDES_TO_EXTERNAL_MODULE_MESSAGE : public REPLY_WITH_ID_AND_TIME {
public:
  NET_INT GlobalId;

  SPEEDES_TO_EXTERNAL_MODULE_MESSAGE():GlobalId(-1){}
  void Init(){
    REPLY_WITH_ID_AND_TIME::Init();
    GlobalId = -1;
  }
};

class SUBSCRIBE_MESSAGE : public GENERIC_MSG_HEADER {
public:
};

class UNSUBSCRIBE_MESSAGE : public GENERIC_MSG_HEADER {
public:
};

class SUBSCRIBED_DATA_MESSAGE : public REPLY_WITH_TIME {
public:
  NET_INT GlobalId;

  SUBSCRIBED_DATA_MESSAGE():GlobalId(-1){}
  void Init(){
    REPLY_WITH_TIME::Init();
    GlobalId = -1;
  }
};


class GO_TO_TIME_REQUEST_MESSAGE : public REPLY_WITH_ID {
public:
  SpSimTime NewTime;
  NET_DOUBLE TimeLag;
  GO_TO_TIME_REQUEST_MESSAGE():NewTime(0), TimeLag(0){}
  void Init(){
    REPLY_WITH_ID::Init();
    NewTime = 0;
    TimeLag = 0;
  }
};

class GO_TO_TIME_GRANT_MESSAGE : public REPLY_WITH_ID_AND_TIME {
};


enum TimeMode{
  IF_IN_PAST_SCHEDULE_AT_GVT,
  IF_IN_PAST_IGNORE
};

class SCHEDULE_EVENT_MESSAGE : public REPLY_WITH_ID_AND_TIME {
public:
  NET_INT LocalId;    //Set in the HostRouter
  NET_INT ObjectType; //Set in the HostRouter
  NET_INT ObjectNode; //Set in the HostRouter

  NET_INT GlobalId;
  NET_INT MessageBytes;
  NET_INT DataBytes;
  TimeMode ScheduleMode;
  NET_INT CancelHandleRequested;

  SCHEDULE_EVENT_MESSAGE():LocalId(-1), ObjectType(-1), ObjectNode(-1),
    GlobalId(-1), MessageBytes(-1), DataBytes(-1),
    ScheduleMode(IF_IN_PAST_IGNORE), CancelHandleRequested(0){}
  void Init(){
    REPLY_WITH_ID_AND_TIME::Init();
    LocalId = -1;
    ObjectType = -1;
    GlobalId = -1;
    MessageBytes = -1;
    DataBytes = -1;
    ScheduleMode = IF_IN_PAST_IGNORE;
    CancelHandleRequested = 0;
  }
};

class SCHEDULE_EVENT_REPLY_MESSAGE : public REPLY_WITH_ID_AND_TIME {
public:
  NET_INT CancelHandle;
  NET_INT ErrorCode;
  SCHEDULE_EVENT_REPLY_MESSAGE():CancelHandle(0), ErrorCode(0){}
  void Init(){
    REPLY_WITH_ID_AND_TIME::Init();
    CancelHandle = 0;
    ErrorCode = 0;
  }
};


class CANCEL_EVENT_MESSAGE : public  REPLY_WITH_ID{
public:
  NET_INT GlobalId;
  NET_INT CancelHandle;
  NET_INT ReplyRequested;
  CANCEL_EVENT_MESSAGE():GlobalId(-1),CancelHandle(-1),ReplyRequested(-1){}
  void Init(){
    REPLY_WITH_ID::Init();
    CancelHandle = -1;
    ReplyRequested = -1;
    GlobalId = -1;
  }
};

class CANCEL_EVENT_REPLY_MESSAGE : public REPLY_WITH_ID {
private:
  NET_INT EightByteAlignmentGarbage;  //geb3 fix
public:
  NET_INT ErrorCode;
  CANCEL_EVENT_REPLY_MESSAGE():EightByteAlignmentGarbage(0),ErrorCode(-1){}
  void Init(){
    REPLY_WITH_ID::Init();
    ErrorCode = -1;
  }
};

#endif
