// SpSubscriptionName.H header file defines a subscription by name. This class
// then manages which types of objects are subscribing and/or which specific
// objects are chosen.

#ifndef SUBSCRIPTION_NAME_OBJ
#define SUBSCRIPTION_NAME_OBJ

#include "SpStdlib.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpString.H"
#include "queue.H"
#include "qitem.H"
#include "SpSubscriptionItem.H"

enum {
  SUBSCRIBE_NONE,
  SUBSCRIBE_ALL,
  SUBSCRIBE_TYPE,
  SUBSCRIBE_OBJECT
};

class SUBSCRIPTION_NAME : public C_ITEM {

  private:

    char *Name;
    C_QUEUE SubItems;

  protected:

  public:

    SUBSCRIPTION_NAME(char *name) {
      Name = strdup(name);
    }

    SUBSCRIPTION_NAME(char *name, SUBSCRIPTION_ITEM *SubItem) {
      Name = strdup(name);
      SubItems.push_bot(SubItem);
    }

    ~SUBSCRIPTION_NAME() {
      if (Name) free(Name);
      Name = NULL;
      while (SubItems.get_length()) {
	delete (SUBSCRIPTION_ITEM *)(SubItems.pop_top());
      }
    }

    virtual char *GetName() {return Name;}
    C_QUEUE *GetSubItems() {return &SubItems;}

    int AddSubItem(SUBSCRIPTION_ITEM *SubItem);
    int RemoveSubItem(SUBSCRIPTION_ITEM *SubItem);

    int Compare(SUBSCRIPTION_ITEM *SubItem) {
      int value = 0;
      SUBSCRIPTION_ITEM *item = (SUBSCRIPTION_ITEM *)SubItems.get_top();
      int len = SubItems.get_length();

      for (int i=0; i<len; i++) {
	if (SubItem->Compare(item)) {
	  value = 1;
	  break;
	}
	item = (SUBSCRIPTION_ITEM *)item->get_link();
      }

      return value;
    }

    int CheckSubSet(SUBSCRIPTION_ITEM *SubItem);

};


#endif
