#include <unistd.h>
#include "SpAssert.H"
#include <sys/wait.h>
#include "SpSignal.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpStdio.H"
#include "SpString.H"
#include <sys/types.h>

void my_freopen(ostrstream &os, const char * const output) {
  assert(os.good());
  char *s = os.str();
  assert(s != NULL);
  FILE *fp = freopen(s, "w",
  		     strcmp(output, "stderr") ? stdout : stderr);
  if (fp == NULL)
  {
    cerr << "freopen could not open " << s;
    cerr.flush();
    perror(": ");
    assert(0);
  }
}

void save_node_output
  (const char * const dir, const int node, const char * const output)
{
  ostrstream os;
  os << dir << "/application_node_" << node << '_' << output << ends;
  my_freopen(os, output);
}

void save_server_output(const char * const dir, const char * const output)
{
  ostrstream os;
  os << dir << "/SpeedesServer_" << output << ends;
  my_freopen(os, output);
}

// syntax for running this program is:
//
// TestSpeedesServer num_nodes SpeedesServer_path application_path output_dir

int main(int argc, char *argv[])
{
  assert(argc == 5);

  const pid_t SpeedesServer_pid = fork();
  if (SpeedesServer_pid == 0)
  {
    save_server_output(argv[4], "stdout");
    save_server_output(argv[4], "stderr");
    
    execl(argv[2], argv[2], "7654", (char*)NULL);
    assert(0);
  }

  int num_nodes;
  istrstream is(argv[1]);
  is >> num_nodes;
  
  int node;
  
  pid_t *application_node_pid = new pid_t[num_nodes];
  for (node = 0; node < num_nodes; node++)
  {
    application_node_pid[node] = fork();
    if (application_node_pid[node] == 0)
    {
      save_node_output(argv[4], node, "stdout");
      save_node_output(argv[4], node, "stderr");
      
      ostrstream os;
      os << num_nodes << ends;

      execl(argv[3], argv[3], "-l", "1", "-n", os.str(), (char*)NULL);
      assert(0);
    }
  }
  
  for (node = 0; node < num_nodes; node++)
  {
    int status;
    const pid_t child = waitpid(application_node_pid[node], &status, 0);
    assert(child != -1);
    if (!WIFEXITED(status))
    {
      cout << "Error: node " << node << " exited with abnormal status of "
	   << status << endl;
      assert(status != 0);
      return status;
    }
  }

  const int failed_p = kill(SpeedesServer_pid, SIGINT);
  assert(!failed_p);
  
  delete [] application_node_pid;
  
  return 0; // 0 means success
}
