// SpFilterTrace.C file

#include "SpStdlib.H"
#include "SpString.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"

/*
** Useful utility for finding strings embedded in another string
*/

char *FindString(char *line, char *string) {
  int i,j,n1,n2;

  n1 = strlen(line);
  n2 = strlen(string);

  for (i=0; i<=n1-n2; i++) {
    for (j=0; j<n2; j++) {
      if (line[i+j] != string[j]) break;
      if (j==n2-1) return &line[i];
    }
  }

  return NULL;
}

/*
** Class for managing a list of lines
*/

class LoL {
  public:
    char Line[256];
    LoL *Next;
    LoL() {Next = NULL;}
};

/*
** FilterTrace: filters a trace file using the command line argument passed
** into the call.
*/

int main(int argc, char **argv) {
  int i;
  int FoundFlag;
  int FoundInTag;
  char Time[256];
  char EvtName[256];
  char ObjName[256];
  char ObjType[256];
  char Handle[256];
  char GlobalId[256];
  char RB_Items[256];
  char EvtCPU[256];
  char TotCmtd[256];
  char TotTime[256];
  char CritPth[256];
  char Tag[256];
  char Line[256];
  LoL *First;
  LoL *Last;
  LoL *Current;

/*
** Make sure the command-line arguments are ok
*/

  if (argc < 3) {
    cerr << "Error, FilterTrace <FileName> <String1> <String2> ... <StringN>"
	 << endl;
    exit(1);
  }

/*
** Open the input file (report error if file does not exist)
*/

  ifstream TraceIn(argv[1]);
  if (!TraceIn) {
    cerr << "Error, cannot open input file " << argv[1] << endl;
  }

/*
** Read in lines until we reach the end of the file
*/

  while (1) {
    TraceIn.getline(Time, 255);
    if (!TraceIn) break;

/*
** These lines are fixed (and known) in the trace file. It is only these lines
** and the tag lines that are used for filtering.
*/

    TraceIn.getline(EvtName, 255);
    TraceIn.getline(ObjName, 255);
    TraceIn.getline(ObjType, 255);
    TraceIn.getline(Handle, 255);
    TraceIn.getline(GlobalId, 255);
    TraceIn.getline(RB_Items, 255);
    TraceIn.getline(EvtCPU, 255);
    TraceIn.getline(TotCmtd, 255);
    TraceIn.getline(TotTime, 255);
    TraceIn.getline(CritPth, 255);
    TraceIn.getline(Tag, 255);

/*
** The tag lines either start with a BeginTag or a NoTag line. If the line is
** a BeginTag, then it is followed by whatever the application wrote using
** the tag utility. It is then followed by an EndTag line.
*/

    if (FindString(Tag, "BeginTag")) {
      First = new LoL();
      Last = First;
      Current = First;
      TraceIn.getline(Current->Line, 255);

      while (!FindString(Current->Line, "EndTag")) {
	Last = new LoL();
	Current->Next = Last;
	Current = Last;
	TraceIn.getline(Current->Line, 255);
      }

    }else{
      First = NULL;
    }

/*
** Now we search through the lines to see if there are matches. There must be
** at least one match for each of the command line arguments for the info to
** pass the output test. In other words, it's a big "and" logical condition.
*/

    FoundFlag = 1;
    for (i=2; i<argc; i++) {
      if (FindString(Time, argv[i])) {
      } else if (FindString(EvtName, argv[i]))  {
      } else if (FindString(ObjName, argv[i]))  {
      } else if (FindString(ObjType, argv[i]))  {
      } else if (FindString(Handle, argv[i]))   {
      } else if (FindString(GlobalId, argv[i])) {
      } else if (FindString(RB_Items, argv[i])) {
      } else if (FindString(EvtCPU, argv[i]))      {
      } else if (FindString(TotCmtd, argv[i]))      {
      } else if (FindString(TotTime, argv[i]))      {
      } else if (FindString(CritPth, argv[i]))      {
      } else if (FindString(Tag, argv[i]))      {
      } else {

	FoundInTag = 0;
	Current = First;
	while (Current) {
	  if (FindString(Current->Line, argv[i])) {
	    FoundInTag = 1;
	    break;
	  }
	  Current = Current->Next;
	}

	FoundFlag = FoundInTag;
	if (!FoundFlag) {
	  break;
	}

      }
    }

/*
** If we pass the output test, then we output the known lines first, and
** then output the tag info. Then, we output the scheduled event info. We
** know we have reached the end of that section when we reach a blank line.
*/

    if (FoundFlag) {
      cout << Time << endl;
      cout << EvtName << endl;
      cout << ObjName << endl;
      cout << ObjType << endl;
      cout << Handle << endl;
      cout << GlobalId << endl;
      cout << RB_Items << endl;
      cout << EvtCPU << endl;
      cout << TotCmtd << endl;
      cout << TotTime << endl;
      cout << CritPth << endl;
      cout << Tag << endl;

      Current = First;
      while (Current) {
	cout << Current->Line << endl;
	Last = Current->Next;
	delete Current;
	Current = Last;
      }

    }

    while (1) {
      TraceIn.getline(Line, 255);
      if (strlen(Line) == 0) break;
      if (FoundFlag) cout << Line << endl;
    }

    if (FoundFlag) cout << endl;

  }

/*
** The is the bottom of the loop where we jump back to work on the next line
*/

  return 0;
}
