// TwoWayCommand.C file

#include "SpWalltime.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpStdlib.H"
#include "SpString.H"
#include "SpSignal.H"

#include "SpEmHostUser.H"

SpEmHostUser *EmHostUser = NULL;

enum {
  GRANULARITY,
  MULTIPLE_OF_WALL_CLOCK,
  SYNCHRONIZED_FROM_START_OF_SIM
};

/*
** Handler routine for when (if) the user hits control C
*/

void sigintfunc(int) {
  cout << "Interupt" << endl;
  signal(SIGINT, sigintfunc);
  if (EmHostUser) delete EmHostUser;
  EmHostUser = NULL;
  exit(1);
}

void CommandLineArgs()
{
  cout<<endl<<"arguments are"<<endl<<endl;
  cout<<"           -m [num] where num is the multiple of wall clock"<<endl;
  cout<<"           -g [num] where num is the granularity of the clock"<<endl;
  cout<<"           --from_start to lock to wall from start of sim"<<endl;
  cout<<endl;
}


int Token(char * string)
{
  if (!strcmp(string, "-g")){
    return GRANULARITY;
  }
  if  (!strcmp(string, "-m")){
    return MULTIPLE_OF_WALL_CLOCK;
  }
  if  (!strcmp(string, "--from_start")){
    return SYNCHRONIZED_FROM_START_OF_SIM;
  }
  cout<<"Unknown argument "<<string<<endl;
  CommandLineArgs();
  exit(0);
  return 0;
}

void HandleCommandLine(double & Granularity, double & MultipleOfWallClock,
		       int & SyncFromStart, int argc, char **argv)
{
  int ArgC = 1;

  while (ArgC < argc){

    switch(Token(argv[ArgC]))
      {
      case GRANULARITY:
	Granularity = strtod(argv[++ArgC], NULL);
	++ArgC;
	break;

      case MULTIPLE_OF_WALL_CLOCK:
	MultipleOfWallClock = strtod(argv[++ArgC], NULL);
	++ArgC;
	break;
      case SYNCHRONIZED_FROM_START_OF_SIM:
	SyncFromStart = 1;
	++ArgC;
	break;
      }
  }
}

int main(int argc, char **argv) {
  double MultipleOfWallClock;
  double NewSimTime;
  double SleepTimeAdjustment = 0;
  double Granularity = .1;
  double SimTime0;
  double WallTime0;
  int SyncFromStart = 0;
  double AdvanceTime;
  double SleepTime;

  signal(SIGINT, sigintfunc);

  if (argc == 1){
    CommandLineArgs();
    exit(0);
  }
  else{
    HandleCommandLine(Granularity, MultipleOfWallClock, SyncFromStart,
		      argc, argv);
  }

  EmHostUser = new SpEmHostUser();

  EmHostUser->Pause();
  EmHostUser->GoToTimeRequest(0, 0);
  EmHostUser->Resume();
  while (!EmHostUser->GoToTimeGranted()){
    if (!EmHostUser->SpeedesExecuting()) {
      exit(1);
    }
  }
  SimTime0 = EmHostUser->GetTime();
  WallTime0 = SpWalltime();
  if (SyncFromStart){
    SimTime0 = 0;
  }
  WallTime0 = SpWalltime();

  cout<<"Starting with SimTime0 = "<<SimTime0<<endl;

  while (EmHostUser->SpeedesExecuting()){
    SleepTime = (Granularity + SleepTimeAdjustment);

    if (SleepTimeAdjustment < 0){
      // If SleepTime < 0, we are running too slow and need to speed up the
      // simulation.
      AdvanceTime = (MultipleOfWallClock - SleepTimeAdjustment)*Granularity;
    }
    else{
      AdvanceTime = (MultipleOfWallClock)*Granularity;
    }

    if (AdvanceTime < 0){
      AdvanceTime = 0;
    }

    EmHostUser->TimeAdvanceRequest(AdvanceTime);

    cout<<"Advance "<<AdvanceTime
	<<" sleeping "<<SleepTime<<" seconds";

    if (SleepTime >= 0){
      struct timeval NOP_Timeout;

      NOP_Timeout.tv_sec = 0;
      NOP_Timeout.tv_usec = (int) ((SleepTime)*1000000);
      select(0, NULL, NULL, NULL, &NOP_Timeout);

      //usleep((int) ((SleepTime)*1000000));
    }
    while (!EmHostUser->TimeAdvanceGranted());
    NewSimTime = EmHostUser->GetTime();
    SleepTimeAdjustment = ((NewSimTime-SimTime0)/MultipleOfWallClock -
			   (SpWalltime() - WallTime0));

    cout << " SleepTimeAdjustment: " << SleepTimeAdjustment << endl;
  }

  return 0;
}
