#include "E_SpInitial.H"
#include "SpAntiMsg.H"
#include "SpEventTypes.H"

E_SpInitial::E_SpInitial() {

  GLOBAL_ID = 0;
      
  cout << "Creating initial event" << endl;
  MESSAGE = (SpMsg *) (new char[sizeof(SpMsg) + sizeof(int)]);

  set_from_SpComm(); // i.e., this message should not be reused by SPEEDES
      
  MESSAGE->time_tag.SetTime(-1.0e20);
  MESSAGE->time_tag.SetPriority1(0x80000000);
  MESSAGE->time_tag.SetPriority2(0x80000000);
  MESSAGE->time_tag.SetCounter(0x80000000);
  MESSAGE->time_tag.SetUniqueId(0x80000000);

  MESSAGE->evtype = -1;

  MESSAGE->obtype = -1;
  MESSAGE->objid = -1;
  MESSAGE->bytes = sizeof(int);

  MESSAGE->EarliestStartTime = 0.0;

  TIME_TAG = MESSAGE->TIME_TAG;
  OBJECT_TYPE = -1;
  LOCAL_ID = -1;
  EVENT_TYPE = -1;
}

#include "SpList.H"
#include "holder.H"

extern SpClusterMgr *GlobalClusterMgr;
extern SpEvent      *CurrentEvent;
extern SpSimObj     *CurrentSimobj;
extern SpTag        *CurrentTag;

void
E_SpInitial::commit(){}

void
E_SpInitial::process(){
  ProcessDM();
  ProcessDDM();
}

#include "SpExecute.H"
#include "SpWalltime.H"
#include "SpScheduleEvent.H"

void
E_SpInitial::ProcessDM(){
  const int num_phase = 8;
  double t[num_phase];
  int i;

  i=0;
  t[i++] = SpWalltime(); ProcessEvent(evtype->get_messid("DEFAULT_SUBSCRIBE_HLA"));
  t[i++] = SpWalltime(); ProcessEvent(evtype->get_messid("DEFAULT_PUBLISH_HLA"));
  t[i++] = SpWalltime(); ProcessEvent(evtype->get_messid("SUBSCRIBE_DECMAN"));
  t[i++] = SpWalltime(); ProcessEvent(evtype->get_messid("PUBLISH_DECMAN"));
  t[i++] = SpWalltime(); ProcessEvent(evtype->get_messid("ADD_SUBSCRIBER_HLA"));
  t[i++] = SpWalltime(); ProcessEvent(evtype->get_messid("NEW_PROXY_LOCAL_PROXYMAN"));
  t[i++] = SpWalltime(); ProcessEvent(evtype->get_messid("DELIVER_PROXY_PTR"));
  t[i++] = SpWalltime();

  cout << "Coord proc node " << get_NODE() << " etime ";
  for (i=1; i<num_phase; i++)
    cout << t[i] - t[i-1] << ", ";
  cout << " total " << t[num_phase-1] - t[0];
  cout << endl;
}

void
E_SpInitial::ProcessDDM(){
  const int num_phase = 9;
  double t[num_phase];
  int i;

  i=0;
  t[i++] = SpWalltime(); ProcessEvent(evtype->get_messid("PublishSpace"));
  t[i++] = SpWalltime(); ProcessEvent(evtype->get_messid("SubscribeSpace"));
  t[i++] = SpWalltime(); ProcessEvent(evtype->get_messid("SubscribeGrid"));
  t[i++] = SpWalltime(); ProcessEvent(evtype->get_messid("PublishGrid"));
  t[i++] = SpWalltime(); ProcessEvent(evtype->get_messid("AddSubscriberSpace"));
  t[i++] = SpWalltime(); ProcessEvent(evtype->get_messid("PublishSpace"));
  t[i++] = SpWalltime(); ProcessEvent(evtype->get_messid("NEW_PROXY_LOCAL_PROXYMAN"));
  t[i++] = SpWalltime(); ProcessEvent(evtype->get_messid("DELIVER_PROXY_PTR"));
  t[i++] = SpWalltime();

  cout << "Coord proc node " << get_NODE() << " etime ";
  for (i=1; i<num_phase; i++)
    cout << t[i] - t[i-1] << ", ";
  cout << " total time " << t[num_phase-1] - t[0];
  cout << endl;
}

void
E_SpInitial::ProcessEvent(int event_type) {
  int len;
  int num_proc, num_sent, num_local, num_rcv;
  char *message;
  SpMsg *header;
  SpList local_no_risk, no_risk;
  C_HOLDER *holder;
  SpEvent  *orig_event  = CurrentEvent;
  SpSimObj *orig_simobj = CurrentSimobj;
  SpTag    *orig_tag    = CurrentTag;

  if (event_type<0) return;

  num_rcv = num_local = num_sent = num_proc = 0;

  while (get_mque()->length()){
    holder  = (C_HOLDER *) get_mque()->pop_top();
    message = holder->get_buff();
    header  = (SpMsg *) message;

    if (header->evtype!=event_type){
      no_risk.Insert(holder);
      continue;
    }

    if (holder->get_node()==get_NODE()){
      local_no_risk.Insert(holder);
      num_local++; num_proc++;
    }
    else{
      SpComm_CoordinatedSend(holder->get_node(), holder->get_size(), holder->get_buff());
      evtype->delete_message(holder);
      num_sent++;
    }
  }

  while ((holder=(C_HOLDER *) no_risk.RemoveTop()))
    get_mque()->push_bot(holder);

  while ((holder=(C_HOLDER *) local_no_risk.RemoveTop())){
    ProcessEvent(holder->get_buff());
    evtype->delete_holder(holder);
  }

  while((message = (char *)SpComm_CoordinatedReceive(len)) != (char *) 0){
    ProcessEvent(message, 1);
    num_proc++;
    num_rcv++;
  }

  cout << "Coord proc "
          << evtype->get_messname(event_type)
          << " node " << get_NODE() << " proc "  << num_proc 
          << " sent " << num_sent   << " local " << num_local 
          << " rcv "  << num_rcv    << endl;

  CurrentEvent  = orig_event;
  CurrentSimobj = orig_simobj;
  CurrentTag    = orig_tag;
}

void 
E_SpInitial::ProcessEvent(char *message, int from_comm){
  SpEvent *eptr;
  SpMsg *header = (SpMsg *) message;

  eptr = (SpEvent *) evtype->message_event(message);
  eptr->set_rbid(header->rbid);
  if (from_comm)
    eptr->set_from_SpComm();
  CurrentEvent  = eptr;
  CurrentSimobj = eptr->get_SIMOBJ();
  CurrentTag    = eptr->GetTag();
  eptr->process();
  TransferMsg(eptr);
  evtype->delete_event(eptr);
}

void
E_SpInitial::TransferMsg(SpEvent *event){
  get_oque()->concat(event->get_oque());
  get_amque()->concat(event->get_amque());
  get_mque()->concat(event->get_mque());
}

/************************************************************************
* Trace : Trace this event						*
************************************************************************/
void 
E_SpInitial::Trace() {
  int i;
  int len;
  char *message;
  C_HOLDER *holder;
  SpMsg *header;
  C_LPTR *lptr;
  SpEvent *event;
  SpAntiMsg *anti;

  cout << TIME_TAG << endl;
  cout << "EvtName: " << evtype->get_messname(EVENT_TYPE) << endl;
  cout << "ObjName: " << SIMOBJ->get_NAME() << endl;
  cout << "ObjType: " << objects->get_obname(OBJECT_TYPE) << endl;
  cout << "Handle:  "
	    << "(" << NODE
	    << "," << OBJECT_TYPE
	    << "," << LOCAL_ID
	    << ")" << endl;
  cout << "GlobId:  " << GLOBAL_ID << endl;

  if (Tag.GetBuff()) {
    cout << "BeginTag:" << endl;
    cout << Tag;
    cout << "EndTag:" << endl;
  }else{
    cout << "NoTag:" << endl;
  }

/*
** Events scheduled locally
*/

  cout << "num local, risk, no risk : " 
       << get_oque()->length() << ", "
       << get_amque()->length() << ", "
       << get_mque()->length() << endl;

  len = get_oque()->length();
  lptr = (C_LPTR *)get_oque()->get_top();
  for (i=0; i<len; i++) {
    event = (SpEvent *)lptr->get_ptr();
    cout << "  LocSched: " << event->get_time_tag() << endl;
    cout << "    EvtName:  "
	      << evtype->get_messname(event->get_EVENT_TYPE()) << endl;

    cout << "    ObjType:  " << event->get_OBJECT_TYPE() << endl;
    cout << "    Handle:  "
	      << "(" << NODE
	      << "," << event->get_OBJECT_TYPE()
	      << "," << event->get_LOCAL_ID()
	      << ")" << endl;

    cout << "    bytes:    " << (event->get_MESSAGE())->bytes
                     + evtype->get_sizeof(event->get_EVENT_TYPE()) << endl;
    lptr = (C_LPTR *)lptr->get_link();
  }

/*
** Messages sent with risk
*/

  len = get_amque()->length();
  anti = (SpAntiMsg *)get_amque()->get_top();
  for (i=0; i<len; i++) {
    cout << "  RskSched: " << anti->get_time_tag() << endl;
    cout << "    EvtName:  "
         << evtype->get_messname(GlobalClusterMgr->ConvertGlobalToLocalEventId(anti->get_etype())) 
         << endl;

    cout << "    Handle:  "
         << "(" << anti->get_node()
         << "," << GlobalClusterMgr->ConvertGlobalToLocalMgrId(anti->get_obtype())
         << "," << anti->get_obid()
         << ")" << endl;

    cout << "    bytes:    "
         << evtype->get_sizeof(GlobalClusterMgr->ConvertGlobalToLocalEventId(anti->get_etype()))
            + anti->get_bytes() << endl;

    anti = (SpAntiMsg *)anti->get_link();
  }

/*
** Messages sent without risk
*/

  len = get_mque()->length();
  holder = (C_HOLDER *)get_mque()->get_top();
  for (i=0; i<len; i++) {
    message = holder->get_buff();
    header = (SpMsg *)message;

    if (N_NODES > 1) {
      cout << "  SafeSched: " << header->time_tag << endl;
    }else{
      cout << "  LocSched: " << header->time_tag << endl;
    }

    cout << "    EvtName:  "
         << evtype->get_messname(header->evtype) << endl;

    cout << "    Handle:  "
         << "(" << holder->get_node()
         << "," << header->obtype
         << "," << header->objid
         << ")" << endl;

    cout << "    bytes:    "
         << evtype->get_sizeof(header->evtype) + header->bytes << endl;

    holder = (C_HOLDER *)holder->get_link();
  }


/*
** Put a blank line after the event info
*/

  cout << endl;
}
