#ifndef SpAltMgr_Header
#define SpAltMgr_Header

#include "SpAlt.H"
#include "SpItemQueue.H"

DEFINE_ITEM_QUEUE( SpAlt )

/************************************************************************
 *
 ************************************************************************/
class SpAltMgr {
public:
  SpAltMgr() : NumAltItems( 0 ) {}

  int GetOptimistic();
  void SetOptimistic();
  void UnsetOptimistic();
  int GetNumAltItems();

  void Append( SpAlt* altItem );
  void Insert( SpAlt* altItem );
  SpAlt* Remove();

  void Cleanup();
  int Rollback();
private:
  static int Optimistic;
  int NumAltItems;
  ITEM_QUEUE( SpAlt ) Queue;
};

/************************************************************************
 *
 ************************************************************************/
inline
int
SpAltMgr::GetOptimistic() {
  return Optimistic;
}

/************************************************************************
 *
 ************************************************************************/
inline
void
SpAltMgr::SetOptimistic() {
  Optimistic = 1;
}

/************************************************************************
 *
 ************************************************************************/
inline
void
SpAltMgr::UnsetOptimistic() {
  Optimistic = 0;
}

/************************************************************************
 *
 ************************************************************************/
inline
int
SpAltMgr::GetNumAltItems() {
  return NumAltItems;
}

/************************************************************************
 *
 ************************************************************************/
inline
void
SpAltMgr::Append( SpAlt* altItem ) {
  NumAltItems++;
  Queue.Append( altItem );
}

/************************************************************************
 *
 ************************************************************************/
inline
void
SpAltMgr::Insert( SpAlt* altItem ) {
  NumAltItems++;
  Queue.Insert( altItem );
}

/************************************************************************
 *
 ************************************************************************/
inline
SpAlt*
SpAltMgr::Remove() {
  NumAltItems--;
  return Queue.Remove();
}

/************************************************************************
 *
 ************************************************************************/
inline
void
SpAltMgr::Cleanup() {
  SpAlt* altItem;

  while( (altItem = Queue.Remove()) ) {
    altItem->Cleanup();
    DEALLOCATE_SpAlt( altItem );
  }

  NumAltItems = 0;
}

/************************************************************************
 *
 ************************************************************************/
inline
int
SpAltMgr::Rollback() {
  int counter = 0;
  SpAlt* altItem;
  ITEM_QUEUE( SpAlt ) reverseQueue;

  while( (altItem = Queue.Remove()) ) {
    altItem->Rollback();
    reverseQueue.Insert( altItem );
    counter++;
  }
  Queue = reverseQueue;

  return counter;
}

extern SpAltMgr* SpCurrentAltMgr;

#endif

