// SpClusterMgr.C source file

#include "SpClusterMgr.H"
#include "SpComm.H"
#include "SpPoDataBase.H"

extern SpPoDataBase* DefaultDataBase;

SpClusterMgr *GlobalClusterMgr = NULL;
SpClusterMgr *SpClusterMgr::ClusterMgr = NULL;


void SpClusterMgr::SetClusterMgr() {
  GlobalClusterMgr = this;
  ClusterMgr = this;
}


SpClusterMgr::SpClusterMgr(): 
  GlobalToLocalMgrId(NULL),
  LocalToGlobalMgrId(NULL),
  NumLocalMgrs(0),
  NumGlobalMgrs(0),
  HighestMgrId(0),
  HighestEventId(0),
  GlobalToLocalEventId(NULL),
  LocalToGlobalEventId(NULL),
  NumLocalEvents(0),
  NumGlobalEvents(0)
{
  SetClusterMgr();
}

inline int SpClusterMgr::SizeOfClusterBuff() {
  int numBytes = sizeof(NET_INT);
  SpClusterInfo *clusterInfo = (SpClusterInfo *)Clusters.GetFirstElement();
  while (clusterInfo) {
    numBytes += clusterInfo->SizeOfBuff();
    clusterInfo = (SpClusterInfo *)Clusters.GetNextElement();
  }
  return numBytes;
}

inline int SpClusterMgr::SizeOfEventBuff() {
  int numBytes = sizeof(NET_INT);
  SpEventInfo *eventInfo = (SpEventInfo *)Events.GetFirstElement();
  while (eventInfo) {
    numBytes += eventInfo->SizeOfBuff();
    eventInfo = (SpEventInfo *)Events.GetNextElement();
  }
  return numBytes;
}

inline int SpClusterMgr::SizeOfMgrBuff() {
  int numBytes = sizeof(NET_INT);
  SpMgrInfo *mgrInfo = (SpMgrInfo *)Mgrs.GetFirstElement();
  while (mgrInfo) {
    numBytes += mgrInfo->SizeOfBuff();
    mgrInfo = (SpMgrInfo *)Mgrs.GetNextElement();
  }
  return numBytes;
}

void SpClusterMgr::GetClusterInfo(char *clusterName,
				  int &numNodesInCluster,
				  int &clusterNodeId) {
  SpClusterInfo *clusterInfo =
    (SpClusterInfo *)Clusters.Find(clusterName);
  if (clusterInfo) {
    numNodesInCluster = clusterInfo->GetNumNodesInCluster();
    clusterNodeId = clusterInfo->GetClusterNodeId();
  }else{
    cout << "Warning, cluster "
	 << clusterName
	 << " has not been plugged in"
	 << endl;
    numNodesInCluster = -1;
    clusterNodeId = -1;
  }
}

int SpClusterMgr::GetNodeId(char *clusterName, int clusterNode) {
  int node = -1;
  SpClusterInfo *clusterInfo =
    (SpClusterInfo *)Clusters.Find(clusterName);
  if (clusterInfo) {
    node = clusterInfo->GetNodeId(clusterNode);
  }else{
    cout << "Warning, cluster "
	 << clusterName
	 << " has not been plugged in"
	 << endl;
  }
  return node;
}

int SpClusterMgr::CheckLocalNodeZero() {
  int retVal = 0;
  SpClusterInfo *clusterInfo = (SpClusterInfo *)Clusters.GetFirstElement();
  while (clusterInfo) {
    if (clusterInfo->GetClusterNodeId() == 0) {
      retVal = 1;
      break;
    }
    clusterInfo = (SpClusterInfo *)Clusters.GetNextElement();
  }
  return retVal;
}

void SpClusterMgr::Print(ostream &out) {

  out << endl;
  out << " ------------------------------ " << endl;
  out << "| Printing Cluster Information |" << endl;
  out << " ------------------------------ " << endl;
  out << endl;

  out << "Number of Clusters = " << Clusters.GetNumElements() << endl;
  out << endl;
  unsigned int maxLen = 0;
  SpClusterInfo *clusterInfo = (SpClusterInfo *)Clusters.GetFirstElement();
  while (clusterInfo) {
    if (maxLen < strlen(clusterInfo->GetClusterName())) {
      maxLen = strlen(clusterInfo->GetClusterName());
    }
    clusterInfo = (SpClusterInfo *)Clusters.GetNextElement();
  }
  clusterInfo = (SpClusterInfo *)Clusters.GetFirstElement();
  while (clusterInfo) {
    clusterInfo->Print(out, maxLen);
    clusterInfo = (SpClusterInfo *)Clusters.GetNextElement();
  }
  out << endl;

  out << "Number of Events   = " << Events.GetNumElements() << endl;
  out << endl;
  SpEventInfo *eventInfo = (SpEventInfo *)Events.GetFirstElement();
  while (eventInfo) {
    eventInfo->Print(out);
    eventInfo = (SpEventInfo *)Events.GetNextElement();
  }
  out << endl;

  out << "Number of Managers = " << Mgrs.GetNumElements() << endl;
  out << endl;
  SpMgrInfo *mgrInfo = (SpMgrInfo *)Mgrs.GetFirstElement();
  while (mgrInfo) {
    mgrInfo->Print(out);
    mgrInfo = (SpMgrInfo *)Mgrs.GetNextElement();
  }
  out << endl;
  out << " ------------------------------------- " << endl;
  out << "| End of Printing Cluster Information |" << endl;
  out << " ------------------------------------- " << endl;
  out << endl;
}


void SpClusterMgr::PlugInCluster(char *clusterName)
{
  if (!Clusters.Find(clusterName)) {
    SpClusterInfo *clusterInfo = new SpClusterInfo();
    clusterInfo->SetClusterName(clusterName);
    Clusters.Insert(clusterInfo, clusterName);
  }
}

void SpClusterMgr::PlugInMgr(char *mgrName, int localMgrId) {
  if (!Mgrs.Find(mgrName)) {
    SpMgrInfo *mgrInfo = new SpMgrInfo();
    mgrInfo->SetMgrName(mgrName);
    mgrInfo->SetLocalMgrId(localMgrId);
    Mgrs.Insert(mgrInfo, mgrName);
    if (localMgrId > HighestMgrId){
      HighestMgrId = localMgrId;
    }
  }
  else{
    cout<<"Warning!!! Plugged in object manager "
	<<mgrName
	<<" more than once!!"
	<<endl;
  }
}

void SpClusterMgr::PlugInEvent(char *eventName, int localEventId) {
  if (!Events.Find(eventName)) {
    SpEventInfo *eventInfo = new SpEventInfo();
    eventInfo->SetEventName(eventName);
    eventInfo->SetLocalEventId(localEventId);
    Events.Insert(eventInfo, eventName);
    if (localEventId > HighestEventId){
      HighestEventId = localEventId;
    }
//     cout<<"eventName "<<eventName
// 	<<"  "<<localEventId
// 	<<"  "<<Events.GetNumElements()<<endl;
  }
  else{
    cout<<"Warning!!! Plugged in event "<<eventName<<" more than once!!"<<endl;
  }
}


void SpClusterMgr::GlobalClusterInfo() {
  int i;

  int numBytes = SizeOfClusterBuff();
  char *buff = new char[numBytes];
  char *b = buff;

  NET_INT numClusters = Clusters.GetNumElements();
  memcpy(b, (char *)&numClusters, sizeof(NET_INT));
  b += sizeof(NET_INT);

  SpClusterInfo *clusterInfo = (SpClusterInfo *)Clusters.GetFirstElement();
  while (clusterInfo) {
    clusterInfo->SetNodeId(SpComm_GetNodeId());
    clusterInfo->AddNodeId(SpComm_GetNodeId()); // Add myself here
    b = clusterInfo->FillBuff(b);
    clusterInfo = (SpClusterInfo *)Clusters.GetNextElement();
  }

  SpComm_CoordinatedSend(numBytes, buff);
  delete [] buff;

  while ( (buff = SpComm_CoordinatedReceive(numBytes)) ) {

    b = buff;
    memcpy((char *)&numClusters, b, sizeof(NET_INT));
    b += sizeof(NET_INT);

    for (i=0; i<numClusters; i++) {
      clusterInfo = new SpClusterInfo();
      b = clusterInfo->InitBuff(b);
      SpClusterInfo *c = (SpClusterInfo *)
	Clusters.Find(clusterInfo->GetClusterName());
      if (c) {
	c->AddNodeId(clusterInfo->GetNodeId());
	delete clusterInfo;
      }else{
	Clusters.Insert(clusterInfo, clusterInfo->GetClusterName());
	clusterInfo->AddNodeId(clusterInfo->GetNodeId());
	clusterInfo->SetNodeId(-1);
	clusterInfo->SetClusterNodeId(-1);
      }
    }

    delete [] buff;

  }

  clusterInfo = (SpClusterInfo *)Clusters.GetFirstElement();
  while (clusterInfo) {
    clusterInfo->Update();
    clusterInfo = (SpClusterInfo *)Clusters.GetNextElement();
  }

};


void SpClusterMgr::GlobalEventInfo() {
  int i;

  NumLocalEvents = Events.GetNumElements();
  LocalToGlobalEventId = new int[HighestEventId+1];

  int numBytes = SizeOfEventBuff();
  char *buff = new char[numBytes];
  char *b = buff;

  NET_INT numEvents = Events.GetNumElements();
  memcpy(b, (char *)&numEvents, sizeof(NET_INT));
  b += sizeof(NET_INT);

  SpEventInfo *eventInfo = (SpEventInfo *)Events.GetFirstElement();
  while (eventInfo) {
    b = eventInfo->FillBuff(b);
    eventInfo = (SpEventInfo *)Events.GetNextElement();
  }

  SpComm_CoordinatedSend(numBytes, buff);
  delete [] buff;

  while ( (buff = SpComm_CoordinatedReceive(numBytes)) ) {

    b = buff;
    memcpy((char *)&numEvents, b, sizeof(NET_INT));
    b += sizeof(NET_INT);

    for (i=0; i<numEvents; i++) {
      eventInfo = new SpEventInfo();
      b = eventInfo->InitBuff(b);
      SpEventInfo *e = (SpEventInfo *)
	Events.Find(eventInfo->GetEventName());
      if (!e) {
	eventInfo->SetLocalEventId(-1);
	Events.Insert(eventInfo, eventInfo->GetEventName());
      }else{
	delete eventInfo;
      }
    }

    delete [] buff;

  }

  NumGlobalEvents = Events.GetNumElements();
  GlobalToLocalEventId = new int[NumGlobalEvents];

  eventInfo = (SpEventInfo *)Events.GetFirstElement();
  i = 0;
  while (eventInfo) {
    if (eventInfo->GetLocalFlag()) {
      eventInfo->SetGlobalEventId(i);
      GlobalToLocalEventId[i] = eventInfo->GetLocalEventId();
      LocalToGlobalEventId[eventInfo->GetLocalEventId()] = i;
    }
    eventInfo = (SpEventInfo *)Events.GetNextElement();
    i++;
  }

};


void SpClusterMgr::GlobalMgrInfo() {
  int i;

  NumLocalMgrs = Mgrs.GetNumElements();
  LocalToGlobalMgrId = new int[HighestMgrId+1];

  int numBytes = SizeOfMgrBuff();
  char *buff = new char[numBytes];
  char *b = buff;

  NET_INT numMgrs = Mgrs.GetNumElements();
  memcpy(b, (char *)&numMgrs, sizeof(NET_INT));
  b += sizeof(NET_INT);

  SpMgrInfo *mgrInfo = (SpMgrInfo *)Mgrs.GetFirstElement();
  while (mgrInfo) {
    b = mgrInfo->FillBuff(b);
    mgrInfo = (SpMgrInfo *)Mgrs.GetNextElement();
  }

  SpComm_CoordinatedSend(numBytes, buff);
  delete [] buff;

  while ( (buff = SpComm_CoordinatedReceive(numBytes)) ) {

    b = buff;
    memcpy((char *)&numMgrs, b, sizeof(NET_INT));
    b += sizeof(NET_INT);

    for (i=0; i<numMgrs; i++) {
      mgrInfo = new SpMgrInfo();
      b = mgrInfo->InitBuff(b);
      SpMgrInfo *m = (SpMgrInfo *)
	Mgrs.Find(mgrInfo->GetMgrName());
      if (!m) {
	mgrInfo->SetLocalMgrId(-1);
	Mgrs.Insert(mgrInfo, mgrInfo->GetMgrName());
      }else{
	delete mgrInfo;
      }
    }

    delete [] buff;

  }

  NumGlobalMgrs = Mgrs.GetNumElements();
  GlobalToLocalMgrId = new int[NumGlobalMgrs];

  mgrInfo = (SpMgrInfo *)Mgrs.GetFirstElement();
  i = 0;
  while (mgrInfo) {
    if (mgrInfo->GetLocalFlag()) {
      mgrInfo->SetGlobalMgrId(i);
      GlobalToLocalMgrId[i] = mgrInfo->GetLocalMgrId();
      LocalToGlobalMgrId[mgrInfo->GetLocalMgrId()] = i;
    }
    mgrInfo = (SpMgrInfo *)Mgrs.GetNextElement();
    i++;
  }
};


void PlugInCluster(char *clusterName) {
  if (!GlobalClusterMgr) GlobalClusterMgr = new SpClusterMgr();
  if (DefaultDataBase == NULL){
    DefaultDataBase = new SpPoDataBase;
  }
  GlobalClusterMgr->PlugInCluster(clusterName);
}


void PlugInMgr(char *mgrName, int localMgrId) {
  if (!GlobalClusterMgr) GlobalClusterMgr = new SpClusterMgr();
  if (DefaultDataBase == NULL){
    DefaultDataBase = new SpPoDataBase;
  }
  GlobalClusterMgr->PlugInMgr(mgrName, localMgrId);
}


void PlugInEvent(char *eventName, int localEventId) {
  if (!GlobalClusterMgr) GlobalClusterMgr = new SpClusterMgr();
  if (DefaultDataBase == NULL){
    DefaultDataBase = new SpPoDataBase;
  }
  GlobalClusterMgr->PlugInEvent(eventName, localEventId);
}


char* SpClusterInfo::FillBuff(char *buff) {
      char *b = buff;
      NET_INT nodeId = NodeId;
      memcpy(b, (char *)&nodeId, sizeof(NET_INT));
      b += sizeof(NET_INT);
      int len = strlen(ClusterName) + 1;
      memcpy(b, ClusterName, len);
      b += len;
      return b;
    }

char* SpClusterInfo::InitBuff(char *buff) {
      char *b = buff;
      NET_INT nodeId;
      memcpy((char *)&nodeId, b, sizeof(NET_INT));
      NodeId = nodeId;
      b += sizeof(NET_INT);
      SetClusterName(b);
      b += strlen(b) + 1;
      return b;
    }

void SpClusterInfo::Print(ostream &out, int maxLen) {
  out << "   "
      << ClusterName
      << ": ";
  for (unsigned int i=0; i < (maxLen-strlen(ClusterName)); i++ ) out << " ";
  out << "Node = "
      << NodeId
      << ", ClusterNode = "
      << ClusterNodeId
      << " Nodes = <";
  if (NodeIds.GetNumElements() < 10) {
    int *element = (int *)NodeIds.GetFirstElement();
    int first = 1;
    while (element) {
      if (!first) out << ", ";
      first = 0;
      out << *element;
      element = (int *)NodeIds.GetNextElement();
    }
  }
  out << ">" << endl;
}

char* SpEventInfo::FillBuff(char *buff) {
  char *b = buff;
  NET_INT localEventId = LocalEventId;
  memcpy(b, (char *)&localEventId, sizeof(NET_INT));
  b += sizeof(NET_INT);
  int len = strlen(EventName) + 1;
  memcpy(b, EventName, len);
  b += len;
  return b;
}

char* SpEventInfo::InitBuff(char *buff) {
  char *b = buff;
  NET_INT localEventId;
  memcpy((char *)&localEventId, b, sizeof(NET_INT));
  LocalEventId = localEventId;
  b += sizeof(NET_INT);
  SetEventName(b);
  b += strlen(b) + 1;
  return b;
}

void SpEventInfo::Print(ostream &out) {
  char s[20];
  sprintf(s, "[%d, %d] ", GlobalEventId, LocalEventId);
  out << "   " << s;
  for (unsigned int i=0; i<13-strlen(s); i++) out << " ";
  out << EventName
      << endl;
}

char* SpMgrInfo::FillBuff(char *buff) {
  char *b = buff;
  NET_INT localMgrId = LocalMgrId;
  memcpy(b, (char *)&localMgrId, sizeof(NET_INT));
  b += sizeof(NET_INT);
  int len = strlen(MgrName) + 1;
  memcpy(b, MgrName, len);
  b += len;
  return b;
}

char* SpMgrInfo::InitBuff(char *buff) {
  char *b = buff;
  NET_INT mgrId;
  memcpy((char *)&mgrId, b, sizeof(NET_INT));
  LocalMgrId = mgrId;
  b += sizeof(NET_INT);
  SetMgrName(b);
  b += strlen(b) + 1;
  return b;
}

void SpMgrInfo::Print(ostream &out) {
  char s[20];
  sprintf(s, "[%d, %d] ", GlobalMgrId, LocalMgrId);
  out << "   " << s;
  for (unsigned int i=0; i<13-strlen(s); i++) out << " ";
  out << MgrName
      << endl;
}
