// SpClusterMgr.H header file

#ifndef SpClusterMgrGate
#define SpClusterMgrGate

#include "SpStdlib.H"
#include "SpStdio.H"
#include "SpString.H"
#include "SpIostream.H"
#include "SpBinaryTree.H"
#include "SpNetTypes.H"

//---------------------------------------------------------------------------
// SpClusterInfo
//---------------------------------------------------------------------------
class SpClusterInfo {

  private:

    int NumNodesInCluster;
    int ClusterNodeId;
    int NodeId;
    char *ClusterName;
    int *NodeIdArray;
    SpBinaryTree NodeIds;

  protected:

  public:

    SpClusterInfo() {
      NumNodesInCluster = 0;
      ClusterNodeId = -1;
      NodeId = -1;
      ClusterName = NULL;
      NodeIdArray = NULL;
    }

    SpClusterInfo(int numNodesInCluster,
		      int clusterNodeId,
		      int nodeId,
		      char *clusterName) {
      NumNodesInCluster = numNodesInCluster;
      ClusterNodeId = clusterNodeId;
      NodeId = nodeId;
      ClusterName = strdup(clusterName);
      AddNodeId(NodeId);
      NodeIdArray = NULL;
    }

    ~SpClusterInfo() {
      if (ClusterName) free(ClusterName);
      if (NodeIdArray) delete [] NodeIdArray;
    }

    int GetNumNodesInCluster() {return NumNodesInCluster;}
    void SetNumNodesInCluster(int numNodesInCluster) {
      NumNodesInCluster = numNodesInCluster;
    }

    int GetClusterNodeId() {return ClusterNodeId;}
    void SetClusterNodeId(int clusterNodeId) {
      ClusterNodeId = clusterNodeId;
    }

    int GetNodeId() {return NodeId;}
    int GetNodeId(int clusterNodeId) {
      int nodeId = -1;
      if ((0 <= clusterNodeId) && (clusterNodeId < NumNodesInCluster)) {
	nodeId = NodeIdArray[clusterNodeId];
      }else{
	cout << "Warning, bad cluster node "
	     << clusterNodeId
	     << " for "
	     << ClusterName
	     << endl;
      }
      return nodeId;
    }

    void SetNodeId(int nodeId) {NodeId = nodeId;}

    char *GetClusterName() {return ClusterName;}
    void SetClusterName(char *clusterName) {
      if (ClusterName) free(ClusterName);
      ClusterName = strdup(clusterName);
    }

    void AddNodeId(int nodeId) {
      int *n = new int();
      *n = nodeId;
      NodeIds.Insert(n, nodeId);
    }

    void Update() {
      NumNodesInCluster = NodeIds.GetNumElements();
      NodeIdArray = new int[NumNodesInCluster];

      int *n = (int *)NodeIds.GetFirstElement();
      for (int i=0; i<NumNodesInCluster; i++) {
	NodeIdArray[i] = *n;
	if (*n == NodeId) {
	  ClusterNodeId = i;
	}
	n = (int *)NodeIds.GetNextElement();
      }
    }

    int SizeOfBuff() {return (sizeof(NET_INT) + strlen(ClusterName) + 1);}

    char *FillBuff(char *buff);

    char *InitBuff(char *buff);

    void Print(ostream &out, int maxLen);

};


//---------------------------------------------------------------------------
// SpEventInfo
//---------------------------------------------------------------------------
class SpEventInfo {

  private:

    int LocalEventId;
    int GlobalEventId;
    int LocalFlag;
    char *EventName;

  protected:

  public:

    SpEventInfo() {
      LocalEventId = -1;
      GlobalEventId = -1;
      EventName = NULL;
    }

    SpEventInfo(int localEventId, char *eventName) {
      LocalEventId = localEventId;
      GlobalEventId = -1;
      LocalFlag = 1;
      EventName = strdup(eventName);
    }

    ~SpEventInfo() {
      if (EventName) free(EventName);
    }

    int GetLocalEventId() {return LocalEventId;}
    void SetLocalEventId(int localEventId) {
      LocalEventId = localEventId;
      LocalFlag = 1;
    }

    int GetGlobalEventId() {return GlobalEventId;}
    void SetGlobalEventId(int globalEventId) {GlobalEventId = globalEventId;}

    int GetLocalFlag() {return LocalFlag;}
    void SetLocalFlag() {LocalFlag = 1;}
    void ResetLocalFlag() {LocalFlag = 0;}

    char *GetEventName() {return EventName;}
    void SetEventName(char *eventName) {
      if (EventName) free(EventName);
      EventName = strdup(eventName);
    }

    int SizeOfBuff() {return (sizeof(NET_INT) + strlen(EventName) + 1);}

    char *FillBuff(char *buff);

    char *InitBuff(char *buff);

    void Print(ostream &out);

};


//---------------------------------------------------------------------------
// SpMgrInfo
//---------------------------------------------------------------------------
class SpMgrInfo {

  private:

    int LocalMgrId;
    int GlobalMgrId;
    int LocalFlag;
    char *MgrName;

  protected:

  public:

    SpMgrInfo() {
      LocalMgrId = -1;
      GlobalMgrId = -1;
      LocalFlag = 0;
      MgrName = NULL;
    }

    SpMgrInfo(int localMgrId, char *mgrName) {
      LocalMgrId = localMgrId;
      GlobalMgrId = -1;
      LocalFlag = 1;
      MgrName = strdup(mgrName);
    }

    ~SpMgrInfo() {
      if (MgrName) free(MgrName);
    }

    int GetLocalMgrId() {return LocalMgrId;}
    void SetLocalMgrId(int localMgrId) {
      LocalMgrId = localMgrId;
      LocalFlag = 1;
    }

    int GetGlobalMgrId() {return GlobalMgrId;}
    void SetGlobalMgrId(int globalMgrId) {GlobalMgrId = globalMgrId;}

    int GetLocalFlag() {return LocalFlag;}
    void SetLocalFlag() {LocalFlag = 1;}
    void ResetLocalFlag() {LocalFlag = 0;}

    char *GetMgrName() {return MgrName;}
    void SetMgrName(char *mgrName) {
      if (MgrName) free(MgrName);
      MgrName = strdup(mgrName);
    }

    int SizeOfBuff() {return (sizeof(NET_INT) + strlen(MgrName) + 1);}

    char *FillBuff(char *buff);

    char *InitBuff(char *buff);

    void Print(ostream &out);

};


//---------------------------------------------------------------------------
// SpClusterMgr
//---------------------------------------------------------------------------
class SpClusterMgr {

  private:

    SpBinaryTree Clusters;
    SpBinaryTree Mgrs;
    SpBinaryTree Events;

    int *GlobalToLocalMgrId;
    int *LocalToGlobalMgrId;
    int NumLocalMgrs;
    int NumGlobalMgrs;

    int HighestMgrId;
    int HighestEventId;

    int *GlobalToLocalEventId;
    int *LocalToGlobalEventId;
    int NumLocalEvents;
    int NumGlobalEvents;

    static SpClusterMgr *ClusterMgr;
    void SetClusterMgr();

    int SizeOfClusterBuff();
    int SizeOfEventBuff();
    int SizeOfMgrBuff();

  protected:

  public:

    SpClusterMgr();

    int ConvertGlobalToLocalEventId(int globalEventId) {
      return GlobalToLocalEventId[globalEventId];
    }

    int ConvertGlobalToLocalMgrId(int localMgrId) {
      int retval=-1;
      if (localMgrId != -1) {
	retval = GlobalToLocalMgrId[localMgrId];
      }
      return retval;
    }

    int ConvertLocalToGlobalEventId(int localEventId) {
      return LocalToGlobalEventId[localEventId];
    }

    int ConvertLocalToGlobalMgrId(int globalMgrId) {
      int retval=-1;
      if (globalMgrId != -1) {
	retval = LocalToGlobalMgrId[globalMgrId];
      }
      return retval;
    }

    SpClusterInfo *GetClusterInfo(char *clusterName) {
      return (SpClusterInfo *)Clusters.Find(clusterName);
    }

    void GetClusterInfo(char *clusterName,
			int &numNodesInCluster,
			int &clusterNodeId);

    int GetNodeId(char *clusterName, int clusterNode);

    void PlugInCluster(char *clusterName);
    void PlugInMgr(char *mgrName, int localMgrId);
    void PlugInEvent(char *eventName, int localEventId);

    void GlobalClusterInfo();
    void GlobalEventInfo();
    void GlobalMgrInfo();

    int CheckLocalNodeZero();

    void Print(ostream &out);
};

void PlugInCluster(char *clusterName);
void PlugInMgr(char *mgrName, int localMgrId);
void PlugInEvent(char *eventName, int localEventId);

#define PLUG_IN_CLUSTER(clusterName)					\
PlugInCluster(#clusterName);						\

#endif
