// SpEventHandlerHolder.C source file that implements the handler 
// holder.

#include "RB_SpDefineClass.H"
#include "SpEventHandlerHolder.H"
#include "RB_SpDefineClass.H"

RB_DEFINE_CLASS(SpEventHandlerDescriptor);
RB_DEFINE_CLASS(SpEventHandlerHolder);
RB_DEFINE_CLASS(SpEventHandlerMethod);


const int INITIAL_EVENT_HANDLER_ARRAY_SIZE = 4;

int SpEventHandlerHolder::Node = 0;


SpEventHandlerHolder::SpEventHandlerHolder(int initialEHarrayPtrLength): EHarrayPtrLength(initialEHarrayPtrLength), Nsubscribed(0) {
  if (initialEHarrayPtrLength > 0) {
    EHarrayPtr = RB_NEW_ARRAY_SpEventHandlerDescriptor(
						       initialEHarrayPtrLength
						       );
  }
  else {
    EHarrayPtr = NULL;
  }
}

void 
SpEventHandlerHolder::SetEventName(char *eventName) {
  if (EventName){ 
    delete [] EventName;
    EventName = NULL;
  }
  char *p = RB_strdup(eventName);
  RB_memcpy(&EventName, &p, sizeof(char *));
}

/*
*****************************************************************************
* AddEventHandler() method adds an event handler to the holder
*****************************************************************************
*/

int SpEventHandlerHolder::AddEventHandler(
					  SpBaseEventHandler *eventHandler,
					  bool subscribed
					  ) {
  SpEventHandlerDescriptor *NewEHarrayPtr;
  int i;

  eventHandler->SetObjectType(ObjectType);
  eventHandler->SetLocalId(LocalId);
  eventHandler->SetGlobalId(GlobalId);
  eventHandler->SetNode(Node);
  eventHandler->SetSimObj(SimObj);

  if (subscribed) Nsubscribed++;

  if (0 == EHarrayPtrLength) {
    NewEHarrayPtr =
      RB_NEW_ARRAY_SpEventHandlerDescriptor(INITIAL_EVENT_HANDLER_ARRAY_SIZE);
    RB_memcpy((char *)&EHarrayPtr,
	      (char *)&NewEHarrayPtr,
	      sizeof(SpEventHandlerDescriptor *));
    EHarrayPtr[0] = eventHandler;
    EHarrayPtr[0].subscribed = subscribed;
    EHarrayPtrLength = INITIAL_EVENT_HANDLER_ARRAY_SIZE;
    return 0;
  }
  else { // EHarrayPtrLength != 0
    for (i=0; i<EHarrayPtrLength; i++) {
      if (EHarrayPtr[i] == NULL) {
	SpEventHandlerDescriptor tmpEventHandlerDescriptor(eventHandler,
							 subscribed);
	RB_memcpy((char *)&EHarrayPtr[i],
		  (char *)&tmpEventHandlerDescriptor,
		  sizeof(SpEventHandlerDescriptor));
	return i;
      }
    }
    
    int oldEHarrayPtrLength = EHarrayPtrLength;
    NewEHarrayPtr =
      RB_NEW_ARRAY_SpEventHandlerDescriptor(
					  EHarrayPtrLength = 
					  oldEHarrayPtrLength*2
					  );
    for (i=0; i<oldEHarrayPtrLength; i++) {
      NewEHarrayPtr[i] = EHarrayPtr[i];
    }
    
    NewEHarrayPtr[oldEHarrayPtrLength] = eventHandler;
    RB_DELETE_ARRAY_SpEventHandlerDescriptor(EHarrayPtr);
    RB_memcpy((char *)&EHarrayPtr,
	      (char *)&NewEHarrayPtr,
	      sizeof(SpEventHandlerDescriptor *));
    
    return (oldEHarrayPtrLength);
    
  }
}

/*
*****************************************************************************
* AddEventHandlerMethod() method adds an event handler to the holder
*****************************************************************************
*/

SpEventHandler *SpEventHandlerHolder::
AddEventHandlerMethod(
		      void *object,
		      EVENT_HANDLER_METHOD_FNC_PTR method
		      ) {
   SpEventHandler *eventHandler;
   
   eventHandler = RB_NEW_SpEventHandlerMethod
     (new SpEventHandlerMethod(object, method));

   AddEventHandler(eventHandler);
   return eventHandler;
}

/*
*****************************************************************************
* RemoveEventHandler() method removes an event handler from the holder and
* returns the index where it was. If flag is set to DIRECTED, the
* first unsubscribed eventhandler is removed, if UNDIRECTED the first
* subscribed, otherwise, the first one occurring is removed, 
* irrespective of mode
*****************************************************************************
*/

int SpEventHandlerHolder::RemoveEventHandler(
					     SpBaseEventHandler *eventHandler,
					     enum SpEventHandlerMode 
					     flag
					     )
{
  int i, j;
  SpEventHandlerDescriptor *NewEHarrayPtr;

  NewEHarrayPtr = RB_NEW_ARRAY_SpEventHandlerDescriptor(EHarrayPtrLength);
  for (i=0; i<EHarrayPtrLength; i++) NewEHarrayPtr[i] = EHarrayPtr[i];

  // we must remove holes in the array to preserve array ordering
  
  for (i=0; i<EHarrayPtrLength; i++) {
    if (NewEHarrayPtr[i] == eventHandler) {
      if (NewEHarrayPtr[i].subscribed && flag == DIRECTED ||
	  !NewEHarrayPtr[i].subscribed && flag == UNDIRECTED) continue;
      if (EHarrayPtr[i].subscribed) Nsubscribed--;
      for (j=i+1; j<EHarrayPtrLength; j++) {
	NewEHarrayPtr[j-1] = EHarrayPtr[j];
      }
      NewEHarrayPtr[j-1] = NULL;
      RB_DELETE_ARRAY_SpEventHandlerDescriptor(EHarrayPtr);
      RB_memcpy((char *)&EHarrayPtr,
		(char *)&NewEHarrayPtr,
		sizeof(SpEventHandlerDescriptor *));
      return i;
    }
  }

  cout << "Warning, could not remove event handler" << endl;
  return -1;

}


/*
*****************************************************************************
* ProcessEventHandlers() method processes the set of event handlers
*****************************************************************************
*/

void SpEventHandlerHolder::ProcessEventHandlers(SpSimTime &timeTag,
						char *data, int dataBytes,
						int externalId, 
						int ehType,
						int undirected) {
  int i;

  for (i=0; i<EHarrayPtrLength; i++) {
    if (EHarrayPtr[i] == NULL) break;

    // Skip event handlers that have been plugged in only for directed event handler events
    if (!EHarrayPtr[i].subscribed && undirected) continue;

    if (((SpBaseEventHandler *) EHarrayPtr[i])->GetType() == ehType)  
      ((SpBaseEventHandler *) EHarrayPtr[i])->
	ProcessEventHandler(EventName, timeTag, data, dataBytes, externalId);
  }

}


SpEventHandlerHolder::~SpEventHandlerHolder() {
   if (EHarrayPtr) delete [] EHarrayPtr;
   if (EventName) delete [] EventName;
}


void RB_Register_SpEventHandlerHolder()
{
  RB_REGISTER_CLASS_SpEventHandlerHolder();
}

void RB_Register_SpEventHandlerDescriptor()
{
  RB_REGISTER_CLASS_SpEventHandlerDescriptor();
}

