// SpEventHandlerHolder.H header file is a holder for event handlers. Note that
// this is not rollbackable yet. The event handlers must be set up during
// initialization. We will make this rollbackable later.

#ifndef EhHo_OBJ
#define EhHo_OBJ

#include "sq_item.H"
#include "SpEventHandler.H"
#include "SpEventHandlerMethod.H"
#include "SpStdlib.H"
#include "SpString.H"
#include "RB_Types.H"
#include "RB_SpDefineClass.H"

#if !defined(__GNUC__) && !defined(__SPEEDES_NT__) && !defined (_BOOL)
#define bool int
#define false ((int) 0)
#define true ((int) 1)
#define _BOOL
#endif  // #ifndef __GNUC__

class SpSimObj;

class SpEventHandlerDescriptor 
{
public:
  SpEventHandlerDescriptor(SpBaseEventHandler *eh=NULL,
			   bool subscribed_local=false): 
    eventHandler(eh), subscribed(subscribed_local) {}
  
  operator SpBaseEventHandler *() {return eventHandler;}
  
  SpEventHandlerDescriptor &operator=(SpEventHandlerDescriptor &ehd) {
    eventHandler = ehd.eventHandler;
    subscribed = ehd.subscribed;
    return *this;
  }
  
  SpBaseEventHandler *operator =(SpBaseEventHandler *eh) {
    eventHandler = eh; 
    return eh;
  }
  
//   int operator ==(SpBaseEventHandler *eh) {
//     return (eventHandler == eh);
//   }
  
  SpBaseEventHandler *eventHandler;
  bool subscribed;
};
    


class SpEventHandlerHolder : public C_SQ_ITEM {


public:
  
  SpEventHandlerHolder()
    : EHarrayPtr(NULL),
      EHarrayPtrLength(0),
      SimObj(NULL),
      EventName(NULL),
      Nsubscribed(0) {}
  
  SpEventHandlerHolder(int numberOfHandlers);
  ~SpEventHandlerHolder();
  
  int AddEventHandler(SpBaseEventHandler *eventHandler,
		      bool subscribed=false);
		      
  SpEventHandler *AddEventHandlerMethod(void *object, 
					    EVENT_HANDLER_METHOD_FNC_PTR method);
  int RemoveEventHandler(SpBaseEventHandler *eventHandler,
			 enum SpEventHandlerMode flag=DONTCARE);
  
  SpBaseEventHandler *GetEventHandler(int i) {return EHarrayPtr[i];}
  
  void SetObjectType(int objectType) {ObjectType = objectType;}
  void SetLocalId(int localId) {LocalId = localId;}
  void SetGlobalId(int globalId) {GlobalId = globalId;}
  void SetNode(int node) {Node = node;}
  
  void SetSimObj(SpSimObj *simObj) {SimObj = simObj;}   
  void SetEventName(char *eventName);
  virtual char *GetName() {return EventName;}
  
  void ProcessEventHandlers(SpSimTime &TimeTag, char *data, int dataBytes,
			    int externalId, int ehType,int undirected);
  int GetNsubscribed() {return Nsubscribed;}
  
private:

  SpEventHandlerDescriptor *EHarrayPtr;
  RB_int EHarrayPtrLength;
  SpSimObj *SimObj;
  int ObjectType;
  int LocalId;
  int GlobalId;
  char *EventName;
  static int Node;
  int Nsubscribed; // number of subscribed event handlers
};

//RB_DEFINE_CLASS(SpEventHandlerHolder);
//RB_DEFINE_CLASS(SpEventHandlerDescriptor);

typedef SpEventHandlerHolder EVENT_HANDLER_HOLDER;

#endif
