//...... types of events (messages)

#ifndef base_evtype_object
#define  base_evtype_object

#include "free_list.H"
#include "holder.H"
#include "SpMsg.H"
#include "SpSimTime.H" //GEB2
#include "SpEvent.H"

class SpList;

/************************************************************************
*       evtype object                                                   *
************************************************************************/
class SpEventMsgMgr {

  private:

    typedef void * (*EVENT_CONSTRUCTOR_FUNCTION)(int n);
    typedef void   (*EVENT_DESTRUCTOR_FUNCTION)(void *e);

    C_QUEUE ToBeCleared;

    int ntypes;

    EVENT_CONSTRUCTOR_FUNCTION *event_constructor;
    EVENT_DESTRUCTOR_FUNCTION *event_destructor;
    TO_BE_DELETED_FUNCTION *ToBeDeleteds;
    char **name;
    int *sizeof_message;
    int *sizeof_event;
    int *neach;
    int *n_events;
    int *n_messages;
    int n_each_holder;
    int n_holders;

    C_QUEUE *qfree_events;
    C_QUEUE *qfree_messages;
    C_QUEUE qfree_holders;

    SpList *events;
    SpList *messages;

    C_QUEUE *qmesstypes;	// queue of all message types

    void *f(int type, int n);

    void compact_events_freelist(int type);
    void compact_messages_freelist(int type);
      
  public:

    SpEventMsgMgr();
    ~SpEventMsgMgr();

    void set_ntypes(int n);
    int get_ntypes() {return ntypes;}

    void set_holders(int n) {n_each_holder = n;}

    int get_messid(char *name);
    char *get_messname(int i) {return name[i];}

    void define_event(char *type, int itype, EVENT_CONSTRUCTOR_FUNCTION, EVENT_DESTRUCTOR_FUNCTION,
		      int sz_e, int sz_m, int n);

//  void define_event(char *type, int itype, TO_BE_DELETED_FUNCTION, int sz_e, int sz_m, int n);

    void *message_event(char *message);
    void *new_event(int type);
    void delete_event(void *event);

    C_HOLDER *new_message(int type);
    C_HOLDER *new_message(const SpSimTime& t, int type, int oty, int oid, int nod);
    C_HOLDER *new_message(const SpSimTime& t, int type, int oty, int oid, int nod,
			  int size, char * &bf);
    void delete_message(C_HOLDER *holder);

    void DefinitelyPutOn_qfree_messages(C_HOLDER *h)
    { qfree_messages[((SpMsg*)(h->get_buff()))->evtype].push_bot(h); }

    
    C_HOLDER *new_holder();
    C_HOLDER *new_holder(char *buff, int size, int node);
    void delete_holder(C_HOLDER *holder) { qfree_holders.push_bot(holder); }

    void print();

    SpMsg *dup_message(SpMsg *header);
    void undup_message(SpMsg *header);
    int get_sizeof(int type) {return sizeof_message[type];}

};
inline void *SpEventMsgMgr::message_event(char *message) {
  SpMsg *header = (SpMsg *)message;
  SpEvent *event = (SpEvent *)new_event(header->evtype);

  event->set_time_tag(header->time_tag);
  event->set_id(header->objid);
  event->set_EVENT_TYPE(header->evtype);
  event->set_MESSAGE(header);
  if (header->bytes == 0)
    event->set_DATA(NULL);
  else
    event->set_DATA(((char*)header) + sizeof_message[header->evtype]);
  event->reset_from_SpComm();
  event->set_OBJECT_TYPE(header->obtype);
  event->set_valid();
  event->reset_processed();
  event->set_rbid(-1);
  event->set_cancel_id(header->cancel_id);
  event->reset_event_canceled();
  event->reset_lazy_proc();
  event->attach_object();
  event->init(header);
  return event;
}

typedef SpEventMsgMgr C_BASE_EVTYPE;

#endif
