// SpEventMsgMgr.C method file

#include "SpAssert.H"
#include "SpStdio.H"
#include "SpStdlib.H"
#include "SpString.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"

#include "SpEventMsgMgr.H"
#include "SpEvent.H"

#define REMOVE_TO_BE_DELETEDS
//#define CONSISTENCY_TEST

/************************************************************************
* SpEventMsgMgr : construct a base event type object			*
************************************************************************/
SpEventMsgMgr::SpEventMsgMgr() {

}

SpEventMsgMgr::~SpEventMsgMgr() {

#ifdef REMOVE_TO_BE_DELETEDS
  ToBeCleared.clear();
  delete [] ToBeDeleteds;
#endif

  delete [] event_constructor;

  for (int i=0; i<ntypes; i++) free(name[i]);
  delete [] name;
  delete [] neach;
  delete [] sizeof_message;
  delete [] sizeof_event;
  delete [] n_messages;
  delete [] n_events;
  delete [] qfree_events;
  delete [] qfree_messages;

}

/************************************************************************
* SpEventMsgMgr : construct a base event type object			*
************************************************************************/
void SpEventMsgMgr::set_ntypes(int n) {
  int i;

  ntypes = n;

  event_constructor = new EVENT_CONSTRUCTOR_FUNCTION[ntypes];

#ifdef REMOVE_TO_BE_DELETEDS
  ToBeDeleteds = new TO_BE_DELETED_FUNCTION[ntypes];
#endif

  name = new char*[ntypes];
  neach = new int[ntypes];
  sizeof_message = new int[ntypes];
  sizeof_event = new int[ntypes];
  n_messages = new int[ntypes];
  n_events = new int[ntypes];
  qfree_events =  new C_QUEUE[ntypes];
  qfree_messages =  new C_QUEUE[ntypes];

  for (i=0; i<ntypes; i++) {
    neach[i] = 0;
    sizeof_message[i] = 0;
    sizeof_event[i] = 0;
    n_messages[i] = 0;
    n_events[i] = 0;
  }

  n_holders = 0;
  n_each_holder = 1000;

}

/************************************************************************
* get_messid : construct a base event type object			*
************************************************************************/
int SpEventMsgMgr::get_messid(char *nm) {
  int i;

  assert(nm != NULL);
  for (i=0; i<ntypes; i++) {
    assert(name[i] != NULL);
    if (!strcmp(name[i], nm)) return i; 
  }

//  fprintf(stderr,"Error, could not find message type %s\n",nm);
  return -1;

}


void SpEventMsgMgr::define_event(char *type, int itype, EVENT_CONSTRUCTOR_FUNCTION f,
		int size_e, int size_m, int n) {

  event_constructor[itype] = f;

#ifdef REMOVE_TO_BE_DELETEDS
  ToBeDeleteds[itype] = NULL;
#endif

  sizeof_message[itype] = size_m;
  sizeof_event[itype] = size_e;
  neach[itype] = n;
  name[itype] = strdup(type);

}

void SpEventMsgMgr::define_event(char *type, int itype, TO_BE_DELETED_FUNCTION f,
		int size_e, int size_m, int n) {

  event_constructor[itype] = NULL;

#ifdef REMOVE_TO_BE_DELETEDS
  ToBeDeleteds[itype] = f;
#endif

  sizeof_message[itype] = size_m;
  sizeof_event[itype] = size_e;
  neach[itype] = n;
  name[itype] = strdup(type);

}

void *SpEventMsgMgr::f(int type, int n) {
  if (event_constructor[type] != NULL) {
    return event_constructor[type](n);
  }else{
#ifdef REMOVE_TO_BE_DELETEDS
    if (ToBeDeleteds[type] != NULL) {
      ARRAY_OF_CHILDREN *tbd = ToBeDeleteds[type](n);
      ToBeCleared.push_bot(tbd);
      return tbd->GetArray();
    }
#endif
    assert(0);
    return NULL;
  }

}

/************************************************************************
* new_event : get a new event from the free list			*
************************************************************************/
void *SpEventMsgMgr::new_event(int type) {
  void *event;
  SpEvent *e;
  int i;
  char *bff;
  SpEvent *free_item;

  if (qfree_events[type].length() == 0) {

    event = f(type, neach[type]);
    e = (SpEvent *)event;

    if (event == NULL) {
      fprintf(stderr,"Error, no more event memory\n");
      exit(1);
    }

    bff = (char *)event;
    for (i=1; i<neach[type]; i++) {
      bff += sizeof_event[type];
      free_item = (SpEvent *)bff;
      free_item->set_free();
      qfree_events[type].push_bot(free_item);
    }

    n_events[type] += neach[type];
    e->set_free();

  }else{

    event = (void *)qfree_events[type].pop_top();

  }

  return event;

}


/************************************************************************
* delete_event : delete an event by putting it in the free list		*
************************************************************************/
void SpEventMsgMgr::delete_event(void *event) {
  int type;
  SpEvent *e;

  e = (SpEvent *)event;

  e->clear_data();
  e->delete_MESSAGE();

  type = e->get_EVENT_TYPE();


  qfree_events[type].push_bot(e);
  if (e->get_free() == 0) {
    e->set_free();
    n_events[type]++;
  }

}

struct char_buffer : public C_ITEM
{
    char_buffer(char *buff) : Buff(buff) {}
    ~char_buffer() { delete [] Buff; }
  private:
    char *Buff;
};

/************************************************************************
* new_message : get a new message from the free list			*
************************************************************************/
C_HOLDER *SpEventMsgMgr::new_message(const SpSimTime& t, int type, int oty, int oid, int nod) { //GEB2
  C_HOLDER *holder = new_message(type);
  holder->set_node(nod);

  SpMsg *header = (SpMsg*) holder->get_buff();

  header->time_tag = t;
  header->obtype = oty;
  header->objid = oid;
  header->ext = 0;
  header->bytes = 0;

  return holder;
}


/************************************************************************
* new_message : get a new message from the free list			*
************************************************************************/
C_HOLDER *SpEventMsgMgr::new_message(int type) { //GEB2
  int size;
  char *buff;
  char *bff;
  int i;
  C_HOLDER *holder;
  SpMsg *header;

  if (qfree_messages[type].length() == 0) {

    size = sizeof_message[type];
    buff = new char[size * neach[type]];

    memset(buff, 0, size*neach[type]); // Fixes a bug when SPEEDES is compiled
				       // as a DLL under NT

#ifdef REMOVE_TO_BE_DELETEDS
    ToBeCleared.push_bot(new char_buffer(buff));
#endif

    if (buff == NULL) {
      fprintf(stderr,"no more message memory\n");
      exit(1);
    }

    bff = buff;
    for (i=1; i<neach[type]; i++) {
      bff += size;
      header = (SpMsg *)bff;
      header->evtype = type;
      holder = (C_HOLDER *)new_holder();
      holder->set_buff(bff);
      holder->set_size(size);
      qfree_messages[type].push_bot(holder);
    }

    n_messages[type] += neach[type];

    header = (SpMsg *)buff;
    header->evtype = type;
    holder = (C_HOLDER *)new_holder();
    holder->set_buff(buff);
    holder->set_size(size);

  }else{

    holder = (C_HOLDER *)qfree_messages[type].pop_top();
    header = (SpMsg *)holder->get_buff();

    if (header->evtype != type) {
      cout << "Error, bad type in SpEventMsgMgr::new_message()" << endl;
    }

  }

  return holder;

}

/************************************************************************
* new_message : get a new message from the free list			*
************************************************************************/
C_HOLDER *SpEventMsgMgr::new_message(const SpSimTime& t, int type, int oty, int oid, int nod, int size, char * &bf) { //GEB2
  int totsize;
  char *buff;
  C_HOLDER *holder;
  SpMsg *header;

  totsize = size + sizeof_message[type];

  buff = new char[totsize];
  bf = buff + sizeof_message[type];

  header = (SpMsg *)buff;
  header->evtype = type;
  header->time_tag = t;
  header->obtype = oty;
  header->objid = oid;
  header->ext = 0;
  header->bytes = size;

  holder = (C_HOLDER *)new_holder(buff, totsize, nod);

  return holder;

}

/************************************************************************
* dup_message : duplicate the message using the free list if possible	*
************************************************************************/
SpMsg *SpEventMsgMgr::dup_message(SpMsg *h) {
  int totsize;
  char *buff;
  char *bff;
  int i;
  int size;
  int type;
  C_HOLDER *holder;
  SpMsg *header;

  type = h->evtype;

  if (h->bytes == 0) {

    totsize = sizeof_message[type];

    if (qfree_messages[type].length() == 0) {

      size = sizeof_message[type];
      buff = new char[size * neach[type]];

      //char_buffers.push_bot(new char_buffer(buff));

      if (buff == NULL) {
        fprintf(stderr,"no more message memory\n");
        exit(1);
      }

      bff = buff;
      for (i=1; i<neach[type]; i++) {
        bff += size;
        header = (SpMsg *)bff;
        header->evtype = type;
        holder = (C_HOLDER *)new_holder(bff, size, 0);
        qfree_messages[type].push_bot(holder);
      }

      n_messages[type] += neach[type];

      header = (SpMsg *)buff;

    }else{

      holder = (C_HOLDER *)qfree_messages[type].pop_top();
      header = (SpMsg *)holder->get_buff();
      delete_holder(holder);

    }

  }else{

    totsize = sizeof_message[type] + h->bytes;
    header = (SpMsg *)new char[totsize];

  }

  memcpy(header, h, totsize);
  return header;

}

/************************************************************************
* undup_message : unduplicate the message using the free list		*
************************************************************************/
void SpEventMsgMgr::undup_message(SpMsg *h) {
  int type;
  int size;
  C_HOLDER *holder;

#ifdef CONSISTENCY_TEST
  int count = 0;
  int len = qfree_messages[h->evtype].get_length();
  C_HOLDER *ho = (C_HOLDER *)qfree_messages[h->evtype].get_top();
  for (int i=0; i<len; i++) {
    if (ho->get_buff() == (char *)h) {
      count++;
    }
    ho = (C_HOLDER *)ho->get_link();
  }
  if (count) {
    cout << "Major problem, header in free list " << count << " times" << endl;
//    while (1);
    return;
  }
#endif

  if (h->bytes == 0) {
    type = h->evtype;
    size = sizeof_message[type];
    holder = new_holder((char *)h, size, 0);
    qfree_messages[type].push_bot(holder);
  }else{
    delete [] (char *)h;
  }

}

/************************************************************************
* delete_message : delete a message by putting it in the free list	*
************************************************************************/
void SpEventMsgMgr::delete_message(C_HOLDER *holder) {
  int type;
  SpMsg *header;

  header = (SpMsg *)holder->get_buff();
  type = header->evtype;

  if (type < 0) {
    delete [] (char *) header;
    delete_holder(holder);
    return;
  }

  if (sizeof_message[type] <= 0) {
    delete [] (char *) header;
    delete_holder(holder);
    return;
  }

  if (header->bytes > 0) {
    delete [] (char *) header;
    delete_holder(holder);
    return;
  }

  /*
    This is commented out since the special case of only one message at a time does not work.
  if (neach[type] == 1) {
    n_messages[type]--;
    delete [] (char *) header;
    delete_holder(holder);
    return;
  }
  */

#ifdef CONSISTENCY_TEST
  int len = qfree_messages[type].get_length();
  int count = 0;
  C_HOLDER *ho = (C_HOLDER *)qfree_messages[type].get_top();
  for (int i=0; i<len; i++) {
    if (ho->get_buff() == (char *)header) {
      count++;
    }
    ho = (C_HOLDER *)ho->get_link();
    if (count) {
      cout << "Major problem, header in free list "
	   << count << " times" << endl;
//      while (1);
      return;
    }
  }
#endif

  qfree_messages[type].push_bot(holder);

}

struct HOLDER_ARRAY_OF_CHILDREN : public C_ITEM
{
    HOLDER_ARRAY_OF_CHILDREN(C_HOLDER *a) : A(a) {}
    ~HOLDER_ARRAY_OF_CHILDREN() { delete [] A; }
  private:
    C_HOLDER *A;
};

C_HOLDER *SpEventMsgMgr::new_holder(char *buff, int size, int node) {
  C_HOLDER *holder = new_holder();
  holder->set(buff,size,node);
  return holder;
}

/************************************************************************
* new_holder : get a new holder from the free list			*
************************************************************************/
C_HOLDER *SpEventMsgMgr::new_holder() {
  C_HOLDER *holder;
  int i;

  if (qfree_holders.length() == 0) {

    holder = new C_HOLDER[n_each_holder];

#ifdef REMOVE_TO_BE_DELETEDS
    ToBeCleared.push_bot(new HOLDER_ARRAY_OF_CHILDREN(holder));
#endif

    if (holder == NULL) {
      fprintf(stderr,"no more holder memory\n");
      exit(1);
    }

    for (i=1; i<n_each_holder; i++) {
      qfree_holders.push_bot(&holder[i]);
    }

    n_holders += n_each_holder;

  }else{

    holder = (C_HOLDER *)qfree_holders.pop_top();

  }

  return holder;

}

#ifdef COMMENTOUT
/************************************************************************
* print : print out the memory usage statistics				*
************************************************************************/
void SpEventMsgMgr::print() {
  int i;
  int j;
  int k;

  fprintf(stderr,"\n");

  j = 0;
  for (i=0; i<ntypes; i++) {

    k = n_events[i]*sizeof_event[i] + n_messages[i]*sizeof_message[i];

    fprintf(stderr,"%3d %-16s   events = %6d free = %6d   messages = %6d free = %6d   bytes = %d\n",
        i, name[i],
	n_events[i], qfree_events[i].length(),
	n_messages[i], qfree_messages[i].length(),
	k);

    j += k;

  }

  k = n_holders * sizeof(C_HOLDER);

  fprintf(stderr,"\n%3d %-20s   created = %6d   free = %6d   bytes = %d\n",
        ntypes, "HOLDER", n_holders, qfree_holders.length(), k);

  j += k;

  fprintf(stderr,"\n    Total bytes = %d\n\n",j);

}
#endif

/************************************************************************
* print : print out the memory usage statistics				*
************************************************************************/
void SpEventMsgMgr::print() {
  int i;
  int Nchars = 0;
  int *NumEvents;
  int *NumFreeEvents;
  int *NumMessages;
  int *NumFreeMessages;
  int *Bytes;
  int TotalBytes = 0;
  int node = SpComm_GetNodeId();
  int NumHolders;
  int NumFreeHolders;

  NumEvents = new int[ntypes];
  NumFreeEvents = new int[ntypes];
  NumMessages = new int[ntypes];
  NumFreeMessages = new int[ntypes];
  Bytes = new int[ntypes];

  for (i=0; i<ntypes; i++) {

    NumEvents[i] = SpComm_GlobalSum(n_events[i]);

    if (NumEvents[i] > 0) {

      NumFreeEvents[i] = SpComm_GlobalSum(qfree_events[i].length());
      NumMessages[i] = SpComm_GlobalSum(n_messages[i]);
      NumFreeMessages[i] = SpComm_GlobalSum(qfree_messages[i].length());
      Bytes[i] = NumEvents[i]*sizeof_event[i]
	       + NumMessages[i]*sizeof_message[i];

      TotalBytes += Bytes[i];

      if (((int)strlen(name[i])) > Nchars) {
	Nchars = strlen(name[i]);
      }

    }

  }

  if (node == 0) {
    cout << "Event Memory usage:" << endl;
    cout << endl;
  }

  for (i=0; i<ntypes; i++) {

    if ((node == 0) && (NumEvents[i] > 0)) {
      cout << setw(Nchars+3) << name[i]
	   << "   Id = " << setw(4) << i
	   << "   Nevents = " << setw(7) << NumEvents[i]
	   << "   Nfree = " << setw(7) << NumFreeEvents[i]
	   << "   Nmessages = " << setw(7) << NumMessages[i]
	   << "   Nfree = " << setw(7) << NumFreeMessages[i]
	   << "   Bytes = " << setw(10) << Bytes[i]
	   << endl;
    }

  }

  NumHolders = SpComm_GlobalSum(n_holders);
  NumFreeHolders = SpComm_GlobalSum(qfree_holders.length());
  TotalBytes += NumHolders * sizeof(C_HOLDER);

  if (node == 0) {

    cout << endl;

    cout << "   Nholders = " << NumHolders
	 << ", Nfree = " << NumFreeHolders
	 << ", Bytes = " << NumHolders * sizeof(C_HOLDER)
	 << endl;

    cout << "   Total bytes = " << TotalBytes << endl;

  }

  delete [] NumEvents;
  delete [] NumFreeEvents;
  delete [] NumMessages;
  delete [] NumFreeMessages;
  delete [] Bytes;

}
