// SpOptMessager.C method file

#include "SpStdio.H"
#include "SpOptMessager.H"
#include "speedes_comm.H"
#include "Merc_types.H"
#include "PO.H"
#include "SpPoDataBase.H"
#include "SpTimeMgtAlg.H"
#include "SpClusterMgr.H"

#include "SpMemory.H"
#include "RB_DefineClass.H"
#ifdef __IBMCPP__
namespace {
#endif
RB_DEFINE_CLASS_PRIMATIVE(char)
#ifdef __IBMCPP__
  }
#endif


extern SpPoDataBase *MessageDataBase;
extern SpClusterMgr *GlobalClusterMgr;

/************************************************************************
* TWOSMESS : TWOSMESS constructor					*
************************************************************************/
SpOptMessager::SpOptMessager(int nd, int n_nds, SpEventTypes *evt) {
  int n_hash;

  node = nd;
  n_nodes = n_nds;
  evtype = evt;
  messcount = 0;

  nanti = 0;

  seed = node;

  n_hash = 999;
  ham = new C_HASH(n_hash);

}

/************************************************************************
* send_messages : send the time warp messages out			*
************************************************************************/
void SpOptMessager::send_messages(SpEvent *event) {
  C_HOLDER *holder;
  SpMsg *header;
  SpAntiMsg *antimess;
  C_QUEUE *mque;
  C_QUEUE *amque;
  int i,len,nd,size;
  char *buff;
  int local_evtype, local_obtype;

  mque = event->get_mque();
  amque = event->get_amque();

  len = mque->length();

  for (i=0; i<len; i++) {

    holder = (C_HOLDER *)mque->pop_top();
    holder->get(buff, size, nd);
    header = (SpMsg *)buff;

    local_evtype = header->evtype;
    local_obtype = header->obtype;

    header->evtype = GlobalClusterMgr->
      ConvertLocalToGlobalEventId(header->evtype);

    header->obtype = GlobalClusterMgr->
      ConvertLocalToGlobalMgrId(header->obtype);

    header->rbid = seed;

    antimess = new_antimess();
    antimess->set_time_tag(header->time_tag);
    antimess->set_obid(header->objid);
    antimess->set_obtype(header->obtype);
    antimess->set_rbid(header->rbid);
    antimess->set_node(nd);
    antimess->set_etype(header->evtype);
    antimess->set_bytes(header->bytes);

    seed += n_nodes;

    SpComm_Send(nd, EVENT_MESSAGE_TYPE, size, (char *)(header));

    amque->push_bot(antimess);
    messcount++;

    // restore local event type & global obj type
    header->evtype = local_evtype;
    header->obtype = local_obtype;

    evtype->delete_message(holder);

  }

}

/************************************************************************
* receive_message : receive an twosmessage if there is one		*
************************************************************************/
SpEvent *SpOptMessager::receive_message(int &status) {
  SpAntiMsg *antimess;
  SpMsg *header;
  SpEvent *event;
  int nbytes;
  
  event = NULL;
  status = 0;

  if ((header = (SpMsg*)
       SpComm_GetPendingMessage(EVENT_MESSAGE_TYPE, nbytes))){

      header->evtype = GlobalClusterMgr->
	ConvertGlobalToLocalEventId(header->evtype);

      header->obtype = GlobalClusterMgr->
	ConvertGlobalToLocalMgrId(header->obtype);

    if (SpTimeMgtAlg::DoingCheckpoints()){
      PO_ADD_char((char*)header, nbytes, MessageDataBase);
    }

    messcount--;

    status = 1;

    antimess = (SpAntiMsg *)ham->dehash(header->rbid);
    if (antimess == NULL) {

      event = (SpEvent *)evtype->message_event((char *)header);
      event->set_from_SpComm();
      event->set_rbid(header->rbid);

      antimess = new_antimess();
      antimess->set_time_tag(header->time_tag);
      antimess->set_obid(header->objid);
      antimess->set_obtype(header->obtype);
      antimess->set_rbid(header->rbid);
      antimess->set_event(event);
      antimess->set_node(ANY);

      ham->add(antimess);

    }else{

      delete_antimess(antimess);
      delete [] ((char*) header);

    }

  }

  return event;

}

/************************************************************************
* send_antimessages : send the time warp antimessages out		*
************************************************************************/
void SpOptMessager::send_antimessages(SpEvent *event) {
  SpAntiMsg *antimess;
  C_QUEUE *amque;
  int i,len;

  amque = event->get_amque();
  len = amque->length();

  for (i=0; i<len; i++) {

    antimess = (SpAntiMsg *)amque->pop_top();

    SpComm_Send(antimess->get_node(), EVENT_ANTIMESSAGE_TYPE,
		sizeof(SpAntiMsg), (char *)(antimess));

    messcount++;

    delete_antimess(antimess);

  }

}

/************************************************************************
* receive_antimessage : receive a time warp antimessage 		*
************************************************************************/
SpEvent *SpOptMessager::receive_antimessage(int &status) {
  SpAntiMsg *antimess1;
  SpAntiMsg *antimess2;
  SpEvent *event;
  int rbid;
  int nbytes;

  event = NULL;
  status = 0;

  if ((antimess1 = (SpAntiMsg*)
       SpComm_GetPendingMessage(EVENT_ANTIMESSAGE_TYPE, nbytes))){

/*
    antimess1->set_etype(GlobalClusterMgr->
      ConvertGlobalToLocalEventId(antimess1->get_etype()));

    antimess1->set_obtype(GlobalClusterMgr->
      ConvertGlobalToLocalMgrId(antimess1->get_obtype()));
*/

    messcount--;
    status = 1;

    rbid = antimess1->get_rbid();
    antimess2 = (SpAntiMsg *)ham->dehash(rbid);

    if (antimess2 != NULL) {

      event = antimess2->get_event();
      event->set_rbid(-1);

      delete_antimess(antimess2);

    }else{

      antimess2 = new_antimess();
      memcpy((char *)antimess2,(char *)antimess1,sizeof(SpAntiMsg));
      ham->add(antimess2);

    }

    delete [] ((char*) antimess1);
  }

  return event;

}

/************************************************************************
* new_antimess : get a new antimessage					*
************************************************************************/
SpAntiMsg *SpOptMessager::new_antimess() {
  SpAntiMsg *antimess;

  if (qfree_antimess.length() == 0) {

    antimess = new SpAntiMsg();
    if (antimess == NULL) {
      fprintf(stderr,"antimess out of memory\n");
    }else{
      nanti++;
    }

  }else{
    antimess = (SpAntiMsg *)qfree_antimess.pop_top();
  }

  return antimess;

}

/************************************************************************
* free_antimess : free the antimessages for this event			*
************************************************************************/
void SpOptMessager::free_antimess(C_QUEUE *qanti, int r) {
  SpAntiMsg *antimess;

  if (qanti->length()) qfree_antimess.concat(qanti);
  if (r >= 0) {
    antimess = (SpAntiMsg *)ham->dehash(r);
    if (antimess != NULL) {
      qfree_antimess.push_bot(antimess);
    }else{
      fprintf(stderr,"error, no antimessage\n");
    }
  }

}

/************************************************************************
* check_messcount : check the global message count			*
************************************************************************/
int SpOptMessager::check_messcount() {

  int nmess = SpComm_GlobalSum(messcount);
  if (nmess == 0) messcount = 0;
  return nmess;

}

/************************************************************************
* print : print out information concerning the twosmess object		*
************************************************************************/
void SpOptMessager::print() {
  int i,len;
  C_QUEUE *qall;
  SpAntiMsg *antimess;
  double tim,tmin;

//...... test code to check the hash object

  ham->combine();
  qall = ham->get_all();
  len = qall->length();

  tmin = 1.0e20;
  antimess = (SpAntiMsg *)qall->ptop();
  for (i=0; i<len; i++) {
    tim = antimess->get_time_tag();
    if (tim < tmin) tmin = tim;
    antimess = (SpAntiMsg *)antimess->get_link();
  }

  fprintf(stderr,"qfree %d, ham %d, tmin %f\n",nanti,len,tmin);

}
