// SpStats.H header file

#ifndef statistics_object
#define statistics_object

#define CYCLE         	0x00000001
#define GVT		0x00000002
#define CPU		0x00000004
#define WALL		0x00000008
#define PHASE1		0x00000010
#define PHASE2		0x00000020
#define EVENTS		0x00000040
#define EVENTSCYCLE	0x00000080
#define ROLLBACKS	0x00000100
#define MESSAGES	0x00000200
#define ANTIMESSAGES	0x00000400
#define CANCELS		0x00000800
#define EXMESSAGES	0x00001000
#define EVTGVT		0x00002000
#define STAR		0x00004000
#define PROCESSED      	0x00008000
#define COMMIT		0x00010000
#define PROCEFF		0x00020000
#define LVT		0x00040000
#define BTW		0x00080000

#include "SpDataParser.H"

/************************************************************************
*	statistics object 						*
************************************************************************/
class SpStats {

  private:

    int cycle;			// current cycle number
    double gvt;			// global virtual time
    double oldgvt;	       	// gvt from the last time
    double oldwall;		// old wall clock from the last time
    double cpu;			// cpu time used
    double wall;		// wall clock time
    double star;		// simulation time advancement rate
    int phase1;			// number of events processed optimistically
    int phase2;			// number of events processed conservatively
    int events;			// total number of processed events
    int nevtgvt;		// total number of gvt processed events
    int rollbacks;		// total number of rollbacks
    int messages;		// total number of inter-node messages
    int antimessages;		// total number of antimessages
    int cancels;		// total number of events canceled
    int exmessages;		// total number of external messages
    double TotalProcessingTime;	// total event processing time
    double TotalCommittedTime;	// total committed event processing time
    double LocalTime;		// local time of my node

    double PhaseBTB;		// time spent in BTB event processing
    double PhaseFBen;		// time spent entering the fuzzy barrier
    double PhaseFBex;		// time spent exiting the fuzzy barrier
    double PhaseGVT;		// time spent updating GVT
    double PhaseCMT;		// time spent committing events

    int print_flag;		// flag that determines what is printed
    double tprint;		// last printout time
    int ReportTime;		// time between reporting info in seconds
    const int n_nodes;		// number of nodes
    const int node;		// my node number

    int output_fd;              // output goes to this file descriptor
    int StatsByNode;

  protected:

  public:

    SpStats(SpDataParser *parser, int n_nodes_arg, int node_arg);

    void set_gvt(double g) {gvt = g;}
    void set_cpu(double c) {cpu += c;}
    void set_wall(double w) {wall = w;}
    void set_star(double s) {star = s;}
    void set_cycle() {cycle++;}
    void set_phase1(int p1) {phase1 = p1;}
    void set_phase2(int p2) {phase2 = p2;}
    void set_events(int e) {events += e;}
    void set_nevtgvt(int e) {nevtgvt = e;}
    void set_rollbacks(int r) {rollbacks += r;}
    void set_messages(int m) {messages += m;}
    void set_antimessages(int a) {antimessages += a;}
    void set_cancels(int c) {cancels = c;}
    void set_exmessages(int e) {exmessages += e;}

    void SetTotalProcessingTime(double totalProcessingTime) {
      TotalProcessingTime = totalProcessingTime;
    }

    void SetTotalCommittedTime(double totalCommittedTime) {
      TotalCommittedTime = totalCommittedTime;
    }

    void SetLocalTime(double localTime) {
      LocalTime = localTime;
    }

    void SetPhaseBTB(double phaseBTB) {PhaseBTB = phaseBTB;}
    void SetPhaseFBen(double phaseFBen) {PhaseFBen = phaseFBen;}
    void SetPhaseFBex(double phaseFBex) {PhaseFBex = phaseFBex;}
    void SetPhaseGVT(double phaseGVT) {PhaseGVT = phaseGVT;}
    void SetPhaseCMT(double phaseCMT) {PhaseCMT = phaseCMT;}

    void set_ReportTime(int r) {ReportTime = r;}

    void print();
    void fast_print(double gvt);
    void WriteNodeStatistics();

};


typedef SpStats C_STATISTICS;

#endif
