// O_SpScatterDecomp.C source file

//#define DEBUG_SCATTER_DECOMP

#include "O_SpScatterDecomp.H"

extern SpSimObj *CurrentSimobj;

//-----------------------------------------------------------------------------
// O_SpScatterDecomp constructor
//-----------------------------------------------------------------------------
O_SpScatterDecomp::O_SpScatterDecomp() {}

//-----------------------------------------------------------------------------
// CreateSimObjs - creates all of the SimObjs using a card dealing
// strategy.
//-----------------------------------------------------------------------------
void O_SpScatterDecomp::CreateSimObjs(int numObjs, int replication) {
  int currentId;
  int i;
  int seed;
  SpSimObj *simObj;
  char *name;
  char s[120];

  SimObjMgrNames->Insert(this, GetSimObjMgrName());
  SimObjMgrTypes->Insert(this, OBJECT_TYPE);

/*
** Find out how many nodes are in our cluster along with our cluster node.
*/

  GetClusterInfo(GetClusterName(), NumClusterNodes, ClusterNode);

  N_TOT = numObjs;
  DecompInfo(N_TOT, NumClusterNodes, ClusterNode,
	     FirstClusterNode, N_LOC, Offset);

/*
** Reserve room for N_LOC objects and then create them in a loop.
*/

  reserve_n_objects(N_LOC);

  currentId = Offset;
  for (i=0; i<N_LOC; i++) {

/*
** Create the simObj
*/

    simObj = CreateSimObj();

/*
** Generate a unique seed for the simObj
*/

    seed = 1 + NumTotCreated + currentId + (N_TOT*replication);
    simObj->set_SEED(seed);

/*
** Pass in the Number Id (e.g., 0,1,2,...,N_TOT-1) for the object
*/

  simObj->SetNumId(currentId);

/*
** Generate the name of the simObj
*/
    if ((name = GetSimObjMgrName()) && (Parser==NULL)) {
      sprintf(s, "%s %d", name, currentId);
      simObj->set_NAME(strdup(s));
    }

/*
** Pass the object into SPEEDES and then increment the currentId
*/

    define_object(simObj);
    currentId += NumClusterNodes;

  }

  TOTAL_OBJECTS += N_TOT;
  NumTotCreated += N_TOT;

/*
** Initialize the simObj
*/

  SpSimObj *tempSimObj = CurrentSimobj;
  if (Parser == NULL) {
    for (i=0; i<N_LOC; i++) {
      simObj = get_obj(i);
      CurrentSimobj = simObj;
      simObj->Init();
    }
  }
  CurrentSimobj = tempSimObj;

}

//-----------------------------------------------------------------------------
// CreateSimObjs - creates all of the SimObjs using a card dealing
// strategy.
//-----------------------------------------------------------------------------
void O_SpScatterDecomp::CreateSimObjs(char *parFileName, int replication) {
  int i, j;
  SpSimObj *simObj;
  void *Mark;

  Parser = new SpDataParser(parFileName);
  Mark = Parser->GoTo("Objects", NULL);
  N_TOT = Parser->GetNumReferences(Mark, "OBJECT");

  CreateSimObjs(N_TOT, replication);

  Parser->GoToFirstReference(Mark, "OBJECT");
  j = 0;
  for (i=0; i<N_TOT; i++) {

    int *num = new int;
    *num = i;
    SimObjNames.Insert(num, Parser->GetReferenceName());

    SpSimObj *tempSimObj = CurrentSimobj;
    if ((i-Offset+NumClusterNodes)%NumClusterNodes == 0) {
      simObj = get_obj(j);
      simObj->set_NAME(Parser->GetReferenceName());
      CurrentSimobj = simObj;
      simObj->InitParser(Parser);
      j++;
    }
    CurrentSimobj = tempSimObj;
    Parser->GoToNextReference(Mark, "OBJECT");
  }

}

//-----------------------------------------------------------------------------
// DecompInfo - returns information needed for decomposing the objects to
// different cluster nodes. In this case, the algorithm is card dealing.
//-----------------------------------------------------------------------------
void O_SpScatterDecomp::DecompInfo(int numTotalObjs,
			    int numClusterNodes,
			    int clusterNode,
			    int &firstClusterNode,
			    int &numLoc,
			    int &offset) {

  firstClusterNode = NumTotCreated % numClusterNodes;
  offset = (clusterNode - firstClusterNode + numClusterNodes)
    % numClusterNodes;
  numLoc = (numTotalObjs - offset) / numClusterNodes;
  if ((numTotalObjs - offset) % numClusterNodes > 0) numLoc++;
}

//-----------------------------------------------------------------------------
// GetObjHandle - returns the object handle for an object who's numerical
// Id is passed in. In this case, the  card dealing algorithm is used to do
// the computation.
//-----------------------------------------------------------------------------
SpObjHandle O_SpScatterDecomp::GetObjHandle(int numId) {

  int cn = (NumClusterNodes + (numId % NumClusterNodes)
	    + (ClusterNode - Offset)) % NumClusterNodes;
  int node = GetNode(GetClusterName(), cn);
  int localId = numId / NumClusterNodes;

  SpObjHandle objHandle;
  objHandle.NodeId = node;
  objHandle.ObjManId = OBJECT_TYPE;
  objHandle.SimObjId = localId;

  return objHandle;

}
