#ifndef RB_SpFrameworkFuncs_Header
#define RB_SpFrameworkFuncs_Header

#include "SpStdlib.H"
#include "SpString.H"

#if defined(__GNUC__) && defined(__INSURE__)
// memcpy is built into g++, which means that there is no standard
// include file which has the prototype of it.  But Insure needs the
// prototype so I write the prototype myself:
void *memcpy(void*, const void*, size_t);
#endif

#include "SpAltMgr.H"
#include "SpAltAbortChange.H"
#include "SpAltBlockChange.H"
#include "SpAltObjectChange.H"
#include "SpAltAntiMsgChange.H"
#include "SpAltMemoryValue.H"
#include "SpAltMemoryExchange.H"
#include "SpAltHostDataSend.H"
#include "SpAltSubscribedDataSend.H"

class SpPoDataBase;
class SpCancelHandle;

extern C_FREE_LIST* SpCurrentFreeList;

SpCancelHandle* RB_NEW_CANCEL_HANDLE();
void RB_DELETE_CANCEL_HANDLE( SpCancelHandle* handle );

/**********************************************************************
 * 
 **********************************************************************/
#define RB_assert( assertion )						\
if( ! ( assertion ) ) {							\
  if( SpCurrentAltMgr->GetOptimistic() ) {				\
    RB_cout								\
      << "********** ASSERTION FAILED **********" << endl		\
      << "Assertion   :  " << #assertion          << endl		\
      << "File Name   :  " << __FILE__            << endl		\
      << "Line Number :  " << __LINE__            << endl;		\
    SpAltAbortChange* altItem = ALLOCATE_SpAltAbortChange();		\
    altItem->Alter();							\
    SpCurrentAltMgr->Append( altItem );					\
  }									\
  else {								\
    cout								\
      << "********** ASSERTION FAILED **********" << endl		\
      << "Assertion   :  " << #assertion          << endl		\
      << "File Name   :  " << __FILE__            << endl		\
      << "Line Number :  " << __LINE__            << endl;		\
    abort();								\
  }									\
}									\
/*********************************************************************/

/**********************************************************************
 * 
 **********************************************************************/
inline
void
AddGvtBarrier( C_QUEUE* queue, C_ITEM* block ) {
  SpAltBlockChange* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltBlockChange();
    altItem->Alter( queue, block );
    SpCurrentAltMgr->Insert( altItem );
  }

  queue->push_bot( block );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_DEFINE_FREE( C_FREE_LIST* freelist ) {
  SpCurrentFreeList = freelist;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
C_ITEM*
RB_FREE_NEW( C_FREE_LIST* freelist, int type ) {
  SpAltObjectChange* altItem;
  C_ITEM* object;

  object = ( C_ITEM* ) freelist->new_object( type );
  if( object == NULL ) {
    return NULL;
  }
  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltObjectChange();
    altItem->Alter( freelist, object, 1 );
    SpCurrentAltMgr->Insert( altItem );
  }

  return object;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
C_ITEM*
RB_FREE_NEW( int type ) {
  return RB_FREE_NEW( SpCurrentFreeList, type );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
C_ITEM*
RB_FREE_GENERATE( C_FREE_LIST* freelist, char* buffer ) {
  SpAltObjectChange* altItem;
  C_ITEM* object;

  object = ( C_ITEM* ) freelist->generate( buffer );
  if( object == NULL ) {
    return NULL;
  }
  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltObjectChange();
    altItem->Alter( freelist, object, 1 );
    SpCurrentAltMgr->Insert( altItem );
  }

  return object;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
C_ITEM*
RB_FREE_GENERATE( char* buffer ) {
  return RB_FREE_GENERATE( SpCurrentFreeList, buffer );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_FREE_DELETE( C_FREE_LIST* freelist, C_ITEM* object ) {
  SpAltObjectChange* altItem;

  if( object == NULL ) {
    return;
  }
  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltObjectChange();
    altItem->Alter( freelist, object, 0 );
    SpCurrentAltMgr->Insert( altItem );
  }else{
    freelist->delete_object( ( C_ITEM* ) object );
  }
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_FREE_DELETE( C_ITEM* object ) {
  RB_FREE_DELETE( SpCurrentFreeList, object );
}

/**********************************************************************
 *
 **********************************************************************/
inline
void*
RB_memcpy( void* src, const void* dest, size_t size ) {
  SpAltMemoryValue* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltMemoryValue();
    altItem->Alter( src, size );
    SpCurrentAltMgr->Insert( altItem );
  }
  return( memcpy( src, dest, size ) );
}

/**********************************************************************
 *
 **********************************************************************/
inline
void
RB_xmem( void* value1, void* value2, size_t size ) {
  SpAltMemoryExchange* altItem;
  void* temp;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltMemoryExchange();
    altItem->Alter( value1, value2, size );
    SpCurrentAltMgr->Insert( altItem );
  }

  temp = malloc( size );
  memcpy( temp, value1, size );
  memcpy( value1, value2, size );
  memcpy( value2, temp, size );
  free( temp );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RBSendData( SpSimTime time, int externalId, char* data, int bytes, int globalId, SpHostUser* hostUser ) {
  SpAltHostDataSend* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltHostDataSend();
    altItem->Alter( hostUser, data, externalId, globalId, bytes, time );
    SpCurrentAltMgr->Append( altItem );
  }
  else {
    hostUser->SendData( time, externalId, data, bytes, globalId );
  }
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RBSendSubscribedData( double time, char* name, char* data, int bytes, int globalId, SpHostUser* hostUser ) {
  SpAltSubscribedDataSend* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltSubscribedDataSend();
    altItem->Alter( hostUser, data, name, globalId, bytes, time );
    SpCurrentAltMgr->Append( altItem );
  }
  else {
    hostUser->SendSubscribedData( time, name, data, bytes, globalId );
  }
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
EXCHANGE( int& value1, int& value2 ) {
  int temp = value1;
  value1 = value2;
  value2 = temp;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
EXCHANGE( int& value1, float& value2 ) {
  float temp = value1;
  value1 = ( int ) value2;
  value2 = temp;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
EXCHANGE( int& value1, double& value2 ) {
  double temp = value1;
  value1 = ( int ) value2;
  value2 = temp;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
EXCHANGE( float& value1, int& value2 ) {
  int temp = ( int ) value1;
  value1 = value2;
  value2 = temp;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
EXCHANGE( float& value1, float& value2 ) {
  float temp = value1;
  value1 = value2;
  value2 = temp;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
EXCHANGE( float& value1, double& value2 ) {
  double temp = value1;
  value1 = value2;
  value2 = temp;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
EXCHANGE( double& value1, int& value2 ) {
  int temp = ( int ) value1;
  value1 = value2;
  value2 = temp;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
EXCHANGE( double& value1, float& value2 ) {
  float temp = value1;
  value1 = value2;
  value2 = temp;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
EXCHANGE( double& value1, double& value2 ) {
  double temp = value1;
  value1 = value2;
  value2 = temp;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
EXCHANGE( unsigned long& value1, unsigned long& value2 ) {
  unsigned long temp = value1;
  value1 = value2;
  value2 = temp;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
EXCHANGE( void*& value1, void*& value2 ) {
  void* temp = value1;
  value1 = value2;
  value2 = temp;
}



char* RB_strdup( char* src );

#endif

