#ifndef RB_SpString_Header
#define RB_SpString_Header

#include "SpString.H"

#include "SpAltMgr.H"
#include "SpAltStringValue.H"

#include "RB_SpDefineClass.H"

/**********************************************************************
 * Class       : RB_SpString
 * Summary     : Rollbackable SpString
 * Description : Use this just as you would a normal character pointer on the
 *              right-hand side of equations when that character pointer is part
 *              of the state of a C_SIMOBJ or non-C_SIMOBJ object (ie, when it
 *              needs to be rollbackable). In addition to regular use on the
 *              right-hand side of equations as a character pointer, this object
 *              can be set equal to another character pointer using the =
 *              operator, and the string that the other character pointer points
 *              to will be rollbackably copied to this character pointer. Memory
 *              management is handled automatically. The = operator is the only
 *              left-hand side of equations operator supported.
 **********************************************************************/
class RB_SpString {
public:
  RB_SpString() : Value( NULL ) {}
  ~RB_SpString() { if (Value) free( Value ); Value = NULL; }
  operator char*() const;
  char* operator =( char* value );
  void Copy( char* value );
private:
  char* Value;
};

/**********************************************************************
 * 
 **********************************************************************/
inline
RB_SpString::operator char*() const {
  return Value;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
char*
RB_SpString::operator =( char* value ) {
  SpAltStringValue* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltStringValue();
    altItem->Alter( & Value );
    SpCurrentAltMgr->Insert( altItem );
  }
  return( Value = ( value ? strdup( value ) : ( char* ) NULL ) );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_SpString::Copy( char* value ) {
  if( Value ) {
    free( Value );
  }
  Value = ( value ? strdup( value ) : ( char* ) NULL );
}

typedef RB_SpString STATE_STRING;
typedef RB_SpString RB_STRING;
typedef RB_SpString RB_string;

//RB_DEFINE_CLASS( RB_SpString );

#endif

