#ifndef RB_double_Header
#define RB_double_Header

#include "SpAltMgr.H"
#include "SpAltDoubleDelta.H"
#include "SpAltDoubleDivisor.H"
#include "SpAltDoubleFactor.H"
#include "SpAltDoubleValue.H"
#include "RB_SpDefineClass.H"

/**********************************************************************
 * Class       : RB_double
 * Summary     : Rollbackable double
 * Description : Use this just as you would a normal double when that double
 *              is part of the state of a C_SIMOBJ or non-C_SIMOBJ object (ie,
 *              when it needs to be rollbackable)
 **********************************************************************/
class RB_double {  
public:
  RB_double() : Value( 0.0 ) {}
  RB_double( double value ) : Value( value ) {}
  operator double() const;
  double operator ++();
  double operator --();
  double operator =( const RB_double& value);
  double operator =( double value );
  double operator +=( double delta );
  double operator -=( double delta );
  double operator *=( double factor );
  double operator /=( double factor );
private:
  double SetDelta( double delta );
  double Value;
};

/**********************************************************************
 * 
 **********************************************************************/
inline
double
RB_double::SetDelta( double delta ) {
  SpAltDoubleDelta* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltDoubleDelta();
    altItem->Alter( & Value, delta );
    SpCurrentAltMgr->Insert( altItem );
  }
  return( Value += delta );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
RB_double::operator double() const {
  return Value;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
double
RB_double::operator ++() {
  return SetDelta( 1.0 );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
double
RB_double::operator --() {
  return SetDelta( - 1.0 );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
double
RB_double::operator =( double value ) {
  SpAltDoubleValue* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltDoubleValue();
    altItem->Alter( & Value );
    SpCurrentAltMgr->Insert( altItem );
  }
  return( Value = value );
}
/**********************************************************************
 * 
 **********************************************************************/
inline
double
RB_double::operator =( const RB_double& value ) {
  SpAltDoubleValue* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltDoubleValue();
    altItem->Alter( & Value );
    SpCurrentAltMgr->Insert( altItem );
  }
  return( Value = value.Value );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
double
RB_double::operator +=( double delta ) {
  return SetDelta( delta );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
double
RB_double::operator -=( double delta ) {
  return SetDelta( - delta );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
double
RB_double::operator *=( double factor ) {
  SpAltDoubleFactor* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltDoubleFactor();
    altItem->Alter( & Value, factor );
    SpCurrentAltMgr->Insert( altItem );
  }
  return( Value *= factor );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
double
RB_double::operator /=( double divisor ) {
  SpAltDoubleDivisor* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltDoubleDivisor();
    altItem->Alter( & Value, divisor );
    SpCurrentAltMgr->Insert( altItem );
  }
  return( Value /= divisor );
}

typedef RB_double STATE_DOUBLE;
typedef RB_double RB_DOUBLE;
typedef RB_double STATE_FLOAT;
typedef RB_double RB_FLOAT;
typedef RB_double RB_float;

//RB_DEFINE_CLASS( RB_double );

#endif

