#ifndef RB_int_Header
#define RB_int_Header

#include "SpAltMgr.H"
#include "SpAltIntAnd.H"
#include "SpAltIntDelta.H"
#include "SpAltIntDivisor.H"
#include "SpAltIntFactor.H"
#include "SpAltIntOr.H"
#include "SpAltIntRemainder.H"
#include "SpAltIntShiftLeft.H"
#include "SpAltIntShiftRight.H"
#include "SpAltIntValue.H"
#include "SpAltIntXor.H"
#include "RB_SpDefineClass.H"

/**********************************************************************
 * Class       : RB_int
 * Summary     : Rollbackable integer
 * Description : Use this just as you would a normal integer when that integer
 *              is part of the state of a C_SIMOBJ or non-C_SIMOBJ object (ie,
 *              when it needs to be rollbackable)
 **********************************************************************/
class RB_int {
public:
  RB_int() : Value( 0 ) {}
  RB_int( int value ) : Value( value ) {}
  operator int() const;
  int operator ++();
  int operator --();
  int operator ++( int );
  int operator --( int );
  int operator =( int value );
  int operator +=( int delta );
  int operator -=( int delta );
  int operator *=( int factor );
  int operator /=( int factor );
  int operator %=( int value );
  int operator ^=( int value );
  int operator &=( int value );
  int operator |=( int value );
  int operator >>=( int value );
  int operator <<=( int value );
  void Exchange( int& var );
private:
  int SetDelta( int delta );
  int Value;
};

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_int::SetDelta( int delta ) {
  SpAltIntDelta* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltIntDelta();
    altItem->Alter( & Value, delta );
    SpCurrentAltMgr->Insert( altItem );
  }
  return( Value += delta );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
RB_int::operator int() const {
  return Value;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_int::operator++() {
  return SetDelta( 1 );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_int::operator --() {
  return SetDelta( - 1 );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_int::operator ++( int ) {
  return( SetDelta( 1 ) - 1 );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_int::operator --( int ) {
  return( SetDelta( - 1 ) + 1 );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_int::operator =( int value ) {
  SpAltIntValue* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltIntValue();
    altItem->Alter( & Value );
    SpCurrentAltMgr->Insert( altItem );
  }
  return( Value = value );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_int::operator +=( int delta ) {
  return SetDelta( delta );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_int::operator -=( int delta ) {
  return SetDelta( - delta );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_int::operator *=( int factor ) {
  SpAltIntFactor* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltIntFactor();
    altItem->Alter( & Value, factor );
    SpCurrentAltMgr->Insert( altItem );
  }
  return( Value *= factor );

}

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_int::operator /=( int divisor ) {
  SpAltIntDivisor* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltIntDivisor();
    altItem->Alter( & Value, divisor );
    SpCurrentAltMgr->Insert( altItem );
  }
  return( Value /= divisor );

}

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_int::operator %=( int divisor ) {
  SpAltIntRemainder* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltIntRemainder();
    altItem->Alter( & Value, divisor );
    SpCurrentAltMgr->Insert( altItem );
  }
  return( Value %= divisor );

}

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_int::operator ^=( int mask ) {
  SpAltIntXor* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltIntXor();
    altItem->Alter( & Value, mask );
    SpCurrentAltMgr->Insert( altItem );
  }
  return( Value ^= mask );

}

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_int::operator &=( int mask ) {
  SpAltIntAnd* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltIntAnd();
    altItem->Alter( & Value, mask );
    SpCurrentAltMgr->Insert( altItem );
  }
  return( Value &= mask );

}

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_int::operator |=( int mask ) {
  SpAltIntOr* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltIntOr();
    altItem->Alter( & Value, mask );
    SpCurrentAltMgr->Insert( altItem );
  }
  return( Value |= mask );

}

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_int::operator <<=( int count ) {
  SpAltIntShiftLeft* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltIntShiftLeft();
    altItem->Alter( & Value, count );
    SpCurrentAltMgr->Insert( altItem );
  }
  return( Value <<= count );

}

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_int::operator >>=( int count ) {
  SpAltIntShiftRight* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltIntShiftRight();
    altItem->Alter( & Value, count );
    SpCurrentAltMgr->Insert( altItem );
  }
  return( Value >>= count );

}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_int::Exchange( int& var ) {
  int temp;
  
  temp = var;
  var = Value;
  Value = temp;
}

typedef RB_int STATE_INT;
typedef RB_int RB_INT;

//RB_DEFINE_CLASS( RB_int );

#endif

