#ifndef RB_ostream_Header
#define RB_ostream_Header

#include "SpAltMgr.H"
#include "SpAltOstreamPut.H"
#include "SpSimTime.H"

/**********************************************************************
 * Class       : RB_ostream
 * Summary     : Rollbackable ostream
 * Description : Use this just as you would a normal ostream when that ostream
 *              is part of the state of a C_SIMOBJ or non-C_SIMOBJ object (ie,
 *              when it needs to be rollbackable)
 **********************************************************************/
class RB_ostream {
  typedef ostream& ( *SpOmanip ) ( ostream& );
public:
  RB_ostream( ostream& ostream );
  RB_ostream& operator <<( char data );
  RB_ostream& operator <<( short data );
  RB_ostream& operator <<( int data );
  RB_ostream& operator <<( long data );
  RB_ostream& operator <<( unsigned char data );
  RB_ostream& operator <<( unsigned short data );
  RB_ostream& operator <<( unsigned int data );
  RB_ostream& operator <<( unsigned long data );
  RB_ostream& operator <<( float data );
  RB_ostream& operator <<( double data );
  RB_ostream& operator <<( void* data );
  RB_ostream& operator <<( char* data );
  RB_ostream& operator <<( const char* data );
  RB_ostream& operator <<( unsigned char* data );
  RB_ostream& operator <<( const SpSimTime data);
  RB_ostream& operator <<( SpOmanip manipulator );
private:
  ostream& Ostream;
};

/**********************************************************************
 * 
 **********************************************************************/
inline
RB_ostream::RB_ostream( ostream& ostream ) :
  Ostream( ostream ) {
}

/**********************************************************************
 *
 **********************************************************************/
#define DEFINE_RB_OSTREAM_TYPE( type )					\
									\
DEFINE_ALT_OSTREAM_TYPE_BY_NAME( type, type )				\
									\
inline									\
RB_ostream&								\
operator <<( RB_ostream& rbOstream, type data ) {			\
  SpAltOstreamPut##type##* altItem;					\
									\
  if( SpCurrentAltMgr->GetOptimistic() ) {				\
    altItem = ALLOCATE_SpAltOstreamPut##type##();			\
    altItem->Alter( & rbOstream.Ostream, data );			\
    SpCurrentAltMgr->Append( altItem );					\
  }									\
  else {								\
    rbOstream.Ostream << data;						\
  }									\
  return rbOstream;							\
}									\
/*********************************************************************/

/**********************************************************************
 *
 **********************************************************************/
#define PLUGIN_RB_OSTREAM_TYPE( type )					\
									\
PLUGIN_MEMPOOL( SpAltOstreamPut##type )					\
/*********************************************************************/

/**********************************************************************
 *
 **********************************************************************/
#define DEFINE_RB_OSTREAM_TYPE_BY_NAME( type, name )			\
									\
inline									\
RB_ostream&								\
RB_ostream::operator <<( type data ) {					\
  SpAltOstreamPut##name##* altItem;					\
									\
  if( SpCurrentAltMgr->GetOptimistic() ) {				\
    altItem = ALLOCATE_SpAltOstreamPut##name##();			\
    altItem->Alter( & this->Ostream, data );				\
    SpCurrentAltMgr->Append( altItem );					\
  }									\
  else {								\
    this->Ostream << data;						\
  }									\
  return *this;								\
}									\
/*********************************************************************/

/**********************************************************************
 *
 **********************************************************************/
#define DEFINE_RB_OSTREAM_QUALIFIED_TYPE_BY_NAME( qual, type, name )	\
									\
inline									\
RB_ostream&								\
RB_ostream::operator <<( qual type data ) {				\
  SpAltOstreamPut##name##* altItem;					\
									\
  if( SpCurrentAltMgr->GetOptimistic() ) {				\
    altItem = ALLOCATE_SpAltOstreamPut##name##();			\
    altItem->Alter( & this->Ostream, data );				\
    SpCurrentAltMgr->Append( altItem );					\
  }									\
  else {								\
    this->Ostream << data;						\
  }									\
  return *this;								\
}									\
/*********************************************************************/

DEFINE_RB_OSTREAM_TYPE_BY_NAME( char, Char )
DEFINE_RB_OSTREAM_TYPE_BY_NAME( short, Short )
DEFINE_RB_OSTREAM_TYPE_BY_NAME( int, Int )
DEFINE_RB_OSTREAM_TYPE_BY_NAME( long, Long )
DEFINE_RB_OSTREAM_QUALIFIED_TYPE_BY_NAME( unsigned, char, UnsignedChar )
DEFINE_RB_OSTREAM_QUALIFIED_TYPE_BY_NAME( unsigned, short, UnsignedShort )
DEFINE_RB_OSTREAM_QUALIFIED_TYPE_BY_NAME( unsigned, int, UnsignedInt )
DEFINE_RB_OSTREAM_QUALIFIED_TYPE_BY_NAME( unsigned, long, UnsignedLong )
DEFINE_RB_OSTREAM_TYPE_BY_NAME( float, Float )
DEFINE_RB_OSTREAM_TYPE_BY_NAME( double, Double )
DEFINE_RB_OSTREAM_TYPE_BY_NAME( void*, VoidPtr )
DEFINE_RB_OSTREAM_TYPE_BY_NAME( char*, CharPtr )
DEFINE_RB_OSTREAM_QUALIFIED_TYPE_BY_NAME( const, char*, ConstCharPtr )
DEFINE_RB_OSTREAM_QUALIFIED_TYPE_BY_NAME( unsigned, char*, UnsignedCharPtr )
DEFINE_RB_OSTREAM_TYPE_BY_NAME( SpOmanip, Omanip )
DEFINE_RB_OSTREAM_QUALIFIED_TYPE_BY_NAME(const, SpSimTime, SpSimTime )

extern RB_ostream RB_cout;
extern RB_ostream RB_cerr;

#endif

