#ifndef RB_voidPtr_Header
#define RB_voidPtr_Header

#include "SpAltMgr.H"
#include "SpAltVoidPtrValue.H"
#include "RB_SpDefineClass.H"

/**********************************************************************
 * Class       : RB_voidPtr
 * Summary     : Rollbackable void pointer
 * Description : Use this just as you would a normal void or character pointer
 *               on the right-hand side of equations when that pointer is part
 *               of the state of a C_SIMOBJ or non-C_SIMOBJ object (ie, when it
 *               needs to be rollbackable). In addition to regular use on the
 *               right-hand side of equations as a void or character pointer,
 *               this object can be rollbackably set equal to another void
 *               pointer using the = operator. The = operator is the only
 *               left-hand side of equations operator supported.
 **********************************************************************/
class RB_voidPtr {
public:
  RB_voidPtr() : Value( NULL ) {}
  operator void*() const;
  void* operator =( void* value );
private:
  void* Value;
};

/**********************************************************************
 * 
 **********************************************************************/
inline
RB_voidPtr::operator void*() const {
  return Value;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_voidPtr::operator =( void* value ) {
  SpAltVoidPtrValue* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltVoidPtrValue();
    altItem->Alter( & Value );
    SpCurrentAltMgr->Insert( altItem );
  }
  return( Value = value );
}

typedef RB_voidPtr STATE_POINTER;
typedef RB_voidPtr RB_POINTER;
typedef RB_voidPtr RB_pointer;

//RB_DEFINE_CLASS( RB_voidPtr );

#endif

