// SpBarrier.H header file defines barriers that limit gvt from advancing

#ifndef SpBarrierGate
#define SpBarrierGate

#include "SpSimTime.H"
#include "SpSimObj.H"
#include "SpObjHandle.H"
#include "SpIostream.H"
#include "SpScheduleEvent.H"

enum SP_BARRIER_REASON {
  SP_BARRIER_GENERIC,	// general purpose barrier
  SP_BARRIER_OUTPUT_MSG,	// outgoing msg may generate input msg
  SP_BARRIER_INPUT_MSG,    	// expecting an input msg
  SP_BARRIER_TIME_GRANT,	// time grant may generate input msg
  SP_BARRIER_NUM_TYPES
};

class SpBarrier : public SpSimTime {

  private:

    SpObjHandle ObjHandle;
    SP_BARRIER_REASON Reason;
    int Gateway;
    SpSimObj *SimObj;
    double TimeLag;

  protected:

  public:

    SpBarrier() {
      Reason = SP_BARRIER_GENERIC;
      Gateway = 0;
      TimeLag = 0.0;
      SimObj = NULL;
    }

    SpBarrier &operator = (SpSimTime &time) {
      *(SpSimTime *)this = time;
      return *this;
    }

    void SetSimObj(SpSimObj *simObj) {
      SimObj = simObj;
      ObjHandle.NodeId = simObj->get_NODE();
      ObjHandle.ObjManId = simObj->get_OBJECT_TYPE();
      ObjHandle.SimObjId = simObj->get_LOCAL_ID();
    }

    SpSimObj *GetSimObj() {return SimObj;}

    void SetObjHandle(SpObjHandle &objHandle) {
      SpSimObjMgr *objMgr = GetSimObjMgr(objHandle.ObjManId);
      SimObj = objMgr->get_obj(objHandle.SimObjId);
      ObjHandle.NodeId = objHandle.NodeId;
      ObjHandle.ObjManId = objHandle.ObjManId;
      ObjHandle.SimObjId = objHandle.SimObjId;
    }

    SpObjHandle &GetObjHandle() {return ObjHandle;}

    void SetGateway(int gateway=1) {Gateway = gateway;}
    void ResetGateway() {Gateway = 0;}
    int GetGateway() {return Gateway;}

    void SetTimeLag(double timeLag) {TimeLag = timeLag;}
    double GetTimeLag() {return TimeLag;}

    SpSimTime GetTimeMinusLag() {
      SpSimTime time;
      time = *this;
      time -= TimeLag;
      return time;
    }

    SpSimTime GetMinGatewayMsgTime() {return SimObj->GetMinGatewayMsgTime();}

    void SetReason(SP_BARRIER_REASON reason) {Reason = reason;}
    SP_BARRIER_REASON GetReason() {return Reason;}
    int CheckReason(SP_BARRIER_REASON reason) {return (reason == GetReason());}

    void Set_GENERIC() {SetReason(SP_BARRIER_GENERIC);}
    int Check_GENERIC() {return CheckReason(SP_BARRIER_GENERIC);}

    void Set_OUTPUT_MSG() {SetReason(SP_BARRIER_OUTPUT_MSG);}
    int Check_OUTPUT_MSG() {return CheckReason(SP_BARRIER_OUTPUT_MSG);}

    void Set_INPUT_MSG() {SetReason(SP_BARRIER_INPUT_MSG);}
    int Check_INPUT_MSG() {return CheckReason(SP_BARRIER_INPUT_MSG);}

    void Set_TIME_GRANT() {SetReason(SP_BARRIER_TIME_GRANT);}
    int Check_TIME_GRANT() {return CheckReason(SP_BARRIER_TIME_GRANT);}

    void Print(ostream &out=cout) {
      out << "Barrier ";
      if (Gateway) cout << "<Gateway> ";
      if (Check_GENERIC()) out << "GENERIC";
      if (Check_OUTPUT_MSG()) out << "OUTPUT_MSG";
      if (Check_INPUT_MSG()) out << "INPUT_MSG";
      if (Check_TIME_GRANT()) out << "TIME_GRANT";
      out << " at " << *this << endl;
    }

};

//RB_DEFINE_CLASS(SpBarrier);

#endif
