#ifndef SpBuiltInMacrosGate
#define SpBuiltInMacrosGate

#include "O_SpBlockDecomp.H"
#include "O_SpScatterDecomp.H"

#include "RB.H"

//-----------------------------------------------------------------------------
// General macro for automatically defining a built-in object manager.
//
// 1. The name is used to generate the O_<name> object manager.
// 2. The simObjClass is used to create simulation objects
// 3. The construction Argument must either be an integer, a function returning
//    an integer, (note, integers are used to define the number of simulation
//    objects to create) or the string name of the "par" file used for
//    construction
// 4. The decompositionType must be either ScatterDecomp or BlockDecomp
// 5. clusterName is used for decomposing the objects to a subset of nodes
//-----------------------------------------------------------------------------
#define DEFINE_AUTOMATIC_OBJECT_MANAGER(name,				\
                                        simObjClass,			\
                                        constructionArgument, 		\
                                        decompositionType, 		\
                                        clusterName) 			\
                                                                        \
RB_DEFINE_CLASS(simObjClass);						\
                                                                        \
class O_##name : public O_Sp##decompositionType {		        \
  private:								\
  protected:								\
  public:								\
    O_##name() {CreateSimObjs(constructionArgument);}			\
    virtual SpSimObj *CreateSimObj() {return new simObjClass();}	\
    virtual void DestroySimObj(SpSimObj *simObj) {			\
      RB_DELETE_##simObjClass((simObjClass *)simObj);  			\
    }									\
    virtual char *GetClusterName() {return #clusterName;}		\
    virtual char *GetSimObjMgrName() {return #name;}			\
};									\
DEFINE_OBJECT_MANAGER_CLUSTER(name, O_##name, clusterName)

//-----------------------------------------------------------------------------
// Macro for automatically defining a built-in scatter decomposition object
// manager.
//
// 1. The name is used to generate the O_<name> object manager.
// 2. The simObjClass is used to create simulation objects
// 3. The construction Argument must either be an integer, a function returning
//    an integer, (note, integers are used to define the number of simulation
//    objects to create) or the string name of the "par" file used for
//    construction
// 4. clusterName is used for decomposing the objects to a subset of nodes
//-----------------------------------------------------------------------------
#define DEFINE_SCATTER_DECOMP_OBJECT_MANAGER(name,			\
                                             simObjClass,		\
                                             constructionArgument, 	\
                                             clusterName) 		\
                                                                        \
DEFINE_AUTOMATIC_OBJECT_MANAGER(name,					\
                                simObjClass,				\
				constructionArgument,			\
			        ScatterDecomp,	 			\
			        clusterName)				\

//-----------------------------------------------------------------------------
// Macro for automatically defining a built-in block decomposition object
// manager.
//
// 1. The name is used to generate the O_<name> object manager.
// 2. The simObjClass is used to create simulation objects
// 3. The construction Argument must either be an integer, a function returning
//    an integer, (note, integers are used to define the number of simulation
//    objects to create) or the string name of the "par" file used for
//    construction
// 4. clusterName is used for decomposing the objects to a subset of nodes
//-----------------------------------------------------------------------------
#define DEFINE_BLOCK_DECOMP_OBJECT_MANAGER(name,			\
                                           simObjClass,			\
                                           constructionArgument, 	\
                                           clusterName) 		\
                                                                        \
DEFINE_AUTOMATIC_OBJECT_MANAGER(name,					\
                                simObjClass,				\
				constructionArgument,	      		\
				BlockDecomp,	 			\
				clusterName)				\



#endif


