// SpComponent.C file defines the base classe that supports adding
// reusable components with associated event handlers to simobjs

#include "SpComponent.H"
#include "SpSimObj.H"
#include "RB_SpDefineClass.H"
RB_DEFINE_CLASS(EventHandlerElement);
RB_DEFINE_CLASS(SpEventHandlerMethod);

SpComponent::SpComponent (): LoadedSimobjData(0) {AddedToSimobj = 0;}

SpComponent::SpComponent (SpSimObj *simobj)
{
  if (simobj){
    LoadSimobjData(simobj);
    AddedToSimobj = 0;
  }
}

SpComponent::~SpComponent ()
{
   if (AddedToSimobj) SimObj->RemoveComponent(this);
   EventHandlerElement *eh =
     (EventHandlerElement *) EventHandlerList.get_top();
   while (eh)
   {
      delete eh;
      eh = (EventHandlerElement *) (++EventHandlerList);
   }
}

void SpComponent::AddEventHandler (char *eventName, SpEventHandler *eventHandler)
{
   eventHandler->SetComponent(this);
   EventHandlerElement *EHelement =
     RB_NEW_EventHandlerElement (new EventHandlerElement(eventName,
							 eventHandler));
   
   EventHandlerList += EHelement;

   if (AddedToSimobj) {
     SimObj->AddEventHandler(eventName, eventHandler);
   }

}

SpEventHandler *SpComponent::AddEventHandlerMethod (char *eventName,
					   EVENT_HANDLER_METHOD_FNC_PTR method,
								 void *objPtr
						 )
{
   SpEventHandler *eventHandler =
     RB_NEW_SpEventHandlerMethod (new SpEventHandlerMethod(objPtr, method));
   AddEventHandler (eventName, eventHandler);

   if (AddedToSimobj) {
     SimObj->AddEventHandler(eventName, eventHandler);
   }

   return eventHandler;
}

void SpComponent::RemoveEventHandler (char *eventName,
				    SpEventHandler *eventHandler)
{
   eventHandler->SetComponent(NULL);
   EventHandlerElement *eh =
     (EventHandlerElement *) EventHandlerList.get_top();
   while (eh)
   {
      if (eh->eventHandler == eventHandler && eh->eventName == eventName)
      {
	 EventHandlerList -= eh;
	 RB_DELETE_EventHandlerElement(eh);
      }
      eh = (EventHandlerElement *) ++EventHandlerList;
   }
}

void SpComponent::LoadSimobjData(SpSimObj *simobj)
{
   SimObj     = simobj;
   ObjectType = simobj->get_OBJECT_TYPE();
   LocalId    = simobj->get_LOCAL_ID();
   GlobalId   = simobj->get_GLOBAL_ID();
   Node       = simobj->get_NODE();

   LoadedSimobjData = 1;
}

void RB_Register_EventHandlerElement()
{
  RB_REGISTER_CLASS_EventHandlerElement();
}


