// SpEvent.C method file

#include "SpStdio.H"
#include "SpTime.H"
#include "speedes_comm.H"
#include "SpEvent.H"
#include "excon.H"
#include "query_data.H"

#include "SpOptMessager.H"
#include "M_SpCancel.H"
#include "M_SpDestroy.H"
#include "host_user.H"
#include "SpObjHandle.H"
#include "SpTag.H"

#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpTimeMgtAlg.H"
#include "SpScheduleEvent.H"
#include "SpClusterMgr.H"
#include "RB_SpDefineClass.H"
#ifdef __IBMCPP__
namespace {
#endif
RB_DEFINE_CLASS_PRIMATIVE(char);
RB_DEFINE_CLASS_PRIMATIVE(int);
#define PO_DEFINED_CHAR
#ifdef __IBMCPP__
}
#endif
#include "RB_SpBinaryTree.H"

EVENT_TYPE(CREATE);
EVENT_TYPE(DESTROY);

//EVENT_TYPE (CREATE);
//EVENT_TYPE (DESTROY);
EVENT_TYPE (CANCEL);

ofstream *TraceOut = NULL;

SpOptMessager *twosmess;
SpEvent *CurrentEvent = NULL;

double SpEvent::spin_time = 0;
double SpEvent::spin_clock = 0;
int SpEvent::EnableLazy = 0;
int SpEvent::one_node_flag = 0;
int SpEvent::traceflag = 0;
char *SpEvent::tracefile = NULL;
FILE *SpEvent::LookaheadFP = NULL;
char *SpEvent::Lookaheadfile = NULL;
int SpEvent::WriteLookahead = 0;
int SpEvent::NODE = 0;
int SpEvent::N_NODES = 0;
int SpEvent::gencid = 0;
C_DBL_QUEUE *SpEvent::dbl_qproc = NULL;
SpMasterSimObjMgr *SpEvent::objects = NULL;
C_RANDOM *SpEvent::RANDOM = NULL;
SpEventTypes *SpEvent::evtype = NULL;
//C_MULTIROUTER *SpEvent::multirouter = NULL; //SpComm Should not be needed
C_HOST_USER *SpEvent::HOST_USER = NULL;

extern SpPoDataBase * MessageDataBase;
extern SpClusterMgr *GlobalClusterMgr;

/************************************************************************
* SpEvent : construct a event object					*
************************************************************************/
SpEvent::SpEvent() : DATA(NULL) {

  evtptr.set_ptr(this);
  dbl_item.set_ptr(this);
  valid = 1;
  rbid = -1;
  cancel_id = -1;
  event_canceled = 0;
  free_flag = 0;
  lazy_flag = 0;
  lazy_proc = 0;

}

/************************************************************************
* delete_MESSAGE : delete the MESSAGE (if it exists)			*
************************************************************************/
void SpEvent::delete_MESSAGE() {

  if (MESSAGE != NULL) {

    if (get_from_SpComm()) {
      delete [] ((char*) MESSAGE);
    }else{
      evtype->undup_message(MESSAGE);
    }
    if (SpTimeMgtAlg::DoingCheckpoints()){
      PO_REMOVE_char((char*) MESSAGE, MessageDataBase);
    }
    MESSAGE = NULL;
  }

}

/************************************************************************
* create : create an object						*
************************************************************************/
SpMsg *SpEvent::create_object(	SpSimTime& timeTag, //GEB2
				int object_type,
				int object_node) {

  return schedule(timeTag, CREATE, object_type, -1, object_node);

}

/************************************************************************
* create : create an object						*
************************************************************************/
SpMsg *SpEvent::create_object(	void *ch,
				SpSimTime& timeTag, //GEB2
				int object_type,
				int object_node) {

  return schedule(ch, timeTag, CREATE, object_type, -1, object_node);

}

/************************************************************************
* create : create an object						*
************************************************************************/
SpMsg *SpEvent::create_object(	SpSimTime& timeTag, //GEB2
				int object_type,
				int object_node,
				int mess_size,
				char *&buff) {

  return schedule(timeTag, CREATE, object_type, -1, object_node,
			mess_size, buff);

}

/************************************************************************
* create : create an object						*
************************************************************************/
SpMsg *SpEvent::create_object(	void *ch,
				SpSimTime& timeTag, //GEB2
				int object_type,
				int object_node,
				int mess_size,
				char *&buff) {

  return schedule(ch, timeTag, CREATE, object_type, -1, object_node,
			mess_size, buff);

}

/************************************************************************
* destroy : destroy an object						*
************************************************************************/
SpMsg *SpEvent::destroy_object(	SpSimTime& timeTag, //GEB2
				int object_type,
				int object_id,
				int object_node) {
  M_SpDestroy *destroy_mess;

  destroy_mess = (M_SpDestroy *)schedule(timeTag, DESTROY,
					  object_type, -1, object_node);
  destroy_mess->OldLocalId = object_id;
  return destroy_mess;

}

/************************************************************************
* destroy : destroy an object						*
************************************************************************/
SpMsg *SpEvent::destroy_object(	void *ch,
				SpSimTime& timeTag, //GEB2
				int object_type,
				int object_id,
				int object_node) {
  M_SpDestroy *destroy_mess;

  destroy_mess = (M_SpDestroy *)schedule(ch, timeTag, DESTROY,
					  object_type, -1, object_node);
  destroy_mess->OldLocalId = object_id;
  return destroy_mess;

}

/************************************************************************
* schedule : schedule an event						*
************************************************************************/
SpMsg *SpEvent::schedule(const SpSimTime& time, int event_type,
			    int object_type, int object_id, int object_node) {
  C_HOLDER *holder;
  SpMsg *header;

  holder = (C_HOLDER *)evtype->new_message(
        time,
        event_type,
        object_type,
        object_id,
        object_node);

  qmessout.push_bot(holder);

  header = (SpMsg *)holder->get_buff();

  header->cancel_id = -1;
//  header->cancel_id = gencid;
//  gencid += N_NODES;

  if (WriteLookahead) {
    WriteLookaheadfile(time, event_type, object_type, object_id, object_node);
  }

  return header;

}

/************************************************************************
* schedule : schedule an event						*
************************************************************************/
SpMsg *SpEvent::schedule(const SpSimTime& time, int event_type,
			    int object_type, int object_id, int object_node,
			    int data_bytes, char * &buff) {
  C_HOLDER *holder;
  SpMsg *header;

  holder = (C_HOLDER *)evtype->new_message(
        time,
        event_type,
        object_type,
        object_id,
        object_node,
	data_bytes,
	buff);

  qmessout.push_bot(holder);

  header = (SpMsg *)holder->get_buff();

  header->cancel_id = -1;
//  header->cancel_id = gencid;
//  gencid += N_NODES;

  if (WriteLookahead) {
    WriteLookaheadfile(time, event_type, object_type, object_id, object_node);
  }

  return header;

}

/************************************************************************
* schedule : schedule an event						*
************************************************************************/
SpMsg *SpEvent::schedule(void *ch, const SpSimTime& time, int event_type,
			    int object_type, int object_id, int object_node) {
  SpMsg *header;
  SpCancelHandle *cancel_handle;

  header = (SpMsg *)schedule(
	time, event_type, object_type, object_id, object_node);

  header->cancel_id = gencid;

//...... fill antimessage information

  cancel_handle = (SpCancelHandle *)ch;

  cancel_handle->set_time_tag(time);
  cancel_handle->set_rbid(gencid);
  cancel_handle->set_obid(object_id);
  cancel_handle->set_obtype(object_type);
  cancel_handle->set_node(object_node);
  cancel_handle->set_event(this);

  gencid += N_NODES;

  if (WriteLookahead) {
    WriteLookaheadfile(time, event_type, object_type, object_id, object_node);
  }

  return header;

}

/************************************************************************
* schedule : schedule an event						*
************************************************************************/
SpMsg *SpEvent::schedule(void *ch, const SpSimTime& time, int event_type,
			    int object_type, int object_id, int object_node,
			    int data_bytes, char * &buff) {
  SpMsg *header;
  SpCancelHandle *cancel_handle;

  header = (SpMsg *)schedule(
	time, event_type, object_type, object_id, object_node,
			data_bytes, buff);

  header->cancel_id = gencid;

//...... fill antimessage information

  cancel_handle = (SpCancelHandle *)ch;

  cancel_handle->set_time_tag(time);
  cancel_handle->set_rbid(gencid);
  cancel_handle->set_obid(object_id);
  cancel_handle->set_obtype(object_type);
  cancel_handle->set_node(object_node);
  cancel_handle->set_event(this);

  gencid += N_NODES;

  if (WriteLookahead) {
    WriteLookaheadfile(time, event_type, object_type, object_id, object_node);
  }

  return header;

}

/************************************************************************
* cancel : cancel an event						*
************************************************************************/
void SpEvent::cancel(void *ch) {
  SpCancelHandle *cancel_handle;
  M_SpCancel *cancel_mess;

  cancel_handle = (SpCancelHandle *)ch;
  cancel_mess = (M_SpCancel *) schedule(
		cancel_handle->get_time_tag(),
		CANCEL,
		cancel_handle->get_obtype(),
		cancel_handle->get_obid(),
		cancel_handle->get_node()	);

  cancel_mess->CANCEL_ID = cancel_handle->get_rbid();

  if (WriteLookahead) {
    WriteLookaheadfile(	cancel_handle->get_time_tag(),
			CANCEL,
			cancel_handle->get_obtype(),
			cancel_handle->get_obid(),
			cancel_handle->get_node()	);
  }

}

/************************************************************************
* set_twosmess : pass the pointer to the twosmess object		*
************************************************************************/
void SpEvent::set_twosmess(void *tm) {

  twosmess = (SpOptMessager *)tm;

}

/************************************************************************
* clear_data : clear the data for this event				*
************************************************************************/
void SpEvent::clear_data() {
  C_HOLDER *holder;

  while (qmessout.length()) {
    holder = (C_HOLDER *)qmessout.pop_top();
    evtype->delete_message(holder);
  }

  if (twosmess) {
    twosmess->free_antimess(&qanti, rbid);
  }

  rbid = -1;
  optevts.reset();

  cleanup();
  Tag.Reset();

}

/************************************************************************
* sendmess : send the messages out					*
************************************************************************/
void SpEvent::sendmess(C_QUEUE *q1nd) {
  int i,len;
  int sz,nd;
  char *buff;
  C_HOLDER *holder;
  int local_evtype, local_obtype;

  len = qmessout.length();
  for (i=0; i<len; i++) {
    holder = (C_HOLDER *)qmessout.pop_top();
    holder->get(buff, sz, nd);
    if (nd == NODE) {
      q1nd->push_bot(holder);
    }else{

      SpMsg *msg = (SpMsg *)buff;

      local_evtype = msg->evtype;
      local_obtype = msg->obtype;

      msg->evtype = GlobalClusterMgr->
	ConvertLocalToGlobalEventId(msg->evtype);

      msg->obtype = GlobalClusterMgr->
	ConvertLocalToGlobalMgrId(msg->obtype);

      if (nd == -1){
	SpComm_CoordinatedSend(sz, buff);
      }
      else{
	SpComm_CoordinatedSend(nd, sz, buff);
      }

      // restore local event type & global obj type
      msg->evtype = local_evtype;
      msg->obtype = local_obtype;

      qmessout.push_bot(holder);
    }
  }

  pop_roll_back();

}

/************************************************************************
* pop_roll_back : pop event off the roll back queue			*
************************************************************************/
void SpEvent::pop_roll_back() {
  C_DBL_QUEUE *q;

  q = SIMOBJ->get_evtq();
  q->pop_top();

/*
  SpEventStatus *temp;
  q = SIMOBJ->get_evtq();
  temp = (SpEventStatus *)q->pop_top();
  if (temp != &objevt) {
    cout << "Error, (EVENT) objevt not right" << endl;
  }
*/

}

/************************************************************************
* check_proc : check if the event is in the proc queue			*
************************************************************************/
int SpEvent::check_proc(int flag) {
  int i,len;
  SpEventStatus *temp;
  C_DBL_QUEUE *q;

  q = SIMOBJ->get_evtq();
  len = q->get_length();
  temp = (SpEventStatus *)q->get_top();

  for (i=0; i<len; i++) {
    if (temp == &objevt) {
      if (flag) {
        if (temp != q->get_bot()) {
          cout << "Error, (EVENT) check proc " << flag << endl;
          return 1;
	}
      }else{
        cout << "Error, (EVENT) check proc " << endl << flag;
        return 1;
      }
    }
    temp = (SpEventStatus *)temp->get_next();
  }
 
  return 0;

}

/************************************************************************
* check_clear : check if the event is clear				*
************************************************************************/
int SpEvent::check_clear() {
  int n;

  n = qmessout.length() + qanti.length() + optevts.length()
	+ AltMgr.GetNumAltItems();

  if (n) {
    cout << "Error, (EVENT) event is not clear for processing" << endl;
  }
 
  return n;

}

/************************************************************************
* local_messages : send out the local messages				*
************************************************************************/
int SpEvent::local_messages() {
  int i,len;
  int sz,nd;
  char *buff;
  SpMsg *header;
  C_HOLDER *holder;
  SpEvent *event;
  C_LPTR *lptr;
  int n_roll;

  n_roll = 0;
  len = qmessout.length();
  for (i=0; i<len; i++) {

    holder = (C_HOLDER *)qmessout.pop_top();
    holder->get(buff, sz, nd);

    if (nd == NODE) {

      header = (SpMsg *)buff;

      if (header->ext == 1) {

        qmessout.push_bot(holder);

      }else{

	if (SpTimeMgtAlg::DoingCheckpoints()){
	  PO_ADD_char(buff, sz, MessageDataBase);
	}
        event = (SpEvent *)evtype->message_event(buff);
        evtype->delete_holder(holder);
	event->evtptr.set_ptr(event);
	lptr = event->get_evtptr();
        optevts.push_bot(lptr);
        event->set_valid_opt();

        n_roll += event->optimistic_insert();

      }

    }else{

      qmessout.push_bot(holder);

    }

  }

  return n_roll;

}

/************************************************************************
* optimistic_insert : optimistically insert events into the event list	*
************************************************************************/
int SpEvent::optimistic_insert() {
  int n_roll;

  n_roll = rollback();
  insert();

  return n_roll;

}

/************************************************************************
* rollback : roll back events if necessary				*
************************************************************************/
int SpEvent::rollback() {
  int n_roll;
  SpEvent *e;
  SpEventStatus *temp;
  C_DBL_QUEUE *q;

  n_roll = 0;
  if (one_node_flag) return n_roll;

  q = SIMOBJ->get_evtq();
  while (q->get_length()) {

    temp = (SpEventStatus *)q->get_bot();

    if (temp->get_valu() >= time_tag) {

      e = (SpEvent *)temp->get_event();
      if (e == this) {
        break;
      }

      if ((temp->get_valu() == time_tag) && (!get_processed())) {
        break;
      }

      q->pop_bot();

      if (e->get_SIMOBJ() != SIMOBJ) {
	cout << "error, SIMOBJ not right" << endl;
      }

      n_roll += e->rollback_event();
      e->reset_processed();

    }else{

      break;

    }

  }

  return n_roll;

}

/************************************************************************
* rollback_event : roll back this event					*
************************************************************************/
int SpEvent::rollback_event() {
  int n_roll;
  C_LPTR *lptr;
  SpEvent *event;

  n_roll = 0;

  if (get_lazy_proc() && get_valid()) {

    init_qroll_backs();
    exchange();
    SpCurrentAltMgr->Rollback();

    if (get_processed()) {
      dbl_item.set_ptr(this);
      dbl_qproc->remove(&dbl_item);
      insert();
    }else{
      cout << "Error, (EVENT) rolling back unprocessed event" << endl;
    }

  }else{

    twosmess->send_antimessages(this);

    init_qroll_backs();
    if (get_processed()) {
      n_roll = 1;
      exchange();
      SpCurrentAltMgr->Rollback();
      if (get_valid()) {
	dbl_item.set_ptr(this);
        dbl_qproc->remove(&dbl_item);
        insert();
      }
    }
  SpCurrentAltMgr->Cleanup();

    while (optevts.length()) {
      lptr = (C_LPTR *)optevts.pop_top();
      event = (SpEvent *)lptr->get_ptr();
      n_roll += event->cancel_event();
    }

    clear_messages();

  }

  return n_roll;

}

/************************************************************************
* lazy_rollback : late roll back this event from lazy cancellation	*
************************************************************************/
int SpEvent::lazy_rollback() {
  int n_roll;
  C_LPTR *lptr;
  SpEvent *event;

  n_roll = 1;

/*
  n_roll += rollback();
  if (n_roll != 1) {
    cout << "Error, (EVENT) lazy rollback" << endl;
  }
*/

  twosmess->send_antimessages(this);
  init_qroll_backs();
  SpCurrentAltMgr->Cleanup();

  while (optevts.length()) {
    lptr = (C_LPTR *)optevts.pop_top();
    event = (SpEvent *)lptr->get_ptr();
    n_roll += event->cancel_event();
  }

  reset_processed();
  clear_messages();
  return n_roll;

}

/************************************************************************
* cancel_event : cancel an event					*
************************************************************************/
int SpEvent::cancel_event() {
  int n_roll;
  SpEvent *e;
  SpEventStatus *temp;
  C_DBL_QUEUE *q;

  reset_valid();
  n_roll = 0;

  if (get_processed()) {

    n_roll = rollback();
    q = SIMOBJ->get_evtq();
    temp = (SpEventStatus *)q->get_bot();
    e = (SpEvent *)temp->get_event();
 
   if (e != this) {
     cout << "error, cancel event not right" << endl;
    }else{
      q->pop_bot();
    }

    n_roll += rollback_event();

  }else{

    if (get_lazy_proc()) {

      n_roll += rollback_event();

    }else{
/*
      if (check_clear()) {
	cout << "Error, (EVENT) cancel event not clear" << endl;
      }
*/
    }

  }
  if (SpTimeMgtAlg::DoingCheckpoints()){
    PO_REMOVE_char((char*) MESSAGE, MessageDataBase);
  }

  return n_roll;

}

/************************************************************************
* calculate_tmin : get the minimum time for a generated message		*
************************************************************************/
void SpEvent::calculate_tmin() {
  int i;
  int len;
  char *buff;
  int size;
  int node;
  C_HOLDER *holder;
  SpMsg *header;

  tmin = 1.0e20;
  len = qmessout.length();
  holder = (C_HOLDER *)qmessout.ptop();
  for (i=0; i<len; i++) {
    holder->get(buff, size, node);
    header = (SpMsg *)buff;
    if (header->time_tag < tmin) tmin = header->time_tag;
    holder = (C_HOLDER *)holder->get_link();
  }

}

/************************************************************************
* clear_messages : clear all of the messages out of this event		*
************************************************************************/
void SpEvent::clear_messages() {
  C_HOLDER *holder;

  Tag.Reset();

  while (qmessout.length()) {
    holder = (C_HOLDER *)qmessout.pop_top();
    evtype->delete_message(holder);
  }

  twosmess->free_antimess(&qanti, -1);
  optevts.reset();

}

/************************************************************************
* set_processed : tell the event that it has been processed		*
************************************************************************/
void SpEvent::set_processed() {
  C_DBL_QUEUE *q;

  if (!one_node_flag) {
    q = SIMOBJ->get_evtq();
    q->push_bot(&objevt);
    objevt.set_processed();
  }

}

/************************************************************************
* attach_object : attach this event to an object			*
************************************************************************/
void SpEvent::attach_object() {

  if (LOCAL_ID < 0) {
    if (OBJECT_TYPE < 0) {
      SIMOBJ = objects;
    }else{
      SIMOBJ = objects->get_objman(OBJECT_TYPE);
    }
  }else{
    if (OBJECT_TYPE < 0) {
      cout << "Error (event): Bad OBJECT_TYPE " << OBJECT_TYPE << endl;
    }else{
      SIMOBJ = objects->get_obj(OBJECT_TYPE, LOCAL_ID);
    }
  }

//...... some quick checks

  if (SIMOBJ == NULL) {
    cout << "Error (event): Attaching to a NULL object" << endl;
  }

  if (SIMOBJ->get_LOCAL_ID() != id) {
    cout << "Error (event): Attaching to a bad object" << endl;
  }

// This error is now ok that we have dynamic object creation which assigns
// local Ids that can jump. N_LOC is the number of local objects, but their Ids
// may not be sequential.
//
//  if (OBJECT_TYPE >= 0) {
//    SpSimObjMgr *oman (SpSimObjMgr *)objects->get_objman(OBJECT_TYPE);
//    if (LOCAL_ID >= oman->get_N_LOC()) {
//      cout << "Error (event): LOCAL_ID "
//	   << LOCAL_ID
//	   << " out of range "
//	   << oman->get_N_LOC() << endl;
//    }
//  }

  if (!one_node_flag) {
    dbl_item.set_ptr(this);
    dbl_item.set_valu(time_tag);
    objevt.set_valu(time_tag);
    objevt.reset_processed();
    objevt.set_event(this);
  }

  GLOBAL_ID = SIMOBJ->get_GLOBAL_ID();

}

/************************************************************************
* spin_loop : spin in a loop to waste time				*
************************************************************************/
void SpEvent::spin_loop(double twait) {
  long timer1;
  long timer2;

  if (twait <= 0) return;
  if (EVENT_TYPE == CANCEL) return;

  spin_time += twait;
  if (spin_clock >= spin_time) return;

  timer1 = clock();
  for (;;) {

    timer2 = clock() - timer1;

    if (timer2 > 0) {
      spin_clock += double(timer2) / 1000000.0;
      if (spin_clock < spin_time) {
	timer1 += timer2;
      }else{
	break;
      }
    }

  }

}

/************************************************************************
* check_canceled : check if the event has been canceled			*
************************************************************************/
int SpEvent::check_canceled(int &n_roll) {
  RB_SpBinaryTree *CancelTree;
  int *element;

  if (!valid) {
    event_canceled = 0;
    return 0;
  }

//...... first check if there are any cancelations pending

  if (SIMOBJ->GetNumCanceled() == 0) {
    event_canceled = 0;
    return 0;
  }

//...... check if this event should be canceled

  CancelTree = SIMOBJ->GetCancelTree();
  element = (int *)CancelTree->Find(cancel_id);
  if (element == NULL) {
     event_canceled = 0;
     return 0;
  }

//...... undo lazy stuff if necessary

  if (get_lazy_proc()) {
    cout << "shouldn't do this" << endl; abort();
    n_roll += lazy_rollback();
  }
  reset_lazy_proc();

//...... process the event by removing the cancelation request

  init_qroll_backs();
  if (AltMgr.GetNumAltItems() != 0) {
     cout << "Error, rollback queue not empty while "
       "removing cancelation request " << AltMgr.GetNumAltItems() << endl;
     SpCurrentAltMgr->Cleanup();
  }

  if (element != (int *)CancelTree->Remove(cancel_id)) {
    cout << "Error, bad remove cancel_id" << endl;
  }

  RB_DELETE_int(element);

  if (CancelTree->Find(cancel_id)) {
     cout << "Error, found multiple cancel_ids" << endl;
     abort();
  }

  event_canceled = 1;
  return 1;

}

/************************************************************************
* open_tracefile : open the tracefile					*
************************************************************************/
void SpEvent::open_tracefile(char *filename) {
  int i;
  char s[120];

  for (i=0; i<120; i++) s[i] = 0;
  sprintf(s,"%s.%d",filename,NODE);

  TraceOut = new ofstream();
  TraceOut->open(s, ios::out);

  cout << "Creating trace file " << s << endl;

}


/************************************************************************
* trace : trace this event						*
************************************************************************/
void SpEvent::trace() {
  int i;
  int len;
  char *message;
  C_HOLDER *holder;
  SpMsg *header;
  C_LPTR *lptr;
  SpEvent *event;
  SpAntiMsg *anti;

  if (!TraceOut) return;

  *TraceOut << TIME_TAG << endl;
  *TraceOut << "EvtName: " << evtype->get_messname(EVENT_TYPE) << endl;
  *TraceOut << "ObjName: " << SIMOBJ->get_NAME() << endl;
  *TraceOut << "ObjType: " << objects->get_obname(OBJECT_TYPE) << endl;
  *TraceOut << "Handle:  "
	    << "(" << NODE
	    << "," << OBJECT_TYPE
	    << "," << LOCAL_ID
	    << ")" << endl;
  *TraceOut << "GlobId:  " << GLOBAL_ID << endl;
  *TraceOut << "RBitems: " << AltMgr.GetNumAltItems() << endl;
  *TraceOut << "EvtCPU:  " << CPU_Time << endl;
  *TraceOut << "TotCmtd: " <<  ProcessingTime << endl;
  *TraceOut << "TotTime: " << TotalTime  << endl;
  *TraceOut << "CritPth: " << CriticalPathTime  << endl;

  if (Tag.GetBuff()) {
    *TraceOut << "BeginTag:" << endl;
    *TraceOut << Tag;
    *TraceOut << "EndTag:" << endl;
  }else{
    *TraceOut << "NoTag:" << endl;
  }

/*
** Events scheduled locally
*/

  len = optevts.length();
  lptr = (C_LPTR *)optevts.get_top();
  for (i=0; i<len; i++) {
    event = (SpEvent *)lptr->get_ptr();
    *TraceOut << "  LocSched: " << event->get_time_tag() << endl;
    *TraceOut << "    EvtName:  "
	      << evtype->get_messname(event->get_EVENT_TYPE()) << endl;

    *TraceOut << "    ObjType:  " << event->get_OBJECT_TYPE() << endl;
    *TraceOut << "    Handle:  "
	      << "(" << NODE
	      << "," << event->get_OBJECT_TYPE()
	      << "," << event->get_LOCAL_ID()
	      << ")" << endl;

    *TraceOut << "    bytes:    " << (event->get_MESSAGE())->bytes
                     + evtype->get_sizeof(event->get_EVENT_TYPE()) << endl;
    lptr = (C_LPTR *)lptr->get_link();
  }

/*
** Messages sent with risk
*/

  len = qanti.length();
  anti = (SpAntiMsg *)qanti.get_top();
  for (i=0; i<len; i++) {
    *TraceOut << "  RskSched: " << anti->get_time_tag() << endl;
    *TraceOut << "    EvtName:  "
	      << evtype->get_messname(GlobalClusterMgr->ConvertGlobalToLocalEventId(anti->get_etype())) << endl;

    *TraceOut << "    Handle:  "
	      << "(" << anti->get_node()
	      << "," << GlobalClusterMgr->ConvertGlobalToLocalMgrId(anti->get_obtype())
	      << "," << anti->get_obid()
	      << ")" << endl;

    *TraceOut << "    bytes:    "
	      << evtype->get_sizeof(GlobalClusterMgr->ConvertGlobalToLocalEventId(anti->get_etype()))
                 + anti->get_bytes() << endl;

    anti = (SpAntiMsg *)anti->get_link();
  }

/*
** Messages sent without risk
*/

  len = qmessout.length();
  holder = (C_HOLDER *)qmessout.get_top();

  for (i=0; i<len; i++) {
    holder = (C_HOLDER *)qmessout.get_top();
    message = holder->get_buff();
    header = (SpMsg *)message;

    if (N_NODES > 1) {
      *TraceOut << "  SafeSched: " << header->time_tag << endl;
    }else{
      *TraceOut << "  LocSched: " << header->time_tag << endl;
    }

    *TraceOut << "    EvtName:  "
	      << evtype->get_messname(header->evtype) << endl;

    *TraceOut << "    Handle:  "
	      << "(" << holder->get_node()
	      << "," << header->obtype
	      << "," << header->objid
	      << ")" << endl;

    *TraceOut << "    bytes:    "
	      << evtype->get_sizeof(header->evtype) + header->bytes << endl;

    holder = (C_HOLDER *)holder->get_link();
  }


/*
** Put a blank line after the event info
*/

  *TraceOut << endl;

}



/************************************************************************
* open_Lookaheadfile : open the lookahead file				*
************************************************************************/
void SpEvent::open_Lookaheadfile(char *filename) {
  int i;
  char s[120];

  for (i=0; i<120; i++) s[i] = 0;
  sprintf(s,"%s.%4.4d",filename,NODE);
  Lookaheadfile = strdup(s);

  LookaheadFP = fopen(Lookaheadfile, "w");
  fprintf(LookaheadFP, "Time1   \tTime2   \tLookahead\tEvType1\tEvType2\tObType1\tObType2\tObID1\tObID2\tNode1\tNode2\tOffNode\n");
  fprintf(stderr,"Opening lookahead file %s\n",Lookaheadfile);

}


/************************************************************************
* WriteLookaheadfile : write out to the lookahead file			*
************************************************************************/
void SpEvent::WriteLookaheadfile(const SpSimTime& t2, int ty2, int ot2, int id2,
				 int nd) { // GEB2
  int RemoteNode;

//...... return if lookahead file writing is not enabled

  if (!WriteLookahead) return;

//...... write out information to the lookahead file

  if (NODE == nd) {
    RemoteNode = 0;
  }else{
    RemoteNode = 1;
  }

  fprintf(LookaheadFP, "%f\t%f\t%f\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\n",
	  (double)time_tag.GetTime(),
	  (double)t2.GetTime(),
	  (double)t2.GetTime()-time_tag.GetTime(), //GEB2
	  (int)EVENT_TYPE,
	  (int)ty2,
	  (int)OBJECT_TYPE,
	  (int)ot2,
	  (int)ID,
	  (int)id2,
	  (int)NODE,
	  (int)nd,
	  (int)RemoteNode);
}

/************************************************************************
* TieBreakGeneratedMessages: breaks simtime ties to guarantee repeatability
************************************************************************/
void SpEvent::TieBreakGeneratedMessages() {
   C_HOLDER *holder;
   SpMsg *header;
   
   int len = qmessout.get_length();
   if (len) {
      
      int Ecounter = TIME_TAG.GetCounter() + 1;
      int counter = SIMOBJ->GetEvCounter();
      
      if (counter < Ecounter) counter = Ecounter;
      SIMOBJ->SetEvCounter(counter + len);

      holder = (C_HOLDER *)qmessout.get_top();
      for (int i=0; i<len; i++) {

	header = (SpMsg *)holder->get_buff();
	if (header->evtype == CANCEL) {

	  // this guarantees that a CANCEL event gets placed on the
	  // event queue prior to the event it is cancelling, by
	  // handling the two possible cases:
	  //
	  //   1) the case where the physical time and user priorities
	  //   of the current event time (TIME_TAG) and the CANCEL
	  //   event time (header->time_tag, i.e., the CANCEL event's
	  //   scheduled time) are equal. In this case, we must assure
	  //   that the CANCEL event occurs before the (same physical
	  //   time) event it intends to cancel, yet is still in the
	  //   future. Thus, we place this CANCEL as little into the
	  //   future as possible by incrementing the UniqueId field
	  //   of the current time (TIME_TAG).
	  //
	  //   2) Otherwise, we zero the tiebreak fields since getting
	  //   the CANCEL to be in the future is not an issue, rather,
	  //   assuring the tiebreak fields to be earlier than the
	  //   event to be cancelled is the issue (as we already know
	  //   that the physical time and user priorities for the
	  //   CANCEL and the event to cancel are equal).

	  if (TIME_TAG.GetTime()      == header->time_tag.GetTime()      &&
	      TIME_TAG.GetPriority1() == header->time_tag.GetPriority1() &&
	      TIME_TAG.GetPriority2() == header->time_tag.GetPriority2() ) {
	    header->time_tag.SetCounter(TIME_TAG.GetCounter());
	    header->time_tag.SetUniqueId(TIME_TAG.GetUniqueId() + 1);
	  }else{
	    header->time_tag.SetCounter(0);
	    header->time_tag.SetUniqueId(0);
	  }

	}else{

	   header->time_tag.SetCounter(counter);
	   header->time_tag.SetUniqueId(GLOBAL_ID);

	}

	holder = (C_HOLDER *)holder->get_link();
	counter++;

      }

   }
   
}



int SpEvent::event_type(char *c) {return evtype->get_messid(c);}
int SpEvent::object_type(char *c) {return objects->object_type(c);}
void *SpEvent::get_manager(int om)
{
   if (om == -2) return objects->get_objman(OBJECT_TYPE);
   return objects->get_objman(om);
}
void SpEvent::memory_print() {evtype->print();}
void SpEvent::print_event_memory() {evtype->print();}


SpObjHandle *SpEvent::GetObjectHandle()
   { return SIMOBJ->GetObjectHandle(); }


void SpEvent::PrintTimeTag()
{
  cout << "Event Time Tag = " << TIME_TAG << endl;
}

void SpEvent::CriticalPathMessages(double earliestStartTime) {
  C_HOLDER *holder;
  SpMsg *header;

  int len = qmessout.get_length();
  if (len) {

    holder = (C_HOLDER *)qmessout.get_top();
    for (int i=0; i<len; i++) {
      header = (SpMsg *)holder->get_buff();
      header->EarliestStartTime = earliestStartTime;
      holder = (C_HOLDER *)holder->get_link();
    }

  }

}

void SpEvent::PrintScheduledEvents()
{
  cout<<"Event of type "<<EVENT_TYPE<<" on object type "<<OBJECT_TYPE
      <<" at time "<<TIME_TAG<<" has events "<<endl;

  char * buff;
  int sz, nd;
  int len = qmessout.get_length();
  C_ITEM* holderPtr = qmessout.get_top();
  for (int i=0; i < len; ++i){
    C_HOLDER* holder = (C_HOLDER*) holderPtr;
    holder->get(buff, sz, nd);
    SpMsg  *msg = (SpMsg *) buff;
    cout<<"---------------------------------"<<endl
	<<"TIME =       "<<msg->time_tag<<endl
	<<"EVENT_TYPE = "<<msg->evtype<<endl;
    //C_ITEM *jj = (C_ITEM*) holder;
    holderPtr = qmessout.next(holderPtr);
  }

};
