// SpEventHandler.H header file defines the base classes that support event
// handling

#ifndef EVT_HND_OBJ
#define EVT_HND_OBJ

#include "SpSimTime.H"
#include "SpRb.H"
#include "SpParmSet.H"
#include "SpEventHandlerTypes.H"
#include "SpString80.H"
#include "RB_Types.H"

class SpSimObj;
class SpComponent;

// Abstract base class (children are SpEventHandler, SpInteraction, and
//                      user-macro-defined typed event handlers)
class SpBaseEventHandler {

public:

  SpBaseEventHandler(enum SpEventHandlerType ehType); // ehType specifies which subclass
  virtual ~SpBaseEventHandler() {}
  
  void SetSimObj(SpSimObj *simObj) {SimObj = simObj;}
  void SetComponent(SpComponent *comp) {Component = comp;}
  void SetObjectType(int objectType) {ObjectType = objectType;}
  void SetLocalId(int localId) {LocalId = localId;}
  void SetGlobalId(int globalId) {GlobalId = globalId;}
  
  void SetNode(int node) {Node = node;}
  
  virtual void ProcessEventHandler(char *eventName, SpSimTime &timeTag,
				   char *data, int dataBytes, int externalId)
    = 0;
  enum SpEventHandlerType GetType() {return EHtype;}

protected:
  
  SpSimObj *SimObj;
  SpComponent *Component;
  enum SpEventHandlerType EHtype;
  
  int ObjectType;
  int LocalId;
  int GlobalId;
  
  static int Node;
  
  
private:
  
  
};

// This just exists to group together the string-based event handlers
// (and distinguish them from event handlers that take type-checked
// arguments, SpTypedEventHandlers)
class SpStringBasedEventHandler: public SpBaseEventHandler
{
public:
  SpStringBasedEventHandler(enum SpEventHandlerType ehType):
    SpBaseEventHandler(ehType) {}
};

// See previous comment
class SpTypedEventHandler: public virtual SpBaseEventHandler
{
public:
  SpTypedEventHandler(): SpBaseEventHandler(TYPEDEVENTHANDLER) {}
};

// legacy event handler, for backwards compatibility
class SpEventHandler: public SpStringBasedEventHandler
{
public:
  SpEventHandler(): SpStringBasedEventHandler(LEGACYEVENTHANDLER) {}
  virtual void ProcessEventHandler(char *eventName, SpSimTime &timeTag,
				   char *data, int dataBytes, 
				   int externalId);
  virtual void Process(char *eventName, SpSimTime &timeTag,
		       char *data, int dataBytes, int externalId);
};

//RB_DEFINE_CLASS(SpEventHandler);

typedef SpEventHandler EVENT_HANDLER;

// temp for SpParamSet:
class SpParamSet_TEMP {public:int a;};

class SpInteractionHandler: public SpStringBasedEventHandler
{
public:
  SpInteractionHandler(): SpStringBasedEventHandler(INTERACTIONHANDLER) {}
  virtual void ProcessEventHandler(char *eventName, SpSimTime &timeTag,
				   char *data, int dataBytes, 
				   int externalId);
  virtual void Process(SpParmSet &paramSet) = 0;
};




#endif
