// SpEventHandlerMethod.H header file defines the macros and the class
// that support event handling by calling an object method as passed
// into its constructor. This file must be included in the *.H file of
// descendents of SpSimObj's who call AddEventHandlerMethod (char
// *eventName, void (*method) (void *, char *, SpSimTime &, char *, int
// ,int), void *objPtr).

#ifndef EVT_HND_MET_OBJ
#define EVT_HND_MET_OBJ

#include "SpEventHandler.H"

/******************************************************************************
* Macro: EVENT_HANDLER_METHOD_H
* Summary: Macro used to declare or define inline an event handler
*          method in a SpSimObj's descendent header file.
* Description: Pass in the class name and the method name of the event
*              handler method. If this is only a declaration, follow
*              this macro with a semicolon. Otherwise, if definition
*              is inline, instead of the semicolon, simply follow this
*              macro, with '{', the method body, and then '}'. It will
*              then resemble the syntax of a normal method definition.
******************************************************************************/
#define EVENT_HANDLER_METHOD_H(className,             \
			       methodName)            \
   static void methodName (void *objPtr,              \
			   char *eventName,           \
			   SpSimTime &timeTag,          \
			   char *data,                \
			   int dataBytes,             \
			   int externalId)            \
   {                                                  \
       ((className *) objPtr)->                       \
          className::methodName (eventName,           \
				 timeTag,             \
				 data,                \
				 dataBytes,           \
				 externalId);         \
   }                                                  \
                                                      \
   void methodName (char *EVENT_NAME,                 \
		    SpSimTime &TIME_TAG,                \
		    char *DATA,                       \
		    int DATA_BYTES,                   \
		    int EXTERNAL_ID)

/******************************************************************************
* Macro: EVENT_HANDLER_METHOD_C
* Summary: Macro used to define an event handler method in a
*          SpSimObj descendent's header file.
* Description: Pass in the class name and the method name of the event
*              handler method.  Follow this macro, with no semicolon,
*              with '{', the method body, and then '}'. It will
*              resemble the syntax of a normal method definition. This
*              macro is not needed if EVENT_HANDLER_METHOD_H defined
*              the method inline (see EVENT_HANDLER_METHOD_H).
******************************************************************************/
#define EVENT_HANDLER_METHOD_C(className, methodName) \
   void className::methodName (char *EVENT_NAME,      \
				 SpSimTime &TIME_TAG,   \
				 char *DATA,          \
				 int DATA_BYTES,      \
				 int EXTERNAL_ID)

/******************************************************************************
* Internal Class: SpEventHandlerMethod
* Summary: Event handler that calls an object method when invoked by a
*          named event
* Description: 
******************************************************************************/

typedef void (*EVENT_HANDLER_METHOD_FNC_PTR) (void *, char *, SpSimTime &, char *, int , int);

class SpEventHandlerMethod: public SpEventHandler {

 public:
  SpEventHandlerMethod(){} //Added just for persistence
   SpEventHandlerMethod (void *objPtr, EVENT_HANDLER_METHOD_FNC_PTR met);

   void Process(char *eventName, SpSimTime &timeTag,
			char *data, int dataBytes, int externalId);

 private:

   void *object;
   EVENT_HANDLER_METHOD_FNC_PTR method;
};

//RB_DEFINE_CLASS(SpEventHandlerMethod); // Change back to WITH_NO_DEFAULT_CONSTRUCTOR
				// when persistence supports it.

typedef SpEventHandlerMethod EVENT_HANDLER_METHOD;

#endif
   
