// SpReportStatus.H header file

#ifndef SpReportStatusGate
#define SpReportStatusGate

#include "SpString.H"
#include "SpMemory.H"
#include "SpIostream.H"
#include "SpNetTypes.H"

enum SP_REPORT_STATUS {
  SP_REPORT_STATUS_PASSED,
  SP_REPORT_STATUS_INCONCLUSIVE,
  SP_REPORT_STATUS_FAILED,
  SP_NUM_REPORT_STATUS
};

class SpReportStatusRecord {

  private:

    char *TestName;
    int StatusCounter[SP_NUM_REPORT_STATUS];
    static char **ReportStatusString;

  protected:

  public:

    SpReportStatusRecord();

    SpReportStatusRecord(char *testName);

    char *GetTestName() {return TestName;}

    void UpdateStatusCounter(SP_REPORT_STATUS status) {
      StatusCounter[status]++;
    }

    int GetStatusCounter(SP_REPORT_STATUS status) {
      return StatusCounter[status];
    }

    SP_REPORT_STATUS GetStatus() {
      SP_REPORT_STATUS retVal;
      if (StatusCounter[SP_REPORT_STATUS_FAILED] > 0) {
	retVal = SP_REPORT_STATUS_FAILED;
      }else{
	if (StatusCounter[SP_REPORT_STATUS_PASSED] > 0) {
	  retVal = SP_REPORT_STATUS_PASSED;
	}else{
	  retVal = SP_REPORT_STATUS_INCONCLUSIVE;
	}
      }
      return retVal;
    }

    char *GetStatusString() {return ReportStatusString[GetStatus()];}

    int GetBuffSize() {
      return strlen(TestName)+1 + SP_NUM_REPORT_STATUS*sizeof(NET_INT);
    }

    char *PackBuff(char *buff) {
      int i;
      NET_INT n;
      strcpy(buff, TestName);
      buff += strlen(buff)+1;

      for (i=0; i<SP_NUM_REPORT_STATUS; i++) {
	n = StatusCounter[i];
	memcpy(buff, (char *)&n, sizeof(NET_INT));
	buff += sizeof(NET_INT);
      }

      return buff;
    }

    char *UnpackBuff(char *buff) {
      int i;
      NET_INT n;

      if (TestName == NULL) strcpy(TestName, buff);
      buff += strlen(buff)+1;

      for (i=0; i<SP_NUM_REPORT_STATUS; i++) {
	memcpy((char *)&n, buff, sizeof(NET_INT));
	StatusCounter[i] += n;
	buff += sizeof(NET_INT);
      }

      return buff;
    }

};

//----------------------------------------------------------------------------
// SpReportStatus() is called by applications or regression tests during
// termination to report status (passed, inconclusive, or failed). SPEEDES
// automatically collects the information across nodes by calling the
// SpGlobalReportStatus() function and then prints out the results on node 0
// using the SpPrintStatus() method.
//----------------------------------------------------------------------------
void SpReportStatus(char *testName, SP_REPORT_STATUS status);
void SpGlobalReportStatus();
void SpPrintReportStatus(ostream &out);

#endif
