#include "SpScheduleEvent.H"
#include "E_SpCreate.H"
#include "E_SpDestroy.H"
#include "M_SpCreate.H"
#include "M_SpDestroy.H"


SpCancelHandle CreateSimObj(const SpSimTime &time,
			    int objManId,
			    int nodeId,
			    SpObjHandle &objHandle,
			    char *data,
			    int dataBytes) {
  
  SpSimObjMgr *simObjMgr = (SpSimObjMgr *)CurrentEvent->get_manager(objManId);
  
  objHandle.ObjManId = objManId;
  objHandle.NodeId = nodeId;
  objHandle.SimObjId = simObjMgr->GenerateNewLocalId();
  
  CREATE_MESS createSimObjMess;
  createSimObjMess.NewLocalId = objHandle.SimObjId;
  
  return ScheduleEvent(time,
		       CREATE,
		       objManId,
		       -1,
		       nodeId,
		       createSimObjMess,
		       data,
		       dataBytes);
  
}

SpCancelHandle DestroySimObj(const SpSimTime &time,
			     const SpObjHandle objHandle) {

  DESTROY_MESS destroySimObjMess;
  destroySimObjMess.OldLocalId = objHandle.SimObjId;

  return ScheduleEvent(time,
		       DESTROY,
		       objHandle.ObjManId,
		       -1,
		       objHandle.NodeId,
		       destroySimObjMess);

}

SpCancelHandle ScheduleEvent(
			     const SpSimTime &time, // IN: Logical time
			     int eventType,	     // IN: Event type Id
			     int simobjType,      // IN: Simulation object
			     //     type
			     int simobjLocalId,   // IN: Simulation object
			     //     local Id
			     int simobjNode,      // IN: Node that the
			     //     simulation object
			     //     is on
			     const SpMsg *header, // IN: Optional
			     //     message for
			     //     scheduling the
			     //     event
			     const char *data, // IN: Variable length 
			     //     data that gets
			     //     passed with the
			     //     event (optional)
			     int dataBytes    // IN: Number of bytes for
			     //     variable length
			     //     data
			     ){
  SpCancelHandle cancelHandle;
  SpMsg *h;
  char *buff;

  if (dataBytes > 0) {
    h = (SpMsg *)CurrentEvent->schedule(&cancelHandle,
					time,
					eventType,
					simobjType,
					simobjLocalId,
					simobjNode,
					dataBytes,
					buff);
    memcpy(buff, data, dataBytes);
  }else{
    h = (SpMsg *)CurrentEvent->schedule(&cancelHandle,
					   time,
					   eventType,
					   simobjType,
					   simobjLocalId,
					   simobjNode);
  }

  h->ExternalId = -1;

  if (header) {
    if (((unsigned int)EventTypes->get_sizeof(eventType)) > sizeof(SpMsg)) {
      memcpy((char *)&h[1],
	     (char *)&header[1],
	     EventTypes->get_sizeof(eventType) - sizeof(SpMsg));
    }
  }

  return cancelHandle;

}


SpCancelHandle ScheduleEvent(
			     const SpSimTime &time, // IN: Logical time
			     int eventType,	     // IN: Event type Id
			     const SpObjHandle &simobj, // IN: Simulation
			     //     object handle
			     const SpMsg *header, // IN: Optional
			     //     message for
			     //     scheduling the
			     //     event
			     const char *data, // IN: Variable length 
			     //     data that gets
			     //     passed with the
			     //     event (optional)
			     int dataBytes    // IN: Number of bytes for
			     //     variable length
			     //     data
			     ){
  SpCancelHandle cancelHandle;
  SpMsg *h;
  char *buff;
  
  if (dataBytes > 0) {
    h = (SpMsg *)CurrentEvent->schedule(&cancelHandle,
					   time,
					   eventType,
					   simobj.ObjManId,
					   simobj.SimObjId,
					   simobj.NodeId,
					   dataBytes,
					   buff);
    memcpy(buff, data, dataBytes);
  }else{
    h = (SpMsg *)CurrentEvent->schedule(&cancelHandle,
					   time,
					   eventType,
					   simobj.ObjManId,
					   simobj.SimObjId,
					   simobj.NodeId);
  }

  h->ExternalId = -1;

  if (header) {
    if (((unsigned int) EventTypes->get_sizeof(eventType)) > sizeof(SpMsg)) {
      memcpy((char *)&h[1],
	     (char *)&header[1],
	     EventTypes->get_sizeof(eventType) - sizeof(SpMsg));
    }
  }

  return cancelHandle;
}

SpCancelHandle ScheduleProcess(
			       const SpSimTime &time, // IN: Logical time
			       int eventType,	     // IN: Event type Id
			       int simobjType,      // IN: Simulation object
			       //     type
			       int simobjLocalId,   // IN: Simulation object
			       //     local Id
			       int simobjNode,      // IN: Node that the
			       //     simulation object
			       //     is on
			       const M_SpProc *message, // IN: Optional
			       //     message for
			       //     scheduling
			       //     the event
			       const char *data,   // IN: Variable length 
			       //     data that gets
			       //     passed with the
			       //     event (optional)
			       int dataBytes    // IN: Number of bytes for
			       //     variable length
			       //     data
				){
   const M_SpProc *actualMessage;
   C_HOLDER *holder;
   
   if (message) actualMessage = message;
   else
   {
     holder = EventTypes->new_message(eventType);
     actualMessage = (M_SpProc*) holder->get_buff();
   }
   ((M_SpProc *)actualMessage)->type = BEGIN_PROCESS;

   SpCancelHandle h =
     ScheduleEvent(time, eventType, simobjType, simobjLocalId,
			simobjNode, actualMessage, data, dataBytes);

   if (!message) EventTypes->DefinitelyPutOn_qfree_messages(holder);

   return h;
}


SpCancelHandle ScheduleEventHandler(
                                 const SpSimTime& time, // IN: Simulated time 
			         const char *eventName,     // IN: String name of
                                                      //     the event handler
				 SpObjHandle *dest,    // IN: object to schedule for  
	                         const char *data,   // IN: Variable length
                                                      //     data that is
                                                      //     passed to the
                                                      //     event handler(s)
	                         int dataBytes, // IN: Number of bytes
			                              //     in the data
				 enum SpEventHandlerType // category of eh
				 ehType
			       ) 
{
  char *buff;
  M_SpEventHandler *eventHandlerEventMess;
  int offset;
  SpCancelHandle cancelHandle;

  //  offset = strlen(eventName) + 1;
  int eventnamesize = strlen(eventName) + 1;
  int alignment = sizeof(double);

  // Align offset to nearest sizeof(double) bytes
  offset = alignment * 
    (eventnamesize/alignment + (eventnamesize%alignment?1:0));
  
//   offset += (sizeof(double) - offset%sizeof(double)) % sizeof(double);
  if (dest) {
    eventHandlerEventMess = (M_SpEventHandler *)
      CurrentEvent->schedule(&cancelHandle,
			     time,
			     EVENT_HANDLER_EVENT,
			     dest->ObjManId,
			     dest->SimObjId,
			     dest->NodeId,
			     dataBytes + offset,
			     buff);

    eventHandlerEventMess->DataBytes = dataBytes;
    eventHandlerEventMess->ExternalId = -1;
    eventHandlerEventMess->EHtype = ehType;
    eventHandlerEventMess->Undirected = 0;
    strcpy(buff, eventName);

    if (dataBytes > 0) {
      buff += offset;
      memcpy(buff, data, dataBytes);
    }
  }
  else {
    SpObjHandle oh;
    oh.ObjManId = -1;
    oh.SimObjId = -1;
    oh.NodeId = GetNode();
    char *datatopass;
    memcpy(&datatopass, &data,sizeof(data)); // suppress compiler warnings
#define  LOCALLYSCHEDULEDEVENTHANDLER 1
    cancelHandle =
      ScheduleScheduleUndirectedEventHandlers(time,oh,eventName,ehType,
					      LOCALLYSCHEDULEDEVENTHANDLER,    
					      datatopass,dataBytes);
#undef LOCALLYSCHEDULEDEVENTHANDLER
  }
  return cancelHandle;
}
