// object.C method file

// I moved the following #include line to before all other #includes
// so that it would compile on the HP Convex/Exemplar computer named
// dax.cmf.nrl.navy.mil.  The old order caused the HP compiler to say
// "signal 11" with no other explanation of what could be wrong, which
// is why I suspect it is a bug in the HP compiler.  There were about
// a total of 38 files in which I had to change the order of includes
// to get it to compile.
#include "SpSimObj.H"
#include "SpStdio.H"
#include "SpSimTime.H"
#include "SpEventTypes.H"
#include "SpMsg.H"
#include "holder.H"
#include "SpEvent.H"
#include "speedes_comm.H"
#include "SpAssert.H"
#include "SpComponent.H"
#include "SpObjHandle.H"
#include "RB_SpContainerFuncs.H"
#include "SpScheduleEvent.H"
#include "PO.H"
#include "RB_SpBinaryTree.H"

PO_DEFINE_CLASS(RB_SpBinaryTree);
#include "RB_SpDefineClass.H"
#include "SubType.H"
#include "EventHandler.H"
#include "EventHandlerHolder.H"

RB_DEFINE_CLASS(SpEventHandlerHolder);

DECLARE_3_ARG(AddSimObjInteraction,SpString80,SpObjHandle,int);
DECLARE_2_ARG(RemoveSimObjInteraction,SpString80,int);

int SpSimObj::done = 0;
int SpSimObj::nlobjs = 0;
int SpSimObj::N_OBJECT_TYPES = 0;
int SpSimObj::N_NODES = 0;
int SpSimObj::NODE = 0;
int SpSimObj::uid_seed = 0;
RB_int SpSimObj::TOTAL_OBJECTS = 0;
void *SpSimObj::evtype = NULL;
SpSimObjMgrTypes *SpSimObj::obman = NULL;
C_QUEUE *SpSimObj::qunblock = NULL;
SpPriorityQueueHeap *SpSimObj::rqueue = NULL;
C_RANDOM *SpSimObj::RANDOM = NULL;
C_QUEUE *SpSimObj::init_events = NULL;
C_QUEUE *SpSimObj::qblock = NULL;

#ifndef INFINITY
#define INFINITY 1.0e20
#endif

SpSimObj *CurrentSimobj = NULL;
extern SpEvent *CurrentEvent;

/************************************************************************
* SpSimObj : construct a generic simulation object			* 
************************************************************************/
SpSimObj::SpSimObj():
   rq_holder(NULL),
   EventHandlerHash(10),
   EvCounter(1),
   CriticalPathTime(0.0),
   ProcessingTime(0.0),
   TotalTime(0.0),
   CommittedTime(0.0),
   TotalEvents(0),
   CommittedEvents(0),
   TIME(-INFINITY),
   ProcCounter(0),
   NAME(NULL),
   OBJECT_TYPE(N_OBJECT_TYPES -1),
   LOCAL_ID(nlobjs),
   GLOBAL_ID(uid_seed),
   CancelTree(new RB_SpBinaryTree)
{

  if (SpCurrentAltMgr->GetOptimistic()) {
    SpCurrentAltMgr->UnsetOptimistic();
    SEED = GLOBAL_ID + 12345;
    SpCurrentAltMgr->SetOptimistic();
  }else{
    SEED = GLOBAL_ID + 12345;
  }

  if (SpPoDataBase::DoNotAllocateMemoryInConstructor){
    //cout<<"Creating a SpSimObj with persistence"<<endl;
  }
  else {
    uid_seed += N_NODES;
    nlobjs++;
    
    SetObjectHandle(NODE, OBJECT_TYPE, LOCAL_ID);
  }
  
  if (done) return;
  
  if (SpPoDataBase::DoNotAllocateMemoryInConstructor != 1){
    RANDOM = new C_RANDOM();
    obman = new SpSimObjMgrTypes();
    qblock = new C_QUEUE();
    qunblock = new C_QUEUE();
    
    done = 1;
  }
}

SpSimObj::~SpSimObj()
{
  EventHandlerHash.ClearAndDestroy();
}

/************************************************************************
* get_manager - get an object manager					*
************************************************************************/
void *SpSimObj::get_manager(int om) {
  return CurrentEvent->get_manager(om);
}

/************************************************************************
* insert_event : insert an event					*
************************************************************************/
void SpSimObj::insert_event(C_SQ_ITEM *event) {
  C_RQ_HOLDER *rqh;

//...... first put the event in a holder for the local rqueue

  rqh = local_rqueue.new_holder();
  rqh->init((void *)event, event->get_time_tag());
  local_rqueue.insert(rqh);

//...... check if this is the only (first) event for the object

  if (rq_holder == NULL) {
    assert(rqueue != NULL);
    rq_holder = rqueue->new_holder();
    rq_holder->init((void *)this, event->get_time_tag());
    rqueue->insert(rq_holder);
    return;
  }

//...... check if this is the earliest event for the object

  if (event->get_time_tag() < rq_holder->get_time_tag()) {
    assert(rqueue != NULL);
    rq_holder->set_pointer(NULL);
    rq_holder = rqueue->new_holder();
    rq_holder->init((void *)this, event->get_time_tag());
    rqueue->insert(rq_holder);
  }

}

/************************************************************************
* get_tblock : get the earliest blocking time				*
************************************************************************/
SpSimTime &SpSimObj::get_tblock() { //GEB5
  static SpSimTime infTime(INFINITY); //GEB5

  C_ITEM *it;

  if (qblock->length() == 0) { 
    return infTime;
  }
  else {
    it = qblock->ptop();
    return (it->get_time_tag());
  }

}

/************************************************************************
* unblock : unblock a socket						*
************************************************************************/
void SpSimObj::unblock(int soc) {
  C_ITEM *it;

//...... new code JSS
  update_block();
  if (qblock->get_item(soc) != NULL) {
    it = new C_ITEM();
    it->set_id(soc);
    qunblock->push_top(it);
  }
//...... end new code

}

/************************************************************************
* update_block : update the blocking lists				*
************************************************************************/
void SpSimObj::update_block() {
  int i,len;
  C_ITEM *item1;
  C_ITEM *item2;

  if (qblock->length() == 0) return;

  len = qunblock->length();
  for (i=0; i<len; i++) {
    item1 = qunblock->pop_top();
    item2 = qblock->deque(item1->get_id());
    if (item2 != NULL) {
      delete item2;
      delete item1;
    }
    else {
      qunblock->push_bot(item1);
    }
  }

}

/************************************************************************
* query : query virtual function					*
************************************************************************/
void SpSimObj::Query(double t, SpNameValueList &NameValueList) {

  if (NAME) NameValueList.SetNameOfList(NAME);

  NameValueList.SetNameValue("Object ID",GLOBAL_ID);
  NameValueList.SetNameValue("Time",t);
  NameValueList.SetNameValue("Pending Events",local_rqueue.get_length());

}

/************************************************************************
* query : query virtual function					*
************************************************************************/
void SpSimObj::NamedQuery(double t, SpNameValueList &NameValueList,
			  char *queryName) {

  if (NAME) NameValueList.SetNameOfList(NAME);

  NameValueList.SetName(queryName);

  NameValueList.SetNameValue("Object ID",GLOBAL_ID);
  NameValueList.SetNameValue("Time",t);
  NameValueList.SetNameValue("Pending Events",local_rqueue.get_length());

}

/************************************************************************
* event_type : get an integer value for an event name			*
************************************************************************/
int SpSimObj::event_type(char *c) {
  SpEventTypes *e;

  e = (SpEventTypes *)evtype;
  return e->get_messid(c);

}

/************************************************************************
* print_event_memory : print out free list event memory statistics	*
************************************************************************/
void SpSimObj::print_event_memory() {
  SpEventTypes *e;

  e = (SpEventTypes *)evtype;
  e->print();

}

/************************************************************************
* schedule : schedule an event                                          *
************************************************************************/
void *SpSimObj::schedule(const SpSimTime& time, int event_type,
			 int object_type, int object_id, int object_node) {


  return CurrentEvent->schedule(time,
				event_type,
				object_type,
				object_id,
				object_node);

}

/************************************************************************
* schedule : schedule an event                                          *
************************************************************************/
void *SpSimObj::schedule(const SpSimTime& time, int event_type,
			 int object_type, int object_id, int object_node,
			 int data_bytes, char *&buff) {


  return CurrentEvent->schedule(time,
				event_type,
				object_type,
				object_id,
				object_node,
				data_bytes,
				buff);

}

/************************************************************************
* combine_total_objects : correct the total number of objects		*
************************************************************************/
void SpSimObj::combine_total_objects() {
  int temp;

  temp = TOTAL_OBJECTS;
  temp = SpComm_GlobalMax(temp);

  if (temp != TOTAL_OBJECTS) {
    fprintf(stderr,"Warning, total objects not quite right... fixing it\n");
    TOTAL_OBJECTS = temp;
  }

  if ((TOTAL_OBJECTS%N_NODES) > NODE) {
    uid_seed = (TOTAL_OBJECTS/N_NODES)*N_NODES + N_NODES + NODE;
  }
  else {
    uid_seed = (TOTAL_OBJECTS/N_NODES)*N_NODES + NODE;
  }

}

/************************************************************************
* AddComponent : adds the specified SpComponent, along with all of its    *
* SpEventHandlers, to the component list                          	*
************************************************************************/
void SpSimObj::AddComponent(SpComponent *component) {

   if (component->AddedToSimobj)
   {
      cout << "Warning: attempted to add a component to a simobj that "
	   << "was all ready attached to that simobj: IGNORED"
	   << endl;
      return;
   }

   // add EventHandlers associated with this SpComponent

   EventHandlerElement *eh =
     (EventHandlerElement *) component->EventHandlerList.get_top();
   while (eh)
   {
      AddEventHandler (eh->eventName, eh->eventHandler);
      eh = (EventHandlerElement *) ++component->EventHandlerList;
   }

   // add SpComponent
   
   ComponentList += component;
   component->AddedToSimobj = 1;
   if (!component->LoadedSimobjData) component->LoadSimobjData(this);
}

/************************************************************************
* RemoveComponent : removes without deleting the specified SpComponent,   *
* along with all of its SpEventHandlers, from the component list         *
************************************************************************/
void SpSimObj::RemoveComponent(SpComponent *component) {

   // remove EventHandlers associated with this SpComponent
   
   if (0 == component->AddedToSimobj)
   {
      cout << "Warning: attempted to remove a component from a simobj that "
	   << "wasn't attached to that simobj: IGNORED"
	   << endl;
	return;
   }
   EventHandlerElement *eh =
     (EventHandlerElement *) component->EventHandlerList.get_top();
   while (eh)
   {
      RemoveEventHandler (eh->eventName, eh->eventHandler);
      eh = (EventHandlerElement *) ++component->EventHandlerList;
   }

   // remove, but do not delete, the SpComponent
   
   ComponentList -= component;
   component->AddedToSimobj = 0;
}

/************************************************************************
* internal method: SetUpEventHandlerHolder: create and initialize or    *
*                  find specified event handler holder                  *
************************************************************************/
SpEventHandlerHolder *SpSimObj::SetUpEventHandlerHolder (char *eventName) {
  SpEventHandlerHolder *eventHandlerHolder=NULL;

  eventHandlerHolder = (SpEventHandlerHolder *)
    EventHandlerHash.find(eventName);

  if (eventHandlerHolder == NULL) {
    
    eventHandlerHolder = RB_NEW_SpEventHandlerHolder();

    eventHandlerHolder->SetSimObj(this);
    eventHandlerHolder->SetObjectType(OBJECT_TYPE);
    eventHandlerHolder->SetLocalId(LOCAL_ID);
    eventHandlerHolder->SetGlobalId(GLOBAL_ID);
    eventHandlerHolder->SetNode(NODE);
    eventHandlerHolder->SetEventName(eventName);

    eventHandlerHolder->set_id(EventHandlerHash.get_hashid(eventName));
    EventHandlerHash += eventHandlerHolder;
  }
  return eventHandlerHolder;
}

/************************************************************************
* AddEventHandler : adds the event handler to the event handler hash   	*
* table.								*
************************************************************************/
void SpSimObj::AddEventHandler(char *eventName, 
			       SpStringBasedEventHandler *eventHandler) {
  SpEventHandlerHolder *eventHandlerHolder;

  eventHandlerHolder = SetUpEventHandlerHolder (eventName);
  eventHandlerHolder->AddEventHandler(eventHandler);
  //cout << "Added event handler for " << eventName << endl;
}

/************************************************************************
* AddSubScribeEventHandler : Same as previous, except it also
* subscribes this event handler to undirected eventhandler events
************************************************************************/
void SpSimObj::AddSubscribeEventHandler(char *eventName, 
			       SpStringBasedEventHandler *eventHandler) {

  SpEventHandlerHolder *eventHandlerHolder;
  eventHandlerHolder = SetUpEventHandlerHolder (eventName);
  if (!eventHandlerHolder->GetNsubscribed()) {
    SpObjHandle mastersimobjmgr;
    mastersimobjmgr.NodeId = get_NODE();
    mastersimobjmgr.ObjManId = -1;
    mastersimobjmgr.SimObjId = -1;
    SpObjHandle thisobj;
    thisobj.NodeId   = get_NODE();
    thisobj.ObjManId = get_OBJECT_TYPE();
    thisobj.SimObjId  = get_LOCAL_ID();
    if (strlen(eventName) > 80) {
      cerr << "Warning: eventName exceeds 80 characters" << endl;
      cerr << "Only the first 80 characters are ";
      cerr << "guaranteed unique" << endl;
    }
    SpString80 stringName = eventName;
    ScheduleAddSimObjInteraction(
				 CurrentEvent->TIME_TAG,
				 mastersimobjmgr,
				 stringName,
				 thisobj,
				 get_GLOBAL_ID()
				 );
  }
  eventHandlerHolder->AddEventHandler(eventHandler,true);

}

/************************************************************************
* AddEventHandlerMethod : adds the event handler method to the event    *
*                         handler hash table.				*
************************************************************************/
SpEventHandler *SpSimObj::AddEventHandlerMethod(char *eventName,
				      EVENT_HANDLER_METHOD_FNC_PTR method,
				      void *objPtr)
{
  SpEventHandlerHolder *eventHandlerHolder;

  eventHandlerHolder = SetUpEventHandlerHolder (eventName);
  return eventHandlerHolder->AddEventHandlerMethod(objPtr, method);
  //cout << "Added event handler method for " << eventName << endl;
}


/************************************************************************
* RemoveEventHandler : removes the event handler from the event handler	*
* hash table.								*
************************************************************************/
void SpSimObj::RemoveEventHandler(char *eventName,
				  SpStringBasedEventHandler *eventHandler,
				  enum SpEventHandlerMode flag) 
{
  SpEventHandlerHolder *eventHandlerHolder=NULL;

  eventHandlerHolder = (SpEventHandlerHolder *)
    EventHandlerHash.find(eventName);
  
  if (eventHandlerHolder == NULL) {
    cout << "Error, could not find event handler holder ";
    cout << eventName << endl;
    return;
  }

  int nsubscribed = eventHandlerHolder->GetNsubscribed();
  eventHandlerHolder->RemoveEventHandler(eventHandler,flag);

  // Unsubscribe from undirected interactions/eventhandler events,
  // if necessary
  if (nsubscribed & !eventHandlerHolder->GetNsubscribed()) {
    SpObjHandle mastersimobjmgr;
    mastersimobjmgr.NodeId = get_NODE();
    mastersimobjmgr.ObjManId = -1;
    mastersimobjmgr.SimObjId = -1;
    ScheduleRemoveSimObjInteraction(
				    GetTime(),
				    mastersimobjmgr,
				    eventName,
				    get_GLOBAL_ID()
				    );
  }
}

/************************************************************************
* ProcessEventHandlers : processes the event handlers for a specific	*
* event type.								*
************************************************************************/
void SpSimObj::ProcessEventHandlers(char *eventName,
				    SpSimTime &timeTag,
				    char *data,
				    int dataBytes,
				    int externalId,
				    int ehType,
				    int undirected) {
  SpEventHandlerHolder *eventHandlerHolder=NULL;
    
  eventHandlerHolder = (SpEventHandlerHolder *)
    EventHandlerHash.find(eventName);

  if (eventHandlerHolder) {
     eventHandlerHolder->ProcessEventHandlers(timeTag,
					      data,
					      dataBytes,
					      externalId,
					      ehType,
					      undirected);
  }
  else {

//    cout << "Warning2, no handlers for event " << eventName << endl;

  }
}

/************************************************************************
* CheckEventHandler : checks if the event handler for the eventName has	*
* been defined								*
************************************************************************/
int SpSimObj::CheckEventHandler(char *eventName) {

  if (EventHandlerHash.find(eventName)) {
    return 1;
  }
  else {
    return 0;
  }

}

void SpSimObj::SetObjectHandle(int nodeId, int objManId, int simObjId)
{
   ObjectHandle.NodeId   = nodeId;
   ObjectHandle.ObjManId = objManId;
   ObjectHandle.SimObjId = simObjId;
}

void SpSimObj::GetObjectHandleAttributes(const SpObjHandle& simobj, int& nodeId,
					  int& objManId, int& simObjId)
{
   nodeId = simobj.NodeId;
   objManId = simobj.ObjManId;
   simObjId = simobj.SimObjId;   
}

SpObjHandle *SpSimObj::GetObjectHandle()
{
   return &ObjectHandle;
}


SpSimTime SpSimObj::GetEarliestNextEventTime () {

  int qMessLen = ((GetEvent())->get_mque())->get_length();
  SpSimTime nextTime = GetTime();

  if(nextTime.GetCounter() < EvCounter)
    nextTime.SetCounter(EvCounter + qMessLen + 1);
  else
    nextTime.SetCounter(nextTime.GetCounter() + qMessLen + 1);

  return nextTime;
}

/******************************************************************************
* Internal Method: get_ncancel
* Summary: Get the number of cancelled events for this object
* Description: 
* Return: the number of cancelled events for this object
******************************************************************************/
int SpSimObj::GetNumCanceled() {return CancelTree->GetNumElements();}

void SpSimObj::AttachCancelTreePointer()
{
  PO_ATTACH_PTR_RB_SpBinaryTree(CancelTree);
}

int SpSimObj::CheckSubType(char *type, int externalId) {
  int RetVal = 0;
  SUB_TYPE *SubType = (SUB_TYPE *)EXTSUBS.get_top();
  for (int i=0; i<EXTSUBS.get_length(); i++) {
    if (SubType->Compare(type, externalId)) {
      RetVal = 1;
      break;
    }
    SubType = (SUB_TYPE *)SubType->get_link();
  }
  return RetVal;
}

int SpSimObj::CheckSubscription(char *name) {
  int RetVal = 0;
  SUB_TYPE *SubType = (SUB_TYPE *)EXTSUBS.get_top();
  for (int i=0; i<EXTSUBS.get_length(); i++) {
    if (SubType->Compare(name)) {
      RetVal = 1;
      break;
    }
    SubType = (SUB_TYPE *)SubType->get_link();
  }
  return RetVal;
}



void SpSimObj::print() {printf("Nothing to print here\n");}

void SpSimObj::terminate(double){}
