// SpSimObjMgr.C method file

#include "SpStdio.H"
#include "SpStdlib.H"
#include "speedes_comm.H"
#include "SpSimObjMgr.H"
#include "SpDataParser.H"
#include "SpObjHandle.H"
#include "PO.H"

#include "RB_SpBinaryTree.H"
#include "RB_SpHashTree.H"
PO_DEFINE_CLASS(RB_SpBinaryTree);
#include "RB_SpDefineClass.H"

#ifdef __IBMCPP__
namespace {
#endif

#ifndef PO_VOID_TYPE_CLASS_DEFINED
RB_DEFINE_CLASS_PRIMATIVE(PO_VOID_TYPE_CLASS);
#define PO_VOID_TYPE_CLASS_DEFINED
#endif

#ifndef PO_DEFINED_CHAR
RB_DEFINE_CLASS_PRIMATIVE(char);
#define PO_DEFINED_CHAR
#endif

#ifdef __IBMCPP__
}
#endif

//#include "multirouter.H"

int SpSimObjMgr::done = 0;
int SpSimObjMgr::ntobj = 0;
int SpSimObjMgr::seed_update = 0;

/************************************************************************
* SpSimObjMgr : construct a objman object					*
************************************************************************/
SpSimObjMgr::SpSimObjMgr(){

  if (SpPoDataBase::DoNotAllocateMemoryInConstructor){
    //cout<<"Creating a SpSimObjMgr with persistence"<<endl;
  }
  else{

#ifdef COMMENT_OUT
    gobj = NULL;
#endif
    LocalSimObjs = new RB_SpHashTree;
    CreationOffset = 0;
    LOCAL_ID = -1;
    OBJECT_TYPE = N_OBJECT_TYPES;
    interact = 0;
    reset_nlobjs();
    GLOBAL_ID = 0x7fffffff - NODE - N_OBJECT_TYPES*N_NODES;
    set_uid_seed(get_uid_seed() - N_NODES);
    N_OBJECT_TYPES++;

    CombinedTotalEvents = 0;
    CombinedCommittedEvents = 0;
    CombinedTotalTime = 0.0;
    CombinedCommittedTime = 0.0;

    if (done) return;

  }
  ntobj = 0;
  seed_update = 1 + NODE;
  
  done = 1;
}

/************************************************************************
* init_events : initialize the events					*
************************************************************************/
void SpSimObjMgr::init_events() {

}

/************************************************************************
* term : Calls the terminate method on each contained simobj before 	*
*	 calling the terminate method on itself. This allows users	*
*	 to have some code executed on their simobjs and object		*
*	 managers to do final data collection.				*
************************************************************************/
void SpSimObjMgr::term(double tend) {

  SpSimObj *simObj = (SpSimObj *)LocalSimObjs->GetFirstElement();
  while (simObj) {
    simObj->terminate(tend);
    simObj = (SpSimObj *)LocalSimObjs->GetNextElement();
  }

#ifdef COMMENT_OUT
  int i;
  for (i=0; i<N_LOC; i++) {
    gobj[i]->terminate(tend);
  }
#endif

  terminate(tend);
}

/************************************************************************
* CriticalPath : Determines the critical path for the simobjs being	*
*		 managed by my object manager. Note, the global sums	*
*		 and maximums are done once by objects (which manages	*
*		 all of the object managers)				*
************************************************************************/
void SpSimObjMgr::ComputeCriticalPath(double &CPU_Time,
				   double &criticalPathTime) {

  SpSimObj *simObj = (SpSimObj *)LocalSimObjs->GetFirstElement();
  while (simObj) {

    if (simObj->GetProcessingTime() > simObj->GetCriticalPathTime()) {
      cout << "Error (ComputeCriticalPath): "
	   << simObj->get_NAME() << ", Processing Time = "
	   << simObj->GetProcessingTime()
	   << ", Critical Path Time = "
	   << simObj->GetCriticalPathTime()
	   << endl;
    }

    CPU_Time += simObj->GetProcessingTime();
    if (criticalPathTime < simObj->GetCriticalPathTime()) {
      criticalPathTime = simObj->GetCriticalPathTime();
    }

    simObj = (SpSimObj *)LocalSimObjs->GetNextElement();
  }

#ifdef COMMENT_OUT
  int i;
  for (i=0; i<N_LOC; i++) {

/*
    cout << gobj[i]->get_NAME() << " critical path time = "
	 << gobj[i]->GetCriticalPathTime() << endl;
*/

    if (gobj[i]->GetProcessingTime() > gobj[i]->GetCriticalPathTime()) {
      cout << "Error (ComputeCriticalPath): "
	   << gobj[i]->get_NAME() << ", Processing Time = "
	   << gobj[i]->GetProcessingTime()
	   << ", Critical Path Time = "
	   << gobj[i]->GetCriticalPathTime()
	   << endl;
    }

    CPU_Time += gobj[i]->GetProcessingTime();
    if (criticalPathTime < gobj[i]->GetCriticalPathTime()) {
      criticalPathTime = gobj[i]->GetCriticalPathTime();
    }
  }
#endif

  CPU_Time += GetProcessingTime();
  if (criticalPathTime < GetCriticalPathTime()) {
    criticalPathTime = GetCriticalPathTime();
  }


}


/************************************************************************
* set_N_LOC : set the number of local objects based on N_TOT		*
************************************************************************/
void SpSimObjMgr::set_N_LOC() {

  N_LOC = N_TOT / N_NODES;
  if ((N_TOT % N_NODES) > NODE) N_LOC++;

}

/************************************************************************
* get_nodeid : get the node and id of an object	given its id		*
************************************************************************/
void SpSimObjMgr::get_nodid(int oi, int &n, int &i) {

  i = oi / N_NODES;
  n = oi % N_NODES;

}

/************************************************************************
* get_oid : get the object id given node and local id			*
************************************************************************/
void SpSimObjMgr::get_oid(int &oi, int n, int i) {

  oi = i*N_NODES + n;

}
/************************************************************************
* get_allids : a buffer of all ids for the old host router		*
************************************************************************/
char *SpSimObjMgr::get_allids() {
  int i,j;
  int size;
  int size1;
  char *buff;
  int *ibuff;
  char *name;

  if (!interact) return NULL;

  size1 = 3*sizeof(int) + 40;
  size = size1 * LocalSimObjs->GetNumElements();

  buff = new char[size];

  i = 0;
  SpSimObj *simObj = (SpSimObj *)LocalSimObjs->GetFirstElement();
  while (simObj) {

    j = i*size1;
    ibuff = (int *)&buff[j];
    ibuff[0] = i;
    ibuff[1] = OBJECT_TYPE;
    ibuff[2] = simObj->get_GLOBAL_ID();

    name = simObj->get_NAME();
    if (name == NULL) {
      strcpy((char *)&ibuff[3], "No Name");
    }else{
      strcpy((char *)&ibuff[3], name);
    }

    simObj = (SpSimObj *)LocalSimObjs->GetNextElement();
    i++;
  }

#ifdef COMMENT_OUT
  for (i=0; i<N_LOC; i++) {
    j = i*size1;
    ibuff = (int *)&buff[j];
    ibuff[0] = i;
    ibuff[1] = OBJECT_TYPE;
    ibuff[2] = gobj[i]->get_GLOBAL_ID();
    name = gobj[i]->get_NAME();
    if (name == NULL) {
      strcpy((char *)&ibuff[3], "No Name");
    }else{
      strcpy((char *)&ibuff[3], name);
    }
  }
#endif

  return buff;

}

/************************************************************************
* set_global_names : construct a table of names for global lookups	*
************************************************************************/
void SpSimObjMgr::set_global_names() {
  int i;
  int hid;
  //int buff_size;
  int message_size;
  char *message;
  char *pointer;
  //C_MULTIROUTER *multirouter; SpComm

//...... construct the table of global names and initialize my names

  global_name = new C_GLOBAL_NAME[N_TOT];

  i = 0;
  SpSimObj *simObj = (SpSimObj *)LocalSimObjs->GetFirstElement();
  while (simObj) {

    if (strlen(simObj->get_NAME()) > MAX_NAME_LENGTH) {
      memcpy(global_name[i].name, simObj->get_NAME(), MAX_NAME_LENGTH);
      global_name[i].name[MAX_NAME_LENGTH-1] = '\0';
    }else{
      strcpy(global_name[i].name, simObj->get_NAME());
    }
    global_name[i].node = NODE;
    global_name[i].lid = i;

    simObj = (SpSimObj *)LocalSimObjs->GetNextElement();
    i++;
  }

#ifdef COMMENT_OUT
  for (i=0; i<N_LOC; i++) {
    if (strlen(gobj[i]->get_NAME()) > MAX_NAME_LENGTH) {
      memcpy(global_name[i].name, gobj[i]->get_NAME(), MAX_NAME_LENGTH);
      global_name[i].name[MAX_NAME_LENGTH-1] = '\0';
    }else{
      strcpy(global_name[i].name, gobj[i]->get_NAME());
    }
    global_name[i].node = NODE;
    global_name[i].lid = i;
  }
#endif

//...... send and receive names with other nodes

  if (N_NODES > 1) {

    SpComm_CoordinatedSend(N_LOC*sizeof(C_GLOBAL_NAME), (char *)global_name);

    pointer = (char *)&global_name[N_LOC];
    while((message = (char *)SpComm_CoordinatedReceive(message_size)) != NULL){
      memcpy(pointer, message, message_size);
      pointer += message_size;
      delete [] message;
    }

    if (pointer != (char *)&global_name[N_TOT]) {
      fprintf(stderr,"Error (objman): set_global_names not lined up\n");
    }

  }

//...... set up the name queue for hashing

  name_queue = new C_QUEUE[N_TOT];
  for (i=0; i<N_TOT; i++) {
    hid = name_hash(global_name[i].name);
    name_queue[hid].push_bot(&global_name[i]);
  }

}

/************************************************************************
* name_hash : convert a name into an integer between 0 and N_TOT	*
************************************************************************/
int SpSimObjMgr::name_hash(char *obj_name) {
  int i;
  int hid;
  char s[MAX_NAME_LENGTH];
  int *pint;
  int nchars;

  pint = (int *)s;
  for (i=0; i<int(MAX_NAME_LENGTH/sizeof(int)); i++) pint[i] = 0;
  nchars = strlen(obj_name);

  if (nchars > MAX_NAME_LENGTH) {
    memcpy(s, obj_name, MAX_NAME_LENGTH);
    s[MAX_NAME_LENGTH-1] = '\0';
  }else{
    strcpy(s, obj_name);
  }

  hid = 0;
  for (i=0; i<int(MAX_NAME_LENGTH/sizeof(int)); i++) hid ^= pint[i];
  hid %= N_TOT;

  return hid;

}

/************************************************************************
* print_global_names : print out the global names			*
************************************************************************/
void SpSimObjMgr::print_global_names() {
  int i,j;
  C_GLOBAL_NAME *gn;

  for (i=0; i<N_TOT; i++) {
    fprintf(stderr,"Name Queue length = %d\n", name_queue[i].get_length());
    gn = (C_GLOBAL_NAME *)name_queue[i].get_top();
    for (j=0; j<name_queue[i].get_length(); j++) {
      cerr << "  " << gn->name << ": Node " << gn->node << ", local id "
	   << gn->lid << "\n";
      gn = (C_GLOBAL_NAME *)gn->get_link();
    }
  }

}

/************************************************************************
* get_nodeid : get the node and local id given an object name		*
************************************************************************/
void SpSimObjMgr::get_nodid(char *name, int &nd, int &lid) {
  int i;
  int len;
  int hid;
  C_GLOBAL_NAME *gn;
  char trunc_name[MAX_NAME_LENGTH];

//...... construct a truncated name

  if (strlen(name) > MAX_NAME_LENGTH) {
    memcpy(trunc_name, name, MAX_NAME_LENGTH);
    trunc_name[MAX_NAME_LENGTH-1] = '\0';
  }else{
    strcpy(trunc_name, name);
  }

//...... find the global_name object with the right name

  hid = name_hash(name);
  gn = (C_GLOBAL_NAME *)name_queue[hid].get_top();
  len = name_queue[hid].get_length();

  for (i=0; i<len; i++) {
    if (!strcmp(gn->name, trunc_name)) {
      nd = gn->node;
      lid = gn->lid;
      return;
    }
    gn = (C_GLOBAL_NAME *)gn->get_link();
  }

//...... error condition, could not find the name

  nd = -1;
  lid = -1;

//  fprintf(stderr,"Error (objman): could not find name %s\n",name);

}


/************************************************************************
* add_object : add an object to speedes					*
************************************************************************/
void SpSimObjMgr::add_object(SpSimObj *simObj) {

  LocalSimObjs->Insert(simObj, simObj->get_LOCAL_ID());

  // To add the dynamic objection creation feature, we made N_LOC be an
  // RB_int, but that causes the g++ compiler to be confused by the
  // overloaded += operator in the line, "TOTAL_OBJECTS += N_LOC".  Such
  // lines typically appear in ObjMgr constructors.  The solution was
  // to change N_LOC back to a regular built-in type int, but that
  // means that in the line, "++N_LOC", no roll back information would
  // be saved, meaning that the old value would NOT be restored since
  // now it is a regular int instead of an RB_int.  Therefore, instead
  // of doing that ++N_LOC line, we do this RB_memcpy which will
  // restore the old value.
  
  const int new_N_LOC = N_LOC + 1;
  RB_memcpy((char*) &N_LOC, (char*) &new_N_LOC, sizeof(N_LOC));

#ifdef COMMENT_OUT
  SpSimObj **new_gobj;

  if (N_LOC >= ndef) {
    ndef *= 2;
    new_gobj = new SpSimObj*[ndef];
    if (new_gobj == NULL) {
      fprintf(stderr,"No more memory: could not add object\n");
      exit(1);
    }
    memcpy((char *)new_gobj, (char *)gobj, N_LOC*sizeof(SpSimObj *));
    delete gobj;
    gobj = new_gobj;
  }

  gobj[simobj->get_LOCAL_ID()] = simobj;
  N_LOC++;
#endif


}

/************************************************************************
* remove_object : remove an object from speedes				*
************************************************************************/
void SpSimObjMgr::remove_object(SpSimObj *simObj) {

#ifdef COMMENT_OUT
  gobj[simobj->get_LOCAL_ID()] = NULL;
#endif

  if (simObj != LocalSimObjs->Remove(simObj->get_LOCAL_ID())) {
    RB_cout << "Error, bad remove_object" << endl;
    return;
  }

  // See the comment in add_object for an explanation of this RB_memcpy:
  const int new_N_LOC = N_LOC - 1;
  RB_memcpy((char*) &N_LOC, (char*) &new_N_LOC, sizeof(N_LOC));

}

/************************************************************************
* deal_me : card deal objects for my node				*
************************************************************************/
int SpSimObjMgr::deal_me(int nobjs) {
  int temp;

  temp = (nobjs) / N_NODES;
  if (nobjs % N_NODES > 0) temp++;

  return temp;

}

/************************************************************************
* deal_me : card deal objects for my node				*
//   nobjs is the total number of objects that the calling manager
// wants distributed to *all* nodes (i.e. the total number of objects
// that need to be managed by some node).  This function returns the
// number of those objects that this node should manage (the other
// nodes will manage the other objects).  The calling manager manages
// objects with ids in the calling manager of offset + N_NODES * j,
// where j is a whole number.
************************************************************************/
int SpSimObjMgr::deal_me(int nobjs, int &offset) {
  int temp;
  int starting_node;

  starting_node = TOTAL_OBJECTS % N_NODES;
  offset = (NODE - starting_node + N_NODES) % N_NODES;
  temp = (nobjs - offset) / N_NODES;
  if ((nobjs - offset) % N_NODES > 0) temp++;

  return temp;
}

void SpSimObjMgr::CardDeal(int totalSimObj,
			SpDynPtrArray &locObj,
			SIMOBJ_CONSTRUCTOR_FUNCTION simobjConstFunc,
			SpDataParser &parser,
			int replication,
			char *objTypeName,
			int useCodeGenParFilePattern)
{
   int offset;

   if (!totalSimObj) return;
   if (useCodeGenParFilePattern)
   {
      parser.MoveOutCompletely();
      parser.MoveIn("Objects");
      N_TOT = parser.GetNumSetsInGroup("OBJECT");
      parser.MoveToFirstSetInGroup("OBJECT");
   }
   else N_TOT = totalSimObj;

   N_LOC = deal_me(N_TOT, offset);

   reserve_n_objects(N_LOC);
   TOTAL_OBJECTS += N_TOT;
   set_interact();

   for (int globalId = offset, localId = 0;
	localId < N_LOC;
	globalId += N_NODES, localId++)
   {
      SpSimObj *newSimobj = simobjConstFunc(parser);
      newSimobj->set_SEED(10000*replication + 12345 + globalId);
      if (useCodeGenParFilePattern)
      {
	 newSimobj->set_NAME(parser.GetSetName());
	 parser.MoveToNextSetInGroup();
      }
      define_object(newSimobj);
      locObj[localId] = newSimobj;
   }
   if (useCodeGenParFilePattern) parser.MoveOutCompletely();
   
   cout << "Card Deal: NODE "
	<< NODE
	<< " Created "
	<< N_LOC
	<< " out of "
	<< N_TOT
	<< " "
	<< objTypeName
	<< " objects"
	<< endl;
}

void SpSimObjMgr::SetObjectHandle(int nodeId, int objManId, int simObjId) {
   SpSimObj::SetObjectHandle(nodeId, objManId, simObjId);
}


void SpSimObjMgr::RegisterSimObjsWithPersistence() {

// NOTE: THIS BREAKS THE PERSISTENCE SUPPORT - WE NEED TO GET A PERSISTENT
// VERSION OF THE RB_SpHashTree TO PROCEED. WHEN IT IS AVAILABLE, WE MIGHT
// NOT EVEN NEED TO PROVIDE THIS ROUTINE.

#ifdef COMMENT_OUT
  PO_ADD(gobj, sizeof(SpSimObj*)*N_LOC);
  for (int i=0; i < N_LOC; ++i){
    PO_ATTACH_PTR(gobj[i]);
    gobj[i]->RegisterWithPersistence();
  }
  PO_ATTACH_PTR(gobj);
#endif

  //This is a fix that works without requiring a RB_SpHashTree to proceed.
  //However, it does not support dynamic object creation/deletion.

  char* poName = new char[strlen(NAME) +1 + 6]; //at most 1M loc simobjs/mgr
  SpSimObj *simObj = (SpSimObj *)LocalSimObjs->GetFirstElement();

  PO_ADD_char(NAME, strlen(NAME)+1);
  PO_ATTACH_PTR_char(NAME);
  for (int i=0; i < N_LOC; ++i){
    sprintf(poName, "%s%d", NAME, i);
    PO_SET_NAME(simObj, poName);
    simObj->AttachCancelTreePointer();
    if (SpPoDataBase::PO_errno){
      SpPoDataBase::PO_perror(poName);
    }
    simObj->RegisterWithPersistence();
    simObj = (SpSimObj *)LocalSimObjs->GetNextElement();
  }
  delete [] poName;
}


void SpSimObjMgr::ResetLocalRqueues() {

  SpSimObj *simObj = (SpSimObj *)LocalSimObjs->GetFirstElement();
  while (simObj) {
    simObj->get_local_rqueue()->clear();
    simObj->ResetRqueueHolder();
    simObj = (SpSimObj *)LocalSimObjs->GetNextElement();
  }
}

void SpSimObjMgr::ResetObjectForRestart()
{
  LocalSimObjs = new RB_SpHashTree;
}


void SpSimObjMgr::reserve_n_objects(int numObjs) {
  if (numObjs) {
    LocalSimObjs->SetArraySize(numObjs);
  }else{
    LocalSimObjs->SetArraySize(1);
  }
}


void SpSimObjMgr::define_object(SpSimObj *simObj ) {
  LocalSimObjs->Insert(simObj, simObj->get_LOCAL_ID());
}

SpSimObj *SpSimObjMgr::get_obj(int localId) {
  return (SpSimObj *)LocalSimObjs->Find(localId);
}

