// speedes.C - SPEEDES class that defines the primary SPEEDES routine:
// Execute()
//
// SPEEDES stands for "Synchronous Parallel Environment for Emulation and
//                     Discrete Event Simulation"

// __OLD_MATHERR_NAMES is defined so it will work on the Origin 2000
// computer named janeway.cmf.nrl.navy.mil
#define __OLD_MATHERR_NAMES
#include "SpMath.H"

#include "SpExecute.H"
#include "SpStdio.H"
#ifndef __SPEEDES_NT__
#include <sys/time.h>
#endif

#include "SpSignal.H"
#include "SpStdlib.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"

#include "SpStats.H"
#include "SpTimeMgtAlg.H"
#include "SpSequential.H"
#include "td_evtq.H"
#include "SpBTW.H"

#include "speedes_comm.H"
#include "SpDataParser.H"

#include "SpWalltime.H"
#include "SpSimTime.H" //GEB2
#include "SpCrashOnDivZero.H"
#include "SpErrorSignals.H"
#include "SpPoDataBase.H"
#include "PO.H"

#include "SpOS.H"

#include "SpPubSubItem.H"

extern SpEvent *CurrentEvent;

#ifdef ERROR_CHECK
// Linux does not appear to support matherr:
#ifndef Linux
// The following is defined only on the HP Convex Exemplar,
// but not on other HPs:
#ifndef __HP_CXD_SPP
int matherr(struct exception *x)
{
  cout << "Error: matherr in " << __FILE__
       << " was called because of an error during a call to "
       << x->name << endl;
  assert(0);
  return 0;
}
#endif
#endif
#endif

#ifndef INFINITY
#define INFINITY 1.0e20
#endif

#ifndef OLD_STYLE_PLUG_IN
void PlugInBuiltInEvents();

#else // i.e., #ifdef OLD_STYLE_PLUG_IN
extern INITIAL_EVENT_TYPE;
int INITIAL_EVENT = INITIAL_EVENT_TYPE;

extern EVENT_HANDLER_EVENT_EVENT_TYPE;
extern INITIAL_EVENT_TYPE;
extern CANCEL_EVENT_TYPE;
extern QUERY_EVENT_TYPE;
extern MONITOR_EVENT_TYPE;
extern CREATE_EVENT_TYPE;
extern DESTROY_EVENT_TYPE;
extern PAUSE_EVENT_TYPE;
extern AddSharedSem_EVENT_TYPE;
extern RequestSharedSem_EVENT_TYPE;

int EVENT_HANDLER_EVENT = EVENT_HANDLER_EVENT_EVENT_TYPE;
int INITIAL_EVENT       = INITIAL_EVENT_TYPE;
int CANCEL              = CANCEL_EVENT_TYPE;
int QUERY               = QUERY_EVENT_TYPE;
int MONITOR             = MONITOR_EVENT_TYPE;
int CREATE              = CREATE_EVENT_TYPE;
int DESTROY             = DESTROY_EVENT_TYPE;
int PAUSE               = PAUSE_EVENT_TYPE;
int AddSharedSem        = AddSharedSem_EVENT_TYPE;
int RequestSharedSem    = RequestSharedSem_EVENT_TYPE;

main (int argc, char **argv)
{
   SPEEDES Speedes;
   Speedes.Execute(argc, argv);
}
#endif

extern int PersistenceEnabled;

SpPoDataBase* MessageDataBase;

SPEEDES *Speedes = NULL;

int ARGC;
char **ARGV;

char *TimeManagementMode;

SpDataParser *speedes_parser = NULL;

void HandleCommandLine(int argc,
		       char **argv,
		       int &nTot,
		       int &nLoc,
		       int &group,
		       char ** pInfo);



void Speedes_sigintfunc(int) {
  if (SpComm_GetNodeId() == 0) cout << endl;
  exit(1);
}

void sigchldfunc(int) {
  cout << "My child exited!" << endl;
}

static char * DirectoryForCheckpointFiles;


/************************************************************************
* SPEEDES class methods 		       				*
************************************************************************/

SPEEDES::SPEEDES(): cumNumObjMan(0), cumNumEvent(0), ExecuteAlreadyCalled(0)
   {}
void SpRB_and_PO_PlugInEngine();

int SPEEDES::Execute(int argc, char **argv)
{
   crash_on_divide_by_zero();
   SpTimeMgtAlg *evtq;
   SpDataParser *parser;
   SpSimTime old_gvt;
   SpSimTime gvt;
   SpSimTime tend;
   double min_step;
   long cpu0;
   long cpu1;
   int scaled_time;
   double scaler;
   int node;
   int n_nodes;
   char *mode;
   int lazy;
   int statistics;
   int test_rollback;
   int check_reptime;
   int trace;
   char *tracefile;
   
   int WriteLookahead;
   char *Lookaheadfile;
   
   int interactive;
   double spin;
   double wallclock0;
   double wallclock;
   double wallgvt;
   double wallgvt0;
   int scale_flag1;
   int optimize_sequential;
   double WallTimeForCheckpoint = -1;
   double SimTimeForCheckpoint = -1;
   double WallForLastCheckpoint = 0;
   double SimTimeForLastCheckpoint = 0;

   if (ExecuteAlreadyCalled)
   {
      cout << "Error: called ExecuteSpeedes() twice" << endl;
      exit(-1);
   }

   ExecuteAlreadyCalled = 1;
   
   //DefaultDataBase = new SpPoDataBase;

   // When I switched speedes to use shared libraries instead of
   // archives, some regression tests crashed on janeway, but when I
   // added this cout, they stopped crashing:
   cout << ' ';

   SpRB_and_PO_PlugInEngine();
   // This database is constructed just so that we set the default add
   // value for databases.  The type "PO_VOID_TYPE_CLASS" has now been
   // defined through the above PO_REGISTER_ATOMIC_CLASSES and we will
   // use this type as the void type.  It is really just a typedef'ed 
   // char type but all I need is something like that.
   SpPoDataBase* jj = new SpPoDataBase;
   delete jj;
   //...... set argv and argc
   
   ARGC = argc;
   ARGV = argv; 
   
   //...... print out the copyright notice
   
   fprintf(stderr,
"SPEEDES Version 0.71\n"
"Released on 3-18-1999\n"
"Copyright 1999 National Aeronautics and Space Administration\n"
"All rights reserved\n"
	   );
   
#ifndef OLD_STYLE_PLUG_IN
    PlugInBuiltInEvents();
    //Speedes->Print();
#endif
   
   //Start up the comm library and fork off the children
   

    int nNode=-1, nLoc=-1, group=-1;
    char * pipeInfo = NULL;
    HandleCommandLine(argc, argv, nNode, nLoc, group, &pipeInfo);

    if (SpComm_StartUp(nNode, nLoc, group, pipeInfo) == ERROR_SP_COMM){
      cerr<<"The comm library initialization failed!!!!!!"<<endl;
      exit(-1);
   }

    SpMainPubSubItem::InitSpMainPubSubItemClass
      (SpComm_GetNodeId(), SpComm_GetNumNodes());
    
#ifndef __SPEEDES_NT__
   ErrorSignals(SpComm_CleanupShm);
         atexit(SpComm_CleanupShm);

   signal(SIGINT, Speedes_sigintfunc);

   signal(SIGCHLD, sigchldfunc);
#endif

   
   //...... initialize parallel hypercube environment variables

   n_nodes = SpComm_GetNumNodes();
   node = SpComm_GetNodeId();

   // plug-in SPEEDES library events (this is a global function)
      
   //...... set up the node information
   
   evtq = NULL; // do this on purpose (static variables set only)
   evtq->set_speedes_object(this);
   evtq->set_node_info(n_nodes, node);
   
   //...... initialize the some simulation values

   gvt = -INFINITY;
   gvt.SetPriority1(0x80000000);
   gvt.SetPriority2(0x80000000);
   gvt.SetCounter(0x80000000);
   gvt.SetUniqueId(0x80000000);

   //...... get speedes parameters out of the parser
   
   parser = new SpDataParser("speedes.par");

   if (parser->GoTo("Checkpoint", NULL) != NULL){
     if (parser->GetLogical("Enable")){
       MessageDataBase = new SpPoDataBase;
       WallTimeForCheckpoint = ceil(parser->GetFloat("WallTimeInterval"));
       SimTimeForCheckpoint = ceil(parser->GetFloat("SimTimeInterval"));
       DirectoryForCheckpointFiles=strdup(parser->GetString("CheckpointPath"));
       if (WallTimeForCheckpoint <= 0){
	 WallTimeForCheckpoint = INFINITY;
       }
       else{
	 evtq->SetDoingCheckpoints();
       }
       if (SimTimeForCheckpoint <= 0){
	 SimTimeForCheckpoint = INFINITY;
       }
       else{
	 evtq->SetDoingCheckpoints();
       }
     }
     else{
       SimTimeForCheckpoint = INFINITY;
       WallTimeForCheckpoint = INFINITY;
     }
   }
   else{
     cout<<"\aParser error!!!!  Should add a checkpointing section"<<endl;
     SimTimeForCheckpoint = INFINITY;
     WallTimeForCheckpoint = INFINITY;
   }

   PersistenceEnabled = evtq->DoingCheckpoints();


   speedes_parser = parser;
   parser->GoTo("parameters", NULL);
   
   tend = parser->GetFloat("tend");
   tend.SetPriority1(0x7fffffff);
   tend.SetPriority2(0x7fffffff);
   tend.SetCounter(0x7fffffff);
   tend.SetUniqueId(0x7fffffff);
  
   min_step = parser->GetFloat("lookahead");
   scaled_time = parser->GetLogical("scaled_time");
   scaler = parser->GetFloat("scaler");
   statistics = parser->GetLogical("statistics");
   lazy = parser->GetLogical("lazy");
   test_rollback = parser->GetLogical("test_rollback");
   check_reptime = parser->GetLogical("check_reptime");
   
   optimize_sequential = parser->GetLogical("optimize_sequential");
   
   interactive = parser->GetLogical("interactive");
   spin = parser->GetFloat("spin");
   
   mode = parser->GetString("mode");
   TimeManagementMode = mode;
   
   evtq = NULL;
   evtq->set_lazy(lazy);
   
   //...... open the tracefile if trace is turned on
   
   
   parser->GoTo("trace", NULL);
   trace = parser->GetLogical("trace");
   tracefile = parser->GetString("tracefile");
   if (trace) {
      evtq->set_trace();
      evtq->set_tracefile(tracefile);
   }
   
   //...... open the Lookahead if Lookahead Statistics is turned on
   
   parser->GoTo("Lookahead", NULL);
   WriteLookahead = parser->GetLogical("WriteLookahead");
   Lookaheadfile = parser->GetString("Lookaheadfile");
   if (WriteLookahead) {
      evtq->set_WriteLookahead();
      evtq->set_Lookaheadfile(Lookaheadfile);
   }
   
   
   //...... create and initialize the event queue
   //...... (keep backwards compatible names)
   
   if ((n_nodes > 1) || (!optimize_sequential)) {
      
      if (!strcasecmp(mode,"Conservative")) evtq = new SpBTW();
      if (!strcasecmp(mode,"FTB")) evtq = new SpBTW();
      if (!strcasecmp(mode,"FIXED_TIME_BUCKETS")) evtq = new SpBTW();
      if (!strcasecmp(mode,"TIME_BUCKETS")) evtq = new SpBTW();
      if (!strcasecmp(mode,"TimeBuckets")) evtq = new SpBTW();
      if (!strcasecmp(mode,"FIXED_TIME_BUCKET")) evtq = new SpBTW();
      if (!strcasecmp(mode,"TIME_BUCKET")) evtq = new SpBTW();
      if (!strcasecmp(mode,"TimeBucket")) evtq = new SpBTW();

      /*
      ** Note that Time Warp and Breathing Time Buckets are provided by
      ** Breathing Time Warp by internally setting the Nrisk value
      ** (see evtq constructor).
      */
      
      if (!strcasecmp(mode,"TW")) evtq = new SpBTW();
      if (!strcasecmp(mode,"TWOS")) evtq = new SpBTW();
      if (!strcasecmp(mode,"TIME_WARP")) evtq = new SpBTW();
      if (!strcasecmp(mode,"TimeWarp")) evtq = new SpBTW();
      
      if (!strcasecmp(mode,"BTB")) evtq = new SpBTW();
      if (!strcasecmp(mode,"BREATHING_TIME_BUCKETS")) evtq = new SpBTW();
      if (!strcasecmp(mode,"BreathingTimeBuckets")) evtq = new SpBTW();
      
      if (!strcasecmp(mode,"BREATHING_TIME_WARP")) evtq = new SpBTW();
      if (!strcasecmp(mode,"BreathingTimeWarp")) evtq = new SpBTW();
      if (!strcasecmp(mode,"BTW")) evtq = new SpBTW();
      
   }else{
      
      evtq = new SpSequential();
      
   }

   if (evtq == NULL) {
      fprintf(stderr,"Error, synchronization protocol unknown\n");
      //comm_cleanup();
      exit(0);
   }
   
   //...... initialize some values for the event queue base class
   
   evtq->set_test_rollback(test_rollback);
   evtq->set_check_reptime(check_reptime);
   evtq->set_minstep(min_step);
   evtq->set_tend(tend);
   evtq->set_gvt(gvt);
   evtq->set_interactive(interactive);
   evtq->set_spin(spin);
   evtq->set_scaled_time(scaled_time);
   
   //...... define the statistics object
   
   parser->GoTo("statistics", NULL);
   SpStats stats(parser, n_nodes, node);
   evtq->set_statistics(&stats);
   
   //...... set up for scaled time
   
   scale_flag1 = 0;
   evtq->set_gvtmax(INFINITY);
   
   if (scaled_time) {
      if (node == 0) {
	 evtq->set_gvtmax(0.0);
	 scale_flag1 = 1;
      }
   }
   
   //...... loop until gvt exceeds the simulation time
   
   SpComm_BarrierSync();
   wallclock0 = SpWalltime();
   
   wallclock = 0.0;
   wallgvt0 = wallclock0;
   wallgvt = wallclock0;
   
   //...... perform the main processing for each cycle
   
   while (gvt < tend) {

//   if (!(((int)gvt)%10)){
//     evtq->RollBack();
//   }
      
      //...... start the cpu clock
      
      cpu0 = clock();
      
      //......  process events phase 1
      
      evtq->temp_process();
      
      //...... if scaled time on more than one processor, limit gvt
      
      if (scale_flag1) {
	 
	 wallgvt = SpWalltime();
	 if (gvt >= 0.0) {
	    evtq->set_gvtmax(gvt.GetTime() + scaler*(wallgvt-wallgvt0));
	 }
	 wallgvt0 = wallgvt;
	 
      }
      
      //...... find the next gvt
      
      evtq->find_gvt();
      old_gvt = gvt;
      gvt = evtq->get_gvt();
      
      //...... process events phase 2
      
      evtq->perm_process();
      
      //...... check for blocked messages
      
      evtq->get_blocked_messages();
      
      //...... stop the clock and do some statistics
      
      cpu1 = clock() - cpu0;
      wallclock = SpWalltime() - wallclock0;


      if (evtq->DoingCheckpoints()){
	int shouldCheckPoint =
	  (wallclock - WallForLastCheckpoint > WallTimeForCheckpoint);
	if (SpComm_GlobalSum(shouldCheckPoint) == n_nodes){
	  evtq->RollBack();
	  Checkpoint(WALL_TIME, wallclock);
	  WallForLastCheckpoint = wallclock;
	}
	
	if (gvt - SimTimeForLastCheckpoint > SimTimeForCheckpoint){
	  evtq->RollBack();
	  Checkpoint(SIM_TIME, gvt);
	  SimTimeForLastCheckpoint = gvt;
	}
      }
      
      if (gvt - SimTimeForLastCheckpoint > SimTimeForCheckpoint){
	evtq->RollBack();
	Checkpoint(SIM_TIME, gvt);
	SimTimeForLastCheckpoint = gvt;
      }
      
      stats.set_cpu(cpu1/1000000.0);
      stats.set_wall(wallclock);
      if (statistics) {
	 evtq->gather_statistics();
	 if (gvt >= tend) stats.set_ReportTime(0);
	 stats.print();
      }else{
	 stats.fast_print(gvt.GetTime());
      }
      
   }
   
   evtq->term(tend);
   
   delete evtq;

   SpComm_CleanupOther();
   
   return 0;
   
}

void SPEEDES::PlugInObjMan(char *name, ObjManConstructFunc conFunc)
{
   ObjectManagerContainer *curObjMan = new ObjectManagerContainer;
   curObjMan->name = strdup(name);
   curObjMan->conFunc = conFunc;
   objMan[cumNumObjMan] = curObjMan;
   cumNumObjMan++;
}

void SPEEDES::PlugInEvent(char *name, EventConstructFunc conFunc,
			  int eSize, int mSize, int numFree)
{
   EventContainer *curEvent = new EventContainer;
   curEvent->name = strdup(name);
   curEvent->conFunc   = conFunc;
   curEvent->eSize     = eSize;
   curEvent->mSize     = mSize;
   curEvent->numFree   = numFree;
   event[cumNumEvent] = curEvent;
   cumNumEvent++;
}

void SPEEDES::Print(ostream &out)
{
   int i, w=0, obmLen, evtLen;

   for (i=0; i<cumNumObjMan; i++)
   {
      obmLen = strlen(((ObjectManagerContainer *) objMan[i])->name);
      if (obmLen>w) w = obmLen;
   }
   for (i=0; i<cumNumEvent;  i++)
   {
      evtLen = strlen(((EventContainer *) event[i])->name);
      if (evtLen>w) w = evtLen;
   }
   
   out << "Object managers -----------------------"
       << "---------------------------------------"
       << endl;

   for (i=0; i<cumNumObjMan; i++)
   {
      out
	<< setw(3)
	<< i
	<< " "
	<< setw(w)
	<< ((ObjectManagerContainer *) objMan[i])->name
	<< endl;
   }

   out << "Events --------------------------------"
       << "---------------------------------------"
       << endl;

   for (i=0; i<cumNumEvent; i++)
   {
      out
	<< setw(3)
	<< i
	<< " "
	<< setw(w)
	<< ((EventContainer *) event[i])->name
	<< setw(8)
	<< ", esize="
	<< setw(4)
	<< ((EventContainer *) event[i])->eSize
	<< setw(8)
	<< ", msize="
	<< setw(4)
	<< ((EventContainer *) event[i])->mSize
	<< setw(8)
	<< ", nfree="
	<< setw(4)
	<< ((EventContainer *) event[i])->numFree
	<< endl;
   }
   out << "---------------------------------------"
       << "---------------------------------------"
       << endl;
   out.flush();
}

int ExecuteSpeedes(int argc, char **argv)
{
   return Speedes->Execute(argc, argv);
}


static inline int WriteBuffer(char* fname, char * buffer)
{
  int rc = 0;
  char* fName =new char[strlen(fname) + strlen(DirectoryForCheckpointFiles)+2];
  sprintf(fName, "%s%s",DirectoryForCheckpointFiles, fname); 
  FILE* outFile = fopen(fName, "w");
  if (!outFile){
    cerr<<"Could not open the checkpoint file "<<fName <<"for writing"
	<<endl;
    rc = 1;
  }
  if (rc == 0){
    int nBytes = fwrite (buffer,
			 sizeof(char),
			 PO_SIZEOF_BUFFER(buffer),
			 outFile);
    fclose(outFile);
    if (nBytes != PO_SIZEOF_BUFFER(buffer)){
      cout<<"Could only write "<<nBytes<<" of "<<PO_SIZEOF_BUFFER(buffer)
	  <<" for the file "<<fName<<endl;
      rc = 1;
    }
  }
  delete []fName;
  return rc;
}

void SPEEDES::Checkpoint(CheckpointReason r,
			 double time,
			 char* CheckpointFname)
{
  char messageFname[256];
  char objectFname[256];
  SpStopWatch watch;

  if (CheckpointFname == NULL){
    cerr<<"You must provide a file name and NULL is _NOT_ a file name"<<endl;
    return;
  }
  switch (r)
    {
    case WALL_TIME:
      sprintf(objectFname, "OBJECTS_%s_WALL_TIME_%d_NODE_%d",
	      CheckpointFname, (int)time, SpComm_GetNodeId());
      sprintf(messageFname, "MESSAGES_%s_WALL_TIME_%d_NODE_%d",
	      CheckpointFname, (int)time, SpComm_GetNodeId());
      break;
    case SIM_TIME:
      sprintf(objectFname, "OBJECTS_%s_SIM_TIME_%d_NODE_%d",
	      CheckpointFname, (int)time, SpComm_GetNodeId());
      sprintf(messageFname, "MESSAGES_%s_SIM_TIME_%d_NODE_%d",
	      CheckpointFname, (int)time, SpComm_GetNodeId());
      break;
    case USER_REQUESTED:
      sprintf(objectFname, "OBJECTS_%s_USER_REQUESTED_%d_NODE_%d",
	      CheckpointFname, (int)time, SpComm_GetNodeId());
      sprintf(messageFname, "MESSAGES_%s_USER_REQUESTED_%d_NODE_%d",
	      CheckpointFname, (int)time, SpComm_GetNodeId());
      break;
    default:
      cerr<<"Error in SPEEDES::Checkpoint(), CheckpointReason is "<<r<<endl;
      return;
    }
  watch.StartTimer();
  char* objectBuffer =
    DefaultDataBase->CreateBuffer(SpPoDataBase::NO_COMPRESSION);
  watch.StopTimer();
  cout<<"Time to create object buffer is "<<watch.get_CumulativeTime()<<endl;
  watch.Reset();
  watch.StartTimer();
  char* messageBuffer =
    MessageDataBase->CreateBuffer(SpPoDataBase::NO_COMPRESSION);
  watch.StopTimer();
  cout<<"Time to create message buffer is "<<watch.get_CumulativeTime()<<endl;
  
  watch.Reset();
  watch.StartTimer();
  if (WriteBuffer(objectFname, objectBuffer)){
    delete [] objectBuffer;
    return;
  }
  delete [] objectBuffer;
  watch.StopTimer();
  cout<<"Time to write object buffer is "<<watch.get_CumulativeTime()<<endl;
  watch.Reset();
  watch.StartTimer();
  WriteBuffer(messageFname, messageBuffer);
  watch.StopTimer();
  cout<<"Time to write message buffer is "<<watch.get_CumulativeTime()<<endl;
  cout.flush();
  delete [] messageBuffer;
}

int GetSimObjType(char *name) { return CurrentEvent->object_type(name); }
int GetEventType (char *name) { return CurrentEvent->event_type (name); }



void CreateDefaultDataBase(){
  if (DefaultDataBase == NULL){
    DefaultDataBase = new SpPoDataBase;
    cout<<" Database created!!!"<<endl;
    SpPoDataBase *jj = new SpPoDataBase;
    delete jj;
  }
}

