// SpExecute.H

#ifndef SPEEDES_OBJ
#define SPEEDES_OBJ

// BEGIN defining the SPEEDES class

#include "SpDynPtrArray.H"
#include "SpMsg.H"
#include "SpStdlib.H"
#include "SpIostream.H"
#include "SpString.H"
#include "SpClusterMgr.H"
#include "SpWarningPatch.H"

typedef void * (*ObjManConstructFunc)();
typedef void * (*EventConstructFunc)(int n);

class ObjectManagerContainer
{
 public:
   char *name;
   ObjManConstructFunc conFunc;
 public:
   ObjectManagerContainer(char *n = NULL, ObjManConstructFunc omcf = NULL):
      name(n), conFunc(omcf) {}
   ObjectManagerContainer *SetName(char *n)
      { if (name) free (name); name = strdup(n); return this; }
   char *GetName() { return name; }
   ObjectManagerContainer *SetConFunc(ObjManConstructFunc omcf)
      { conFunc = omcf; return this; }
   ObjManConstructFunc GetConFunc() { return conFunc; }
};

struct EventContainer
{
   char *name;
   EventConstructFunc conFunc;
   int eSize;
   int mSize;
   int numFree;
};

class SPEEDES
{
  enum CheckpointReason {
    WALL_TIME,
    SIM_TIME,
    USER_REQUESTED
  };

   SpDynPtrArray objMan;
   int cumNumObjMan;

   SpDynPtrArray event;
   int cumNumEvent;

   int ExecuteAlreadyCalled;
   
 public:
   SPEEDES();
   void PlugInObjMan(char *name, ObjManConstructFunc conFunc);
   void PlugInEvent(char *name, EventConstructFunc conFunc,
		    int eSize, int mSize, int numFree);
   int GetNumObjManTypes() { return cumNumObjMan; }
   char *GetObjManName(int id)
      { return ((ObjectManagerContainer *)objMan[id])->name; }
   ObjManConstructFunc GetObjManConFunc(int id)
      { return ((ObjectManagerContainer *) objMan[id])->conFunc; }
   int GetNumEventTypes() { return cumNumEvent; }
   char *GetEventName(int id) { return ((EventContainer *)event[id])->name; }
   EventConstructFunc GetEventConFunc(int id) { return ((EventContainer *)
						       event[id])->conFunc; }
   int GetEventSize(int id)   { return ((EventContainer *)event[id])->eSize; }
   int GetMessageSize(int id) { return ((EventContainer *)event[id])->mSize; }
   int GetNumFree(int id)     { return ((EventContainer *)event[id])->numFree; }

   void Print(ostream &out = cout);   
   int Execute(int argc = 0, char **argv = NULL);   

  void Checkpoint(CheckpointReason, double time,
		  char* CheckpointFname="CheckPoint");

};

extern SPEEDES *Speedes;

int ExecuteSpeedes(int argc = 0, char **argv = NULL);

////////////////////////////////////////
/////// begin internal macros: //////////
////////////////////////////////////////

int GetSimObjType(char *name);
int GetEventType (char *name);

#define _DEFINE_TYPE_INT(eventOrSimObj, name)                      \
class name##_TYPE                                                  \
{                                                                  \
 private:                                                          \
   int name##_INT;                                                 \
 public:                                                           \
   name##_TYPE(): name##_INT(-1){}                                 \
   operator int()                                                  \
   {                                                               \
      if (name##_INT == -1)                                        \
	 name##_INT = Get##eventOrSimObj##Type(#name);             \
      return name##_INT;                                           \
   }                                                               \
};                                                                 \
static name##_TYPE name;                                           \
WARNING_REMOVAL(name)

#define _PO_DEFINE_BEGIN(name, class, arg, array)               \
   static inline void *new_##name(arg) { return PO_NEW_##class(); }
      
#define _DEFINE_BEGIN(name, class, arg, array)               \
   static inline void *new_##name(arg) { return new class##array } \
   WARNING_REMOVAL(new_##name)

#define _DEFINE_EVENT(name, eclass, mclass, nfree)      \
   _DEFINE_BEGIN(name, eclass, int n, [n];);            \
   static inline int name##_esize() { return sizeof(eclass); } \
   static inline int name##_msize() { return sizeof(mclass); } \
   static inline int name##_nfree() { return nfree; }          \
   WARNING_REMOVAL(name##_esize);                       \
   WARNING_REMOVAL(name##_msize);                       \
   WARNING_REMOVAL(name##_nfree);                       \

void CreateDefaultDataBase();

#define _PLUG_IN_BEGIN(name) \
   CreateDefaultDataBase(); \
   if (!Speedes) Speedes = new SPEEDES; \

// -----------------------------------------------------------------------
// -----------------------------------------------------------------------
// begin user macros: DEFINEs and DECLAREs
// (must have at least one DEFINE when using a corresponding DECLARE):
// -----------------------------------------------------------------------
// -----------------------------------------------------------------------

/******************************************************************************
* Macro: DEFINE_OBJECT_MANAGER
* Summary: Specifies an object manager as a prerequisite to calling
*          PLUG_IN_OBJECT_MANAGER
* Description: When using the SPEEDES plug-in system, each object manager
*              and event must be defined and plugged in to SPEEDES. This
*              macro is used to define the name (string name and global
*              integer name) and the class of object managers, presumed
*              to be linked in by the make system. This define may be included
*              multiple times in different files, but must be included in the
*              file that calls the corresponding PLUG_IN_OBJECT_MANAGER macro.
******************************************************************************/
#define DEFINE_OBJECT_MANAGER(name, class) \
DEFINE_OBJECT_MANAGER_CLUSTER(name, class, SPEEDES)

#define DEFINE_OBJECT_MANAGER_CLUSTER(name, class, cluster) \
   _DEFINE_BEGIN(name, class, void, ;)     \
   _DEFINE_TYPE_INT(SimObj, name);	   \
   static char *name##ClusterName=#cluster; \
   WARNING_REMOVAL(name##ClusterName)

#define PO_DEFINE_OBJECT_MANAGER(name, class) \
   _PO_DEFINE_BEGIN(name, class, void, ;)     \
   _DEFINE_TYPE_INT(SimObj, name);            \
   static char *name##ClusterName="SPEEDES";  \
   WARNING_REMOVAL(name##ClusterName)

#define DEFINE_OBJ_MGR(name, class) DEFINE_OBJECT_MANAGER(name, class)
#define PO_DEFINE_OBJ_MGR(name, class) PO_DEFINE_OBJECT_MANAGER(name, class)

/******************************************************************************
* Macro: DEFINE_EVENT
* Summary: Specifies an event as a prerequisite to calling PLUG_IN_EVENT
* Description: When using the SPEEDES plug-in system, each object manager
*              and event must be defined and plugged in to SPEEDES. This
*              macro is used to define the name (string name and global
*              integer name) and the class of the event, as well as the class
*              of the associated message (use SpMsg is there is no
*              associated message class) and the number of events and messages
*              to be created at a time by the freelists. The event and message
*              classes are assumed to be linked in by the make system. This
*              define may be included multiple times in different files, but
*              must be included in the file that calls the corresponding
*              PLUG_IN_EVENT macro.
******************************************************************************/
#define DEFINE_EVENT(name, eclass, mclass, nfree) \
   _DEFINE_EVENT(name, eclass, mclass, nfree)     \
   _DEFINE_TYPE_INT(Event, name);

// -----------------------------------------------------------------------
// PLUG_INs:
// -----------------------------------------------------------------------

/******************************************************************************
* Macro: PLUG_IN_OBJECT_MANAGER
* Summary: Tells SPEEDES to add an object manager to the application
*          at run-time, specified by object manager name
* Description: When using the SPEEDES plug-in system, each object manager
*              and event must be defined and plugged in to SPEEDES. This
*              macro is used to plug-in an object manager, presumed to be
*              DEFINEd earlier in the same file. This macro is typically
*              called in the user's main() routine, along with other calls
*              to PLUG_IN_OBJECT_MANAGER and PLUG_IN_EVENT, followed by a
*              call to ExecuteSpeedes(argc, argv);.
******************************************************************************/
#define PLUG_IN_OBJECT_MANAGER(name)                  \
   CreateDefaultDataBase() ;\
   PlugInCluster(name##ClusterName); 		      \
   _PLUG_IN_BEGIN(name)                               \
   Speedes->PlugInObjMan(#name, new_##name);

/******************************************************************************
* Macro: PLUG_IN_OBJECT_MANAGER
* Summary: Tells SPEEDES to add an object manager to the application
*          at run-time, specified by object manager name
* Description: When using the SPEEDES plug-in system, each object manager
*              and event must be defined and plugged in to SPEEDES. This
*              macro is used to plug-in an object manager, presumed to be
*              DEFINEd earlier in the same file. This macro is typically
*              called in the user's main() routine, along with other calls
*              to PLUG_IN_OBJECT_MANAGER and PLUG_IN_EVENT, followed by a
*              call to ExecuteSpeedes(argc, argv);.
******************************************************************************/
#define PLUG_IN_AUTOMATIC_OBJECT_MANAGER(name)        \
   CreateDefaultDataBase(); \
   PlugInCluster(name##ClusterName); 		      \
   _PLUG_IN_BEGIN(name)                               \
   Speedes->PlugInObjMan(#name, new_##name);          \
   RB_REGISTER_CLASS_O_##name();

/******************************************************************************
* Macro: PLUG_IN_EVENT
* Summary: Tells SPEEDES to add an event to the application at run-time,
*          specified by object manager name
* Description: When using the SPEEDES plug-in system, each object manager
*              and event must be defined and plugged in to SPEEDES. This
*              macro is used to plug-in an event, presumed to be
*              DEFINEd earlier in the same file. This macro is typically
*              called in the user's main() routine, along with other calls
*              to PLUG_IN_OBJECT_MANAGER and PLUG_IN_EVENT, followed by a
*              call to ExecuteSpeedes(argc, argv);.
******************************************************************************/
#define PLUG_IN_EVENT(name)     \
   _PLUG_IN_BEGIN(name)         \
   Speedes->PlugInEvent(#name,  \
               new_##name,      \
               name##_esize(),  \
               name##_msize(),  \
               name##_nfree());

/******************************************************************************
* Macro: PLUG_IN_EVENT
* Summary: Tells SPEEDES to add an event to the application at run-time,
*          specified by object manager name
* Description: When using the SPEEDES plug-in system, each object manager
*              and event must be defined and plugged in to SPEEDES. This
*              macro is used to plug-in an event, presumed to be
*              DEFINEd earlier in the same file. This macro is typically
*              called in the user's main() routine, along with other calls
*              to PLUG_IN_OBJECT_MANAGER and PLUG_IN_EVENT, followed by a
*              call to ExecuteSpeedes(argc, argv);.
******************************************************************************/
#define PLUG_IN_EVENT_METHOD_PROCESS(className, name)     \
   _PLUG_IN_BEGIN(name)         \
   Speedes->PlugInEvent(#name,  \
               new_##name,      \
               name##_esize(),  \
               name##_msize(),  \
               name##_nfree()); \
   RB_REGISTER_CLASS_LV_##className##_##name();

// temporary code for obsolete process model interface:

#define DEFINE_EVENT_NO_GLOBAL(name, eclass, mclass, nfree) \
   void *new_##name(int n) { return new eclass[n]; }        \
   static inline int name##_esize() { return sizeof(eclass); }     \
   static inline int name##_msize() { return sizeof(mclass); }     \
   static inline int name##_nfree() { return nfree; }
#define PLUG_IN_EVENT_NO_GLOBAL(name) \
   CreateDefaultDataBase(); \
   if (!Speedes)                      \
     Speedes = new SPEEDES;           \
   Speedes->PlugInEvent(#name,        \
               new_##name,            \
               name##_esize(),        \
               name##_msize(),        \
               name##_nfree());

// Event Methods - use these macros to DEFINE and DECLARE events as methods on
//                 SpSimObjs

// begin event method internal macros:

#define _PAIR(id, str1, str2) str2 str1##id
#define  _0_COMMA_PAIRS(v)
#define  _1_COMMA_PAIRS(v, t1) \
     _PAIR(1, v, t1)
#define  _2_COMMA_PAIRS(v, t1, t2) \
     _PAIR(1, v, t1), _PAIR(2, v, t2)
#define  _3_COMMA_PAIRS(v, t1, t2, t3) \
     _2_COMMA_PAIRS(v, t1, t2), _PAIR(3, v, t3)
#define  _4_COMMA_PAIRS(v, t1, t2, t3, t4) \
     _3_COMMA_PAIRS(v, t1, t2, t3), _PAIR(4, v, t4)
#define  _5_COMMA_PAIRS(v, t1, t2, t3, t4, t5) \
     _4_COMMA_PAIRS(v, t1, t2, t3, t4), _PAIR(5, v, t5)
#define  _6_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6) \
     _5_COMMA_PAIRS(v, t1, t2, t3, t4, t5), _PAIR(6, v, t6)
#define  _7_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7) \
     _6_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6), _PAIR(7, v, t7)
#define  _8_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8) \
     _7_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7), _PAIR(8, v, t8)
#define  _9_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9) \
     _8_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8), _PAIR(9, v, t9)
#define  _10_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10) \
     _9_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9),         \
     _PAIR(10, v, t10)
#define  _11_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11)   \
     _10_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10),           \
     _PAIR(11, v, t11)
#define  _12_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,    \
     t12) _11_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11), \
     _PAIR(12, v, t12)
#define  _13_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,    \
     t12, t13) _12_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10,   \
	t11, t12),                                                       \
     _PAIR(13, v, t13)
#define  _14_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,    \
     t12, t13, t14) _13_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9,   \
	t10, t11, t12, t13),                                             \
    _PAIR(14, v, t14)
#define  _15_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,    \
     t12, t13, t14, t15) _14_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8,  \
	t9, t10, t11, t12, t13, t14),                                    \
     _PAIR(15, v, t15)
#define  _16_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,    \
     t12, t13, t14, t15, t16) _15_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, \
	t8, t9, t10, t11, t12, t13, t14, t15),                           \
     _PAIR(16, v, t16)
#define  _17_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,    \
     t12, t13, t14, t15, t16, t17) _16_COMMA_PAIRS(v, t1, t2, t3, t4, t5,    \
	t6, t7, t8, t9, t10, t11, t12, t13, t14, t15, t16),              \
     _PAIR(17, v, t17)
#define  _18_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,    \
     t12, t13, t14, t15, t16, t17, t18) _17_COMMA_PAIRS(v, t1, t2, t3, t4,   \
	t5, t6, t7, t8, t9, t10, t11, t12, t13, t14, t15, t16, t17),     \
     _PAIR(18, v, t18)
#define  _19_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,    \
     t12, t13, t14, t15, t16, t17, t18, t19) _18_COMMA_PAIRS(v, t1, t2, t3,  \
	t4, t5, t6, t7, t8, t9, t10, t11, t12, t13, t14, t15, t16, t17,   \
	t18),                                                            \
     _PAIR(19, v, t19)
#define  _20_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,    \
     t12, t13, t14, t15, t16, t17, t18, t19, t20) _19_COMMA_PAIRS(v, t1, t2, \
	t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13, t14, t15, t16,    \
	t17, t18, t19),                                                  \
     _PAIR(20, v, t20)

#define  _0_SEMIC_PAIRS(v)
#define  _1_SEMIC_PAIRS(v, t1) \
     _PAIR(1, v, t1);
#define  _2_SEMIC_PAIRS(v, t1, t2) \
     _PAIR(1, v, t1); _PAIR(2, v, t2);
#define  _3_SEMIC_PAIRS(v, t1, t2, t3) \
     _2_SEMIC_PAIRS(v, t1, t2) _PAIR(3, v, t3);
#define  _4_SEMIC_PAIRS(v, t1, t2, t3, t4) \
     _3_SEMIC_PAIRS(v, t1, t2, t3) _PAIR(4, v, t4);
#define  _5_SEMIC_PAIRS(v, t1, t2, t3, t4, t5) \
     _4_SEMIC_PAIRS(v, t1, t2, t3, t4) _PAIR(5, v, t5);
#define  _6_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6) \
     _5_SEMIC_PAIRS(v, t1, t2, t3, t4, t5) _PAIR(6, v, t6);
#define  _7_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7) \
     _6_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6) _PAIR(7, v, t7);
#define  _8_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8) \
     _7_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7) _PAIR(8, v, t8);
#define  _9_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9) \
     _8_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8) _PAIR(9, v, t9);
#define  _10_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10) \
     _9_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9)         \
     _PAIR(10, v, t10);
#define  _11_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11)  \
     _10_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10)           \
     _PAIR(11, v, t11);
#define  _12_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,  \
     t12) _11_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11) \
     _PAIR(12, v, t12);
#define  _13_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,    \
     t12, t13) _12_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10,   \
	t11, t12)                                                       \
     _PAIR(13, v, t13);
#define  _14_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,    \
     t12, t13, t14) _13_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9,   \
	t10, t11, t12, t13)                                             \
    _PAIR(14, v, t14);
#define  _15_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,    \
     t12, t13, t14, t15) _14_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8,  \
	t9, t10, t11, t12, t13, t14)                                    \
     _PAIR(15, v, t15);
#define  _16_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,    \
     t12, t13, t14, t15, t16) _15_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, \
	t8, t9, t10, t11, t12, t13, t14, t15)                           \
     _PAIR(16, v, t16);
#define  _17_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,    \
     t12, t13, t14, t15, t16, t17) _16_SEMIC_PAIRS(v, t1, t2, t3, t4, t5,    \
	t6, t7, t8, t9, t10, t11, t12, t13, t14, t15, t16)              \
     _PAIR(17, v, t17);
#define  _18_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,    \
     t12, t13, t14, t15, t16, t17, t18) _17_SEMIC_PAIRS(v, t1, t2, t3, t4,   \
	t5, t6, t7, t8, t9, t10, t11, t12, t13, t14, t15, t16, t17)     \
     _PAIR(18, v, t18);
#define  _19_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,    \
     t12, t13, t14, t15, t16, t17, t18, t19) _18_SEMIC_PAIRS(v, t1, t2, t3,  \
	t4, t5, t6, t7, t8, t9, t10, t11, t12, t13, t14, t15, t16, t17,   \
	t18)                                                            \
     _PAIR(19, v, t19);
#define  _20_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,    \
     t12, t13, t14, t15, t16, t17, t18, t19, t20) _19_SEMIC_PAIRS(v, t1, t2, \
	t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13, t14, t15, t16,    \
	t17, t18, t19)                                                  \
     _PAIR(20, v, t20);

#define _INIT_LIST(id) v##id##(a##id##)
#define  _0_INIT_LIST
#define  _1_INIT_LIST    _INIT_LIST(1)
#define  _2_INIT_LIST  _1_INIT_LIST, _INIT_LIST(2)
#define  _3_INIT_LIST  _2_INIT_LIST, _INIT_LIST(3)
#define  _4_INIT_LIST  _3_INIT_LIST, _INIT_LIST(4)
#define  _5_INIT_LIST  _4_INIT_LIST, _INIT_LIST(5)
#define  _6_INIT_LIST  _5_INIT_LIST, _INIT_LIST(6)
#define  _7_INIT_LIST  _6_INIT_LIST, _INIT_LIST(7)
#define  _8_INIT_LIST  _7_INIT_LIST, _INIT_LIST(8)
#define  _9_INIT_LIST  _8_INIT_LIST, _INIT_LIST(9)
#define _10_INIT_LIST  _9_INIT_LIST, _INIT_LIST(10)
#define _11_INIT_LIST _10_INIT_LIST, _INIT_LIST(11)
#define _12_INIT_LIST _11_INIT_LIST, _INIT_LIST(12)
#define _13_INIT_LIST _12_INIT_LIST, _INIT_LIST(13)
#define _14_INIT_LIST _13_INIT_LIST, _INIT_LIST(14)
#define _15_INIT_LIST _14_INIT_LIST, _INIT_LIST(15)
#define _16_INIT_LIST _15_INIT_LIST, _INIT_LIST(16)
#define _17_INIT_LIST _16_INIT_LIST, _INIT_LIST(17)
#define _18_INIT_LIST _17_INIT_LIST, _INIT_LIST(18)
#define _19_INIT_LIST _18_INIT_LIST, _INIT_LIST(19)
#define _20_INIT_LIST _19_INIT_LIST, _INIT_LIST(20)

#define _VAR_LIST(id) v##id
#define  _0_VAR_LIST
#define  _1_VAR_LIST    _VAR_LIST(1)
#define  _2_VAR_LIST  _1_VAR_LIST, _VAR_LIST(2)
#define  _3_VAR_LIST  _2_VAR_LIST, _VAR_LIST(3)
#define  _4_VAR_LIST  _3_VAR_LIST, _VAR_LIST(4)
#define  _5_VAR_LIST  _4_VAR_LIST, _VAR_LIST(5)
#define  _6_VAR_LIST  _5_VAR_LIST, _VAR_LIST(6)
#define  _7_VAR_LIST  _6_VAR_LIST, _VAR_LIST(7)
#define  _8_VAR_LIST  _7_VAR_LIST, _VAR_LIST(8)
#define  _9_VAR_LIST  _8_VAR_LIST, _VAR_LIST(9)
#define _10_VAR_LIST  _9_VAR_LIST, _VAR_LIST(10)
#define _11_VAR_LIST _10_VAR_LIST, _VAR_LIST(11)
#define _12_VAR_LIST _11_VAR_LIST, _VAR_LIST(12)
#define _13_VAR_LIST _12_VAR_LIST, _VAR_LIST(13)
#define _14_VAR_LIST _13_VAR_LIST, _VAR_LIST(14)
#define _15_VAR_LIST _14_VAR_LIST, _VAR_LIST(15)
#define _16_VAR_LIST _15_VAR_LIST, _VAR_LIST(16)
#define _17_VAR_LIST _16_VAR_LIST, _VAR_LIST(17)
#define _18_VAR_LIST _17_VAR_LIST, _VAR_LIST(18)
#define _19_VAR_LIST _18_VAR_LIST, _VAR_LIST(19)
#define _20_VAR_LIST _19_VAR_LIST, _VAR_LIST(20)

#define _MSG_ARG_LIST(name, id) ((M_##name##_ARG *) MESSAGE)->v##id
#define  _0_MSG_ARG_LIST(name)
#define  _1_MSG_ARG_LIST(name)    _MSG_ARG_LIST(name, 1)
#define  _2_MSG_ARG_LIST(name)  _1_MSG_ARG_LIST(name), _MSG_ARG_LIST(name, 2)
#define  _3_MSG_ARG_LIST(name)  _2_MSG_ARG_LIST(name), _MSG_ARG_LIST(name, 3)
#define  _4_MSG_ARG_LIST(name)  _3_MSG_ARG_LIST(name), _MSG_ARG_LIST(name, 4)
#define  _5_MSG_ARG_LIST(name)  _4_MSG_ARG_LIST(name), _MSG_ARG_LIST(name, 5)
#define  _6_MSG_ARG_LIST(name)  _5_MSG_ARG_LIST(name), _MSG_ARG_LIST(name, 6)
#define  _7_MSG_ARG_LIST(name)  _6_MSG_ARG_LIST(name), _MSG_ARG_LIST(name, 7)
#define  _8_MSG_ARG_LIST(name)  _7_MSG_ARG_LIST(name), _MSG_ARG_LIST(name, 8)
#define  _9_MSG_ARG_LIST(name)  _8_MSG_ARG_LIST(name), _MSG_ARG_LIST(name, 9)
#define _10_MSG_ARG_LIST(name)  _9_MSG_ARG_LIST(name), _MSG_ARG_LIST(name, 10)
#define _11_MSG_ARG_LIST(name) _10_MSG_ARG_LIST(name), _MSG_ARG_LIST(name, 11)
#define _12_MSG_ARG_LIST(name) _11_MSG_ARG_LIST(name), _MSG_ARG_LIST(name, 12)
#define _13_MSG_ARG_LIST(name) _12_MSG_ARG_LIST(name), _MSG_ARG_LIST(name, 13)
#define _14_MSG_ARG_LIST(name) _13_MSG_ARG_LIST(name), _MSG_ARG_LIST(name, 14)
#define _15_MSG_ARG_LIST(name) _14_MSG_ARG_LIST(name), _MSG_ARG_LIST(name, 15)
#define _16_MSG_ARG_LIST(name) _15_MSG_ARG_LIST(name), _MSG_ARG_LIST(name, 16)
#define _17_MSG_ARG_LIST(name) _16_MSG_ARG_LIST(name), _MSG_ARG_LIST(name, 17)
#define _18_MSG_ARG_LIST(name) _17_MSG_ARG_LIST(name), _MSG_ARG_LIST(name, 18)
#define _19_MSG_ARG_LIST(name) _18_MSG_ARG_LIST(name), _MSG_ARG_LIST(name, 19)
#define _20_MSG_ARG_LIST(name) _19_MSG_ARG_LIST(name), _MSG_ARG_LIST(name, 20)

#define _ARG_LIST(args, id) args->v##id
#define  _0_ARG_LIST(args)
#define  _1_ARG_LIST(args)    _ARG_LIST(args, 1)
#define  _2_ARG_LIST(args)  _1_ARG_LIST(args), _ARG_LIST(args, 2)
#define  _3_ARG_LIST(args)  _2_ARG_LIST(args), _ARG_LIST(args, 3)
#define  _4_ARG_LIST(args)  _3_ARG_LIST(args), _ARG_LIST(args, 4)
#define  _5_ARG_LIST(args)  _4_ARG_LIST(args), _ARG_LIST(args, 5)
#define  _6_ARG_LIST(args)  _5_ARG_LIST(args), _ARG_LIST(args, 6)
#define  _7_ARG_LIST(args)  _6_ARG_LIST(args), _ARG_LIST(args, 7)
#define  _8_ARG_LIST(args)  _7_ARG_LIST(args), _ARG_LIST(args, 8)
#define  _9_ARG_LIST(args)  _8_ARG_LIST(args), _ARG_LIST(args, 9)
#define _10_ARG_LIST(args)  _9_ARG_LIST(args), _ARG_LIST(args, 10)
#define _11_ARG_LIST(args) _10_ARG_LIST(args), _ARG_LIST(args, 11)
#define _12_ARG_LIST(args) _11_ARG_LIST(args), _ARG_LIST(args, 12)
#define _13_ARG_LIST(args) _12_ARG_LIST(args), _ARG_LIST(args, 13)
#define _14_ARG_LIST(args) _13_ARG_LIST(args), _ARG_LIST(args, 14)
#define _15_ARG_LIST(args) _14_ARG_LIST(args), _ARG_LIST(args, 15)
#define _16_ARG_LIST(args) _15_ARG_LIST(args), _ARG_LIST(args, 16)
#define _17_ARG_LIST(args) _16_ARG_LIST(args), _ARG_LIST(args, 17)
#define _18_ARG_LIST(args) _17_ARG_LIST(args), _ARG_LIST(args, 18)
#define _19_ARG_LIST(args) _18_ARG_LIST(args), _ARG_LIST(args, 19)
#define _20_ARG_LIST(args) _19_ARG_LIST(args), _ARG_LIST(args, 20)

/*********************************************************************
 * These macros produce classes that encapsulate arguments. For example,
 * _2_ARG_CLASS(foo, int, double) would generate:
 *
 * class foo_ARG
 * {
 *  public:
 *    int    v1;
 *    double v2;
 *    foo_ARG() {};
 *    foo_ARG(int a1,  double a1): v1(a1), v2(a2){};
 * };
 *********************************************************************/
#define _DECLARE_ARG_CLASS(name, numArgs, decls, constructParams, \
                           sep, parent)                           \
class name##_ARG sep parent                                       \
{                                                                 \
 public:                                                          \
   decls                                                          \
   name##_ARG() {};                                               \
   name##_ARG(constructParams): _##numArgs##_INIT_LIST {}         \
};

#define _DECLARE_0_ARG_CLASS(name, sep, parent) \
  class name##_ARG sep parent {};
#define _DECLARE_1_ARG_CLASS(name, t1, sep, parent) \
  _DECLARE_ARG_CLASS(name, 1,          \
  _1_SEMIC_PAIRS(v, t1),               \
  _1_COMMA_PAIRS(a, t1), sep, parent)
#define _DECLARE_2_ARG_CLASS(name, t1, t2, sep, parent) \
  _DECLARE_ARG_CLASS(name, 2,              \
  _2_SEMIC_PAIRS(v, t1, t2),               \
  _2_COMMA_PAIRS(a, t1, t2), sep, parent)
#define _DECLARE_3_ARG_CLASS(name, t1, t2, t3, sep, parent) \
  _DECLARE_ARG_CLASS(name, 3,                  \
  _3_SEMIC_PAIRS(v, t1, t2, t3),               \
  _3_COMMA_PAIRS(a, t1, t2, t3), sep, parent)
#define _DECLARE_4_ARG_CLASS(name, t1, t2, t3, t4, sep, parent) \
  _DECLARE_ARG_CLASS(name, 4,                      \
  _4_SEMIC_PAIRS(v, t1, t2, t3, t4),               \
  _4_COMMA_PAIRS(a, t1, t2, t3, t4), sep, parent)
#define _DECLARE_5_ARG_CLASS(name, t1, t2, t3, t4, t5, sep, parent) \
  _DECLARE_ARG_CLASS(name, 5,                          \
  _5_SEMIC_PAIRS(v, t1, t2, t3, t4, t5),               \
  _5_COMMA_PAIRS(a, t1, t2, t3, t4, t5), sep, parent)
#define _DECLARE_6_ARG_CLASS(name, t1, t2, t3, t4, t5, t6, sep, parent) \
  _DECLARE_ARG_CLASS(name, 6,                              \
  _6_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6),               \
  _6_COMMA_PAIRS(a, t1, t2, t3, t4, t5, t6), sep, parent)
#define _DECLARE_7_ARG_CLASS(name, t1, t2, t3, t4, t5, t6, t7, sep, parent) \
  _DECLARE_ARG_CLASS(name, 7,                                  \
  _7_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7),               \
  _7_COMMA_PAIRS(a, t1, t2, t3, t4, t5, t6, t7), sep, parent)
#define _DECLARE_8_ARG_CLASS(name, t1, t2, t3, t4, t5, t6, t7, t8, sep, parent) \
  _DECLARE_ARG_CLASS(name, 8,                                      \
  _8_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8),               \
  _8_COMMA_PAIRS(a, t1, t2, t3, t4, t5, t6, t7, t8), sep, parent)
#define _DECLARE_9_ARG_CLASS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, sep, \
     parent)                                                                \
  _DECLARE_ARG_CLASS(name, 9,                                               \
  _9_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9),                    \
  _9_COMMA_PAIRS(a, t1, t2, t3, t4, t5, t6, t7, t8, t9), sep, parent)
#define _DECLARE_10_ARG_CLASS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, \
     sep, parent)                                                            \
  _DECLARE_ARG_CLASS(name, 10,                                               \
  _10_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10),               \
  _10_COMMA_PAIRS(a, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10), sep, parent)
#define _DECLARE_11_ARG_CLASS(name, t1, t2, t3, t4, t5, t6, t7, t8, \
     t9, t10, t11, sep, parent)                                     \
  _DECLARE_ARG_CLASS(name, 11,                                      \
  _11_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11), \
  _11_COMMA_PAIRS(a, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11), sep, parent)
#define _DECLARE_12_ARG_CLASS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9,  \
     t10, t11, t12, sep, parent)                                         \
  _DECLARE_ARG_CLASS(name, 12,                                           \
  _12_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12), \
  _12_COMMA_PAIRS(a, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12), \
     sep, parent)
#define _DECLARE_13_ARG_CLASS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10,  \
     t11, t12, t13, sep, parent)                                              \
  _DECLARE_ARG_CLASS(name, 13,                                                \
  _13_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13), \
  _13_COMMA_PAIRS(a, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13), \
     sep, parent)
#define _DECLARE_14_ARG_CLASS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10,  \
     t11, t12, t13, t14, sep, parent)                                         \
  _DECLARE_ARG_CLASS(name, 14,                                                \
  _14_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
    t14),                                                                     \
  _14_COMMA_PAIRS(a, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
    t14), sep, parent)
#define _DECLARE_15_ARG_CLASS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10,  \
     t11, t12, t13, t14, t15, sep, parent)                                    \
  _DECLARE_ARG_CLASS(name, 15,                                                \
  _15_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
    t14, t15),                                                                \
  _15_COMMA_PAIRS(a, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
    t14, t15), sep, parent)
#define _DECLARE_16_ARG_CLASS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10,  \
     t11, t12, t13, t14, t15, t16, sep, parent)                               \
  _DECLARE_ARG_CLASS(name, 16,                                                \
  _16_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
    t14, t15, t16),                                                           \
  _16_COMMA_PAIRS(a, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
    t14, t15, t16), sep, parent)
#define _DECLARE_17_ARG_CLASS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10,  \
     t11, t12, t13, t14, t15, t16, t17, sep, parent)                          \
  _DECLARE_ARG_CLASS(name, 17,                                                \
  _17_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
    t14, t15, t16, t17),                                                      \
  _17_COMMA_PAIRS(a, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
    t14, t15, t16, t17), sep, parent)
#define _DECLARE_18_ARG_CLASS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10,  \
     t11, t12, t13, t14, t15, t16, t17, t18, sep, parent)                     \
  _DECLARE_ARG_CLASS(name, 18,                                                \
  _18_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
    t14, t15, t16, t17, t18),                                                 \
  _18_COMMA_PAIRS(a, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
    t14, t15, t16, t17, t18), sep, parent)
#define _DECLARE_19_ARG_CLASS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10,  \
     t11, t12, t13, t14, t15, t16, t17, t18, t19, sep, parent)                \
  _DECLARE_ARG_CLASS(name, 19,                                                \
  _19_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
    t14, t15, t16, t17, t18, t19),                                            \
  _19_COMMA_PAIRS(a, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
    t14, t15, t16, t17, t18, t19), sep, parent)
#define _DECLARE_20_ARG_CLASS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10,  \
     t11, t12, t13, t14, t15, t16, t17, t18, t19, t20, sep, parent)           \
  _DECLARE_ARG_CLASS(name, 20,                                                \
  _20_SEMIC_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
    t14, t15, t16, t17, t18, t19, t20),                                       \
  _20_COMMA_PAIRS(a, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
    t14, t15, t16, t17, t18, t19, t20), sep, parent)

#define _DEFINE_SCHEDULE_FUNCS(name, numArgs, argParams)          \
_DEFINE_TYPE_INT(Event, name);                                    \
static inline SpCancelHandle Schedule_##name(const SpSimTime &a,  \
                                           const SpObjHandle &b,  \
                     			   argParams,             \
                                           const char *f = NULL,  \
                                           int g = 0)             \
  { return ScheduleProcess (a, name, b, M_##name##_ARG(           \
			    _##numArgs##_VAR_LIST  ), f, g); }    \
static inline SpCancelHandle Schedule##name(const SpSimTime &a,   \
                                            const SpObjHandle &b, \
                     			    argParams,            \
                                            const char *f = NULL, \
                                            int g = 0)            \
  { return ScheduleProcess (a, name, b, M_##name##_ARG(           \
			    _##numArgs##_VAR_LIST  ), f, g); }    \
WARNING_REMOVAL(Schedule_##name)                                  \
WARNING_REMOVAL(Schedule##name)

#define _DEFINE_0_ARG_SCHEDULE_FUNCS(name) \
_DEFINE_TYPE_INT(Event, name);                                    \
static inline SpCancelHandle Schedule_##name(const SpSimTime &a,  \
                                           const SpObjHandle &b,  \
                                           const char *f = NULL,  \
                                           int g = 0)             \
  { return ScheduleProcess (a, name, b, NULL, f, g); }            \
static inline SpCancelHandle Schedule##name(const SpSimTime &a,   \
                                            const SpObjHandle &b, \
                                            const char *f = NULL, \
                                            int g = 0)            \
  { return ScheduleProcess (a, name, b, NULL, f, g); }            \
WARNING_REMOVAL(Schedule_##name)                                  \
WARNING_REMOVAL(Schedule##name)
#define _DEFINE_1_ARG_SCHEDULE_FUNCS(name, t1) \
  _DEFINE_SCHEDULE_FUNCS(name, 1, _1_COMMA_PAIRS(v, t1))
#define _DEFINE_2_ARG_SCHEDULE_FUNCS(name, t1, t2) \
  _DEFINE_SCHEDULE_FUNCS(name, 2, _2_COMMA_PAIRS(v, t1, t2))
#define _DEFINE_3_ARG_SCHEDULE_FUNCS(name, t1, t2, t3) \
  _DEFINE_SCHEDULE_FUNCS(name, 3, _3_COMMA_PAIRS(v, t1, t2, t3))
#define _DEFINE_4_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4) \
  _DEFINE_SCHEDULE_FUNCS(name, 4, _4_COMMA_PAIRS(v, t1, t2, t3, t4))
#define _DEFINE_5_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5) \
  _DEFINE_SCHEDULE_FUNCS(name, 5, _5_COMMA_PAIRS(v, t1, t2, t3, t4, t5))
#define _DEFINE_6_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6) \
  _DEFINE_SCHEDULE_FUNCS(name, 6, _6_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6))
#define _DEFINE_7_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7) \
  _DEFINE_SCHEDULE_FUNCS(name, 7, _7_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7))
#define _DEFINE_8_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8)      \
  _DEFINE_SCHEDULE_FUNCS(name, 8, _8_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, \
     t8))
#define _DEFINE_9_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9)  \
  _DEFINE_SCHEDULE_FUNCS(name, 9, _9_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6, t7, \
     t8, t9))
#define _DEFINE_10_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, \
     t10)                                                                       \
  _DEFINE_SCHEDULE_FUNCS(name, 10, _10_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6,   \
     t7, t8, t9, t10))
#define _DEFINE_11_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, \
     t10, t11)                                                                  \
  _DEFINE_SCHEDULE_FUNCS(name, 11, _11_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6,   \
     t7, t8, t9, t10, t11))
#define _DEFINE_12_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, \
     t10, t11, t12)                                                             \
  _DEFINE_SCHEDULE_FUNCS(name, 12, _12_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6,   \
     t7, t8, t9, t10, t11, t12))
#define _DEFINE_13_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, \
     t10, t11, t12, t13)                                                        \
  _DEFINE_SCHEDULE_FUNCS(name, 13, _13_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6,   \
     t7, t8, t9, t10, t11, t12, t13))
#define _DEFINE_14_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, \
     t10, t11, t12, t13, t14)                                                   \
  _DEFINE_SCHEDULE_FUNCS(name, 14, _14_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6,   \
     t7, t8, t9, t10, t11, t12, t13, t14))
#define _DEFINE_15_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, \
     t10, t11, t12, t13, t14, t15)                                              \
  _DEFINE_SCHEDULE_FUNCS(name, 15, _15_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6,   \
     t7, t8, t9, t10, t11, t12, t13, t14, t15))
#define _DEFINE_16_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, \
     t10, t11, t12, t13, t14, t15, t16)                                         \
  _DEFINE_SCHEDULE_FUNCS(name, 16, _16_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6,   \
     t7, t8, t9, t10, t11, t12, t13, t14, t15, t16))
#define _DEFINE_17_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, \
     t10, t11, t12, t13, t14, t15, t16, t17)                                    \
  _DEFINE_SCHEDULE_FUNCS(name, 17, _17_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6,   \
     t7, t8, t9, t10, t11, t12, t13, t14, t15, t16, t17))
#define _DEFINE_18_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, \
     t10, t11, t12, t13, t14, t15, t16, t17, t18)                               \
  _DEFINE_SCHEDULE_FUNCS(name, 18, _18_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6,   \
     t7, t8, t9, t10, t11, t12, t13, t14, t15, t16, t17, t18))
#define _DEFINE_19_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, \
     t10, t11, t12, t13, t14, t15, t16, t17, t18, t19)                          \
  _DEFINE_SCHEDULE_FUNCS(name, 19, _19_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6,   \
     t7, t8, t9, t10, t11, t12, t13, t14, t15, t16, t17, t18, t19))
#define _DEFINE_20_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, \
     t10, t11, t12, t13, t14, t15, t16, t17, t18, t19, t20)                     \
  _DEFINE_SCHEDULE_FUNCS(name, 20, _20_COMMA_PAIRS(v, t1, t2, t3, t4, t5, t6,   \
     t7, t8, t9, t10, t11, t12, t13, t14, t15, t16, t17, t18, t19, t20))

/*********************************************************************
 * This macro produces an event method class. For example,
 * _DEFINE_EVENT_METHOD(name, sclass, smethod, 2) would generate:
 *
 * class E_##name: public SpEvent   
 * {                            
 *  public:                     
 *    void process()            
 *    {                         
 *       ((sclass *) SIMOBJ)->smethod(((M_##name *) MESSAGE)->v1 ,
 * 	  			      ((M_##name *) MESSAGE)->v2);
 *    }
 * };
 **********************************************************************/
     
#define _DEFINE_EVENT_METHOD(name, nfree, sclass, smethod, numArgs) \
class E_##name: public SpEvent                                       \
{                                                                    \
 public:                                                             \
   void process()                                                    \
   {                                                                 \
      ((sclass *) SIMOBJ)->smethod(_##numArgs##_MSG_ARG_LIST(name)); \
   }                                                                 \
};                                                                   \
_DEFINE_EVENT(name, E_##name, M_##name##_ARG, nfree)

/******************************************************************************
* Macro: DECLARE_*_ARG
* Summary: Allows a file to schedule an event specified by the global
*          integer ID, n, as these macro's first arguments. Then,
*          depending on the value of '*' (0-20), the following
*          arguments represent the types of the method's arguments,
*          respectively.
* Description: This macro is used to provide the global integer ID and
*              message prototypes, without requiring the other
*              information required by the corresponding
*              DEFINE_EVENT_METHOD_*_ARG macro. However, using
*              DEFINE_EVENT_METHOD_*_ARG instead of DECLARE_*_ARG for
*              providing these prototypes is valid and equivalent.
*
* Macro: DEFINE_EVENT_METHOD_*_ARG
* Summary: Specifies an event method on a SpSimObj as a prerequisite
*          to calling PLUG_IN_EVENT
* Description: When using the SPEEDES plug-in system, each object
*              manager and event must be defined and plugged in to
*              SPEEDES. These macros are used to define the name of
*              the event, n, (string name and global integer name),
*              the number of events and messages to be created at a
*              time by the freelists, f, the class name to whom the
*              method belongs, c, and the method name, m. This class
*              and method is presumed to be linked in by the make
*              system. Then, depending on the value of '*' (0-20), the
*              following arguments represent the types of the method's
*              arguments, respectively. An event to call this method
*              is generated, along with a message type ('M_<name of
*              event>, n>') consisting of attributes for each method
*              argument of consistent type ('v1' through 'v<number of
*              arguments>'). This define may be included multiple times
*              in different files, but must be included in the file
*              that calls the corresponding PLUG_IN_EVENT macro.
******************************************************************************/
#define DECLARE_0_ARG(name)                         \
  _DECLARE_0_ARG_CLASS(M_##name, :public, M_SpProc) \
  _DEFINE_0_ARG_SCHEDULE_FUNCS(name)
#define DECLARE(name) DECLARE_0_ARG(name)
#define DEFINE_EVENT_METHOD_0_ARG(name, nfree, sclass, smethod) \
  DECLARE_0_ARG(name)                                           \
  _DEFINE_EVENT_METHOD(name, nfree, sclass, smethod, 0)
#define DECLARE_1_ARG(name, t1)                                 \
  _DECLARE_1_ARG_CLASS(M_##name, t1, :public, M_SpProc)         \
  _DEFINE_1_ARG_SCHEDULE_FUNCS(name, t1)
#define DEFINE_EVENT_METHOD_1_ARG(name, nfree, sclass, smethod, t1) \
  DECLARE_1_ARG(name, t1)                                           \
  _DEFINE_EVENT_METHOD(name, nfree, sclass, smethod, 1)
#define DECLARE_2_ARG(name, t1, t2)                                 \
  _DECLARE_2_ARG_CLASS(M_##name, t1, t2, :public, M_SpProc)         \
  _DEFINE_2_ARG_SCHEDULE_FUNCS(name, t1, t2)
#define DEFINE_EVENT_METHOD_2_ARG(name, nfree, sclass, smethod, t1, t2) \
  DECLARE_2_ARG(name, t1, t2)                                           \
  _DEFINE_EVENT_METHOD(name, nfree, sclass, smethod, 2)
#define DECLARE_3_ARG(name, t1, t2, t3)                                 \
  _DECLARE_3_ARG_CLASS(M_##name, t1, t2, t3, :public, M_SpProc)         \
  _DEFINE_3_ARG_SCHEDULE_FUNCS(name, t1, t2, t3)
#define DEFINE_EVENT_METHOD_3_ARG(name, nfree, sclass, smethod, t1, t2, t3)     \
  DECLARE_3_ARG(name, t1, t2, t3)                                               \
  _DEFINE_EVENT_METHOD(name, nfree, sclass, smethod, 3)
#define DECLARE_4_ARG(name, t1, t2, t3, t4)                                     \
  _DECLARE_4_ARG_CLASS(M_##name, t1, t2, t3, t4, :public, M_SpProc)             \
  _DEFINE_4_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4)
#define DEFINE_EVENT_METHOD_4_ARG(name, nfree, sclass, smethod, t1, t2, t3, t4) \
  DECLARE_4_ARG(name, t1, t2, t3, t4)                                           \
  _DEFINE_EVENT_METHOD(name, nfree, sclass, smethod, 4)
#define DECLARE_5_ARG(name, t1, t2, t3, t4, t5)                                 \
  _DECLARE_5_ARG_CLASS(M_##name, t1, t2, t3, t4, t5, :public, M_SpProc)         \
  _DEFINE_5_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5)
#define DEFINE_EVENT_METHOD_5_ARG(name, nfree, sclass, smethod, t1, t2, t3, t4, \
				  t5)                                           \
  DECLARE_5_ARG(name, t1, t2, t3, t4, t5)                                       \
  _DEFINE_EVENT_METHOD(name, nfree, sclass, smethod, 5)
#define DECLARE_6_ARG(name, t1, t2, t3, t4, t5, t6)                             \
  _DECLARE_6_ARG_CLASS(M_##name, t1, t2, t3, t4, t5, t6, :public, M_SpProc)     \
  _DEFINE_6_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6)
#define DEFINE_EVENT_METHOD_6_ARG(name, nfree, sclass, smethod, t1, t2, t3, t4, \
				  t5, t6)                                       \
  DECLARE_6_ARG(name, t1, t2, t3, t4, t5, t6)                                   \
  _DEFINE_EVENT_METHOD(name, nfree, sclass, smethod, 6)
#define DECLARE_7_ARG(name, t1, t2, t3, t4, t5, t6, t7)                         \
  _DECLARE_7_ARG_CLASS(M_##name, t1, t2, t3, t4, t5, t6, t7, :public, M_SpProc) \
  _DEFINE_7_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7)
#define DEFINE_EVENT_METHOD_7_ARG(name, nfree, sclass, smethod, t1, t2, t3, t4, \
				  t5, t6, t7)                                   \
  DECLARE_7_ARG(name, t1, t2, t3, t4, t5, t6, t7)                               \
  _DEFINE_EVENT_METHOD(name, nfree, sclass, smethod, 7)
#define DECLARE_8_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8)                     \
  _DECLARE_8_ARG_CLASS(M_##name, t1, t2, t3, t4, t5, t6, t7, t8, :public,       \
		       M_SpProc)                                                \
  _DEFINE_8_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8)
#define DEFINE_EVENT_METHOD_8_ARG(name, nfree, sclass, smethod, t1, t2, t3, t4, \
				  t5, t6, t7, t8)                               \
  DECLARE_8_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8)                           \
  _DEFINE_EVENT_METHOD(name, nfree, sclass, smethod, 8)
#define DECLARE_9_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8, t9)                 \
  _DECLARE_9_ARG_CLASS(M_##name, t1, t2, t3, t4, t5, t6, t7, t8, t9, :public,   \
     M_SpProc)                                                                  \
  _DEFINE_9_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9)
#define DEFINE_EVENT_METHOD_9_ARG(name, nfree, sclass, smethod, t1, t2, t3, t4, \
				  t5, t6, t7, t8, t9)                           \
  DECLARE_9_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8, t9)                       \
  _DEFINE_EVENT_METHOD(name, nfree, sclass, smethod, 9)
#define DECLARE_10_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10)           \
  _DECLARE_10_ARG_CLASS(M_##name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10,      \
			:public, M_SpProc)                                      \
  _DEFINE_10_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10)
#define DEFINE_EVENT_METHOD_10_ARG(name, nfree, sclass, smethod, t1, t2, t3,    \
				   t4, t5, t6, t7, t8, t9, t10)                 \
  DECLARE_10_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10)                 \
  _DEFINE_EVENT_METHOD(name, nfree, sclass, smethod, 10)
#define DECLARE_11_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11)      \
  _DECLARE_11_ARG_CLASS(M_##name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, \
     :public, M_SpProc)                                                         \
  _DEFINE_11_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10,  \
				t11)
#define DEFINE_EVENT_METHOD_11_ARG(name, nfree, sclass, smethod, t1, t2, t3,    \
				   t4, t5, t6, t7, t8, t9, t10, t11)            \
  DECLARE_11_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11)            \
  _DEFINE_EVENT_METHOD(name, nfree, sclass, smethod, 11)
#define DECLARE_12_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12) \
  _DECLARE_12_ARG_CLASS(M_##name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, \
			t12, :public, M_SpProc)                                 \
  _DEFINE_12_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10,  \
				t11, t12)
#define DEFINE_EVENT_METHOD_12_ARG(name, nfree, sclass, smethod, t1, t2, t3,    \
				   t4, t5, t6, t7, t8, t9, t10, t11, t12)       \
  DECLARE_12_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12)       \
  _DEFINE_EVENT_METHOD(name, nfree, sclass, smethod, 12)
#define DECLARE_13_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,      \
		       t12, t13)                                                \
  _DECLARE_13_ARG_CLASS(M_##name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, \
			t12, t13, :public, M_SpProc)                            \
  _DEFINE_13_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10,  \
				t11, t12, t13)
#define DEFINE_EVENT_METHOD_13_ARG(name, nfree, sclass, smethod, t1, t2, t3,    \
				   t4, t5, t6, t7, t8, t9, t10, t11, t12, t13)  \
  DECLARE_13_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13)  \
  _DEFINE_EVENT_METHOD(name, nfree, sclass, smethod, 13)
#define DECLARE_14_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,      \
		       t12, t13, t14)                                           \
  _DECLARE_14_ARG_CLASS(M_##name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, \
			t12, t13, t14, :public, M_SpProc)                       \
  _DEFINE_14_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10,  \
				t11, t12, t13, t14)
#define DEFINE_EVENT_METHOD_14_ARG(name, nfree, sclass, smethod, t1, t2, t3,    \
				   t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
				   t14)                                         \
  DECLARE_14_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
		 t14)                                                           \
  _DEFINE_EVENT_METHOD(name, nfree, sclass, smethod, 14)
#define DECLARE_15_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,      \
		       t12, t13, t14, t15)                                      \
  _DECLARE_15_ARG_CLASS(M_##name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, \
			t12, t13, t14, t15, :public, M_SpProc)                  \
  _DEFINE_15_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10,  \
				t11, t12, t13, t14, t15)
#define DEFINE_EVENT_METHOD_15_ARG(name, nfree, sclass, smethod, t1, t2, t3,    \
				   t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
				   t14, t15)                                    \
  DECLARE_15_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
		 t14, t15)                                                      \
  _DEFINE_EVENT_METHOD(name, nfree, sclass, smethod, 15)
#define DECLARE_16_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,      \
		       t12, t13, t14, t15, t16)                                 \
  _DECLARE_16_ARG_CLASS(M_##name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, \
			t12, t13, t14, t15, t16, :public, M_SpProc)             \
  _DEFINE_16_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10,  \
				t11, t12, t13, t14, t15, t16)
#define DEFINE_EVENT_METHOD_16_ARG(name, nfree, sclass, smethod, t1, t2, t3,    \
				   t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
				   t14, t15, t16)                               \
  DECLARE_16_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
		 t14, t15, t16)                                                 \
  _DEFINE_EVENT_METHOD(name, nfree, sclass, smethod, 16)
#define DECLARE_17_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,      \
		       t12, t13, t14, t15, t16, t17)                            \
  _DECLARE_17_ARG_CLASS(M_##name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, \
			t12, t13, t14, t15, t16, t17, :public, M_SpProc)        \
  _DEFINE_17_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10,  \
				t11, t12, t13, t14, t15, t16, t17)
#define DEFINE_EVENT_METHOD_17_ARG(name, nfree, sclass, smethod, t1, t2, t3,    \
				   t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
				   t14, t15, t16, t17)                          \
  DECLARE_17_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
		 t14, t15, t16, t17)                                            \
  _DEFINE_EVENT_METHOD(name, nfree, sclass, smethod, 17)
#define DECLARE_18_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,      \
		       t12, t13, t14, t15, t16, t17, t18)                       \
  _DECLARE_18_ARG_CLASS(M_##name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, \
			t12, t13, t14, t15, t16, t17, t18, :public, M_SpProc)   \
  _DEFINE_18_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10,  \
				t11, t12, t13, t14, t15, t16, t17, t18)
#define DEFINE_EVENT_METHOD_18_ARG(name, nfree, sclass, smethod, t1, t2, t3,    \
				   t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
				   t14, t15, t16, t17, t18)                     \
  DECLARE_18_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
		 t14, t15, t16, t17, t18)                                       \
  _DEFINE_EVENT_METHOD(name, nfree, sclass, smethod, 18)
#define DECLARE_19_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,      \
		       t12, t13, t14, t15, t16, t17, t18, t19)                  \
  _DECLARE_19_ARG_CLASS(M_##name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, \
			t12, t13, t14, t15, t16, t17, t18, t19, :public,        \
     M_SpProc)                                                                  \
  _DEFINE_19_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10,  \
				t11, t12, t13, t14, t15, t16, t17, t18, t19)
#define DEFINE_EVENT_METHOD_19_ARG(name, nfree, sclass, smethod, t1, t2, t3,    \
				   t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
				   t14, t15, t16, t17, t18, t19)                \
  DECLARE_19_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
		 t14, t15, t16, t17, t18, t19)                                  \
  _DEFINE_EVENT_METHOD(name, nfree, sclass, smethod, 19)
#define DECLARE_20_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11,      \
		       t12, t13, t14, t15, t16, t17, t18, t19, t20)             \
  _DECLARE_20_ARG_CLASS(M_##name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, \
			t12, t13, t14, t15, t16, t17, t18, t19, t20, :public,   \
     M_SpProc)                                                                  \
  _DEFINE_20_ARG_SCHEDULE_FUNCS(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10,  \
				t11, t12, t13, t14, t15, t16, t17, t18, t19, t20)
#define DEFINE_EVENT_METHOD_20_ARG(name, nfree, sclass, smethod, t1, t2, t3,    \
				   t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
				   t14, t15, t16, t17, t18, t19, t20)           \
  DECLARE_20_ARG(name, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
		 t14, t15, t16, t17, t18, t19, t20)                             \
  _DEFINE_EVENT_METHOD(name, nfree, sclass, smethod, 20)

#define DEFINE_ASK_EVENT_METHOD(name, nfree, sclass, smethod, numArgs)   \
class E_##name##_ASK: public SpEvent                                     \
{                                                                        \
 public:                                                                 \
   void process()                                                        \
   {                                                                     \
      int offset = (M_SpProc *)GetMsg()->GetOffset();                    \
      char *data = (char*) GetData();                                    \
      SpAskReturn *ret = (SpAskReturn *) data[offset];                   \
      A_##name *args = (A_##name *) data[offset() - ret->GetAskBytes()]; \
      ((sclass *) SIMOBJ)->smethod(_##numArgs##_ARG_LIST(args));         \
   }                                                                     \
};                                                                       \
_DEFINE_ASK_EVENT(name, E_##name##_ASK, nfree)                           \
GetAskRet_##name(t1 &a1) \
   { a1 = ((A_##name *) (GetData() + ((SpProc *) GetMsg())->GetOffset()))->a1; }

   
   
#define EVENT_TYPE(name) \
   DECLARE_0_ARG(name)
#define OBJECT_MANAGER_TYPE(name) \
   _DEFINE_TYPE_INT(SimObj, name);
#define OBJ_MGR_TYPE(name) OBJECT_MANAGER_TYPE(name)
#define OM_TYPE(name) OBJECT_MANAGER_TYPE(name)
#define O_TYPE(name) OBJECT_MANAGER_TYPE(name)
#define E_TYPE(name) EVENT_TYPE(name)
#define SCHEDULE(name) EVENT_TYPE(name)
#define SCHEDULE_0_ARG(name) EVENT_TYPE(name)
#define SCHEDULE_1_ARG(name, t1) DECLARE_1_ARG(name, t1)
#define SCHEDULE_2_ARG(name, t1, t2) DECLARE_2_ARG(name, t1, t2)
#define SCHEDULE_3_ARG(n, t1, t2, t3) DECLARE_3_ARG(n, t1, t2, t3)
#define SCHEDULE_4_ARG(n, t1, t2, t3, t4) DECLARE_4_ARG(n, t1, t2, t3, t4)
#define SCHEDULE_5_ARG(n, t1, t2, t3, t4, t5) \
   DECLARE_5_ARG(n, t1, t2, t3, t4, t5)
#define SCHEDULE_6_ARG(n, t1, t2, t3, t4, t5, t6) \
   DECLARE_6_ARG(n, t1, t2, t3, t4, t5, t6)
#define SCHEDULE_7_ARG(n, t1, t2, t3, t4, t5, t6, t7) \
   DECLARE_7_ARG(n, t1, t2, t3, t4, t5, t6, t7)
#define SCHEDULE_8_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8) \
   DECLARE_8_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8)
#define SCHEDULE_9_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8, t9) \
   DECLARE_9_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8, t9)
#define SCHEDULE_10_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10) \
   DECLARE_10_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10)
#define SCHEDULE_11_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11) \
   DECLARE_11_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11)
#define SCHEDULE_12_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12) \
   DECLARE_12_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12)
#define SCHEDULE_13_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, \
		       t13)                                                   \
   DECLARE_13_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13)
#define SCHEDULE_14_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, \
		       t13, t14)                                              \
   DECLARE_14_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
		  t14)
#define SCHEDULE_15_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, \
		       t13, t14, t15)                                         \
   DECLARE_15_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
		  t14, t15)
#define SCHEDULE_16_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, \
		       t13, t14, t15, t16)                                    \
   DECLARE_16_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
		  t14, t15, t16)
#define SCHEDULE_17_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, \
		       t13, t14, t15, t16, t17)                               \
   DECLARE_17_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
		  t14, t15, t16, t17)
#define SCHEDULE_18_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, \
		       t13, t14, t15, t16, t17, t18)                          \
   DECLARE_18_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
		  t14, t15, t16, t17, t18)
#define SCHEDULE_19_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, \
		       t13, t14, t15, t16, t17, t18, t19)                     \
   DECLARE_19_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
		  t14, t15, t16, t17, t18, t19)
#define SCHEDULE_20_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, \
		       t13, t14, t15, t16, t17, t18, t19, t20)                \
   DECLARE_20_ARG(n, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13,  \
		  t14, t15, t16, t17, t18, t19, t20)

#endif
