// SpProc.H header file
// Jim Brutocao
// 1/22/99

// Description: defines macros to support process-model simulations in SPEEDES

#ifndef SpProc_H
#define SpProc_H

#ifndef DEBUG_PROCESS_MODEL
#define _M_C_
#endif

#include "SpProcStackElement.H"
#include "SpWarningPatch.H"
#include "RB_SpContainerFuncs.H"
#include "SpScheduleEvent.H"
#include "SpStdlib.H"

/***************************************************************************
* internal (non-user) macros:                                              *
***************************************************************************/

#define P_MSG (*(M_SpProc *) CurrentEvent->get_MESSAGE())
#define FIRST_ENTRY (P_MSG.type == BEGIN_PROCESS)
#define COMMA ,
     
#define _JUMP(rLabel) case rLabel: goto label_##rLabel; _M_C_
   
#define  _0_JUMPS  _JUMP (0)
#define  _1_JUMPS  _0_JUMPS _JUMP (1)
#define  _2_JUMPS  _1_JUMPS _JUMP (2)
#define  _3_JUMPS  _2_JUMPS _JUMP (3)
#define  _4_JUMPS  _3_JUMPS _JUMP (4)
#define  _5_JUMPS  _4_JUMPS _JUMP (5)
#define  _6_JUMPS  _5_JUMPS _JUMP (6)
#define  _7_JUMPS  _6_JUMPS _JUMP (7)
#define  _8_JUMPS  _7_JUMPS _JUMP (8)
#define  _9_JUMPS  _8_JUMPS _JUMP (9)
#define _10_JUMPS  _9_JUMPS _JUMP (10)
#define _11_JUMPS _10_JUMPS _JUMP (11)
#define _12_JUMPS _11_JUMPS _JUMP (12)
#define _13_JUMPS _12_JUMPS _JUMP (13)
#define _14_JUMPS _13_JUMPS _JUMP (14)
#define _15_JUMPS _14_JUMPS _JUMP (15)
#define _16_JUMPS _15_JUMPS _JUMP (16)
#define _17_JUMPS _16_JUMPS _JUMP (17)
#define _18_JUMPS _17_JUMPS _JUMP (18)
#define _19_JUMPS _18_JUMPS _JUMP (19)
#define _20_JUMPS _19_JUMPS _JUMP (20)
#define _21_JUMPS _20_JUMPS _JUMP (21)
#define _22_JUMPS _21_JUMPS _JUMP (22)
#define _23_JUMPS _22_JUMPS _JUMP (23)
#define _24_JUMPS _23_JUMPS _JUMP (24)
#define _25_JUMPS _24_JUMPS _JUMP (25)
#define _26_JUMPS _25_JUMPS _JUMP (26)
#define _27_JUMPS _26_JUMPS _JUMP (27)
#define _28_JUMPS _27_JUMPS _JUMP (28)
#define _29_JUMPS _28_JUMPS _JUMP (29)
#define _30_JUMPS _29_JUMPS _JUMP (30)

#define _DUMMY \
{

#define _DECLARE_LV_CLASS(className, methodName, stackType)  \
class LV_##className##_##methodName: public stackType        _M_C_ \
{                                                            _M_C_ \
 public:                                                     _M_C_ \
   LV_##className##_##methodName(SpObjHandle *simobj,        _M_C_ \
				 int procType):              _M_C_ \
      stackType##(simobj, procType) {}                       _M_C_ \
      LV_##className##_##methodName(){}

      
  //Change the following RB_DEFINE_CLASS() to RB_DEFINE_CLASS_WITH_NO_DEFAULT_CONSTRUCTOR
  //when persistence supports it.
#define _BEGIN_SET_UP(className, methodName)                 \
};                                                           _M_C_ \
RB_DEFINE_CLASS\
(LV_##className##_##methodName);                             _M_C_ \
                                                             _M_C_ \
void className##::##methodName##()                           _M_C_ \
{                                                            _M_C_ \
   LV_##className##_##methodName *LV;

#define _ESTABLISH_LV(className, methodName)                 \
   if FIRST_ENTRY                                            _M_C_ \
     LV = (LV_##className##_##methodName *)                  _M_C_ \
        RB_NEW_LV_##className##_##methodName                 _M_C_ \
	   (new LV_##className##_##methodName                _M_C_ \
	    (GetObjectHandle(),                              _M_C_ \
	     CurrentEvent->get_EVENT_TYPE()));               _M_C_ \
   else LV = (LV_##className##_##methodName *)               _M_C_ \
	  P_MSG.reentryLV;

#define _END_SET_UP(name, numReentries)                      \
   if FIRST_ENTRY goto label_BEGIN;                          _M_C_ \
   switch (LV->reentryLabel)                                 _M_C_ \
   {                                                         _M_C_ \
      _##numReentries##_JUMPS                                _M_C_ \
      default: LV->AbortProc(#name);                         _M_C_ \
   }                                                         _M_C_ \
   label_BEGIN:                                              _M_C_ \
   LV->InitProc();                                           _M_C_ \
   label_0:

#define _BLOCK(rLabel)                                       \
   LV->SaveContext(rLabel);                                  _M_C_ \
   return;                                                   _M_C_ \
   label_##rLabel:

#define _REQUEST_SEMAPHORE(method, amountIfAny, commaIfAny,  \
                           rLabel, semaphore, timeOut, sf)   \
   if (!(semaphore##).##method (amountIfAny commaIfAny       _M_C_ \
			     CurrentEvent->get_EVENT_TYPE(), _M_C_ \
			     LV, timeOut)                    _M_C_ \
       && timeOut != 0)                                      _M_C_ \
   {                                                         _M_C_ \
      if (LV->finiteTimeOut = (timeOut > 0))                 _M_C_ \
      {                                                      _M_C_ \
         LV->SemaphorePtr = &(semaphore);                    _M_C_ \
	 LV->ScheduleTimeOut(GetTime(), P_MSG, timeOut);     _M_C_ \
      }                                                      _M_C_ \
      _BLOCK (rLabel);                                       _M_C_ \
      if (LV->finiteTimeOut)                                 _M_C_ \
	LV->HandleSemReturn(P_MSG, sf);                      _M_C_ \
   }                                                         _M_C_ \
   else sf = 1;

// new process model (no success flag):
#define __REQUEST_SEMAPHORE(method, amountIfAny, commaIfAny, \
                           rLabel, semaphore, timeOut)       \
   if (!(semaphore##).##method (amountIfAny commaIfAny       _M_C_ \
			     CurrentEvent->get_EVENT_TYPE(), _M_C_ \
			     LV, timeOut)                    _M_C_ \
       && timeOut != 0)                                      _M_C_ \
   {                                                         _M_C_ \
      if (LV->finiteTimeOut = (timeOut > 0))                 _M_C_ \
      {                                                      _M_C_ \
         LV->SemaphorePtr = &(semaphore);                    _M_C_ \
	 LV->ScheduleTimeOut(GetTime(), P_MSG, timeOut);     _M_C_ \
      }                                                      _M_C_ \
      _BLOCK (rLabel);                                       _M_C_ \
      if (LV->finiteTimeOut)                                 _M_C_ \
	LV->HandleSemReturn(P_MSG);                          _M_C_ \
   }

/***************************************************************************
* process model (user) macros:                                             *
***************************************************************************/

#define PROCESS_LV(className, methodName)                    \
_DECLARE_LV_CLASS(className, methodName, SpProcStackElement)

#define BEGIN_PROCESS(className, methodName, numReentries)   \
   _BEGIN_SET_UP (className, methodName)                     _M_C_ \
   _ESTABLISH_LV (className, methodName)                     _M_C_ \
   if (FIRST_ENTRY) LV->EstablishNewStack();                 _M_C_ \
   _END_SET_UP (className, numReentries)

#define END_PROCESS(className, methodName)                   \
   WARNING_REMOVAL_label_END                                 _M_C_ \
   label_END:                                                _M_C_ \
   LV_##className##_##methodName *LVtemp =                   _M_C_ \
      (LV_##className##_##methodName *)                      _M_C_ \
         RB_POP_TOP (LV->stackPtr->sePtr);                   _M_C_ \
   RB_DELETE_LV_##className##_##methodName (LVtemp);         _M_C_ \
   LV->CleanupProc();                                        _M_C_ \
}

#define WAIT(rLabel, duration)                               \
   LV->ScheduleWaitReturn(P_MSG, GetTime() + duration);      _M_C_ \
   _BLOCK (rLabel);

#define WAIT_UNTIL(rLabel, scheduleTime)                     \
/* The if_expression is a work around for an insure bug similar to that described in testSEMAPHORE/S_ship.C */                                    \
   int if_expression;                                        _M_C_ \
   if_expression = GetTime() < SpSimTime(scheduleTime);      _M_C_ \
   if (if_expression)                                        _M_C_ \
   {                                                         _M_C_ \
      LV->ScheduleWaitReturn(P_MSG, scheduleTime);           _M_C_ \
      _BLOCK (rLabel);                                       _M_C_ \
   }

#define RETURN goto label_END

#define REQUEST_RESOURCE(rLabel, semaphore, amount, timeOut, successFlag) \
   _REQUEST_SEMAPHORE(RequestResource, amount, COMMA,        _M_C_ \
		      rLabel, semaphore, timeOut, successFlag)

// This macro passes a zero and a plus symbol to _REQUEST_SEMAPHORE as
// the second and third arguments.  The _REQUEST_SEMAPHORE macro will
// put these arguments together before a variable which is of type int.
// So this means that we are in effect adding 0 to that int.  Such an
// addition will not change the value where that int is since adding 0
// to something does not change it.  I do this because other callers of
// _REQUEST_SEMAPHORE need to pass in some things which make sense,
// but this caller below does NOT need to really pass in anything.  But
// the HP compiler complained about not passing in anything to the second
// and third arguments, so I satisfied the HP compiler by passing in 0
// and +.
#define WAIT_FOR_LOGICAL(rLabel, semaphore, timeOut,         \
		         successFlag)                        \
   _REQUEST_SEMAPHORE(WaitForLogical,0,+, rLabel, semaphore,   _M_C_ \
		      timeOut, successFlag)

#define WAIT_FOR_COUNTER(rLabel, semaphore, timeOut, successFlag) \
   WAIT_FOR_LOGICAL(rLabel, semaphore, timeOut, successFlag)

//----------------------------------------------------------------------------
// NEW PROCESS MODEL INTERFACE (eventually replace everything above this line)
//----------------------------------------------------------------------------
// SpProc.H header file
// Jim Brutocao
// 10/28/98
// Description: defines macros to support process-model methods in SPEEDES

#define WAIT_FOR(rLabel, semaphore, timeOut)                  \
   __REQUEST_SEMAPHORE(WaitForLogical,0,+, rLabel, semaphore, _M_C_ \
		      timeOut)
#define WAIT_FOR_RESOURCE(rLabel, semaphore, amount, timeOut) \
   __REQUEST_SEMAPHORE(RequestResource, amount, COMMA,        _M_C_ \
		      rLabel, semaphore, timeOut)      

#define P_VAR                                                \
{                                                            _M_C_ \
   SpProcStackElement *LV;                                   _M_C_ \
   if FIRST_ENTRY                                            _M_C_ \
   {                                                         _M_C_ \
      LV = RB_NEW_SpProcStackElement (new SpProcStackElement _M_C_ \
	 (CurrentEvent->GetObjectHandle(),                   _M_C_ \
	    CurrentEvent->get_EVENT_TYPE()));                _M_C_ \
      LV->EstablishNewStack();                               _M_C_ \
   }                                                         _M_C_ \
   else LV = (SpProcStackElement *) P_MSG.reentryLV;         _M_C_ \
   int saveVarStateCalled = 0;                               _M_C_ \
   /* This line gets rid of a warning saying it is not used.*/ _M_C_ \
   saveVarStateCalled=saveVarStateCalled;

#define P_LV(type, var)                                      \
   type var;                                                 _M_C_ \
   if ((++saveVarStateCalled) && FIRST_ENTRY)              _M_C_ \
      LV->SaveVarState(&var, sizeof(type), FIRST_ENTRY);

#define P_BEGIN(numReentries)                                \
   if FIRST_ENTRY LV->ConsolidateVarsState();                _M_C_ \
   else                                                      _M_C_ \
   {                                                         _M_C_ \
      LV->LoadVarsState();                                   _M_C_ \
      switch (LV->reentryLabel)                              _M_C_ \
      {                                                      _M_C_ \
         _##numReentries##_JUMPS                             _M_C_ \
         default: LV->AbortProc();                           _M_C_ \
      }                                                      _M_C_ \
   }                                                         _M_C_ \
   LV->InitProc();                                           _M_C_ \
   label_0:
   // label_0 is needed for rollbacks since rolling back the first
   // wait sets reentry to 0 and when reprocessing, the process model
   // must not call InitProc() again

#define P_END                                                \
   WARNING_REMOVAL_label_END                                 _M_C_ \
   label_END:                                                _M_C_ \
   SpProcStackElement *SEtemp = (SpProcStackElement *)       _M_C_ \
      RB_POP_TOP (LV->stackPtr->sePtr);                      _M_C_ \
   RB_DELETE_SpProcStackElement(SEtemp);                     _M_C_ \
   LV->CleanupProc();                                        _M_C_ \
}
//   (E_Ask *) GeEvent()->SetFinishedProc();                  _M_C_


#define ASK_1_ARG(rLabel, delay, askObj, askName, p1,        \
                  userData, userBytes)                       \
   LV->SetUpAskData(A_##askName##(pi), sizeof(A_##askName),  _M_C_ \
	            userData, userBytes, data, bytes);       _M_C_ \
   ScheduleEvent(GetTime() + delay, askObj, MESSAGE, data,   _M_C_ \
		 bytes);                                     _M_C_ \
   _BLOCK (rLabel);                                          _M_C_ \
   GetAskRet_##askName(p1);

#endif
