// SpProcSem.C

#include "SpProcSem.H"
#include "SpScheduleEvent.H"
extern SpEvent *CurrentEvent;
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "RB_Types.H"
#include "RB_SpContainerFuncs.H"
#include "RB_SpDefineClass.H"

RB_DEFINE_CLASS(SpWaitingItem);
RB_DEFINE_CLASS(SpIntWaitingItem);
RB_DEFINE_CLASS(SpFloatWaitingItem);
RB_DEFINE_CLASS(SpLogicalSem);
RB_DEFINE_CLASS(SpIntSem);
RB_DEFINE_CLASS(SpFloatSem);

SpWaitingItem::SpWaitingItem (int et, SpProcStackElement *lv, double tof):
   eventType (et), LVptr (lv), timeOut (tof) {}

SpIntWaitingItem::SpIntWaitingItem (int ra, int et, SpProcStackElement *lv,
				    int tof):
   SpWaitingItem(et, lv, tof), requestAmount (ra) {}

SpFloatWaitingItem::SpFloatWaitingItem (double ra, int et,
					SpProcStackElement *lv, int tof):
   SpWaitingItem(et, lv, tof), requestAmount (ra) {}

SpSharedSemWaitingItem::SpSharedSemWaitingItem (double ra,
						int et,
						SpProcStackElement *lv,
						int tof,
						SpObjHandle oh,
						double rl):
   SpFloatWaitingItem(ra, et, lv, tof), obj(oh), reentryLookahead(rl) {}

#define REMOVE_BLOCKING_PROCESS_TEMPLATE(classType, waitingItemType)     \
void classType::RemoveBlockingProcess (SpProcStackElement *LVptr)        \
{                                                                        \
   waitingItemType *waitingItem;                                         \
									 \
   waitingItem = (waitingItemType *) blockingProcessList.GetTop();       \
   while (waitingItem)                                                   \
   {                                                                     \
      if (waitingItem->LVptr == LVptr)                                   \
      {                                                                  \
	 blockingProcessList -= waitingItem;                             \
	 RB_DELETE_##waitingItemType (waitingItem);                      \
	 return;                                                         \
      }                                                                  \
      else waitingItem = (waitingItemType *) ++blockingProcessList;      \
   }                                                                     \
}

REMOVE_BLOCKING_PROCESS_TEMPLATE(SpLogicalSem, SpWaitingItem)
REMOVE_BLOCKING_PROCESS_TEMPLATE(SpIntSem,     SpIntWaitingItem)
REMOVE_BLOCKING_PROCESS_TEMPLATE(SpFloatSem,   SpFloatWaitingItem)

double SpResourceSem::operator = (double d)
{
   cout << "resource semaphores cannot use \"=\" operator with operand " << d
	<< endl;
   exit (-1);
   return d;
}

void SpProcSem::ScheduleSuccessfulSemReturnEvent(SpProcStackElement *LVptr,
						 int eventType)
{
   void *cancelHandleVoidPtr = LVptr->TimeOutCancelHandle;
   if (LVptr->finiteTimeOut)
     CancelEvent((SpCancelHandle *) cancelHandleVoidPtr);

   C_HOLDER *holder = EventTypes->new_message(eventType);
   M_SpProc *reentryMsg = (M_SpProc*) holder->get_buff();
   reentryMsg->type = SUCCESSFUL_SEMAPHORE_RETURN;
   reentryMsg->reentryLV = LVptr;
   ScheduleEvent(CurrentEvent->TIME_TAG,
		 eventType,
		 CurrentEvent->get_OBJECT_TYPE (),
		 CurrentEvent->LOCAL_ID,
		 CurrentEvent->get_NODE (),
		 reentryMsg);
   EventTypes->DefinitelyPutOn_qfree_messages(holder);
}

int SpLogicalSem::Toggle (int newStateIfSetting)
{
   SpWaitingItem *waitingItem;

   if (0 != state) state = 0;
   else
   {
      state = newStateIfSetting;
      waitingItem = (SpWaitingItem *) blockingProcessList.GetTop();
      while (waitingItem)
      {
	 ScheduleSuccessfulSemReturnEvent(waitingItem->LVptr,
					  waitingItem->eventType);
	 blockingProcessList -= waitingItem;
	 RB_DELETE_SpWaitingItem(waitingItem);
	 waitingItem = (SpWaitingItem *) blockingProcessList.GetTop();
      }
   }
   return state;
}

int SpLogicalSem::WaitForLogical (int eventType, SpProcStackElement *LVptr,
				  double timeOut)
{
   if (state) return 1;
   else if (timeOut == 0) return 0;
   else
   {
      SpWaitingItem *waitingItem = RB_NEW_SpWaitingItem
	(new SpWaitingItem (eventType, LVptr, timeOut));
      blockingProcessList += waitingItem;
      return 0;
   }
}

#define RELEASE_WAITING_ITEMS_TEMPLATE(waitingItemType)               \
waitingItemType *waitingItem =                                        \
   (waitingItemType *) blockingProcessList.GetTop();                  \
while (waitingItem)                                                   \
{                                                                     \
   if (waitingItem->requestAmount <= amount)                          \
   {                                                                  \
      amount -= waitingItem->requestAmount;                           \
      ScheduleSuccessfulSemReturnEvent(waitingItem->LVptr,            \
				       waitingItem->eventType);       \
      blockingProcessList -= waitingItem;                             \
      RB_DELETE_##waitingItemType(waitingItem);                       \
      waitingItem = (waitingItemType *) blockingProcessList.GetTop(); \
   }                                                                  \
   else return amount;                                                \
}

int SpIntSem::AddResource (int val)
{
   amount += val;
   RELEASE_WAITING_ITEMS_TEMPLATE(SpIntWaitingItem);
   return amount;
}

double SpFloatSem::AddResource (double val)
{
   amount += val;
   RELEASE_WAITING_ITEMS_TEMPLATE(SpFloatWaitingItem);
   return amount;
}

int SpIntSem::RequestResource (int requestAmount, int eventType,
			       SpProcStackElement *LVptr, int timeOut)
{
   if (requestAmount <= amount)
   {
      amount -= requestAmount;
      return 1;
   }
   else if (!timeOut) return 0;
   else
   {
      SpIntWaitingItem *waitingItem = RB_NEW_SpIntWaitingItem
	(new SpIntWaitingItem (requestAmount, eventType, LVptr, timeOut));
      blockingProcessList += waitingItem;
      return 0;
   }
}

int SpFloatSem::RequestResource (double requestAmount, int eventType,
				 SpProcStackElement *LVptr, int timeOut)
{
   if (requestAmount <= amount)
   {
      amount -= requestAmount;
      return 1;
   }
   else if (!timeOut) return 0;
   else
   {
      SpFloatWaitingItem *waitingItem = RB_NEW_SpFloatWaitingItem
	(new SpFloatWaitingItem (requestAmount, eventType, LVptr, timeOut));
      blockingProcessList += waitingItem;
      return 0;
   }
}
void RB_Register_SpWaitingItem()
{
  RB_REGISTER_CLASS_SpWaitingItem();
}

void RB_Register_SpIntWaitingItem()
{
  RB_REGISTER_CLASS_SpIntWaitingItem();
}

void RB_Register_SpFloatWaitingItem()
{
  RB_REGISTER_CLASS_SpFloatWaitingItem();
}

void RB_Register_SpLogicalSem()
{
  RB_REGISTER_CLASS_SpLogicalSem();
}

void RB_Register_SpIntSem()
{
  RB_REGISTER_CLASS_SpIntSem();
}

void RB_Register_SpFloatSem()
{
  RB_REGISTER_CLASS_SpFloatSem();
}

