// SpProcSem.H

#ifndef SpProcSem_Classes
#define SpProcSem_Classes

#include "SpAssert.H"
#include "SpProcStackElement.H"
#include "xqueue.H"
#include "xlist.H"
#include "M_SpProc.H"

// SpProcSem support structures:

struct SpWaitingItem
{
  SpWaitingItem(){} //Need to remove when persistence no longer needs
   SpWaitingItem (int et, SpProcStackElement *lv, double tof);
   int eventType;
   SpProcStackElement *LVptr;
   double timeOut;      
};

struct SpIntWaitingItem: public SpWaitingItem
{
  SpIntWaitingItem(){} //Need to remove when persistence no longer needs
   SpIntWaitingItem (int ra, int et, SpProcStackElement *lv, int tof);
   int requestAmount;
};

struct SpFloatWaitingItem: public SpWaitingItem
{
  SpFloatWaitingItem(){} //Need to remove when persistence no longer needs
   SpFloatWaitingItem (double ra, int et, SpProcStackElement *lv, int tof);
   double requestAmount;
};

struct SpSharedSemWaitingItem: public SpFloatWaitingItem, public C_SQ_ITEM
{
  SpSharedSemWaitingItem(){} //Need to remove when persistence no longer needs
   SpSharedSemWaitingItem (double ra, int et, SpProcStackElement *lv, int tof,
			   SpObjHandle oh, double rl);
   SpObjHandle obj;
   double reentryLookahead;
};

// SpProcSem class declarations:

// Abstract base class
class SpProcSem: public C_SQ_ITEM
{
 protected:
   XLIST blockingProcessList;
 public:
   //SpProcSem() {}
   virtual void RemoveBlockingProcess (SpProcStackElement *LVptr) = 0;
   void ScheduleSuccessfulSemReturnEvent(SpProcStackElement *LVptr,
					 int eventType);
   double operator += (double d);
   double operator -= (double d);
   double operator *= (double d);
   double operator /= (double d);
};

class SpLogicalSem: public SpProcSem
{
   RB_int state;
   void RemoveBlockingProcess (SpProcStackElement *LVptr);
 public:
   SpLogicalSem (int a = 0)     { state = a; }
   int Set(int newState = 1)    { return (*this = newState); }
   int Unset()                  { return (*this = 0); }
   int GetState ()              { return state; }
   int IsSet ()                 { return state; }
   int operator =  (int d)      { if ((d && 0 == state) || (!d && 0 != state))
                                  Toggle(d); else state = d; return d; }
   int operator ++ ()           { int val = state + 1; *this = val; return val; }
   int operator ++ (int)        { int val = state; *this = val + 1; return val; }
   int operator -- ()           { int val = state; if (0 < state)
				  { val = state - 1; *this = val; } return val; }
   int operator -- (int)        { int val = state; if (0 < state)
                                  { *this = val - 1; } return val; }
   int operator += (int amount) { int val = state + amount; *this = val;
                                  return val;}
   int operator -= (int amount) { int val = state; if (amount <= state)
                                  { val = state - amount; *this = val; }
                                  return val; }
   operator int ()              { return state; }
   void PrintState ()           { cout << "state = " << state << endl; }
   int  Toggle(int newStateIfSetting = 1);
   int  WaitForLogical (int eventType, SpProcStackElement *LVptr,
                        double timeOut = 0);
};

typedef SpLogicalSem SpCounterSem;

// Abstract base class
class SpResourceSem: public SpProcSem
{   
 public:
   double operator = (double d);
};

class SpIntSem: public SpResourceSem
{
   RB_int amount;
 public:
   SpIntSem (int a) { amount = a; }
   SpIntSem () {}
   int AddResource (int val);
   int RequestResource (int requestAmount, int eventType,
			SpProcStackElement *LVptr, int timeOut = 0);
   void RemoveBlockingProcess (SpProcStackElement *LVptr);
   int operator += (int d) { return AddResource (d); }
   operator int () { return amount; }
};

class SpFloatSem: public SpResourceSem
{
   RB_double amount;
 public:
   SpFloatSem (double a) { amount = a; }
   SpFloatSem () {}
   double AddResource (double val);
   int RequestResource (double requestAmount, int eventType,
			SpProcStackElement *LVptr, int timeOut = 0);
   void RemoveBlockingProcess (SpProcStackElement *LVptr);
   double operator += (double d) { return AddResource (d); }
   operator double () { return amount; }
};

// rollback items
//
//RB_DEFINE_CLASS_WITH_NO_DEFAULT_CONSTRUCTOR(SpWaitingItem);
//RB_DEFINE_CLASS_WITH_NO_DEFAULT_CONSTRUCTOR(SpIntWaitingItem);
//RB_DEFINE_CLASS_WITH_NO_DEFAULT_CONSTRUCTOR(SpFloatWaitingItem);
//RB_DEFINE_CLASS(SpWaitingItem);       // WITH_NO_DEFAULT_CONSTRUCTOR
//RB_DEFINE_CLASS(SpIntWaitingItem); // WITH_NO_DEFAULT_CONSTRUCTOR
//RB_DEFINE_CLASS(SpFloatWaitingItem); // WITH_NO_DEFAULT_CONSTRUCTOR


//RB_DEFINE_CLASS(SpLogicalSem);
//RB_DEFINE_CLASS(SpIntSem);
//RB_DEFINE_CLASS(SpFloatSem);

typedef SpLogicalSem LogicalSemaphore;
typedef SpIntSem     IntegerSemaphore;
typedef SpFloatSem   FloatSemaphore;

#endif
