#include "SpProcStackElement.H"
#include "RB_Types.H"
#include "RB_SpContainerFuncs.H"
#include "xqueue.H"
#include "SpAntiMsg.H"
#include "S_SpProc.H"
#include "SpProcSem.H"
#include "M_SpProcShSem.H"
#include "SpTimeMgtAlg.H"
#include "SpScheduleEvent.H"
#include "RB_SpDefineClass.H"
#include "SpAntiMsg.H"
#include "SpBinaryTree.H"
#include "SpPoDataBase.H"

#include "RB_SpBinaryTree.H"

#ifdef __IBMCPP__
namespace {
#endif

#ifndef PO_DEFINED_CHAR
RB_DEFINE_CLASS_PRIMATIVE(char);
#define PO_DEFINED_CHAR
#endif

#ifdef __IBMCPP__
}
#endif
RB_DEFINE_CLASS(Stack);
RB_DEFINE_CLASS(C_XQUEUE);
RB_DEFINE_CLASS(C_CANCEL_HANDLE);

// this class holds a stack memory segment for process model local
// variables which are saved during waits and resored upon reentry
class VarElement
{
   void *Buffer;    // allocated memory for saving the stack memory segment
   int   Size;      // size (which increases during stack segment consolidation)
   int   Offset;    // offset from the beginning of the stack
   int   Allocated; // whether the buffer has been allocated
 public:
  VarElement(): Allocated(0) {} //Added just for persistence
   VarElement(int initialSize): Size(initialSize), Offset(-10000), Allocated(0) {}
   ~VarElement() { if (Allocated) RB_DELETE_ARRAY_char ((char *) Buffer); }
   void AllocateBuffer() { Buffer = RB_NEW_ARRAY_char(Size); Allocated = 1; }
   void SetOffset(int offset) { Offset = offset; }
   void SetSize(int size) { Size = size; }
   int GetOffset() { return Offset; }
   int GetSize() { return Size; }
   void *GetBuffer() { return Buffer; }
};

RB_DEFINE_CLASS(VarElement); //Needs to be _WITH_NO_DEFAULT_CONSTRUCTOR

SpAskReturn::SpAskReturn(int askBytes): AskBytes(askBytes)
{
   RetEventType = GetEventType();
   RetObjHandle = GetObjHandle();
}
   
SpProcStackElement::SpProcStackElement(SpObjHandle *simObj, int eventType):
   ProcId(CurrentSimobj->RegisterProc()),
   EventType(eventType),
   Simobj(simObj),
   Random(ProcId ^ CurrentSimobj->GetGlobalId()),
   FirstVarPtr_1(NULL),
   FirstVarPtr_2(NULL),
   Vars(new RB_SpBinaryTree)
{
}
SpProcStackElement::~SpProcStackElement(){delete Vars;}

void SpProcStackElement::SaveVarState(void *varPtr, int size, int isFirstEntry)
{
   if (isFirstEntry)
   {
      VarElement *varElementPtr = RB_NEW_VarElement(new VarElement(size));
      Vars->Insert(varElementPtr, (int)varPtr);
      Cycle = 1;
      if (!FirstVarPtr_1)
      {
	 Cycle = 1;
	 FirstVarPtr_1 = varPtr;
      }
      return;
   }

   if (!FirstVarPtr_2 && (FirstVarPtr_1 != varPtr)) FirstVarPtr_2 = varPtr;
   if (varPtr == FirstVarPtr_1) Cycle = 1;
   else if (varPtr == FirstVarPtr_2) Cycle = 2;
   else
   {
      cout << "WARNING: should never happen: apparently more than "
	   << "two stack positions for processes"
	   << endl;
   }
}

void SpProcStackElement::ConsolidateVarsState()
{
   // set up the first memory segment pair out of the Vars tree
   VarElement *varElement1 = (VarElement *) Vars->GetFirstElement();
   if (!varElement1) return;

   // get the pointer to the first local var on the stack
   int varPtr1 = (int) Vars->GetCurrentKey();

   // set the offset of the first local var on the stack relative to
   // the first local var specified in the source code
   varElement1->SetOffset(varPtr1 - (int)FirstVarPtr_1);

   // consolidate contiguous memory segments into single items, or if
   // not contiguous, set the stack memory offsets between segments
   VarElement *varElement2 = (VarElement *)(Vars->GetNextElement());
   while (varElement2)
   {
      int varPtr2 = (int) Vars->GetCurrentKey();

      // if contiguous, consolidate the two memory segments into one
      if ((varPtr2 - varPtr1) == varElement1->GetSize())
      {
	 varElement1->SetSize(varElement1->GetSize() + varElement2->GetSize());
	 varElement2 = (VarElement *)(Vars->GetNextElement());
	 RB_DELETE_VarElement ((VarElement *) Vars->Remove(varPtr2));
      }
      
      // else, set the stack offset of this non-contiguous memory segment 
      else
      {
	 varElement2->SetOffset(varPtr2 - (int)FirstVarPtr_1);

	 // setup next memory segment pair
	 varPtr1 = varPtr2;
	 varElement1 = varElement2;
	 varElement2 = (VarElement *)(Vars->GetNextElement());
      }
   }

   // allocate memory for each consolidated memory segment
   // (must wait until after consolidating to allocate)
   VarElement *varElement = (VarElement *) Vars->GetFirstElement();
   while (varElement)
   {
      varElement->AllocateBuffer();
      varElement = (VarElement *)(Vars->GetNextElement());
   }
}

void SpProcStackElement::SaveContext(int rLabel)
{
   // set the reentry label for when this process returns
   reentryLabel = rLabel;

   // save the stack as formulated with the initial SaveVarState() and
   // ConsolidateVarsState() calls
   VarElement *varElement = (VarElement *) Vars->GetFirstElement();

   int stackOffset = 0;
   if (Cycle == 2) stackOffset = (int) FirstVarPtr_2 - (int) FirstVarPtr_1;
   while (varElement)
   {
      RB_memcpy(varElement->GetBuffer(),
		(void *)((int) Vars->GetCurrentKey() + stackOffset),
		varElement->GetSize());
      varElement = (VarElement *)(Vars->GetNextElement());
   }
}

void SpProcStackElement::LoadVarsState()
{
   VarElement *varElement = (VarElement *) Vars->GetFirstElement();

   void *firstVarPtr = (Cycle == 1) ? FirstVarPtr_1 : FirstVarPtr_2;
   while (varElement)
   {
      memcpy((void *)((int)firstVarPtr + varElement->GetOffset()),
	     varElement->GetBuffer(),
	     varElement->GetSize());
      varElement = (VarElement *)(Vars->GetNextElement());
   }
}

void SpProcStackElement::EstablishNewStack()
{
   stackPtr = RB_NEW_Stack();
   stackPtr->sePtr = RB_NEW_C_XQUEUE ();
   RB_PUSH_TOP (((S_SpProc *) CurrentSimobj)->threadList, stackPtr);
}

void SpProcStackElement::InitProc() { RB_PUSH_TOP (stackPtr->sePtr, this); }

void SpProcStackElement::AbortProc(char *name)
{
   cout << "unknown reentry order in "
	<< name
	<< " process event: "
	<< reentryLabel
	<< endl;
   exit (-1);
}

void SpProcStackElement::ScheduleTimeOut(const SIMTIME &currentTime,
					 M_SpProc &msg, double timeOut)
{
   SpCancelHandle *cancelHandleMemPtr;
   SpCancelHandle  cancelHandleValue;
   int cancelHandleFreeId;
   SpSimTime scheduleTime = currentTime;
   int oldType = msg.type; // for rollbacks
   SpProcStackElement *oldReentryLV = msg.reentryLV; // for rollbacks

   scheduleTime += timeOut;
   msg.type = TIMED_OUT_SEMAPHORE_RETURN;
   msg.reentryLV = this;
   cancelHandleValue = ReScheduleEvent (scheduleTime, &msg);
   msg.type = oldType;
   msg.reentryLV = oldReentryLV;

   cancelHandleMemPtr = RB_NEW_CANCEL_HANDLE();
   cancelHandleFreeId = cancelHandleMemPtr->get_freeid();
   *cancelHandleMemPtr = cancelHandleValue;
   cancelHandleMemPtr->set_freeid(cancelHandleFreeId);
   TimeOutCancelHandle = cancelHandleMemPtr;
}

void SpProcStackElement::HandleSemReturn(M_SpProc &msg)
{
   void *cancelHandleVoidPtr = TimeOutCancelHandle;
   SpCancelHandle *timeOutCancelHandle;

   timeOutCancelHandle = (SpCancelHandle *) cancelHandleVoidPtr;
   RB_DELETE_CANCEL_HANDLE(timeOutCancelHandle);

   if (msg.type == TIMED_OUT_SEMAPHORE_RETURN)
     SemaphorePtr->RemoveBlockingProcess(this);
}

void SpProcStackElement::CleanupProc()
{
   if (stackPtr->sePtr->get_length () == 0)
   {
      RB_DELETE_C_XQUEUE (stackPtr->sePtr);
      *((S_SpProc *) CurrentSimobj)->threadList -= (C_SQ_ITEM *) stackPtr;
      RB_DELETE_Stack (stackPtr);
   }
   while (Vars->GetNumElements())
     RB_DELETE_VarElement ((VarElement *) Vars->RemoveFirstElement());
}

void SpProcStackElement::ScheduleWaitReturn(M_SpProc &msg,
					    const SpSimTime &scheduleTime)
{
   int oldType = msg.type; // for rollbacks
   SpProcStackElement *oldReentryLV = msg.reentryLV; // for rollbacks

   msg.type = WAIT_RETURN;
   msg.reentryLV = this;
   ReScheduleEvent (scheduleTime, &msg);
   msg.type = oldType;
   msg.reentryLV = oldReentryLV;
}

// for backward compatibility:
void SpProcStackElement::HandleSemReturn(M_SpProc &msg, int &successFlag)
{
   void *cancelHandleVoidPtr = TimeOutCancelHandle;
   SpCancelHandle *timeOutCancelHandle;

   timeOutCancelHandle = (SpCancelHandle *) cancelHandleVoidPtr;
   RB_DELETE_CANCEL_HANDLE(timeOutCancelHandle);

   if (msg.type == TIMED_OUT_SEMAPHORE_RETURN)
     SemaphorePtr->RemoveBlockingProcess(this);
   successFlag = (msg.type == SUCCESSFUL_SEMAPHORE_RETURN);
}

void StackElement::SetUpAskData(void *askData, int askBytes,
				void *userData, int userBytes,
				void *&scheduleData, int &scheduleBytes)
{
   scheduleBytes = GetDataBytes() + askBytes + sizeof(SpAskReturn) + userBytes;
   scheduleData = malloc(scheduleBytes);

   memcpy(scheduleData, GetData(), GetDataBytes());

   int i = GetDataBytes();
   memcpy(&((char*)scheduleData)[i], askData, askBytes);

   /* plug this warning
   i += askBytes;
   memcpy(&((char*)scheduleData)[i], &SpAskReturn(askBytes),
	  sizeof(SpAskReturn));
   */
   i += sizeof(SpAskReturn);
   memcpy(&((char*)scheduleData)[i], userData, userBytes);
}

RB_DEFINE_CLASS(SpProcStackElement);

void RB_Register_VarElement()
{
  RB_REGISTER_CLASS_VarElement();
}

void RB_Register_Stack()
{
  RB_REGISTER_CLASS_Stack();
}

void RB_Register_SpProcStackElement()
{
  RB_REGISTER_CLASS_SpProcStackElement();
}


