// SpProcStackElement.H header file (types for SpProc.H and reentryMsg.H)

#ifndef ProcessTypes_H
#define ProcessTypes_H

#include "SpObjHandle.H"
#include "SpRandom.H"
class RB_SpBinaryTree;

class SpEvent;
struct M_SpProc;
class SpProcSem;

#include "SpAntiMsg.H"
#include "SpNetTypes.H"

/***************************************************************************
 * process model support structures:                                        *
 ***************************************************************************/

enum
{
   BEGIN_PROCESS,
   END_PROCESS,
   WAIT_RETURN,
   SUCCESSFUL_SEMAPHORE_RETURN,
   TIMED_OUT_SEMAPHORE_RETURN
};

class SpAskReturn
{
 private:
   int RetEventType;
   SpObjHandle RetObjHandle;
   int AskBytes;
 public:
   SpAskReturn(){} // Added just for persistence
   SpAskReturn(int askBytes);
   int GetRetEventType()  { return RetEventType; }
   SpObjHandle GetRetObjHandle()  { return RetObjHandle; }
   int GetAskArgBytes() { return AskBytes; }
};

struct Reentry_T
{
  SpObjHandle obj;
  NET_INT eventType;
  NET_DOUBLE lookahead;
};

struct Stack: public C_SQ_ITEM
{
   C_XQUEUE *sePtr; // list of SpProcStackElement's
};

//RB_DEFINE_CLASS(Stack); // Change back to
                                     // _WITH_NO_DEFAULT_CONSTRUCTOR when
				     // persistence supports it.
//RB_DEFINE_CLASS(Stack);

struct SpProcStackElement: public C_SQ_ITEM
{
   SpProcStackElement():Vars(NULL){} //Need to remove when persistence supports
   ~SpProcStackElement();
   SpProcStackElement(SpObjHandle *simobj, int eventType);
   void SaveVarState(void *varPtr, int size, int isFirstEntry);
   void ConsolidateVarsState();
   void LoadVarsState();
   void InitProc();
   void AbortProc(char *name = NULL);
   void EstablishNewStack();
   void SaveContext(int rLabel);
   void ScheduleTimeOut(const SpSimTime &currentTime,
			M_SpProc &msg, double timeOut);
   void HandleSemReturn(M_SpProc &msg);
   void ScheduleWaitReturn(M_SpProc &msg, const SpSimTime &scheduleTime);
   void CleanupProc();
   void SetUpAskData(void *askData, int askBytes,
		     void *userData, int userBytes,
		     void *&scheduleData, int &scheduleBytes);

   // for backward compatibility:
   void HandleSemReturn(M_SpProc &msg, int &successFlag);

   // the following are used for the present event

   int           ProcId;
   int           EventType;         // enum of event (event id) for this process
   Stack        *stackPtr;          // stack on which is this SpProcStackElement

   SpObjHandle  *Simobj;             // simobj on which this process acts
   SpEvent      *Event;              // this pointer of the process of this stack
   SpProcSem    *SemaphorePtr;       // save the semaphore for semaphore timeouts

   RB_int       finiteTimeOut;       // whether a semaphore block has a time out
   RB_pointer   TimeOutCancelHandle; // cancel handle for semaphore timeouts
   RB_SpRandom  Random;              // the simobj's random number generator class

   RB_int       reentryLabel;        // integer to goto correct point in the code
                                     //    (set before blocking, use at reentry)
   RB_SpBinaryTree* Vars;            // local variable info
   void *FirstVarPtr_1;              // ptr to the first stack variable listed
                                     //    in the source code
   void *FirstVarPtr_2;              // Previous frst var ptr. This is relavent
                                     //    during the first save when the stack
                                     //    location changes
   RB_int Cycle;                     // whether in cycle 1 or 2, corresponding
                                     //    to FirstVarPtr_1 or FirstVarPtr_2,
                                     //    respectively
};

//RB_DEFINE_CLASS(SpProcStackElement);

typedef SpProcStackElement StackElement;

#endif
