// SpBTW.C method file

#include "SpStdio.H"
#include "SpTime.H"
#include "speedes_comm.H"
#include "SpBTW.H"
#include "host_user.H"
#include "SpBarrier.H"
#include "SpClusterMgr.H"

#include "Merc_types.H"

#include "SpWalltime.H"
#include "RB_SpBinaryTree.H"

#define OPTIMIZE
#ifndef INFINITY
#define INFINITY 1.0e20
#endif

#define BROAD

extern SpEvent *CurrentEvent;
extern SpSimObj *CurrentSimobj;
extern SpTag *CurrentTag;

extern int SpGateway;

extern char *TimeManagementMode;
extern RB_SpBinaryTree *Barriers;
extern SpSimTime *IgnoreTimeGrantGvt;
extern SpClusterMgr *GlobalClusterMgr;

/************************************************************************
* SpBTW : construct an event queue object				*
************************************************************************/
SpBTW::SpBTW() {
  SpEvent *event;

/*
  printf("BREATHING TIME WARP EVTQ created with Time Management: %s\n",
	 TimeManagementMode);

  if (node == 0) {
    printf("Tgvt=%f, Tasb=%f, Ngvt=%d, Nrisk=%d, Nopt=%d\n",
	Tgvt, Tasb, Ngvt, Nrisk, Nopt);
  }
*/
  
  SpCurrentAltMgr->SetOptimistic();

  while (n_items) {
    event = (SpEvent *)pop_top();
    event->insert();
  }

  localt = -INFINITY;

  twosmess = new SpOptMessager(node, n_nodes, evtype);
  SpEvent::set_twosmess(twosmess);
  SpEvent::set_one_node_flag(0);

  gvtupdate = 0;
  nevtgvt = 0;

  SpEvent::set_qproc(&dbl_qproc);
  if (lazy) SpEvent::enable_lazy();

  asbtime = SpWalltime();
  asb_cycle = 0;
  nasb = 0;

  StopWatchBTW.UseWallTime();
  NumProcConservative = 0;
  IgnoreTimeGrantGvt = new SpSimTime();
  *IgnoreTimeGrantGvt = 1.0e20;

}

/************************************************************************
* temp_process - process events in temporary storage			*
************************************************************************/
void SpBTW::temp_process() {
  C_DBL_ITEM *dbl_item;
  SpEvent *event;
  int num;
  int i, j;
  double t0;
  double t1;
  double Tgvt_Dummy;
  double Tasb_Dummy;
  double BTB_Time;
  double FBen_Time = 0;
  double FBex_Time;

/*
** Start timeer for BTW mode
*/

  BTB_Time = -1.0;
  StopWatchBTW.StartTimer();

//...... initialize some values

  t0 = SpWalltime();
  t1 = t0;

  if (Tgvt <=0) {
    Tgvt_Dummy = 1.0e20;
  }else{
    Tgvt_Dummy = Tgvt;
  }

  if (Tasb < 0) {
    Tasb_Dummy = 1.0e20;
  }else{
    Tasb_Dummy = Tasb;
  }

  asb_cycle++;

  n_temp = 0;
  n_perm = 0;
  n_roll = 0;
  n_not_valid = 0;
  mess_count = 0;
  n_mrec = 0;
  n_messrec = 0;
  n_amessrec = 0;
  n_exmess = 0;

  NumProcConservative = 0;
  NoValidEvents = 0;
  GvtPlusL = gvt;
  GvtPlusL.SetTime(gvt+minstep);

//...... localt must be gvt (don't play with this) and tmin set to tmin

  tmin = tend;
  localt = gvt;

//...... loop over the already processed events to find the starting horizon

  event_horizon = tend;
  glob_eh = tend;

  j = 0;
  num = dbl_qproc.get_length();
  dbl_item = (C_DBL_ITEM *)dbl_qproc.get_top();
  for (i=0; i<num; i++) {
    event = (SpEvent *)dbl_item->get_ptr();
    if (event->get_valid()) {
      if (GvtPlusL < event->get_time_tag()) {
	j++;
      }else{
	NumProcConservative++;
      }
      if (j < Nrisk) {
        twosmess->send_messages(event);
        mess_count++;
      }else{
        if (event_horizon < event->get_time_tag()) break;
        tmin = event->get_tmin();
        if (tmin < event_horizon) event_horizon = tmin;
      }
    }
    dbl_item = dbl_item->get_next();
  }

//...... now process events according to the Breathing Time Warp strategy

  state = 0;

// The FB_Flag makes sure we do at least one iteration before exiting
// the processing loop. We do not try to enter the fuzzy barrier until the
// flag gets set.

  int FB_Flag = 0;

  for (;;) {

    if (Tgvt > 0) t1 = SpWalltime();

//...... check if we are ready to make a GVT request on this node

    if (state == 0) {
	if (	(Ngvt <= dbl_qproc.get_length()) ||
		(NoValidEvents) ||
      		(Ngvt < n_mrec) ||
      		((t1 - t0) > Tgvt_Dummy) ||
      		(event_horizon < localt)	) {
	  if (FB_Flag) {
	    SpComm_EnterFuzzyBarrier();
	    FBen_Time = StopWatchBTW.StopTimer(); // Timer for entering F.B.
	    state = 1;
	  }else{
	    if (rqueue.get_time() > gvt) {
	      FB_Flag = 1;
	    }
	  }
	}
    }

//...... check if all of the nodes have made their request

    if (state == 1) {
      if (SpComm_ExitFuzzyBarrier()){
	FBex_Time = StopWatchBTW.StopTimer();
	if (BTB_Time == -1.0)  BTB_Time = FBex_Time;
	statistics->SetPhaseBTB(BTB_Time);
	statistics->SetPhaseFBen(FBen_Time);
	statistics->SetPhaseFBex(FBex_Time);
	break;
      }
    }

/*
** Timer for BTB statistics
*/

    if (BTB_Time == -1.0) {
      if (dbl_qproc.get_length() >= Nrisk) {
	BTB_Time = StopWatchBTW.StopTimer();
      }
    }

//...... process and event, read messages, and then read antimessages

    process_event();
    SpComm_Receive();
    messages();
    antimessages();

#ifdef BROAD

//...... read incoming messages

    int nbytes;
    while ((abroadin = (ABROAD *)
	    SpComm_GetPendingMessage(ASYNC_BROADCAST_MESSAGE_TYPE,
				     nbytes))){

      if (abroadin->cycle == asb_cycle) {
        if (abroadin->time < glob_eh) {
          glob_eh = abroadin->time;
	  if (glob_eh < event_horizon) event_horizon = glob_eh;
        }
        nasb++;
      }

    }

//...... send out messages if necessary

    if (event_horizon < localt) {
      if (event_horizon < glob_eh) {
	if ((Tgvt <= 0) && (Tasb > 0)) t1 = SpWalltime();
        if ((t1-asbtime) > Tasb_Dummy) {
	  abroadout.time = event_horizon; //This is the buffer to send
	  abroadout.cycle = asb_cycle;
	  SpComm_Send(ASYNC_BROADCAST_MESSAGE_TYPE,
		      sizeof(ABROAD),
		      (char *)&abroadout);
	  asbtime = t1;
	  glob_eh = event_horizon;


        }
      }
    }

#endif


  }

}

/************************************************************************
* process_event - process the current event optimistically		*
************************************************************************/
void SpBTW::process_event() {
  SpEvent *lastevent;
  SpEvent *event;
  C_DBL_ITEM *dbl_item;
  SpSimTime tlastevent; //GEB2
  double CPU_Time;
  SpSimObjMgr *objman;
  int conservativeEventFlag;

//...... get the next event (unless we have gone Nopt events beyond GVT)

  for (;;) {

//...... check if we have already reached Nopt events

    if (Nopt <= (dbl_qproc.get_length() - NumProcConservative)) {
      if ((GvtPlusL.GetTime()) < (rqueue.get_time()).GetTime()) {
	dbl_item = (C_DBL_ITEM *)dbl_qproc.get_bot();
	lastevent = (SpEvent *)dbl_item->get_ptr();
	tlastevent = lastevent->get_time_tag();
	if (tlastevent.GetTime() < (rqueue.get_time()).GetTime()) {
	  if (0.0 <= rqueue.get_time()) {
	    NoValidEvents = 1;
	    return;
	  }
	}
      }
    }

    if ((Nopt == 0) &&
	(0.0 <= rqueue.get_time()) &&
	(rqueue.get_time() != gvt) &&
	(GvtPlusL <= rqueue.get_time())) {
      NoValidEvents = 1;
      return;
    }

    event = get_next_event();

    if (event == NULL) {
      localt = INFINITY;
      NoValidEvents = 1;
#ifdef NOISY
      fprintf(stderr,"No more events on node %d\n",node);
#endif
      return;
    }

    if (event->get_time_tag() <= (GvtPlusL)) {
      if (GvtPlusL < 0.0) {
	conservativeEventFlag = 0;
      }else{
	conservativeEventFlag = 1;
      }
    }else{
      conservativeEventFlag = 0;
    }

    if ((Nopt == 0) &&
	(!conservativeEventFlag) &&
	(0.0 <= event->get_time_tag())) {
      cout << "Error, should not be processing optimistically" << endl;
      cout << "GvtPlusL: " << GvtPlusL << endl;
      cout << "Event time: " << event->get_time_tag() << endl;
    }

    CurrentEvent = event;
    CurrentSimobj = event->get_SIMOBJ();
    CurrentTag = event->GetTag();
    CurrentTag->Reset();
    event->init_qroll_backs();

    if (event->check_canceled(n_roll)) {

      event->set_processed();
      event->calculate_tmin();
      dbl_item = event->get_dbl_item();
      dbl_item->set_ptr(event);
      dbl_item->set_valu(event->get_time_tag());
      dbl_qproc.insert(dbl_item);

    }else{

      if (event->get_valid()) {
        break;
      }

      evtype->delete_event(event);

    }

  }

//...... make sure that this event is not in the rollback list

//  event->check_proc(0);

//...... make sure that this event does not have a time tag less than GVT

  if (event->get_time_tag() < gvt) {
    cout << "---| Warning (BTW_EVTQ): Event Time < GVT" << endl;
    cout << "---| Event type = " << event->get_EVENT_TYPE() << endl;
    cout << "---| Object type = " << event->get_OBJECT_TYPE() << endl;
    cout << "---| Tevt= " << event->get_time_tag() << endl;
    cout << "---| GVT = " << gvt << endl;
    cout << "---| ";
    event->PrintTimeTag();
  }

//...... check if it is appropriate to process this event

  if (event->get_time_tag() > tend) {
    localt = INFINITY;
    event->insert();
    NoValidEvents = 1;
#ifdef NOISY
    cout << "Event time is beyond tend on node" << node << endl;
#endif
    return;
  }

//...... define the local simulation time

#ifdef COMMENTOUT
  if (n_nodes == 1) {
    if (event->get_time_tag() < localt) {
      cerr << "Error, (BREATHING TIME WARP) events out of order\nlocalt = "
	   << localt << "\nEvent = " << (event->get_time_tag()) << endl;
      abort();
    }
  }
#endif

  localt = event->get_time_tag();

//...... process the event

  if (event->get_lazy_proc()) {
    if (event->lazy()) {
      event->init_qroll_backs();
      SpCurrentAltMgr->Rollback();
      event->exchange();
    }else{
      n_roll += event->lazy_rollback();
      event->init_qroll_backs();
//      event->check_clear();

      StopWatch.StartTimer();
 
     if (conservativeEventFlag) {
//	SpCurrentAltMgr->UnsetOptimistic();
	SpCurrentAltMgr->SetOptimistic();
      }else{
	SpCurrentAltMgr->SetOptimistic();
      }

      event->temp_process();
      event->TieBreakGeneratedMessages();
      event->spin_loop(spin);
      event->exchange();
      CPU_Time = StopWatch.StopTimer();

      TotalProcessingTime += CPU_Time;

      event->SetCPU_Time(CPU_Time);

      NumEventsProcessed[event->get_EVENT_TYPE()]++;
      CPU_EventTime[event->get_EVENT_TYPE()] += CPU_Time;

      CurrentSimobj->IncrementTotalTime(CPU_Time);
      CurrentSimobj->IncrementTotalEvents();

      CurrentSimobj->IncrementProcessingTime(CPU_Time);
      CurrentSimobj->IncrementCriticalPathTime(
		(CurrentEvent->get_MESSAGE())->EarliestStartTime,
		CPU_Time);

      event->CriticalPathMessages(CurrentSimobj->GetCriticalPathTime());

      event->SetProcessingTime(CurrentSimobj->GetProcessingTime());
      event->SetCriticalPathTime(CurrentSimobj->GetCriticalPathTime());
      event->SetTotalTime(CurrentSimobj->GetTotalTime());

      objman = (SpSimObjMgr *)objects->
	get_objman(CurrentSimobj->get_OBJECT_TYPE());

      if (objman) {
	objman->IncrementCombinedTotalEvents();
	objman->IncrementCombinedTotalTime(CPU_Time);
      }

    }

  }else{
    event->init_qroll_backs();
//    event->check_clear();

    StopWatch.StartTimer();

    if (conservativeEventFlag) {
//      SpCurrentAltMgr->UnsetOptimistic();
      SpCurrentAltMgr->SetOptimistic();
    }else{
      SpCurrentAltMgr->SetOptimistic();
    }

    event->temp_process();
    event->TieBreakGeneratedMessages();
    event->spin_loop(spin);
    event->exchange();
    CPU_Time = StopWatch.StopTimer();

    TotalProcessingTime += CPU_Time;

    event->SetCPU_Time(CPU_Time);

    NumEventsProcessed[event->get_EVENT_TYPE()]++;
    CPU_EventTime[event->get_EVENT_TYPE()] += CPU_Time;

    CurrentSimobj->IncrementTotalTime(CPU_Time);
    CurrentSimobj->IncrementTotalEvents();

    CurrentSimobj->IncrementProcessingTime(CPU_Time);
    CurrentSimobj->IncrementCriticalPathTime(
		(CurrentEvent->get_MESSAGE())->EarliestStartTime,
		CPU_Time);

    event->CriticalPathMessages(CurrentSimobj->GetCriticalPathTime());
    
    event->SetProcessingTime(CurrentSimobj->GetProcessingTime());
    event->SetCriticalPathTime(CurrentSimobj->GetCriticalPathTime());
    event->SetTotalTime(CurrentSimobj->GetTotalTime());

    objman = (SpSimObjMgr *)objects->
      get_objman(CurrentSimobj->get_OBJECT_TYPE());

    if (objman) {
      objman->IncrementCombinedTotalEvents();
      objman->IncrementCombinedTotalTime(CPU_Time);
    }

  }

//...... test state change in rollback

  if (test_rollback && (!conservativeEventFlag)) {
    event->init_qroll_backs();
    event->exchange();
    SpCurrentAltMgr->Rollback();
    SpCurrentAltMgr->Rollback();
    event->exchange();
  }

//...... mark the event as processed

  event->calculate_tmin();
  tmin = event->get_tmin();
  if (tmin < event->get_time_tag()){
    cout << "Error, (BREATHING TIME WARP) event generated in past: "
	 << endl
	 << "event time = "
	 << event->get_time_tag()
	 << endl
	 << "      tmin = "
	 << tmin
	 << endl;
    event->PrintScheduledEvents();
  }

  event->set_processed();
  if (event->get_lazy()) event->set_lazy_proc();

/*
** Modify message time tags if minstep (lookahead) is not zero
*/

  if (minstep > 0.0) {
    C_QUEUE *qmess = event->get_mque();
    int len = qmess->length();
    C_HOLDER *holder = (C_HOLDER *)qmess->get_top();
    for (int i=0; i<len; i++) {
      char *message = holder->get_buff();
      SpMsg *header = (SpMsg *)message;
      if (node != holder->get_node()) {
	if (header->time_tag < (event->get_time_tag() + minstep)) {
	  header->time_tag.SetTime(event->get_time_tag() + minstep);
	}
      }
      holder = (C_HOLDER *)holder->get_link();
    }
  }

//...... send the messages if appropriate

  n_roll += event->local_messages();

  if (n_nodes > 1) {
    if ((((dbl_qproc.get_length() - NumProcConservative) < Nrisk)
	&& !gvtupdate) || (conservativeEventFlag)){
      twosmess->send_messages(event);
      mess_count++;
    }
  }

  event->calculate_tmin();
  tmin = event->get_tmin();
  if (tmin < event_horizon) event_horizon = tmin;

  dbl_item = event->get_dbl_item();
  dbl_item->set_ptr(event);
  dbl_item->set_valu(event->get_time_tag());
  dbl_qproc.insert(dbl_item);

  if (conservativeEventFlag) NumProcConservative++;
  n_temp++;
  if (gvtupdate) nevtgvt++;

//...... make sure that this event is now in the rollback list

//  event->check_proc(1);

}

/************************************************************************
* messages - receive incomming messages 				*
************************************************************************/
void SpBTW::messages() {
  SpEvent *event;
  int status;

//...... check if there are any incoming messages

  event = twosmess->receive_message(status);

  while (status) {

    n_mrec++;
    n_messrec++;

    if (event != NULL) {
      n_roll += event->optimistic_insert();
      NumMessages[event->get_EVENT_TYPE()]++;
    }

    event = twosmess->receive_message(status);

  }

//...... handle externally generated events

  event = external_event(GvtPlusL);
  while (event != NULL) {
    n_exmess++;
    n_roll += event->optimistic_insert();
    event = external_event(GvtPlusL);
  }

#ifdef COMMENTOUT
  tblock = objects->get_tblock();
  if (tblock < event_horizon) event_horizon = tblock;
#endif

}

/************************************************************************
* antimessages - receive incomming antimessages 			*
************************************************************************/
void SpBTW::antimessages() {
  SpEvent *event;
  int status;

  event = twosmess->receive_antimessage(status);

  while (status) {
    n_mrec++;
    n_amessrec++;
    if (event != NULL) {
      n_roll += event->cancel_event();
      NumAntimessages[event->get_EVENT_TYPE()]++;
    }
    event = twosmess->receive_antimessage(status);
  }

}

/************************************************************************
* find_gvt - find the minimum global time for safe processing		*
************************************************************************/
void SpBTW::find_gvt() {
  int nits;
  SpSimTime *NamedPauseTime;
  SpSimTime *BarrierTime;
  SpBarrier *barrier;
  double GVT_Time;

//...... flush out all of the messages

  gvtupdate = 1;
  nits = 0;

  for (;;) {

    SpComm_Receive();
    messages();
    antimessages();

    SpComm_EnterFuzzyBarrier();
    while (!SpComm_ExitFuzzyBarrier()) {
      process_event();

      SpComm_Receive();
      messages();
      antimessages();
    }
    
    nits++;

    if (twosmess->check_messcount() == 0) break;

  }

  statistics->set_nevtgvt(nevtgvt);

/*
  printf("Number of GVT iterations = %d, produced %d events\n", nits, nevtgvt);
*/

  if (!PauseFlag) {
    gvt = rqueue.get_time();
//    cout << "gvt from rqueue = " << gvt << endl;
  }

  if (NamedPauses.GetNumElements()) {
    NamedPauseTime = (SpSimTime *)NamedPauses.GetFirstElement();
    while (NamedPauseTime) {
      if (*NamedPauseTime < gvt) gvt = *NamedPauseTime;
      NamedPauseTime = (SpSimTime *) ++NamedPauses;
    }
  }

  if (event_horizon < gvt) {
    gvt = event_horizon;
//    cout << "gvt from event horizon" << gvt << endl;
  }

  if (gvt > tend) gvt = tend;
  if (gvt > get_gvtmax()) {
    gvt = get_gvtmax();
    gvt.SetPriority1(0x80000000);
    gvt.SetPriority2(0x80000000);
    gvt.SetCounter  (0x80000000);
    gvt.SetUniqueId (0x80000000);
  }
  
//...... check if there are any barriers set up for external messages

  if (GvtBarriers.GetNumElements()) {
    BarrierTime = (SpSimTime *)GvtBarriers.GetFirstElement();
    while (BarrierTime) {
      if (*BarrierTime < gvt) gvt = *BarrierTime;
      BarrierTime = (SpSimTime *) ++GvtBarriers;
    }
  }

/*
** IMPORTANT NOTE: THIS MUST BE THE LAST THING DONE IN CALCULATING GVT
** BEFORE THE GLOBAL MIN!!!
**
** Do special barriers that are constructed by events in a rollbackable manner.
** the challenging thing is to handle gateway barriers special. When a
** gateway barrier it established because of a time grant, then we want to
** capture the earliest outgoing message to the gateway. This is very delicate
** stuff.
*/

  *IgnoreTimeGrantGvt = gvt;

  if (Barriers->GetNumElements()) {

    barrier = (SpBarrier *)Barriers->GetFirstElement();
    while (barrier) {

      if (barrier->GetGateway()) {
	if (barrier->Check_TIME_GRANT()) {
	  if (barrier->GetMinGatewayMsgTime() < *IgnoreTimeGrantGvt) {
	    *IgnoreTimeGrantGvt = barrier->GetMinGatewayMsgTime();
	  }
	}else{
	  if (barrier->GetTimeMinusLag() < *IgnoreTimeGrantGvt) {
	    *IgnoreTimeGrantGvt = barrier->GetTimeMinusLag();
	  }
	}
      }else{
	if (*barrier < *IgnoreTimeGrantGvt) {
	  *IgnoreTimeGrantGvt = *barrier;
	}
      }

      if (*barrier < gvt) gvt = *barrier;
      barrier = (SpBarrier *) ++(*Barriers);

    }

  }

/*
** GlobalMin for updating gvt
*/

  gvt = SpComm_GlobalMin(gvt);

/*
** Update IgnoreTimeGrantGvt - only do this only if we are supporting the
** HLA gateway.
*/

  if (SpGateway) {
    *IgnoreTimeGrantGvt = SpComm_GlobalMin(*IgnoreTimeGrantGvt);
  }

  gvtupdate = 0;

  GVT_Time = StopWatchBTW.StopTimer();
  statistics->SetPhaseGVT(GVT_Time);

}

/************************************************************************
* perm_process - carry out the events and make them permanent		*
************************************************************************/
void SpBTW::perm_process() {
  C_DBL_ITEM *dbl_item;
  SpEvent *event = NULL;
  int num;
  int i;
  char *message;
  SpMsg *header;
  C_HOLDER *holder;
  int size;
  double CPU_Time;
  SpSimObjMgr *objman;
  double CMT_Time;

  num = dbl_qproc.get_length();
  n_temp = num;

//...... loop over all of the events that can be processed

  for (i=0; i<num; i++) {

    dbl_item = (C_DBL_ITEM *)dbl_qproc.pop_top();
    event = (SpEvent *)dbl_item->get_ptr();

    if (event->get_valid()) {
      if (gvt < event->get_time_tag()) {
        dbl_qproc.push_top(dbl_item);
        break;
      }else{
	if (event->get_processed()) {
	  if (event->get_event_canceled()){
	    ++n_cancel;
	  }
	  else{
	    event->perm_process();

	    CurrentSimobj = event->get_SIMOBJ();
	    CPU_Time = event->GetCPU_Time();
	    TotalCommittedTime += CPU_Time;
	    CurrentSimobj->IncrementCommittedTime(CPU_Time);
	    CurrentSimobj->IncrementCommittedEvents();

	    NumEventsCommitted[event->get_EVENT_TYPE()]++;
	    CPU_EventTimeCommitted[event->get_EVENT_TYPE()] += CPU_Time;

	    objman = (SpSimObjMgr *)objects->
	      get_objman(CurrentSimobj->get_OBJECT_TYPE());

	    if (objman) {
	      objman->IncrementCombinedCommittedEvents();
	      objman->IncrementCombinedCommittedTime(CPU_Time);
	    }

            n_perm++;
	  }
	  event->trace();
	  event->sendmess(&q1nd);
        }else{
	  if (event->get_time_tag() == gvt) {
	    dbl_qproc.push_top(dbl_item);
	    break;
	  }else{
	    fprintf(stderr,
	    "Error, (BREATHING TIME WARP) event not processed %f %f\n",
		    event->get_time_tag()+0.0, gvt+0.0);
	  }
        }
      }
    }

    event->init_qroll_backs();
    SpCurrentAltMgr->Cleanup();
    evtype->delete_event(event);

  }

  while (q1nd.length()) {
    holder = (C_HOLDER *)q1nd.pop_top();
    message = holder->get_buff();
    header = (SpMsg *)message;
    if (header->ext) {
      n_exmess++;
      host_user->send_message((C_EM_HEADER *)message);
      evtype->delete_message(holder);
    }else{
      event = (SpEvent *)evtype->message_event(message);
      evtype->delete_message(holder);
    }
    n_roll += event->optimistic_insert();
  }

//...... get the incomming messages and turn them into future events

  if (n_nodes > 1) {

    while((message = (char *)SpComm_CoordinatedReceive(size)) != NULL) {

      header = (SpMsg *)message;

      header->evtype = GlobalClusterMgr->
	ConvertGlobalToLocalEventId(header->evtype);

      header->obtype = GlobalClusterMgr->
	ConvertGlobalToLocalMgrId(header->obtype);

      if (header->ext) {
	n_exmess++;
        host_user->send_message((C_EM_HEADER *)message);
      }else{
	n_messrec++;
        event = (SpEvent *)evtype->message_event(message);
	event->set_from_SpComm();
        n_roll += event->optimistic_insert();
	NumCoordMessages[event->get_EVENT_TYPE()]++;
      }
    }

  }

  CMT_Time = StopWatchBTW.StopTimer();
  statistics->SetPhaseCMT(CMT_Time);

}


int SpBTW::RollBack()
{
  int nRoll = 0;

  while (dbl_qproc.get_length()){
    C_DBL_ITEM* dblPtr = dbl_qproc.get_bot();
    SpEvent* eventPtr = (SpEvent *)dblPtr->get_ptr();
    C_DBL_QUEUE * q = eventPtr->get_SIMOBJ()->get_evtq();
    while (!q->get_length()){
      if (dblPtr == dbl_qproc.get_top()){
	return nRoll;
      }
      dblPtr = dblPtr->get_previous();
      eventPtr = (SpEvent *)dblPtr->get_ptr();
      q = eventPtr->get_SIMOBJ()->get_evtq();
    }
    while (q->get_length()){
      SpEventStatus* temp = (SpEventStatus *)q->get_bot();
      SpEvent *e = (SpEvent *)temp->get_event();
      q->pop_bot();
      nRoll += e->rollback_event();
      e->reset_processed();
    }
  }
  //cout<<"Rolled back "<<nRoll<<" events on node "<<SpComm_GetNodeId()<<endl;
  return nRoll;
}


int SpBTW::RollsForward()
{
  cout<<"SpBTW::RollForward() Not yet implemented!!!!!!!!!!!!!"<<endl;
  return 0;
}
