// SpSequential.C method file

#include "SpStdio.H"
#include "speedes_comm.H"
#include "SpSequential.H"
#include "host_user.H"
#include "E_SpInitial.H"

#define OPTIMIZE
#ifndef INFINITY
#define INFINITY 1.0e20
#endif

extern SpEvent *CurrentEvent;
extern SpSimObj *CurrentSimobj;
extern SpTag *CurrentTag;

/************************************************************************
* SpSequential : construct an event queue object			*
************************************************************************/
SpSequential::SpSequential() {
  SpEvent *event;

  printf("SEQUENTIAL event queue created\n");

  while (n_items) {
    event = (SpEvent *)pop_top();
    event->insert();
  }

  localt = -INFINITY;
  localt.SetPriority1(0x80000000);
  localt.SetPriority2(0x80000000);
  localt.SetCounter(0x80000000);
  localt.SetUniqueId(0x80000000);

}

/************************************************************************
* temp_process - process events in temporary storage			*
************************************************************************/
void SpSequential::temp_process() {
  int i,len;
  double tblock;
  SpEvent *event;
  SpEvent *new1_event;
  SpEvent *ext_event;
  C_QUEUE *qmess;
  C_HOLDER *holder;
  SpMsg *header;
  char *message;
  SpSimObj *obj;
  double tmin;
  double CPU_Time;
  SpSimObjMgr *objman;

//...... loop until the speedes tree needs to be resorted

  n_temp = 0;
  n_perm = 0;
  tblock = objects->get_tblock();

  while (n_temp < 500) {

//...... process the next event

    event = get_next_event();

//...... check for termination conditions

    if (event == NULL) {
      localt = tend;
      return;
    }

    if (event->get_time_tag() > tend) {
      localt = tend;
      return;
    }

    if (event->get_time_tag() > tblock) {
      event->insert();
      return;
    }

//...... define simulation time to be the time of this event

    if (localt > event->get_time_tag()) {
      fprintf(stderr,"Error, event list out of order\n");
    }

    localt = event->get_time_tag();

//...... check if the event is canceled

    if (event->check_canceled(n_roll)) {

      n_cancel++;

    }else{

//...... check reptime

      obj = event->get_SIMOBJ();
      if (obj->get_TIME() == event->get_time_tag()) {
        if (check_reptime) {
          if (event->get_time_tag() > 0.0) {
	    fprintf(stderr,"Warning, events %4d %4d with same time tag %f\n",
		    obj->get_last_etype(),
		    event->get_EVENT_TYPE(),
		    double(event->get_time_tag())); //GEB2
          }
        }
      }
      obj->set_TIME(event->get_time_tag());
      obj->set_last_etype(event->get_EVENT_TYPE());

//...... phase one event processing and delta exchange

      CurrentEvent = event;
      CurrentSimobj = event->get_SIMOBJ();
      CurrentTag = event->GetTag();
      CurrentTag->Reset();

      event->init_qroll_backs();

      StopWatch.StartTimer();
      event->temp_process();
      event->TieBreakGeneratedMessages();
      event->spin_loop(spin);
      event->exchange();
      CPU_Time = StopWatch.StopTimer();

      TotalProcessingTime += CPU_Time;

      event->SetCPU_Time(CPU_Time);

      NumEventsProcessed[event->get_EVENT_TYPE()]++;
      CPU_EventTime[event->get_EVENT_TYPE()] += CPU_Time;

      CurrentSimobj->IncrementTotalTime(CPU_Time);
      CurrentSimobj->IncrementTotalEvents();

      CurrentSimobj->IncrementProcessingTime(CPU_Time);
      CurrentSimobj->IncrementCriticalPathTime(
		(CurrentEvent->get_MESSAGE())->EarliestStartTime,
		CPU_Time);

      event->CriticalPathMessages(CurrentSimobj->GetCriticalPathTime());

      event->SetProcessingTime(CurrentSimobj->GetProcessingTime());
      event->SetCriticalPathTime(CurrentSimobj->GetCriticalPathTime());
      event->SetTotalTime(CurrentSimobj->GetTotalTime());

      objman = (SpSimObjMgr *)objects->
	get_objman(CurrentSimobj->get_OBJECT_TYPE());

      if (objman) {
	objman->IncrementCombinedTotalEvents();
	objman->IncrementCombinedTotalTime(CPU_Time);
      }

      n_temp++;

/*
** Trap the event and print out messages generated if it is the right type.
** This should not be deleted since it is useful for debugging. By setting
** the TrappedEventName, people can printout (or trap) their events.
*/

#ifdef COMMENTOUT

      char *TrappedEventName = "RadarPrecomp";

      if (!strcmp(evtype->get_messname(event->get_EVENT_TYPE()),
		  TrappedEventName)) {
        qmess = event->get_mque();
	cout << "Printing trapped event "
	     << evtype->get_messname(event->get_EVENT_TYPE())
	     << " : "
	     << qmess->length() << " messages generated" << endl;

	len = qmess->length();
	holder = (C_HOLDER *)qmess->get_top();
	for (i=0; i<len; i++) {
	  message = holder->get_buff();
	  header = (SpMsg *)message;
	  cout << "   Message[" << i << "] = "
	       << evtype->get_messname(header->evtype)
	       << endl;
	  holder = (C_HOLDER *)holder->get_link();
	}
      }

#endif

//...... test rollback if flag set

      if (test_rollback && (event->get_EVENT_TYPE() != INITIAL_EVENT)) {

        qmess = event->get_mque();
        len = qmess->length();
        event->exchange();
        SpCurrentAltMgr->Rollback();
        SpCurrentAltMgr->Cleanup();
        event->clear_messages();

	StopWatch.StartTimer();
        event->temp_process();
	event->TieBreakGeneratedMessages();
	event->spin_loop(spin);
        event->exchange();
	CPU_Time = StopWatch.StopTimer();

	TotalProcessingTime += CPU_Time;

	event->SetCPU_Time(CPU_Time);

	NumEventsProcessed[event->get_EVENT_TYPE()]++;
	CPU_EventTime[event->get_EVENT_TYPE()] += CPU_Time;

	CurrentSimobj->IncrementTotalTime(CPU_Time);
	CurrentSimobj->IncrementTotalEvents();

	CurrentSimobj->IncrementProcessingTime(CPU_Time);
	CurrentSimobj->IncrementCriticalPathTime(
		(CurrentEvent->get_MESSAGE())->EarliestStartTime,
		CPU_Time);

	event->CriticalPathMessages(CurrentSimobj->GetCriticalPathTime());

	event->SetProcessingTime(CurrentSimobj->GetProcessingTime());
	event->SetCriticalPathTime(CurrentSimobj->GetCriticalPathTime());
	event->SetTotalTime(CurrentSimobj->GetTotalTime());

	objman = (SpSimObjMgr *)objects->
	  get_objman(CurrentSimobj->get_OBJECT_TYPE());

	if (objman) {
	  objman->IncrementCombinedTotalEvents();
	  objman->IncrementCombinedTotalTime(CPU_Time);
	}

        qmess = event->get_mque();
        if (qmess->length() != len) {
	  cout << "Test rollback error generated by "
	       << evtype->get_messname(event->get_EVENT_TYPE())
	       << " : "
	       << len << " messages the first time, "
	       << qmess->length() << " messages the second time" << endl;

	  len = qmess->length();
	  holder = (C_HOLDER *)qmess->get_top();
	  for (i=0; i<len; i++) {
	    message = holder->get_buff();
	    header = (SpMsg *)message;
	    cout << "   Message[" << i << "] = "
		 << evtype->get_messname(header->evtype)
		 << endl;
	    holder = (C_HOLDER *)holder->get_link();
	  }
        }

      }

//...... test for messages generated in the past error

      if (check_reptime) {
        event->calculate_tmin();
        tmin = event->get_tmin();
        if (tmin < event->get_time_tag()) {
          fprintf(stderr,"Error, a message was generated in my past\n");
        }
      }

//...... phase two event processing

      event->perm_process();

      TotalCommittedTime += CPU_Time;
      CurrentSimobj->IncrementCommittedTime(CPU_Time);
      CurrentSimobj->IncrementCommittedEvents();

      NumEventsCommitted[event->get_EVENT_TYPE()]++;
      CPU_EventTimeCommitted[event->get_EVENT_TYPE()] += CPU_Time;

      objman = (SpSimObjMgr *)objects->
	get_objman(CurrentSimobj->get_OBJECT_TYPE());

      if (objman) {
	objman->IncrementCombinedCommittedEvents();
	objman->IncrementCombinedCommittedTime(CPU_Time);
      }

      event->trace();
      if (!event->void_event()) n_perm++;

//...... insert any new events back into the speedes tree

      qmess = event->get_mque();
      len = qmess->length();

      for (i=0; i<len; i++) {
        holder = (C_HOLDER *)qmess->pop_top();
        message = holder->get_buff();
        header = (SpMsg *)message;
        if (header->ext) {
          host_user->send_message((C_EM_HEADER *)message);
          evtype->delete_message(holder);
        }else{
          new1_event = (SpEvent *)evtype->message_event(message);
          evtype->delete_holder(holder);
          new1_event->insert();
        }
      }

    }

//...... handle externally generated events

    for (;;) {
      ext_event = external_event(localt);
      if (ext_event != NULL) {
        ext_event->insert();
      }else{
        break;
      }
    }

//...... put the event back into the free list

    SpCurrentAltMgr->Cleanup();
    evtype->delete_event(event);

//...... check blocking conditions

    tblock = objects->get_tblock();

  }

}

/************************************************************************
* find_gvt - find the minimum global time for safe processing		*
************************************************************************/
void SpSequential::find_gvt() {

  gvt = localt;
  objects->update_block();

}

/************************************************************************
* perm_process - carry out the events and make them permanent		*
************************************************************************/
void SpSequential::perm_process() {

}

/************************************************************************
* get_blocked_messages - get messages that are blocking SPEEDES		*
************************************************************************/
void SpSequential::get_blocked_messages() {
  SpEvent *ext_event;

  while (objects->get_tblock() <= gvt) {

    ext_event = external_event(gvt);

    if (ext_event != NULL) {
      rqueue.insert(ext_event);
    }

    objects->update_block();

  }

}
