// SpTimeMgtAlg.C method file

#include "SpAddPaddingToAlign.H"
#include "SpEventHandlerTypes.H"
#include "SpStdio.H"
#include "SpMath.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"

#include "speedes_comm.H"
#include "SpTimeMgtAlg.H"
#include "xlist.H"
#include "SpHostUser.H"
#include "SpNameHandle.H"
#include "SpHostRouterMsgs.H"
#include "SpNameValueList.H"
#include "M_SpQuery.H"
#include "host_user.H"
#include "M_SpSubscribe.H"
#include "M_SpMasterSimObjMgrUnsubscribe.H"
#include "M_SpEventHandler.H"
#include "SpSubscriptionItem.H"
#include "SpOS.H"
#include "E_SpInitial.H"
#include "SpTag.H"
#include "SpClientSock.H"
#include "RB_SpBinaryTree.H"
#include "SpBarrier.H"
#include "SpClusterMgr.H"
#include "SpReportStatus.H"
#include "PO.H"

PO_DEFINE_CLASS(SpMasterSimObjMgr);
#ifdef __IBMCPP__
namespace {
#endif

#ifndef PO_DEFINED_CHAR
PO_DEFINE_CLASS(char)
#define PO_DEFINED_CHAR
#endif

#ifdef __IBMCPP__
  }
#endif

#define OPTIMIZE
#ifndef INFINITY
#define INFINITY 1.0e20
#endif

extern int SpTimeMgtAlg_Restart;
extern char* SpTimeMgtAlg_RestartType;
extern double SpTimeMgtAlg_RestartTime;

extern int GroupId;
extern SpClusterMgr *GlobalClusterMgr;

int polling_counter;
int ExternalEventCounter = 1;
SpSimTime *IgnoreTimeGrantGvt = NULL;

SPEEDES *SpTimeMgtAlg::Speedes = NULL;
char * SpTimeMgtAlg::mr_type = NULL;
int SpTimeMgtAlg::node = 0;
int SpTimeMgtAlg::n_nodes = 0;
int SpTimeMgtAlg::mr_bytes = 0;
int SpTimeMgtAlg::mr_thresh = 0;
int SpTimeMgtAlg::lazy = 0;
int SpTimeMgtAlg::trace = 0;
int SpTimeMgtAlg::WriteLookahead = 0;
char *SpTimeMgtAlg::tracefile = NULL;
char *SpTimeMgtAlg::Lookaheadfile = NULL;
int SpTimeMgtAlg::rewind_enable = 0;
int SpTimeMgtAlg::rewind_buff_size = 0;
int SpTimeMgtAlg::CheckPointing = 0;

int SpGateway = 0;

SpMasterSimObjMgr *MasterSimObjMgr = NULL;

extern SpDataParser *speedes_parser;
extern char *TimeManagementMode;
extern SpEvent *CurrentEvent;
SpTag *CurrentTag = NULL;

SpHostUser *HostUser = NULL;
SpEventTypes *EventTypes = NULL;

void AtExitEvtqFunc(void) {
  if (HostUser) delete HostUser;
  HostUser = NULL;
}

RB_SpBinaryTree *Barriers = NULL;

/************************************************************************
* SpTimeMgtAlg : construct an event queue object				*
************************************************************************/
SpTimeMgtAlg::SpTimeMgtAlg() :
  KillSPEEDES(0),
  OldestExternallyScheduledEvent(0),
  Restart(SpTimeMgtAlg_Restart),
  RestartType(strdup(SpTimeMgtAlg_RestartType)),
  RestartTime(SpTimeMgtAlg_RestartTime)
{
  SpEvent *event;
  SpSimObj *simobj;
  //XLIST *xlist;
  int size;
  char *buff;
  int i,j, len;
  char *message;
  C_HOLDER *holder;
  SpMsg *header;
  SpSimObjMgr *objman;
  SpNameHandle *NameHandle;
  C_QUEUE NameList;
  char *PauseName;
  char *PauseTime;
  SpSimTime *NamedPauseTime;

  atexit(AtExitEvtqFunc);

  printf("EVTQ created \n");
  printf("Time Management Mode: %s\n", TimeManagementMode);

  Barriers = new RB_SpBinaryTree();

//...... set up stopwatch clock or wall time setting

  speedes_parser->GoTo("statistics", NULL);

  if (speedes_parser->GetString("Timer")) {

    if (!strcmp("WallClock", speedes_parser->GetString("Timer"))) {
      StopWatch.UseWallTime();
    } else if (!strcmp("CPU", speedes_parser->GetString("Timer"))) {
      StopWatch.UseClockTime();
    }

  }else{

    StopWatch.UseClockTime();

  }

  MemoryUsage = speedes_parser->GetLogical("MemoryUsage");
  MessageSending = speedes_parser->GetLogical("MessageSending");
  EventProcessing = speedes_parser->GetLogical("EventProcessing");
  ObjectProcessing = speedes_parser->GetLogical("ObjectProcessing");
  CriticalPath = speedes_parser->GetLogical("CriticalPath");


/*
** HLA Gateway enabled
*/

  if (speedes_parser->GoTo("HLA", NULL) == NULL)
  {
    cerr << "Error: speedes.par does not have an HLA section." << endl;
    SpGateway = 0;
  }
  else
    SpGateway = speedes_parser->GetLogical("Gateway");
  
//...... get stuff out of the speedes parser

  speedes_parser->GoTo("gvt_parameters", NULL);
  Tasb = speedes_parser->GetFloat("Tasb");
  Tgvt = speedes_parser->GetFloat("Tgvt");
  Ngvt = speedes_parser->GetInt("Ngvt");
  Nrisk = speedes_parser->GetInt("Nrisk");
  Nopt = speedes_parser->GetInt("Nopt");

  if (!strcasecmp(TimeManagementMode, "BTB")) Nrisk = 0;
  if (!strcasecmp(TimeManagementMode, "BREATHING_TIME_BUCKETS")) Nrisk = 0;
  if (!strcasecmp(TimeManagementMode, "BreathingTimeBuckets")) Nrisk = 0;

  if (!strcasecmp(TimeManagementMode, "TW")) Nrisk = 0x7fffffff;
  if (!strcasecmp(TimeManagementMode, "TWOS")) Nrisk = 0x7fffffff;
  if (!strcasecmp(TimeManagementMode, "TIME_WARP")) Nrisk = 0x7fffffff;
  if (!strcasecmp(TimeManagementMode, "TimeWarp")) Nrisk = 0x7fffffff;

  if (!strcasecmp(TimeManagementMode,"Conservative")) Nopt = 0;
  if (!strcasecmp(TimeManagementMode,"FTB")) Nopt = 0;
  if (!strcasecmp(TimeManagementMode,"FIXED_TIME_BUCKETS")) Nopt = 0;
  if (!strcasecmp(TimeManagementMode,"TIME_BUCKETS")) Nopt = 0;
  if (!strcasecmp(TimeManagementMode,"TimeBuckets")) Nopt = 0;
  if (!strcasecmp(TimeManagementMode,"FIXED_TIME_BUCKET")) Nopt = 0;
  if (!strcasecmp(TimeManagementMode,"TIME_BUCKET")) Nopt = 0;
  if (!strcasecmp(TimeManagementMode,"TimeBucket")) Nopt = 0;

  if (node == 0) {
    void *PauseMark = speedes_parser->GoTo("NamedPauses", NULL);
    if (PauseMark) {
      speedes_parser->GetFirstDefine(PauseName, PauseTime);
      for (i=0; i<speedes_parser->GetNdefines(); i++) {
	cout << "Initial named Pause "
	     << PauseName
	     << " "
	     << PauseTime
	     << endl;
	NamedPauseTime = new SpSimTime(atof(PauseTime));
#ifdef DONT_KNOW_IF_THIS_IS_THE_RIGHT_FIX
	NamedPauseTime->SetCounter  (0x7FFFFFFF); 
	NamedPauseTime->SetUniqueId (0x7FFFFFFF); 
#endif
	NamedPauses.Insert(NamedPauseTime, PauseName);
	speedes_parser->GetNextDefine(PauseName, PauseTime);
      }
    }
  }

//...... initialize some quantities

  simobj = NULL;
  event = NULL;
  //xlist = NULL;

  PauseFlag = 0;
  polling_counter = 0;
  n_temp = 0;
  n_perm = 0;
  n_roll = 0;
  n_cancel = 0;
  n_messrec = 0;
  n_amessrec = 0;
  n_exmess = 0;

  TotalProcessingTime = 0.0;
  TotalCommittedTime = 0.0;

  host_user = new C_HOST_USER();
  host_user->set_node(node);

  E_SpInitial *InitialEvent = new E_SpInitial();
  CurrentEvent = InitialEvent;

  InitialEvent->init_qroll_backs();
  
  evtype = new SpEventTypes();
  EventTypes = evtype;
  simobj->set_evtype(evtype);

  InitialEvent->set_evtype(evtype);

// Plug in the SPEEDES cluster default

  PLUG_IN_CLUSTER(SPEEDES);

// Get the cluster event info

  for (i=0; i<EventTypes->get_ntypes(); i++) {
    GlobalClusterMgr->PlugInEvent(EventTypes->get_messname(i), i);
  }
  GlobalClusterMgr->GlobalEventInfo();


// Get the cluster/node info

  GlobalClusterMgr->GlobalClusterInfo();

  if (Restart){
    GetObjectsFromCheckpoint(InitialEvent);
    objects->ResetLocalRqueues();
    objects->PersistenceRestartResetObjects();
  }
  else{
    if (CheckPointing){
      objects = PO_NEW_SpMasterSimObjMgr();
      objects->PersistenceRegisterPointers();
    }
    else{
      objects = new SpMasterSimObjMgr();
      MasterSimObjMgr = objects;
    }
  }

// Get the cluster mgr info

  for (i=0; i<objects->GetNmanagers(); i++) {
    GlobalClusterMgr->PlugInMgr(objects->GetManagerName(i), i);
  }
  GlobalClusterMgr->GlobalMgrInfo();

// Print out the cluster info

  if (GlobalClusterMgr->CheckLocalNodeZero()) GlobalClusterMgr->Print(cout);

  InitialEvent->InitSimObj(objects);
  InitialEvent->InitEventType(INITIAL_EVENT);

  buff = objects->get_allids(size);
  host_user->send_node_objects(size, buff);
  delete [] buff;

/*
** New SPEEDES Host User code connects to the SpeedesServer (which supports
** the host router) and passes the names and Ids of all of its named simobjs.
*/

  char * hrMachineName;
  int hrPort;

  speedes_parser->GoTo("SpeedesServer", "HostRouter", NULL);
  if (speedes_parser->GetLogical("UseDefaults")){
    speedes_parser->GoTo("SpeedesServer", NULL);
    hrMachineName = speedes_parser->GetString("DefaultMachineName");
    hrPort = speedes_parser->GetInt("DefaultPort");
  }
  else{
    hrMachineName = speedes_parser->GetString("MachineName");
    hrPort = speedes_parser->GetInt("Port");
  }


  HostUser = new SpHostUser(hrMachineName, hrPort, GroupId);

  if (HostUser->ConnectionStatus() == GOOD_CONNECTION_STATUS){
    cout<<"Created the SPEEDES Host User"<<endl;
  }

  for (i=0; i<objects->GetNmanagers(); i++) {
    objman = (SpSimObjMgr *)objects->get_objman(i);
/*
    cout << "Node " << node
	 << ", Type " << i
	 << ", Name " << objman->get_NAME()
	 << ", GlobalId " << objman->get_GLOBAL_ID()
	 << endl;
*/
    NameHandle = new SpNameHandle(node,
				 i,
				 -1,
				 objman->get_GLOBAL_ID(),
				 objman->get_NAME());
    NameList.push_bot(NameHandle);

    for (j=0; j<objman->get_N_LOC(); j++) {
      simobj = objman->get_obj(j);
      if (simobj->get_NAME()) {
	NameHandle = new SpNameHandle(node,
				     i,
				     j,
				     simobj->get_GLOBAL_ID(),
				     simobj->get_NAME());
	NameList.push_bot(NameHandle);
      }
    }
  }

  HostUser->SpeedesConnection(n_nodes, node, &NameList);
  HostUser->Flush();

  while ((NameHandle = (SpNameHandle *)NameList.pop_top())) delete NameHandle;

/*
** Done connecting up the new SPEEDES Host User
*/

//  evtype = new SpEventTypes();

//  EventTypes = evtype;

#ifdef OPTIMIZE
  if (n_nodes == 1) {
    one_node_flag = 1;
  }else{
    one_node_flag = 0;
  }
#else
  one_node_flag = 0;
#endif

/*
** The comm library gets to decide what type of multirouter to return
*/

  simobj->set_rqueue(&rqueue);

  event->set_one_node_flag(one_node_flag);
  event->set_HOST_USER(host_user);
  event->set_objects(objects);
  event->set_RANDOM(simobj->get_RANDOM());
  event->set_evtype(evtype);

  if (trace) {
    event->set_trace();
    event->open_tracefile(tracefile);
  }else{
    event->reset_trace();
  }

  if (WriteLookahead) {
    event->set_WriteLookahead();
    event->open_Lookaheadfile(Lookaheadfile);
  }else{
    event->reset_WriteLookahead();
  }

  gvt = -INFINITY;
  gvt.SetPriority1(0x80000000);
  gvt.SetPriority2(0x80000000);
  gvt.SetCounter(0x80000000);
  gvt.SetUniqueId(0x80000000);

  gvtmax = INFINITY;
  localt = INFINITY;

  printf("initializing events...\n");

  simobj->set_init_event_queue((C_QUEUE *)this);
//  simobj->set_evtype(evtype);

  if (Restart){
    GetMessagesFromCheckpoint(InitialEvent, len);
  }
  else{
    objects->init_events();

    //...... turn the initial messages into events

    len = n_items;
    for (i=0; i<len; i++) {
      holder = (C_HOLDER *)pop_top();
      message = holder->get_buff();
      header = (SpMsg *)message;
      header->cancel_id = -1;
      event = (SpEvent *)evtype->message_event(message);
      push_bot(event);
      evtype->delete_message(holder);
    }
    push_top(InitialEvent);
  }
  sort();

  SpComm_BarrierSync();

  QUERY = evtype->get_messid("QUERY");
  MONITOR = evtype->get_messid("MONITOR");
  EVENT_HANDLER_EVENT = evtype->get_messid("EVENT_HANDLER_EVENT");

  //event->set_EVENT_HANDLER_EVENT(EVENT_HANDLER_EVENT);

  C_QUEUE *qmess;
  qmess = InitialEvent->get_mque();

  len += qmess->get_length();
  len = SpComm_GlobalSum(len);

  printf("Created %d total initial events\n",len);

/*
** Create the CPU_EventTime and CPU_EventTime2 arrays for statistics
** gathering of processing times for each kind of event. Note, the
** CPU_EventTime2 variable is used to obtain standard deviations
*/

  NumEventsProcessed = new int[evtype->get_ntypes()];
  NumEventsCommitted = new int[evtype->get_ntypes()];
  CPU_EventTime = new double[evtype->get_ntypes()];
  CPU_EventTimeCommitted = new double[evtype->get_ntypes()];

  NumMessages = new int[evtype->get_ntypes()];
  NumCoordMessages = new int[evtype->get_ntypes()];
  NumAntimessages = new int[evtype->get_ntypes()];

  for (i=0; i<evtype->get_ntypes(); i++) {
    NumEventsProcessed[i] = 0;
    NumEventsCommitted[i] = 0;
    CPU_EventTime[i] = 0.0;
    CPU_EventTimeCommitted[i] = 0.0;
    NumMessages[i] = 0;
    NumCoordMessages[i] = 0;
    NumAntimessages[i] = 0;
  }

}

SpTimeMgtAlg::~SpTimeMgtAlg()
{
  delete evtype;
  delete [] NumEventsProcessed;
  delete [] NumEventsCommitted;
  delete [] CPU_EventTime;
  delete [] CPU_EventTimeCommitted;
  delete [] NumMessages;
  delete [] NumCoordMessages;
  delete [] NumAntimessages;
}

/************************************************************************
* term - terminate the simulation and print out some processing		*
*	 statistics							*
************************************************************************/
void SpTimeMgtAlg::term(double tnd) {
  int i;
  int Nchars = 0;
  double averageCPU;
  double totalCPU = 0.0;
  double totalCPUCommitted = 0.0;
  double CPU_Time;
  double criticalPathTime;
  double maxSpeedup;
  SpSimObjMgr *objman;
  int CombinedTotalEvents;
  int totalEventsCommitted = 0;
  int CombinedCommittedEvents;
  double CombinedTotalTime;
  double CombinedCommittedTime;
  double TotalTime = 0.0;
  int Nrollbacks1 = 0;
  int Nrollbacks2 = 0;
  int TotalMessages = 0;
  int TotalCoordMessages = 0;
  int TotalAntimessages = 0;
  int AnythingPrinted = 0;
/*
** Terminate all of the objects and object managers
*/

  TotalCommittedTime = 0.0;

  objects->term(tnd);
  SpGlobalReportStatus();
  SpPrintReportStatus(cout);

/*
** Return if we don't need to print out any statistics
*/

  if ( !(CriticalPath
      || MemoryUsage
      || (MessageSending && (n_nodes > 1))
      || ObjectProcessing
      || EventProcessing)) {

    return;
  }

/*
** Figure out what is the largest event or object manager name (for clean
** printing and lining up things). Then print out the first line to start
** the printout section.
*/

  for (i=0; i<evtype->get_ntypes(); i++) {
    if (NumEventsProcessed[i]) {
      if (((int)strlen(evtype->get_messname(i))) > Nchars) {
	Nchars = strlen(evtype->get_messname(i));
      }
    }
  }

  for (i=0; i<objects->GetNmanagers(); i++) {
    objman = (SpSimObjMgr *)objects->get_objman(i);
    if (((int)strlen(objman->get_NAME())) > Nchars) {
      Nchars = strlen(objman->get_NAME());
    }
  }

  if (node == 0) {
    cout << endl;
    for (i=0; i<Nchars + 132; i++) cout << "-";
    cout << endl;
  }


/*
** Print out the memory usage statistics
*/

  if (MemoryUsage) {
    evtype->print();
    AnythingPrinted = 1;
  }

/*
** Print out the message sending statistics 
*/

  if (MessageSending && (n_nodes > 1)) {

    if (node == 0) {
      if (AnythingPrinted) cout << endl;
      AnythingPrinted = 1;
      cout << "Message sending statistics by event types:" << endl;
      cout << endl;
    }

    for (i=0; i<evtype->get_ntypes(); i++) {

      NumMessages[i] = SpComm_GlobalSum(NumMessages[i]);
      NumAntimessages[i] = SpComm_GlobalSum(NumAntimessages[i]);
      NumCoordMessages[i] = SpComm_GlobalSum(NumCoordMessages[i]);

      TotalMessages += NumMessages[i];
      TotalCoordMessages += NumCoordMessages[i];
      TotalAntimessages += NumAntimessages[i];

      if ((node == 0) && (NumMessages[i] +
			  NumCoordMessages[i] +
			  NumAntimessages[i] > 0)) {

	cout << setw(Nchars+3) << evtype->get_messname(i)
	     << "   Nmess = " << setw(10) << NumMessages[i]
	     << "   Ncoord = " << setw(10) << NumCoordMessages[i]
	     << "   Nanti = " << setw(10) << NumAntimessages[i]
	     << endl;
      }

    }

    if (node ==0) {
      cout << endl;
      cout << "   Number of asynchronous messages = " << TotalMessages << endl;
      cout << "   Number of coordinated messages = "
	   << TotalCoordMessages << endl;
      cout << "   Number of antimessages = " << TotalAntimessages << endl;
      cout << "   Total number of messages = "
	   << TotalMessages + TotalCoordMessages + TotalAntimessages
	   << endl;
    }

  }

/*
** Print out the event processing time statistics
*/

  if (EventProcessing || CriticalPath) {

    if ((node == 0) && (EventProcessing)) {
      if (AnythingPrinted) cout << endl;
      AnythingPrinted = 1;
      cout << "Event processing times by event types in seconds:" << endl;
      cout << endl;
    }

    for (i=0; i<evtype->get_ntypes(); i++) {

      NumEventsProcessed[i] = SpComm_GlobalSum(NumEventsProcessed[i]);

      if ((NumEventsProcessed[i])
	  && (strcmp(evtype->get_messname(i), "INITIAL_EVENT"))) {

	NumEventsCommitted[i] = SpComm_GlobalSum(NumEventsCommitted[i]);
	CPU_EventTime[i] = SpComm_GlobalSum(CPU_EventTime[i]);
	CPU_EventTimeCommitted[i] =
	  SpComm_GlobalSum(CPU_EventTimeCommitted[i]);

	Nrollbacks1 += NumEventsProcessed[i] - NumEventsCommitted[i];

	averageCPU = CPU_EventTimeCommitted[i] / NumEventsCommitted[i];
	totalCPU += CPU_EventTime[i];
	totalCPUCommitted += CPU_EventTimeCommitted[i];
	totalEventsCommitted += NumEventsCommitted[i];

	if ((node == 0) && (EventProcessing)) {
	  cout << setw(Nchars+3) << evtype->get_messname(i)
	       << "   Nproc = " << setw(10) << NumEventsProcessed[i]
	       << "   Ncmtd = " << setw(10) << NumEventsCommitted[i]
	       << "   Tcpu = " << setw(11) << CPU_EventTime[i]
	       << "   Tcmtd = " << setw(11) << CPU_EventTimeCommitted[i]
	       << "   <Tcmtd> = " << setw(11) << averageCPU
	       << "   Eff = " << setw(11)
	       << (CPU_EventTimeCommitted[i] + 1.0e-20)
	               / (CPU_EventTime[i] + 1.0e-20)
	       << endl;
	}
      }
    }

  }

/*
** Print out the processing time statistics by object manager type
*/

  if (ObjectProcessing || CriticalPath) {

    if ((node == 0) && (ObjectProcessing)) {
      if (AnythingPrinted) cout << endl;
      AnythingPrinted = 1;
      cout << "Simobj processing times by object manager type in seconds:"
	   << endl;
      cout << endl;
    }


    for (i=0; i<objects->GetNmanagers(); i++) {
      objman = (SpSimObjMgr *)objects->get_objman(i);

      CombinedTotalEvents =
	SpComm_GlobalSum(objman->GetCombinedTotalEvents());

      CombinedCommittedEvents =
	SpComm_GlobalSum(objman->GetCombinedCommittedEvents());

      CombinedTotalTime =
	SpComm_GlobalSum(objman->GetCombinedTotalTime());

      CombinedCommittedTime =
	SpComm_GlobalSum(objman->GetCombinedCommittedTime());

      TotalTime += CombinedTotalTime;
      TotalCommittedTime += CombinedCommittedTime;

      Nrollbacks2 += CombinedTotalEvents - CombinedCommittedEvents;

      if ((node == 0) && (ObjectProcessing)) {
	cout << setw(Nchars+3) << objman->get_NAME()
	     << "   Nproc = " << setw(10) << CombinedTotalEvents
	     << "   Ncmtd = " << setw(10) << CombinedCommittedEvents
	     << "   Tcpu = " << setw(11) << CombinedTotalTime
	     << "   Tcmtd = " << setw(11) << CombinedCommittedTime
	     << "   <Tcmtd> = " << setw(11);

	if (CombinedCommittedEvents == 0)
	  cout << 0.0;
	else
	  cout << CombinedCommittedTime / CombinedCommittedEvents;

	cout << "   Eff = " << setw(11)
	     << (CombinedCommittedTime + 1.0e-20)
	          / (CombinedTotalTime + 1.0e-20)
	     << endl;
      }
    }

    if (ObjectProcessing && EventProcessing) {
      if (Nrollbacks1 != Nrollbacks2) {
	cout << "error, bad number of rollbacks" << endl;
      }
    }

  }

/*
** Compute the critical path and print it out only if we have done the event
** and object analysis.
*/

  if (CriticalPath) {

    objects->ComputeCriticalPath(CPU_Time, criticalPathTime, maxSpeedup);

    if (node == 0) {
      if (AnythingPrinted) cout << endl;
      AnythingPrinted = 1;
      cout << "Critical path analysis:" << endl;
      cout << endl;
      cout << "   Number of events committed = "
	   << totalEventsCommitted << endl;
      cout << "   Number of events rolled back = " << Nrollbacks1 << endl;

      cout << "   Total event processing time = " << totalCPU << endl;
      cout << "   Total committed event processing time = "
	   << TotalCommittedTime << endl;

      if (TotalTime != 0) {
	cout << "   Total processing efficiency = "
	     << TotalCommittedTime / TotalTime << endl;
      }
      cout << "   Critical path processing time = "
	   << criticalPathTime << endl;
      cout << "   Maximum possible speedup = " << maxSpeedup << endl;
    }

  }

/*
** Print out the final line of the statistics section
*/

  if (node == 0) {
    for (i=0; i<Nchars + 132; i++) cout << "-";
    cout << endl;
    cout << endl;
  }

} 

/************************************************************************
* set_node_info - set up the node information				*
************************************************************************/
void SpTimeMgtAlg::set_node_info(int nds, int nd) {
  SpSimObj *simobj;
  SpEvent *event;
  //C_CYCLE *cycle;

  node = nd;
  n_nodes = nds;

  event = NULL;
  event->set_NODE(node);
  event->set_N_NODES(n_nodes);

  simobj = NULL;
  simobj->set_NODE(node);
  simobj->set_N_NODES(n_nodes);

}

/************************************************************************
* temp_process - process events in temporary storage			*
************************************************************************/
void SpTimeMgtAlg::temp_process() {
}

/************************************************************************
* find_gvt - find the minimum global time for safe processing	*
************************************************************************/
void SpTimeMgtAlg::find_gvt() {

}

/************************************************************************
* perm_process - carry out the events and make them permanent		*
************************************************************************/
void SpTimeMgtAlg::perm_process() {
}

/************************************************************************
* external_event - check if there are external messages			*
************************************************************************/
SpEvent *SpTimeMgtAlg::external_event(SpSimTime& time) {

  SpEvent *ext_event = NULL;

  SpMsg *Header = NULL;;
  M_SpEventHandler *EventHandlerEventMess;
  SUBSCRIBE_MESS *Subscribe_Mess;
  UNSUBSCRIBE_MESS *UnSubscribe_Mess;

  GENERIC_MSG_HEADER *Message;
  SUBSCRIBE_MESSAGE *SubscribeMessage;
  UNSUBSCRIBE_MESSAGE *UnSubscribeMessage;
  TWO_WAY_COMMAND_MESSAGE *TwoWayCommandMessage;
  QUERY_MESSAGE *QueryMessage;
  COMMAND_MESSAGE *CommandMessage;
  NAMED_PAUSE_MESSAGE *NamedPauseMessage;
  NAMED_RESUME_MESSAGE *NamedResumeMessage;

  int LocalId;
  int ObjectType;
  int GlobalId;
  int ExternalId;
  int DataBytes;
  int HandlerNum;
  int EventType;
  int Offset;
  int BarrierMode;

  double TimeLag;
  double OldTime;
  double NewTime;

  SpSimTime EventTime;
  SpSimTime *NamedPauseTime;
  SpSimTime *BarrierTime;

  char *Buff;
  char *Data;
  char *CommandName;

  C_QUEUE *qblock;
  C_ITEM *item;
  SUBSCRIPTION_ITEM *SubItem;
  char *NamedPause;
  char *NamedResume;

  if ((Message = HostUser->GetMessage())) {

    HandlerNum = Message->handlerNum;

    switch (HandlerNum) {

/*
** Handle an incoming query
*/

      case QUERY_MESSAGE_ID:

	cout << "Got a query message" << endl;
	{

	QueryMessage = (QUERY_MESSAGE *)Message;

	M_SpQuery *QueryMess = (M_SpQuery *)
	  new char[sizeof(M_SpQuery) + QueryMessage->bytes];

	QueryMess->time_tag = time;
	QueryMess->evtype = QUERY;
	QueryMess->obtype = QueryMessage->ObjectType;
	QueryMess->objid = QueryMessage->LocalId;
	QueryMess->ExternalId = QueryMessage->ExternalId;
	QueryMess->bytes = QueryMessage->bytes;
	memcpy((char *)&QueryMess[1], (char *)&QueryMessage[1],
	       QueryMessage->bytes);

	ext_event = (SpEvent *)evtype->message_event((char *)QueryMess);
	}

	delete [] (char *)QueryMessage;

	return ext_event;

/*
** Handle an incoming command
*/

      case COMMAND_MESSAGE_ID:

	cout << "Got a command message" << endl;

	CommandMessage = (COMMAND_MESSAGE *)Message;

	ExternalId = CommandMessage->ExternalId;
	LocalId = CommandMessage->LocalId;
	ObjectType = CommandMessage->ObjectType;
	GlobalId = CommandMessage->GlobalId;
	if(CommandMessage->Time < time)
	  EventTime = time;
	else
	  EventTime = CommandMessage->Time;

	Buff = (char *)&CommandMessage[1];

	if ((GlobalId < 0) && (CommandMessage->Node < 0)) {
	  Buff += strlen(Buff) + 1;
	}

	CommandName = Buff;
	Buff += strlen(Buff) + 1;
	Data = Buff;

	DataBytes = CommandMessage->bytes - (Data-(char *)CommandMessage);

/*
	cout << "ObjectType: " << ObjectType << endl;
	cout << "LocalId: " << LocalId << endl;
	cout << "GlobalId: " << GlobalId << endl;
	cout << "Command Name: " << CommandName << endl;
	cout << "Command Data: " << Data << endl;
	cout << " Data Bytes: " << DataBytes << endl;
*/

	EventType = evtype->get_messid(CommandName);
//	cout << "EventType: " << EventType << endl;

	if (EventType >= 0) {

	  if (CommandMessage->ReturnStatus) {
	    HostUser->CommandReply(time, ExternalId, COMMAND_STATUS_GOOD);
	  }

	  Header = (SpMsg *)new char[sizeof(SpMsg) + DataBytes];
	  memcpy((char *)&Header[1], Data, DataBytes);

	  Header->time_tag = time;
	  //(Header->time_tag).SetCounter
	  //  (time.GetCounter() + ExternalEventCounter);
	  //ExternalEventCounter++;
	  //(Header->time_tag).SetUniqueId(-ExternalId);

	  Header->evtype = EventType;
	  Header->obtype = ObjectType;
	  Header->objid = LocalId;
//	  Header->ExternalId = -1;
	  Header->ExternalId = ExternalId;
	  Header->bytes = DataBytes;
	  Header->EarliestStartTime = -INFINITY;

	  ext_event = (SpEvent *)evtype->message_event((char *)Header);

	  //delete [] (char *)Header;

	}else{

	  EventType = evtype->get_messid("EVENT_HANDLER_EVENT");

	  if (EventType < 0) {

	    if (CommandMessage->ReturnStatus) {
	      HostUser->CommandReply(time, ExternalId,
				     COMMAND_STATUS_BAD_EVENT);
	    }

	    cout << "Warning, could not schedule command " << CommandName
		 << endl;

	  }else{


	    if (CommandMessage->ReturnStatus) {
	      HostUser->CommandReply(time, ExternalId, COMMAND_STATUS_GOOD);
	    }

	    Offset = strlen(CommandName) + 1;
	    Offset = SpAddPaddingToAlign(Offset, sizeof(double));

	    EventHandlerEventMess = (M_SpEventHandler *)
	      new char[sizeof(M_SpEventHandler)
		      + Offset + DataBytes];

	    Buff = (char *)&EventHandlerEventMess[1];
	    strcpy(Buff, CommandName);
	    memcpy(&Buff[Offset], Data, DataBytes);

	    EventHandlerEventMess->time_tag = EventTime;
	    //(EventHandlerEventMess->time_tag).SetCounter
	    //  (time.GetCounter() + ExternalEventCounter);
	    //ExternalEventCounter++;
	    //(EventHandlerEventMess->time_tag).SetUniqueId(-ExternalId);

	    EventHandlerEventMess->Undirected = 0;
	    EventHandlerEventMess->EHtype = LEGACYEVENTHANDLER;
	    EventHandlerEventMess->evtype = EventType;
	    EventHandlerEventMess->obtype = CommandMessage->ObjectType;
	    EventHandlerEventMess->objid = CommandMessage->LocalId;
	    EventHandlerEventMess->rbid = -1;
	    EventHandlerEventMess->bytes = Offset + DataBytes;
	    EventHandlerEventMess->DataBytes = DataBytes;
//	    EventHandlerEventMess->ExternalId = -1;
	    EventHandlerEventMess->ExternalId = ExternalId;

	    ext_event = (SpEvent *)evtype->
	      message_event((char *)EventHandlerEventMess);

	    //delete [] (char *)EventHandlerEventMess;

	  }

	}

	delete [] (char *)CommandMessage;

	break;

/*
** Handle an incoming two-way command
*/

      case TWO_WAY_COMMAND_MESSAGE_ID:

	cout << "Got a two-way command message" << endl;

	TwoWayCommandMessage = (TWO_WAY_COMMAND_MESSAGE *)Message;

	ExternalId = TwoWayCommandMessage->ExternalId;
	LocalId = TwoWayCommandMessage->LocalId;
	ObjectType = TwoWayCommandMessage->ObjectType;
	GlobalId = TwoWayCommandMessage->GlobalId;
	TimeLag = TwoWayCommandMessage->TimeLag;
	BarrierMode = TwoWayCommandMessage->BarrierMode;

	if (BarrierMode != NO_BARRIER) {

	  BarrierTime = (SpSimTime *)GvtBarriers.Find(ExternalId);

	  if (!BarrierTime) {
//	    BarrierTime = new SpSimTime(gvt);
	    BarrierTime = new SpSimTime(time);
	    GvtBarriers.Insert(BarrierTime, ExternalId);
	  }

	  if (BarrierMode == AGGRESSIVE_BARRIER) {
//	    EventTime = gvt;
	    EventTime = time;
//	    *BarrierTime = gvt + TimeLag;
	    *BarrierTime = time + TimeLag;
	  }

	  if (BarrierMode == SYNCHRONIZED_BARRIER) {
	    EventTime = *BarrierTime;
	    *BarrierTime += TimeLag;
	  }

#ifdef COMMENTOUT
	  qblock = objects->get_qblock();
	  item = qblock->deque(ExternalId);

	  if (!item) {
	    item = new C_ITEM();
	    item->set_id(ExternalId);
//	    EventTime = gvt;
	    EventTime = time;
	  }else{
	    if (BarrierMode == AGGRESSIVE_BARRIER) {
//	      EventTime = gvt;
	      EventTime = time;
	    }
	    if (BarrierMode == SYNCHRONIZED_BARRIER) {
	      EventTime = item->get_time_tag();
	    }
	  }

	  FutureTime = EventTime;
	  FutureTime.SetTime(EventTime + TimeLag);
	  SetBlock(FutureTime, ExternalId, item);
#endif

	}else{

//	  EventTime = gvt;
	  EventTime = time;

	}

	Buff = (char *)&TwoWayCommandMessage[1];

	if (GlobalId < 0) {
	  Buff += strlen(Buff) + 1;
	}

	CommandName = Buff;
	Buff += strlen(Buff) + 1;
	Data = Buff;

	DataBytes = TwoWayCommandMessage->bytes -
	  (Data-(char *)TwoWayCommandMessage);


	cout << "ObjectType: " << ObjectType << endl;
	cout << "LocalId: " << LocalId << endl;
	cout << "GlobalId: " << GlobalId << endl;
	cout << "Command Name: " << CommandName << endl;
	cout << "Command Data: " << Data << endl;
	cout << " Data Bytes: " << DataBytes << endl;
	cout << " BarrierMode: " << BarrierMode << endl;
	cout << " Time Lag: " << TimeLag << endl;

	EventType = evtype->get_messid(CommandName);
//	cout << "EventType: " << EventType << endl;

	if (EventType >= 0) {

	  HostUser->CommandReply(time, ExternalId, COMMAND_STATUS_GOOD);

	  Header = (SpMsg *)new char[sizeof(SpMsg) + DataBytes];
	  memcpy((char *)&Header[1], Data, DataBytes);

	  Header->time_tag = EventTime;
	  Header->evtype = EventType;
	  Header->obtype = ObjectType;
	  Header->objid = LocalId;
	  Header->ExternalId = ExternalId;
	  Header->bytes = DataBytes;
	  Header->EarliestStartTime = -INFINITY;

	  ext_event = (SpEvent *)evtype->message_event((char *)Header);

	  //delete [] (char *)Header;

	}else{

	  EventType = evtype->get_messid("EVENT_HANDLER_EVENT");

	  if (EventType < 0) {
	    HostUser->CommandReply(time, ExternalId,
				     COMMAND_STATUS_BAD_EVENT);
	    cout << "Warning, could not schedule command " << CommandName
		 << endl;
	  }else{

	    HostUser->CommandReply(time, ExternalId, COMMAND_STATUS_GOOD);

	    Offset = strlen(CommandName) + 1;
	    Offset += (sizeof(double) - Offset%sizeof(double))
	      % sizeof(double);

	    EventHandlerEventMess = (M_SpEventHandler *)
	      new char[sizeof(M_SpEventHandler)
		      + Offset + DataBytes];

	    Buff = (char *)&EventHandlerEventMess[1];
	    strcpy(Buff, CommandName);
	    memcpy(&Buff[Offset], Data, DataBytes);

	    EventHandlerEventMess->time_tag = EventTime;
	    EventHandlerEventMess->EHtype = LEGACYEVENTHANDLER;
	    EventHandlerEventMess->evtype = EventType;
	    EventHandlerEventMess->obtype = TwoWayCommandMessage->ObjectType;
	    EventHandlerEventMess->objid = TwoWayCommandMessage->LocalId;
	    EventHandlerEventMess->rbid = -1;
	    EventHandlerEventMess->bytes = Offset + DataBytes;
	    EventHandlerEventMess->DataBytes = DataBytes;
	    EventHandlerEventMess->ExternalId =
	      TwoWayCommandMessage->ExternalId;

	    ext_event = (SpEvent *)evtype->
	      message_event((char *)EventHandlerEventMess);

	    //delete [] (char *)EventHandlerEventMess;

	  }

	}

	delete [] (char *)TwoWayCommandMessage;

	break;

      case SCHEDULE_EVENT_ID:{
	static int externalRollbackId = SpComm_GetNodeId();
	SCHEDULE_EVENT_MESSAGE* msg = (SCHEDULE_EVENT_MESSAGE*)Message;
	char* eventName = (char *) &msg[1];
	char* msgData = eventName + strlen(eventName) + 1;
        char* extraData = msgData + msg->MessageBytes;
	int eventType = CurrentSimobj->event_type(eventName);
	SpCancelHandle *ch;
	int rc = EVENT_SCHEDULED_AT_GVT;
	int inPast = 0;
	SpSimTime schedTime = msg->Time;

//	if ((msg->Time < (double)gvt)){
	if ((msg->Time < (double)time)){
	  if (msg->ScheduleMode == IF_IN_PAST_IGNORE){
	    rc = NOT_SCHEDULED_BECAUSE_IN_PAST;
	    schedTime = time;
	  }
	  else{
	    inPast = 1;
	    schedTime = (double)time;
	  }
	}
	schedTime.SetPriority1(gvt.GetPriority1());
	schedTime.SetPriority2(gvt.GetPriority2());
	schedTime.SetUniqueId(MaxPriority1);
	schedTime.SetCounter(externalRollbackId);
	if (rc != NOT_SCHEDULED_BECAUSE_IN_PAST){
	  if (eventType != -1){

	    Header = (SpMsg *)new char[msg->MessageBytes + msg->DataBytes];
	    memcpy((char*)Header, msgData, msg->MessageBytes);
	    memcpy(((char *)Header) + msg->MessageBytes,
		   extraData, msg->DataBytes);

	    Header->bytes = msg->DataBytes;
	    Header->evtype = eventType;
	    rc = (inPast) ? EVENT_SCHEDULED_AT_GVT : EVENT_SCHEDULED;
	  }
	  else{
	    Offset = strlen(eventName)+1;
	    Offset+= (sizeof(double) - Offset%sizeof(double)) % sizeof(double);

	    EventHandlerEventMess = (M_SpEventHandler *)
	      new char[sizeof(M_SpEventHandler)
		      + Offset + msg->DataBytes];

	    Buff = (char *)&EventHandlerEventMess[1];
	    strcpy(Buff, eventName);
	    memcpy(&Buff[Offset], extraData, msg->DataBytes);

	    EventHandlerEventMess->bytes = Offset + msg->DataBytes;
	    EventHandlerEventMess->DataBytes = msg->DataBytes;
	    Header = (SpMsg*)EventHandlerEventMess;

	    Header->bytes = Offset + msg->DataBytes;

	    Header->evtype = evtype->get_messid("EVENT_HANDLER_EVENT");
	    rc=inPast? EVENT_HANDLER_SCHEDULED_AT_GVT :EVENT_HANDLER_SCHEDULED;
	  }
	  Header->time_tag = schedTime;
	  Header->obtype = msg->ObjectType;
	  Header->objid = msg->LocalId;
	  Header->rbid = externalRollbackId;
	  Header->EarliestStartTime = -INFINITY;

	  ext_event = (SpEvent *)evtype->message_event((char *)Header);
	}
	else{
	  //cout<<"!!!!!!!!!!!!!!!Event not scheduled because in past"<<endl;
	}

	if (msg->CancelHandleRequested){
	  if (Header){
	    ch = (SpCancelHandle*)twosmess->new_antimess();
	    ch->set_time_tag(Header->time_tag);
	    ch->set_event(ext_event);
	    ExternallyScheduledEvents.Insert((void*)ch, externalRollbackId);
	    externalRollbackId +=n_nodes;
	  }

	  HostUser->ScheduleEventReply(rc, schedTime,
				       msg->ExternalId,
				       externalRollbackId - n_nodes);
	}
	  

	if (OldestExternallyScheduledEvent > msg->Time){
	  OldestExternallyScheduledEvent = schedTime;
	}
//	else if (OldestExternallyScheduledEvent < gvt - 100){
	else if (OldestExternallyScheduledEvent < time - 100){
	  CleanUpExternallyScheduledEvents();
	}
	delete [] (char *) Message;
      }
      ext_event->SetCriticalPathTime(0);

      break;


      case CANCEL_EVENT_ID:
	{
	  static int cancelEventType = CurrentSimobj->event_type("CANCEL");
	  CleanUpExternallyScheduledEvents();

	  int rc = CANCELED_EVENT_OK;
	  CANCEL_EVENT_MESSAGE* msg = (CANCEL_EVENT_MESSAGE*)Message;
	  int rbid = msg->CancelHandle;
	  SpCancelHandle* ch =
	    (SpCancelHandle*)ExternallyScheduledEvents.Remove(rbid);
	  if (ch){
	    n_roll += ch->get_event()->cancel_event();
	    delete ch;
	  }
	  else{
	    rc = CANCEL_EVENT_NOT_FOUND;
	  }
	  if (msg->ReplyRequested){
	    HostUser->CancelReply(msg->ExternalId, rc);
	  }
	  
	}
	break;

/*
** Handle a time message
*/

      case TIME_MESSAGE_ID:
//	HostUser->TimeReply(gvt, ((TIME_MESSAGE *)Message)->ExternalId);
	HostUser->TimeReply(time, ((TIME_MESSAGE *)Message)->ExternalId);
	delete [] (char *)Message;

	break;

/*
** Handle a MinMsgTime message
*/

      case MIN_MSG_TIME_ID:
	HostUser->MinMsgTimeReply
	  (*IgnoreTimeGrantGvt, ((MIN_MSG_TIME_MESSAGE *)Message)->ExternalId);

	delete [] (char *)Message;

	break;

/*
** Handle a advance request message
*/

      case TIME_ADVANCE_REQUEST_MESSAGE_ID:

	ExternalId = ((TIME_ADVANCE_REQUEST_MESSAGE *)Message)->ExternalId;
	TimeLag = ((TIME_ADVANCE_REQUEST_MESSAGE *)Message)->TimeLag;
	OldTime = ((TIME_ADVANCE_REQUEST_MESSAGE *)Message)->OldTime;

#ifdef COMMENTOUT
//	cout << "Time Advance Request gvt " << gvt << endl;
	cout << "Time Advance Request time " << time << endl;
	cout << "Time Advance Request OldTime " << OldTime << endl;
	cout << "Time Advance Request TimeLag " << TimeLag << endl;
	cout << "NumBarriers = " << GvtBarriers.GetNumElements() << endl;
#endif

//	if (OldTime < gvt) {
	if (OldTime < time) {

//	  HostUser->TimeAdvanceGrant(gvt, ExternalId);
	  HostUser->TimeAdvanceGrant(time, ExternalId);
	  delete [] (char *)Message;

	  BarrierTime = (SpSimTime *)GvtBarriers.Find(ExternalId);
	  if (BarrierTime) {
//	    cout << "Found BarrierTime" << endl;
//	    BarrierTime->SetTime(gvt+TimeLag);
	    BarrierTime->SetTime(time+TimeLag);
	  }else{
//	    cout << "Did not find BarrierTime" << endl;
//	    BarrierTime = new SpSimTime(gvt+TimeLag);
	    BarrierTime = new SpSimTime(time+TimeLag);
	    GvtBarriers.Insert(BarrierTime, ExternalId);
	  }

	}else{

//	  cout << "Scheduling Time Advance Grant" << endl;
	  HostUser->ScheduleTimeAdvanceGrant(TimeLag, ExternalId);
	  BarrierTime = (SpSimTime *)GvtBarriers.Find(ExternalId);
	  if (BarrierTime) {
	    BarrierTime->SetTime(OldTime+TimeLag);
	  }
	  else{
	    BarrierTime = new SpSimTime(OldTime+TimeLag);
	    GvtBarriers.Insert(BarrierTime, ExternalId);
	  }

	}

	break;

    case GO_TO_TIME_REQUEST_MESSAGE_ID:
      
      
      ExternalId = ((GO_TO_TIME_REQUEST_MESSAGE *)Message)->ExternalId;
      TimeLag =    ((GO_TO_TIME_REQUEST_MESSAGE *)Message)->TimeLag;
      NewTime =    ((GO_TO_TIME_REQUEST_MESSAGE *)Message)->NewTime;
 
//      if (NewTime < gvt) {
      if (NewTime < time) {

//	HostUser->GrantGoToTime(gvt, ExternalId);
	HostUser->GrantGoToTime(time, ExternalId);
	delete [] (char *)Message;
//	NewTime = gvt+TimeLag;
	NewTime = time+TimeLag;
      }else{
	HostUser->ScheduleGoToTimeGrant(NewTime, TimeLag, ExternalId);
      }
      qblock = objects->get_qblock();
      item = qblock->deque(ExternalId);
 
      if (!item) {
	item = new C_ITEM();
	item->set_id(ExternalId);
      }
      item->set_time_tag(NewTime);
 
      SetBlock(item->get_time_tag(), ExternalId, item);

      break;
 
/*
** Handle a disconnect barrier message
*/

      case DISCONNECT_BARRIER_MESSAGE_ID:

	cout << "Got a disconnect barrier message" << endl;

	ExternalId = ((DISCONNECT_BARRIER_MESSAGE *)Message)->ExternalId;
	delete [] (char *)Message;

	BarrierTime = (SpSimTime *)GvtBarriers.Remove(ExternalId);
	if (BarrierTime) {
	  delete BarrierTime;
	}else{
	  cout << "Error in disconnecting barrier: " << ExternalId
	       << " could not be found" << endl;
	}

#ifdef COMMENTOUT
	qblock = objects->get_qblock();
	item = qblock->deque(ExternalId);

	if (item) {
	  delete item;
	}else{
	  cout << "Error in disconnecting barrier: " << ExternalId
	       << " could not be found" << endl;
	}
#endif

	HostUser->UnScheduleTimeAdvanceGrant(ExternalId);
	HostUser->UnScheduleGoToTimeGrant(ExternalId);

	break;

/*
** Handle a Pause message by incrementing the PauseFlag
*/

      case PAUSE_MESSAGE_ID:

	cout << "Got a Pause message" << endl;
	PauseFlag++;

	break;

/*
** Handle a Resume message by decrimenting the PauseFlag. Note that we never
** want the PauseFlag to go negative.
*/

      case RESUME_MESSAGE_ID:

	cout << "Got a Resume message" << endl;
	if (PauseFlag > 0) PauseFlag--;

	break;

/*
** Handle a Named Pause message
*/

      case NAMED_PAUSE_MESSAGE_ID:

	NamedPauseMessage = (NAMED_PAUSE_MESSAGE *)Message;
	NamedPause = (char *)&NamedPauseMessage[1];
	cout << "Got a NamedPause message: " << NamedPause
	     << " at time " << NamedPauseMessage->Time << endl;

//	if (gvt < NamedPauseMessage->Time) {
	if (time < NamedPauseMessage->Time) {
	  NamedPauses.Insert(new SpSimTime(NamedPauseMessage->Time), NamedPause);
	}else{
//	  NamedPauses.Insert(new SpSimTime(gvt), NamedPause);
	  NamedPauses.Insert(new SpSimTime(time), NamedPause);
	}

	break;

/*
** Handle a Resume message by decrimenting the PauseFlag. Note that we never
** want the PauseFlag to go negative.
*/

      case NAMED_RESUME_MESSAGE_ID:

	NamedResumeMessage = (NAMED_RESUME_MESSAGE *)Message;
	NamedResume = (char *)&NamedResumeMessage[1];
	cout << "Got a Named Resume message: " << NamedResume << endl;

	NamedPauseTime = (SpSimTime *)(NamedPauses.Remove(NamedResume));

	if (NamedPauseTime) {
	  cout << "Removed named pause " << NamedResume
	       << " at time " << *NamedPauseTime
	       << endl;
	  delete NamedPauseTime;
	}else{
	  cout << "Could not remove named pause " << NamedResume << endl;
	}

	break;


/*
** Handle a Kill SPEEDES message.
*/

      case KILL_SPEEDES_MESSAGE_ID:

	cout << "Got a Kill SPEEDES message" << endl;
	KillSPEEDES = 1;

	break;

/*
** Handle a subscribe message.
*/

      case SUBSCRIBE_MESSAGE_ID:

	cout << "Got a subscribe message" << endl;

	SubscribeMessage = (SUBSCRIBE_MESSAGE *)Message;
	SubItem = new SUBSCRIPTION_ITEM();
	SubItem->UnPackBuffer((char *)&SubscribeMessage[1]);

	Subscribe_Mess = (SUBSCRIBE_MESS *)
	  new char[sizeof(SUBSCRIBE_MESS) + SubscribeMessage->bytes];

	memcpy((char *)&Subscribe_Mess[1],
	       (char *)&SubscribeMessage[1],
	       SubscribeMessage->bytes);

	Subscribe_Mess->time_tag = time;
	Subscribe_Mess->evtype = evtype->get_messid("SUBSCRIBE");
	Subscribe_Mess->obtype = SubItem->GetObjectType();
	Subscribe_Mess->objid = SubItem->GetLocalId();
	Subscribe_Mess->ExternalId = SubItem->GetExternalId();
	Subscribe_Mess->bytes = SubscribeMessage->bytes;

	ext_event = (SpEvent *)evtype->message_event((char *)Subscribe_Mess);

	// comment out since Header has not been assigned.
	//delete [] (char *)Header;

	delete SubItem;
	delete [] (char *)Message;

	break;


/*
** Handle an UnSubscribe message.
*/

      case UNSUBSCRIBE_MESSAGE_ID:

	cout << "Got an UnSubscribe message" << endl;

	UnSubscribeMessage = (UNSUBSCRIBE_MESSAGE *)Message;
	SubItem = new SUBSCRIPTION_ITEM();
	SubItem->UnPackBuffer((char *)&UnSubscribeMessage[1]);

	UnSubscribe_Mess = (UNSUBSCRIBE_MESS *)
	  new char[sizeof(UNSUBSCRIBE_MESS) + UnSubscribeMessage->bytes];


	memcpy((char *)&UnSubscribe_Mess[1],
	       (char *)&UnSubscribeMessage[1],
	       UnSubscribeMessage->bytes);

	UnSubscribe_Mess->time_tag = time;
	UnSubscribe_Mess->evtype = evtype->get_messid("UNSUBSCRIBE");
	UnSubscribe_Mess->obtype = SubItem->GetObjectType();
	UnSubscribe_Mess->objid = SubItem->GetLocalId();
	UnSubscribe_Mess->ExternalId = SubItem->GetExternalId();
	UnSubscribe_Mess->bytes = UnSubscribeMessage->bytes;

	ext_event = (SpEvent *)evtype->message_event((char *)UnSubscribe_Mess);

	// comment out since Header has not been assigned.
	//delete [] (char *)Header;

	delete SubItem;
	delete [] (char *)Message;

	break;

/*
** Default should never happen.
*/

      default:

	cout << "Error, unknown external message type " << HandlerNum << endl;
	break;


    } // end of switch/case statement

  } // end of if message found block

/*
** Return the external event (call the old external event routine for backward
** compatibility if there are no new external events with newer host router)
*/

  if (ext_event) {
    return ext_event;
  }else{
    return old_external_event(time);
  }

}


/************************************************************************
* old_external_event - check if there are external messages			*
************************************************************************/
SpEvent *SpTimeMgtAlg::old_external_event(SpSimTime& time) { //GEB2
  C_EM_HEADER *EM_header;
  C_EM_COMMAND *EM_command;
  SpEvent *ext_event = NULL;
  char *message;
  C_ITEM *item;

//...... this is a poor kludge to get around polling

/*
  polling_counter++;
  if (polling_counter < 10) return NULL;
  polling_counter = 0;
*/

//...... check external commands messages

  EM_header = host_user->get_message();
  if (EM_header == NULL) return NULL;

  message = (char *)EM_header;

  if (EM_header->EM_interaction & EM_QUERY) {
    EM_header->time_tag = time;
    EM_header->evtype = QUERY;
    ext_event = (SpEvent *)evtype->message_event(message);
  }

  if (EM_header->EM_interaction & EM_MONITOR) {
    EM_header->time_tag = time;
    EM_header->evtype = MONITOR;
    ext_event = (SpEvent *)evtype->message_event(message);
  }

  if (EM_header->EM_interaction & EM_COMMAND) {
    EM_command = (C_EM_COMMAND *)EM_header;
    EM_header->time_tag = time;
    EM_header->evtype = evtype->get_messid(EM_command->EM_command);
    ext_event = (SpEvent *)evtype->message_event(message);
  }

  if (EM_header->EM_interaction & EM_MODULE) {

    if (EM_header->EM_interaction & EM_DISCONNECT) {

        item = new C_ITEM();
        item->set_id(EM_header->EM_socket);
        objects->add_unblock(item);

        objects->update_block();
        ext_event = NULL;

        fprintf(stderr,"DISCONNECTING EXTERNAL MODULE\n");

    }else{

      EM_command = (C_EM_COMMAND *)EM_header;
      if (EM_header->EM_interaction & EM_AGGRESSIVE)
		EM_header->time_tag = time;
      if (EM_header->time_tag < time) EM_header->time_tag = time;
      EM_header->evtype = evtype->get_messid(EM_command->EM_command);
      ext_event = (SpEvent *)evtype->message_event(message);

      if (EM_header->EM_interaction & EM_BLOCKING) {
//        objects->unblock(EM_header->EM_socket);
        item = new C_ITEM();
        item->set_id(EM_header->EM_socket);
        objects->add_unblock(item);
        objects->update_block();
      }

//      fprintf(stderr,"CONNECTING EXTERNAL MODULE\n");

    }

  }

  //delete message; //RVI: I don't think we should be deleting here...
  return ext_event;

}

/************************************************************************
* get_blocked_messages - get messages that are blocking SPEEDES		*
************************************************************************/
void SpTimeMgtAlg::get_blocked_messages() {
  int ExternalId;
  double TimeLag;
  static SpSimTime OldGvt = -1.0e20;

  if (SpComm_GlobalSum(KillSPEEDES)) {
    delete HostUser;
    HostUser = NULL;
    exit(0);
  }

  if (OldGvt == gvt) return;


/*
** Flush out any messages through the host router if necessary and then do a
** global reduction to make sure that all messages have been sent and
** flushed before sending out time messages. The global reduction also checks
** if we should exit SPEEDES because of a killSPEEDES request.
*/

  HostUser->Flush();
/*
** Send out any scheduled time advance grants here
*/

  while (HostUser->GetScheduledTimeAdvanceGrant(TimeLag, ExternalId)) {
    HostUser->TimeAdvanceGrant(gvt, ExternalId);

#ifdef COMMENTOUT
    cout << "Removing Sheduled Time Advance Grant with TimeLag"
	 << TimeLag << endl;
#endif

    SpSimTime *BarrierTime = (SpSimTime *)GvtBarriers.Find(ExternalId);
    *BarrierTime = OldGvt + TimeLag;
    if (*BarrierTime < gvt){
      *BarrierTime = OldGvt + TimeLag;
    }

#ifdef COMMENTOUT
    C_QUEUE *qblock = objects->get_qblock();
    C_ITEM *item = qblock->deque(ExternalId);

    if (!item) {
      item = new C_ITEM();
      item->set_id(ExternalId);
    }

    item->set_time_tag(gvt+TimeLag);
    qblock->insert(item);
#endif
  }

  /*
  ** Send out any go to time requests here if we have reached their times
  */
  
  double blkTime;
  int NumGoTo = HostUser->GetNumGoToTimeGrants();
  for (int jj = 0; jj < NumGoTo; ++jj){
    HostUser->GetScheduledGoToTimeGrant(blkTime, TimeLag, ExternalId);
    if (gvt >= blkTime){
      HostUser->GrantGoToTime(gvt, ExternalId);
      C_QUEUE *qblock = objects->get_qblock();
      C_ITEM *item = qblock->deque(ExternalId);
      item->set_time_tag(gvt+TimeLag);
      qblock->insert(item);
    }
    else{
      // Stick it back on the list for processing since we have not reached
      // the goto time yet.
      HostUser->ScheduleGoToTimeGrant(blkTime, TimeLag, ExternalId);
    }
  }
  
  OldGvt = gvt;

/*
** Stuff for the old host router
*/

  host_user->flush();

}

/************************************************************************
* get_next_event - get the next event out of the event list		*
************************************************************************/
SpEvent *SpTimeMgtAlg::get_next_event() {
  SpEvent *event;
  SpSimObj *obj;
  SpPriorityQueueHeap *lrq;
  C_RQ_HOLDER *rq_holder;
  C_RQ_HOLDER *lrqh;

//...... loop until an event is found (or the queue of objects is empty)

  while (rqueue.get_length()) {

//...... get the object

    rq_holder = (C_RQ_HOLDER *)rqueue.pop();
    if (rq_holder == NULL) return NULL;

    obj = (SpSimObj *)rq_holder->get_pointer();

//...... a NULL obj means try again

    if (obj == NULL) {

      rqueue.delete_holder(rq_holder);

    }else{

//...... get the local rqueue from the object and the holder for its event

      lrq = obj->get_local_rqueue();
      lrqh = (C_RQ_HOLDER *)lrq->pop();

//...... make sure that there is a holder for the event

      if (lrqh) {

//...... get the event and then delete the holder

        event = (SpEvent *)lrqh->get_pointer();
	rqueue.delete_holder(lrqh);

//...... check if the object has any more events

        if (lrq->get_length() == 0) {
	  rqueue.delete_holder(rq_holder);
          obj->set_rq_holder(NULL);
	}else{
	  rq_holder->set_time_tag(lrq->get_time());
	  rqueue.insert(rq_holder);
	}

        if (event) return event;

      }

    }

  }

  return NULL;

}


/************************************************************************
* gather_statistics - gather statistics for performance analysis	*
************************************************************************/
void SpTimeMgtAlg::gather_statistics() {

  statistics->set_cycle();
  statistics->set_gvt(gvt);
  statistics->set_phase1(n_temp);
  statistics->set_phase2(n_perm);
  statistics->set_events(n_perm);
  statistics->set_rollbacks(n_roll);
  statistics->set_cancels(n_cancel);
  statistics->set_messages(n_messrec);
  statistics->set_antimessages(n_amessrec);
  statistics->set_exmessages(n_exmess);

  statistics->SetTotalProcessingTime(TotalProcessingTime);
  statistics->SetTotalCommittedTime(TotalCommittedTime);
  statistics->SetLocalTime(rqueue.get_time());

}
void SpTimeMgtAlg::SetBlock(SpSimTime& time, int externalId, C_ITEM* item){

  C_QUEUE* qblock = objects->get_qblock();
  double blockTime;
  double gtLagTime;
  int GoToTime = HostUser->FindScheduledGoToTimeGrant(blockTime,
							gtLagTime,
							externalId);
  
  if (!item){
    item = qblock->deque(externalId);
    if (!item) {
      item = new C_ITEM();
      item->set_id(externalId);
    }
  }

  if ((blockTime <= time) && (GoToTime)){
    item->set_time_tag(blockTime);
  }
  else{
    item->set_time_tag(time);
  }
  qblock->insert(item);
}

void SpTimeMgtAlg::CleanUpExternallyScheduledEvents()
{
  SpDynPtrArray listToBeRemoved;
  int i;
  int rbid;
  int numToBeRemoved = 0;
  int numInList = ExternallyScheduledEvents.GetNumElements();
  OldestExternallyScheduledEvent = INFINITY;
  SpCancelHandle *ch = 
    (SpCancelHandle*)ExternallyScheduledEvents.GetFirstElement();
  
  for (i=0; i < numInList; ++i){
    if (ch->get_time_tag() < gvt){
      listToBeRemoved[numToBeRemoved++] = ch;
    }
    else if (ch->get_time_tag() < OldestExternallyScheduledEvent){
      OldestExternallyScheduledEvent = ch->get_time_tag();
    }
    ch =(SpCancelHandle*) ExternallyScheduledEvents.GetNextElement();
  }
  for (i=0; i < numToBeRemoved; ++i){
    rbid = ((SpCancelHandle*)(listToBeRemoved[i]))->get_rbid();
    ExternallyScheduledEvents.Remove(rbid);
    twosmess->delete_antimess((SpAntiMsg*)(listToBeRemoved[i]));
  }
}

extern SpPoDataBase *MessageDataBase;

void SpTimeMgtAlg::GetObjectsFromCheckpoint(SpEvent*)
{
  char tempBuffer[6*sizeof(int)];
  int bufferSize;
  char* persistenceBuffer;
  double time = RestartTime;
  char fName[1024];
  FILE* inFile = NULL;
  int i;

  for (i=0; i < 7200; ++i){
    sprintf(fName, "OBJECTS_CheckPoint_%s_TIME_%d_NODE_%d",
	    RestartType, (int)(time-i), SpComm_GetNodeId());
    inFile = fopen(fName, "rb");
    if (inFile){
      cout<<"Closest checkpointed time is "<<time-i
	  <<" with checkpoint file name "<<fName<<endl;
      time -= i;
      break;
    }
  }
  if (inFile == NULL){
    cout<<"Could not find a checkpoint file at time "<<time
	<<" or up to two hours prior to it.  Exiting"<<endl;
    exit(-2);
  }
  RestartTime = time;
    
  fread(tempBuffer, sizeof(char), 6*sizeof(int), inFile);
  bufferSize = PO_BUFFER_SIZE(tempBuffer);
  ::rewind(inFile); //:: means use the global name space
  persistenceBuffer = new char[bufferSize];
  fread(persistenceBuffer, sizeof(char), bufferSize, inFile);
  PO_RESTORE(persistenceBuffer);
  fclose(inFile);
  delete [] persistenceBuffer;

  objects = PO_GET_FIRST_SpMasterSimObjMgr(); //There should only be one;

  for (i=0; i < objects->GetNmanagers(); ++i){
    SpSimObjMgr* mgr = (SpSimObjMgr*)objects->get_objman(i);
    char* name = mgr->get_NAME();
    char* poName = new char [strlen(name) +1 + 6];
    mgr->reserve_n_objects(mgr->get_N_LOC());
    for (int j=0; j < mgr->get_N_LOC(); ++j){
      sprintf(poName, "%s%d", name, j);
      SpSimObj* object = (SpSimObj*)PO_LOOKUP_NAME(poName);
      if (object){
	mgr->define_object(object);
      }
      else{
	cout<<"Could not find object "<<poName<<" for restart"<<endl;
      }
    }
    delete [] poName;
  }

}


void SpTimeMgtAlg::GetMessagesFromCheckpoint(SpEvent* InitialEvent, int &len)
{
  char tempBuffer[6*sizeof(int)];
  int bufferSize;
  char* persistenceBuffer;
  char fName[1024];
  FILE* inFile;

  len = 0;
  sprintf(fName, "MESSAGES_CheckPoint_%s_TIME_%d_NODE_%d",
	  RestartType, (int)RestartTime, SpComm_GetNodeId());
  inFile = fopen(fName, "rb");
  if (inFile == NULL){
    cout<<"Could not open "<<fName<<endl;
    exit(-1);
  }
  fread(tempBuffer, sizeof(char), 6*sizeof(int), inFile);
  bufferSize = PO_BUFFER_SIZE(tempBuffer);
  ::rewind(inFile); //:: means use the global name space
  persistenceBuffer = new char[bufferSize];
  fread(persistenceBuffer, sizeof(char), bufferSize, inFile);
  PO_RESTORE(persistenceBuffer, MessageDataBase);
  fclose(inFile);
  delete [] persistenceBuffer;

  char* msg = PO_GET_FIRST_BLOCK_char(MessageDataBase);
  while (msg){
    SpMsg* header = (SpMsg *)msg;
    header->cancel_id = -1;
    SpEvent* event = (SpEvent *)evtype->message_event(msg);
    push_bot(event);
    msg = PO_GET_NEXT_BLOCK_char(MessageDataBase);
    ++len;
  }
  push_top(InitialEvent);
}
