// SpTimeMgtAlg.H header file

#ifndef evtq_object
#define evtq_object

#include "speedes_comm.H"
#include "queue.H"
#include "SpPriorityQueueHeap.H"
#include "SpMasterSimObjMgr.H"
#include "SpEventTypes.H"
#include "random.H"
#include "SpEvent.H"
#include "SpStats.H"
#include "SpOptMessager.H"
//#include "cycle.H"
#include "rewind.H"
#include "SpSimTime.H" //GEB2
#include "SpBinaryTree.H"
#include "SpStopWatch.H"

/************************************************************************
*	evtq object 							*
************************************************************************/
class SpTimeMgtAlg : public C_QUEUE {

  friend void HandleCommandLine(int argc,
				char **argv,
				int &nTot,
				int &nLoc,
				int &group,
				char ** pInfo);
    
  private:

  protected:

    SpPriorityQueueHeap rqueue;	// event list data structure

    double minstep;		// minimum time step
    SpSimTime localt;		// local safe time to process  GEB2
    SpSimTime gvt;	        // global virtual time for all nodes GEB2
    SpSimTime gvtmax;		// maximum global virtual time GEB2
    SpSimTime tend;   	        // end time for simulation GEB2
    double gvt_rate;		// gvt update rate
    int scaled_time;		// scaled time flag
    int n_max;			// max number of temporary events
    int n_min;			// min number of temporary events
    double Tasb;		// Wall time for asynchronous broadcasts
    double Tgvt;		// Wall time before GVT request
    int Ngvt;			// # of events before GVT request
    int Nrisk;			// # of events beyond GVT allowed with risk
    int Nopt;			// # optimistic events allowed beyond GVT
    int n_temp;			// number of temporary events processed
    int n_perm;			// number of permanent events processed
    int n_roll;			// number of events rolled back
    int n_cancel;		// number of events canceled
    int n_exmess;		// number of messages received
    int n_messrec;		// number of messages received
    int n_amessrec;		// number of antimessages received
    int one_node_flag;		// one node flag
    int test_rollback;		// test rollback flag (used by seq_evtq)
    int check_reptime;		// test if events have same time tag
    int interactive;		// interactive simulation flag
    double spin;		// spin loop variable in sec

    int MemoryUsage;		// print out memory statistics
    int MessageSending;		// print out message statistics
    int EventProcessing;	// print out event processing statistics
    int ObjectProcessing;	// print out object processing statistics
    int CriticalPath;		// print out critical path statistics

    SpStats *statistics;	// statistics object
    SpOptMessager *twosmess;	// message and antimessage facility
    SpMasterSimObjMgr *objects;	// manager of all objects in the simulation
    SpEventTypes *evtype;	// event type and message type manager
    C_RANDOM *random;		// random number generator
    C_QUEUE q1nd;		// queue of messages for one node
    C_HOST_USER *host_user;	// host user object for interactive messaging
    C_REWIND *rewind;		// rewind object
    int QUERY;			// id for query events
    int MONITOR;		// id for monitor events
    //int SUBSCRIBE;		// id for subscribe events
    //int UNSUBSCRIBE;		// id for unsubscribe events
    int EVENT_HANDLER_EVENT;   	// id for event handler event

    SpBinaryTree NamedPauses;	// named pause container class
    SpBinaryTree GvtBarriers;	// barriers used to limit gvt
    SpStopWatch StopWatch;	// usefull stopwatch for doing timings
    int *NumEventsProcessed;  	// array number of events processed by type
    int *NumEventsCommitted;  	// array number of events committed by type
    double *CPU_EventTime;  	// array of summed event processing times
    double *CPU_EventTimeCommitted; // array of summed committed event
                                    // processing times

    double TotalProcessingTime;	// total amount of event processing time
    double TotalCommittedTime;	// total amount of committed processing time

    int *NumMessages;		// array of message counters
    int *NumCoordMessages;	// array of coordinated message counters
    int *NumAntimessages;	// array of antimessage counters

    int PauseFlag;		// Flag indicating whether SPEEDES is paused
    int KillSPEEDES;		// Flag for killing SPEEDES processes

    static SPEEDES *Speedes;    // global SPEEDES object
    static int node;		// node
    static int n_nodes;		// number of nodes

    static char *mr_type;	// type of multirouter
    static int mr_bytes;	// number of bytes for multirouter
    static int mr_thresh;	// multirouter send thresh in bytes
    static int lazy;		// flag for lazy cancellation
    static int rewind_enable;	// flag for rewind enabling
    static int rewind_buff_size; // buffer size for rewind
    static int trace;		// trace flag
    static char *tracefile;	// name of trace file
    static int WriteLookahead;	// write lookahead file flag
    static char *Lookaheadfile;	// name of output lookahead file
    double OldestExternallyScheduledEvent;

    SpBinaryTree ExternallyScheduledEvents;

    void CleanUpExternallyScheduledEvents();
    static int CheckPointing;
    int Restart;
    char* RestartType;
    double RestartTime;


    void GetObjectsFromCheckpoint(SpEvent* InitialEvent);
    void GetMessagesFromCheckpoint(SpEvent* InitialEvent, int& len);

  public:

    SpTimeMgtAlg();
    virtual ~SpTimeMgtAlg();
    void set_speedes_object(SPEEDES *sp) { Speedes = sp; }
    void set_node_info(int nds, int nd);
    void set_statistics(SpStats *s) {statistics = s;}
    void gather_statistics();
    void set_minstep(double t) {minstep = t;}
    void set_test_rollback(int trb) {
		test_rollback = trb;
		if (trb) SpCurrentAltMgr->SetOptimistic();
	}
    void set_check_reptime(int cr) {check_reptime = cr;}
    void set_lazy(int l) {lazy = l;}

    void set_scaled_time(int s) {scaled_time = s;}
    int get_scaled_time() {return scaled_time;}

    void set_interactive(int i) {interactive = i;}
    int get_interactive() {return interactive;}

    void set_spin(double s) {spin = s;}
    double get_spin() {return spin;}

    void set_rewind_enable(int r) {rewind_enable = r;}
    void set_rewind_buff_size(int s) {rewind_buff_size = s;}

    void save_rewind();

    void set_n_max(int nmax) {n_max = nmax;}
    void set_n_min(int nmin) {n_min = nmin;}

    void SetBlock(SpSimTime& time, int externalId, C_ITEM* item = NULL);

    void set_trace() {trace = 1;}
    void reset_trace() {trace = 0;}
    void set_tracefile(char *tf) {tracefile = tf;}

    void set_WriteLookahead() {WriteLookahead = 1;}
    void reset_WriteLookahead() {WriteLookahead = 0;}
    void set_Lookaheadfile(char *lf) {Lookaheadfile = lf;}

    void set_tend(SpSimTime& t) {tend = t;}  //GEB2
    void set_gvt(SpSimTime& t) {gvt = t;}    //GEB2
    void set_gvt_rate(double t) {gvt_rate = t;}
    SpSimTime& get_gvt() {return gvt;}       //GEB2
    SpSimTime& get_gvtmax() {return gvtmax;}  //GEB2
    void set_gvtmax(SpSimTime& gmax) {gvtmax = gmax;}  //GEB2
    void set_gvtmax(double gmax) {
      gvtmax.SetTime(gmax);
      gvtmax.SetPriority1(MaxPriority1);
      gvtmax.SetPriority2(MaxPriority2);
    }  //GEB2

    int get_ntemp() {return n_temp;}
    int get_nperm() {return n_perm;}
    int get_nroll() {return n_roll;}
    int get_ncancel() {return n_cancel;}
    void print() {objects->print();}

    SpEvent *get_next_event();

    void set_multirouter_type(char *typ) {mr_type = typ;}
    void set_multirouter_bufsize(int nb) {mr_bytes = nb;}
    void set_multirouter_thresh(int thr) {mr_thresh = thr;}

    SpEvent *old_external_event(SpSimTime& time); //GEB2
    SpEvent *external_event(SpSimTime& time);

    virtual void get_blocked_messages();
    virtual void temp_process();
    virtual void perm_process();
    virtual void find_gvt();

    void term(double tnd);

    virtual int RollBack(){return 0;};	//Rolls back all events to GVT
    virtual int RollsForward(){return 0;}; //Rolls all events forward as far as possible

  static int DoingCheckpoints(){ return CheckPointing;}
  void SetDoingCheckpoints(){ CheckPointing=1;}
  void UnsetDoingCheckpoints(){ CheckPointing=0;}

};


typedef SpTimeMgtAlg C_EVTQ;

#endif
