// dbl_queue.C method file

#include "SpStdio.H"
#include "dbl_queue.H"

/***********************************************************************
* C_DBL_QUEUE : construct q dbl_queue object				*
************************************************************************/
C_DBL_QUEUE::C_DBL_QUEUE() {

  length = 0;
  top = NULL;
  bot = NULL;

}

/************************************************************************
* push_top : pushes an item onto the top of the dbl_queue		*
************************************************************************/
void C_DBL_QUEUE::push_top(C_DBL_ITEM *dbl_item) {

  if (length == 0) {
    bot = dbl_item;
    top = dbl_item;
    dbl_item->set_next(dbl_item);
    dbl_item->set_previous(dbl_item);
    length = 1;
  }else{
    top->set_previous(dbl_item);
    dbl_item->set_next(top);
    top = dbl_item;
    length++;
  }

}

/************************************************************************
* push_bot : pushes an item onto the bottom of the dbl_queue		*
************************************************************************/
void C_DBL_QUEUE::push_bot(C_DBL_ITEM *dbl_item) {

  if (length == 0) {
    bot = dbl_item;
    top = dbl_item;
    dbl_item->set_next(dbl_item);
    dbl_item->set_previous(dbl_item);
    length = 1;
  }else{
    bot->set_next(dbl_item);
    dbl_item->set_previous(bot);
    bot = dbl_item;
    length++;
  }

}

/************************************************************************
* pop_bot : pops an item off of the bottom of the dbl_queue		*
************************************************************************/
C_DBL_ITEM *C_DBL_QUEUE::pop_bot() {
  C_DBL_ITEM *dbl_item = NULL;

  if (length != 0) {
    dbl_item = bot;
    bot = bot->get_previous();
    length--;
  }
  return dbl_item;
}

/************************************************************************
* insert : inserts an item into the  dbl_queue				*
************************************************************************/
void C_DBL_QUEUE::insert(C_DBL_ITEM *dbl_item) {
  C_DBL_ITEM *temp;
  int i;
  SpSimTime v; //GEB2

//...... case where the list is empty

  if (length == 0) {
    top = dbl_item;
    bot = dbl_item;
    length = 1;
    return;
  }

  v = dbl_item->get_valu();

//...... item has minimum value in list

  if (v < top->get_valu()) {
    push_top(dbl_item);
    return;
  }

//...... item has maximum value (or equal to) in list

  if (v >= bot->get_valu()) {
    push_bot(dbl_item);
    return;
  }

//...... insert from the bottom up (better than top down for near fifo queues)

  temp = bot->get_previous();
  for (i=1; i<length; i++) {
    if (v >= temp->get_valu()) {
      dbl_item->set_previous(temp);
      dbl_item->set_next(temp->get_next());
      temp->set_next(dbl_item);
      temp = dbl_item->get_next();
      temp->set_previous(dbl_item);
      length++;
      return;
   }
    temp = temp->get_previous();
  }

  fprintf(stderr,"ERROR IN DBL_QUEUE : INSERT\n");
  dbl_item->print();

}

/************************************************************************
* remove : remove an item from the  dbl_queue				*
************************************************************************/
void C_DBL_QUEUE::remove(C_DBL_ITEM *dbl_item) {
  C_DBL_ITEM *previous;
  C_DBL_ITEM *next;

  if (length == 0) return;

  if (length == 1) {
    if (dbl_item != top) return;
    top = NULL;
    bot = NULL;
    length = 0;
    return;
  }

  if (dbl_item == top) {
    top = dbl_item->get_next();
    top->set_previous(NULL);
    length--;
    return;
  }

  if (dbl_item == bot) {
    bot = dbl_item->get_previous();
    bot->set_next(NULL);
    length--;
    return;
  }

  previous = dbl_item->get_previous();
  next = dbl_item->get_next();
  previous->set_next(next);
  next->set_previous(previous);
  length--;

}
/************************************************************************
* check_order : check the order of the  dbl_queue			*
************************************************************************/
void C_DBL_QUEUE::check_order() {
  C_DBL_ITEM *temp;
  int i;
  SpSimTime v; //GEB2

  if (length <= 1) return;

  v = top->get_valu();
  temp = top->get_next();

  for (i=1; i<length; i++) {
    if (v > temp->get_valu()) {
      fprintf(stderr,"DBL_QUEUE check order error\n");
    }
    v = temp->get_valu();
    temp = temp->get_next();
  }

}

