// hash.C method file

#include "SpStdio.H"
#include "SpString.H"

#include "hash.H"

/************************************************************************
* C_HASH : construct a hash object					*
************************************************************************/
C_HASH::C_HASH(int n) {

  size = n;
  q = new C_QUEUE[size];
  nitems = 0;

//  printf("HASH object created\n");

}

/************************************************************************
* reset : reset the hash object 					*
************************************************************************/
void C_HASH::reset() {
  int i;

  for (i=0; i<size; i++) q[i].reset();
  nitems = 0;

}

/************************************************************************
* add : add the item to the hash 					*
************************************************************************/
void C_HASH::add(C_ITEM *item) {
  int mid;

  mid = (item->get_id())%size;
  q[mid].push_bot(item);
  nitems++;

}

/************************************************************************
* get_hashid : get the hash Id based on a name				*
************************************************************************/
int C_HASH::get_hashid(char *name) {
  int HashId;
  int len;
  int i;

  HashId = 0;
  len = strlen(name);
  for (i=0; i<len; i++) {
    HashId ^= name[i] << ((i*8)%32);
  }

  return (HashId & 0x7fffffff);

}

/************************************************************************
* get : get an item from the hash using the item's name			*
************************************************************************/
C_ITEM *C_HASH::get(char *name) {
  int hid;
  int mid;
  int i;
  int len;
  C_ITEM *item;

  hid = get_hashid(name);
  mid = hid%size;

  item = q[mid].get_top();
  len = q[mid].get_length();
  for (i=0; i<len; i++) {
    if (!strcmp(item->GetName(), name)) return item;
    item = item->get_link();
  }

  return NULL;

}

/************************************************************************
* combine : combine all of the items in a queue				*
************************************************************************/
void C_HASH::combine() {
  int i;

  qall.reset();
  for (i=0; i<size; i++) {
    qall.join(q+i);
  }

}

/************************************************************************
* remove : remove this item from the hash				*
************************************************************************/
void C_HASH::remove(C_ITEM *item) {
  int mid;
  int len;

  mid = (item->get_id())%size;
  len = q[mid].length();
  q[mid].remove(item);
  if (q[mid].length() != len) {
    nitems--;
    return;
  }

  fprintf(stderr,"ERROR, could not remove the hash id\n");

}

/************************************************************************
* remove : remove the item with this id from the hash			*
************************************************************************/
void C_HASH::remove(int id) {
  int mid;
  int i;
  int len;
  C_ITEM *item;

  mid = (id)%size;
  item = q[mid].ptop();
  len = q[mid].length();
  for (i=0; i<len; i++) {
    if (item->get_id() == id) {
      q[mid].remove(item);
      nitems--;
      return;
    }
    item = item->get_link();
  }
  fprintf(stderr,"ERROR, could not remove the hash id\n");
}

/************************************************************************
* dehash : remove this item from the hash but dont delete it		*
************************************************************************/
void C_HASH::dehash(C_ITEM *item) {
  int mid;
  int len;

  mid = (item->get_id())%size;
  len = q[mid].length();
  q[mid].deque(item);
  if (q[mid].length() != len) {
    nitems--;
    return;
  }

  fprintf(stderr,"ERROR, could not remove the hash id\n");

}

/************************************************************************
* dehash : remove the item using id from the hash but dont delete it	*
************************************************************************/
C_ITEM *C_HASH::dehash(int id) {
  int mid;
  C_ITEM *item;

  mid = (id)%size;
  item = q[mid].deque(id);
  if (item != NULL) nitems--;
  return item;

}

/************************************************************************
* printh								*
************************************************************************/
void C_HASH::printh() {
  int i;

  for (i=0; i<size; i++) {
    if (q[i].length()) {
      fprintf(stderr,"HASH[%d]\n",i);
      q[i].print();
    }
  }

}



