// hash.H header file

#ifndef hash_object
#define hash_object

#include "queue.H"
#include "qitem.H"

/*****************************************************************************
* Class: C_HASH
* Summary: An array of queues with array like access to	items
* Description: 
*****************************************************************************/
class C_HASH  {

 public:

/******************************************************************************
* Method: dehash
* Summary: 
* Description: 
* Return: None
******************************************************************************/
  void dehash(C_ITEM *item);
  
/******************************************************************************
* Method: dehash
* Summary: 
* Description: 
* Return: C_ITEM
******************************************************************************/
  C_ITEM *dehash(int id);

/******************************************************************************
* Method: reset
* Summary: Clear the hash table.  Does NOT delete items.
* Description: Clear every queue in the hash array, then set length of
*              the hash array to zero
* Return: None
******************************************************************************/
  void reset();

/******************************************************************************
* Method: add
* Summary: Add a specified C_ITEM to the hash table
* Description: Push specified item to the bottom of the appropriate queue in
*              the hash array
* Return: None
******************************************************************************/
  void add(C_ITEM *item // C_ITEM to add
	   );

/******************************************************************************
* Method: get
* Summary: Return the C_ITEM in this hash table specified by item's ID
* Description: Gets the item from the appropriate queue in the hash array
* Return: The C_ITEM in this hash table specified by item's ID
******************************************************************************/
  C_ITEM *get(int hid // ID of C_ITEM to get
	      ) {int mid = hid%size; return q[mid].get_item(hid);}

/******************************************************************************
* Method: get
* Summary: Return the C_ITEM in this hash table specified by item's name
* Description: Gets the item from the appropriate queue in the hash array
* Return: The C_ITEM in this hash table specified by item's name
******************************************************************************/
  C_ITEM *get(char *name // name of C_ITEM to get
	      );

/******************************************************************************
* Method: get_hashid
* Summary: Return a C_ITEM's ID in this hash table specified by its name
* Description: 
* Return: The C_ITEM's ID in this hash table specified by its name
******************************************************************************/
  int get_hashid(char *name // name of C_ITEM from whom to get ID
		 );

/******************************************************************************
* Method: combine
* Summary: Concatonate each queue in this hash array into one queue, leaving
*          the original hash table intact
* Description: The resulting single queue can then be retrieved with the method
*              get_all () and printed with printq ().
* Return: None
******************************************************************************/
  void combine();

/******************************************************************************
* Method: get_all
* Summary: Get the combined queue generated by the last call to combine ()
* Description: 
* Return: The combined queue generated by the last call to combine ()
******************************************************************************/
  C_QUEUE *get_all() {return &qall;}

/******************************************************************************
* Method: remove
* Summary: Remove (delete) the specified item from this hash table
* Description: 
* Return: None
******************************************************************************/
  void remove(C_ITEM *item  // C_ITEM to remove
	      );

/******************************************************************************
* Method: remove
* Summary: Remove (delete) an item from this hash table specified by its ID
* Description: 
* Return: None
******************************************************************************/
  void remove(int id  // ID of C_ITEM to remove
	      );

/******************************************************************************
* Method: get_size
* Summary: Get the number of queues in this hash array
* Description: 
* Return: The number of queues in this hash array
******************************************************************************/
  int get_size() {return size;}

/******************************************************************************
* Method: get_nitems
* Summary: Get the number of C_ITEMs in this hash table
* Description: 
* Return: The number of C_ITEMs in this hash table
******************************************************************************/
  int get_nitems() {return nitems;}

/******************************************************************************
* Method: printq
* Summary: Print each C_ITEM in the combined hash table
* Description: Unlike printh (), which simply prints the hash table, this
*              method relies on first calling combine () which concatonates the
*              hash array into one combined queue. The resulting combined queue
*              is printed with this method by calling the virtual function
*              print () on all of the C_ITEMs in the combined queue.
* Return: None
******************************************************************************/
  void printq() {qall.print();}

/******************************************************************************
* Method: printh
* Summary: Print each C_ITEM in the hash table
* Description: This prints the hash table by calling the virtual function print
*              () on all of the C_ITEMs in the hash table. Unlike printq (),
*              this method does not rely on first calling combine ().
* Return: None
******************************************************************************/
  void printh();
  

/******************************************************************************
* Constructor Method: 
* Summary: Constructs a C_HASH with a hash array of specified length
* Description: 
******************************************************************************/
  C_HASH(int n  // length of hash array
	 );
   
  ~C_HASH() {if (q) delete [] q;}

 protected:

  int size;		// size of queue array
  int nitems;		// total number of items in the hash
  C_QUEUE *q;		// array of queues (dimension size)
  C_QUEUE qall;	        // a queue of concatonated hash queues.

};


#endif
