// Filename: holder.H

#ifndef holder_object
#define holder_object

#include "qitem.H"

/*****************************************************************************
* Class: C_HOLDER 
* Summary: A C_ITEM with an size-unlimited character buffer in it and the
*          node to which it is supposed to arrive when sent as a message
*****************************************************************************/
class C_HOLDER : public C_ITEM {

 public:

/******************************************************************************
* Method: get_buff
* Summary: Get the character buffer of which this object is a holder, in bytes
* Description: 
* Return: The buffer
******************************************************************************/
  char *get_buff() {return buff;}

/******************************************************************************
* Method: get_size
* Summary: Get the size of the character buffer of which this object is a
*          holder, in bytes
* Description: 
* Return: The size of the buffer, in bytes
******************************************************************************/
  int get_size() {return size;}

/******************************************************************************
* Method: set_node
* Summary: Set the node to which this C_HOLDER is supposed to arrive when sent
*          as a message
* Description: 
* Return: None
******************************************************************************/
  void set_node(int n) {node = n;}

  void set_size(int s) {size = s;}

  void set_buff(char *b) { buff = b; }
    
/******************************************************************************
* Method: set
* Summary: Set all the data in this C_HOLDER
* Description: 
* Return: None
******************************************************************************/
  void set(char *b, // IN: character buffer of which this object is a holder
	   int s,   // IN: size of character buffer, in bytes
	   int n    // IN: node to which this C_HOLDER is supposed to arrive
	            // when sent as a message
	   ){

    buff = b;
    size = s;
    node = n;
  }


/******************************************************************************
* Method: get
* Summary: Get all the data in this C_HOLDER
* Description: 
* Return: None
******************************************************************************/
  void get(char *&b, // OUT: character buffer of which this object is a holder
	   int &s,   // OUT: size of character buffer, in bytes
	   int &n    // OUT: node to which this C_HOLDER is supposed to arrive
	             // when sent as a message
	   );

/******************************************************************************
* Method: get_node
* Summary: Get the node to which this C_HOLDER is supposed to arrive when sent
*          as a message
* Description: 
* Return: The node to which this C_HOLDER is supposed to arrive when sent as a
*         message
******************************************************************************/
  int get_node() {return node;}

/******************************************************************************
* Constructor Method: 
* Summary: Initialize the data in the object
* Description: 
******************************************************************************/
  C_HOLDER(char *b, // IN: character buffer of which this object is a holder
	   int s,   // IN: size of character buffer, in bytes
	   int n    // IN: node to which this C_HOLDER is supposed to arrive
	            // when sent as a message
	   );

  C_HOLDER() {}

 private:

  char *buff;			// pointer to a buffer
  int size;			// size of the buffer
  int node;			// node that the buffer should be sent to

};

#endif
