/************************************************************************
*   HOST_ROUTER: routes messages between SPEEDES and external modules	*
************************************************************************/

#ifdef SGI
#include <unistd.h>
#endif

#include "bfr.H"
#include "queue.H"
#include "hash.H"
#include "SpNodeId.H"
#include "SpSocketId.H"
#include "SpMsg.H"
#include "SpSocketUtil.H"

#include "SpStdio.H"
#include "SpStdlib.H"
#include <sys/types.h>
#include "SpErrno.H"
#include "SpString.H"
#include "SpSignal.H" 
#include "SpMemory.H"

#ifndef __IBMCPP__
#include <sys/time.h>
#include <sys/file.h>
#include <sys/socket.h> 
#include <netdb.h>
#include <netinet/in.h>
#endif

#define TRUE 1
#define PARSER_FILE
#define DEBUG

#ifdef PARSER_FILE
#include "SpDataParser.H"
#endif

#include "SpOS.H"

void init();
void node_objects();
void service_messages();
void open_connect();
void close_connect(SpSocketId *sockid);
int read_data(SpSocketId *sockid, int socket, char *buffer, int length);
int write_data(SpSocketId *sockid, int socket, char *buffer, int length);
int broken_pipe;




#if defined(SUN3) || defined(SunOS)
void sigpipefunc(int)
{
    fprintf( stderr, "Broken Pipe\n" );
    signal( SIGPIPE, sigpipefunc ); 
    broken_pipe = 1;
} 

void sigintfunc(int)
{
    fprintf( stderr, "SIGINT received\n" );
    signal( SIGINT, sigintfunc ); 
} 

void sigurgfunc(int)
{
    fprintf( stderr, "SIGURG received\n" );
    signal( SIGURG, sigurgfunc ); 
} 
#endif





#ifdef SGI
void sigpipefunc(int)
{
    fprintf( stderr, "Broken Pipe\n" );
    signal( SIGPIPE, sigpipefunc ); 
    broken_pipe = 1;
} 

void sigintfunc(int)
{
    fprintf( stderr, "SIGINT received\n" );
    signal( SIGINT, sigintfunc ); 
} 

void sigurgfunc(int)
{
    fprintf( stderr, "SIGURG received\n" );
    signal( SIGURG, sigurgfunc ); 
} 
#endif




#define Crosfile "Cros.resrc"
#define hostfile "SPEEDES.resrc"


int Cros_Server_Port;

#ifdef PARSER_FILE
char *SPEEDES_Server_Name;
char *Cros_Server_Name;
#else
char SPEEDES_Server_Name[40];
char Cros_Server_Name[40];
#endif

int SPEEDES_Server_Port;
int n_hash;
C_HASH *hash_objects;
int Hock;
struct sockaddr_in Host_Server;
struct timeval timeout;
fd_set p_readfds;
fd_set p_writefds;
C_QUEUE qEM;

int n_nodes;
int *Nodock;
int Nodsock;
int i,j,k,len;

#include "SpCrashOnDivZero.H"

/************************************************************************
* main routine								*
************************************************************************/
int main() {
  crash_on_divide_by_zero();

  init();
  node_objects();
  service_messages();
  printf("Exiting host_router\n");

  return 1;

}

/************************************************************************
* initialize								*
************************************************************************/
void init() {
  int length;

//...... get Cros parameters

#ifdef PARSER_FILE
  SpDataParser parser("speedes.par");

  parser.GoTo("parameters", NULL);
  n_nodes = parser.GetInt("n_nodes");

  parser.GoTo("Host_Router", NULL);
  SPEEDES_Server_Name = parser.GetString("Host_Router_Name");
  SPEEDES_Server_Port = parser.GetInt("Host_Router_Port");
  n_hash = parser.GetInt("Host_Router_Hash");

#else
  FILE *f;
  f = fopen(Crosfile,"r");
  if (f == NULL) {
    printf("cannot find %s\n",Crosfile);
    exit(1);
  }else{
    fscanf(f,"%s\n",Cros_Server_Name);
    fscanf(f,"%d\n",&Cros_Server_Port);
    fscanf(f,"%d\n",&n_nodes);
    fclose(f);
  }

  f = fopen(hostfile,"r");
  if (f == NULL) {
    printf("cannot find %s\n",hostfile);
    exit(1);
  }else{
    fscanf(f,"%s\n",SPEEDES_Server_Name);
    fscanf(f,"%d\n",&SPEEDES_Server_Port);
    fscanf(f,"%d\n",&n_hash);
    fclose(f);
  }
#endif

  hash_objects = new C_HASH(n_hash);

  timeout.tv_sec = 100;
  timeout.tv_usec = 0;

//...... Setup signal catching routines.

#if defined(SUN3) || defined(SunOS)
  signal( SIGPIPE, sigpipefunc ); 
//  signal( SIGINT, sigintfunc ); 
  signal( SIGURG, sigurgfunc ); 
#endif

#ifdef SGI
  signal( SIGPIPE, sigpipefunc ); 
//  signal( SIGINT, sigintfunc ); 
  signal( SIGURG, sigurgfunc ); 
#endif



  broken_pipe = 0;

//...... Clean out FD masks.

  memset( &p_readfds,  0, sizeof(p_readfds) );
  memset( &p_writefds, 0, sizeof(p_writefds) );

//...... Open socket stream to receive connection requests

  Hock = socket( AF_INET, SOCK_STREAM, 0 );
  if (Hock < 0 ) {
    perror( "opening socket stream" );
    exit(1);
  }
    
  Host_Server.sin_family = AF_INET;
  Host_Server.sin_addr.s_addr = INADDR_ANY;
  Host_Server.sin_port = htons(SPEEDES_Server_Port);

  if (bind(Hock, (struct sockaddr *)&Host_Server, sizeof Host_Server) < 0) {
    perror("binding stream socket");
    exit(1);
  }

  listen(Hock,5);

//...... Put main sock in list of file descriptors for select

  FD_SET( Hock, &p_readfds );

  length = sizeof Host_Server;
  if (GETSOCKNAME(Hock, (struct sockaddr *)&Host_Server, &length) < 0) {
    perror("getting socket name");
    exit(1);
  }

  set_nodelay(Hock);

  printf("Host Server Socket port #%d created\n",ntohs(Host_Server.sin_port));

  for (i=0; i<n_nodes; i++) {
    open_connect();
  }

}


/************************************************************************
* node_objects - read in the objects from each node                     *
************************************************************************/
void node_objects() {
  int bytes;
  char *buff = NULL;
  int *uid;
  //int status;
  int n;
  SpNodeId *node_id;
  SpSocketId *sockid;
  int socket;
  int size1;

  sockid = (SpSocketId *)qEM.ptop();
  size1 = 3*sizeof(int) + 40;

  for (i=0; i<n_nodes; i++) {

    socket = sockid->get_id();
    /*status = */socket_read(socket,(char *)&bytes,sizeof(int));
    n = bytes / size1;

    if (bytes != 0) {
      buff = new char[bytes];
      /*status = */socket_read(socket,buff,bytes);
    }

    uid = (int *)buff;
    node_id = new SpNodeId[n];

    for (j=0; j<n; j++) {

      k = j*size1;
      uid = (int *)&buff[k];

      node_id[j].set_local_id(uid[0]);
      node_id[j].set_type(uid[1]);
      node_id[j].set_id(uid[2]);
      node_id[j].set_name((char *)&uid[3]);

      node_id[j].set_socket(socket);
      hash_objects->add(&node_id[j]);

    }

    delete buff;
    printf("%d node_objects on socket %d\n",n,socket);

    sockid = (SpSocketId *)sockid->get_link();

  }

  hash_objects->combine();

}


/************************************************************************
* service_messages : main loop that services messages                   *
************************************************************************/
void service_messages() {
  int i;
  int len;
  int j,jlen;
  int sock;
  int status;
  int nbytes;
  int totbytes;
  fd_set readfds;
  fd_set writefds;
  char *buff;
  char *fullbuff;
  int object_id;
  int outsock;
  C_EM_COMMAND *EM_command;
  C_EM_HEADER EM_header;
  C_EM_HEADER *head;
  SpSocketId *out_sockid;
  SpSocketId *sockid;
  SpSocketId *nextsockid;
  SpNodeId *node_id;
  C_BFR *bfr;
  C_QUEUE *qbuf;
  C_QUEUE *qall;

  while(1) {

//...... Setup select flags for all read sockets and any blocked write sockets.

    memcpy( (char *)&readfds, (char *)&p_readfds, sizeof( p_readfds ) );
    memcpy( (char *)&writefds, (char *)&p_writefds, sizeof( p_writefds ) );

//...... check if there is incomming socket data

    status = SELECT(FD_SETSIZE, &readfds, &writefds, NULL, &timeout);

    if ( status< 0) {
      perror( "select" );
      exit(1);
    }

//...... If there is data on the main socket we have a new connection.

    while (FD_ISSET(Hock, &readfds)) { 
      open_connect();
      memcpy( (char *)&readfds, (char *)&p_readfds, sizeof( p_readfds ) );
      memcpy( (char *)&writefds, (char *)&p_writefds, sizeof( p_writefds ) );
      status = SELECT(FD_SETSIZE, &readfds, &writefds, NULL, &timeout);
    }

//...... loop through existing sockets and read data (or break connection)

    len = qEM.length();
    sockid = (SpSocketId *)qEM.ptop();

    len = qEM.length();
    for (i=0; i<len; i++) {

      nextsockid = (SpSocketId *)sockid->get_link();
      sock = sockid->get_id();

      if (FD_ISSET(sock, &readfds)) {

        status = read_data(sockid, sock, (char *)&EM_header, sizeof(EM_header));

        if (status > 0) {

	  sockid->set_EMi(EM_header.EM_interaction);
	  sockid->set_objid(EM_header.objid);

          nbytes = EM_header.data_bytes;
	  totbytes = sizeof(C_EM_HEADER) + nbytes;
	  fullbuff = new char[totbytes];
	  memcpy(fullbuff, (char *)&EM_header, sizeof(C_EM_HEADER));
	  head = (C_EM_HEADER *)fullbuff;

	  if (nbytes != 0) {
	    buff = &fullbuff[sizeof(C_EM_HEADER)];
	    status = read_data(sockid, sock, buff, nbytes);
	  }

	  if (status > 0) {

//...... query for object id given a name

	    if (head->EM_interaction & EM_GETID) {

	      EM_command = (C_EM_COMMAND *)head;
	      EM_command->objid = -1;


	      if (strcmp(EM_command->EM_command, "FLUSH_SOCKET")) {

	        qall = hash_objects->get_all();
	        node_id = (SpNodeId *)qall->ptop();
	        jlen = qall->length();

	        for (j=0; j<jlen; j++) {
		  if (node_id->check_name(EM_command->EM_command)) {
		    EM_command->objid = node_id->get_id();
		    break;
		  }else{
		    node_id = (SpNodeId *)node_id->get_link();
	          }
	        }

	      }

	      set_blocking(sock);
	      status = write(sock, (char *)EM_command, sizeof(C_EM_COMMAND));
	      delete fullbuff;

	    }else{

//...... normal message

	      object_id = head->objid;
	      if (object_id >= 0) {
	        node_id = (SpNodeId *)hash_objects->get(object_id);
	        if (node_id != NULL) {
		  head->objid = node_id->get_local_id();
		  head->obtype = node_id->get_type();
	          head->EM_socket = sock;
	          outsock = node_id->get_socket();
	        }else{
	          printf("Error, bad object id %d\n",object_id);
	          outsock = -1;
	        }
	      }else{
	        outsock = head->EM_socket;
	      }

	      out_sockid = (SpSocketId *)qEM.get_item(outsock);

	      if (out_sockid != NULL) {
	        qbuf = out_sockid->get_qbuf();
	        bfr = new C_BFR(fullbuff,totbytes);
	        qbuf->push_bot(bfr);
	        FD_SET(outsock, &p_writefds);
	      }else{
	        delete fullbuff;
	      }
	    }

	  }else{

	    delete fullbuff;

	  }

	}

      }

      sockid = nextsockid;

    }

//...... output stuff now

    sockid = (SpSocketId *)qEM.ptop();
    len = qEM.length();
    for (i=0; i<len; i++) {

      nextsockid = (SpSocketId *)sockid->get_link();
      sock = sockid->get_id();

      if (FD_ISSET(sock, &writefds)) {

	qbuf = sockid->get_qbuf();
	while (1) {

	  if (qbuf->length() == 0) {

	    printf("Error, bfr does not exist\n");

	  }else{

	    bfr = (C_BFR *)qbuf->pop_top();
	    buff = bfr->get_start_buff();
	    nbytes = bfr->get_nbytes_left();

	    status = write_data(sockid, sock, buff, nbytes);

	    if (status != nbytes) {

	      if (status >= 0) {
	        bfr->update_start(status);
	        qbuf->push_top(bfr);
	      }
	      break;

	    }else{

	      buff = bfr->get_buffer();
	      nbytes = bfr->get_nbytes();

	      delete buff;
	      delete bfr;

	      if (qbuf->length() == 0) {
		FD_CLR(sock,&p_writefds);
		break;
	      }

	    }


	  }

	}

      }

      sockid = nextsockid;

    }

  }

}

/************************************************************************
* open_connect : open a connection to the host router                   *
************************************************************************/
void open_connect() {	
  int new_sock;
  SpSocketId *sockid;

//...... Accept new connection

  new_sock = ACCEPT(Hock, (struct sockaddr *)0, (int *)0);
  if (new_sock < 0) {
    perror( "accept" );
    exit(1);
  }

  set_nodelay(new_sock);
  fprintf(stderr, "Connecting socket number %d\n", new_sock);

  sockid = new SpSocketId();
  sockid->set_id(new_sock);
  qEM.push_bot(sockid);

//...... Add new socket number to active file descrptors 

  FD_SET(new_sock, &p_readfds );

}

/************************************************************************
* close_connect : close a connection to the host router                 *
************************************************************************/
void close_connect(SpSocketId *sockid) {
  int socket;
  int objid;
  int outsock;
  int len;
  int i;
  int interaction;
  //int nbytes;
  char *buff;
  C_EM_HEADER disc;
  C_QUEUE *qbuf;
  C_BFR *bfr;
  SpSocketId *out_sockid;
  SpNodeId *node_id;

//...... delete the outgoing buffered messages

  qbuf = sockid->get_qbuf();
  len = qbuf->length();
  for (i=0; i<len; i++) {
    bfr = (C_BFR *)qbuf->pop_top();
    buff = bfr->get_buffer();
    //nbytes = bfr->get_nbytes();
    delete buff;
    delete bfr;
  }

  socket = sockid->get_id();

  if (sockid->get_EMi() & EM_CONNECT) {

    interaction = sockid->get_EMi() & (EM_MODULE | EM_MONITOR);

    objid = sockid->get_objid();
    if (objid != -1) {
      node_id = (SpNodeId *)hash_objects->get(objid);
      if (node_id != NULL) {
        outsock = node_id->get_socket();
        out_sockid = (SpSocketId *)qEM.get_item(outsock);

        disc.objid = node_id->get_local_id();
        disc.obtype = node_id->get_type();
        disc.EM_interaction = interaction | EM_DISCONNECT | EM_AGGRESSIVE;
        disc.EM_socket = socket;
        disc.data_bytes = 0;

        write_data(out_sockid, outsock, (char *)&disc, sizeof(C_EM_HEADER));
      }
    }

  }

  qEM.remove(socket);

  close(socket);
  FD_CLR( socket, &p_readfds );
  FD_CLR( socket, &p_writefds );

  fprintf(stderr, "Closing socket number %d\n", socket);

}

/************************************************************************
* read_data : read data from the socket					*
************************************************************************/
int read_data(SpSocketId *sockid, int socket, char *buffer, int length) {
  int start;
  int status;

  set_blocking(socket);
  start = 0;

  while (start != length) {
    status = read(socket, &buffer[start], length-start);

    if (status <= 0) {

      if (errno == EWOULDBLOCK) {
	status = 0;

	close_connect(sockid);
        return -1;

      }else{
	close_connect(sockid);
        return -1;
      }

    }

    start += status;

  }

  return 1;

}			

/************************************************************************
* write_data : write data from the socket				*
************************************************************************/
int write_data(SpSocketId *, int socket, char *buffer, int length) {
  int start;
  int status;

  set_nonblocking(socket);
  start = 0;

  while (start != length) {
    status = write(socket, &buffer[start], length-start);

    if (status <= 0) {
      break;
    }

    start += status;

  }

  set_blocking(socket);
  return start;


}			

