// Host_User.C method file

#include "host_user.H"

#ifdef HOST

#include "SpStdio.H"
#include "SpStdlib.H"
#include "SpMemory.H"

#include "query_data.H"
#include "holder.H"
#include "defunc.H"

#ifdef PARSER_FILE
#include "SpDataParser.H"
#endif

#include "SpSocketUtil.H"

/************************************************************************
* C_HOST_USER : get the Host_User parameters				*
************************************************************************/
C_HOST_USER::C_HOST_USER() {

  node = -1;
  connected = 0;
  looping = 0;
  input_file_flag = 0;
  output_file_flag = 0;

#ifdef PARSER_FILE
  SpDataParser parser("speedes.par");

  parser.GoTo("Host_Router", NULL);
  Host_Server_Name = parser.GetString("Host_Router_Name");
  Host_Server_Port = parser.GetInt("Host_Router_Port");
#else
  FILE *f;
  f = fopen(hostfile,"r");
  if (f == NULL) {
    printf("cannot find %s\n",hostfile);
    return;
  }else{
    fscanf(f,"%s\n",Host_Server_Name);
    fscanf(f,"%d\n",&Host_Server_Port);
    fclose(f);
  }
#endif

//...... Create my node socket named "Nodock"

  Nodock = socket(AF_INET, SOCK_STREAM,0);
  if (Nodock < 0) {
    perror("opening stream socket");
    return;
  }

//...... Connect Host_User socket (Nodock) to Host_Server socket (Crock)

  Host_Server.sin_family = AF_INET;
  hostent *hp = get_host_by_name(Host_Server_Name);
  if (hp == 0) {
    fprintf(stderr,"%s : Unknown Host\n",Host_Server_Name);
    return;
  }

//  bcopy((char *)hp->h_addr, (char *)&Host_Server.sin_addr, hp->h_length);
  memcpy((char *)&Host_Server.sin_addr, (char *)hp->h_addr, hp->h_length);
  Host_Server.sin_port = htons(Host_Server_Port);

  if (connect(Nodock,(struct sockaddr *)&Host_Server, sizeof Host_Server) < 0) {
    perror("connecting stream socket");
    return;
  }

  connected = 1;
  set_nodelay(Nodock);

  printf("HOST_USER object created\n");

}

/************************************************************************
* C_HOST_USER : get the Host_User parameters				*
************************************************************************/
C_HOST_USER::C_HOST_USER(char *fname) {

  node = -1;
  connected = 1;
  looping = 0;
  input_file_flag = 1;
  output_file_flag = 0;

  input_file = strdup(fname);
  fp_input = fopen(fname,"rb");

  printf("HOST_USER object reading from file %s\n",input_file);

}

/************************************************************************
* set_output_file : record all messages to a file			*
************************************************************************/
void C_HOST_USER::set_output_file(char *fname) {

  output_file = strdup(fname);
  fp_output = fopen(fname,"wb");
  output_file_flag = 1;

  printf("HOST_USER object recording all messages to file %s\n",output_file);

}


/************************************************************************
* get_message - gets a message from the host				*
************************************************************************/
C_EM_HEADER *C_HOST_USER::get_message(int message_type) {
  int i;
  char *buff;
  int bytes;
  int status;
  C_EM_HEADER EM_header;
  C_EM_HEADER *p_EM_header;
  int bflag;
  C_HOLDER *holder;

//...... check if message waiting in the message queue

  for (i=0; i<message_queue.get_length(); i++) {
    holder = (C_HOLDER *)message_queue.pop_top();
    p_EM_header = (C_EM_HEADER *)holder->get_buff();
    if (p_EM_header->EM_interaction & message_type) {
      delete holder;
      return p_EM_header;
    }else{
      message_queue.push_bot(holder);
    }
  }

  if (input_file_flag) {

//...... read from an input file

    status = fread((char *)&EM_header,sizeof(C_EM_HEADER),1,fp_input);
    if (status == 0) return NULL;

    bytes = EM_header.data_bytes;
    buff = new char[sizeof(C_EM_HEADER) + bytes];

    if (buff == NULL) {
      fprintf(stderr,"Error (host_user) could not create %d bytes\n",
		sizeof(SpMsg) + bytes);
      abort();
    }

    p_EM_header = (C_EM_HEADER *)buff;
    memcpy(buff, (char *)&EM_header, sizeof(C_EM_HEADER));

    if (bytes != 0) {
      status = fread(&buff[sizeof(C_EM_HEADER)], bytes, 1, fp_input);
    }

  }else{

//...... read from a socket

    if (!connected) return NULL;
    bflag = 0;

    set_nonblocking(Nodock);
    status = nb_socket_read(Nodock,(char *)&EM_header,sizeof(C_EM_HEADER));

    if ((status > 0 ) && (status != sizeof(C_EM_HEADER))) {
      set_blocking(Nodock);
      buff = (char *)&EM_header;
      status = socket_read(Nodock, &buff[status], sizeof(C_EM_HEADER) - status);
      bflag = 1;
    }

    if (status != 0) {

      if (bflag == 0) set_blocking(Nodock);
      bytes = EM_header.data_bytes;
      buff = new char[sizeof(C_EM_HEADER) + bytes];

      if (buff == NULL) {
        fprintf(stderr,"Error (host_user) could not create %d bytes\n",
		sizeof(SpMsg) + bytes);
	abort();
      }

      p_EM_header = (C_EM_HEADER *)buff;
      memcpy(buff, (char *)&EM_header, sizeof(C_EM_HEADER));

      if (bytes != 0) {
        status = socket_read(Nodock, &buff[sizeof(C_EM_HEADER)], bytes);
      }

    }else{

      return NULL;

    }

  }

//...... if the output file flag is set, write out the message to the file

  if (output_file_flag) {
    status = fwrite(buff,sizeof(C_EM_HEADER)+EM_header.data_bytes,1,fp_output);
    fflush(fp_output);
    if (status < 0) {
      fprintf(stderr,"Error (host_user) could not write to disk\n");
    }
  }

//...... make sure that the message is of the right type

  if (p_EM_header->EM_interaction & message_type) {
    return p_EM_header;
  }else{
    holder = new C_HOLDER();
    holder->set((char *)p_EM_header, -1, -1);
    message_queue.push_bot(holder);
    p_EM_header = NULL;
  }

  return p_EM_header;

}

/************************************************************************
* send_message - sends a message to the host				*
************************************************************************/
void C_HOST_USER::send_message(C_EM_HEADER *EM_header) {
  char *buff;
  int bytes;
  int status;
  int start;

  if (!connected) return;
  if (input_file_flag) return;

  buff = (char *)EM_header;
  bytes = sizeof(C_EM_HEADER) + EM_header->data_bytes;
  bytes_flush += bytes;

  if (node == -1) {
    EM_header->EM_socket = Nodock;
  }

  start = 0;
  while (start != bytes) {
    status = nb_socket_write(Nodock,&buff[start],bytes-start);
    start += status;
  }

}

/************************************************************************
* send_node_objects - sends node objects to the host			*
************************************************************************/
void C_HOST_USER::send_node_objects(int size, char *node_objects) {
  //int status;

  if (!connected) return;

  /*status = */socket_write(Nodock, (char *)&size, sizeof(int));
  /*status = */socket_write(Nodock, (char *)node_objects, size);

}

/************************************************************************
* getid - get the global id for an object				*
************************************************************************/
int C_HOST_USER::getid(char *name) {
  C_EM_COMMAND EM_command;
  C_EM_COMMAND *EM_command1;
  int id;

  if (!connected) return -1;

  EM_command.time_tag = -1.0e20;
  EM_command.evtype = -1;
  EM_command.obtype = -1;
  EM_command.objid = -1;
  EM_command.ext = 1;
  EM_command.rbid = -1;
  EM_command.EM_socket = -1;
  EM_command.EM_interaction = EM_GETID;
  EM_command.data_bytes = sizeof(C_EM_COMMAND) - sizeof(C_EM_HEADER);

  strcpy(EM_command.EM_command, name);

  send_message(&EM_command);

  while (1) {
    EM_command1 = (C_EM_COMMAND *)get_message(EM_GETID);
    if (EM_command1) break;
  }

  id = EM_command1->objid;
  delete EM_command1;

  return id;

}

/************************************************************************
* flush - flush out all of the messages 				*
************************************************************************/
void C_HOST_USER::flush() {
  //int id;

  if (!connected) return;
  if (bytes_flush == 0) return;

  //id = getid("FLUSH_SOCKET");
  bytes_flush = 0;

}

/************************************************************************
* query - query an object						*
************************************************************************/
C_EM_HEADER *C_HOST_USER::query(int oid) {
  C_EM_HEADER *EM_header;

  if (!connected) return NULL;

  header.time_tag = -1.0e20;
  header.evtype = -1;
  header.obtype = -1;
  header.objid = oid;
  header.ext = 1;
  header.rbid = -1;
  header.EM_socket = -1;
  header.EM_interaction = (EM_QUERY + EM_AGGRESSIVE);
  header.data_bytes = 0;

  send_message(&header);
  while (1) {
    EM_header = get_message();
    if (EM_header != NULL) break;
  }

  print_query(EM_header);
  return EM_header;

}

/************************************************************************
* query - query an object						*
************************************************************************/
C_QUERY_DATA *C_HOST_USER::query(int oid, int &n) {
  C_EM_HEADER *EM_header;
  char *buff;
  C_QUERY_DATA *query_data;
  int size;

  if (!connected) return NULL;

  header.time_tag = -1.0e20;
  header.evtype = -1;
  header.obtype = -1;
  header.objid = oid;
  header.ext = 1;
  header.rbid = -1;
  header.EM_socket = -1;
  header.EM_interaction = (EM_QUERY + EM_AGGRESSIVE);
  header.data_bytes = 0;

  send_message(&header);
  while (1) {
    EM_header = get_message();
    if (EM_header != NULL) break;
  }

  buff = (char *)EM_header;
  buff += sizeof(C_EM_HEADER);

  size = EM_header->data_bytes;
  n = size / sizeof(C_QUERY_DATA);

  query_data = new C_QUERY_DATA[n];;
  memcpy((char *)query_data, buff, size);
  delete EM_header;

  return query_data;

}

/************************************************************************
* monitor - monitor an object						*
************************************************************************/
C_EM_HEADER *C_HOST_USER::monitor(int oid) {
  C_EM_HEADER *EM_header;

  if (!connected) return NULL;

  if (!looping) {
    header.time_tag = -1.0e20;
    header.evtype = -1;
    header.obtype = -1;
    header.objid = oid;
    header.ext = 1;
    header.rbid = -1;
    header.EM_socket = -1;
    header.EM_interaction = (EM_MONITOR + EM_CONNECT + EM_AGGRESSIVE);
    header.data_bytes = 0;

    send_message(&header);
    looping = 1;
    conid = oid;
  }

  EM_header = get_message();

  if (EM_header != NULL) {
    print_query(EM_header);
  }

  return EM_header;

}

/************************************************************************
* monitor - monitor an object						*
************************************************************************/
C_QUERY_DATA *C_HOST_USER::monitor(int oid, int &n) {
  C_EM_HEADER *EM_header;
  C_EM_HEADER *EM_header1;
  char *buff;
  C_QUERY_DATA *query_data;
  int size;

  if (!connected) return NULL;

  if (!looping) {
    header.time_tag = -1.0e20;
    header.evtype = -1;
    header.obtype = -1;
    header.objid = oid;
    header.ext = 1;
    header.rbid = -1;
    header.EM_socket = -1;
    header.EM_interaction = (EM_MONITOR + EM_CONNECT + EM_AGGRESSIVE);
    header.data_bytes = 0;

    send_message(&header);
    looping = 1;
    conid = oid;
  }

  EM_header = get_message();
  while (EM_header) {
    EM_header1 = get_message();
    if (EM_header1) {
      delete EM_header;
      EM_header = EM_header1;
    }else{
      break;
    }
  }

  if (EM_header == NULL) {
    n = 0;
    return NULL;
  }

  buff = (char *)EM_header;
  buff += sizeof(C_EM_HEADER);

  size = EM_header->data_bytes;
  n = size / sizeof(C_QUERY_DATA);

  query_data = new C_QUERY_DATA[n];;
  memcpy((char *)query_data, buff, size);
  delete EM_header;

  return query_data;

}

/************************************************************************
* disconnect - disconnect an object					*
************************************************************************/
void C_HOST_USER::disconnect() {

  if (!connected) return;
  if (looping) {
    header.time_tag = -1.0e20;
    header.evtype = -1;
    header.obtype = -1;
    header.objid = conid;
    header.ext = 1;
    header.rbid = -1;
    header.EM_socket = -1;
    header.EM_interaction = (EM_MONITOR + EM_DISCONNECT + EM_AGGRESSIVE);
    header.data_bytes = 0;

    send_message(&header);
    looping = 0;
  }

}

/************************************************************************
* module - control an external module from a simuilation object		*
************************************************************************/
C_EM_HEADER *C_HOST_USER::module(C_EM_HEADER *EM_header) {
  C_EM_HEADER *EM_header1;

  if (!connected) return NULL;

  if (!looping) {
    looping = 1;
    EM_header->time_tag = -1.0e20;
    EM_header->obtype = -1;
    EM_header->ext = 0;
    EM_header->rbid = -1;
    EM_header->EM_socket = -1;
    EM_header->EM_interaction = ( EM_MODULE + EM_CONNECT + EM_AGGRESSIVE);
  }else{
    EM_header->obtype = -1;
    EM_header->ext = 1;
    EM_header->rbid = -1;
    EM_header->EM_socket = -1;
    EM_header->EM_interaction = (EM_MODULE + EM_CONNECT + EM_BLOCKING);
  }

  send_message(EM_header);

  while (1) {
    EM_header1 = get_message();
    if (EM_header1 != NULL) break;
  }

  return EM_header1;

}

/************************************************************************
* send_module - send a message out to SPEEDES from a module		*
************************************************************************/
void C_HOST_USER::send_module(C_EM_HEADER *EM_header) {

  if (!connected) return;

  if (!looping) {
    looping = 1;
    EM_header->time_tag = -1.0e20;
    EM_header->obtype = -1;
    EM_header->ext = 0;
    EM_header->rbid = -1;
    EM_header->EM_socket = -1;
    EM_header->EM_interaction = ( EM_MODULE + EM_CONNECT + EM_AGGRESSIVE);
  }else{
    EM_header->obtype = -1;
    EM_header->ext = 1;
    EM_header->rbid = -1;
    EM_header->EM_socket = -1;
    EM_header->EM_interaction =
	(EM_MODULE + EM_CONNECT + EM_BLOCKING + EM_AGGRESSIVE);
  }

  send_message(EM_header);

}

/************************************************************************
* blocking_module - synchronize an external module			*
************************************************************************/
C_EM_HEADER *C_HOST_USER::blocking_module(C_EM_HEADER *EM_header) {
  C_EM_HEADER *EM_header1;

  if (!connected) return NULL;

  if (!looping) {
    looping = 1;
    EM_header->obtype = -1;
    EM_header->ext = 0;
    EM_header->rbid = -1;
    EM_header->EM_socket = -1;
    EM_header->EM_interaction = ( EM_MODULE + EM_CONNECT + EM_BLOCKING);
  }else{
    EM_header->obtype = -1;
    EM_header->ext = 1;
    EM_header->rbid = -1;
    EM_header->EM_socket = -1;
    EM_header->EM_interaction = (EM_MODULE + EM_CONNECT + EM_BLOCKING);
  }

  send_message(EM_header);

  while (1) {
    EM_header1 = get_message();
    if (EM_header1 != NULL) break;
  }

  return EM_header1;

}

/************************************************************************
* aggressive_module - aggressively synchronize an external module	*
************************************************************************/
C_EM_HEADER *C_HOST_USER::aggressive_module(C_EM_HEADER *EM_header) {
  C_EM_HEADER *EM_header1;

  if (!connected) return NULL;

  if (!looping) {
    looping = 1;
    EM_header->obtype = -1;
    EM_header->ext = 0;
    EM_header->rbid = -1;
    EM_header->EM_socket = -1;
    EM_header->EM_interaction = 
	(EM_MODULE + EM_CONNECT + EM_BLOCKING + EM_AGGRESSIVE);
  }else{
    EM_header->obtype = -1;
    EM_header->ext = 1;
    EM_header->rbid = -1;
    EM_header->EM_socket = -1;
    EM_header->EM_interaction =
	(EM_MODULE + EM_CONNECT + EM_BLOCKING + EM_AGGRESSIVE);
  }

  send_message(EM_header);

  while (1) {
    EM_header1 = get_message();
    if (EM_header1 != NULL) break;
  }

  return EM_header1;

}

/************************************************************************
* get_module - get data (non-blocking) from SPEEDES			*
************************************************************************/
C_EM_HEADER *C_HOST_USER::get_module() {
  C_EM_HEADER *EM_header1;

  EM_header1 = get_message();
  return EM_header1;

}

/************************************************************************
* command - issue a command for an object				*
************************************************************************/
void C_HOST_USER::command(C_EM_HEADER *EM_header) {

  if (!connected) return;

  EM_header->time_tag = -1.0e20;
  EM_header->evtype = -1;
  EM_header->obtype = -1;
  EM_header->ext = 1;
  EM_header->rbid = -1;
  EM_header->EM_socket = -1;
  EM_header->EM_interaction = (EM_COMMAND + EM_AGGRESSIVE);

  send_message(EM_header);

}

/************************************************************************
* print_query - print out received data					*
************************************************************************/
void C_HOST_USER::print_query(C_EM_HEADER *EM_header) {
  int i,n,size;
  C_QUERY_DATA *query_data;
  char *buff;

  if (EM_header == NULL) return;

  buff = (char *)EM_header;
  buff += sizeof(C_EM_HEADER);
  query_data = (C_QUERY_DATA *)buff;

  size = EM_header->data_bytes;
  n = size / sizeof(C_QUERY_DATA);

  fprintf(stderr,"\n");
  for (i=0; i<n; i++) {
    query_data[i].print();
  }
  fprintf(stderr,"\n");

}


#endif
