// queue.C method file

#include "SpStdio.H"
#include "queue.H"

int C_QUEUE::cutoff = 15;

/************************************************************************
* C_QUEUE : construct q queue object					*
************************************************************************/
C_QUEUE::C_QUEUE() {

//...... initially top and bot point to null, and there are no items

  top = NULL;
  bot = NULL;
  n_items = 0;
}

/************************************************************************
* init : initialize the queue object					*
************************************************************************/
void C_QUEUE::init(C_ITEM *t, C_ITEM *b, int len) {

  top = t;
  bot = b;
  n_items = len;

}

/************************************************************************
* remove : remove (delete) an item from the queue			*
************************************************************************/
void C_QUEUE::remove(C_ITEM *item) {
  C_ITEM *temp1, *temp2, *temp3;
  int i;

  if (n_items == 0) return;

  if (n_items == 1) {
    if (item == top) {
      delete item;
      clear();
    }
    return;
  }

  if (item == top) {
    top = top->get_link();
    n_items--;
    delete item;
    return;
  }

  temp1 = top;
  temp2 = temp1->get_link();
  for (i=1; i<n_items; i++) {
    if (temp2 == item) {
      temp3 = temp2->get_link();
      temp1->put_link(temp3);
      if (bot == item) bot = temp1;
      n_items--;
      delete item;
      return;
    }else{
      temp1 = temp2;
      temp2 = temp2->get_link();
    }
  }

}

/************************************************************************
* remove : remove (delete) an item from the queue			*
************************************************************************/
void C_QUEUE::remove(int id) {
  C_ITEM *item1, *item2;
  int i;

//...... special case when there are no items

  if (n_items == 0) return;

//...... special case when there is only one item

  if (n_items == 1) {
    if (id == top->get_id()) {
      delete top;
      clear();
    }
    return;
  }

//...... now there are more than 1 items so check if it is the first one

  if (id == top->get_id()) {
    item1 = top->get_link();
    n_items--;
    delete top;
    top = item1;
    return;
  }

//...... loop through the rest of the items to find the right id.

  item1 = top;
  item2 = top->get_link();
  for (i=1; i<n_items; i++) {
    if (id == item2->get_id()) {
      item1->put_link(item2->get_link());
      if (item2 == bot) bot = item1;
      delete item2;
      n_items--;
      break;
    }
    item1 = item2;
    item2 = item2->get_link();
  }

}

/************************************************************************
* deque : remove (but don't delete) an item from the queue		*
************************************************************************/
void C_QUEUE::deque(C_ITEM *item) {
  C_ITEM *temp1, *temp2, *temp3;
  int i;

  if (n_items == 0) return;

  if (n_items == 1) {
    if (top == item) {
      reset();
    }
    return;
  }

  if (item == top) {
    top = top->get_link();
    n_items--;
    return;
  }

  temp1 = top;
  temp2 = temp1->get_link();
  for (i=1; i<n_items; i++) {
    if (temp2 == item) {
      temp3 = temp2->get_link();
      temp1->put_link(temp3);
      if (bot == item) bot = temp1;
      n_items--;
      return;
    }else{
      temp1 = temp2;
      temp2 = temp2->get_link();
    }
  }

}

/************************************************************************
* deque : remove (but don't delete) an item from the queue		*
************************************************************************/
C_ITEM *C_QUEUE::deque(int rid) {
  C_ITEM *temp1, *temp2, *temp3, *item;
  int i;

  item = NULL;

  if (n_items == 0) {
//    fprintf(stderr,"deque error, no items\n");
    return item;
  }

  if (n_items == 1) {
    if (top->get_id() == rid) {
      item = top;
      reset();
    }else{
//      fprintf(stderr,"deque error, 1 item, wrong id\n");
    }
    return item;
  }

  if (top->get_id() == rid) {
    item = top;
    top = top->get_link();
    n_items--;
    return item;
  }

  temp1 = top;
  temp2 = temp1->get_link();
  for (i=1; i<n_items; i++) {
    if (temp2->get_id() == rid) {
      temp3 = temp2->get_link();
      temp1->put_link(temp3);
      if (bot->get_id() == rid) bot = temp1;
      n_items--;
      return temp2;
    }else{
      temp1 = temp2;
      temp2 = temp2->get_link();
    }
  }

//  fprintf(stderr,"deque error, id not found\n");
  return item;

}

/************************************************************************
* get_item : get the pointer to the item with the right id		*
************************************************************************/
C_ITEM *C_QUEUE::get_item(int id) {
  int i;
  C_ITEM *item;

  item = top;
  for (i=0; i<n_items; i++) {
    if (item->get_id() == id) return item;
    item = item->get_link();
  }

  return NULL;

}

/************************************************************************
* concat : concat an input queue to this queue (reset the input queue)	*
************************************************************************/
void C_QUEUE::concat(C_QUEUE *q) {
  join(q);
  q->reset();
}

/************************************************************************
* join : join the input queue to this queue but don't reset		*
************************************************************************/
void C_QUEUE::join(C_QUEUE *q) {
  C_ITEM *temp;
  if (q->length() == 0) return;
  if (n_items == 0) {
    top = q->ptop();
    bot = q->pbot();
    n_items = q->length();
  }else{
    temp = q->ptop();
    bot->put_link(temp);
    bot = q->pbot();
    n_items += q->length();
  }
}

/************************************************************************
* insert : insert the item into the queue (low on top)			*
************************************************************************/
void C_QUEUE::insert(C_ITEM *item) {
  C_ITEM *temp1;
  C_ITEM *temp2;
  SpSimTime t; // GEB2
  int i;

  if (n_items == 0) {
    top = item;
    bot = item;
    n_items = 1;
    return;
  }

  t = item->get_time_tag();

  if (t <= top->get_time_tag()) {
    push_top(item);
    return;
  }

  if (t > bot->get_time_tag()) {
    push_bot(item);
    return;
  }

  temp1 = top;
  for (i=1; i<n_items; i++) {
    temp2 = temp1->get_link();
    if (t <= temp2->get_time_tag()) {
      temp1->put_link(item);
      item->put_link(temp2);
      n_items++;
      return;
    }
    temp1 = temp2;
  }

  fprintf(stderr,"ERROR IN QUEUE : INSERT\n");
  item->print();
  print();
}

/************************************************************************
* insert_after : insert item into the queue (low on top)			*
************************************************************************/
void C_QUEUE::insert_after(C_ITEM *item) {
  C_ITEM *temp1;
  C_ITEM *temp2;
  SpSimTime t;    //GEB2
  int i;

  if (n_items == 0) {
    top = item;
    bot = item;
    n_items = 1;
    return;
  }

  t = item->get_time_tag();

  if (t < top->get_time_tag()) {
    push_top(item);
    return;
  }

  if (t >= bot->get_time_tag()) {
    push_bot(item);
    return;
  }

  temp1 = top;
  for (i=1; i<n_items; i++) {
    temp2 = temp1->get_link();
    if (t < temp2->get_time_tag()) {
      temp1->put_link(item);
      item->put_link(temp2);
      n_items++;
      return;
    }
    temp1 = temp2;
  }

  fprintf(stderr,"ERROR IN QUEUE : INSERT\n");
  item->print();
  print();
}

/************************************************************************
* merge : merge the input queue into this queue in time order		*
************************************************************************/
void C_QUEUE::merge(C_QUEUE *q) {
  C_ITEM *qtop;
  C_ITEM *thistop;
  C_ITEM *mitem;
  SpSimTime qval,thisval; //GEB2
  int qlen,thislen;

//...... case where input q is empty

  if (q->length() == 0) return;

//...... case where this q is empty

  if (n_items == 0) {
    top = q->ptop();
    bot = q->pbot();
    n_items = q->length();
    q->reset();
    return;
  }

//...... get the top of each queue

  thistop = top;
  thislen = n_items;
  //  thisval = thistop->get_valu();  GEB2: change to get_time_tag:
  thisval = thistop->get_time_tag();

  qtop = q->ptop();
  qlen = q->length();
  //  qval = qtop->get_valu();  GEB2: change to get_time_tag:
  qval = qtop->get_time_tag();

  n_items = thislen + qlen;

//...... set the merged queue to the smaller of the two values

  if (qval < thisval) {

    top = qtop;
    qtop = qtop->get_link();
    qlen--;

    if (qlen == 0) {
      top->put_link(thistop);
      q->reset();
      return;
    }

    qval = qtop->get_time_tag(); //GEB2: was get_valu

  }else{

    top = thistop;
    thistop = thistop->get_link();
    thislen--;

    if (thislen == 0) {
      top->put_link(qtop);
      bot = q->pbot();
      q->reset();
      return;
    }

    thisval = thistop->get_time_tag(); //GEB2: was get_valu

  }

  mitem = top;

  while (1) {

    if (qval < thisval) {

      mitem->put_link(qtop);
      mitem = qtop;
      qtop = qtop->get_link();
      qlen--;

      if (qlen == 0) {
        mitem->put_link(thistop);
        q->reset();
        return;
      }

      qval = qtop->get_time_tag(); //GEB2: was get_valu

    }else{

      mitem->put_link(thistop);
      mitem = thistop;
      thistop = thistop->get_link();
      thislen--;

      if (thislen == 0) {
        mitem->put_link(qtop);
        bot = q->pbot();
        q->reset();
        return;
      }

      thisval = thistop->get_time_tag(); //GEB2: was get_valu

    }

  }


}

/************************************************************************
* copy : copy this queue						*
************************************************************************/
C_QUEUE *C_QUEUE::copy() {
  C_QUEUE *q;
  C_ITEM *temp1;
  C_ITEM *temp2;

  q = new C_QUEUE();
  temp1 = top;
  while(n_items > 0) {
    temp2 = temp1->copy();
    q->push_bot(temp2);
    temp1 = temp1->get_link();
  }
  return q;
}

/************************************************************************
* print : print the items in the queue					*
************************************************************************/
void C_QUEUE::print() {
  C_ITEM *temp;
  int i;
  if (n_items == 0) return;
//  fprintf(stderr,"printing queue...\n");
  temp = top;
  for (i=0; i<n_items; i++) {
    temp->print();
    temp = temp->get_link();
  }

}

/************************************************************************
* printstream : print the items in the queue using stream io		*
************************************************************************/
void C_QUEUE::printstream(ostream &out) {
  C_ITEM *temp;
  int i;
  if (n_items == 0) return;
  temp = top;
  for (i=0; i<n_items; i++) {
    temp->printstream(out);
    temp = temp->get_link();
  }

}

/************************************************************************
* reset : reset the queue without deleting the items			*
************************************************************************/
void C_QUEUE::reset() {
  top = NULL;
  bot = NULL;
  n_items = 0;
}

/************************************************************************
* clear : clear the queue by deleting all the items			*
************************************************************************/
void C_QUEUE::clear() {
  C_ITEM *temp;
  while(n_items > 0) {
    temp = pop_top();
    delete temp;
  }
}

/************************************************************************
* sort : sort the queue using merge and recursion			*
************************************************************************/
void C_QUEUE::sort() {
  SpSimTime val;    //GEB2: was double
  C_QUEUE q ;
  C_ITEM *item;
  C_ITEM *next_item;
  C_ITEM *item1;
  C_ITEM *item2;
  C_ITEM *top2;
  C_ITEM *bot2;
  int i, len1, len2;

//...... if n_items is less than a certain cutoff, insertion sort is faster

//  if (n_items <= cutoff) {
  if (n_items <= 15) {

//...... if there is 0 or 1 item in the queue, it is already sorted

    if (n_items <= 1) return;

//...... take the first item and then loop through the n-1 other items

    top2 = top; 
    bot2 = top;
    next_item = top->get_link();
    for (i= 1; i<n_items; i++) {
      item = next_item;
      next_item = item->get_link();
      val = item->get_time_tag();  //GEB2: was get_valu

//...... if val >= top value then put on the top

      if (val <= top2->get_time_tag()) { //GEB2: was get_valu
        item->put_link(top2);
	top2 = item;
      }else{

//...... if val <= bot value then put on the bottom

	if (val >= bot2->get_time_tag()) { //GEB: was get_valu
	  bot2->put_link(item);
	  bot2 = item;
	}else{

//...... otherwise, loop through the already sorted items and insert

	  item1 = top2;
          while (1) {
	    item2 = item1->get_link();
	    if (val <= item2->get_time_tag()) { //GEB: was get_valu
	      item->put_link(item2);
	      item1->put_link(item);
	      break;
	    }
	    item1 = item2;
	  }
	}
      }
    }

//...... reset the object top and bot pointers to the sorted pointers

    top = top2;
    bot = bot2;
    return;
  }

//...... enough items to warrent merge sort so divide the queue into two halves

  len1 = n_items / 2;
  len2 = n_items - len1;
  bot2 = bot;

  item = top;
  for (i=1; i<len1; i++) item = item->get_link();
  bot = item;
  n_items = len1;

  top2 = item->get_link();
  q.init(top2,bot2,len2);

//...... recursively sort me (the upper half)

  sort();

//...... recursevely sort the bottom half

  q.sort();

//...... merge the two halves

  merge(&q);

}

/************************************************************************
* push_top : pushes an item onto the top of the queue           *
************************************************************************/

void C_QUEUE::push_top(C_ITEM *item) {
  C_ITEM *temp;
  if (n_items == 0) {
    bot = item;
    top = item;
    n_items = 1;
  }else{
    temp = top;
    top = item;
    top->put_link(temp);
    n_items++;
  }
}


/************************************************************************
* fast_pop_top : pops an item off of the top of the queue but loses *
*        track of the number of items in the queue      *
************************************************************************/

C_ITEM *C_QUEUE::fast_pop_top() {
  C_ITEM *temp;
  temp = top;
  top = top->get_link();
  return temp;
}


