// td_evtq.C method file

#include "SpStdio.H"
#include "speedes_comm.H"
#include "td_evtq.H"
#include "host_user.H"

#define OPTIMIZE
#ifndef INFINITY
#define INFINITY 1.0e20
#endif

/************************************************************************
* C_TD_EVTQ : construct an event queue object				*
************************************************************************/
C_TD_EVTQ::C_TD_EVTQ() {
  SpEvent *event;

  printf("TIME BUCKET EVTQ created\n");

  SpCurrentAltMgr->UnsetOptimistic();

  while (n_items) {
    event = (SpEvent *)pop_top();
    event->insert();
  }

  localt = -INFINITY;
  gvt = -INFINITY;

  twosmess = new SpOptMessager(node, n_nodes, evtype);
  event->set_twosmess(twosmess);

  event->set_qproc(&dbl_qproc);

  n_roll = 0;
  n_not_valid = 0;

}

/************************************************************************
* temp_process - process events in temporary storage			*
************************************************************************/
void C_TD_EVTQ::temp_process() {
  double tupper;

//...... first get the starting time of the earliest event in the simulation

  localt = rqueue.get_time();
  localt = SpComm_GlobalMin(localt);
  //scombine((int *)&localt,MINDBL,sizeof(double),1);
  event_horizon = localt + minstep;
  if (scaled_time) {
    //tupper = gvtmax;
    tupper = SpComm_GlobalMin(gvtmax);
    //scombine((int *)&tupper,MINDBL,sizeof(double),1);
    if (event_horizon > tupper) event_horizon = tupper;
  }
  if (event_horizon > tend) event_horizon = tend;

  tmin = localt;
  gvt = localt;

  tblock = objects->get_tblock();

//...... initialize some values

  n_temp = 0;
  n_perm = 0;
  n_messrec = 0;
  n_exmess = 0;

//...... loop over the already processed events to find the starting horizon

  while (localt <= event_horizon) {
    process_event();
    messages();
  }

}

/************************************************************************
* process_event - process the current event optimistically		*
************************************************************************/
void C_TD_EVTQ::process_event() {
  C_QUEUE *mq;
  SpEvent *event;
  C_DBL_ITEM *dbl_item;

//...... get the next event out of the rqueue

  for(;;) {

    event = get_next_event();

    if (event == NULL) {
      localt = event_horizon;
      return;
    }

    if (localt > event->get_time_tag()) {
      fprintf(stderr,"Error, (TD evtq) time decreasing...\n");
    }

    localt = event->get_time_tag();

    if (localt < tmin) {
      fprintf(stderr,"Error, (TD evtq) local time < tmin\n");
    }

    if (localt > event_horizon) {
      event->insert();
      return;
    }

    if (localt > tblock) {
      localt = tblock;
      event->insert();
      return;
    }

    if (event->check_canceled(n_roll)) {
      event->set_processed();
      event->calculate_tmin();
      dbl_item = event->get_dbl_item();
      dbl_item->set_valu(event->get_time_tag());
      dbl_qproc.insert(dbl_item);
    }else{
      if (event->get_valid()) break;
      evtype->delete_event(event);
    }

  }

//...... make sure that the event is cleaned up

  mq = event->get_mque();
  if (mq->length()) {
    fprintf(stderr,"Error, (TD_EVTQ) event not clean\n");
  }

//...... process the event

  event->init_qroll_backs();
  event->temp_process();
  event->spin_loop(spin);
  event->exchange();

//...... mark the event as processed

  event->set_processed();

//...... convert the local messages to events right away

  n_roll += event->local_messages();

//...... check that tmin is ok

  event->calculate_tmin();

  if (event->get_tmin() < localt) {
    fprintf(stderr,"Error, (TB) event generated in past\n");
  }

  if (event->get_tmin() < event_horizon) {
    if ((event->get_tmin() - localt) < minstep) {
      fprintf(stderr,"Error, (TB) event generated too early\n");
    }else{
      fprintf(stderr,"Error, (TB) bad algorithm somewhere\n");
    }
  }

//...... put the event in the processed event queue

  dbl_item = event->get_dbl_item();
  dbl_item->set_valu(event->get_time_tag());
  dbl_qproc.insert(dbl_item);

  n_temp++;

}

/************************************************************************
* messages - receive incomming messages 				*
************************************************************************/
void C_TD_EVTQ::messages() {
  SpEvent *event;

  if (!interactive) return;

//...... handle externally generated events

  event = external_event(localt);
  while (event != NULL) {
    n_exmess++;
    event->insert();
    event = external_event(localt);
  }

  tblock = objects->get_tblock();

}


/************************************************************************
* find_gvt - find the minimum global time for safe processing		*
************************************************************************/
void C_TD_EVTQ::find_gvt() {

  gvt = event_horizon;

}

/************************************************************************
* perm_process - carry out the events and make them permanent		*
************************************************************************/
void C_TD_EVTQ::perm_process() {
  C_DBL_ITEM *dbl_item;
  SpEvent *event;
  int num;
  int i;
  char *message;
  SpMsg *header;
  C_HOLDER *holder;
  int size;

  //if (n_nodes > 1) multirouter->reset(); SpComm
  num = dbl_qproc.get_length();
  n_temp = num;

//...... loop over all of the events that can be processed

  for (i=0; i<num; i++) {

    dbl_item = (C_DBL_ITEM *)dbl_qproc.pop_top();
    event = (SpEvent *)dbl_item->get_ptr();

    if (event->get_valid()) {
      if (event->get_processed()) {
        event->perm_process();
	event->trace();
        event->sendmess(&q1nd);
	if (event->void_event()) {
	  if (event->get_event_canceled()) n_cancel++;
	}else{
          n_perm++;
	}
      }else{
	fprintf(stderr,"Error, (TD) event not processed\n");
      }
    }

    event->init_qroll_backs();
    SpCurrentAltMgr->Cleanup();
    evtype->delete_event(event);

  }

//...... get messages from the q1nd message queue

  while (q1nd.length()) {
    holder = (C_HOLDER *)q1nd.pop_top();
    message = holder->get_buff();
    header = (SpMsg *)message;
    if (header->ext) {
      n_exmess++;
      host_user->send_message((C_EM_HEADER *)message);
      evtype->delete_message(holder);
    }else{
      n_messrec++;
      event = (SpEvent *)evtype->message_event(message);
      evtype->delete_message(holder);
      event->insert();
    }
  }

//...... get the incomming messages and turn them into future events

  if (n_nodes > 1) {

    //multirouter->multirout(); SpComm

    //while((message = multirouter->getmess(size)) != NULL) {
    while((message = (char *)SpComm_CoordinatedReceive(size)) != NULL) {
      header = (SpMsg *)message;
      if (header->ext) {
        n_exmess++;
        host_user->send_message((C_EM_HEADER *)message);
      }else{
        n_messrec++;
        event = (SpEvent *)evtype->message_event(message);
        event->insert();
      }
      delete [] message;
    }

  }

}

/************************************************************************
* get_blocked_messages - get messages that are blocking SPEEDES		*
************************************************************************/
void C_TD_EVTQ::get_blocked_messages() {

  if (objects->get_tblock() <= gvt) {
    messages();
    objects->update_block();
  }

}



