// xhash.C method file

#include "SpStdio.H"
#include "SpString.H"
#include "xhash.H"
#include "RB_SpDefineClass.H"
RB_DEFINE_CLASS(RB_hash)

/************************************************************************
* C_XHASH : construct a xhash object					*
************************************************************************/
C_XHASH::C_XHASH(int n) {

  size = n;
  nitems = 0;
  q = new C_XQUEUE[size];

}

C_XHASH::~C_XHASH()
{
  delete [] q;
}

/************************************************************************
* reset : reset the xhash object 					*
************************************************************************/
void C_XHASH::reset() {
  int i;

  for (i=0; i<size; i++) q[i].reset();
  nitems = 0;

}

void C_XHASH::ClearAndDestroy() {
  int i;

  for (i=0; i<size; i++) q[i].ClearAndDestroy();
  nitems = 0;

}

/************************************************************************
* combine : combine all of the items into a single xqueue		*
************************************************************************/
void C_XHASH::combine() {
  int i;

  qall.reset();
  for (i=0; i<size; i++) {
    qall.join(&q[i]);
  }

}

/************************************************************************
* add : add the item to the xhash 					*
************************************************************************/
void C_XHASH::add(C_SQ_ITEM *item) {
  int mid;

  mid = (item->get_id())%size;
  q[mid].push_bot(item);
  nitems++;

}
/************************************************************************
* get_hashid : get the hash Id based on a name				*
************************************************************************/
int C_XHASH::get_hashid(char *name) {
  int HashId;
  int len;
  int i;

  HashId = 0;
  len = strlen(name);
  for (i=0; i<len; i++) {
    HashId ^= name[i] << ((i*8)%32);
  }

  return (HashId & 0x7fffffff);

}

/************************************************************************
* find : find an item from the hash using the item's name      		*
************************************************************************/
C_SQ_ITEM *C_XHASH::find(char *name) {
  int hid;
  int mid;
  int i;
  int len;
  C_SQ_ITEM *item;

  hid = get_hashid(name);
  mid = hid%size;

  item = (C_SQ_ITEM *)q[mid].get_top();
  len = q[mid].get_length();
  for (i=0; i<len; i++) {
    if (!strcmp(item->GetName(), name)) return item;
    item = (C_SQ_ITEM *)item->get_link();
  }

  return NULL;

}

/************************************************************************
* remove : remove this item from the xhash				*
************************************************************************/
void C_XHASH::remove(C_SQ_ITEM *item) {
  int mid;
  int len;

  mid = (item->get_id())%size;
  len = q[mid].get_length();
  q[mid].remove(item);
  if (q[mid].get_length() != len) {
    nitems--;
    return;
  }

  fprintf(stderr,"ERROR, could not remove the xhash id\n");

}

/************************************************************************
* remove : remove the item with this id from the xhash			*
************************************************************************/
C_SQ_ITEM *C_XHASH::remove(int id) {
  int mid;
  int i;
  int len;
  C_SQ_ITEM *item;

  mid = (id)%size;
  item = q[mid].get_top();
  len = q[mid].get_length();
  for (i=0; i<len; i++) {
    if (item->get_id() == id) {
      q[mid].remove(item);
      nitems--;
      return item;
    }
    item = (C_SQ_ITEM *)item->get_link();
  }

  return NULL;

}

/************************************************************************
* exchange : exchange the item						*
************************************************************************/
void C_XHASH::exchange(C_SQ_ITEM *x_item) {

  if (x_item->get_processed()) {
    add(x_item);
    x_item->set_xremove();
  }else{
    remove(x_item);
    x_item->set_xinsert();
  }

}

/************************************************************************
* printh								*
************************************************************************/
void C_XHASH::printh() {
  int i;

  for (i=0; i<size; i++) {
    if (q[i].get_length()) {
      fprintf(stderr,"XHASH[%d]\n",i);
      q[i].print();
    }
  }

}



void RB_Register_RB_hash()
{
  RB_REGISTER_CLASS_RB_hash();
}
