// xhash.H header file

#ifndef xhash_object
#define xhash_object

#include "xqueue.H"
#include "RB_SpDefineClass.H"

class C_XHASH  {

 public:

/******************************************************************************
* Method: reset
* Summary: Clear the hash table
* Description: Remove (without deleting) every item in the hash table
* Return: None
******************************************************************************/
  void reset();

/******************************************************************************
* Method: ClearAndDestroy
* Summary: Clear this hash table and delete all of the items
* Description:
* Return: None
******************************************************************************/
  void ClearAndDestroy();

/******************************************************************************
* Method: add
* Summary: Add a specified C_SQ_ITEM to the hash table
* Description: Push specified item to the bottom of the appropriate queue in
*              the hash array
* Return: None
******************************************************************************/
  void add(C_SQ_ITEM *item // C_SQ_ITEM to add
	   );

/******************************************************************************
* Overloaded Method: find
* Summary: Return the C_SQ_ITEM in this hash table specified by item's ID
* Description: Gets the item from the appropriate queue in the hash array
* Return: The C_SQ_ITEM in this hash table specified by item's ID
******************************************************************************/
  C_SQ_ITEM *find(int hid  // ID of item to find
		  ) {	int mid = hid%size;
		  return q[mid].find(hid); }

/******************************************************************************
* Overloaded Method: find
* Summary: Return the C_SQ_ITEM in this hash table specified by item's name
* Description: Gets the item from the appropriate queue in the hash array
* Return: The C_SQ_ITEM in this hash table specified by item's name
******************************************************************************/
  C_SQ_ITEM *find(char *name);

/******************************************************************************
* Method: get_hashid
* Summary: Passes in the name that is used for hashing and returns the hash
*          id that is used for the array index.
* Return: The hash id that is used for the array index
******************************************************************************/
  int get_hashid(char *name);

/******************************************************************************
* Method: combine
* Summary: Concatonate each queue in this hash array into one queue, leaving
*          the original hash table intact
* Description: The resulting single queue can then be retrieved with the method
*              get_all () and printed with printq ().
* Return: None
******************************************************************************/
  void combine();

/******************************************************************************
* Method: get_all
* Summary: Get the combined queue generated by the last call to combine ()
* Description: 
* Return: The combined queue generated by the last call to combine ()
******************************************************************************/
  C_XQUEUE *get_all() {return &qall;}

/******************************************************************************
* Overloaded Method: GetQ
* Summary: Get the Xqueue for the given hashid
* Description:
* Return: The Xqueue for the given hashid
******************************************************************************/
  C_XQUEUE *GetQ(int hashid = 0) {return &q[hashid%size];}

/******************************************************************************
* Overloaded Method: GetQ
* Summary: Get the Xqueue for the given name
* Description:
* Return: The Xqueue for the given name
******************************************************************************/
  C_XQUEUE *GetQ(char *name) {
    int hashid = get_hashid(name);
    return &q[hashid%size];
  }

/******************************************************************************
* Method: remove
* Summary: Remove (but don't delete) the specified item from this hash table
* Description: 
* Return: None
******************************************************************************/
  void remove(C_SQ_ITEM *item  // INOUT: C_SQ_ITEM to remove
	      );

/******************************************************************************
* Method: remove
* Summary: Remove (but don't delete) an item from this hash table specified
*          by its ID
* Description: 
* Return: None
******************************************************************************/
  C_SQ_ITEM *remove(int id  // IN: ID of C_SQ_ITEM to remove
		    );

/******************************************************************************
* Method: get_size
* Summary: Get the number of queues in this hash array
* Description: 
* Return: The number of queues in this hash array
******************************************************************************/
  int get_size() {return size;}

/******************************************************************************
* Method: get_length
* Summary: Get the number of C_SQ_ITEMs in this hash table
* Description: 
* Return: The number of C_SQ_ITEMs in this hash table
******************************************************************************/
  int get_length() {return nitems;}

/******************************************************************************
* Method: printq
* Summary: Print each C_SQ_ITEM in the combined hash table
* Description: Unlike printh (), which simply prints the hash table, this
*              method relies on first calling combine () which concatonates the
*              hash array into one combined queue. The resulting combined queue
*              is printed with this method by calling the virtual function
*              print () on all of the C_SQ_ITEMs in the combined queue.
* Return: None
******************************************************************************/
  void printq() {qall.print();}

/******************************************************************************
* Method: printh
* Summary: Print each C_SQ_ITEM in the hash table
* Description: This prints the hash table by calling the virtual function print
*              () on all of the C_SQ_ITEMs in the hash table. Unlike printq (),
*              this method does not rely on first calling combine ().
* Return: None
******************************************************************************/
  void printh();
  
/******************************************************************************
* Internal Method: exchange
* Summary: Exchange a specified C_SQ_ITEM for rollback purposes
* Description: 
* Return: None
******************************************************************************/
  void exchange(C_SQ_ITEM *x_item  // IN: C_SQ_ITEM to exchange
		);

/******************************************************************************
* Constructor Method: 
* Summary: Constructs a C_XHASH with a hash array of specified length
* Description: 
******************************************************************************/
  C_XHASH(int n = 100  // length of hash array
	 );

  ~C_XHASH();

  protected:

    int size;		// size of queue array
    int nitems;		// total number of items in the xhash
    C_XQUEUE *q;	// array of queues (dimension size)
    C_XQUEUE qall;	// a queue of concatonated xhash queues.

};

typedef C_XHASH RB_hash;
//RB_DEFINE_CLASS(RB_hash);

#endif
