// xqueue.C method file

#include "SpStdio.H"
#include "xqueue.H"
#include "RB_SpDefineClass.H"
RB_DEFINE_CLASS(C_XQUEUE)

/************************************************************************
* C_XQUEUE : construct q xqueue object					*
************************************************************************/
C_XQUEUE::C_XQUEUE() {

  length = 0;
  top = NULL;
  bot = NULL;

}

void C_XQUEUE::ClearAndDestroy()
{
  while (get_length())
    delete pop_top();
}

/************************************************************************
* push_top : pushes an item onto the top of the xqueue			*
************************************************************************/
void C_XQUEUE::push_top(C_SQ_ITEM *x_item) {

  if (length == 0) {
    bot = x_item;
    top = x_item;
    length = 1;
  }else{
    top->set_backlink(x_item);
    x_item->set_link(top);
    top = x_item;
    length++;
  }

}

/************************************************************************
* push_bot : pushes an item onto the bottom of the xqueue		*
************************************************************************/
void C_XQUEUE::push_bot(C_SQ_ITEM *x_item) {

  if (length == 0) {
    bot = x_item;
    top = x_item;
    length = 1;
  }else{
    bot->set_link(x_item);
    x_item->set_backlink(bot);
    bot = x_item;
    length++;
  }

}

/************************************************************************
* pop_top : pops an item off of the top of the xqueue			*
************************************************************************/
C_SQ_ITEM *C_XQUEUE::pop_top() {
  C_SQ_ITEM *x_item;

  if (length == 0) {
    return NULL;
  }else{
    x_item = top;
    top = (C_SQ_ITEM *)top->get_link();
    length--;
    return x_item;
  }

}

/************************************************************************
* pop_bot : pops an item off of the bottom of the xqueue		*
************************************************************************/
C_SQ_ITEM *C_XQUEUE::pop_bot() {
  C_SQ_ITEM *x_item;

  if (length == 0) {
    return NULL;
  }else{
    x_item = bot;
    bot = bot->get_backlink();
    length--;
    return x_item;
  }

}

/************************************************************************
* remove : remove an item from the  xqueue				*
************************************************************************/
void C_XQUEUE::remove(C_SQ_ITEM *x_item) {
  C_SQ_ITEM *previous;
  C_SQ_ITEM *next;

  if (length == 0) return;

  if (length == 1) {
    if (x_item != top) return;
    top = NULL;
    bot = NULL;
    length = 0;
    return;
  }

  if (x_item == top) {
    top = (C_SQ_ITEM *)x_item->get_link();
    length--;
    return;
  }

  if (x_item == bot) {
    bot = x_item->get_backlink();
    length--;
    return;
  }

  previous = x_item->get_backlink();
  next = (C_SQ_ITEM *)x_item->get_link();
  previous->set_link(next);
  next->set_backlink(previous);
  length--;

}

/************************************************************************
* exchange : exchange an item from the xqueue				*
************************************************************************/
void C_XQUEUE::exchange(C_SQ_ITEM *x_item) {

  if (x_item->get_processed()) {
    push_bot(x_item);
    x_item->set_xremove();
  }else{
    remove(x_item);
    x_item->set_xinsert();
  }

}

/************************************************************************
* find : find an item (don't remove) in the xqueue			*
************************************************************************/
C_SQ_ITEM * C_XQUEUE::find(int id) {
  C_SQ_ITEM *x_item;

  if (length == 0) return NULL;

  x_item = top;
  
  while (1) {
    if (x_item->get_id() == id) return x_item;
    if (x_item == bot) break;
    x_item = (C_SQ_ITEM *)x_item->get_link();
  }

  return NULL;
}

/************************************************************************
* join : join the input queue to this queue but don't reset		*
************************************************************************/
void C_XQUEUE::join(C_XQUEUE *q) {
  C_SQ_ITEM *temp;

  if (q->get_length() == 0) return;

  if (length == 0) {

    top = q->get_top();
    bot = q->get_bot();
    length = q->get_length();

  }else{

    bot->set_link(q->get_top());
    temp = q->get_top();
    temp->set_backlink(bot);
    bot = q->get_bot();
    length += q->get_length();

  }

}

/************************************************************************
* print : print the queue						*
************************************************************************/
void C_XQUEUE::print() {
  int i;
  C_SQ_ITEM *temp;

  fprintf(stderr,"Printing XQUEUE with %d elements\n",length);
  temp = top;
  for (i=0; i<length; i++) {
    temp->print();
    temp = (C_SQ_ITEM *)temp->get_link();
  }

}

/************************************************************************
* check_integrity : check the integrity of the queue			*
************************************************************************/
int C_XQUEUE::check_integrity() {
  int i;
  C_SQ_ITEM *temp;

  if (length == 0) return 1;

  temp = top;
  for (i=0; i<(length-1); i++) {
    temp = (C_SQ_ITEM *)temp->get_link();
  }

  if (temp == bot) return 1;
  fprintf(stderr,"ERROR Xqueue failed integrity test\n");
  return 0;

}


void RB_Register_C_XQUEUE()
{
  RB_REGISTER_CLASS_C_XQUEUE();
}
