// SpExtFedStateMgr.C source file

#include "SpExtFedStateMgr.H"
#include "SpExtMsgSimObjIds.H"
#include "SpExtFedRecvInteractionEvent.H"

DEFINE_STATE_MGR_EVENT(ExtFedRecvInteraction, SpExtFedRecvInteractionEvent);


SpExtFedStateMgr::SpExtFedStateMgr (char *machineName, int port, int groupId) :
  SpFedStateMgr(machineName, port, groupId)
{
  FedAmb = NULL;
}

SpExtFedStateMgr::SpExtFedStateMgr(SpDataParser *parser) :
  SpFedStateMgr(parser)
{
  FedAmb = NULL;
}

SpExtFedStateMgr::SpExtFedStateMgr(char *fileName) :
  SpFedStateMgr(fileName)
{
  FedAmb = NULL;
}


char* SpExtFedStateMgr::Connect(const char *fedName, const char *fedexName,
				SpSimTime connectTime)
{
  char *data;
  SpSimTime time;
  int dataBytes;

  FedName = strdup(fedName);
  NodeId = EmHostUser->GetNextNode();
  MgrId = EmHostUser->ObjectManagerType("SpFedGhost");

  dataBytes = strlen(fedName) + strlen(fedexName) + 2;
  data = new char[dataBytes];
  strcpy (data, fedName);
  strcpy (data+(strlen(fedName)+1), fedexName);

  EmHostUser->SendCommand("CreateFedGhost",
			  NodeId, MgrId, -1,
			  connectTime,
			  data, dataBytes);
  delete [] data;

  data = NULL;
  while (!data) {
    data = EmHostUser->ReceiveNamedData("SimObj Ids",
					time,
					dataBytes,
					GlobalId);
  }

  SpExtMsgSimObjIds *extMsgSimObjIds = (SpExtMsgSimObjIds *)data;
  NodeId = extMsgSimObjIds->NodeId;
  MgrId = extMsgSimObjIds->MgrId;
  LocalId = extMsgSimObjIds->LocalId;

  char *FED = strdup(data + sizeof(SpExtMsgSimObjIds));
  delete [] data;

  TimeGranted = time;

  cout << "Received SimObj Ids = <"
       << NodeId << ", "
       << MgrId << ", "
       << LocalId << ", "
       << GlobalId << ">"
       << endl;

  // return FED location
  return FED;
}

void SpExtFedStateMgr::Disconnect() {

  SpSimTime eventTime = TimeGranted;
  eventTime += TimeLag;

  EmHostUser->SendCommand("Resign Fed",
			  NodeId,
			  MgrId,
			  LocalId,
			  eventTime,
			  NULL, 0);
}


/*
*****************************************************************************
* Subscribe Interaction method
*****************************************************************************
*/

void SpExtFedStateMgr::SubscribeInteraction(char *className) {

  SpSimTime eventTime = TimeGranted;
  eventTime += TimeLag;
  EmHostUser->SendCommand("Subscribe Interaction",
			  NodeId,
			  MgrId,
			  LocalId,
			  eventTime,
			  className,
			  strlen(className) + 1);

  // plug in a handler for this interaction class
  this->PlugInEvent(className, NewExtFedRecvInteraction);
}

/*
*****************************************************************************
* Next event request method
*****************************************************************************
*/

void SpExtFedStateMgr::NextEventRequest(SpSimTime &time) {

  if(time >= FutureEvents.get_time()) {
    // can grant time from here
    TimeGranted = FutureEvents.get_time();
    TimeGrantedFlag = 1;
    cout << "Received TimeAdvanceGrant for " << TimeGranted << endl;
    return;
  }

  EmHostUser->SendCommand("Next Event Request",
			  NodeId,
			  MgrId,
			  LocalId,
			  TimeGranted,
			  (char *)&time,
			  sizeof(SpSimTime));
  TimeRequested = time;
  TimeGrantedFlag = 0;
}
