// SpFedStateMgr.H header file

#ifndef DEF_SpFedStateMgr
#define DEF_SpFedStateMgr

#include "SpStdarg.H"
#include "SpStdlib.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"

#include "SpDataParser.H"
#include "SpEmHostUser.H"
#include "SpStateMgrEvent.H"
#include "SpObjProxy.H"
#include "SpPriorityQueueHeap.H"
#include "xqueue.H"
#include "xhash.H"
#include "SpGlobalState.H"
#include "F_SpProxyItem.H"
#include "free_list.H"
#include "SpSimTime.H"
#include "SpBinaryTree.H"
#include "SpHashTree.H"
#include "SpParmSet.H"
#include "SpList.H"

enum HLA_MODE {
  HLA_MODE_GATEWAY,
  HLA_MODE_EXTERNAL_FEDERATE,
  HLA_MODE_DIRECT_FEDERATE
};

class SpEventConstructorsByName {
  public:
    SpList EventNameConstructors;
};

class SpEventNameConstructor {
  public:
    SpStateMgrEvent *(*f)();
};

class SpFedStateMgr {

  private:
  
  protected:
    
    double TimeLag;		      	// Max time lag with SPEEDES
    SpSimTime TimeGranted;		// Time granted by SPEEDES
    SpSimTime TimeRequested;            // Most recent time request time
    int TimeGrantedFlag;		// Flag indicating time was granted

    void ValidateGrant();               // make sure grant arrived after event

    SpBinaryTree EventConstructors;	// Hash table of event constructors

    HLA_MODE Mode;	       		// Mode for the FedStateMgr

    char *FedName;			// Name of the federate

    int NodeId;				// Node of the FedGhost
    int MgrId;				// Object manager type for the FedGhost
    int LocalId;			// Local Id for the FedGhost
    int GlobalId;			// Global Id for the FedGhost

    SpNameValueList *ObjectTypes;       // List of object type ids & names
    SpNameValueList *ObjectInstances;   // List of object names & ids

    void *UserState;                    // User-defined state variables
    C_XHASH *ObjectProxyIdHash;         // Rollbackable hash table of proxies
    C_XHASH *ObjectProxyNameHash;       // Rollbackable named hash table
    C_XHASH *ObjectProxyTypeHash;       // Rollbackable typed hash table
    C_XQUEUE *CurrentQ;                 // Current Xqueue in hash table
    SpObjProxy *CurrentObjectProxy;	// Current Object Proxy
    F_SpProxyItem *CurrentProxyItem;	// Current proxy item holds a proxy
  
    SpPriorityQueueHeap FutureEvents;	// Future unprocessed events
    C_XQUEUE ProcessedEvents;	      	// List of processed events
    SpStateMgrEvent *CurrentEvent;  	// Last event processed
    SpEmHostUser *EmHostUser;	      	// Host User for external modules
    SpGlobalState *GlobalState;        	// Global state shared by events

    void Init();		    	// called internally by constructors
    SpObjProxy *GetNextProxyByType();	// Gets the next proxy by type

  public:

    SpFedStateMgr(char *machineName, int port, int groupId);
    SpFedStateMgr(SpDataParser *parser);
    SpFedStateMgr(char *fileName);

    virtual ~SpFedStateMgr();

    void PlugInEvent(char *eventName, SpStateMgrEvent *(*f)());

    void GoToTime(SpSimTime time);

    SpSimTime GetLastProcessedEventTime() {
      SpStateMgrEvent *LastProcessedEvent =
	(SpStateMgrEvent *)ProcessedEvents.get_bot();
      return LastProcessedEvent->get_time_tag();
    }

    int GetNumProxies() {return ObjectProxyIdHash->get_length();}

    SpObjProxy *GetFirstProxy() {
      CurrentQ = NULL;
      ObjectProxyIdHash->combine();
      C_XQUEUE *Qall = ObjectProxyIdHash->get_all();
      CurrentProxyItem = (F_SpProxyItem *)Qall->get_top();
      if (CurrentProxyItem) {
	CurrentObjectProxy = CurrentProxyItem->GetObjectProxy();
      }else{
	CurrentObjectProxy = NULL;
      }
      return CurrentObjectProxy;
    }
  
    SpObjProxy *GetNextProxy();

    int GetNumProxies(char *type) {
      int counter = 0;
      SpObjProxy *ObjectProxy;
      C_XQUEUE *Q = ObjectProxyTypeHash->GetQ(type);
      F_SpProxyItem *ProxyItem = (F_SpProxyItem *)Q->get_top();
      for (int i=0; i<Q->get_length(); i++) {
	ObjectProxy = ProxyItem->GetObjectProxy();
	if (!strcmp(type, ObjectProxy->ObjProxyGetClassName())) counter++;
	ProxyItem = (F_SpProxyItem *)ProxyItem->get_link();
      }
      return counter;
    }
  
    SpObjProxy *GetFirstProxy(char *type) {
      SpObjProxy *ObjectProxy;
      SpObjProxy *ReturnedObjectProxy = NULL;
      CurrentQ = ObjectProxyTypeHash->GetQ(type);
      F_SpProxyItem *ProxyItem = (F_SpProxyItem *)CurrentQ->get_top();
      for (int i=0; i<CurrentQ->get_length(); i++) {
	ObjectProxy = ProxyItem->GetObjectProxy();
	if (!strcmp(type, ObjectProxy->ObjProxyGetClassName())) {
	  CurrentObjectProxy = ObjectProxy;
	  CurrentProxyItem = ProxyItem;
	  ReturnedObjectProxy = ObjectProxy;
	  break;
	}
	ProxyItem = (F_SpProxyItem *)ProxyItem->get_link();
      }
      return ReturnedObjectProxy;
    }
  
    SpObjProxy *GetProxy(char *objectName) {
      SpObjProxy *ObjectProxy;
      F_SpProxyItem *ProxyItem = (F_SpProxyItem *)
	ObjectProxyNameHash->find(objectName);
      if (ProxyItem) {
	ObjectProxy = ProxyItem->GetObjectProxy();
      }else{
	ObjectProxy = NULL;
      }
      return ObjectProxy;
    }

    SpEmHostUser *GetEmHostUser() {return EmHostUser;}

    void ScheduleEvent(SpStateMgrEvent *event, SpSimTime time) {
      if (event->get_link()) {
	cout << "Warning (StateManager): Posting event twice?" << endl;
      }else{
	event->set_time_tag(time);
	SpStateMgrEvent *LastProcessedEvent =
	  (SpStateMgrEvent *)ProcessedEvents.get_bot();
	if (time < LastProcessedEvent->get_time_tag()) {
	  cout << "Warning (StateManager): Posting an event in the past"
	       << endl;
	}else{
	  FutureEvents.insert(event);
	}
      }
    }
  
    void SetUserState(void *state) {UserState = state;}
    void *GetUserState(){ return UserState; }

    int GetGlobalId() { return GlobalId; }

    /*** ***/

    void ReadMessages();

    void Subscribe(char *className);
    virtual void SubscribeInteraction(char *className);

    void SendInteraction(char *className, SpParmSet &paramSet,
			 SpSimTime eventTime=-1.0e20);

    virtual void NextEventRequest (SpSimTime &time);
    void SetTimeLag (double timeLag);
    double GetTimeLag () { return TimeLag; }

    int TimeAdvanceGranted();
    SpSimTime& GetTimeGranted() {return TimeGranted;}

    char* GetObjectName(int instId);
    int GetObjectId(const char *instName);
    char* GetObjectMgrName(int typeId);
    int GetObjectMgrId(const char *typeName);
};

#endif

