// ReflectAttributeEvent.C source file

#include "SpReflectAttributesEvent.H"
#include "F_SpProxyItem.H"

#include "SpObjFactory.H"
extern SpObjFactory *ObjFact;

#include "SpDynItem.H"
#include "SpFreeDynAttributes.H"
extern SpFreeDynAttributes *FreeDynamicAttributes;

void SpReflectAttributesEvent::Process() {
  int i;
  int Natts;
  int Nbytes;
  int OpWarning;
  SpObjProxy *ObjectProxy;
  SpObjProxy *Op;
  ATTRIBUTE_MESSAGE_INFO *Info;
  char *Buff;
  BASE_ATTRIBUTE *Attribute;
  PROXY_LEVEL_INFO *ProxyLevelInfo;
  int AttributeIndex;
  int ArrayIndex;
  int Level;
  OBJECT_ATTRIBUTE *ObjectAttribute;
  LIST_ATTRIBUTE *ListAttribute;
  SpDynItem *DynamicItem;
  BASE_DYNAMIC_ATTRIBUTE *DynamicAttribute;

  cout << "Processing: SpReflectAttributesEvent at time " << time_tag << endl;
  GlobalState->IncrementNumberOfReflections();

  F_SpProxyItem *ProxyItem = (F_SpProxyItem *)ObjectProxyIdHash->find(GlobalId);
  if (ProxyItem) {
    cout << "Got it" << endl;
    ObjectProxy = ProxyItem->GetObjectProxy();
  }else{
    cout << "Error, could not find Object Proxy with Id " << GlobalId << endl;
    return;
  }

/*
** Go through the message and update the attributes for the Object Proxy
*/

  Buff = new char[DataBytes];
  memcpy(Buff, Data, DataBytes);
  Data = Buff;
//  Buff = Data;
  Nbytes = DataBytes;
  Natts = 1000000;

//  cout << "Number of reflected attributes = " << Natts << endl;

  for (i=0; i<Natts; i++) {

    ProxyLevelInfo = (PROXY_LEVEL_INFO *)Buff;
    Level = ProxyLevelInfo->Level;

    Buff += Level*sizeof(PROXY_LEVEL_INFO);
    Info = (ATTRIBUTE_MESSAGE_INFO *)Buff;

    AttributeIndex = Info->GetAttributeIndex();
    ArrayIndex = Info->GetArrayIndex();

/*
    cout << endl;
    for (int j=0; j<Level; j++) ProxyLevelInfo[j].print();
*/

    Op = ObjectProxy->FindProxy(ProxyLevelInfo);

    if (Op) {
      Attribute = Op->GetAttribute(AttributeIndex, ArrayIndex);
      if (Attribute) {

	if (Info->GetAddRemove() == NEITHER) {
/*
** Handle a Normal Attribute (i.e., anything that does not add or remove an
** item from a list attribute.
*/

/*
	  cout << i << " Reflecting Normal Attribute: Index " << AttributeIndex
	       << ", ArrayIndex " << ArrayIndex
	       << ", Bytes " << Info->GetNbytes()
	       << endl;
*/

	  Attribute->EmptyMessage(Buff);

	}else{

/*
	  cout << i <<  " Reflecting list Attribute: Index " << AttributeIndex
	       << ", ArrayIndex " << ArrayIndex
	       << ", Bytes " << Info->GetNbytes()
	       << endl;
*/

	  if (Attribute->GetType() == LIST_ATTRIBUTE_TYPE) {

/*
** Handle a list attribute that contains object attributes.
*/

	    ListAttribute = (LIST_ATTRIBUTE *)Attribute;

	    if (Info->GetAddRemove() == OBJECT_PROXY_ADD) {

//	      cout << "Adding an Object Attribute to a List" << endl;

	      //ObjectAttribute = (OBJECT_ATTRIBUTE *) // bad lines
	      //  ObjFact->CreateAttribute(OBJECT_ATTRIBUTE_TYPE,1);

	      ObjectAttribute = new OBJECT_ATTRIBUTE; // better line
	      ObjectAttribute->EmptyMessage(Buff);

	      *ListAttribute += ObjectAttribute;

	    }else{

//	      cout << "Removing an Object Attribute from a List" << endl;

	      ObjectAttribute =
		ListAttribute->FindObjectAttribute(Info->GetAddRemoveId());

	      if (ObjectAttribute) {
		*ListAttribute -= ObjectAttribute;
	      }else{
/*
		cout << "Error (ReflectAttributes): Object not in list"
		     << endl;
*/
	      }

	    }

	  }else{

/*
** Handle a dynamic attribute that contains dynamic items.
*/

	    DynamicAttribute = (BASE_DYNAMIC_ATTRIBUTE *)Attribute;

	    if (Info->GetAddRemove() == OBJECT_PROXY_ADD) {

//	      cout << "Adding a dynamic item to a Dynamic Attribute" << endl;

	      DynamicItem = (SpDynItem *)
		FreeDynamicAttributes->generate((char *)&Info[1]);

	      *DynamicAttribute += DynamicItem;

	    }else{
/*
	      cout << "Removing a dynamic item from a Dynamic Attribute"
		   << endl;
*/

	      DynamicItem = 
		DynamicAttribute->FindDynamicItem(Info->GetAddRemoveId());

	      if (DynamicItem) {
		*DynamicAttribute -= DynamicItem;
	      }else{
		cout << "Error (ReflectAttributes): "
		     << "Dynamic Item not in Dynamic Attribute"
		     << endl;
	      }

	    }

	  }

	}

      }else{
/*
** Error: Could not find Attribute in Op
*/
	  cout << "Error (ReflectAttributes): Could Not Find Attribute"
	       << endl;
      }

    }else{

/*
** Warning: Could not find Object Proxy (Op) but that may be Ok since the
** Object Proxy will be created later in this event in an Object Attribute.
** This should not be a problem, but we should verify to make sure.
*/

      OpWarning = 1;

    }

/*
** End of the loop where we advance Buff by the number of bytes in the Info
*/

    Buff += Info->GetNbytes();

    if (int(Buff-Data) == Nbytes) {
      break;
    }

    if (int(Buff-Data) > Nbytes) {
      cout << "Error (ReflectAttributesEvent): bad number of bytes" << endl;
      break;
    }

  }

  if (OpWarning) {

    Buff = Data;
    for (i=0; i<Natts; i++) {

      ProxyLevelInfo = (PROXY_LEVEL_INFO *)Buff;
      Level = ProxyLevelInfo->Level;

      Buff += Level*sizeof(PROXY_LEVEL_INFO);
      Info = (ATTRIBUTE_MESSAGE_INFO *)Buff;

      AttributeIndex = Info->GetAttributeIndex();
      ArrayIndex = Info->GetArrayIndex();

      Op = ObjectProxy->FindProxy(ProxyLevelInfo);

      if (Op == NULL) {
	cout << "Warning (ReflectAttributesEvent): OpWarning set" << endl;
      }

      Buff += Info->GetNbytes();

      if (int(Buff-Data) == Nbytes) {
	break;
      }

      if (int(Buff-Data) > Nbytes) {
	cout << "Error (ReflectAttributesEvent): bad number of bytes" << endl;
	break;
      }

    }

  }

//  ObjectProxy->PrintStream();

}
