// SpStateMgrEvent.C file

#include "SpStateMgrEvent.H"

void *SpStateMgrEvent::UserState = NULL;
C_XHASH *SpStateMgrEvent::ObjectProxyIdHash = NULL;
C_XHASH *SpStateMgrEvent::ObjectProxyNameHash = NULL;
C_XHASH *SpStateMgrEvent::ObjectProxyTypeHash = NULL;
SpEmHostUser *SpStateMgrEvent::EmHostUser = NULL;
SpGlobalState *SpStateMgrEvent::GlobalState = NULL;
SpPriorityQueueHeap *SpStateMgrEvent::FutureEvents = NULL;
C_XQUEUE *SpStateMgrEvent::ProcessedEvents = NULL;

/*
*****************************************************************************
* ProcessEvent() method sets up the rollback mechanisms and then calls the
* virtual function, Process(), that provides specific processing for the event
*****************************************************************************
*/

void SpStateMgrEvent::ProcessEvent() {

  InitRollBackQueue();

  switch (ProcessedFlag) {

    case EVENT_PROCESSED:
      cout << "Time " << TIME_TAG << endl;
      cout << "Warning, Processing an event twice... "
	   << "doing nothing" << endl;
      break;

    case EVENT_UNPROCESSED:
      Process();
      ProcessedFlag = EVENT_PROCESSED;
      cout << AltMgr.GetNumAltItems() << " rollback items created" << endl;
      break;

    case EVENT_ROLLEDBACK:
      cout << "Time " << TIME_TAG << endl;
      cout << "Warning, processing an event that was rolled back... "
	   << "Rolling forward" << endl;
      RollForwardEvent();
      break;

  }

}

/*
*****************************************************************************
* RollbackEvent() method rolls back an already processed event
*****************************************************************************
*/

void SpStateMgrEvent::RollBackEvent() {

  InitRollBackQueue();

  switch (ProcessedFlag) {

    case EVENT_PROCESSED:
      SpCurrentAltMgr->Rollback();
      ProcessedFlag = EVENT_ROLLEDBACK;
      break;

    case EVENT_UNPROCESSED:
      cout << "Time " << TIME_TAG << endl;
      cout << "Warning, rolling back an event that was never processed... "
	   << "doing nothing" << endl;
      break;

    case EVENT_ROLLEDBACK:
      cout << "Time " << TIME_TAG << endl;
      cout << "Warning, rolling back an event twice... "
	   << "doing nothing" << endl;
      break;

  }

}

/*
*****************************************************************************
* RollForwardEvent() method rolls an event forward that has already been
* processed and rolled back
*****************************************************************************
*/

void SpStateMgrEvent::RollForwardEvent() {

  InitRollBackQueue();

  switch (ProcessedFlag) {

    case EVENT_PROCESSED:
      cout << "Time " << TIME_TAG << endl;
      cout << "Warning, rolling forward a processed event... "
	   << "doing nothing" << endl;
      break;

    case EVENT_UNPROCESSED:
      cout << "Time " << TIME_TAG << endl;
      cout << "Warning, rolling forward an unprocessed event... "
	   << "processing the event" << endl;
      ProcessEvent();
      break;

    case EVENT_ROLLEDBACK:
      SpCurrentAltMgr->Rollback();
      ProcessedFlag = EVENT_PROCESSED;
      break;

  }

}

