#ifndef RBSpKey_H
#define RBSpKey_H


#include "SpString.H"
#include "SpStdlib.H"
#include "SpPoDataBase.H"
#include "SpPersistenceBaseClass.H"
#include "SpMempool.H"
#include "SpKeyMode.H"

extern SpPoDataBase* DefaultDataBase;

class RBSpKey : private SpPersistenceBaseClass{
private:
protected:

public:
  RBSpKey(){}
  virtual ~RBSpKey(){}
  virtual RBSpKey * Clone(){ return NULL;}
  virtual KEY_MODE GetMode() const{return NO_KEY_MODE;}
  int operator <(const RBSpKey & y)const { return this->lt(&y);}
  int operator >(const RBSpKey &y)const { return this->gt(&y);}
  int operator ==(const RBSpKey &y)const { return this->eq(&y);}
  int operator >=(const RBSpKey &y)const { return this->geq(&y);}
  int operator <=(const RBSpKey &y)const { return this->leq(&y);}

  virtual int lt(const RBSpKey*)const;
  virtual int gt(const RBSpKey*)const;
  virtual int eq(const RBSpKey*)const;
  virtual int leq(const RBSpKey*)const;
  virtual int geq(const RBSpKey*)const;
};


class RBSpDoubleKey :public RBSpKey{
private:
  double Val;

public:
  RBSpDoubleKey(){;}
  RBSpDoubleKey(double v){Val = v;}

  virtual KEY_MODE GetMode() const{return DOUBLE_KEY_MODE;}
  virtual RBSpKey * Clone(){ return new RBSpDoubleKey(*this);}
  virtual operator double () {return Val;}
  int operator <(const RBSpDoubleKey & y)const { return (Val < y.Val);}
  int operator >(const RBSpDoubleKey & y)const { return (Val > y.Val);}
  int operator ==(const RBSpDoubleKey & y)const { return (Val == y.Val);}
  int operator >=(const RBSpDoubleKey & y)const { return (Val >= y.Val);}
  int operator <=(const RBSpDoubleKey & y)const { return (Val <= y.Val);}

  virtual int lt(const RBSpKey*y)const
    {return this->Val < ((RBSpDoubleKey*)y)->Val;}
  virtual int gt(const RBSpKey*y)const
    {return this->Val > ((RBSpDoubleKey*)y)->Val;}
  virtual int eq(const RBSpKey*y)const
    {return this->Val == ((RBSpDoubleKey*)y)->Val;}
  virtual int leq(const RBSpKey*y)const
    {return this->Val <= ((RBSpDoubleKey*)y)->Val;}
  virtual int geq(const RBSpKey*y)const
    {return this->Val >= ((RBSpDoubleKey*)y)->Val;}

};


class RBSpStringKey :public RBSpKey{
protected:
  char* Val;

public:
  virtual KEY_MODE GetMode() const{return STRING_KEY_MODE;}
  RBSpStringKey(){Val = NULL;}
  RBSpStringKey(const char * s){Val = strdup(s);}
  RBSpStringKey(const RBSpStringKey &s){Val = strdup(s.Val);}
  RBSpStringKey & operator = (const RBSpStringKey &s)
    {
      if (this != &s){
	free(Val);
	Val = strdup(s.Val);
      }
      return *this;
    }
  virtual ~RBSpStringKey (){free(Val);}

  virtual RBSpKey * Clone(){ return new RBSpStringKey(*this);}
  virtual operator char* () {return Val;}

  virtual int lt(const RBSpKey*y)const 
    {return strcmp(this->Val, ((RBSpStringKey*)y)->Val) < 0;}
  virtual int gt(const RBSpKey*y)const
    {return strcmp(this->Val, ((RBSpStringKey*)y)->Val) > 0;}
  virtual int eq(const RBSpKey*y)const
    {return strcmp(this->Val, ((RBSpStringKey*)y)->Val) == 0;}
  virtual int leq(const RBSpKey*y)const
    {return strcmp(this->Val, ((RBSpStringKey*)y)->Val) > -1;}
  virtual int geq(const RBSpKey*y)const
    {return strcmp(this->Val, ((RBSpStringKey*)y)->Val) < 1;}

};



class RBSpPersistentStringKey :public RBSpStringKey{
private:
  int PointerRecorded;
  static int PO_charId;

  void AttachPointer();

public:
  virtual KEY_MODE GetMode() const{return STRING_KEY_MODE;}
  RBSpPersistentStringKey():PointerRecorded(0){}
  RBSpPersistentStringKey(const char * s)
    {
      if (PO_charId == -1){
	PO_charId = DefaultDataBase->GetClassId("char");
      }
      DefaultDataBase->Add((void*&)Val, strlen(s)+1, PO_charId);
      if (PointerRecorded == 0){
	AttachPointer();
      }
    }

  RBSpPersistentStringKey(RBSpPersistentStringKey &s)
    {
      if (PO_charId == -1){
	PO_charId = DefaultDataBase->GetClassId("char");
      }
      DefaultDataBase->Add((void*&)Val, strlen(s)+1, PO_charId);
      if (PointerRecorded == 0){
	AttachPointer();
      }
    }


  RBSpPersistentStringKey& operator = (RBSpPersistentStringKey &s)
    {
      if (this != &s){
	free(Val);
	Val = strdup(s.Val);
      }
      if (PO_charId == -1){
	PO_charId = DefaultDataBase->GetClassId("char");
      }
      DefaultDataBase->Add((void*&)Val, strlen(s)+1, PO_charId);
      if (PointerRecorded == 0){
	AttachPointer();
      }
      return *this;
    }
  
  virtual ~RBSpPersistentStringKey ()
    {
      free(Val);
      Val = NULL;
      DefaultDataBase->Remove((void*&)Val);
    }

  virtual RBSpKey * Clone(){ return new RBSpPersistentStringKey(*this);}
};



//DEFINE_MEMPOOL( SpStringKey, 100 );
//DEFINE_MEMPOOL( SpIntervalKey, 100 );
//DEFINE_MEMPOOL( SpDoubleKey, 100 );

#endif
