#include "RB_SpBinaryTree.H"
#include "RB_SpDefineClass.H"


#ifndef OLD_FUNCTIONALITY

// BinaryTree.C method file

#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "RB_SpBinaryTree.H"

const double RB_TREE_BALANCE_RATIO = 20.0;
const int RB_TREE_CUTOFF = 40;
const int RB_N_SPLAY = 100;


RB_SpBinaryTree* junkTreeInstance = NULL;

PO_DEFINE_CLASS(RBSpPersistentStringKey);
PO_DEFINE_CLASS(RBSpDoubleKey);
RB_DEFINE_CLASS(RB_SpBinaryTree);









int BaseRB_SpTreeItem::operator == (const BaseRB_SpTreeItem &item) const
{return this->eq(&item);}

int BaseRB_SpTreeItem::operator < (const BaseRB_SpTreeItem &item) const
{return this->lt(&item);}

int BaseRB_SpTreeItem::operator <= (const BaseRB_SpTreeItem &item) const
{return this->leq(&item);}

int BaseRB_SpTreeItem::operator > (const BaseRB_SpTreeItem &item) const
{return this->gt(&item);}

int BaseRB_SpTreeItem::operator >= (const BaseRB_SpTreeItem &item) const
{return this->geq(&item);}

void BaseRB_SpTreeItem::SetTieBreaker(int t){TieBreaker = t;}
void BaseRB_SpTreeItem::SetNleft(int nl){Nleft = nl;}
void BaseRB_SpTreeItem::SetNright(int nr){Nright = nr;}
int BaseRB_SpTreeItem::GetTieBreaker(){return TieBreaker;}


void PersistentRB_SpTreeItem::Cleanup(){
  switch(NodeKey->GetMode()){
  case DOUBLE_KEY_MODE:
    GlobalSpPoMempool->ReturnItem(DoubleKeyFreeId, (RBSpKey*)NodeKey);
    break;
  case STRING_KEY_MODE:
    GlobalSpPoMempool->ReturnItem(StringKeyFreeId, (RBSpKey*)NodeKey);
    break;
  default:
    cout<<"Unknown key type for RB_SpBinaryTree: "<<NodeKey->GetMode()<<endl;
    break;
  }
}

int PersistentRB_SpTreeItem::lt(const BaseRB_SpTreeItem* item) const{
  return NodeKey->lt(((PersistentRB_SpTreeItem*)item)->NodeKey);
}
int PersistentRB_SpTreeItem::leq(const BaseRB_SpTreeItem* item) const{
  return NodeKey->leq(((PersistentRB_SpTreeItem*)item)->NodeKey);
}
int PersistentRB_SpTreeItem::eq(const BaseRB_SpTreeItem* item) const{
  return NodeKey->eq(((PersistentRB_SpTreeItem*)item)->NodeKey);
}
int PersistentRB_SpTreeItem::geq(const BaseRB_SpTreeItem* item) const{
  return NodeKey->geq(((PersistentRB_SpTreeItem*)item)->NodeKey);
}
int PersistentRB_SpTreeItem::gt(const BaseRB_SpTreeItem* item) const{
  return NodeKey->gt(((PersistentRB_SpTreeItem*)item)->NodeKey);
}


BaseRB_SpTreeItem *PersistentRB_SpTreeItem::GetParent(){return Parent;}

BaseRB_SpTreeItem *PersistentRB_SpTreeItem::GetLeftChild()
{return LeftChild;}

BaseRB_SpTreeItem *PersistentRB_SpTreeItem::GetRightChild()
{return RightChild;}

RBSpKey * PersistentRB_SpTreeItem::GetNodeKey(){return NodeKey;}

void PersistentRB_SpTreeItem::SetParent(BaseRB_SpTreeItem* t)
{ Parent=(PersistentRB_SpTreeItem*)t;}

void PersistentRB_SpTreeItem::SetLeftChild(BaseRB_SpTreeItem*t)
{LeftChild=(PersistentRB_SpTreeItem*)t;}

void PersistentRB_SpTreeItem::SetRightChild(BaseRB_SpTreeItem*t)
{ RightChild=(PersistentRB_SpTreeItem*)t;}

void PersistentRB_SpTreeItem::SetNodeKey(RBSpKey* k){NodeKey = k;}

void* PersistentRB_SpTreeItem::GetElement() const {return Element;}

void PersistentRB_SpTreeItem::SetElement(void* e)  {Element = (char*)e;}

double PersistentRB_SpTreeItem::DoubleVal(){
  return (double)*(RBSpDoubleKey*)((RBSpKey*)NodeKey);
}




void NonPersistentRB_SpTreeItem::Cleanup(){
  switch(NodeKey->GetMode()){
  case DOUBLE_KEY_MODE:
    GlobalSpPoMempool->ReturnItem(DoubleKeyFreeId, NodeKey);
    break;
  case STRING_KEY_MODE:
    GlobalSpPoMempool->ReturnItem(StringKeyFreeId, NodeKey);
    break;
  default:
    cout<<"Unknown key type for RB_SpBinaryTree: "<<NodeKey->GetMode()<<endl;
    break;
  }
}

int NonPersistentRB_SpTreeItem::lt(const BaseRB_SpTreeItem* item) const{
  return NodeKey->lt(((NonPersistentRB_SpTreeItem*)item)->NodeKey);
}
int NonPersistentRB_SpTreeItem::leq(const BaseRB_SpTreeItem* item) const{
  return NodeKey->leq(((NonPersistentRB_SpTreeItem*)item)->NodeKey);
}
int NonPersistentRB_SpTreeItem::eq(const BaseRB_SpTreeItem* item) const{
  return NodeKey->eq(((NonPersistentRB_SpTreeItem*)item)->NodeKey);
}
int NonPersistentRB_SpTreeItem::geq(const BaseRB_SpTreeItem* item) const{
  return NodeKey->geq(((NonPersistentRB_SpTreeItem*)item)->NodeKey);
}
int NonPersistentRB_SpTreeItem::gt(const BaseRB_SpTreeItem* item) const{
  return NodeKey->gt(((NonPersistentRB_SpTreeItem*)item)->NodeKey);
}


BaseRB_SpTreeItem *NonPersistentRB_SpTreeItem::GetParent()
{return Parent;}

BaseRB_SpTreeItem *NonPersistentRB_SpTreeItem::GetLeftChild()
{return LeftChild;}

BaseRB_SpTreeItem *NonPersistentRB_SpTreeItem::GetRightChild()
{return RightChild;}

RBSpKey * NonPersistentRB_SpTreeItem::GetNodeKey(){return NodeKey;}

void NonPersistentRB_SpTreeItem::SetParent(BaseRB_SpTreeItem* t)
{ Parent = t;}

void NonPersistentRB_SpTreeItem::SetLeftChild(BaseRB_SpTreeItem* t){
  LeftChild = t;
}

void NonPersistentRB_SpTreeItem::SetRightChild(BaseRB_SpTreeItem* t){
  RightChild = t;
}

void NonPersistentRB_SpTreeItem::SetNodeKey(RBSpKey* k){NodeKey = k;}

void* NonPersistentRB_SpTreeItem::GetElement() const {return Element;}
void NonPersistentRB_SpTreeItem::SetElement(void* e)  {Element = e;}

double NonPersistentRB_SpTreeItem::DoubleVal(){
  return (double)*(RBSpDoubleKey*)NodeKey;
}
























/*
** Insert an element into the binary tree
*/

void RB_SpBinaryTree::InternalInsert(void *element, RBSpKey * key, int tieBreaker) {
  BaseRB_SpTreeItem *node;
  BaseRB_SpTreeItem *left;
  BaseRB_SpTreeItem *right;
  BaseRB_SpTreeItem *item;

  if (element == NULL) {
    cout << "Warning, tring to insert a NULL element into a RB_SpBinaryTree"
	 << endl;
    return;
  }

  if ((KeyMode != key->GetMode()) || (KeyMode == NO_KEY_MODE)){
    KeyMode = key->GetMode();
  }

  item = NewTreeItem();
  item->Init();
  item->SetElement(element);
  item->SetNodeKey(Clone(key));

  if (tieBreaker == -1) {
    item->SetTieBreaker(TieBreaker);
    LastTieBreaker = TieBreaker;
    TieBreaker++;
  }else{
    item->SetTieBreaker(tieBreaker);
    LastTieBreaker = tieBreaker;
  }

  CurrentNode = item;

  if (FirstNode) {
    if (*item < *FirstNode) FirstNode = item;
  }else{
    FirstNode = item;
  }

  if (LastNode) {
    if (*item >= *LastNode) LastNode = item;
  }else{
    LastNode = item;
  }

  NumElements++;

  if (Root == NULL) {
    Root = item;
    return;
  }

  node = Root;

//  int InsertCounter = 0;

  while (1) {

//    InsertCounter++;

    if (Mode == BALANCED_TREE_MODE) {
      node = RoughBalance(node);
    }

    if (*item < *node) {
      node->SetNleft(node->GetNleft() + 1);
      left = node->GetLeftChild();
      if (left) {
	node = left;
      }else{
	node->SetLeftChild(item);
	item->SetParent(node);

	if (Mode == SPLAY_TREE_MODE) {
	  Splay(item);
	}

	break;
      }
    }else{
      node->SetNright(node->GetNright() + 1);
      right = node->GetRightChild();
      if (right) {
	node = right;
      }else{
	node->SetRightChild(item);
	item->SetParent(node);

	if (Mode == SPLAY_TREE_MODE) {
	  Splay(item);
	}

	break;
      }
    }

  }

//  cout << "InsertCounter = " << InsertCounter << endl;

}

/*
** Find an element
*/

void *RB_SpBinaryTree::Find(const RBSpKey * key) {
  #ifdef __INSURE__
  // Without this definition of a ostream object, there will be a
  // crash when using INSURE++.  This must be a bug in INSURE++.  The
  // ifdef prevents this code from slowing down the program when you
  // are NOT compiling with INSURE++.  Note that I only define this
  // variable, but do not use it, since defining it was sufficient to
  // stop INSURE++ from crashing.
  ostrstream os;
  #endif

  BaseRB_SpTreeItem *node;
  BaseRB_SpTreeItem *lChildFind;
  void* rc = NULL;

  if ((KeyMode != key->GetMode()) && (KeyMode != NO_KEY_MODE)){
    cout<<"Trying to do a find with the wrong key type"<<endl;
  }
  else{
    
    node = Find(Root, key);
    if (node) {
      
      if ((lChildFind = Find(node->GetLeftChild(), key))){
	node = lChildFind;
      }
      if (Mode == SPLAY_TREE_MODE) {
	Splay(node);
      }
      CurrentNode = node;
      rc = node->GetElement();
    }
  }
  return rc;
}


BaseRB_SpTreeItem *RB_SpBinaryTree::Find(BaseRB_SpTreeItem *subTree,
					 const RBSpKey* key) {
  BaseRB_SpTreeItem *node;
  BaseRB_SpTreeItem *lChildFind;
  BaseRB_SpTreeItem* rc = NULL;

  if (subTree != NULL){
    node = subTree;
    while (node) {
      
      if (Mode == BALANCED_TREE_MODE) {
	node = RoughBalance(node);
      }
      
      if (*(node->GetNodeKey()) == *key) {
	if ((lChildFind = Find(node->GetLeftChild(), key))){
	  node = lChildFind;
	}
	rc = node;
	break;
      }
      
      if (*key < *(node->GetNodeKey())) {
	node = node->GetLeftChild();
	if (node == NULL){
	  break;
	}
      }else{
	node = node->GetRightChild();
	if (node == NULL){
	  break;
	}
      }
    }
  }
  return rc;
}

/*
** Remove an element
*/

void RB_SpBinaryTree::InternalRemoveNode(BaseRB_SpTreeItem *node) {
  BaseRB_SpTreeItem *left;
  BaseRB_SpTreeItem *right;
  BaseRB_SpTreeItem *parent;

  UpdateFirstLastNode(node);

  if (CurrentNode == node) GetNextElement();
  
  if (node == Root) {

//	cout << "Removing root" << endl;

    if (node->GetLeftChild()) {
      Root = node->GetLeftChild();
      Root->SetParent(NULL);
      InsertSubTree(Root, node->GetRightChild());
    }else{
      if (node->GetRightChild()) {
	Root = node->GetRightChild();
	Root->SetParent(NULL);
      }else{
	Root = NULL;
      }
    }

  }else{

    parent = node->GetParent();
    left = node->GetLeftChild();
    right = node->GetRightChild();
    
    if (parent == NULL) {
      cout << "Bad parent..." << endl;
    }

    if (parent->GetLeftChild() == node) {
      parent->SetLeftChild(left);
      if (left) {
	parent->SetNleft( 1 + left->GetNleft() + left->GetNright());
	left->SetParent(parent);
      }else{
	parent->SetNleft(0);
      }
      InsertSubTree(parent, right);
    }else{
      if (parent->GetRightChild() != node) cout << "Bad children" << endl;
      parent->SetRightChild(right);
      if (right) {
	parent->SetNright(1 + right->GetNleft() + right->GetNright());
	right->SetParent(parent);
      }else{
	parent->SetNright(0);
      }	  
      InsertSubTree(parent, left);
    }
    
  }
  
  NumElements--;
  
  if (NumElements == 0) {
    FirstNode = NULL;
    LastNode = NULL;
  }
  
  DeleteTreeItem(node);
  
}

/*
** Check the integrity of the tree
*/

int RB_SpBinaryTree::CheckIntegrity() {
  void *element;
  int i;
  BaseRB_SpTreeItem *keyItem;
  int retVal = 1;
  int nl, nr;
  BaseRB_SpTreeItem *left;
  BaseRB_SpTreeItem *right;

  GetFirstElement();
  if (CurrentNode != FirstNode) {
    retVal = 0;
    cout << "Error, bad FirstNode" << endl;
  }

  GetLastElement();
  if (CurrentNode != LastNode) {
    retVal = 0;
    cout << "Error, bad LastNode" << endl;
  }

  element = GetFirstElement();
  keyItem = GetCurrentNode();
  i = 0;
  while (element) {

    nl = 0;
    left = CurrentNode->GetLeftChild();
    if (left) nl = left->GetNleft() + left->GetNright() + 1;
    if (CurrentNode->GetNleft() != nl) {
      cout << "Bad right/left" << endl;
      retVal = 0;
    }

    nr = 0;
    right = CurrentNode->GetRightChild();
    if (right) nr = right->GetNleft() + right->GetNright() + 1;
    if (CurrentNode->GetNright() != nr) {
      cout << "Bad right/left " << endl;
      retVal = 0;
    }

    if (*keyItem > *GetCurrentNode()) {
      cout << "Tree not sorted" << endl;
      retVal = 0;
    }
    keyItem = GetCurrentNode();
    element = GetNextElement();
    i++;
  }

  if (i != NumElements) {
    cout << "Bad integrity: wrong number of elements ("
	 << NumElements << ", " << i << ")" << endl;
    retVal = 0;
    return retVal;
  }else{
    return retVal;
  }

}

/*
** Get the first element
*/

void RB_SpBinaryTree::InsertSubTree(BaseRB_SpTreeItem *startingNode, BaseRB_SpTreeItem *subTree) {
  BaseRB_SpTreeItem *node;
  BaseRB_SpTreeItem *left;
  BaseRB_SpTreeItem *right;

  if (startingNode == NULL) {
    cout << "Error, starting node is NULL" << endl;
    return;
  }
  if (subTree == NULL) {
/*
    cout << "subTree = NULL: "
	 << "Nleft = " << startingNode->Nleft
	 << ", Nright = " << startingNode->Nright << endl;
*/
    CorrectNchildrenUp(startingNode);
    return;
  }

  node = startingNode;

  while (1) {
    if (*subTree < *node) {
      left = node->GetLeftChild();
      if (left) {
	node = left;
      }else{
	node->SetLeftChild(subTree);
	subTree->SetParent(node);
	node->SetNleft(1 + subTree->GetNleft() + subTree->GetNright());
	CorrectNchildrenUp(node);
/*
	if (Mode == SPLAY_TREE_MODE) {
	  Splay(node);
	}
*/
	break;
      }
    }else{
      right = node->GetRightChild();
      if (right) {
	node = right;
      }else{
	node->SetRightChild(subTree);
	subTree->SetParent(node);
	node->SetNright(1 + subTree->GetNleft() + subTree->GetNright());
	CorrectNchildrenUp(node);
/*
	if (Mode == SPLAY_TREE_MODE) {
	  Splay(node);
	}
*/
	break;

      }
    }
  }

}

/*
** Correct the GetNleft() and Nright values when adding a subTree to a node
*/
void RB_SpBinaryTree::CorrectNchildrenUp(BaseRB_SpTreeItem *it) {
  BaseRB_SpTreeItem *node = it;
  BaseRB_SpTreeItem *parent;

  while (node != Root) {

    parent = node->GetParent();

    if (parent->GetLeftChild() == node) {
      parent->SetNleft(1 + node->GetNleft() + node->GetNright());
    }else{
      parent->SetNright(1 + node->GetNleft() + node->GetNright());
    }

    node = parent;
  }

}

/*
** Get the first element
*/

void *RB_SpBinaryTree::GetFirstElement() {

  CurrentNode = Root;
  if (CurrentNode == NULL) return NULL;

  while (CurrentNode->GetLeftChild()) CurrentNode = CurrentNode->GetLeftChild();
  return CurrentNode->GetElement();

}

/*
** Get the last element
*/

void *RB_SpBinaryTree::GetLastElement() {

  CurrentNode = Root;
  if (CurrentNode == NULL) return NULL;

  while (CurrentNode->GetRightChild()) CurrentNode = CurrentNode->GetRightChild();
  return CurrentNode->GetElement();

}

/*
** Get the next element
*/

void *RB_SpBinaryTree::GetNextElement() {
  BaseRB_SpTreeItem *parent;

  if (CurrentNode == NULL) return NULL;

  if (CurrentNode == Root) {
    CurrentNode = CurrentNode->GetRightChild();
    if (CurrentNode == NULL) return NULL;
    while (CurrentNode->GetLeftChild()) CurrentNode = CurrentNode->GetLeftChild();
    return CurrentNode->GetElement();
  }

  if (CurrentNode->GetRightChild()) {
    CurrentNode = CurrentNode->GetRightChild();
    while (CurrentNode->GetLeftChild()) CurrentNode = CurrentNode->GetLeftChild();
    return CurrentNode->GetElement();
  }

  while (1) {

    if (CurrentNode == Root) {
      CurrentNode = NULL;
      return NULL;
    }

    parent = CurrentNode->GetParent();
    if (parent->GetLeftChild() == CurrentNode) {
      CurrentNode = parent;
      return CurrentNode->GetElement();
    }

    CurrentNode = parent;

  }

}

/*
** Get the previous element
*/

void *RB_SpBinaryTree::GetPreviousElement() {
  BaseRB_SpTreeItem *parent;

  if (CurrentNode == NULL) return NULL;

  if (CurrentNode == Root) {
    CurrentNode = CurrentNode->GetLeftChild();
    if (CurrentNode == NULL) return NULL;
    while (CurrentNode->GetRightChild()) CurrentNode = CurrentNode->GetRightChild();
    return CurrentNode->GetElement();
  }

  if (CurrentNode->GetLeftChild()) {
    CurrentNode = CurrentNode->GetLeftChild();
    while (CurrentNode->GetRightChild()) CurrentNode = CurrentNode->GetRightChild();
    return CurrentNode->GetElement();
  }

  while (1) {

    if (CurrentNode == Root) {
      CurrentNode = NULL;
      return NULL;
    }

    parent = CurrentNode->GetParent();
    if (parent->GetRightChild() == CurrentNode) {
      CurrentNode = parent;
      return CurrentNode->GetElement();
    }

    CurrentNode = parent;

  }

}

/*
** Rotate this node left
*/

BaseRB_SpTreeItem *RB_SpBinaryTree::RotateLeft(BaseRB_SpTreeItem *item) {
  BaseRB_SpTreeItem *X = item;
  BaseRB_SpTreeItem *Y;
  BaseRB_SpTreeItem *B;
  BaseRB_SpTreeItem *parent;

//  cout << "Rotating Left" << endl;

  if (X == NULL) {
    cout << "Error, X = NULL in RotateLeft" << endl;
    return NULL;
  }

  Y = X->GetRightChild();

  if (Y) {

    parent = X->GetParent();

    B = Y->GetLeftChild();

    Y->SetLeftChild(X);
    X->SetRightChild(B);

    if (B) {
      B->SetParent(X);
      X->SetNright(1 + B->GetNleft() + B->GetNright());
    }else{
      X->SetNright(0);
    }

    Y->SetNleft(1 + X->GetNleft() + X->GetNright());

    Y->SetParent(parent);

    if (item != Root) {
      if (parent->GetRightChild() == X) {
	parent->SetRightChild(Y);
      }else{
	parent->SetLeftChild(Y);
      }
    }else{
      Root = Y;
    }

    X->SetParent(Y);
    Y->SetLeftChild(X);

  }else{

    return X;

  }

  return Y;

}

/*
** Rotate this node right
*/

BaseRB_SpTreeItem *RB_SpBinaryTree::RotateRight(BaseRB_SpTreeItem *item) {
  BaseRB_SpTreeItem *Y = item;
  BaseRB_SpTreeItem *X;
  BaseRB_SpTreeItem *B;
  BaseRB_SpTreeItem *parent;

//  cout << "Rotating Right" << endl;

  if (Y == NULL) {
    cout << "Error, Y = NULL in RotateRight" << endl;
    return NULL;
  }

  X = Y->GetLeftChild();

  if (X) {

    parent = Y->GetParent();

    B = X->GetRightChild();

    X->SetRightChild(Y);
    Y->SetLeftChild(B);

    if (B) {
      B->SetParent(Y);
      Y->SetNleft(1 + B->GetNleft() + B->GetNright());
    }else{
      Y->SetNleft(0);
    }

    X->SetNright(1 + Y->GetNleft() + Y->GetNright());

    X->SetParent(parent);

    if (item != Root) {
      if (parent->GetLeftChild() == Y) {
	parent->SetLeftChild(X);
      }else{
	parent->SetRightChild(X);
      }
    }else{
      Root = X;
    }

    Y->SetParent(X);
    X->SetRightChild(Y);

  }else{

    return Y;

  }

  return X;

}

/*
** Perform the splay operation
*/

void RB_SpBinaryTree::Splay(BaseRB_SpTreeItem *node) {
  BaseRB_SpTreeItem *parent;
  BaseRB_SpTreeItem *grandParent;
  int nSplay = 0;
  int parentFlag;
  int grandParentFlag;

  while(node != Root) {

    parent = node->GetParent();

    if (parent == Root) {

      if (parent->GetLeftChild() == node) {
        RotateRight(parent);
      }else{
        RotateLeft(parent);
      }

      break;

    }else{

      if (parent->GetLeftChild() == node) {
        parentFlag = 0;
      }else{
        parentFlag = 1;
      }

      grandParent = parent->GetParent();
      if (grandParent->GetLeftChild() == parent) {
        grandParentFlag = 0;
      }else{
        grandParentFlag = 1;
      }

      if ((parentFlag + grandParentFlag) == 1) {

//...... Zig Zag

        if (parentFlag == 0) {
          RotateRight(parent);
          RotateLeft(grandParent);
        }else{
          RotateLeft(parent);
          RotateRight(grandParent);
        }

      }else{

//...... Zig Zig

        if (parentFlag == 0) {
          RotateRight(grandParent);
          RotateRight(parent);
        }else{
          RotateLeft(grandParent);
          RotateLeft(parent);
        }

      }

    }

    if (nSplay >= RB_N_SPLAY) break;
    nSplay++;

  }

}

int RB_SpBinaryTree::TreeItemID = -1;
int RB_SpBinaryTree::SpStringKeyID = -1;
int RB_SpBinaryTree::SpDoubleKeyId = -1;


const static int FREE_LIST_SIZE = 100;

void RB_SpBinaryTree::AllocateMoreTreeItems()
{
  extern int PersistenceEnabled;
  if (GlobalSpPoMempool == NULL){
    GlobalSpPoMempool = new SpPoMempool;
  }
  if (PersistenceEnabled){
    void* jj = PO_NEW_ARRAY_PersistentRB_SpTreeItem(FREE_LIST_SIZE);
    TreeItemFreeId =
      GlobalSpPoMempool->AddChunk(jj,
				  TreeItemFreeId,
				  FREE_LIST_SIZE,
				  sizeof(PersistentRB_SpTreeItem));
  }
  else{
    void* jj =new NonPersistentRB_SpTreeItem[FREE_LIST_SIZE];
    TreeItemFreeId =
      GlobalSpPoMempool->AddChunk(jj,
				  TreeItemFreeId,
				  FREE_LIST_SIZE,
				  sizeof(NonPersistentRB_SpTreeItem));
  }
}



void RB_SpBinaryTree::AllocateMoreSpStringKeys()
{
  extern int PersistenceEnabled;
  int StringKeyId = BaseRB_SpTreeItem::GetStringKeyFreeId();
  if (GlobalSpPoMempool == NULL){
    GlobalSpPoMempool = new SpPoMempool;
  }
  if (PersistenceEnabled){
    void* jj = PO_NEW_ARRAY_RBSpPersistentStringKey(FREE_LIST_SIZE);
    StringKeyId =
      GlobalSpPoMempool->AddChunk(jj,
				  StringKeyId,
				  FREE_LIST_SIZE,
				  sizeof(RBSpStringKey));
  }
  else{
    void* jj = new RBSpStringKey[FREE_LIST_SIZE];
    StringKeyId = GlobalSpPoMempool->AddChunk(jj,
					      StringKeyId,
					      FREE_LIST_SIZE,
					      sizeof(RBSpStringKey));
  }
  BaseRB_SpTreeItem::SetStringKeyFreeId(StringKeyId);
}



void RB_SpBinaryTree::AllocateMoreSpDoubleKeys()
{
  extern int PersistenceEnabled;
  int DoubleKeyId = BaseRB_SpTreeItem::GetDoubleKeyFreeId();
  if (GlobalSpPoMempool == NULL){
    GlobalSpPoMempool = new SpPoMempool;
  }
  if (PersistenceEnabled){
    void* jj = PO_NEW_ARRAY_RBSpDoubleKey(FREE_LIST_SIZE);
    DoubleKeyId =
      GlobalSpPoMempool->AddChunk(jj,
				  DoubleKeyId,
				  FREE_LIST_SIZE,
				  sizeof(RBSpDoubleKey));
  }
  else{
    void* jj = new RBSpDoubleKey[FREE_LIST_SIZE];
    DoubleKeyId = GlobalSpPoMempool->AddChunk(jj,
					      DoubleKeyId,
					      FREE_LIST_SIZE,
					      sizeof(RBSpDoubleKey));
  }
  BaseRB_SpTreeItem::SetDoubleKeyFreeId(DoubleKeyId);
}



RBSpKey* RB_SpBinaryTree::Clone(RBSpKey* key){
  RBSpKey* rc = NULL;
  RBSpStringKey* tempStringRc;
  RBSpDoubleKey* tempDoubleRc;
  switch (key->GetMode()){
  case DOUBLE_KEY_MODE:
    tempDoubleRc = junkTreeInstance->NewSpDoubleKey();
    *tempDoubleRc = *(RBSpDoubleKey*)key;
    rc = tempDoubleRc;
    break;
  case STRING_KEY_MODE:
    tempStringRc = junkTreeInstance->NewSpStringKey();
    *tempStringRc = *(RBSpStringKey*)key;
    rc = tempStringRc;
  }
  return rc;
}










void BaseRB_SpTreeItem::ARGH(int i)const
{
  cout<<"AAAAAARRRRRRGGGGGGHHHHH SHOULD NOT HAPPEN!!!!-->"<<i<<endl;
}

void BaseRB_SpTreeItem::Cleanup(){ARGH(7);}


int BaseRB_SpTreeItem::lt(const BaseRB_SpTreeItem* )const{ARGH(1);return 0;}
int BaseRB_SpTreeItem::leq(const BaseRB_SpTreeItem* )const{ARGH(2);return 0;}
int BaseRB_SpTreeItem::eq(const BaseRB_SpTreeItem* )const{ARGH(3);return 0;}
int BaseRB_SpTreeItem::geq(const BaseRB_SpTreeItem* )const{ARGH(5);return 0;}
int BaseRB_SpTreeItem::gt(const BaseRB_SpTreeItem* )const{ARGH(6);return 0;}


BaseRB_SpTreeItem *BaseRB_SpTreeItem::GetParent(){ARGH(8);return NULL;};
BaseRB_SpTreeItem *BaseRB_SpTreeItem::GetLeftChild(){ARGH(9);return NULL;};
BaseRB_SpTreeItem *BaseRB_SpTreeItem::GetRightChild(){ARGH(10);return NULL;};
RBSpKey * BaseRB_SpTreeItem::GetNodeKey(){ARGH(11);return NULL;};
void BaseRB_SpTreeItem::SetParent(BaseRB_SpTreeItem*){ARGH(12);};
void BaseRB_SpTreeItem::SetLeftChild(BaseRB_SpTreeItem*){ARGH(13);};
void BaseRB_SpTreeItem::SetRightChild(BaseRB_SpTreeItem*){ARGH(14);};
void BaseRB_SpTreeItem::SetNodeKey(RBSpKey*){ARGH(15);};
void* BaseRB_SpTreeItem::GetElement() const {ARGH(16); return NULL;}
void BaseRB_SpTreeItem::SetElement(void* ) {ARGH(17);}

double BaseRB_SpTreeItem::DoubleVal(){ARGH(18);return -17171717171717.17;}




int BaseRB_SpTreeItem::DoubleKeyFreeId = -1;
int BaseRB_SpTreeItem::StringKeyFreeId = -1;
int RB_SpBinaryTree::TreeItemFreeId = -1;
#endif


void RB_Register_RB_SpBinaryTree()
{
  RB_REGISTER_CLASS_RB_SpBinaryTree();
}

