#ifndef RB_SpBinaryTree_Header
#define RB_SpBinaryTree_Header

#include "RB_SpDefineClass.H"

#include "SpAltMgr.H"
#include "SpAltTreeChange.H"

//#define OLD_FUNCTIONALITY
#ifdef OLD_FUNCTIONALITY

#include "SpBinaryTree.H"


/**********************************************************************
 * 
 **********************************************************************/
class RB_SpBinaryTree : public SpBinaryTree {
  friend class RB_SpPriorityTree;
public:
  void Insert( void *item, SpKey* key );
  void* Remove( SpKey* key );
  void Insert( void* item, double keyNum );
  void* Remove( double keyNum );
  void Insert( void* item, char* keyName );
  void* Remove( char *keyName );
  void* RemoveFirstElement();
  void* RemoveFirstElement( double& keyNum );

  int GetNumElements(){return SpBinaryTree::GetNumElements();}
  int CheckIntegrity(){return SpBinaryTree::CheckIntegrity();}
  void SetNormalTreeMode(){ SpBinaryTree::SetNormalTreeMode();}
  void SetSplayTreeMode(){ SpBinaryTree::SetSplayTreeMode();}
  void SetBalancedTreeMode(){ SpBinaryTree::SetBalancedTreeMode();}

  int IsNormalTreeMode(){return SpBinaryTree::IsNormalTreeMode();}
  int IsSplayTreeMode(){return SpBinaryTree::IsSplayTreeMode();}
  int IsBalancedTreeMode(){return SpBinaryTree::IsBalancedTreeMode();}

  void * Find(const SpKey* key){return SpBinaryTree::Find(key);}
  void * Find(const SpKey& key){return SpBinaryTree::Find(key);}
  void * Find(const double key){return SpBinaryTree::Find(key);}
  void * Find(const char* key){return SpBinaryTree::Find(key);}

  void * GetFirstElement(){return SpBinaryTree::GetFirstElement();}
  void * GetCurrentElement(){return SpBinaryTree::GetCurrentElement();}
  void * GetLastElement(){return SpBinaryTree::GetLastElement();}

  void * GetNextElement(){return SpBinaryTree::GetNextElement();}
  void * GetPreviousElement(){return SpBinaryTree::GetPreviousElement();}

  void * operator ++(){return SpBinaryTree::operator++();}
  void * operator ++(int){return SpBinaryTree::operator++(1);}

  void * operator --(){return SpBinaryTree::operator--();}
  void * operator --(int){return SpBinaryTree::operator--(1);}
  double GetCurrentKey() { return SpBinaryTree::GetCurrentKey();}
  char* GetCurrentKeyName() { return SpBinaryTree::GetCurrentKeyName();}
  static SpKey* Clone(SpKey*k ){return k->Clone();}

protected:
  void Insert( void* item, double keyNum, char* keyName );
  void* Remove( double keyNum, char* keyName );
  void RemoveNode( TREE_ITEM* node );
};

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_SpBinaryTree::Insert( void* item, SpKey* key ) {
  SpAltTreeChange* altItem;

  SpBinaryTree::Insert( item, key );

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltTreeChange();
    altItem->Alter( this, item, key, GetLastTieBreaker(), 1 );
    SpCurrentAltMgr->Insert( altItem  );
  }
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpBinaryTree::Remove( SpKey* key ) {
  SpAltTreeChange* altItem;

  void* item = SpBinaryTree::Remove( key );

  if( item && SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltTreeChange();
    altItem->Alter( this, item, key, GetLastTieBreaker(), 0 );
    SpCurrentAltMgr->Insert( altItem  );
  }

  return item;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_SpBinaryTree::Insert( void* item, double keyNum, char* keyName ) {
  SpKey* key;

  if( keyName ){
    key = new SpStringKey( keyName );
  }
  else{
    key = new SpDoubleKey( keyNum );
  }

  Insert( item, key );
  delete key;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpBinaryTree::Remove( double keyNum, char* keyName ) {
  SpKey* key;

  if( keyName ){
    key = new SpStringKey( keyName );
  }
  else{
    key = new SpDoubleKey( keyNum );
  }

  void* item = Remove( key );
  delete key;

  return item;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_SpBinaryTree::Insert( void* item, double keyNum ) {
  Insert( item, keyNum, NULL );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpBinaryTree::Remove( double keyNum ) {
  return Remove( keyNum, NULL );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_SpBinaryTree::Insert( void* item, char* keyName ) {
  Insert( item, -1, keyName );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpBinaryTree::Remove( char *keyName ) {
  return Remove( -1, keyName );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpBinaryTree::RemoveFirstElement() {
  void* item = NULL;

  if( FirstNode ) {
    SpKey* key = ( FirstNode->NodeKey )->Clone();
    item = Remove( key );
    delete key;
  }

  return item;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpBinaryTree::RemoveFirstElement( double& keyNum ) {
  void* item = NULL;

  keyNum = 1.0e20;
  if( FirstNode ) {
    keyNum = * ( SpDoubleKey* ) FirstNode->NodeKey;
    item = Remove( keyNum );
  }

  return item;
}

#else //#ifdef OLD_FUNCTIONALITY

#include "RB_SpDefineClass.H"
#include "SpHolderItem.H"
#include "RBSpKey.H"
#include "SpPersistenceBaseClass.H"
#include "SpTreeMode.H"

extern int PersistenceEnabled;

extern const double RB_TREE_BALANCE_RATIO;
extern const int RB_TREE_CUTOFF;
extern const int RB_N_SPLAY;

class BaseRB_SpTreeItem;
PO_DEFINE_SMART_POINTERS_NO_ADDITIONAL_OPERATORS(BaseRB_SpTreeItem)

class BaseRB_SpTreeItem : private SpPersistenceBaseClass{

private:
  void ARGH(int i)const;

  int Nleft;
  int Nright;
  int TieBreaker;

protected:
  static int DoubleKeyFreeId;
  static int StringKeyFreeId;

public:
  BaseRB_SpTreeItem() {Init();}

  virtual void Init() {
    Nleft = 0;
    Nright = 0;
    TieBreaker = 0;
  }

  virtual void Cleanup();

  int operator == (const BaseRB_SpTreeItem &item) const;
  int operator < (const BaseRB_SpTreeItem &item) const;
  int operator <= (const BaseRB_SpTreeItem &item) const;
  int operator > (const BaseRB_SpTreeItem &item) const;
  int operator >= (const BaseRB_SpTreeItem &item) const;

  virtual int lt(const BaseRB_SpTreeItem* item) const;
  virtual int leq(const BaseRB_SpTreeItem* item) const;
  virtual int eq(const BaseRB_SpTreeItem* item) const;
  virtual int geq(const BaseRB_SpTreeItem* item) const;
  virtual int gt(const BaseRB_SpTreeItem* item) const;


  virtual BaseRB_SpTreeItem *GetParent();
  virtual BaseRB_SpTreeItem *GetLeftChild();
  virtual BaseRB_SpTreeItem *GetRightChild();
  virtual RBSpKey * GetNodeKey();
  virtual void SetParent(BaseRB_SpTreeItem*);
  virtual void SetLeftChild(BaseRB_SpTreeItem*);
  virtual void SetRightChild(BaseRB_SpTreeItem*);
  virtual void SetNodeKey(RBSpKey*);
  virtual void* GetElement() const;
  virtual void SetElement(void* e);
  virtual double DoubleVal();

  void SetTieBreaker(int t);
  void SetNleft(int nl);
  void SetNright(int nr);
  int GetTieBreaker();
  int GetNleft(){return Nleft;}
  int GetNright(){return Nright;}

  static int GetDoubleKeyFreeId(){return DoubleKeyFreeId;}
  static int GetStringKeyFreeId(){return StringKeyFreeId;}
  static void SetDoubleKeyFreeId(int id){DoubleKeyFreeId = id;}
  static void SetStringKeyFreeId(int id){StringKeyFreeId = id;}
};

class PersistentRB_SpTreeItem;
PO_DEFINE_SMART_POINTERS_NO_ADDITIONAL_OPERATORS(PersistentRB_SpTreeItem);
PO_DEFINE_CLASS(RBSpKey);

#ifdef __IBMCPP__
namespace {
#endif

#ifndef PO_DEFINED_CHAR
RB_DEFINE_CLASS_PRIMATIVE(char);
#define PO_DEFINED_CHAR
#endif

#ifdef __IBMCPP__
}
#endif

class PersistentRB_SpTreeItem : public BaseRB_SpTreeItem {

private:

  PO_PTR_char Element;

  PO_PTR_PersistentRB_SpTreeItem Parent;
  PO_PTR_PersistentRB_SpTreeItem LeftChild;
  PO_PTR_PersistentRB_SpTreeItem RightChild;
  PO_PTR_RBSpKey NodeKey;

public:
  PersistentRB_SpTreeItem() {Init();}

  void Init() {
    Element = (char*)NULL;
    Parent = (PersistentRB_SpTreeItem*)NULL;
    LeftChild = (PersistentRB_SpTreeItem*)NULL;
    RightChild = (PersistentRB_SpTreeItem*)NULL;
    NodeKey = (RBSpKey*)NULL;
    BaseRB_SpTreeItem::Init();
  }

  virtual void Cleanup();
  virtual int lt(const BaseRB_SpTreeItem* item) const;
  virtual int leq(const BaseRB_SpTreeItem* item) const;
  virtual int eq(const BaseRB_SpTreeItem* item) const;
  virtual int geq(const BaseRB_SpTreeItem* item) const;
  virtual int gt(const BaseRB_SpTreeItem* item) const;
  virtual BaseRB_SpTreeItem *GetParent();

  BaseRB_SpTreeItem *GetLeftChild();
  BaseRB_SpTreeItem *GetRightChild();
  RBSpKey * GetNodeKey();
  void SetParent(BaseRB_SpTreeItem* t);
  void SetLeftChild(BaseRB_SpTreeItem*t);
  void SetRightChild(BaseRB_SpTreeItem*t);
  void SetNodeKey(RBSpKey* k);
  void* GetElement() const;
  void SetElement(void* e);
  double DoubleVal();
};

PO_DEFINE_CLASS_NO_SMART_POINTERS(PersistentRB_SpTreeItem);

class NonPersistentRB_SpTreeItem : public BaseRB_SpTreeItem {

private:

  void *Element;

  BaseRB_SpTreeItem* Parent;
  BaseRB_SpTreeItem* LeftChild;
  BaseRB_SpTreeItem* RightChild;
  RBSpKey *NodeKey;

public:
  NonPersistentRB_SpTreeItem() {Init();}

  void Init() {
    Element = NULL;
    Parent = NULL;
    LeftChild = NULL;
    RightChild = NULL;
    NodeKey = NULL;
    BaseRB_SpTreeItem::Init();
  }

  virtual void Cleanup();

  virtual int lt(const BaseRB_SpTreeItem* item) const;
  virtual int leq(const BaseRB_SpTreeItem* item) const;
  virtual int eq(const BaseRB_SpTreeItem* item) const;
  virtual int geq(const BaseRB_SpTreeItem* item) const;
  virtual int gt(const BaseRB_SpTreeItem* item) const;

  BaseRB_SpTreeItem *GetParent();
  BaseRB_SpTreeItem *GetLeftChild();
  BaseRB_SpTreeItem *GetRightChild();
  RBSpKey * GetNodeKey();
  void SetParent(BaseRB_SpTreeItem* t);
  void SetLeftChild(BaseRB_SpTreeItem* t);
  void SetRightChild(BaseRB_SpTreeItem* t);
  void SetNodeKey(RBSpKey* k);
  void* GetElement() const;
  void SetElement(void* e);
  double DoubleVal();
};


extern SpPoMempool* GlobalSpPoMempool;


class RB_SpBinaryTree {
  friend class SpAltTreeChange;
public:
  RB_SpBinaryTree();
  int GetLastTieBreaker() {return LastTieBreaker;}
  void Insert( void *item, RBSpKey* key );
  void* Remove( RBSpKey* key );
  void Insert( void* item, double keyNum );
  void* Remove( double keyNum );
  void Insert( void* item, char* keyName );
  void* Remove( char *keyName );
  void* RemoveFirstElement();
  void* RemoveFirstElement( double& keyNum );

  int GetNumElements(){return NumElements;}
  int CheckIntegrity();
  void SetNormalTreeMode(){Mode = NORMAL_TREE_MODE;}
  void SetSplayTreeMode(){Mode = BALANCED_TREE_MODE;}
  void SetBalancedTreeMode(){Mode = SPLAY_TREE_MODE;}

  int IsNormalTreeMode(){return (Mode == NORMAL_TREE_MODE);}
  int IsSplayTreeMode(){return (Mode == BALANCED_TREE_MODE);}
  int IsBalancedTreeMode(){return (Mode == SPLAY_TREE_MODE);}

  void * Find(const RBSpKey* key);
  void * Find(const RBSpKey& key) {return Find(&key);}
  void *Find(const double key) 
    {
      void * rc = NULL;
      if (KeyMode == DOUBLE_KEY_MODE){
	RBSpDoubleKey newKey(key);
	rc = Find(&newKey);
      }
      else if (KeyMode != NO_KEY_MODE) {
	cout<<"Tried to find a double key when not in double key mode."<<endl;
      }
      return rc;
    }
  void *Find(const char * keyName) 
    {
      void * rc = NULL;
      if (KeyMode == STRING_KEY_MODE){
	RBSpStringKey newKey(keyName);
	rc = Find(&newKey);
      }
      else if (KeyMode != NO_KEY_MODE) {
	cout<<"Tried to find a string key when not in string key mode."<<endl;
      }
      return rc;
    }

  void * GetFirstElement();
  void * GetCurrentElement() {
    void *e=NULL;
    if (CurrentNode) e = CurrentNode->GetElement();
    return e;
  }
  void * GetLastElement();

  void * GetNextElement();
  void * GetPreviousElement();

  void * operator ++(){GetCurrentElement(); return GetNextElement();}
  void * operator ++(int){
    void* rc = GetCurrentElement();
    GetNextElement();
    return rc;
  }
  void * operator --(){GetCurrentElement(); return GetPreviousElement();}
  void * operator --(int){
    void* rc = GetCurrentElement();
    GetPreviousElement();
    return rc;
  }
  double GetCurrentKey();
  char* GetCurrentKeyName();
  BaseRB_SpTreeItem* GetFirstNode(){return FirstNode;}
  BaseRB_SpTreeItem* GetLastNode(){return LastNode;}
  BaseRB_SpTreeItem* GetCurrentNode(){ return CurrentNode;}
  static RBSpKey* Clone(RBSpKey*);

  void InternalInsert(RBSpKey*);
  void InternalRemove(RBSpKey* key){
    if (KeyMode == key->GetMode()){
      void *element = Find(key);
      if (element) {
	InternalRemoveNode(CurrentNode);
      }
    }
    else{
      if (KeyMode != NO_KEY_MODE){
	cout<<"Tried to remove when in the wrong mode"<<endl;
      }
    }
    if (NumElements == 0){
      KeyMode = NO_KEY_MODE;
    }
  }



protected:
  void Insert( void* item, double keyNum, char* keyName );
  void* Remove( double keyNum, char* keyName );
  BaseRB_SpTreeItem *Find (BaseRB_SpTreeItem* subTree,
			   const RBSpKey* key);
  

  KEY_MODE KeyMode;
  
  BaseRB_SpTreeItem *Root;
  BaseRB_SpTreeItem *CurrentNode;
  BaseRB_SpTreeItem *FirstNode;
  BaseRB_SpTreeItem *LastNode;
  int NumElements;
  int Mode;
  int TieBreaker;
  int LastTieBreaker;

  void InternalInsert(void *element, RBSpKey * key, int tieBreaker);
  void InternalRemoveNode(BaseRB_SpTreeItem *node);
  
  BaseRB_SpTreeItem* NewTreeItem();
  void DeleteTreeItem(BaseRB_SpTreeItem*);
  void AllocateMoreTreeItems();
  
  RBSpStringKey* NewSpStringKey();
  void DeleteSpStringKey(RBSpStringKey*);
  void AllocateMoreSpStringKeys();
  
  RBSpDoubleKey* NewSpDoubleKey();
  void DeleteSpDoubleKey(RBSpDoubleKey*);
  void AllocateMoreSpDoubleKeys();
  static int TreeItemFreeId;


  BaseRB_SpTreeItem *RotateLeft(BaseRB_SpTreeItem *it);
  BaseRB_SpTreeItem *RotateRight(BaseRB_SpTreeItem *it);
  void Splay(BaseRB_SpTreeItem *node);

  int GetFreeId(RBSpKey*);
  
  BaseRB_SpTreeItem *RoughBalance(BaseRB_SpTreeItem *it) {
    BaseRB_SpTreeItem *node = it;
    int Nleft = node->GetNleft();
    int Nright = node->GetNright();
    if (Nright+Nleft > RB_TREE_CUTOFF) {
      if (Nright < Nleft) {
	if (Nleft/(Nright+1) > RB_TREE_BALANCE_RATIO) node = RotateRight(node);
      }else{
	if (Nright/(Nleft+1) > RB_TREE_BALANCE_RATIO) node = RotateLeft(node);
      }
    }
    return node;
  }
  void CorrectNchildrenUp(BaseRB_SpTreeItem *it);
  
  void InsertSubTree(BaseRB_SpTreeItem *startingNode, BaseRB_SpTreeItem *subTree);

  void UpdateFirstLastNode(BaseRB_SpTreeItem *node) {
    if (NumElements == 0) {
      FirstNode = NULL;
      LastNode = NULL;
    }
    if (node == FirstNode) {
      CurrentNode = FirstNode;
      GetNextElement();
      FirstNode = CurrentNode;
    }
    if (node == LastNode) {
      CurrentNode = LastNode;
      GetPreviousElement();
      LastNode = CurrentNode;
    }
  }
  
  static int TreeItemID;
  static int SpStringKeyID;
  static int SpDoubleKeyId;
};

inline BaseRB_SpTreeItem* RB_SpBinaryTree::NewTreeItem(){
  BaseRB_SpTreeItem* rc =
    (BaseRB_SpTreeItem*)GlobalSpPoMempool->RemoveItem(TreeItemFreeId);
  if (rc == NULL){
    AllocateMoreTreeItems();
    rc = (BaseRB_SpTreeItem*)GlobalSpPoMempool->RemoveItem(TreeItemFreeId);
  }
  return rc;
}

inline void RB_SpBinaryTree::DeleteTreeItem(BaseRB_SpTreeItem *item){
  GlobalSpPoMempool->ReturnItem(TreeItemFreeId, item);
}


inline RBSpStringKey* RB_SpBinaryTree::NewSpStringKey(){
  RBSpStringKey* rc =
    (RBSpStringKey*)GlobalSpPoMempool->
    RemoveItem(BaseRB_SpTreeItem::GetStringKeyFreeId());
  if (rc == NULL){
    AllocateMoreSpStringKeys();
    rc = (RBSpStringKey*)GlobalSpPoMempool->
      RemoveItem(BaseRB_SpTreeItem::GetStringKeyFreeId());
  }
  return rc;
}

inline void RB_SpBinaryTree::DeleteSpStringKey(RBSpStringKey *item){
  GlobalSpPoMempool->ReturnItem(BaseRB_SpTreeItem::GetStringKeyFreeId(), item);
}


inline RBSpDoubleKey* RB_SpBinaryTree::NewSpDoubleKey(){
  RBSpDoubleKey* rc =
    (RBSpDoubleKey*)GlobalSpPoMempool->
    RemoveItem(BaseRB_SpTreeItem::GetDoubleKeyFreeId());
  if (rc == NULL){
    AllocateMoreSpDoubleKeys();
    rc = (RBSpDoubleKey*)GlobalSpPoMempool->
      RemoveItem(BaseRB_SpTreeItem::GetDoubleKeyFreeId());
  }
  return rc;
}

inline void RB_SpBinaryTree::DeleteSpDoubleKey(RBSpDoubleKey *item){
  GlobalSpPoMempool->ReturnItem(BaseRB_SpTreeItem::GetDoubleKeyFreeId(), item);
}




inline RB_SpBinaryTree::RB_SpBinaryTree():
  KeyMode(NO_KEY_MODE),
  Root(NULL),
  CurrentNode(NULL),
  FirstNode(NULL),
  LastNode(NULL),
  NumElements(0),
  Mode(NORMAL_TREE_MODE),
  TieBreaker(0),
  LastTieBreaker(0)
{
  if (DefaultDataBase == NULL){
    DefaultDataBase = new SpPoDataBase;
  }
}

inline double RB_SpBinaryTree::GetCurrentKey()
{
  double retVal = -1;
  if (CurrentNode){
    if (KeyMode == DOUBLE_KEY_MODE){
      retVal = *((RBSpDoubleKey*)CurrentNode->GetNodeKey());
    }
    else{
      cout<<"Called GetCurrentKey when not in double key mode"<<endl;
    }
  }
  return retVal;
}
  
inline char* RB_SpBinaryTree::GetCurrentKeyName()
{
  char *retVal = NULL;
  if (CurrentNode){
    if (KeyMode == STRING_KEY_MODE){
      retVal = (char *)(*((RBSpStringKey*)(CurrentNode->GetNodeKey())));
    }
    else{
      cout<<"Not in string key mode"<<endl;
    }
  }
  return retVal;
}

inline
void
RB_SpBinaryTree::Insert( void* item, RBSpKey* key ) {
  SpAltTreeChange* altItem;

  InternalInsert( item, key,  GetLastTieBreaker());

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltTreeChange();
    altItem->Alter( this, item, key, GetLastTieBreaker(), 1 );
    SpCurrentAltMgr->Insert( altItem  );
  }
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpBinaryTree::Remove( RBSpKey* key ) {
  SpAltTreeChange* altItem;

  void* item = Find(key);
  if (item){
    InternalRemoveNode( CurrentNode);
    if(SpCurrentAltMgr->GetOptimistic() ) {
      altItem = ALLOCATE_SpAltTreeChange();
      altItem->Alter( this, item, key, GetLastTieBreaker(), 0 );
      SpCurrentAltMgr->Insert( altItem  );
    }
  }

  return item;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_SpBinaryTree::Insert( void* item, double keyNum, char* keyName ) {
  RBSpKey* key;

  if( keyName ){
    key = new RBSpStringKey( keyName );
  }
  else{
    key = new RBSpDoubleKey( keyNum );
  }

  Insert( item, key );
  delete key;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpBinaryTree::Remove( double keyNum, char* keyName ) {
  RBSpKey* key;

  if( keyName ){
    key = new RBSpStringKey( keyName );
  }
  else{
    key = new RBSpDoubleKey( keyNum );
  }

  void* item = Remove( key );
  delete key;

  return item;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_SpBinaryTree::Insert( void* item, double keyNum ) {
  Insert( item, keyNum, NULL );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpBinaryTree::Remove( double keyNum ) {
  return Remove( keyNum, NULL );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_SpBinaryTree::Insert( void* item, char* keyName ) {
  Insert( item, -1, keyName );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpBinaryTree::Remove( char *keyName ) {
  return Remove( -1, keyName );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpBinaryTree::RemoveFirstElement() {
  void* item = NULL;

  if( FirstNode ) {
    RBSpKey* key = Clone(FirstNode->GetNodeKey());
    item = Remove( key );
    switch (key->GetMode()){
    case DOUBLE_KEY_MODE:
      DeleteSpDoubleKey((RBSpDoubleKey*)key);
      break;
    case STRING_KEY_MODE:
      DeleteSpStringKey((RBSpStringKey*)key);
      break;
    }
  }

  return item;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpBinaryTree::RemoveFirstElement( double& keyNum ) {
  void* item = NULL;

  keyNum = 1.0e20;
  if( FirstNode ) {
    keyNum = * ( RBSpDoubleKey* ) FirstNode->GetNodeKey();
    item = Remove( keyNum );
  }

  return item;
}


#endif //OLD_FUNCTIONALITY

typedef RB_SpBinaryTree RB_BINARY_TREE;

//RB_DEFINE_CLASS( RB_SpBinaryTree )

#endif

