#ifndef RB_SpContainerFuncs_Header
#define RB_SpContainerFuncs_Header

#include "xhash.H"
#include "xqueue.H"
#include "xpriority.H"

#include "SpAltMgr.H"
#include "SpAltHashChange.H"
#include "SpAltQueueChange.H"
#include "SpAltPriorityChange.H"

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_PUSH_BOT( C_XQUEUE* queue, C_SQ_ITEM *item )
{
  SpAltQueueChange* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltQueueChange();
    altItem->Alter( queue, item, 1, 0 );
    SpCurrentAltMgr->Insert( altItem );
  }

  queue->push_bot( item );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_PUSH_TOP( C_XQUEUE* queue, C_SQ_ITEM *item )
{
  SpAltQueueChange* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltQueueChange();
    altItem->Alter( queue, item, 1, 1 );
    SpCurrentAltMgr->Insert( altItem );
  }

  queue->push_top( item );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_PUSH( C_XQUEUE* queue, C_SQ_ITEM* item )
{
  RB_PUSH_BOT( queue, item );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
C_SQ_ITEM*
RB_POP_TOP( C_XQUEUE* queue )
{
  SpAltQueueChange* altItem;
  C_SQ_ITEM* item;

  item = queue->pop_top();
  if( item == NULL ) {
    return NULL;
  }

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltQueueChange();
    altItem->Alter( queue, item, 0, 1 );
    SpCurrentAltMgr->Insert( altItem );
  }

  return item;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
C_SQ_ITEM*
RB_POP_BOT( C_XQUEUE* queue )
{
  SpAltQueueChange* altItem;
  C_SQ_ITEM* item;

  item = queue->pop_bot();
  if( item == NULL ) {
    return NULL;
  }

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltQueueChange();
    altItem->Alter( queue, item, 0, 0 );
    SpCurrentAltMgr->Insert( altItem );
  }

  return item;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
C_SQ_ITEM*
RB_POP( C_XQUEUE* queue )
{
  return RB_POP_TOP( queue );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_INSERT( C_XQUEUE* queue, C_SQ_ITEM* item )
{
  RB_PUSH( queue, item );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_INSERT( C_XHASH* hash, C_SQ_ITEM* item )
{
  SpAltHashChange* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltHashChange();
    altItem->Alter( hash, item, 1 );
    SpCurrentAltMgr->Insert( altItem );
  }

  hash->add( item );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_INSERT( C_XPRIORITY* priority, C_SQ_ITEM* item )
{
  SpAltPriorityChange* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltPriorityChange();
    altItem->Alter( priority, item, 1 );
    SpCurrentAltMgr->Insert( altItem );
  }

  priority->insert( item );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
C_SQ_ITEM*
RB_REMOVE( C_XQUEUE* queue, C_SQ_ITEM* item )
{
  SpAltQueueChange* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltQueueChange();
    altItem->Alter( queue, item, 0, 1 );
    SpCurrentAltMgr->Insert( altItem );
  }

  queue->remove( item );

  return item;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
C_SQ_ITEM*
RB_REMOVE( C_XQUEUE* queue, int index )
{
  C_SQ_ITEM* item;

  item = queue->find( index );

  if( item == NULL ) {
    return NULL;
  }

  return RB_REMOVE( queue, item );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
C_SQ_ITEM*
RB_REMOVE( C_XQUEUE* queue )
{
  return RB_POP( queue );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
C_SQ_ITEM*
RB_REMOVE( C_XHASH* hash, C_SQ_ITEM* item )
{
  SpAltHashChange* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltHashChange();
    altItem->Alter( hash, item, 0 );
    SpCurrentAltMgr->Insert( altItem );
  }

  hash->remove( item );

  return item;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
C_SQ_ITEM*
RB_REMOVE( C_XHASH* hash, int index )
{
  C_SQ_ITEM* item;

  item = hash->find( index );
  if( item == NULL ) {
    return NULL;
  }

  return RB_REMOVE( hash, item );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
C_SQ_ITEM*
RB_REMOVE( C_XHASH* hash, char*  name )
{
  C_SQ_ITEM* item;

  item = hash->find( name );

  if( item == NULL ) {
    return NULL;
  }

  return RB_REMOVE( hash, item );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
C_SQ_ITEM*
RB_REMOVE( C_XPRIORITY* priority )
{
  SpAltPriorityChange* altItem;
  C_SQ_ITEM* item;

  item = priority->pop();
  if( item == NULL ) {
    return NULL;
  }

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltPriorityChange();
    altItem->Alter( priority, item, 0 );
    SpCurrentAltMgr->Insert( altItem );
  }

  return item;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
operator +=( C_XQUEUE& queue, C_SQ_ITEM* item ) {
  RB_INSERT( &queue, item );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
operator -=( C_XQUEUE& queue, C_SQ_ITEM* item ) {
  RB_REMOVE( &queue, item );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
operator +=( C_XHASH& hash, C_SQ_ITEM* item ) {
  RB_INSERT( &hash, item );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
operator -=( C_XHASH& hash, C_SQ_ITEM* item ) {
  RB_REMOVE( &hash, item );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
operator -=( C_XHASH& hash, int index ) {
  RB_REMOVE( &hash, index );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
operator +=( C_XPRIORITY& priority, C_SQ_ITEM* item ) {
  RB_INSERT( &priority, item );
}

#endif

