#include "RB_SpHashTree.H"

/**********************************************************************
 * 
 **********************************************************************/
RB_SpHashTree::~RB_SpHashTree() {
  delete [] BinaryTree;
}

/**********************************************************************
 * 
 **********************************************************************/
RB_SpHashTree::RB_SpHashTree( int arraySize )
  : CurrentArrayIndex(0), BinaryTree((RB_SpBinaryTree *)0){
  if (SpPoDataBase::DoNotAllocateMemoryInConstructor){
  }
  else{
    BinaryTree = new RB_SpBinaryTree[arraySize];
    NumElements = 0;
    ArraySize = arraySize;
    for( int i = 0; i < arraySize; i++ ) {
      BinaryTree[i].SetBalancedTreeMode();
    }
  }
}

/**********************************************************************
 * 
 **********************************************************************/
void
RB_SpHashTree::SetArraySize( int arraySize ) {
  ArraySize = arraySize;
  if (BinaryTree) delete [] BinaryTree;
  BinaryTree = new RB_SpBinaryTree[arraySize];
  for( int i = 0; i < arraySize; i++ ) {
    BinaryTree[i].SetBalancedTreeMode();
  }
}


 double RB_SpHashTree::GetCurrentKey(){
  return BinaryTree[CurrentArrayIndex].GetCurrentKey();
}

 char* RB_SpHashTree::GetCurrentKeyName(){
  return BinaryTree[CurrentArrayIndex].GetCurrentKeyName();
}

/**********************************************************************
 * 
 **********************************************************************/

void
RB_SpHashTree::SetNormalTreeMode() {
  for( int i = 0; i < ArraySize; i++ ) {
    BinaryTree[i].SetNormalTreeMode();
  }
}

/**********************************************************************
 * 
 **********************************************************************/

void
RB_SpHashTree::SetBalancedTreeMode() {
  for( int i = 0; i < ArraySize; i++ ) {
    BinaryTree[i].SetBalancedTreeMode();
  }
}

/**********************************************************************
 * 
 **********************************************************************/

void
RB_SpHashTree::SetSplayTreeMode() {
  for( int i = 0; i < ArraySize; i++ ) {
    BinaryTree[i].SetSplayTreeMode();
  }
}

/**********************************************************************
 * 
 **********************************************************************/

int
RB_SpHashTree::IsNormalTreeMode() {
  return BinaryTree[0].IsNormalTreeMode();
}

/**********************************************************************
 * 
 **********************************************************************/

int
RB_SpHashTree::IsBalancedTreeMode() {
  return BinaryTree[0].IsBalancedTreeMode();
}

/**********************************************************************
 * 
 **********************************************************************/

int
RB_SpHashTree::IsSplayTreeMode() {
  return BinaryTree[0].IsSplayTreeMode();
}


/**********************************************************************
 * 
 **********************************************************************/

RB_SpBinaryTree*
RB_SpHashTree::GetBinaryTree( int arrayIndex ) {
  return &BinaryTree[arrayIndex];
}

/**********************************************************************
 * 
 **********************************************************************/

int
RB_SpHashTree::CheckIntegrity() {
  int retVal = 0;
  int numElements = 0;
  for( int i = 0; i < ArraySize; i++ ) {
    numElements += BinaryTree[i].GetNumElements();
    retVal |= BinaryTree[i].CheckIntegrity();
  }
  if( numElements != NumElements ) {
    cout << "Error, Bad integrity NumElements "
	 << numElements << " not equal to " << NumElements
	 << endl;
  }
  return retVal;
}

/**********************************************************************
 * 
 **********************************************************************/

void
RB_SpHashTree::Insert( void* element, int key ) {
  BinaryTree[key%ArraySize].Insert( element, key );
  NumElements = NumElements + 1;
}

/**********************************************************************
 * 
 **********************************************************************/

void
RB_SpHashTree::Insert( void* element, char* keyName ) {
  int hashId = GetHashId( keyName );
  BinaryTree[hashId%ArraySize].Insert( element, keyName );
  NumElements = NumElements + 1;
}

/**********************************************************************
 * 
 **********************************************************************/

void*
RB_SpHashTree::Find( int key ) {
  return BinaryTree[key%ArraySize].Find( key );
}

/**********************************************************************
 * 
 **********************************************************************/

void*
RB_SpHashTree::Find( char* keyName ) {
  int hashId = GetHashId( keyName );
  return BinaryTree[hashId%ArraySize].Find( keyName );
}

/**********************************************************************
 * 
 **********************************************************************/

void*
RB_SpHashTree::Remove( int key ) {
  void* retVal = BinaryTree[key%ArraySize].Remove( key );
  if( retVal ) {
    NumElements = NumElements - 1;
  }
  return retVal;
}

/**********************************************************************
 * 
 **********************************************************************/

void*
RB_SpHashTree::Remove( char* keyName ) {
  int hashId = GetHashId( keyName );
  void* retVal = BinaryTree[hashId%ArraySize].Remove( keyName );
  if( retVal ) {
    NumElements = NumElements - 1;
  }
  return retVal;
}

/**********************************************************************
 * 
 **********************************************************************/

void*
RB_SpHashTree::GetFirstElement() {
  void* element = NULL;
  for( int i = 0; i < ArraySize; i++ ) {
    element = BinaryTree[i].GetFirstElement();
    if( element ) {
      CurrentArrayIndex = i;
      break;
    }
  }
  return element;
}

/**********************************************************************
 * 
 **********************************************************************/

void*
RB_SpHashTree::GetLastElement() {
  void* element;
  for( int i = ArraySize - 1; i >= 0; i-- ) {
    element = BinaryTree[i].GetLastElement();
    if( element ) {
      CurrentArrayIndex = i;
      break;
    }
  }
  return element;
}

/**********************************************************************
 * 
 **********************************************************************/

void*
RB_SpHashTree::GetNextElement() {
  void* element = NULL;
  for( int i = CurrentArrayIndex; i < ArraySize; i++ ) {
    element = BinaryTree[i].GetNextElement();
    if( element ) {
      CurrentArrayIndex = i;
      break;
    }
    if( i < ArraySize - 1 ) {
      element = BinaryTree[i+1].GetFirstElement();
    }
    if( element ) {
      CurrentArrayIndex = i+1;
      break;
    }
  }
  return element;
}

/**********************************************************************
 * 
 **********************************************************************/

void*
RB_SpHashTree::GetPreviousElement() {
  void* element = NULL;
  for( int i = CurrentArrayIndex; i >= 0; i-- ) {
    element = BinaryTree[i].GetPreviousElement();
    if( element ) {
      CurrentArrayIndex = i;
      break;
    }
    if( i > 0 ) {
      element = BinaryTree[i-1].GetLastElement();
    }
    if( element ) {
      CurrentArrayIndex = i-1;
      break;
    }
  }
  return element;
}

/**********************************************************************
 * 
 **********************************************************************/

void*
RB_SpHashTree::GetCurrentElement() {
  return BinaryTree[CurrentArrayIndex].GetCurrentElement();
}

