#ifndef RB_SpHashTree_Header
#define RB_SpHashTree_Header

#include "RB_SpBinaryTree.H"
#include "RB_Types.H"
#include "PO.H"


/**********************************************************************
 * 
 **********************************************************************/
class RB_SpHashTree {
public:
  RB_SpHashTree( int arraySize = 100);
  ~RB_SpHashTree();
  void SetArraySize( int arraySize );
  void SetNormalTreeMode();
  void SetBalancedTreeMode();
  void SetSplayTreeMode();
  int IsNormalTreeMode();
  int IsBalancedTreeMode();
  int IsSplayTreeMode();
  int GetNumElements();
  int GetArraySize();
  RB_SpBinaryTree* GetBinaryTree();
  RB_SpBinaryTree* GetBinaryTree( int arrayIndex );
  int GetHashId( char* keyName );
  int CheckIntegrity();
  void Insert( void* element, int key );
  void Insert( void* element, char *keyName );
  void* Find( int key );
  void* Find( char* keyName );
  void* Remove( int key );
  void* Remove( char *keyName );
  void* GetFirstElement();
  void* GetLastElement();
  void* GetNextElement();
  void* GetPreviousElement();
  void* GetCurrentElement();
  void* operator ++();
  void* operator --();
  void* operator ++(int);
  void* operator --(int);
  void operator +=( void* ptr );
  void* operator -=( void* ptr );
  double GetCurrentKey();
  char* GetCurrentKeyName();
private:
  RB_SpBinaryTree* BinaryTree;
  RB_int NumElements;
  int ArraySize;
  int CurrentArrayIndex;
};

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_SpHashTree::GetNumElements() {
  return NumElements;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_SpHashTree::GetArraySize() {
  return ArraySize;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
RB_SpBinaryTree*
RB_SpHashTree::GetBinaryTree() {
  return BinaryTree;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_SpHashTree::GetHashId( char* keyName ) {
  int hashId = 0;
  int len = strlen( keyName );
  for( int i = 0; i < len; i++ ) {
    hashId ^= keyName[i] << ( ( i * 8 ) % 32 );
  }
  return ( hashId & 0x7fffffff );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpHashTree::operator ++() {
  return GetNextElement();
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpHashTree::operator --() {
  return GetPreviousElement();
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpHashTree::operator ++(int) {
  void* element = GetCurrentElement();
  GetNextElement();
  return element;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpHashTree::operator --(int) {
  void* element = GetCurrentElement();
  GetPreviousElement();
  return element;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_SpHashTree::operator +=( void* ptr ) {
  Insert( ptr, int( ptr ) );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpHashTree::operator -=( void* ptr ) {
  return Remove( int( ptr ) );
}

typedef RB_SpHashTree RB_HASH_TREE;

#endif

