#ifndef RB_SpList_Header
#define RB_SpList_Header

//#define USE_OLD_FUNCTIONALITY
#ifndef USE_OLD_FUNCTIONALITY

#include "SpHolderItem.H"
#include "RB_Types.H"
#include "SpPersistenceBaseClass.H"
#include "SpPoDataBase.H"

extern SpPoDataBase* DefaultDataBase;

extern SpPoMempool* GlobalSpPoMempool;

#ifdef __IBMCPP__
namespace {
#endif

#ifndef PO_VOID_TYPE_CLASS_DEFINED
typedef char PO_VOID_TYPE_CLASS;
RB_DEFINE_CLASS_PRIMATIVE(PO_VOID_TYPE_CLASS);
#define PO_VOID_TYPE_CLASS_DEFINED
#endif
#ifdef __IBMCPP__
typedef PO_PTR_PO_VOID_TYPE_CLASS PO_PTR_VOID;
}
#else
typedef PO_PTR_PO_VOID_TYPE_CLASS PO_PTR_VOID;
#endif



class RB_SpListItem;
PO_DEFINE_SMART_POINTERS_NO_ADDITIONAL_OPERATORS(RB_SpListItem);

class RB_SpListItem : public SpPersistenceBaseClass{
public:
  virtual void SetForwardLink(void*){};
  virtual RB_SpListItem* GetForwardLink(){return NULL;};

  virtual void SetBackwardLink(void*){};
  virtual RB_SpListItem* GetBackwardLink(){return NULL;};

  virtual void SetReference(void*){};
  virtual void* GetReference(){return NULL;};

};
class PersistentRB_SpListItem;
PO_DEFINE_SMART_POINTERS_NO_ADDITIONAL_OPERATORS(PersistentRB_SpListItem);

class PersistentRB_SpListItem : public RB_SpListItem{
private:
  PO_PTR_VOID Reference;
  PO_PTR_PersistentRB_SpListItem ForwardLink;
  PO_PTR_PersistentRB_SpListItem BackwardLink;
  public:
  PersistentRB_SpListItem():
    Reference((char*)NULL),
    ForwardLink((PersistentRB_SpListItem*)NULL),
    BackwardLink((PersistentRB_SpListItem*)NULL) {}
  void Reset(){
    Reference = (char*)NULL;
    ForwardLink = (PersistentRB_SpListItem*)NULL;
    BackwardLink = (PersistentRB_SpListItem*)NULL;
  }
public:
  virtual void SetForwardLink(void* t)
    {ForwardLink = (PersistentRB_SpListItem*)t;}
  virtual RB_SpListItem* GetForwardLink(){return ForwardLink;}

  virtual void SetBackwardLink(void* t)
    {BackwardLink = (PersistentRB_SpListItem*)t;}
  virtual RB_SpListItem* GetBackwardLink(){return BackwardLink;};


  virtual void SetReference(void* t){Reference = (char*)t;}
  virtual void* GetReference(){return Reference;};

};

PO_DEFINE_CLASS_NO_SMART_POINTERS(PersistentRB_SpListItem);


class NonPersistentRB_SpListItem : public RB_SpListItem{
private:
  void* Reference;
  NonPersistentRB_SpListItem* ForwardLink;
  NonPersistentRB_SpListItem* BackwardLink;
  public:
  NonPersistentRB_SpListItem():
    Reference((char*)NULL),
    ForwardLink((NonPersistentRB_SpListItem*)NULL),
    BackwardLink((NonPersistentRB_SpListItem*)NULL) {}
  void Reset(){
    Reference = (char*)NULL;
    ForwardLink = (NonPersistentRB_SpListItem*)NULL;
    BackwardLink = (NonPersistentRB_SpListItem*)NULL;
  }
public:
  virtual void SetForwardLink(void* t)
    {ForwardLink = (NonPersistentRB_SpListItem*)t;}
  virtual RB_SpListItem* GetForwardLink(){return ForwardLink;}

  virtual void SetBackwardLink(void* t)
    {BackwardLink = (NonPersistentRB_SpListItem*)t;}
  virtual RB_SpListItem* GetBackwardLink(){return BackwardLink;};

  virtual void SetReference(void* t){Reference=(char*)t;}
  virtual void* GetReference(){return Reference;};

};




#include "SpAltMgr.H"
#include "SpAltListChange.H"

/**********************************************************************
 * 
 **********************************************************************/
class RB_SpList  {
  friend class SpAltListChange;
protected:
  enum {
    TOP_FLAG,
    BOT_FLAG,
    FREE_LIST_SIZE = 200
  };

public:
  // If you call Insert (or other methods which modify this object) on
  // the return value of this function, and the event is rolled back,
  // then the item you inserted will NOT be removed.  I.e. the state
  // of this RB_SpList will NOT be restored!  So be very careful if
  // you decide to call this dangerous function.  Some acceptable
  // uses of it would be to only call methods on the return value that
  // do NOT modify the RB_SpList, like GetNumElements.
  //SpList &GetNonRbVersion() { return *this; }
  RB_SpList();
  void* RB_Insert( void* item, int TopBotFlag );
  void* Insert( void* item, int TopBotFlag  = BOT_FLAG);
  void RB_Remove( void* item );
  void* RB_add( void* item, int TopBotFlag  = BOT_FLAG );
  void RB_remove( void* item );
  void Remove( void* item );
  void remove( void* item );
  void* operator +=( void* item );
  void operator -=( void* item );
  void* GetFirstElement();
  void* GetLastElement();
  void* push_top( void* item );
  void* push_bot( void* item );
  void* pop_top();
  void* pop_bot();
  int GetNumElements();
  int get_length();
  void* GetTop();
  void* get_top();
  void* GetBot();
  void* get_bot();
  void* GetNextElement();
  void* GetPreviousElement();
  void* get_next();
  void* get_previous();
  void* PushBot(void*);
  void* push_bot(void);

  // Warning!  The following iteration operations are NOT rollbackable!
  void* operator ++();
  void* operator --();

private:
  RB_SpListItem* RemoveTopItem();
  RB_SpListItem* RemoveBotItem();
  void InternalRemove(void* item);
  void* InsertBot(void* item);
  void* InsertTop(void* item);
  void InsertTopItem(RB_SpListItem* item);
  void InsertBotItem(RB_SpListItem* item);
  void SetNumElements(int l) {length = l;}

  int length;			   // length of the queue
  PO_PTR_RB_SpListItem top;	   // pointer to the top of the queue
  PO_PTR_RB_SpListItem bot;	   // pointer to the bottom of the queue
  RB_SpListItem *current;          // current pointer in the list


  RB_SpListItem *NewItem();
  void AllocateMoreItems();
  void DeleteItem(RB_SpListItem *item);
  static int PoFreeId;
};

inline RB_SpListItem *RB_SpList::NewItem(){
  RB_SpListItem* rc = (RB_SpListItem*)GlobalSpPoMempool->RemoveItem(PoFreeId);
  if (rc == NULL){
    AllocateMoreItems();
    rc = (RB_SpListItem*)GlobalSpPoMempool->RemoveItem(PoFreeId);
  }
  return rc;
}

inline void RB_SpList::DeleteItem(RB_SpListItem *item){
  GlobalSpPoMempool->ReturnItem(PoFreeId, item);
}


/******************************************************************************
* Method: GetNumElements
* Summary: Get the number of elements in the list
* Description: 
* Return: The number of elements in the list
******************************************************************************/
inline int RB_SpList::GetNumElements() {return length;}

   /* backward compatibility */
inline int RB_SpList::get_length() { return length; }

/******************************************************************************
* Method: GetFirstElement
* Summary: Get a void pointer to the first element in the list
* Description: 
* Return: A void pointer to the first element in the list
******************************************************************************/
inline void *RB_SpList::GetFirstElement() {
  void *Ref;
  if (length) {
    current = top;
    Ref = top->GetReference();
  }else{
    current = NULL;
    Ref = NULL;
  }
  return Ref;
}

   /* backward compatibility */
inline void *RB_SpList::GetTop() { return GetFirstElement(); }
inline void *RB_SpList::get_top() { return GetFirstElement(); }

/******************************************************************************
* Method: GetLastElement
* Summary: Get a void pointer to the last element in the list
* Description: 
* Return: A void pointer to the last element in the list
******************************************************************************/
inline void *RB_SpList::GetLastElement() {
  void *Ref;
  if (length) {
    current = bot;
    Ref = bot->GetReference();
  }else{
    current = NULL;
    Ref = NULL;
  }
  return Ref;
}

/* backward compatibility */
inline void *RB_SpList::GetBot() { return GetLastElement(); }
inline void *RB_SpList::get_bot() { return GetLastElement(); }



/******************************************************************************
* Method: InsertTopItem
* Summary: Insert a list item to the top of the list
* Description: 
* Return:
******************************************************************************/
inline void RB_SpList::InsertTopItem(RB_SpListItem *item){
  if (length == 0) {
    bot = item;
    top = item;
    length = 1;
  }else{
    top->SetBackwardLink(item);
    item->SetForwardLink(top);
    top = item;
    length++;
  }
}


/*****************************************************************************
* Method: InsertTop
* Summary: Insert an element on the top of the list
* Description: 
* Return: Pointer to the list item that contains the inserted element
******************************************************************************/
inline void *RB_SpList::InsertTop(void *v) {
  RB_SpListItem *item = NewItem();
  item->SetReference(v);
  InsertTopItem(item);
  return item;
}
   
/******************************************************************************
* Method: InsertBotItem
* Summary: Insert a list item to the bottom of the list
* Description: 
* Return:
******************************************************************************/
inline void RB_SpList::InsertBotItem(RB_SpListItem *item){
  if (length == 0) {
    bot = item;
    top = item;
    length = 1;
  }else{
    bot->SetForwardLink(item);
    item->SetBackwardLink(bot);
    bot = item;
    length++;
  }
}
   

/******************************************************************************
* Method: InsertBot
* Summary: Insert an element on the bottom of the list
* Description: 
* Return: Pointer to the removed element
******************************************************************************/
inline void *RB_SpList::InsertBot(void *v) {
  RB_SpListItem *item = NewItem();
  item->SetReference(v);
  InsertBotItem(item);
  return item;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpList::RB_Insert( void* item, int TopBotFlag) {
  SpAltListChange* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltListChange();
    altItem->Alter( this, item, TopBotFlag, 1 );
    SpCurrentAltMgr->Insert( altItem  );
  }

  if( TopBotFlag == TOP_FLAG ) {
    return InsertTop( item );
  }else{
    return InsertBot( item );
  }
}

/**********************************************************************
 * 
 **********************************************************************/
   /* backward compatibility */

inline
void*
RB_SpList::push_top( void* item ) {
  return RB_Insert( item, TOP_FLAG );
}

inline
void*
RB_SpList::push_bot( void* item ) {
  return RB_Insert( item, BOT_FLAG );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpList::Insert( void* item, int TopBotFlag ) {
  return RB_Insert( item, TopBotFlag );
}


/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_SpList::RB_Remove( void* item ) {
  SpAltListChange* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltListChange();
    altItem->Alter( this, item, TOP_FLAG, 0 );
    SpCurrentAltMgr->Insert( altItem  );
  }

  InternalRemove( item );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_SpList::Remove( void* item ) {
  RB_Remove( item );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpList::RB_add( void* item, int TopBotFlag) {
  return RB_Insert( item, TopBotFlag );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_SpList::RB_remove( void* item ) { 
  RB_Remove( item );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_SpList::remove( void* item ) {
  RB_Remove( item );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpList::operator +=( void* item ) {
  return RB_add( item);
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_SpList::operator -=( void* item ) {
  RB_Remove( item );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpList::pop_top() {
  void* item = GetFirstElement();
  RB_Remove( item );
  return item;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpList::pop_bot(){
  void* item = GetLastElement();
  RB_Remove( item );
  return item;
}




/******************************************************************************
* Constructor: RB_SpList
* Summary: Constructs the SpList (no arguments required)
* Description:
* Return:
******************************************************************************/
inline RB_SpList::RB_SpList():
  length(0)
{
  if (DefaultDataBase == NULL){
    DefaultDataBase = new SpPoDataBase();
  }
  top = (RB_SpListItem*)NULL;
  bot = (RB_SpListItem*)NULL;
  current = (RB_SpListItem*)NULL;
};
   
/******************************************************************************
* Method: GetNextElement
* Summary: Get a void pointer to the next element in the list
* Description: 
* Return: A void pointer to the next element in the list
******************************************************************************/
inline void *RB_SpList::GetNextElement() {
  void* rc = NULL;
  if (current != bot) {
    if (current != NULL){
      current = current->GetForwardLink();
    }
    else{
      current = top;
    }
    if (current){
      rc = current->GetReference();
    }
  }
  return rc;
}
   
   /* backward compatibility */
inline void *RB_SpList::get_next() { return GetNextElement(); }

/******************************************************************************
* Method: GetPreviousElement
* Summary: Get a void pointer to the previous element in the list
* Description: 
* Return: A void pointer to the previous element in the list
******************************************************************************/
inline void *RB_SpList::GetPreviousElement() {
  void* rc = NULL;
  if (current != top) {
    if (current != NULL){
      current = current->GetBackwardLink();
      rc = current->GetReference();
    }
    else{
      current = bot;
    }
    if (current){
      rc = current->GetReference();
    }
  }
  return rc;
}
   
   /* backward compatibility */
inline void *RB_SpList::get_previous() { return GetPreviousElement(); }

/******************************************************************************
* Operator: ++
* Summary: iterator method that returns the next element in the list
* Description:
* Return: The next element in the list
******************************************************************************/
inline void *RB_SpList::operator ++ () {return GetNextElement();}
   
/******************************************************************************
* Operator: --
* Summary: Returns the previous element in the list
* Description:
* Return: The previous element in the list (NULL if no previous element)
******************************************************************************/
inline void *RB_SpList::operator -- () {return GetPreviousElement();}





typedef RB_SpList C_XLIST;
typedef RB_SpList XLIST;
typedef RB_SpList RB_List;
typedef RB_SpList RB_list;
typedef RB_SpList RB_LIST;


#else

#include "SpList.H"

#include "SpAltMgr.H"
#include "SpAltListChange.H"

/**********************************************************************
 * 
 **********************************************************************/
class RB_SpList : protected SpList {
  friend class SpAltListChange;
public:
  // If you call Insert (or other methods which modify this object) on
  // the return value of this function, and the event is rolled back,
  // then the item you inserted will NOT be removed.  I.e. the state
  // of this RB_SpList will NOT be restored!  So be very careful if
  // you decide to call this dangerous function.  Some acceptable
  // uses of it would be to only call methods on the return value that
  // do NOT modify the RB_SpList, like GetNumElements.
  SpList &GetNonRbVersion() { return *this; }
  void InternalRemove(void* item){SpList::Remove(item);}
  RB_SpList():PoFreeId(-1) {}
  void* RB_Insert( void* item, int TopBotFlag );
  void* Insert( void* item, int TopBotFlag );
  void* Insert( void* item );
  void RB_Remove( void* item );
  void* RB_add( void* item, int TopBotFlag );
  void RB_remove( void* item );
  void Remove( void* item );
  void remove( void* item );
  void* operator +=( void* item );
  void operator -=( void* item );
  void* GetFirstElement();
  void* GetLastElement();
  void* push_top( void* item );
  void* push_bot( void* item );
  void* pop_top();
  void* pop_bot();
  int GetNumElements();
  int get_length();
  void* GetTop();
  void* get_top();
  void* get_bot();
  SpList::GetNextElement;
  SpList::GetPreviousElement;
  SpList::get_next;
  SpList::get_previous;
  //virtual SpList_ITEM *NewItem();
  //virtual void DeleteItem(SpList_ITEM *item);

  // Warning!  The following iteration operations are NOT rollbackable!
  void* operator ++();
  void* operator --();

private:
  int PoFreeId;
};

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpList::RB_Insert( void* item, int TopBotFlag = BOT_FLAG ) {
  SpAltListChange* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltListChange();
    altItem->Alter( this, item, TopBotFlag, 1 );
    SpCurrentAltMgr->Insert( altItem  );
  }

  if( TopBotFlag == TOP_FLAG ) {
    return InsertTop( item );
  }else{
    return InsertBot( item );
  }
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpList::Insert( void* item, int TopBotFlag ) {
  return RB_Insert( item, TopBotFlag );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpList::Insert( void* item ) {
  return RB_Insert( item, BOT_FLAG );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_SpList::RB_Remove( void* item ) {
  SpAltListChange* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltListChange();
    altItem->Alter( this, item, TOP_FLAG, 0 );
    SpCurrentAltMgr->Insert( altItem  );
  }

  SpList::Remove( item );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_SpList::Remove( void* item ) {
  RB_Remove( item );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpList::RB_add( void* item, int TopBotFlag = BOT_FLAG ) {
  return RB_Insert( item, TopBotFlag );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_SpList::RB_remove( void* item ) { 
  RB_Remove( item );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_SpList::remove( void* item ) {
  RB_Remove( item );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpList::operator +=( void* item ) {
  return RB_add( item );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_SpList::operator -=( void* item ) {
  RB_Remove( item );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpList::GetFirstElement(){
  return SpList::GetFirstElement();
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpList::GetLastElement() {
  return SpList::GetLastElement();
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpList::push_top( void* item ) {
  return RB_Insert( item, TOP_FLAG );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpList::push_bot( void* item ) {
  return RB_Insert( item, BOT_FLAG );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpList::pop_top() {
  void* item = GetFirstElement();
  RB_Remove( item );
  return item;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpList::pop_bot(){
  void* item = GetLastElement();
  RB_Remove( item );
  return item;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_SpList::GetNumElements(){
  return SpList::GetNumElements();
}

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_SpList::get_length(){
  return SpList::get_length();
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpList::GetTop() {
  return SpList::GetTop();
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpList::get_top() {
  return SpList::get_top();
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpList::get_bot() {
  return SpList::get_bot();
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpList::operator ++() {
  return SpList::operator++();
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpList::operator --() {
  return SpList::operator--();
}

typedef RB_SpList C_XLIST;
typedef RB_SpList XLIST;
typedef RB_SpList RB_List;
typedef RB_SpList RB_list;
typedef RB_SpList RB_LIST;

#endif

#endif

