#ifndef RB_SpPriorityTree_Header
#define RB_SpPriorityTree_Header

#include "RB_SpBinaryTree.H"

/**********************************************************************
 * 
 **********************************************************************/
class RB_SpPriorityTree {
public:
  RB_SpPriorityTree() { BinaryTree.SetBalancedTreeMode(); }
  int CheckIntegrity();
  int GetNumElements();
  int IsNormalTreeMode();
  int IsBalancedTreeMode();
  int IsSplayTreeMode();
  void SetNormalTreeMode();
  void SetBalancedTreeMode();
  void SetSplayTreeMode();
  void* Find( double priority );
  double GetCurrentPriority();
  void* Insert( void* item, double priority );
  void* Remove( double& priority );
  void* Remove();
  int Retract( void* retractionHandle );
protected:
  RB_SpBinaryTree BinaryTree;
};

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_SpPriorityTree::CheckIntegrity() {
  return BinaryTree.CheckIntegrity();
}

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_SpPriorityTree::GetNumElements() {
  return BinaryTree.GetNumElements();
}

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_SpPriorityTree::IsNormalTreeMode() {
  return BinaryTree.IsNormalTreeMode();
}

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_SpPriorityTree::IsBalancedTreeMode() {
  return BinaryTree.IsBalancedTreeMode();
}

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_SpPriorityTree::IsSplayTreeMode() {
  return BinaryTree.IsSplayTreeMode();
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_SpPriorityTree::SetNormalTreeMode() {
  BinaryTree.SetNormalTreeMode();
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_SpPriorityTree::SetBalancedTreeMode() {
  BinaryTree.SetBalancedTreeMode();
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_SpPriorityTree::SetSplayTreeMode() {
  BinaryTree.SetSplayTreeMode();
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpPriorityTree::Find( double priority ) {
  return BinaryTree.Find( priority );
}

/**********************************************************************
 * 
 **********************************************************************/
inline
double
RB_SpPriorityTree::GetCurrentPriority() {
  BaseRB_SpTreeItem* top = BinaryTree.GetFirstNode();
  return * ( SpDoubleKey* ) top->GetNodeKey();
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpPriorityTree::Insert( void* item, double priority ) {
  BinaryTree.Insert( item, priority );
  return BinaryTree.GetCurrentNode();
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpPriorityTree::Remove( double& priority ) {
  void* element = BinaryTree.RemoveFirstElement( priority );
  return element;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void*
RB_SpPriorityTree::Remove() {
  void* element = BinaryTree.RemoveFirstElement();
  return element;
}

/**********************************************************************
 * 
 **********************************************************************/
inline
int
RB_SpPriorityTree::Retract( void* retractionHandle ) {
  int n = BinaryTree.GetNumElements();
  RBSpKey* key =
    (( BaseRB_SpTreeItem* ) retractionHandle)->GetNodeKey()->Clone();
  BinaryTree.Remove( key );
  delete key;
  return( n - BinaryTree.GetNumElements() );
}

typedef RB_SpPriorityTree RB_PRIORITY_TREE;

#endif

