#include "SpHiGridRegionVec.H"
#include "SpPubSubItem.H"

SpHiGridRegionVec::SpHiGridRegionVec
  (const int num_grids, const int numDimRegions)
  : MaxNum(num_grids), NextIndexToSet(-1), IdCoverage(NULL), Region(NULL),
    NumDimRegions(numDimRegions)
{
  assert(MaxNum >= 0); 
  assert(NumDimRegions >= 0);
  
  if (MaxNum >= 1)
    alloc();
}

SpHiGridRegionVec::~SpHiGridRegionVec()
{
  if (IdCoverage) free(IdCoverage);
  if (Region) free(Region);
}

static inline void *saferealloc(void *v, const size_t s)
{
  if (v)
    return realloc(v, s);
  else
    return malloc(s);
} 

void SpHiGridRegionVec::alloc()
{
  IdCoverage = (GridIdCoverage*)
    saferealloc(IdCoverage, MaxNum * sizeof(GridIdCoverage));

  if (NumDimRegions >= 1)
    Region = (DimRegion*)
      saferealloc(Region, MaxNum * NumDimRegions * sizeof(DimRegion));
}

SpHiGridRegionVec &SpHiGridRegionVec::set_next()
{
  ++NextIndexToSet;
  if (NextIndexToSet == MaxNum)
  {
    cerr << "Warning: I'm taking a lot of time reallocating the arrays!"
	 << endl;
    assert(MaxNum >= 1);
    MaxNum *= 2;
    alloc();
  }
  return *this;
}

void SpHiGridRegionVec::Init_PubSub_Item
  (MAIN_PUBSUB_ITEM * const MainPubSubItem, const int CurrentIndex) const
{
  assert(CurrentIndex >= 0);
  MainPubSubItem->AllocateNumDimensions(NumDimRegions);
  const DimRegion * const reg = &(get_DimRegion(CurrentIndex, 0));
  for (int i = 0; i < NumDimRegions; i++)
  {
    MainPubSubItem
      ->SetLHR(i, reg[i].get_Lo(), reg[i].get_Hi(), reg[i].get_Res());
  }
}

SpHiGridRegionVec &SpHiGridRegionVec::copy_grid
  (const SpHiGridRegionVecReader &reader)
{
  assert(NextIndexToSet >= 0);
  assert(NumDimRegions == reader.get_HGRVector()->NumDimRegions);
  const DimRegion * const source = reader.get_currentDimRegion();
  DimRegion * const target = &(get_DimRegion(NextIndexToSet, 0));
  for (int i = 0; i < NumDimRegions; i++)
    target[i] = source[i];

  IdCoverage[NextIndexToSet].Id = reader.id();
  IdCoverage[NextIndexToSet].GridCovered_p = reader.gridCovered_p();

  return *this;
}

SpHiGridRegionVec &SpHiGridRegionVec::SetRegionsToConcatenationOf
  (const SpHiGridRegionVecReader &reader1,
   const SpHiGridRegionVecReader &reader2)
{
  assert(NextIndexToSet >= 0);
  assert(reader1.get_HGRVector()->NumDimRegions +
	 reader2.get_HGRVector()->NumDimRegions == NumDimRegions);
  int target_index = 0;
  CopyRegion(target_index, reader1);
  CopyRegion(target_index, reader2);
  return *this;
}

void SpHiGridRegionVec::CopyRegion
  (int &target_index, const SpHiGridRegionVecReader &source)
{
  const int sourceNumDimRegions = source.get_HGRVector()->NumDimRegions;
  if (sourceNumDimRegions >= 1)
  {
    const DimRegion * const sourceDimRegion = source.get_currentDimRegion();
    DimRegion * const targetDimRegion = &(get_DimRegion(NextIndexToSet, 0));
    for (int i = 0; i < sourceNumDimRegions; i++)
      targetDimRegion[target_index++] = sourceDimRegion[i];
  }
} 
