#ifndef SpKey_H
#define SpKey_H


#include "SpString.H"
#include "SpStdlib.H"
#include "SpPoDataBase.H"
#include "SpMempool.H"
#include "SpKeyMode.H"

class SpKey{
private:
protected:

public:
  SpKey(){}
  virtual ~SpKey(){}
  virtual SpKey * Clone(){ return NULL;}
  virtual KEY_MODE GetMode() const{return NO_KEY_MODE;}
  int operator <(const SpKey & y)const { return this->lt(&y);}
  int operator >(const SpKey &y)const { return this->gt(&y);}
  int operator ==(const SpKey &y)const { return this->eq(&y);}
  int operator >=(const SpKey &y)const { return this->geq(&y);}
  int operator <=(const SpKey &y)const { return this->leq(&y);}

  virtual int lt(const SpKey*)const;
  virtual int gt(const SpKey*)const;
  virtual int eq(const SpKey*)const;
  virtual int leq(const SpKey*)const;
  virtual int geq(const SpKey*)const;
};


class SpDoubleKey :public SpKey{
private:
  double Val;

public:
  SpDoubleKey(){;}
  SpDoubleKey(double v){Val = v;}

  virtual KEY_MODE GetMode() const{return DOUBLE_KEY_MODE;}
  virtual SpKey * Clone(){ return new SpDoubleKey(*this);}
  virtual operator double () {return Val;}
  int operator <(const SpDoubleKey & y)const { return (Val < y.Val);}
  int operator >(const SpDoubleKey & y)const { return (Val > y.Val);}
  int operator ==(const SpDoubleKey & y)const { return (Val == y.Val);}
  int operator >=(const SpDoubleKey & y)const { return (Val >= y.Val);}
  int operator <=(const SpDoubleKey & y)const { return (Val <= y.Val);}

  virtual int lt(const SpKey*y)const
    {return this->Val < ((SpDoubleKey*)y)->Val;}
  virtual int gt(const SpKey*y)const
    {return this->Val > ((SpDoubleKey*)y)->Val;}
  virtual int eq(const SpKey*y)const
    {return this->Val == ((SpDoubleKey*)y)->Val;}
  virtual int leq(const SpKey*y)const
    {return this->Val <= ((SpDoubleKey*)y)->Val;}
  virtual int geq(const SpKey*y)const
    {return this->Val >= ((SpDoubleKey*)y)->Val;}

};

class SpStringKey :public SpKey{
private:
  char* Val;

public:
  virtual KEY_MODE GetMode() const{return STRING_KEY_MODE;}
  SpStringKey(){Val = NULL;}
  SpStringKey(const char * s){Val = strdup(s);}
  SpStringKey(const SpStringKey &s){Val = strdup(s.Val);}
  SpStringKey & operator = (const SpStringKey &s)
    {
      if (this != &s){
	free(Val);
	Val = strdup(s.Val);
      }
      return *this;
    }
  virtual ~SpStringKey (){free(Val);}

  virtual SpKey * Clone(){ return new SpStringKey(*this);}
  virtual operator char* () {return Val;}

  virtual int lt(const SpKey*y)const 
    {return strcmp(this->Val, ((SpStringKey*)y)->Val) < 0;}
  virtual int gt(const SpKey*y)const
    {return strcmp(this->Val, ((SpStringKey*)y)->Val) > 0;}
  virtual int eq(const SpKey*y)const
    {return strcmp(this->Val, ((SpStringKey*)y)->Val) == 0;}
  virtual int leq(const SpKey*y)const
    {return strcmp(this->Val, ((SpStringKey*)y)->Val) > -1;}
  virtual int geq(const SpKey*y)const
    {return strcmp(this->Val, ((SpStringKey*)y)->Val) < 1;}

};


class SpIntervalKey :public SpKey{
private:
  int Low;
  int High;

public:
  virtual KEY_MODE GetMode() const{return INTERVAL_KEY_MODE;}
  SpIntervalKey(){;}
  SpIntervalKey(int l){Low = High = l;}
  SpIntervalKey(int l, int h){Low = l; High = h;}
  
  int Min(){return Low;}
  int Max(){return High;}
  
  virtual SpKey * Clone(){ return new SpIntervalKey(*this);}
  SpIntervalKey& operator()(int l, int h){Low = l; High = h; return *this;}
  
  virtual int lt(const SpKey*y)const
    {return this->High < ((SpIntervalKey*)y)->Low;}
  virtual int gt(const SpKey*y)const
    {return this->Low > ((SpIntervalKey*)y)->High;}
  virtual int eq(const SpKey*y)const
    { return (((this->High >= ((SpIntervalKey*)y)->High) &&
	       (this->Low <= ((SpIntervalKey*)y)->Low)) ||
	      ((this->High <= ((SpIntervalKey*)y)->High) &&
	       (this->Low >= ((SpIntervalKey*)y)->Low)));}
  virtual int leq(const SpKey*y)const
    {return this->High <= ((SpIntervalKey*)y)->Low;}
  virtual int geq(const SpKey*y)const
    {return this->Low >= ((SpIntervalKey*)y)->High;}
};


#endif
