// SpList.H header file

#ifndef SpList_object
#define SpList_object

#include "SpStdlib.H"  //Gets us the NULL macro
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"

#define SpList_FREE_SIZE 200

class SpList_ITEM {
private:
  friend class SpList;
  void *Reference;
  SpList_ITEM *ForwardLink;
  SpList_ITEM *BackwardLink;
public:
  SpList_ITEM(): Reference(NULL), ForwardLink(NULL), BackwardLink(NULL) {}
  void Reset(){ Reference = ForwardLink = BackwardLink = NULL;}
};

/************************************************************************
*	SpList - doubly linked list data structure			*
************************************************************************/
class SpList {
  protected:
    enum {
      TOP_FLAG,
      BOT_FLAG
    };

 public:

/******************************************************************************
* Method: Reset
* Summary: Clear the list pointers without deleting anything the list points to
* Description: Set length to zero, and the top and bottom pointers to NULL
* Return: None
******************************************************************************/
   void Reset();

   /* backward compatibility */
   void reset () { Reset(); }

/******************************************************************************
* Method: GetNumElements
* Summary: Get the number of elements in the list
* Description: 
* Return: The number of elements in the list
******************************************************************************/
   int GetNumElements() {return length;}

   /* backward compatibility */
   int GetLength() {return length;}
   int get_length() { return length; }

/******************************************************************************
* Method: GetFirstElement
* Summary: Get a void pointer to the first element in the list
* Description: 
* Return: A void pointer to the first element in the list
******************************************************************************/
   void *GetFirstElement() {
      void *Ref;
      if (length) {
	 current = top;
	 Ref = top->Reference;
      }else{
	 current = NULL;
	 Ref = NULL;
      }
      return Ref;
   }

   /* backward compatibility */
   void *GetTop() { return GetFirstElement(); }
   void *get_top() { return GetFirstElement(); }

/******************************************************************************
* Method: GetLastElement
* Summary: Get a void pointer to the last element in the list
* Description: 
* Return: A void pointer to the last element in the list
******************************************************************************/
   void *GetLastElement() {
      void *Ref;
      if (length) {
	 current = bot;
	 Ref = bot->Reference;
      }else{
	 current = NULL;
	 Ref = NULL;
      }
      return Ref;
   }
   
   /* backward compatibility */
   void *GetBot() { return GetLastElement(); }
   void *get_bot() { return GetLastElement(); }

/******************************************************************************
* Method: GetNextElement
* Summary: Get a void pointer to the next element in the list
* Description: 
* Return: A void pointer to the next element in the list
******************************************************************************/
   void *GetNextElement() {
      if (current == bot) {
	 return NULL;
      }else{
	 current = current->ForwardLink;
	 return current->Reference;
      }
   }
   
   /* backward compatibility */
   void *GetNext() { return GetNextElement(); }
   void *get_next() { return GetNextElement(); }

/******************************************************************************
* Method: GetPreviousElement
* Summary: Get a void pointer to the previous element in the list
* Description: 
* Return: A void pointer to the previous element in the list
******************************************************************************/
   void *GetPreviousElement() {
      if (current == top) {
	 return NULL;
      }else{
	 current = current->BackwardLink;
	 return current->Reference;
      }
   }
   
   /* backward compatibility */
   void *GetPrevious() { return GetPreviousElement(); }
   void *get_previous() { return GetPreviousElement(); }

/******************************************************************************
* Method: Insert
* Summary: Insert an element on the top of the list (same as InsertTop)
* Description: 
* Return: Pointer to the list item that contains the inserted element
******************************************************************************/
   void *Insert(void *v, int TopOrBot) {
     if (TopOrBot == TOP_FLAG)
       return InsertTop(v);
     else
       return InsertBot(v);
   }
   void *Insert(void *v) { return Insert(v, TOP_FLAG); }

/******************************************************************************
* Method: InsertTop
* Summary: Insert an element on the top of the list
* Description: 
* Return: Pointer to the list item that contains the inserted element
******************************************************************************/
   void *InsertTop(void *v) {
      SpList_ITEM *item = NewItem();
      item->Reference = v;
      push_item_top(item);
      return item;
   }
   
   /* backward compatibility */
   void *PushTop(void *v) { return InsertTop(v); }
   void *push_top(void *v) { return InsertTop(v); }

/******************************************************************************
* Method: InsertBot
* Summary: Insert an element on the bottom of the list
* Description: 
* Return: Pointer to the removed element
******************************************************************************/
   void *InsertBot(void *v) {
      SpList_ITEM *item = NewItem();
      item->Reference = v;
      ItemPushBot(item);
      return item;
   }

   /* backward compatibility */
   void *PushBot(void *v) { return InsertBot(v); }
   void *push_bot(void *v) { return InsertBot(v); }

/******************************************************************************
* Method: Remove
* Summary: Remove an element from the top of the list (same as RemoveTop)
* Description: 
* Return: Pointer to the removed element
******************************************************************************/
  void *Remove() {return RemoveTop();}

/******************************************************************************
* Method: RemoveTop
* Summary: Remove an element from the top of the list
* Description: 
* Return: Pointer to the removed element
******************************************************************************/
   void *RemoveTop() {
      void *Reference;
      SpList_ITEM *item = pop_item_top();
      if (item) {
	 Reference = item->Reference;
	 DeleteItem(item);
      }else{
	 Reference = NULL;
      }
      return Reference;
   }
   
   /* backward compatibility */
   void *PopTop() { return RemoveTop(); }
   void *pop_top() { return RemoveTop(); }

/******************************************************************************
* Method: RemoveBot
* Summary: Remove an element from the bottom of the list
* Description: 
* Return: Pointer to the removed element
******************************************************************************/
   void *RemoveBot() {
      void *Reference;
      SpList_ITEM *item = pop_item_bot();
      if (item) {
	 Reference = item->Reference;
	 DeleteItem(item);
      }else{
	 Reference = NULL;
      }
      return Reference;
   }
   
   /* backward compatibility */
   void *PopBot() { return RemoveBot(); }
   void *pop_bot() { return RemoveBot(); }

   /*
   ** SpList_ITEM inline methods
   */

  private:
/******************************************************************************
* Method: RemoveItem
* Summary: Remove a list item from the top of the list (see RemoveTopItem)
* Description: 
* Return: Pointer to the removed list item
******************************************************************************/
   SpList_ITEM *RemoveItem(){return RemoveTopItem();}

/******************************************************************************
* Method: RemoveTopItem
* Summary: Remove a list item from the top of the list
* Description: 
* Return: Pointer to the removed list item
******************************************************************************/
   SpList_ITEM *RemoveTopItem(){
      SpList_ITEM *item;
      if (length == 0) {
	 return NULL;
      }else{
	 item = top;
	 if (current == top){
	   current = top->ForwardLink;
	 }
	 top = top->ForwardLink;
	 length--;
	 if (length == 0){
	   top = bot = current = NULL;
	 }
	 return item;
      }
   }
   
   /* backward compatibility */
   SpList_ITEM *ItemPopTop(){return RemoveTopItem();}
   SpList_ITEM *pop_item_top(){ return RemoveTopItem(); }

/******************************************************************************
* Method: RemoveBotItem
* Summary: Remove a list item from the bottom of the list
* Description: 
* Return: Pointer to the removed list item
******************************************************************************/
   SpList_ITEM *RemoveBotItem(){
      SpList_ITEM *item;
      if (length == 0) {
	 return NULL;
      }else{
	 item = bot;
	 if (current == bot){
	   current = bot->BackwardLink;
	 }
	 bot = bot->BackwardLink;
	 length--;
	 if (length == 0){
	   top = bot = current = NULL;
	 }
	 return item;
      }
   }
   
   /* backward compatibility */
   SpList_ITEM *ItemPopBot(){return RemoveBotItem();}
   SpList_ITEM *pop_item_bot(){ return RemoveBotItem(); }

/******************************************************************************
* Method: InsertItem
* Summary: Insert a list item to the top of the list (same as InsertTopItem)
* Description: 
* Return:
******************************************************************************/
   void InsertItem(SpList_ITEM *item){InsertTopItem(item);}

/******************************************************************************
* Method: InsertTopItem
* Summary: Insert a list item to the top of the list
* Description: 
* Return:
******************************************************************************/
   void InsertTopItem(SpList_ITEM *item){
      if (length == 0) {
	 bot = item;
	 top = item;
	 length = 1;
      }else{
	 top->BackwardLink = item;
	 item->ForwardLink = top;
	 top = item;
	 length++;
      }
   }
   
   /* backward compatibility */
   void ItemPushTop(SpList_ITEM *item){InsertTopItem(item);}
   void push_item_top(SpList_ITEM *item){ InsertTopItem(item); }

/******************************************************************************
* Method: InsertBotItem
* Summary: Insert a list item to the bottom of the list
* Description: 
* Return:
******************************************************************************/
   void InsertBotItem(SpList_ITEM *item){
      if (length == 0) {
	 bot = item;
	 top = item;
	 length = 1;
      }else{
	 bot->ForwardLink = item;
	 item->BackwardLink = bot;
	 bot = item;
	 length++;
      }
   }
   
   /* backward compatibility */
   void ItemPushBot(SpList_ITEM *item){InsertBotItem(item);}
   void push_item_bot(SpList_ITEM *item){InsertBotItem(item); }

  public:

/******************************************************************************
* Method: Remove
* Summary: Removes an element from the list
* Description: 
* Return:
******************************************************************************/
   void Remove(void *v);
   
   /* backward compatibility */
   void remove(void *v) { Remove(v); }

/******************************************************************************
* Method: PrintStream
* Summary: Prints summary information about the list to an ostream
* Description: 
* Return:
******************************************************************************/
   void PrintStream(ostream &out) {
      out << "SpList has " << length << " elements" << endl;
   }
   
   /*
   ** Free LIST Methods
   */
   
  private:
/******************************************************************************
* Method: NewItem
* Summary: Creates a new list item from the internal free list. The new list
*          item is retrieved from the free list instead of being newed. If
*          the free list is empty, then the list item is newed.
* Description: The list uses a free list to manage the list item classes
*              that contain the elements in the list. SpList items are
*              reused in the free list to improve performance
* Return: The new list item
******************************************************************************/
   virtual SpList_ITEM *NewItem() {
      SpList_ITEM *item;
      if (FreeList->get_length()) {
	item = FreeList->pop_item_top();
	item->Reset();
	return item;
      }else{
	item = new SpList_ITEM[SpList_FREE_SIZE];
	for (int i=2; i<SpList_FREE_SIZE; i++) {
	  item[i-1].ForwardLink = &item[i];
	}
	FreeList->set(SpList_FREE_SIZE-1, &item[1],&item[SpList_FREE_SIZE-1]);
	return &item[0];
      }
   }
   
/******************************************************************************
* Method: DeleteItem
* Summary: Deletes a list item using the internal free list. The list item is
*          returned to the free list instead of being actually deleted.
* Description: The list uses a free list to manage the list item classes
*              that contain the elements in the list.SpList items are
*              reused in the free list to improve performance
* Return: The new list item
******************************************************************************/
   void DeleteItem(SpList_ITEM *item) {FreeList->push_item_bot(item);}

  public:

   /*
   ** Overloaded operators
   */
   
/******************************************************************************
* Operator: ++
* Summary: iterator method that returns the next element in the list
* Description:
* Return: The next element in the list
******************************************************************************/
   void *operator ++ () {return get_next();}
   
/******************************************************************************
* Operator: --
* Summary: Returns the previous element in the list
* Description:
* Return: The previous element in the list (NULL if no previous element)
******************************************************************************/
   void *operator -- () {return get_previous();}


/******************************************************************************
* Operator: +=
* Summary: Inserts an element to the bottom of the list
* Description:
* Return: The list item container
******************************************************************************/
   void *operator += (void *v) {
      return InsertBot(v);
   }
/******************************************************************************
* Operator: -=
* Summary: Removes an element from the list
* Description:
* Return:
******************************************************************************/
   void operator -= (void *v) {
      Remove(v);
   }
   
/******************************************************************************
* Constructor: SpList
* Summary: Constructs the SpList (no arguments required)
* Description:
* Return:
******************************************************************************/
   SpList();
   
 private:

/******************************************************************************
* Internal Method: set
* Summary: Set the length, and the top and bottom pointers
* Description: 
* Return: None
******************************************************************************/
   void set(int l,        // new length
	    SpList_ITEM *t, //
	    SpList_ITEM *b  //
	    ) {
      length = l; top = t; bot = b;
   }

   void SetNumElements(int l) {length = l;}
   void SetLength(int l) {length = l;}
   void set_length(int l) {length = l;}

   int length;			// length of the queue
   SpList_ITEM *top;		// pointer to the top of the queue
   SpList_ITEM *bot;		// pointer to the bottom of the queue
   SpList_ITEM *current;          // current pointer in the list

   static SpList *FreeList;    // free list items
   static int done;		// done flag for first time

};

typedef SpList List;

/*
typedef SpList LIST;

	y.tab.C is generated from Yacc.Y which includes SpHashTree.H
which includes SpBinaryTree.H which includes SpPoDataBase.H which
includes SpList.H which says:

typedef SpList LIST;

	And y.tab.C says:

#define LIST    310

	Which means SpList.H is equivalent to saying:

typedef SpList 310;

	Which is a syntax error.  Therefore I commented the typedef
out.  But that is bad since it makes SPEEDES be backwards
incompatible.  So we need another solution.
	I think I can blame the author of pi for not following the
naming conventions by making a #define name not start with something
like "Sp" or "Pi".  Anyway, bison has an option which allows you to
specify a prefix for about 5 functions, but NOT #define names.  I
cannot find any other option to bison which could easily solve this
problem.  We might be able to have SpPoDataBase.H stop including
SpList.H, if we made about 3 functions in SpPoDataBase.H be non-inline.
But there may be other paths to the including of SpList.H, and the
other paths may be harder to avoid.
*/

#endif
