// SpNameValueList.C file

#include "SpStdlib.H"
#include "SpNetTypes.H"
#include "SpNameValueList.H"

static int NameValueCompare(const void *n1, const void *n2) {

  NAME_VALUE *NameValue1 = (NAME_VALUE *)n1;
  NAME_VALUE *NameValue2 = (NAME_VALUE *)n2;

  if (NameValue1->Name == NULL) {
    if (NameValue2->Name == NULL) {
      return 0;
    }else{
      return -1;
    }
  }else{
    if (NameValue2->Name == NULL) {
      return 1;
    }else{
      int status = strcmp(NameValue1->Name, NameValue2->Name);
      if (status > 0) return 1;
      if (status < 0) return -1;
      return 0;
    }
  }

}

SpNameValueList::SpNameValueList(int n) {
  NameOfList = strdup("UnNamedList");
  NameValue = new NAME_VALUE[n];
  Nitems = n;
  Iterator = 0;
}

SpNameValueList::~SpNameValueList() {
  if (NameOfList) free(NameOfList);
  if (NameValue) delete [] NameValue;
}

void SpNameValueList::Resize(int s) {
  int i;

  NAME_VALUE *NewNameValue = new NAME_VALUE[s];

  for (i=0; i<Nitems; i++) {
    NewNameValue[i].Name = NameValue[i].Name;
    NameValue[i].Name = NULL;
    NewNameValue[i].Value = NameValue[i].Value;
  }

  if (NameValue) delete [] NameValue;
  NameValue = NewNameValue;
  Nitems = s;

}

void SpNameValueList::SetNameValue(int i, char *N, double V) {
  if (i < Nitems) {
    if (NameValue[i].Name) free(NameValue[i].Name);
    NameValue[i].Name = strdup(N);
    NameValue[i].Value = V;
  }else{
    cout << "Error in SpNameValueList: bad index " << i << endl;
  }
}

void SpNameValueList::SetNameValue(char *N, double V) {
  if (Iterator >= Nitems) Resize(2*Nitems + 1);
  SetNameValue(Iterator, N, V);
  Iterator++;
}

void SpNameValueList::PackBuffer(char *Buffer, int Size) {
  char *Buff = Buffer;
  NET_INT tNetInt;
  NET_DOUBLE tNetDouble;

  strcpy(Buff, NameOfList);
  Buff += strlen(NameOfList) + 1;

  tNetInt = Nitems;
  memcpy(Buff, (char *)&tNetInt, sizeof(NET_INT));
  Buff += sizeof(NET_INT);

  tNetInt = Size; 
  memcpy(Buff, (char *)&tNetInt, sizeof(NET_INT));
  Buff += sizeof(NET_INT);

  for (int i=0; i<Nitems; i++) {
    if (NameValue[i].Name) {
      tNetInt = i; 
      memcpy(Buff, (char *)&tNetInt, sizeof(NET_INT));
      Buff += sizeof(NET_INT);

      strcpy(Buff, NameValue[i].Name);
      Buff += strlen(NameValue[i].Name) + 1;

      tNetDouble = NameValue[i].Value;
      memcpy(Buff, (char *)(&tNetDouble), sizeof(NET_DOUBLE));
      Buff += sizeof(NET_DOUBLE);
    }
  }

  if (Buff-Buffer - Size) {
    cout << "Error, bad NameValueList packing" << endl;
  }

}

void SpNameValueList::UnPackBuffer(char *Buffer) {
  char *Buff = Buffer;
  int Size;
  int Index;
  NET_INT tNetInt;
  NET_DOUBLE tNetDouble;

  if (NameOfList) free(NameOfList);
  if (NameValue) delete [] NameValue;

  NameOfList = strdup(Buff);
  Buff += strlen(Buffer) + 1;

  memcpy((char *)&tNetInt, Buff, sizeof(NET_INT));
  Nitems = tNetInt;
  Buff += sizeof(NET_INT);

  memcpy((char *)&tNetInt, Buff, sizeof(NET_INT));
  Size = tNetInt; 
  Buff += sizeof(NET_INT);

  NameValue = new NAME_VALUE[Nitems];
  Iterator = 0;

  while (Buff-Buffer < Size) {
    memcpy((char *)&tNetInt, Buff, sizeof(NET_INT));
    Index = tNetInt; 
    Buff += sizeof(NET_INT);

    NameValue[Index].Name = strdup(Buff);
    Buff += strlen(Buff) + 1;

    memcpy((char *)&tNetDouble, Buff, sizeof(NET_DOUBLE));
    NameValue[Index].Value = tNetDouble;
    Buff += sizeof(NET_DOUBLE);
  }

}

void SpNameValueList::Sort() {
  qsort((char *)NameValue, Nitems, sizeof(NAME_VALUE), NameValueCompare);
}
