// SpNameValueList.H header file that defines a class that holds a variable
// number of named values

#ifndef SpNameValueList_OBJECT
#define SpNameValueList_OBJECT

#include "SpString.H"
#include "SpNetTypes.H"
#include "SpStdlib.H"


class NAME_VALUE {
  public:
    char *Name;
    double Value;
    NAME_VALUE() {Name = NULL; Value = 0.0;}
    ~NAME_VALUE() {if (Name) free(Name);}
};

class SpNameValueList {

  private:

    char *NameOfList;
    NAME_VALUE *NameValue;
    int Nitems;
    int Iterator;

  protected:

  public:

    SpNameValueList() {
      NameOfList = strdup("UnNamedList");
      NameValue = NULL;
      Nitems = 0;
      Iterator = 0;
    }

    SpNameValueList(char *Buffer) {
      NameOfList = NULL;
      NameValue = NULL;
      Nitems = 0;
      Iterator = 0;
      UnPackBuffer(Buffer);
    }

    SpNameValueList(int n);

    ~SpNameValueList();

    void SetNitems(int n) {
      if (NameValue) delete [] NameValue;
      NameValue = new NAME_VALUE[n];
    }

    void SetNameOfList(char *NOL) {
      if (NameOfList) free(NameOfList);
      NameOfList = strdup(NOL);
    }

    void SetNameValue(int i, char *N, double V);
    void SetNameValue(char *N, double V);
    void SetName(char *N) {
      SetNameValue(N, 12345e67);
    }

    void ResetIterator() {Iterator = 0;}
    void Resize(int s);

    char *GetNameOfList() {return NameOfList;}
    char *GetName(int i) {return NameValue[i].Name;}
    double GetValue(int i) {return NameValue[i].Value;}

    void GetNameValue(int i, char *&N, double &V) {
      N = NameValue[i].Name;
      V = NameValue[i].Value;
    }

    void GetNameValue(char *&N, double &V) {
      if (Iterator < Nitems) {
	N = NameValue[Iterator].Name;
	V = NameValue[Iterator].Value;
	Iterator++;
      }else{
	N = NULL;
	V = -1.0e20;
      }
    }

    double GetValue(char *N, int &NameFound) {
      double V = 0.0;
      NameFound = 0;
      int i;
      for (i=0; i<Nitems; i++) {
	if (!strcmp(N,NameValue[i].Name)) {
	  V = NameValue[i].Value;
	  NameFound = 1;
	  break;
	}
      }
      if (i== Nitems) {
	cout << "Error in SpNameValueList: bad name " << N << endl;
      }
      return V;
    }

    int GetNitems() {return Nitems;}

    int Allign8Bytes(int size){ return size + (8 - size%8);}

    int GetSizeOfBuffer() {
      int Size = strlen(NameOfList) + 1 + 2*sizeof(NET_INT);
      int i;
      for (i=0; i<Nitems; i++) {
	if (NameValue[i].Name) Size += sizeof(NET_INT)
				 + strlen(NameValue[i].Name) + 1
				 + sizeof(NET_DOUBLE);
      }
      return Size;
    }

    char *MakeBuffer(int &Size) {
      if (Size == 0) Size = GetSizeOfBuffer();
      Size = GetSizeOfBuffer();
      char *Buffer = new char[Size];
      PackBuffer(Buffer, Size);
      return Buffer;
    }

    void PackBuffer(char *Buffer, int Size);
    void UnPackBuffer(char *Buffer);

    void Print(ostream &out = cout) {
      out << "Printing NameList: " << NameOfList << endl;
      int i;
      for (i=0; i<Nitems; i++) {
	if (NameValue[i].Name) {
	  out << "[" << i << "] " << NameValue[i].Name;
	  if (NameValue[i].Value != 12345e67) {
	   out << " = "
	       << NameValue[i].Value;
	  }
	  out << endl;
	}
      }
    }

    void Sort();

};

typedef SpNameValueList NAME_VALUE_LIST;

#endif
