// SpParmSetElement source file

#include "SpParmSet.H"

SpParmSetElement::SpParmSetElement () {
  Data = NULL;
}

SpParmSetElement::SpParmSetElement (int val) {
  Data = new SpParmSetIntData(val);
}

SpParmSetElement::SpParmSetElement (float val) {
  Data = new SpParmSetFloatData(val);
}

SpParmSetElement::SpParmSetElement (double val) {
  Data = new SpParmSetDoubleData(val);
}

SpParmSetElement::SpParmSetElement (char* val) {
  Data = new SpParmSetStringData(val);
}

SpParmSetElement::SpParmSetElement (char* val, int size) {
  Data = new SpParmSetBufferData(val, size);
}

SpParmSetElement::SpParmSetElement (const SpParmSetElement& elem) {
  if(elem.Data)
    Data = elem.Data->Clone();
  else
    Data = NULL;
}

SpParmSetElement::~SpParmSetElement () {
  if(Data)  delete Data;
}

int
SpParmSetElement::GetPackLength (SpParmSetPackMethod pack) {
  if(!Data) return 0;
  return Data->GetPackLength(pack);
}
#ifdef SPPARMSET_USE_XDR
void
SpParmSetElement::PackBuffer (SpParmSetPackMethod pack,
			      char* buff, int& offset,
			      XDR* xdrHandle) {
#else
void
SpParmSetElement::PackBuffer (SpParmSetPackMethod pack,
			      char* buff, int& offset) {
#endif
  if(Data)
#ifdef SPPARMSET_USE_XDR
    Data->PackBuffer(pack, buff, offset, xdrHandle);
#else
    Data->PackBuffer(pack, buff, offset);
#endif
  else {    // need to account for odd case where element has no type
    NET_INT netnone;
    int intnone;
    switch(pack) {
    case PACK_NET_TYPES:
      netnone = (int) NONE_DATA;
      memcpy(buff+offset, (char*)&netnone, sizeof(NET_INT));
      offset += sizeof(NET_INT);
      break;
    case PACK_NATIVE:
      intnone = (int) NONE_DATA;
      memcpy(buff+offset, (char*)&intnone, sizeof(int));
      offset += sizeof(int);
      break;
    case PACK_XDR:
#ifdef SPPARMSET_USE_XDR
      intnone = (int) NONE_DATA;
      xdr_int(xdrHandle, &intnone);
#endif
      break;
    }
  }
}
#ifdef SPPARMSET_USE_XDR
void
SpParmSetElement::UnpackBuffer (SpParmSetPackMethod pack,
				char* buff, int& offset,
				XDR* xdrHandle) {

  Data = SpParmSetDataItem::UnpackBuffer(pack, buff, offset, xdrHandle);
}
#else
void
SpParmSetElement::UnpackBuffer (SpParmSetPackMethod pack,
				char* buff, int& offset) {

  Data = SpParmSetDataItem::UnpackBuffer(pack, buff, offset);
}
#endif

/****************************/
#ifdef SPPARMSET_USE_XDR
SpParmSetDataItem*
SpParmSetDataItem::UnpackBuffer (SpParmSetPackMethod pack, 
				 char* buff, int& offset,
				 XDR* xdrHandle) {
#else
SpParmSetDataItem*
SpParmSetDataItem::UnpackBuffer (SpParmSetPackMethod pack, 
				 char* buff, int& offset) {
#endif
  NET_INT netint;
  int normint;
  SpParmSetDataType datatype;

  switch(pack) {
  case PACK_NET_TYPES:
    memcpy((char*)&netint, buff+offset, sizeof(NET_INT));
    offset += sizeof(NET_INT);
    datatype = (SpParmSetDataType)(int) netint;
    break;
  case PACK_NATIVE:
    memcpy((char*)&normint, buff+offset, sizeof(int));
    offset += sizeof(int);
    datatype = (SpParmSetDataType) normint;
    break;
  case PACK_XDR:
#ifdef SPPARMSET_USE_XDR
    xdr_enum(xdrHandle, (int*)&datatype);
#endif
    break;
  }
#ifdef SPPARMSET_USE_XDR
  switch(datatype) {
  case INT_DATA:
    return new SpParmSetIntData(pack, buff, offset, xdrHandle);
  case FLOAT_DATA:
    return new SpParmSetFloatData(pack, buff, offset, xdrHandle);
  case DOUBLE_DATA:
    return new SpParmSetDoubleData(pack, buff, offset, xdrHandle);
  case STRING_DATA:
    return new SpParmSetStringData(pack, buff, offset, xdrHandle);
  case BUFFER_DATA:
    return new SpParmSetBufferData(pack, buff, offset, xdrHandle);
  case NONE_DATA:
    break;
  }
#else
  switch(datatype) {
  case INT_DATA:
    return new SpParmSetIntData(pack, buff, offset);
  case FLOAT_DATA:
    return new SpParmSetFloatData(pack, buff, offset);
  case DOUBLE_DATA:
    return new SpParmSetDoubleData(pack, buff, offset);
  case STRING_DATA:
    return new SpParmSetStringData(pack, buff, offset);
  case BUFFER_DATA:
    return new SpParmSetBufferData(pack, buff, offset);
  case NONE_DATA:
    break;
  }
#endif
  return NULL;
}

/****************************/
#ifdef SPPARMSET_USE_XDR
SpParmSetIntData::SpParmSetIntData (SpParmSetPackMethod pack, 
				    char* buff, int& offset,
				    XDR* xdrHandle) {
#else
SpParmSetIntData::SpParmSetIntData (SpParmSetPackMethod pack, 
				    char* buff, int& offset) {
#endif
  NET_INT netint;
  int normint;

  switch(pack) {
  case PACK_NET_TYPES:
    memcpy((char*)&netint, buff+offset, sizeof(NET_INT));
    offset += sizeof(NET_INT);
    Value = netint;
    break;
  case PACK_NATIVE:
    memcpy((char*)&normint, buff+offset, sizeof(int));
    offset += sizeof(int);
    Value = normint;
    break;
  case PACK_XDR:
#ifdef SPPARMSET_USE_XDR
    xdr_int(xdrHandle, &Value);
#endif
    break;
  }
}

int
SpParmSetIntData::GetPackLength (SpParmSetPackMethod pack) {
  switch(pack) {
  case PACK_NET_TYPES:
    return sizeof(NET_INT) * 2;  // includes type data, not decoded here
  case PACK_NATIVE:
    return sizeof(int) * 2;      // includes type data, not decoded here
  case PACK_XDR:
    return sizeof(int) * 2;      // includes type data, not decoded here
  }
  return 0;
}
#ifdef SPPARMSET_USE_XDR
void
SpParmSetIntData::PackBuffer (SpParmSetPackMethod pack, 
			      char* buff, int& offset,
			      XDR* xdrHandle) {
#else
void
SpParmSetIntData::PackBuffer (SpParmSetPackMethod pack, 
			      char* buff, int& offset) {
#endif
  NET_INT netint;
  int normint;

  switch(pack) {
  case PACK_NET_TYPES:
    netint = (int) INT_DATA;
    memcpy(buff+offset, (char*)&netint, sizeof(NET_INT));
    offset += sizeof(NET_INT);
    netint = Value;
    memcpy(buff+offset, (char*)&netint, sizeof(NET_INT));
    offset += sizeof(NET_INT);
    break;
  case PACK_NATIVE:
    normint = (int) INT_DATA;
    memcpy(buff+offset, (char*)&normint, sizeof(int));
    offset += sizeof(int);
    memcpy(buff+offset, (char*)&Value, sizeof(int));
    offset += sizeof(int);
    break;
  case PACK_XDR:
#ifdef SPPARMSET_USE_XDR
    normint = (int) INT_DATA;
    xdr_enum(xdrHandle, &normint);
    xdr_int(xdrHandle, &Value);
#endif
    break;
  }
}
#ifdef SPPARMSET_USE_XDR
SpParmSetFloatData::SpParmSetFloatData (SpParmSetPackMethod pack, 
					char* buff, int& offset,
					XDR* xdrHandle) {
#else
SpParmSetFloatData::SpParmSetFloatData (SpParmSetPackMethod pack, 
					char* buff, int& offset) {
#endif
  NET_DOUBLE netdouble;
  float normfloat;

  switch(pack) {
  case PACK_NET_TYPES:
    memcpy((char*)&netdouble, buff+offset, sizeof(NET_DOUBLE));
    offset += sizeof(NET_DOUBLE);
    Value = (float) netdouble;
    break;
  case PACK_NATIVE:
    memcpy((char*)&normfloat, buff+offset, sizeof(int));
    offset += sizeof(float);
    Value = normfloat;
    break;
  case PACK_XDR:
#ifdef SPPARMSET_USE_XDR
    xdr_float(xdrHandle, &Value);
#endif
    break;
  }
}

int
SpParmSetFloatData::GetPackLength (SpParmSetPackMethod pack) {
  switch(pack) {
  case PACK_NET_TYPES:
    return sizeof(NET_INT) + sizeof(NET_DOUBLE);  // includes type data
  case PACK_NATIVE:
    return sizeof(int) + sizeof(float);           // includes type data
  case PACK_XDR:
    return sizeof(int) + sizeof(float);           // includes type data
  }
  return 0;
}
#ifdef SPPARMSET_USE_XDR
void
SpParmSetFloatData::PackBuffer (SpParmSetPackMethod pack, 
				char* buff, int& offset,
				XDR* xdrHandle) {
#else
void
SpParmSetFloatData::PackBuffer (SpParmSetPackMethod pack, 
				char* buff, int& offset) {
#endif
  NET_INT netdata;
  int normdata;
  NET_DOUBLE netdouble;

  switch(pack) {
  case PACK_NET_TYPES:
    netdata = (int) FLOAT_DATA;
    memcpy(buff+offset, (char*)&netdata, sizeof(NET_INT));
    offset += sizeof(NET_INT);
    netdouble = Value;
    memcpy(buff+offset, (char*)&netdouble, sizeof(NET_DOUBLE));
    offset += sizeof(NET_DOUBLE);
    break;
  case PACK_NATIVE:
    normdata = (int) FLOAT_DATA;
    memcpy(buff+offset, (char*)&normdata, sizeof(int));
    offset += sizeof(int);
    memcpy(buff+offset, (char*)&Value, sizeof(float));
    offset += sizeof(float);
    break;
  case PACK_XDR:
#ifdef SPPARMSET_USE_XDR
    normdata = (int) FLOAT_DATA;
    xdr_enum(xdrHandle, (int*)&normdata);
    xdr_float(xdrHandle, &Value);
#endif
    break;
  }
}
#ifdef SPPARMSET_USE_XDR
SpParmSetDoubleData::SpParmSetDoubleData (SpParmSetPackMethod pack, 
					  char* buff, int& offset,
					  XDR* xdrHandle) {
#else
SpParmSetDoubleData::SpParmSetDoubleData (SpParmSetPackMethod pack, 
					  char* buff, int& offset) {
#endif
  NET_DOUBLE netdouble;
  double normdouble;

  switch(pack) {
  case PACK_NET_TYPES:
    memcpy((char*)&netdouble, buff+offset, sizeof(NET_DOUBLE));
    offset += sizeof(NET_DOUBLE);
    Value = netdouble;
    break;
  case PACK_NATIVE:
    memcpy((char*)&normdouble, buff+offset, sizeof(double));
    offset += sizeof(double);
    Value = normdouble;
    break;
  case PACK_XDR:
#ifdef SPPARMSET_USE_XDR
    xdr_double(xdrHandle, &Value);
#endif
    break;
  }
}

int
SpParmSetDoubleData::GetPackLength (SpParmSetPackMethod pack) {
  switch(pack) {
  case PACK_NET_TYPES:
    return sizeof(NET_INT) + sizeof(NET_DOUBLE);  // includes type data
  case PACK_NATIVE:
    return sizeof(int) + sizeof(double);  // includes type data
  case PACK_XDR:
    return sizeof(int) + sizeof(double);  // includes type data
  }
  return 0;
}
#ifdef SPPARMSET_USE_XDR
void
SpParmSetDoubleData::PackBuffer (SpParmSetPackMethod pack, 
				 char* buff, int& offset,
				 XDR* xdrHandle) {
#else
void
SpParmSetDoubleData::PackBuffer (SpParmSetPackMethod pack, 
				 char* buff, int& offset) {
#endif
  NET_INT netdata;
  int intdata;
  NET_DOUBLE netdouble;

  switch(pack) {
  case PACK_NET_TYPES:
    netdata = (int) DOUBLE_DATA;
    memcpy(buff+offset, (char*)&netdata, sizeof(NET_INT));
    offset += sizeof(NET_INT);
    netdouble = Value;
    memcpy(buff+offset, (char*)&netdouble, sizeof(NET_DOUBLE));
    offset += sizeof(NET_DOUBLE);
    break;
  case PACK_NATIVE:
    intdata = (int) DOUBLE_DATA;
    memcpy(buff+offset, (char*)&intdata, sizeof(int));
    offset += sizeof(int);
    memcpy(buff+offset, (char*)&Value, sizeof(double));
    offset += sizeof(double);
    break;
  case PACK_XDR:
#ifdef SPPARMSET_USE_XDR
    intdata = (int) DOUBLE_DATA;
    xdr_enum(xdrHandle, (int*)&intdata);
    xdr_double(xdrHandle, &Value);
#endif
    break;
  }
}
#ifdef SPPARMSET_USE_XDR
SpParmSetStringData::SpParmSetStringData (SpParmSetPackMethod pack, 
					  char* buff, int& offset,
					  XDR* xdrHandle) {
#else
SpParmSetStringData::SpParmSetStringData (SpParmSetPackMethod pack, 
					  char* buff, int& offset) {
#endif
  int strlength = strlen(buff+offset)+1;

  switch(pack) {
  case PACK_NET_TYPES:
    Value = new char[strlength];
    strcpy(Value, buff+offset);
    offset += strlength;
    break;
  case PACK_NATIVE:
    Value = new char[strlength];
    strcpy(Value, buff+offset);
    offset += strlength;
    break;
  case PACK_XDR:
#ifdef SPPARMSET_USE_XDR
    xdr_int(xdrHandle, &strlength);
    Value = new char[strlength];
    xdr_string(xdrHandle, &Value, strlength);
#endif
    break;
  }
}

int
SpParmSetStringData::GetPackLength (SpParmSetPackMethod pack) {
  int strlength = strlen(Value)+1;
  switch(pack) {
  case PACK_NET_TYPES:
    return sizeof(NET_INT) + strlength;
  case PACK_NATIVE:
    return sizeof(int) + strlength;
  case PACK_XDR:
#ifdef SPPARMSET_USE_XDR
    return sizeof(int) * 2 + strlength;
#else
    return 0;
#endif
  }
  return 0;
}
#ifdef SPPARMSET_USE_XDR
void
SpParmSetStringData::PackBuffer (SpParmSetPackMethod pack, 
				 char* buff, int& offset,
				 XDR* xdrHandle) {
#else
void
SpParmSetStringData::PackBuffer (SpParmSetPackMethod pack, 
				 char* buff, int& offset) {
#endif
  NET_INT netdata;
  int intdata;
  int strlength = strlen(Value)+1;

  switch(pack) {
  case PACK_NET_TYPES:
    netdata = (int) STRING_DATA;
    memcpy(buff+offset, (char*)&netdata, sizeof(NET_INT));
    offset += sizeof(NET_INT);
    memcpy(buff+offset, Value, strlength);
    offset += strlength;
    break;
  case PACK_NATIVE:
    intdata = (int) STRING_DATA;
    memcpy(buff+offset, (char*)&intdata, sizeof(int));
    offset += sizeof(int);
    memcpy(buff+offset, Value, strlength);
    offset += strlength;
    break;
  case PACK_XDR:
#ifdef SPPARMSET_USE_XDR
    intdata = (int) STRING_DATA;
    xdr_enum(xdrHandle, (int*)&intdata);
    xdr_int(xdrHandle, &strlength);
    xdr_string(xdrHandle, &Value, strlength);
#endif
    break;
  }
}
#ifdef SPPARMSET_USE_XDR
SpParmSetBufferData::SpParmSetBufferData (SpParmSetPackMethod pack, 
					  char* buff, int& offset,
					  XDR* xdrHandle) {
#else
SpParmSetBufferData::SpParmSetBufferData (SpParmSetPackMethod pack, 
					  char* buff, int& offset) {
#endif
  NET_INT netint;
  int normint;

  switch(pack) {
  case PACK_NET_TYPES:
    memcpy((char*)&netint, buff+offset, sizeof(NET_INT));
    offset += sizeof(NET_INT);
    Size = netint;
    Value = new char[Size];
    memcpy(Value, buff+offset, Size);
    offset += Size;
    break;
  case PACK_NATIVE:
    memcpy((char*)&normint, buff+offset, sizeof(int));
    offset += sizeof(int);
    Size = normint;
    Value = new char[Size];
    memcpy(Value, buff+offset, Size);
    offset += Size;
    break;
  case PACK_XDR:
#ifdef SPPARMSET_USE_XDR
    xdr_u_int(xdrHandle, &Size);
    Value = new char[Size];
    xdr_bytes(xdrHandle, &Value, &Size, Size);
#endif
    break;
  }
}

int
SpParmSetBufferData::GetPackLength (SpParmSetPackMethod pack) {
  switch(pack) {
  case PACK_NET_TYPES:
    return sizeof(NET_INT) * 2 + Size;
  case PACK_NATIVE:
    return sizeof(int) * 2 + Size;
  case PACK_XDR:
#ifdef SPPARMSET_USE_XDR
    return sizeof(int) * 2 + Size;
#else
    return 0;
#endif
  }
  return 0;
}
#ifdef SPPARMSET_USE_XDR
void
SpParmSetBufferData::PackBuffer (SpParmSetPackMethod pack, 
				 char* buff, int& offset,
				 XDR* xdrHandle) {
#else
void
SpParmSetBufferData::PackBuffer (SpParmSetPackMethod pack, 
				 char* buff, int& offset) {
#endif
  NET_INT netint;
  int normint;

  switch(pack) {
  case PACK_NET_TYPES:
    netint = (int) BUFFER_DATA;
    memcpy(buff+offset, (char*)&netint, sizeof(NET_INT));
    offset += sizeof(NET_INT);
    netint = Size;
    memcpy(buff+offset, (char*)&netint, sizeof(NET_INT));
    offset += sizeof(NET_INT);
    memcpy(buff+offset, Value, Size);
    offset += Size;
    break;
  case PACK_NATIVE:
    normint = (int) BUFFER_DATA;
    memcpy(buff+offset, (char*)&normint, sizeof(int));
    offset += sizeof(int);
    memcpy(buff+offset, (char*)&Size, sizeof(int));
    offset += sizeof(int);
    memcpy(buff+offset, Value, Size);
    offset += Size;
    break;
  case PACK_XDR:
#ifdef SPPARMSET_USE_XDR
    normint = (int) BUFFER_DATA;
    xdr_enum(xdrHandle, (int*)&normint);
    xdr_u_int(xdrHandle, &Size);
    xdr_bytes(xdrHandle, &Value, &Size, Size);
#endif
    break;
  }
}
