// SpPriorityQueueHeap.C method file

#include "SpStdio.H"
#include "SpStdlib.H"
#include "SpMath.H"
#include "SpPriorityQueueHeap.H"

#ifndef INFINITY
#define INFINITY 1.0e20
#endif
#define TEMPQ

C_QUEUE *SpPriorityQueueHeap::qfree = NULL;
C_QUEUE *SpPriorityQueueHeap::qfree_holder = NULL;
int SpPriorityQueueHeap::done = 0;
int SpPriorityQueueHeap::n_created = 0;
int SpPriorityQueueHeap::n_created_holder = 0;
C_QUEUE SpPriorityQueueHeap::ToBeCleared;

/************************************************************************
* SpPriorityQueueHeap : construct a rqueue object					*
************************************************************************/
SpPriorityQueueHeap::SpPriorityQueueHeap() {
  n = 0;


  n_inserts = 0;
  n_pops = 0;
  n_loops = 0;
  sort_size = 0;
  n_sorts = 0;
  insert_size = 0;
  Ncomb = NCOMB;

  min_time = 1.0e20;

  if (done) return;
  done = 1;

  n_created = 0;
  qfree = new C_QUEUE();

  qfree_holder = new C_QUEUE();
  n_created_holder = 0;

}

SpPriorityQueueHeap::~SpPriorityQueueHeap()
{
// There is a serious bug with this - it screws up dynamic object creation
//  ToBeCleared.clear();
}

/************************************************************************
* init : initialize the rqueue						*
************************************************************************/
void SpPriorityQueueHeap::init() {
  C_SQ_ITEM *it;

  it = pop();
  if (it) insert(it);
  /*
  it = free_new();
  free_delete(it);
  */

  n_inserts = 0;
  n_loops = 0;
  sort_size = 0;
  n_sorts = 0;
  insert_size = 0;
  n_pops = 0;

}

/************************************************************************
* insert : insert an item into the rqueue				*
************************************************************************/
void SpPriorityQueueHeap::insert(C_SQ_ITEM *it) {

  n++;

#ifdef TEMPQ

  temp_queue.push_bot(it);
  if (it->get_time_tag() < min_time) min_time = it->get_time_tag();

#else

  C_SQ_ITEM *new_it;
  C_SQ_ITEM *next_it;
  C_QUEUE *q;

  if (queue.get_length() > Ncomb) {
    new_it = free_new();
    next_it = (C_SQ_ITEM *)queue.get_top();
    new_it->set_time_tag(next_it->get_time_tag());
    q = new_it->get_queue();
    q->init(queue.get_top(), queue.get_bot(), queue.get_length());
    queue.init(new_it, new_it, 1);
  }

  insert_size += queue.get_length();
  n_inserts++;
  queue.insert(it);

#endif
}

/************************************************************************
* pop : pop the next item out of the speedes queue			*
************************************************************************/
C_SQ_ITEM *SpPriorityQueueHeap::pop() {
  C_SQ_ITEM *new_it;
  C_SQ_ITEM *new_it1;
  C_SQ_ITEM *next_it;
  C_SQ_ITEM *next_it1;
  C_SQ_ITEM *it;
  C_QUEUE *q;
  C_QUEUE *q1;

//...... check if the queue is empty

  if (queue.get_length() == 0) {

    if (temp_queue.get_length() == 0) {
      return NULL;
    }

    n_pops++;

    n_sorts++;
    sort_size += temp_queue.get_length();
    temp_queue.sort();

    queue.init(	temp_queue.get_top(),
		temp_queue.get_bot(),
		temp_queue.get_length());

    temp_queue.reset();
    min_time = 1.0e20;

    it = (C_SQ_ITEM *)queue.pop_top();
    n--;
    return it;

  }

  n_pops++;

//...... check if temp queue has the next event

  it = (C_SQ_ITEM *)queue.get_top();
  if (min_time < it->get_time_tag()) {

    n_sorts++;
    sort_size += temp_queue.get_length();
    temp_queue.sort();

    it = (C_SQ_ITEM *)temp_queue.pop_top();

    if (queue.get_length() > Ncomb) {
      new_it = free_new();
      next_it = (C_SQ_ITEM *)queue.get_top();
      new_it->set_time_tag(next_it->get_time_tag());
      q = new_it->get_queue();
      q->init(queue.get_top(), queue.get_bot(), queue.get_length());
      queue.init(new_it, new_it, 1);
    }

    if (temp_queue.get_length() > Ncomb) {

      new_it = free_new();
      next_it = (C_SQ_ITEM *)temp_queue.get_top();
      new_it->set_time_tag(next_it->get_time_tag());
      q = new_it->get_queue();
      q->init(	temp_queue.get_top(),
		temp_queue.get_bot(),
		temp_queue.get_length());
      insert_size += queue.get_length();
      n_inserts++;
      queue.insert(new_it);
      temp_queue.reset();

    }else{

      if (temp_queue.get_length()) {
        queue.merge(&temp_queue);
      }

    }

    min_time = 1.0e20;
    n--;
    return it;

  }


//...... get the top item off of the main queue

  it = (C_SQ_ITEM *)queue.pop_top();

//...... create a new item if necessary

  if (queue.get_length() > Ncomb) {
    new_it = free_new();
    next_it = (C_SQ_ITEM *)queue.get_top();
    new_it->set_time_tag(next_it->get_time_tag());
    q = new_it->get_queue();
    q->init(queue.get_top(), queue.get_bot(), queue.get_length());
    queue.init(new_it, new_it, 1);
  }

//...... untangle the recursive queue

  q = it->get_queue();
  while (q->get_length() != 0) {

    new_it = (C_SQ_ITEM *)q->pop_top();

    if (q->get_length()) {

      n_loops++;

      if (q->get_length() == 1) {
        insert_size += queue.get_length();
        n_inserts += q->get_length();
	queue.insert(q->pop_top());
        free_delete(it);
      }else{
        next_it = (C_SQ_ITEM *)q->get_top();
        it->set_time_tag(next_it->get_time_tag());
        insert_size += queue.get_length();
        n_inserts++;
        queue.insert(it);
      }

      if (queue.get_length() > Ncomb) {
        new_it1 = free_new();
        next_it1 = (C_SQ_ITEM *)queue.get_top();
        new_it1->set_time_tag(next_it1->get_time_tag());
        q1 = new_it1->get_queue();
        q1->init(queue.get_top(), queue.get_bot(), queue.get_length());
        queue.init(new_it1, new_it1, 1);
      }

    }else{

      free_delete(it);

    }

    it = new_it;
    q = new_it->get_queue();

  }

  n--;
  return it;

}

/************************************************************************
* get_time : get the time of the next event in the speedes queue	*
************************************************************************/
SpSimTime& SpPriorityQueueHeap::get_time() { //GEB2
  C_SQ_ITEM *it = (C_SQ_ITEM *)queue.get_top();

  if (it == NULL) {
    return min_time;
  }else{
    if (it->get_time_tag() < min_time) {
      return it->get_time_tag();
    }else{
      return min_time;
    }
  }
}

struct SQ_ITEM_ARRAY_OF_CHILDREN : public C_ITEM
{
    SQ_ITEM_ARRAY_OF_CHILDREN(C_SQ_ITEM *a) : A(a) {}
    ~SQ_ITEM_ARRAY_OF_CHILDREN() { delete [] A; }
  private:
    C_SQ_ITEM *A;
};

/************************************************************************
* free_new : create a new item for the recursive queue			*
************************************************************************/
C_SQ_ITEM *SpPriorityQueueHeap::free_new() {
  C_SQ_ITEM *it;
  int i;

  if (qfree->get_length()) {
    return (C_SQ_ITEM *)qfree->pop_top();
  }

  it = new C_SQ_ITEM[100];
  ToBeCleared.push_bot(new SQ_ITEM_ARRAY_OF_CHILDREN(it));
  if (it == NULL) {
    fprintf(stderr,"Error (RQUEUE) no more memory\n");
    exit(1);
  }
  n_created += 100;

  for (i=1; i<99; i++) {
    it[i].set_link(&it[i+1]);
    qfree->init(&it[1], &it[99], 99);
  }

  return it;

}

struct RQ_HOLDER_ARRAY_OF_CHILDREN : public C_ITEM
{
    RQ_HOLDER_ARRAY_OF_CHILDREN(C_RQ_HOLDER *a) : A(a) {}
    ~RQ_HOLDER_ARRAY_OF_CHILDREN() { delete [] A; }
  private:
    C_RQ_HOLDER *A;
};

/************************************************************************
* new_holder : create a new holder for the recursive queue		*
************************************************************************/
C_RQ_HOLDER *SpPriorityQueueHeap::new_holder() {
  C_RQ_HOLDER *rqh;
  int i;

  if (qfree_holder->get_length()) {
    return (C_RQ_HOLDER *)qfree_holder->pop_top();
  }

/*
  printf("Rqueue: Creating new set of holders (%d already exist)\n",
	n_created_holder);
*/

  rqh = new C_RQ_HOLDER[1000];
  ToBeCleared.push_bot(new RQ_HOLDER_ARRAY_OF_CHILDREN(rqh));
  
  if (rqh == NULL) {
    fprintf(stderr,"Error (RQUEUE) no more holder memory\n");
    exit(1);
  }
  n_created_holder += 1000;

  for (i=1; i<999; i++) {
    rqh[i].set_link(&rqh[i+1]);
    qfree_holder->init(&rqh[1], &rqh[999], 999);
  }

  return rqh;

}

/************************************************************************
* get_average_comparisons : should return log2(n)			*
************************************************************************/
double SpPriorityQueueHeap::get_average_comparisons() {
  double av_insert;
  double temp;

  av_insert = get_average_insert();
  temp = log(av_insert)/log(2.0);
//  if (temp < 1.0) temp = 1.0;

#ifdef TEMPQ
  temp *= get_average_loop();
  temp += log(get_average_sort()) / log(2.0);
#else
  temp *= 1 + get_average_loop();
#endif

  return temp;

}

/************************************************************************
* print : print method							*
************************************************************************/
void SpPriorityQueueHeap::print() {
  C_SQ_ITEM *sq_item;
  C_QUEUE q;

  while ((sq_item = pop()) != NULL) {
    sq_item->print();
    q.push_bot(sq_item);
  }

  while (q.get_length() > 0) insert((C_SQ_ITEM *)q.pop_top());

}
