// SpPriorityQueueHeap.H header file

#ifndef rqueue_object
#define rqueue_object

#include "queue.H"
#include "sq_item.H"
#include "SpSimTime.H" //GEB2

#define NCOMB 40

class C_RQ_HOLDER : public C_SQ_ITEM {

  private:

    void *pointer;

  public:

    void init(void *ptr, const SpSimTime &time) {pointer = ptr; time_tag = time;}
    void set_pointer(void *ptr) {pointer = ptr;}
    void *get_pointer() {return pointer;}

};

/************************************************************************
*	rqueue - singly linked list data structure			*
************************************************************************/
class SpPriorityQueueHeap {

  private:

    static C_QUEUE ToBeCleared;
    C_QUEUE queue;
    C_QUEUE temp_queue;
    SpSimTime min_time; //GEB2

    int n;

    int n_inserts;
    int n_pops;
    int n_loops;
    int sort_size;
    int n_sorts;
    int insert_size;
    int Ncomb;

//...... static free list quantities

    static int done;
    static C_QUEUE *qfree;
    static int n_created;
    static C_QUEUE *qfree_holder;
    static int n_created_holder;

  protected:

    C_SQ_ITEM *free_new();
    void free_delete(C_SQ_ITEM *it) {qfree->push_bot(it);}

  public:

    void clear(){
      queue.reset();
      temp_queue.reset();
      n = n_inserts = n_pops = n_loops = sort_size = n_sorts = insert_size = 0;
      Ncomb = NCOMB;
      min_time = 1.0e20;
    }

    SpPriorityQueueHeap();
    ~SpPriorityQueueHeap();
    void init();
    void insert(C_SQ_ITEM *it);
    C_SQ_ITEM *pop();
    SpSimTime& get_time();  //GEB2
    int get_length() {return n;}
    void set_Ncomb(int nc) {Ncomb = nc;}

//...... RQ holder free list methods

    static C_RQ_HOLDER *new_holder();
    void delete_holder(C_RQ_HOLDER *rqh) {qfree_holder->push_bot(rqh);}

//...... statistics

    int get_n_created() {return n_created;}
    int get_n_active() {return (n_created - qfree->get_length());}
    int get_n_inserts() {return n_inserts;}
    int get_n_loops() {return n_loops;}
    int get_sort_size() {return sort_size;}
    int get_insert_size() {return insert_size;}

    double get_average_sort() {
	return (double(sort_size)/double(n_sorts+0.000001));}

    double get_average_insert() {
	return (double(insert_size)/double(n_inserts));}

    double get_average_loop() {
	return (double(n_loops)/double(n_pops));}

    double get_average_comparisons();

//...... print method

    void print();

};

typedef SpPriorityQueueHeap C_RQUEUE;

#endif

