// SpPriorityTree.H header file defines a priority queue based on a binary
// tree.

#ifndef SpPriorityTree_OBJ
#define SpPriorityTree_OBJ

#include "SpBinaryTree.H"

class SpPriorityTree {

  private:

  protected:

    SpBinaryTree BinaryTree;

  public:

    SpPriorityTree() {
      BinaryTree.SetBalancedTreeMode();
    }

    void SetNormalTreeMode() {BinaryTree.SetNormalTreeMode();}
    void SetBalancedTreeMode() {BinaryTree.SetBalancedTreeMode();}
    void SetSplayTreeMode() {BinaryTree.SetSplayTreeMode();}

    int IsNormalTreeMode() {return BinaryTree.IsNormalTreeMode();}
    int IsBalancedTreeMode() {return BinaryTree.IsBalancedTreeMode();}
    int IsSplayTreeMode() {return BinaryTree.IsSplayTreeMode();}

    void *Insert(void *element, double priority) {
      BinaryTree.Insert(element, priority);
      return BinaryTree.GetCurrentNode();
    }

    void *Remove() {
      void *element = BinaryTree.RemoveFirstElement();
      return element;
    }

    void *Remove(double &priority) {
      void *element = BinaryTree.RemoveFirstElement(priority);
      return element;
    }

    int Retract(void *RetractionHandle) {
      int n = BinaryTree.GetNumElements();
      BinaryTree.RemoveNode((TREE_ITEM *)RetractionHandle);
      return (BinaryTree.GetNumElements() - n);
    }

    int CheckIntegrity() {return BinaryTree.CheckIntegrity();}

    int GetNumElements() {return BinaryTree.GetNumElements();}

    
    double GetCurrentPriority() {
      TREE_ITEM *top = BinaryTree.GetFirstNode();
      return top->DoubleVal();
    }

};

typedef SpPriorityTree PRIORITY_TREE;

#endif
