// SpPubSubItem.H header file

#ifndef PubSubItem_object
#define PubSubItem_object

#include "qitem.H"
#include "SpNetTypes.H"
#include "SpStdlib.H"

/*
** This class provides the items that go in the hierarchical grids.
** This represents both publishers and subscribers.
*/

class SpPubSubItem : public C_ITEM {

  private:
    SpPubSubItem   (SpPubSubItem&) {abort();}
    void operator=(SpPubSubItem&) {abort();}

    NET_INT node;
    NET_INT type;
    NET_INT local_id;
    NET_INT class_type;
    int Mark;
    SpPubSubItem *MainPubSubItem;

  static int Ncreated;

  protected:

  public:

    SpPubSubItem()
      : node(-1), type(-1), local_id(-1), class_type(-1), Mark(-99999), MainPubSubItem(this){
      Ncreated++;
    }

    SpPubSubItem(SpPubSubItem *PubSubItem);
    void init(SpPubSubItem *PubSubItem);
    virtual ~SpPubSubItem() {Ncreated--;}

    void init(int n, int t , int lid, int gid, int ct) {
      node = n;
      type = t;
      local_id = lid;
      id = gid;
      class_type = ct;
    }

    void get_all(int &n, int &t , int &lid, int &gid, int &ct) {
      n   = node;
      t   = type;
      lid = local_id;
      gid = id;
      ct  = class_type;
    }

    void set_node(int n) {node = n;}
    void set_local_id(int lid) {local_id = lid;}
    void set_type(int t) {type = t;}
    void set_global_id(int gid) {id = gid;}
    void set_class_type(int v) { class_type = v;}

    int get_node() const {return node;}
    int get_type() const {return type;}
    int get_local_id() const {return local_id;}
    int get_global_id() const {return id;}
    int get_class_type(void) const { return (class_type); }

    int get_Ncreated() {return Ncreated;}
    SpPubSubItem *get_MainPubSubItem() {return MainPubSubItem;}

    int get_Mark() {return Mark;}
    void set_Mark(int M) {Mark = M;}

    static int GetGlobalId(void *item) { return (((SpPubSubItem *) item)->get_global_id()); }
};

//RB_DEFINE_CLASS(SpPubSubItem);

class MAIN_PUBSUB_ITEM : public SpPubSubItem {

  private:

    static int CurrentUniqueKey, NumNodes;
    NET_INT UniqueKey;
    NET_INT Ndimensions;
    NET_DOUBLE *Lo;
    NET_DOUBLE *Hi;
    NET_DOUBLE *Res;

  protected:

  public:

    MAIN_PUBSUB_ITEM(void);
    MAIN_PUBSUB_ITEM(int uniqueKey);
    ~MAIN_PUBSUB_ITEM(void) { if (Lo) delete [] Lo; }

    // You must call this function before using this class in any
    // other way. If you are using this class within a SPEEDES node,
    // then you do not need to call it since it is already called in
    // SpExecute.C
    static void InitSpMainPubSubItemClass(int nodeId, int numNodes) {
      CurrentUniqueKey = nodeId;
      NumNodes = numNodes;
    }

    void SetUniqueKey(void);
    int GetUniqueKey() const { return UniqueKey; }

    void AllocateNumDimensions(int n) {
      if (Lo) delete [] Lo;
      Ndimensions = n;
      Lo  = (n>0) ? new NET_DOUBLE[3*n] : (NET_DOUBLE *) 0;
      Hi  = (n>0) ? &Lo[Ndimensions]    : (NET_DOUBLE *) 0;
      Res = (n>0) ? &Hi[Ndimensions]    : (NET_DOUBLE *) 0;
      for (int i=0; i<Ndimensions; i++) Res[i] = -1.0;
    }

    int GetNdimensions() {return Ndimensions;}

    void SetLHR(int d, double l, double h, double r) {
      Lo[d] = l;
      Hi[d] = h;
      Res[d] = r;
    }

    void GetLHR(int d, double &l, double &h, double &r) {
      l = Lo[d];
      h = Hi[d];
      r = Res[d];
    }

    double GetLo(int i) {return Lo[i];}
    double GetHi(int i) {return Hi[i];}
    double GetRes(int i) {return Res[i];}

    void SetLo(int i, double lo) {Lo[i] = lo;}
    void SetHi(int i, double hi) {Hi[i] = hi;}
    void SetRes(int i, double res) {Res[i] = res;}

    //double *GetLo() {return Lo;}
    //double *GetHi() {return Hi;}
    //double *GetRes() {return Res;}

    void PrintThis(ostream &os) const;

    void CreateMessage(char* &msg, int &len);
    static void CreateObject(const char *msg, int &obj_bytes, MAIN_PUBSUB_ITEM &item);

    int operator==(const MAIN_PUBSUB_ITEM& i) const;
};

typedef MAIN_PUBSUB_ITEM SpMainPubSubItem;

//RB_DEFINE_CLASS(MAIN_PUBSUB_ITEM);

inline ostream & operator<<(ostream &os, const MAIN_PUBSUB_ITEM &in)
{
  in.PrintThis(os);
  return os;
}

typedef SpPubSubItem PUBSUB_ITEM;

#endif
