// SpSetElement.C
// Implementation of SpSetElement subclass methods

#include "SpStdlib.H"
#define __SpSet_Internal__
#include "SpSetElement.H"
#include "SpSet.H"
#define __STRICT_ANSI__
#include "SpNetTypes.H"
#undef __STRICT_ANSI__
#include "SpAddPaddingToAlign.H"

int SpSetEnum::operator()(const char *key) const
{
  for(int k=0; k<Nlabels;k++) {
    if (!strcmp(key,*(Label+k))) return k;
  }
  cerr << "Enum string " << key << " not found" << endl;
  abort();  
  return 0;
}

SpSetElement::~SpSetElement(void)
{
}

SpSetString::SpSetString(char *val): SpSetElement(STRING_ET,true) 
{
  if (val != NULL) {
    Val = strdup(val);
  }
  else {
    Val = NULL;
  }
}

SpSetEnum::SpSetEnum(SpSetEnum &orig): SpSetElement(ENUM_ET)
{
  *this = *(SpSetEnum *)orig.Original;    // shallow copy
  Original = orig.Original;
}

SpSetInt::SpSetInt(SpSetInt &orig): SpSetElement(INT_ET)
{
  *this = *(SpSetInt *)orig.Original;    // shallow copy
  Original = orig.Original;
}

SpSetFloat::SpSetFloat(SpSetFloat &orig): SpSetElement(FLOAT_ET)
{
  *this = *(SpSetFloat *)orig.Original;    // shallow copy
  Original = orig.Original;
}

SpSetString::SpSetString(SpSetString &orig): SpSetElement(STRING_ET)
{
  *this = *(SpSetString *)orig.Original;    // shallow copy
  Original = orig.Original;
}

SpSetLogical::SpSetLogical(SpSetLogical &orig): SpSetElement(LOGICAL_ET)
{
  *this = *(SpSetLogical *)orig.Original;    // shallow copy
  Original = orig.Original;
}

SpSetVoidPtr::SpSetVoidPtr(SpSetVoidPtr &orig): SpSetElement(VOIDPTR_ET)
{
  *this = *(SpSetVoidPtr *)orig.Original;    // shallow copy
  Original = orig.Original;
}

SpSetArray::SpSetArray(SpSetArray &orig): SpSetElement(ARRAY_ET)
{
  *this = *(SpSetArray *)orig.Original;    // shallow copy
  Original = orig.Original;
}

char *SpSetEnum::Pack(char *buffer)
{
  
  return buffer;
}

SpSetElement *SpSetInt::Clone()
{
  SpSetInt *retval = new SpSetInt;
  *retval = Val;
  return retval;
}

SpSetElement *SpSetFloat::Clone()
{
  SpSetFloat *retval = new SpSetFloat;
  *retval = Val;
  return retval;
}

SpSetElement *SpSetLogical::Clone()
{
  SpSetLogical *retval = new SpSetLogical;
  *retval = Val;
  return retval;
}

SpSetElement *SpSetString::Clone()
{
  SpSetString *retval = new SpSetString;
  *retval = Val;
  return retval;
}

SpSetElement *SpSetArray::Clone()
{
  SpSetArray *retval;
  int k;
  switch(ArrayType) {
  case INT_ET:
    retval = new SpSetArray(INT_ET,ArrayLength);
    memcpy(retval->Val,Val,ArrayLength*sizeof(int));
    break;
  case FLOAT_ET:
    retval = new SpSetArray(FLOAT_ET,ArrayLength);
    memcpy(retval->Val,Val,ArrayLength*sizeof(double));
    break;
  case STRING_ET:
    retval = new SpSetArray(STRING_ET,ArrayLength);
    for (k=0; k<ArrayLength; k++) {
      ((char **)(retval->Val))[k] = strdup(((char **)Val)[k]);
    }
    break;
  case LOGICAL_ET:
    retval = new SpSetArray(LOGICAL_ET,ArrayLength);
    memcpy(retval->Val,Val,ArrayLength*sizeof(bool));
    break;
  case NONE_ET:
    retval = new SpSetArray(NONE_ET,ArrayLength);
    for (k=0; k<ArrayLength; k++) {
      ((SpSetElement **)retval->Val)[k] = ((SpSetElement **)Val)[k]->Clone();
    }
    break;
  default:
    cout << "SpSetArray:::Clone: Don't know how ";
    cout << "to clone array of type ";
    cout << SpSetElementName[ArrayType] << endl;
    return NULL;
  }
  return retval;
}

char *SpSetInt::Pack(char *buffer)
{
  char *offset = buffer;
  if (!IsRef()) {
    char char_init = (char) Init;
    NET_INT net_int_value = Val;
    memset(offset,char_init,sizeof(char_init));
    offset += sizeof(char_init);
    memcpy(offset,&net_int_value,sizeof(net_int_value));
    offset += sizeof(net_int_value);
  }
  else {
    offset += 2;
    memset(offset,(char) INT_ET, sizeof(char));
  }
  return buffer;
}

int SpSetInt::Size()
{
  int retval = 0;
  if (!IsRef()) {
    retval += 1 + sizeof(NET_INT); // One byte for Init field
  }
  else {
    retval = 3;
  }
  return retval;
}

SpSetElement *SpSetInt::Restore(char *buffer)
{
  char *offset = buffer;
  char char_init;
  memcpy(&char_init,offset,sizeof(char_init));
  offset += sizeof(char_init);
  Init = (bool) char_init;
  NET_INT net_int_val;
  memcpy(&net_int_val,offset,sizeof(net_int_val));
  Val = net_int_val;
  return this;
}

char *SpSetFloat::Pack(char *buffer)
{
  // Layout of buffer returned by SpSetFloat::Pack() method:
  //
  // <buffer> := <char init field><NET_INT value>
  
  char *offset = buffer;
  if (!IsRef()) {
    char char_init = (char) Init;
    NET_DOUBLE net_float_value = Val;

    memset(buffer,char_init,sizeof(char_init));
    offset += sizeof(char_init);
    memcpy(offset,&net_float_value,sizeof(net_float_value));
  }
  else {
    offset += 2;
    memset(offset,(char) FLOAT_ET, sizeof(char));
  }
  return buffer;
}

int SpSetFloat::Size()
{
  int retval = 0;
  if (!IsRef()) {
    retval += 1 + sizeof(NET_DOUBLE); // One byte for Init field
  }
  else {
    retval = 3;
  }
  return retval;
}

SpSetElement *SpSetFloat::Restore(char *buffer)
{
  char *offset = buffer;
  char char_init;
  memcpy(&char_init,offset,sizeof(char_init));
  offset += sizeof(char_init);
  Init = (bool) char_init;
  NET_DOUBLE net_float_val;
  memcpy(&net_float_val,offset,sizeof(net_float_val));
  Val = net_float_val;
  return this;
}

char *SpSetArray::Pack(char *buffer)
{
  // Layout of buffer returned by SpSetArray::Pack()
  // 
  // <buffer> := <header><series of records95>
  // <header> := <char element type><array length>
  // <series of records> := <series of records><record>
  // <record> := <int record | float record | string record | 
  //              logical record>
  // <int record> := <NET_INT value>
  // <float record> := <NET_DOUBLE value>
  // <string record> := <null terminated char string>
  // <logical record> := <(char) 0 or (char) 1>
  // Note: all records are required to be of uniform type

  char *offset = buffer;
  if (!IsRef()) {
    memset(offset,(char)ArrayType,sizeof(ArrayType));
    offset += sizeof(char);
    NET_INT net_int_length = ArrayLength;
    memcpy(offset,&net_int_length,sizeof(net_int_length));
    offset += sizeof(net_int_length);

    int k;
    switch (ArrayType) {
    case INT_ET:
      for (k=0; k<ArrayLength; k++) {
	NET_INT net_int_value = ((int *)Val)[k];
	memcpy(offset,&net_int_value,sizeof(NET_INT));
	offset += sizeof(net_int_value);
      }
      break;
    case FLOAT_ET:
      for (k=0; k<ArrayLength; k++) {
	NET_DOUBLE net_float_value = ((double *)Val)[k];
	memcpy(offset,&net_float_value,sizeof(NET_DOUBLE));
	offset += sizeof(net_float_value);
      }
      break;
    case LOGICAL_ET:
      for (k=0; k<ArrayLength; k++) {
	memset(offset,(char)((bool *)Val)[k],sizeof(char));
	offset += sizeof(char);
      }
      break;
    case STRING_ET:
      for (k=0; k<ArrayLength; k++) {
	memcpy(offset,((char **)Val)[k],1+strlen(((char **)Val)[k]));
	offset += 1+strlen(((char **)Val)[k]);
      }
      break;
    default:
      cerr << "Don't know how to Pack() array type ";
      cerr << SpSetElementName[(int) ArrayType] << endl;
      abort();
      break;
    }

  }
  else {
    offset++;
    memset(offset,(char) ARRAY_ET,sizeof(char));
    offset += sizeof(char);
    memset(offset,(char) ArrayType,sizeof(char));
  }
  return buffer;
}

int SpSetArray::Size() 
{
  int k;
  int retval = 0;
  if (!IsRef()) {
    int datasize;
    switch (ArrayType) {
    case INT_ET:
      datasize = ArrayLength * sizeof(NET_INT);
      break;
    case FLOAT_ET:
      datasize = ArrayLength * sizeof(NET_DOUBLE);
      break;
    case LOGICAL_ET:
      datasize = ArrayLength * sizeof(char);
      break;
    case STRING_ET:
      datasize = 0;
      for(k=0;k<ArrayLength; k++) {
	datasize += 1+strlen(((char **)Val)[k]);
      }
    default:
      cerr << "Don't know how to calculate size of array of type ";
      cerr << SpSetElementName[(int) ArrayType] << endl;
      abort();
    }
    retval += sizeof(char) + sizeof(NET_INT) + datasize;
  }
  else {
    retval = 3;
  }
  return retval;
}

SpSetElement *SpSetArray::Restore(char *buffer)
{
  char *offset = buffer;
  char etype;
  memcpy(&etype,offset,sizeof(char));
  offset += sizeof(char);
  ArrayType = (ElementType) etype;

  NET_INT net_int_length;
  memcpy(&net_int_length,offset,sizeof(net_int_length));
  offset += sizeof(net_int_length);
  ArrayLength = net_int_length;

  if (Val) {
    delete [] Val;
  }

  int k;
  switch(ArrayType) {
    
  case INT_ET:
    Val = new int[ArrayLength];
    for (k=0;k<ArrayLength; k++) {
      NET_INT net_int_val;
      memcpy(&net_int_val,offset,sizeof(net_int_val));
      offset += sizeof(net_int_val);
      ((int *)Val)[k] = (int) net_int_val;
    }
    break;
  case FLOAT_ET:
    Val = new double[ArrayLength];
    for (k=0;k<ArrayLength; k++) {
      NET_DOUBLE net_float_val;
      memcpy(&net_float_val,offset,sizeof(net_float_val));
      offset += sizeof(net_float_val);
      ((double *)Val)[k] = (int) net_float_val;
    }
    break;
  case LOGICAL_ET:
    Val = new bool[ArrayLength];
    for (k=0;k<ArrayLength; k++) {
      char char_bool; 
      memcpy(&char_bool,offset,sizeof(char_bool));
      offset += sizeof(char_bool);
      ((bool *)Val)[k] = (bool) char_bool;
    }
    break;
  case STRING_ET:
    Val = new char *[ArrayLength];
    for (k=0;k<ArrayLength; k++) {
      ((char **)Val)[k] = strdup(offset);
      offset += 1+strlen(((char **)Val)[k]);
    }    
    break;
  default:
    cerr << "Don't know how to handle array ";
    cerr << "of type " << SpSetElementName[(int) ArrayType] << endl;
    abort();
  }

  return this;
}

char *SpSetVoidPtr::Pack(char *buffer)
{
  // Layout of buffer returned by this method
  //
  // <buffer> := <init><length><data>
  // <init>   := <char>
  // <length> := <NET_INT>

  char *offset = buffer;
  if (!IsRef()) {
    char char_init = (char) Init;
    memset(offset,char_init,sizeof(char_init));
    offset += sizeof(char_init);
    NET_INT net_int_length = Length;
    memcpy(offset,&net_int_length,sizeof(net_int_length));
    offset += sizeof(net_int_length);
    if (Init) {
      memcpy(offset,Val,Length);
    }
  }
  else {
    offset += 2;
    memset(offset,(char) VOIDPTR_ET,sizeof(char)); 
  }
  return buffer;
}

int SpSetVoidPtr::Size() 
{
  int retval;
  if (!IsRef()) {
    retval = 1+ sizeof(NET_INT) + Length;
  }
  else {
    retval = 3;
  }
  return retval;
}

SpSetElement *SpSetVoidPtr::Restore(char *buffer)
{
  char *offset = buffer;
  char char_init;
  memcpy(&char_init,offset,sizeof(char_init));
  Init = (bool) char_init;
  NET_INT net_int_length;
  memcpy(&net_int_length,offset,sizeof(NET_INT));
  offset += sizeof(net_int_length);
  Length = net_int_length;
  if (Init) {
    memcpy(Val,offset,Length);
  }
  else {
    Val = NULL;
  }
  return this;
}

char *SpSetString::Pack(char *buffer)
{

  char *offset = buffer;
  if (!IsRef()) {
    char char_init = (char) Init;
    memset(offset,char_init,sizeof(char_init));
    offset += sizeof(char_init);
    if (Init) {
      memcpy(offset,Val,1+strlen(Val));
    }
    else {
      memset(offset,0,1);
    }
  }
  else {
    offset += 2;
    memset(offset,(char)STRING_ET, sizeof(char));
  }
  return buffer;
}

int SpSetString::Size()
{
  int retval=0;
  if (!IsRef()) {
    retval += 2+strlen(Val); // char_init + string + terminating null
  }
  else {
    retval = 3;
  }
  return retval;
}

SpSetElement *SpSetString::Restore(char *buffer)
{
  char *offset = buffer;
  char char_init;
  memcpy(&char_init,offset,sizeof(char_init));
  offset += sizeof(char_init);
  Init = (bool) char_init;
  if (Init) {
    Val = strdup(offset);
  }
  else {
    Val = NULL;
  }
  return this;
}

char *SpSetLogical::Pack(char *buffer)
{
  char *offset = buffer;
  if (!IsRef()) {
    char char_init = (char) Init;
    memset(offset,char_init,sizeof(char_init));
    offset += sizeof(char_init);
    memset(offset,(char) Val,sizeof(char));
  }
  else {
    offset += 2;
    memset(buffer,(char)LOGICAL_ET,sizeof(char));
  }
  return buffer;
}

int SpSetLogical::Size() 
{
  int retval = 0;
  if (!IsRef()) {
    retval += 2;
  }
  else {
    retval += 3;
  }
  return retval;
}

SpSetElement *SpSetLogical::Restore(char *buffer)
{
  char *offset = buffer;
  char char_init;
  memcpy(&char_init,offset,sizeof(char_init));
  offset += sizeof(char_init);
  Init = (bool) char_init;
  if (Init) {
    char char_val;
    memcpy(&char_val,offset,sizeof(char));
    Val = (bool) char_val;
  }
  return this;
}

SpSetVoidPtr::~SpSetVoidPtr() 
{
  if (this == Original) {
    free(Val);
  }
}

SpSetEnum::~SpSetEnum()
{
  int k;
  if (this == Original) {
    for(k=0; k<Nlabels; k++) free(*(Label+k));
  }

}


SpSetArray::SpSetArray(ElementType type, int arraylength):
  SpSetElement(ARRAY_ET,true), ArrayLength(arraylength), ArrayType(type), TotalLength(arraylength+5) {
  int k;
  switch (ArrayType) {
  case INT_ET:
    Val = new int[TotalLength];
    for (k=0;k<ArrayLength;k++) ((int *)Val)[k] = 0;
    break;
  case FLOAT_ET:
    Val = new double[TotalLength];
    for (k=0;k<ArrayLength;k++) ((double *)Val)[k] = 0.;
    break;
  case LOGICAL_ET:
    Val = new bool[TotalLength];
    for (k=0;k<ArrayLength;k++) ((bool *)Val)[k] = false;
    break;
  case STRING_ET:
    Val = new char *[TotalLength];
    for (k=0;k<ArrayLength;k++) ((char **)Val)[k] = NULL;
    break;
  case NONE_ET:
    Val = new SpSetElement *[TotalLength];
    for (k=0;k<ArrayLength;k++) ((SpSetElement **)Val)[k] = NULL;
    break;
  default:
    cerr << "Don't know how to handle array ";
    cerr << "of type " << SpSetElementName[(int) type] << endl;
    abort();
  }
}

SpSetArray *SpSetArray::Push(int i)
{
  int *tmpint;
  SpSetElement **tmpelement; 
  switch (ArrayType) {
  case INT_ET:
    if (TotalLength == ArrayLength) {
      TotalLength += 5;
      tmpint = new int[TotalLength];
      memcpy(tmpint,Val,ArrayLength*sizeof(int));
      delete [] Val;
      Val = tmpint;
    }
    ((int *)Val)[ArrayLength] = i;
    break;
  case NONE_ET:
    if (TotalLength == ArrayLength) {
      TotalLength += 5;
      tmpelement = new SpSetElement *[TotalLength];
      memcpy(tmpelement,Val,ArrayLength*sizeof(SpSetElement *));
      delete [] Val;
      Val = tmpelement;
    }
    ((SpSetElement **)Val)[ArrayLength] = new SpSetInt(i);
    break;
  default:
    cerr << "SpSetArray: cannot push integer onto array of ";
    cerr << "type " << SpSetElementName[ArrayType] << endl;
    return NULL;
  }
  ArrayLength++;
  return this;
}

SpSetArray *SpSetArray::Push(double d)
{
  double *tmpdbl;
  SpSetElement **tmpelement;
  switch (ArrayType) {
  case FLOAT_ET:
    if (TotalLength == ArrayLength) {
      TotalLength += 5;
      tmpdbl = new double[TotalLength];
      memcpy(tmpdbl,Val,ArrayLength*sizeof(double));
      delete [] Val;
      Val = tmpdbl;
    }
    ((double *)Val)[ArrayLength] = d;
    break;
  case NONE_ET:
    if (TotalLength == ArrayLength) {
      TotalLength += 5;
      tmpelement = new SpSetElement *[TotalLength];
      memcpy(tmpelement,Val,ArrayLength*sizeof(SpSetElement *));
      delete [] Val;
      Val = tmpelement;
    }
    ((SpSetElement **)Val)[ArrayLength] = new SpSetFloat(d);
    break;
  default:
    cerr << "SpSetArray: cannot push integer onto array of ";
    cerr << "type " << SpSetElementName[ArrayType] << endl;
    return NULL;
  }
  ArrayLength++;
  return this;
}

SpSetArray *SpSetArray::Push(bool b)
{
  bool *tmpbool;
  SpSetElement **tmpelement;
  switch (ArrayType) {
  case LOGICAL_ET:
    if (TotalLength == ArrayLength) {
      TotalLength += 5;
      tmpbool = new bool[TotalLength];
      memcpy(tmpbool,Val,ArrayLength*sizeof(bool));
      delete [] Val;
      Val = tmpbool;
    }
    ((bool *)Val)[ArrayLength] = b;
    break;
  case NONE_ET:
    if (TotalLength == ArrayLength) {
      TotalLength += 5;
      tmpelement = new SpSetElement *[TotalLength];
      memcpy(tmpelement,Val,ArrayLength*sizeof(SpSetElement *));
      delete [] Val;
      Val = tmpelement;
    }
    ((SpSetElement **) Val)[ArrayLength] = new SpSetLogical(b);
    break;
  default:
    cerr << "SpSetArray: cannot push integer onto array of ";
    cerr << "type " << SpSetElementName[ArrayType] << endl;
    return NULL;
  }
  ArrayLength++;
  return this;
}

SpSetArray *SpSetArray::Push(char *s)
{
  char **tmpchar;
  SpSetElement **tmpelement;
  switch (ArrayType) {
  case STRING_ET:
    if (TotalLength == ArrayLength) {
      TotalLength += 5;
      tmpchar = new char *[TotalLength];
      memcpy(tmpchar,Val,ArrayLength*sizeof(char *));
      delete [] Val;
      Val = tmpchar;
    }
    ((char **)Val)[ArrayLength] = strdup(s);
    break;
  case NONE_ET:
    if (TotalLength == ArrayLength) {
      TotalLength += 5;
      tmpelement = new SpSetElement *[TotalLength];
      memcpy(tmpelement,Val,ArrayLength*sizeof(SpSetElement *));
      delete [] Val;
      Val = tmpelement;
    }
    ((SpSetElement **) Val)[ArrayLength] = new SpSetString(s);
    break;
  default:
    cerr << "SpSetArray: cannot push integer onto array of ";
    cerr << "type " << SpSetElementName[ArrayType] << endl;
    return NULL;
  }
  ArrayLength++;
  return this;
}

SpSetArray *SpSetArray::Push(SpSetElement &element) 
{
  // If the type of this array is a standard built-in
  // (i.e., INT_ET,LOGICAL_ET,FLOAT_ET,STRING_ET) then
  // add the value of the element. If the type is NONE_ET,
  // add the element itself to the array'

  // Resize array if necessary
  int *tmpint;
  bool *tmpbool;
  double *tmpfloat;
  char **tmpchar;
  int k;
  SpSetElement **tmpelement;

  if (TotalLength == ArrayLength) {
    TotalLength += 5;
    switch (ArrayType) {
    case INT_ET:
      tmpint = new int[TotalLength];
      memcpy(tmpint,Val,ArrayLength*sizeof(int));
      delete [] Val;
      Val = tmpint;
      break;
    case STRING_ET:
      tmpchar = new char*[TotalLength];
      for (k=0;k<ArrayLength;k++) {
	tmpchar[k] = strdup(((char **)Val)[k]);
      }
      delete [] Val;
      Val = tmpchar;
      break;
    case LOGICAL_ET:
      tmpbool = new bool[TotalLength];
      memcpy(tmpbool,Val,ArrayLength*sizeof(bool));
      delete [] Val;
      Val = tmpbool;
      break;
    case FLOAT_ET:
      tmpfloat = new double[TotalLength];
      memcpy(tmpfloat,Val,ArrayLength*sizeof(double));
      delete [] Val;
      Val = tmpfloat;
      break;
    case NONE_ET:
      tmpelement = new SpSetElement *[TotalLength];
      memcpy(tmpelement,Val,ArrayLength*sizeof(SpSetElement *));
      delete [] Val;
      Val = tmpelement;
      break;
    default:
      cerr << "Unknown type \"" << SpSetElementName[ArrayType] << "\" at ";
      cerr << __FILE__ << ", " << __LINE__ << endl;
      break;
    }
  }

  // Append new array element
  switch (ArrayType) {
  case INT_ET:
    ((int *)Val)[ArrayLength] = element;
    break;
  case STRING_ET:
    ((char **)Val)[ArrayLength] = element;
    break;
  case LOGICAL_ET:
    ((bool *)Val)[ArrayLength] = element;
    break;
  case FLOAT_ET:
    ((double *)Val)[ArrayLength] = element;
    break;
  case NONE_ET:
    ((SpSetElement **)Val)[ArrayLength] = &element;
    break;
  }

  ArrayLength++;
  return this;
}

SpSetArray::~SpSetArray()
{
  if (Original == this) {
    if (Val) {
      delete [] (void *)Val;
    }
  }
}

SpSetString::~SpSetString() {
  if (Original == this) {
    if (Val) {
      free(Val);
    }
  }
}

SpSetString::operator char *() {
  if (Val != NULL) {
    // Returns private data! Regrettably required to be consistent
    // with handling of arrays. Don't attempt to free this pointer!
    return Val;
  }
  else {
    cerr << "Attempt to access an uninitialized element" << endl;
    abort();
  }
  return NULL;
}

SpSetElement &SpSetArray::operator [] (int i) {
  if (i < ArrayLength) {
    switch(ArrayType) {
    case INT_ET:
      return *(new SpSetInt(((int *)Val)[i]));
    case FLOAT_ET:
      return  *(new SpSetFloat(((double *)Val)[i]));
    case LOGICAL_ET:
      return  *(new SpSetLogical(((bool *)Val)[i]));
    case STRING_ET:
      return  *(new SpSetString(((char **)Val)[i]));
    case NONE_ET:
      return *(((SpSetElement **)Val)[i]);
    default:
      cerr << "SpSetArray: Attempt to dereference of unknown ";
      cerr << "elementtype " << ArrayType << endl;
    }
  }
  else {
    cerr << "SpSetArray: Out-of-bounds array dereference";
    cerr << endl;
  }
  abort();
  static SpSetElement this_variable_is_here_so_we_can_return_something_to_get_rid_of_compiler_warning;
  return this_variable_is_here_so_we_can_return_something_to_get_rid_of_compiler_warning;
}
    
SpSetArray::operator int *() {
  if (ArrayType == INT_ET) {
    return (int *)Val;
  }
  else {
    return (int *)NULL;
  }
}

SpSetArray::operator double *() {
  if (ArrayType == FLOAT_ET) {
    return (double *)Val;
  }
  else {
    return (double *)NULL;
  }
}

SpSetArray::operator char **() {
  if (ArrayType == STRING_ET) {
    return (char **)Val;
  }
  else {
    return (char **)NULL;
  }
}

SpSetArray::operator bool *() {
  if (ArrayType == LOGICAL_ET) {
    return (bool *)Val;
  }
  else {
      return (bool *)NULL;
  }
}
  
SpSetElement &SpSetNone::operator=(int rval) 
{
  SpSetInt *retval = new SpSetInt(rval);
  Parent->Namespace->Remove(Name);
  Parent->Namespace->Insert(retval,Name);
  delete this;
  return *retval;
}

SpSetElement &SpSetNone::operator=(double rval) 
{
  SpSetFloat *retval = new SpSetFloat(rval);
  Parent->Namespace->Remove(Name);
  Parent->Namespace->Insert(retval,Name);
  delete this;
  return *retval;
}

SpSetElement &SpSetNone::operator=(char *rval) 
{
  SpSetString *retval = new SpSetString(rval);
  Parent->Namespace->Remove(Name);
  Parent->Namespace->Insert(retval,Name);
  delete this;
  return *retval;
}

SpSetElement &SpSetNone::operator=(bool rval) 
{
  SpSetLogical *retval = new SpSetLogical(rval);
  Parent->Namespace->Remove(Name);
  Parent->Namespace->Insert(retval,Name);
  delete this;
  return *retval;
}

// Protected method to determine which element to pack,
// self or original
SpSetElement *SpSetElement::packwhich(char *name,SpSet *top)
{
  SpSetElement *retval = Original;
  if (IsRef()) {
    if (!top->inscope(Original,name)) {
      retval = this;
    }
  }
  return retval;
}

