// SpPubSubItem.C file

#ifndef __SPEEDES_NT__
#include <values.h>
#else
#define INTBITS 32
#define MININT      (1 << (INTBITS - 1))
#define MAXINT      (~MININT)
#endif
#include "SpAssert.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpPubSubItem.H"
#include "SpString.H"
#include "SpMath.H"
#include "SpPoDataBase.H"
#include "RB_SpDefineClass.H"

RB_DEFINE_CLASS(SpPubSubItem);
RB_DEFINE_CLASS(MAIN_PUBSUB_ITEM);

int SpPubSubItem::Ncreated = 0;

SpPubSubItem::SpPubSubItem(SpPubSubItem *PubSubItem){
  memcpy(this, PubSubItem, sizeof(SpPubSubItem));
  Ncreated++;
}

void SpPubSubItem::init(SpPubSubItem *PubSubItem) {
  memcpy(this, PubSubItem, sizeof(SpPubSubItem));
  Ncreated++;
}

// This is initialized by InitSpMainPubSubItemClass.  If you are using
// this class within a SPEEDES node, then you do not need to call it
// since it is already called in SpExecute.C
int MAIN_PUBSUB_ITEM::CurrentUniqueKey = -1;
int MAIN_PUBSUB_ITEM::NumNodes;

MAIN_PUBSUB_ITEM::MAIN_PUBSUB_ITEM()
{
  Lo=Hi=Res=NULL;
  Ndimensions = 0;
  if (SpPoDataBase::DoNotAllocateMemoryInConstructor != 1){
    SetUniqueKey();
  }
}

void
MAIN_PUBSUB_ITEM::SetUniqueKey(){

//#ifdef ERROR_CHECK
  assert(CurrentUniqueKey >= 0);
  assert(CurrentUniqueKey < MAXINT - NumNodes);
//#endif

  UniqueKey = CurrentUniqueKey;
  CurrentUniqueKey += NumNodes;
}

MAIN_PUBSUB_ITEM::MAIN_PUBSUB_ITEM(int uniqueKey)
  : UniqueKey(uniqueKey) {
  Lo=Hi=Res=NULL;
  Ndimensions = 0;
}

void MAIN_PUBSUB_ITEM::PrintThis(ostream &os) const {
  os << "global id is " << get_global_id() << '\n';
  for (int i=0; i<Ndimensions; i++) {
    os << "MAIN_PUBSUB_ITEM(" << i << "):";

    if (Res[i] == -1)
    {
      os << " Don't Care" << endl;
    }
    else
    {
      os << setprecision(5) << resetiosflags(ios::left)
         << " Lo = " << setw(12) << Lo[i]
         << ", Hi =  " << setw(12) << Hi[i]
         << ", Res = " << setw(12) << Res[i]
         << endl;
    }
  }
}

void
MAIN_PUBSUB_ITEM::CreateMessage(char* &msg, int &len){

  len = (2 * sizeof(NET_INT)) + (sizeof(NET_DOUBLE) * 3 * Ndimensions);

  msg = new char [len];
  char *p = msg;

  memcpy(p, &UniqueKey,   sizeof(NET_INT)); p += sizeof(NET_INT);
  memcpy(p, &Ndimensions, sizeof(NET_INT)); p += sizeof(NET_INT);
  if (Ndimensions>0){
    memcpy(p, Lo,  sizeof(NET_DOUBLE)*Ndimensions); p += (sizeof(NET_DOUBLE)*Ndimensions);
    memcpy(p, Hi,  sizeof(NET_DOUBLE)*Ndimensions); p += (sizeof(NET_DOUBLE)*Ndimensions);
    memcpy(p, Res, sizeof(NET_DOUBLE)*Ndimensions);
  }
}

void
MAIN_PUBSUB_ITEM::CreateObject(const char *msg, int &obj_bytes, MAIN_PUBSUB_ITEM &item){
  const char *p = msg;
  int nbytes;

  obj_bytes = 0;

  NET_INT    num_dimensions, unique_key;
  NET_DOUBLE *hi  = (NET_DOUBLE *) 0;
  NET_DOUBLE *lo  = (NET_DOUBLE *) 0;
  NET_DOUBLE *res = (NET_DOUBLE *) 0;

  nbytes = sizeof(NET_INT);
  memcpy(&unique_key, p, nbytes); p += nbytes; obj_bytes += nbytes;

  nbytes = sizeof(NET_INT);
  memcpy(&num_dimensions, p, nbytes); p += nbytes; obj_bytes += nbytes;

  if (num_dimensions>0){
    lo  = new NET_DOUBLE [num_dimensions * 3];
    hi  = &(lo[num_dimensions]);
    res = &(hi[num_dimensions]);
    nbytes = sizeof(NET_DOUBLE)*num_dimensions*3;
    memcpy(lo, p, nbytes); obj_bytes += nbytes; p += nbytes;
  }

  item.UniqueKey = unique_key;
  item.AllocateNumDimensions(num_dimensions);
  for (int i=0; i<num_dimensions; i++)
    item.SetLHR(i, lo[i], hi[i], res[i]);

  if (lo) delete [] lo;
}

int 
MAIN_PUBSUB_ITEM::operator==(const MAIN_PUBSUB_ITEM& p) const{
  int i;
  double d;
  const double e = 1E-5;
  int not_equal = 1;

  if (Ndimensions!=p.Ndimensions) return (!not_equal);

  not_equal = 0;
  for (i=0; (i<Ndimensions) && !not_equal; i++){
    d = fabs(Lo[i]  - p.Lo[i]);  if (d>e) not_equal = 1; 
    d = fabs(Hi[i]  - p.Hi[i]);  if (d>e) not_equal = 1;
    d = fabs(Res[i] - p.Res[i]); if (d>e) not_equal = 1;
  }

  return (!not_equal);
}

void RB_Register_SpPubSubItem()
{
  RB_REGISTER_CLASS_SpPubSubItem();
}


void RB_Register_MAIN_PUBSUB_ITEM()
{
  RB_REGISTER_CLASS_MAIN_PUBSUB_ITEM();
}


