// SpBasetype.C method file

#include "SpStdio.H"
#include "SpStdlib.H"

#include "SpBasetype.H"

#include "gentype.H"
#include "genint.H"
#include "genfloat.H"
#include "genstring.H"
#include "genlogical.H"
#include "genenum.H"

/************************************************************************
* BASETYPE : BASETYPE constructor					*
************************************************************************/
SpBasetype::SpBasetype(char *string) {

  setname(string);
  ntypes = 0;
  nenum = 0;

}

SpBasetype::~SpBasetype()
{
  elements.clear();
}

/************************************************************************
* get_enumid : get an enumerated id (0 through n)			*
************************************************************************/
int SpBasetype::get_enumid(char *s) {
  int i,len;
  C_NAMIT *namit;

  len = elements.length();
  namit = (C_NAMIT *)elements.ptop();
  for (i=0; i<len; i++) {
    if (!strcmp(s,namit->get_name())) {
      return i;
    }
    namit = (C_NAMIT *)namit->get_link();
  }

  fprintf(stderr,"Error in enumerated basetype %s, could not find %s\n",
	name,s);
  exit(1);
  return -1;

}

/************************************************************************
* get_element : get a generic element based on the name			*
************************************************************************/
C_NAMIT *SpBasetype::get_element(char *s) {
  int i,len;
  C_NAMIT *namit;

  len = elements.length();
  namit = (C_NAMIT *)elements.ptop();
  for (i=0; i<len; i++) {
    if (!strcmp(s,namit->get_name())) {
      return namit;
    }
    namit = (C_NAMIT *)namit->get_link();
  }

  fprintf(stderr,"Error: basetype %s does not have element %s\n",name,s);
  exit(1);
  return NULL;

}

/************************************************************************
* exists : check if element exists					*
************************************************************************/
int SpBasetype::exists(char *s) {
  int i,len;
  C_NAMIT *namit;

  len = elements.length();
  namit = (C_NAMIT *)elements.ptop();
  for (i=0; i<len; i++) {
    if (!strcmp(s,namit->get_name())) {
      return 1;
    }
    namit = (C_NAMIT *)namit->get_link();
  }

  return 0;

}

/************************************************************************
* get_first_type : get the first type element				*
************************************************************************/
SpBasetype *SpBasetype::get_first_type() {

  counter = 0;
  next_pointer = (void *)elements.ptop();
  return get_next_type();

}

/************************************************************************
* get_next_type : get the first type element				*
************************************************************************/
SpBasetype *SpBasetype::get_next_type() {
  SpBasetype *basetype;
  C_GENTYPE *gentype;

  gentype = (C_GENTYPE *)next_pointer;

  if (counter < ntypes) {
    counter++;
    basetype = gentype->get_basetype();
    next_pointer = (void *)gentype->get_link();
    return basetype;
  }else{
    return NULL;
  }

}

/************************************************************************
* get_type : get a type element						*
************************************************************************/
SpBasetype *SpBasetype::get_type(char *s) {
  C_GENTYPE *gentype;

  gentype = (C_GENTYPE *)get_element(s);
  return gentype->get_basetype();

}

/************************************************************************
* get_int : get an int element						*
************************************************************************/
int SpBasetype::get_int(char *s, int i) {
  C_GENINT *genint;
  char S[120];

  if (i>=0) {
    sprintf(S, "%s[%d]", s, i);
    genint = (C_GENINT *)get_element(S);
  }else{
    genint = (C_GENINT *)get_element(s);
  }

  return genint->get_id();

}

/************************************************************************
* get_float : get a float element					*
************************************************************************/
float SpBasetype::get_float(char *s, int i) {
  C_GENFLOAT *genfloat;
  char S[120];

  if (i>=0) {
    sprintf(S, "%s[%d]", s, i);
    genfloat = (C_GENFLOAT *)get_element(S);
  }else{
    genfloat = (C_GENFLOAT *)get_element(s);
  }

  return genfloat->GetDoubleValue(); //GEB4: use get_valu

}

/************************************************************************
* get_string : get a string element					*
************************************************************************/
char *SpBasetype::get_string(char *s, int i) {
  C_GENSTRING *genstring;
  char S[120];

  if (i>=0) {
    sprintf(S, "%s[%d]", s, i);
    genstring = (C_GENSTRING *)get_element(S);
  }else{
    genstring = (C_GENSTRING *)get_element(s);
  }

  return genstring->get_string();

}

/************************************************************************
* get_logical : get a logical element					*
************************************************************************/
int SpBasetype::get_logical(char *s, int i) {
  C_GENLOGICAL *genlogical;
  char S[120];

  if (i>=0) {
    sprintf(S, "%s[%d]", s, i);
    genlogical = (C_GENLOGICAL *)get_element(S);
  }else{
    genlogical = (C_GENLOGICAL *)get_element(s);
  }

  return genlogical->get_id();

}

/************************************************************************
* get_enum : get an enum element					*
************************************************************************/
int SpBasetype::get_enum(char *s) {
  C_GENENUM *genenum;

  genenum = (C_GENENUM *)get_element(s);
  return genenum->get_id();

}

/************************************************************************
* get_enum : get an enum element					*
************************************************************************/
char *SpBasetype::get_enum(int i) {
  int j;
  C_NAMIT *namit;
  C_GENENUM *genenum;

  namit = (C_NAMIT *)elements.get_top();
  for (j=0; j<elements.get_length(); j++) {
    if (namit->CheckEnum()) {
      genenum = (C_GENENUM *)namit;
      if (genenum->get_id()== i) return genenum->get_name();
    }
    namit = (C_NAMIT *)namit->get_link();
  }

  return NULL;

}

/************************************************************************
* get_array_size : get the size of an array    				*
************************************************************************/
int SpBasetype::get_array_size(char *s) {
  char S[120];
  int i;

  i = 0;
  while (1) {
    sprintf(S, "%s[%d]", s, i);
    if (!exists(S)) {
      return i;
    }
    i++;
  }

}


/************************************************************************
* printfile : print to a file						*
************************************************************************/
void SpBasetype::printfile(FILE *fp) {
  int i,len;
  C_ITEM *item;

  fprintf(fp,"%s {\n",name);

  len = elements.length();
  item = elements.ptop();
  for (i=0; i<len; i++) {
    item->printfile(fp);
    item = item->get_link();
  }

  fprintf(fp,"}\n\n");

}


