// SpBitField.H header file defines sets of bits for multicasting in
// SPEEDES. 

#ifndef _SpBitField_H
#define _SpBitField_H

#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpString.H"

// Implements a class for storing a string of on/off flags as 
// single bits


class SpBitField 
{
  friend class RB_SpBitField;
  static unsigned int bitsinbyte;
public:
  
  SpBitField(): nBytes(0), n_bits(0), BitField(NULL) {}
  
  SpBitField(int Nbits): n_bits(Nbits) {
    nBytes = n_bits/bitsinbyte + ((n_bits%bitsinbyte)?1:0);       
    BitField = new char[nBytes];
    if (!BitField) {
      cout << "Error, could not create memory for BitField" << endl;
    }else{
      Reset();
    }
  }
  
  ~SpBitField() {
    //      cout << "Deleting BitField" << endl;
    if (BitField) delete [] BitField;
  }
  
  void Reset() {
    memset(BitField,0,nBytes);
  }
  
  void Reset(int bit) {
    if (bit < n_bits) {
      int i = bit / bitsinbyte;
      int j = bit % bitsinbyte;
      if (BitField[i] & (1<<j)) {
	BitField[i] &= ~(1<<j);
      }
    }else{
      cout << "Warning (BitField): bad bit " << bit << endl;
    }
  }
  
  void Set(int bit) {
    if (bit < n_bits) {
      int i = bit / bitsinbyte;
      int j = bit % bitsinbyte;
      BitField[i] |= 1<<j;
    }else{
      cout << "Warning (BitField): bad bit " << bit << endl;
    }
  }
  
  int Check(int bit) {
    int i = bit / bitsinbyte;
    int j = bit % bitsinbyte;
    return (BitField[i] & (1<<j));
  }
  
  void print(ostream &out) {
    out << "Printing the BitField Object for "
	<< n_bits << " bits" << endl;
    for (int i=0; i<nBytes; i++) {
      if (BitField[i]) {
	for (int j=0; j<(int)bitsinbyte; j++) {
	  if (BitField[i] & (1<<j)) {
	    out << "Bit " << i*bitsinbyte + j << endl;
	  }
	}
      }
    }
  }
  
  void SetNbits(int Nbits) {
    char *buff = NULL;
    int nold;
    if (BitField) buff = BitField;
    n_bits = Nbits;
    nold = nBytes;
    nBytes = n_bits/bitsinbyte + (n_bits%bitsinbyte?1:0);
    BitField = new char[nBytes];
    if (!BitField) {
      cout << "Error, could not create memory for BitField" << endl;
      if (buff) delete buff;
    }else{
      Reset();
      if (buff) {
	if (nBytes < nold) {
	  memcpy(BitField, buff, nBytes);
	}else{
	  memcpy(BitField, buff, nold);
	}
      }
    }
  }
  
  int GetNbits() {return n_bits;}
  int GetNbytes() {return nBytes;}
  char *GetBitField() {return BitField;}
  
  void operator += (int bit) {Set(bit);}
  void operator -= (int bit) {Reset(bit);}
  int operator == (int bit) {return Check(bit);}
  
  int GetFirstBit() {
    for (int i=0; i<nBytes; i++) {
      if (BitField[i]) {
	for (int j=0; j<(int)bitsinbyte; j++) {
	  if (BitField[i] & (1<<j)) {
	    return i*bitsinbyte + j;
	  }
	}
      }
    }
    return -1;
  }
  
  int GetNextBit(int bit) {
    int k = 0;

    int j = bit % bitsinbyte;
    int byteoffset = bit/bitsinbyte;
    
    int i = ((((j+1)%bitsinbyte)!=0)? byteoffset:(byteoffset+1));

    if (i>=nBytes) return -1;

    for (k=(j+1)%bitsinbyte; k<(int)bitsinbyte; k++) {
      if (BitField[i] & (1<<k)) {
	return (i*bitsinbyte) + k;
      }
    }

    k = i+1;
    for (i=k; i<nBytes; i++) {
      if (BitField[i]) {
	for (j=0; j<(int)bitsinbyte; j++) {
	  if (BitField[i] & (1<<j)) {
	    return (i*bitsinbyte) + j;
	  }
	}
      }
    }
    return -1;
  }
  
  char * PackBuffer(int & Nbytes){
    Nbytes = 3*sizeof(int) + nBytes;
    char * rc = new char[Nbytes];
    int * tempInt = (int*)rc;
    
    tempInt[0] = nBytes;
    tempInt[1] = n_bits;
    tempInt[2] = Bytes;
    memcpy(rc + 3*sizeof(int), BitField, nBytes);
    return rc;
  }
  
  void UnpackBuffer(char * buff){
    int * tempInt = (int*)buff;
    
    nBytes = tempInt[0];
    n_bits = tempInt[1];
    Bytes = tempInt[2];
    if (BitField == NULL){
      BitField = new char[nBytes];
    }
    memcpy(BitField, buff+3*sizeof(int), nBytes);
  }

protected:
  
private:
  
  int nBytes;
  int n_bits;
  int Bytes;
  char *BitField;
};

#endif // #ifndef _SpBitField_H
