#ifndef COMMSERVERMESSAGES_H
#define COMMSERVERMESSAGES_H

#include "SpSimTime.H"

/*************************************************************************
 * enum: 
 * Summary: Handler types for the Client Server code
 * Description: The ClientServer code needes to know what handlers are
 *              it will encounter.  This simply enumerates them so that
 *              I can create apropriate handlers.
*************************************************************************/
enum {DUMMY_CONN_MESSAGE = 0,       // Dummy connection handler that should be
                                    // ignored
      NORMAL_MESSAGE,               // Normal message handler
      MESSAGE_BROADCAST,            // Broadcast handler
      MESSAGE_MULTICAST,            // Multicast handler
      GLOBAL_SUM,                   // Global Sum handler
      GLOBAL_MIN,                   // Global Min handler
      GLOBAL_MAX,                   // Global Max handler
      NON_BLOCKING_SYNCHRONIZATION, // Non-blocking sync handler
      BLOCKING_SYNCHRONIZATION,     // Blocking sync handler
      TCPIP_STARTUP,                // Startup message handler
      DEBUG_MESSAGE,                // Debugging message handler
      NUMHANDLERS                   // Number of handlers
};





/*************************************************************************
 * Class: REDOP_INFO
 * Summary: structure for TCPIP reduction information
 * Description: Reductions that do sum's, mins, or maxes need to send thes
 *              same info.  That is encapsulated here and uses NET_INTs
 *              ( and soon NetDoubles) to achieve heterogeneous operations
 * Examples: Look SpCommTCP.C and CommmServer.C
*************************************************************************/
class REDOP_INFO{
private:
  NET_INT Int;
  NET_DOUBLE Double;
  SpSimTime Stime;
public:

  /*************************************************************************
   * Method: init
   * Summary: initialize REDOP_INFO to certain values
   * Description: Before sending a REDOP_INFO, we need to set the values
   *              which is what is passed in here.
   * Return: void
   *************************************************************************/
  void Init(int I,     //In: Integer value
	    double d,  //In: double value
	    SpSimTime s  //In: SpSimTime value
	    )
  {
    Int = I; 
    Double = d; 
    Stime = s;
  };

  /*************************************************************************
   * Method: REDOP_INFO
   * Summary: basic constructor with no parameters
   * Description: basic constructor with no parameters.  Does nothing/.
   * Return: REDOP_INFO
   *************************************************************************/
  REDOP_INFO(){};

  /*************************************************************************
   * Method: REDOP_INFO
   * Summary: constructs a REDOP_INFO object with particular values
   * Description: Again, just like init, we can create a REDOP_INFO
   *              from the start to have the correct values.
   * Return: the constructed REDOP_INFO
   *************************************************************************/
  REDOP_INFO(int I,    //In: The integer value for the reduction
	     double d, //In: the double value for the reduction
	     SpSimTime S //In: the SpSimTime value for the reduction
	     ){
    this->Init(I, d, S);
  };

  /*************************************************************************
   * Method: GetInto
   * Summary: returns the integer value of REDOP_INFO
   * Description: returns the integer value of REDOP_INFO
   * Return: returns the integer value of REDOP_INFO
   *************************************************************************/
  int GetInt(){
    return Int;
  }

  /*************************************************************************
   * Method: GetDouble
   * Summary: returns the double value of REDOP_INFO
   * Description: returns the double value of REDOP_INFO
   * Return: returns the double value of REDOP_INFO
   *************************************************************************/
  double GetDouble(){ 
    return Double;
  }

  /*************************************************************************
   * Method: GetSimTime
   * Summary: returns the SpSimTime value of REDOP_INFO
   * Description: returns the SpSimTime value of REDOP_INFO
   * Return: returns the SpSimTime value of REDOP_INFO
   *************************************************************************/
  SpSimTime GetSimTime(){
    return Stime;
  }
};

#endif

