// SpDataParserTypes.C file

#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpStdlib.H"

#include "SpDataParserTypes.H"

BASE_DATA_TYPE::~BASE_DATA_TYPE()
{
  Qtypes.clear();
}

BASE_DATA_TYPE::BASE_DATA_TYPE(C_QUEUE *Qatoms) :
                              FUNDAMENTAL_DATA_TYPE(Qatoms) {
  SpATOM *Atom;
  INT_DATA_TYPE *IntDataType;
  FLOAT_DATA_TYPE *FloatDataType;
  LOGICAL_DATA_TYPE *LogicalDataType;
  STRING_DATA_TYPE *StringDataType;
  DEFINE_DATA_TYPE *DefineDataType;
  ENUM_DATA_TYPE *EnumDataType;
  BASE_DATA_TYPE *BaseDataType;
  REFERENCE_DATA_TYPE *ReferenceDataType;

  Nenum = 0;
  Nreference = 0;
  LastReference = NULL;

  Atom = (SpATOM *)Qatoms->pop_top();
  if (!strcmp(Atom->GetSymbol(), "{")) {

    Type = NULL;
    delete Atom;

  }else{

    Type = Name;
    Name = strdup(Atom->GetSymbol());
    delete Atom;

    Atom = (SpATOM *)Qatoms->pop_top();
    if (strcmp(Atom->GetSymbol(), "{")) {
      cout << "Error (BASE_DATA_TYPE): Expect { but got "
	   << Atom->GetSymbol() << endl;
      exit(0);
    }
    delete Atom;

  }

/*
** Now parse through the symbols. Note, we return when we receive a "}" here
*/

  while (1) {

    Atom = (SpATOM *)Qatoms->pop_top();

    if (!strcmp(Atom->GetSymbol(), "}")) {
      delete Atom;
      return;
    }

    while (1) {

      if (!strcasecmp(Atom->GetSymbol(), "int")) {
        delete Atom;
        IntDataType = new INT_DATA_TYPE(Qatoms);
	Qtypes.push_bot(IntDataType);
        break;
      }

      if (!strcasecmp(Atom->GetSymbol(), "float")) {
        delete Atom;
        FloatDataType = new FLOAT_DATA_TYPE(Qatoms);
	Qtypes.push_bot(FloatDataType);
        break;
      }

      if (!strcasecmp(Atom->GetSymbol(), "logical")) {
        delete Atom;
        LogicalDataType = new LOGICAL_DATA_TYPE(Qatoms);
	Qtypes.push_bot(LogicalDataType);
        break;
      }

      if (!strcasecmp(Atom->GetSymbol(), "string")) {
        delete Atom;
        StringDataType = new STRING_DATA_TYPE(Qatoms);
	Qtypes.push_bot(StringDataType);
        break;
      }

      if (!strcasecmp(Atom->GetSymbol(), "define")) {
        delete Atom;
        DefineDataType = new DEFINE_DATA_TYPE(Qatoms);
	Qtypes.push_bot(DefineDataType);
        break;
      }

      if (!strcasecmp(Atom->GetSymbol(), "enum")) {
        delete Atom;
        EnumDataType = new ENUM_DATA_TYPE(Qatoms, Nenum);
	Qtypes.push_bot(EnumDataType);
	Nenum++;
        break;
      }

      if (!strcasecmp(Atom->GetSymbol(), "reference")) {
        delete Atom;
        ReferenceDataType = new REFERENCE_DATA_TYPE(Qatoms);
	Qtypes.push_bot(ReferenceDataType);
	Nreference++;
        break;
      }

      Qatoms->push_top(Atom);
      BaseDataType = new BASE_DATA_TYPE(Qatoms);
      Qtypes.push_bot(BaseDataType);
      break;

    }

  }


}


FUNDAMENTAL_DATA_TYPE::FUNDAMENTAL_DATA_TYPE(C_QUEUE *Qatoms) {
  SpATOM *Atom;
  Atom = (SpATOM *)Qatoms->pop_top();
  Name = strdup(Atom->GetSymbol());
  delete Atom;
}

FUNDAMENTAL_DATA_TYPE::~FUNDAMENTAL_DATA_TYPE()
{
  free(Name);
}

INT_DATA_TYPE::INT_DATA_TYPE(C_QUEUE *Qatoms) : FUNDAMENTAL_DATA_TYPE(Qatoms) {
  SpATOM *Atom;
  Atom = (SpATOM *)Qatoms->pop_top();
  Value = atoi(Atom->GetSymbol());
  delete Atom;
}


FLOAT_DATA_TYPE::FLOAT_DATA_TYPE(C_QUEUE *Qatoms) : FUNDAMENTAL_DATA_TYPE(Qatoms) {
  SpATOM *Atom;
  Atom = (SpATOM *)Qatoms->pop_top();
  Value = atof(Atom->GetSymbol());
  delete Atom;
}


LOGICAL_DATA_TYPE::LOGICAL_DATA_TYPE(C_QUEUE *Qatoms) : FUNDAMENTAL_DATA_TYPE(Qatoms) {
  SpATOM *Atom;
  Atom = (SpATOM *)Qatoms->pop_top();

  if (!strcasecmp(Atom->GetSymbol(), "t")) {
    Value = 1;
    delete Atom;
    return;
  }

  if (!strcasecmp(Atom->GetSymbol(), "f")) {
    Value = 0;
    delete Atom;
    return;
  }

  delete Atom;
  cout << "Error (LOGICAL_DATA_TYPE): Bad Value" << endl;
  exit(0);

}


STRING_DATA_TYPE::STRING_DATA_TYPE(C_QUEUE *Qatoms) : FUNDAMENTAL_DATA_TYPE(Qatoms) {
  SpATOM *Atom;
  Atom = (SpATOM *)Qatoms->pop_top();
  Value = strdup(Atom->GetSymbol());
  delete Atom;
}

STRING_DATA_TYPE::~STRING_DATA_TYPE()
{
  free (Value);
}

DEFINE_DATA_TYPE::DEFINE_DATA_TYPE(C_QUEUE *Qatoms) : FUNDAMENTAL_DATA_TYPE(Qatoms) {
  SpATOM *Atom;
  Atom = (SpATOM *)Qatoms->pop_top();
  Value = strdup(Atom->GetSymbol());
  delete Atom;
}


ENUM_DATA_TYPE::ENUM_DATA_TYPE(C_QUEUE *Qatoms, int n) :
                                      FUNDAMENTAL_DATA_TYPE(Qatoms) {
  Value = n;
}


C_QUEUE *REFERENCE_DATA_TYPE::Qref;

REFERENCE_DATA_TYPE::REFERENCE_DATA_TYPE(C_QUEUE *Qatoms) :
                                      FUNDAMENTAL_DATA_TYPE(Qatoms) {
  REF_ITEM *RefItem;
  SpATOM *Atom;
  Atom = (SpATOM *)Qatoms->pop_top();
  Type = Name;
  Name = strdup(Atom->GetSymbol());
  Value = NULL;
  delete Atom;

  RefItem = new REF_ITEM();
  RefItem->RDT = (REFERENCE_DATA_TYPE *)this;
  Qref->push_bot(RefItem);

}




