// SpDataParserTypes.H file

#ifndef DataParserTypes_Object
#define DataParserTypes_Object

#include "queue.H"
#include "SpOS.H"
#include "SpStdlib.H"
#include "SpAssert.H"
#include "SpString.H"

class SpATOM : public C_ITEM {
  private:
    char *Symbol;
  public:
    SpATOM(char *s) {Symbol = strdup(s);}
    virtual ~SpATOM() {free(Symbol);}
    char *GetSymbol() {return Symbol;}
};


class BASE_DATA_TYPE;

class FUNDAMENTAL_DATA_TYPE : public C_ITEM {
  protected:

    // This data member would be "ContractNumber", in
    // the example, "int ContractNumber 1701"
    char *Name;

  public:
    FUNDAMENTAL_DATA_TYPE(C_QUEUE *Qatoms);
    virtual ~FUNDAMENTAL_DATA_TYPE();
    
    char *GetName() {return Name;}

    virtual void printlevel(ostream &out, int level) {
      for (int i=0; i<level; i++) out << "  ";
      out << "Fundamental Data Type: Name " << Name << endl;
    }

    virtual FUNDAMENTAL_DATA_TYPE *Find(char *FindName) {
      cout << "Error (FUNDAMENTAL_DATA_TYPE): Trying to find "
	   << FindName << endl;
      return NULL;
    }

    virtual char *GetType() {return NULL;}

    virtual int GetNtypes(char *T) {
      cout << "ERROR (FUNDAMENTAL_DATA_TYPE): GetNtypes " << T << endl;
      return 0;
    }

    virtual BASE_DATA_TYPE *GetBDT() {
      cout << "ERROR (FUNDAMENTAL_DATA_TYPE): Bad call to GetBDT" << endl;
      abort();
      return NULL;
    }

};


class BASE_DATA_TYPE : public FUNDAMENTAL_DATA_TYPE {
  private:
    int Nenum;
    int Nreference;
    char *Type;
    C_QUEUE Qtypes;
    FUNDAMENTAL_DATA_TYPE *LastReference;

  public:
    BASE_DATA_TYPE(C_QUEUE *Qatoms);
    ~BASE_DATA_TYPE();

    int GetNenum() {return Nenum;}
    int GetNreference() {return Nreference;}
    C_QUEUE *GetQtypes() {return &Qtypes;}

    int GetArraySize(char *n) {
      char s[120];
      int ArraySize;

      if (Exists(n)) {
	cout << "Error (BASE_DATA_TYPE): " << n << " is not an array" << endl;
	exit(1);
      }

      for (ArraySize=0; ArraySize<1000000; ArraySize++) {
	sprintf(s,"%s[%d]", n, ArraySize);
	if (!Exists(s)) break;
      }
      return ArraySize;

    }

    virtual BASE_DATA_TYPE *GetBDT() {return this;}

    virtual int GetNtypes(char *T) {
      int Ntypes = 0;
      FUNDAMENTAL_DATA_TYPE *FDT = (FUNDAMENTAL_DATA_TYPE *)Qtypes.get_top();
      for (int i=0; i<Qtypes.get_length(); i++) {
	 if (FDT->GetType() != NULL && !strcasecmp(FDT->GetType(), T))
	   Ntypes++; // bug fix JSB
	 FDT = (FUNDAMENTAL_DATA_TYPE *)FDT->get_link();
      }
      return Ntypes;
    }

    virtual char *GetType() {return Type;}

    virtual void printlevel(ostream &out, int level) {
      int i;
      for (i=0; i<level; i++) out << "  ";
      if (Type == NULL) {
        out << Name << " {" << endl;
      }else{
        out << Type << " " << Name << " {" << endl;
      }
      FUNDAMENTAL_DATA_TYPE *FDT = (FUNDAMENTAL_DATA_TYPE *)Qtypes.get_top();
      for (i=0; i<Qtypes.get_length(); i++) {
	FDT->printlevel(out, level+1);
	FDT = (FUNDAMENTAL_DATA_TYPE *)FDT->get_link();
      }
      for (i=0; i<level; i++) out << "  ";
      out << "}" << endl;
    }

    virtual FUNDAMENTAL_DATA_TYPE *Find(char *FindName) {
      FUNDAMENTAL_DATA_TYPE *FDT = (FUNDAMENTAL_DATA_TYPE *)Qtypes.get_top();
      for (int i=0; i<Qtypes.get_length(); i++) {
	if (!strcmp(FDT->GetName(), FindName)) return FDT;
	FDT = (FUNDAMENTAL_DATA_TYPE *)FDT->get_link();
      }
      cout << "Error (BASE_DATA_TYPE): Trying to find " << FindName << endl;
      return NULL;
    }

    virtual FUNDAMENTAL_DATA_TYPE *FindFirstDefine() {
      FUNDAMENTAL_DATA_TYPE *FDT = (FUNDAMENTAL_DATA_TYPE *)Qtypes.get_top();
      for (int i=0; i<Qtypes.get_length(); i++) {
	if (FDT->GetType() != NULL && !strcasecmp(FDT->GetType(), "define"))
	  return FDT; // bug fix JSB
	FDT = (FUNDAMENTAL_DATA_TYPE *)FDT->get_link();
      }
      return NULL;
    }

    virtual FUNDAMENTAL_DATA_TYPE *FindNextDefine(FUNDAMENTAL_DATA_TYPE *Start) {
      if (Start == NULL) return NULL;
      FUNDAMENTAL_DATA_TYPE *FDT = Start;
      while (FDT != Qtypes.get_bot()) {
	FDT = (FUNDAMENTAL_DATA_TYPE *)FDT->get_link();
	if (FDT->GetType() != NULL && !strcasecmp(FDT->GetType(), "define"))
	  return FDT; // bug fix JSB
      }
      return NULL;
    }

    int GetNumReferences(char *ReferenceName) {
      if (ReferenceName == NULL)
	return Qtypes.get_length();
	  
      int NumReferences = 0;
      FUNDAMENTAL_DATA_TYPE *FDT = (FUNDAMENTAL_DATA_TYPE *)Qtypes.get_top();
      for (int i=0; i<Qtypes.get_length(); i++) {
	char *FDT_type = FDT->GetType();
	if ((FDT_type != NULL) && (!strcasecmp(FDT_type, ReferenceName)))
	  ++NumReferences;
	FDT = (FUNDAMENTAL_DATA_TYPE *)FDT->get_link();
      }
      return NumReferences;
    }

    FUNDAMENTAL_DATA_TYPE *GoToFirstReference(char *ReferenceName) {
      FUNDAMENTAL_DATA_TYPE *FDT = (FUNDAMENTAL_DATA_TYPE *)Qtypes.get_top();
      if (ReferenceName == NULL)
      {
	LastReference = FDT;
	return FDT;
      }
      for (int i=0; i<Qtypes.get_length(); i++) {
	LastReference = FDT;
	if (FDT->GetType() != NULL &&
	    !strcasecmp(FDT->GetType(), ReferenceName))
	  return FDT; // bug fix JSB
	FDT = (FUNDAMENTAL_DATA_TYPE *)FDT->get_link();
      }
      return NULL;
    }

    FUNDAMENTAL_DATA_TYPE *GoToNextReference(char *ReferenceName) {
      if (LastReference == NULL) return NULL;
      if (LastReference == (FUNDAMENTAL_DATA_TYPE *)Qtypes.get_bot()) {
	LastReference = NULL;
	return LastReference;
      }

      LastReference = (FUNDAMENTAL_DATA_TYPE *)LastReference->get_link();
      if (ReferenceName == NULL)
        return LastReference;
      for (;;) {
	if (LastReference->GetType() != NULL &&
	    !strcasecmp(LastReference->GetType(), ReferenceName))
	  return LastReference; // bug fix JSB
	if (LastReference == (FUNDAMENTAL_DATA_TYPE *)Qtypes.get_bot()) break;
	LastReference = (FUNDAMENTAL_DATA_TYPE *)LastReference->get_link();
      }
      LastReference = NULL;
      return LastReference;
    }
   
    int Exists(char *FindName) {
      FUNDAMENTAL_DATA_TYPE *FDT = (FUNDAMENTAL_DATA_TYPE *)Qtypes.get_top();
      for (int i=0; i<Qtypes.get_length(); i++) {
	if (!strcmp(FDT->GetName(), FindName)) return 1;
	FDT = (FUNDAMENTAL_DATA_TYPE *)FDT->get_link();
      }
      return 0;
    }

};


class INT_DATA_TYPE : public FUNDAMENTAL_DATA_TYPE {
  private:
    int Value;
  public:
    INT_DATA_TYPE(C_QUEUE *Qatoms);
    int GetValue() {return Value;}
    virtual char *GetType() {return "int";}
    virtual void printlevel(ostream &out, int level) {
      for (int i=0; i<level; i++) out << "  ";
      out << "int " << Name << " " << Value << endl;
    }
};


class FLOAT_DATA_TYPE : public FUNDAMENTAL_DATA_TYPE {
  private:
    double Value;
  public:
    FLOAT_DATA_TYPE(C_QUEUE *Qatoms);
    double GetValue() {return Value;}
    virtual char *GetType() {return "float";}
    virtual void printlevel(ostream &out, int level) {
      for (int i=0; i<level; i++) out << "  ";
      out << "float " << Name << " " << Value << endl;
    }
};


class LOGICAL_DATA_TYPE : public FUNDAMENTAL_DATA_TYPE {
  private:
    int Value;
  public:
    LOGICAL_DATA_TYPE(C_QUEUE *Qatoms);
    int GetValue() {return Value;}
    virtual char *GetType() {return "logical";}
    virtual void printlevel(ostream &out, int level) {
      for (int i=0; i<level; i++) out << "  ";
      if (Value) {
        out << "logical " << Name << " T"<< endl;
      }else{
        out << "logical " << Name << " F" << endl;
      }
    }
};


class STRING_DATA_TYPE : public FUNDAMENTAL_DATA_TYPE {
  private:
    char *Value;
  public:
    STRING_DATA_TYPE(C_QUEUE *Qatoms);
    ~STRING_DATA_TYPE();
    char *GetValue() {return Value;}
    virtual char *GetType() {return "string";}
    virtual void printlevel(ostream &out, int level) {
      for (int i=0; i<level; i++) out << "  ";
      out << "string " << Name << " " << Value << endl;
    }
};


class DEFINE_DATA_TYPE : public FUNDAMENTAL_DATA_TYPE {
  private:
    char *Value;
  public:
    DEFINE_DATA_TYPE(C_QUEUE *Qatoms);
    ~DEFINE_DATA_TYPE() { free(Value); }
    char *GetValue() {return Value;}
    virtual char *GetType() {return "define";}
    virtual void printlevel(ostream &out, int level) {
      for (int i=0; i<level; i++) out << "  ";
      out << "define " << Name << " " << Value << endl;
    }
};


class ENUM_DATA_TYPE : public FUNDAMENTAL_DATA_TYPE {
  private:
    int Value;
  public:
    ENUM_DATA_TYPE(C_QUEUE *Qatoms, int n);
    int GetValue() {return Value;}
    virtual char *GetType() {return "enum";}
    virtual void printlevel(ostream &out, int level) {
      for (int i=0; i<level; i++) out << "  ";
      out << "enum " << Name << " " << Value << endl;
    }
};


class REFERENCE_DATA_TYPE : public FUNDAMENTAL_DATA_TYPE {
  private:
    BASE_DATA_TYPE *Value;
    char *Type;
    static C_QUEUE *Qref;
  public:

    REFERENCE_DATA_TYPE(C_QUEUE *Qatoms);
    ~REFERENCE_DATA_TYPE() { free(Type); }

    BASE_DATA_TYPE *GetValue() {return Value;}
    void SetValue(BASE_DATA_TYPE *BDT) {Value = BDT;}
    void SetQref(C_QUEUE *Q) {Qref = Q;}

    virtual char *GetType() {return Type;}
    virtual void printlevel(ostream &out, int level) {
      for (int i=0; i<level; i++) out << "  ";
      out << "reference " << Type << " " << Name << endl;
    }

    virtual FUNDAMENTAL_DATA_TYPE *Find(char *FindName) {
      return Value->Find(FindName);
    }

    virtual int GetNtypes(char *T) {
      return Value->GetNtypes(T);
    }

    virtual BASE_DATA_TYPE *GetBDT() {return Value;}

};


class REF_ITEM : public C_ITEM {
  public:
    REFERENCE_DATA_TYPE *RDT;
};


#endif
