// SpDestination.H header file defines sets of nodes for multicasting in
// SPEEDES. It can be also used for defining sets of nodes

#ifndef DestinationObject
#define DestinationObject

#include "SpBitField.H"

/*
** Class definition for defining destinations when multicasting. Note that
** destination-based multicasting is not the normal way people thing about
** multicasting with multicast groups. However, this is the way SPEEDES
** defines multicasting. Optimizations for supporting this service can be
** made available on shared memory, networks, etc., so it is useful to provide
** this capability.
*/

class SpDestination: public SpBitField
{
public:
  
  SpDestination(): SpBitField() {}
  
  SpDestination(int Nnodes): SpBitField(Nnodes) {}
  
  void Reset() {
    SpBitField::Reset();
  }
  
  void Reset(int node) {
    SpBitField::Reset(node);
  }
  
  void Set(int node) {
    SpBitField::Set(node);
  }
  
  int Check(int node) {
    return SpBitField::Check(node);
  }
  
  void print(ostream &out) {
    SpBitField::print(out);
  }
  
  void SetNnodes(int nNodes) {
    SpBitField::SetNbits(nNodes);
  }
  
  int GetNnodes() {return SpBitField::GetNbits();}
  int GetNbytes() {return SpBitField::GetNbytes();}
  char *GetDestination() {return SpBitField::GetBitField();}
  
  void operator += (int node) {SpBitField::Set(node);}
  void operator -= (int node) {SpBitField::Reset(node);}
  int operator == (int node) {return SpBitField::Check(node);}
  
  int GetFirstNode() {
    return SpBitField::GetFirstBit();
  }
  
  int GetNextNode(int node) {
    return SpBitField::GetNextBit(node);
  }
  
  char * PackBuffer(int & Nbytes){
    return SpBitField::PackBuffer(Nbytes);
  }
  
  void UnpackBuffer(char * buff){
    SpBitField::UnpackBuffer(buff);
  }
};

typedef SpDestination DESTINATION;


#endif
