#include "SpSignal.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpAssert.H"
#include "SpStdlib.H"
#include "SpErrorSignals.H"
#include "SpComm.H"
#include "SpDataParser.H"

static void (*ErrorSignalsHandler)();

static void sigerrorfunc(int sig)
{
  ErrorSignals(ErrorSignalsHandler);
  
  cerr << "Got signal ";
  switch (sig)
  {
#if (!defined (__IBMCPP__) & !defined (__SPEEDES_NT__))
    case    SIGHUP: cerr <<    "SIGHUP"; break;
    case   SIGQUIT: cerr <<   "SIGQUIT"; break;
    case    SIGILL: cerr <<    "SIGILL"; break;
    case   SIGABRT: cerr <<   "SIGABRT"; break;
    case    SIGFPE: cerr <<    "SIGFPE"; break;
    case   SIGSEGV: cerr <<   "SIGSEGV"; break;
    case   SIGPIPE: cerr <<   "SIGPIPE"; break;
    case   SIGALRM: cerr <<   "SIGALRM"; break;
    case   SIGTERM: cerr <<   "SIGTERM"; break;
    case    SIGBUS: cerr <<    "SIGBUS"; break;
    case     SIGIO: cerr <<     "SIGIO"; break;
    case SIGVTALRM: cerr << "SIGVTALRM"; break;
           default: cerr <<         sig; break;
#endif
  }
  cerr << endl;
  
  ErrorSignalsHandler();
  
  void (*result)(int) = signal(SIGABRT, SIG_DFL);
#if (!defined (__IBMCPP__) & !defined (__SPEEDES_NT__))
  if (result == SIG_ERR)
    cerr << "Warning: Could not reset signal to default." << endl;
  else
    cerr << "Successfully reset SIGABRT to default." << endl;
#endif
    
  abort();
  cerr << "\n\n!!!!!!!! Should not be here\n\n" << endl;
}

static void set_sigerrorfunc(int sig)
{
  void (*result)(int) = signal(sig, sigerrorfunc);
#if (!defined (__IBMCPP__) & !defined (__SPEEDES_NT__))
  assert(result != SIG_ERR);
#endif
}

void ErrorSignals(void (*handler)())
{
  ErrorSignalsHandler = handler;
  SpDataParser parser("speedes.par");
  int haveSignalsSection = (parser.GoTo("SignalHandlers") != NULL);
  
#if (!defined (__IBMCPP__) & !defined (__SPEEDES_NT__))
  if (haveSignalsSection){
    if (parser.GetLogical("SIGHUP")){
      set_sigerrorfunc(SIGHUP);
    }
    if (parser.GetLogical("SIGQUIT")){
      set_sigerrorfunc(SIGQUIT);
    }
    if (parser.GetLogical("SIGILL")){
      set_sigerrorfunc(SIGILL);
    }
    if (parser.GetLogical("SIGABRT")){
      set_sigerrorfunc(SIGABRT);
    }
    if (parser.GetLogical("SIGFPE")){
      set_sigerrorfunc(SIGFPE);
    }
    if (parser.GetLogical("SIGSEGV")){
      set_sigerrorfunc(SIGSEGV);
    }
    if (parser.GetLogical("SIGPIPE")){
      set_sigerrorfunc(SIGPIPE);
    }
    if (parser.GetLogical("SIGALRM")){
      set_sigerrorfunc(SIGALRM);
    }
    if (parser.GetLogical("SIGTERM")){
      set_sigerrorfunc(SIGTERM);
    }
    if (parser.GetLogical("SIGBUS")){
      set_sigerrorfunc(SIGBUS);
    }
    //set_sigerrorfunc(SIGSTKFLT);
    if (parser.GetLogical("SIGIO")){
      set_sigerrorfunc(SIGIO);
    }
    //set_sigerrorfunc(SIGXCPU);
    //set_sigerrorfunc(SIGXFSZ);
    if (parser.GetLogical("SIGVTALRM")){
      set_sigerrorfunc(SIGVTALRM);
    }
  }
  else{
    set_sigerrorfunc(SIGHUP);
    set_sigerrorfunc(SIGQUIT);
    set_sigerrorfunc(SIGILL);
    set_sigerrorfunc(SIGABRT);
    set_sigerrorfunc(SIGFPE);
    set_sigerrorfunc(SIGSEGV);
    set_sigerrorfunc(SIGPIPE);
    set_sigerrorfunc(SIGALRM);
    set_sigerrorfunc(SIGTERM);
    set_sigerrorfunc(SIGBUS);
    //set_sigerrorfunc(SIGSTKFLT);
    set_sigerrorfunc(SIGIO);
    //set_sigerrorfunc(SIGXCPU);
    //set_sigerrorfunc(SIGXFSZ);
    set_sigerrorfunc(SIGVTALRM);
  }
#endif
}
