// SpNameHandle.H header file is a class used by the Host User to receive the
// names and handles <Node, Object Type, Local Id> of each object instance.

#ifndef SpNameHandle_OBJECT
#define SpNameHandle_OBJECT

#include "SpString.H"
#include "SpMemory.H"
#include "SpStdlib.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"

#include "qitem.H"
#include "SpNetTypes.H"

#define EIGHT_BYTE_ALIGNMENT 8

class SpNameHandle : public C_ITEM {

  private:

    NET_INT Node;		// Node number for the object
    NET_INT ObjectType;	// Type of object
    NET_INT LocalId;	// Local Id for the object
    NET_INT GlobalId;	// GlobalId for the object
    char *Name;		// Name of the object

  protected:

  public:

    SpNameHandle() {
      Node = 0;
      ObjectType = 0;
      LocalId = 0;
      GlobalId = 0;
      id = 0;
      Name = NULL;
    }

    SpNameHandle(SpNameHandle *NameHandle) {
      memcpy((char *)this, (char *)NameHandle, sizeof(SpNameHandle));
      Name = strdup(NameHandle->GetName());
    }

    SpNameHandle(int Nd, int Otype, int Lid, int Gid, char *Nm) {
      SetValues(Nd, Otype, Lid, Gid, Nm);
    }

    virtual ~SpNameHandle() {
      if (Name) {
	free(Name);
      }else{
	cout << "Funny thing, trying to free a Name that is NULL" << endl;
      }
      Name = NULL;
    }

    void SetValues(int Nd, int Otype, int Lid, int Gid, char *Nm) {
      if (Nm) {
	Node = Nd;
	ObjectType = Otype;
	LocalId = Lid;
	id = Gid;
	GlobalId = Gid;
	Name = strdup(Nm);
      }else{
	cout << "Error, NULL Name for SpNameHandle" << endl;
      }
    }

    void SetNode(int Nd) {Node = Nd;}
    void SetObjectType(int Otype) {ObjectType = Otype;}
    void SetLocalId(int Lid) {LocalId = Lid;}
    void SetGlobalId(int Gid) {id = Gid; GlobalId = Gid;}
    void SetName(char *Nm)
    {
      if (Name != NULL)
      {
        free(Name);
      }
      Name = strdup(Nm);
    }


    void GetValues(int &Nd, int &Otype, int &Lid, int &Gid, char *&Nm) {
      Nd = Node;
      Otype = ObjectType;
      Lid = LocalId;
      Gid = GlobalId;
      Nm = Name;
    }

    int GetNode() {return Node;}
    int GetObjectType() {return ObjectType;}
    int GetLocalId() {return LocalId;}
    int GetGlobalId() {return GlobalId;}
    char *GetName() {return Name;}

    int GetBufferSize() {
      int s = sizeof(SpNameHandle) + strlen(Name) + 1;
      if (sizeof(SpNameHandle)%EIGHT_BYTE_ALIGNMENT){
	s += EIGHT_BYTE_ALIGNMENT - sizeof(SpNameHandle)%EIGHT_BYTE_ALIGNMENT;
      }
      if (s%EIGHT_BYTE_ALIGNMENT){
	s += EIGHT_BYTE_ALIGNMENT - s%EIGHT_BYTE_ALIGNMENT;
      }
      return s;
    }

    char *FillBuffer(char *Buff) {
      int s1 = sizeof(SpNameHandle);
      if (sizeof(SpNameHandle)%EIGHT_BYTE_ALIGNMENT) {
	s1 += EIGHT_BYTE_ALIGNMENT - sizeof(SpNameHandle)%EIGHT_BYTE_ALIGNMENT;
      }
      memcpy(Buff, (char *)this, sizeof(SpNameHandle));
      int s2 = strlen(Name) + 1;
      memcpy(&Buff[s1], Name, s2);
      int s3 = s1 + s2;
      if (s3%EIGHT_BYTE_ALIGNMENT) {
	s3 += EIGHT_BYTE_ALIGNMENT - s3%EIGHT_BYTE_ALIGNMENT;
	for (int i=s1+s2; i<s3; i++) Buff[i] = 0;
      }
      return &Buff[s3];
    }

    char *EmptyBuffer(char *Buff) {
      int s = sizeof(SpNameHandle);
      char *bf1 = (char *)this + sizeof(C_DUMMY_V);
      char *bf2 = Buff + sizeof(C_DUMMY_V);
      memcpy(bf1, bf2, s - sizeof(C_DUMMY_V));
      if (sizeof(SpNameHandle)%EIGHT_BYTE_ALIGNMENT){
	s += EIGHT_BYTE_ALIGNMENT - sizeof(SpNameHandle)%EIGHT_BYTE_ALIGNMENT;
      }
      Name = strdup(&Buff[s]);
      return &Buff[GetBufferSize()];
    }

    virtual void printstream(ostream &out) {
      out << "NameHandle: " << Name << "(" << GetGlobalId() << ") <"
	  << GetNode() << ", "
	  << GetObjectType() << ", "
	  << GetLocalId() << ">" 
	  << endl;
    }

};

typedef SpNameHandle NAME_HANDLE;

#endif
