#include "SpNetTypes.H"

#if HETEROGENEOUS_DATA_TYPE

void NET_INT::BigEndianToLittleEndian (void * v) const{
    NET_INT_TYPE *val = (NET_INT_TYPE*)v;
#if BASE_INT_IS_EIGHT_BYTES
#ifdef HP_NO_GNU
    *val >>= 32;
#endif
#endif
    *val = ntohl(*val);
  }

  
void NET_INT::LittleEndianToBigEndian(void * val) const{
  
  char * ptr = (char *)val;
  char temp = *ptr;
  ptr[0] = ptr[3];
  ptr[3] = temp;
  temp = ptr[1];
  ptr[1] = ptr[2];
  ptr[2] = temp;
  
#ifndef HP_NO_GNU
#if BASE_INT_IS_EIGHT_BYTES
  NET_INT_TYPE *v = (NET_INT_TYPE *)val;
  *v >>= 32;
#endif
#endif
}


void NET_INT::Convert() const
{
  //Not our machine type so now we have to convert to our type
  void * V = (void *)&Val;
  if (*TypeStatus == SpEndianNature::BigEndian){
    BigEndianToLittleEndian(V);
  }
  else{
#ifdef HP_NO_GNU
#if BASE_INT_IS_EIGHT_BYTES
    /* Since the HP does not have an 8 byte int without GNU, we have
       an additional overhad of the EightByteGarbage.  A little endian
       machine will have it's actual info in these 4 bytes rather than
       in the upper 4 bytes in val */
    
    V = (char *)&EightByteGarbage;
#endif
#else
    
    V = (char *)&Val;
#endif
    LittleEndianToBigEndian(V);
    
  }
  V = (void *)TypeStatus;
  SetEndianNature(V);
}

void NET_DOUBLE::BigEndianToLittleEndian(void * v) const{
  char * ptr = (char *)v;

  char temp = *ptr;
  ptr[0] = ptr[7];
  ptr[7] = temp;

  temp = ptr[1];
  ptr[1] = ptr[6];
  ptr[6] = temp;

  temp = ptr[2];
  ptr[2] = ptr[5];
  ptr[5] = temp;

  temp = ptr[3];
  ptr[3] = ptr[4];
  ptr[4] = temp;

}

void NET_DOUBLE::Convert() const
{
  //Not our machine type so now we have to convert to our type
  void * V = (void *)&Val;
  if (*TypeStatus == SpEndianNature::BigEndian){
    BigEndianToLittleEndian(V);
  }
  else{
    LittleEndianToBigEndian(V);
  }
  V = (void *)TypeStatus;
  SetEndianNature(V);
}


void NET_INT::Display()
{
  char * v = (char *)&Val;
  cout<<Val<<"====";
  for (unsigned int i=0; i < sizeof(NET_INT_TYPE); ++i) cout<<(int)*v++<<" ";
  cout<<endl;
}

#endif
