#ifndef NETTYPES_H
#define NETTYPES_H
#ifndef __SPEEDES_NT__
#include <sys/types.h>
#include <netinet/in.h>
#endif
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpEndianNature.H"

/* Hetero ints only use 4 of the 8 bytes that are allocated.  This (on
  most 64 bit machines, is the right 4 bytes of the NET_INT_TYPE val.
  On an HP, we sort of fudge this and only use 4 bytes (the largest
  integer allowed).  */


/* Set the following to 1 if you want to use 8 bytes integers as the
 * base for NET_INTS.  Otherwise, NET_INTS will default to having a 4
 * byte integer.  */
#define BASE_INT_IS_EIGHT_BYTES 0


/* If you are running across multiple architectures and need
 * heterogeneous operations, set the following to 1.  Otherwise, if
 * all the machines are homogeneous in nature, set it to 0
 * This supercedes BASE_INT_IS_EIGHT_BYTES!  If you run heterogeneous
 * then NET_INTS are typedefed to int and NET_DOUBLES are typedef'ed
 * to double.
 */

#define HETEROGENEOUS_DATA_TYPE 0


// geb: Why is this here??? #if HETEROGENEOUS_DATA_TYPE



#if BASE_INT_IS_EIGHT_BYTES

#ifdef __GNUC__  //GNU machine always has 8 byte integers if I need them
#define BASE_INT_TYPE_DEFINED
typedef long long int NET_INT_TYPE;
#endif

#ifndef BASE_INT_TYPE_DEFINED
#ifdef sun
#define BASE_INT_TYPE_DEFINED
typedef long long int NET_INT_TYPE;
#endif
#endif

#ifndef BASE_INT_TYPE_DEFINED
#ifdef __sparc
#define BASE_INT_TYPE_DEFINED
typedef long long int NET_INT_TYPE;
#endif
#endif

#ifndef BASE_INT_TYPE_DEFINED
#ifdef hpux
#define HP_NO_GNU //On a HP without GNU meaning we don't have 8 byte integers
#define BASE_INT_TYPE_DEFINED
typedef long int NET_INT_TYPE;
#endif
#endif

#ifndef BASE_INT_TYPE_DEFINED
#define BASE_INT_TYPE_DEFINED
typedef long long int NET_INT_TYPE;
#endif



#else //BASE_INT_IS_EIGHT_BYTES

#define BASE_INT_TYPE_DEFINED
typedef int NET_INT_TYPE;

#endif //BASE_INT_IS_EIGHT_BYTES



#if HETEROGENEOUS_DATA_TYPE



class NET_INT{

private:


  void Display();


#if BASE_INT_IS_EIGHT_BYTES
#ifdef HP_NO_GNU
  /* This must be first in the structure so that the "upper 4 bytes"
     get copied into val rather than into EightByteGarbage for HP's
     since NET_INT_TYPE is only 4 bytes on HP's w/o gcc.
     */
  NET_INT_TYPE EightByteGarbage;
#endif
#endif

  NET_INT_TYPE Val;
  char TypeStatus[sizeof(NET_INT_TYPE)];
#ifdef HP_NO_GNU
  char TypeStatusJunk[sizeof(NET_INT_TYPE)];
#endif



  void BigEndianToLittleEndian(void* val) const;
  void LittleEndianToBigEndian(void * val) const;

  void SetEndianNature(void * s) const{
    char * ss = (char *)s;
    *ss = SpEndianNature::GetEndianNature();
  }


  void Set(int HostVal) {
    *TypeStatus = SpEndianNature::GetEndianNature();
    Val = HostVal;
#ifdef HP_NO_GNU
#if BASE_INT_IS_EIGHT_BYTES
    EightByteGarbage = 0;
#endif
#else
#endif

  }

  void Convert() const;

  int Get() const{
    if (*TypeStatus != SpEndianNature::GetEndianNature()){
      Convert();
    }
    return Val;
  };

public:

  NET_INT(int a = 0){
    Set(a);
  }


  operator NET_INT_TYPE () const { //Convert to an int
    Get();
    return Val;
  }


  NET_INT& operator = (const NET_INT &netint) {
    Val = netint.Val; 
    TypeStatus[0] = netint.TypeStatus[0]; // don't forget this!
    return *this; 
  }
  int operator = (int x)   {Set(x); return x;}
  int operator -- ()       {*this -= 1; return Val;}
  int operator ++ ()       {*this += 1; return Val;}
  int operator -- (int)    {*this -= 1; return (Val+1);} // postfix
  int operator ++ (int)    {*this += 1; return (Val-1);} // postfix
  int operator += (int x)  { Get(); Val += x; return Val;}
  int operator -= (int x)  { Get(); Val -= x; return Val;}
  int operator *= (int x)  { Get(); Val *= x; return Val;}
  int operator /= (int x)  { Get(); Val /= x; return Val;}
  int operator %= (int x)  { Get(); Val %= x; return Val;}
  int operator ^= (int x)  { Get(); Val ^= x; return Val;}
  int operator &= (int x)  { Get(); Val &= x; return Val;}
  int operator |= (int x)  { Get(); Val |= x; return Val;}
  int operator >>= (int x) { Get(); Val >>= x; return Val;}
  int operator <<= (int x) { Get(); Val <<= x; return Val;}

};




class NET_DOUBLE{

private:

  double Val;
  char TypeStatus[sizeof(double)];


  void BigEndianToLittleEndian(void * v) const;

  void LittleEndianToBigEndian(void * v) const{
    BigEndianToLittleEndian(v);
  }

  void Set(double HostVal) {
    *TypeStatus = SpEndianNature::GetEndianNature();
    Val = HostVal;
  }

  void SetEndianNature(void * s) const{
    char * ss = (char *)s;
    *ss = SpEndianNature::GetEndianNature();
  }

  void Convert() const;

  double Get () const{
    if (*TypeStatus != SpEndianNature::GetEndianNature()){
      Convert();
    }
    return Val;
  };

public:

  NET_DOUBLE(double a = 0){
    Set(a);
  }

  operator double () const { //Convert to a double
    Get();
    return Val;
  }


  NET_DOUBLE& operator = (const NET_DOUBLE &netdoub) {
    Val = netdoub.Val; 
    TypeStatus[0] = netdoub.TypeStatus[0]; // don't forget this!
    return *this; 
  }
  double operator = (double x)  {Set(x); return x;}
  double operator ++ ()         {*this+=1;return Val;}
  double operator -- ()         {*this-=1;return Val;}
  double operator ++ (int)      {*this+=1;return (Val-1);}
  double operator -- (int)      {*this-=1;return (Val+1);}
  double operator += (double x) { Get(); Val += x; return Val;}
  double operator -= (double x) { Get(); Val -= x; return Val;}
  double operator *= (double x) { Get(); Val *= x; return Val;}
  double operator /= (double x) { Get(); Val /= x; return Val;}


};

typedef NET_DOUBLE NET_FLOAT;



#else // Homogeneous code:

#if 0

class NET_INT{

private:

  int Val;

public:

  NET_INT(int a = 0):Val(a){}

  NET_INT& operator = (const NET_INT &netint):Val(netint.Val){ return *this;}

  operator NET_INT_TYPE () const { //Convert to an int
    return Val;
  }

  int operator = (int x)   {Val = x; return x;}
  int operator -- ()       {*this -= 1; return Val;}
  int operator ++ ()       {*this += 1; return Val;}
  int operator -- (int)    {*this -= 1; return (Val+1);} // postfix
  int operator ++ (int)    {*this += 1; return (Val-1);} // postfix
  int operator += (int x)  {Val += x; return Val;}
  int operator -= (int x)  {Val -= x; return Val;}
  int operator *= (int x)  {Val *= x; return Val;}
  int operator /= (int x)  {Val /= x; return Val;}
  int operator %= (int x)  {Val %= x; return Val;}
  int operator ^= (int x)  {Val ^= x; return Val;}
  int operator &= (int x)  {Val &= x; return Val;}
  int operator |= (int x)  {Val |= x; return Val;}
  int operator >>= (int x) {Val >>= x; return Val;}
  int operator <<= (int x) {Val <<= x; return Val;}

};

class NET_DOUBLE{

private:

  double Val;

public:

  NET_DOUBLE(double a = 0):Val(a){}

  NET_DOUBLE& operator = (const NET_DOUBLE &netdoub):Val(netdoub.Val)
    { return *this; }

  operator double () const { //Convert to a double
    return Val;
  }

  double operator = (double x)  {Val = x; return x;}
  double operator ++ ()         {*this+=1;return Val;}
  double operator -- ()         {*this-=1;return Val;}
  double operator ++ (int)      {*this+=1;return (Val-1);}
  double operator -- (int)      {*this-=1;return (Val+1);}
  double operator += (double x) {Val += x; return Val;}
  double operator -= (double x) {Val -= x; return Val;}
  double operator *= (double x) {Val *= x; return Val;}
  double operator /= (double x) {Val /= x; return Val;}


};


#else
typedef double NET_DOUBLE;
typedef int NET_INT;

#endif

#endif //HETEROGENEOUS_DATA_TYPE

#endif
