#include "SpStdio.H"
#include "SpStdlib.H"
#include "SpErrno.H"
#include <sys/types.h>
#ifndef __SPEEDES_NT__
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <fcntl.h>
#else
#include <winsock.h>
#include <io.h>
#define ENOTSOCK                WSAENOTSOCK
#define EWOULDBLOCK             WSAEWOULDBLOCK
#endif
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"

#include "SpOS.H"
#include "SpSocketUtil.H"

struct hostent *get_host_by_name(char *name)
{
  return gethostbyname(name);
}

void read_extmessage(int operation);

/************************************************************************
*	set the socket to no delay (faster but not as efficient)	*
************************************************************************/
int set_nodelay(int sockno) {
  int flag;

#ifndef __SPEEDES_NT__
  flag = 1;

  if (setsockopt(	sockno,
			IPPROTO_TCP,
			TCP_NODELAY,
			(char *)&flag, sizeof(int)) < 0) {
    perror("_set_nodelay_");
    return -1;
  }

#ifndef Linux
  if (setsockopt(	sockno,
  			SOL_SOCKET,
			SO_USELOOPBACK,
			(char *)&flag, sizeof(int)) < 0) {
    perror("_set_nodelay_");
    return -1;
  }
#endif

#endif //__SPEEDES_NT__
  return 0;

}


/************************************************************************
*	blocking write							*
************************************************************************/
int socket_write(int sock, char *buff, int bytes) {
  int start;
  int status;

  start = 0;

  while (start < bytes) {

    status = write(sock, &buff[start], bytes-start);

    if (status < 0) {
      printf("Bad Status in blocking write %d\n", status);
      break;
    }

    start += status;

  }

  return status;

}


/************************************************************************
*	non-blocking write						*
************************************************************************/
int nb_socket_write(int sock,char *buff, int bytes) {
  int status;

  status = write(sock, buff, bytes);

  if (status < 0) {
    if (errno == EWOULDBLOCK) status = 0;
    printf("Bad Status in non-blocking write %d\n", status);
  }

  return status;

}


/************************************************************************
*	blocking read							*
************************************************************************/
int socket_read(int sock, char *buff, int bytes) {
  int start;
  int status;

  start = 0;

  while (start < bytes) {

    status = read(sock, &buff[start], bytes-start);

    if (status < 0) {
      printf("Bad Status in blocking read %d\n", status);
      return status;
    }else{
      start += status;
    }

  }

  return start;

}

/************************************************************************
*	non-blocking read						*
************************************************************************/
int nb_socket_read(int sock, char *buff, int bytes) {
  int status;

  status = read(sock, buff, bytes);

  if (status < 0) {
    //printf("Bad Status in non-blocking read %d\n", status);
    if (errno == EWOULDBLOCK) status = 0;
  }

  return status;

}


/************************************************************************
*	set a socket to be non-blocking					*
************************************************************************/
int set_nonblocking(int p_socket) {
  int fcntlarg;

#ifndef __SPEEDES_NT__
  if ((fcntlarg = fcntl(p_socket, F_GETFL, 0)) < 0) {
    return 0;
  }

//  fcntlarg |= O_NDELAY;
  fcntlarg |= O_NDELAY;
  if (fcntl(p_socket, F_SETFL, fcntlarg) < 0) {
    return 0;
  }

#endif
  return 1;

}

/************************************************************************
*	set a socket to be blocking					*
************************************************************************/
int set_blocking(int p_socket) {
  int fcntlarg;

#ifndef __SPEEDES_NT__
  if ((fcntlarg = fcntl(p_socket, F_GETFL, 0)) < 0) {
    return 0;
  }

//  fcntlarg &= ~O_NDELAY;
  fcntlarg &= ~O_NDELAY;
  if (fcntl(p_socket, F_SETFL, fcntlarg) < 0) {
    return 0;
  }

#endif
  return 1;

}



/*************** GEB7: 4/4/97 ********************/
int setSockBuffsize(int fd, int sendOrRecvBuff, int sizeInBytes) { //GEB7
   int ret, len=sizeof sizeInBytes;

   if (sizeInBytes > MAX_SOCK_BUFF) {
      cerr << "setSockBuffsize:  size > MAX_SOCK_BUFF;  reducing to MAX."  << endl;
      sizeInBytes = MAX_SOCK_BUFF;
   }

   if ((ret=setsockopt(fd, SOL_SOCKET, sendOrRecvBuff, (char *) &sizeInBytes, len)) < 0) {
      perror("setsockopt failure");
   }

   return ret;
}

int setSendBuffsize(int fd, int sizeInBytes) { //GEB7
   return setSockBuffsize(fd, SO_SNDBUF, sizeInBytes);
}

int setRecvBuffsize(int fd, int sizeInBytes) { //GEB7
   return setSockBuffsize(fd, SO_RCVBUF, sizeInBytes);
}
