// SpTag.H file defines a tag that can be used for supporting user-specified
// output to the trace file

#ifndef _SpTagObj
#define _SpTagObj

#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpString.H"


#define DEFAULT_TAG_BUFF_SIZE 256

class SpTag {

  private:

    char *Buff;
    int BuffSize;
    ostrstream *Out;
    SpTag *GlobalTag;

  protected:

  public:

    SpTag() {
      BuffSize = DEFAULT_TAG_BUFF_SIZE;
      Buff = NULL;
      Out = NULL;
    }

    SpTag(int buffSize) {
      BuffSize = buffSize;
      Buff = NULL;
      Out = NULL;
    }

    void SetBuffSize(int buffSize) {
      BuffSize = buffSize;
    }

    void Init();

    void Reset() {
      if (Buff) {
	delete [] Buff;
	Buff = NULL;
	delete Out;
	Out = NULL;
      }
    }

    char *GetBuff() {return Buff;}

    ostrstream &operator << (char *val);

    ostrstream &operator << (int val);

    ostrstream &operator << (unsigned int val);

    ostrstream &operator << (long val);

    ostrstream &operator << (unsigned long val);

    ostrstream &operator << (short val);

    ostrstream &operator << (unsigned short val);

    ostrstream &operator << (double val);

    ostrstream &operator << (float val);

    ostrstream &operator << (void *p);

};

ostream &operator << (ostream &out, SpTag tag);

#endif

//#define TEST_SP_TAG
#ifdef TEST_SP_TAG
main() {
  SpTag Tag;

  Tag << 1 << " Hello World " << 1 << " &tag " << &Tag << endl;
  Tag << 2.0 << " Hello World " << 2.0 << endl;

  cout << Tag;
  Tag.Reset();

  Tag << 12345 << endl;
  cout << "Hi " << Tag;

}
#endif
