 #ifndef RB_SpDefineClass_Header
#define RB_SpDefineClass_Header

#include "SpIostream.H"

#include "SpAltMgr.H"
#include "SpAltBufferChange.H"
#include "PO.H"
extern int PersistenceEnabled;
#define PERSISTENCE_ENABLED

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_NEW_HELPER( void* object, void ( *deleteFunction )( void* ) )
{
  SpAltBufferChange* altItem;

  if( object && SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltBufferChange();
    altItem->Alter( deleteFunction, object, 1 );
    SpCurrentAltMgr->Insert( altItem );
  }
}

/**********************************************************************
 * 
 **********************************************************************/
inline
void
RB_DELETE_HELPER( void* object, void ( *deleteFunction )( void* ) )
{
  SpAltBufferChange* altItem;

  if( object && SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltBufferChange();
    altItem->Alter( deleteFunction, object, 0 );
    SpCurrentAltMgr->Insert( altItem );
  }
  else {
    deleteFunction( object );
  }
}
#ifdef PERSISTENCE_ENABLED
#define RB_NEW(type)                                                    \
static									\
inline									\
type*									\
RB_NEW_##type( type* object ) {						\
  RB_NEW_HELPER( object, delete_single_##type );			\
  if (PersistenceEnabled){                                              \
    PO_ADD_##type(object);                                              \
  }								        \
  return object;							\
}

#define RB_DELETE(type)                                                 \
static									\
inline									\
void									\
RB_DELETE_##type( type* object ) {					\
  RB_DELETE_HELPER( object, delete_single_##type );			\
  if (PersistenceEnabled){                                              \
    PO_REMOVE_##type(object);                                           \
  }								        \
}

#else
#define RB_NEW(type)                                                    \
inline									\
type*									\
RB_NEW_##type( type* object ) {						\
  RB_NEW_HELPER( object, delete_single_##type );			\
  return object;							\
}

#define RB_DELETE(type)                                                 \
inline									\
void									\
RB_DELETE_##type( type* object ) {					\
  RB_DELETE_HELPER( object, delete_single_##type );			\
}

#endif
/**********************************************************************
 * Pass this macro the name of a class for which you would like to
 * make dynamically allocated objects or arrays.  This macro will
 * create functions which are safe to use in events that may be
 * rolled back. See following comment blocks for an explanation of
 * the functions created by this macro. Note that the class you pass
 * to this macro must have a constructor that can be called without
 * arguments.
 **********************************************************************/
#define RB_NEW_SINGLE_MACROS( type )					\
									\
inline									\
void									\
delete_single_##type( void* object ) {					\
  delete ( ( type* ) object );						\
}									\
									\
inline									\
void									\
delete_single_##type##_PTR( void* object ) {				\
  delete ( ( type** ) object );						\
}									\
									\
RB_NEW(type)                                                            \
									\
inline									\
type**									\
RB_NEW_##type##_PTR() {							\
  type** object = new type*;						\
  RB_NEW_HELPER( object, delete_single_##type##_PTR );			\
  return object;							\
}									\
									\
RB_DELETE(type)                                                         \
									\
inline									\
void									\
RB_DELETE_##type##_PTR( type** object ) {				\
  RB_DELETE_HELPER( object, delete_single_##type##_PTR );		\
}									\
									\
inline void RB_NEW_SINGLE_MACROS_warning_removal_function_##type() {	\
  /* These warning removal functions should never be */			\
  /* called.  Their purpose is to get rid of warnings */		\
  type* t = NULL;							\
  RB_NEW_##type( t );							\
  RB_NEW_##type##_PTR();						\
  RB_DELETE_##type( t );						\
  type** td = NULL;							\
  RB_DELETE_##type##_PTR( td );						\
									\
  void* v = RB_NEW_SINGLE_MACROS_warning_removal_function_##type;	\
									\
  /* This "v = v" line gets rid of the warning: */			\
  /* "variable "v"  was set but never used" */				\
  v = v;								\
}									\
/*********************************************************************/

/**********************************************************************
 * 
 **********************************************************************/
#define RB_NEW_SINGLE_MACRO_WITH_NO_ARGUMENT_CONSTRUCTOR( type )				\
												\
inline												\
type*												\
RB_NEW_##type() {										\
  type* object = new type;									\
  RB_NEW_HELPER( object, delete_single_##type );						\
  return object;										\
}												\
												\
inline												\
void												\
RB_NEW_SINGLE_MACRO_WITH_NO_ARGUMENT_CONSTRUCTOR_warning_removal_function_##type() {		\
  RB_NEW_##type();										\
  void* v = RB_NEW_SINGLE_MACRO_WITH_NO_ARGUMENT_CONSTRUCTOR_warning_removal_function_##type;	\
												\
  /* This "v = v" line gets rid of the warning: */						\
  /* "variable "v"  was set but never used" */							\
  v = v;											\
}												\
/*********************************************************************/


#ifdef PERSISTENCE_ENABLED
#define RB_NEW_ARRAY(type)                                             \
static									\
inline									\
type*									\
RB_NEW_ARRAY_##type( size_t num_##type ) {				\
  type* object = new type[num_##type];					\
  if( object == NULL ) {						\
    cerr << "new operator returned NULL during call to "		\
         << "RB_NEW_ARRAY_" << #type << endl;				\
    abort();								\
  }									\
  RB_NEW_HELPER( object, delete_##type##_array );			\
  if (PersistenceEnabled){                                              \
    PO_ADD_##type(object, num_##type);                                  \
  }								        \
  return object;							\
}									\

#define RB_DELETE_ARRAY(type)                                           \
static									\
inline									\
void									\
RB_DELETE_ARRAY_##type( type* object ) {				\
  RB_DELETE_HELPER( object, delete_##type##_array );			\
  if (PersistenceEnabled){                                              \
    PO_REMOVE_##type(object);                                           \
  }								        \
}									\

#else
#define RB_NEW_ARRAY(type)                                             \
inline									\
type*									\
RB_NEW_ARRAY_##type( size_t num_##type ) {				\
  type* object = new type[num_##type];					\
  if( object == NULL ) {						\
    cerr << "new operator returned NULL during call to "		\
         << "RB_NEW_ARRAY_" << #type << endl;				\
    abort();								\
  }									\
  RB_NEW_HELPER( object, delete_##type##_array );			\
  return object;							\
}									\

#define RB_DELETE_ARRAY(type)                                           \
inline									\
void									\
RB_DELETE_ARRAY_##type( type* object ) {				\
  RB_DELETE_HELPER( object, delete_##type##_array );			\
}									\

#endif


/**********************************************************************
 * 
 **********************************************************************/
#define RB_NEW_ARRAY_MACROS( type )					\
									\
inline									\
void									\
delete_##type##_array( void* array ) {					\
  delete [] ( ( type* ) array );					\
}									\
									\
inline									\
void									\
delete_##type##_array_PTR( void* array ) {				\
  delete [] ( ( type** ) array );					\
}									\
									\
RB_NEW_ARRAY(type)                                                      \
									\
inline									\
type**									\
RB_NEW_ARRAY_##type##_PTR( size_t num_##type ) {			\
  type** object = new type*[num_##type];				\
  if( object == NULL ) {						\
    cerr << "new operator returned NULL during call to "		\
         << "RB_NEW_ARRAY_" << #type << "_PTR" << endl;			\
    abort();								\
  }									\
  RB_NEW_HELPER( object, delete_##type##_array_PTR );			\
  return object;							\
}									\
									\
RB_DELETE_ARRAY(type)                                                   \
									\
inline									\
void									\
RB_DELETE_ARRAY_##type##_PTR( type** object ) {				\
  RB_DELETE_HELPER( object, delete_##type##_array_PTR );		\
}									\
									\
inline									\
void									\
RB_NEW_ARRAY_MACROS_warning_removal_function_##type() {			\
  RB_NEW_ARRAY_##type( 0 );						\
  RB_NEW_ARRAY_##type##_PTR( 0 );					\
  type* t = NULL;							\
  RB_DELETE_ARRAY_##type( t );						\
  type** td = NULL;							\
  RB_DELETE_ARRAY_##type##_PTR( td );					\
  void* v = RB_NEW_ARRAY_MACROS_warning_removal_function_##type;	\
									\
  /* This "v = v" line gets rid of the warning: */			\
  /* "variable "v"  was set but never used" */				\
  v = v;								\
}									\
/*********************************************************************/


#define RB_REGISTER_CLASS(ClassName) \
static void RB_REGISTER_CLASS_##ClassName () _M_C_ \
{ _M_C_ \
  PO_REGISTER_CLASS_##ClassName(); _M_C_ \
} _M_C_ \


/**********************************************************************
 * 
 **********************************************************************/
#define RB_DEFINE_CLASS_PRIMATIVE( type )				\
									\
PO_DEFINE_CLASS_NO_ARROW_OPERATORS( type )                              \
RB_NEW_SINGLE_MACROS( type )						\
RB_NEW_SINGLE_MACRO_WITH_NO_ARGUMENT_CONSTRUCTOR( type )		\
RB_NEW_ARRAY_MACROS( type )						\
RB_REGISTER_CLASS( type )                                               \
/*********************************************************************/

/**********************************************************************
 * 
 **********************************************************************/
#define RB_DEFINE_CLASS( type )						\
									\
PO_DEFINE_CLASS( type )                                                 \
RB_NEW_SINGLE_MACROS( type )						\
RB_NEW_SINGLE_MACRO_WITH_NO_ARGUMENT_CONSTRUCTOR( type )		\
RB_NEW_ARRAY_MACROS( type )						\
RB_REGISTER_CLASS( type )                                               \
/*********************************************************************/

/**********************************************************************
 * 
 **********************************************************************/
#define RB_DEFINE_CLASS_WITH_NO_DEFAULT_CONSTRUCTOR( type )		\
									\
RB_NEW_SINGLE_MACROS( type )						\
/*********************************************************************/



/**********************************************************************
 * Inline: T* RB_NEW_T()
 * Summary: Rollbackably creates a new object of type T
 * Description: T may be char or int or float or double, or anything
 *              else passed to macro RB_DEFINE_CLASS
 * Example: During the processing of an event, the line "C_SQ_ITEM
 *          *cancel_item = RB_NEW_C_SQ_ITEM();" assigns to cancel_item
 *          a new object of type C_SQ_ITEM.  It is similar to
 *          "C_SQ_ITEM *cancel_item = new C_SQ_ITEM;", except that if
 *          the event is rolled back, the memory will be deleted
 *          automatically by SPEEDES.
 * Return: a new object of type T
 **********************************************************************/

/**********************************************************************
 * Inline: T* RB_NEW_T(T *t)
 * Summary: Records that t should be deleted if the event is rolled back.
 * Description: T may be char or int or float or double, or anything
 *              else passed to macro RB_DEFINE_CLASS
 * Example: During the processing of an event, the line
 *          "ResourceWaitingItem *rwItem =
 *          RB_NEW_ResourceWaitingItem(new ResourceWaitingItem (obj,
 *          eventType, reentryLA, LVptr, requestAmount));" tells
 *          SPEEDES that if the event gets rolled back, then SPEEDES
 *          should automatically delete rwItem.  It is similar to
 *          "ResourceWaitingItem *rwItem = new ResourceWaitingItem
 *          (obj, eventType, reentryLA, LVptr, requestAmount);", except
 *          that if the event is rolled back, rwItem will be deleted
 *          automatically by SPEEDES.
 * Return: t
 **********************************************************************/

/**********************************************************************
 * Inline: T* RB_NEW_ARRAY_T(size_t n)
 * Summary: Rollbackably creates a new array of n objects of type T
 * Description: T may be char or int or float or double, or anything
 *              else passed to macro RB_DEFINE_CLASS
 * Example: During the processing of an event, the line "C_SQ_ITEM
 *          *link1 = RB_NEW_ARRAY_C_SQ_ITEM(3);" assigns to link1 an
 *          array of 3 C_SQ_ITEM objects.  It is similar to "C_SQ_ITEM
 *          *link1 = new C_SQ_ITEM[3];", except that if the event is
 *          rolled back, the memory will be deleted automatically by
 *          SPEEDES.
 * Return: a new array of Ts
 **********************************************************************/

/**********************************************************************
 * Inline: void RB_DELETE_T(T *t)
 * Summary: Rollbackably deletes an object of type T
 * Description: T may be char or int or float or double, or anything
 *              else passed to macro RB_DEFINE_CLASS
 * Example: During the processing of an event, the code "C_SQ_ITEM
 *          *cancel_item = new C_SQ_ITEM; ...;
 *          RB_DELETE_C_SQ_ITEM(cancel_item);" tells SPEEDES that
 *          cancel_item should be deleted, but SPEEDES will not delete
 *          it immediately.  Rather, SPEEDES will wait until the event
 *          is committed to actually do the delete.  This way if the
 *          event gets rolled back, then SPEEDES will decide NOT to
 *          delete the memory.  That code is similar to "delete
 *          cancel_item;", except that if the event is rolled back, the
 *          memory will NOT have actually been deleted.  Do NOT pass
 *          this function an array (example of arrays are something
 *          returned by RB_NEW_ARRAY_T or new T[]).
 * Return: void
 **********************************************************************/

/**********************************************************************
 * Inline: void RB_DELETE_ARRAY_T(T *t)
 * Summary: Rollbackably deletes an array of objects of type T
 * Description: T may be char or int or float or double, or anything
 *              else passed to macro RB_DEFINE_CLASS
 * Example: During the processing of an event, the code "C_SQ_ITEM
 *          *cancel_item = new C_SQ_ITEM[3]; ...;
 *          RB_DELETE_ARRAY_C_SQ_ITEM(cancel_item);" tells SPEEDES that
 *          cancel_item should be deleted with brackets, but SPEEDES
 *          will not delete it immediately.  Rather, SPEEDES will wait
 *          until the event is committed to actually do the delete.
 *          This way if the event gets rolled back, then SPEEDES will
 *          decide NOT to delete the memory.  That code is similar to
 *          "delete [] cancel_item;", except that if the event is
 *          rolled back, the memory will NOT have actually been
 *          deleted.  Do NOT pass this function a single non-array
 *          object (an example of a single non-array object is something
 *          returned by RB_NEW_T to by new T(...));
 * Return: void
 **********************************************************************/






#endif

