// convert.C method file

#include "SpMath.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"

#include "SpProxyDefines.H"
#include "SpConvert.H"

double gsam = GSAM;

void set_gsam(double g) {gsam = g;}
double get_gsam() {return gsam;}

double Squared(double X) {
  return (X*X);
}

void Sum(double X1[3], double X2[3], double X3[3]) {
  X3[0] = X1[0] + X2[0];
  X3[1] = X1[1] + X2[1];
  X3[2] = X1[2] + X2[2];
}

void Difference(double X1[3], double X2[3], double X3[3]) {
  X3[0] = X1[0] - X2[0];
  X3[1] = X1[1] - X2[1];
  X3[2] = X1[2] - X2[2];
}

double Magnitude(double X[3]) {
  return sqrt(X[0]*X[0] + X[1]*X[1] + X[2]*X[2]);
}

#if HETEROGENEOUS_DATA_TYPE
double Magnitude(NET_DOUBLE X[3]) {
  return sqrt(X[0]*X[0] + X[1]*X[1] + X[2]*X[2]);
}
#endif

double DotProduct(double X1[3], double X2[3]) {
  return X1[0]*X2[0] + X1[1]*X2[1] + X1[2]*X2[2];
}

#if HETEROGENEOUS_DATA_TYPE
double DotProduct(NET_DOUBLE X1[3], double X2[3]) {
  return X1[0]*X2[0] + X1[1]*X2[1] + X1[2]*X2[2];
}

double DotProduct(double X1[3], NET_DOUBLE X2[3]) {
  return X1[0]*X2[0] + X1[1]*X2[1] + X1[2]*X2[2];
}

double DotProduct(NET_DOUBLE X1[3], NET_DOUBLE X2[3]) {
  return X1[0]*X2[0] + X1[1]*X2[1] + X1[2]*X2[2];
}
#endif

void CrossProduct(double X1[3], double X2[3], double X3[3]) {
  X3[0] = X1[1]*X2[2] - X1[2]*X2[1];
  X3[1] = X1[2]*X2[0] - X1[0]*X2[2];
  X3[2] = X1[0]*X2[1] - X1[1]*X2[0];
}

#if HETEROGENEOUS_DATA_TYPE
void CrossProduct(NET_DOUBLE X1[3], NET_DOUBLE X2[3], NET_DOUBLE X3[3]) {
  X3[0] = X1[1]*X2[2] - X1[2]*X2[1];
  X3[1] = X1[2]*X2[0] - X1[0]*X2[2];
  X3[2] = X1[0]*X2[1] - X1[1]*X2[0];
}

void CrossProduct(double X1[3], NET_DOUBLE X2[3], NET_DOUBLE X3[3]) {
  X3[0] = X1[1]*X2[2] - X1[2]*X2[1];
  X3[1] = X1[2]*X2[0] - X1[0]*X2[2];
  X3[2] = X1[0]*X2[1] - X1[1]*X2[0];
}

void CrossProduct(NET_DOUBLE X1[3], NET_DOUBLE X2[3], double X3[3]) {
  X3[0] = X1[1]*X2[2] - X1[2]*X2[1];
  X3[1] = X1[2]*X2[0] - X1[0]*X2[2];
  X3[2] = X1[0]*X2[1] - X1[1]*X2[0];
}

void CrossProduct(NET_DOUBLE X1[3], double X2[3], double X3[3]) {
  X3[0] = X1[1]*X2[2] - X1[2]*X2[1];
  X3[1] = X1[2]*X2[0] - X1[0]*X2[2];
  X3[2] = X1[0]*X2[1] - X1[1]*X2[0];
}

void CrossProduct(NET_DOUBLE X1[3], double X2[3], NET_DOUBLE X3[3]) {
  X3[0] = X1[1]*X2[2] - X1[2]*X2[1];
  X3[1] = X1[2]*X2[0] - X1[0]*X2[2];
  X3[2] = X1[0]*X2[1] - X1[1]*X2[0];
}

void CrossProduct(double X1[3], NET_DOUBLE X2[3], double X3[3]) {
  X3[0] = X1[1]*X2[2] - X1[2]*X2[1];
  X3[1] = X1[2]*X2[0] - X1[0]*X2[2];
  X3[2] = X1[0]*X2[1] - X1[1]*X2[0];
}
#endif

void VectorCopy(double Xout[3], double Xin[3]) {
  Xout[0] = Xin[0];
  Xout[1] = Xin[1];
  Xout[2] = Xin[2];
}

#if HETEROGENEOUS_DATA_TYPE
void VectorCopy(double Xout[3], NET_DOUBLE Xin[3]) {
  Xout[0] = Xin[0];
  Xout[1] = Xin[1];
  Xout[2] = Xin[2];
}

void VectorCopy(NET_DOUBLE Xout[3], NET_DOUBLE Xin[3]) {
  Xout[0] = Xin[0];
  Xout[1] = Xin[1];
  Xout[2] = Xin[2];
}


void VectorCopy(NET_DOUBLE Xout[3], double Xin[3]) {
  Xout[0] = Xin[0];
  Xout[1] = Xin[1];
  Xout[2] = Xin[2];
}
#endif




void Normalize(double X[3]) {
  double Mag = Magnitude(X);
  X[0] /= Mag;
  X[1] /= Mag;
  X[2] /= Mag;
}

#if HETEROGENEOUS_DATA_TYPE
void Normalize(NET_DOUBLE X[3]) {
  double Mag = Magnitude(X);
  X[0] /= Mag;
  X[1] /= Mag;
  X[2] /= Mag;
}
#endif

void ScalarMultiply(double X[3], double Scaler) {
  X[0] *= Scaler;
  X[1] *= Scaler;
  X[2] *= Scaler;
}


#if HETEROGENEOUS_DATA_TYPE
void EARTH_to_ECR(NET_DOUBLE Earth[3], NET_DOUBLE RetEcr[3])
{
  double E[3];
  double R[3];
  int i;
  for (i=0; i < 3; ++i){
    E[i] = Earth[i];
    R[i] = RetEcr[i];
  }
  EARTH_to_ECR(E, R);
  for (i=0; i < 3; ++i){
    Earth[i ] = E[i];
    RetEcr[i] = R[i];
  }

}

void EARTH_to_ECR(double Earth[3], NET_DOUBLE RetEcr[3])
{
  double R[3];
  int i;
  for (i=0; i < 3; ++i){
    R[i] = RetEcr[i];
  }
  EARTH_to_ECR(Earth, R);
  for (i=0; i < 3; ++i){
    RetEcr[i] = R[i];
  }

}

void EARTH_to_ECR(NET_DOUBLE Earth[3], double RetEcr[3])
{
  double E[3];
  int i;
  for (i=0; i < 3; ++i){
    E[i] = Earth[i];
  }
  EARTH_to_ECR(E, RetEcr);
  for (i=0; i < 3; ++i){
    Earth[i] = E[i];
  }

}
#endif

/************************************************************************
* EARTH_to_ECR - transforms EARTH coordinates to ECR			*
************************************************************************/
void EARTH_to_ECR(double Earth[3], double RetEcr[3]) {
  double cos_lat;
  double sidereal_angle;
  double Ecr[3];

  cos_lat = cos(Earth[0]);
  sidereal_angle = Earth[1];

  Ecr[0] = cos_lat * cos(sidereal_angle) * (RE + Earth[2]);
  Ecr[1] = cos_lat * sin(sidereal_angle) * (RE + Earth[2]);
  Ecr[2] = sin(Earth[0]) * (RE + Earth[2]);

  RetEcr[0] = Ecr[0];
  RetEcr[1] = Ecr[1];
  RetEcr[2] = Ecr[2];

}

#if HETEROGENEOUS_DATA_TYPE
void EARTH_to_ECI(double t, NET_DOUBLE Earth[3], NET_DOUBLE RetEcr[3])
{
  double E[3];
  double R[3];
  int i;
  for (i=0; i < 3; ++i){
    E[i] = Earth[i];
    R[i] = RetEcr[i];
  }
  EARTH_to_ECI(t, E, R);
  for (i=0; i < 3; ++i){
    Earth[i ] = E[i];
    RetEcr[i] = R[i];
  }
}

void EARTH_to_ECI(double t, double Earth[3], NET_DOUBLE RetEcr[3])
{
  double R[3];
  int i;
  for (i=0; i < 3; ++i){
    R[i] = RetEcr[i];
  }
  EARTH_to_ECI(t, Earth, R);
  for (i=0; i < 3; ++i){
    RetEcr[i] = R[i];
  }
}

void EARTH_to_ECI(double t, NET_DOUBLE Earth[3], double RetEcr[3])
{
  double E[3];
  int i;
  for (i=0; i < 3; ++i){
    E[i] = Earth[i];
  }
  EARTH_to_ECI(t, E, RetEcr);
  for (i=0; i < 3; ++i){
    Earth[i] = E[i];
  }
}
#endif

/************************************************************************
* EARTH_to_ECI - transforms EARTH coordinates to ECI			*
************************************************************************/
void EARTH_to_ECI(double t, double Earth[3], double RetEcr[3]) {
  double cos_lat;
  double sidereal_angle;
  double Ecr[3];

  cos_lat = cos(Earth[0]);
  sidereal_angle = Earth[1] + gsam + t*OMEGAE;

  Ecr[0] = (cos_lat * cos(sidereal_angle)) * (RE + Earth[2]);
  Ecr[1] = (cos_lat * sin(sidereal_angle)) * (RE + Earth[2]);
  Ecr[2] = sin(Earth[0]) * (RE + Earth[2]);

  RetEcr[0] = Ecr[0];
  RetEcr[1] = Ecr[1];
  RetEcr[2] = Ecr[2];

}


#if HETEROGENEOUS_DATA_TYPE
void ECR_to_EARTH(NET_DOUBLE Earth[3], NET_DOUBLE RetEcr[3])
{
  double E[3];
  double R[3];
  int i;
  for (i=0; i < 3; ++i){
    E[i] = Earth[i];
    R[i] = RetEcr[i];
  }
  ECR_to_EARTH(E, R);
  for (i=0; i < 3; ++i){
    Earth[i ] = E[i];
    RetEcr[i] = R[i];
  }

}

void ECR_to_EARTH(double Earth[3], NET_DOUBLE RetEcr[3])
{
  double R[3];
  int i;
  for (i=0; i < 3; ++i){
    R[i] = RetEcr[i];
  }
  ECR_to_EARTH(Earth, R);
  for (i=0; i < 3; ++i){
    RetEcr[i] = R[i];
  }

}

void ECR_to_EARTH(NET_DOUBLE Earth[3], double RetEcr[3])
{
  double E[3];
  int i;
  for (i=0; i < 3; ++i){
    E[i] = Earth[i];
  }
  ECR_to_EARTH(E, RetEcr);
  for (i=0; i < 3; ++i){
    Earth[i] = E[i];
  }

}
#endif


/************************************************************************
* ECR_to_EARTH - transforms ECR coordinates to EARTH			*
************************************************************************/
void ECR_to_EARTH(double Ecr[3], double RetEarth[3]) {
  double rp;
  double Earth[3];

  rp = hypot(Ecr[0], Ecr[1]);

  if (fabs(rp) < 0.001) {
    cout << "Warning, hypot not working" << endl;
    rp = sqrt(Ecr[0]*Ecr[0] + Ecr[1]*Ecr[1]);
  }

  if (fabs(rp) < 0.001) {
    cout << "Warning, input for ECI to EARTH is bad" << endl;
    Earth[0] = PI/2;
  }else{
    Earth[0] = atan(Ecr[2]/rp);
  }

  Earth[1] = atan2(Ecr[1],Ecr[0]);
  Earth[1] = fmod(Earth[1],TWOPI);
  if (Earth[1] < -M_PI) Earth[1] += TWOPI;
  if (Earth[1] >  M_PI) Earth[1] -= TWOPI;

  Earth[2] = Magnitude(Ecr) - RE;

  if (Earth[2] > 50000.0) {
    cout << "Warning, Altitude too big" << endl;
  }

  if (Earth[2] < -100.0) {
    cout << "Warning, Altitude too small" << endl;
  }

  RetEarth[0] = Earth[0];
  RetEarth[1] = Earth[1];
  RetEarth[2] = Earth[2];

}

/************************************************************************
* ECI_to_EARTH : transforms ECI coordinates to EARTH			*
************************************************************************/
void ECI_to_EARTH(double t, double Eci[3], double RetEarth[3]) {
  double rp;
  double Earth[3];

  rp = hypot(Eci[0], Eci[1]);

  if (fabs(rp) < 0.001) {
    cout << "Warning, hypot not working" << endl;
    rp = sqrt(Eci[0]*Eci[0] + Eci[1]*Eci[1]);
  }

  if (fabs(rp) < 0.001) {
    cout << "Warning, input for ECI to EARTH is bad" << endl;
    Earth[0] = PI/2;
  }else{
    Earth[0] = atan(Eci[2]/rp);
  }

  Earth[1] = atan2(Eci[1],Eci[0]) - gsam - t*OMEGAE;
  Earth[1] = fmod(Earth[1],TWOPI);
  if (Earth[1] < -M_PI) Earth[1] += TWOPI;
  if (Earth[1] >  M_PI) Earth[1] -= TWOPI;

  Earth[2] = Magnitude(Eci) - RE;

  if (Earth[2] > 50000.0) {
    cout << "Warning, Altitude too big" << endl;
  }

  if (Earth[2] < -100.0) {
    cout << "Warning, Altitude too small" << endl;
  }

  RetEarth[0] = Earth[0];
  RetEarth[1] = Earth[1];
  RetEarth[2] = Earth[2];

}

#if HETEROGENEOUS_DATA_TYPE
void ECI_to_ECR(double t, NET_DOUBLE X[3])
{
  double x[3];
  int i;
  for (i=0; i < 3; ++i){
    x[i] = X[i];
  }
  ECI_to_ECR(t, x);
  for (i=0; i < 3; ++i){
    X[i] = x[i];
  }
} 

void ECI_to_ECR(double t, NET_DOUBLE X[3], NET_DOUBLE V[3])
{
  double x[3];
  double v[3];
  int i;
  for (i=0; i < 3; ++i){
    v[i] = V[i];
    x[i] = X[i];
  }
  ECI_to_ECR(t, x, v);
  for (i=0; i < 3; ++i){
    V[i] = v[i];
    X[i] = x[i];
  }
} 


void ECI_to_ECR(double t, NET_DOUBLE X[3], NET_DOUBLE V[3], NET_DOUBLE A[3])
{
  double x[3];
  double v[3];
  double a[3];
  int i;
  for (i=0; i < 3; ++i){
    v[i] = V[i];
    a[i] = A[i];
    x[i] = X[i];
  }
  ECI_to_ECR(t, x, v, a);
  for (i=0; i < 3; ++i){
    A[i] = a[i];
    V[i] = v[i];
    X[i] = x[i];
  }
} 
#endif


/************************************************************************
* ECI_to_ECR : convert from eci to ecr					*
************************************************************************/
void ECI_to_ECR(double t, double x[3]) {
  double co,so;
  double x0, x1;

//...... compute rotation cosine and sine

  co = cos(t*OMEGAE);
  so = sin(t*OMEGAE);

//...... rotate x and y from the earth

  x0 = co*x[0] + so*x[1];
  x1 = - so*x[0] + co*x[1];

//...... store the values

  x[0] = x0;
  x[1] = x1;

}

/************************************************************************
* ECI_to_ECR : convert from eci to ecr					*
************************************************************************/
void ECI_to_ECR(double t, double x[3], double v[3]) {
  double co,so;
  double x0, x1;
  double v0, v1;

//...... compute rotation cosine and sine

  co = cos(t*OMEGAE);
  so = sin(t*OMEGAE);

//...... rotate x and y from the earth

  x0 = co*x[0] + so*x[1];
  x1 = - so*x[0] + co*x[1];

//...... rotate vx and vy from the earth

  v0 = co*v[0] + so*v[1]   - OMEGAE*(so*x[0] - co*x[1]);
  v1 = - so*v[0] + co*v[1] - OMEGAE*(co*x[0] + so*x[1]);

//...... store the values

  x[0] = x0;
  x[1] = x1;

  v[0] = v0;
  v[1] = v1;

}

/************************************************************************
* ECI_to_ECR : convert from eci to ecr					*
************************************************************************/
void ECI_to_ECR(double t, double x[3], double v[3], double a[3]) {
  double co,so;
  double x0, x1;
  double v0, v1;
  double a0, a1;

//...... compute rotation cosine and sine

  co = cos(t*OMEGAE);
  so = sin(t*OMEGAE);

//...... rotate x and y from the earth

  x0 = co*x[0] + so*x[1];
  x1 = - so*x[0] + co*x[1];

//...... rotate vx and vy from the earth

  v0 = co*v[0] + so*v[1]   - OMEGAE*(so*x[0] - co*x[1]);
  v1 = - so*v[0] + co*v[1] - OMEGAE*(co*x[0] + so*x[1]);

//...... rotate ax and ay from the earth

  a0 = co*a[0] + so*a[1] - 2*OMEGAE*(so*v[0] - co*v[1])
    - (OMEGAE*OMEGAE)*(co*x[0] + so*x[1]);

  a1 = -so*a[0] + co*a[1] - 2*OMEGAE*(co*v[0] + so*v[1])
    + (OMEGAE*OMEGAE)*(so*x[0] - co*x[1]);

//...... store the values

  x[0] = x0;
  x[1] = x1;

  v[0] = v0;
  v[1] = v1;

  a[0] = a0;
  a[1] = a1;

}

/************************************************************************
* ECR_to_ECI : convert from ecr to eci					*
************************************************************************/
void ECR_to_ECI(double t, double x[3]) {
  double co,so;
  double x0, x1;

//...... compute rotation cosine and sine

  co = cos(t*OMEGAE);
  so = sin(t*OMEGAE);

//...... rotate x and y from the earth

  x0 = co*x[0] - so*x[1];
  x1 = so*x[0] + co*x[1];

//...... store the values

  x[0] = x0;
  x[1] = x1;

}

/************************************************************************
* ECR_to_ECI : convert from ecr to eci					*
************************************************************************/
void ECR_to_ECI(double t, double x[3], double v[3]) {
  double co,so;
  double x0, x1;
  double v0, v1;

//...... compute rotation cosine and sine

  co = cos(t*OMEGAE);
  so = sin(t*OMEGAE);

//...... rotate x and y from the earth

  x0 = co*x[0] - so*x[1];
  x1 = so*x[0] + co*x[1];

//...... rotate vx and vy from the earth

  v0 = co*v[0] - so*v[1] - OMEGAE*(so*x[0] + co*x[1]);
  v1 = so*v[0] + co*v[1] - OMEGAE*(-co*x[0] + so*x[1]);

//...... store the values

  x[0] = x0;
  x[1] = x1;

  v[0] = v0;
  v[1] = v1;

}

/************************************************************************
* ECR_to_ECI : convert from ecr to eci					*
************************************************************************/
void ECR_to_ECI(double t, double x[3], double v[3], double a[3]) {
  double co,so;
  double x0, x1;
  double v0, v1;
  double a0, a1;

//...... compute rotation cosine and sine

  co = cos(t*OMEGAE);
  so = sin(t*OMEGAE);

//...... rotate x and y from the earth

  x0 = co*x[0] - so*x[1];
  x1 = so*x[0] + co*x[1];

//...... rotate vx and vy from the earth

  v0 = co*v[0] - so*v[1] - OMEGAE*(so*x[0] + co*x[1]);
  v1 = so*v[0] + co*v[1] - OMEGAE*(-co*x[0] + so*x[1]);

//...... rotate vx and vy from the earth

  a0 = co*a[0] - so*a[1] - 2*OMEGAE*(so*v[0] + co*v[1])
    - (OMEGAE*OMEGAE)*(co*x[0] - so*x[1]);

  a1 = so*a[0] + co*a[1] - 2*OMEGAE*(-co*v[0] + so*v[1])
    - (OMEGAE*OMEGAE)*(so*x[0]+co*x[1]);

//...... store the values

  x[0] = x0;
  x[1] = x1;

  v[0] = v0;
  v[1] = v1;

  a[0] = a0;
  a[1] = a1;

}

/************************************************************************
* Convert : convert from system to another     				*
************************************************************************/
void Convert(double X1[3], int PT1,
	     double X2[3], int PT2,
	     double t) {

  switch(PT1) {

    case EARTH:
      switch (PT2) {

	case EARTH:
	  VectorCopy(X2, X1);
	  break;

	case ECR:
	  EARTH_to_ECR(X1, X2);
	  break;

	case ECI:
	  EARTH_to_ECI(t, X1, X2);
	  break;

      }
      break;

    case ECR:
      switch (PT2) {

	case EARTH:
	  ECR_to_EARTH(X1, X2);
	  break;

	case ECR:
	  VectorCopy(X2, X1);
	  break;

	case ECI:
	  VectorCopy(X2, X1);
	  ECR_to_ECI(t, X2);
	  break;

      }
      break;

    case ECI:
      switch (PT2) {

	case EARTH:
	  ECI_to_EARTH(t, X1, X2);
	  break;

	case ECR:
	  VectorCopy(X2, X1);
	  ECI_to_ECR(t, X2);
	  break;

	case ECI:
	  VectorCopy(X2, X1);
	  break;

      }
      break;

  }

}

/************************************************************************
* Convert : convert from system to another     				*
************************************************************************/
void Convert(double X1[3], double V1[3], int PT1,
	     double X2[3], double V2[3], int PT2,
	     double t) {

  switch(PT1) {

    case EARTH:
      switch (PT2) {

	case EARTH:
	  VectorCopy(X2, X1);
	  break;

	case ECR:
	  EARTH_to_ECR(X1, X2);
	  break;

	case ECI:
	  EARTH_to_ECI(t, X1, X2);
	  break;

      }
      break;

    case ECR:
      switch (PT2) {

	case EARTH:
	  ECR_to_EARTH(X1, X2);
	  break;

	case ECR:
	  VectorCopy(X2, X1);
	  VectorCopy(V2, V1);
	  break;

	case ECI:
	  VectorCopy(X2, X1);
	  VectorCopy(V2, V1);
	  ECR_to_ECI(t, X2, V2);
	  break;

      }
      break;

    case ECI:
      switch (PT2) {

	case EARTH:
	  ECI_to_EARTH(t, X1, X2);
	  break;

	case ECR:
	  VectorCopy(X2, X1);
	  VectorCopy(V2, V1);
	  ECI_to_ECR(t, X2, V2);
	  break;

	case ECI:
	  VectorCopy(X2, X1);
	  VectorCopy(V2, V1);
	  break;

      }
      break;

  }

}

/************************************************************************
* Convert : convert from system to another     				*
************************************************************************/
void Convert(double X1[3], double V1[3], double A1[3], int PT1,
	     double X2[3], double V2[3], double A2[3], int PT2,
	     double t) {

  switch(PT1) {

    case EARTH:
      switch (PT2) {

	case EARTH:
	  VectorCopy(X2, X1);
	  break;

	case ECR:
	  EARTH_to_ECR(X1, X2);
	  break;

	case ECI:
	  EARTH_to_ECI(t, X1, X2);
	  break;

      }
      break;

    case ECR:
      switch (PT2) {

	case EARTH:
	  ECR_to_EARTH(X1, X2);
	  break;

	case ECR:
	  VectorCopy(X2, X1);
	  VectorCopy(V2, V1);
	  VectorCopy(A2, A1);
	  break;

	case ECI:
	  VectorCopy(X2, X1);
	  VectorCopy(V2, V1);
	  VectorCopy(A2, A1);
	  ECR_to_ECI(t, X2, V2, A2);
	  break;

      }
      break;

    case ECI:
      switch (PT2) {

	case EARTH:
	  ECI_to_EARTH(t, X1, X2);
	  break;

	case ECR:
	  VectorCopy(X2, X1);
	  VectorCopy(V2, V1);
	  VectorCopy(A2, A1);
	  ECI_to_ECR(t, X2, V2, A2);
	  break;

	case ECI:
	  VectorCopy(X2, X1);
	  VectorCopy(V2, V1);
	  VectorCopy(A2, A1);
	  break;

      }
      break;

  }

}

#if HETEROGENEOUS_DATA_TYPE
void Convert(NET_DOUBLE X1[3], int PT1,
	     NET_DOUBLE X2[3], int PT2,
	     double t){
  double x[3];
  x[0] = X1[0];
  x[1] = X1[1];
  x[2] = X1[2];
  double x2[3];
  x2[0] = X2[0];
  x2[1] = X2[1];
  x2[2] = X2[2];
  Convert(x, PT1, x2, PT2, t);
  X2[0] = x2[0];
  X2[1] = x2[1];
  X2[2] = x2[2];
}

void Convert(NET_DOUBLE X1[3], double V1[3], int PT1,
	     double X2[3], double V2[3], int PT2,
	     double t){
  double x[3];
  x[0] = X1[0];
  x[1] = X1[1];
  x[2] = X1[2];
  double v[3];
  v[0] = V1[0];
  v[1] = V1[1];
  v[2] = V1[2];
  Convert(x, v, PT1, X2, V2, PT2, t);

}

void Convert(NET_DOUBLE X1[3], double V1[3], double A1[3], int PT1,
	     double X2[3], double V2[3], double A2[3], int PT2,
	     double t){
  double x[3];
  x[0] = X1[0];
  x[1] = X1[1];
  x[2] = X1[2];
  double v[3];
  v[0] = V1[0];
  v[1] = V1[1];
  v[2] = V1[2];
  double a[3];
  a[0] = A1[0];
  a[1] = A1[1];
  a[2] = A1[2];
  Convert(x, v, a, PT1, X2, V2, A2, PT2, t);

}
#endif
