// convert.H header file

#ifndef convert_functions
#define convert_functions

#include "SpMath.H"
#include "SpString.H"
#include "SpProxyDefines.H"
#include "SpNetTypes.H"

// EARTH: Latitude, Longitude, Altitude
// ECR: X, Y, Z in a rotating earth coordinate system
// ECI: X, Y, Z in an inertial coordinate system

void set_gsam(double g);
double get_gsam();

void Sum(double X1[3], double X2[3], double X3[3]);
void Difference(double X1[3], double X2[3], double X3[3]);
double Magnitude(double X[3]);
#if HETEROGENEOUS_DATA_TYPE
double Magnitude(NET_DOUBLE X[3]);
#endif
double Squared(double X);

double DotProduct(double X1[3], double X2[3]);
#if HETEROGENEOUS_DATA_TYPE
double DotProduct(NET_DOUBLE X1[3], double X2[3]);
double DotProduct(double X1[3], NET_DOUBLE X2[3]);
double DotProduct(NET_DOUBLE X1[3], NET_DOUBLE X2[3]);
#endif

void CrossProduct(double X1[3], double X2[3], double X3[3]);
#if HETEROGENEOUS_DATA_TYPE
void CrossProduct(NET_DOUBLE X1[3], NET_DOUBLE X2[3], NET_DOUBLE X3[3]);
void CrossProduct(double X1[3], NET_DOUBLE X2[3], NET_DOUBLE X3[3]);
void CrossProduct(NET_DOUBLE X1[3], double X2[3], NET_DOUBLE X3[3]);
void CrossProduct(double X1[3], NET_DOUBLE X2[3], double X3[3]);
void CrossProduct(NET_DOUBLE X1[3], NET_DOUBLE X2[3], double X3[3]);
void CrossProduct(NET_DOUBLE X1[3], double X2[3], double X3[3]);
#endif

void VectorCopy(double Xout[3], double Xin[3]);
#if HETEROGENEOUS_DATA_TYPE
void VectorCopy(double Xout[3], NET_DOUBLE Xin[3]);
void VectorCopy(NET_DOUBLE Xout[3], NET_DOUBLE Xin[3]);
void VectorCopy(NET_DOUBLE Xout[3], double Xin[3]);
#endif

void Normalize(double X[3]);
#if HETEROGENEOUS_DATA_TYPE
void Normalize(NET_DOUBLE X[3]);
#endif
void ScalarMultiply(double X[3], double Scaler);

void EARTH_to_ECR(double Earth[3], double Ecr[3]);
#if HETEROGENEOUS_DATA_TYPE
void EARTH_to_ECR(NET_DOUBLE Earth[3], double Ecr[3]);
void EARTH_to_ECR(double Earth[3], NET_DOUBLE Ecr[3]);
void EARTH_to_ECR(NET_DOUBLE Earth[3], NET_DOUBLE Ecr[3]);
#endif

void EARTH_to_ECI(double t, double Earth[3], double Eci[3]);
#if HETEROGENEOUS_DATA_TYPE
void EARTH_to_ECI(double t, NET_DOUBLE Earth[3], double Eci[3]);
void EARTH_to_ECI(double t, double Earth[3], NET_DOUBLE Eci[3]);
void EARTH_to_ECI(double t, NET_DOUBLE Earth[3], NET_DOUBLE Eci[3]);
#endif

void ECR_to_EARTH(double Ecr[3], double Earth[3]);
#if HETEROGENEOUS_DATA_TYPE
void ECR_to_EARTH(NET_DOUBLE Ecr[3], double Earth[3]);
void ECR_to_EARTH(double Ecr[3], NET_DOUBLE Earth[3]);
void ECR_to_EARTH(NET_DOUBLE Ecr[3], NET_DOUBLE Earth[3]);
#endif

void ECI_to_EARTH(double t, double Eci[3], double Earth[3]);
#if HETEROGENEOUS_DATA_TYPE
void ECI_to_EARTH(double t, NET_DOUBLE Eci[3], double Earth[3]);
void ECI_to_EARTH(double t, double Eci[3], NET_DOUBLE Earth[3]);
void ECI_to_EARTH(double t, NET_DOUBLE Eci[3], NET_DOUBLE Earth[3]);
#endif

void ECI_to_ECR(double t, double x[3]);
void ECI_to_ECR(double t, double x[3], double v[3]);
void ECI_to_ECR(double t, double x[3], double v[3], double a[3]);
#if HETEROGENEOUS_DATA_TYPE
void ECI_to_ECR(double t, NET_DOUBLE x[3]);
void ECI_to_ECR(double t, NET_DOUBLE x[3], NET_DOUBLE v[3]);
void ECI_to_ECR(double t, NET_DOUBLE x[3], NET_DOUBLE v[3], NET_DOUBLE a[3]);
#endif

void ECR_to_ECI(double t, double x[3]);
#if HETEROGENEOUS_DATA_TYPE
void ECR_to_ECI(double t, NET_DOUBLE x[3]);
#endif
void ECR_to_ECI(double t, double x[3], double v[3]);

#if HETEROGENEOUS_DATA_TYPE
void ECR_to_ECI(double t, NET_DOUBLE x[3], NET_DOUBLE v[3]);
void ECR_to_ECI(double t, NET_DOUBLE x[3], NET_DOUBLE v[3], NET_DOUBLE a[3]);
#endif
void ECR_to_ECI(double t, double x[3], double v[3], double a[3]);

void Convert(double X1[3], int PT1,
	     double X2[3], int PT2,
	     double t=0);

void Convert(double X1[3], double V1[3], int PT1,
	     double X2[3], double V2[3], int PT2,
	     double t=0);

void Convert(double X1[3], double V1[3], double A1[3], int PT1,
	     double X2[3], double V2[3], double A2[3], int PT2,
	     double t=0);

#if HETEROGENEOUS_DATA_TYPE
void Convert(NET_DOUBLE X1[3], int PT1,
	     NET_DOUBLE X2[3], int PT2,
	     double t=0);

void Convert(NET_DOUBLE X1[3], double V1[3], int PT1,
	     double X2[3], double V2[3], int PT2,
	     double t=0);

void Convert(NET_DOUBLE X1[3], double V1[3], double A1[3], int PT1,
	     double X2[3], double V2[3], double A2[3], int PT2,
	     double t=0);
#endif

inline double NauticalMiles_to_Kilometers(double nauticalMiles) {
  return nauticalMiles * Kilometers_per_NauticalMile;
}

inline double Kilometers_to_NauticalMiles(double kilometers) {
  return kilometers * NauticalMiles_per_Kilometer;
}

inline double Miles_to_Kilometers(double miles) {
  return miles * Kilometers_per_Mile;
}

inline double Kilometers_to_Miles(double kilometers) {
  return kilometers * Miles_per_Kilometer;
}

inline int GetUnitDistance(char *unitDistance) {
  int UnitDistance = -1;

  if (!strcmp("NauticalMiles", unitDistance)) {
    UnitDistance = 0;
  }

  else if (!strcmp("Miles", unitDistance)) {
    UnitDistance = 1;
  }

  else if (!strcmp("Kilometers", unitDistance)) {
    UnitDistance = 2;
  }

  return UnitDistance;
}

inline int GetUnitTime(char *unitTime) {
  int UnitTime = -1;

  if (!strcmp("Days", unitTime)) {
    UnitTime = 0;
  }

  else if (!strcmp("Hours", unitTime)) {
    UnitTime = 1;
  }

  else if (!strcmp("Minutes", unitTime)) {
    UnitTime = 2;
  }

  else if (!strcmp("Seconds", unitTime)) {
    UnitTime = 3;
  }

  return UnitTime;

}

inline double UnitDistance_to_Kilometers(double distance, int unitDistance) {
  double Kilometers = 0.0;

  if (unitDistance == 0) {
    Kilometers = NauticalMiles_to_Kilometers(distance);
  }

  else if (unitDistance == 1) {
    Kilometers = Miles_to_Kilometers(distance);
  }

  else if (unitDistance == 2) {
    Kilometers = distance;
  }

  return Kilometers;
}

inline double Kilometers_to_UnitDistance(double kilometers, int unitDistance) {
  double distance = 0.0;

  if (unitDistance == 0) {
    distance = Kilometers_to_NauticalMiles(kilometers);
  }

  else if (unitDistance == 1) {
    distance = Kilometers_to_Miles(kilometers);
  }

  else if (unitDistance == 2) {
    distance = kilometers;
  }

  return distance;
}

inline double KilometersPerSecond(double speed, int unitDistance,
				  int unitTime) {
  double TimeConvert = INFINITY;
  if (unitTime == 0) TimeConvert = Days_per_Second;
  else if (unitTime == 1) TimeConvert = Hours_per_Second;
  else if (unitTime == 2) TimeConvert = Minutes_per_Second;
  else if (unitTime == 3) TimeConvert = 1;

  double DistanceConvert = INFINITY;
  if (unitDistance == 0) DistanceConvert = Kilometers_per_NauticalMile;
  else if (unitDistance == 1) DistanceConvert = Kilometers_per_Mile;
  else if (unitDistance == 2) DistanceConvert = 1;

  return (speed * DistanceConvert * TimeConvert);

}

#endif
