// SpDataParser.C file


#include "SpDataParser.H"
#include "SpDataParserTypes.H"
#include "SpStdlib.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpAssert.H"

struct MARK: public C_ITEM
{
   void *val;
};

// #define DEBUG_SpDataParser

SpDataParser::~SpDataParser()
{
  Qtypes.clear();
  free(FileName);
}

/************************************************************************
* Constructor for the SpDataParser object				*
************************************************************************/
SpDataParser::SpDataParser(const char *F) {
  char Symbol[256];
  char ch;
  SpATOM *atom;
  BASE_DATA_TYPE *BaseDataType;
  REFERENCE_DATA_TYPE *ReferenceDataType;
  REF_ITEM *RefItem;
  int i;
  C_QUEUE Qref;

/*
** Initialize the GoTo BASE DATA TYPE (BDT) pointer
*/

  GoToBDT = NULL;

/*
** First open the file and make sure that it is there
*/

  FileName = strdup(F);
  ifstream InputStream(FileName);

  if (!InputStream.good()) {
    cerr << "Error DataParser: Cannot open " << FileName << endl;
    abort();
  }

/*
** Now, read the file and store the symbols in a queue. The comments
** are stripped away here.
*/

  while (!InputStream.eof()) {

    Symbol[0] = 0;
    InputStream >> Symbol;
    if (Symbol[0] == 0) break;

    if (Symbol[0] == '/' && Symbol[1] == '/') {
      while (!InputStream.eof()) {
	InputStream.get(ch);
	if (ch == '\n') break;
      }
    }else{
#ifdef DEBUG_SpDataParser
      cout << Symbol << endl;
#endif

// Hunt for embedded "{" or "}" symbols

      if (strlen(Symbol) > 1) {

	int j = 0;
	int len = strlen(Symbol);
	int found;
	int k;

	for (i=0; i<len; i++) {

	  found = 0;

	  if (Symbol[j] == '{') {
	    Symbol[j] = 0;
	    if (j) {
	      atom = new SpATOM(Symbol);
	      Qatoms.push_bot(atom);
	    }
	    atom = new SpATOM("{");
	    Qatoms.push_bot(atom);
	    if (i == len-1) {
	      Symbol[0] = 0;
	    }else{
	      for (k=0; k<len-i-1; k++) Symbol[k] = Symbol[j+k+1];
	      Symbol[len-i-1] = 0;
	    }
	    found = 1;
	    j = 0;
	    len--;
	  }

	  if (Symbol[j] == '}') {
	    Symbol[j] = 0;
	    if (j) {
	      atom = new SpATOM(Symbol);
	      Qatoms.push_bot(atom);
	    }
	    atom = new SpATOM("}");
	    Qatoms.push_bot(atom);
	    if (i == len-1) {
	      Symbol[0] = 0;
	    }else{
	      for (k=0; k<len-i-1; k++) Symbol[k] = Symbol[j+k+1];
	      Symbol[len-i-1] = 0;
	    }
	    found = 1;
	    j = 0;
	    len--;
	  }

	  if (!found) j++;

	}

      }

      if (strlen(Symbol)) {
	atom = new SpATOM(Symbol);
	Qatoms.push_bot(atom);
      }

    }

  }

#ifdef DEBUG_SpDataParser
  cout << Qatoms.get_length() << " atoms in Qatoms" << endl;
#endif

/*
** Go through the symbols stored in the queue and construct the
** types. First, we set the Qref in the ReferenceDataType object. Note
** that the ReferenceDataType object is not every instantiated even though
** we are calling its method. That is ok, since the method only sets a
** static member.
*/

  ReferenceDataType = NULL;
  ReferenceDataType->SetQref(&Qref);

  while (Qatoms.get_length()) {
    BaseDataType = new BASE_DATA_TYPE(&Qatoms);
    Qtypes.push_bot(BaseDataType);
  }

/*
** Set up the references and make sure that they are correct
*/

  while (Qref.get_length()) {

    RefItem = (REF_ITEM *)Qref.pop_top();
    ReferenceDataType = RefItem->RDT;

#ifdef DEBUG_SpDataParser
    cout << "Setting up Reference for " << ReferenceDataType->GetName()
	 << endl;
#endif

    delete RefItem;

    int found = 0;
    BaseDataType = (BASE_DATA_TYPE *)Qtypes.get_top();
    for (i=0; i<Qtypes.get_length(); i++) {
      if (!strcmp(BaseDataType->GetName(),
		  ReferenceDataType->GetName())) {
	ReferenceDataType->SetValue(BaseDataType);
	found = 1;
	break;
      }
      BaseDataType = (BASE_DATA_TYPE *)BaseDataType->get_link();
    }

    if (!found) {
      cout << "Error (DataParser): Could not find reference "
	   << ReferenceDataType->GetName() << endl;
      exit(0);
    }

  }

}


/************************************************************************
* Get an integer value from the parser					*
************************************************************************/
int SpDataParser::GetInt(char *n, int Index) {
  INT_DATA_TYPE *IntDataType;
  BASE_DATA_TYPE *BDT;
  char s[120];

  BDT = GoToBDT;

  if (Index == -1) {
    IntDataType = (INT_DATA_TYPE *)BDT->Find(n);
  }else{
    sprintf(s,"%s[%d]", n, Index);
    IntDataType = (INT_DATA_TYPE *)BDT->Find(s);
  }

  if (IntDataType != NULL){
    return IntDataType->GetValue();
  }
  else{
    return -1;
  }

}


/************************************************************************
* Get the name of the current reference in the parser			*
************************************************************************/
char *SpDataParser::GetReferenceName() {
  BASE_DATA_TYPE *BDT;

  BDT = GoToBDT;
  if (BDT == NULL) return NULL;
  return BDT->GetName();

}


/************************************************************************ 
* Get the type of the current reference in the parser			*
************************************************************************/
char *SpDataParser::GetType() {
  BASE_DATA_TYPE *BDT;
 
  BDT = GoToBDT;
  if (BDT == NULL) return NULL;
  return BDT->GetType();
 
}


/************************************************************************
* Get an enum value from the parser					*
************************************************************************/
int SpDataParser::GetEnum(char *n, int Index) {
  ENUM_DATA_TYPE *EnumDataType;
  BASE_DATA_TYPE *BDT;
  char s[120];

  BDT = GoToBDT;

  if (Index == -1) {
    EnumDataType = (ENUM_DATA_TYPE *)BDT->Find(n);
  }else{
    sprintf(s,"%s[%d]", n, Index);
    EnumDataType = (ENUM_DATA_TYPE *)BDT->Find(s);
  }
  if (EnumDataType){
    return EnumDataType->GetValue();
  }
  else{
    return -1;
  }

}


/************************************************************************
* Get a logical value from the parser					*
************************************************************************/
int SpDataParser::GetLogical(char *n, int Index) {
  LOGICAL_DATA_TYPE *LogicalDataType;
  BASE_DATA_TYPE *BDT;
  char s[120];

  BDT = GoToBDT;

  if (Index == -1) {
    LogicalDataType = (LOGICAL_DATA_TYPE *)BDT->Find(n);
  }else{
    sprintf(s,"%s[%d]", n, Index);
    LogicalDataType = (LOGICAL_DATA_TYPE *)BDT->Find(s);
  }

  if (LogicalDataType){
    return LogicalDataType->GetValue();
  }
  else{
    return -1;
  }

}


/************************************************************************
* Get a float value from the parser					*
************************************************************************/
double SpDataParser::GetFloat(char *n, int Index) {
  FLOAT_DATA_TYPE *FloatDataType;
  BASE_DATA_TYPE *BDT;
  char s[120];

  BDT = (BASE_DATA_TYPE *)GoToBDT;

  if (Index == -1) {
    FloatDataType = (FLOAT_DATA_TYPE *)BDT->Find(n);
  }else{
    sprintf(s,"%s[%d]", n, Index);
    FloatDataType = (FLOAT_DATA_TYPE *)BDT->Find(s);
  }

  if (FloatDataType){
    return FloatDataType->GetValue();
  }
  else{
    return -1.0;
  }

}


/************************************************************************
* Get a string value from the parser					*
************************************************************************/
char *SpDataParser::GetString(char *n, int Index) {
  STRING_DATA_TYPE *StringDataType;
  BASE_DATA_TYPE *BDT;
  char s[120];

  BDT = GoToBDT;

  if (Index == -1) {
    StringDataType = (STRING_DATA_TYPE *)BDT->Find(n);
  }else{
    sprintf(s,"%s[%d]", n, Index);
    StringDataType = (STRING_DATA_TYPE *)BDT->Find(s);
  }

  if (StringDataType){
    return StringDataType->GetValue();
  }
  else{
    return "You Did Not Give The Requested Attribute In The Par File";
  }

}


/************************************************************************
* Get a define value from the parser					*
************************************************************************/
char *SpDataParser::GetDefine(char *n, int Index) {
  DEFINE_DATA_TYPE *DefineDataType;
  BASE_DATA_TYPE *BDT;
  char s[120];

  BDT = GoToBDT;

  if (Index == -1) {
    DefineDataType = (DEFINE_DATA_TYPE *)BDT->Find(n);
  }else{
    sprintf(s,"%s[%d]", n, Index);
    DefineDataType = (DEFINE_DATA_TYPE *)BDT->Find(s);
  }

  if (DefineDataType){
    return DefineDataType->GetValue();
  }
  else{
    return NULL;
  }

}

/************************************************************************
* Get first define value from the parser				*
************************************************************************/
void SpDataParser::GetFirstDefine(char *&n, char *&v) {
  DEFINE_DATA_TYPE *DefineDataType;
  BASE_DATA_TYPE *BDT;

  if (GoToBDT == NULL) {
    cout << "Error (GetFirstDefine): Mark is NULL" << endl;
    return;
  }

  BDT = GoToBDT;
  DefineDataType = (DEFINE_DATA_TYPE *)BDT->FindFirstDefine();
  DefPtr = DefineDataType;

  if (DefPtr == NULL) {
    n = NULL;
    v = NULL;
  }else{
    n = DefineDataType->GetName();
    v = DefineDataType->GetValue();
  }

}


/************************************************************************
* Get next define value from the parser					*
************************************************************************/
void SpDataParser::GetNextDefine(char *&n, char *&v) {
  DEFINE_DATA_TYPE *DefineDataType;
  BASE_DATA_TYPE *BDT;

  BDT = GoToBDT;

  DefineDataType = (DEFINE_DATA_TYPE *)BDT->FindNextDefine(DefPtr);
  DefPtr = DefineDataType;

  if (DefPtr == NULL) {
    n = NULL;
    v = NULL;
  }else{
    n = DefineDataType->GetName();
    v = DefineDataType->GetValue();
  }

}


/************************************************************************
* GoTo an item in the parser						*
************************************************************************/
void *SpDataParser::GoTo(char *FirstName, ...) {
  BASE_DATA_TYPE *BDT;
  char *NextName;
  int i;
  va_list va;
  int found;

/*
** First, find the BASE_DATA_TYPE in the parser's Qtypes list
*/

  found = 0;
  BDT = (BASE_DATA_TYPE *)Qtypes.get_top();
  for (i=0; i<Qtypes.get_length(); i++) {
    if (!strcmp(BDT->GetName(), FirstName)) {
      found = 1;
      break;
    }
    BDT = (BASE_DATA_TYPE *)BDT->get_link();
  }

  if (!found) {
//    cout << "Error (SpDataParser,top of GoTo): Could not find " << FirstName
//	 << " in GoTo" << endl;
    GoToBDT = NULL;
    return GoToBDT;
  }

/*
** Now go through the variable argument list and find the rest
*/

  va_start(va, FirstName);
  while (1) {
    NextName = va_arg(va, char *);
    if (NextName == NULL) break;

#ifdef DEBUG_SpDataParser
    cout << "Trying to find " << NextName << " in " << BDT->GetName() << endl;
#endif

    BDT = (BASE_DATA_TYPE*) BDT->Find(NextName);

    if (BDT == NULL) {
      cout << "Error (SpDataParser,bottom of GoTo): Could not find " << NextName
	   << " in GoTo" << endl;
      GoToBDT = NULL;
      return GoToBDT;
    }

  }
  va_end(va);

  GoToBDT = BDT->GetBDT();
  return GoToBDT;

}


/************************************************************************ 
* GetNumElements							*
************************************************************************/
int SpDataParser::GetNumElements(void *Mark, char *ReferenceName) {
  FUNDAMENTAL_DATA_TYPE *FDT;
  BASE_DATA_TYPE *BDT;

  if (Mark == NULL) {
    cout << "Error (GetNumReferences): Mark is NULL" << endl;
    return 0;
  }

  FDT = (FUNDAMENTAL_DATA_TYPE *)Mark;
  BDT = FDT->GetBDT();

  return BDT->GetNumReferences(ReferenceName);
}


/************************************************************************ 
* Find          							*
************************************************************************/
void *SpDataParser::Find(void *Mark, char *ReferenceName) {
  FUNDAMENTAL_DATA_TYPE *FDT;
  BASE_DATA_TYPE *BDT;

  if (Mark == NULL) {
    cout << "Error (Find): Mark is NULL" << endl;
    return NULL;
  }

  FDT = (FUNDAMENTAL_DATA_TYPE *)Mark;
  BDT = FDT->GetBDT();

  FDT = BDT->Find(ReferenceName);
  if (FDT) GoToBDT = (BASE_DATA_TYPE*) FDT;

  return (void *)FDT;

}

/************************************************************************ 
* GoToFirstElement							*
************************************************************************/
void *SpDataParser::GoToFirstElement(void *Mark, char *ReferenceName) {
  FUNDAMENTAL_DATA_TYPE *FDT;
  BASE_DATA_TYPE *BDT;

  if (Mark == NULL) {
    cout << "Error (GoToNextReference): Mark is NULL" << endl;
    return NULL;
  }

  FDT = (FUNDAMENTAL_DATA_TYPE *)Mark;
  BDT = FDT->GetBDT();

  FDT = BDT->GoToFirstReference(ReferenceName);
  if (FDT) GoToBDT = (BASE_DATA_TYPE*) FDT;

  return (void *)FDT;

}

/************************************************************************ 
* GoToNextElement							*
************************************************************************/
void *SpDataParser::GoToNextElement(void *Mark, char *ReferenceName) {
  FUNDAMENTAL_DATA_TYPE *FDT;
  BASE_DATA_TYPE *BDT;

  if (Mark == NULL) {
    cout << "Error (GoToNextReference): Mark is NULL" << endl;
    return NULL;
  }

  FDT = (FUNDAMENTAL_DATA_TYPE *)Mark;
  BDT = FDT->GetBDT();

  FDT = BDT->GoToNextReference(ReferenceName);
  if (FDT) GoToBDT = (BASE_DATA_TYPE*) FDT;

  return (void *)FDT;

}

/************************************************************************
* GoToFirstReference							*
************************************************************************/
void *SpDataParser::GoToFirstReference(void *Mark, char *ReferenceName) {
  FUNDAMENTAL_DATA_TYPE *FDT;
  BASE_DATA_TYPE *BDT;

  if (Mark == NULL) {
    cout << "Error (GoToNextReference): Mark is NULL" << endl;
    return NULL;
  }

  FDT = (FUNDAMENTAL_DATA_TYPE *)Mark;
  BDT = FDT->GetBDT();

  FDT = BDT->GoToFirstReference(ReferenceName);
  if (FDT) GoToBDT = FDT->GetBDT();

  return (void *)FDT;

}


/************************************************************************
* GoToNextReference							*
************************************************************************/
void *SpDataParser::GoToNextReference(void *Mark, char *ReferenceName) {
  FUNDAMENTAL_DATA_TYPE *FDT;
  BASE_DATA_TYPE *BDT;

  if (Mark == NULL) {
    cout << "Error (GoToNextReference): Mark is NULL" << endl;
    return NULL;
  }

  FDT = (FUNDAMENTAL_DATA_TYPE *)Mark;
  BDT = FDT->GetBDT();

  FDT = BDT->GoToNextReference(ReferenceName);
  if (FDT) GoToBDT = FDT->GetBDT();
  return (void *)FDT;

}

/************************************************************************
* GetArraySize								*
************************************************************************/
int SpDataParser::GetArraySize(char *Name) {
  BASE_DATA_TYPE *BDT;

  BDT = GoToBDT->GetBDT();

  if (GoToBDT == NULL) {
    cout << "Error (GoToNextReference): Mark is NULL" << endl;
    return 0;
  }

  return BDT->GetArraySize(Name);

}


/************************************************************************
* GetNtypes								*
************************************************************************/
int SpDataParser::GetNtypes(char *Type) {
  BASE_DATA_TYPE *BDT;

  BDT = GoToBDT;
  if (BDT) {
    return BDT->GetNtypes(Type);
  }else{
    cout << "Error (GetNtypes): Mark is NULL" << endl;
    return 0;
  }

}

int SpDataParser::CheckStack() {
  MARK * m = (MARK *)markStack.get_top();
  return (GoToBDT == m->val);
}

char *SpDataParser::GetSetName() {
  MARK * m = (MARK *)markStack.get_top();
  if (!CheckStack()) cout << "DATA PARSER ERROR: GetSetName()" << endl;
  return ((BASE_DATA_TYPE *)(m->val))->GetName();
//  return GetReferenceName ();
}


/************************************************************************
* Print the SpDataParser object						*
************************************************************************/
void SpDataParser::print(ostream &out) {

  out << "Data Parser: "
      << FileName << endl << endl;

  BASE_DATA_TYPE *BDT = (BASE_DATA_TYPE *)Qtypes.get_top();
  for (int i=0; i<Qtypes.get_length(); i++) {
    BDT->printlevel(out, 0);
    BDT = (BASE_DATA_TYPE *)BDT->get_link();
    if (i < Qtypes.get_length()-1) out << endl;
  }
}

#define CURRENT_STACK_MARK ((MARK *) markStack.get_top ())->val
#define CASTED_CURRENT_STACK_MARK (BASE_DATA_TYPE *) CURRENT_STACK_MARK

int SpDataParser::MoveIn (char *setName)
{
   void *newMarkVal;
   
   if (markStack.get_length () == 0) newMarkVal = GoTo (setName, NULL);
   else newMarkVal = Find(CURRENT_STACK_MARK, setName);
   if (newMarkVal)
   {
      MARK *newMark = new MARK;
      newMark->val = newMarkVal;
      markStack.push_top (newMark);
      GoToBDT = CASTED_CURRENT_STACK_MARK;
      if (!CheckStack()) cout << "DATA PARSER ERROR: MoveIn()" << endl;
      return 1;
   }
   return 0;
}

void SpDataParser::MoveOut (int numLevels)
{
   for (int i=0; i<numLevels; i++)
   {
      if (markStack.get_length() > 1) delete markStack.pop_top ();
      else
      {
	 markStack.clear ();
	 return;
      }
   }
   GoToBDT = CASTED_CURRENT_STACK_MARK;
   if (!CheckStack()) cout << "DATA PARSER ERROR: MoveOut()" << endl;
}

void SpDataParser::MoveOutCompletely ()
{
   markStack.clear ();
   GoToBDT = NULL;
}

int SpDataParser::GetNumSetsInGroup(char *ReferenceName)
{
   return GetNumElements(CURRENT_STACK_MARK, ReferenceName);
}

int SpDataParser::MoveToFirstSetInGroup(char *ReferenceName)
{
   strcpy (setMarkRefName, ReferenceName);
   void *newMarkVal = GoToFirstReference(CURRENT_STACK_MARK, setMarkRefName);
   if (newMarkVal)
   {
      MARK *newMark = new MARK;
      newMark->val = newMarkVal;
      markStack.push_top (newMark);
      GoToBDT = CASTED_CURRENT_STACK_MARK;
      if (!CheckStack()) cout << "DATA PARSER ERROR: MoveToFirstSetInGroup()"
			      << endl;
      return 1;
   }
   return 0;
}

int SpDataParser::MoveToNextSetInGroup()
{
   
   MARK *mark = (MARK *) markStack.pop_top ();
   mark->val = GoToNextReference(CURRENT_STACK_MARK, setMarkRefName);
   markStack.push_top ((C_ITEM *) mark);
   GoToBDT = CASTED_CURRENT_STACK_MARK;
   if (!CheckStack()) cout << "DATA PARSER ERROR: MoveToNextSetInGroup()"
			   << endl;
   if (mark->val) return 1;
   return 0;
}

#ifdef SpDataParser_TESTING

#include "SpCrashOnDivZero.H"

/************************************************************************
* Test program for standalone testing					*
************************************************************************/
int main() {
  crash_on_divide_by_zero();

  SpDataParser *Parser;
  char F[80];
  void *Mark;

  cout << "Enter name of input file: ";
  cin >> F;

  Parser = new SpDataParser(F);
  Parser->print(cout);

  if (!strcmp(F, "t.par")) {
// Test String
    Parser->GoTo("InterestSpaces",
	       NULL);

    cout << "GetString " << Parser->GetString("SpaceSet") << endl;

// Test Int
    Parser->GoTo("InterestSpaces",
	       "WorldWarIISpace",
	       "WWIICategories",
	       NULL);

    cout << "GetInt " << Parser->GetInt("Ncategories") << endl;

// Test Logical
    Parser->GoTo("InterestSpaces",
	       "WorldWarIISpace",
	       "Europe",
	       "Altitude",
	       NULL);

    cout << "GetLogical " << Parser->GetLogical("Distribute") << endl;

// Test Enum
    Parser->GoTo("InterestSpaces",
	       "WorldWarIISpace",
	       "ShipTypes",
	       NULL);

    cout << "GetEnum " << Parser->GetEnum("AirCraftCarrier") << endl;

// Test Float
    Parser->GoTo("InterestSpaces",
	       "WorldWarIISpace",
	       "Europe",
	       "LatLng",
	       "Latitude",
	       NULL);

    cout << "GetFloat " << Parser->GetFloat("Lo") << endl;

// Test Array Size
    Parser->GoTo("InterestSpaces",
	       "WorldWarIISpace",
	       "Europe",
	       "LatLng",
	       NULL);

    cout << "GetArraySize " << Parser->GetArraySize("Resolution") << endl;
    cout << "GetFloat(Resolution, 3) " << Parser->GetFloat("Resolution", 3)
	 << endl;

    cout << "GetFloat(Resolution, 1) " << Parser->GetFloat("Resolution", 1)
	 << endl;

// Test N types
    Parser->GoTo("InterestSpaces",
	       "WorldWarIISpace",
	       NULL);

    cout << "GetInt(test) " << Parser->GetInt("test") << endl;
    cout << "GetNtypes(ENUMTYPE) " << Parser->GetNtypes("ENUMTYPE") << endl;

// Test References
    Mark = Parser->GoTo("InterestSpaces",
	       "WorldWarIISpace",
	       NULL);

    Parser->GoToFirstReference(Mark, "ENUMTYPE");

    cout << "GetEnum(PtBoat) " << Parser->GetEnum("PtBoat") << endl;
    cout << "GetEnum(BattleShip) " << Parser->GetEnum("BattleShip") << endl;

    Parser->GoToNextReference(Mark, "ENUMTYPE");
    cout << "GetEnum(Allies) " << Parser->GetEnum("Allies") << endl;
    cout << "GetEnum(Axis) " << Parser->GetEnum("Axis") << endl;

    if (Parser->GoToNextReference(Mark, "ENUMTYPE")) {
      cout << "We have a problem with GoToNextReference" << endl;
    }

  }

}
#endif
