// free_list.C method file

#include "SpStdio.H"
#include "SpStdlib.H"
#include "SpAssert.H"

#include "SpMemory.H"

#include "free_list.H"
#include "qitem.H"

/************************************************************************
* FREE_LIST : FREE_LIST constructor					*
************************************************************************/
C_FREE_LIST::C_FREE_LIST() {
  ntypes = 0;
}

C_FREE_LIST::~C_FREE_LIST() {
  ToBeCleared.clear();
  delete [] size;
  delete [] neach;
  delete [] ncreated;
  for (int i=0; i<ntypes; i++)
    free(name[i]);
  delete [] name;
  delete [] constructor;
  delete [] ToBeDeleteds;
  delete [] qfree;
}

/************************************************************************
* set_ntypes : set (and create) n types of freelists			*
************************************************************************/
void C_FREE_LIST::set_ntypes(int n) {
  int i;
  C_QUEUE *qfree1;
  int *size1;
  int *neach1;
  FREE_OBJECT_CONSTRUCTOR_FUNCTION *constructor1;
  TO_BE_DELETED_FUNCTION *ToBeDeleteds1;
  int *ncreated1;
  int ntypes1;
  char **name1;

  if (ntypes == 0) {

    ntypes = n;

    size = new int[ntypes];
    neach = new int[ntypes];
    ncreated = new int[ntypes];
    name = new char*[ntypes];
    constructor = new FREE_OBJECT_CONSTRUCTOR_FUNCTION[ntypes];
    ToBeDeleteds = new TO_BE_DELETED_FUNCTION[ntypes];
    qfree = new C_QUEUE[ntypes];

    for (i=0; i<ntypes; i++) {
      ncreated[i] = 0;
      name[i] = NULL;
      size[i] = 0;
      neach[i] = 0;
      constructor[i] = NULL;
      ToBeDeleteds[i] = NULL;
    }

  }else{

    ntypes1 = n;

    size1 = new int[ntypes1];
    neach1 = new int[ntypes1];
    ncreated1 = new int[ntypes1];
    name1 = new char*[ntypes1];
    constructor1 = new FREE_OBJECT_CONSTRUCTOR_FUNCTION[ntypes1];
    ToBeDeleteds1 = new TO_BE_DELETED_FUNCTION[ntypes1];
    qfree1 = new C_QUEUE[ntypes1];

    for (i=ntypes; i<ntypes1; i++) {
      ncreated1[i] = 0;
      name1[i] = NULL;
      size1[i] = 0;
      neach1[i] = 0;
      constructor1[i] = NULL;
      ToBeDeleteds1[i] = NULL;
    }

    memcpy((char *)qfree1, (char *)qfree, ntypes*sizeof(C_QUEUE));
    memcpy((char *)size1, (char *)size, ntypes*sizeof(int));
    memcpy((char *)neach1, (char *)neach, ntypes*sizeof(int));
    memcpy((char *)constructor1, (char *)constructor, ntypes*sizeof(FREE_OBJECT_CONSTRUCTOR_FUNCTION));
    memcpy((char *)ToBeDeleteds1, (char *)ToBeDeleteds, ntypes*sizeof(TO_BE_DELETED_FUNCTION));
    memcpy((char *)ncreated1, (char *)ncreated, ntypes*sizeof(int));
    memcpy((char *)name1, (char *)name, ntypes*sizeof(char *));

    delete [] qfree;
    delete [] size;
    delete [] neach;
    delete [] constructor;
    delete [] ToBeDeleteds;
    delete [] ncreated;
    delete [] name;

    ntypes = ntypes1;
    qfree = qfree1;
    size = size1;
    neach = neach1;
    constructor = constructor1;
    ToBeDeleteds = ToBeDeleteds1;
    ncreated = ncreated1;
    name = name1;

  }

}

/************************************************************************
* get_size : get the size of an object					*
************************************************************************/
int C_FREE_LIST::get_size(C_ITEM *freebase) {
  int type;

  // JBrutocao - need error checking if the object is not a C_ITEM
  type = freebase->get_freeid();
  // JBrutocao - error checking code (added freeid = -1 to qitem constructor as well)
  if (type == -1) return -1;
  return size[type];

}

void *C_FREE_LIST::f(int type, int n)
{
  if (constructor[type] != NULL)
    return constructor[type](n);
  else if (ToBeDeleteds[type] != NULL)
  {
    ARRAY_OF_CHILDREN *tbd = ToBeDeleteds[type](n);
    ToBeCleared.push_bot(tbd);
    return tbd->GetArray();
  }
  else
  {
    assert(0);
    return NULL;
  }
}

/************************************************************************
* new_object : get or construct a new object				*
************************************************************************/
void *C_FREE_LIST::new_object(int type) {
  C_ITEM *freebase;
  void *object;
  char *bff;
  int i;

  if (qfree[type].length() == 0) {

    if (neach[type] <= 0) {
      object = f(type, 1);
      if (object == NULL) {
        fprintf(stderr,"no more object memory (1)\n");
        abort();
      }
      ncreated[type]++;
      freebase = (C_ITEM *)object;
      freebase->set_freeid(type);
      return object;
    }else{
      object = f(type, neach[type]);
    }

    if (object == NULL) {
      fprintf(stderr,"no more object memory (2)\n");
      abort();
    }else{

      bff = (char *)object;
      for (i=1; i<neach[type]; i++) {
        bff += size[type];
        freebase = (C_ITEM *)bff;
        freebase->set_freeid(type);
        qfree[type].push_bot(freebase);
      }

      freebase = (C_ITEM *)object;
      freebase->set_freeid(type);
      ncreated[type] += neach[type];

    }

  }else{

    freebase = qfree[type].pop_top();
    freebase->set_freeid(type);
    object = (void *)freebase;

  }

  return object;

}

/************************************************************************
* delete_object : delete an object by putting it back into a free list	*
************************************************************************/
void C_FREE_LIST::delete_object(void *object) {
  C_ITEM *freebase;

  if (object == NULL) {
    fprintf(stderr,"Error, (free list) deleting NULL object\n");
    exit(1);
  }

  freebase = (C_ITEM *)object;

  if (neach[freebase->get_freeid()] > 0) {
    qfree[freebase->get_freeid()].push_bot(freebase);
  }else{
    ncreated[freebase->get_freeid()]--;
    assert(name[freebase->get_freeid()] != NULL);
    cerr << "\nIn name = " << name[freebase->get_freeid()]
	 << ", you cannot delete a type of void*" << endl;
    assert(0);
    //delete [] object;
  }

}

/************************************************************************
* delete_object : delete an object by putting it back into a free list	*
************************************************************************/
void C_FREE_LIST::delete_object(void *object, int type) {
  C_ITEM *freebase;

  if (object == NULL) {
    fprintf(stderr,"Error, (free list) deleting NULL object\n");
    exit(1);
  }

  freebase = (C_ITEM *)object;

  if (neach[type] > 0) {
    freebase->set_freeid(type);
    qfree[type].push_bot(freebase);
  }else{
    ncreated[freebase->get_freeid()]--;
    assert(0); // cannot delete void*: delete object;
  }

}

/************************************************************************
* delete_list : add a list of objects back into a free list		*
************************************************************************/
void C_FREE_LIST::delete_list(C_QUEUE *q, int type) {
  int i,len;
  C_ITEM *item;

  len = q->length();
  if (len != 0) {

    if ((q->ptop() == NULL) || (q->pbot() == NULL)) {

      fprintf(stderr,"Error, (free list) deleting bad list\n");
      exit(1);

    }else{

      if (type == -1) {

	for (i=0; i<len; i++) {
	  item = q->pop_top();

	  if (neach[item->get_freeid()] > 0) {
	    qfree[item->get_freeid()].push_bot(item);
	  }else{
	    ncreated[item->get_freeid()]--;
	    delete item;
	  }

	}

      }else{

	if (neach[type] > 0) {
          qfree[type].concat(q);
	}else{
	  for (i=0; i<len; i++) {
	    item = q->pop_top();
	    delete item;
	  }
	  ncreated[type]-=len;
	}

      }

    }

  }

}

/************************************************************************
* generate : generate an object from a buffer				*
************************************************************************/
void *C_FREE_LIST::generate(char *buff) {
  C_ITEM *fb;
  void *obj;
  char *ptr;
  int size_DV;

  fb = (C_ITEM *)buff;
  obj = new_object(fb->get_freeid());

  ptr = (char *)obj;
  size_DV = sizeof(C_DUMMY_V);

  memcpy(ptr+size_DV, buff+size_DV, size[fb->get_freeid()]-size_DV);

  return obj;

}

/************************************************************************
* print : print out status information for the free list		*
************************************************************************/
void C_FREE_LIST::print(ostream &out) {
  int i;
  int j;
  char s[120];

  out << endl;

  j = 0;
  for (i=0; i<ntypes; i++) {
    sprintf(s,
	    "%3d %-16s   size = %6d   created = %6d   free = %6d   bytes = %d",
	    i, name[i], size[i], ncreated[i], qfree[i].length(),
	    ncreated[i]*size[i]);
    out << s << endl;
    j += ncreated[i]*size[i];
  }

  out << endl << "    Total bytes = " << j << endl << endl;
}

  void C_FREE_LIST::set_type(char *nam,        //
		int type,         //
		TO_BE_DELETED_FUNCTION f, //
		int sz,           //
		int n             //
		)
  {
    assert(nam != NULL);
    constructor[type] = NULL;
    ToBeDeleteds[type] = f;
    size[type] = sz;
    name[type] = strdup(nam);
    neach[type] = n;
  }


  void C_FREE_LIST::set_type(char *nam,        //
		int type,         //
		FREE_OBJECT_CONSTRUCTOR_FUNCTION f, //
		int sz,           //
		int n             //
		)
  {
    assert(nam != NULL);
    constructor[type] = f;
    ToBeDeleteds[type] = NULL;
    size[type] = sz;
    name[type] = strdup(nam);
    neach[type] = n;
  }

void C_FREE_LIST::CopyType(char *NewName, // name of new freelist type
		int NewType,   // type ID of new type
		int OldType)  {
  assert(NewName != NULL);
    constructor[NewType] = constructor[OldType];
    ToBeDeleteds[NewType] = ToBeDeleteds[OldType];
    size[NewType] = size[OldType];
    name[NewType] = (char *)strdup(NewName);
    neach[NewType] = neach[OldType];
  }
