#ifndef __SpGeoFunctions__
#define __SpGeoFunctions__

#include "SpMath.H"

/* Changed so as not to cause conflicts with NSS */
//const double pi = 3.141592653589793;
//const double two_pi = 2 * pi;

#ifndef PI
#define PI 3.141592653589793
#endif

#ifndef TWO_PI
#define TWO_PI (2.0*PI)
#endif

static const double
  north_pole_rad_lat =  PI / 2,
  south_pole_rad_lat = -PI / 2;

static const double radius_of_earth = 6371; //In km; it's the mean radius of the earth

inline static double RAD_to_DEG(const double radians)
{ return (180*(radians)/PI); }

inline static double DEG_to_RAD(const double degrees)
{ return (PI*(degrees)/180); }

// Pass in the radius of a sphere and the distance from the center of
// this sphere to the center of the earth, and the return value is
// half of the number of latitude radians in which the sphere exists.
double get_delta_rad_lat_of_sphere
  (const double sphere_radius,
   const double distance_from_center_of_sphere_to_center_of_earth);

// Pass in the radius of a sphere and the distance from the center of
// this sphere to the axis of the earth, and the return value is
// half of the number of longitude radians in which the sphere exists.
double get_delta_rad_lng_of_sphere
  (const double sphere_radius,
   const double distance_from_center_of_sphere_to_axis_of_earth);

// Pass in the latitude in radians of a point, and pass in the
// distance from that point to the center of the earth, and the return
// value is the distance of this point from the axis of the earth.
double get_distance_from_axis
  (const double rad_lat,
   const double distance_from_input_point_to_center_of_earth);

// Returns by reference delta_rad_lat and delta_rad_lng which are
// obtained from the functions get_delta_rad_lat_of_sphere and
// get_delta_rad_lng_of_sphere if possible.  But, if the sphere
// contains the center of the earth then for any latitude and any
// longitude, there is a point within the sphere with that latitude
// and longitude, which means this function just returns radians
// representing half of a full circle, namely pi.  If the sphere
// contains the axis of the earth then for any longitude, there is a
// point within the sphere with that longitude, which means this
// function just returns in delta_rad_lng, radians representing half
// of a full circle, namely pi.
void get_delta_rad_angles_of_sphere
  (const double rad_lat,
   const double sphere_radius,
   const double distance_from_center_of_sphere_to_axis_of_earth,
   double &delta_rad_lat,
   double &delta_rad_lng);

#endif
