
#include "SpComm.H"
#include "SpDataParser.H"
#include "SpPoDataBase.H"
#include "PO.H"
#include "SpTimeMgtAlg.H"



int SpTimeMgtAlg_Restart;
char* SpTimeMgtAlg_RestartType = "";
double SpTimeMgtAlg_RestartTime;
extern int SpServerPortNum;
extern char* SpServerMachineName;
extern int SpGatherStatistics;

const char *CmdLineArgs[]={
  "nnodes",
  "lnodes",
  "group",
  "pipe",
  "help",
  "restart",
  "PrintDatabases",
  NULL
};

enum {
  NNODES,
  LNODES,
  GROUP,
  PIPE,
  HELP,
  RESTART,
  PRINT_DATA_BASES
};


void CommandLineHelp()
{
  cout<<"Command line arguments are:"<<endl
      <<endl
      <<" -nnodes <total number of nodes>"<<endl
      <<" -lnodes <total number of local nodes>"<<endl
      <<" -group <your group ID"<<endl
      <<" -restart <type> <time> where type is WALL or SIM"<<endl
      <<" -PrintDatabases <type> <time> where type is WALL or SIM"<<endl
      <<" -help  Displays this message"<<endl
      <<endl
      <<" If nnodes is not given and lnodes is given, that number is"<<endl
      <<" taken to be the total number of nodes "<<endl
      <<" If none of the arguments are given, the proper values are"<<endl
      <<" read from the speedes.par file"<<endl;
  exit(0);
}


int MyStrcmpOnLength(const char * str1, const char * str2, const int length){
  int rc = 0;
  for (int i=0; (i < length) && (!rc); ++i){
    if (*str1 != *str2){
      if       (*str1 < *str2) rc = -1;
      else if  (*str1 > *str2) rc = 1;
    }
    ++str1;
    ++str2;
  }
  return rc;
}






void BadArgument(const char * arg, const char * errorMessage = NULL)
{
  cout<<"Bad command line argument: "<<arg<<endl;
  if (errorMessage!= NULL){
    cout<<errorMessage<<endl;
  }
  CommandLineHelp();
  exit(-1);
}



/*
  findverb is the workhorse of the whole process, it returns two things:
  
  1) a pointer to the first non-whitespace character after a verb is
  found, or NULL if it can't parse a verb from the string.
  
  2) the number of the verb in the verbs[] table. this should be ignored
     if the pointer returned is NULL.
*/
int FindVerb(char *p)
{
  int i = 0;
  int verb = -1;

  if (*p != '-'){
    BadArgument(p);
  }
  else{
    ++p; //Skip over the '-'
    while (CmdLineArgs[i] != NULL){
      if (MyStrcmpOnLength(CmdLineArgs[i], p, strlen(p)) == 0){
	/* Found a match */
	if (verb == -1){
	  verb = i;
	}
	else{			/* Found a second match!! Bad!!! */
	  char matches[256];
	  sprintf(matches, "At least two matches!!!!: %s %s",
		  CmdLineArgs[i], CmdLineArgs[verb]);
	  BadArgument(p, matches);
	}
      }
      ++i;
    }
  }
  return verb;
}

int PrintDataBase(char* fName)
{
  FILE* inFile = fopen(fName, "rb");
  char tempBuffer[6*sizeof(int)];
  int bufferSize;
  char* persistenceBuffer;
  SpPoDataBase* printDataBase = new SpPoDataBase;
  int rc = 0;

  if (inFile == NULL){
    cout<<"Could not open "<<fName<<endl;
    rc = 1;
  }
  if (!rc){
    fread(tempBuffer, sizeof(char), 6*sizeof(int), inFile);
    bufferSize = PO_BUFFER_SIZE(tempBuffer);
    rewind(inFile);
    persistenceBuffer = new char[bufferSize];
    fread(persistenceBuffer, sizeof(char), bufferSize, inFile);
    PO_RESTORE(persistenceBuffer, printDataBase);
    cout<<"/////////////////////////////////////////////////////////////////////////////"<<endl;
    cout<<"//// Printing database for "<<fName<<"////"<<endl;
    cout<<"/////////////////////////////////////////////////////////////////////////////"<<endl;
    printDataBase->Print();
    fclose(inFile);
    delete [] persistenceBuffer;
  }
  delete printDataBase;
  return rc;
}



void PrintDataBases(char* restartType, char* time)
{
  char name[256];

  for (int i=0; i < 128; ++i){
    sprintf(name, "OBJECTS_CheckPoint_%s_TIME_%s_NODE_%d",
	    restartType, time, i);
    if (PrintDataBase(name)){
      break;
    }

    sprintf(name, "MESSAGES_CheckPoint_%s_TIME_%s_NODE_%d",
	    restartType, time, i);
    if (PrintDataBase(name)){
      break;
    }
  }
}


void HandleCommandLine(int argc,
		       char **argv,
		       int &nTot,
		       int &nLoc,
		       int &group,
		       char ** pInfo)
{
  int verb;
  char s[50];
  char * pipeInfo = NULL;
  nLoc = 0;
  nTot = 0;
  int i = 1;
  group = -1;

  if (argc > 1){


    //First handle old stype command line aruments
    if (('0' <= argv[1][0]) && ('9' >=  argv[1][0])){
      nTot = atoi(argv[1]);
      char junkString[256];
      sprintf(junkString, "%d",nTot);
      if (strcmp(junkString, argv[1])){
	cout<<"Bad command line "<<argv[1]<<endl;
	exit(17);
      }
      i = 2;
      if (argc > 2){
	if (('0' <= argv[2][0]) && ('9' >=  argv[2][0])){ 
	  nLoc = nTot;
	  nTot = atoi(argv[2]); 
	  char junkString[256]; 
	  sprintf(junkString, "%d",nTot); 
	  if (strcmp(junkString, argv[2])){ 
	    cout<<"Bad command line "<<argv[2]<<endl; 
	    exit(17);
	  }
	  i = 3;
	  if (argc > 3){
	    if (('0' <= argv[3][0]) && ('9' >=  argv[3][0])){
	      group = atoi(argv[2]);
	      char junkString[256];
	      sprintf(junkString, "%d",group);
	      if (strcmp(junkString, argv[3])){
		cout<<"Bad command line "<<argv[3]<<endl;
		exit(17);
	      }
	      i = 4;
	    }
	  }
	}
      }
    }


    for (; i < argc; ++i){
      char * restartType;
      int nTotGivenTwice = 0;
      int nLocGivenTwice=0;
      verb = FindVerb(argv[i]);
      switch(verb){
      case NNODES:

	if (nTot != 0){
	  cout<<"Cannot use both old and new command line arguments!!!"
	      <<endl;
	  exit(17);
	}
	if (nTotGivenTwice){
	  cout<<"Total number of nodes can only be given once"<<endl;
	  exit(17);
	}
	nTotGivenTwice=1;
	nTot = atoi(argv[++i]);
	sprintf(s, "%d",nTot);
	if (strcmp(s, argv[i])){
	  sprintf(s, "-nnodes %d",nTot);
	  BadArgument(s);
	}
	break;
      case LNODES:
        if (nLoc != 0){ 
          cout<<"Cannot use both old and new command line arguments!!!" 
              <<endl; 
	  exit(17);
        } 
        if (nLocGivenTwice){ 
          cout<<"Total number of nodes can only be given once"<<endl; 
          exit(17); 
        } 
        nLocGivenTwice=1; 
	nLoc = atoi(argv[++i]);
	sprintf(s, "%d", nLoc);
	if (strcmp(s, argv[i])){
	  sprintf(s, "-nnodes %d", nLoc);
	  BadArgument(s);
	}
	break;
      case GROUP:
	group = atoi(argv[++i]);
	sprintf(s, "%d", group);
	if (strcmp(s, argv[i])){
	  sprintf(s, "-nnodes %d", group);
	    BadArgument(s);
	}
	break;
      case PIPE:
	*pInfo = argv[++i];
	break;
      case RESTART:
	SpTimeMgtAlg_Restart = 1;
	SpTimeMgtAlg_RestartType = argv[++i];
	if (!strcmp(SpTimeMgtAlg_RestartType, "WALL")){
	  cout<<"Restarting from Wall time ";
	}
	else if (!strcmp(SpTimeMgtAlg_RestartType, "SIM")){
	  cout<<"Restarting from Sim time ";
	}
	else{
	  cout<<"Don't know how to restart with parameter "<<argv<<endl;
	  exit(-1);
	}
	SpTimeMgtAlg_RestartTime = atof(argv[++i]);
	cout<<SpTimeMgtAlg_RestartTime<<endl;
	break;
      case PRINT_DATA_BASES:
	restartType = argv[++i];
	PrintDataBases(restartType, argv[++i]);
	break;
      case HELP:
	CommandLineHelp();
	break;
      }
    }
    if (nTot == 0){
      nTot = nLoc;
    }
    else if (nLoc == 0){
      nLoc = nTot;
    }
    else if ((nTot == 0) && (nLoc == 0)){
      cout<<"Cannot run SPEEDES with no nodes!"<<endl;
      exit(-1);
    }
    if ((nLoc > nTot) && (nLoc > 0) && (nTot > 0)){
      cout<<"Cannot have more local nodes than global node"<<endl;
      exit(-1);
    }
    if ((pipeInfo == NULL) && (nTot > 1)){ /*Need to store argv[0] for spawn()*/
      *pInfo = argv[0];
    }
    if (group == -1){
      SpDataParser parser("speedes.par");
      parser.GoTo("SpeedesServer", NULL);
      group = parser.GetInt("Group");
    }
  }
  else{				/* No arguments given so go to the
				   speedes.par file for all the info */
    SpDataParser *parser = new SpDataParser("speedes.par");
    parser->GoTo("SpeedesServer", NULL);
    group = parser->GetInt("Group");
    parser->GoTo("parameters", NULL);
    nTot = parser->GetInt("n_nodes");
    nLoc = nTot;
    *pInfo = argv[0];
    delete parser;
  }
  cout.flush();

  if (nTot == 0){
    SpDataParser *parser = new SpDataParser("speedes.par");
    parser->GoTo("SpeedesServer", NULL);
    group = parser->GetInt("Group");
    parser->GoTo("parameters", NULL);
    nTot = parser->GetInt("n_nodes");
    nLoc = nTot;
    *pInfo = argv[0];
    delete parser;
  }

  if (nTot != nLoc){
    SpDataParser Parser("speedes.par");

    Parser.GoTo("SpeedesServer", "SpeedesComm", NULL);

    if (Parser.GetLogical("UseDefaults")) {
      Parser.GoTo("SpeedesServer", NULL);
      char *Ptr = Parser.GetString("DefaultMachineName");
      SpServerMachineName = new char [strlen(Ptr)+1];
      strcpy(SpServerMachineName, Ptr);
      SpGatherStatistics = Parser.GetLogical("DefaultStatistics");
      SpServerPortNum = Parser.GetInt("DefaultPort");
    }
    else{
      Parser.GoTo("SpeedesServer", NULL);
      char *Ptr = Parser.GetString("MachineName");
      SpServerMachineName = new char [strlen(Ptr)+1];
      strcpy(SpServerMachineName, Ptr);
      SpGatherStatistics = Parser.GetLogical("Statistics");
      SpServerPortNum = Parser.GetInt("Port");
    }
  }
}

