%{
//
// Rules for a reentrant scanner for the speedes par file format
//

#include <stdiostream.h>
//#include <cstring>
#include "SpSet.H"
#include "SpParser.tab.h"
static inline void *saferealloc(void *v, const size_t s)
{
  if (v)
    return realloc(v, s);
  else
    return malloc(s);
} 

#define __DONT_INCLUDE_FLEXLEXER__
#define __DONT_INCLUDE_y_tab_h__
#ifdef __GNUC__
#pragma implementation "SpParserFlexLexer.H"
#endif // #ifdef __GNUC__
#include "SpParserFlexLexer.H"
#include "SpParserCommon.H"
static int matches_builtin_keyword(const char *);

#ifdef YY_MAIN
YYSTYPE lval;
YYSTYPE *lvalp = &lval;
const char *alwaysnull = NULL;
#define INTRO alwaysnull
struct SpParserCommon commonstruct;
struct SpParserCommon *commondata = &commonstruct; 
extern "C" {
int yylex(void);
}
#else
#undef YY_DECL
#define YY_DECL int SpParserFlexLexer::yylex(YYSTYPE *lvalp, void *commondata)
#define INTRO ((struct SpParserCommon *) commondata)->intro
#define SCANNER ((struct SpParserCommon *) commondata)->scanner
#endif

%}

%option yyclass="SpParserFlexLexer"
%option noyywrap
%option c++

DQUOTE               \x22 


BSTRING              [[:alpha:]]+[[:alnum:]_]* 

GSTRING              [^[:space:]"]+

LOPENPROTO           \x28


LCLOSEPROTO          \x29


LOPENSET             \x7b


LCLOSESET            \x7d


LINT                 [[\-+]?[0-9]+


LBOOL                (F)|(false)|(true)|(T)


LFLOAT               ([\-+]?[0-9]*\.[0-9]*([Ee][+-]([0-9]{1,3}))?)|([[:space:]]*[-+]?[0-9]+[Ee][+-]?([0-9]{1,3}))


KINT                 int


KFLOAT               float


KSTRING              string


KBOOL                logical


KREF                 reference


KKWD                keyword


KINCLUDE       include


SEP                  [^_/0-9A-Za-z\{\}\"\)]+ 


SSEP                 [^_0-9A-Za-z\(\)\{\}\"]+ 


CCOMMENT             \/\/


%x LSTRING KWDPROTO TYPELIST LCOMMENT SFLOAT SSTRING SBOOL SREF SINT SSTRINGLNAME SFLOATLNAME SBOOLLNAME SINTLNAME SINCLUDE

%%

<LCOMMENT>\n %{
	BEGIN(INITIAL);
             %}

<INITIAL,KWDPROTO,TYPELIST,SFLOAT,SINCLUDE,SSTRING,SBOOL,SREF,SINT,SSTRINGLNAME,SFLOATLNAME,SBOOLLNAME,SINTLNAME>{SEP} %{
	// throwaway
      %}

{KKWD} %{
	BEGIN(KWDPROTO);
                 %}

<KWDPROTO>{BSTRING} %{
	lvalp->stringval = strdup(YYText());
	BEGIN(TYPELIST);

#ifndef YY_MAIN
	return UKWD;
#endif
                  %}

<TYPELIST>{SSEP} %{
	// throw away
                 %}

<TYPELIST>{KINT} %{
	lvalp->intval = INT;

#ifndef YY_MAIN
	return BUILTINTYPE;
#endif
                 %}

<TYPELIST>{KFLOAT} %{
	lvalp->intval = FLOAT;

#ifndef YY_MAIN
	return BUILTINTYPE;
#endif
%}

<TYPELIST>{KSTRING} %{
	lvalp->intval = STRING;

#ifndef YY_MAIN
	return BUILTINTYPE;
#endif
                    %}

<TYPELIST>{KBOOL} %{
	lvalp->intval = TOKBOOL;

#ifndef YY_MAIN
	return BUILTINTYPE;
#endif
                  %}

<TYPELIST>{KREF} %{
	lvalp->intval = REFNAME;

#ifndef YY_MAIN
	return BUILTINTYPE;
#endif
                 %}

<TYPELIST>{LCLOSEPROTO} %{
	BEGIN(INITIAL);
                        %}

<KWDPROTO,TYPELIST,INITIAL>{CCOMMENT} %{
	BEGIN(LCOMMENT);
                                              %}

<LCOMMENT>[^\n]* %{
	lvalp->stringval = strdup(YYText());

#ifndef YY_MAIN
	return COMMENT;
#endif
                 %}

{LOPENSET} %{

#ifndef YY_MAIN
	return OPENSET;
#endif
           %}

{LCLOSESET} %{

#ifndef YY_MAIN
	return CLOSESET;
#endif
            %}

<INITIAL,SSTRINGLNAME>{DQUOTE} %{
	BEGIN(LSTRING);
	lvalp->stringval = NULL;
         %}

<LSTRING>[^"]* %{
              {
	int initiallength = ((lvalp->stringval)?strlen(lvalp->stringval):0);
        lvalp->stringval = (char *) saferealloc(lvalp->stringval,
	                                  initiallength + strlen(YYText())+1);
        if (!initiallength) {
           lvalp->stringval[0] = '\0';
        }
	strcat(lvalp->stringval,YYText());
              }
               %}

<LSTRING>\" %{
        // If escaped, include in existing string
        {
	int initiallength = ((lvalp->stringval)?strlen(lvalp->stringval):0);
        if (initiallength && (*(lvalp->stringval+strlen(lvalp->stringval)-1) == '\\')) {
           lvalp->stringval = (char *) saferealloc((void *) lvalp->stringval,
                                               2+initiallength);
           (lvalp->stringval)[initiallength+1] = '\0';
           (lvalp->stringval)[initiallength] = '"';
    }
    else {
      BEGIN(INITIAL);
#ifdef YY_MAIN
      in_include = false;
#endif
#ifndef YY_MAIN
      if (in_include) {
          in_include = false;
          return INCLUDE;
      }
      else {
          return STRING;
      }
#endif
    }
    }
            %}


<LSTRING>{DQUOTE} %{
	BEGIN(INITIAL);
#ifdef YY_MAIN
              in_include = false;
#endif
#ifndef YY_MAIN
              if (in_include) {
                 in_include = false;
                 return INCLUDE;
              }
              else {
                  return STRING;
              }
#endif
                  %}

<INITIAL,SBOOLLNAME>{LBOOL}/{SEP} %{
	if (*YYText() == 'F' || *YYText() == 'f') {
	   lvalp->boolval = false;
	}
	else {
	   lvalp->boolval = true;
        }
        BEGIN(INITIAL);
#ifndef YY_MAIN
	return TOKBOOL;
#endif
              %}


{KINCLUDE} %{
          // Expect an included set or file
          BEGIN(SINCLUDE);
          %}

<SINCLUDE>{BSTRING} %{
                   lvalp->stringval = strdup(YYText());
                   BEGIN(INITIAL);
                   in_include = false;
#                  ifndef YY_MAIN                 
                     return INCLUDE;
#                  endif // #ifndef YY_MAIN
           %}

<SINCLUDE>{DQUOTE} %{
                 in_include = true;
                 lvalp->stringval = NULL;
                BEGIN(LSTRING);
          %}

{KSTRING} %{
          // Expect a string-valued variable
          BEGIN(SSTRING);
          %}

<SSTRING>{BSTRING} %{
                   lvalp->stringval = strdup(YYText());
                   BEGIN(SSTRINGLNAME);
#                  ifndef YY_MAIN                 
                     return LNAME;
#                  endif // #ifndef YY_MAIN
                   %}

<SSTRING,SSTRINGLNAME,SFLOAT,SFLOATLNAME,SINT,SINTLNAME,SBOOL,SBOOLLNAME,SREF>{CCOMMENT} %{
                                LexerError("SpParser: Inline comment not allowed in typed statement" );
                                                               %}
          

<SSTRINGLNAME>{GSTRING} %{
                                 lvalp->stringval = strdup(YYText());
                                 BEGIN(INITIAL);
#                  ifndef YY_MAIN
                     return STRING;
#                  endif // #ifndef YY_MAIN
                   %}

                       
                                      
{KFLOAT} %{
          // Expect a float-valued variable
          BEGIN(SFLOAT);
         %}

<SFLOAT>{BSTRING} %{
                   lvalp->stringval = strdup(YYText());
                   BEGIN(SFLOATLNAME);
#                  ifndef YY_MAIN                 
                     return LNAME;
#                  endif // #ifndef YY_MAIN
                   %}



{KREF}   %{
          // Expect a reference variable
          BEGIN(SREF);
         %}

<SREF>{BSTRING} %{
                 lvalp->stringval = strdup(YYText());
                 BEGIN(INITIAL);
#ifndef YY_MAIN
                  return REFNAME;
#endif // #ifndef YY_MAIN
                                %}

{KBOOL}  %{
          // Expect a boolean-valued variable
          BEGIN(SBOOL);
         %}

<SBOOL>{BSTRING} %{
                   lvalp->stringval = strdup(YYText());
                   BEGIN(SBOOLLNAME);
#                  ifndef YY_MAIN                 
                     return LNAME;
#                  endif // #ifndef YY_MAIN
                   %}

{KINT}   %{
          // Expect an integer-valued variable
          BEGIN(SINT);
         %}

<SINT>{BSTRING} %{
                lvalp->stringval = strdup(YYText());
               BEGIN(SINTLNAME);
#             ifndef YY_MAIN
                 return LNAME;
#             endif // #ifndef YY_MAIN
                                   %}              

{BSTRING}/{SEP}{BSTRING} %{
               // First lookup in the list of builtin keywords
               if (matches_builtin_keyword(YYText())) REJECT;

	// Now look in the lexically tied-in table of user keywords 
	// This lets users override a keyword (other than "false" or "true")
	//struct kwd_struct *kwd_def;
               lvalp->stringval = strdup(YYText());
#ifndef YY_MAIN
	if (INTRO && (/*kwd_def =*/ (struct kwd_struct *) INTRO->Find(YYText()))) {
                 return UKWD;
              }

              return REFNAME;
#            endif // #ifndef YY_MAIN               
%}

{BSTRING} %{
	// Look up in the lexically tied-in table of user keywords first
	// This lets users override a keyword (other than "false" or "true")
              {
	//struct kwd_struct *kwd_def;
	   lvalp->stringval = strdup(YYText());
#ifndef YY_MAIN
	if (INTRO && (/*kwd_def =*/ (struct kwd_struct *) INTRO->Find(YYText()))) {
	   // Read in a userdefined kwd statement 
                 return UKWD;
        }



	   return LNAME;
#endif // #IFndef YY_MAIN
           }
        %}

<INITIAL,SINTLNAME>{LINT} %{
	sscanf(YYText(),"%d",&lvalp->intval);
              BEGIN(INITIAL);
#ifndef YY_MAIN
	return INT;
#endif
       %}


<INITIAL,SFLOATLNAME>{LFLOAT} %{
        (istrstream(YYText())) >> lvalp->floatval;
	//sscanf(YYText(),"%f",&lvalp->floatval);
              BEGIN(INITIAL);
#ifndef YY_MAIN
	return FLOAT;
#endif
         %}
 
<SFLOATLNAME>{LINT} %{
        (istrstream(YYText())) >> lvalp->floatval;
	//sscanf(YYText(),"%f",&lvalp->floatval);
              BEGIN(INITIAL);
#ifndef YY_MAIN
	return FLOAT;
#endif
         %}



<<EOF>> %{
	return LEOF;
        %}
%%

SpParserFlexLexer::SpParserFlexLexer(istream *input): yyFlexLexer(input)
{
}

    
static int matches_builtin_keyword(const char *word)
{
   char *keyword[] = {"keyword","int","float","string","reference","logical","include"};
   int k;
   for (k=0;k<7;k++) if (!strcmp(word,keyword[k])) return 1;
   return 0;
}


#ifdef YY_MAIN
extern "C" {
int yylex() 
{
    memset(commondata,0,sizeof(commonstruct));
    SpParserFlexLexer *myscanner;
    istream *inputfile = new ifstream("test.par");
    if (inputfile->bad()) {
	inputfile = &cin;
    }
    else {
        printf("Using mytestfile...\n");
    }
    myscanner = new SpParserFlexLexer(inputfile); 
    myscanner->set_debug(1);
    myscanner->yylex();
}
}
#else
extern "C" {
int yylex(YYSTYPE *lvalp,void *commondata) 
{
    return SCANNER->yylex(lvalp,commondata);
}
}
#endif // #ifdef YY_MAIN
