#include "RB_Types.H"
#include "SpAltPOAddRemove.H"
#include "PO.H"
#include "SpBinaryTree.H"
#include "SpPoBufferHeader.H"
#include "SpMemBlkHolder.H"

PO_DEFINE_CLASS_NO_ARROW_OPERATORS(PO_VOID_TYPE_CLASS)
PO_DEFINE_CLASS_NO_ARROW_OPERATORS(PO_SPEEDES_MSG_TYPE)

class SpPoDataBase;

char* CouldNotFindObject="Could not find object in list of all objects";

void
RBPO_ADD( void* chunk,
	  SpPoDataBase* db,
	  int type,
	  int nBytes){
  SpAltPoAddRemove* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltPoAddRemove();
    altItem->Alter( chunk, SpAltPoAddRemove::ADD, db, type, nBytes);
    SpCurrentAltMgr->Append( altItem );
  }
}


void
RBPO_REMOVE( void* chunk,
	     SpPoDataBase* db,
	     int type,
	     int nBytes){
  SpAltPoAddRemove* altItem;

  if( SpCurrentAltMgr->GetOptimistic() ) {
    altItem = ALLOCATE_SpAltPoAddRemove();
    altItem->Alter( chunk, SpAltPoAddRemove::REMOVE, db, type, nBytes);
    SpCurrentAltMgr->Append( altItem );
  }
}



void* PO_GET_FIRST(char* Class, SpPoDataBase *db)
{
  db->SetGetNextClassType(db->GetClassId(Class));
  return db->GetFirstClass(db->GetGetNextClassType());
}

void* PO_GET_NEXT(char * Class, SpPoDataBase *db)
{
  void* rc = NULL;
  if (db->GetClassId(Class) != db->GetGetNextClassType()){
    cout<<"Did a GetFirst() with class type "
	<<db->GetClassTypeName(db->GetGetNextClassType())
	<<" and a get next with type "<<Class<<endl;
  }
  else{
    rc = db->GetNextClass();
  }
  return rc;
}

void* PO_GET_FIRST_BLOCK(char* Class, SpPoDataBase *db)
{
  db->SetGetNextClassType(db->GetClassId(Class));
  return db->GetFirstClassBlock(db->GetGetNextClassType());
}

void* PO_GET_NEXT_BLOCK(char * Class, SpPoDataBase *db)
{
  void* rc = NULL;
  if (db->GetClassId(Class) != db->GetGetNextClassType()){
    cout<<"Did a GetFirst() with class type "
	<<db->GetClassTypeName(db->GetGetNextClassType())
	<<" and a get next with type "<<Class<<endl;
  }
  else{
    rc = db->GetNextClassBlock();
  }
  return rc;
}

int PO_GET_NUMBER_OF(char* Class, SpPoDataBase *db)
{
  int rc = db->GetNumberOf(db->GetClassId(Class));

  if (rc == -1){
    rc = 0;
  }
  return rc;
}

void  PO_RESTORE(char* buffer, SpPoDataBase *db)
{
  db->ReconstructBuffer(buffer);
}

void PO_RECONSTRUCT_BUFFER(char* buffer, SpPoDataBase *db)
{
  db->ReconstructBuffer(buffer);
}

char* PO_CREATE_BUFFER(int compressionType, SpPoDataBase *db)
{
  return db->CreateBuffer(compressionType);
}

int PO_BUFFER_SIZE(char* buffer){
  return ((SpPoBufferHeader*)buffer)->BuffSize;
}

int PO_SIZEOF_BUFFER(char* buffer){
  return *(int*)buffer;
}

void* PO_GET_NEW_ID(char * className,
		    int & PoId,
		    void* (singleFunc) (void),
		    void* (arrayFunc) (int),
		    int objectSize)
{
  if (DefaultDataBase == NULL){
    DefaultDataBase = new SpPoDataBase();
  }
  PoId = DefaultDataBase->GetClassId(className);
  DefaultDataBase->RegisterSingleConstructor(PoId, singleFunc);
  DefaultDataBase->RegisterArrayConstructor(PoId, arrayFunc);
  DefaultDataBase->RegisterSize(PoId,objectSize);
  DefaultDataBase->SetAddToVtableList();
  DefaultDataBase->SetAddToVtableType(PoId);
  void * jj = singleFunc();
  DefaultDataBase->MakeVtableIntoOffsets((void**) &jj);
  DefaultDataBase->UnsetAddToVtableList();
  DefaultDataBase->UnsetAddToVtableType();
  return jj;
}

void FriendOfSpPoDataBase::RecordOffset(void** tInt)
{
  void*&t = *tInt;
  DefaultDataBase->UnsetAddToVtableList();
  SpIntervalKey ptrKey((int)&t, (int)&t);
  DefaultDataBase->SetAddToVtableList();
  SpMemBlkHolder* holder =
    (SpMemBlkHolder*)DefaultDataBase->ListOfAllPointers->Find(ptrKey);
  if (holder){
    holder->SetPointer(t);
  }
}

void RB_Register_PO_VOID_TYPE_CLASS()
{
  PO_REGISTER_CLASS_PO_VOID_TYPE_CLASS();
}


void RB_Register_PO_SPEEDES_MSG_TYPE()
{
  PO_REGISTER_CLASS_PO_SPEEDES_MSG_TYPE();
}


int AttachPointer(void*& ptr, SpPoDataBase* db)
{
  int rc = 0;
  SpMemBlkHolder* thisObject =
    (SpMemBlkHolder*) db->LookupAddress((int)&ptr, (int)&ptr);
  if (thisObject){
    thisObject->SetPointer(ptr);
  }
  else {
    cout<<CouldNotFindObject<<endl;
    rc = 1;
  }
  return rc;
}


int DetachPointer(void*&ptr, SpPoDataBase* db)
{
  int rc = 0;
  SpMemBlkHolder* thisObject =
    (SpMemBlkHolder*) db->LookupAddress((int)&ptr, (int)&ptr);
  if (thisObject){
    thisObject->UnsetPointer(ptr);
  }
  else{
    cout<<CouldNotFindObject<<endl;
    rc = 1;
  }
  return rc;
}


void* PO_ATTACH_PTR_VOID(char*& ptr, SpPoDataBase* db)
{
  if (AttachPointer((void*&) ptr, db)){
    return NULL;
  }
  else{
    return ptr;
  }
}
