#ifndef PO_H
#define PO_H

#include "SpStdlib.H"
#include "SpPoDataBase.H"


#ifndef DEBUG_PROCESS_MODEL
#define _M_C_
#endif


extern int PersistenceEnabled;

/**********************************************************************
 * 
 **********************************************************************/

void RBPO_ADD( void* chunk,
	       SpPoDataBase* db,
	       int type,
	       int nBytes);


void RBPO_REMOVE( void* chunk,
		  SpPoDataBase* db,
		  int type,
		  int nBytes);


typedef char PO_VOID_TYPE_CLASS;
typedef char PO_SPEEDES_MSG_TYPE;

#define DefDB DefaultDataBase

extern SpPoDataBase* DefaultDataBase;

void* PO_ATTACH_PTR_VOID(char*& ptr, SpPoDataBase* db);

#ifdef __GNUC__

#define PO_ATTACH_PTR(ptr) _M_C_ \
  PO_ATTACH_PTR_VOID(((char*&)(void*&)ptr), DefDB)

#else


#ifdef __VACPP__
#define PO_ATTACH_PTR(ptr) _M_C_ \
  PO_ATTACH_PTR_VOID(((char*&)ptr), DefDB)
#else
#define PO_ATTACH_PTR(ptr) _M_C_ \
  PO_ATTACH_PTR_VOID(((char*&)ptr), DefDB)
#endif



#endif

#define PO_ATTACH_PTR_DB(ptr, db) _M_C_ \
  PO_ATTACH_PTR_PO_VOID_TYPE_CLASS((void*&)ptr, db)

#define PO_DETACH_PTR(ptr) _M_C_ \
  PO_DETACH_PTR_PO_VOID_TYPE_CLASS((PO_VOID_TYPE_CLASS*&)ptr, DefDB)

#define PO_DETACH_PTR_DB(ptr, db) _M_C_ \
  PO_DETACH_PTR_PO_VOID_TYPE_CLASS((void*&)ptr, db)

inline void* PO_ADD(void *ptr,
		    int nbytes,
		    int classID = SpPoDataBase::DefaultAddType,
		    SpPoDataBase *db = DefDB)
{
  void* rc = NULL;
  if (db){
    rc =  db->Add(ptr, nbytes, classID);
  }
  return rc;
}


inline static void* PO_REMOVE(void *ptr, SpPoDataBase *db = DefDB)
{
  void* rc = NULL;
  if (db){
    rc =  db->Remove(ptr);
  }
  return rc;
}

int PO_SIZEOF_BUFFER(char* buffer);

int PO_BUFFER_SIZE(char* buffer);

char* PO_CREATE_BUFFER(int compressionType=
		       SpPoDataBase::NO_COMPRESSION,
		       SpPoDataBase *db = DefDB);

void PO_RECONSTRUCT_BUFFER(char* buffer, SpPoDataBase *db = DefDB);

void  PO_RESTORE(char* buffer, SpPoDataBase *db =DefDB);

int PO_GET_NUMBER_OF(char* Class, SpPoDataBase *db = DefDB);

void* PO_GET_FIRST(char* Class, SpPoDataBase *db=DefDB);

void* PO_GET_NEXT(char * Class, SpPoDataBase *db=DefDB);

void* PO_GET_FIRST_BLOCK(char* Class, SpPoDataBase *db = DefDB);

void* PO_GET_NEXT_BLOCK(char * Class, SpPoDataBase *db = DefDB);

inline static void PO_SET_NAME(void* ptr, char * name, SpPoDataBase *db = DefDB)
{
  if (db){
    db->SetName(ptr, name);
  }
}

inline static void * PO_LOOKUP_NAME(char* name, SpPoDataBase *db = DefDB)
{
  void* rc = NULL;
  if (db){
    rc =  db->LookupName(name);
  }
  return rc;
}


void* PO_GET_NEW_ID(char* className,
		    int & PoId,
		    void* (singleFunc) (void),
		    void* (arrayFunc)(int),
		    int objectSize);

/* end of non macro generated public persistence calls */


#define PO_GET_CLASS_ID(ClassName) _M_C_ \
  static void *VOID_RC_PO_NEW_##ClassName##()                           _M_C_ \
  {                                                                     _M_C_ \
     SpPoDataBase::DoNotAllocateMemoryInConstructor = 1;                _M_C_ \
     void* rc = (void *) new ClassName [1];                             _M_C_ \
     SpPoDataBase::DoNotAllocateMemoryInConstructor = 0;                _M_C_ \
     return rc;                                                         _M_C_ \
  }                                                                     _M_C_ \
                                                                        _M_C_ \
                                                                        _M_C_ \
  static void *VOID_RC_PO_NEW_ARRAY_##ClassName##(int arraySize)        _M_C_ \
  {                                                                     _M_C_ \
     SpPoDataBase::DoNotAllocateMemoryInConstructor = 1;                _M_C_ \
     void* rc = (void*) new ClassName [arraySize];                      _M_C_ \
     SpPoDataBase::DoNotAllocateMemoryInConstructor = 0;                _M_C_ \
     return rc;                                                         _M_C_ \
  }                                                                     _M_C_ \
                                                                        _M_C_ \
                                                                        _M_C_ \
  static void PO_GET_NEW_ID_##ClassName##()                             _M_C_ \
  {                                                                     _M_C_ \
    if (PersistenceEnabled){                                            _M_C_ \
      delete [](ClassName*)PO_GET_NEW_ID(#ClassName,                    _M_C_ \
  	    	                         PO_##ClassName##_ID,           _M_C_ \
  	                                 VOID_RC_PO_NEW_##ClassName,    _M_C_ \
    	                                 VOID_RC_PO_NEW_ARRAY_##ClassName,_M_C_ \
    	                                 sizeof(ClassName));            _M_C_ \
    }                                                                   _M_C_ \
  }                                                                     _M_C_ \
                                                                        _M_C_ \
                                                                        _M_C_ \
  inline static int PO_GET_##ClassName##_ID()                           _M_C_ \
  {                                                                     _M_C_ \
    if (PO_##ClassName##_ID  == -1){                                    _M_C_ \
      PO_GET_NEW_ID_##ClassName##();                                    _M_C_ \
    }                                                                   _M_C_ \
    return PO_##ClassName##_ID;                                         _M_C_ \
  }



#define PO_ADD_CLASS(ClassName) _M_C_ \
  inline static ClassName* PO_ADD_##ClassName (ClassName * ptr,         _M_C_ \
		                               int arraySize = 1,       _M_C_ \
		                               SpPoDataBase *db = DefDB)_M_C_ \
  {                                                                     _M_C_ \
    PO_ADD((void *)ptr, sizeof(ClassName)*arraySize,                    _M_C_ \
           PO_GET_##ClassName##_ID(), db);                              _M_C_ \
    return ptr;                                                         _M_C_ \
  }                                                                     _M_C_ \
                                                                        _M_C_ \
                                                                        _M_C_ \
  inline static ClassName* RB_PO_ADD_##ClassName ( ClassName * ptr,     _M_C_ \
		                              int arraySize = 1,        _M_C_ \
		                              SpPoDataBase *db = DefDB) _M_C_ \
  {                                                                     _M_C_ \
    PO_ADD((void *)ptr, sizeof(ClassName)*arraySize,                    _M_C_ \
           PO_GET_##ClassName##_ID(), db);                              _M_C_ \
    RBPO_ADD(ptr, db, PO_GET_##ClassName##_ID(),                        _M_C_ \
             sizeof(ClassName)*arraySize);                              _M_C_ \
    return ptr; _M_C_ \
  }

#define PO_NEW_CLASS(ClassName)                                         _M_C_ \
  inline static ClassName *PO_NEW_##ClassName (SpPoDataBase *db = DefDB)_M_C_ \
  {                                                                     _M_C_ \
    db->ResetPostponedAttachList();                                     _M_C_ \
    ClassName * NewObject = new ClassName [1];                          _M_C_ \
    PO_ADD_##ClassName (NewObject, 1, db);                              _M_C_ \
    return NewObject;                                                   _M_C_ \
  }


#define PO_NEW_ARRAY_CLASS(ClassName)                                   _M_C_ \
  inline static ClassName *PO_NEW_ARRAY_##ClassName (int arraySize,     _M_C_ \
				              SpPoDataBase *db = DefDB) _M_C_ \
  {                                                                     _M_C_ \
    db->ResetPostponedAttachList();                                     _M_C_ \
    ClassName * NewObject = new ClassName[arraySize];                   _M_C_ \
    PO_ADD_##ClassName (NewObject, arraySize, db);                      _M_C_ \
    return NewObject;                                                   _M_C_ \
  }

#define PO_DELETE_CLASS(ClassName)                                      _M_C_ \
  inline static int PO_DELETE_##ClassName (ClassName *ptr,              _M_C_ \
                                           SpPoDataBase *db = DefDB)    _M_C_ \
  {                                                                     _M_C_ \
    int rc = 1;                                                         _M_C_ \
    if (PO_REMOVE(ptr, db)){                                            _M_C_ \
      delete [] ptr;                                                    _M_C_ \
      rc = 0;                                                           _M_C_ \
    }                                                                   _M_C_ \
    return rc;                                                          _M_C_ \
  }

#define PO_DELETE_ARRAY_CLASS(ClassName)                                _M_C_ \
  inline static int PO_DELETE_ARRAY_##ClassName ( ClassName *ptr,       _M_C_ \
				              SpPoDataBase *db = DefDB) _M_C_ \
  {                                                                     _M_C_ \
    int rc = 1;                                                         _M_C_ \
    if (PO_REMOVE(ptr, db)){                                            _M_C_ \
      delete [] ptr;                                                    _M_C_ \
      rc = 0;                                                           _M_C_ \
    }                                                                   _M_C_ \
    return rc;                                                          _M_C_ \
  }

#define PO_REMOVE_CLASS(ClassName)                                      _M_C_ \
  inline static ClassName* PO_REMOVE_##ClassName (ClassName * ptr,      _M_C_ \
				              SpPoDataBase *db = DefDB) _M_C_ \
  {                                                                     _M_C_ \
    PO_REMOVE(ptr, db);                                                 _M_C_ \
    return ptr;                                                         _M_C_ \
  }                                                                     _M_C_ \
                                                                        _M_C_ \
  inline static ClassName* RB_PO_REMOVE_##ClassName (ClassName * ptr,   _M_C_ \
				              SpPoDataBase *db = DefDB) _M_C_ \
  {                                                                     _M_C_ \
    PO_REMOVE(ptr, db);                                                 _M_C_ \
    RBPO_REMOVE(ptr, db, PO_GET_##ClassName##_ID(),                     _M_C_ \
                db->GetSize(ptr));                                      _M_C_ \
    return ptr;                                                         _M_C_ \
  }

int AttachPointer(void*& ptr, SpPoDataBase* db);

#define PO_ATTACH_PTR_CLASS(ClassName)                                  _M_C_ \
  inline int PO_ATTACH_PTR_##ClassName (ClassName *&ptr,                _M_C_ \
				        SpPoDataBase *db = DefDB)       _M_C_ \
  {                                                                     _M_C_ \
    return AttachPointer((void*&)ptr, db);                              _M_C_ \
  }

int DetachPointer(void*& ptr, SpPoDataBase* db);

#define PO_DETACH_PTR_CLASS(ClassName)                                  _M_C_ \
  inline int PO_DETACH_PTR_##ClassName (ClassName *&ptr,                _M_C_ \
				        SpPoDataBase *db = DefDB)       _M_C_ \
  {                                                                     _M_C_ \
    return DetachPointer((void*&)ptr, db);                              _M_C_ \
  }

#define PO_SET_NAME_CLASS(ClassName)                                    _M_C_ \
  inline static void PO_SET_NAME_##ClassName (ClassName *ptr,           _M_C_ \
				       char * name,                     _M_C_ \
				       SpPoDataBase *db = DefDB)        _M_C_ \
  {                                                                     _M_C_ \
    void * voidPtr = (void *)ptr;                                       _M_C_ \
    db->SetName(voidPtr, name);                                         _M_C_ \
  }

#define PO_LOOKUP_NAME_CLASS(ClassName)                                 _M_C_ \
  inline static ClassName *PO_LOOKUP_NAME_##ClassName(char * name,      _M_C_ \
				 SpPoDataBase *db = DefDB)              _M_C_ \
  {                                                                     _M_C_ \
    return ( ClassName *)db->LookupName(name);                          _M_C_ \
  }

#define PO_GET_NUMBER_OF_CLASS( ClassName )                             _M_C_ \
  inline static int PO_GET_NUMBER_OF_##ClassName(SpPoDataBase *db = DefDB)_M_C_ \
  {                                                                     _M_C_ \
    return db->GetNumberOf(PO_GET_##ClassName##_ID());                  _M_C_ \
  }

#define PO_GET_FIRST_CLASS( ClassName )                                 _M_C_ \
  inline static ClassName * PO_GET_FIRST_##ClassName(SpPoDataBase *db = DefDB) _M_C_ \
  {                                                                     _M_C_ \
    return ( ClassName *) _M_C_ \
      db->GetFirstClass(PO_GET_##ClassName##_ID());                     _M_C_ \
  }

#define PO_GET_NEXT_CLASS( ClassName )                                  _M_C_ \
  inline static ClassName * PO_GET_NEXT_##ClassName(SpPoDataBase *db = DefDB) _M_C_ \
  {                                                                     _M_C_ \
    return ( ClassName *) db->GetNextClass();                           _M_C_ \
  }

#define PO_GET_FIRST_BLOCK_CLASS( ClassName )                           _M_C_ \
  inline static ClassName * PO_GET_FIRST_BLOCK_##ClassName(SpPoDataBase *db = DefDB) _M_C_ \
  {                                                                     _M_C_ \
    return ( ClassName *)                                               _M_C_ \
      db->GetFirstClassBlock(PO_GET_##ClassName##_ID());                _M_C_ \
  }

#define PO_GET_NEXT_BLOCK_CLASS( ClassName )                            _M_C_ \
  inline static ClassName * PO_GET_NEXT_BLOCK_##ClassName(SpPoDataBase *db = DefDB) _M_C_ \
  {                                                                     _M_C_ \
    return ( ClassName *) db->GetNextClassBlock();                      _M_C_ \
  }

#define PO_CREATE_BUFFER_CLASS( ClassName )                             _M_C_ \
  inline static char * PO_CREATE_BUFFER_##ClassName##(int compressionType, _M_C_ \
				               ClassName * ptr,         _M_C_ \
					       SpPoDataBase *db = DefDB)_M_C_ \
  {                                                                     _M_C_ \
    return db->CreateBuffer(compressionType,                            _M_C_ \
                                  ptr);                                 _M_C_ \
  }
 
#define PO_RECONSTRUCT_CLASS( ClassName )                               _M_C_ \
  inline static ClassName * PO_RECONSTRUCT_##ClassName##(char * buffer, _M_C_ \
					      SpPoDataBase *db = DefDB) _M_C_ \
  {                                                                     _M_C_ \
    db->ReconstructBuffer(buffer);                                      _M_C_ \
    return                                                              _M_C_ \
     ( ClassName *)db->GetFirstClass(db->GetClassId( #ClassName ));     _M_C_ \
  }


class FriendOfSpPoDataBase{
public:
  FriendOfSpPoDataBase():OffsetRecorded(0){}
  void AttachPointer(void **tInt){
    if ((OffsetRecorded == 0) && (PersistenceEnabled)){
      RecordOffset(tInt);
    }
  }

protected:
private:
  void RecordOffset(void** tInt);
  int OffsetRecorded;
};


#define _BASE_CREATE_SMART_POINTERS( ClassName)  _M_C_ \
class PO_PTR_##ClassName : protected FriendOfSpPoDataBase{ _M_C_ \
private: _M_C_ \
  ClassName * Value; _M_C_ \
protected: _M_C_ \
public: _M_C_ \
  PO_PTR_##ClassName() {;} _M_C_ \
  PO_PTR_##ClassName(ClassName *t){ _M_C_ \
    Value=t; _M_C_ \
    AttachPointer((void**)&(Value)); _M_C_ \
  } _M_C_ \
 _M_C_ \
  ClassName * operator = (ClassName * t){ _M_C_ \
    Value=t; _M_C_ \
    AttachPointer((void**)&(Value)); _M_C_ \
    return Value; _M_C_ \
  } _M_C_ \
 _M_C_ \
  ClassName * operator = (PO_PTR_##ClassName & t){ _M_C_ \
    Value = t.Value; _M_C_ \
    AttachPointer((void**)&(Value)); _M_C_ \
    return Value; _M_C_ \
  } _M_C_ \
 _M_C_ \
  operator ClassName *() const { return Value;} _M_C_ \
 _M_C_ \

#define _SMART_POINTER_ARROW_OPERATOR( ClassName)  _M_C_ \
  ClassName * operator -> () const { return Value; }  _M_C_ \
 _M_C_

#define _SMART_POINTERS_UNARY_OPERATORS_PROTOTYPES( ClassName)  _M_C_ \
  ClassName * operator += (int i); _M_C_ \
 _M_C_ \
  ClassName * operator -= (int i); _M_C_ \
 _M_C_ \
  ClassName * operator ++ (); _M_C_ \
 _M_C_ \
  ClassName * operator -- (); _M_C_ \
 _M_C_ \
  ClassName * operator ++ (int); _M_C_ \
 _M_C_ \
  ClassName * operator -- (int); _M_C_ \
 _M_C_ \
}; _M_C_


#define _SMART_POINTERS_UNARY_OPERATORS_STATIC( ClassName)  _M_C_ \

#define UNDEFINED_FOR_NOW_SINCE_I_DONT_KNOW_HOW_IT_CAN_BE_DONE _M_C_ \
ClassName * PO_PTR_##ClassName::operator += (int i){ _M_C_ \
  Value += i; _M_C_ \
  AttachPointer((void**)&(Value)); _M_C_ \
  return Value; _M_C_ \
} _M_C_ \
_M_C_ \
ClassName * PO_PTR_##ClassName::operator -= (int i){ _M_C_ \
  Value -= i; _M_C_ \
  AttachPointer((void**)&(Value)); _M_C_ \
  return Value; _M_C_ \
} _M_C_ \
_M_C_ \
ClassName * PO_PTR_##ClassName::operator ++ (){ _M_C_ \
  Value++; _M_C_ \
  AttachPointer((void**)&(Value)); _M_C_ \
  return Value; _M_C_ \
} _M_C_ \
_M_C_ \
ClassName * PO_PTR_##ClassName::operator -- (){ _M_C_ \
  Value--; _M_C_ \
  AttachPointer((void**)&(Value)); _M_C_ \
  return Value; _M_C_ \
} _M_C_ \
_M_C_ \
ClassName * PO_PTR_##ClassName::operator ++ (int){ _M_C_ \
  Value++; _M_C_ \
  AttachPointer((void**)&(Value)); _M_C_ \
  return Value - 1; _M_C_ \
} _M_C_ \
_M_C_ \
ClassName * PO_PTR_##ClassName::operator -- (int){ _M_C_ \
  Value--; _M_C_ \
  AttachPointer((void**)&(Value)); _M_C_ \
  return Value + 1; _M_C_ \
}



#define _SMART_POINTERS_UNARY_OPERATORS( ClassName)  _M_C_ \
  ClassName * operator += (int i){ _M_C_ \
    Value += i; _M_C_ \
    AttachPointer((void**)&(Value)); _M_C_ \
    return Value; _M_C_ \
  } _M_C_ \
 _M_C_ \
  ClassName * operator -= (int i){ _M_C_ \
    Value -= i; _M_C_ \
    AttachPointer((void**)&(Value)); _M_C_ \
    return Value; _M_C_ \
  } _M_C_ \
 _M_C_ \
  ClassName * operator ++ (){ _M_C_ \
    Value++; _M_C_ \
    AttachPointer((void**)&(Value)); _M_C_ \
    return Value; _M_C_ \
  } _M_C_ \
 _M_C_ \
  ClassName * operator -- (){ _M_C_ \
    Value--; _M_C_ \
    AttachPointer((void**)&(Value)); _M_C_ \
    return Value; _M_C_ \
  } _M_C_ \
 _M_C_ \
  ClassName * operator ++ (int){ _M_C_ \
    Value++; _M_C_ \
    AttachPointer((void**)&(Value)); _M_C_ \
    return Value - 1; _M_C_ \
  } _M_C_ \
 _M_C_ \
  ClassName * operator -- (int){ _M_C_ \
    Value--; _M_C_ \
    AttachPointer((void**)&(Value)); _M_C_ \
    return Value + 1; _M_C_ \
  } _M_C_ \
 _M_C_ \
}; _M_C_


#define PO_CREATE_SMART_POINTERS( ClassName ) _M_C_ \
  _BASE_CREATE_SMART_POINTERS( ClassName) _M_C_ \
  _SMART_POINTER_ARROW_OPERATOR( ClassName)  _M_C_ \
  _SMART_POINTERS_UNARY_OPERATORS( ClassName)

#define PO_CREATE_SMART_POINTERS_NO_ARROW_OPERATOR( ClassName)  _M_C_ \
  _BASE_CREATE_SMART_POINTERS( ClassName) _M_C_ \
  _SMART_POINTERS_UNARY_OPERATORS( ClassName )

#define PO_CREATE_SMART_POINTERS_PROTOTYPES_FOR_UNARY_OPERATORS( ClassName ) _M_C_ \
  _BASE_CREATE_SMART_POINTERS( ClassName) _M_C_ \
  _SMART_POINTER_ARROW_OPERATOR( ClassName)  _M_C_ \
  _SMART_POINTERS_UNARY_OPERATORS_PROTOTYPES( ClassName)

#define PO_CREATE_SMART_POINTER_POINTERS(ClassName) _M_C_ \
class PO_PTR_PTR_##ClassName : protected FriendOfSpPoDataBase{ _M_C_ \
private: _M_C_ \
  PO_PTR_##ClassName* Value; _M_C_ \
protected: _M_C_ \
public: _M_C_ \
  PO_PTR_PTR_##ClassName(){} _M_C_ \
  PO_PTR_PTR_##ClassName(PO_PTR_##ClassName*t){ _M_C_ \
    Value=t; _M_C_ \
    PO_ATTACH_PTR_PO_PTR_##ClassName( Value ); _M_C_ \
  } _M_C_ \
  PO_PTR_##ClassName * operator = (PO_PTR_##ClassName * t){ _M_C_ \
        Value=t; _M_C_ \
    PO_ATTACH_PTR_PO_PTR_##ClassName( Value ); _M_C_ \
    return Value; _M_C_ \
  } _M_C_ \
 _M_C_ \
  operator PO_PTR_##ClassName * (){return Value;} _M_C_ \
  PO_PTR_##ClassName operator [] (int i){return Value[i];} _M_C_ \
}; _M_C_ \



#define PO_REGISTER_CLASS(ClassName) _M_C_ \
inline static void PO_REGISTER_CLASS_##ClassName() _M_C_ \
{ _M_C_ \
  PO_GET_##ClassName##_ID(); _M_C_ \
  PO_GET_PO_PTR_##ClassName##_ID(); _M_C_ \
} _M_C_ \



  




#define PO_DEFINE_PERSISTENT_CLASS(ClassName) _M_C_ \
static int PO_##ClassName##_ID = -1; _M_C_ \
 _M_C_ \
 _M_C_ \
 _M_C_ \
PO_GET_CLASS_ID(ClassName) _M_C_ \
 _M_C_ \
PO_ADD_CLASS( ClassName ) _M_C_ \
 _M_C_ \
PO_REMOVE_CLASS( ClassName ) _M_C_ \
 _M_C_ \
 _M_C_ \
 _M_C_ \
PO_NEW_CLASS( ClassName ) _M_C_ \
 _M_C_ \
PO_NEW_ARRAY_CLASS( ClassName ) _M_C_ \
 _M_C_ \
PO_DELETE_CLASS( ClassName ) _M_C_ \
 _M_C_ \
PO_DELETE_ARRAY_CLASS( ClassName ) _M_C_ \
 _M_C_ \
 _M_C_ \
 _M_C_ \
PO_SET_NAME_CLASS( ClassName ) _M_C_ \
 _M_C_ \
PO_LOOKUP_NAME_CLASS( ClassName ) _M_C_ \
 _M_C_ \
 _M_C_ \
 _M_C_ \
PO_ATTACH_PTR_CLASS( ClassName ) _M_C_ \
 _M_C_ \
PO_DETACH_PTR_CLASS( ClassName ) _M_C_ \
 _M_C_ \
 _M_C_ \
 _M_C_ \
PO_CREATE_BUFFER_CLASS( ClassName ) _M_C_ \
 _M_C_ \
PO_RECONSTRUCT_CLASS( ClassName ) _M_C_ \
 _M_C_ \
 _M_C_ \
 _M_C_ \
PO_GET_NUMBER_OF_CLASS( ClassName ) _M_C_ \
 _M_C_ \
PO_GET_FIRST_CLASS( ClassName ) _M_C_ \
 _M_C_ \
PO_GET_NEXT_CLASS( ClassName ) _M_C_ \
 _M_C_ \
PO_GET_FIRST_BLOCK_CLASS( ClassName ) _M_C_ \
 _M_C_ \
PO_GET_NEXT_BLOCK_CLASS( ClassName ) _M_C_ \
 _M_C_ \


#define _PO_WARNING_REMOVAL_FCTN(name) _M_C_ \
  v = &name; _M_C_ \

#define _PO_WARNING_REMOVAL(ClassName) _M_C_ \
static void _PO_WARNING_REMOVAL_##ClassName() {  _M_C_ \
  void* v; _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_REMOVE_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_NEW_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_NEW_ARRAY_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_DELETE_ARRAY_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_SET_NAME_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_ATTACH_PTR_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_DETACH_PTR_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_CREATE_BUFFER_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_RECONSTRUCT_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_GET_NUMBER_OF_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_GET_FIRST_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_GET_NEXT_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_GET_FIRST_BLOCK_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_GET_NEXT_BLOCK_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_REMOVE_PO_PTR_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_NEW_ARRAY_PO_PTR_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_DELETE_ARRAY_PO_PTR_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_SET_NAME_PO_PTR_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_LOOKUP_NAME_PO_PTR_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_DETACH_PTR_PO_PTR_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_CREATE_BUFFER_PO_PTR_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_RECONSTRUCT_PO_PTR_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_GET_NUMBER_OF_PO_PTR_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_GET_FIRST_PO_PTR_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_GET_NEXT_PO_PTR_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_GET_FIRST_BLOCK_PO_PTR_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(PO_GET_NEXT_BLOCK_PO_PTR_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(_PO_WARNING_REMOVAL_##ClassName) _M_C_ \
  _PO_WARNING_REMOVAL_FCTN(RB_PO_ADD_##ClassName)                  \
  _PO_WARNING_REMOVAL_FCTN(PO_REGISTER_CLASS_##ClassName)            \
  _PO_WARNING_REMOVAL_FCTN(RB_PO_REMOVE_##ClassName)            \
  _PO_WARNING_REMOVAL_FCTN(PO_LOOKUP_NAME_##ClassName)            \
  _PO_WARNING_REMOVAL_FCTN(RB_PO_ADD_PO_PTR_##ClassName)            \
  _PO_WARNING_REMOVAL_FCTN(RB_PO_REMOVE_PO_PTR_##ClassName)            \
  v = v; _M_C_ \
}


#define PO_DEFINE_SMART_POINTERS(ClassName) _M_C_ \
PO_CREATE_SMART_POINTERS(ClassName) _M_C_ \
PO_DEFINE_PERSISTENT_CLASS(PO_PTR_##ClassName) _M_C_ \
PO_CREATE_SMART_POINTER_POINTERS(ClassName) _M_C_ \


#define PO_DEFINE_SMART_POINTERS_NO_ADDITIONAL_OPERATORS(ClassName) _M_C_ \
PO_CREATE_SMART_POINTERS_PROTOTYPES_FOR_UNARY_OPERATORS(ClassName) _M_C_ \
PO_DEFINE_PERSISTENT_CLASS(PO_PTR_##ClassName) _M_C_ \
PO_CREATE_SMART_POINTER_POINTERS(ClassName) _M_C_ \


#define PO_DEFINE_CLASS(ClassName) _M_C_ \
PO_DEFINE_PERSISTENT_CLASS(ClassName) _M_C_ \
PO_DEFINE_SMART_POINTERS(ClassName) _M_C_ \
PO_REGISTER_CLASS(ClassName) _M_C_ \
_PO_WARNING_REMOVAL(ClassName) _M_C_ \


#define PO_DEFINE_CLASS_NO_SMART_POINTERS(ClassName) _M_C_ \
PO_DEFINE_PERSISTENT_CLASS(ClassName) _M_C_ \
PO_REGISTER_CLASS(ClassName) _M_C_ \
_PO_WARNING_REMOVAL(ClassName) _M_C_ \

#define PO_DEFINE_CLASS_SMART_POINTERS_DECLARED_PREVIOUSLY(ClassName) _M_C_ \
PO_DEFINE_PERSISTENT_CLASS(ClassName) _M_C_ \
PO_REGISTER_CLASS(ClassName) _M_C_ \
_SMART_POINTERS_UNARY_OPERATORS_STATIC(ClassName) _M_C_ \
_PO_WARNING_REMOVAL(ClassName)

#define PO_DEFINE_CLASS_NO_ARROW_OPERATORS(ClassName) _M_C_ \
PO_DEFINE_PERSISTENT_CLASS(ClassName) _M_C_ \
PO_CREATE_SMART_POINTERS_NO_ARROW_OPERATOR(ClassName) _M_C_ \
PO_DEFINE_PERSISTENT_CLASS(PO_PTR_##ClassName) _M_C_ \
PO_REGISTER_CLASS(ClassName) _M_C_ \
PO_CREATE_SMART_POINTER_POINTERS(ClassName) _M_C_ \
_PO_WARNING_REMOVAL(ClassName) _M_C_ \


//PO_DEFINE_CLASS_NO_ARROW_OPERATORS(long)
//PO_DEFINE_CLASS_NO_ARROW_OPERATORS(int)
//PO_DEFINE_CLASS_NO_ARROW_OPERATORS(char)
//PO_DEFINE_CLASS_NO_ARROW_OPERATORS(double)
//PO_DEFINE_CLASS_NO_ARROW_OPERATORS(float)



#endif
