#include "SpStdlib.H"
#include "SpIostream.H"
#include "SpStdio.H"
#include "SpPoDataBase.H"
#include "SpPersistencePtr.H"
#include "SpList.H"
#include "SpPersistenceBaseClass.H"
#include "SpNetTypes.H"
#include "SpPoMemoryMapping.H"
#include "Splzhuf.H"
#include "SpPoBufferHeader.H"
#include "SpClassNameAndId.H"
#include "PO.H"
#include "SpPoNameAndAddress.H"
#include "SpBinaryTree.H"
#include "SpMemBlkHolder.H"
#include "SpDynPtrArray.H"

PO_DEFINE_CLASS_NO_ARROW_OPERATORS(PO_VOID_TYPE_CLASS);

int PersistenceEnabled  = 0;

SpDynPtrArray *SpPoDataBase::VtableOffsetValues = NULL;
SpBinaryTree *SpPoDataBase::ClassIDs = NULL;
SpDynPtrArray *SpPoDataBase::SizeList = NULL;
SpDynPtrArray *SpPoDataBase::ArrayConstructors = NULL;
SpDynPtrArray *SpPoDataBase::SingleConstructors = NULL;

int SpPoDataBase::DoNotAllocateMemoryInConstructor = 0;

SpPoDataBase* DefaultDataBase = NULL;


int SpPoDataBase::PO_errno = 0;
int SpPoDataBase::NumClasses = 0;
int SpPoDataBase::DefaultAddType = -1;



inline int SpPoDataBase::NumVtables(int classId)
{
  return ((SpList*)(*VtableOffsetValues)[classId])->GetNumElements();
}


void SpPoDataBase::InternalStoreVtableInfo(void **pObject){
  int * ptr = new int;
  *ptr = (int)(*pObject);
  ((SpList*)(*VtableOffsetValues)[AddToVtableType])->Insert((void*)ptr);
};


void SpPoDataBase::IncrementCount(int classId, int numItems)
{
  (*ClassCounts)[classId] = (void*)(((int)(*ClassCounts)[classId]) + numItems);
};

void SpPoDataBase::DecrementCount(int classId, int numItems)
{
  (*ClassCounts)[classId] = (void*)(((int)(*ClassCounts)[classId]) - numItems);
};


void SpPoDataBase::PO_perror(char* msg){

  switch(PO_errno){
  case PO_NoError:
    cout<<"No error "<<msg<<endl;
    break;
  case PO_RemovedAddressNotInDataBase:
    cout<<"Tried to Remove an address that has not been added! "<<msg<<endl;
    break;
  case PO_ClassNotRegistered:
    cout<<"Class has not been registered this database "<<msg<<endl;
    break;
  case PO_SetPointerObjectNotInDatabase:
    cout<<"Tried to set the pointer within an object that was not in the "
	<<"database "<<msg<<endl;
    break;
  case PO_UnSetPointerObjectNotInDatabase:
    cout<<"Tried to UnSet the pointer within an object that was not in the "
	<<"database "<<msg<<endl;
    break;
  case PO_AddAddressTwice:
    cout<<"Tried to add an address twice "<<msg<<endl;
    break;
  default:
    cout<<"Unknown error "<<PO_errno<<" "<<msg<<endl;
    break;
  }
}

SpPoDataBase::SpPoDataBase() :
  ClassCounts(new SpDynPtrArray),
  ListOfClassesByClassType(new SpDynPtrArray),
  AddToVtableList(0),
  AddToVtableType(-1),
  PtrsByName(new SpBinaryTree),
  ListOfAllPointers(new SpBinaryTree),
  CurrentItemPointer(NULL),
  PostponedAttachList(new SpList)
{
  if (ClassIDs == NULL){
    ClassIDs = new SpBinaryTree;
    ClassIDs->SetBalancedTreeMode();
    SizeList = new SpDynPtrArray;
    SingleConstructors = new SpDynPtrArray;
    ArrayConstructors = new SpDynPtrArray;
    VtableOffsetValues = new SpDynPtrArray;
  }

  for (int i=99; i >= 0; --i){ // Initialize in reverse to reduce # of times
    // we enlarge the array.
    (*ClassCounts)[i] = NULL;
  }
  if (VtableOffsetValues == NULL){
    for (int i=99; i >= 0; --i){ //Ditto
    }
  }
  PtrsByName->SetBalancedTreeMode();
  //ListOfAllPointers->SetSplayTreeMode();
  ListOfAllPointers->SetBalancedTreeMode();
  if ((DefaultAddType == -1) && (DefaultDataBase != NULL)){
    int OldPersistenceEnabled = PersistenceEnabled;
    PersistenceEnabled = 1;
    PO_REGISTER_CLASS_PO_VOID_TYPE_CLASS();
    PersistenceEnabled = OldPersistenceEnabled;
    DefaultAddType = GetClassId("PO_VOID_TYPE_CLASS");
  }
  
}



SpPoDataBase::~SpPoDataBase()
{
  int i;
  int len = NumClasses;
  for (i=0; i < len; ++i){
    SpBinaryTree * classList = (SpBinaryTree*)(*ListOfClassesByClassType)[i];
    if (classList){
      int nOfThisClass = classList->GetNumElements();
      for (int j=0; j < nOfThisClass; ++j){
	(*classList) -= classList->GetFirstElement();
      }
      delete classList;
    }
  }
  len = PtrsByName->GetNumElements();
  for (i=0; i < len; ++i){
    delete (SpPoNameAndAddress*)PtrsByName->RemoveFirstElement();
  }

  len = ListOfAllPointers->GetNumElements();
  for (i=0; i < len; ++i){
    delete ((SpMemBlkHolder*)ListOfAllPointers->RemoveFirstElement());
  }
  delete ListOfAllPointers;
  delete PtrsByName;
  delete ClassCounts;
  delete ListOfClassesByClassType;
  delete PostponedAttachList;
}


const char* SpPoDataBase::GetClassTypeName(int classId)
{
  SpClassNameAndId *nameAndId = (SpClassNameAndId*)ClassIDs->GetFirstElement();

  while (nameAndId){
    if (nameAndId->GetId() == classId){
      return nameAndId->GetName();
    }
    nameAndId = (SpClassNameAndId*)ClassIDs->GetNextElement();
  }
  return NULL;
}


int SpPoDataBase::GetSize(void * Pointer)
{
  static SpIntervalKey key;
  int rc = -1;
  key((int)Pointer, (int)Pointer);
  SpMemBlkHolder* PtrList = (SpMemBlkHolder*) ListOfAllPointers->Find(key);
  if (PtrList){
    rc = PtrList->GetSize();
  }
  return rc;
}




int SpPoDataBase::GetClassId(char * name)
{
  int classId = -1;

  /* first see if this class name has been given an ID already */

  SpClassNameAndId* id = (SpClassNameAndId*)ClassIDs->Find(name);
  if (id != NULL){
    classId = id->GetId();
  }

  if (classId == -1){
    classId = NumClasses;
    
    id = new SpClassNameAndId;
    id->SetName(name);
    id->SetId(classId);
    ClassIDs->Insert(id, name);

    (*ClassCounts)[classId] = (void*)0;
    (*ListOfClassesByClassType)[classId] = new SpBinaryTree;
    ((SpBinaryTree*)((*ListOfClassesByClassType)[classId]))->
      SetBalancedTreeMode();
    (*VtableOffsetValues)[classId] = new SpList;
    ++NumClasses;
  }
  return classId;
};





int SpPoDataBase::GetNumberOf(int classId)
{
  if (classId < (*ClassCounts).GetArraySize()){
    return (int)(*ClassCounts)[classId];
  }
  else{
    PO_errno = PO_ClassNotRegistered;
    return -1;
  }
};


void* SpPoDataBase::GetFirstClassBlock(int classId){
  void * rc = NULL;
  CurrentClassForGetNextClass = classId;
  if ((*ClassCounts)[classId] != NULL){
    SpMemBlkHolder* pPtr = (SpMemBlkHolder*) 
      ((SpBinaryTree*)(*ListOfClassesByClassType)[classId])->GetFirstElement();
    if (pPtr){
      rc = pPtr->GetBaseAddr();
    }
    GetNextObjectSize = (int)(*SizeList)[CurrentClassForGetNextClass];
    GetNextNumArrayElements = pPtr->GetSize()/GetNextObjectSize;
  }
  else{
    return NULL;
  }
  CurrentItemPointer = rc;
  return rc;


}

void* SpPoDataBase::GetNextClassBlock()
{
  SpMemBlkHolder* ListObject = (SpMemBlkHolder*)
    ((SpBinaryTree*)(*ListOfClassesByClassType)[CurrentClassForGetNextClass])
    ->GetNextElement();
  if (ListObject){
    return ListObject->GetBaseAddr();
  }
  else{
    return NULL;
  }
}

static int currentArrayElement;

void* SpPoDataBase::GetFirstClass(int classId)
{
  void * rc = NULL;
  currentArrayElement = 1;
  CurrentClassForGetNextClass = classId;
  if ((*ClassCounts)[classId] != NULL){
    SpMemBlkHolder* pPtr = (SpMemBlkHolder*) 
      ((SpBinaryTree*)(*ListOfClassesByClassType)[classId])->GetFirstElement();
    if (pPtr){
      rc = pPtr->GetBaseAddr();
    }
    GetNextObjectSize = (int)(*SizeList)[CurrentClassForGetNextClass];
    GetNextNumArrayElements = pPtr->GetSize()/GetNextObjectSize;
  }
  else{
    return NULL;
  }
  CurrentItemPointer = rc;
  return rc;
};


void* SpPoDataBase::GetNextClass()
{
  void * rc = NULL;

  if (CurrentItemPointer){
    
    if (currentArrayElement < GetNextNumArrayElements){
      ++currentArrayElement;
      CurrentItemPointer  = ((char*)CurrentItemPointer) + GetNextObjectSize;
      rc = CurrentItemPointer;
    }
    else{
      SpMemBlkHolder* ListObject = (SpMemBlkHolder*)
	((SpBinaryTree*)(*ListOfClassesByClassType)[CurrentClassForGetNextClass])
	->GetNextElement();
      currentArrayElement = 1;
      if (ListObject != NULL){
	rc = ListObject->GetBaseAddr();
	GetNextObjectSize = (int)(*SizeList)[CurrentClassForGetNextClass];
	GetNextNumArrayElements = ListObject->GetSize()/GetNextObjectSize;
	CurrentItemPointer = rc;
      }
    }
  }
  return rc;
};





void SpPoDataBase::SetName(void *&ptr, char * name)
{
  if (name){
    static SpIntervalKey key;
    SpMemBlkHolder* addrPtr;
    SpPoNameAndAddress*  nameAndAddress= NULL;
    key((int)ptr, (int)ptr);

    //See if we are setting the name of a valid object
    addrPtr = (SpMemBlkHolder*)ListOfAllPointers->Find(key);
    if (addrPtr == NULL){
      PO_errno = PO_ClassNotRegistered;
      return;
    }
    //Now see if someone has already set this name
    nameAndAddress = (SpPoNameAndAddress*)PtrsByName->Find(name);
    if (nameAndAddress == NULL){
      nameAndAddress = new SpPoNameAndAddress(ptr, name);  
      PtrsByName->Insert(nameAndAddress, name);
    }
    else{
      nameAndAddress->SetPtr(ptr);
    }
  }
};





void* SpPoDataBase::LookupName(char * name)
{
  void* rc = NULL;
  SpPoNameAndAddress* blk = (SpPoNameAndAddress*) PtrsByName->Find(name);
  if (blk){
    rc = blk->GetPtr();
  }
  return rc;
};





void  SpPoDataBase::SetPointer(void *&ptr)
{
  SpIntervalKey ptrKey((int)ptr);
   SpMemBlkHolder*c = (SpMemBlkHolder*)ListOfAllPointers->Find(ptrKey);
  if (c){
    c->SetPointer(ptr);
  }
  else{
    PO_errno = PO_SetPointerObjectNotInDatabase;
  }
}


void  SpPoDataBase::UnsetPointer(void *&ptr)
{
  static SpIntervalKey key;

  key((int)&ptr, (int)&ptr);
   SpMemBlkHolder* AddrPtr = (SpMemBlkHolder*)ListOfAllPointers->Find(key);

  if (AddrPtr){
    AddrPtr->UnsetPointer(ptr);
  }
  else{
    PO_errno = PO_UnSetPointerObjectNotInDatabase;
    cout<<"Delaying attachment of pointer"<<endl;
    PostponedAttachList->Insert(new void**(&ptr));
  }
};



inline int SpPoDataBase::HasVtables(int ClassId)
{
  return ((SpList*)(*VtableOffsetValues)[ClassId])->GetNumElements() != 0;
}



void* SpPoDataBase::Add(void *&ptr, unsigned int size, int classID)
{
  if (classID == -1){
    return NULL;
  }

  SpMemBlkHolder* addrPtr;
  void * rc = NULL;

  static SpIntervalKey key;
  key = (int)ptr;

  if ((addrPtr = (SpMemBlkHolder*) ListOfAllPointers->Find(key))){
    PO_errno = PO_AddAddressTwice;
  }
  else{
    rc = ptr;
    SpBinaryTree* classList;
    classList = (SpBinaryTree *)(*ListOfClassesByClassType)[classID];
    if (classList == NULL){
      classList = new SpBinaryTree;
      classList->SetBalancedTreeMode();
      (*ListOfClassesByClassType)[classID] = classList;
    }
    addrPtr = new SpMemBlkHolder(size, classID);
    addrPtr->SetBaseAddr(ptr);
    key((int)ptr, ((int)ptr) + size - 1);
    ListOfAllPointers->Insert(addrPtr, key);
    (*classList) += addrPtr;
    IncrementCount(classID, size/(int)((*SizeList)[classID]));
  }
  int len = PostponedAttachList->GetNumElements();
  int maxPtr = ((int)ptr) + size - 1;
  for (int i=0; i < len; ++i){
    void** delayedAdd = ((void**)PostponedAttachList->Remove());
    int addr = (int)*delayedAdd;
    if ((addr >= (int)ptr) && (addr < maxPtr)){
      cout<<"Attaching delayed pointer"<<endl;
      addrPtr->SetPointer(*delayedAdd);
    }
    delete delayedAdd;
  }
  return rc;
}



void* SpPoDataBase::Remove(void *&ptr)
{
  SpMemBlkHolder *tAddrPtr;
  static SpIntervalKey key;
  void* rc = NULL;

  key((int)ptr, (int)ptr);
  tAddrPtr = (SpMemBlkHolder *)ListOfAllPointers->Find(key);
  if (tAddrPtr == NULL){
    if (ptr != NULL){ //Don't complain about trying to remove a NULL
      PO_errno = PO_RemovedAddressNotInDataBase;
    }
  }
  else{
    int classId = tAddrPtr->GetType();
    SpBinaryTree* ClassList =
      (SpBinaryTree *)(*ListOfClassesByClassType)[classId];
    if (ClassList != NULL){
      tAddrPtr = (SpMemBlkHolder*) ((*ClassList) -= tAddrPtr);
      if (tAddrPtr){
	rc = ptr;
	key((int)tAddrPtr->GetBaseAddr(),(int)tAddrPtr->GetBaseAddr());
	ListOfAllPointers->Remove(key);
	int classSize = (int)((*SizeList)[classId]);
	int numItems = tAddrPtr->GetSize()/classSize;
	DecrementCount(classId, numItems);
	delete tAddrPtr;
      }
    }
    if (rc == NULL){
      cout<<"INTERNAL ERROR!!"
	  <<"   Could not find object in list of pointers by type!!!!"<<endl;
    }
  }
  return rc;
};



static inline int Allign8Bytes(int val) {return val + (8 - val%8);}


int SpPoDataBase::GetBufferSize(int &numTypes, int & numItems,
				int &numNamedObjects){
   SpMemBlkHolder* AddrPtr =
    (SpMemBlkHolder*)ListOfAllPointers->GetFirstElement();
  int buffSize = 0;
  SpClassNameAndId* currentNameAndId;

  //Get amount of memory required for all the type names
  currentNameAndId = (SpClassNameAndId*)ClassIDs->GetFirstElement();
  while (currentNameAndId){
    ++numTypes;
    buffSize += (4*sizeof(int) +
		 Allign8Bytes(strlen(currentNameAndId->GetName())+1));
    currentNameAndId = (SpClassNameAndId*)ClassIDs->GetNextElement();
  }
  
  // Get the amount of space required for all the memory blocks, their
  // pointers, and their names if they are named
  
  if (ListOfAllPointers->GetNumElements()){
    while (AddrPtr != NULL){
      ++numItems;
      buffSize += Allign8Bytes(AddrPtr->GetSize());
      if (AddrPtr->GetPtrTree()){
	buffSize += sizeof(int)*AddrPtr->GetPtrTree()->GetNumElements();
      }
      buffSize += 4*sizeof(int);
      //buffSize += 6*sizeof(int);
      AddrPtr = (SpMemBlkHolder*)ListOfAllPointers->GetNextElement();
    }
  }
  numNamedObjects = PtrsByName->GetNumElements();
  SpPoNameAndAddress* namedObject =
    (SpPoNameAndAddress*)PtrsByName->GetFirstElement();
  for (int jj=0; jj < numNamedObjects; ++jj){
    int slen = strlen(namedObject->GetName())+1;
    buffSize += Allign8Bytes(slen) + sizeof(void*) + 2*sizeof(int);
    namedObject = (SpPoNameAndAddress*)PtrsByName->GetNextElement();
  }

  buffSize += sizeof(SpPoBufferHeader);
  return buffSize;
}


char* SpPoDataBase::CreateBuffer(int compressionType)
{
  //Initially compressionType will be ignored
  SpMemBlkHolder * AddrPtr =
    (SpMemBlkHolder*)ListOfAllPointers->GetFirstElement();
  int buffSize = 0;
  int numItems = 0;
  int numTypes = 0;
  int numNamedObjects = 0;
  char * buffer;
  int * tIntPtr;
  int i;
  SpPoBufferHeader *hdr;
  SpClassNameAndId *currentNameAndId;

  buffSize=GetBufferSize(numTypes, numItems, numNamedObjects);
  char * returnedBuffer = buffer = new char[buffSize];

  //#define DEBUGGING
#ifdef DEBUGGING
  memset(returnedBuffer, 0, buffSize);
#endif

  // Now start copying things in to the buffer so that we can save it
  // First, the header information
  hdr = (SpPoBufferHeader*)buffer;
  hdr->BuffSize = buffSize;
  hdr->CompressionType = compressionType;
  hdr->NumItems = numItems;
  hdr->NumNamedObjects = numNamedObjects;
  hdr->NumTypes = numTypes;
  buffer += 6*sizeof(int);


  // Second, copy in all the names with their corresponding ID/type

  currentNameAndId = (SpClassNameAndId*)ClassIDs->GetFirstElement();
  int len = ClassIDs->GetNumElements();
  for (int zz=0; zz < len; ++zz){
    int id = currentNameAndId->GetId();
    memcpy(buffer, (char *) &id, sizeof(int));
    buffer += 2*sizeof(int);
    memcpy(buffer, currentNameAndId->GetName(),
	   strlen(currentNameAndId->GetName())+1);
    buffer += Allign8Bytes(strlen(currentNameAndId->GetName())+1);
    currentNameAndId = (SpClassNameAndId*)ClassIDs->GetNextElement();
  }



  // Third, start copying in the objects followed by their pointers

  AddrPtr = (SpMemBlkHolder*)ListOfAllPointers->GetFirstElement();
  while (AddrPtr){
    
    tIntPtr = (int *) buffer;
    tIntPtr[0] = 0;
    tIntPtr[1] = AddrPtr->GetType();
    tIntPtr[2] = AddrPtr->GetSize();
    tIntPtr[3] = (int)AddrPtr->GetBaseAddr();
    buffer += 4*sizeof(int);

    PERSISTENCE_PTR * listPtr = NULL;
    if (AddrPtr->GetPtrTree()){
      tIntPtr[0] = AddrPtr->GetPtrTree()->GetNumElements();
      listPtr = (PERSISTENCE_PTR *)(AddrPtr->GetPtrTree()->GetFirstElement());
    }
    
    memcpy(buffer, (char *) AddrPtr->GetBaseAddr(), AddrPtr->GetSize());
    buffer += Allign8Bytes(AddrPtr->GetSize());
    memcpy(buffer, (char*)&tIntPtr[0], sizeof(int));
    int* tIntPtr2 = (int*)buffer;
    for (i=0; i < tIntPtr[0]; ++i){
      *tIntPtr2 = listPtr->GetOffset();
      ++tIntPtr2;
       int offset = listPtr->GetOffset();
       memcpy(buffer, (char *) &offset, sizeof(int));
       buffer += sizeof(int);
      listPtr = (PERSISTENCE_PTR *)(AddrPtr->GetPtrTree()->GetNextElement());
    }
    //    buffer = (char*)tIntPtr2;
    AddrPtr = (SpMemBlkHolder*)ListOfAllPointers->GetNextElement();
  }
  //Fourth, copy in the names of the named objects along with their addresses

  SpPoNameAndAddress *namedObject = (SpPoNameAndAddress*)
    PtrsByName->GetFirstElement();
  while (namedObject != NULL){
    void*& addr = namedObject->GetPtr();
    memcpy(buffer, (char *)&addr, sizeof(int));
    buffer += 2*sizeof(int);
    memcpy(buffer,
	   (char*)namedObject->GetName(),
	   strlen((char*)namedObject->GetName())+1);
    buffer += Allign8Bytes(strlen((char*)namedObject->GetName())+1);
    namedObject = (SpPoNameAndAddress*)PtrsByName->GetNextElement();
  }
  int NewSize = CompressBuffer(returnedBuffer, buffSize, compressionType);
  hdr->BuffSize = NewSize + 6*sizeof(int);
  return returnedBuffer;
};

SpBinaryTree AddressMappingTree;

int * ClassIdMappingTable = NULL;

void SpPoDataBase::AddToClassIdMappingTable(char *name, int oldClassId)
{

  int newClassId = -1;
  SpClassNameAndId *nameAndId = (SpClassNameAndId*)ClassIDs->Find(name);

  if (nameAndId){
    newClassId = nameAndId->GetId();
    ClassIdMappingTable[oldClassId] = newClassId;
  }
  else{
    cout<<"Tried to reconstruct class "<<name<<" but this class"<<endl
	<<"has not yet been registered in this database.  This class"<<endl
	<<"will be restore as char buffers!!!!!!\a"<<endl;
    nameAndId = (SpClassNameAndId*)ClassIDs->Find("PO_VOID_TYPE_CLASS");
    newClassId = nameAndId->GetId();
    ClassIdMappingTable[oldClassId] = newClassId;
  }
}

/* reads through the buffer, pulling out class type names and their
   corresponding old class id.  These are then inserted into the
   ClassIdMappingTable and a pointer to the remainder of the buffer is
   returned.*/

char * SpPoDataBase::SetupClassIdMappingTable(char * buff, int nTypes){
  int oldClassId;

  for (int i=0; i < nTypes; ++i){
    oldClassId = *((int *)buff);
    buff += 2*sizeof(int);
    AddToClassIdMappingTable(buff, oldClassId);
    buff += Allign8Bytes(strlen(buff)+1);
  }
  return buff;
}

char *SpPoDataBase::ReconstructObjects(char * buff, int nItems)
{
  static SpIntervalKey key;
  static const int vtableSize = sizeof(DUMMY_CLASS_JUST_TO_GET_SIZE_OF_VTABLE);
  int baseAddr;
  int type;
  int numPtrs;
  void * newObjectPtr;
  int numOfType;
  int newType;
  int nBytes;
  int i, j;
  int * tIntPtr;
  int offset;
  SpMemBlkHolder* addrPtr;
  char * oldObjectPtr;
  SpIntervalKey oldRange;
  SpIntervalKey newRange;
  SpPoMemoryMapping * mMap;
  char * tObjectPtr;
  int objectSize;
  int numVtables;
  char* oldClassWithVtables;

  for (i=0; i < nItems; ++i){
    tIntPtr = (int*)buff;
    numPtrs = tIntPtr[0];
    type = tIntPtr[1];
    nBytes = tIntPtr[2];
    baseAddr = tIntPtr[3];
    
    buff += 4*sizeof(int);
    
    
    newType = ClassIdMappingTable[type];
    objectSize = (int)(*SizeList)[newType];
    numOfType = nBytes/objectSize;
    void* (*f)(int) =(void* (*)(int))(*ArrayConstructors)[newType];
    newObjectPtr = f(numOfType);

    oldRange(baseAddr, baseAddr+nBytes);
    newRange((int)newObjectPtr, ((int)newObjectPtr)+nBytes);
    mMap = new SpPoMemoryMapping(oldRange, newRange);
    AddressMappingTree.Insert(mMap, oldRange);

    IncrementCount(newType, numOfType);
    oldClassWithVtables = new char[objectSize];
    memcpy(oldClassWithVtables, newObjectPtr, objectSize);

    oldObjectPtr = buff;
    numVtables = 0;
    SpList * vtableList = NULL;
    if ((*VtableOffsetValues)[newType] != NULL){
      vtableList = (SpList*)(*VtableOffsetValues)[newType];
      numVtables = vtableList->GetNumElements();
    }
    SpBinaryTree * classList =
      (SpBinaryTree *) (*ListOfClassesByClassType)[newType];
    if (classList == NULL){
      classList = new SpBinaryTree;
      classList->SetBalancedTreeMode();
      (*ListOfClassesByClassType)[newType] = classList;
    }
    static SpIntervalKey key;
    tObjectPtr = (char*)newObjectPtr;
    memcpy(newObjectPtr, oldObjectPtr, nBytes);
    if (numVtables){
      for (int kk=0; kk < numOfType; ++kk){
	int * vtableOffset = (int *) vtableList->GetFirstElement();
	for (int jj=0; jj < numVtables; ++jj){
	  memcpy(tObjectPtr + *vtableOffset,
		 oldClassWithVtables+*vtableOffset,
		 vtableSize);
	  vtableOffset = (int*)vtableList->GetNextElement();
	}
	//((SpMemBlkHolder*)(tObjectPtr))->SetType(newType);
	tObjectPtr += objectSize;
      }

      addrPtr = new SpMemBlkHolder(nBytes, newType);
      addrPtr->SetBaseAddr(newObjectPtr);
      (*classList) += addrPtr;
    }
    else{
      addrPtr = new SpMemBlkHolder(nBytes, newType);
      (*classList) += addrPtr;
      addrPtr->SetBaseAddr(newObjectPtr);
    }
    delete [] oldClassWithVtables;
    key((int)newObjectPtr,(int)newObjectPtr + nBytes);
    ListOfAllPointers->Insert(addrPtr, key);

    //Now that the objects have been created, start inserting all of the
    //pointers
    buff += Allign8Bytes(nBytes);
    tIntPtr = (int *) buff;

    for (j=0; j < numPtrs; ++j){
      PERSISTENCE_PTR *pPtr = new PERSISTENCE_PTR;
      offset = *tIntPtr;
      ++tIntPtr;
      pPtr->SetOffset(offset);
      addrPtr->AddPersistentPointer(pPtr);
    }
    buff = (char*)tIntPtr;
  }	 	 
  return buff;
}






void SpPoDataBase::UpdatePointers()
{
  SpPoMemoryMapping * mMap;
  void * newAddress;
  int diff;
  SpMemBlkHolder* tPersPtrList = (SpMemBlkHolder*)
    ListOfAllPointers->GetFirstElement();
  SpIntervalKey key;
  int ptr;

  while (tPersPtrList){
    if (tPersPtrList->GetPtrTree()){
      int numPtrs = tPersPtrList->GetPtrTree()->GetNumElements();
      PERSISTENCE_PTR* pPtr = (PERSISTENCE_PTR*)
	tPersPtrList->GetPtrTree()->GetFirstElement();
      for (int j=0; j < numPtrs; ++j){
	//First, find the memory mapping for the old pointer
	memcpy(&ptr, ((char*)tPersPtrList->GetBaseAddr())+pPtr->GetOffset(),
	       sizeof(int));
	if (ptr){
	  key(ptr, ptr);
	  mMap = (SpPoMemoryMapping*)AddressMappingTree.Find(key);
	  //Find the address of the pointer within the new object
	  if (mMap){
	    diff = ptr - mMap->OldRange.Min();
	    newAddress = (void*)(mMap->NewRange.Min()+ diff);
	    
	    //Update the pointer within our object
	    memcpy(((char*)tPersPtrList->GetBaseAddr())+pPtr->GetOffset(),
		   &newAddress, sizeof(void*));
	    
	    //Next, update the pointer list about to where we point
	  }
	  else{
	    cout<<"\a!!!!!!!!!!!!!!!!!!!!!!!!Could not remap a pointer!!!!"
		<<endl;
	  }
	}
	pPtr=(PERSISTENCE_PTR*) tPersPtrList->GetPtrTree()->GetNextElement();
      }// for
    }// if
    tPersPtrList=(SpMemBlkHolder*)ListOfAllPointers->GetNextElement();
  } //while
}





char * SpPoDataBase::ReconstructNamedObjects(char *buff, int nNamedObjects)
{
  for (int i=0; i < nNamedObjects; ++i){
    static SpIntervalKey key;
    SpPoMemoryMapping * mMap;
    SpPoNameAndAddress* nameAndAddress;

    key(*((int *)buff), *((int *)buff));
    buff += 2*sizeof(int);
    mMap = (SpPoMemoryMapping*) AddressMappingTree.Find(key);
    if (!mMap){
      cout<<"Unable to look up the address "<<key.Min()
	  <<" when reconstructing named object "<<buff<<endl;
    }
    else{
      int diff = key.Min() - (mMap->OldRange).Min();
      void *nAddr = (void *)((mMap->NewRange).Min() + diff);
      nameAndAddress = new SpPoNameAndAddress(nAddr, buff);
      PtrsByName->Insert(nameAndAddress, buff);
    }
    buff += Allign8Bytes(strlen(buff)+1);
  }
  return buff;
}


int SpPoDataBase::ReconstructBuffer(char *compressedBuffer)
{
  SpPoBufferHeader* hdr = (SpPoBufferHeader*)compressedBuffer;
  int compressionType = hdr->CompressionType;
  int numItems = hdr->NumItems;
  int numNamedObjects = hdr->NumNamedObjects;
  int numTypes = hdr->NumTypes;
  char* trueBuffer = DecompressBuffer((char*)&hdr[1],
				      hdr->BuffSize - 6*sizeof(int),
				      compressionType);
  SpPoDataBase* tempDB = DefaultDataBase;

  if (trueBuffer == NULL){
    cout<<"Unable to decompress buffer"<<endl;
    return -1;
  }

  DefaultDataBase = this;
  char* tPtr = trueBuffer;

  ClassIdMappingTable = new int[numTypes];

  tPtr = SetupClassIdMappingTable(tPtr, numTypes);



  tPtr = ReconstructObjects(tPtr, numItems);

  //Then go through the tree and update all the pointers to the new values
  UpdatePointers();

  tPtr = ReconstructNamedObjects(tPtr, numNamedObjects);

  delete [] ClassIdMappingTable;

  //If we really decompressed it, delete it!
  if (trueBuffer != compressedBuffer + 6*sizeof(int)){
    delete [] trueBuffer;
  }

  SpPoMemoryMapping * mMap =
    (SpPoMemoryMapping*)AddressMappingTree.RemoveFirstElement();
  while (mMap){
    delete mMap;
    mMap = (SpPoMemoryMapping*)AddressMappingTree.RemoveFirstElement();
  }
  ClassIdMappingTable = NULL;
  DefaultDataBase  = tempDB;
  return 0;
};





void  SpPoDataBase::RegisterSingleConstructor(int classId,
					      void* (func)())
{
  (*SingleConstructors)[classId] = func;
};





void SpPoDataBase::RegisterArrayConstructor(int classId,
					    void*(func)(int))
{
  (*ArrayConstructors)[classId] = func;
}


void SpPoDataBase::RegisterSize(int classId, int size)
{
  (*SizeList)[classId] = (void*) size;
}


int SpPoDataBase::CompressBuffer(char *buff, int size, int compressionType)
{
  int rc;
  int* tInt = (int*)buff;
  char * compressedBuffer;
  switch (compressionType)
    {
    case NO_COMPRESSION:
      rc = *(int *) buff - 6*sizeof(int);
      tInt[1] = NO_COMPRESSION;
      break;
    case BUILT_IN_LZH:
      compressedBuffer = (char*)SpCompress(buff+6*sizeof(int),
					   size - 6*sizeof(int), rc);
      if (rc < size){
	memcpy(buff + 6*sizeof(int), compressedBuffer, rc);
	tInt[1] = BUILT_IN_LZH;
      }
      else{
	int* tInt = (int*)buff;
	tInt[1] = NO_COMPRESSION;
	rc = *(int *) buff - 6*sizeof(int);
      }
      delete [] compressedBuffer;
    }
  return rc;
  //Real buffer is at buff + 6*sizeof(int);
}


char *SpPoDataBase::DecompressBuffer(char *buffer, int size, int cType)
{
  char * rc = NULL;
  switch(cType){
  case NO_COMPRESSION:
    rc = buffer;
    break;
  case BUILT_IN_LZH:
    rc = SpDecompress((char*)buffer, size, size);
    break;
  default:
    cout<<"Compression type "<<cType<<" not supported"<<endl;
    rc = (char*)(((int)rc) * size); //prevent warning in compilation
    break;
  }
  return rc;
}

char *SpPoDataBase::CreateBuffer(int, void*)
{
  cout<<'\a'<<"Not supported yet"<<endl;
  return NULL;
}

void printPtrs(SpDynPtrArray &ptrArray,
	       int numPointers,
	       int objectSize,
	       int baseAddr)
{
  PERSISTENCE_PTR *p;
  if (numPointers){
  cout<<"Memory block has "<<numPointers<<" pointer"
      <<((numPointers>1) ? "s" : "")<<endl;
    cout<<"      /--------------------------------------------------\\"<<endl;
    cout<<"     /--------------------Pointers------------------------\\"<<endl;
    cout<<"    /------------------------------------------------------\\"<<endl;
    cout<<"    |    Array    |   offset of pointer   |   destination  |"<<endl;
    cout<<"    |   element   |    within element     |   of pointer   |"<<endl;
    for (int i=0; i < numPointers; ++i){
      p = (PERSISTENCE_PTR*)ptrArray[i];
      cout<<"    |"<<setw(7);
      cout<<p->GetOffset()/objectSize<<"      |";
      cout<<setw(12);
      cout<<p->GetOffset()%objectSize<<"           |";
      cout<<setw(13);
      cout<<(int)(*((void**)(baseAddr+(int)p->GetOffset())))
	  <<"   |"<<endl;
    }
    cout<<"     ------------------------------------------------------"<<endl;
  }
  else{
    cout<<"Memory block has no pointers"<<endl;
  }
}


char *PO_stringify(int i)
{
  static char string[10];
  switch (i)
    {
    case 1:
      strcpy(string, "1st");
      break;
    case 2:
      strcpy(string, "2nd");
      break;
    case 3:
      strcpy(string, "3rd");
      break;
    default:
      sprintf(string, "%dth", i);
      break;
    }
  return string;
}



void SpPoDataBase::PrintBlock(SpMemBlkHolder* AddrPtr,
			      SpDynPtrArray* classIDs)
{
  int numPointers = 0;
  int elementSize = (int)(*SizeList)[AddrPtr->GetType()];
  static SpDynPtrArray ptrArray;

  cout<<"----------------------------------------"
      <<"---------------------------------------"<<endl;
  cout<<"-------------------------------Begin mem"
      <<"ory block------------------------------"<<endl;
  cout<<"----------------------------------------"
      <<"---------------------------------------"<<endl;
  cout<<"Class type: "<<(char*)(*classIDs)[AddrPtr->GetType()]
      <<endl;
  int ArraySize = AddrPtr->GetSize()/elementSize;
  if (ArraySize == 1){
    cout<<"Single item"<<endl;
  }
  else{
    cout<<"Array of "<<ArraySize<<" items"<<endl;
  }
  cout<<"Base address: "<<(int) AddrPtr->GetBaseAddr()<<endl;
  cout<<"Memory block size: "<<(int) AddrPtr->GetSize()<<endl;

  if (AddrPtr->GetPtrTree()){
    ptrArray[numPointers] = AddrPtr->GetPtrTree()->GetFirstElement();
    while (ptrArray[numPointers]){
      ++numPointers;
      ptrArray[numPointers] = AddrPtr->GetPtrTree()->GetNextElement();
    }
  }
  printPtrs(ptrArray, numPointers, elementSize, (int)AddrPtr->GetBaseAddr());
}


int SpPoDataBase::FindDanglingPointers()
{
  /*
  int numBlocks = ListOfAllPointers->GetNumElements();
  SpMemBlkHolder* AddrPtr=(SpMemBlkHolder*)ListOfAllPointers->GetFirstElement();
  for (int i=0; i < numBlocks; ++i){
    if (AddrPtr->GetPtrTree()){
      for (int j=0; j < AddrPtr->GetPtrTree()->GetNumElements(); ++j){
	SpIntervalKey key;
	key(fill, fill);
	if (ListOfAllPointers->Find(key) == NULL){
	  cout<<"Have a dangling pointer offset "<<AddrPtr->Get
	  */
  return 17;
}


int SpPoDataBase::FindWildPointers()
{


  return 17;
}


int SpPoDataBase::VerifyDatabase()
{

  return 17;
}

void SpPoDataBase::Print()
{
  int i=0;
  int j;
  SpDynPtrArray classIDs;
  cout<<"::::::::::::::::::::::::::::::::::::::::"
      <<":::::::::::::::::::::::::::::::::::::::"<<endl;
  cout<<":::::::::::::::::::::::::::::::Static cl"
      <<"ass info:::::::::::::::::::::::::::::::"<<endl;
  cout<<"::::::::::::::::::::::::::::::::::::::::"
      <<":::::::::::::::::::::::::::::::::::::::"<<endl;

  SpClassNameAndId * nameAndId = (SpClassNameAndId*)ClassIDs->GetFirstElement();
  while (nameAndId != NULL){
    i = nameAndId->GetId();
    classIDs[nameAndId->GetId()] = nameAndId->GetName();
    SpList* list = (SpList*)(*VtableOffsetValues)[i];
    cout<< "Class name: "<<nameAndId->GetName()<<endl;
    cout<< "Class size: "<<(int)(*SizeList)[i]<<endl;
    cout<< "Number virtual function tables: "<<list->GetNumElements()<<endl;
    int * offset = (int*)list->GetFirstElement();
    for (j = 0; j < list->GetNumElements(); ++j){
      cout<<"     vtable at offset "<<*offset<<endl;
      offset = (int*) list->GetNextElement();
    }
    nameAndId = (SpClassNameAndId*)ClassIDs->GetNextElement();
    cout<<endl;
  }
  cout<<"::::::::::::::::::::::::::::::::::::::::"
      <<":::::::::::::::::::::::::::::::::::::::"<<endl;
  cout<<":::::::::::::::::::::::::::::::Dynamic c"
      <<"lass info::::::::::::::::::::::::::::::"<<endl;
  cout<<"::::::::::::::::::::::::::::::::::::::::"
      <<":::::::::::::::::::::::::::::::::::::::"<<endl;

   SpMemBlkHolder* AddrPtr =
    (SpMemBlkHolder*)ListOfAllPointers->GetFirstElement();
  //int arrayElement = 0;
  //int numPointers = 0;
  SpDynPtrArray ptrArray;
  //int elementSize = 0;
  // Get the amount of space required for all the memory blocks, their
  // pointers, and their names if they are named
  
  while (AddrPtr != NULL){

    PrintBlock(AddrPtr, &classIDs);
    AddrPtr = (SpMemBlkHolder*)ListOfAllPointers->GetNextElement();
    /*
      I commented this out since arrayElement is never used.
    if (AddrPtr){
      arrayElement = 0;
    }
    */
  }

  cout<<"----------------------------------------"
      <<"---------------------------------------"<<endl;


}

void *SpPoDataBase::LookupAddress(const SpKey& k)
{return ListOfAllPointers->Find(k);}

void *SpPoDataBase::LookupAddress(int low, int high)
{
  static SpIntervalKey address;
  address(low, high);
  return ListOfAllPointers->Find(address);}

void *SpPoDataBase::LookupAddress(void * ptr){
  static SpIntervalKey LookupAddressKey;
  LookupAddressKey((int)ptr, (int)ptr);
  return ListOfAllPointers->Find(LookupAddressKey);
}



void SpPoDataBase::MakeVtableIntoOffsets(void * *pBaseAddr){
  void * baseAddr = *pBaseAddr;
  SpList * lst = (SpList*)(*VtableOffsetValues)[AddToVtableType];
  int len = lst->GetNumElements();
  void * vtableAddr = lst->GetFirstElement();
  for (int i=0; i < len; ++i){
    *(int*)vtableAddr = ((int)*(int*)vtableAddr - (int)baseAddr);
    vtableAddr = lst->GetNextElement();
  }
  vtableAddr = lst->GetFirstElement();

  // Set the first item to zero so I can identify types that inherit from
  // SpPersistenceBaseClass
  for (int j=0; j < len; ++j){
    if ((*(int*)vtableAddr) == 0){
      if (j == 0){
	break;
      }
      void* tempVtableAddr = lst->GetFirstElement();
      *(int*)(lst->GetFirstElement()) = *(int*)vtableAddr;
      *(int*)vtableAddr = *(int*)tempVtableAddr;
    }
    vtableAddr = lst->GetNextElement();
  }
}

void SpPoDataBase::ResetPostponedAttachList()
{
  int len = PostponedAttachList->GetNumElements();

  for (int i=0; i < len; ++i){
    delete (void**) PostponedAttachList->Remove();
  }
}
