#ifndef SpPoDataBase_H
#define SpPoDataBase_H

class SpDynPtrArray;
class SpMemBlkHolder;
class SpBinaryTree;
class SpKey;
class SpList;

class SpPoDataBase{
  friend class FriendOfSpPoDataBase;
public:

  enum ErrorCodes{
    PO_NoError = 0,
    PO_RemovedAddressNotInDataBase,
    PO_ClassNotRegistered,
    PO_SetPointerObjectNotInDatabase,
    PO_UnSetPointerObjectNotInDatabase,
    PO_AddAddressTwice,
    PO_NumErrorCodes
  };

  enum {
    NO_COMPRESSION,
    BUILT_IN_LZH
  };

  SpPoDataBase();
  ~SpPoDataBase();

  void  SetName(void * &ptr, char * name); //Sets name of given object
  void* LookupName(char * name);          //Returns object with given name

  void SetPointer(void *&ptr);   //Sets a pointer as pointing to some memory.
  void UnsetPointer(void *&ptr); //unsets a pointer from pointing to some
                                   //memory.

  void* Add(void *&ptr,unsigned int size,int classId);// Adds a piece of memory
                                                      // to the data base
  void* Remove(void *&ptr);	// Removes a piece of memory to the data
				// data base
  char* CreateBuffer(int compressionType = NO_COMPRESSION); //Turns database into a buffer
  char* CreateBuffer(int compressionType, void * ptr); //Turns database into a buffer
  int  ReconstructBuffer(char *); //Reconstructs database from a buffer


  void Print(); //Prints out the database for inspection






  /* The following should never be used by "normal" users*/



  int   GetNumberOf(int classId);  //returns number of items of a given class
  void* GetFirstClass(int classId);//returns first instance of a given class
  void* GetNextClass();            //returns next instance of a given class
  void* GetFirstClassBlock(int classId);//returns first instance of a given class
  void* GetNextClassBlock();           //returns next instance of a given class


  int GetSize(void *ptr);	//Returns the size of memory pointed to by ptr

  void IncrementCount(int classId, int numItems = 1); //Increment number of items
                                                   //of a given class
  void DecrementCount(int classId, int numItems = 1); //Decrement number of items
                                                   //of a given class

  int GetClassId(char *);	//Returns an ID for a given class ID >0
  const char* GetClassTypeName(int classId);

  void RegisterSingleConstructor(int classId, void* ()
#ifdef HPUX
				 ()
#endif
				 );
    //Registers how to construct a single object of type classId
  void RegisterArrayConstructor(int classId, void* (func) (int));
    //Registers how to construct an array of objects of type classId
  void RegisterSize(int classId, int size);



  void StoreVtableInfo(void **pObject);
  void MakeVtableIntoOffsets(void **pBaseAddr);
  void *LookupAddress(const SpKey& k);
  void *LookupAddress(int low, int high);
  void *LookupAddress(void * ptr);
  int NumVtables(int classId);
  void SetAddToVtableList(){AddToVtableList = 1;}
  void UnsetAddToVtableList(){AddToVtableList = 0;}
  int AddToVtableListIsSet(){return AddToVtableList == 1;}
  void SetAddToVtableType(int type){AddToVtableType = type;}
  void UnsetAddToVtableType(){AddToVtableType = -1;}


  static int PO_errno;
  static int DefaultAddType;
  static int DoNotAllocateMemoryInConstructor;


  static void PO_perror(char* msg);
  void SetGetNextClassType(int val){GetNextClassType = val;}
  int GetGetNextClassType(){return GetNextClassType;}

  int VerifyDatabase();

  int FindDanglingPointers();
  int FindWildPointers();
  void ResetPOErrno(){PO_errno = 0;}
  void ResetPostponedAttachList();

protected:

  void PrintBlock(SpMemBlkHolder* AddrPtr, SpDynPtrArray* classIDs);

  int HasVtables(int ClassId);

  void  AddToClassIdMappingTable(char *name, int oldClassId);
  char* SetupClassIdMappingTable(char * buff, int nTypes);
  char* ReconstructNamedObjects(char *buff, int nNamedObjects);
  char* ReconstructObjects(char * buff, int nItems);
  int   CompressBuffer(char *buffer, int size,
		       int compressionType= NO_COMPRESSION);
  char* DecompressBuffer(char *buffer, int size, int compressionType);
  int GetBufferSize(int& numTypes, int& numItems, int& numNamedObjects);

  void UpdatePointers();
  void InternalStoreVtableInfo(void **pObject);


  SpBinaryTree* ListOfAllPointers;
  SpDynPtrArray* ClassCounts;	// (void*)'s in list will be used as ints

  static SpBinaryTree* ClassIDs;
  static SpDynPtrArray* SizeList; // (void*)'s in list will be used as ints
  static SpDynPtrArray* SingleConstructors;
  static SpDynPtrArray* ArrayConstructors;
  static SpDynPtrArray* VtableOffsetValues; //array of lists of vtable offsets
  static int NumClasses;


  SpList* PostponedAttachList;
  SpDynPtrArray* ListOfClassesByClassType;
  SpBinaryTree* PtrsByName;

  int CurrentClassForGetNextClass;
  int AddToVtableList;
  int AddToVtableType;

  int GetNextNumArrayElements;
  int GetNextObjectSize;

  void * CurrentItemPointer;
  void * CurrentArrayPointer;

  SpMemBlkHolder * CurrentClassInBlock;
  int CurrentBlockSize;
  int SizeOfClassInCurrentBlock;

  int GetNextClassType;

};


inline void SpPoDataBase::StoreVtableInfo(void **pObject){
  if (AddToVtableListIsSet()){
    if (AddToVtableType != -1){
      InternalStoreVtableInfo(pObject);
    }
  }
};


#endif
